# 1 "/builddir/build/BUILD/cp2k-2024.1/src/minimax/minimax_exp_k53.F" 1
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright 2000-2024 CP2K developers group <https://cp2k.org>                                   !
!                                                                                                  !
!   SPDX-License-Identifier: GPL-2.0-or-later                                                      !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines to calculate the minimax coefficients in order to
!>        approximate 1/x as a sum over exponential functions
!>        1/x ~ SUM_{i}^{K} w_i EXP(-a_i * x) for x belonging to [1:Rc].
!>        This module contains coefficients for minimax approximations with 1 <= k <= 53.
!>        Generated from data from
!>        http://www.mis.mpg.de/scicomp/EXP_SUM/1_x
!>        See also https://doi.org/10.1007/s00791-018-00308-4
!>        This module should not be modified manually and should not be used anywhere
!>        except in main minimax module.
!>        This file was created using the scripts in cp2k/tools/minimax_tools.
! **************************************************************************************************
MODULE minimax_exp_k53
   USE kinds,                           ONLY: dp

   IMPLICIT NONE
   PRIVATE
   PUBLIC :: R_max, R_mm, err_mm, get_minimax_coeff_low, k_max, k_mm, k_p, n_approx

   INTEGER, PARAMETER :: n_approx = 1221
   INTEGER, PARAMETER :: n_k = 53
   INTEGER, PARAMETER :: k_max = 53
   REAL(KIND=dp), PARAMETER :: R_max = 4.0E+12_dp

   INTEGER, PARAMETER, DIMENSION(n_k + 1) :: k_p = &
                                             [1, 10, 23, 42, 64, &
                                              92, 121, 154, 183, 213, &
                                              246, 284, 323, 363, 406, &
                                              453, 501, 549, 597, 648, &
                                              699, 748, 797, 848, 902, &
                                              957, 1011, 1051, 1057, 1060, &
                                              1065, 1070, 1076, 1083, 1092, &
                                              1104, 1106, 1108, 1111, 1115, &
                                              1120, 1127, 1137, 1139, 1141, &
                                              1144, 1148, 1153, 1160, 1170, &
                                              1207, 1209, 1212, 1222]

   INTEGER, PARAMETER, DIMENSION(n_approx) :: k_mm = &
                                              [1, 1, 1, 1, 1, &
                                               1, 1, 1, 1, 2, &
                                               2, 2, 2, 2, 2, &
                                               2, 2, 2, 2, 2, &
                                               2, 2, 3, 3, 3, &
                                               3, 3, 3, 3, 3, &
                                               3, 3, 3, 3, 3, &
                                               3, 3, 3, 3, 3, &
                                               3, 4, 4, 4, 4, &
                                               4, 4, 4, 4, 4, &
                                               4, 4, 4, 4, 4, &
                                               4, 4, 4, 4, 4, &
                                               4, 4, 4, 5, 5, &
                                               5, 5, 5, 5, 5, &
                                               5, 5, 5, 5, 5, &
                                               5, 5, 5, 5, 5, &
                                               5, 5, 5, 5, 5, &
                                               5, 5, 5, 5, 5, &
                                               5, 6, 6, 6, 6, &
                                               6, 6, 6, 6, 6, &
                                               6, 6, 6, 6, 6, &
                                               6, 6, 6, 6, 6, &
                                               6, 6, 6, 6, 6, &
                                               6, 6, 6, 6, 6, &
                                               7, 7, 7, 7, 7, &
                                               7, 7, 7, 7, 7, &
                                               7, 7, 7, 7, 7, &
                                               7, 7, 7, 7, 7, &
                                               7, 7, 7, 7, 7, &
                                               7, 7, 7, 7, 7, &
                                               7, 7, 7, 8, 8, &
                                               8, 8, 8, 8, 8, &
                                               8, 8, 8, 8, 8, &
                                               8, 8, 8, 8, 8, &
                                               8, 8, 8, 8, 8, &
                                               8, 8, 8, 8, 8, &
                                               8, 8, 9, 9, 9, &
                                               9, 9, 9, 9, 9, &
                                               9, 9, 9, 9, 9, &
                                               9, 9, 9, 9, 9, &
                                               9, 9, 9, 9, 9, &
                                               9, 9, 9, 9, 9, &
                                               9, 9, 10, 10, 10, &
                                               10, 10, 10, 10, 10, &
                                               10, 10, 10, 10, 10, &
                                               10, 10, 10, 10, 10, &
                                               10, 10, 10, 10, 10, &
                                               10, 10, 10, 10, 10, &
                                               10, 10, 10, 10, 10, &
                                               11, 11, 11, 11, 11, &
                                               11, 11, 11, 11, 11, &
                                               11, 11, 11, 11, 11, &
                                               11, 11, 11, 11, 11, &
                                               11, 11, 11, 11, 11, &
                                               11, 11, 11, 11, 11, &
                                               11, 11, 11, 11, 11, &
                                               11, 11, 11, 12, 12, &
                                               12, 12, 12, 12, 12, &
                                               12, 12, 12, 12, 12, &
                                               12, 12, 12, 12, 12, &
                                               12, 12, 12, 12, 12, &
                                               12, 12, 12, 12, 12, &
                                               12, 12, 12, 12, 12, &
                                               12, 12, 12, 12, 12, &
                                               12, 12, 13, 13, 13, &
                                               13, 13, 13, 13, 13, &
                                               13, 13, 13, 13, 13, &
                                               13, 13, 13, 13, 13, &
                                               13, 13, 13, 13, 13, &
                                               13, 13, 13, 13, 13, &
                                               13, 13, 13, 13, 13, &
                                               13, 13, 13, 13, 13, &
                                               13, 13, 14, 14, 14, &
                                               14, 14, 14, 14, 14, &
                                               14, 14, 14, 14, 14, &
                                               14, 14, 14, 14, 14, &
                                               14, 14, 14, 14, 14, &
                                               14, 14, 14, 14, 14, &
                                               14, 14, 14, 14, 14, &
                                               14, 14, 14, 14, 14, &
                                               14, 14, 14, 14, 14, &
                                               15, 15, 15, 15, 15, &
                                               15, 15, 15, 15, 15, &
                                               15, 15, 15, 15, 15, &
                                               15, 15, 15, 15, 15, &
                                               15, 15, 15, 15, 15, &
                                               15, 15, 15, 15, 15, &
                                               15, 15, 15, 15, 15, &
                                               15, 15, 15, 15, 15, &
                                               15, 15, 15, 15, 15, &
                                               15, 15, 16, 16, 16, &
                                               16, 16, 16, 16, 16, &
                                               16, 16, 16, 16, 16, &
                                               16, 16, 16, 16, 16, &
                                               16, 16, 16, 16, 16, &
                                               16, 16, 16, 16, 16, &
                                               16, 16, 16, 16, 16, &
                                               16, 16, 16, 16, 16, &
                                               16, 16, 16, 16, 16, &
                                               16, 16, 16, 16, 16, &
                                               17, 17, 17, 17, 17, &
                                               17, 17, 17, 17, 17, &
                                               17, 17, 17, 17, 17, &
                                               17, 17, 17, 17, 17, &
                                               17, 17, 17, 17, 17, &
                                               17, 17, 17, 17, 17, &
                                               17, 17, 17, 17, 17, &
                                               17, 17, 17, 17, 17, &
                                               17, 17, 17, 17, 17, &
                                               17, 17, 17, 18, 18, &
                                               18, 18, 18, 18, 18, &
                                               18, 18, 18, 18, 18, &
                                               18, 18, 18, 18, 18, &
                                               18, 18, 18, 18, 18, &
                                               18, 18, 18, 18, 18, &
                                               18, 18, 18, 18, 18, &
                                               18, 18, 18, 18, 18, &
                                               18, 18, 18, 18, 18, &
                                               18, 18, 18, 18, 18, &
                                               18, 19, 19, 19, 19, &
                                               19, 19, 19, 19, 19, &
                                               19, 19, 19, 19, 19, &
                                               19, 19, 19, 19, 19, &
                                               19, 19, 19, 19, 19, &
                                               19, 19, 19, 19, 19, &
                                               19, 19, 19, 19, 19, &
                                               19, 19, 19, 19, 19, &
                                               19, 19, 19, 19, 19, &
                                               19, 19, 19, 19, 19, &
                                               19, 19, 20, 20, 20, &
                                               20, 20, 20, 20, 20, &
                                               20, 20, 20, 20, 20, &
                                               20, 20, 20, 20, 20, &
                                               20, 20, 20, 20, 20, &
                                               20, 20, 20, 20, 20, &
                                               20, 20, 20, 20, 20, &
                                               20, 20, 20, 20, 20, &
                                               20, 20, 20, 20, 20, &
                                               20, 20, 20, 20, 20, &
                                               20, 20, 20, 21, 21, &
                                               21, 21, 21, 21, 21, &
                                               21, 21, 21, 21, 21, &
                                               21, 21, 21, 21, 21, &
                                               21, 21, 21, 21, 21, &
                                               21, 21, 21, 21, 21, &
                                               21, 21, 21, 21, 21, &
                                               21, 21, 21, 21, 21, &
                                               21, 21, 21, 21, 21, &
                                               21, 21, 21, 21, 21, &
                                               21, 21, 22, 22, 22, &
                                               22, 22, 22, 22, 22, &
                                               22, 22, 22, 22, 22, &
                                               22, 22, 22, 22, 22, &
                                               22, 22, 22, 22, 22, &
                                               22, 22, 22, 22, 22, &
                                               22, 22, 22, 22, 22, &
                                               22, 22, 22, 22, 22, &
                                               22, 22, 22, 22, 22, &
                                               22, 22, 22, 22, 22, &
                                               22, 23, 23, 23, 23, &
                                               23, 23, 23, 23, 23, &
                                               23, 23, 23, 23, 23, &
                                               23, 23, 23, 23, 23, &
                                               23, 23, 23, 23, 23, &
                                               23, 23, 23, 23, 23, &
                                               23, 23, 23, 23, 23, &
                                               23, 23, 23, 23, 23, &
                                               23, 23, 23, 23, 23, &
                                               23, 23, 23, 23, 23, &
                                               23, 23, 24, 24, 24, &
                                               24, 24, 24, 24, 24, &
                                               24, 24, 24, 24, 24, &
                                               24, 24, 24, 24, 24, &
                                               24, 24, 24, 24, 24, &
                                               24, 24, 24, 24, 24, &
                                               24, 24, 24, 24, 24, &
                                               24, 24, 24, 24, 24, &
                                               24, 24, 24, 24, 24, &
                                               24, 24, 24, 24, 24, &
                                               24, 24, 24, 24, 24, &
                                               24, 25, 25, 25, 25, &
                                               25, 25, 25, 25, 25, &
                                               25, 25, 25, 25, 25, &
                                               25, 25, 25, 25, 25, &
                                               25, 25, 25, 25, 25, &
                                               25, 25, 25, 25, 25, &
                                               25, 25, 25, 25, 25, &
                                               25, 25, 25, 25, 25, &
                                               25, 25, 25, 25, 25, &
                                               25, 25, 25, 25, 25, &
                                               25, 25, 25, 25, 25, &
                                               25, 26, 26, 26, 26, &
                                               26, 26, 26, 26, 26, &
                                               26, 26, 26, 26, 26, &
                                               26, 26, 26, 26, 26, &
                                               26, 26, 26, 26, 26, &
                                               26, 26, 26, 26, 26, &
                                               26, 26, 26, 26, 26, &
                                               26, 26, 26, 26, 26, &
                                               26, 26, 26, 26, 26, &
                                               26, 26, 26, 26, 26, &
                                               26, 26, 26, 26, 26, &
                                               27, 27, 27, 27, 27, &
                                               27, 27, 27, 27, 27, &
                                               27, 27, 27, 27, 27, &
                                               27, 27, 27, 27, 27, &
                                               27, 27, 27, 27, 27, &
                                               27, 27, 27, 27, 27, &
                                               27, 27, 27, 27, 27, &
                                               27, 27, 27, 27, 27, &
                                               28, 28, 28, 28, 28, &
                                               28, 29, 29, 29, 30, &
                                               30, 30, 30, 30, 31, &
                                               31, 31, 31, 31, 32, &
                                               32, 32, 32, 32, 32, &
                                               33, 33, 33, 33, 33, &
                                               33, 33, 34, 34, 34, &
                                               34, 34, 34, 34, 34, &
                                               34, 35, 35, 35, 35, &
                                               35, 35, 35, 35, 35, &
                                               35, 35, 35, 36, 36, &
                                               37, 37, 38, 38, 38, &
                                               39, 39, 39, 39, 40, &
                                               40, 40, 40, 40, 41, &
                                               41, 41, 41, 41, 41, &
                                               41, 42, 42, 42, 42, &
                                               42, 42, 42, 42, 42, &
                                               42, 43, 43, 44, 44, &
                                               45, 45, 45, 46, 46, &
                                               46, 46, 47, 47, 47, &
                                               47, 47, 48, 48, 48, &
                                               48, 48, 48, 48, 49, &
                                               49, 49, 49, 49, 49, &
                                               49, 49, 49, 49, 50, &
                                               50, 50, 50, 50, 50, &
                                               50, 50, 50, 50, 50, &
                                               50, 50, 50, 50, 50, &
                                               50, 50, 50, 50, 50, &
                                               50, 50, 50, 50, 50, &
                                               50, 50, 50, 50, 50, &
                                               50, 50, 50, 50, 50, &
                                               50, 51, 51, 52, 52, &
                                               52, 53, 53, 53, 53, &
                                               53, 53, 53, 53, 53, &
                                               53]

   REAL(KIND=dp), PARAMETER, DIMENSION(n_approx) :: R_mm = &
                                                    [2.0E+00_dp, 3.0E+00_dp, 4.0E+00_dp, 5.0E+00_dp, 6.0E+00_dp, &
                                                     7.0E+00_dp, 8.0E+00_dp, 9.0E+00_dp, 1.0E+01_dp, 2.0E+00_dp, &
                                                     3.0E+00_dp, 4.0E+00_dp, 5.0E+00_dp, 6.0E+00_dp, 7.0E+00_dp, &
                                                     8.0E+00_dp, 9.0E+00_dp, 1.0E+01_dp, 2.0E+01_dp, 3.0E+01_dp, &
                                                     4.0E+01_dp, 5.0E+01_dp, 2.0E+00_dp, 3.0E+00_dp, 4.0E+00_dp, &
                                                     5.0E+00_dp, 6.0E+00_dp, 7.0E+00_dp, 8.0E+00_dp, 9.0E+00_dp, &
                                                     1.0E+01_dp, 2.0E+01_dp, 3.0E+01_dp, 4.0E+01_dp, 5.0E+01_dp, &
                                                     6.0E+01_dp, 7.0E+01_dp, 8.0E+01_dp, 9.0E+01_dp, 1.0E+02_dp, &
                                                     2.0E+02_dp, 2.0E+00_dp, 3.0E+00_dp, 4.0E+00_dp, 5.0E+00_dp, &
                                                     6.0E+00_dp, 7.0E+00_dp, 8.0E+00_dp, 9.0E+00_dp, 1.0E+01_dp, &
                                                     2.0E+01_dp, 3.0E+01_dp, 4.0E+01_dp, 5.0E+01_dp, 6.0E+01_dp, &
                                                     7.0E+01_dp, 8.0E+01_dp, 9.0E+01_dp, 1.0E+02_dp, 2.0E+02_dp, &
                                                     3.0E+02_dp, 4.0E+02_dp, 5.0E+02_dp, 2.0E+00_dp, 3.0E+00_dp, &
                                                     4.0E+00_dp, 5.0E+00_dp, 6.0E+00_dp, 7.0E+00_dp, 8.0E+00_dp, &
                                                     9.0E+00_dp, 1.0E+01_dp, 2.0E+01_dp, 3.0E+01_dp, 4.0E+01_dp, &
                                                     5.0E+01_dp, 6.0E+01_dp, 7.0E+01_dp, 8.0E+01_dp, 9.0E+01_dp, &
                                                     1.0E+02_dp, 2.0E+02_dp, 3.0E+02_dp, 4.0E+02_dp, 5.0E+02_dp, &
                                                     6.0E+02_dp, 7.0E+02_dp, 8.0E+02_dp, 9.0E+02_dp, 1.0E+03_dp, &
                                                     2.0E+03_dp, 2.0E+00_dp, 3.0E+00_dp, 4.0E+00_dp, 5.0E+00_dp, &
                                                     6.0E+00_dp, 7.0E+00_dp, 8.0E+00_dp, 9.0E+00_dp, 1.0E+01_dp, &
                                                     2.0E+01_dp, 3.0E+01_dp, 4.0E+01_dp, 5.0E+01_dp, 6.0E+01_dp, &
                                                     7.0E+01_dp, 8.0E+01_dp, 9.0E+01_dp, 1.0E+02_dp, 2.0E+02_dp, &
                                                     3.0E+02_dp, 4.0E+02_dp, 5.0E+02_dp, 6.0E+02_dp, 7.0E+02_dp, &
                                                     8.0E+02_dp, 9.0E+02_dp, 1.0E+03_dp, 2.0E+03_dp, 3.0E+03_dp, &
                                                     2.0E+00_dp, 3.0E+00_dp, 4.0E+00_dp, 5.0E+00_dp, 6.0E+00_dp, &
                                                     7.0E+00_dp, 8.0E+00_dp, 9.0E+00_dp, 1.0E+01_dp, 2.0E+01_dp, &
                                                     3.0E+01_dp, 4.0E+01_dp, 5.0E+01_dp, 6.0E+01_dp, 7.0E+01_dp, &
                                                     8.0E+01_dp, 9.0E+01_dp, 1.0E+02_dp, 2.0E+02_dp, 3.0E+02_dp, &
                                                     4.0E+02_dp, 5.0E+02_dp, 6.0E+02_dp, 7.0E+02_dp, 8.0E+02_dp, &
                                                     9.0E+02_dp, 1.0E+03_dp, 2.0E+03_dp, 3.0E+03_dp, 4.0E+03_dp, &
                                                     5.0E+03_dp, 6.0E+03_dp, 7.0E+03_dp, 1.0E+01_dp, 2.0E+01_dp, &
                                                     3.0E+01_dp, 4.0E+01_dp, 5.0E+01_dp, 6.0E+01_dp, 7.0E+01_dp, &
                                                     8.0E+01_dp, 9.0E+01_dp, 1.0E+02_dp, 2.0E+02_dp, 3.0E+02_dp, &
                                                     4.0E+02_dp, 5.0E+02_dp, 6.0E+02_dp, 7.0E+02_dp, 8.0E+02_dp, &
                                                     9.0E+02_dp, 1.0E+03_dp, 2.0E+03_dp, 3.0E+03_dp, 4.0E+03_dp, &
                                                     5.0E+03_dp, 6.0E+03_dp, 7.0E+03_dp, 8.0E+03_dp, 9.0E+03_dp, &
                                                     1.0E+04_dp, 2.0E+04_dp, 1.0E+01_dp, 2.0E+01_dp, 3.0E+01_dp, &
                                                     4.0E+01_dp, 5.0E+01_dp, 6.0E+01_dp, 7.0E+01_dp, 8.0E+01_dp, &
                                                     9.0E+01_dp, 1.0E+02_dp, 2.0E+02_dp, 3.0E+02_dp, 4.0E+02_dp, &
                                                     5.0E+02_dp, 6.0E+02_dp, 7.0E+02_dp, 8.0E+02_dp, 9.0E+02_dp, &
                                                     1.0E+03_dp, 2.0E+03_dp, 3.0E+03_dp, 4.0E+03_dp, 5.0E+03_dp, &
                                                     6.0E+03_dp, 7.0E+03_dp, 8.0E+03_dp, 9.0E+03_dp, 1.0E+04_dp, &
                                                     2.0E+04_dp, 3.0E+04_dp, 1.0E+01_dp, 2.0E+01_dp, 3.0E+01_dp, &
                                                     4.0E+01_dp, 5.0E+01_dp, 6.0E+01_dp, 7.0E+01_dp, 8.0E+01_dp, &
                                                     9.0E+01_dp, 1.0E+02_dp, 2.0E+02_dp, 3.0E+02_dp, 4.0E+02_dp, &
                                                     5.0E+02_dp, 6.0E+02_dp, 7.0E+02_dp, 8.0E+02_dp, 9.0E+02_dp, &
                                                     1.0E+03_dp, 2.0E+03_dp, 3.0E+03_dp, 4.0E+03_dp, 5.0E+03_dp, &
                                                     6.0E+03_dp, 7.0E+03_dp, 8.0E+03_dp, 9.0E+03_dp, 1.0E+04_dp, &
                                                     2.0E+04_dp, 3.0E+04_dp, 4.0E+04_dp, 5.0E+04_dp, 6.0E+04_dp, &
                                                     1.0E+01_dp, 2.0E+01_dp, 3.0E+01_dp, 4.0E+01_dp, 5.0E+01_dp, &
                                                     6.0E+01_dp, 7.0E+01_dp, 8.0E+01_dp, 9.0E+01_dp, 1.0E+02_dp, &
                                                     2.0E+02_dp, 3.0E+02_dp, 4.0E+02_dp, 5.0E+02_dp, 6.0E+02_dp, &
                                                     7.0E+02_dp, 8.0E+02_dp, 9.0E+02_dp, 1.0E+03_dp, 2.0E+03_dp, &
                                                     3.0E+03_dp, 4.0E+03_dp, 5.0E+03_dp, 6.0E+03_dp, 7.0E+03_dp, &
                                                     8.0E+03_dp, 9.0E+03_dp, 1.0E+04_dp, 2.0E+04_dp, 3.0E+04_dp, &
                                                     4.0E+04_dp, 5.0E+04_dp, 6.0E+04_dp, 7.0E+04_dp, 8.0E+04_dp, &
                                                     9.0E+04_dp, 1.0E+05_dp, 2.0E+05_dp, 1.0E+01_dp, 2.0E+01_dp, &
                                                     3.0E+01_dp, 4.0E+01_dp, 5.0E+01_dp, 6.0E+01_dp, 7.0E+01_dp, &
                                                     8.0E+01_dp, 9.0E+01_dp, 1.0E+02_dp, 2.0E+02_dp, 3.0E+02_dp, &
                                                     4.0E+02_dp, 5.0E+02_dp, 6.0E+02_dp, 7.0E+02_dp, 8.0E+02_dp, &
                                                     9.0E+02_dp, 1.0E+03_dp, 2.0E+03_dp, 3.0E+03_dp, 4.0E+03_dp, &
                                                     5.0E+03_dp, 6.0E+03_dp, 7.0E+03_dp, 8.0E+03_dp, 9.0E+03_dp, &
                                                     1.0E+04_dp, 2.0E+04_dp, 3.0E+04_dp, 4.0E+04_dp, 5.0E+04_dp, &
                                                     6.0E+04_dp, 7.0E+04_dp, 8.0E+04_dp, 9.0E+04_dp, 1.0E+05_dp, &
                                                     2.0E+05_dp, 3.0E+05_dp, 1.0E+01_dp, 2.0E+01_dp, 3.0E+01_dp, &
                                                     4.0E+01_dp, 5.0E+01_dp, 6.0E+01_dp, 7.0E+01_dp, 8.0E+01_dp, &
                                                     9.0E+01_dp, 1.0E+02_dp, 2.0E+02_dp, 3.0E+02_dp, 4.0E+02_dp, &
                                                     5.0E+02_dp, 6.0E+02_dp, 7.0E+02_dp, 8.0E+02_dp, 9.0E+02_dp, &
                                                     1.0E+03_dp, 2.0E+03_dp, 3.0E+03_dp, 4.0E+03_dp, 5.0E+03_dp, &
                                                     6.0E+03_dp, 7.0E+03_dp, 8.0E+03_dp, 9.0E+03_dp, 1.0E+04_dp, &
                                                     2.0E+04_dp, 3.0E+04_dp, 4.0E+04_dp, 5.0E+04_dp, 6.0E+04_dp, &
                                                     7.0E+04_dp, 8.0E+04_dp, 9.0E+04_dp, 1.0E+05_dp, 2.0E+05_dp, &
                                                     3.0E+05_dp, 4.0E+05_dp, 1.0E+01_dp, 2.0E+01_dp, 3.0E+01_dp, &
                                                     4.0E+01_dp, 5.0E+01_dp, 6.0E+01_dp, 7.0E+01_dp, 8.0E+01_dp, &
                                                     9.0E+01_dp, 1.0E+02_dp, 2.0E+02_dp, 3.0E+02_dp, 4.0E+02_dp, &
                                                     5.0E+02_dp, 6.0E+02_dp, 7.0E+02_dp, 8.0E+02_dp, 9.0E+02_dp, &
                                                     1.0E+03_dp, 2.0E+03_dp, 3.0E+03_dp, 4.0E+03_dp, 5.0E+03_dp, &
                                                     6.0E+03_dp, 7.0E+03_dp, 8.0E+03_dp, 9.0E+03_dp, 1.0E+04_dp, &
                                                     2.0E+04_dp, 3.0E+04_dp, 4.0E+04_dp, 5.0E+04_dp, 6.0E+04_dp, &
                                                     7.0E+04_dp, 8.0E+04_dp, 9.0E+04_dp, 1.0E+05_dp, 2.0E+05_dp, &
                                                     3.0E+05_dp, 4.0E+05_dp, 5.0E+05_dp, 6.0E+05_dp, 7.0E+05_dp, &
                                                     1.0E+01_dp, 2.0E+01_dp, 3.0E+01_dp, 4.0E+01_dp, 5.0E+01_dp, &
                                                     6.0E+01_dp, 7.0E+01_dp, 8.0E+01_dp, 9.0E+01_dp, 1.0E+02_dp, &
                                                     2.0E+02_dp, 3.0E+02_dp, 4.0E+02_dp, 5.0E+02_dp, 6.0E+02_dp, &
                                                     7.0E+02_dp, 8.0E+02_dp, 9.0E+02_dp, 1.0E+03_dp, 2.0E+03_dp, &
                                                     3.0E+03_dp, 4.0E+03_dp, 5.0E+03_dp, 6.0E+03_dp, 7.0E+03_dp, &
                                                     8.0E+03_dp, 9.0E+03_dp, 1.0E+04_dp, 2.0E+04_dp, 3.0E+04_dp, &
                                                     4.0E+04_dp, 5.0E+04_dp, 6.0E+04_dp, 7.0E+04_dp, 8.0E+04_dp, &
                                                     9.0E+04_dp, 1.0E+05_dp, 2.0E+05_dp, 3.0E+05_dp, 4.0E+05_dp, &
                                                     5.0E+05_dp, 6.0E+05_dp, 7.0E+05_dp, 8.0E+05_dp, 9.0E+05_dp, &
                                                     1.0E+06_dp, 2.0E+06_dp, 1.0E+01_dp, 2.0E+01_dp, 3.0E+01_dp, &
                                                     4.0E+01_dp, 5.0E+01_dp, 6.0E+01_dp, 7.0E+01_dp, 8.0E+01_dp, &
                                                     9.0E+01_dp, 1.0E+02_dp, 2.0E+02_dp, 3.0E+02_dp, 4.0E+02_dp, &
                                                     5.0E+02_dp, 6.0E+02_dp, 7.0E+02_dp, 8.0E+02_dp, 9.0E+02_dp, &
                                                     1.0E+03_dp, 2.0E+03_dp, 3.0E+03_dp, 4.0E+03_dp, 5.0E+03_dp, &
                                                     6.0E+03_dp, 7.0E+03_dp, 8.0E+03_dp, 9.0E+03_dp, 1.0E+04_dp, &
                                                     2.0E+04_dp, 3.0E+04_dp, 4.0E+04_dp, 5.0E+04_dp, 6.0E+04_dp, &
                                                     7.0E+04_dp, 8.0E+04_dp, 9.0E+04_dp, 1.0E+05_dp, 2.0E+05_dp, &
                                                     3.0E+05_dp, 4.0E+05_dp, 5.0E+05_dp, 6.0E+05_dp, 7.0E+05_dp, &
                                                     8.0E+05_dp, 9.0E+05_dp, 1.0E+06_dp, 2.0E+06_dp, 3.0E+06_dp, &
                                                     2.0E+01_dp, 3.0E+01_dp, 4.0E+01_dp, 5.0E+01_dp, 6.0E+01_dp, &
                                                     7.0E+01_dp, 8.0E+01_dp, 9.0E+01_dp, 1.0E+02_dp, 2.0E+02_dp, &
                                                     3.0E+02_dp, 4.0E+02_dp, 5.0E+02_dp, 6.0E+02_dp, 7.0E+02_dp, &
                                                     8.0E+02_dp, 9.0E+02_dp, 1.0E+03_dp, 2.0E+03_dp, 3.0E+03_dp, &
                                                     4.0E+03_dp, 5.0E+03_dp, 6.0E+03_dp, 7.0E+03_dp, 8.0E+03_dp, &
                                                     9.0E+03_dp, 1.0E+04_dp, 2.0E+04_dp, 3.0E+04_dp, 4.0E+04_dp, &
                                                     5.0E+04_dp, 6.0E+04_dp, 7.0E+04_dp, 8.0E+04_dp, 9.0E+04_dp, &
                                                     1.0E+05_dp, 2.0E+05_dp, 3.0E+05_dp, 4.0E+05_dp, 5.0E+05_dp, &
                                                     6.0E+05_dp, 7.0E+05_dp, 8.0E+05_dp, 9.0E+05_dp, 1.0E+06_dp, &
                                                     2.0E+06_dp, 3.0E+06_dp, 4.0E+06_dp, 4.0E+01_dp, 5.0E+01_dp, &
                                                     6.0E+01_dp, 7.0E+01_dp, 8.0E+01_dp, 9.0E+01_dp, 1.0E+02_dp, &
                                                     2.0E+02_dp, 3.0E+02_dp, 4.0E+02_dp, 5.0E+02_dp, 6.0E+02_dp, &
                                                     7.0E+02_dp, 8.0E+02_dp, 9.0E+02_dp, 1.0E+03_dp, 2.0E+03_dp, &
                                                     3.0E+03_dp, 4.0E+03_dp, 5.0E+03_dp, 6.0E+03_dp, 7.0E+03_dp, &
                                                     8.0E+03_dp, 9.0E+03_dp, 1.0E+04_dp, 2.0E+04_dp, 3.0E+04_dp, &
                                                     4.0E+04_dp, 5.0E+04_dp, 6.0E+04_dp, 7.0E+04_dp, 8.0E+04_dp, &
                                                     9.0E+04_dp, 1.0E+05_dp, 2.0E+05_dp, 3.0E+05_dp, 4.0E+05_dp, &
                                                     5.0E+05_dp, 6.0E+05_dp, 7.0E+05_dp, 8.0E+05_dp, 9.0E+05_dp, &
                                                     1.0E+06_dp, 2.0E+06_dp, 3.0E+06_dp, 4.0E+06_dp, 5.0E+06_dp, &
                                                     6.0E+06_dp, 5.0E+01_dp, 6.0E+01_dp, 7.0E+01_dp, 8.0E+01_dp, &
                                                     9.0E+01_dp, 1.0E+02_dp, 2.0E+02_dp, 3.0E+02_dp, 4.0E+02_dp, &
                                                     5.0E+02_dp, 6.0E+02_dp, 7.0E+02_dp, 8.0E+02_dp, 9.0E+02_dp, &
                                                     1.0E+03_dp, 2.0E+03_dp, 3.0E+03_dp, 4.0E+03_dp, 5.0E+03_dp, &
                                                     6.0E+03_dp, 7.0E+03_dp, 8.0E+03_dp, 9.0E+03_dp, 1.0E+04_dp, &
                                                     2.0E+04_dp, 3.0E+04_dp, 4.0E+04_dp, 5.0E+04_dp, 6.0E+04_dp, &
                                                     7.0E+04_dp, 8.0E+04_dp, 9.0E+04_dp, 1.0E+05_dp, 2.0E+05_dp, &
                                                     3.0E+05_dp, 4.0E+05_dp, 5.0E+05_dp, 6.0E+05_dp, 7.0E+05_dp, &
                                                     8.0E+05_dp, 9.0E+05_dp, 1.0E+06_dp, 2.0E+06_dp, 3.0E+06_dp, &
                                                     4.0E+06_dp, 5.0E+06_dp, 6.0E+06_dp, 7.0E+06_dp, 8.0E+06_dp, &
                                                     9.0E+06_dp, 1.0E+07_dp, 6.0E+01_dp, 7.0E+01_dp, 8.0E+01_dp, &
                                                     9.0E+01_dp, 1.0E+02_dp, 2.0E+02_dp, 3.0E+02_dp, 4.0E+02_dp, &
                                                     5.0E+02_dp, 6.0E+02_dp, 7.0E+02_dp, 8.0E+02_dp, 9.0E+02_dp, &
                                                     1.0E+03_dp, 2.0E+03_dp, 3.0E+03_dp, 4.0E+03_dp, 5.0E+03_dp, &
                                                     6.0E+03_dp, 7.0E+03_dp, 8.0E+03_dp, 9.0E+03_dp, 1.0E+04_dp, &
                                                     2.0E+04_dp, 3.0E+04_dp, 4.0E+04_dp, 5.0E+04_dp, 6.0E+04_dp, &
                                                     7.0E+04_dp, 8.0E+04_dp, 9.0E+04_dp, 1.0E+05_dp, 2.0E+05_dp, &
                                                     3.0E+05_dp, 4.0E+05_dp, 5.0E+05_dp, 6.0E+05_dp, 7.0E+05_dp, &
                                                     8.0E+05_dp, 9.0E+05_dp, 1.0E+06_dp, 2.0E+06_dp, 3.0E+06_dp, &
                                                     4.0E+06_dp, 5.0E+06_dp, 6.0E+06_dp, 7.0E+06_dp, 8.0E+06_dp, &
                                                     9.0E+06_dp, 1.0E+07_dp, 2.0E+07_dp, 9.0E+01_dp, 1.0E+02_dp, &
                                                     2.0E+02_dp, 3.0E+02_dp, 4.0E+02_dp, 5.0E+02_dp, 6.0E+02_dp, &
                                                     7.0E+02_dp, 8.0E+02_dp, 9.0E+02_dp, 1.0E+03_dp, 2.0E+03_dp, &
                                                     3.0E+03_dp, 4.0E+03_dp, 5.0E+03_dp, 6.0E+03_dp, 7.0E+03_dp, &
                                                     8.0E+03_dp, 9.0E+03_dp, 1.0E+04_dp, 2.0E+04_dp, 3.0E+04_dp, &
                                                     4.0E+04_dp, 5.0E+04_dp, 6.0E+04_dp, 7.0E+04_dp, 8.0E+04_dp, &
                                                     9.0E+04_dp, 1.0E+05_dp, 2.0E+05_dp, 3.0E+05_dp, 4.0E+05_dp, &
                                                     5.0E+05_dp, 6.0E+05_dp, 7.0E+05_dp, 8.0E+05_dp, 9.0E+05_dp, &
                                                     1.0E+06_dp, 2.0E+06_dp, 3.0E+06_dp, 4.0E+06_dp, 5.0E+06_dp, &
                                                     6.0E+06_dp, 7.0E+06_dp, 8.0E+06_dp, 9.0E+06_dp, 1.0E+07_dp, &
                                                     2.0E+07_dp, 3.0E+07_dp, 1.0E+02_dp, 2.0E+02_dp, 3.0E+02_dp, &
                                                     4.0E+02_dp, 5.0E+02_dp, 6.0E+02_dp, 7.0E+02_dp, 8.0E+02_dp, &
                                                     9.0E+02_dp, 1.0E+03_dp, 2.0E+03_dp, 3.0E+03_dp, 4.0E+03_dp, &
                                                     5.0E+03_dp, 6.0E+03_dp, 7.0E+03_dp, 8.0E+03_dp, 9.0E+03_dp, &
                                                     1.0E+04_dp, 2.0E+04_dp, 3.0E+04_dp, 4.0E+04_dp, 5.0E+04_dp, &
                                                     6.0E+04_dp, 7.0E+04_dp, 8.0E+04_dp, 9.0E+04_dp, 1.0E+05_dp, &
                                                     2.0E+05_dp, 3.0E+05_dp, 4.0E+05_dp, 5.0E+05_dp, 6.0E+05_dp, &
                                                     7.0E+05_dp, 8.0E+05_dp, 9.0E+05_dp, 1.0E+06_dp, 2.0E+06_dp, &
                                                     3.0E+06_dp, 4.0E+06_dp, 5.0E+06_dp, 6.0E+06_dp, 7.0E+06_dp, &
                                                     8.0E+06_dp, 9.0E+06_dp, 1.0E+07_dp, 2.0E+07_dp, 3.0E+07_dp, &
                                                     4.0E+07_dp, 2.0E+02_dp, 3.0E+02_dp, 4.0E+02_dp, 5.0E+02_dp, &
                                                     6.0E+02_dp, 7.0E+02_dp, 8.0E+02_dp, 9.0E+02_dp, 1.0E+03_dp, &
                                                     2.0E+03_dp, 3.0E+03_dp, 4.0E+03_dp, 5.0E+03_dp, 6.0E+03_dp, &
                                                     7.0E+03_dp, 8.0E+03_dp, 9.0E+03_dp, 1.0E+04_dp, 2.0E+04_dp, &
                                                     3.0E+04_dp, 4.0E+04_dp, 5.0E+04_dp, 6.0E+04_dp, 7.0E+04_dp, &
                                                     8.0E+04_dp, 9.0E+04_dp, 1.0E+05_dp, 2.0E+05_dp, 3.0E+05_dp, &
                                                     4.0E+05_dp, 5.0E+05_dp, 6.0E+05_dp, 7.0E+05_dp, 8.0E+05_dp, &
                                                     9.0E+05_dp, 1.0E+06_dp, 2.0E+06_dp, 3.0E+06_dp, 4.0E+06_dp, &
                                                     5.0E+06_dp, 6.0E+06_dp, 7.0E+06_dp, 8.0E+06_dp, 9.0E+06_dp, &
                                                     1.0E+07_dp, 2.0E+07_dp, 3.0E+07_dp, 4.0E+07_dp, 5.0E+07_dp, &
                                                     6.0E+07_dp, 7.0E+07_dp, 2.0E+02_dp, 3.0E+02_dp, 4.0E+02_dp, &
                                                     5.0E+02_dp, 6.0E+02_dp, 7.0E+02_dp, 8.0E+02_dp, 9.0E+02_dp, &
                                                     1.0E+03_dp, 2.0E+03_dp, 3.0E+03_dp, 4.0E+03_dp, 5.0E+03_dp, &
                                                     6.0E+03_dp, 7.0E+03_dp, 8.0E+03_dp, 9.0E+03_dp, 1.0E+04_dp, &
                                                     2.0E+04_dp, 3.0E+04_dp, 4.0E+04_dp, 5.0E+04_dp, 6.0E+04_dp, &
                                                     7.0E+04_dp, 8.0E+04_dp, 9.0E+04_dp, 1.0E+05_dp, 2.0E+05_dp, &
                                                     3.0E+05_dp, 4.0E+05_dp, 5.0E+05_dp, 6.0E+05_dp, 7.0E+05_dp, &
                                                     8.0E+05_dp, 9.0E+05_dp, 1.0E+06_dp, 2.0E+06_dp, 3.0E+06_dp, &
                                                     4.0E+06_dp, 5.0E+06_dp, 6.0E+06_dp, 7.0E+06_dp, 8.0E+06_dp, &
                                                     9.0E+06_dp, 1.0E+07_dp, 2.0E+07_dp, 3.0E+07_dp, 4.0E+07_dp, &
                                                     5.0E+07_dp, 6.0E+07_dp, 7.0E+07_dp, 8.0E+07_dp, 9.0E+07_dp, &
                                                     1.0E+08_dp, 2.0E+02_dp, 3.0E+02_dp, 4.0E+02_dp, 5.0E+02_dp, &
                                                     6.0E+02_dp, 7.0E+02_dp, 8.0E+02_dp, 9.0E+02_dp, 1.0E+03_dp, &
                                                     2.0E+03_dp, 3.0E+03_dp, 4.0E+03_dp, 5.0E+03_dp, 6.0E+03_dp, &
                                                     7.0E+03_dp, 8.0E+03_dp, 9.0E+03_dp, 1.0E+04_dp, 2.0E+04_dp, &
                                                     3.0E+04_dp, 4.0E+04_dp, 5.0E+04_dp, 6.0E+04_dp, 7.0E+04_dp, &
                                                     8.0E+04_dp, 9.0E+04_dp, 1.0E+05_dp, 2.0E+05_dp, 3.0E+05_dp, &
                                                     4.0E+05_dp, 5.0E+05_dp, 6.0E+05_dp, 7.0E+05_dp, 8.0E+05_dp, &
                                                     9.0E+05_dp, 1.0E+06_dp, 2.0E+06_dp, 3.0E+06_dp, 4.0E+06_dp, &
                                                     5.0E+06_dp, 6.0E+06_dp, 7.0E+06_dp, 8.0E+06_dp, 9.0E+06_dp, &
                                                     1.0E+07_dp, 2.0E+07_dp, 3.0E+07_dp, 4.0E+07_dp, 5.0E+07_dp, &
                                                     6.0E+07_dp, 7.0E+07_dp, 8.0E+07_dp, 9.0E+07_dp, 1.0E+08_dp, &
                                                     2.0E+08_dp, 4.0E+02_dp, 5.0E+02_dp, 6.0E+02_dp, 7.0E+02_dp, &
                                                     8.0E+02_dp, 9.0E+02_dp, 1.0E+03_dp, 2.0E+03_dp, 3.0E+03_dp, &
                                                     4.0E+03_dp, 5.0E+03_dp, 6.0E+03_dp, 7.0E+03_dp, 8.0E+03_dp, &
                                                     9.0E+03_dp, 1.0E+04_dp, 2.0E+04_dp, 3.0E+04_dp, 4.0E+04_dp, &
                                                     5.0E+04_dp, 6.0E+04_dp, 7.0E+04_dp, 8.0E+04_dp, 9.0E+04_dp, &
                                                     1.0E+05_dp, 2.0E+05_dp, 3.0E+05_dp, 4.0E+05_dp, 5.0E+05_dp, &
                                                     6.0E+05_dp, 7.0E+05_dp, 8.0E+05_dp, 9.0E+05_dp, 1.0E+06_dp, &
                                                     2.0E+06_dp, 3.0E+06_dp, 4.0E+06_dp, 5.0E+06_dp, 6.0E+06_dp, &
                                                     7.0E+06_dp, 8.0E+06_dp, 9.0E+06_dp, 1.0E+07_dp, 2.0E+07_dp, &
                                                     3.0E+07_dp, 4.0E+07_dp, 5.0E+07_dp, 6.0E+07_dp, 7.0E+07_dp, &
                                                     8.0E+07_dp, 9.0E+07_dp, 1.0E+08_dp, 2.0E+08_dp, 3.0E+08_dp, &
                                                     1.0E+04_dp, 2.0E+04_dp, 3.0E+04_dp, 4.0E+04_dp, 5.0E+04_dp, &
                                                     6.0E+04_dp, 7.0E+04_dp, 8.0E+04_dp, 9.0E+04_dp, 1.0E+05_dp, &
                                                     2.0E+05_dp, 3.0E+05_dp, 4.0E+05_dp, 5.0E+05_dp, 6.0E+05_dp, &
                                                     7.0E+05_dp, 8.0E+05_dp, 9.0E+05_dp, 1.0E+06_dp, 2.0E+06_dp, &
                                                     3.0E+06_dp, 4.0E+06_dp, 5.0E+06_dp, 6.0E+06_dp, 7.0E+06_dp, &
                                                     8.0E+06_dp, 9.0E+06_dp, 1.0E+07_dp, 2.0E+07_dp, 3.0E+07_dp, &
                                                     4.0E+07_dp, 5.0E+07_dp, 6.0E+07_dp, 7.0E+07_dp, 8.0E+07_dp, &
                                                     9.0E+07_dp, 1.0E+08_dp, 2.0E+08_dp, 3.0E+08_dp, 4.0E+08_dp, &
                                                     1.0E+08_dp, 2.0E+08_dp, 3.0E+08_dp, 4.0E+08_dp, 5.0E+08_dp, &
                                                     6.0E+08_dp, 7.0E+08_dp, 8.0E+08_dp, 9.0E+08_dp, 7.0E+08_dp, &
                                                     8.0E+08_dp, 9.0E+08_dp, 1.0E+09_dp, 2.0E+09_dp, 7.0E+08_dp, &
                                                     8.0E+08_dp, 9.0E+08_dp, 1.0E+09_dp, 2.0E+09_dp, 7.0E+08_dp, &
                                                     8.0E+08_dp, 9.0E+08_dp, 1.0E+09_dp, 2.0E+09_dp, 3.0E+09_dp, &
                                                     7.0E+08_dp, 8.0E+08_dp, 9.0E+08_dp, 1.0E+09_dp, 2.0E+09_dp, &
                                                     3.0E+09_dp, 4.0E+09_dp, 7.0E+08_dp, 8.0E+08_dp, 9.0E+08_dp, &
                                                     1.0E+09_dp, 2.0E+09_dp, 3.0E+09_dp, 4.0E+09_dp, 5.0E+09_dp, &
                                                     6.0E+09_dp, 7.0E+08_dp, 8.0E+08_dp, 9.0E+08_dp, 1.0E+09_dp, &
                                                     2.0E+09_dp, 3.0E+09_dp, 4.0E+09_dp, 5.0E+09_dp, 6.0E+09_dp, &
                                                     7.0E+09_dp, 8.0E+09_dp, 9.0E+09_dp, 1.0E+10_dp, 2.0E+10_dp, &
                                                     1.0E+10_dp, 2.0E+10_dp, 1.0E+10_dp, 2.0E+10_dp, 3.0E+10_dp, &
                                                     1.0E+10_dp, 2.0E+10_dp, 3.0E+10_dp, 4.0E+10_dp, 1.0E+10_dp, &
                                                     2.0E+10_dp, 3.0E+10_dp, 4.0E+10_dp, 5.0E+10_dp, 1.0E+10_dp, &
                                                     2.0E+10_dp, 3.0E+10_dp, 4.0E+10_dp, 5.0E+10_dp, 6.0E+10_dp, &
                                                     7.0E+10_dp, 1.0E+10_dp, 2.0E+10_dp, 3.0E+10_dp, 4.0E+10_dp, &
                                                     5.0E+10_dp, 6.0E+10_dp, 7.0E+10_dp, 8.0E+10_dp, 9.0E+10_dp, &
                                                     1.0E+11_dp, 1.0E+11_dp, 2.0E+11_dp, 1.0E+11_dp, 2.0E+11_dp, &
                                                     1.0E+11_dp, 2.0E+11_dp, 3.0E+11_dp, 1.0E+11_dp, 2.0E+11_dp, &
                                                     3.0E+11_dp, 4.0E+11_dp, 1.0E+11_dp, 2.0E+11_dp, 3.0E+11_dp, &
                                                     4.0E+11_dp, 5.0E+11_dp, 1.0E+11_dp, 2.0E+11_dp, 3.0E+11_dp, &
                                                     4.0E+11_dp, 5.0E+11_dp, 6.0E+11_dp, 7.0E+11_dp, 1.0E+11_dp, &
                                                     2.0E+11_dp, 3.0E+11_dp, 4.0E+11_dp, 5.0E+11_dp, 6.0E+11_dp, &
                                                     7.0E+11_dp, 8.0E+11_dp, 9.0E+11_dp, 1.0E+12_dp, 2.0E+08_dp, &
                                                     3.0E+08_dp, 4.0E+08_dp, 5.0E+08_dp, 6.0E+08_dp, 7.0E+08_dp, &
                                                     8.0E+08_dp, 9.0E+08_dp, 1.0E+09_dp, 2.0E+09_dp, 3.0E+09_dp, &
                                                     4.0E+09_dp, 5.0E+09_dp, 6.0E+09_dp, 7.0E+09_dp, 8.0E+09_dp, &
                                                     9.0E+09_dp, 1.0E+10_dp, 2.0E+10_dp, 3.0E+10_dp, 4.0E+10_dp, &
                                                     5.0E+10_dp, 6.0E+10_dp, 7.0E+10_dp, 8.0E+10_dp, 9.0E+10_dp, &
                                                     1.0E+11_dp, 2.0E+11_dp, 3.0E+11_dp, 4.0E+11_dp, 5.0E+11_dp, &
                                                     6.0E+11_dp, 7.0E+11_dp, 8.0E+11_dp, 9.0E+11_dp, 1.0E+12_dp, &
                                                     2.0E+12_dp, 1.0E+12_dp, 2.0E+12_dp, 1.0E+12_dp, 2.0E+12_dp, &
                                                     3.0E+12_dp, 4.0E+11_dp, 5.0E+11_dp, 6.0E+11_dp, 7.0E+11_dp, &
                                                     8.0E+11_dp, 9.0E+11_dp, 1.0E+12_dp, 2.0E+12_dp, 3.0E+12_dp, &
                                                     4.0E+12_dp]

   REAL(KIND=dp), PARAMETER, DIMENSION(n_approx) :: err_mm = &
                                                    [2.128E-02_dp, 4.358E-02_dp, 5.960E-02_dp, 7.075E-02_dp, 7.825E-02_dp, &
                                                     8.288E-02_dp, 8.516E-02_dp, 8.556E-02_dp, 8.556E-02_dp, 2.080E-04_dp, &
                                                     1.035E-03_dp, 2.191E-03_dp, 3.437E-03_dp, 4.659E-03_dp, 5.811E-03_dp, &
                                                     6.878E-03_dp, 7.857E-03_dp, 8.752E-03_dp, 1.448E-02_dp, 1.699E-02_dp, &
                                                     1.784E-02_dp, 1.785E-02_dp, 1.834E-06_dp, 2.223E-05_dp, 7.279E-05_dp, &
                                                     1.500E-04_dp, 2.463E-04_dp, 3.553E-04_dp, 4.718E-04_dp, 5.924E-04_dp, &
                                                     7.145E-04_dp, 1.819E-03_dp, 2.627E-03_dp, 3.215E-03_dp, 3.659E-03_dp, &
                                                     4.001E-03_dp, 4.271E-03_dp, 4.485E-03_dp, 4.655E-03_dp, 4.789E-03_dp, &
                                                     5.052E-03_dp, 1.542E-08_dp, 4.556E-07_dp, 2.311E-06_dp, 6.258E-06_dp, &
                                                     1.246E-05_dp, 2.079E-05_dp, 3.098E-05_dp, 4.273E-05_dp, 5.577E-05_dp, &
                                                     2.169E-04_dp, 3.795E-04_dp, 5.230E-04_dp, 6.469E-04_dp, 7.541E-04_dp, &
                                                     8.474E-04_dp, 9.293E-04_dp, 1.002E-03_dp, 1.066E-03_dp, 1.456E-03_dp, &
                                                     1.628E-03_dp, 1.695E-03_dp, 1.700E-03_dp, 1.261E-10_dp, 9.088E-09_dp, &
                                                     7.139E-08_dp, 2.543E-07_dp, 6.143E-07_dp, 1.185E-06_dp, 1.982E-06_dp, &
                                                     3.004E-06_dp, 4.243E-06_dp, 2.521E-05_dp, 5.336E-05_dp, 8.266E-05_dp, &
                                                     1.110E-04_dp, 1.377E-04_dp, 1.626E-04_dp, 1.858E-04_dp, 2.074E-04_dp, &
                                                     2.274E-04_dp, 3.707E-04_dp, 4.554E-04_dp, 5.117E-04_dp, 5.517E-04_dp, &
                                                     5.811E-04_dp, 6.031E-04_dp, 6.193E-04_dp, 6.309E-04_dp, 6.385E-04_dp, &
                                                     6.428E-04_dp, 1.012E-12_dp, 1.780E-10_dp, 2.167E-09_dp, 1.016E-08_dp, &
                                                     2.976E-08_dp, 6.643E-08_dp, 1.246E-07_dp, 2.076E-07_dp, 3.173E-07_dp, &
                                                     2.880E-06_dp, 7.379E-06_dp, 1.285E-05_dp, 1.872E-05_dp, 2.471E-05_dp, &
                                                     3.066E-05_dp, 3.648E-05_dp, 4.213E-05_dp, 4.760E-05_dp, 9.217E-05_dp, &
                                                     1.235E-04_dp, 1.467E-04_dp, 1.649E-04_dp, 1.795E-04_dp, 1.915E-04_dp, &
                                                     2.016E-04_dp, 2.102E-04_dp, 2.177E-04_dp, 2.570E-04_dp, 2.646E-04_dp, &
                                                     8.020E-15_dp, 3.444E-12_dp, 6.498E-11_dp, 4.007E-10_dp, 1.424E-09_dp, &
                                                     3.677E-09_dp, 7.741E-09_dp, 1.417E-08_dp, 2.344E-08_dp, 3.252E-07_dp, &
                                                     1.008E-06_dp, 1.973E-06_dp, 3.121E-06_dp, 4.382E-06_dp, 5.711E-06_dp, &
                                                     7.077E-06_dp, 8.458E-06_dp, 9.841E-06_dp, 2.261E-05_dp, 3.297E-05_dp, &
                                                     4.141E-05_dp, 4.842E-05_dp, 5.438E-05_dp, 5.952E-05_dp, 6.401E-05_dp, &
                                                     6.799E-05_dp, 7.153E-05_dp, 9.365E-05_dp, 1.047E-04_dp, 1.110E-04_dp, &
                                                     1.146E-04_dp, 1.162E-04_dp, 1.163E-04_dp, 1.716E-09_dp, 3.640E-08_dp, &
                                                     1.365E-07_dp, 3.003E-07_dp, 5.155E-07_dp, 7.701E-07_dp, 1.054E-06_dp, &
                                                     1.361E-06_dp, 1.683E-06_dp, 2.016E-06_dp, 5.498E-06_dp, 8.724E-06_dp, &
                                                     1.157E-05_dp, 1.407E-05_dp, 1.630E-05_dp, 1.829E-05_dp, 2.008E-05_dp, &
                                                     2.172E-05_dp, 2.321E-05_dp, 3.341E-05_dp, 3.926E-05_dp, 4.316E-05_dp, &
                                                     4.596E-05_dp, 4.807E-05_dp, 4.969E-05_dp, 5.096E-05_dp, 5.195E-05_dp, &
                                                     5.271E-05_dp, 5.392E-05_dp, 1.248E-10_dp, 4.045E-09_dp, 1.836E-08_dp, &
                                                     4.539E-08_dp, 8.456E-08_dp, 1.344E-07_dp, 1.933E-07_dp, 2.598E-07_dp, &
                                                     3.325E-07_dp, 4.103E-07_dp, 1.327E-06_dp, 2.292E-06_dp, 3.209E-06_dp, &
                                                     4.061E-06_dp, 4.848E-06_dp, 5.577E-06_dp, 6.252E-06_dp, 6.881E-06_dp, &
                                                     7.468E-06_dp, 1.180E-05_dp, 1.456E-05_dp, 1.654E-05_dp, 1.804E-05_dp, &
                                                     1.923E-05_dp, 2.021E-05_dp, 2.103E-05_dp, 2.172E-05_dp, 2.232E-05_dp, &
                                                     2.546E-05_dp, 2.611E-05_dp, 9.021E-12_dp, 4.470E-10_dp, 2.456E-09_dp, &
                                                     6.822E-09_dp, 1.380E-08_dp, 2.333E-08_dp, 3.525E-08_dp, 4.932E-08_dp, &
                                                     6.532E-08_dp, 8.303E-08_dp, 3.186E-07_dp, 5.986E-07_dp, 8.850E-07_dp, &
                                                     1.165E-06_dp, 1.434E-06_dp, 1.691E-06_dp, 1.935E-06_dp, 2.167E-06_dp, &
                                                     2.389E-06_dp, 4.140E-06_dp, 5.360E-06_dp, 6.284E-06_dp, 7.020E-06_dp, &
                                                     7.626E-06_dp, 8.137E-06_dp, 8.576E-06_dp, 8.958E-06_dp, 9.296E-06_dp, &
                                                     1.133E-05_dp, 1.230E-05_dp, 1.282E-05_dp, 1.307E-05_dp, 1.312E-05_dp, &
                                                     6.492E-13_dp, 4.918E-11_dp, 3.270E-10_dp, 1.021E-09_dp, 2.241E-09_dp, &
                                                     4.031E-09_dp, 6.398E-09_dp, 9.323E-09_dp, 1.278E-08_dp, 1.673E-08_dp, &
                                                     7.613E-08_dp, 1.557E-07_dp, 2.430E-07_dp, 3.326E-07_dp, 4.221E-07_dp, &
                                                     5.101E-07_dp, 5.960E-07_dp, 6.795E-07_dp, 7.605E-07_dp, 1.445E-06_dp, &
                                                     1.963E-06_dp, 2.375E-06_dp, 2.715E-06_dp, 3.004E-06_dp, 3.254E-06_dp, &
                                                     3.474E-06_dp, 3.669E-06_dp, 3.844E-06_dp, 4.975E-06_dp, 5.583E-06_dp, &
                                                     5.971E-06_dp, 6.242E-06_dp, 6.437E-06_dp, 6.580E-06_dp, 6.683E-06_dp, &
                                                     6.754E-06_dp, 6.795E-06_dp, 6.807E-06_dp, 4.654E-14_dp, 5.389E-12_dp, &
                                                     4.337E-11_dp, 1.522E-10_dp, 3.625E-10_dp, 6.939E-10_dp, 1.157E-09_dp, &
                                                     1.756E-09_dp, 2.490E-09_dp, 3.357E-09_dp, 1.812E-08_dp, 4.032E-08_dp, &
                                                     6.645E-08_dp, 9.463E-08_dp, 1.238E-07_dp, 1.533E-07_dp, 1.829E-07_dp, &
                                                     2.122E-07_dp, 2.412E-07_dp, 5.025E-07_dp, 7.157E-07_dp, 8.936E-07_dp, &
                                                     1.046E-06_dp, 1.178E-06_dp, 1.295E-06_dp, 1.400E-06_dp, 1.495E-06_dp, &
                                                     1.582E-06_dp, 2.171E-06_dp, 2.514E-06_dp, 2.748E-06_dp, 2.920E-06_dp, &
                                                     3.053E-06_dp, 3.159E-06_dp, 3.246E-06_dp, 3.318E-06_dp, 3.379E-06_dp, &
                                                     3.630E-06_dp, 3.630E-06_dp, 3.326E-15_dp, 5.888E-13_dp, 5.735E-12_dp, &
                                                     2.262E-11_dp, 5.847E-11_dp, 1.191E-10_dp, 2.085E-10_dp, 3.295E-10_dp, &
                                                     4.836E-10_dp, 6.716E-10_dp, 4.301E-09_dp, 1.041E-08_dp, 1.812E-08_dp, &
                                                     2.683E-08_dp, 3.618E-08_dp, 4.594E-08_dp, 5.593E-08_dp, 6.605E-08_dp, &
                                                     7.623E-08_dp, 1.741E-07_dp, 2.601E-07_dp, 3.351E-07_dp, 4.013E-07_dp, &
                                                     4.604E-07_dp, 5.138E-07_dp, 5.624E-07_dp, 6.070E-07_dp, 6.481E-07_dp, &
                                                     9.429E-07_dp, 1.127E-06_dp, 1.257E-06_dp, 1.357E-06_dp, 1.437E-06_dp, &
                                                     1.503E-06_dp, 1.558E-06_dp, 1.605E-06_dp, 1.646E-06_dp, 1.877E-06_dp, &
                                                     1.966E-06_dp, 1.984E-06_dp, 2.371E-16_dp, 6.414E-14_dp, 7.562E-13_dp, &
                                                     3.352E-12_dp, 9.405E-12_dp, 2.038E-11_dp, 3.749E-11_dp, 6.169E-11_dp, &
                                                     9.369E-11_dp, 1.340E-10_dp, 1.018E-09_dp, 2.681E-09_dp, 4.926E-09_dp, &
                                                     7.587E-09_dp, 1.055E-08_dp, 1.372E-08_dp, 1.706E-08_dp, 2.050E-08_dp, &
                                                     2.403E-08_dp, 6.017E-08_dp, 9.426E-08_dp, 1.253E-07_dp, 1.536E-07_dp, &
                                                     1.794E-07_dp, 2.032E-07_dp, 2.252E-07_dp, 2.457E-07_dp, 2.648E-07_dp, &
                                                     4.081E-07_dp, 5.029E-07_dp, 5.731E-07_dp, 6.283E-07_dp, 6.735E-07_dp, &
                                                     7.114E-07_dp, 7.440E-07_dp, 7.723E-07_dp, 7.973E-07_dp, 9.491E-07_dp, &
                                                     1.023E-06_dp, 1.066E-06_dp, 1.092E-06_dp, 1.105E-06_dp, 1.108E-06_dp, &
                                                     1.708E-17_dp, 6.972E-15_dp, 9.947E-14_dp, 4.956E-13_dp, 1.509E-12_dp, &
                                                     3.479E-12_dp, 6.723E-12_dp, 1.152E-11_dp, 1.811E-11_dp, 2.667E-11_dp, &
                                                     2.403E-10_dp, 6.888E-10_dp, 1.336E-09_dp, 2.140E-09_dp, 3.067E-09_dp, &
                                                     4.091E-09_dp, 5.190E-09_dp, 6.349E-09_dp, 7.555E-09_dp, 2.074E-08_dp, &
                                                     3.407E-08_dp, 4.673E-08_dp, 5.861E-08_dp, 6.972E-08_dp, 8.014E-08_dp, &
                                                     8.993E-08_dp, 9.917E-08_dp, 1.079E-07_dp, 1.762E-07_dp, 2.238E-07_dp, &
                                                     2.604E-07_dp, 2.898E-07_dp, 3.145E-07_dp, 3.356E-07_dp, 3.539E-07_dp, &
                                                     3.702E-07_dp, 3.847E-07_dp, 4.770E-07_dp, 5.262E-07_dp, 5.577E-07_dp, &
                                                     5.797E-07_dp, 5.959E-07_dp, 6.079E-07_dp, 6.169E-07_dp, 6.235E-07_dp, &
                                                     6.280E-07_dp, 6.311E-07_dp, 1.000E-18_dp, 7.562E-16_dp, 1.306E-14_dp, &
                                                     7.312E-14_dp, 2.417E-13_dp, 5.927E-13_dp, 1.203E-12_dp, 2.147E-12_dp, &
                                                     3.492E-12_dp, 5.298E-12_dp, 5.663E-11_dp, 1.766E-10_dp, 3.616E-10_dp, &
                                                     6.025E-10_dp, 8.902E-10_dp, 1.217E-09_dp, 1.576E-09_dp, 1.962E-09_dp, &
                                                     2.371E-09_dp, 7.134E-09_dp, 1.229E-08_dp, 1.739E-08_dp, 2.232E-08_dp, &
                                                     2.704E-08_dp, 3.154E-08_dp, 3.584E-08_dp, 3.995E-08_dp, 4.388E-08_dp, &
                                                     7.587E-08_dp, 9.939E-08_dp, 1.180E-07_dp, 1.334E-07_dp, 1.465E-07_dp, &
                                                     1.579E-07_dp, 1.679E-07_dp, 1.769E-07_dp, 1.850E-07_dp, 2.389E-07_dp, &
                                                     2.694E-07_dp, 2.899E-07_dp, 3.048E-07_dp, 3.164E-07_dp, 3.256E-07_dp, &
                                                     3.331E-07_dp, 3.393E-07_dp, 3.445E-07_dp, 3.658E-07_dp, 3.659E-07_dp, &
                                                     8.202E-17_dp, 1.711E-15_dp, 1.077E-14_dp, 3.863E-14_dp, 1.008E-13_dp, &
                                                     2.150E-13_dp, 4.004E-13_dp, 6.723E-13_dp, 1.050E-12_dp, 1.332E-11_dp, &
                                                     4.519E-11_dp, 9.771E-11_dp, 1.693E-10_dp, 2.579E-10_dp, 3.613E-10_dp, &
                                                     4.776E-10_dp, 6.052E-10_dp, 7.426E-10_dp, 2.450E-09_dp, 4.425E-09_dp, &
                                                     6.461E-09_dp, 8.484E-09_dp, 1.046E-08_dp, 1.239E-08_dp, 1.425E-08_dp, &
                                                     1.606E-08_dp, 1.780E-08_dp, 3.261E-08_dp, 4.404E-08_dp, 5.336E-08_dp, &
                                                     6.124E-08_dp, 6.807E-08_dp, 7.409E-08_dp, 7.947E-08_dp, 8.433E-08_dp, &
                                                     8.877E-08_dp, 1.193E-07_dp, 1.375E-07_dp, 1.501E-07_dp, 1.597E-07_dp, &
                                                     1.672E-07_dp, 1.734E-07_dp, 1.785E-07_dp, 1.829E-07_dp, 1.867E-07_dp, &
                                                     2.076E-07_dp, 2.148E-07_dp, 2.155E-07_dp, 1.584E-15_dp, 6.165E-15_dp, &
                                                     1.711E-14_dp, 3.834E-14_dp, 7.420E-14_dp, 1.292E-13_dp, 2.079E-13_dp, &
                                                     3.128E-12_dp, 1.155E-11_dp, 2.636E-11_dp, 4.750E-11_dp, 7.459E-11_dp, &
                                                     1.071E-10_dp, 1.445E-10_dp, 1.864E-10_dp, 2.322E-10_dp, 8.396E-10_dp, &
                                                     1.590E-09_dp, 2.396E-09_dp, 3.220E-09_dp, 4.043E-09_dp, 4.858E-09_dp, &
                                                     5.659E-09_dp, 6.445E-09_dp, 7.213E-09_dp, 1.399E-08_dp, 1.948E-08_dp, &
                                                     2.409E-08_dp, 2.807E-08_dp, 3.157E-08_dp, 3.471E-08_dp, 3.754E-08_dp, &
                                                     4.013E-08_dp, 4.251E-08_dp, 5.942E-08_dp, 6.997E-08_dp, 7.755E-08_dp, &
                                                     8.339E-08_dp, 8.810E-08_dp, 9.202E-08_dp, 9.534E-08_dp, 9.822E-08_dp, &
                                                     1.007E-07_dp, 1.157E-07_dp, 1.227E-07_dp, 1.265E-07_dp, 1.284E-07_dp, &
                                                     1.289E-07_dp, 9.826E-16_dp, 2.902E-15_dp, 6.829E-15_dp, 1.376E-14_dp, &
                                                     2.480E-14_dp, 4.110E-14_dp, 7.333E-13_dp, 2.946E-12_dp, 7.099E-12_dp, &
                                                     1.331E-11_dp, 2.154E-11_dp, 3.170E-11_dp, 4.367E-11_dp, 5.732E-11_dp, &
                                                     7.251E-11_dp, 2.874E-10_dp, 5.708E-10_dp, 8.873E-10_dp, 1.220E-09_dp, &
                                                     1.560E-09_dp, 1.902E-09_dp, 2.244E-09_dp, 2.582E-09_dp, 2.918E-09_dp, &
                                                     5.993E-09_dp, 8.602E-09_dp, 1.086E-08_dp, 1.284E-08_dp, 1.462E-08_dp, &
                                                     1.623E-08_dp, 1.770E-08_dp, 1.906E-08_dp, 2.032E-08_dp, 2.955E-08_dp, &
                                                     3.555E-08_dp, 3.997E-08_dp, 4.346E-08_dp, 4.631E-08_dp, 4.872E-08_dp, &
                                                     5.079E-08_dp, 5.261E-08_dp, 5.421E-08_dp, 6.415E-08_dp, 6.923E-08_dp, &
                                                     7.240E-08_dp, 7.454E-08_dp, 7.603E-08_dp, 7.706E-08_dp, 7.773E-08_dp, &
                                                     7.806E-08_dp, 7.811E-08_dp, 4.914E-16_dp, 1.215E-15_dp, 2.549E-15_dp, &
                                                     4.754E-15_dp, 8.114E-15_dp, 1.717E-13_dp, 7.507E-13_dp, 1.910E-12_dp, &
                                                     3.723E-12_dp, 6.213E-12_dp, 9.371E-12_dp, 1.318E-11_dp, 1.760E-11_dp, &
                                                     2.261E-11_dp, 9.824E-11_dp, 2.046E-10_dp, 3.281E-10_dp, 4.617E-10_dp, &
                                                     6.011E-10_dp, 7.438E-10_dp, 8.883E-10_dp, 1.033E-09_dp, 1.179E-09_dp, &
                                                     2.564E-09_dp, 3.793E-09_dp, 4.886E-09_dp, 5.868E-09_dp, 6.761E-09_dp, &
                                                     7.580E-09_dp, 8.337E-09_dp, 9.041E-09_dp, 9.700E-09_dp, 1.467E-08_dp, &
                                                     1.803E-08_dp, 2.057E-08_dp, 2.260E-08_dp, 2.430E-08_dp, 2.574E-08_dp, &
                                                     2.700E-08_dp, 2.812E-08_dp, 2.911E-08_dp, 3.548E-08_dp, 3.893E-08_dp, &
                                                     4.118E-08_dp, 4.279E-08_dp, 4.401E-08_dp, 4.495E-08_dp, 4.571E-08_dp, &
                                                     4.631E-08_dp, 4.679E-08_dp, 4.794E-08_dp, 9.103E-16_dp, 1.600E-15_dp, &
                                                     4.017E-14_dp, 1.910E-13_dp, 5.132E-13_dp, 1.040E-12_dp, 1.790E-12_dp, &
                                                     2.767E-12_dp, 3.972E-12_dp, 5.400E-12_dp, 7.044E-12_dp, 3.354E-11_dp, &
                                                     7.325E-11_dp, 1.212E-10_dp, 1.745E-10_dp, 2.313E-10_dp, 2.905E-10_dp, &
                                                     3.513E-10_dp, 4.131E-10_dp, 4.755E-10_dp, 1.095E-09_dp, 1.671E-09_dp, &
                                                     2.196E-09_dp, 2.678E-09_dp, 3.122E-09_dp, 3.535E-09_dp, 3.921E-09_dp, &
                                                     4.283E-09_dp, 4.624E-09_dp, 7.274E-09_dp, 9.131E-09_dp, 1.057E-08_dp, &
                                                     1.174E-08_dp, 1.273E-08_dp, 1.358E-08_dp, 1.433E-08_dp, 1.500E-08_dp, &
                                                     1.560E-08_dp, 1.958E-08_dp, 2.184E-08_dp, 2.336E-08_dp, 2.448E-08_dp, &
                                                     2.535E-08_dp, 2.605E-08_dp, 2.662E-08_dp, 2.711E-08_dp, 2.752E-08_dp, &
                                                     2.955E-08_dp, 2.976E-08_dp, 3.153E-16_dp, 9.385E-15_dp, 4.857E-14_dp, &
                                                     1.377E-13_dp, 2.905E-13_dp, 5.151E-13_dp, 8.162E-13_dp, 1.196E-12_dp, &
                                                     1.655E-12_dp, 2.192E-12_dp, 1.144E-11_dp, 2.620E-11_dp, 4.472E-11_dp, &
                                                     6.588E-11_dp, 8.892E-11_dp, 1.133E-10_dp, 1.388E-10_dp, 1.649E-10_dp, &
                                                     1.916E-10_dp, 4.674E-10_dp, 7.350E-10_dp, 9.860E-10_dp, 1.221E-09_dp, &
                                                     1.440E-09_dp, 1.647E-09_dp, 1.842E-09_dp, 2.026E-09_dp, 2.201E-09_dp, &
                                                     3.602E-09_dp, 4.619E-09_dp, 5.422E-09_dp, 6.089E-09_dp, 6.658E-09_dp, &
                                                     7.155E-09_dp, 7.596E-09_dp, 7.992E-09_dp, 8.351E-09_dp, 1.079E-08_dp, &
                                                     1.223E-08_dp, 1.322E-08_dp, 1.397E-08_dp, 1.457E-08_dp, 1.505E-08_dp, &
                                                     1.546E-08_dp, 1.580E-08_dp, 1.611E-08_dp, 1.780E-08_dp, 1.848E-08_dp, &
                                                     1.868E-08_dp, 2.191E-15_dp, 1.234E-14_dp, 3.693E-14_dp, 8.101E-14_dp, &
                                                     1.481E-13_dp, 2.405E-13_dp, 3.597E-13_dp, 5.065E-13_dp, 6.813E-13_dp, &
                                                     3.898E-12_dp, 9.359E-12_dp, 1.648E-11_dp, 2.485E-11_dp, 3.415E-11_dp, &
                                                     4.417E-11_dp, 5.475E-11_dp, 6.578E-11_dp, 7.715E-11_dp, 1.993E-10_dp, &
                                                     3.230E-10_dp, 4.422E-10_dp, 5.557E-10_dp, 6.636E-10_dp, 7.663E-10_dp, &
                                                     8.641E-10_dp, 9.576E-10_dp, 1.047E-09_dp, 1.782E-09_dp, 2.334E-09_dp, &
                                                     2.779E-09_dp, 3.154E-09_dp, 3.479E-09_dp, 3.765E-09_dp, 4.021E-09_dp, &
                                                     4.253E-09_dp, 4.464E-09_dp, 5.936E-09_dp, 6.834E-09_dp, 7.471E-09_dp, &
                                                     7.960E-09_dp, 8.353E-09_dp, 8.679E-09_dp, 8.956E-09_dp, 9.194E-09_dp, &
                                                     9.404E-09_dp, 1.065E-08_dp, 1.124E-08_dp, 1.157E-08_dp, 1.176E-08_dp, &
                                                     1.184E-08_dp, 1.185E-08_dp, 5.111E-16_dp, 3.131E-15_dp, 9.895E-15_dp, &
                                                     2.278E-14_dp, 4.254E-14_dp, 7.080E-14_dp, 1.081E-13_dp, 1.549E-13_dp, &
                                                     2.116E-13_dp, 1.327E-12_dp, 3.341E-12_dp, 6.071E-12_dp, 9.364E-12_dp, &
                                                     1.310E-11_dp, 1.720E-11_dp, 2.159E-11_dp, 2.621E-11_dp, 3.103E-11_dp, &
                                                     8.488E-11_dp, 1.418E-10_dp, 1.981E-10_dp, 2.528E-10_dp, 3.055E-10_dp, &
                                                     3.562E-10_dp, 4.051E-10_dp, 4.521E-10_dp, 4.975E-10_dp, 8.804E-10_dp, &
                                                     1.178E-09_dp, 1.423E-09_dp, 1.632E-09_dp, 1.816E-09_dp, 1.979E-09_dp, &
                                                     2.126E-09_dp, 2.260E-09_dp, 2.384E-09_dp, 3.262E-09_dp, 3.815E-09_dp, &
                                                     4.216E-09_dp, 4.528E-09_dp, 4.783E-09_dp, 4.997E-09_dp, 5.180E-09_dp, &
                                                     5.340E-09_dp, 5.481E-09_dp, 6.349E-09_dp, 6.791E-09_dp, 7.067E-09_dp, &
                                                     7.254E-09_dp, 7.385E-09_dp, 7.478E-09_dp, 7.540E-09_dp, 7.575E-09_dp, &
                                                     7.583E-09_dp, 1.192E-16_dp, 7.939E-16_dp, 2.649E-15_dp, 6.284E-15_dp, &
                                                     1.221E-14_dp, 2.083E-14_dp, 3.246E-14_dp, 4.734E-14_dp, 6.566E-14_dp, &
                                                     4.514E-13_dp, 1.192E-12_dp, 2.234E-12_dp, 3.526E-12_dp, 5.023E-12_dp, &
                                                     6.692E-12_dp, 8.504E-12_dp, 1.044E-11_dp, 1.247E-11_dp, 3.612E-11_dp, &
                                                     6.222E-11_dp, 8.870E-11_dp, 1.149E-10_dp, 1.405E-10_dp, 1.655E-10_dp, &
                                                     1.897E-10_dp, 2.133E-10_dp, 2.362E-10_dp, 4.347E-10_dp, 5.939E-10_dp, &
                                                     7.278E-10_dp, 8.439E-10_dp, 9.467E-10_dp, 1.039E-09_dp, 1.123E-09_dp, &
                                                     1.200E-09_dp, 1.272E-09_dp, 1.791E-09_dp, 2.127E-09_dp, 2.376E-09_dp, &
                                                     2.573E-09_dp, 2.735E-09_dp, 2.873E-09_dp, 2.992E-09_dp, 3.096E-09_dp, &
                                                     3.190E-09_dp, 3.779E-09_dp, 4.094E-09_dp, 4.298E-09_dp, 4.444E-09_dp, &
                                                     4.553E-09_dp, 4.638E-09_dp, 4.705E-09_dp, 4.759E-09_dp, 4.802E-09_dp, &
                                                     4.898E-09_dp, 7.087E-16_dp, 1.749E-15_dp, 3.501E-15_dp, 6.122E-15_dp, &
                                                     9.739E-15_dp, 1.446E-14_dp, 2.036E-14_dp, 1.534E-13_dp, 4.246E-13_dp, &
                                                     8.214E-13_dp, 1.327E-12_dp, 1.924E-12_dp, 2.602E-12_dp, 3.347E-12_dp, &
                                                     4.152E-12_dp, 5.009E-12_dp, 1.536E-11_dp, 2.727E-11_dp, 3.968E-11_dp, &
                                                     5.218E-11_dp, 6.458E-11_dp, 7.679E-11_dp, 8.879E-11_dp, 1.005E-10_dp, &
                                                     1.120E-10_dp, 2.144E-10_dp, 2.992E-10_dp, 3.719E-10_dp, 4.359E-10_dp, &
                                                     4.932E-10_dp, 5.451E-10_dp, 5.928E-10_dp, 6.368E-10_dp, 6.777E-10_dp, &
                                                     9.820E-10_dp, 1.185E-09_dp, 1.338E-09_dp, 1.460E-09_dp, 1.562E-09_dp, &
                                                     1.650E-09_dp, 1.726E-09_dp, 1.793E-09_dp, 1.854E-09_dp, 2.246E-09_dp, &
                                                     2.464E-09_dp, 2.609E-09_dp, 2.715E-09_dp, 2.797E-09_dp, 2.863E-09_dp, &
                                                     2.916E-09_dp, 2.961E-09_dp, 2.999E-09_dp, 3.179E-09_dp, 3.190E-09_dp, &
                                                     2.010E-12_dp, 6.528E-12_dp, 1.195E-11_dp, 1.774E-11_dp, 2.368E-11_dp, &
                                                     2.966E-11_dp, 3.561E-11_dp, 4.152E-11_dp, 4.735E-11_dp, 5.310E-11_dp, &
                                                     1.057E-10_dp, 1.506E-10_dp, 1.899E-10_dp, 2.250E-10_dp, 2.567E-10_dp, &
                                                     2.858E-10_dp, 3.126E-10_dp, 3.375E-10_dp, 3.609E-10_dp, 5.381E-10_dp, &
                                                     6.595E-10_dp, 7.525E-10_dp, 8.280E-10_dp, 8.916E-10_dp, 9.465E-10_dp, &
                                                     9.947E-10_dp, 1.038E-09_dp, 1.076E-09_dp, 1.333E-09_dp, 1.480E-09_dp, &
                                                     1.581E-09_dp, 1.656E-09_dp, 1.715E-09_dp, 1.763E-09_dp, 1.803E-09_dp, &
                                                     1.837E-09_dp, 1.866E-09_dp, 2.028E-09_dp, 2.086E-09_dp, 2.094E-09_dp, &
                                                     1.159E-09_dp, 1.284E-09_dp, 1.341E-09_dp, 1.371E-09_dp, 1.384E-09_dp, &
                                                     1.385E-09_dp, 9.203E-10_dp, 9.227E-10_dp, 9.227E-10_dp, 6.020E-10_dp, &
                                                     6.083E-10_dp, 6.130E-10_dp, 6.162E-10_dp, 6.188E-10_dp, 3.912E-10_dp, &
                                                     3.969E-10_dp, 4.015E-10_dp, 4.053E-10_dp, 4.177E-10_dp, 2.536E-10_dp, &
                                                     2.581E-10_dp, 2.619E-10_dp, 2.651E-10_dp, 2.814E-10_dp, 2.837E-10_dp, &
                                                     1.642E-10_dp, 1.676E-10_dp, 1.705E-10_dp, 1.730E-10_dp, 1.870E-10_dp, &
                                                     1.925E-10_dp, 1.938E-10_dp, 1.062E-10_dp, 1.087E-10_dp, 1.109E-10_dp, &
                                                     1.128E-10_dp, 1.238E-10_dp, 1.289E-10_dp, 1.316E-10_dp, 1.329E-10_dp, &
                                                     1.331E-10_dp, 6.860E-11_dp, 7.044E-11_dp, 7.203E-11_dp, 7.343E-11_dp, &
                                                     8.181E-11_dp, 8.593E-11_dp, 8.842E-11_dp, 9.004E-11_dp, 9.109E-11_dp, &
                                                     9.171E-11_dp, 9.194E-11_dp, 9.194E-11_dp, 6.365E-11_dp, 6.382E-11_dp, &
                                                     4.353E-11_dp, 4.452E-11_dp, 2.962E-11_dp, 3.110E-11_dp, 3.121E-11_dp, &
                                                     2.011E-11_dp, 2.147E-11_dp, 2.194E-11_dp, 2.197E-11_dp, 1.364E-11_dp, &
                                                     1.476E-11_dp, 1.525E-11_dp, 1.548E-11_dp, 1.554E-11_dp, 9.246E-12_dp, &
                                                     1.013E-11_dp, 1.055E-11_dp, 1.079E-11_dp, 1.094E-11_dp, 1.102E-11_dp, &
                                                     1.104E-11_dp, 6.262E-12_dp, 6.943E-12_dp, 7.284E-12_dp, 7.494E-12_dp, &
                                                     7.635E-12_dp, 7.733E-12_dp, 7.801E-12_dp, 7.845E-12_dp, 7.866E-12_dp, &
                                                     7.869E-12_dp, 5.591E-12_dp, 5.633E-12_dp, 3.938E-12_dp, 4.047E-12_dp, &
                                                     2.766E-12_dp, 2.902E-12_dp, 2.919E-12_dp, 1.940E-12_dp, 2.063E-12_dp, &
                                                     2.107E-12_dp, 2.113E-12_dp, 1.360E-12_dp, 1.462E-12_dp, 1.507E-12_dp, &
                                                     1.529E-12_dp, 1.534E-12_dp, 9.524E-13_dp, 1.035E-12_dp, 1.074E-12_dp, &
                                                     1.096E-12_dp, 1.110E-12_dp, 1.117E-12_dp, 1.118E-12_dp, 6.667E-13_dp, &
                                                     7.317E-13_dp, 7.639E-13_dp, 7.837E-13_dp, 7.969E-13_dp, 8.060E-13_dp, &
                                                     8.121E-13_dp, 8.158E-13_dp, 8.172E-13_dp, 8.172E-13_dp, 4.429E-14_dp, &
                                                     5.718E-14_dp, 6.780E-14_dp, 7.692E-14_dp, 8.496E-14_dp, 9.218E-14_dp, &
                                                     9.874E-14_dp, 1.048E-13_dp, 1.103E-13_dp, 1.515E-13_dp, 1.790E-13_dp, &
                                                     1.998E-13_dp, 2.167E-13_dp, 2.309E-13_dp, 2.432E-13_dp, 2.540E-13_dp, &
                                                     2.637E-13_dp, 2.724E-13_dp, 3.314E-13_dp, 3.664E-13_dp, 3.911E-13_dp, &
                                                     4.100E-13_dp, 4.253E-13_dp, 4.380E-13_dp, 4.488E-13_dp, 4.582E-13_dp, &
                                                     4.664E-13_dp, 5.171E-13_dp, 5.430E-13_dp, 5.594E-13_dp, 5.708E-13_dp, &
                                                     5.792E-13_dp, 5.855E-13_dp, 5.903E-13_dp, 5.940E-13_dp, 5.966E-13_dp, &
                                                     5.992E-13_dp, 4.321E-13_dp, 4.407E-13_dp, 3.120E-13_dp, 3.245E-13_dp, &
                                                     3.251E-13_dp, 2.025E-13_dp, 2.086E-13_dp, 2.133E-13_dp, 2.170E-13_dp, &
                                                     2.202E-13_dp, 2.228E-13_dp, 2.251E-13_dp, 2.371E-13_dp, 2.405E-13_dp, &
                                                     2.405E-13_dp]

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param i ...
!> \param aw ...
! **************************************************************************************************
   SUBROUTINE get_minimax_coeff_low(i, aw)
      INTEGER, INTENT(IN)                                :: i
      REAL(KIND=dp), DIMENSION(k_mm(i)*2), INTENT(OUT)   :: aw

      SELECT CASE (i)

      CASE (1)

         aw(:) = & ! a
            [0.7151291879699049_dp, &
             ! w
             2.000945890508967_dp]

      CASE (2)

         aw(:) = & ! a
            [0.5970833669667293_dp, &
             ! w
             1.737635742582102_dp]

      CASE (3)

         aw(:) = & ! a
            [0.5323920576673996_dp, &
             ! w
             1.601503072363919_dp]

      CASE (4)

         aw(:) = & ! a
            [0.49316306690405565_dp, &
             ! w
             1.5216296303352121_dp]

      CASE (5)

         aw(:) = & ! a
            [0.4688474749950737_dp, &
             ! w
             1.4730921650589592_dp]

      CASE (6)

         aw(:) = & ! a
            [0.4545496122972507_dp, &
             ! w
             1.4448869123259267_dp]

      CASE (7)

         aw(:) = & ! a
            [0.4476963168465264_dp, &
             ! w
             1.4314535341251469_dp]

      CASE (8)

         aw(:) = & ! a
            [0.44649260629880866_dp, &
             ! w
             1.429099786979271_dp]

      CASE (9)

         aw(:) = & ! a
            [0.4464926062994778_dp, &
             ! w
             1.4290997869805788_dp]

      CASE (10)

         aw(:) = & ! a
            [0.4072692399097063_dp, 2.4428035233716465_dp, &
             ! w
             1.0724401884999553_dp, 3.2918896543848644_dp]

      CASE (11)

         aw(:) = & ! a
            [0.32451524499523515_dp, 2.035129346922339_dp, &
             ! w
             0.8627352798623843_dp, 2.8723614320211905_dp]

      CASE (12)

         aw(:) = & ! a
            [0.2748454853769434_dp, 1.8026054538227607_dp, &
             ! w
             0.7381177111390874_dp, 2.6510075653784217_dp]

      CASE (13)

         aw(:) = & ! a
            [0.2412189720039558_dp, 1.6503932424456509_dp, &
             ! w
             0.654413821579126_dp, 2.5128880395953934_dp]

      CASE (14)

         aw(:) = & ! a
            [0.21677626323385438_dp, 1.5424317633410847_dp, &
             ! w
             0.5939961099849468_dp, 2.417973551456907_dp]

      CASE (15)

         aw(:) = & ! a
            [0.19814373071471802_dp, 1.461666823401535_dp, &
             ! w
             0.5482353429308007_dp, 2.348496068608947_dp]

      CASE (16)

         aw(:) = & ! a
            [0.18344320998999283_dp, 1.3988865294263442_dp, &
             ! w
             0.5123441656997128_dp, 2.2953108404122684_dp]

      CASE (17)

         aw(:) = & ! a
            [0.17153706461006926_dp, 1.3486468789878938_dp, &
             ! w
             0.4834323382159764_dp, 2.2532140530550535_dp]

      CASE (18)

         aw(:) = & ! a
            [0.16169273989379984_dp, 1.307514133065234_dp, &
             ! w
             0.4596450781301019_dp, 2.2190206174604987_dp]

      CASE (19)

         aw(:) = & ! a
            [0.11377055334818492_dp, 1.112242083226372_dp, &
             ! w
             0.34568839455577144_dp, 2.0588931989746815_dp]

      CASE (20)

         aw(:) = & ! a
            [0.09798314302456711_dp, 1.0494466863388991_dp, &
             ! w
             0.30892698728560847_dp, 2.00757054644027_dp]

      CASE (21)

         aw(:) = & ! a
            [0.09318559554449489_dp, 1.0304678137783845_dp, &
             ! w
             0.297838783224004_dp, 1.9920021768207374_dp]

      CASE (22)

         aw(:) = & ! a
            [0.09310617256883674_dp, 1.0301539335400713_dp, &
             ! w
             0.2976555432327633_dp, 1.9917443679019897_dp]

      CASE (23)

         aw(:) = & ! a
            [0.2874933084171688_dp, 1.6046370491507347_dp, 4.506486437639924_dp, &
             ! w
             0.7470594125158281_dp, 1.9588903206974837_dp, 4.161788370499587_dp]

      CASE (24)

         aw(:) = & ! a
            [0.22719994406103003_dp, 1.290960582912292_dp, 3.7605449827660413_dp, &
             ! w
             0.5926726335139195_dp, 1.6100816777502402_dp, 3.6515813642302706_dp]

      CASE (25)

         aw(:) = & ! a
            [0.190820901834524_dp, 1.1043195988478427_dp, 3.334024351088452_dp, &
             ! w
             0.49973085683672597_dp, 1.4071850340285912_dp, 3.383430297960359_dp]

      CASE (26)

         aw(:) = & ! a
            [0.16607313750141492_dp, 0.9784372085453776_dp, 3.0530159808682455_dp, &
             ! w
             0.4365792105033684_dp, 1.2723305210744182_dp, 3.2151540832501007_dp]

      CASE (27)

         aw(:) = & ! a
            [0.14797546872033812_dp, 0.8869636853585635_dp, 2.851850660065598_dp, &
             ! w
             0.39043661216740244_dp, 1.1753669469725918_dp, 3.0981972422591575_dp]

      CASE (28)

         aw(:) = & ! a
            [0.13407826516448915_dp, 0.8170825513653607_dp, 2.699682964801345_dp, &
             ! w
             0.35503082305979605_dp, 1.1018792658187146_dp, 3.011330116355922_dp]

      CASE (29)

         aw(:) = & ! a
            [0.12302217745120096_dp, 0.7617320987617905_dp, 2.579950751689479_dp, &
             ! w
             0.32688491641152817_dp, 1.0440277074411328_dp, 2.943745649391346_dp]

      CASE (30)

         aw(:) = & ! a
            [0.11398687647296017_dp, 0.716672821402158_dp, 2.4829070960193653_dp, &
             ! w
             0.30390175304476086_dp, 0.9971555180958862_dp, 2.8893365676093965_dp]

      CASE (31)

         aw(:) = & ! a
            [0.10644554843011034_dp, 0.6791933646735415_dp, 2.402416310116676_dp, &
             ! w
             0.2847348675836868_dp, 0.9583115118002355_dp, 2.844377225487889_dp]

      CASE (32)

         aw(:) = & ! a
            [0.06790136894058926_dp, 0.4897948875089661_dp, 1.996201580919574_dp, &
             ! w
             0.18718659453518208_dp, 0.7633857188437595_dp, 2.6169711142660255_dp]

      CASE (33)

         aw(:) = & ! a
            [0.0526373510053126_dp, 0.4159550868841871_dp, 1.835980917898024_dp, &
             ! w
             0.14893502807079687_dp, 0.6872993149781067_dp, 2.5251745372914685_dp]

      CASE (34)

         aw(:) = & ! a
            [0.044324908247723685_dp, 0.3760267128377693_dp, 1.7481061682069727_dp, &
             ! w
             0.1282763604743763_dp, 0.6457800255491173_dp, 2.4737466774457357_dp]

      CASE (35)

         aw(:) = & ! a
            [0.039082098432306915_dp, 0.3509230683778315_dp, 1.6921947262166477_dp, &
             ! w
             0.11533107566965783_dp, 0.6194232578183114_dp, 2.440503632678777_dp]

      CASE (36)

         aw(:) = & ! a
            [0.03548785756339287_dp, 0.33373310920779525_dp, 1.6535367452293752_dp, &
             ! w
             0.1065007940761847_dp, 0.6012169550303839_dp, 2.417244683821898_dp]

      CASE (37)

         aw(:) = & ! a
            [0.0328950169895425_dp, 0.32133337767322623_dp, 1.6254323321274151_dp, &
             ! w
             0.10015530773686102_dp, 0.5879845498502224_dp, 2.4001809699912178_dp]

      CASE (38)

         aw(:) = & ! a
            [0.030966040572607583_dp, 0.3121038074526534_dp, 1.6043802276101924_dp, &
             ! w
             0.09544856149463261_dp, 0.5780719785823824_dp, 2.3873083432732534_dp]

      CASE (39)

         aw(:) = & ! a
            [0.029508058669578938_dp, 0.3051223245003003_dp, 1.5883739921280895_dp, &
             ! w
             0.09189920985752194_dp, 0.5705336658200345_dp, 2.3774665399131085_dp]

      CASE (40)

         aw(:) = & ! a
            [0.028403806288507458_dp, 0.2998300711146907_dp, 1.5761907929646568_dp, &
             ! w
             0.08921571278587176_dp, 0.5647942698268908_dp, 2.369942830136091_dp]

      CASE (41)

         aw(:) = & ! a
            [0.026349272848825177_dp, 0.2899685548514263_dp, 1.5533661852864638_dp, &
             ! w
             0.08423369646700837_dp, 0.5540364331569211_dp, 2.355768842927211_dp]

      CASE (42)

         aw(:) = & ! a
            [0.22252821013426086_dp, 1.2118271769428668_dp, 3.187481792030648_dp, &
             6.739887374833184_dp, &
             ! w
             0.5753086924118194_dp, 1.4324382751398494_dp, 2.601504267946919_dp, &
             4.837282274455988_dp]

      CASE (43)

         aw(:) = & ! a
            [0.17527105138036778_dp, 0.9635165380263817_dp, 2.5825322433881674_dp, &
             5.633585403420891_dp, &
             ! w
             0.45408560693614775_dp, 1.1521496451107505_dp, 2.170818682065073_dp, &
             4.2622256210860705_dp]

      CASE (44)

         aw(:) = & ! a
            [0.14671950647800755_dp, 0.8142596432786335_dp, 2.2245993655671508_dp, &
             5.000745372447783_dp, &
             ! w
             0.38090986643822805_dp, 0.98502848337898_dp, 1.9245224600485114_dp, &
             3.961347573512278_dp]

      CASE (45)

         aw(:) = & ! a
            [0.1272851568278456_dp, 0.7128840158683422_dp, 1.9837003473579684_dp, &
             4.582740315117001_dp, &
             ! w
             0.33111140294414865_dp, 0.8719992880247881_dp, 1.7621447067201108_dp, &
             3.7722154835897865_dp]

      CASE (46)

         aw(:) = & ! a
            [0.11306686288605404_dp, 0.6387959915512463_dp, 1.8086875076613338_dp, &
             4.282405371555544_dp, &
             ! w
             0.2946777813201027_dp, 0.78961356147058_dp, 1.645743241597089_dp, &
             3.640143062267599_dp]

      CASE (47)

         aw(:) = & ! a
            [0.10214418415560887_dp, 0.5819143686208856_dp, 1.6748643119314177_dp, &
             4.054266665255846_dp, &
             ! w
             0.26668604228316484_dp, 0.7264788634360871_dp, 1.5575123311575398_dp, &
             3.541449719416932_dp]

      CASE (48)

         aw(:) = & ! a
            [0.09345102611264643_dp, 0.5366603519779132_dp, 1.5686956963951062_dp, &
             3.8739609091686398_dp, &
             ! w
             0.24440500260960255_dp, 0.6763200419930573_dp, 1.4879051296978405_dp, &
             3.464152056212784_dp]

      CASE (49)

         aw(:) = & ! a
            [0.08634379244020035_dp, 0.4996723020499699_dp, 1.4820860449391133_dp, &
             3.727163003860314_dp, &
             ! w
             0.2261863808727144_dp, 0.6353676827863498_dp, 1.4313117857432134_dp, &
             3.4014990384232084_dp]

      CASE (50)

         aw(:) = & ! a
            [0.08040902441555287_dp, 0.46879309466888663_dp, 1.409871537871361_dp, &
             3.6048552773416724_dp, &
             ! w
             0.21097144230678308_dp, 0.6012085214120437_dp, 1.3842061450697847_dp, &
             3.3493765842990433_dp]

      CASE (51)

         aw(:) = & ! a
            [0.04996987754589871_dp, 0.3105631707014664_dp, 1.0398249107852826_dp, &
             2.9750469432139246_dp, &
             ! w
             0.13291800704161472_dp, 0.42649085342514365_dp, 1.1416477627542354_dp, &
             3.078067534660662_dp]

      CASE (52)

         aw(:) = & ! a
            [0.03774937905247896_dp, 0.24717316890019844_dp, 0.8901553582336263_dp, &
             2.7154179633949425_dp, &
             ! w
             0.1016013598614176_dp, 0.35645000765859947_dp, 1.0408682232432722_dp, &
             2.962067059993243_dp]

      CASE (53)

         aw(:) = & ! a
            [0.030971882763324972_dp, 0.21207745700491543_dp, 0.806200001127913_dp, &
             2.567223570690613_dp, &
             ! w
             0.0842613451067412_dp, 0.3175083683168244_dp, 0.9827580985120555_dp, &
             2.8938728508266505_dp]

      CASE (54)

         aw(:) = & ! a
            [0.026607943720184272_dp, 0.1895065774790961_dp, 0.7515202985781508_dp, &
             2.469348057181892_dp, &
             ! w
             0.0731190495973734_dp, 0.29232694965731365_dp, 0.9440309351700997_dp, &
             2.8478366618531807_dp]

      CASE (55)

         aw(:) = & ! a
            [0.023541991865702758_dp, 0.17366047680464378_dp, 0.7126891672717475_dp, &
             2.399053964746976_dp, &
             ! w
             0.0653078213757229_dp, 0.2745449432568481_dp, 0.9160037860065624_dp, &
             2.8142148838640844_dp]

      CASE (56)

         aw(:) = & ! a
            [0.021260416555732336_dp, 0.16187242115691308_dp, 0.6835043049845141_dp, &
             2.3457307234433387_dp, &
             ! w
             0.059507528591388775_dp, 0.261239936416624_dp, 0.894605770150256_dp, &
             2.7883700207922746_dp]

      CASE (57)

         aw(:) = & ! a
            [0.019491643253607_dp, 0.1527342674760213_dp, 0.6606717166460769_dp, &
             2.3036894119803124_dp, &
             ! w
             0.055020249665865696_dp, 0.25086805101326054_dp, 0.877642602016386_dp, &
             2.7677729356352665_dp]

      CASE (58)

         aw(:) = & ! a
            [0.018077782916019164_dp, 0.14542839561841847_dp, 0.6422668197625268_dp, &
             2.2695770442654135_dp, &
             ! w
             0.051440413450525975_dp, 0.24253168056726374_dp, 0.8638140458215766_dp, &
             2.750910552833004_dp]

      CASE (59)

         aw(:) = & ! a
            [0.016920441215969285_dp, 0.13944586119810226_dp, 0.6270840649394218_dp, &
             2.241276902218253_dp, &
             ! w
             0.04851544830010834_dp, 0.23567104569970013_dp, 0.8522950664164703_dp, &
             2.736815432948413_dp]

      CASE (60)

         aw(:) = & ! a
            [0.011401253796233394_dp, 0.1108289024285984_dp, 0.5527015496667355_dp, &
             2.1002718748225484_dp, &
             ! w
             0.034652854429156214_dp, 0.20226202325315296_dp, 0.7941993308902318_dp, &
             2.6650476801253453_dp]

      CASE (61)

         aw(:) = & ! a
            [0.009566774796103916_dp, 0.10123798051872637_dp, 0.5269302725083039_dp, &
             2.050377006102403_dp, &
             ! w
             0.030084225422181535_dp, 0.1907535375929733_dp, 0.7733295903723639_dp, &
             2.6389901708906205_dp]

      CASE (62)

         aw(:) = & ! a
            [0.00891495987318472_dp, 0.09781045618420699_dp, 0.5175886051578179_dp, &
             2.032141433665363_dp, &
             ! w
             0.028466124477088454_dp, 0.1865878374088192_dp, 0.7656573116813944_dp, &
             2.6293741631869985_dp]

      CASE (63)

         aw(:) = & ! a
            [0.008868536798788046_dp, 0.09756583504881955_dp, 0.5169189990764869_dp, &
             2.0308311261282253_dp, &
             ! w
             0.028350982424774226_dp, 0.18628933863708444_dp, 0.7651050731282705_dp, &
             2.6286812529073846_dp]

      CASE (64)

         aw(:) = & ! a
            [0.18161223444789812_dp, 0.9776830415195706_dp, 2.5060472629612187_dp, &
             4.995923809435138_dp, 9.078298047219704_dp, &
             ! w
             0.4683556115061091_dp, 1.1387420740801897_dp, 1.95318188970844_dp, &
             3.1153038431409947_dp, 5.397941851023563_dp]

      CASE (65)

         aw(:) = & ! a
            [0.1427866528332531_dp, 0.7731653372376538_dp, 2.0043671769709315_dp, &
             4.06918816955637_dp, 7.599088437689025_dp, &
             ! w
             0.36871557779169384_dp, 0.9070179518196302_dp, 1.5911924263942965_dp, &
             2.6280030891313686_dp, 4.771668084689307_dp]

      CASE (66)

         aw(:) = & ! a
            [0.1193145236895995_dp, 0.6498239284000503_dp, 1.7039690954687403_dp, &
             3.5232332927125776_dp, 6.75309425511802_dp, &
             ! w
             0.3085046174714676_dp, 0.7677830644995077_dp, 1.377717871730129_dp, &
             2.3531329292269723_dp, 4.445303796640638_dp]

      CASE (67)

         aw(:) = & ! a
            [0.10333512117595546_dp, 0.5658998554798739_dp, 1.500259486444618_dp, &
             3.156388982390542_dp, 6.193584172208591_dp, &
             ! w
             0.2675144938257101_dp, 0.6731648185410648_dp, 1.2341495375414908_dp, &
             2.173074998760648_dp, 4.240049104595772_dp]

      CASE (68)

         aw(:) = & ! a
            [0.09164442849128951_dp, 0.5044896950537755_dp, 1.3514626526197429_dp, &
             2.8899298640154387_dp, 5.790782426391747_dp, &
             ! w
             0.23752137119510233_dp, 0.6039528783972736_dp, 1.1298096471277477_dp, &
             2.0442936920871437_dp, 4.096349191177988_dp]

      CASE (69)

         aw(:) = & ! a
            [0.08266403865575715_dp, 0.4572948614526955_dp, 1.2372195309898968_dp, &
             2.686051267949959_dp, 5.484100671224871_dp, &
             ! w
             0.2144769745022462_dp, 0.5507585162114406_dp, 1.04995382939453_dp, &
             1.9466601062734827_dp, 3.98859281695391_dp]

      CASE (70)

         aw(:) = & ! a
            [0.07551738726841437_dp, 0.41971519861212586_dp, 1.1462958376485461_dp, &
             2.524114942076429_dp, 5.241123104161658_dp, &
             ! w
             0.19613395743207157_dp, 0.5083904835637346_dp, 0.9865228049566487_dp, &
             1.8695023711995027_dp, 3.9038727784001863_dp]

      CASE (71)

         aw(:) = & ! a
            [0.06967518048938522_dp, 0.3889757291641015_dp, 1.0719353806517393_dp, &
             2.3918121851951577_dp, 5.042803535262377_dp, &
             ! w
             0.18113555675752338_dp, 0.4737218817910962_dp, 0.9347038834466117_dp, &
             1.8066027816920285_dp, 3.8349340375191976_dp]

      CASE (72)

         aw(:) = & ! a
            [0.06479731629497418_dp, 0.3632942397911341_dp, 1.009808499526931_dp, &
             2.2813077174805523_dp, 4.877153951942815_dp, &
             ! w
             0.16861003324210627_dp, 0.44474622263913965_dp, 0.8914300222454683_dp, &
             1.7540776690871602_dp, 3.777359367020908_dp]

      CASE (73)

         aw(:) = & ! a
            [0.03979103551863197_dp, 0.23131332253827694_dp, 0.6899245534043174_dp, &
             1.709843508205596_dp, 4.01492257681544_dp, &
             ! w
             0.10434198456045442_dp, 0.2955585665005043_dp, 0.6678313388646437_dp, &
             1.478881661645208_dp, 3.472870891680209_dp]

      CASE (74)

         aw(:) = & ! a
            [0.029757786779991018_dp, 0.17814645562861608_dp, 0.5601138852938733_dp, &
             1.4739028120126259_dp, 3.6519366545145373_dp, &
             ! w
             0.07852022967252373_dp, 0.2352291127246095_dp, 0.5755235459038333_dp, &
             1.36036132731287_dp, 3.3389033426820007_dp]

      CASE (75)

         aw(:) = & ! a
            [0.024190180350097283_dp, 0.14857337492633746_dp, 0.4872759750389151_dp, &
             1.339149099031898_dp, 3.4411575727997534_dp, &
             ! w
             0.06418148380506787_dp, 0.2015553177348804_dp, 0.5226910879403149_dp, &
             1.2901011954205592_dp, 3.258411518133752_dp]

      CASE (76)

         aw(:) = & ! a
            [0.02059851693641806_dp, 0.12946507144298772_dp, 0.439797239670509_dp, &
             1.2499372840249323_dp, 3.299772274458194_dp, &
             ! w
             0.054928968471960854_dp, 0.17972260846046623_dp, 0.48760926843571956_dp, &
             1.2421875649935035_dp, 3.2030368198904093_dp]

      CASE (77)

         aw(:) = & ! a
            [0.018067709913646333_dp, 0.11598435008616298_dp, 0.40601702895000163_dp, &
             1.1856093034676984_dp, 3.1967392715623495_dp, &
             ! w
             0.0484092394476768_dp, 0.16426671642208504_dp, 0.4622317543222329_dp, &
             1.2068045801673655_dp, 3.1618887677979015_dp]

      CASE (78)

         aw(:) = & ! a
            [0.016177447463648808_dp, 0.10590557466186795_dp, 0.38055858744901055_dp, &
             1.1365624686784197_dp, 3.1174919552876026_dp, &
             ! w
             0.04354046414237496_dp, 0.15267140778657548_dp, 0.4428218963958594_dp, &
             1.1792923812761709_dp, 3.129743372172042_dp]

      CASE (79)

         aw(:) = & ! a
            [0.014705949753103624_dp, 0.09805286816785048_dp, 0.3605733280656045_dp, &
             1.0976668403036145_dp, 3.0541811446566616_dp, &
             ! w
             0.03975143315365374_dp, 0.14360627651378796_dp, 0.42738350642551687_dp, &
             1.1571122334202288_dp, 3.1037318963686116_dp]

      CASE (80)

         aw(:) = & ! a
            [0.013524447064889847_dp, 0.09174282146012296_dp, 0.3444004900958718_dp, &
             1.0659073186356656_dp, 3.0021576019667893_dp, &
             ! w
             0.03671031159339888_dp, 0.1362975085252603_dp, 0.41474261404916296_dp, &
             1.1387451937085928_dp, 3.0821271502266367_dp]

      CASE (81)

         aw(:) = & ! a
            [0.01255272754046799_dp, 0.08654939487728926_dp, 0.33100124293713873_dp, &
             1.0393832416814095_dp, 2.9584699518773347_dp, &
             ! w
             0.03421030348746315_dp, 0.1302622960820673_dp, 0.404158415645879_dp, &
             1.1232181675451096_dp, 3.0638169728620763_dp]

      CASE (82)

         aw(:) = & ! a
            [0.0077919805414365445_dp, 0.061030287502729146_dp, 0.26354517613629047_dp, &
             0.9023059551184773_dp, 2.7287535886135674_dp, &
             ! w
             0.021992413199290763_dp, 0.10020642248192244_dp, 0.34896373518542456_dp, &
             1.0398862719837947_dp, 2.9648211490348504_dp]

      CASE (83)

         aw(:) = & ! a
            [0.006003383449447628_dp, 0.051381028213895265_dp, 0.2369999919944957_dp, &
             0.846323438825162_dp, 2.6327559231103215_dp, &
             ! w
             0.01742823123743174_dp, 0.08855164468272898_dp, 0.32612062131196123_dp, &
             1.004141456348347_dp, 2.9219747649465218_dp]

      CASE (84)

         aw(:) = & ! a
            [0.005055044709474687_dp, 0.04623422613950649_dp, 0.2224845241613079_dp, &
             0.8150922595119978_dp, 2.578567083457729_dp, &
             ! w
             0.01501876146021437_dp, 0.08222284510376696_dp, 0.31328309235675417_dp, &
             0.9837032818773332_dp, 2.897369577036015_dp]

      CASE (85)

         aw(:) = & ! a
            [0.004470091020653244_dp, 0.043042233518954816_dp, 0.21332569755781333_dp, &
             0.7951322401352172_dp, 2.543679303681629_dp, &
             ! w
             0.013537186918918044_dp, 0.07824517136303201_dp, 0.30503984426641567_dp, &
             0.9704412982176837_dp, 2.88136144708144_dp]

      CASE (86)

         aw(:) = & ! a
            [0.004079351161198711_dp, 0.04089971785147186_dp, 0.20710038537556108_dp, &
             0.7814442554474885_dp, 2.5196339666338297_dp, &
             ! w
             0.012549702236901892_dp, 0.07554802779982046_dp, 0.29936846535124767_dp, &
             0.9612526221486811_dp, 2.870250216273631_dp]

      CASE (87)

         aw(:) = & ! a
            [0.0038074691411251675_dp, 0.039402728467349026_dp, 0.202709695765557_dp, &
             0.7717281393047815_dp, 2.5025046733436875_dp, &
             ! w
             0.011863663762419518_dp, 0.07364871574594964_dp, 0.29533335962239976_dp, &
             0.9546824601007418_dp, 2.8622952932595034_dp]

      CASE (88)

         aw(:) = & ! a
            [0.0036161175535254345_dp, 0.038345504980682704_dp, 0.1995869766812679_dp, &
             0.7647855301606952_dp, 2.4902332388884396_dp, &
             ! w
             0.011381349717183978_dp, 0.07229930785193518_dp, 0.2924451938945003_dp, &
             0.949963002876562_dp, 2.856575920404516_dp]

      CASE (89)

         aw(:) = & ! a
            [0.0034843237838238315_dp, 0.0376153534981393_dp, 0.19741919379317616_dp, &
             0.7599497569265251_dp, 2.481669850578368_dp, &
             ! w
             0.011049399698694484_dp, 0.07136320544570081_dp, 0.2904310341712172_dp, &
             0.9466633544503644_dp, 2.8525745409845267_dp]

      CASE (90)

         aw(:) = & ! a
            [0.0034004600489073895_dp, 0.037149816796615_dp, 0.19603210903786736_dp, &
             0.7568483876518314_dp, 2.476170832968167_dp, &
             ! w
             0.01083827265702328_dp, 0.07076449780867321_dp, 0.2891382091047488_dp, &
             0.9445417262916082_dp, 2.8500005633404064_dp]

      CASE (91)

         aw(:) = & ! a
            [0.0033529112764591387_dp, 0.03688552380863027_dp, 0.1952428895765193_dp, &
             0.7550812995393414_dp, 2.4730352459564653_dp, &
             ! w
             0.010718601062497496_dp, 0.07042393570386862_dp, 0.28840120533084923_dp, &
             0.9433310104981166_dp, 2.8485313369244247_dp]

      CASE (92)

         aw(:) = & ! a
            [0.15344232732696164_dp, 0.8208030144142543_dp, 2.076326637759574_dp, &
             4.037544944825249_dp, 6.9549672561485965_dp, 11.489101449975879_dp, &
             ! w
             0.3951503586666183_dp, 0.9483584803953559_dp, 1.5813287669193052_dp, &
             2.3797403876283134_dp, 3.548133935573261_dp, 5.8817260741536765_dp]

      CASE (93)

         aw(:) = & ! a
            [0.12050409838853515_dp, 0.6471706374763834_dp, 1.649502383645565_dp, &
             3.2457669505020714_dp, 5.688110896918401_dp, 9.628826117919456_dp, &
             ! w
             0.31060826041851003_dp, 0.7514093561343357_dp, 1.272007612541384_dp, &
             1.9595162535981736_dp, 3.01771994304632_dp, 5.212640690904524_dp]

      CASE (94)

         aw(:) = & ! a
            [0.10058381094183846_dp, 0.5423052006448188_dp, 1.39268940689611_dp, &
             2.773384987559996_dp, 4.944377020913284_dp, 8.56530729992123_dp, &
             ! w
             0.259492533708378_dp, 0.6326994281541639_dp, 1.0873449642129085_dp, &
             1.714381139255248_dp, 2.721852627098835_dp, 4.865160915450924_dp]

      CASE (95)

         aw(:) = & ! a
            [0.08702163406475558_dp, 0.47091367520479227_dp, 1.2180676458000685_dp, &
             2.4535517473926842_dp, 4.445294940570353_dp, 7.861429581448342_dp, &
             ! w
             0.22469037674868492_dp, 0.5519095844142892_dp, 0.9622019394266136_dp, &
             1.5504627855961601_dp, 2.5290208569930264_dp, 4.646614158574707_dp]

      CASE (96)

         aw(:) = & ! a
            [0.07709989750137178_dp, 0.41866258611030904_dp, 1.0902916798550335_dp, &
             2.220067742808422_dp, 4.082849994717455_dp, 7.3540700896552895_dp, &
             ! w
             0.19922622623379935_dp, 0.49276258797330597_dp, 0.8707649507441914_dp, &
             1.431682309041832_dp, 2.391339554164293_dp, 4.493362869549154_dp]

      CASE (97)

         aw(:) = & ! a
            [0.06947912297479582_dp, 0.3785036647871915_dp, 0.9920629035718725_dp, &
             2.0408032638639466_dp, 3.8053977452053824_dp, 6.967212719012414_dp, &
             ! w
             0.17966376455572852_dp, 0.447278321906676_dp, 0.8005082241463485_dp, &
             1.3408813762157024_dp, 2.2869350476400565_dp, 4.378184574358267_dp]

      CASE (98)

         aw(:) = & ! a
            [0.06341523419030931_dp, 0.34652627576455186_dp, 0.9138093905878505_dp, &
             1.8980781549742471_dp, 3.5848315846970733_dp, 6.6602309134426045_dp, &
             ! w
             0.16409461040061746_dp, 0.41103624081412404_dp, 0.7445371549477633_dp, &
             1.268755832367187_dp, 2.2043138776461757_dp, 4.287402567458178_dp]

      CASE (99)

         aw(:) = & ! a
            [0.0584588699633234_dp, 0.3203701144935931_dp, 0.849763287841353_dp, &
             1.7812813101063347_dp, 3.40442656718486_dp, 6.409267895239723_dp, &
             ! w
             0.1513663895024808_dp, 0.38137058330649815_dp, 0.6987115135433852_dp, &
             1.2097848806132692_dp, 2.1368239386439773_dp, 4.213342398454894_dp]

      CASE (100)

         aw(:) = & ! a
            [0.05432125985000594_dp, 0.2985187647358117_dp, 0.7962225930216239_dp, &
             1.683622952477423_dp, 3.2535534765211933_dp, 6.199310760184936_dp, &
             ! w
             0.1407385838051434_dp, 0.356569420068569_dp, 0.6603802616738013_dp, &
             1.1604682915072342_dp, 2.0803301182531366_dp, 4.1513349861478135_dp]

      CASE (101)

         aw(:) = & ! a
            [0.033125384072048854_dp, 0.18624531899278388_dp, 0.5202262124793207_dp, &
             1.1783399761432671_dp, 2.468203283261333_dp, 5.09898230836826_dp, &
             ! w
             0.08624967411800354_dp, 0.228755231383876_dp, 0.4619231653435139_dp, &
             0.9029137915898263_dp, 1.7807264424526321_dp, 3.8200570280772443_dp]

      CASE (102)

         aw(:) = & ! a
            [0.02463386018680619_dp, 0.14102557634568977_dp, 0.4082545329312478_dp, &
             0.970749927523341_dp, 2.1393343031375327_dp, 4.629695544074457_dp, &
             ! w
             0.06438753922607744_dp, 0.17699255332037328_dp, 0.3803989314000022_dp, &
             0.7936758834219183_dp, 1.6486087888192316_dp, 3.671677753256644_dp]

      CASE (103)

         aw(:) = & ! a
            [0.019927292158899282_dp, 0.11586876096539546_dp, 0.345547325025503_dp, &
             0.8529186964399603_dp, 1.949300779528476_dp, 4.35439826891279_dp, &
             ! w
             0.05225747423429781_dp, 0.14807969939302706_dp, 0.3341395292959525_dp, &
             0.7296921875931182_dp, 1.5689274489411984_dp, 3.5813517206284184_dp]

      CASE (104)

         aw(:) = & ! a
            [0.016893837552218283_dp, 0.09960917776823842_dp, 0.3047652371799654_dp, &
             0.7753183769485388_dp, 1.8222505520282237_dp, 4.1681458423137245_dp, &
             ! w
             0.044433418238675776_dp, 0.12933083480399385_dp, 0.30367117547648526_dp, &
             0.6863985328086_dp, 1.513850308723576_dp, 3.5185424210031164_dp]

      CASE (105)

         aw(:) = & ! a
            [0.014757794449453762_dp, 0.08813396805017469_dp, 0.2758135968716969_dp, &
             0.7195968816061252_dp, 1.7298527418488276_dp, 4.031389035577503_dp, &
             ! w
             0.03892075392146144_dp, 0.11606060550997924_dp, 0.2817827958041293_dp, &
             0.6545864705958214_dp, 1.4727168297337245_dp, 3.4714354715327134_dp]

      CASE (106)

         aw(:) = & ! a
            [0.013163067275079546_dp, 0.0795507056312876_dp, 0.25403700481526015_dp, &
             0.677249443803188_dp, 1.6588625672007167_dp, 3.925479637456699_dp, &
             ! w
             0.034803217566193485_dp, 0.10610858572603164_dp, 0.2651359722976122_dp, &
             0.6299268945664218_dp, 1.4404181621079915_dp, 3.434327679345151_dp]

      CASE (107)

         aw(:) = & ! a
            [0.011921868297062481_dp, 0.07285946900570793_dp, 0.23697015769013335_dp, &
             0.6437479496493271_dp, 1.6021676951562196_dp, 3.84032618251257_dp, &
             ! w
             0.031597334451890974_dp, 0.09833125045975417_dp, 0.25195493890390575_dp, &
             0.6100807113831754_dp, 1.4141485422756008_dp, 3.404069746353219_dp]

      CASE (108)

         aw(:) = & ! a
            [0.010925238490803518_dp, 0.0674791066323321_dp, 0.22317706781392338_dp, &
             0.6164401736272703_dp, 1.55556859114024_dp, 3.7699289128884463_dp, &
             ! w
             0.02902243561550214_dp, 0.09206299625144375_dp, 0.24120038626341075_dp, &
             0.5936581793011397_dp, 1.392217727402881_dp, 3.378756504485369_dp]

      CASE (109)

         aw(:) = & ! a
            [0.010105361344629787_dp, 0.06304737930382137_dp, 0.21176070087633833_dp, &
             0.5936601207247773_dp, 1.51640746079713_dp, 3.7104671222364463_dp, &
             ! w
             0.026903750347277787_dp, 0.08688840346469896_dp, 0.23221984689982497_dp, &
             0.5797744666263794_dp, 1.3735366550622796_dp, 3.357156154496523_dp]

      CASE (110)

         aw(:) = & ! a
            [0.006074556160583506_dp, 0.04116568929465924_dp, 0.15431499997262116_dp, &
             0.47583402653062223_dp, 1.3088413484383163_dp, 3.3901114378459396_dp, &
             ! w
             0.016484631406817916_dp, 0.06110676281193284_dp, 0.18555599341475487_dp, &
             0.5047457845186151_dp, 1.2702004011589922_dp, 3.237028368952973_dp]

      CASE (111)

         aw(:) = & ! a
            [0.0045403772662585915_dp, 0.032777244620163834_dp, 0.1315230758019218_dp, &
             0.4270458936595013_dp, 1.2198581690530463_dp, 3.2496928841682373_dp, &
             ! w
             0.012522087399164902_dp, 0.051044965137218956_dp, 0.16606657745371223_dp, &
             0.4717249694785839_dp, 1.2233401903869243_dp, 3.182181287962263_dp]

      CASE (112)

         aw(:) = & ! a
            [0.0037126126629825553_dp, 0.028227596539976623_dp, 0.11885555456344414_dp, &
             0.3992215234884477_dp, 1.168114174117086_dp, 3.1670518413818134_dp, &
             ! w
             0.010387584330417182_dp, 0.045509430030571195_dp, 0.15488496346531724_dp, &
             0.4522533544737275_dp, 1.1952746785512347_dp, 3.149214520189165_dp]

      CASE (113)

         aw(:) = & ! a
            [0.0031893814523236115_dp, 0.025338999611455863_dp, 0.11066043250792441_dp, &
             0.3808951402331244_dp, 1.1335878719950365_dp, 3.111472453208487_dp, &
             ! w
             0.009040751564373343_dp, 0.0419530265693815_dp, 0.14748815509853327_dp, &
             0.43914352406806945_dp, 1.1761879348950757_dp, 3.126742024435687_dp]

      CASE (114)

         aw(:) = & ! a
            [0.0028268458406437787_dp, 0.023329602254469856_dp, 0.10487319183563992_dp, &
             0.3677786713597671_dp, 1.108641964111255_dp, 3.0710862053749617_dp, &
             ! w
             0.00810913855401357_dp, 0.03945410196052096_dp, 0.1421765733804508_dp, &
             0.42961079747665915_dp, 1.1622091695018335_dp, 3.1102558315953166_dp]

      CASE (115)

         aw(:) = & ! a
            [0.0025601001960133666_dp, 0.021845744822363813_dp, 0.10054616021655595_dp, &
             0.35786818283376687_dp, 1.0896558648067596_dp, 3.040214954555503_dp, &
             ! w
             0.007424732308214915_dp, 0.037592711108251284_dp, 0.1381527530049599_dp, &
             0.42232059870164745_dp, 1.151460470160175_dp, 3.097562766245219_dp]

      CASE (116)

         aw(:) = & ! a
            [0.0023553678339827302_dp, 0.02070300562775958_dp, 0.0971788429385239_dp, &
             0.3500900854010226_dp, 1.0746683097025738_dp, 3.0157614053961606_dp, &
             ! w
             0.006900154654307798_dp, 0.03614838297117291_dp, 0.13498801054740742_dp, &
             0.41654403926734435_dp, 1.1429068035001628_dp, 3.087451474766092_dp]

      CASE (117)

         aw(:) = & ! a
            [0.0021932677980895975_dp, 0.01979537163607368_dp, 0.09448030910814234_dp, &
             0.3438129085859722_dp, 1.062515385363092_dp, 2.9958772123567106_dp, &
             ! w
             0.006485309446126698_dp, 0.034993586030055_dp, 0.13242947055069354_dp, &
             0.4118457658248257_dp, 1.135925466915433_dp, 3.0791919318758856_dp]

      CASE (118)

         aw(:) = & ! a
            [0.0020618552752413614_dp, 0.019057395573237303_dp, 0.09226917084160827_dp, &
             0.3386390121450569_dp, 1.0524587572217026_dp, 2.979384587274735_dp, &
             ! w
             0.0061493505796972655_dp, 0.03404912883205661_dp, 0.13031750436562978_dp, &
             0.4079481557970085_dp, 1.1301170634334539_dp, 3.0723153218216233_dp]

      CASE (119)

         aw(:) = & ! a
            [0.0014764458510144259_dp, 0.015736804651964958_dp, 0.08209752558374095_dp, &
             0.31445243244806775_dp, 1.0049446929502894_dp, 2.90097660575702_dp, &
             ! w
             0.0046569405523143575_dp, 0.029724286090695502_dp, 0.12040494837441577_dp, &
             0.389411597615533_dp, 1.1022784500339817_dp, 3.0392951653690305_dp]

      CASE (120)

         aw(:) = & ! a
            [0.00138020194524145_dp, 0.015183614381572768_dp, 0.08036151185525324_dp, &
             0.310255608632774_dp, 0.9966114862688712_dp, 2.887139653946674_dp, &
             ! w
             0.004412235352657983_dp, 0.028989220321940696_dp, 0.11867791139299246_dp, &
             0.38613951254853507_dp, 1.0973265611814316_dp, 3.033410611192584_dp]

      CASE (121)

         aw(:) = & ! a
            [0.1328537955998788_dp, 0.7079130219116617_dp, 1.7769359317465696_dp, &
             3.4093982762602817_dp, 5.73641479371703_dp, 9.022775927136047_dp, &
             13.953318359742749_dp, &
             ! w
             0.34183027419014206_dp, 0.8139373975876502_dp, 1.3352973461439932_dp, &
             1.9505008993945934_dp, 2.744440392392776_dp, 3.9247848246178934_dp, &
             6.309956840322047_dp]

      CASE (122)

         aw(:) = & ! a
            [0.1042559111992259_dp, 0.5571299562861916_dp, 1.4060269600791764_dp, &
             2.720266979333603_dp, 4.631373848989341_dp, 7.4035412706077315_dp, &
             11.706208031960966_dp, &
             ! w
             0.2684261108739403_dp, 0.6428494301097513_dp, 1.0661026045730193_dp, &
             1.5833307838749986_dp, 2.279828399460775_dp, 3.3594350485176023_dp, &
             5.603774368653575_dp]

      CASE (123)

         aw(:) = & ! a
            [0.08695630513426755_dp, 0.46599558018018294_dp, 1.1823447473116389_dp, &
             2.3066801208193026_dp, 3.9742794392437286_dp, 6.455652812958656_dp, &
             10.422112499317336_dp, &
             ! w
             0.22402959509378434_dp, 0.5395688431206702_dp, 0.9044919147174741_dp, &
             1.3657443364559092_dp, 2.01152465270939_dp, 3.0469664839439665_dp, &
             5.238102296673424_dp]

      CASE (124)

         aw(:) = & ! a
            [0.07517801205393558_dp, 0.40394057377108084_dp, 1.0300953839123785_dp, &
             2.0257208844822334_dp, 3.5300416178849785_dp, 5.8202600945046035_dp, &
             9.571879302298791_dp, &
             ! w
             0.19380100796794852_dp, 0.4692435586300529_dp, 0.7946376596686956_dp, &
             1.2188241997775222_dp, 1.8330663504453952_dp, 2.844141721918306_dp, &
             5.0081428000286365_dp]

      CASE (125)

         aw(:) = & ! a
            [0.06656176398587473_dp, 0.35852406376155277_dp, 0.9186334445646663_dp, &
             1.820181207842435_dp, 3.205920829057035_dp, 5.35889789623272_dp, &
             8.958513851633413_dp, &
             ! w
             0.17168478698145412_dp, 0.4177512782833647_dp, 0.7142178763847493_dp, &
             1.1116696392634222_dp, 1.7040902635819437_dp, 2.6995153646546877_dp, &
             4.846714032397592_dp]

      CASE (126)

         aw(:) = & ! a
            [0.05994431184969806_dp, 0.3236222394928578_dp, 0.8329240783297642_dp, &
             1.6621487568152902_dp, 2.9570712920382243_dp, 5.005596919496737_dp, &
             8.490350769144383_dp, &
             ! w
             0.15469632923142762_dp, 0.3781553650192849_dp, 0.6523457119440436_dp, &
             1.0293972455516656_dp, 1.605588731877836_dp, 2.5898198883040795_dp, &
             4.725197806343358_dp]

      CASE (127)

         aw(:) = & ! a
            [0.05467935470040509_dp, 0.2958355411939927_dp, 0.7646359311879659_dp, &
             1.536208458069161_dp, 2.7588796360606658_dp, 4.724545136307037_dp, &
             8.118442969694092_dp, &
             ! w
             0.14117768906298914_dp, 0.3466092767684028_dp, 0.603008829024842_dp, &
             0.9638568128475085_dp, 1.5273335779508246_dp, 2.502917653016165_dp, &
             4.629251753453269_dp]

      CASE (128)

         aw(:) = & ! a
            [0.05037653168752645_dp, 0.27311145827271915_dp, 0.7087438967352819_dp, &
             1.4330831595065083_dp, 2.596601846366239_dp, 4.494470571023396_dp, &
             7.814060433661144_dp, &
             ! w
             0.1301275574325502_dp, 0.3207919173372999_dp, 0.5625882013965461_dp, &
             0.910172757571702_dp, 1.4632925602129687_dp, 2.431819911650354_dp, &
             4.550838356843569_dp]

      CASE (129)

         aw(:) = & ! a
            [0.04678497191969506_dp, 0.2541310917703071_dp, 0.6620206366128168_dp, &
             1.3468236945914096_dp, 2.4608175652259985_dp, 4.301871868914413_dp, &
             7.559127174440305_dp, &
             ! w
             0.12090244093952099_dp, 0.2992121471107249_dp, 0.5287623608132956_dp, &
             0.8652325406126139_dp, 1.4096599453909606_dp, 2.372198230000475_dp, &
             4.4850705608595405_dp]

      CASE (130)

         aw(:) = & ! a
            [0.02839787501935781_dp, 0.15669465451979245_dp, 0.4212835366409848_dp, &
             0.9007103394850067_dp, 1.7551920884002583_dp, 3.2943879648814645_dp, &
             6.216752116933116_dp, &
             ! w
             0.07364094379374109_dp, 0.1880960775887743_dp, 0.3535856122925518_dp, &
             0.6310263698265569_dp, 1.1268723268357688_dp, 2.0532076958089367_dp, &
             4.131214614090865_dp]

      CASE (131)

         aw(:) = & ! a
            [0.02104159749657532_dp, 0.11751591067464534_dp, 0.3237851994719232_dp, &
             0.7182511074334368_dp, 1.4622083631638707_dp, 2.868129060797868_dp, &
             5.639128283714038_dp, &
             ! w
             0.05470812366039994_dp, 0.14316733618034405_dp, 0.2818519061350474_dp, &
             0.5330147207850684_dp, 1.0042678371684879_dp, 1.9102105077080431_dp, &
             3.97078392780554_dp]

      CASE (132)

         aw(:) = & ! a
            [0.01696894434764736_dp, 0.09574540214390341_dp, 0.2692943329972381_dp, &
             0.6152724295093569_dp, 1.2943422432026803_dp, 2.6197801492130472_dp, &
             5.297964533423825_dp, &
             ! w
             0.04421634439751523_dp, 0.11809953230597657_dp, 0.24136520239356096_dp, &
             0.47639520405533553_dp, 0.9312065231173353_dp, 1.8229592256705782_dp, &
             3.8722630501939475_dp]

      CASE (133)

         aw(:) = & ! a
            [0.01434661695411904_dp, 0.08168663696512406_dp, 0.23392918256659598_dp, &
             0.547817369616115_dp, 1.18289671872084_dp, 2.452604838409686_dp, &
             5.065840953958105_dp, &
             ! w
             0.037455686176677_dp, 0.10185840534484672_dp, 0.21485110258922235_dp, &
             0.43851262496240967_dp, 0.8811012601672848_dp, 1.7621025260651733_dp, &
             3.8032676122312763_dp]

      CASE (134)

         aw(:) = & ! a
            [0.012501624790553638_dp, 0.07177123210528669_dp, 0.20887384143197713_dp, &
             0.4996135480465156_dp, 1.1023091213725187_dp, 2.3303116968910063_dp, &
             4.894562576543979_dp, &
             ! w
             0.03269607742097257_dp, 0.09037208740586535_dp, 0.19590828145261074_dp, &
             0.4109239117501715_dp, 0.8438800600320582_dp, 1.7163114229596177_dp, &
             3.751205173302536_dp]

      CASE (135)

         aw(:) = & ! a
            [0.011125203816029645_dp, 0.06435860149721041_dp, 0.1900644749502538_dp, &
             0.46313529471763043_dp, 1.0406826317548055_dp, 2.2358669464308134_dp, &
             4.761334307488529_dp, &
             ! w
             0.029143346494564064_dp, 0.08176423354275007_dp, 0.18157540048038484_dp, &
             0.3896901972802879_dp, 0.8147620352906384_dp, 1.6801229596135674_dp, &
             3.7099722173278535_dp]

      CASE (136)

         aw(:) = & ! a
            [0.010054615399715751_dp, 0.05858247648365217_dp, 0.17535029502975125_dp, &
             0.43438561313685364_dp, 0.991657281318819_dp, 2.1600909301499334_dp, &
             4.65378692704619_dp, &
             ! w
             0.02637871732005645_dp, 0.07504215090863822_dp, 0.17027978818667666_dp, &
             0.37269963545122914_dp, 0.7911419754631662_dp, 1.6505212211957283_dp, &
             3.6761865418821604_dp]

      CASE (137)

         aw(:) = & ! a
            [0.009195480405483587_dp, 0.05393966047189065_dp, 0.16347931911580282_dp, &
             0.4110293908202165_dp, 0.9514935033869456_dp, 2.097545864519508_dp, &
             4.564548258392167_dp, &
             ! w
             0.024159221705021952_dp, 0.06962821946939234_dp, 0.16110291618092115_dp, &
             0.35870654013259184_dp, 0.7714604530426776_dp, 1.6256811668432707_dp, &
             3.647795593113763_dp]

      CASE (138)

         aw(:) = & ! a
            [0.008489082134020809_dp, 0.05011664201496171_dp, 0.15366998285929265_dp, &
             0.3916035531845051_dp, 0.9178336466850517_dp, 2.0447798621093103_dp, &
             4.488912743216377_dp, &
             ! w
             0.022333642551176373_dp, 0.06516201073040385_dp, 0.15346948447776387_dp, &
             0.3469229620284128_dp, 0.754717937038961_dp, 1.604422129926041_dp, &
             3.6234684847661183_dp]

      CASE (139)

         aw(:) = & ! a
            [0.005021341168426651_dp, 0.03125464109942908_dp, 0.10459702700841456_dp, &
             0.2920522758702768_dp, 0.7407504784499062_dp, 1.7609744335543205_dp, &
             4.075903696914512_dp, &
             ! w
             0.013361554718382558_dp, 0.04297284694243602_dp, 0.11430293990815156_dp, &
             0.28388812669341895_dp, 0.6622322841999484_dp, 1.4847175320092705_dp, &
             3.4859753729916254_dp]

      CASE (140)

         aw(:) = & ! a
            [0.003702129096267596_dp, 0.02402035633785203_dp, 0.08527629374441756_dp, &
             0.25124270854584935_dp, 0.6652246780099562_dp, 1.6360585970149728_dp, &
             3.8902964899531036_dp, &
             ! w
             0.009943477668559816_dp, 0.03435111065922189_dp, 0.09818698596300013_dp, &
             0.2563375496711153_dp, 0.6200522686047906_dp, 1.4287348681193341_dp, &
             3.4213578343787923_dp]

      CASE (141)

         aw(:) = & ! a
            [0.0029885877621226805_dp, 0.020086057792296826_dp, 0.07455690900558905_dp, &
             0.2279899926003867_dp, 0.6211585820580099_dp, 1.5618502570945418_dp, &
             3.778745593374526_dp, &
             ! w
             0.008093917805684206_dp, 0.02961377120275156_dp, 0.08897105543150366_dp, &
             0.24003462235871895_dp, 0.5945144839955654_dp, 1.3943733291175233_dp, &
             3.3815881302823882_dp]

      CASE (142)

         aw(:) = & ! a
            [0.002535797864365192_dp, 0.017578759091398184_dp, 0.06761316861077361_dp, &
             0.21262834171859676_dp, 0.5915602300563042_dp, 1.5113889410817196_dp, &
             3.7022956650360443_dp, &
             ! w
             0.006920304813525343_dp, 0.026567898179860355_dp, 0.08286374742528071_dp, &
             0.22897941405638794_dp, 0.576932442449835_dp, 1.370498834102023_dp, &
             3.3539045747733187_dp]

      CASE (143)

         aw(:) = & ! a
            [0.002220603088272845_dp, 0.015827051545107267_dp, 0.06269465031628192_dp, &
             0.20157811436912737_dp, 0.5700007444812679_dp, 1.4742950212653216_dp, &
             3.645772538526923_dp, &
             ! w
             0.006103579691650011_dp, 0.0244232192416007_dp, 0.07845898683705749_dp, &
             0.22087024192514393_dp, 0.5638921330027344_dp, 1.352672001922338_dp, &
             3.3332048728001094_dp]

      CASE (144)

         aw(:) = & ! a
            [0.001987459760689592_dp, 0.014527207706942338_dp, 0.05900120240242023_dp, &
             0.193175465285863_dp, 0.5534433535878025_dp, 1.4456024138634962_dp, &
             3.601854362338122_dp, &
             ! w
             0.005499733478671553_dp, 0.02282055828311453_dp, 0.07510209589065722_dp, &
             0.21460869698738236_dp, 0.5537362828757469_dp, 1.338715519022809_dp, &
             3.3169815640530342_dp]

      CASE (145)

         aw(:) = & ! a
            [0.0018074376308293873_dp, 0.013520634224809635_dp, 0.056111063613119946_dp, &
             0.1865310578633184_dp, 0.5402437912418699_dp, 1.4225951349686723_dp, &
             3.5665103665179134_dp, &
             ! w
             0.005033717325569482_dp, 0.02157155721867786_dp, 0.0724425268654194_dp, &
             0.20959525364169268_dp, 0.5455484720427803_dp, 1.3274159182138672_dp, &
             3.303834943091577_dp]

      CASE (146)

         aw(:) = & ! a
            [0.001663906158190309_dp, 0.012715963868680322_dp, 0.05377913143326573_dp, &
             0.1811218120543097_dp, 0.5294244944099309_dp, 1.4036450526560744_dp, &
             3.5373112546752496_dp, &
             ! w
             0.0046623733840567665_dp, 0.020567251935995977_dp, 0.07027373966039487_dp, &
             0.2054711368999697_dp, 0.5387744815015728_dp, 1.3180345908931348_dp, &
             3.2929120497385744_dp]

      CASE (147)

         aw(:) = & ! a
            [0.0015465907100920744_dp, 0.012056643172650844_dp, 0.05185248993325443_dp, &
             0.17661795467136937_dp, 0.5203635254756013_dp, 1.3877091798262136_dp, &
             3.5126938632453744_dp, &
             ! w
             0.0043590321081121135_dp, 0.019739937643250896_dp, 0.06846531409283038_dp, &
             0.20200683741399486_dp, 0.5330567089683809_dp, 1.3100924544159886_dp, &
             3.2836589740207804_dp]

      CASE (148)

         aw(:) = & ! a
            [0.0009839513388777788_dp, 0.008865469103519473_dp, 0.04226812138861273_dp, &
             0.15367245013063802_dp, 0.4733864179916387_dp, 1.304067249716343_dp, &
             3.382502968419956_dp, &
             ! w
             0.0029076880267603932_dp, 0.015659795689673513_dp, 0.05920955020430181_dp, &
             0.1838875600867601_dp, 0.5027200930136522_dp, 1.2675787498436362_dp, &
             3.2340325901955294_dp]

      CASE (149)

         aw(:) = & ! a
            [0.0007821171870305443_dp, 0.007701702998949635_dp, 0.03862942520670454_dp, &
             0.14468193071323893_dp, 0.4545657667811826_dp, 1.2700397189863826_dp, &
             3.329040949966218_dp, &
             ! w
             0.002389135991366753_dp, 0.014127070818368568_dp, 0.0555607731872177_dp, &
             0.1765507108276268_dp, 0.49021766809565676_dp, 1.2498655911588725_dp, &
             3.213306254204669_dp]

      CASE (150)

         aw(:) = & ! a
            [0.0006828418611695833_dp, 0.0071228279773048585_dp, 0.03678031908136637_dp, &
             0.14004272890256222_dp, 0.444751966020422_dp, 1.2521695310166052_dp, &
             3.30084297204394_dp, &
             ! w
             0.0021345887454745104_dp, 0.01335161150081422_dp, 0.05367243060786458_dp, &
             0.17270668365493763_dp, 0.48361359837566975_dp, 1.2404615898649594_dp, &
             3.202290193666039_dp]

      CASE (151)

         aw(:) = & ! a
            [0.0006308136329875133_dp, 0.006817140620478282_dp, 0.03579160751106932_dp, &
             0.13754098637797435_dp, 0.4394293305525137_dp, 1.2424396843361853_dp, &
             3.2854539018961137_dp, &
             ! w
             0.002001311088035267_dp, 0.012937890869550871_dp, 0.05265247259283306_dp, &
             0.1706164960435683_dp, 0.48000667433517574_dp, 1.2353112694406894_dp, &
             3.196253285166892_dp]

      CASE (152)

         aw(:) = & ! a
            [0.0006085749350024497_dp, 0.006685902159141663_dp, 0.03536430717421001_dp, &
             0.1364550050371723_dp, 0.4371119961322448_dp, 1.2381951172423498_dp, &
             3.278732451656292_dp, &
             ! w
             0.0019443655262837806_dp, 0.012759280727803383_dp, 0.05220935096523234_dp, &
             0.16970529863934425_dp, 0.4784306996286164_dp, 1.2330577486624195_dp, &
             3.193610952471498_dp]

      CASE (153)

         aw(:) = & ! a
            [0.0006068004605425494_dp, 0.006675414550308602_dp, 0.03533008531165831_dp, &
             0.1363679064445001_dp, 0.43692597185507065_dp, 1.2378541669818244_dp, &
             3.278192298140961_dp, &
             ! w
             0.001939822244031596_dp, 0.012744981074293062_dp, 0.052173800858709714_dp, &
             0.16963212116290538_dp, 0.47830405076036075_dp, 1.2328765453712285_dp, &
             3.193398415618935_dp]

      CASE (154)

         aw(:) = & ! a
            [0.041096167724934025_dp, 0.22159210039399246_dp, 0.5688643520073686_dp, &
             1.1299764765876945_dp, 1.9937655342879093_dp, 3.3208105201716793_dp, &
             5.409823121081442_dp, 8.948554779106718_dp, &
             ! w
             0.10602646910567383_dp, 0.25852905669687154_dp, 0.4435891050076718_dp, &
             0.6929754149010114_dp, 1.0604512558549937_dp, 1.6406263617959378_dp, &
             2.636487509406833_dp, 4.787520763764322_dp]

      CASE (155)

         aw(:) = & ! a
            [0.02486266414966956_dp, 0.13560482860785025_dp, 0.3561108775384803_dp, &
             0.7334936098496532_dp, 1.3603355122458833_dp, 2.4047037240593268_dp, &
             4.17436963925409_dp, 7.36122931342208_dp, &
             ! w
             0.06430916462238742_dp, 0.160465799927234_dp, 0.28808044473936617_dp, &
             0.48165041896975763_dp, 0.7991696608085076_dp, 1.3381266056699614_dp, &
             2.301417441287326_dp, 4.41406270779838_dp]

      CASE (156)

         aw(:) = & ! a
            [0.01837515238630235_dp, 0.10109103962531879_dp, 0.2701415651193927_dp, &
             0.5719241549530382_dp, 1.0991792961155158_dp, 2.020941274862189_dp, &
             3.647536911785666_dp, 6.673761667175686_dp, &
             ! w
             0.047619383133358346_dp, 0.1209066723887145_dp, 0.2245974893577788_dp, &
             0.39396856430486504_dp, 0.6878199173590028_dp, 1.2046845144004612_dp, &
             2.149404405565857_dp, 4.243245954411484_dp]

      CASE (157)

         aw(:) = & ! a
            [0.01478688449158526_dp, 0.08193908775465297_dp, 0.22218948449558504_dp, &
             0.4811176249313396_dp, 0.9506792305778743_dp, 1.7993973054918035_dp, &
             3.3386937535433483_dp, 6.265728126848089_dp, &
             ! w
             0.0383807437367189_dp, 0.09887277234911501_dp, 0.18889510206656607_dp, &
             0.34383252442421686_dp, 0.6224537875740375_dp, 1.1241351163372293_dp, &
             2.0558777824169634_dp, 4.137687467758895_dp]

      CASE (158)

         aw(:) = & ! a
            [0.012478362825083333_dp, 0.06958512369187221_dp, 0.1911241780349973_dp, &
             0.42187946175263485_dp, 0.8527570534377368_dp, 1.651381610017021_dp, &
             3.129749265706254_dp, 5.986990637663619_dp, &
             ! w
             0.03243314247535478_dp, 0.08461681524598316_dp, 0.16560014581798357_dp, &
             0.3106025687325339_dp, 0.57813757860161_dp, 1.0683503216704084_dp, &
             1.990224850235617_dp, 4.0633908158171135_dp]

      CASE (159)

         aw(:) = & ! a
            [0.010855351097420482_dp, 0.060880248851111846_dp, 0.1691520259232484_dp, &
             0.37971034141061927_dp, 0.7823660354159746_dp, 1.5437686299849278_dp, &
             2.9762479203500893_dp, 5.780606277045765_dp, &
             ! w
             0.02824933969172581_dp, 0.07454603913718215_dp, 0.1490178202155422_dp, &
             0.28660129449280364_dp, 0.545507820648827_dp, 1.026582992088914_dp, &
             1.9405634730900532_dp, 4.00708798269101_dp]

      CASE (160)

         aw(:) = & ! a
            [0.009645335139589121_dp, 0.0543778231378708_dp, 0.15268317853560984_dp, &
             0.34791260514597394_dp, 0.7288141658794776_dp, 1.4610854083998512_dp, &
             2.8572596946046263_dp, 5.61957954529671_dp, &
             ! w
             0.025128643705808987_dp, 0.06700649821821521_dp, 0.1365152448171116_dp, &
             0.26826059735147795_dp, 0.5201604752923397_dp, 0.9936935209922502_dp, &
             1.9011380241938254_dp, 3.962328550369193_dp]

      CASE (161)

         aw(:) = & ! a
            [0.008704741723334132_dp, 0.04931450289315343_dp, 0.1398190631110848_dp, &
             0.32293423980780295_dp, 0.6864049880172786_dp, 1.3950320723685947_dp, &
             2.7614744811077356_dp, 5.489232398769128_dp, &
             ! w
             0.02270175596341297_dp, 0.061123900907299696_dp, 0.12669535656869585_dp, &
             0.2536760883721414_dp, 0.49971581877262344_dp, 0.9668633389980442_dp, &
             1.8687591398468626_dp, 3.92552895170279_dp]

      CASE (162)

         aw(:) = & ! a
            [0.007950339699079541_dp, 0.0452471530544418_dp, 0.1294552708085345_dp, &
             0.30270343388656173_dp, 0.6518003973409469_dp, 1.3407130798147102_dp, &
             2.6821760146617595_dp, 5.380800620728163_dp, &
             ! w
             0.020754516169563704_dp, 0.05638989857987046_dp, 0.11874302529615148_dp, &
             0.2417297705831964_dp, 0.48275998370961964_dp, 0.9443959761885042_dp, &
             1.8414905485795192_dp, 3.8945094081684952_dp]

      CASE (163)

         aw(:) = & ! a
            [0.007330367889227327_dp, 0.041899841948996194_dp, 0.12090283062560349_dp, &
             0.2859242916917437_dp, 0.6229028130797221_dp, 1.2950331295281308_dp, &
             2.6150915210491834_dp, 5.288681179279969_dp, &
             ! w
             0.019153697215708195_dp, 0.052487499170134476_dp, 0.11214831768328601_dp, &
             0.23171781920936582_dp, 0.46839273666695513_dp, 0.9251988674368323_dp, &
             1.8180761380444024_dp, 3.8678537676116576_dp]

      CASE (164)

         aw(:) = & ! a
            [0.004292369050515735_dp, 0.025415422465390292_dp, 0.07833933625763832_dp, &
             0.20079259921259504_dp, 0.4726395102313897_dp, 1.0517208224751955_dp, &
             2.2505781328076893_dp, 4.781104701694306_dp, &
             ! w
             0.011299708072695298_dp, 0.03315384599330144_dp, 0.07869657572341132_dp, &
             0.1789636493609062_dp, 0.3899113717033012_dp, 0.8174921432572277_dp, &
             1.6846079041638187_dp, 3.715537026453917_dp]

      CASE (165)

         aw(:) = & ! a
            [0.003140000416042068_dp, 0.019110817347133886_dp, 0.06173604440993229_dp, &
             0.16642818361581635_dp, 0.4095542480247393_dp, 0.9458455055554492_dp, &
             2.08738342309079_dp, 4.549417092859298_dp, &
             ! w
             0.008314770073237271_dp, 0.02568102055845356_dp, 0.0651854465211138_dp, &
             0.15634108007807077_dp, 0.3545343631542286_dp, 0.7671776007943374_dp, &
             1.620923646350249_dp, 3.6426182431755443_dp]

      CASE (166)

         aw(:) = & ! a
            [0.002517652556701411_dp, 0.015687085526692218_dp, 0.05258053100648433_dp, &
             0.14701518319112136_dp, 0.37301617466915876_dp, 0.8831960386953908_dp, &
             1.9892072955102489_dp, 4.408493070315262_dp, &
             ! w
             0.006700887588905575_dp, 0.02159084964403569_dp, 0.057542315578722775_dp, &
             0.1430624602634638_dp, 0.33318420591832787_dp, 0.7362076092535838_dp, &
             1.5812540713021126_dp, 3.5971080384364664_dp]

      CASE (167)

         aw(:) = & ! a
            [0.0021229535274735287_dp, 0.013506419943363183_dp, 0.04667371867458424_dp, &
             0.1342539621944194_dp, 0.3485587576491087_dp, 0.8406244096896625_dp, &
             1.921727596399514_dp, 4.3108989561714495_dp, &
             ! w
             0.005676595472761695_dp, 0.018968428553127636_dp, 0.052510258462381006_dp, &
             0.13408946335725638_dp, 0.3184836575206165_dp, 0.7145967548360082_dp, &
             1.5533459981559692_dp, 3.565046943939176_dp]

      CASE (168)

         aw(:) = & ! a
            [0.001848157812366753_dp, 0.011982819812136185_dp, 0.0425003466075096_dp, &
             0.1250994074781487_dp, 0.33076492891273124_dp, 0.8092949962970134_dp, &
             1.8716385878240696_dp, 4.23804767180299_dp, &
             ! w
             0.004963122756087272_dp, 0.017125452463649007_dp, 0.04889484508034009_dp, &
             0.1275139287006729_dp, 0.3075600247602414_dp, 0.6983780043252399_dp, &
             1.5322730134232336_dp, 3.540812525952162_dp]

      CASE (169)

         aw(:) = & ! a
            [0.0016447651044251732_dp, 0.010851661468111345_dp, 0.03937106572628469_dp, &
             0.11814688325978337_dp, 0.3170960230793337_dp, 0.7850072328236102_dp, &
             1.8325419240849563_dp, 4.18093122478255_dp, &
             ! w
             0.004434875104399002_dp, 0.01574993184106714_dp, 0.0461450866606819_dp, &
             0.12243364470001045_dp, 0.29902783933586385_dp, 0.6856106320477459_dp, &
             1.5156041931287734_dp, 3.5216267135009915_dp]

      CASE (170)

         aw(:) = & ! a
            [0.001487556523225969_dp, 0.009974982906957941_dp, 0.03692406090015132_dp, &
             0.1126502931174693_dp, 0.30618595287517036_dp, 0.7654746988259726_dp, &
             1.800923440813932_dp, 4.134571976484762_dp, &
             ! w
             0.004026499991842654_dp, 0.014678676911974713_dp, 0.043968217801443873_dp, &
             0.1183595924121216_dp, 0.29212454992692743_dp, 0.6752146704956449_dp, &
             1.5019777460020225_dp, 3.505931608619109_dp]

      CASE (171)

         aw(:) = & ! a
            [0.0013620554429901596_dp, 0.009273407969141249_dp, 0.034949866661316824_dp, &
             0.10817313344692848_dp, 0.29722666439401046_dp, 0.7493320377090902_dp, &
             1.7746689959691648_dp, 4.095960310062159_dp, &
             ! w
             0.003700459858514899_dp, 0.013817517288333215_dp, 0.04219286808922606_dp, &
             0.11500074858119679_dp, 0.28639069776653364_dp, 0.6665335644303946_dp, &
             1.4905611878984848_dp, 3.492773945607544_dp]

      CASE (172)

         aw(:) = & ! a
            [0.0012593274957499998_dp, 0.008697850453973345_dp, 0.03331819801427228_dp, &
             0.10444133341331838_dp, 0.2897059260448648_dp, 0.7357066637534232_dp, &
             1.752418955036336_dp, 4.0631526124948225_dp, &
             ! w
             0.003433576050391069_dp, 0.013108057719402625_dp, 0.040711374217526033_dp, &
             0.11217166530926562_dp, 0.2815304358415968_dp, 0.6591413831361904_dp, &
             1.4808119960150754_dp, 3.481532103291602_dp]

      CASE (173)

         aw(:) = & ! a
            [0.0007623416133137365_dp, 0.005890810975737877_dp, 0.025149545186670872_dp, &
             0.08523766862430776_dp, 0.25013407265478554_dp, 0.6627794284196206_dp, &
             1.6318321257986887_dp, 3.883916595608218_dp, &
             ! w
             0.0021430462268822023_dp, 0.009593451326400984_dp, 0.03305675473132598_dp, &
             0.09713197940984239_dp, 0.25518467736496675_dp, 0.6185015005689335_dp, &
             1.4267377848347798_dp, 3.419075109519514_dp]

      CASE (174)

         aw(:) = & ! a
            [0.0005779647722766573_dp, 0.004834508041275553_dp, 0.02194359693177028_dp, &
             0.07739916773698584_dp, 0.23349001096806937_dp, 0.6314114180649208_dp, &
             1.579120648939428_dp, 3.8047625323126018_dp, &
             ! w
             0.0016651967973811464_dp, 0.008234495899626803_dp, 0.029913258361711683_dp, &
             0.09072287364059083_dp, 0.24367301553143572_dp, 0.6004207194957538_dp, &
             1.402406578053114_dp, 3.390910901261706_dp]

      CASE (175)

         aw(:) = & ! a
            [0.0004803056376499483_dp, 0.004269208788726529_dp, 0.02018253421181372_dp, &
             0.07299928590414934_dp, 0.22399874724744046_dp, 0.6133156517874289_dp, &
             1.548459707964839_dp, 3.7584802839867817_dp, &
             ! w
             0.0014125661829260472_dp, 0.007493769268018652_dp, 0.02814264460179232_dp, &
             0.08704429226649443_dp, 0.23698080612286268_dp, 0.5898121473669076_dp, &
             1.3880475721468792_dp, 3.3742707589985144_dp]

      CASE (176)

         aw(:) = & ! a
            [0.0004196991604112789_dp, 0.0039154579369630625_dp, 0.019060101588172385_dp, &
             0.07015487445149451_dp, 0.21780047958598356_dp, 0.6014111142968014_dp, &
             1.5281833987584514_dp, 3.727773036068565_dp, &
             ! w
             0.0012560128074441002_dp, 0.007023904657541629_dp, 0.026995379037448926_dp, &
             0.08463234039908399_dp, 0.2325573562750163_dp, 0.5827590540484507_dp, &
             1.3784658861922865_dp, 3.363158703600496_dp]

      CASE (177)

         aw(:) = & ! a
            [0.0003786335074948956_dp, 0.0036740934531544313_dp, 0.01828370455906452_dp, &
             0.06816734757077088_dp, 0.21343856074098708_dp, 0.592990481271668_dp, &
             1.5137888616481403_dp, 3.7059238120115063_dp, &
             ! w
             0.0011500481004688015_dp, 0.006699933093528048_dp, 0.026192434291328334_dp, &
             0.08293030429864064_dp, 0.22941829836160357_dp, 0.577733583351278_dp, &
             1.371621300415861_dp, 3.355216834827016_dp]

      CASE (178)

         aw(:) = & ! a
            [0.00034930745430810777_dp, 0.003500714479076156_dp, 0.0177200698247083_dp, &
             0.06671353770291188_dp, 0.210231191212566_dp, 0.5867753361711184_dp, &
             1.5031361158199719_dp, 3.6897272261912297_dp, &
             ! w
             0.0010744332166565012_dp, 0.006465273349026221_dp, 0.02560440150682541_dp, &
             0.08167628463428958_dp, 0.22709597493762274_dp, 0.5740045808031631_dp, &
             1.366532924057776_dp, 3.349310441935163_dp]

      CASE (179)

         aw(:) = & ! a
            [0.00032771667857610175_dp, 0.003372431857593268_dp, 0.017299566753370164_dp, &
             0.06562262596516362_dp, 0.20781482681240834_dp, 0.5820796197507841_dp, &
             1.4950714931789086_dp, 3.6774503027702363_dp, &
             ! w
             0.0010187920024821151_dp, 0.006290492121673012_dp, 0.025162751693917528_dp, &
             0.08073012427677075_dp, 0.22533830676734343_dp, 0.5711759383022554_dp, &
             1.3626676872789059_dp, 3.344822524860974_dp]

      CASE (180)

         aw(:) = & ! a
            [0.0003116085515013382_dp, 0.003276327027022856_dp, 0.016982472114569446_dp, &
             0.06479629033151992_dp, 0.20597888244829302_dp, 0.5785040412974864_dp, &
             1.4889211961369415_dp, 3.6680786204192275_dp, &
             ! w
             0.0009772948930853393_dp, 0.006158854585985739_dp, 0.02482797926328938_dp, &
             0.08001041559645473_dp, 0.2239981341538948_dp, 0.5690154793075478_dp, &
             1.359712304782684_dp, 3.341390279927907_dp]

      CASE (181)

         aw(:) = & ! a
            [0.00029964174428975153_dp, 0.003204686902545453_dp, 0.01674488767701508_dp, &
             0.06417502390498837_dp, 0.20459531728809519_dp, 0.5758049688355265_dp, &
             1.4842730758101719_dp, 3.6609907150717964_dp, &
             ! w
             0.0009464736592531919_dp, 0.006060314212579553_dp, 0.024576150889317736_dp, &
             0.07946757918251539_dp, 0.22298546196725477_dp, 0.5673808117367402_dp, &
             1.3574743205521425_dp, 3.3387907492178366_dp]

      CASE (182)

         aw(:) = & ! a
            [0.00028127089949768415_dp, 0.0030942621869177655_dp, 0.016376552972548756_dp, &
             0.06320813546148488_dp, 0.20243640179262407_dp, 0.5715854881795668_dp, &
             1.4769970948293096_dp, 3.6498865034832644_dp, &
             ! w
             0.0008991679845521025_dp, 0.005907694052883593_dp, 0.02418399221444927_dp, &
             0.07861972999994296_dp, 0.2214005640066087_dp, 0.5648187066649002_dp, &
             1.353963338681178_dp, 3.334711769038868_dp]

      CASE (183)

         aw(:) = & ! a
            [0.03664663996285433_dp, 0.196621045088126_dp, 0.49988724525066536_dp, &
             0.9775594233646503_dp, 1.6858162243103594_dp, 2.7212944265320904_dp, &
             4.248587129766745_dp, 6.565902165746518_dp, 10.361975057294288_dp, &
             ! w
             0.09444081799100229_dp, 0.22798682547002244_dp, 0.3835833446278581_dp, &
             0.580876962423688_dp, 0.8513259963968657_dp, 1.2461698737081928_dp, &
             1.8552456496972611_dp, 2.878254284214635_dp, 5.064864774966903_dp]

      CASE (184)

         aw(:) = & ! a
            [0.02211601616717754_dp, 0.11970056172643999_dp, 0.3095577652107071_dp, &
             0.6220398529933125_dp, 1.114702211483489_dp, 1.8885600218294079_dp, &
             3.1147249042730754_dp, 5.0981371610313975_dp, 8.527471659429366_dp, &
             ! w
             0.05710689683806502_dp, 0.14030460024012567_dp, 0.24428884251928015_dp, &
             0.39000280939785126_dp, 0.6117598936983984_dp, 0.9636093286970012_dp, &
             1.5367428589142293_dp, 2.5294131477214146_dp, 4.67401573757719_dp]

      CASE (185)

         aw(:) = & ! a
            [0.01631421657105954_dp, 0.0888739577095855_dp, 0.23282412387217624_dp, &
             0.47762157517138293_dp, 0.8804879707385603_dp, 1.542669021349391_dp, &
             2.636221658061287_dp, 4.468355767093642_dp, 7.729002946792687_dp, &
             ! w
             0.042186813259380146_dp, 0.10501161699893571_dp, 0.18760642523953358_dp, &
             0.311262968587439_dp, 0.5109461867562471_dp, 0.8412092785410361_dp, &
             1.3943134568732372_dp, 2.3697175189404875_dp, 4.494047315704001_dp]

      CASE (186)

         aw(:) = & ! a
            [0.01310769278658543_dp, 0.07178932791056176_dp, 0.1901024737453173_dp, &
             0.39670951451930686_dp, 0.7480739001355727_dp, 1.3446599602615252_dp, &
             2.3582852576121467_dp, 4.097389192379444_dp, 7.253327633896303_dp, &
             ! w
             0.033935259759454994_dp, 0.08538754041252856_dp, 0.1558201128416026_dp, &
             0.26654675338405603_dp, 0.4525182570791364_dp, 0.7683860176045938_dp, &
             1.3074948146510295_dp, 2.2708533075573003_dp, 4.382308901189268_dp]

      CASE (187)

         aw(:) = & ! a
            [0.011046162027061476_dp, 0.060780358590888095_dp, 0.16246957067470932_dp, &
             0.3440842840849094_dp, 0.661228691056707_dp, 1.2133287214430295_dp, &
             2.1716507047182065_dp, 3.8454420774551985_dp, 6.927402028599623_dp, &
             ! w
             0.028627268985310098_dp, 0.07270850592790482_dp, 0.13513575605066283_dp, &
             0.23710656804635652_dp, 0.41333272977734137_dp, 0.7184889744300452_dp, &
             1.2469241956836628_dp, 2.201122430059486_dp, 4.303366802485584_dp]

      CASE (188)

         aw(:) = & ! a
            [0.00959767293150908_dp, 0.053030158150609935_dp, 0.1429528058023013_dp, &
             0.30672868385932517_dp, 0.5991056928552874_dp, 1.1184415035302333_dp, &
             2.0353775012644673_dp, 3.659744161164704_dp, 6.685457309051659_dp, &
             ! w
             0.024895972176476603_dp, 0.06376235712417291_dp, 0.12044910511050758_dp, &
             0.2159745478825277_dp, 0.38473919314424293_dp, 0.6814342485253309_dp, &
             1.201306390551467_dp, 2.1481696096724234_dp, 4.243353313165919_dp]

      CASE (189)

         aw(:) = & ! a
            [0.008518360871648093_dp, 0.047245437550452345_dp, 0.12834328815009152_dp, &
             0.27863568352195117_dp, 0.5520531959128403_dp, 1.0459297104212417_dp, &
             1.9302841753638598_dp, 3.515390795108448_dp, 6.496258291887618_dp, &
             ! w
             0.022114523859998472_dp, 0.05707172554263855_dp, 0.10940265203680985_dp, &
             0.19991824725567156_dp, 0.36269407103471935_dp, 0.6524437931310794_dp, &
             1.165208706530346_dp, 2.1059899831229174_dp, 4.195511171460698_dp]

      CASE (190)

         aw(:) = & ! a
            [0.007679789483948467_dp, 0.04274416689670455_dp, 0.1169452389354415_dp, &
             0.25662289986184644_dp, 0.514940659935099_dp, 0.9882757618211389_dp, &
             1.8460526692834496_dp, 3.398895119289925_dp, 6.342795057626663_dp, &
             ! w
             0.019952678430106358_dp, 0.05185630039334263_dp, 0.10074660869725498_dp, &
             0.18721649412691854_dp, 0.3450262461891317_dp, 0.6289183401844696_dp, &
             1.1356374020567792_dp, 2.0712467277560673_dp, 4.156078967198704_dp]

      CASE (191)

         aw(:) = & ! a
            [0.007007525840355502_dp, 0.03913062459196527_dp, 0.10777290996272314_dp, &
             0.23883615703861727_dp, 0.4847687047212495_dp, 0.941062053978471_dp, &
             1.7765815928488142_dp, 3.3022331320416156_dp, 6.214897269621191_dp, &
             ! w
             0.018218993905498363_dp, 0.04766271101136857_dp, 0.09375242604685922_dp, &
             0.17686154628295433_dp, 0.33045366257578834_dp, 0.6093043202241387_dp, &
             1.1107832014454115_dp, 2.0419093040247858_dp, 4.122764944772186_dp]

      CASE (192)

         aw(:) = & ! a
            [0.0064552940709981695_dp, 0.03615852291388057_dp, 0.10021175216681184_dp, &
             0.22411697154250507_dp, 0.4596569999787067_dp, 0.9015052273114635_dp, &
             1.7180040401346317_dp, 3.2202910181231106_dp, 6.106054374868566_dp, &
             ! w
             0.016794417745794436_dp, 0.04420843632131467_dp, 0.08796472897000776_dp, &
             0.1682207519727353_dp, 0.31816462771886816_dp, 0.5926071686898263_dp, &
             1.0894766158992761_dp, 2.016657531293518_dp, 4.094077743972719_dp]

      CASE (193)

         aw(:) = & ! a
            [0.003753505907258484_dp, 0.02155182075421551_dp, 0.06273631153032878_dp, &
             0.1500627747875536_dp, 0.3306054926762188_dp, 0.6934050066019539_dp, &
             1.4030303812801839_dp, 2.7716770734119853_dp, 5.502433394755472_dp, &
             ! w
             0.00981702804109182_dp, 0.02714183122363002_dp, 0.05885554054999135_dp, &
             0.12336869370144365_dp, 0.2520334255785889_dp, 0.4999564200366364_dp, &
             0.9685402483680957_dp, 1.8714190637073345_dp, 3.9288430703647426_dp]

      CASE (194)

         aw(:) = & ! a
            [0.00273146050814806_dp, 0.015983936853035185_dp, 0.04822855460935647_dp, &
             0.12059563145779052_dp, 0.27738623596598116_dp, 0.6044051058271435_dp, &
             1.2638987330530245_dp, 2.5683342984551687_dp, 5.223861350006336_dp, &
             ! w
             0.007172698427828807_dp, 0.020576588872870942_dp, 0.04727844477596849_dp, &
             0.10455119508323607_dp, 0.22278452857668699_dp, 0.4572258464535361_dp, &
             0.9110357207216384_dp, 1.801108799828981_dp, 3.848690607070614_dp]

      CASE (195)

         aw(:) = & ! a
            [0.002180530109983482_dp, 0.012966829796073573_dp, 0.04027380641050435_dp, &
             0.10410674805475155_dp, 0.24688474333689828_dp, 0.5522250628274896_dp, &
             1.1807304388190987_dp, 2.444927751309307_dp, 5.05303534364627_dp, &
             ! w
             0.005745521949355099_dp, 0.016995748484371104_dp, 0.04079999911086326_dp, &
             0.09363900972689708_dp, 0.2052932850347775_dp, 0.43106468673745896_dp, &
             0.8752159980258892_dp, 1.7568568252994703_dp, 3.798180091949992_dp]

      CASE (196)

         aw(:) = & ! a
            [0.0018316193162704233_dp, 0.011048247676452443_dp, 0.03516488856787591_dp, &
             0.0933425211230552_dp, 0.22661138549778412_dp, 0.5169700044462441_dp, &
             1.1237641613889138_dp, 2.359501711743573_dp, 4.9339327011643235_dp, &
             ! w
             0.00484084226221103_dp, 0.014706549877645558_dp, 0.03656917957236209_dp, &
             0.0863219618381845_dp, 0.19331294173316757_dp, 0.41285608032480475_dp, &
             0.8499869026869297_dp, 1.7254635478924916_dp, 3.762313206988196_dp]

      CASE (197)

         aw(:) = & ! a
            [0.0015889645189828857_dp, 0.009709423586791393_dp, 0.0315686052432081_dp, &
             0.08566105902673506_dp, 0.21193512199224865_dp, 0.4911220300723129_dp, &
             1.0815587767859631_dp, 2.2957009271064512_dp, 4.844498056227317_dp, &
             ! w
             0.004211216499798775_dp, 0.013101744341948425_dp, 0.033548366727585155_dp, &
             0.08098808669378582_dp, 0.1844390570715673_dp, 0.39920566599403257_dp, &
             0.8309034361257308_dp, 1.701587525270117_dp, 3.735014597504688_dp]

      CASE (198)

         aw(:) = & ! a
            [0.0014095059934701932_dp, 0.008716394943563546_dp, 0.028880260117061844_dp, &
             0.07985102079144676_dp, 0.20070186121842573_dp, 0.4711331220524638_dp, &
             1.04864495451299_dp, 2.245626201329628_dp, 4.774002223938378_dp, &
             ! w
             0.0037452979413056405_dp, 0.011906536498876437_dp, 0.03126205748967432_dp, &
             0.0768822495241899_dp, 0.1775211901125934_dp, 0.38846227970546593_dp, &
             0.8157768943467678_dp, 1.6825795696161268_dp, 3.713268572163096_dp]

      CASE (199)

         aw(:) = & ! a
            [0.0012708763612674769_dp, 0.007947321379349486_dp, 0.026783343381280474_dp, &
             0.07527226607849304_dp, 0.19175946058786122_dp, 0.45508305566151963_dp, &
             1.0220320315021252_dp, 2.2049227056952825_dp, 4.716496405330148_dp, &
             ! w
             0.0033852098812279124_dp, 0.010977408073676732_dp, 0.02945905545806215_dp, &
             0.07359810659635183_dp, 0.17192991097391452_dp, 0.3797108500746659_dp, &
             0.8033826246878993_dp, 1.6669487573286486_dp, 3.695376880011522_dp]

      CASE (200)

         aw(:) = & ! a
            [0.001160248421106101_dp, 0.007332176435702436_dp, 0.02509509829800203_dp, &
             0.07155197410263119_dp, 0.18442989266337936_dp, 0.441830506360945_dp, &
             0.9999272003534081_dp, 2.1709623019090607_dp, 4.6683738668041475_dp, &
             ! w
             0.003097742469161467_dp, 0.010231663890554594_dp, 0.027993090102620757_dp, &
             0.07089526892841691_dp, 0.16728774916681835_dp, 0.3723968344810379_dp, &
             0.7929727209927074_dp, 1.6537803753979143_dp, 3.680297016053351_dp]

      CASE (201)

         aw(:) = & ! a
            [0.0010697132697980306_dp, 0.006827699189441219_dp, 0.023702137973609167_dp, &
             0.06845703145726359_dp, 0.1782853335724231_dp, 0.43064914538427485_dp, &
             0.9811811804065309_dp, 2.142050481175247_dp, 4.627299876929942_dp, &
             ! w
             0.0028624090304025795_dp, 0.009618099703285324_dp, 0.026772710656902746_dp, &
             0.06862135720991581_dp, 0.163352677554946_dp, 0.3661615356237251_dp, &
             0.7840601806864651_dp, 1.6424763479445015_dp, 3.6673470190152675_dp]

      CASE (202)

         aw(:) = & ! a
            [0.0006313454889490319_dp, 0.004366846985787416_dp, 0.016754567817872237_dp, &
             0.05258401592915444_dp, 0.1459750048379582_dp, 0.3706381471395139_dp, &
             0.8789214132697631_dp, 1.9823948779811633_dp, 4.398639838848585_dp, &
             ! w
             0.0017219984012311613_dp, 0.006588442625411695_dp, 0.020496274793963855_dp, &
             0.05652945093522863_dp, 0.14192707509516378_dp, 0.3315985863927448_dp, &
             0.7339849281745117_dp, 1.5784277751334237_dp, 3.5938760022078236_dp]

      CASE (203)

         aw(:) = & ! a
            [0.0004677826770463522_dp, 0.003436913173081709_dp, 0.014027484844973091_dp, &
             0.04608466268371336_dp, 0.13227034370139945_dp, 0.34446469222973564_dp, &
             0.8333436224059947_dp, 1.9100798947374522_dp, 4.293972597070242_dp, &
             ! w
             0.0012962831308852621_dp, 0.005418064450388919_dp, 0.017913634087816786_dp, &
             0.051323222153869655_dp, 0.13240949292960194_dp, 0.3158801151655617_dp, &
             0.7108062282245082_dp, 1.5484532018565422_dp, 3.5594302119364722_dp]

      CASE (204)

         aw(:) = & ! a
            [0.0003803762514503546_dp, 0.0029354867547009725_dp, 0.012518840564158371_dp, &
             0.0423977620853282_dp, 0.12434109876798174_dp, 0.32908940347314647_dp, &
             0.8062550457285699_dp, 1.866728574542357_dp, 4.2308745838119775_dp, &
             ! w
             0.0010688850352942277_dp, 0.0047768063467026785_dp, 0.016443757155660196_dp, &
             0.04828720609205642_dp, 0.12676597891135466_dp, 0.30644185437677185_dp, &
             0.6967562347704147_dp, 1.5301758899530689_dp, 3.538405395173158_dp]

      CASE (205)

         aw(:) = & ! a
            [0.0003254404360101519_dp, 0.0026180514674484442_dp, 0.01154517732693377_dp, &
             0.03997641243207779_dp, 0.11906408321953327_dp, 0.31875343796582456_dp, &
             0.7879045136257067_dp, 1.837195090818923_dp, 4.187731284690917_dp, &
             ! w
             0.0009260627770325223_dp, 0.004365663939146157_dp, 0.015476137494080599_dp, &
             0.046256432129645954_dp, 0.12294951395737774_dp, 0.3000063118609165_dp, &
             0.6871166160166533_dp, 1.517587078958083_dp, 3.523914413006404_dp]

      CASE (206)

         aw(:) = & ! a
            [0.00028751792539579725_dp, 0.002397572916353228_dp, 0.010858482130977168_dp, &
             0.038246225642043115_dp, 0.11525649534425735_dp, 0.3112408729002721_dp, &
             0.7744924306712251_dp, 1.8155216145202242_dp, 4.1559868654318946_dp, &
             ! w
             0.0008275446967731432_dp, 0.00407709282091288_dp, 0.014783476806786546_dp, &
             0.04478585386996454_dp, 0.12016387661043228_dp, 0.295280971897028_dp, &
             0.6800069561584169_dp, 1.508276110585112_dp, 3.513191249963715_dp]

      CASE (207)

         aw(:) = & ! a
            [0.0002596868790846979_dp, 0.0022348903483392375_dp, 0.010345396602249564_dp, &
             0.03694010841533558_dp, 0.11236044056791297_dp, 0.3054946294132512_dp, &
             0.7641900750527943_dp, 1.798821603444244_dp, 4.131477966912905_dp, &
             ! w
             0.000755293632344843_dp, 0.0038622717168615727_dp, 0.014259845002795556_dp, &
             0.04366428520270354_dp, 0.1180264413253419_dp, 0.2916386422069089_dp, &
             0.6745080482776007_dp, 1.5010591549296193_dp, 3.5048765058635483_dp]

      CASE (208)

         aw(:) = & ! a
            [0.00023836476106365418_dp, 0.0021096524885158997_dp, 0.00994622307968993_dp, &
             0.03591545744278059_dp, 0.11007474202476876_dp, 0.30093910080735065_dp, &
             0.7559949292957713_dp, 1.7855045682887565_dp, 4.111902905373047_dp, &
             ! w
             0.0006999750349259217_dp, 0.0036956292786038195_dp, 0.013848581672983534_dp, &
             0.04277719760189193_dp, 0.11632771186630943_dp, 0.2887334228096179_dp, &
             0.670110049997968_dp, 1.4952771916809655_dp, 3.498213004817574_dp]

      CASE (209)

         aw(:) = & ! a
            [0.00022150414516648346_dp, 0.00201018703456342_dp, 0.00962632871459135_dp, &
             0.03508860562192356_dp, 0.10822110981725729_dp, 0.2972311371099197_dp, &
             0.7493060755473707_dp, 1.7746133568189388_dp, 4.095872911844918_dp, &
             ! w
             0.0006562555949134862_dp, 0.0035623967101711023_dp, 0.013516391665643545_dp, &
             0.04205655242733982_dp, 0.1149422647181856_dp, 0.28635696925323484_dp, &
             0.6665045419456763_dp, 1.4905304766796543_dp, 3.492741194242436_dp]

      CASE (210)

         aw(:) = & ! a
            [0.00020784696525288592_dp, 0.0019292989209037795_dp, 0.00936414928646571_dp, &
             0.03440695455808136_dp, 0.10668662080214998_dp, 0.2941521685207763_dp, &
             0.7437390948258376_dp, 1.7655336768516126_dp, 4.082494835296907_dp, &
             ! w
             0.0006208594515187709_dp, 0.003453412616156619_dp, 0.013242316836217145_dp, &
             0.041459127716502404_dp, 0.11378993066689655_dp, 0.2843755027087347_dp, &
             0.6634927640743865_dp, 1.4865608254662939_dp, 3.488164179515476_dp]

      CASE (211)

         aw(:) = & ! a
            [0.0001468747327408845_dp, 0.0015634875681823865_dp, 0.008151590050869005_dp, &
             0.031203069191438823_dp, 0.09939208476306977_dp, 0.2793935723291259_dp, &
             0.7168880009925573_dp, 1.7215411744800457_dp, 4.017487616446215_dp, &
             ! w
             0.00046301946625643294_dp, 0.0029518667511140365_dp, 0.011951279637165236_dp, &
             0.038608343330302074_dp, 0.10824204346377927_dp, 0.27477195324743986_dp, &
             0.6488225767108784_dp, 1.4671639226752504_dp, 3.465786603779763_dp]

      CASE (212)

         aw(:) = & ! a
            [0.00013620787587729222_dp, 0.0014984233172990983_dp, 0.007930484313668681_dp, &
             0.030608852526075443_dp, 0.09802331751259105_dp, 0.2766007114067538_dp, &
             0.7117747229082216_dp, 1.7131253759530025_dp, 4.005015434301447_dp, &
             ! w
             0.0004354299056155451_dp, 0.002860852031658403_dp, 0.011711291382515341_dp, &
             0.038071387509506595_dp, 0.10718758733492406_dp, 0.27293433084542407_dp, &
             0.6460013499115237_dp, 1.4634218815677373_dp, 3.461467001531876_dp]

      CASE (213)

         aw(:) = & ! a
            [0.03306976242736113_dp, 0.17680784378283929_dp, 0.4464853448685271_dp, &
             0.863799409509969_dp, 1.4665304201179228_dp, 2.31721140013242_dp, &
             3.517208942969788_dp, 5.232873217999519_dp, 7.761728767892475_dp, &
             11.795294978905565_dp, &
             ! w
             0.08515488845145727_dp, 0.2041221377379093_dp, 0.3387624704046451_dp, &
             0.5020218012768304_dp, 0.7136389791460016_dp, 1.00422833835908_dp, &
             1.4225951252393652_dp, 2.055377543910107_dp, 3.1013168391488204_dp, &
             5.321560423899623_dp]

      CASE (214)

         aw(:) = & ! a
            [0.0199190782019327_dp, 0.1072334417383125_dp, 0.2744085978174513_dp, &
             0.5421416040798438_dp, 0.9483605618497398_dp, 1.5578821538499603_dp, &
             2.4764307129347025_dp, 3.8758067491779373_dp, 6.058295080044497_dp, &
             9.71184803990698_dp, &
             ! w
             0.05137209581340257_dp, 0.12485945800871932_dp, 0.21283441302264827_dp, &
             0.32879262838087114_dp, 0.49422642050095056_dp, 0.74206692724155_dp, &
             1.1227892497020227_dp, 1.7234803772519538_dp, 2.7403911047158_dp, &
             4.91502735537284_dp]

      CASE (215)

         aw(:) = & ! a
            [0.014672089347410964_dp, 0.07938691179467121_dp, 0.20517497813391075_dp, &
             0.41184592426970557_dp, 0.736738817100995_dp, 1.2445399693315784_dp, &
             2.041220998765312_dp, 3.2995507853921113_dp, 5.323729672106679_dp, &
             8.80146136115717_dp, &
             ! w
             0.037882989131618645_dp, 0.09301632450371458_dp, 0.1617466336454914_dp, &
             0.25765519527407565_dp, 0.40269419497450526_dp, 0.6301925993044387_dp, &
             0.9910799607619161_dp, 1.57344701648294_dp, 2.5740012849105933_dp, &
             4.726857011736634_dp]

      CASE (216)

         aw(:) = & ! a
            [0.011774025170233755_dp, 0.0639700951282584_dp, 0.1666917616889322_dp, &
             0.3390289700500115_dp, 0.6176072358757309_dp, 1.0663680454451372_dp, &
             1.7906093738966222_dp, 2.9631424352797224_dp, 4.889379313076166_dp, &
             8.257531682354081_dp, &
             ! w
             0.03042835426798262_dp, 0.0753370830029148_dp, 0.1331690520655589_dp, &
             0.21744811936857542_dp, 0.3501479372604659_dp, 0.5645330070305195_dp, &
             0.9118556551925829_dp, 1.4812981052443768_dp, 2.4704999769201472_dp, &
             4.609597851526631_dp]

      CASE (217)

         aw(:) = & ! a
            [0.009911839853047714_dp, 0.05404467176447483_dp, 0.141834651822087_dp, &
             0.2917769967006175_dp, 0.5397886064780781_dp, 0.9489117072037068_dp, &
             1.6235559881076618_dp, 2.736311064442644_dp, 4.593474331342694_dp, &
             7.883962032792184_dp, &
             ! w
             0.025636076499812976_dp, 0.06392887219780452_dp, 0.11461312317687036_dp, &
             0.1910990220813975_dp, 0.315211226733076_dp, 0.5200333749196062_dp, &
             0.8571142202739548_dp, 1.416643388358046_dp, 2.397230987996018_dp, &
             4.526514342364133_dp]

      CASE (218)

         aw(:) = & ! a
            [0.008604064608033616_dp, 0.04706270872618831_dp, 0.12429966311710704_dp, &
             0.25830664284518123_dp, 0.484331180051568_dp, 0.8645073678078468_dp, &
             1.5023398624500846_dp, 2.570112871924677_dp, 4.374813518228905_dp, &
             7.6060959473532055_dp, &
             ! w
             0.022269224904038827_dp, 0.05588806630464082_dp, 0.10146365565522789_dp, &
             0.17226893055438913_dp, 0.2899119020042487_dp, 0.4872765569924376_dp, &
             0.8161879928937111_dp, 1.3677291102282898_dp, 2.341423980323161_dp, &
             4.463197532355998_dp]

      CASE (219)

         aw(:) = & ! a
            [0.007630040579213754_dp, 0.04185500587435238_dp, 0.1111879397803506_dp, &
             0.23318545543856892_dp, 0.4424728513495299_dp, 0.8003165994032522_dp, &
             1.4093598104639633_dp, 2.4415582485058356_dp, 4.2044568362569175_dp, &
             7.3884269231485025_dp, &
             ! w
             0.019760738475349357_dp, 0.04988023165757547_dp, 0.09159139706158959_dp, &
             0.1580208841785473_dp, 0.27053665184305753_dp, 0.4618333919085842_dp, &
             0.7839904120304717_dp, 1.3288768434592462_dp, 2.296856375044337_dp, &
             4.412613921753437_dp]

      CASE (220)

         aw(:) = & ! a
            [0.006873584630849889_dp, 0.037805265177748806_dp, 0.10096864775958635_dp, &
             0.21353808268362326_dp, 0.4095631635599569_dp, 0.749498579552089_dp, &
             1.3351857641416567_dp, 2.3382521089266644_dp, 4.066705245618029_dp, &
             7.211595244532029_dp, &
             ! w
             0.01781196457799392_dp, 0.04520104299920789_dp, 0.08386868600678984_dp, &
             0.14679334916722228_dp, 0.25509977725182514_dp, 0.44131140027782323_dp, &
             0.7577390458093131_dp, 1.2969458549711415_dp, 2.260063004909375_dp, &
             4.370842386544174_dp]

      CASE (221)

         aw(:) = & ! a
            [0.006267382968008953_dp, 0.03455606410432884_dp, 0.09275251241973845_dp, &
             0.1976907281888568_dp, 0.3828877720426687_dp, 0.7080447357505528_dp, &
             1.2742625091008632_dp, 2.25284939047651_dp, 3.952204179611082_dp, &
             7.06401285647989_dp, &
             ! w
             0.01624982988730891_dp, 0.0414415425972865_dp, 0.07763876472913307_dp, &
             0.13767351439365783_dp, 0.24243359586166527_dp, 0.42428979480948564_dp, &
             0.735762739448808_dp, 1.2700319870608947_dp, 2.228931348991005_dp, &
             4.335490774803464_dp]

      CASE (222)

         aw(:) = & ! a
            [0.0057695950967900245_dp, 0.03188503735236182_dp, 0.0859854511862231_dp, &
             0.1845984841571088_dp, 0.3607477727164536_dp, 0.6734363415369501_dp, &
             1.2230817507171463_dp, 2.1806860194693054_dp, 3.8549835889359443_dp, &
             6.938254312061828_dp, &
             ! w
             0.014966733330367177_dp, 0.03834702878119511_dp, 0.07249146439732221_dp, &
             0.13008928187098504_dp, 0.2318018689487819_dp, 0.409864313242222_dp, &
             0.716987073246654_dp, 1.2469010750155314_dp, 2.20208574116089_dp, &
             4.305000487897624_dp]

      CASE (223)

         aw(:) = & ! a
            [0.0033373898429685275_dp, 0.018782920102500628_dp, 0.05255601754553826_dp, &
             0.1191576576823401_dp, 0.24812207557829966_dp, 0.4935977722085606_dp, &
             0.9512698505332244_dp, 1.7897496715257404_dp, 3.3196100790475946_dp, &
             6.2374001617059704_dp, &
             ! w
             0.008691576585670246_dp, 0.023096915721798953_dp, 0.046761944252935964_dp, &
             0.09121522731605315_dp, 0.17546964148618782_dp, 0.3309435153894121_dp, &
             0.6115252378444787_dp, 1.11442752382245_dp, 2.0466210507551135_dp, &
             4.1283173287555055_dp]

      CASE (224)

         aw(:) = & ! a
            [0.002419472338931235_dp, 0.013804633431198317_dp, 0.03969259714997238_dp, &
             0.09342229848814294_dp, 0.2024464290319135_dp, 0.41810184647482834_dp, &
             0.8332962399975895_dp, 1.6150345745196448_dp, 3.0746551660908783_dp, &
             5.911309202082852_dp, &
             ! w
             0.006319480627819706_dp, 0.017256004783303094_dp, 0.03664720484282423_dp, &
             0.07523541712130324_dp, 0.15108705453980023_dp, 0.29520356140995513_dp, &
             0.5620186886646714_dp, 1.050575234327714_dp, 1.9705361918601552_dp, &
             4.041763024591411_dp]

      CASE (225)

         aw(:) = & ! a
            [0.001925490917079591_dp, 0.011112891434512117_dp, 0.03267157966957963_dp, &
             0.07914231127444575_dp, 0.17654897955074694_dp, 0.3743260077419205_dp, &
             0.763465141856196_dp, 1.5097794912507614_dp, 2.9250174225165053_dp, &
             5.710152417296363_dp, &
             ! w
             0.005041488711366136_dp, 0.01408000545000744_dp, 0.031036726307924655_dp, &
             0.06608638494458426_dp, 0.13667658371254043_dp, 0.27352566637805875_dp, &
             0.5313705135370472_dp, 1.0104385988466746_dp, 1.9222796166689597_dp, &
             3.986827718496853_dp]

      CASE (226)

         aw(:) = & ! a
            [0.001613055445090334_dp, 0.009404070986281243_dp, 0.028179289645629872_dp, &
             0.0698808951661137_dp, 0.15946969745111075_dp, 0.34497335680324653_dp, &
             0.7159420334350729_dp, 1.4372483830045268_dp, 2.8208885751505575_dp, &
             5.569217376576335_dp, &
             ! w
             0.004232468560920475_dp, 0.012054637074887036_dp, 0.02739864157379049_dp, &
             0.06000636980438821_dp, 0.12688158041115283_dp, 0.2585247942583659_dp, &
             0.5098606010576789_dp, 0.981968501582794_dp, 1.8878328906613233_dp, &
             3.9475916782101654_dp]

      CASE (227)

         aw(:) = & ! a
            [0.0013960043685149544_dp, 0.008213254887607688_dp, 0.025027267413297496_dp, &
             0.06330693283229066_dp, 0.1471799227081379_dp, 0.3235737232994006_dp, &
             0.6808942948679826_dp, 1.3832418428333828_dp, 2.7427730207536514_dp, &
             5.462943326466918_dp, &
             ! w
             0.003670030390081848_dp, 0.010637799788337269_dp, 0.02481632131404231_dp, &
             0.05560394069299469_dp, 0.11966564665897712_dp, 0.24732344582669916_dp, &
             0.4936263432805594_dp, 0.9603068597394765_dp, 1.8614967394761157_dp, &
             3.9175799473154904_dp]

      CASE (228)

         aw(:) = & ! a
            [0.0012356310955025927_dp, 0.007331032472325871_dp, 0.022677688858641562_dp, &
             0.05835663352935596_dp, 0.13781848901562838_dp, 0.30709662141570726_dp, &
             0.6536553453439855_dp, 1.3409422950604044_dp, 2.6812225946439_dp, &
             5.378859676938215_dp, &
             ! w
             0.003254204988938095_dp, 0.009584571527771694_dp, 0.022871636879711024_dp, &
             0.05223282184648336_dp, 0.11406297049810572_dp, 0.2385322997992477_dp, &
             0.48077611616865584_dp, 0.9430492834876176_dp, 1.840433117231212_dp, &
             3.8935670915039187_dp]

      CASE (229)

         aw(:) = & ! a
            [0.0011118450661519537_dp, 0.006648463900226815_dp, 0.020849619765312084_dp, &
             0.05447016243526779_dp, 0.13039557675356525_dp, 0.29391196427837873_dp, &
             0.631688149053242_dp, 1.3066088455436546_dp, 2.6310144253057905_dp, &
             5.31003632366897_dp, &
             ! w
             0.0029330749052116586_dp, 0.008767208268500323_dp, 0.02134462743679683_dp, &
             0.04954768010270035_dp, 0.1095487420615465_dp, 0.23138584059176653_dp, &
             0.4702560460707865_dp, 0.9288450627750301_dp, 1.8230399046731272_dp, &
             3.873731693471814_dp]

      CASE (230)

         aw(:) = & ! a
            [0.0010131325949005471_dp, 0.006102993111564896_dp, 0.019381144616221816_dp, &
             0.05132266342783186_dp, 0.12433148787755646_dp, 0.2830557631852255_dp, &
             0.6134788317812593_dp, 1.2779919332611942_dp, 2.588988376603919_dp, &
             5.252261882250294_dp, &
             ! w
             0.002676872297502817_dp, 0.00811218941641512_dp, 0.020107699851274852_dp, &
             0.04734541295731025_dp, 0.10580994489661319_dp, 0.2254222361170528_dp, &
             0.4614245274188429_dp, 0.916866220993235_dp, 1.8083310522241476_dp, &
             3.8569524630728167_dp]

      CASE (231)

         aw(:) = & ! a
            [0.0009323990731474346_dp, 0.005656005251340616_dp, 0.018171976481109728_dp, &
             0.048711670607687746_dp, 0.11926194585764506_dp, 0.27391713557144975_dp, &
             0.598060633554161_dp, 1.2536454845296414_dp, 2.5531023689261665_dp, &
             5.202804863940397_dp, &
             ! w
             0.0024672464939504654_dp, 0.0075740393679528135_dp, 0.019081354523931775_dp, &
             0.04549790683754734_dp, 0.10264676840998525_dp, 0.22034384503099905_dp, &
             0.45386487819252747_dp, 0.9065719038631264_dp, 1.7956602125219736_dp, &
             3.842494156645649_dp]

      CASE (232)

         aw(:) = & ! a
            [0.000542169647820812_dp, 0.003480673073208015_dp, 0.012180844662780915_dp, &
             0.035434707946528865_dp, 0.09280878667460729_dp, 0.22514555176454076_dp, &
             0.5142100271408127_dp, 1.1191922258880422_dp, 2.352574322476114_dp, &
             4.924224886750447_dp, &
             ! w
             0.001452680781671318_dp, 0.004929747195516357_dp, 0.013855151302821705_dp, &
             0.035746821242962044_dp, 0.08549708589224828_dp, 0.19223374522917086_dp, &
             0.4113201462352774_dp, 0.8478938111702771_dp, 1.722869004271629_dp, &
             3.7593552654949773_dp]

      CASE (233)

         aw(:) = & ! a
            [0.0003967302317420766_dp, 0.0026604835814450934_dp, 0.009848928612008827_dp, &
             0.030048816748303474_dp, 0.08166219386211819_dp, 0.20393288868216913_dp, &
             0.4767849144334122_dp, 1.0579265306400314_dp, 2.2597542577639733_dp, &
             4.7939093124023815_dp, &
             ! w
             0.001073880461895617_dp, 0.003915231331134727_dp, 0.011728458153009026_dp, &
             0.03157155026457845_dp, 0.07788195876052818_dp, 0.17940022094088975_dp, &
             0.3914628342219864_dp, 0.8200399624470611_dp, 1.6879552117417538_dp, &
             3.719424268448312_dp]

      CASE (234)

         aw(:) = & ! a
            [0.00031890135131914344_dp, 0.002218040837918263_dp, 0.008562208380223413_dp, &
             0.02699845169830664_dp, 0.07520683111420867_dp, 0.1914251600871076_dp, &
             0.45439737636288313_dp, 1.0208561605370439_dp, 2.20310436847618_dp, &
             4.7139161935991885_dp, &
             ! w
             0.0008710258738679723_dp, 0.0033608319216372057_dp, 0.010520901561447929_dp, &
             0.029131879268255015_dp, 0.073342527978235_dp, 0.1716321958691351_dp, &
             0.37929632229606325_dp, 0.8028142821879096_dp, 1.6662385945818383_dp, &
             3.69456689051089_dp]

      CASE (235)

         aw(:) = & ! a
            [0.00026986356050743616_dp, 0.0019374881805125897_dp, 0.007731658122650176_dp, &
             0.024991954089039963_dp, 0.07089423488187012_dp, 0.18296615376928954_dp, &
             0.439108530776862_dp, 0.9953449891370985_dp, 2.163893145603384_dp, &
             4.658334784577247_dp, &
             ! w
             0.0007431794401095744_dp, 0.0030055420983453205_dp, 0.00972495697615314_dp, &
             0.02749231394684896_dp, 0.07025059678298039_dp, 0.16628654509079477_dp, &
             0.37085524684180093_dp, 0.7907883037645156_dp, 1.6510183817369137_dp, &
             3.6771356408478857_dp]

      CASE (236)

         aw(:) = & ! a
            [0.00023590547799222017_dp, 0.0017421610274382282_dp, 0.0071448724389958735_dp, &
             0.023553433794488387_dp, 0.06776606583548084_dp, 0.17677413911353548_dp, &
             0.4278362266663701_dp, 0.9764291736433911_dp, 2.134695487213037_dp, &
             4.616830999915808_dp, &
             ! w
             0.000654646086815901_dp, 0.0027559343285720023_dp, 0.009153365507686223_dp, &
             0.026297880154494843_dp, 0.06797568401762288_dp, 0.16232353809633152_dp, &
             0.36455975071295516_dp, 0.7817777963990976_dp, 1.639581912179826_dp, &
             3.664032236363803_dp]

      CASE (237)

         aw(:) = & ! a
            [0.00021088927916765285_dp, 0.0015975896285734075_dp, 0.006705129014537052_dp, &
             0.022462537087632566_dp, 0.06537177483149383_dp, 0.17200075407454837_dp, &
             0.41909753981230863_dp, 0.9617003167186978_dp, 2.111885654080337_dp, &
             4.584336718927926_dp, &
             ! w
             0.0005894337117817894_dp, 0.0025697219209575373_dp, 0.008719298224014746_dp, &
             0.025380603698704273_dp, 0.0662150923425635_dp, 0.15923835005988818_dp, &
             0.3596358039121037_dp, 0.7747051418395825_dp, 1.6305850484217983_dp, &
             3.653720535967333_dp]

      CASE (238)

         aw(:) = & ! a
            [0.00019163628898460425_dp, 0.0014858556424114495_dp, 0.0063615903805791605_dp, &
             0.02160185169515269_dp, 0.0634683844611125_dp, 0.16818392811964739_dp, &
             0.41207819655857164_dp, 0.9498272260709857_dp, 2.093449438711582_dp, &
             4.558026872726239_dp, &
             ! w
             0.000539255212760206_dp, 0.002424792621295801_dp, 0.008376428653471312_dp, &
             0.024649441921090424_dp, 0.0648029218064254_dp, 0.1567518884914158_dp, &
             0.3556524160150504_dp, 0.7689669348259185_dp, 1.623272536629018_dp, &
             3.645337053467284_dp]

      CASE (239)

         aw(:) = & ! a
            [0.00017632786429649504_dp, 0.0013966738489163582_dp, 0.006084786373808149_dp, &
             0.02090251844390575_dp, 0.06191195625921248_dp, 0.165047673834177_dp, &
             0.4062886092368535_dp, 0.9400053232670332_dp, 2.0781646409684527_dp, &
             4.536182664138307_dp, &
             ! w
             0.0004993672899010215_dp, 0.00230838510826411_dp, 0.008097554874040448_dp, &
             0.02405024342240155_dp, 0.06363958536256346_dp, 0.1546954156882667_dp, &
             0.3523475646338255_dp, 0.7641947978821464_dp, 1.6171820561365569_dp, &
             3.638352948786651_dp]

      CASE (240)

         aw(:) = & ! a
            [0.0001638448168262917_dp, 0.0013236958800674286_dp, 0.005856363480641509_dp, &
             0.02032121754616425_dp, 0.06061117238246225_dp, 0.16241571930937454_dp, &
             0.401414372352973_dp, 0.9317156300794015_dp, 2.0652402323623247_dp, &
             4.51768916274079_dp, &
             ! w
             0.0004668500190071136_dp, 0.0022125830949226376_dp, 0.007865544601183026_dp, &
             0.02354853492163057_dp, 0.06266121898451606_dp, 0.1529600992483177_dp, &
             0.34955143608834877_dp, 0.7601490813086588_dp, 1.6120121524724762_dp, &
             3.632423318172745_dp]

      CASE (241)

         aw(:) = & ! a
            [0.0001044471836183439_dp, 0.0009720821957014101_dp, 0.004725443567268713_dp, &
             0.01737862810508732_dp, 0.0539181294703908_dp, 0.1487055525765664_dp, &
             0.37578058267914366_dp, 0.8877951802573104_dp, 1.9963843960430068_dp, &
             4.418804877219708_dp, &
             ! w
             0.0003122983278971261_dp, 0.00174201049818798_dp, 0.006687907463837846_dp, &
             0.020953148975112912_dp, 0.05753309764487348_dp, 0.14377294911196278_dp, &
             0.3346309405621727_dp, 0.7384301152245304_dp, 1.5841532058029584_dp, &
             3.6004513613032674_dp]

      CASE (242)

         aw(:) = & ! a
            [8.35622857860373e-05_dp, 0.0008458181869110869_dp, 0.004303148689255562_dp, &
             0.016247236036916173_dp, 0.051290689636432255_dp, 0.1432399532634682_dp, &
             0.36544035137816383_dp, 0.8699163624140205_dp, 1.9681651743163817_dp, &
             4.3780996775916_dp, &
             ! w
             0.0002580490502292502_dp, 0.0015680060081697754_dp, 0.006233523092257048_dp, &
             0.019927658890088462_dp, 0.05547349612734545_dp, 0.14003731505420994_dp, &
             0.328505211932211_dp, 0.7294474860966533_dp, 1.5725780732758836_dp, &
             3.587157358005245_dp]

      CASE (243)

         aw(:) = & ! a
            [7.37267894495509e-05_dp, 0.0007855766690852342_dp, 0.004097611065260232_dp, &
             0.01568891066720469_dp, 0.04998161435070802_dp, 0.14049764336646903_dp, &
             0.3602244356841656_dp, 0.8608605962688444_dp, 1.9538283459889698_dp, &
             4.357378217156472_dp, &
             ! w
             0.00023251538169232424_dp, 0.0014836646911268203_dp, 0.006008925962568857_dp, &
             0.0194152714557574_dp, 0.054436739383579465_dp, 0.13814632846470795_dp, &
             0.3253907902446524_dp, 0.724865397386506_dp, 1.5666612323006026_dp, &
             3.5803595542324804_dp]

      CASE (244)

         aw(:) = & ! a
            [6.924968347849166e-05_dp, 0.0007579413760384286_dp, 0.004002314214743724_dp, &
             0.015428199217226279_dp, 0.04936735277326966_dp, 0.13920627880981634_dp, &
             0.3577616046206043_dp, 0.8565758106159339_dp, 1.947034404522222_dp, &
             4.347548930323242_dp, &
             ! w
             0.0002208934821832957_dp, 0.0014446374339168884_dp, 0.005903963658001153_dp, &
             0.019174503377867066_dp, 0.05394773475305085_dp, 0.13725188732361046_dp, &
             0.3239144208051743_dp, 0.7226896593627993_dp, 1.5638487674530244_dp, &
             3.5771277827054706_dp]

      CASE (245)

         aw(:) = & ! a
            [6.844390138830545e-05_dp, 0.0007529515955878696_dp, 0.003985035599836499_dp, &
             0.015380799076269474_dp, 0.04925546526265801_dp, 0.1389707382811856_dp, &
             0.35731192893101843_dp, 0.8557928463574525_dp, 1.9457922045681697_dp, &
             4.345751081595511_dp, &
             ! w
             0.00021880174937623272_dp, 0.0014375664500890272_dp, 0.0058848743519528705_dp, &
             0.01913062408222225_dp, 0.053858487442246424_dp, 0.13708846821179418_dp, &
             0.3236444502032896_dp, 0.7222915374704968_dp, 1.5633339391232082_dp, &
             3.5765361929512145_dp]

      CASE (246)

         aw(:) = & ! a
            [0.030130988522655035_dp, 0.16068099452459852_dp, 0.40377505508058487_dp, &
             0.7751832494867651_dp, 1.3015684467174886_dp, 2.0256723917946107_dp, &
             3.014381717181716_dp, 4.371828907798731_dp, 6.265001798289237_dp, &
             8.99099924987998_dp, 13.245430010609466_dp, &
             ! w
             0.07754180309997503_dp, 0.1849123891694236_dp, 0.3038503047979931_dp, &
             0.44332561047408736_dp, 0.616437795661255_dp, 0.8427130343156188_dp, &
             1.151683744372129_dp, 1.5901310214760866_dp, 2.2427211746433686_dp, &
             3.3086113851271755_dp, 5.560937167681976_dp]

      CASE (247)

         aw(:) = & ! a
            [0.018121079712724322_dp, 0.09717503809053926_dp, 0.24679851729920788_dp, &
             0.4817566785490354_dp, 0.8282771166045614_dp, 1.3300696805608714_dp, &
             2.0567305766295165_dp, 3.1164004963694247_dp, 4.680507613564424_dp, &
             7.0492221261381145_dp, 10.911604422719243_dp, &
             ! w
             0.04669379968228601_dp, 0.11260338022623151_dp, 0.18903914404436575_dp, &
             0.28514829660319224_dp, 0.41504033333966167_dp, 0.5999483249149236_dp, &
             0.8710338480743368_dp, 1.2759883242857373_dp, 1.8993330622314406_dp, &
             2.9368796146680767_dp, 5.140078286417148_dp]

      CASE (248)

         aw(:) = & ! a
            [0.013332142168378771_dp, 0.07178446747588552_dp, 0.18375050791754996_dp, &
             0.3631906488145699_dp, 0.6356426510512829_dp, 1.0444692700724252_dp, &
             1.6592152251340613_dp, 2.58830483947198_dp, 4.004145202137294_dp, &
             6.2084129877437375_dp, 9.888554680124999_dp, &
             ! w
             0.03438538281941105_dp, 0.0836001053696773_dp, 0.1425876773323265_dp, &
             0.22044407827130794_dp, 0.33155187607849695_dp, 0.4974967407138458_dp, &
             0.7496946753853811_dp, 1.1363279712558223_dp, 1.7427500950293133_dp, &
             2.7645508616331735_dp, 4.944458579778126_dp]

      CASE (249)

         aw(:) = & ! a
            [0.010688453945637362_dp, 0.057739605989888636_dp, 0.14875435974316353_dp, &
             0.2970663897990964_dp, 0.5275519950057115_dp, 0.8829075520464834_dp, &
             1.4319631560494335_dp, 2.282648411744307_dp, 3.6076327856256674_dp, &
             5.7096768822342785_dp, 9.275873043842799_dp, &
             ! w
             0.027587389940205532_dp, 0.0675179313170176_dp, 0.11665978806309565_dp, &
             0.18400492341463712_dp, 0.283949354121628_dp, 0.4380335127182524_dp, &
             0.6776838734799814_dp, 1.0515697426467314_dp, 1.6460013694707742_dp, &
             2.6569503323667347_dp, 4.822200407234935_dp]

      CASE (250)

         aw(:) = & ! a
            [0.008990500158375417_dp, 0.04870417709230819_dp, 0.12617621226056372_dp, &
             0.2542357904458434_dp, 0.4571594292572616_dp, 0.7769092949171713_dp, &
             1.2814468596702209_dp, 2.078030149813022_dp, 3.3393616918465425_dp, &
             5.369054124614137_dp, 8.854288513719101_dp, &
             ! w
             0.023219556679335892_dp, 0.057151318173521565_dp, 0.09985528545071565_dp, &
             0.1602055619046999_dp, 0.25250354389485224_dp, 0.3981171244943777_dp, &
             0.6284424002596707_dp, 0.9926088267135148_dp, 1.577815572456583_dp, &
             2.580559279052842_dp, 4.735374139934867_dp]

      CASE (251)

         aw(:) = & ! a
            [0.00779855030728434_dp, 0.04235242011099923_dp, 0.11026525511118422_dp, &
             0.22394750615744496_dp, 0.40713580038316777_dp, 0.7010697302865118_dp, &
             1.1728396207014495_dp, 1.9290152408304664_dp, 3.1422391785388206_dp, &
             5.116820253028923_dp, 8.540207013979234_dp, &
             ! w
             0.02015234701752812_dp, 0.04985139674349512_dp, 0.08796626461982837_dp, &
             0.14325152757656873_dp, 0.22986665559520938_dp, 0.3689707967994121_dp, &
             0.5919302613549222_dp, 0.9482919217093115_dp, 1.5260458052374906_dp, &
             2.522236190314505_dp, 4.6690760135659035_dp]

      CASE (252)

         aw(:) = & ! a
            [0.006911123151973878_dp, 0.037617544377860225_dp, 0.09837885543161767_dp, &
             0.20124939360566238_dp, 0.3694783097483234_dp, 0.6436206330275096_dp, &
             1.0899407695335288_dp, 1.814350827026615_dp, 2.9893905748368823_dp, &
             4.919951938106212_dp, 8.293825536803077_dp, &
             ! w
             0.017868083742747738_dp, 0.04440159746301195_dp, 0.07905343422266112_dp, &
             0.13046141332449654_dp, 0.21262346265016102_dp, 0.34648736986481427_dp, &
             0.5633974185429599_dp, 0.9132730201360275_dp, 1.4848031156704544_dp, &
             2.4755646905106925_dp, 4.61602072885565_dp]

      CASE (253)

         aw(:) = & ! a
            [0.006222156846201917_dp, 0.03393745119365244_dp, 0.0891223549172806_dp, &
             0.1835224902391903_dp, 0.33994441123141017_dp, 0.5983039637414346_dp, &
             1.0240978685121114_dp, 1.7226238644716445_dp, 2.8662979670373296_dp, &
             4.760509608503989_dp, 8.093419785392712_dp, &
             ! w
             0.016094195143138396_dp, 0.04016017419248495_dp, 0.07209069721209424_dp, &
             0.12041126251744641_dp, 0.19895221819667774_dp, 0.3284599168601711_dp, &
             0.5402638660330097_dp, 0.8846145384436354_dp, 1.4508208171080412_dp, &
             2.436965669932821_dp, 4.572141876812747_dp]

      CASE (254)

         aw(:) = & ! a
            [0.005670214189250698_dp, 0.03098626556992178_dp, 0.08168603873992042_dp, &
             0.16924340664004847_dp, 0.3160607081649595_dp, 0.5614600806021925_dp, &
             0.9702293534113281_dp, 1.6470963177289164_dp, 2.764343118490741_dp, &
             4.627789942664766_dp, 7.925974405427083_dp, &
             ! w
             0.014672760181410385_dp, 0.036754698281782155_dp, 0.0664809481604872_dp, &
             0.11226966842914492_dp, 0.18778449751520476_dp, 0.31358447094093145_dp, &
             0.5209897018134803_dp, 0.860545213956893_dp, 1.4221134659936936_dp, &
             2.4042539239717673_dp, 4.53495589588831_dp]

      CASE (255)

         aw(:) = & ! a
            [0.00521711579001643_dp, 0.02856132097318766_dp, 0.07556567145631886_dp, &
             0.15746184069012953_dp, 0.296280949416954_dp, 0.5307942425023797_dp, &
             0.9251356226523106_dp, 1.5835045879490817_dp, 2.678045059132043_dp, &
             4.514954717187926_dp, 7.7831431411206164_dp, &
             ! w
             0.013505622891399446_dp, 0.033953336487602744_dp, 0.06185157183615044_dp, &
             0.10551618687024221_dp, 0.1784485164513013_dp, 0.3010351533892806_dp, &
             0.5045904598592531_dp, 0.8399224594117822_dp, 1.3973915223155493_dp, &
             2.376004757996431_dp, 4.502843252178717_dp]

      CASE (256)

         aw(:) = & ! a
            [0.003005695991918856_dp, 0.01668589503036214_dp, 0.045412158513915044_dp, &
             0.09886416435482326_dp, 0.19648852690779_dp, 0.3731861006919959_dp, &
             0.6885800478141613_dp, 1.2431291028181768_dp, 2.2076885241065622_dp, &
             3.8906895639880266_dp, 6.984079759654594_dp, &
             ! w
             0.007804654340042264_dp, 0.020179039473228334_dp, 0.03881773764993935_dp, &
             0.07123828209592745_dp, 0.12967996499604464_dp, 0.23339590799829762_dp, &
             0.41368187645494986_dp, 0.7229529177297216_dp, 1.2547944496604302_dp, &
             2.2115303605695886_dp, 4.315875214695467_dp]

      CASE (257)

         aw(:) = & ! a
            [0.0021727378031523377_dp, 0.012186596428899627_dp, 0.033865629464411746_dp, &
             0.07603076681008096_dp, 0.1565903606314177_dp, 0.308173763767149_dp, &
             0.58778690361397_dp, 1.0936113751271432_dp, 1.9954974609163636_dp, &
             3.6029405911917065_dp, 6.609943656627655_dp, &
             ! w
             0.005654329530181336_dp, 0.014923845603807212_dp, 0.029841061750614697_dp, &
             0.057386404242061054_dp, 0.1090165155437166_dp, 0.20338355405894115_dp, &
             0.3717369601140614_dp, 0.667264631875859_dp, 1.1853197469033225_dp, &
             2.1303494112152292_dp, 4.22357887008893_dp]

      CASE (258)

         aw(:) = & ! a
            [0.0017251108842001406_dp, 0.00975870366315658_dp, 0.027586318979494935_dp, &
             0.06344777081151343_dp, 0.13418908694113843_dp, 0.2708946251364909_dp, &
             0.5287828806322352_dp, 1.0044247033581155_dp, 1.8668735332583042_dp, &
             3.426263796963456_dp, 6.378103336168637_dp, &
             ! w
             0.004497626052151732_dp, 0.012074116805691842_dp, 0.024895278858103238_dp, &
             0.04954771657188593_dp, 0.09695687927319423_dp, 0.1853828690454563_dp, &
             0.34600844390650487_dp, 0.632482450036626_dp, 1.1413356299301325_dp, &
             2.078553347438098_dp, 4.1646774569138225_dp]

      CASE (259)

         aw(:) = & ! a
            [0.0014423138924763025_dp, 0.008219798589139462_dp, 0.02358071050996141_dp, &
             0.05533215139217533_dp, 0.11952504895081274_dp, 0.2460990065813065_dp, &
             0.48893784639873406_dp, 0.9433780372993746_dp, 1.7778182811830074_dp, &
             3.3028236886633358_dp, 6.21507446845864_dp, &
             ! w
             0.003766288741081957_dp, 0.010260753270741874_dp, 0.021706183150243887_dp, &
             0.04438379397805724_dp, 0.08883047269225096_dp, 0.1730191077007709_dp, &
             0.32805941902406205_dp, 0.6079089945236059_dp, 1.1099636389059035_dp, &
             2.041404419000145_dp, 4.122422598454609_dp]

      CASE (260)

         aw(:) = & ! a
            [0.0012460394897290097_dp, 0.007148780139643052_dp, 0.020777498809389484_dp, &
             0.04959830683148656_dp, 0.10903590899972522_dp, 0.2281340299783253_dp, &
             0.45972326499025856_dp, 0.8981462859609084_dp, 1.7112485746366775_dp, &
             3.209906633515137_dp, 6.091753199276357_dp, &
             ! w
             0.003258376259568787_dp, 0.008994542514435296_dp, 0.019453474665386623_dp, &
             0.04067031008258711_dp, 0.08288224484454552_dp, 0.1638367468622256_dp, &
             0.3145704196522124_dp, 0.5892630713497109_dp, 1.0859847078241962_dp, &
             2.0128882604504184_dp, 4.089979761465004_dp]

      CASE (261)

         aw(:) = & ! a
            [0.0011011368472607192_dp, 0.006356190430447665_dp, 0.01869276431717682_dp, &
             0.04529792991109369_dp, 0.10108540562921922_dp, 0.21437076457370233_dp, &
             0.43712168928244916_dp, 0.8628534639195983_dp, 1.6589340653420583_dp, &
             3.1364760958078275_dp, 5.993909833379412_dp, &
             ! w
             0.0028831912480045086_dp, 0.008054789696030526_dp, 0.017764191867862825_dp, &
             0.037842624769486494_dp, 0.07828705211822772_dp, 0.15665995872237118_dp, &
             0.303927329683743_dp, 0.5744369494539767_dp, 1.0668055886728782_dp, &
             1.9900009359991648_dp, 4.063935518109033_dp]

      CASE (262)

         aw(:) = & ! a
            [0.0009893735940206593_dp, 0.005743567250685712_dp, 0.01707414274036477_dp, &
             0.04193357594617004_dp, 0.09480761228138741_dp, 0.20340337145131307_dp, &
             0.4189621037780273_dp, 0.8342932954015233_dp, 1.6163458927661858_dp, &
             3.076417755654109_dp, 5.913622018341871_dp, &
             ! w
             0.002593667981520902_dp, 0.007326535559007862_dp, 0.016442723854887428_dp, &
             0.03560084983612468_dp, 0.07459962007051006_dp, 0.15084498084425932_dp, &
             0.2952357646445964_dp, 0.5622513167694062_dp, 1.050964696914549_dp, &
             1.9710423046491485_dp, 4.042357758909767_dp]

      CASE (263)

         aw(:) = & ! a
            [0.0009003070544349657_dp, 0.0052544185410531965_dp, 0.01577639480864869_dp, &
             0.0392174231619656_dp, 0.08969753873619522_dp, 0.19440455604236653_dp, &
             0.40395550124062374_dp, 0.8105465859229404_dp, 1.5807541895705421_dp, &
             3.026025363731146_dp, 5.846068063666509_dp, &
             ! w
             0.0023628382210883927_dp, 0.006743690025702634_dp, 0.015375912461056013_dp, &
             0.033769523602153365_dp, 0.07155586763212222_dp, 0.1460054304956694_dp, &
             0.2879536312425964_dp, 0.5519856851543958_dp, 1.037563739502719_dp, &
             1.9549638627492085_dp, 4.0240548291653635_dp]

      CASE (264)

         aw(:) = & ! a
            [0.0008275060405979829_dp, 0.0048538992490044595_dp, 0.014709683757547234_dp, &
             0.03697056590495137_dp, 0.08543906157163482_dp, 0.186852311634262_dp, &
             0.39128223664162726_dp, 0.7903843504629097_dp, 1.550400192178853_dp, &
             2.98289946493133_dp, 5.788115495076612_dp, &
             ! w
             0.0021740872264102923_dp, 0.0062654140734613555_dp, 0.014493434412544949_dp, &
             0.03223848021628171_dp, 0.06898803495334238_dp, 0.14189336224047105_dp, &
             0.28173013348362336_dp, 0.5431705675223087_dp, 1.0260142895533086_dp, &
             1.9410767529159068_dp, 4.0082438733806685_dp]

      CASE (265)

         aw(:) = & ! a
            [0.00047633867638452546_dp, 0.002909964655167804_dp, 0.009457493514563445_dp, &
             0.025652164792170237_dp, 0.06344016008115329_dp, 0.14691472784904946_dp, &
             0.32287836144553367_dp, 0.6796431373233776_dp, 1.3812552178932358_dp, &
             2.739865543352038_dp, 5.458966455736076_dp, &
             ! w
             0.0012623682047825586_dp, 0.003925652675398484_dp, 0.01004658375330077_dp, &
             0.024242501459288118_dp, 0.05518051512738748_dp, 0.11927023809408377_dp, &
             0.2468429160149552_dp, 0.49298608655375087_dp, 0.9594751280677348_dp, &
             1.8604943398396532_dp, 3.916442326880172_dp]

      CASE (266)

         aw(:) = & ! a
            [0.00034586004935876314_dp, 0.002180068448344695_dp, 0.0074338002448057355_dp, &
             0.021122671958534404_dp, 0.05429114786508466_dp, 0.1297318847429076_dp, &
             0.2925888019277062_dp, 0.6294124318896298_dp, 1.3030115311584665_dp, &
             2.6257268942687264_dp, 5.302769236620561_dp, &
             ! w
             0.000922883899415914_dp, 0.003034681058102913_dp, 0.008264552339266687_dp, &
             0.020862922814715763_dp, 0.04910404525683061_dp, 0.10899880862103302_dp, &
             0.2305969093077354_dp, 0.46912654138356547_dp, 0.9273310285024502_dp, &
             1.8211892895434076_dp, 3.871623662017366_dp]

      CASE (267)

         aw(:) = & ! a
            [0.0002761320530126241_dp, 0.0017871801758091305_dp, 0.0063237062667975815_dp, &
             0.018574957233231215_dp, 0.04902279026213492_dp, 0.11963878645348468_dp, &
             0.27450111774609626_dp, 0.5990053817796526_dp, 1.2551208194212886_dp, &
             2.5552713219979286_dp, 5.205792916917917_dp, &
             ! w
             0.0007412355914717759_dp, 0.0025500918289050423_dp, 0.007260652343017782_dp, &
             0.018898305189684796_dp, 0.04549056619090046_dp, 0.10278288849115716_dp, &
             0.22062449477394547_dp, 0.4543089394242019_dp, 0.9071881653619919_dp, &
             1.7964239050734692_dp, 3.843367561798431_dp]

      CASE (268)

         aw(:) = & ! a
            [0.0002322146395909956_dp, 0.0015383078777110473_dp, 0.0056097155499255456_dp, &
             0.01690524019317347_dp, 0.0455116261284354_dp, 0.11281827094933433_dp, &
             0.2621385842803853_dp, 0.5780283644600492_dp, 1.2218328763756419_dp, &
             2.5060164008391586_dp, 5.137732154838399_dp, &
             ! w
             0.0006267332758852317_dp, 0.002240499431309514_dp, 0.006601770852296073_dp, &
             0.017580377812750667_dp, 0.043028675267022684_dp, 0.09849704584143885_dp, &
             0.21368155960769558_dp, 0.4439105540875763_dp, 0.892965923250978_dp, &
             1.778872245964063_dp, 3.8233333460474905_dp]

      CASE (269)

         aw(:) = & ! a
            [0.0002017946813401587_dp, 0.0013650957954943518_dp, 0.005106343053963922_dp, &
             0.015710315576460424_dp, 0.04296620994180534_dp, 0.1078215546982082_dp, &
             0.2530041706297298_dp, 0.5624205865907776_dp, 1.1969260235286665_dp, &
             2.469004640142788_dp, 5.086440738844717_dp, &
             ! w
             0.0005473782929676091_dp, 0.0020234317344929253_dp, 0.006129638812465468_dp, &
             0.01662026786477423_dp, 0.041214281305774245_dp, 0.09531008242567778_dp, &
             0.2084812698209121_dp, 0.4360759297357786_dp, 0.8822012112666164_dp, &
             1.7655503222027713_dp, 3.8081221072884683_dp]

      CASE (270)

         aw(:) = & ! a
            [0.00017937019485878274_dp, 0.0012368771763232084_dp, 0.004729537850198795_dp, &
             0.014804694996910868_dp, 0.04101686411438877_dp, 0.10396278870461453_dp, &
             0.24590220936842966_dp, 0.550218803660413_dp, 1.1773684055533487_dp, &
             2.4398440817448357_dp, 5.045937842151792_dp, &
             ! w
             0.0004888589654993824_dp, 0.0018616953743137686_dp, 0.005771405969229113_dp, &
             0.01588216091220091_dp, 0.03980659796626245_dp, 0.09282001291518131_dp, &
             0.20439489927666582_dp, 0.42989081733198803_dp, 0.8736723641609264_dp, &
             1.7549721269953933_dp, 3.7960404785939112_dp]

      CASE (271)

         aw(:) = & ! a
            [0.0001620957580603921_dp, 0.0011377383639357614_dp, 0.004435299760668608_dp, &
             0.014090043792072175_dp, 0.039465191380223995_dp, 0.10086997697308225_dp, &
             0.24017835055121126_dp, 0.5403404704563021_dp, 1.161477862487281_dp, &
             2.416086212663772_dp, 5.012878062561856_dp, &
             ! w
             0.00044376817842623015_dp, 0.0017359012756834168_dp, 0.005488426431554195_dp, &
             0.01529279060151841_dp, 0.03867412614923515_dp, 0.09080518586078573_dp, &
             0.20107298702184065_dp, 0.42484366957051944_dp, 0.8666922993395992_dp, &
             1.7462992361734815_dp, 3.786132649859132_dp]

      CASE (272)

         aw(:) = & ! a
            [0.00014834488073842516_dp, 0.0010585558771085244_dp, 0.0041982030737258555_dp, &
             0.013508914964261778_dp, 0.03819408501640941_dp, 0.09832159671271519_dp, &
             0.23544005124366202_dp, 0.532132194386579_dp, 1.148233976425562_dp, &
             2.396239934234456_dp, 4.985218745726289_dp, &
             ! w
             0.000407869187231851_dp, 0.0016348875607537822_dp, 0.005258108118650664_dp, &
             0.014808725799872507_dp, 0.037738122389550766_dp, 0.08913182093553668_dp, &
             0.19830326562500195_dp, 0.4206221039756088_dp, 0.8608396444681432_dp, &
             1.7390161743430637_dp, 3.7778109552097634_dp]

      CASE (273)

         aw(:) = & ! a
            [0.00013711729501785012_dp, 0.0009937039401902875_dp, 0.0040024563087556735_dp, &
             0.013025287217430374_dp, 0.03712946550086021_dp, 0.09617646982553832_dp, &
             0.2314355666053776_dp, 0.5251727289213195_dp, 1.1369760622900191_dp, &
             2.3793366605690114_dp, 4.961630042198732_dp, &
             ! w
             0.0003785550695185705_dp, 0.0015517451432116936_dp, 0.005066275361096803_dp, &
             0.014402394221929217_dp, 0.03694817371211211_dp, 0.08771369686719463_dp, &
             0.19594815380436012_dp, 0.4170227140801809_dp, 0.855839089207469_dp, &
             1.7327854324885192_dp, 3.770690478597186_dp]

      CASE (274)

         aw(:) = & ! a
            [8.322776802917285e-05_dp, 0.0006789846232488049_dp, 0.0030262161445555554_dp, &
             0.010550805930632207_dp, 0.03157256088269674_dp, 0.08480476655604943_dp, &
             0.20994413805118853_dp, 0.4874504767676322_dp, 1.075470439657678_dp, &
             2.2864325945367705_dp, 4.831457842787824_dp, &
             ! w
             0.00023787241240500552_dp, 0.0011410910795990356_dp, 0.004082078710881861_dp, &
             0.012267293008636779_dp, 0.03272778769529988_dp, 0.0800397386211778_dp, &
             0.18307197907468373_dp, 0.39717840089040185_dp, 0.828091222345798_dp, &
             1.6980729553696199_dp, 3.731000175331714_dp]

      CASE (275)

         aw(:) = & ! a
            [6.348423252433597e-05_dp, 0.0005614631095692756_dp, 0.0026458704217153886_dp, &
             0.009551452042587599_dp, 0.029267376085073665_dp, 0.07999047048687996_dp, &
             0.2006996270874012_dp, 0.47101726159173085_dp, 1.0484057106225948_dp, &
             2.245240143304216_dp, 4.773448082651668_dp, &
             ! w
             0.00018637903410584144_dp, 0.0009832196596205357_dp, 0.0036830423775804813_dp, &
             0.011374078549377407_dp, 0.030923673307909894_dp, 0.0767048805977973_dp, &
             0.1774024737754178_dp, 0.3883477284904973_dp, 0.8156424850848444_dp, &
             1.6824209713496043_dp, 3.71309132927725_dp]

      CASE (276)

         aw(:) = & ! a
            [5.314101534529174e-05_dp, 0.0004990776322345713_dp, 0.002438805589025952_dp, &
             0.008996731690895392_dp, 0.027969831614632127_dp, 0.07725217190606437_dp, &
             0.19539875448635483_dp, 0.4615336980048271_dp, 1.0327073988310767_dp, &
             2.2212559751175838_dp, 4.739586108324459_dp, &
             ! w
             0.000159425762683865_dp, 0.0008978503681720132_dp, 0.003461078195377784_dp, &
             0.010869229496105194_dp, 0.029892678888969553_dp, 0.0747830911785287_dp, &
             0.1741134978491549_dp, 0.3831974542878289_dp, 0.8083521459057916_dp, &
             1.67323127405907_dp, 3.7025727660653884_dp]

      CASE (277)

         aw(:) = & ! a
            [4.681778510201952e-05_dp, 0.0004605499591683695_dp, 0.0023087081448212162_dp, &
             0.00864381633616296_dp, 0.027137006340929144_dp, 0.07548306313644564_dp, &
             0.19195672251086449_dp, 0.45535106823555815_dp, 1.0224408958727922_dp, &
             2.2055334225191645_dp, 4.717353301417677_dp, &
             ! w
             0.00014295682470358455_dp, 0.0008444325077663087_dp, 0.003319674228999414_dp, &
             0.010544379264967827_dp, 0.029224680527137968_dp, 0.073531409505792_dp, &
             0.17196246801986348_dp, 0.3798178817818088_dp, 0.803556031079179_dp, &
             1.6671760315071371_dp, 3.695640346653618_dp]

      CASE (278)

         aw(:) = & ! a
            [4.2632368619756146e-05_dp, 0.0004348420268111736_dp, 0.002220813641088391_dp, &
             0.008403298331247156_dp, 0.026565961593869418_dp, 0.07426458692330412_dp, &
             0.18957782609900978_dp, 0.4510664232845759_dp, 1.015310830149632_dp, &
             2.1945965730454664_dp, 4.701871305613612_dp, &
             ! w
             0.0001320587393026254_dp, 0.0008084406119011537_dp, 0.003223216313602195_dp, &
             0.010321263666998209_dp, 0.02876370960494463_dp, 0.07266456142158251_dp, &
             0.17046857081193098_dp, 0.3774654487154653_dp, 0.8002117847669223_dp, &
             1.662949265201805_dp, 3.6908005282052048_dp]

      CASE (279)

         aw(:) = & ! a
            [3.975563513065617e-05_dp, 0.00041705908651804344_dp, 0.0021594527670394047_dp, &
             0.00823433029790558_dp, 0.026163048204589357_dp, 0.07340212074387423_dp, &
             0.1878898612124454_dp, 0.4480203556554623_dp, 1.0102341915301305_dp, &
             2.186800620586824_dp, 4.690827158504661_dp, &
             ! w
             0.0001245689706270913_dp, 0.0007833608835555495_dp, 0.0031554085100756854_dp, &
             0.010163653271231704_dp, 0.028436982689144606_dp, 0.07204859873545982_dp, &
             0.16940491540987362_dp, 0.3757878335510951_dp, 0.7978239360570349_dp, &
             1.6599289698604585_dp, 3.6873417688751546_dp]

      CASE (280)

         aw(:) = & ! a
            [3.777242526030486e-05_dp, 0.00040473828737865177_dp, 0.0021166501617549322_dp, &
             0.008115929324496608_dp, 0.025879831930912478_dp, 0.07279448736437288_dp, &
             0.18669855364741078_dp, 0.44586758164032736_dp, 1.0066424418124102_dp, &
             2.1812804438583977_dp, 4.6830027640905465_dp, &
             ! w
             0.00011940550723828523_dp, 0.0007658891868451849_dp, 0.0031078712414421917_dp, &
             0.01005277280986663_dp, 0.028206572662436515_dp, 0.07161342895673377_dp, &
             0.16865238213255387_dp, 0.3745995653998444_dp, 0.7961311097004466_dp, &
             1.6577866046054526_dp, 3.684888195236902_dp]

      CASE (281)

         aw(:) = & ! a
            [3.646338596501122e-05_dp, 0.0003965757940554431_dp, 0.0020881565457395784_dp, &
             0.00803685787822609_dp, 0.02569027722232841_dp, 0.07238714938904642_dp, &
             0.18589895912279036_dp, 0.4444212696436607_dp, 1.0042275510594376_dp, &
             2.177566882833669_dp, 4.677737119450149_dp, &
             ! w
             0.00011599715550524298_dp, 0.0007542687198627458_dp, 0.0030761142316467717_dp, &
             0.009978518315797633_dp, 0.02805200984838512_dp, 0.07132113851027284_dp, &
             0.16814642492721013_dp, 0.3738000108488575_dp, 0.7949913534217391_dp, &
             1.6563436290123126_dp, 3.6832355145521425_dp]

      CASE (282)

         aw(:) = & ! a
            [3.571906216558647e-05_dp, 0.0003919230767898415_dp, 0.0020718641379196025_dp, &
             0.007991553201685386_dp, 0.025581519027222978_dp, 0.07215319936467512_dp, &
             0.1854393637310578_dp, 0.4435894412459486_dp, 1.0028379881400717_dp, &
             2.175429273739434_dp, 4.674705369175517_dp, &
             ! w
             0.00011405903698367499_dp, 0.0007476279138218427_dp, 0.0030579149948176663_dp, &
             0.00993589863751133_dp, 0.02796320103187524_dp, 0.07115305887788502_dp, &
             0.16785529241781696_dp, 0.37333970513687387_dp, 0.7943349361889072_dp, &
             1.6555123760808559_dp, 3.6822834226183354_dp]

      CASE (283)

         aw(:) = & ! a
            [3.550348465924681e-05_dp, 0.0003905739575885531_dp, 0.0020671330057150483_dp, &
             0.00797838466010411_dp, 0.02554988645008824_dp, 0.07208512275482128_dp, &
             0.18530557881445084_dp, 0.4433472301960455_dp, 1.0024332764988158_dp, &
             2.174806566306048_dp, 4.67382205788286_dp, &
             ! w
             0.00011349768781191745_dp, 0.0007457000167406751_dp, 0.00305262454150563_dp, &
             0.009923500448762425_dp, 0.02793735378855931_dp, 0.07110412222577314_dp, &
             0.16777050240847788_dp, 0.37320560837397676_dp, 0.794143662646913_dp, &
             1.6552701124330567_dp, 3.6820059233978353_dp]

      CASE (284)

         aw(:) = & ! a
            [0.02767312926220816_dp, 0.14728661310153135_dp, 0.3687581752417785_dp, &
             0.7039332404419827_dp, 1.172365441249489_dp, 1.8045683154882166_dp, &
             2.646952430489296_dp, 3.7693192226293997_dp, 5.27735243988307_dp, &
             7.338174973596355_dp, 10.248846166739662_dp, 14.70999228737189_dp, &
             ! w
             0.07118467663536959_dp, 0.16908953223627346_dp, 0.2757926387649255_dp, &
             0.3977636094386997_dp, 0.54410574348806_dp, 0.7279147631051287_dp, &
             0.9682939484624312_dp, 1.2936786640404163_dp, 1.7493131133110282_dp, &
             2.418770402622885_dp, 3.5024366140575562_dp, 5.785552407989267_dp]

      CASE (285)

         aw(:) = & ! a
            [0.016621986833340294_dp, 0.08887641901396574_dp, 0.22446216418490886_dp, &
             0.43430080891188433_dp, 0.7372692134906315_dp, 1.1640767884289862_dp, &
             1.76265927443391_dp, 2.6056037700656303_dp, 3.802129378655715_dp, &
             5.522886860404714_dp, 8.066542148084695_dp, 12.124597250003708_dp, &
             ! w
             0.04280253466281205_dp, 0.10261564373695731_dp, 0.1703341135440457_dp, &
             0.2524250366464258_dp, 0.35854064076296327_dp, 0.5029449307365429_dp, &
             0.7061343335450083_dp, 0.9975940043005186_dp, 1.4229943025732463_dp, &
             2.0653230769823834_dp, 3.120898077906089_dp, 5.351475189683586_dp]

      CASE (286)

         aw(:) = & ! a
            [0.012217590485622118_dp, 0.06554383522960264_dp, 0.16659237203973748_dp, &
             0.32558270571482617_dp, 0.5606935393689713_dp, 0.9021601691438875_dp, &
             1.397705445731634_dp, 2.119903667111354_dp, 3.1783136457891827_dp, &
             4.744534993463762_dp, 7.1183009631362095_dp, 10.988266599537688_dp, &
             ! w
             0.031484785005288264_dp, 0.07598807870443566_dp, 0.12776669458080814_dp, &
             0.19317510324206744_dp, 0.2819959712025566_dp, 0.4087743922620433_dp, &
             0.5942255934432832_dp, 0.8680605358456985_dp, 1.2764587208952323_dp, &
             1.903008836197448_dp, 2.943222654358237_dp, 5.149016836830881_dp]

      CASE (287)

         aw(:) = & ! a
            [0.009787250213222088_dp, 0.05264669751460476_dp, 0.1345082558306065_dp, &
             0.2650565360294569_dp, 0.46186942457185204_dp, 0.7545813463702145_dp, &
             1.190270456056405_dp, 1.8408744447358316_dp, 2.8156533824155634_dp, &
             4.286609772539828_dp, 6.554364102152325_dp, 10.306420412806078_dp, &
             ! w
             0.025237117064509148_dp, 0.061238802880255835_dp, 0.10405085556044878_dp, &
             0.15990524330983047_dp, 0.23857118674932304_dp, 0.35458212495967756_dp, &
             0.5285932206532509_dp, 0.7904471850597391_dp, 1.1868782571026184_dp, &
             1.8022353684717614_dp, 2.831945778444912_dp, 5.022182355440811_dp]

      CASE (288)

         aw(:) = & ! a
            [0.008226923180261456_dp, 0.04435476571968886_dp, 0.11382938114522073_dp, &
             0.225910987843479_dp, 0.3976625203871757_dp, 0.6581133163982475_dp, &
             1.0535979897201335_dp, 1.6552845367932827_dp, 2.571982747933977_dp, &
             3.9759001427243166_dp, 6.1684054473566166_dp, 9.836513308538686_dp, &
             ! w
             0.021224651302760023_dp, 0.05173983688640228_dp, 0.08870403403015269_dp, &
             0.13823277713967072_dp, 0.21002162997077486_dp, 0.31848388786257004_dp, &
             0.48414783544765566_dp, 0.736976098003317_dp, 1.1242221404269488_dp, &
             1.730958028955162_dp, 2.75276052749726_dp, 4.931935957863172_dp]

      CASE (289)

         aw(:) = & ! a
            [0.0071319631843329125_dp, 0.0385288520026887_dp, 0.09926949796073953_dp, &
             0.19826580289961176_dp, 0.35213340700048434_dp, 0.5893267643752341_dp, &
             0.955441720557511_dp, 1.5208802148863778_dp, 2.3939769277419587_dp, &
             3.7470487790881046_dp, 5.882102309653825_dp, 9.485972661585034_dp, &
             ! w
             0.018408098383937928_dp, 0.045056081276050215_dp, 0.07786113143898855_dp, &
             0.122831059386808_dp, 0.18956093445734312_dp, 0.2923043033618055_dp, &
             0.4514544295919487_dp, 0.6970886541216346_dp, 1.076924350663281_dp, &
             1.6766859343620684_dp, 2.6921885124990736_dp, 4.862917181590072_dp]

      CASE (290)

         aw(:) = & ! a
            [0.006317001356062951_dp, 0.034188087739164805_dp, 0.08840086940238588_dp, &
             0.1775737848854597_dp, 0.31792844578021495_dp, 0.5373835811949759_dp, &
             0.8808348910539449_dp, 1.4179620136198705_dp, 2.256637850632205_dp, &
             3.5692344888793874_dp, 5.6583087139239066_dp, 9.210677590301046_dp, &
             ! w
             0.01631125461774546_dp, 0.04006973564565049_dp, 0.0697424933664369_dp, &
             0.1112380896871292_dp, 0.1740396434997305_dp, 0.2722299633243372_dp, &
             0.4260759427480407_dp, 0.6657626611697179_dp, 1.0394169828252717_dp, &
             1.6333462645016368_dp, 2.6436380689380594_dp, 4.807608641074463_dp]

      CASE (291)

         aw(:) = & ! a
            [0.005684475858101966_dp, 0.030815812656619954_dp, 0.07994292383209656_dp, &
             0.16143190116044875_dp, 0.291152971588611_dp, 0.49652875773662286_dp, &
             0.8218028957748642_dp, 1.335985987778696_dp, 2.1465138972910536_dp, &
             3.425780060211239_dp, 5.476821513760035_dp, 8.986526926934294_dp, &
             ! w
             0.014683440378687724_dp, 0.0361914353823823_dp, 0.06340726558223926_dp, &
             0.10214785527990394_dp, 0.1617806182160849_dp, 0.2562191165902424_dp, &
             0.40561646277705926_dp, 0.6402570726581068_dp, 1.0086292613465813_dp, &
             1.5975625195441152_dp, 2.6034273539351465_dp, 4.761810257792545_dp]

      CASE (292)

         aw(:) = & ! a
            [0.005177883653409764_dp, 0.028112585847666887_dp, 0.0731525476894913_dp, &
             0.14844332176028877_dp, 0.26953853518176357_dp, 0.4634018779090883_dp, &
             0.7736733807857686_dp, 1.2687470578926419_dp, 2.055648798399003_dp, &
             3.3067709916274914_dp, 5.3255742673945_dp, 8.79907055107436_dp, &
             ! w
             0.013379449457657585_dp, 0.03307930263759621_dp, 0.0583083622425076_dp, &
             0.09479860562631726_dp, 0.1518020216293596_dp, 0.24306983539405944_dp, &
             0.38865380693947577_dp, 0.6189288727611131_dp, 0.9827038031723337_dp, &
             1.5672784492918068_dp, 2.5693057474657723_dp, 4.722954399845834_dp]

      CASE (293)

         aw(:) = & ! a
            [0.004762117357297583_dp, 0.025892231189378046_dp, 0.06756718406524467_dp, &
             0.13773723574799498_dp, 0.25166827727237495_dp, 0.4358989325752496_dp, &
             0.7335111188980595_dp, 1.2123305594792853_dp, 1.978999373338425_dp, &
             3.2058935052153563_dp, 5.196850709872084_dp, 8.639036294627692_dp, &
             ! w
             0.01230904592078374_dp, 0.03052060654576384_dp, 0.05410461513536041_dp, &
             0.08871396312775404_dp, 0.14348763916152588_dp, 0.23202358585890717_dp, &
             0.37428328477056766_dp, 0.6007236370303556_dp, 0.9604392175073997_dp, &
             1.5411563493400489_dp, 2.5398045743466136_dp, 4.689365635807877_dp]

      CASE (294)

         aw(:) = & ! a
            [0.002734772548216079_dp, 0.015033916035287382_dp, 0.040110919550103284_dp, &
             0.08469311373235947_dp, 0.16209254870784667_dp, 0.29585299509549345_dp, &
             0.525182239750685_dp, 0.9139677614786229_dp, 1.5660334617940184_dp, &
             2.6532976247909676_dp, 4.481951863229659_dp, 7.740949632915444_dp, &
             ! w
             0.0070860134344731774_dp, 0.01796384745053746_dp, 0.03326483860847221_dp, &
             0.05805971178818426_dp, 0.10057714481103108_dp, 0.17333339296520517_dp, &
             0.2957310433985091_dp, 0.49871603403504905_dp, 0.83315161426675_dp, &
             1.3896105342194978_dp, 2.367294950712571_dp, 4.4930548591707975_dp]

      CASE (295)

         aw(:) = & ! a
            [0.0019724053061639557_dp, 0.010930068316984159_dp, 0.029639159205617847_dp, &
             0.06417083780725863_dp, 0.126692767641025_dp, 0.2389728369268008_dp, &
             0.43796655665209683_dp, 0.7852367571901353_dp, 1.3828078771058658_dp, &
             2.402073550428113_dp, 4.150438718514592_dp, 7.318326982604391_dp, &
             ! w
             0.005119598334286014_dp, 0.01318928537450709_dp, 0.02519760960181168_dp, &
             0.045838577075039494_dp, 0.08274424240086012_dp, 0.14782334419627308_dp, &
             0.26017272629411214_dp, 0.45093045744030513_dp, 0.7718498864664847_dp, &
             1.3151307967790185_dp, 2.281566539053284_dp, 4.395551521193392_dp]

      CASE (296)

         aw(:) = & ! a
            [0.00156320176896099_dp, 0.008719429282445866_dp, 0.023961109876658067_dp, &
             0.0529226964629682_dp, 0.1069822237653646_dp, 0.20669168458800416_dp, &
             0.38747343994181427_dp, 0.7092770597551052_dp, 1.272817610450666_dp, &
             2.249019230469243_dp, 3.94605690383392_dp, 7.055504983503552_dp, &
             ! w
             0.00406322999100781_dp, 0.0106063465272963_dp, 0.020775731297102012_dp, &
             0.038990282127968_dp, 0.07246291949761123_dp, 0.13270410923150439_dp, &
             0.23859287372275573_dp, 0.4213494413144101_dp, 0.7332827906426044_dp, &
             1.2677049097445012_dp, 2.2266091411905946_dp, 4.333056111653106_dp]

      CASE (297)

         aw(:) = & ! a
            [0.0013049277071097104_dp, 0.007320172278045236_dp, 0.020347839561300884_dp, &
             0.04570046774727458_dp, 0.09416383067198496_dp, 0.1853856908569084_dp, &
             0.35364802698136605_dp, 0.6576800518814916_dp, 1.1971736850908161_dp, &
             2.1426416636948535_dp, 3.802802086222116_dp, 6.8701597226829625_dp, &
             ! w
             0.003396042287459571_dp, 0.008965861319430082_dp, 0.017936773769428257_dp, &
             0.03451341490836681_dp, 0.06559528308063958_dp, 0.12240405308635692_dp, &
             0.22364618721018073_dp, 0.4005751196598172_dp, 0.7058874382446316_dp, &
             1.2337273932034403_dp, 2.1870446405759956_dp, 4.288066679809135_dp]

      CASE (298)

         aw(:) = & ! a
            [0.0011258192253550162_dp, 0.006347475918930572_dp, 0.017824504943091377_dp, &
             0.040617411866614016_dp, 0.08504422686915565_dp, 0.17004360275241995_dp, &
             0.3290007972164119_dp, 0.6196717314799713_dp, 1.1409124547374707_dp, &
             2.0628733271753124_dp, 3.6946810222503697_dp, 6.72961535723366_dp, &
             ! w
             0.0029330978418186607_dp, 0.007822193312381915_dp, 0.015938966070157382_dp, &
             0.03131422784213807_dp, 0.06060218290721298_dp, 0.11480104916419112_dp, &
             0.21247322229705415_dp, 0.38488070274759967_dp, 0.6850094387100483_dp, &
             1.2076616461896144_dp, 2.1565782421799327_dp, 4.253422306083899_dp]

      CASE (299)

         aw(:) = & ! a
            [0.000993683472150266_dp, 0.005628371952263191_dp, 0.01595141882853389_dp, &
             0.036817978762499176_dp, 0.07816337835263072_dp, 0.15834925822365145_dp, &
             0.3100284584585824_dp, 0.5901523820112894_dp, 1.096873044119201_dp, &
             2.0000185378949156_dp, 3.6090366250071586_dp, 6.617868081069637_dp, &
             ! w
             0.002591396443098623_dp, 0.006974568747265797_dp, 0.014445878248169917_dp, &
             0.028891038917439545_dp, 0.05676567335226572_dp, 0.10888726021450111_dp, &
             0.20369414604773922_dp, 0.3724435069704642_dp, 0.6683474147663148_dp, &
             1.1867479034725879_dp, 2.1320583937567417_dp, 4.22553878529586_dp]

      CASE (300)

         aw(:) = & ! a
            [0.0008918316487520122_dp, 0.00507304241614829_dp, 0.014499582701534557_dp, &
             0.03385446798099944_dp, 0.0727516916962805_dp, 0.14907043771770526_dp, &
             0.2948485991588153_dp, 0.5663551610182705_dp, 1.0611357875910474_dp, &
             1.9487292245922745_dp, 3.53884368903063_dp, 6.525994088567763_dp, &
             ! w
             0.002327893648941141_dp, 0.0063185218754966575_dp, 0.013281445340449002_dp, &
             0.02697866054417605_dp, 0.05370079969460482_dp, 0.10411442500731535_dp, &
             0.19654885277745274_dp, 0.362248823900893_dp, 0.6546090522727667_dp, &
             1.1694264755675832_dp, 2.1116976718957297_dp, 4.202383696636389_dp]

      CASE (301)

         aw(:) = & ! a
            [0.0008107099970118849_dp, 0.004629993113525699_dp, 0.013337360969949264_dp, &
             0.03146842376005929_dp, 0.06836197307409192_dp, 0.1414853034908997_dp, &
             0.2823490730713297_dp, 0.5466319853569387_dp, 1.031348518967244_dp, &
             1.9057753817936651_dp, 3.4798372032583975_dp, 6.448554963488439_dp, &
             ! w
             0.0021179393643737904_dp, 0.0057940548940096075_dp, 0.012344024954260667_dp, &
             0.02542260545143525_dp, 0.051180511115307005_dp, 0.10015525589131392_dp, &
             0.1905788931856401_dp, 0.3536793060741089_dp, 0.6430024712651331_dp, &
             1.1547365191625083_dp, 2.0943916499248556_dp, 4.1827013514861955_dp]

      CASE (302)

         aw(:) = & ! a
            [0.0007444374958604835_dp, 0.004267483163526778_dp, 0.01238341904919077_dp, &
             0.029499518285170032_dp, 0.06471515255300529_dp, 0.13514013046615872_dp, &
             0.2718256002117981_dp, 0.5299318074134547_dp, 1.0060016003098822_dp, &
             1.869072580692834_dp, 3.42925262833547_dp, 6.382014136726755_dp, &
             ! w
             0.001946354897725961_dp, 0.005364126523625145_dp, 0.011570567870099505_dp, &
             0.02412623896654783_dp, 0.04906122192210912_dp, 0.09680064117270458_dp, &
             0.1854888010464085_dp, 0.3463341660098126_dp, 0.6330104554036071_dp, &
             1.1420476000340243_dp, 2.0794138334222345_dp, 4.1656658270973_dp]

      CASE (303)

         aw(:) = & ! a
            [0.0004253461425896496_dp, 0.0025124573159036588_dp, 0.007711033123349251_dp, &
             0.019666768891738344_dp, 0.04606996645974101_dp, 0.10193456174953619_dp, &
             0.21557145284519655_dp, 0.43896854975244937_dp, 0.8656822115642425_dp, &
             1.6631068381675098_dp, 3.142330171348081_dp, 6.001714225275986_dp, &
             ! w
             0.001119157921380432_dp, 0.0032686761628313806_dp, 0.007708739991268936_dp, &
             0.017432462565101902_dp, 0.03778046628285069_dp, 0.07850101026832947_dp, &
             0.15715272263412114_dp, 0.3047343816013479_dp, 0.5755987681276237_dp, &
             1.0683278448896232_dp, 1.991827320223211_dp, 4.066017442783831_dp]

      CASE (304)

         aw(:) = & ! a
            [0.0003071487155038261_dp, 0.0018562367167545763_dp, 0.005926962713218073_dp, &
             0.01578531225429062_dp, 0.038431367060375164_dp, 0.08784952530017573_dp, &
             0.19096824701843967_dp, 0.3981208018572958_dp, 0.8012392220997078_dp, &
             1.566739379849031_dp, 3.0061217791432275_dp, 5.8193331245322035_dp, &
             ! w
             0.0008121033978678006_dp, 0.002475891449187961_dp, 0.0061840387301675456_dp, &
             0.014647615943609724_dp, 0.032879924438971635_dp, 0.07027723931085329_dp, &
             0.14405963768242985_dp, 0.2850571929983485_dp, 0.5479078906024393_dp, &
             1.0322349251234493_dp, 1.948564005643397_dp, 4.016770558125771_dp]

      CASE (305)

         aw(:) = & ! a
            [0.00024410845884712435_dp, 0.001503965954085424_dp, 0.00495431741873133_dp, &
             0.013620324176661923_dp, 0.034069498653249104_dp, 0.07963642987258447_dp, &
             0.1763621932285932_dp, 0.37349824006702287_dp, 0.7618906435501059_dp, &
             1.5072708912101296_dp, 2.9213720831136536_dp, 5.705202579937019_dp, &
             ! w
             0.0006481129780097904_dp, 0.0020466770735909014_dp, 0.005333058743876843_dp, &
             0.013042184119851218_dp, 0.02998380183138278_dp, 0.06532249265267938_dp, &
             0.1360452092785601_dp, 0.2728507539383112_dp, 0.5305383007341931_dp, &
             1.0094001089269478_dp, 1.9210522466051994_dp, 3.985441972650604_dp]

      CASE (306)

         aw(:) = & ! a
            [0.00020445702266883183_dp, 0.001281256958284562_dp, 0.004331576125033463_dp, &
             0.012209581331053501_dp, 0.031178020766606883_dp, 0.07411040155756039_dp, &
             0.16641059131119462_dp, 0.3565435846098175_dp, 0.7345542696373139_dp, &
             1.465654518888358_dp, 2.861727762472194_dp, 5.6245660855886985_dp, &
             ! w
             0.0005448622032481286_dp, 0.001773432389097132_dp, 0.004778122743988018_dp, &
             0.011970700521301728_dp, 0.028017415245052215_dp, 0.06191327500495124_dp, &
             0.13047012959118923_dp, 0.2642812084313893_dp, 0.5182503659100501_dp, &
             0.9931501845600177_dp, 1.9014045488724225_dp, 3.963061960527494_dp]

      CASE (307)

         aw(:) = & ! a
            [0.00017701712992867238_dp, 0.0011264749626048653_dp, 0.003894061476197579_dp, &
             0.01120414798695001_dp, 0.029089343238320638_dp, 0.07007272500776505_dp, &
             0.1590694792498803_dp, 0.3439359505070408_dp, 0.7140902598402066_dp, &
             1.4343293737380105_dp, 2.816642092608492_dp, 5.563433674386706_dp, &
             ! w
             0.00047335471896030166_dp, 0.001582388137128785_dp, 0.004382311140362497_dp, &
             0.011192673691802966_dp, 0.02657093384013896_dp, 0.05938011981654246_dp, &
             0.12629341090014687_dp, 0.25781656328144686_dp, 0.5089272241359735_dp, &
             0.9807660789125298_dp, 1.8863908554847064_dp, 3.9459563742834196_dp]

      CASE (308)

         aw(:) = & ! a
            [0.00015680129241092775_dp, 0.0010120177099265388_dp, 0.003567435469483656_dp, &
             0.01044442956687408_dp, 0.027493628154220823_dp, 0.06695941940829239_dp, &
             0.15336555072636895_dp, 0.3340773637667245_dp, 0.6980031233317447_dp, &
             1.4095969802818262_dp, 2.780925706814588_dp, 5.514893295772328_dp, &
             ! w
             0.00042064061830746427_dp, 0.001440362888929025_dp, 0.004083002663268586_dp, &
             0.010595792883007661_dp, 0.025449706287729155_dp, 0.05740079837258294_dp, &
             0.12300842150184972_dp, 0.25270414931933_dp, 0.5015205762704295_dp, &
             0.9708929815731903_dp, 1.874395859947349_dp, 3.932287409204537_dp]

      CASE (309)

         aw(:) = & ! a
            [0.00014123339422979187_dp, 0.0009235844571854585_dp, 0.0033129136408568405_dp, &
             0.009846227409444923_dp, 0.026225443989018295_dp, 0.06446605435754707_dp, &
             0.14876840474135875_dp, 0.3260898308896925_dp, 0.6849120562819437_dp, &
             1.389398856075759_dp, 2.751676990172356_dp, 5.475067612347207_dp, &
             ! w
             0.0003800261919516157_dp, 0.0013301003573762626_dp, 0.0038471561499707706_dp, &
             0.010119789362668696_dp, 0.02454788741911878_dp, 0.0557982667134647_dp, &
             0.12033438658185705_dp, 0.24852373674307937_dp, 0.495441460839257_dp, &
             0.9627660079132289_dp, 1.864504966724854_dp, 3.9210143330472396_dp]

      CASE (310)

         aw(:) = & ! a
            [0.00012884258237781313_dp, 0.0008529882715494499_dp, 0.003108152580403656_dp, &
             0.009360560312193842_dp, 0.02518757156783348_dp, 0.0624120873998581_dp, &
             0.14496100163188208_dp, 0.3194449964601701_dp, 0.6739813444718515_dp, &
             1.372483152762555_dp, 2.7271246942905365_dp, 5.441583486158302_dp, &
             ! w
             0.00034768720363469267_dp, 0.0012416884882636282_dp, 0.003655546666553874_dp, &
             0.00972909509388717_dp, 0.023802329627452105_dp, 0.054465997559836665_dp, &
             0.11810116915020036_dp, 0.24501915646440897_dp, 0.4903290170640597_dp, &
             0.9559146236107918_dp, 1.856154176832162_dp, 3.9114952214588152_dp]

      CASE (311)

         aw(:) = & ! a
            [0.00011872520483091538_dp, 0.0007951872127706306_dp, 0.002939312555228744_dp, &
             0.008956825862427074_dp, 0.024318736549766254_dp, 0.06068283979249307_dp, &
             0.1417406039031726_dp, 0.3138030384443519_dp, 0.664670854136532_dp, &
             1.358037548959736_dp, 2.7061158915034746_dp, 5.412892843244912_dp, &
             ! w
             0.00032127284046629584_dp, 0.001169004120712643_dp, 0.00349615969152817_dp, &
             0.009401211526504768_dp, 0.02317272329168829_dp, 0.053335496528745405_dp, &
             0.11619870985460978_dp, 0.242023840172228_dp, 0.48594762151819965_dp, &
             0.9500306179525972_dp, 1.848973326909618_dp, 3.903308669782224_dp]

      CASE (312)

         aw(:) = & ! a
            [7.008775402149382e-05_dp, 0.000514603341447927_dp, 0.002098964346237113_dp, &
             0.0068925799028730994_dp, 0.01977566522553707_dp, 0.051476174041109345_dp, &
             0.12434197056428334_dp, 0.28295181355533494_dp, 0.6132492392975714_dp, &
             1.277605091192352_dp, 2.588407362029324_dp, 5.251456362040042_dp, &
             ! w
             0.00019418571177335728_dp, 0.000810864352634179_dp, 0.002679316147965226_dp, &
             0.00767337842141309_dp, 0.01978936954067177_dp, 0.047167641120222_dp, &
             0.10568998356363599_dp, 0.22530693312206004_dp, 0.46128519787253686_dp, &
             0.9166901694437996_dp, 1.808120167667663_dp, 3.856714277189032_dp]

      CASE (313)

         aw(:) = & ! a
            [5.2138943102830404e-05_dp, 0.000409306526224381_dp, 0.0017704069260422984_dp, &
             0.0060528420260062485_dp, 0.017868708168537473_dp, 0.047515945104408215_dp, &
             0.1167105055867729_dp, 0.26920332146424586_dp, 0.5900340204014407_dp, &
             1.2409091384781628_dp, 2.534271438492348_dp, 5.176802135930046_dp, &
             ! w
             0.00014726020232260983_dp, 0.0006729441549679708_dp, 0.0023457790853374234_dp, &
             0.0069407052297282635_dp, 0.01831668661440417_dp, 0.044428524253441744_dp, &
             0.10094689382399291_dp, 0.21766012257552927_dp, 0.44987888309975665_dp, &
             0.9011384471340759_dp, 1.7889649083208523_dp, 3.8348545304578163_dp]

      CASE (314)

         aw(:) = & ! a
            [4.261493254491972e-05_dp, 0.0003527727371090675_dp, 0.0015893014787284403_dp, &
             0.005579256145782302_dp, 0.016774516774911593_dp, 0.0452133875137842_dp, &
             0.11222692983810825_dp, 0.26105766370605243_dp, 0.5761848412587852_dp, &
             1.2188966194989865_dp, 2.5016599323530255_dp, 5.131701437624496_dp, &
             ! w
             0.0001223662032403646_dp, 0.0005975720984372261_dp, 0.0021572482646542556_dp, &
             0.006518103345192513_dp, 0.01745528607734258_dp, 0.04280911567719277_dp, &
             0.09811848665806323_dp, 0.21306778903382625_dp, 0.4429888502949645_dp, &
             0.8917020906086252_dp, 1.7773100118148164_dp, 3.821549847575351_dp]

      CASE (315)

         aw(:) = & ! a
            [3.666426779379303e-05_dp, 0.0003171191362127202_dp, 0.0014728761601310882_dp, &
             0.005270021984096421_dp, 0.016051805758001095_dp, 0.04367929297588455_dp, &
             0.10921931955714444_dp, 0.25556352388338877_dp, 0.566802085460325_dp, &
             1.2039297761375631_dp, 2.4794259728407133_dp, 5.10089572719918_dp, &
             ! w
             0.00010681809466084738_dp, 0.0005493937032447625_dp, 0.002033953285766026_dp, &
             0.0062380475151624214_dp, 0.01687918929910473_dp, 0.04171847231017417_dp, &
             0.09620291520770906_dp, 0.20994326535366173_dp, 0.4382833418657671_dp, &
             0.8852388105310183_dp, 1.7693129347305854_dp, 3.812418836388984_dp]

      CASE (316)

         aw(:) = & ! a
            [3.258095852489616e-05_dp, 0.00029246253796716984_dp, 0.001391158355165531_dp, &
             0.005050458025198609_dp, 0.015534368103430863_dp, 0.04257403706841105_dp, &
             0.10704184852347523_dp, 0.25157023878774487_dp, 0.5599607543798089_dp, &
             1.1929889984088595_dp, 2.4631412952139593_dp, 5.078303234541634_dp, &
             ! w
             9.615286207387757e-05_dp, 0.0005157153072668187_dp, 0.001946310791824602_dp, &
             0.006037067822516981_dp, 0.01646302515775481_dp, 0.04092664681854212_dp, &
             0.09480659986183144_dp, 0.20765823148465828_dp, 0.4348328508627572_dp, &
             0.8804895241309951_dp, 1.7634291082281814_dp, 3.8056996902973355_dp]

      CASE (317)

         aw(:) = & ! a
            [2.9604747988173302e-05_dp, 0.000274370388233659_dp, 0.001330478862820812_dp, &
             0.004885958496957802_dp, 0.015144216737676134_dp, 0.041736684063799184_dp, &
             0.10538604301713107_dp, 0.24852461973746406_dp, 0.5547304198922046_dp, &
             1.1846084193789386_dp, 2.4506489987511735_dp, 5.060954915172068_dp, &
             ! w
             8.838149728758597e-05_dp, 0.0004907838512223435_dp, 0.0018805905358133131_dp, &
             0.0058852634473607794_dp, 0.01614710680853166_dp, 0.04032325906037048_dp, &
             0.09373934249682181_dp, 0.2059073593406457_dp, 0.4321835987443645_dp, &
             0.8768373538065267_dp, 1.7589001291082953_dp, 3.8005271312550946_dp]

      CASE (318)

         aw(:) = & ! a
            [2.7343719827419234e-05_dp, 0.00026054521624525645_dp, 0.0012836543895089704_dp, &
             0.004758109011011953_dp, 0.014839452347731865_dp, 0.04108012688474337_dp, &
             0.10408395577720814_dp, 0.24612402742747283_dp, 0.5506000240940426_dp, &
             1.1779802600289322_dp, 2.440757527008835_dp, 5.047207751457457_dp, &
             ! w
             8.247882492859621e-05_dp, 0.000471590688787289_dp, 0.001829477013847291_dp, &
             0.005766521594319684_dp, 0.015899015506388867_dp, 0.03984799549445157_dp, &
             0.09289670046973406_dp, 0.20452228055868693_dp, 0.430084490791883_dp, &
             0.8739400353462454_dp, 1.7553045113560795_dp, 3.796420182689478_dp]

      CASE (319)

         aw(:) = & ! a
            [2.5574733151017644e-05_dp, 0.00024967287838820624_dp, 0.0012465281404348222_dp, &
             0.004656145008377167_dp, 0.014595392765368722_dp, 0.04055274491356749_dp, &
             0.10303558367238869_dp, 0.2441875616515203_dp, 0.5472631222222715_dp, &
             1.1726189198158181_dp, 2.432749168846063_dp, 5.03607077412063_dp, &
             ! w
             7.786134633878762e-05_dp, 0.0004564015830582648_dp, 0.0017886893751454932_dp, &
             0.0056713282825404665_dp, 0.015699488388695025_dp, 0.039464839261616654_dp, &
             0.09221605744482542_dp, 0.20340172537792053_dp, 0.42838409351553175_dp, &
             0.8715907221401209_dp, 1.7523871953418695_dp, 3.793087723378363_dp]

      CASE (320)

         aw(:) = & ! a
            [2.416123354593282e-05_dp, 0.00024094573061246368_dp, 0.0012165203508505715_dp, &
             0.004573330630431546_dp, 0.014396497097563623_dp, 0.040121878233495056_dp, &
             0.10217741169459016_dp, 0.24259997023082036_dp, 0.5445239847651634_dp, &
             1.168213603282065_dp, 2.426163849504265_dp, 5.02690808444167_dp, &
             ! w
             7.417208130549883e-05_dp, 0.00044414355192226227_dp, 0.0017555465728734156_dp, &
             0.005593682473722872_dp, 0.015536312172565325_dp, 0.03915086280491433_dp, &
             0.09165742495916719_dp, 0.20248085367321886_dp, 0.4269852391095038_dp, &
             0.8696564606558904_dp, 1.7499840776440918_dp, 3.7903424576358176_dp]

      CASE (321)

         aw(:) = & ! a
            [1.894164014898326e-05_dp, 0.00020834134170772181_dp, 0.001102573174965553_dp, &
             0.004255354573720039_dp, 0.013626917540883616_dp, 0.03844527490385057_dp, &
             0.09882343542158212_dp, 0.2363736023713517_dp, 0.5337511717653998_dp, &
             1.1508488521066906_dp, 2.400161597773302_dp, 4.990687337267393_dp, &
             ! w
             6.054825300265897e-05_dp, 0.0003977512966816704_dp, 0.0016281580671739016_dp, &
             0.005292662029315792_dp, 0.014899928076456488_dp, 0.037920841856142196_dp, &
             0.08946113299959055_dp, 0.1988498907193317_dp, 0.4214565362923436_dp, &
             0.8619976422582177_dp, 1.7404580408085224_dp, 3.779458581884993_dp]

      CASE (322)

         aw(:) = & ! a
            [1.8934012296256358e-05_dp, 0.0002082931461534274_dp, 0.0011024023520154727_dp, &
             0.004254873447773686_dp, 0.013625745873543612_dp, 0.03844271076566545_dp, &
             0.09881828764566687_dp, 0.23636401719642108_dp, 0.5337345457964074_dp, &
             1.1508219943995057_dp, 2.4001213099764143_dp, 4.990631150994172_dp, &
             ! w
             6.0528329999281435e-05_dp, 0.00039768193515047327_dp, 0.0016279651361197933_dp, &
             0.005292202990367738_dp, 0.014898953114611908_dp, 0.03791895060651046_dp, &
             0.0894577452373523_dp, 0.19884427477508254_dp, 0.42144796502526766_dp, &
             0.8619857428372386_dp, 1.7404432216810901_dp, 3.779441649091308_dp]

      CASE (323)

         aw(:) = & ! a
            [0.025586820710741212_dp, 0.13597714276535305_dp, 0.3394815035908448_dp, &
             0.645238040957444_dp, 1.06803887406634_dp, 1.6304515349860733_dp, &
             2.3659824690662763_dp, 3.3237148488909694_dp, 4.575354823060675_dp, &
             6.227412457617611_dp, 8.446967333451797_dp, 11.53142920465396_dp, &
             16.187093383419732_dp, &
             ! w
             0.06579511197609775_dp, 0.15581449726553434_dp, 0.25269312295697643_dp, &
             0.3612509019945492_dp, 0.48806016096110766_dp, 0.6422740734747383_dp, &
             0.8368561900828182_dp, 1.0903726366052773_dp, 1.430281496814054_dp, &
             1.9007339485469747_dp, 2.584815795943897_dp, 3.6846225781604858_dp, &
             5.9974170499464545_dp]

      CASE (324)

         aw(:) = & ! a
            [0.015352742449764926_dp, 0.08190582986667876_dp, 0.20597658744065508_dp, &
             0.39588447355131207_dp, 0.6656749056064382_dp, 1.0376871217418018_dp, &
             1.546150772948807_dp, 2.24196428609261_dp, 3.199493073654813_dp, &
             4.528287494545492_dp, 6.398138729903282_dp, 9.106778496873513_dp, &
             13.3491215794235_dp, &
             ! w
             0.03951386830420629_dp, 0.09430497087174285_dp, 0.15519500637374517_dp, &
             0.22691929188907797_dp, 0.31632490899004906_dp, 0.4333263112160413_dp, &
             0.5919767118483865_dp, 0.8118655465128661_dp, 1.1210827091354272_dp, &
             1.5638845436284246_dp, 2.2224139656716746_dp, 3.2940777647231294_dp, &
             5.5510419006212235_dp]

      CASE (325)

         aw(:) = & ! a
            [0.01127574349610271_dp, 0.06032239787221326_dp, 0.1525007237397826_dp, &
             0.2955277282819509_dp, 0.5027902095349923_dp, 0.7960938616773462_dp, &
             1.2093501035922478_dp, 1.7933089972016882_dp, 2.622363474304417_dp, &
             3.8064590061885735_dp, 5.516260496116213_dp, 8.050119245723234_dp, &
             12.098989522348345_dp, &
             ! w
             0.02903917514657867_dp, 0.06969309707580232_dp, 0.11591945784093333_dp, &
             0.1723198518541611_dp, 0.24577780538823762_dp, 0.34640460833112163_dp, &
             0.4884592767890423_dp, 0.6916744519623734_dp, 0.9843875484296041_dp, &
             1.4113359869286717_dp, 2.0550145974924656_dp, 3.1115307366236418_dp, &
             5.342250701176915_dp]

      CASE (326)

         aw(:) = & ! a
            [0.009026898154561844_dp, 0.04839933033898771_dp, 0.12288198160156376_dp, &
             0.2397391142273434_dp, 0.41182175891984385_dp, 0.6603888791946663_dp, &
             1.018784626228823_dp, 1.5371450840589502_dp, 2.2893425823845694_dp, &
             3.3851999418594696_dp, 4.995853148865419_dp, 7.4203059140681455_dp, &
             11.347630826479701_dp, &
             ! w
             0.023259371522787082_dp, 0.05607233406744155_dp, 0.09407216033479125_dp, &
             0.14173660359977427_dp, 0.20591056376397773_dp, 0.2967053050336383_dp, &
             0.4283329410906127_dp, 0.6205113431853985_dp, 0.9018119630642786_dp, &
             1.3175182943399095_dp, 1.950675047069855_dp, 2.996916913306896_dp, &
             5.211189067278923_dp]

      CASE (327)

         aw(:) = & ! a
            [0.007583566224204744_dp, 0.04073763079402279_dp, 0.10380803619051947_dp, &
             0.2037030441964128_dp, 0.3528297904687777_dp, 0.5719362889920455_dp, &
             0.8937509529926758_dp, 1.3676956728177228_dp, 2.0670104178735342_dp, &
             3.101260624440397_dp, 4.641888042160206_dp, 6.988499295038847_dp, &
             10.829141964301934_dp, &
             ! w
             0.01954876830141318_dp, 0.047306690644147206_dp, 0.07995321519886125_dp, &
             0.12185677751129535_dp, 0.17979436712684913_dp, 0.26379797808027833_dp, &
             0.3879558647919772_dp, 0.571946108592381_dp, 0.8445649615904518_dp, &
             1.2516047076139492_dp, 1.8766592674405786_dp, 2.9152010854831847_dp, &
             5.117788931492246_dp]

      CASE (328)

         aw(:) = & ! a
            [0.006571003540270054_dp, 0.03535697894808773_dp, 0.09038798510022897_dp, &
             0.17828212453182066_dp, 0.31107006018785655_dp, 0.5090319817929683_dp, &
             0.804294662529962_dp, 1.2455733860384253_dp, 1.9054792787505055_dp, &
             2.8932856501032744_dp, 4.3806489535904545_dp, 6.667715839723064_dp, &
             10.441936803488176_dp, &
             ! w
             0.016944972378415024_dp, 0.0411429334765565_dp, 0.069989346884536_dp, &
             0.10775614466512438_dp, 0.1611404671865264_dp, 0.24006224528660414_dp, &
             0.3584665626662519_dp, 0.5359938990470181_dp, 0.8016474287450003_dp, &
             1.201672057883574_dp, 1.820170175732486_dp, 2.85259527233136_dp, &
             5.0462646056798155_dp]

      CASE (329)

         aw(:) = & ! a
            [0.00581756844645099_dp, 0.03134961581118217_dp, 0.08037675834692043_dp, &
             0.1592740538724147_dp, 0.27974646477337584_dp, 0.4616472946686746_dp, &
             0.7365358864903887_dp, 1.152461515922996_dp, 1.7814423255914706_dp, &
             2.732456742950032_dp, 4.1773180520687125_dp, 6.416654069276692_dp, &
             10.137558331678514_dp, &
             ! w
             0.0150071017032093_dp, 0.03654719753795217_dp, 0.06253657270889061_dp, &
             0.09716122078717658_dp, 0.14703419585189478_dp, 0.22195120873509197_dp, &
             0.3357153160293236_dp, 0.5079360945962476_dp, 0.7678029510712538_dp, &
             1.1619597933934014_dp, 1.774971707625038_dp, 2.802347223798614_dp, &
             4.988882893643294_dp]

      CASE (330)

         aw(:) = & ! a
            [0.005232938971153595_dp, 0.028237522193697503_dp, 0.07259067018283133_dp, &
             0.14445951077075755_dp, 0.2552626614849805_dp, 0.42446316473997137_dp, &
             0.6830922302920382_dp, 1.0785818503932039_dp, 1.6824002197305863_dp, &
             2.603239326783796_dp, 4.013029479317588_dp, 6.2128284447499835_dp, &
             9.88951840694477_dp, &
             ! w
             0.013503115172407102_dp, 0.03297459314353523_dp, 0.056726387700893124_dp, &
             0.08886720981520872_dp, 0.1359255258845317_dp, 0.20757045072921196_dp, &
             0.31747132217542084_dp, 0.48521293975581137_dp, 0.7401508114959612_dp, &
             1.129281387058064_dp, 1.737590139062592_dp, 2.7606812801987437_dp, &
             4.941319591053454_dp]

      CASE (331)

         aw(:) = & ! a
            [0.004764812206990202_dp, 0.025743717023161722_dp, 0.06634309763098721_dp, &
             0.13254918901322424_dp, 0.23552556282645784_dp, 0.39437719281092637_dp, &
             0.6396459009904628_dp, 1.018194486068431_dp, 1.6009832616680564_dp, &
             2.4964300537331274_dp, 3.8765534322194464_dp, 6.042797435469854_dp, &
             9.681926134606167_dp, &
             ! w
             0.012298624445397305_dp, 0.03010915114692234_dp, 0.05205409224682944_dp, &
             0.0821720281454225_dp, 0.12690822677767305_dp, 0.19580758337318957_dp, &
             0.3024160839681164_dp, 0.46629843854623454_dp, 0.7169579916993148_dp, &
             1.1017045394902962_dp, 1.7059067569607391_dp, 2.725287468572967_dp, &
             4.900929525247936_dp]

      CASE (332)

         aw(:) = & ! a
            [0.004380695877988043_dp, 0.023696032529779535_dp, 0.061206817958545164_dp, &
             0.12273976774091432_dp, 0.21922884725225694_dp, 0.36944915428665787_dp, &
             0.6034892573939584_dp, 0.9676872342224387_dp, 1.5325340755817476_dp, &
             2.4061866587862055_dp, 3.760730311086456_dp, 5.897958210150131_dp, &
             9.504578462029908_dp, &
             ! w
             0.011310131876868038_dp, 0.027754333129107387_dp, 0.04820514535602219_dp, &
             0.07663699929378934_dp, 0.11941431361329644_dp, 0.1859624116085837_dp, &
             0.28971433902840205_dp, 0.45021808459781604_dp, 0.6971087365781358_dp, &
             1.0779765276442093_dp, 1.6785411291166155_dp, 2.69465685384593_dp, &
             4.865985188743286_dp]

      CASE (333)

         aw(:) = & ! a
            [0.0025091447649231518_dp, 0.013693954323940598_dp, 0.03600534963956549_dp, &
             0.0742866820971901_dp, 0.13795424770764012_dp, 0.24347562392253488_dp, &
             0.41776894357208094_dp, 0.7035481042365874_dp, 1.1680019135509343_dp, &
             1.91726565781939_dp, 3.1235654399781727_dp, 5.090979407316672_dp, &
             8.506780966608327_dp, &
             ! w
             0.0064910169387615785_dp, 0.01621706891887387_dp, 0.029181234262053293_dp, &
             0.04891003792535065_dp, 0.08111585359226478_dp, 0.13432719007941266_dp, &
             0.22123864262390655_dp, 0.36128939869381554_dp, 0.5848980597832798_dp, &
             0.9414285557062445_dp, 1.5190239900002864_dp, 2.5149049645269637_dp, &
             4.661108986894671_dp]

      CASE (334)

         aw(:) = & ! a
            [0.001806344761894647_dp, 0.009921598041914002_dp, 0.02642548679866468_dp, &
             0.05564594784382461_dp, 0.1061317262233611_dp, 0.19297872173170383_dp, &
             0.341247835854634_dp, 0.5915378195802624_dp, 1.0090348395240913_dp, &
             1.698503088979307_dp, 2.8320126081133545_dp, 4.714893936031113_dp, &
             8.035297141565342_dp, &
             ! w
             0.004679508185843388_dp, 0.01184276418242282_dp, 0.02185653875493935_dp, &
             0.03797172723427941_dp, 0.06546164362694586_dp, 0.11231904712276647_dp, &
             0.1908397923835511_dp, 0.32036737167335694_dp, 0.5316673394828265_dp, &
             0.875036346274796_dp, 1.4400626713409013_dp, 2.4250786861151266_dp, &
             4.558840325251003_dp]

      CASE (335)

         aw(:) = & ! a
            [0.0014294954818313335_dp, 0.007892561660233931_dp, 0.021243656118378387_dp, &
             0.0454727229384415_dp, 0.08853415433528523_dp, 0.16457980344733686_dp, &
             0.29740499192555203_dp, 0.5261556981667116_dp, 0.9146013961828088_dp, &
             1.5664758893112252_dp, 2.6536401907715668_dp, 4.482250787728166_dp, &
             7.741239498452869_dp, &
             ! w
             0.003707454782787126_dp, 0.009481200335284304_dp, 0.01785788759977449_dp, &
             0.031890560274698374_dp, 0.056536120284522984_dp, 0.09942866225608747_dp, &
             0.17259663294033112_dp, 0.2952899874667848_dp, 0.49846247913640174_dp, &
             0.8330134160758221_dp, 1.3895433034020468_dp, 2.367271349260771_dp, &
             4.493058443621265_dp]

      CASE (336)

         aw(:) = & ! a
            [0.0011918361572482998_dp, 0.006609790028561528_dp, 0.017952705675068747_dp, &
             0.03896391960290669_dp, 0.07715294167358444_dp, 0.14596684889042683_dp, &
             0.2682605794022303_dp, 0.4820900184432985_dp, 0.8501377564276386_dp, &
             1.475315159861829_dp, 2.5292704379654323_dp, 4.318759829944416_dp, &
             7.533387743823241_dp, &
             ! w
             0.00309407700416737_dp, 0.00798376946113414_dp, 0.015299313447531767_dp, &
             0.027940576171147948_dp, 0.05062324037940678_dp, 0.0907199318854559_dp, &
             0.1600583567295716_dp, 0.27780021778331876_dp, 0.4750130531377467_dp, &
             0.8030283835320917_dp, 1.353217071700287_dp, 2.3255281638367187_dp, &
             4.445569586002092_dp]

      CASE (337)

         aw(:) = & ! a
            [0.001027137675099082_dp, 0.005718969305900617_dp, 0.01565839219507016_dp, &
             0.03439710451347879_dp, 0.06909330507123403_dp, 0.13263989876743576_dp, &
             0.24715361813290734_dp, 0.4498264599858398_dp, 0.8024646171256475_dp, &
             1.4072970218639687_dp, 2.43576833316227_dp, 4.195096891786353_dp, &
             7.375469199813955_dp, &
             ! w
             0.0026687973590316587_dp, 0.006941272928014548_dp, 0.013504119704610738_dp, &
             0.02513307772763668_dp, 0.04635227962651341_dp, 0.08433232366424552_dp, &
             0.15074019422638985_dp, 0.26465568169652737_dp, 0.4572195938687417_dp, &
             0.7800936606603105_dp, 1.3252656634607674_dp, 2.2933018738763806_dp, &
             4.408913503078078_dp]

      CASE (338)

         aw(:) = & ! a
            [0.0009057063557367582_dp, 0.0050609742799299145_dp, 0.013957892715389362_dp, &
             0.03099289087333813_dp, 0.06303652695019825_dp, 0.12252994015290104_dp, &
             0.23098802484999723_dp, 0.4248923236469685_dp, 0.765318137452345_dp, &
             1.3539125860337489_dp, 2.3619293562503416_dp, 4.0969574805808335_dp, &
             7.249693224218248_dp, &
             ! w
             0.0023551074366806287_dp, 0.006169560864039329_dp, 0.012165999842184043_dp, &
             0.023016426049836298_dp, 0.04308816974049331_dp, 0.07938914807328556_dp, &
             0.14345232336954264_dp, 0.2542818368570687_dp, 0.44306747733264_dp, &
             0.7617342359688349_dp, 1.302781312838692_dp, 2.267308148921753_dp, &
             4.3793496679198665_dp]

      CASE (339)

         aw(:) = & ! a
            [0.0008121562645573277_dp, 0.004553233604141542_dp, 0.01264163804743066_dp, &
             0.028344203975865997_dp, 0.058289724018512935_dp, 0.11454112796017396_dp, &
             0.21810891379792788_dp, 0.40487450869670616_dp, 0.7352886271457775_dp, &
             1.3104923188652018_dp, 2.301561095349978_dp, 4.0163901817262655_dp, &
             7.146127573918624_dp, &
             ! w
             0.0021133505242918287_dp, 0.005572909503115044_dp, 0.011124932829870654_dp, &
             0.02135276049935338_dp, 0.04049232245911701_dp, 0.07541647517567895_dp, &
             0.13754327950438497_dp, 0.24580700768708516_dp, 0.4314308537768419_dp, &
             0.7465562294807135_dp, 1.284117051126591_dp, 2.245681072627058_dp, &
             4.354753793838828_dp]

      CASE (340)

         aw(:) = & ! a
            [0.0007376832054758359_dp, 0.004148437245624365_dp, 0.011589277560737764_dp, &
             0.026216438624069104_dp, 0.05445141244764643_dp, 0.10803396170468815_dp, &
             0.20754286697441598_dp, 0.38834227253882253_dp, 0.710339190947803_dp, &
             1.2742276377578197_dp, 2.2509169842175876_dp, 3.9485610579777064_dp, &
             7.058712356075641_dp, &
             ! w
             0.0019208272844967538_dp, 0.005096388346275755_dp, 0.010288681461498514_dp, &
             0.020003964851027276_dp, 0.03836598994370699_dp, 0.07213283056214909_dp, &
             0.1326220401624146_dp, 0.23870321832702932_dp, 0.4216224758355981_dp, &
             0.733703324253776_dp, 1.268256298266233_dp, 2.2272659740952503_dp, &
             4.3338116476564155_dp]

      CASE (341)

         aw(:) = & ! a
            [0.0006768696090223793_dp, 0.003817439892751027_dp, 0.01072651642580349_dp, &
             0.024464299529105633_dp, 0.05127171840130901_dp, 0.1026078910917859_dp, &
             0.1986759226082186_dp, 0.37438694444199266_dp, 0.6891677591849794_dp, &
             1.243312621519654_dp, 2.207575740341867_dp, 3.8903332120321474_dp, &
             6.983503032159278_dp, &
             ! w
             0.0017635661455619365_dp, 0.004706110756320155_dp, 0.009600116197389492_dp, &
             0.01888391584913048_dp, 0.036584034840318434_dp, 0.06935918825624371_dp, &
             0.12843766518029287_dp, 0.23262901684656914_dp, 0.41319494401699847_dp, &
             0.7226149903814322_dp, 1.25453089790273_dp, 2.211302264573346_dp, &
             4.315657718002286_dp]

      CASE (342)

         aw(:) = & ! a
            [0.00038452714734484467_dp, 0.002218600873633601_dp, 0.006518782483552045_dp, &
             0.015779014074440583_dp, 0.035172364824389984_dp, 0.0745116209124148_dp, &
             0.15177184191745516_dp, 0.2991167658475698_dp, 0.5729793346936465_dp, &
             1.071061805318608_dp, 1.962978129683943_dp, 3.5583604501642894_dp, &
             6.551558946577899_dp, &
             ! w
             0.0010067371248319557_dp, 0.0028099546056502035_dp, 0.00618802787423932_dp, &
             0.013163809149326671_dp, 0.02720166466963974_dp, 0.05437575286824172_dp, &
             0.10534346464598406_dp, 0.19848144775358342_dp, 0.365055625720691_dp, &
             0.6584198791878357_dp, 1.1742485462233792_dp, 2.117375670194502_dp, &
             4.20884361574561_dp]

      CASE (343)

         aw(:) = & ! a
            [0.0002765305534557809_dp, 0.0016230480146192058_dp, 0.004924194174712558_dp, &
             0.012392626875699645_dp, 0.028674604784592495_dp, 0.06277505950348358_dp, &
             0.1315527145078089_dp, 0.26575193734456104_dp, 0.5202041093762643_dp, &
             0.9911579072113549_dp, 1.8474989877676289_dp, 3.399439816082922_dp, &
             6.3427254429993996_dp, &
             ! w
             0.0007266173762481529_dp, 0.0020964974323044224_dp, 0.00485817671567068_dp, &
             0.010822599427408885_dp, 0.02318605457776981_dp, 0.0477278172991908_dp, &
             0.09478793298707364_dp, 0.18247331930475158_dp, 0.3419906840673959_dp, &
             0.6270943201002195_dp, 1.1345214719922043_dp, 2.07051977057129_dp, &
             4.155550678612812_dp]

      CASE (344)

         aw(:) = & ! a
            [0.0002190369434955462_dp, 0.0013041638214468583_dp, 0.004059496158467338_dp, &
             0.010519066359753125_dp, 0.024997669627579076_dp, 0.05599099115221941_dp, &
             0.11964332368032968_dp, 0.24577445856885485_dp, 0.488152408807871_dp, &
             0.9420361988247535_dp, 1.775783724038284_dp, 3.299957459828327_dp, &
             6.211259721847333_dp, &
             ! w
             0.0005772981384290037_dp, 0.0017117321655301405_dp, 0.004122409744541786_dp, &
             0.009485681360429935_dp, 0.020831898560872888_dp, 0.04374884426072721_dp, &
             0.08836114093280086_dp, 0.17258330268935843_dp, 0.3275606215791396_dp, &
             0.607288120762629_dp, 1.1091987143511128_dp, 2.040510996781803_dp, &
             4.121413407696248_dp]

      CASE (345)

         aw(:) = & ! a
            [0.00018292554248688548_dp, 0.0011029602809106097_dp, 0.003508204077638724_dp, &
             0.009305558270018882_dp, 0.022575689079456448_dp, 0.05145322081026685_dp, &
             0.11157011580444981_dp, 0.2320750581708474_dp, 0.4659542555149446_dp, &
             0.907726951170141_dp, 1.7253408354432977_dp, 3.2295976939344446_dp, &
             6.117918237876528_dp, &
             ! w
             0.00048341836983747076_dp, 0.001467549224849752_dp, 0.0036457480254511606_dp, &
             0.00859915348968931_dp, 0.01924178101950627_dp, 0.04102220781900407_dp, &
             0.0839043742330625_dp, 0.16565495050506554_dp, 0.3173632327682836_dp, &
             0.5931886289713928_dp, 1.0910702058875275_dp, 2.018956366734576_dp, &
             4.096889505593968_dp]

      CASE (346)

         aw(:) = & ! a
            [0.000157963947532522_dp, 0.0009633488427175757_dp, 0.0031222338920011534_dp, &
             0.008444768319610179_dp, 0.020834486239889226_dp, 0.048151695748445335_dp, &
             0.10563562134899177_dp, 0.2219158164417318_dp, 0.4493679596948343_dp, &
             0.8819265141436959_dp, 1.6872058594615873_dp, 3.1761840167015074_dp, &
             6.046850903162634_dp, &
             ! w
             0.000418472794854246_dp, 0.001297273226305447_dp, 0.0033075235899535918_dp, &
             0.007958461857768825_dp, 0.01807633560260295_dp, 0.0390017527131793_dp, &
             0.08057188460113408_dp, 0.16043433301035423_dp, 0.3096284138665074_dp, &
             0.5824345405064077_dp, 1.0771838414761132_dp, 2.002403861388036_dp, &
             4.07805426041055_dp]

      CASE (347)

         aw(:) = & ! a
            [0.00013959111039492074_dp, 0.0008602478231696832_dp, 0.002834933533612028_dp, &
             0.007796820896148651_dp, 0.019509164130539908_dp, 0.0456141150523087_dp, &
             0.10103628350635624_dp, 0.2139863759050597_dp, 0.4363438111546552_dp, &
             0.8615632167206827_dp, 1.6569797098202836_dp, 3.1337076649686693_dp, &
             5.99020477695265_dp, &
             ! w
             0.0003706377943815103_dp, 0.0011709747407894105_dp, 0.003052834447777809_dp, &
             0.007468703767697578_dp, 0.017175337038177165_dp, 0.03742598157368391_dp, &
             0.07795396841799117_dp, 0.15630785109678166_dp, 0.30348236203184703_dp, &
             0.5738515410841367_dp, 1.0660630760791807_dp, 1.989121140896082_dp, &
             4.0629379599812605_dp]

      CASE (348)

         aw(:) = & ! a
            [0.00012545332134134083_dp, 0.0007806785450082562_dp, 0.002611617669561367_dp, &
             0.007288228687817971_dp, 0.018458979830445036_dp, 0.04358679859808994_dp, &
             0.09733625074458488_dp, 0.207569816731476_dp, 0.4257518170980608_dp, &
             0.8449324268358553_dp, 1.6322074462041654_dp, 3.0988005415161073_dp, &
             5.9435640977043915_dp, &
             ! w
             0.0003338078810012544_dp, 0.0010731162787732957_dp, 0.002852841923913739_dp, &
             0.0070792216098078805_dp, 0.016452071309619343_dp, 0.03615180925671417_dp, &
             0.07582438457542363_dp, 0.1529339780982986_dp, 0.29843532311328425_dp, &
             0.5667775223001789_dp, 1.0568715828439852_dp, 1.9781243638999457_dp, &
             4.050421851496485_dp]

      CASE (349)

         aw(:) = & ! a
            [0.00011420782823794646_dp, 0.000717219178008899_dp, 0.002432350146565218_dp, &
             0.0068763947394231825_dp, 0.017601559600451056_dp, 0.041919905687346874_dp, &
             0.09427596163377339_dp, 0.20223610579221138_dp, 0.4169099038558701_dp, &
             0.8309996676277617_dp, 1.611392549114466_dp, 3.069402153219648_dp, &
             5.904220564916988_dp, &
             ! w
             0.00030449799865156393_dp, 0.0009947872650128496_dp, 0.002690831402232914_dp, &
             0.0067602481509608766_dp, 0.015854998355485473_dp, 0.0350934163362851_dp, &
             0.07404640435482042_dp, 0.15010495924541264_dp, 0.29418770846220355_dp, &
             0.5608055691529965_dp, 1.049093494657459_dp, 1.9688053441501094_dp, &
             4.039814271278962_dp]

      CASE (350)

         aw(:) = & ! a
            [0.00010503032595626133_dp, 0.0006653037355264198_dp, 0.002284808747141548_dp, &
             0.00653479029689385_dp, 0.01688516727545049_dp, 0.04051858357228697_dp, &
             0.09168996420612356_dp, 0.19770945978443402_dp, 0.40937831727498003_dp, &
             0.819094877997119_dp, 1.5935618858617477_dp, 3.0441685261321627_dp, &
             5.8704038773647556_dp, &
             ! w
             0.00028056761655329845_dp, 0.0009304915861452455_dp, 0.0025563936765912246_dp, &
             0.006493022545935985_dp, 0.015351319502275317_dp, 0.034195774587582224_dp, &
             0.072531801119424_dp, 0.14768599429281498_dp, 0.2905441729112578_dp, &
             0.5556692375929922_dp, 1.0423899529883427_dp, 1.9607638827744525_dp, &
             4.0306601210273865_dp]

      CASE (351)

         aw(:) = & ! a
            [6.0956302303357914e-05_dp, 0.0004138256560681806_dp, 0.0015544459121753231_dp, &
             0.004798361250838413_dp, 0.013155891287071456_dp, 0.03307746300802235_dp, &
             0.07772964065380825_dp, 0.1729320306822195_dp, 0.3676675397432615_dp, &
             0.7525109383547484_dp, 1.4930192950672836_dp, 2.9009782543357345_dp, &
             5.677660607120728_dp, &
             ! w
             0.0001654908148945107_dp, 0.0006151784699279699_dp, 0.001871869030778897_dp, &
             0.005090033368441002_dp, 0.012648088621176327_dp, 0.029294877358506985_dp, &
             0.0641449465226693_dp, 0.13413056599077305_dp, 0.2699174240072517_dp, &
             0.5263420537685121_dp, 1.0038603898502862_dp, 1.9143610028934042_dp, &
             3.9778209431449656_dp]

      CASE (352)

         aw(:) = & ! a
            [4.468259611956098e-05_dp, 0.00031958277172101316_dp, 0.0012704370862732053_dp, &
             0.004095065206711756_dp, 0.011592626388800938_dp, 0.029870575454816518_dp, &
             0.07157599078255672_dp, 0.16180480043873144_dp, 0.3486416674339012_dp, &
             0.7217412890707888_dp, 1.4460582142493035_dp, 2.833542511723205_dp, &
             5.586367411415857_dp, &
             ! w
             0.00012292811035454133_dp, 0.000494404960373647_dp, 0.0015937643014944905_dp, &
             0.004495058589365715_dp, 0.011466756376369864_dp, 0.02710293595495223_dp, &
             0.0603224187630089_dp, 0.1278538769406165_dp, 0.2602377796731581_dp, &
             0.5124251314491098_dp, 0.9854183262584159_dp, 1.8920351868576104_dp, &
             3.9523876209156668_dp]

      CASE (353)

         aw(:) = & ! a
            [3.602414857739837e-05_dp, 0.0002689177215294932_dp, 0.0011138900112606656_dp, &
             0.0036977107134903097_dp, 0.010691796563079364_dp, 0.02799370344199565_dp, &
             0.06792933792151704_dp, 0.15514312043586892_dp, 0.3371541164686315_dp, &
             0.7030312092079485_dp, 1.4173371437396478_dp, 2.7921147310098906_dp, &
             5.53011043527255_dp, &
             ! w
             0.00010026840907033134_dp, 0.0004284578838812178_dp, 0.0014363039742978636_dp, &
             0.004150066629245997_dp, 0.010770327177645532_dp, 0.0257941271944965_dp, &
             0.05801624766424963_dp, 0.12403429301516422_dp, 0.2543043980292382_dp, &
             0.5038426226723924_dp, 0.9739918162410033_dp, 1.8781632305683442_dp, &
             3.936580836490376_dp]

      CASE (354)

         aw(:) = & ! a
            [3.059290317116564e-05_dp, 0.00023687354026147777_dp, 0.0010129814071288485_dp, &
             0.0034370642238080416_dp, 0.010092843649304992_dp, 0.026732628752206854_dp, &
             0.06545858862460072_dp, 0.15059872432458815_dp, 0.329273330106885_dp, &
             0.6901352746833209_dp, 1.3974653098481216_dp, 2.7633665911407808_dp, &
             5.490992564617657_dp, &
             ! w
             8.605164165141342e-05_dp, 0.00038623147208030476_dp, 0.0013328549100884563_dp, &
             0.003919749917228889_dp, 0.010300190614591443_dp, 0.02490300747167536_dp, &
             0.056435183477473685_dp, 0.12140061114751656_dp, 0.25019343575120073_dp, &
             0.4978723341593861_dp, 0.9660184769062037_dp, 1.8684653230203156_dp, &
             3.925528352086655_dp]

      CASE (355)

         aw(:) = & ! a
            [2.6845896830691438e-05_dp, 0.00021461241100040283_dp, 0.0009418044037313119_dp, &
             0.0032507493716804788_dp, 0.009660358732766935_dp, 0.025814963509048904_dp, &
             0.06364959420170253_dp, 0.147254847407397_dp, 0.3234505236479869_dp, &
             0.6805743737332881_dp, 1.3826914922067293_dp, 2.7419477342100405_dp, &
             5.461804767163133_dp, &
             ! w
             7.624345752222808e-05_dp, 0.00035659607805726403_dp, 0.0012588177132532594_dp, &
             0.0037529578889870484_dp, 0.009956926167738479_dp, 0.024248270310051_dp, &
             0.055267631418727196_dp, 0.11944758096159838_dp, 0.24713418393642417_dp, &
             0.49341644271764507_dp, 0.9600541594749636_dp, 1.8612010493317928_dp, &
             3.917248355526611_dp]

      CASE (356)

         aw(:) = & ! a
            [2.4094823232259868e-05_dp, 0.00019816925495368652_dp, 0.0008885607718760437_dp, &
             0.0031098892132738133_dp, 0.009330785724583677_dp, 0.025111426490404324_dp, &
             0.062256093556386755_dp, 0.14466904915153894_dp, 0.3189334519863107_dp, &
             0.6731379052209211_dp, 1.3711757555495363_dp, 2.7252247966433902_dp, &
             5.438990337108309_dp, &
             ! w
             6.904283412341186e-05_dp, 0.0003345150123097334_dp, 0.0012027886212585387_dp, &
             0.0036255705603485647_dp, 0.009693084355477577_dp, 0.023742566479366768_dp, &
             0.05436230874510011_dp, 0.11792829638341537_dp, 0.24474789910260672_dp, &
             0.4899329193331565_dp, 0.9553832519272085_dp, 1.8555060801891419_dp, &
             3.910756415013032_dp]

      CASE (357)

         aw(:) = & ! a
            [2.1984307370857297e-05_dp, 0.00018548696533501942_dp, 0.000847051537777818_dp, &
             0.0029991098457322163_dp, 0.009069920827295656_dp, 0.024551838658194044_dp, &
             0.061143457040534074_dp, 0.1425980180708223_dp, 0.3153063777625399_dp, &
             0.6671540370846212_dp, 1.3618935366720428_dp, 2.7117275232706946_dp, &
             5.420559873935379_dp, &
             ! w
             6.351947555942654e-05_dp, 0.0003173554870172747_dp, 0.0011586892549084967_dp, &
             0.0035245614263889643_dp, 0.009482800720598632_dp, 0.023337933345164923_dp, &
             0.05363564328753332_dp, 0.11670566218661632_dp, 0.24282337996420006_dp, &
             0.487118429227304_dp, 0.9516041653640989_dp, 1.8508945552459837_dp, &
             3.9054990836020873_dp]

      CASE (358)

         aw(:) = & ! a
            [2.031156839834793e-05_dp, 0.00017538664846030618_dp, 0.0008136848678657221_dp, &
             0.0029094046335054503_dp, 0.008857546946392335_dp, 0.024094424360485655_dp, &
             0.06023108615116179_dp, 0.1408954044595645_dp, 0.3123182459571927_dp, &
             0.6622157063979266_dp, 1.3542223373494233_dp, 2.700560711742634_dp, &
             5.405300304241755_dp, &
             ! w
             5.9142380513518004e-05_dp, 0.00030359853043067684_dp, 0.0011229551350652564_dp, &
             0.003442209358442474_dp, 0.009310627355174086_dp, 0.023005556667744488_dp, &
             0.05303718866658264_dp, 0.11569658666450071_dp, 0.2412321732593248_dp, &
             0.4847879398770228_dp, 0.9484713730416796_dp, 1.8470690280275346_dp, &
             3.9011375187965016_dp]

      CASE (359)

         aw(:) = & ! a
            [1.895213311382775e-05_dp, 0.00016714197755685704_dp, 0.0007862261856894293_dp, &
             0.002835117155929301_dp, 0.00868087251405237_dp, 0.02371259819497883_dp, &
             0.059467448761291966_dp, 0.1394672790474465_dp, 0.3098074145316526_dp, &
             0.6580601213399029_dp, 1.347759389326439_dp, 2.6911441396230593_dp, &
             5.392424421173327_dp, &
             ! w
             5.5585566252102124e-05_dp, 0.0002923026615596326_dp, 0.0010933457681329864_dp, &
             0.0033736175164245734_dp, 0.00916670704022603_dp, 0.022726961980710926_dp, &
             0.05253447378770823_dp, 0.11484741580812091_dp, 0.23989110364890742_dp, &
             0.4828213524067168_dp, 0.9458252126439093_dp, 1.843835848614256_dp, &
             3.8974510845721055_dp]

      CASE (360)

         aw(:) = & ! a
            [1.2631096573461333e-05_dp, 0.00012823114964889898_dp, 0.0006533583639144626_dp, &
             0.00246894477740814_dp, 0.007798455611566926_dp, 0.02178661814865923_dp, &
             0.055585788243469034_dp, 0.13216284059551617_dp, 0.2968998479855184_dp, &
             0.6366075269218335_dp, 1.3142815347706613_dp, 2.6422386687752124_dp, &
             5.325432923990921_dp, &
             ! w
             3.905245929050013e-05_dp, 0.0002379839928024588_dp, 0.000947155383419069_dp, &
             0.003029866845825191_dp, 0.008437929334998805_dp, 0.02130504814726735_dp, &
             0.04995246800201147_dp, 0.11046334300216694_dp, 0.23293749231829983_dp, &
             0.47258778725902156_dp, 0.9320171019691014_dp, 1.8269360619592532_dp, &
             3.8781787480793275_dp]

      CASE (361)

         aw(:) = & ! a
            [1.0714409072106346e-05_dp, 0.00011616318173188649_dp, 0.0006107789893315344_dp, &
             0.002348904241494349_dp, 0.007504570726957518_dp, 0.02113765842610693_dp, &
             0.05426601592774816_dp, 0.12966135357159264_dp, 0.2924534242934999_dp, &
             0.6291816606770987_dp, 1.3026475945125682_dp, 2.6251922772109966_dp, &
             5.302034557078597_dp, &
             ! w
             3.4038922791420766e-05_dp, 0.000220701484722969_dp, 0.0008991359743572639_dp, &
             0.0029149378829502514_dp, 0.008191276251697455_dp, 0.020819342283799648_dp, &
             0.049064011688295966_dp, 0.108945747440993_dp, 0.2305184102400087_dp, &
             0.46901297791286584_dp, 0.9271782595854164_dp, 1.8210023119996728_dp, &
             3.8714105759372774_dp]

      CASE (362)

         aw(:) = & ! a
            [1.0347391661194427e-05_dp, 0.0001138316914487579_dp, 0.0006024601498553105_dp, &
             0.0023252777266436543_dp, 0.007446434516969326_dp, 0.02100880434492859_dp, &
             0.05400321926097351_dp, 0.12916211533151045_dp, 0.2915643750527972_dp, &
             0.6276946181536831_dp, 1.3003150067055842_dp, 2.6217712688276107_dp, &
             5.297335744076939_dp, &
             ! w
             3.3078584812160216e-05_dp, 0.0002173321916008654_dp, 0.0008896788976872138_dp, &
             0.002892175605306567_dp, 0.008142234962777087_dp, 0.020722487658696522_dp, &
             0.04888643421469716_dp, 0.10864184956111259_dp, 0.2300332300492383_dp, &
             0.46829507330399833_dp, 0.9262055359666433_dp, 1.8198087510110939_dp, &
             3.8700490617962435_dp]

      CASE (363)

         aw(:) = & ! a
            [0.023793678248788978_dp, 0.12629699549283505_dp, 0.31461372374455443_dp, &
             0.595950838226835_dp, 0.9817872624012357_dp, 1.489306648805245_dp, &
             2.1434983518707615_dp, 2.98017722599483_dp, 4.050339314522109_dp, &
             5.426906118962724_dp, 7.216799744882609_dp, 9.58704299766279_dp, &
             12.835745300094228_dp, 17.67532495211082_dp, &
             ! w
             0.06116729495997983_dp, 0.14450858936283956_dp, 0.23330879114734795_dp, &
             0.33125541607836817_dp, 0.44324687555769954_dp, 0.5759073006895733_dp, &
             0.7383747431819776_dp, 0.9433758083290422_dp, 1.2089173154542734_dp, &
             1.5616542881502753_dp, 2.0450063722789924_dp, 2.741983515989405_dp, &
             3.85667343119404_dp, 6.198179189871694_dp]

      CASE (364)

         aw(:) = & ! a
            [0.014264102057225569_dp, 0.07596385221308057_dp, 0.19039853035057047_dp, &
             0.36406075206460514_dp, 0.6076945939410506_dp, 0.9380409019477502_dp, &
             1.3802978464353806_dp, 1.9713646649512173_dp, 2.7641415826304563_dp, &
             3.8339843611569067_dp, 5.290378851043468_dp, 7.3023268339139555_dp, &
             10.167121192094225_dp, 14.583797074551883_dp, &
             ! w
             0.036697133494631565_dp, 0.08727260903458155_dp, 0.14265932551810612_dp, &
             0.20642712177378034_dp, 0.28358973861827835_dp, 0.3812489108142731_dp, &
             0.5093454281471229_dp, 0.6815036880705535_dp, 0.9164311767479404_dp, &
             1.2411293792614433_dp, 1.6988916309365063_dp, 2.371479369041959_dp, &
             3.457751863425088_dp, 5.74024728694743_dp]

      CASE (365)

         aw(:) = & ! a
            [0.010469205433632786_dp, 0.055885282300705635_dp, 0.1406960623012082_dp, &
             0.2708801030466687_dp, 0.45658265580748264_dp, 0.713999068872431_dp, &
             1.0679417676762393_dp, 1.5550829406200728_dp, 2.228212327235486_dp, &
             3.1627623992935905_dp, 4.468652779969455_dp, 6.315645590979072_dp, &
             9.001211922079047_dp, 13.219418343427552_dp, &
             ! w
             0.02694857193241024_dp, 0.06439208283566383_dp, 0.10620510716864073_dp, &
             0.15582262426848184_dp, 0.21823871034834755_dp, 0.30067604733989817_dp, &
             0.4132531737378125_dp, 0.5697250681720737_dp, 0.7889239435759887_dp, &
             1.0981226631535514_dp, 1.5410283271380334_dp, 2.1995131318883794_dp, &
             3.2707215533727316_dp, 5.525548933707593_dp]

      CASE (366)

         aw(:) = & ! a
            [0.008376635467243343_dp, 0.04479919644102725_dp, 0.11319023446557304_dp, &
             0.21914567583183742_dp, 0.37233917623427065_dp, 0.5884689484069489_dp, &
             0.8918422272547547_dp, 1.318584467247321_dp, 1.9209270418492403_dp, &
             2.773912789743336_dp, 3.987478435091623_dp, 5.731868093843135_dp, &
             8.304953050914408_dp, 12.398249969934723_dp, &
             ! w
             0.02157140484578431_dp, 0.05173884593742352_dp, 0.08595433665197048_dp, &
             0.1275363375090656_dp, 0.18142368247815055_dp, 0.25483450865047297_dp, &
             0.3578645442824938_dp, 0.5042194995002234_dp, 0.7127864027157677_dp, &
             1.0111326193554142_dp, 1.4434560698218157_dp, 2.0919781635689985_dp, &
             3.153048296436189_dp, 5.390553588406537_dp]

      CASE (367)

         aw(:) = & ! a
            [0.007033979537092302_dp, 0.03767845776050892_dp, 0.09548945031954283_dp, &
             0.18576388023211604_dp, 0.31779274694968784_dp, 0.5068344028592785_dp, &
             0.7766828152420552_dp, 1.162849189524159_dp, 1.7169175230748923_dp, &
             2.5134448279597255_dp, 3.6622634753022947_dp, 5.333968785846338_dp, &
             7.8268675578464855_dp, 11.830958084696892_dp, &
             ! w
             0.018120390399748474_dp, 0.04360094729961982_dp, 0.07288184584725375_dp, &
             0.10918243318101181_dp, 0.15737533850088423_dp, 0.2246218843896931_dp, &
             0.32092402692893673_dp, 0.45989685263698105_dp, 0.6604755354010544_dp, &
             0.9505074044318574_dp, 1.3746503676746198_dp, 2.015510837239436_dp, &
             3.0690175985474153_dp, 5.294223339872018_dp]

      CASE (368)

         aw(:) = & ! a
            [0.006092281782467828_dp, 0.0326796427354352_dp, 0.08304320403684408_dp, &
             0.16223734107994103_dp, 0.27923353713325183_dp, 0.4489001056044332_dp, &
             0.6945418267921846_dp, 1.0510655723020723_dp, 1.5694187175490266_dp, &
             2.323670781635799_dp, 3.4235072283402714_dp, 5.039793906527383_dp, &
             7.471257554117014_dp, 11.406912447727011_dp, &
             ! w
             0.015699437791355096_dp, 0.03788173333538656_dp, 0.06366548182511249_dp, &
             0.09618473759480022_dp, 0.14024328897973412_dp, 0.20292278108490766_dp, &
             0.2941076299401058_dp, 0.42731861873832927_dp, 0.621538679846388_dp, &
             0.9048675733944339_dp, 1.3223740652817815_dp, 1.9570374697495083_dp, &
             3.004553995154926_dp, 5.22037367009797_dp]

      CASE (369)

         aw(:) = & ! a
            [0.005391735866188923_dp, 0.028957954603636153_dp, 0.07376354638484661_dp, &
             0.14466052258820267_dp, 0.25034735943968095_dp, 0.40534378380722114_dp, &
             0.6324982596023326_dp, 0.9661477674983822_dp, 1.4566423523519518_dp, &
             2.1775887642399563_dp, 3.238510386071566_dp, 4.810487992794182_dp, &
             7.192639119323395_dp, 11.073305447184763_dp, &
             ! w
             0.013898109410051639_dp, 0.03361953824395259_dp, 0.05677783232042905_dp, &
             0.0864325676208157_dp, 0.1273193360592727_dp, 0.18643070735698647_dp, &
             0.27352872709342707_dp, 0.40204660101994955_dp, 0.5910134830489486_dp, &
             0.868752818168269_dp, 1.2806972849123448_dp, 1.910175277355787_dp, &
             2.9527569215812304_dp, 5.161069674409225_dp]

      CASE (370)

         aw(:) = & ! a
            [0.00484825925160011_dp, 0.026068629301107003_dp, 0.06655004077408314_dp, &
             0.1309719715774268_dp, 0.22779534002542953_dp, 0.37122571954561345_dp, &
             0.5836889969345771_dp, 0.8989929313055354_dp, 1.3669355474682265_dp, &
             2.060691761648574_dp, 3.0896188502677946_dp, 4.624971917108687_dp, &
             6.966228180332507_dp, 10.80125317612083_dp, &
             ! w
             0.012500421264673638_dp, 0.030307691418891464_dp, 0.05141246143779468_dp, &
             0.07880839701021243_dp, 0.11716490991039365_dp, 0.1733827405509416_dp, &
             0.25710460680089164_dp, 0.3816850457961545_dp, 0.5661965543161639_dp, &
             0.8391601113814531_dp, 1.2463316648187273_dp, 1.8713633171352735_dp, &
             2.9097642414871805_dp, 5.111871222135591_dp]

      CASE (371)

         aw(:) = & ! a
            [0.004413168781924726_dp, 0.023754010752811715_dp, 0.06076457786681532_dp, &
             0.11997470973785124_dp, 0.20963557196151325_dp, 0.34366743099740693_dp, &
             0.5441054324421765_dp, 0.844268319686792_dp, 1.2934459449065359_dp, &
             1.9644120588566827_dp, 2.9663588112266797_dp, 4.470684882051529_dp, &
             6.7771978107945285_dp, 10.573418176382921_dp, &
             ! w
             0.011381304125104947_dp, 0.027652477612770014_dp, 0.04710099946812233_dp, &
             0.07266153181156448_dp, 0.1089398770918275_dp, 0.16274559652344747_dp, &
             0.2436082848337518_dp, 0.36481237099885855_dp, 0.5454701187758315_dp, &
             0.8142773411254591_dp, 1.217278942214352_dp, 1.8384268358954554_dp, &
             2.873211216315124_dp, 5.070060574632702_dp]

      CASE (372)

         aw(:) = & ! a
            [0.004056224537639509_dp, 0.021853974309257018_dp, 0.05601024565262235_dp, &
             0.11092329056913866_dp, 0.19465685777434072_dp, 0.3208703362477144_dp, &
             0.511237203549927_dp, 0.7986245539444243_dp, 1.2318541070578792_dp, &
             1.8833267887511218_dp, 2.8620729340610342_dp, 4.339610870617615_dp, &
             6.616051909273061_dp, 10.378663116498753_dp, &
             ! w
             0.010463061182274542_dp, 0.025471251874033252_dp, 0.043551694367213026_dp, &
             0.06758571090396688_dp, 0.10211824396619247_dp, 0.15387007964196028_dp, &
             0.2322649456760445_dp, 0.3505246753474143_dp, 0.5277975088592344_dp, &
             0.7929346987934134_dp, 1.1922412421604374_dp, 1.8099472513036932_dp, &
             2.8415520082506047_dp, 5.033861770652594_dp]

      CASE (373)

         aw(:) = & ! a
            [0.002318228823055896_dp, 0.012582373148663025_dp, 0.03271956185523208_dp, &
             0.06632553676179402_dp, 0.12025455907099865_dp, 0.20636944282486241_dp, &
             0.3438043885629948_dp, 0.5622945792714699_dp, 0.9074155937680068_dp, &
             1.4488825900310236_dp, 2.294363498102311_dp, 3.615926928952993_dp, &
             5.715774128586636_dp, 9.28056534370184_dp, &
             ! w
             0.0059897754697515_dp, 0.014799327003276113_dp, 0.02605247704571643_dp, &
             0.042272541370318134_dp, 0.06752646013008466_dp, 0.1078388793634545_dp, &
             0.17190192927606873_dp, 0.27254095421632174_dp, 0.4291033113493527_dp, &
             0.6713790246552778_dp, 1.047359969242586_dp, 1.6432750177593765_dp, &
             2.6552117669183684_dp, 4.8210711548907295_dp]

      CASE (374)

         aw(:) = & ! a
            [0.001666361328381412_dp, 0.009091717350004885_dp, 0.02389076500814641_dp, &
             0.049245714464806946_dp, 0.09133750366284342_dp, 0.16096835309780874_dp, &
             0.27577340973498193_dp, 0.46366223695930214_dp, 0.768288083326851_dp, &
             1.2576815493486098_dp, 2.0385090183531704_dp, 3.282955139328998_dp, &
             5.294434037924235_dp, 8.75989883375243_dp, &
             ! w
             0.00431050555030777_dp, 0.010762996303283687_dp, 0.019344699253129392_dp, &
             0.03236862423762432_dp, 0.05358025504012204_dp, 0.08856507291533267_dp, &
             0.14560785425842068_dp, 0.23730641760902718_dp, 0.38305910746031374_dp, &
             0.6131010801761819_dp, 0.9763482420250424_dp, 1.5602761224141384_dp, &
             2.561660651628839_dp, 4.714407504454591_dp]

      CASE (375)

         aw(:) = & ! a
            [0.001317125050908727_dp, 0.007216593069539949_dp, 0.01912487995618562_dp, &
             0.03995618721778877_dp, 0.07543526951836034_dp, 0.13563310227543893_dp, &
             0.2371598253031562_dp, 0.40667738866120773_dp, 0.686500757576153_dp, &
             1.1434436672367445_dp, 1.883389915309971_dp, 3.07851501265906_dp, &
             5.033059170837773_dp, 8.434387218341376_dp, &
             ! w
             0.0034102793689438035_dp, 0.00858770676444383_dp, 0.015694822236987458_dp, &
             0.02689698752610825_dp, 0.045704926161504116_dp, 0.07740221972288289_dp, &
             0.13000398878108615_dp, 0.21594125807286957_dp, 0.3546111560919621_dp, &
             0.5765112081449788_dp, 0.9311714010393723_dp, 1.5069617811410725_dp, &
             2.501262800806158_dp, 4.645597053062268_dp]

      CASE (376)

         aw(:) = & ! a
            [0.0010970334583767882_dp, 0.00603234068032961_dp, 0.016103087843071007_dp, &
             0.03402965537383759_dp, 0.06519718711042698_dp, 0.1191293407601384_dp, &
             0.2116738402433356_dp, 0.36856118641837915_dp, 0.6310911840532925_dp, &
             1.0651302197216859_dp, 1.7759219386721268_dp, 2.935585417502371_dp, &
             4.848974717177059_dp, 8.203864682046829_dp, &
             ! w
             0.002842666311093802_dp, 0.007210324160903711_dp, 0.013365727519587825_dp, &
             0.0233612862120285_dp, 0.040526365073116286_dp, 0.06992153045699963_dp, &
             0.11936365985825013_dp, 0.2011495081893048_dp, 0.3346541988448492_dp, &
             0.5505486506765955_dp, 0.8988127429424152_dp, 1.4685085046890018_dp, &
             2.457538764929661_dp, 4.595806077787992_dp]

      CASE (377)

         aw(:) = & ! a
            [0.0009445997794898478_dp, 0.00521065038284475_dp, 0.013999399805948753_dp, &
             0.029881653864532533_dp, 0.057975120278581475_dp, 0.10737211368193411_dp, &
             0.19332191162222_dp, 0.34082012830256314_dp, 0.5903549288936565_dp, &
             1.0070203076872235_dp, 1.695519386926231_dp, 2.827895180373993_dp, &
             4.709483364001195_dp, 8.028441638713057_dp, &
             ! w
             0.002449376298727717_dp, 0.0062525417642426495_dp, 0.011735380214284632_dp, &
             0.02085937855571093_dp, 0.036808222546315036_dp, 0.0644690856910467_dp, &
             0.11150331075715983_dp, 0.19009445632931343_dp, 0.3195868584534787_dp, &
             0.530774348128065_dp, 0.8739864827484358_dp, 1.4388462588116058_dp, &
             2.4237120872806432_dp, 4.557297766285848_dp]

      CASE (378)

         aw(:) = & ! a
            [0.0008322695037266466_dp, 0.0046041779357021044_dp, 0.01244213757850956_dp, &
             0.026796443186325423_dp, 0.052566141462612966_dp, 0.09849126392519111_dp, &
             0.17933349904469312_dp, 0.31948641371298403_dp, 0.5587657706361328_dp, &
             0.9616156945673737_dp, 1.632272275282876_dp, 2.7426945317859523_dp, &
             4.598610352054523_dp, 7.888529242760485_dp, &
             ! w
             0.0021594492963257278_dp, 0.005544274203005808_dp, 0.010522682293089395_dp, &
             0.01898044963568778_dp, 0.0339808497464532_dp, 0.06027111111097718_dp, &
             0.10538515841359175_dp, 0.1814084463685946_dp, 0.3076509595262928_dp, &
             0.5149978645662089_dp, 0.8540615398602766_dp, 1.4149345700297018_dp, &
             2.3963776420823_dp, 4.526187182673834_dp]

      CASE (379)

         aw(:) = & ! a
            [0.0007457708419311654_dp, 0.004136510586912845_dp, 0.011238112552192312_dp, &
             0.024400748456895504_dp, 0.04833979007819415_dp, 0.09149991383771948_dp, &
             0.16823452346758364_dp, 0.3024303189839189_dp, 0.5333315311260183_dp, &
             0.9248228623483392_dp, 1.5807297911277949_dp, 2.6729249179396786_dp, &
             4.5074643277946995_dp, 7.773179960349463_dp, &
             ! w
             0.0019361209248683394_dp, 0.004997180620004319_dp, 0.009580991751005498_dp, &
             0.017508740203720244_dp, 0.03174197023294575_dp, 0.05691172464721022_dp, &
             0.10044431634744969_dp, 0.17433856609666365_dp, 0.2978690083437283_dp, &
             0.5019910235145302_dp, 0.8375524167267339_dp, 1.3950481721402253_dp, &
             2.3735982301458436_dp, 4.500265207123504_dp]

      CASE (380)

         aw(:) = & ! a
            [0.0006769402078263541_dp, 0.0037638910093807645_dp, 0.010276475257263445_dp, &
             0.022479744521262293_dp, 0.044931604380811674_dp, 0.08582407353530463_dp, &
             0.15916145186976766_dp, 0.28839478925987994_dp, 0.5122730073839316_dp, &
             0.8941908057078767_dp, 1.5376077062077285_dp, 2.614310174320446_dp, &
             4.430634642330474_dp, 7.675709147970693_dp, &
             ! w
             0.0017583561916395855_dp, 0.004560605163462098_dp, 0.008825895255928024_dp, &
             0.016319294769126736_dp, 0.029914903195384646_dp, 0.05414515705762229_dp, &
             0.09634342234963152_dp, 0.1684309556033988_dp, 0.28964696629880016_dp, &
             0.491002198766957_dp, 0.8235447309834687_dp, 1.3781205638529523_dp, &
             2.3541736567668394_dp, 4.478163654314514_dp]

      CASE (381)

         aw(:) = & ! a
            [0.0006207557929943959_dp, 0.0034593749933095317_dp, 0.00948884518046029_dp, &
             0.02090054667171633_dp, 0.0421152120539184_dp, 0.08110525250097157_dp, &
             0.15157149669631095_dp, 0.2765844306591898_dp, 0.4944571460594124_dp, &
             0.8681492267517379_dp, 1.500790443602208_dp, 2.564082798645993_dp, &
             4.364606150011944_dp, 7.5917614754962415_dp, &
             ! w
             0.0016132124497670363_dp, 0.00420331796732863_dp, 0.00820516814263875_dp, &
             0.015334374359018485_dp, 0.028388806800141737_dp, 0.0518157306160421_dp, &
             0.0928668402052448_dp, 0.16339320159873136_dp, 0.2825994389709872_dp, &
             0.48154081648437635_dp, 0.8114387138683526_dp, 1.3634496747327067_dp, &
             2.3373123988437143_dp, 4.458980609708005_dp]

      CASE (382)

         aw(:) = & ! a
            [0.0003510368022941439_dp, 0.001991370297862904_dp, 0.005660955552562638_dp, &
             0.013120833766572953_dp, 0.027978805324674217_dp, 0.05691665897610339_dp, &
             0.11184337459575114_dp, 0.21354011284411326_dp, 0.39763380584307784_dp, &
             0.7243220713184966_dp, 1.2945431454910699_dp, 2.2792966515798136_dp, &
             3.986588778702906_dp, 7.10774102165579_dp, &
             ! w
             0.0009157569104258144_dp, 0.002472193770080037_dp, 0.005147847078762964_dp, &
             0.010354088235585797_dp, 0.020440799298929627_dp, 0.03936088428275196_dp, &
             0.07385970672203142_dp, 0.1353151948841132_dp, 0.24264830676571894_dp, &
             0.42710268361778675_dp, 0.7409052083663801_dp, 1.27715698373519_dp, &
             2.23760787462859_dp, 4.345574578366373_dp]

      CASE (383)

         aw(:) = & ! a
            [0.00025163195510559877_dp, 0.0014463966580943882_dp, 0.00421928634217767_dp, &
             0.010119565206226807_dp, 0.022352100147247436_dp, 0.04696516665200164_dp, &
             0.09497611733572855_dp, 0.18599495643301286_dp, 0.3542301980022189_dp, &
             0.6583664765398028_dp, 1.1980745909205335_dp, 2.1438544125498438_dp, &
             3.804408366070667_dp, 6.872224936180135_dp, &
             ! w
             0.0006582780968990025_dp, 0.0018238955102013433_dp, 0.0039689243401756155_dp, &
             0.00834715035690808_dp, 0.017091734724666522_dp, 0.03391225628989364_dp, &
             0.06528123379627732_dp, 0.12229910480859243_dp, 0.22368972347688337_dp, &
             0.40073684144048183_dp, 0.7061527331295446_dp, 1.234083194357158_dp, &
             2.187472583120084_dp, 4.288560018892773_dp]

      CASE (384)

         aw(:) = & ! a
            [0.00019879789350306175_dp, 0.001155266613227247_dp, 0.0034409962587201824_dp, &
             0.008471121868838132_dp, 0.01919615146362964_dp, 0.04126544677289401_dp, &
             0.08512799329935912_dp, 0.16963442951218227_dp, 0.3280570368244356_dp, &
             0.6180615250656186_dp, 1.138440292460392_dp, 2.0593131981714614_dp, &
             3.689818329462071_dp, 6.723268180635575_dp, &
             ! w
             0.0005212677491514955_dp, 0.001475424141415463_dp, 0.003321571161298502_dp, &
             0.007212088049560924_dp, 0.015145745290034129_dp, 0.030676415146612054_dp, &
             0.06009357511543142_dp, 0.11430448144640727_dp, 0.21188560202242127_dp, &
             0.38412470507388513_dp, 0.6840362336317067_dp, 1.2064609599820673_dp, &
             2.155181281852188_dp, 4.25183836321432_dp]

      CASE (385)

         aw(:) = & ! a
            [0.00016565509837365063_dp, 0.0009719040519362572_dp, 0.0029465717920729977_dp, &
             0.00740937623356772_dp, 0.01713075134685372_dp, 0.03747754007118164_dp, &
             0.07849213724446259_dp, 0.15847540973525662_dp, 0.3100135420433681_dp, &
             0.5900153839041758_dp, 1.0966083910583349_dp, 1.999608115087129_dp, &
             3.608459123387142_dp, 6.617103943396409_dp, &
             ! w
             0.0004352427581767984_dp, 0.0012548576180906005_dp, 0.0029046595378673015_dp, &
             0.006464558611470215_dp, 0.013839272006944079_dp, 0.028470484166535776_dp, &
             0.05651194178611248_dp, 0.10872444766731394_dp, 0.2035678909477675_dp, &
             0.3723216049428288_dp, 0.6682119500204294_dp, 1.1865913419286378_dp, &
             2.1318813562162844_dp, 4.225340930493407_dp]

      CASE (386)

         aw(:) = & ! a
            [0.00014276954290540903_dp, 0.0008448608471347978_dp, 0.0026014627740257174_dp, &
             0.006659644816094836_dp, 0.015653379916881753_dp, 0.034735064921288035_dp, &
             0.07363590112605035_dp, 0.1502321267433044_dp, 0.296575178434725_dp, &
             0.5689780633687209_dp, 1.0650374633237811_dp, 1.9543163700723007_dp, &
             3.5464904138886957_dp, 6.536008202050467_dp, &
             ! w
             0.0003757964231191714_dp, 0.0011013938591955834_dp, 0.0026102631943494842_dp, &
             0.005927146483372808_dp, 0.01288600465322186_dp, 0.026841997756471673_dp, &
             0.0538421401369422_dp, 0.1045302971427401_dp, 0.19727052443603674_dp, &
             0.36332888268997277_dp, 0.6560911744653187_dp, 1.1713098067261307_dp, &
             2.1139192519072916_dp, 4.204913003426183_dp]

      CASE (387)

         aw(:) = & ! a
            [0.00012593990399240612_dp, 0.0007511606113890969_dp, 0.0023452526412624827_dp, &
             0.006097436900567215_dp, 0.014533480971501483_dp, 0.032635362170355266_dp, &
             0.06988524500770414_dp, 0.143817073802175_dp, 0.2860481254040681_dp, &
             0.55240381972184_dp, 1.0400418005232401_dp, 1.9183099739324458_dp, &
             3.4970663804887807_dp, 6.471180050842394_dp, &
             ! w
             0.00033205236024469117_dp, 0.0009877883512064986_dp, 0.0023894838975166595_dp, &
             0.005518054937658007_dp, 0.012151608899108974_dp, 0.02557555583391717_dp, &
             0.051749652677722136_dp, 0.10122106011972111_dp, 0.19227284584225376_dp, &
             0.35615595907222597_dp, 0.6463818938212137_dp, 1.1590285100821416_dp, &
             2.0994562104528667_dp, 4.188463742112101_dp]

      CASE (388)

         aw(:) = & ! a
            [0.00011299976677305081_dp, 0.0006789271013530575_dp, 0.0021465653166690152_dp, &
             0.005657578532329506_dp, 0.01364910453916546_dp, 0.03096318849755199_dp, &
             0.06687628439022514_dp, 0.1386378928877156_dp, 0.2775023855512693_dp, &
             0.53888499004393_dp, 1.0195707670958232_dp, 1.8887208600165935_dp, &
             3.4563421263845617_dp, 6.417661514462016_dp, &
             ! w
             0.00029839899635634423_dp, 0.0008999186728744625_dp, 0.00221672913738762_dp, &
             0.005193839750728901_dp, 0.011563728001605426_dp, 0.024553799052682058_dp, &
             0.050050465373926975_dp, 0.098518862889518_dp, 0.18817218892600346_dp, &
             0.3502457688554335_dp, 0.6383535275892052_dp, 1.148845805752548_dp, &
             2.0874455976907558_dp, 4.174803092036887_dp]

      CASE (389)

         aw(:) = & ! a
            [0.00010271406841920615_dp, 0.0006213754330576075_dp, 0.0019873999833586_dp, &
             0.00530241009100983_dp, 0.012929154766809487_dp, 0.029591932684118336_dp, &
             0.06439320066489618_dp, 0.13434064202061552_dp, 0.270378577477527_dp, &
             0.527569888660758_dp, 1.0023772597521854_dp, 1.863797196012145_dp, &
             3.4219610092511665_dp, 6.372406081992173_dp, &
             ! w
             0.0002716356155158933_dp, 0.0008296965023838829_dp, 0.002077213899147062_dp, &
             0.004929082317867688_dp, 0.011079530851376154_dp, 0.023706599039416767_dp, &
             0.04863374933347189_dp, 0.09625518934642596_dp, 0.18472285946332168_dp, &
             0.3452565701047218_dp, 0.6315558344955864_dp, 1.1402040860907414_dp, &
             2.0772388473542316_dp, 4.163193618228598_dp]

      CASE (390)

         aw(:) = & ! a
            [9.432504161652162e-05_dp, 0.0005743348523820048_dp, 0.0018566487446835668_dp, &
             0.005008541893268789_dp, 0.012329127593858291_dp, 0.02844172371620226_dp, &
             0.06229888069160124_dp, 0.13069903077317394_dp, 0.26431705530112054_dp, &
             0.5179082172316746_dp, 0.9876519994633036_dp, 1.842398000470507_dp, &
             3.392383628136573_dp, 6.333419494072734_dp, &
             ! w
             0.00024979755372032573_dp, 0.0007721386329534672_dp, 0.0019617570439813717_dp, &
             0.0047078208097291975_dp, 0.010671852299116087_dp, 0.022989121646123302_dp, &
             0.04742817482302329_dp, 0.09432095411354324_dp, 0.18176499644466768_dp, &
             0.3409650001887979_dp, 0.6256934039371225_dp, 1.132736399634179_dp, &
             2.0684084060135146_dp, 4.153149177338219_dp]

      CASE (391)

         aw(:) = & ! a
            [5.411491253465531e-05_dp, 0.00034712951476670404_dp, 0.0012134406295876315_dp, &
             0.0035264446522853933_dp, 0.0092288791875334_dp, 0.022372536105391687_dp, &
             0.05104857330359674_dp, 0.11083599543130185_dp, 0.2308180075705037_dp, &
             0.4639056583092208_dp, 0.904547009867386_dp, 1.720649638824277_dp, &
             3.223037137645256_dp, 6.1091989539711875_dp, &
             ! w
             0.00014496785037750894_dp, 0.0004912763971936477_dp, 0.001378902188720698_dp, &
             0.0035543625962753084_dp, 0.008495076648918189_dp, 0.019085692482120022_dp, &
             0.04076655101195879_dp, 0.08349013523408717_dp, 0.16501058496368268_dp, &
             0.3164121765948108_dp, 0.591869702979274_dp, 1.0893701786456509_dp, &
             2.0169320276678824_dp, 4.094586174432079_dp]

      CASE (392)

         aw(:) = & ! a
            [3.9300884986618324e-05_dp, 0.0002623153389586003_dp, 0.0009655315853114956_dp, &
             0.002932105025064018_dp, 0.007940134743043187_dp, 0.019772649033196633_dp, &
             0.04610736632184178_dp, 0.1019272143223431_dp, 0.21552364928462556_dp, &
             0.4388730436176661_dp, 0.8655243916325448_dp, 1.6628681648767378_dp, &
             3.141992361576881_dp, 6.001262340418874_dp, &
             ! w
             0.00010626070791558328_dp, 0.00038450847708988684_dp, 0.0011446716833487984_dp, &
             0.0030688076099272287_dp, 0.007547733414566504_dp, 0.01734252518854569_dp, &
             0.03772804497713254_dp, 0.0784608084602891_dp, 0.15711015119496588_dp, &
             0.3046803290832214_dp, 0.5755275193529117_dp, 1.0682374296805368_dp, &
             1.991720174776034_dp, 4.065895935435006_dp]

      CASE (393)

         aw(:) = & ! a
            [3.142101757888054e-05_dp, 0.00021678515541478832_dp, 0.0008294336079992923_dp, &
             0.0025975470555735874_dp, 0.0071990446914907535_dp, 0.018251475926555048_dp, &
             0.04317507971664544_dp, 0.09657783357055981_dp, 0.20624840209242398_dp, &
             0.42356357914236825_dp, 0.8414881261873081_dp, 1.6270668754178648_dp, &
             3.0915458270867107_dp, 5.933860549831819_dp, &
             ! w
             8.564695144358155e-05_dp, 0.00032643277553823013_dp, 0.001012593835495053_dp, &
             0.002787620676381317_dp, 0.006988737096983311_dp, 0.016298896981613548_dp, &
             0.0358871886029567_dp, 0.07538326056743468_dp, 0.15223427026439756_dp, &
             0.29738659791331695_dp, 0.5653048703970436_dp, 1.0549552392198862_dp, &
             1.9758295507429005_dp, 4.047809842991139_dp]

      CASE (394)

         aw(:) = & ! a
            [2.6474838971722706e-05_dp, 0.0001879969436971919_dp, 0.0007418596247248935_dp, &
             0.0023783053577616318_dp, 0.0067060613090639725_dp, 0.017227402137173373_dp, &
             0.041181819263646564_dp, 0.09291231300375254_dp, 0.19984996655960596_dp, &
             0.4129422999765327_dp, 0.8247322320497623_dp, 1.6020104793093735_dp, &
             3.05613055764379_dp, 5.8864401090495555_dp, &
             ! w
             7.26984225218588e-05_dp, 0.0002893195987653552_dp, 0.0009259204579195344_dp, &
             0.002599686145099633_dp, 0.006610308897826231_dp, 0.015585353790063133_dp, &
             0.03461837903925443_dp, 0.07324767730849191_dp, 0.14883123612108737_dp, &
             0.29227091156461876_dp, 0.5581052043607662_dp, 1.0455708717528165_dp, &
             1.9645808458523943_dp, 4.035005340705382_dp]

      CASE (395)

         aw(:) = & ! a
            [2.3058410580510703e-05_dp, 0.0001679900094432579_dp, 0.0006801140639941704_dp, &
             0.002221505974928107_dp, 0.006349440376324858_dp, 0.016479905649325548_dp, &
             0.03971633009346449_dp, 0.09020126203347086_dp, 0.1950940150662858_dp, &
             0.405014251405954_dp, 0.8121807184817234_dp, 1.5831863205703816_dp, &
             3.0294635397994294_dp, 5.850676859825858_dp, &
             ! w
             6.375060493842179e-05_dp, 0.000263293987103535_dp, 0.0008638617226868241_dp, &
             0.0024632632619747065_dp, 0.006332963662235677_dp, 0.015058530391730383_dp, &
             0.033675957456170844_dp, 0.07165348269041043_dp, 0.14628005112992484_dp, &
             0.28842179110719324_dp, 0.5526715091664561_dp, 1.0384716324095022_dp, &
             1.956059273582706_dp, 4.025304212199482_dp]

      CASE (396)

         aw(:) = & ! a
            [2.054611343465688e-05_dp, 0.00015319900277873115_dp, 0.0006339033028677352_dp, &
             0.0021027869766665195_dp, 0.0060769589706749635_dp, 0.015904694498627883_dp, &
             0.03858217339550542_dp, 0.08809333525543948_dp, 0.1913816774420092_dp, &
             0.39880551631622985_dp, 0.8023239547983557_dp, 1.5683698969982658_dp, &
             3.008436820114962_dp, 5.822443073454802_dp, &
             ! w
             5.716893018939122e-05_dp, 0.0002439023638760942_dp, 0.00081682963820073_dp, &
             0.0023587466952448306_dp, 0.0061188749018843535_dp, 0.014649494795265394_dp, &
             0.032940791327922415_dp, 0.07040498978480411_dp, 0.14427542711723315_dp, &
             0.2853886853964313_dp, 0.5483795422191684_dp, 1.0328537538974532_dp, &
             1.949308444668341_dp, 4.017618333143779_dp]

      CASE (397)

         aw(:) = & ! a
            [1.8615070842341552e-05_dp, 0.0001417759283012718_dp, 0.0005978334511040867_dp, &
             0.00200921574748141_dp, 0.005860576829904252_dp, 0.015445239861383164_dp, &
             0.03767203736127576_dp, 0.08639533327067402_dp, 0.18838179696402568_dp, &
             0.39377495810645957_dp, 0.7943197190147671_dp, 1.5563159555568638_dp, &
             2.991305970347473_dp, 5.799417598842914_dp, &
             ! w
             5.2109167115011715e-05_dp, 0.00022882219912923564_dp, 0.0007797299594009687_dp, &
             0.0022755684800263992_dp, 0.005947441689789478_dp, 0.01432039943309562_dp, &
             0.03234703367047513_dp, 0.06939342662672476_dp, 0.1426468346138291_dp, &
             0.2829188574091156_dp, 0.5448778979814616_dp, 1.0282635227986774_dp, &
             1.9437875950719539_dp, 4.01133238422056_dp]

      CASE (398)

         aw(:) = & ! a
            [1.7081046562876867e-05_dp, 0.00013266232049979044_dp, 0.0005687853175852535_dp, &
             0.0019332310575148547_dp, 0.005683745218129161_dp, 0.015067924260827526_dp, &
             0.03692170023011739_dp, 0.08499101296010061_dp, 0.18589421020436803_dp, &
             0.38959420630714_dp, 0.787655216417328_dp, 1.5462641811150883_dp, &
             2.97700354361372_dp, 5.780177883924808_dp, &
             ! w
             4.808929949886569e-05_dp, 0.00021671595392277037_dp, 0.0007495819023385534_dp, &
             0.002207471801872518_dp, 0.005806365989541935_dp, 0.01404850497867388_dp, &
             0.03185490574916339_dp, 0.0685527727285889_dp, 0.14129035145734162_dp, &
             0.28085774823845083_dp, 0.5419510362511929_dp, 1.0244220015019012_dp, &
             1.9391638143706962_dp, 4.006067527339242_dp]

      CASE (399)

         aw(:) = & ! a
            [1.583091392566193e-05_dp, 0.00012520603813060788_dp, 0.0005448204397084741_dp, &
             0.0018700887819500696_dp, 0.005535996249593052_dp, 0.014751340367424074_dp, &
             0.036290041440967064_dp, 0.08380560818355155_dp, 0.1837896860325149_dp, &
             0.38605057200929904_dp, 0.7819973916253243_dp, 1.53771962729618_dp, &
             2.964833415574557_dp, 5.7637950604709385_dp, &
             ! w
             4.481321399041367e-05_dp, 0.0002067554141339402_dp, 0.0007245138992999495_dp, &
             0.0021504886393591276_dp, 0.005687791620873926_dp, 0.013819201766022637_dp, &
             0.031438733990656526_dp, 0.06784025662859389_dp, 0.14013843256663194_dp, &
             0.2791046160922478_dp, 0.5394581270165297_dp, 1.0211465978723961_dp, &
             1.9352189367908148_dp, 4.001575478322477_dp]

      CASE (400)

         aw(:) = & ! a
            [9.879098487133625e-06_dp, 8.921120682367847e-05_dp, 0.00042591995967115117_dp, &
             0.0015496891813780627_dp, 0.004773671233411263_dp, 0.01309699436300676_dp, &
             0.03295591404984426_dp, 0.07749731631973529_dp, 0.1725140716994954_dp, &
             0.36695624751661815_dp, 0.7513652468643365_dp, 1.4912765733052293_dp, &
             2.8984822282192_dp, 5.674287619811761_dp, &
             ! w
             2.9217197910873096e-05_dp, 0.00015774239553229696_dp, 0.0005970685206904868_dp, &
             0.00185517048136989_dp, 0.0050649977665210314_dp, 0.012602399661328837_dp, &
             0.02921201049634923_dp, 0.06400179605954581_dp, 0.1338968878030112_dp, &
             0.26955866677335194_dp, 0.5258281092343448_dp, 1.0031812141289602_dp, &
             1.9135401620880086_dp, 3.976885986963906_dp]

      CASE (401)

         aw(:) = & ! a
            [7.753277327156187e-06_dp, 7.605204330358708e-05_dp, 0.0003806666145700972_dp, &
             0.0014239784197318734_dp, 0.004467972401590203_dp, 0.012422665107787384_dp, &
             0.031579448787572406_dp, 0.07486609461124191_dp, 0.1677709581830264_dp, &
             0.35886717527466633_dp, 0.7383113916383872_dp, 1.4713888415525989_dp, &
             2.869962763920454_dp, 5.635715122518086_dp, &
             ! w
             2.3648127307189253e-05_dp, 0.00013928634487856857_dp, 0.0005469375253764436_dp, &
             0.0017360956451197699_dp, 0.004809547223882456_dp, 0.012096771438445859_dp, &
             0.028277071837355484_dp, 0.062376320873386674_dp, 0.13123476221452038_dp, &
             0.26546235450160005_dp, 0.5199495630844788_dp, 0.9954024569825356_dp, &
             1.9041314494641703_dp, 3.9661684959592667_dp]

      CASE (402)

         aw(:) = & ! a
            [6.696075236885778e-06_dp, 6.940993090965878e-05_dp, 0.00035732186352038304_dp, &
             0.0013581310319656613_dp, 0.004306130142902661_dp, 0.01206283349398925_dp, &
             0.03084043432481198_dp, 0.07344644767858383_dp, 0.16520150343654666_dp, &
             0.3544703630207691_dp, 0.7311960172904285_dp, 1.4605235201037703_dp, &
             2.8543539073991235_dp, 5.614578248780441_dp, &
             ! w
             2.08782574912506e-05_dp, 0.00012981230044218096_dp, 0.0005206466240105108_dp, &
             0.0016728965705399956_dp, 0.004672840820625864_dp, 0.011824481632581034_dp, &
             0.027771081411684398_dp, 0.06149301612074717_dp, 0.12978317872441927_dp, &
             0.2632222906375393_dp, 0.5167271419666196_dp, 0.9911304562174131_dp, &
             1.8989584974329314_dp, 3.960275403412971_dp]

      CASE (403)

         aw(:) = & ! a
            [6.120260181978378e-06_dp, 6.57560511305431e-05_dp, 0.00034430982484925335_dp, &
             0.0013211024594849182_dp, 0.004214562828893384_dp, 0.01185833139217857_dp, &
             0.030418971726480306_dp, 0.07263456481617418_dp, 0.1637287034613831_dp, &
             0.35194534500283137_dp, 0.7271033023740112_dp, 1.4542657643355081_dp, &
             2.845355161303602_dp, 5.6023840927620085_dp, &
             ! w
             1.936916825247699e-05_dp, 0.00012454582047396193_dp, 0.0005058519763189434_dp, &
             0.0016370898195280585_dp, 0.004595022046215853_dp, 0.011668932264647821_dp, &
             0.027481213087688194_dp, 0.060985829088612185_dp, 0.12894808457200949_dp, &
             0.26193148531711846_dp, 0.5148677455824445_dp, 0.9886628439296682_dp, &
             1.8959685802046633_dp, 3.9568690606905945_dp]

      CASE (404)

         aw(:) = & ! a
            [5.836358908510162e-06_dp, 6.394342296992747e-05_dp, 0.0003378055982068063_dp, &
             0.0013025006111422447_dp, 0.004168404779948954_dp, 0.01175498460808096_dp, &
             0.03020556908165388_dp, 0.07222284103882821_dp, 0.16298086688612895_dp, &
             0.3506618758582416_dp, 0.7250211411664764_dp, 1.451079851462804_dp, &
             2.8407712217760275_dp, 5.59617003566072_dp, &
             ! w
             1.862490515335481e-05_dp, 0.00012191714741546814_dp, 0.000498416832147413_dp, &
             0.0016190262440005966_dp, 0.004555660808632603_dp, 0.01159009815579477_dp, &
             0.027334074828345777_dp, 0.06072804982782488_dp, 0.1285231908735246_dp, &
             0.2612741320539892_dp, 0.513920118410828_dp, 0.9874045096933182_dp, &
             1.8944433646506675_dp, 3.9551313639097376_dp]

      CASE (405)

         aw(:) = & ! a
            [5.779310360886807e-06_dp, 6.357821325660413e-05_dp, 0.00033649100160858427_dp, &
             0.0012987332474693207_dp, 0.0041590436292552775_dp, 0.01173400402932722_dp, &
             0.03016221209209259_dp, 0.07213913881528675_dp, 0.16282875663814733_dp, &
             0.35040070791304617_dp, 0.724597300191373_dp, 1.4504311395006662_dp, &
             2.839837629245966_dp, 5.5949042429760905_dp, &
             ! w
             1.847532346082452e-05_dp, 0.00012138616450563895_dp, 0.0004969107802932144_dp, &
             0.0016153616951963146_dp, 0.004547667195500915_dp, 0.011574075510211794_dp, &
             0.027304150630042847_dp, 0.06067559706847823_dp, 0.1284366973240137_dp, &
             0.2611402694575219_dp, 0.5137270825249222_dp, 0.9871481159900667_dp, &
             1.8941325485709735_dp, 3.9547772395130827_dp]

      CASE (406)

         aw(:) = & ! a
            [0.022236374209981875_dp, 0.11791766919744567_dp, 0.2932181611149397_dp, &
             0.5539290777205831_dp, 0.9091507799137611_dp, 1.372290803221144_dp, &
             1.9625022213863008_dp, 2.7067233856900885_dp, 3.642533914113716_dp, &
             4.822283592881374_dp, 6.319658729817172_dp, 8.241777536872023_dp, &
             10.755677335590809_dp, 14.160475342580211_dp, 19.175063516833696_dp, &
             ! w
             0.057151196247983434_dp, 0.13476098203537124_dp, 0.21679241734108728_dp, &
             0.30613005549755173_dp, 0.4065242973198528_dp, 0.5229143871001193_dp, &
             0.6619515007536857_dp, 0.8326876570348027_dp, 1.0475551694593583_dp, &
             1.3240406119874788_dp, 1.6881745466647176_dp, 2.1829224353081296_dp, &
             2.8914979429762138_dp, 4.020155735554085_dp, 6.389664773839351_dp]

      CASE (407)

         aw(:) = & ! a
            [0.013319986361304756_dp, 0.07083579672152751_dp, 0.17707544003870992_dp, &
             0.3372094560469261_dp, 0.5596497412025824_dp, 0.8572651840652747_dp, &
             1.249100413601347_dp, 1.7626491360944359_dp, 2.436770909892311_dp, &
             3.3256722098160103_dp, 4.50520271619247_dp, 6.0845927692710795_dp, &
             8.232190523166691_dp, 11.24526638554403_dp, 15.827490930567276_dp, &
             ! w
             0.03425707558406306_dp, 0.08123895745525056_dp, 0.13208806420884975_dp, &
             0.1895620574511708_dp, 0.25743475526325454_dp, 0.3409324413871414_dp, &
             0.44727250246519884_dp, 0.5862378479869452_dp, 0.7709416195632693_dp, &
             1.0193233364053462_dp, 1.3575648828950586_dp, 1.8283241651243347_dp, &
             2.513297129714605_dp, 3.613022574802447_dp, 5.920293613177617_dp]

      CASE (408)

         aw(:) = & ! a
            [0.009770683987955369_dp, 0.052065600250376584_dp, 0.13064783675261413_dp, &
             0.2502493155231965_dp, 0.41874801446408033_dp, 0.6484848684705119_dp, &
             0.9580829940330154_dp, 1.374754414402287_dp, 1.9371818079908711_dp, &
             2.6994958848433686_dp, 3.7376744793080214_dp, 5.161396389003349_dp, &
             7.139628574735659_dp, 9.969393775979658_dp, 14.348477927481863_dp, &
             ! w
             0.025140460599849_dp, 0.05986150445520789_dp, 0.09807687730060566_dp, &
             0.14241735374966083_dp, 0.19660642495789066_dp, 0.2659327500342726_dp, &
             0.3577519270311488_dp, 0.48197465447938836_dp, 0.6518014776973908_dp, &
             0.8853077937671971_dp, 1.208952690788188_dp, 1.6657167955120669_dp, &
             2.3371854613987044_dp, 3.421831279185492_dp, 5.700050886680071_dp]

      CASE (409)

         aw(:) = & ! a
            [0.007814083335386305_dp, 0.0417064349886099_dp, 0.10497252408020417_dp, &
             0.20202015970696996_dp, 0.34031535059655627_dp, 0.5317496824765323_dp, &
             0.7944899302065639_dp, 1.1552671367902654_dp, 1.6522434451894115_dp, &
             2.3390686809253647_dp, 3.291457699350984_dp, 4.619206665040386_dp, &
             6.491733402413081_dp, 9.206204449567815_dp, 13.45724445221323_dp, &
             ! w
             0.020113453882336103_dp, 0.04804689001617416_dp, 0.07920474912528992_dp, &
             0.11611217015946855_dp, 0.16242908854580934_dp, 0.22342941349560405_dp, &
             0.3064598852976081_dp, 0.4213902130844776_dp, 0.5814003222278704_dp, &
             0.8046842905645214_dp, 1.118016679838287_dp, 1.5647904440413158_dp, &
             2.2267586807525968_dp, 3.301329533845546_dp, 5.561372468379086_dp]

      CASE (410)

         aw(:) = & ! a
            [0.0065589737028395635_dp, 0.03505508535410041_dp, 0.08845958576104579_dp, &
             0.17092819133140455_dp, 0.2895971214269864_dp, 0.45597509431125866_dp, &
             0.6877928532962936_dp, 1.0112643890368855_dp, 1.46396209397375_dp, &
             2.098978792068047_dp, 2.9916804481492303_dp, 4.251871991065483_dp, &
             6.049327218771119_dp, 8.681471187710688_dp, 12.84095643519424_dp, &
             ! w
             0.016888050445770582_dp, 0.040452365710777934_dp, 0.0670338169082043_dp, &
             0.0990695634544053_dp, 0.1401553278525078_dp, 0.19551843921678796_dp, &
             0.2724397623186262_dp, 0.38069885879339954_dp, 0.5334384650052095_dp, &
             0.7489681402627489_dp, 1.0543592358076195_dp, 1.4933981322491248_dp, &
             2.148075195978754_dp, 3.2151626673716205_dp, 5.462301709306798_dp]

      CASE (411)

         aw(:) = & ! a
            [0.0056788702569020915_dp, 0.03038732653250226_dp, 0.07685464920229881_dp, &
             0.1490327845948825_dp, 0.2537854202184757_dp, 0.4022900808658429_dp, &
             0.6118736806237611_dp, 0.9082474519754038_dp, 1.3284052305443363_dp, &
             1.9248935323363436_dp, 2.7727204966996775_dp, 3.9816530453020174_dp, &
             5.721754809996626_dp, 8.29073958875012_dp, 12.379918566854883_dp, &
             ! w
             0.014625921787252189_dp, 0.035117506444987685_dp, 0.05846019571054044_dp, &
             0.08701656853891046_dp, 0.12432078010685567_dp, 0.17553968940573733_dp, &
             0.2478661231331975_dp, 0.3509786266364573_dp, 0.49798500909446447_dp, &
             0.7073027698955995_dp, 1.0062683560617502_dp, 1.4390235276810452_dp, &
             2.0878094629421766_dp, 3.1489876316118783_dp, 5.386280084251533_dp]

      CASE (412)

         aw(:) = & ! a
            [0.005024274436918441_dp, 0.026913143108849905_dp, 0.0682062998374951_dp, &
             0.13268617365179883_dp, 0.22698556198400419_dp, 0.36199116993088_dp, &
             0.5546589943832143_dp, 0.8302275837852374_dp, 1.2251474742957151_dp, &
             1.7914531635436965_dp, 2.6038082335791435_dp, 3.771918204892345_dp, &
             5.466087357167651_dp, 7.984318517067509_dp, 12.016957660758704_dp, &
             ! w
             0.01294313857608113_dp, 0.031143396217599553_dp, 0.05205762286983721_dp, &
             0.07798404343584268_dp, 0.1123987204649185_dp, 0.1604023598955141_dp, &
             0.22909264975941185_dp, 0.3280493759577527_dp, 0.4703507500452663_dp, &
             0.6745122915667454_dp, 0.968104164489216_dp, 1.3955858527294684_dp, &
             2.039444860968879_dp, 3.0957651278694307_dp, 5.325182143344801_dp]

      CASE (413)

         aw(:) = & ! a
            [0.00451653735813712_dp, 0.024216696398001444_dp, 0.06148639749263848_dp, &
             0.11996390252722296_dp, 0.20608254876660995_dp, 0.330470376569777_dp, &
             0.5097429488165323_dp, 0.7686998504455994_dp, 1.1432889758299511_dp, &
             1.68507155299078_dp, 2.468385959962654_dp, 3.602863723807878_dp, &
             5.259012010579803_dp, 7.735111843845486_dp, 11.720786699599223_dp, &
             ! w
             0.011637695929719756_dp, 0.028056563128754045_dp, 0.047073462681424004_dp, &
             0.07093023318086474_dp, 0.10304837479507904_dp, 0.1484608343730584_dp, &
             0.2141699383993229_dp, 0.30966329040709384_dp, 0.4479943584047823_dp, &
             0.6477664122253113_dp, 0.9367556700653058_dp, 1.3597062225198744_dp, &
             1.9993413337952957_dp, 3.051552625398909_dp, 5.27445878592156_dp]

      CASE (414)

         aw(:) = & ! a
            [0.0041101270238754406_dp, 0.022057130001892678_dp, 0.05609892843875579_dp, &
             0.10974902632500383_dp, 0.18926575471008383_dp, 0.3050446106979406_dp, &
             0.473388051967266_dp, 0.7186895877224461_dp, 1.0764340399712193_dp, &
             1.5977464735297437_dp, 2.3566601731303014_dp, 3.4627257320720157_dp, &
             5.086623533173339_dp, 7.526898140387448_dp, 11.472616340715833_dp, &
             ! w
             0.010592634434090541_dp, 0.02558260773574769_dp, 0.043070796710343195_dp, &
             0.0652489962492778_dp, 0.09548747656979081_dp, 0.1387519341256913_dp, &
             0.20195228350980532_dp, 0.2944915569107426_dp, 0.4294022509808315_dp, &
             0.6253657255484111_dp, 0.9103406991822888_dp, 1.3293276205302154_dp, &
             1.9652732153735464_dp, 3.013934377400175_dp, 5.231323963368795_dp]

      CASE (415)

         aw(:) = & ! a
            [0.0037767636667319993_dp, 0.020284783350103888_dp, 0.051673260806997495_dp, &
             0.1013462127489563_dp, 0.17540646964753462_dp, 0.28403858719520364_dp, &
             0.44325614892799137_dp, 0.677077058827935_dp, 1.0205594158683728_dp, &
             1.5244258129485722_dp, 2.262423525735423_dp, 3.344018507192219_dp, &
             4.940040599987986_dp, 7.349282402405416_dp, 11.26037224551846_dp, &
             ! w
             0.009735303525025218_dp, 0.02355093562907654_dp, 0.039777578408647846_dp, &
             0.06056213114892813_dp, 0.08922669612322283_dp, 0.13067133729833505_dp, &
             0.19171789934992506_dp, 0.2816922494803035_dp, 0.41360870552369505_dp, &
             0.6062180228309941_dp, 0.8876415482272523_dp, 1.303112114550926_dp, &
             1.9357875430219507_dp, 2.9813305239671295_dp, 5.193956481924646_dp]

      CASE (416)

         aw(:) = & ! a
            [0.0021545257831798764_dp, 0.011643650096875076_dp, 0.03002159361496933_dp, &
             0.06002975766012449_dp, 0.10678656324749078_dp, 0.17905295169846994_dp, &
             0.29081998970365036_dp, 0.46348600584637584_dp, 0.7291591909739005_dp, &
             1.1357391474843714_dp, 1.7548471860364_dp, 2.695116115806408_dp, &
             4.1281869152446315_dp, 6.354665973727979_dp, 10.061464318030847_dp, &
             ! w
             0.005561444579369837_dp, 0.013622443101964081_dp, 0.02357504877630784_dp, &
             0.03727297185434665_dp, 0.05767502313601273_dp, 0.08915470712053893_dp, &
             0.13789213957374424_dp, 0.21270878404495483_dp, 0.32649369729261973_dp, &
             0.4983953354087469_dp, 0.7575386633983471_dp, 1.150704588625524_dp, &
             1.7626459217391768_dp, 2.7889482630945155_dp, 4.973804296754859_dp]

      CASE (417)

         aw(:) = & ! a
            [0.0015467042849176954_dp, 0.008395281647213891_dp, 0.021833498105180888_dp, &
             0.04426551232805565_dp, 0.08027493484236987_dp, 0.1377946023687582_dp, &
             0.22961694638506064_dp, 0.3756070866312024_dp, 0.6061451901696485_dp, &
             0.9674225764379704_dp, 1.5296736237746063_dp, 2.4008394966710966_dp, &
             3.7528692640339534_dp, 5.887488128860728_dp, 9.491335915341685_dp, &
             ! w
             0.00399637634878453_dp, 0.009875100905968779_dp, 0.017387396634635732_dp, &
             0.028220974798764165_dp, 0.04509585062219922_dp, 0.07203996361679202_dp, &
             0.11485953787837302_dp, 0.18209812208238063_dp, 0.2865444576764014_dp, &
             0.44750175734874115_dp, 0.6945970376068834_dp, 1.0754922855491433_dp, &
             1.675987119509339_dp, 2.6919852463147227_dp, 4.863059946671454_dp]

      CASE (418)

         aw(:) = & ! a
            [0.0012213088418495538_dp, 0.006652212840138926_dp, 0.017421036052630262_dp, &
             0.03571531963858667_dp, 0.06576084772314154_dp, 0.11492088242881206_dp, &
             0.19516517038963463_dp, 0.3253123101170386_dp, 0.5345514314848429_dp, &
             0.8678657429638242_dp, 1.3944649527164186_dp, 2.2217212147441034_dp, &
             3.5217221726634937_dp, 5.59697902058961_dp, 9.134180652860735_dp, &
             ! w
             0.0031580665533451928_dp, 0.0078585696681408_dp, 0.014029668666190268_dp, &
             0.02324497834093027_dp, 0.0380499461297273_dp, 0.06222888856781097_dp, &
             0.10133876158503932_dp, 0.16373235062412314_dp, 0.26210759057693517_dp, &
             0.4158369767907243_dp, 0.6548585051238817_dp, 1.0274341017750817_dp, &
             1.6201338736773288_dp, 2.6292153158930103_dp, 4.791440233102354_dp]

      CASE (419)

         aw(:) = & ! a
            [0.0010163654596489673_dp, 0.0055523279214757_dp, 0.014627213267154257_dp, &
             0.03027301304168235_dp, 0.05645108226259123_dp, 0.10009822610470911_dp, &
             0.1725709802348771_dp, 0.2919086142773966_dp, 0.4864039250725653_dp, &
             0.8001129679399844_dp, 1.3014354844107494_dp, 2.097263745799141_dp, &
             3.359747856790814_dp, 5.391995406737326_dp, 8.880844953668603_dp, &
             ! w
             0.0026298460390556484_dp, 0.0065832442531841_dp, 0.011891760353531402_dp, &
             0.02004290428535919_dp, 0.033446472957147144_dp, 0.05570395478367397_dp, &
             0.09218995507492977_dp, 0.1511113941004667_dp, 0.2450839575131379_dp, &
             0.3935114036889439_dp, 0.6265467178768755_dp, 0.9928992278647892_dp, &
             1.5797450704374967_dp, 2.583677532614033_dp, 4.73951497299954_dp]

      CASE (420)

         aw(:) = & ! a
            [0.0008744958294475132_dp, 0.0047897439227468615_dp, 0.012684537183536447_dp, &
             0.026471524446639985_dp, 0.04990479669928629_dp, 0.08958471195173273_dp, &
             0.15638598932792558_dp, 0.2677347395888904_dp, 0.45121163531576486_dp, &
             0.7501246982213589_dp, 1.2322057652185714_dp, 2.0039333213152113_dp, &
             3.237480334364189_dp, 5.236429810163087_dp, 8.68780371413627_dp, &
             ! w
             0.0022640577225830362_dp, 0.005697330305768576_dp, 0.010398105760650445_dp, &
             0.017785216586660536_dp, 0.03015875363482246_dp, 0.050976629501325935_dp, &
             0.08547163636018314_dp, 0.14173226938992878_dp, 0.2322996708593134_dp, &
             0.3765894209406676_dp, 0.6049134328425472_dp, 0.966333675141072_dp, &
             1.5485236223280243_dp, 2.5483852865162735_dp, 4.699290104861745_dp]

      CASE (421)

         aw(:) = & ! a
            [0.0007699970490877884_dp, 0.004227263039996045_dp, 0.011247967102625068_dp, &
             0.02364907508575807_dp, 0.04501573330675333_dp, 0.08167319397751645_dp, &
             0.14410347433199278_dp, 0.24923170950722728_dp, 0.42405163921518896_dp, &
             0.7112470016203498_dp, 1.177982638095075_dp, 1.9303738608286636_dp, &
             3.140593991068739_dp, 5.1126181849943695_dp, 8.53366064618028_dp, &
             ! w
             0.001994538063877984_dp, 0.005042790711367482_dp, 0.009288972459623018_dp, &
             0.016095093047755187_dp, 0.027669961291400685_dp, 0.04735487643408512_dp, &
             0.08026756989727749_dp, 0.1343967267489981_dp, 0.22221568187459867_dp, &
             0.3631409974153666_dp, 0.587607506425437_dp, 0.94496571152293_dp, &
             1.5233094274329126_dp, 2.5198228955401025_dp, 4.666746342486291_dp]

      CASE (422)

         aw(:) = & ! a
            [0.0006895611127582456_dp, 0.003793771263036526_dp, 0.010138302361517542_dp, &
             0.02146095802251129_dp, 0.041205263921542445_dp, 0.07546556482301157_dp, &
             0.134395064162407_dp, 0.23449817276910057_dp, 0.4022718771382512_dp, &
             0.6798660151460763_dp, 1.133953764561905_dp, 1.8703271767807597_dp, &
             3.061146420673472_dp, 5.010718338220209_dp, 8.40644792105068_dp, &
             ! w
             0.0017870212715594368_dp, 0.004537600273700881_dp, 0.008429030056570658_dp, &
             0.014775062127092764_dp, 0.025706911662564964_dp, 0.04446868748654226_dp, &
             0.07608185116897298_dp, 0.12844879827351857_dp, 0.21398087768135024_dp, &
             0.35208925142447856_dp, 0.5733069875972592_dp, 0.927227144514161_dp, &
             1.5023065731054268_dp, 2.495987956892389_dp, 4.639595890201489_dp]

      CASE (423)

         aw(:) = & ! a
            [0.0006255780084365578_dp, 0.003448563506901705_dp, 0.009252790051906358_dp, &
             0.019709005547203024_dp, 0.0381394934269199_dp, 0.07044087422946572_dp, &
             0.12648528169149412_dp, 0.22241633640795855_dp, 0.38430207682454826_dp, &
             0.6538272042959847_dp, 1.0972315718134094_dp, 1.8200161955372063_dp, &
             2.994319698223544_dp, 4.9247349165318255_dp, 8.298851830464779_dp, &
             ! w
             0.001621908467086356_dp, 0.004134753100077638_dp, 0.007740459671259435_dp, &
             0.013710985682103825_dp, 0.024110475583018008_dp, 0.04210037037509712_dp, &
             0.07261969647900664_dp, 0.12349486157075798_dp, 0.20708025408375177_dp, &
             0.3427777482480648_dp, 0.5612008838336487_dp, 0.9121508407153058_dp, &
             1.4844032778764127_dp, 2.4756385101249463_dp, 4.616420489631633_dp]

      CASE (424)

         aw(:) = & ! a
            [0.0005733678386173403_dp, 0.0031665856795662917_dp, 0.00852808551109951_dp, &
             0.018270763473988684_dp, 0.03561138654189378_dp, 0.06627456654781967_dp, &
             0.11988821705853822_dp, 0.2122814420767167_dp, 0.36914601818150405_dp, &
             0.6317554902189304_dp, 1.0659628718482381_dp, 1.7770046979357534_dp, &
             2.936992734626107_dp, 4.8507701846577405_dp, 8.20610444968673_dp, &
             ! w
             0.0014871445606266375_dp, 0.0038052862730260425_dp, 0.007175164832047987_dp, &
             0.012831987215754884_dp, 0.022781135733066275_dp, 0.040112571829962605_dp, &
             0.06969345214233551_dp, 0.11928232476685569_dp, 0.20118097215415276_dp, &
             0.3347795767086457_dp, 0.5507588922017482_dp, 0.8991015649090148_dp, &
             1.4688669259558778_dp, 2.4579547964873605_dp, 4.596284356252853_dp]

      CASE (425)

         aw(:) = & ! a
            [0.00032302400345152345_dp, 0.001809571450258827_dp, 0.005016189032917566_dp, &
             0.011221288784793264_dp, 0.023017041072208755_dp, 0.04511475652147593_dp, &
             0.08570449837387419_dp, 0.15873709987044146_dp, 0.28760821760041116_dp, &
             0.5110180594894183_dp, 0.8923202702949009_dp, 1.534945362091075_dp, &
             2.6106708645312535_dp, 4.425849001568645_dp, 7.669625966066977_dp, &
             ! w
             0.0008404149095994822_dp, 0.0022127172568996187_dp, 0.004404415771810006_dp, &
             0.008425277738802644_dp, 0.01592959737226886_dp, 0.02959388006800317_dp, &
             0.05385111124331897_dp, 0.09601861984218273_dp, 0.16802320266993814_dp, &
             0.28911105098501644_dp, 0.49030195578440955_dp, 0.8226597473598823_dp, &
             1.3770544599011862_dp, 2.352951908922952_dp, 4.476775135243211_dp]

      CASE (426)

         aw(:) = & ! a
            [0.0002309479118724317_dp, 0.001307288415953211_dp, 0.0037002878105363976_dp, &
             0.008525783297610773_dp, 0.018065957266800242_dp, 0.036533906615871956_dp, &
             0.07140833938868062_dp, 0.13568854796893848_dp, 0.25157213145227064_dp, &
             0.4563724471127593_dp, 0.8120423512111979_dp, 1.4209334533330518_dp, &
             2.454524013880335_dp, 4.219933248501357_dp, 7.4072221155356335_dp, &
             ! w
             0.0006021998451342177_dp, 0.0016187261267440272_dp, 0.0033452393365352765_dp, &
             0.006674211828226234_dp, 0.013086819211628463_dp, 0.025059091042315712_dp, &
             0.04679690043877067_dp, 0.08536753743553713_dp, 0.15246529091869512_dp, &
             0.2672108204861979_dp, 0.46074954971779497_dp, 0.7846876596876935_dp, &
             1.3308927436829059_dp, 2.299805756787061_dp, 4.4163157384143945_dp]

      CASE (427)

         aw(:) = & ! a
            [0.0001820782633968936_dp, 0.0010395111289733386_dp, 0.002992518041134211_dp, &
             0.00705453203212621_dp, 0.01531164122266209_dp, 0.03166333811769612_dp, &
             0.06313673858107553_dp, 0.1221173957421379_dp, 0.23001680215937964_dp, &
             0.42322173412829067_dp, 0.7627288970210382_dp, 1.3501284949179844_dp, &
             2.356651925392222_dp, 4.089910831961853_dp, 7.240637442438634_dp, &
             ! w
             0.0004756369084263005_dp, 0.001300352080479198_dp, 0.002767286459160053_dp, &
             0.005692847078211061_dp, 0.01145011201060995_dp, 0.02238848101496398_dp, &
             0.04256339138227245_dp, 0.07887048489516325_dp, 0.1428378098730393_dp, &
             0.25348444003380916_dp, 0.44201784909441627_dp, 0.760390204298733_dp, &
             1.3011427010528307_dp, 2.265417075418835_dp, 4.377202122006528_dp]

      CASE (428)

         aw(:) = & ! a
            [0.00015145635064703637_dp, 0.0008711260387687643_dp, 0.002544236167079167_dp, &
             0.006111572326002756_dp, 0.013520155993096897_dp, 0.028447563369405873_dp, &
             0.05759880321508503_dp, 0.11291655201288686_dp, 0.21523828131969222_dp, &
             0.40026579443472193_dp, 0.7282787055024561_dp, 1.3002835551871088_dp, &
             2.2873048529384317_dp, 3.9973073734529803_dp, 7.1215488770918425_dp, &
             ! w
             0.00039626775058836133_dp, 0.0010992903894092524_dp, 0.0023969412766162127_dp, &
             0.005050871742805682_dp, 0.010358340399326064_dp, 0.020578367630362422_dp, &
             0.03965559019306404_dp, 0.07435653129978974_dp, 0.1360810125724632_dp, &
             0.24376413894234467_dp, 0.42864805165100434_dp, 0.742932031095579_dp, &
             1.2796585097760809_dp, 2.240512293189428_dp, 4.348877998943321_dp]

      CASE (429)

         aw(:) = & ! a
            [0.00013033103406794383_dp, 0.0007546147404157981_dp, 0.0022321298968959614_dp, &
             0.005448428722604167_dp, 0.012245002457801791_dp, 0.02613113290971849_dp, &
             0.05356582832282522_dp, 0.10615046394251602_dp, 0.2042761733917355_dp, &
             0.3831072245356745_dp, 0.7023544400177716_dp, 1.2625539646581452_dp, &
             2.2345533125141976_dp, 3.9265876434223164_dp, 7.030343965244846_dp, &
             ! w
             0.0003414757206547955_dp, 0.0009596636276879567_dp, 0.0021365303957790903_dp, &
             0.004591778079753565_dp, 0.009565621418706611_dp, 0.019247852256063992_dp, &
             0.037496331172369966_dp, 0.07097499857232124_dp, 0.1309799832669097_dp, &
             0.2363753024370087_dp, 0.41842364841844976_dp, 0.7295129765209069_dp, &
             1.2630808671191904_dp, 2.2212530740729632_dp, 4.326975585389564_dp]

      CASE (430)

         aw(:) = & ! a
            [0.00011480821110753842_dp, 0.0006687803525803461_dp, 0.0020009370081711234_dp, &
             0.004952879233832503_dp, 0.011282327983739444_dp, 0.024364935643347487_dp, &
             0.05046316608457395_dp, 0.10090370653772263_dp, 0.19571590322897758_dp, &
             0.36962513440574296_dp, 0.6818738938022625_dp, 1.2326061877222592_dp, &
             2.192515387605805_dp, 3.87005289218891_dp, 6.957266894726307_dp, &
             ! w
             0.0003011910603773181_dp, 0.0008564729701576366_dp, 0.001941949039919808_dp, &
             0.004243812767569963_dp, 0.008957291926001311_dp, 0.01821665289633639_dp, &
             0.03580899730849507_dp, 0.06831372308127896_dp, 0.12694034675372626_dp, &
             0.2304915026927615_dp, 0.41024232450349957_dp, 0.7187313324855581_dp, &
             1.2497198359483892_dp, 2.2057032208277745_dp, 4.30929205486255_dp]

      CASE (431)

         aw(:) = & ! a
            [0.00010288125515705975_dp, 0.0006026772000226355_dp, 0.001822006664337531_dp, &
             0.004566347413687135_dp, 0.010524744801794937_dp, 0.022963211356889522_dp, &
             0.047982064100025854_dp, 0.09667998948287258_dp, 0.18878427387251762_dp, &
             0.358651662541209_dp, 0.6651285739047629_dp, 1.20802403908763_dp, &
             2.157895226097974_dp, 3.8233719362259295_dp, 6.896813043249191_dp, &
             ! w
             0.0002702223025865614_dp, 0.0007767771880651209_dp, 0.001790180565135051_dp, &
             0.0039690416370657304_dp, 0.008471888750834608_dp, 0.017386988046962246_dp, &
             0.03444206730612096_dp, 0.06614499357639096_dp, 0.1236312094677251_dp, &
             0.22564948578040867_dp, 0.403482391855154_dp, 0.7097924619431463_dp, &
             1.2386136266646492_dp, 2.1927583985724617_dp, 4.294571189852885_dp]

      CASE (432)

         aw(:) = & ! a
            [9.340687201826213e-05_dp, 0.0005500573606271143_dp, 0.0016789264798342002_dp, &
             0.004255075995680376_dp, 0.009909868948963068_dp, 0.021817123424575507_dp, &
             0.045940142089935354_dp, 0.09318394358039717_dp, 0.18301796157967218_dp, &
             0.34948266839017517_dp, 0.6510827012489292_dp, 1.1873356150225314_dp, &
             2.128676796058516_dp, 3.783886720061892_dp, 6.84559604089572_dp, &
             ! w
             0.0002456102897028887_dp, 0.0007131730524152648_dp, 0.001667963204438964_dp, &
             0.0037453590318413544_dp, 0.00807317743648729_dp, 0.016700649976084902_dp, &
             0.03330460517970574_dp, 0.06433116993653835_dp, 0.12085128264622179_dp, &
             0.22156584058117093_dp, 0.3977616018252502_dp, 0.7022056961576497_dp, &
             1.2291664486410006_dp, 2.181733284169996_dp, 4.282033460231659_dp]

      CASE (433)

         aw(:) = & ! a
            [8.568398259700287e-05_dp, 0.0005070831754268762_dp, 0.0015615827771326923_dp, &
             0.00399814853161422_dp, 0.009398769381362596_dp, 0.020858242041970152_dp, &
             0.0442219155511724_dp, 0.09022732758979565_dp, 0.17811997012053413_dp, &
             0.34166444021898185_dp, 0.6390657647645775_dp, 1.16958416047751_dp, &
             2.103545046297059_dp, 3.7498584977918745_dp, 6.801396144000696_dp, &
             ! w
             0.00022553973536760532_dp, 0.0006611042867166782_dp, 0.0015670812397325414_dp, &
             0.0035589310274069556_dp, 0.007738253960003522_dp, 0.016120538591168715_dp, &
             0.03233824599695246_dp, 0.06278338066607869_dp, 0.11846991021405431_dp, &
             0.21805572258820682_dp, 0.39282956855052187_dp, 0.6956484406594744_dp, &
             1.2209854833881975_dp, 2.172175315798922_dp, 4.271164158873577_dp]

      CASE (434)

         aw(:) = & ! a
            [4.8739190172902635e-05_dp, 0.00030010585008340243_dp, 0.0009876738995310717_dp, &
             0.002712785098174194_dp, 0.006780417981129713_dp, 0.015838848845221645_dp, &
             0.03505674762899372_dp, 0.07419657960460874_dp, 0.15118187349477635_dp, &
             0.29812570854867027_dp, 0.5714107430373463_dp, 1.068696241936488_dp, &
             1.9595754022136687_dp, 3.5536970419800036_dp, 6.5454496859105245_dp, &
             ! w
             0.00012938678278271335_dp, 0.00040815193181071137_dp, 0.001062214062252585_dp, &
             0.0025953168134927747_dp, 0.0059625322233694805_dp, 0.012982592757148673_dp, &
             0.027023043168572618_dp, 0.05414698872008696_dp, 0.1050139537402174_dp, &
             0.19800028720439314_dp, 0.3643742896013531_dp, 0.6575035004193269_dp, &
             1.1730934362815018_dp, 2.1160177713084845_dp, 4.207299694970199_dp]

      CASE (435)

         aw(:) = & ! a
            [3.5169359346347595e-05_dp, 0.0002231914509169778_dp, 0.0007685348839439618_dp, &
             0.0022034376739169226_dp, 0.0057045634082146255_dp, 0.013710371978094897_dp, &
             0.031064716844303373_dp, 0.06705280643482078_dp, 0.13893935298222163_dp, &
             0.27799940351856034_dp, 0.5396720621929763_dp, 1.02076429001262_dp, &
             1.890448314510032_dp, 3.4587222980581394_dp, 6.420791960269669_dp, &
             ! w
             9.398706889507707e-05_dp, 0.00031270814722341594_dp, 0.0008619223237327935_dp, &
             0.002194170537323732_dp, 0.005196273282319835_dp, 0.011590120665894764_dp, &
             0.02460940238679822_dp, 0.050147397997076994_dp, 0.09867528571252786_dp, &
             0.18841094455177004_dp, 0.3505909246252059_dp, 0.6388233010426897_dp, &
             1.1494424326720414_dp, 2.0881498525600604_dp, 4.175604138897015_dp]

      CASE (436)

         aw(:) = & ! a
            [2.7962527836574136e-05_dp, 0.00018200952935360935_dp, 0.0006489023697117784_dp, &
             0.0019185523076751792_dp, 0.005089358226233148_dp, 0.012470406549450759_dp, &
             0.028702769593906517_dp, 0.0627704515067165_dp, 0.1315181973788553_dp, &
             0.26568150949006847_dp, 0.5200852967808487_dp, 0.9909733978710837_dp, &
             1.8472289819832828_dp, 3.399065544970394_dp, 6.342231455315601_dp, &
             ! w
             7.515949613893913e-05_dp, 0.00026103760606393373_dp, 0.0007497640070200919_dp, &
             0.0019630356691321526_dp, 0.004745588219399585_dp, 0.01075792076659256_dp, &
             0.02314781051220141_dp, 0.047698255802753414_dp, 0.09475624512994404_dp, &
             0.18243211814053142_dp, 0.3419342955770852_dp, 0.6270188755573953_dp, &
             1.13442609846837_dp, 2.0704073288446456_dp, 4.155422960799509_dp]

      CASE (437)

         aw(:) = & ! a
            [2.3442211847579637e-05_dp, 0.00015601207874397938_dp, 0.0005722023469705627_dp, &
             0.0017325672580031299_dp, 0.004681307602456673_dp, 0.011637162572818341_dp, &
             0.027098380528090743_dp, 0.05983525087597221_dp, 0.12639251721341946_dp, &
             0.257117676564694_dp, 0.5063906918916145_dp, 0.9700434173842551_dp, &
             1.8167422775653321_dp, 3.3568508191244555_dp, 6.286515566528177_dp, &
             ! w
             6.333851992393269e-05_dp, 0.0002281248435418086_dp, 0.000676464514215691_dp, &
             0.0018089259063906505_dp, 0.004440787185753832_dp, 0.010188848719609588_dp, &
             0.022139253940655763_dp, 0.0459952627953969_dp, 0.09201316146721433_dp, &
             0.17822329561444525_dp, 0.33581010830815566_dp, 0.6186327602932166_dp, &
             1.1237237866556156_dp, 2.0577381714562333_dp, 4.14101145084166_dp]

      CASE (438)

         aw(:) = & ! a
            [2.0320706770976913e-05_dp, 0.0001379618934911562_dp, 0.0005182537641554639_dp, &
             0.0015998516778860995_dp, 0.004386538648843667_dp, 0.01102922216417733_dp, &
             0.025918232747603794_dp, 0.05766148963760496_dp, 0.12257466425850463_dp, &
             0.25070751453997564_dp, 0.496096720771317_dp, 0.9542539897212147_dp, &
             1.7936743370670443_dp, 3.324833677063283_dp, 6.244188525243287_dp, &
             ! w
             5.516954006827813e-05_dp, 0.00020509748079726464_dp, 0.0006241102339083461_dp, &
             0.001697163125294396_dp, 0.004217353118586344_dp, 0.009768201045540496_dp, &
             0.021388647742569756_dp, 0.044720531100293504_dp, 0.0899497540634171_dp, &
             0.17504377138331076_dp, 0.33116649166532336_dp, 0.6122542075015512_dp, &
             1.1155639626400287_dp, 2.0480651502889393_dp, 4.1300074845427135_dp]

      CASE (439)

         aw(:) = & ! a
            [1.8025083924085845e-05_dp, 0.00012462455979027703_dp, 0.00047794208773949565_dp, &
             0.0014994930548697962_dp, 0.004161413461356942_dp, 0.010561202669204389_dp, &
             0.025003792418828705_dp, 0.055968059141112844_dp, 0.11958690105092243_dp, &
             0.2456715923201127_dp, 0.4879826826781143_dp, 0.9417729177120128_dp, &
             1.775396810166009_dp, 3.2994185175201984_dp, 6.210545634506915_dp, &
             ! w
             4.915855213036491e-05_dp, 0.00018796748269911816_dp, 0.0005844875268236697_dp, &
             0.0016115425732556384_dp, 0.004044712467570624_dp, 0.009441020488982278_dp, &
             0.02080166234517673_dp, 0.043719135072158646_dp, 0.08832247940785155_dp, &
             0.17252783018099083_dp, 0.3274812974448913_dp, 0.6071797272127996_dp, &
             1.1090601210810358_dp, 2.040346649221291_dp, 4.12122653593576_dp]

      CASE (440)

         aw(:) = & ! a
            [1.626001874645348e-05_dp, 0.00011432648923585893_dp, 0.00044650807772755104_dp, &
             0.0014204384211977804_dp, 0.003982602417064132_dp, 0.010187006714168015_dp, &
             0.02426875122889233_dp, 0.05460082055979655_dp, 0.11716565051382832_dp, &
             0.24157757391557835_dp, 0.4813683310828_dp, 0.9315751173489706_dp, &
             1.7604342451050394_dp, 3.2785815369593756_dp, 6.182933782310514_dp, &
             ! w
             4.4534822175707674e-05_dp, 0.00017466099457205613_dp, 0.0005532529872607035_dp, &
             0.0015433653863258104_dp, 0.003906271733703253_dp, 0.00917721994149124_dp, &
             0.020326277742617055_dp, 0.042905102977692434_dp, 0.08699545650927189_dp, &
             0.17047046236969857_dp, 0.3244606279917579_dp, 0.6030119662734568_dp, &
             1.1037101603888286_dp, 2.0339917573641815_dp, 4.113996572297717_dp]

      CASE (441)

         aw(:) = & ! a
            [1.4857159349583318e-05_dp, 0.00010611045268422314_dp, 0.00042120733950326653_dp, &
             0.001356246798130767_dp, 0.003836378751366743_dp, 0.00987928956091117_dp, &
             0.02366156195708651_dp, 0.05346718475694228_dp, 0.11515179812886821_dp, &
             0.23816333421069838_dp, 0.4758396579070398_dp, 0.9230346418150228_dp, &
             1.7478831877213807_dp, 3.261080806738854_dp, 6.1597223444523195_dp, &
             ! w
             4.085869045258197e-05_dp, 0.00016398652220510584_dp, 0.0005278742235130223_dp, &
             0.0014874958480147098_dp, 0.0037921455828952323_dp, 0.008958748589080953_dp, &
             0.019931104482529532_dp, 0.042226300456844605_dp, 0.08588592092988677_dp, &
             0.16874630477225577_dp, 0.32192413907723844_dp, 0.5995063826849762_dp, &
             1.0992043860297382_dp, 2.0286355378079897_dp, 4.10790256253564_dp]

      CASE (442)

         aw(:) = & ! a
            [1.3713201078198846e-05_dp, 9.938731807296921e-05_dp, 0.00040033849607394297_dp, &
             0.0013028886787681718_dp, 0.0037140838982274832_dp, 0.009620682410111355_dp, &
             0.02314928988686572_dp, 0.05250769666415305_dp, 0.11344273080991822_dp, &
             0.23525920440588335_dp, 0.47112782773955647_dp, 0.9157439170756236_dp, &
             1.7371540183027678_dp, 3.2461043412074573_dp, 6.139843829293329_dp, &
             ! w
             3.786021104247671e-05_dp, 0.00015520778775327015_dp, 0.000506766192819174_dp, &
             0.0014406853409254816_dp, 0.003696032942457307_dp, 0.008774030326622207_dp, &
             0.01959590850322764_dp, 0.04164897434399873_dp, 0.08494009065360072_dp, &
             0.1672736315408742_dp, 0.31975392819781684_dp, 0.5965027187924734_dp, &
             1.095339467866635_dp, 2.024038131358753_dp, 4.10267171027402_dp]

      CASE (443)

         aw(:) = & ! a
            [8.241390969460819e-06_dp, 6.682828547515902e-05_dp, 0.0002964966549912458_dp, &
             0.0010306941688953423_dp, 0.0030780394598367983_dp, 0.00825525528467757_dp, &
             0.02041166047918162_dp, 0.04732896673960058_dp, 0.10414123315508589_dp, &
             0.21934170448715923_dp, 0.4451461228571678_dp, 0.8753351088762175_dp, &
             1.6774340998452855_dp, 3.1624654656299014_dp, 6.02856867999779_dp, &
             ! w
             2.3509713948686844e-05_dp, 0.00011193516343481133_dp, 0.00039887745388847906_dp, &
             0.0011959084983082509_dp, 0.003185370388606981_dp, 0.007780425656002086_dp, &
             0.017774776728682256_dp, 0.03848607637085464_dp, 0.07972142358903424_dp, &
             0.15909821973325272_dp, 0.30764243135637825_dp, 0.5796652215027419_dp, &
             1.0735996091356506_dp, 1.9981255460279612_dp, 4.073185575994722_dp]

      CASE (444)

         aw(:) = & ! a
            [6.2431686640910065e-06_dp, 5.468320791513948e-05_dp, 0.00025609051506611936_dp, &
             0.0009209692722278494_dp, 0.00281479824198553_dp, 0.007678677508051289_dp, &
             0.019237161869981093_dp, 0.04507835139776771_dp, 0.10005536512621487_dp, &
             0.21228613435150717_dp, 0.43354067323594847_dp, 0.8571675697157035_dp, &
             1.6504395708290165_dp, 3.1245000757193746_dp, 5.977909953952341_dp, &
             ! w
             1.826746824496202e-05_dp, 9.531844182955848e-05_dp, 0.0003552642447686069_dp, &
             0.001093901083308695_dp, 0.00296802102213176_dp, 0.007350645409534533_dp, &
             0.01697677748977904_dp, 0.0370851818495071_dp, 0.07738893505811961_dp, &
             0.1554156882809298_dp, 0.30215037102183157_dp, 0.5719871863122427_dp, &
             1.0636430911665204_dp, 1.9862275462579313_dp, 4.059644734426645_dp]

      CASE (445)

         aw(:) = & ! a
            [5.195444951992844e-06_dp, 4.822212668040772e-05_dp, 0.00023404332711157215_dp, &
             0.0008599278390227917_dp, 0.0026662941765864825_dp, 0.007349978123699891_dp, &
             0.01856207171556509_dp, 0.043776099124357835_dp, 0.09767816779619294_dp, &
             0.2081621149455093_dp, 0.4267305715755695_dp, 0.8464714069361969_dp, &
             1.6345029021996433_dp, 3.1020386494445855_dp, 5.947894003280663_dp, &
             ! w
             1.551888511787385e-05_dp, 8.631401153933488e-05_dp, 0.0003309648392076759_dp, &
             0.0010361587098921638_dp, 0.0028436283944012924_dp, 0.007102607500969437_dp, &
             0.016513140130819788_dp, 0.0362667645379296_dp, 0.0760199281015208_dp, &
             0.15324570067029347_dp, 0.2989030437384331_dp, 0.5674343373675752_dp, &
             1.0577261185866327_dp, 1.9791475037579467_dp, 4.051586430200116_dp]

      CASE (446)

         aw(:) = & ! a
            [4.552150151574599e-06_dp, 4.421084777083031e-05_dp, 0.0002201145769694895_dp, &
             0.0008208720232572033_dp, 0.002570421715333931_dp, 0.007136351257862949_dp, &
             0.018121029796242263_dp, 0.04292174992456705_dp, 0.09611318317440591_dp, &
             0.2054392332913893_dp, 0.4222231162936867_dp, 0.8393771127857702_dp, &
             1.623914646674579_dp, 3.0870953868842137_dp, 5.927906145240573_dp, &
             ! w
             1.383120882196413e-05_dp, 8.064969674292988e-05_dp, 0.00031540243318754985_dp, &
             0.000998802769663293_dp, 0.0027625888028654623_dp, 0.0069401552242630395_dp, &
             0.016208194317830372_dp, 0.03572659643246346_dp, 0.07511371637549026_dp, &
             0.1518056921914172_dp, 0.2967435010669072_dp, 0.5644011641235893_dp, &
             1.0537786810559537_dp, 1.974420231979491_dp, 4.046205680024345_dp]

      CASE (447)

         aw(:) = & ! a
            [4.122470352568141e-06_dp, 4.1507834020308295e-05_dp, 0.00021060778347810542_dp, &
             0.0007939753098109034_dp, 0.0025039817880868555_dp, 0.006987617677166264_dp, &
             0.017812852658809695_dp, 0.042323040132568476_dp, 0.09501384762109422_dp, &
             0.2035226829492591_dp, 0.4190450724687312_dp, 0.834368015227247_dp, &
             1.6164297185912617_dp, 3.0765221474157727_dp, 5.91375447663908_dp, &
             ! w
             1.270379239433333e-05_dp, 7.679490203563814e-05_dp, 0.0003046778256007781_dp, &
             0.0009728781787346506_dp, 0.002706074097468122_dp, 0.006826447492646356_dp, &
             0.01599412356822144_dp, 0.03534648805097493_dp, 0.07447474056025505_dp, &
             0.1507885861468661_dp, 0.29521592904369187_dp, 0.562252974965816_dp, &
             1.050980317614839_dp, 1.9710671362765142_dp, 4.042388915888311_dp]

      CASE (448)

         aw(:) = & ! a
            [3.8221418129924285e-06_dp, 3.9605078721605024e-05_dp, 0.0002038507295928549_dp, &
             0.0007747318622394548_dp, 0.002456229125364483_dp, 0.006880356521467069_dp, &
             0.017590024549201822_dp, 0.04188923185699184_dp, 0.0942159226543805_dp, &
             0.20212958905835693_dp, 0.4167322018194978_dp, 0.8307188025957016_dp, &
             1.6109721781890427_dp, 3.0688076941703457_dp, 5.903424355831322_dp, &
             ! w
             1.1915614169497044e-05_dp, 7.406073411343107e-05_dp, 0.00029700113056061725_dp, &
             0.0009542263774814714_dp, 0.002665270010945913_dp, 0.006744130073567943_dp, &
             0.01583882052745671_dp, 0.03507025017926313_dp, 0.07400969910517785_dp, &
             0.1500474266391113_dp, 0.2941016146429678_dp, 0.5606845466315471_dp, &
             1.0489357829343278_dp, 1.9686163018769662_dp, 4.039599098908012_dp]

      CASE (449)

         aw(:) = & ! a
            [3.608537740339574e-06_dp, 3.824407439216674e-05_dp, 0.00019898201087107598_dp, &
             0.0007607979526876089_dp, 0.002421534945568358_dp, 0.006802232903652828_dp, &
             0.01742741491753449_dp, 0.041572169332743804_dp, 0.09363199234526508_dp, &
             0.20110902193548585_dp, 0.4150362892748818_dp, 0.8280409847947445_dp, &
             1.6069648893909732_dp, 3.06314047039987_dp, 5.89583302293874_dp, &
             ! w
             1.1354903809493574e-05_dp, 7.209361960912407e-05_dp, 0.00029144057318994843_dp, &
             0.000940665064529573_dp, 0.0026355249568126965_dp, 0.006684004826496181_dp, &
             0.01572520905868895_dp, 0.03486791105427211_dp, 0.07366870003184767_dp, &
             0.14950346336821788_dp, 0.29328314067500394_dp, 0.5595317683699012_dp, &
             1.047432312892957_dp, 1.9668135128869835_dp, 4.037546918031682_dp]

      CASE (450)

         aw(:) = & ! a
            [3.4584295676527143e-06_dp, 3.7283389858028125e-05_dp, 0.00019552643654497995_dp, &
             0.0007508724217801809_dp, 0.002396759555128402_dp, 0.006746341842291622_dp, &
             0.017310916375668207_dp, 0.0413447587833187_dp, 0.09321278334088909_dp, &
             0.20037577676117063_dp, 0.41381702779984175_dp, 0.8261147226712228_dp, &
             1.6040809666808717_dp, 3.0590604918593285_dp, 5.890366474092025_dp, &
             ! w
             1.0960779696264931e-05_dp, 7.069896989431445e-05_dp, 0.00028747860410465636_dp, &
             0.0009309756360462223_dp, 0.0026142314993089742_dp, 0.006640901003258657_dp, &
             0.01564366799263008_dp, 0.03472255287623384_dp, 0.0734235382725954_dp, &
             0.14911211924735612_dp, 0.2926939688804001_dp, 0.5587015537904995_dp, &
             1.0463491355034387_dp, 1.965514404952094_dp, 4.036068073463604_dp]

      CASE (451)

         aw(:) = & ! a
            [3.359044487633565e-06_dp, 3.664525662033725e-05_dp, 0.00019322206826094169_dp, &
             0.0007442365167359122_dp, 0.0023801663474433475_dp, 0.006708860705186679_dp, &
             0.017232713424910437_dp, 0.04119198146784048_dp, 0.09293096926312501_dp, &
             0.19988258085084912_dp, 0.41299654611678777_dp, 0.8248179681623964_dp, &
             1.6021388900621432_dp, 3.056312285460253_dp, 5.886683650439728_dp, &
             ! w
             1.06997832196879e-05_dp, 6.976964094308859e-05_dp, 0.00028482927867991807_dp, &
             0.0009244837486827712_dp, 0.0025999456482000308_dp, 0.006611953048067482_dp, &
             0.015588862131619368_dp, 0.034624789851089674_dp, 0.07325855975868562_dp, &
             0.1488486445351574_dp, 0.2922971450475792_dp, 0.558142191475726_dp, &
             1.0456191489815219_dp, 1.9646387631965558_dp, 4.035071257281259_dp]

      CASE (452)

         aw(:) = & ! a
            [3.291667899337863e-06_dp, 3.6211649942309474e-05_dp, 0.0001916520413872042_dp, &
             0.0007397073815295243_dp, 0.0023688276459558682_dp, 0.006683226213018714_dp, &
             0.017179191903350474_dp, 0.041087365085774624_dp, 0.09273790706811486_dp, &
             0.19954458193741717_dp, 0.41243407495664974_dp, 0.8239287603035587_dp, &
             1.6008068838318577_dp, 3.054427060491967_dp, 5.884156997235583_dp, &
             ! w
             1.0522817667026174e-05_dp, 6.91367855732188e-05_dp, 0.00028302083846862936_dp, &
             0.000920046489898947_dp, 0.0025901722444491987_dp, 0.006592135178427996_dp, &
             0.015551321239944434_dp, 0.03455779403365817_dp, 0.07314545968838804_dp, &
             0.14866796426944323_dp, 0.2920249462823512_dp, 0.5577584133433053_dp, &
             1.0451182163084207_dp, 1.9640378140134063_dp, 4.034387153432591_dp]

      CASE (453)

         aw(:) = & ! a
            [0.02123626989477938_dp, 0.11253862971525486_dp, 0.2794938163318228_dp, &
             0.5270043571293699_dp, 0.8626802745521075_dp, 1.2975630678973789_dp, &
             1.84713674133089_dp, 2.5327235062448628_dp, 3.383372551099275_dp, &
             4.438456018478218_dp, 5.751498661890009_dp, 7.396553979358261_dp, &
             9.480410156069674_dp, 12.169734076683913_dp, 15.765064211048774_dp, &
             20.994933219479922_dp, &
             ! w
             0.05457228894042105_dp, 0.12850668039658322_dp, 0.2062107508418175_dp, &
             0.29006703140987156_dp, 0.38310839254799145_dp, 0.48921483128497045_dp, &
             0.6134551742856799_dp, 0.7625177069692254_dp, 0.9452813262658691_dp, &
             1.173706845606069_dp, 1.464534719177301_dp, 1.8429759059688344_dp, &
             2.3515079137842094_dp, 3.073922472654218_dp, 4.220025719875798_dp, &
             6.627386625306971_dp]

      CASE (454)

         aw(:) = & ! a
            [0.01249335598340411_dp, 0.06636347934040579_dp, 0.16554012109404947_dp, &
             0.3142122161559276_dp, 0.5190964445166673_dp, 0.790303614338604_dp, &
             1.14256205246528_dp, 1.5968455073257304_dp, 2.182439219932948_dp, &
             2.9396026075790025_dp, 3.9233621214215906_dp, 5.209757485204127_dp, &
             6.907690198558372_dp, 9.185012378465641_dp, 12.33931597149999_dp, &
             17.07928007504425_dp, &
             ! w
             0.0321225885764125_dp, 0.07600169220553715_dp, 0.1230393637303321_dp, &
             0.1754101491069477_dp, 0.23602314019090215_dp, 0.30881881858347765_dp, &
             0.39915952341596667_dp, 0.5142450102957286_dp, 0.6635919448742423_dp, &
             0.8598082538447277_dp, 1.1202030710426285_dp, 1.4703548620534355_dp, &
             1.9525233214943742_dp, 2.6485569231618076_dp, 3.7608140990416503_dp, &
             6.092181707707883_dp]

      CASE (455)

         aw(:) = & ! a
            [0.009159784924232068_dp, 0.04874124565626381_dp, 0.12198111404576532_dp, &
             0.232692995418522_dp, 0.38712219115716845_dp, 0.5948850330998255_dp, &
             0.8702777453847756_dp, 1.233954861820021_dp, 1.7149841111626252_dp, &
             2.3534803627430483_dp, 3.204432603105609_dp, 4.344054149447819_dp, &
             5.881684905596398_dp, 7.985635754045998_dp, 10.952845642003844_dp, &
             15.485273091084188_dp, &
             ! w
             0.023560919726193764_dp, 0.05594131340168916_dp, 0.09116326281033685_dp, &
             0.1312859571784798_dp, 0.17915330472587593_dp, 0.23872704337829892_dp, &
             0.31546694974446954_dp, 0.4166893847022847_dp, 0.5519940219255626_dp, &
             0.7340636247311381_dp, 0.9803606385561896_dp, 1.3167257512871589_dp, &
             1.785640077047676_dp, 2.4686440241868266_dp, 3.565729803472949_dp, &
             5.866704271620807_dp]

      CASE (456)

         aw(:) = & ! a
            [0.007322560141813404_dp, 0.039019303320244614_dp, 0.09790678446590025_dp, &
             0.187522875978312_dp, 0.3137542183667891_dp, 0.4858136923194133_dp, &
             0.7175823469735936_dp, 1.029286936368008_dp, 1.4495317958082097_dp, &
             2.017951611756463_dp, 2.7891432547193737_dp, 3.8391679753262133_dp, &
             5.277547112934667_dp, 7.272981854778001_dp, 10.122302228244262_dp, &
             14.523744726813701_dp, &
             ! w
             0.018841222624291037_dp, 0.04486054935433717_dp, 0.07349307310161883_dp, &
             0.10670408062782458_dp, 0.14727159188773734_dp, 0.19913285414852644_dp, &
             0.2677423025669557_dp, 0.3603915702110762_dp, 0.4866411695538599_dp, &
             0.6591954549030602_dp, 0.8956803685351775_dp, 1.2222440383488673_dp, &
             1.6816930580726246_dp, 2.3555756876213545_dp, 3.4425924383299704_dp, &
             5.724556568819595_dp]

      CASE (457)

         aw(:) = & ! a
            [0.006144279628950153_dp, 0.03277911224232179_dp, 0.08243143408680947_dp, &
             0.15842545956161763_dp, 0.2663633767071308_dp, 0.41512300978947586_dp, &
             0.6182082205518974_dp, 0.8954087492618855_dp, 1.274817140373277_dp, &
             1.7955250328439663_dp, 2.511671361173349_dp, 3.499086710699372_dp, &
             4.867379267125952_dp, 6.785582282555889_dp, 9.550605650905268_dp, &
             13.858296262592724_dp, &
             ! w
             0.015813720032822463_dp, 0.037740980322357426_dp, 0.062106580525873915_dp, &
             0.09079863087031755_dp, 0.1265341518126833_dp, 0.17320811624424157_dp, &
             0.23622773902770974_dp, 0.3228120296047017_dp, 0.4424551929503153_dp, &
             0.6078787063087151_dp, 0.8368663078342602_dp, 1.1558545652842125_dp, &
             1.6079713225351366_dp, 2.274872303451242_dp, 3.3544410264652536_dp, &
             5.622910311607047_dp]

      CASE (458)

         aw(:) = & ! a
            [0.005318206626187421_dp, 0.028401132818993903_dp, 0.07156053986480068_dp, &
             0.13794842392532944_dp, 0.23293358815672274_dp, 0.36510922545412866_dp, &
             0.5476393721885707_dp, 0.7998966011997658_dp, 1.1494732580738143_dp, &
             1.6349308451367353_dp, 2.309957873522667_dp, 3.2501382678703843_dp, &
             4.565119676968007_dp, 6.424218342852793_dp, 9.12449771873384_dp, &
             13.36013347283737_dp, &
             ! w
             0.013690837729939604_dp, 0.032741734632585666_dp, 0.05409121573881286_dp, &
             0.07956259327454886_dp, 0.11181728320952655_dp, 0.15470082310158226_dp, &
             0.2135556027713844_dp, 0.29551330198325365_dp, 0.41000015242482113_dp, &
             0.5697558701772301_dp, 0.7927078885207337_dp, 1.1055499871864973_dp, &
             1.5517056878524866_dp, 2.212974007155654_dp, 3.286677802349345_dp, &
             5.544848727853524_dp]

      CASE (459)

         aw(:) = & ! a
            [0.004703903160134626_dp, 0.025143461350440856_dp, 0.06346245800840158_dp, &
             0.1226699481548257_dp, 0.20793818507598555_dp, 0.32761384294384377_dp, &
             0.4945540693857813_dp, 0.727742231258249_dp, 1.0543005913145043_dp, &
             1.512294703134273_dp, 2.154989672686737_dp, 3.0577259786550077_dp, &
             4.3301629301894_dp, 6.141859693294843_dp, 8.79006196623135_dp, &
             12.967709297738798_dp, &
             ! w
             0.01211194222848858_dp, 0.02901895327156339_dp, 0.04810931086327413_dp, &
             0.07115084222869922_dp, 0.10075425998731947_dp, 0.14071320966019513_dp, &
             0.19629857621820135_dp, 0.27455289222346774_dp, 0.38483957921000617_dp, &
             0.5399170453962834_dp, 0.7578406304494023_dp, 1.0655310964555629_dp, &
             1.5066799629278596_dp, 2.1632416780817367_dp, 3.23213398805937_dp, &
             5.482067206815196_dp]

      CASE (460)

         aw(:) = & ! a
            [0.00422749413273055_dp, 0.022615645154839315_dp, 0.05717240489148407_dp, &
             0.1107855007815738_dp, 0.1884582617757678_dp, 0.29832047060331274_dp, &
             0.452950710213161_dp, 0.6709720585108714_dp, 0.9790711090780677_dp, &
             1.414855708895712_dp, 2.0311982614953905_dp, 2.903205265470869_dp, &
             4.140529901781275_dp, 5.912941751200072_dp, 8.517877774821814_dp, &
             12.647326367116241_dp, &
             ! w
             0.010887306886137257_dp, 0.02612826632233438_dp, 0.04345528500785436_dp, &
             0.06458788949871225_dp, 0.09209034960236005_dp, 0.12970422989847916_dp, &
             0.18262748876520904_dp, 0.25781680518497263_dp, 0.3645792437854545_dp, &
             0.515691279201699_dp, 0.729321297550378_dp, 1.0325907226271627_dp, &
             1.4694338205014248_dp, 2.1219626037538277_dp, 3.186791641347356_dp, &
             5.429913659794066_dp]

      CASE (461)

         aw(:) = & ! a
            [0.003846215689887253_dp, 0.020591567605920486_dp, 0.052131236756410314_dp, &
             0.10124813840432649_dp, 0.17279805880579355_dp, 0.27471750430074443_dp, &
             0.41933078052311024_dp, 0.6249280273155943_dp, 0.9177937594965698_dp, &
             1.3351147388188598_dp, 1.9294001189108119_dp, 2.775533661043749_dp, &
             3.9831504408499843_dp, 5.722203938971686_dp, 8.290322245610936_dp, &
             12.378740047913006_dp, &
             ! w
             0.009907094205358409_dp, 0.02381219843928404_dp, 0.03971970223882869_dp, &
             0.059306507415729695_dp, 0.08509411213091364_dp, 0.12077294875471438_dp, &
             0.17146923106208586_dp, 0.24405896397709637_dp, 0.34779893650915916_dp, &
             0.49548237669575296_dp, 0.7053775625652202_dp, 1.0047843223684194_dp, &
             1.4378576455423158_dp, 2.0868647293083935_dp, 3.14818751482594_dp, &
             5.385537697274526_dp]

      CASE (462)

         aw(:) = & ! a
            [0.0035335098097274502_dp, 0.018930747017319782_dp, 0.04799132910586563_dp, &
             0.09340633146197712_dp, 0.15990097589914684_dp, 0.25523772056519983_dp, &
             0.3915075062326464_dp, 0.5866925245564957_dp, 0.8667062212594495_dp, &
             1.2683475596560745_dp, 1.8437892243513303_dp, 2.6677035867305716_dp, &
             3.849699295046004_dp, 5.559891904882398_dp, 8.096096154837575_dp, &
             12.148935783294984_dp, &
             ! w
             0.009103084606394137_dp, 0.021910709462719523_dp, 0.03664772176776826_dp, &
             0.054952977443978805_dp, 0.07930837363114634_dp, 0.1133547294055117_dp, &
             0.16214897233480427_dp, 0.23249210399071235_dp, 0.333595921973549_dp, &
             0.4782687636968508_dp, 0.6848674256397951_dp, 0.9808514234763911_dp, &
             1.4105774826377881_dp, 2.056463551205302_dp, 3.1147098104802877_dp, &
             5.347075263990651_dp]

      CASE (463)

         aw(:) = & ! a
            [0.0020125660928861262_dp, 0.010839302580047023_dp, 0.02776073727683414_dp, &
             0.05491535485485394_dp, 0.09621390677878901_dp, 0.15826858590132964_dp, &
             0.25155176221863657_dp, 0.3918898940757439_dp, 0.6026248579910799_dp, &
             0.9178653621745613_dp, 1.3873828979151186_dp, 2.0842135216668547_dp, &
             3.1175469236537974_dp, 4.658456142887671_dp, 7.006244025097976_dp, &
             10.848771624757788_dp, &
             ! w
             0.005191005039162004_dp, 0.012627761284911658_dp, 0.02156064071318729_dp, &
             0.033384664006361114_dp, 0.050295370164638205_dp, 0.07553538085132828_dp, &
             0.11363368103946318_dp, 0.1708798850219695_dp, 0.256195443232919_dp, &
             0.3824592769646401_dp, 0.5685478192743597_dp, 0.842929673227647_dp, &
             1.2513442211943193_dp, 1.8774322383842335_dp, 2.916747074100502_dp, &
             5.120036241800754_dp]

      CASE (464)

         aw(:) = & ! a
            [0.0014432114855624366_dp, 0.007801531779809701_dp, 0.020125469832939162_dp, &
             0.04027487358928148_dp, 0.07172700433101158_dp, 0.12043784221977033_dp, &
             0.19592250866598496_dp, 0.3127387451926994_dp, 0.4927264320226538_dp, &
             0.7683981194605783_dp, 1.1880336160297063_dp, 1.823551281651308_dp, &
             2.7837030596469976_dp, 4.239997459981806_dp, 6.4927254010870055_dp, &
             10.228937612952235_dp, &
             ! w
             0.003725556416991007_dp, 0.009130298705641785_dp, 0.015817113603965635_dp, &
             0.025046690407301304_dp, 0.038832707327519145_dp, 0.060151024090465445_dp, &
             0.09320815067576782_dp, 0.1440089851094002_dp, 0.22130250395541864_dp, &
             0.33796141989191814_dp, 0.513089899907686_dp, 0.7756846839561413_dp, &
             1.172301678749089_dp, 1.7874363147483472_dp, 2.8166376845013903_dp, &
             5.005481438637751_dp]

      CASE (465)

         aw(:) = & ! a
            [0.001138607606637662_dp, 0.006172997812692638_dp, 0.016016909835216055_dp, &
             0.03235243234811286_dp, 0.058370258317666956_dp, 0.09957778943079486_dp, &
             0.16483155756290288_dp, 0.2678221908723548_dp, 0.42936026818301354_dp, &
             0.680845171457615_dp, 1.0694838116053031_dp, 1.6663482146929018_dp, &
             2.5798010052893874_dp, 3.9815934666964297_dp, 6.172734113102047_dp, &
             9.839972278801387_dp, &
             ! w
             0.0029411704155330163_dp, 0.007250637976088699_dp, 0.0127075060808784_dp, &
             0.0204817457415187_dp, 0.032455233648936765_dp, 0.051412045731351635_dp, &
             0.08133999621352078_dp, 0.1280531259067859_dp, 0.20017159948716043_dp, &
             0.31053680925413074_dp, 0.4783756936706183_dp, 0.7330239927551989_dp, &
             1.121605973954248_dp, 1.7292644544768974_dp, 2.751681290901997_dp, &
             4.931240718471879_dp]

      CASE (466)

         aw(:) = & ! a
            [0.0009468599039160355_dp, 0.005146160887302762_dp, 0.013418571153765537_dp, &
             0.027319234990486427_dp, 0.049828746034209906_dp, 0.08611936057782459_dp, &
             0.14455569569598634_dp, 0.2381830158803071_dp, 0.3870414798533882_dp, &
             0.6216847902584319_dp, 0.9884865630628115_dp, 1.5578388987534184_dp, &
             2.4377619194799007_dp, 3.8001562576159_dp, 5.946586577844916_dp, &
             9.56369803004218_dp, &
             ! w
             0.002447213193628341_dp, 0.006063096162127432_dp, 0.010731255021633994_dp, &
             0.017554051482094653_dp, 0.02831105177630225_dp, 0.04564029610353025_dp, &
             0.07336883599358635_dp, 0.11716847667926884_dp, 0.18555466002953247_dp, &
             0.2913293199019894_dp, 0.4537929108325465_dp, 0.7025225505971376_dp, &
             1.0850730482430841_dp, 1.6871056785919718_dp, 2.704472005383909_dp, &
             4.877325329085324_dp]

      CASE (467)

         aw(:) = & ! a
            [0.0008141838379395475_dp, 0.0044346795523015625_dp, 0.011613638985669604_dp, &
             0.02380926309685988_dp, 0.04383837761827239_dp, 0.07660898723413151_dp, &
             0.1300986481276461_dp, 0.2168455261816723_dp, 0.3562805725175744_dp, &
             0.5782804460787271_dp, 0.9285403059074989_dp, 1.4768849008164218_dp, &
             2.3310307802189953_dp, 3.662975489583911_dp, 5.7747339253388486_dp, &
             9.352938686108484_dp, &
             ! w
             0.0021053197367968384_dp, 0.005238881032241541_dp, 0.009352740756311956_dp, &
             0.015495836964274748_dp, 0.025364843770632677_dp, 0.041481888656963145_dp, &
             0.06754908621571802_dp, 0.10912528140851005_dp, 0.1746371114059134_dp, &
             0.27684504062967474_dp, 0.4350961478436266_dp, 0.6791506110084174_dp, &
             1.0569067323081625_dp, 1.6544567764087874_dp, 2.6678292216763007_dp, &
             4.83550044561988_dp]

      CASE (468)

         aw(:) = & ! a
            [0.0007164949012989691_dp, 0.003910186382166121_dp, 0.010280107246459396_dp, &
             0.021207046467061254_dp, 0.03937487847212398_dp, 0.0694755195241084_dp, &
             0.11917077023509802_dp, 0.20058512720646549_dp, 0.33264946256516004_dp, &
             0.5446788598533016_dp, 0.8817981251009139_dp, 1.4133459060584153_dp, &
             2.24676696687539_dp, 3.5541234904565746_dp, 5.637806288890546_dp, &
             9.184482411224847_dp, &
             ! w
             0.0018535144254540458_dp, 0.004630391151922406_dp, 0.008330547866480902_dp, &
             0.013959036425840211_dp, 0.023143355601873458_dp, 0.038310745695463555_dp, &
             0.06306228606568692_dp, 0.10286331693679289_dp, 0.16606321357597628_dp, &
             0.2653813785938354_dp, 0.42019534692385324_dp, 0.6604101630611989_dp, &
             1.0342079766011074_dp, 1.6280489633923891_dp, 2.6381353298352224_dp, &
             4.8016213243881145_dp]

      CASE (469)

         aw(:) = & ! a
            [0.0006413268094817408_dp, 0.0035061734626053672_dp, 0.009250850744827762_dp, &
             0.01919232793589403_dp, 0.035903368401840566_dp, 0.06389445073545688_dp, &
             0.11056285245895803_dp, 0.187686236929584_dp, 0.3137737097128846_dp, &
             0.5176624942979188_dp, 0.8439866686004004_dp, 1.3616601833297932_dp, &
             2.1778821657413934_dp, 3.464758288901994_dp, 5.52500051528786_dp, &
             9.045335753558277_dp, &
             ! w
             0.001659711097960326_dp, 0.004161063262449229_dp, 0.007539020145854125_dp, &
             0.012761570676967888_dp, 0.02139721225887122_dp, 0.03579364194675145_dp, &
             0.05946778021054195_dp, 0.09780530353975819_dp, 0.15908716848022886_dp, &
             0.25599310022175054_dp, 0.4079207043380653_dp, 0.6448932537552953_dp, &
             1.015333520252577_dp, 1.6060220091447996_dp, 2.613327684839822_dp, &
             4.773326490075026_dp]

      CASE (470)

         aw(:) = & ! a
            [0.0005815528588767196_dp, 0.0031845872823828758_dp, 0.00843009848407876_dp, &
             0.01758116593260212_dp, 0.03311565064103493_dp, 0.059388592600538914_dp, &
             0.1035710999862279_dp, 0.1771438940187625_dp, 0.2982529920600857_dp, &
             0.4953210601193777_dp, 0.8125524576330856_dp, 1.3184844662501274_dp, &
             2.12009233578979_dp, 3.389510703935226_dp, 5.429730790303327_dp, &
             8.927553354440402_dp, &
             ! w
             0.0015055631595250563_dp, 0.0037870450506321844_dp, 0.00690596865384866_dp, &
             0.011798367130382651_dp, 0.019981560165404876_dp, 0.03373529707545055_dp, &
             0.05650478663366711_dp, 0.09360638685684552_dp, 0.15325966504078803_dp, &
             0.24810642080935616_dp, 0.3975573296197845_dp, 0.6317344227996358_dp, &
             0.9992684300875393_dp, 1.5872230602253214_dp, 2.5921258825123528_dp, &
             4.749150852289011_dp]

      CASE (471)

         aw(:) = & ! a
            [0.00053279144924341_dp, 0.002922014073309145_dp, 0.007758841968395698_dp, &
             0.01625999258428908_dp, 0.03082088275147343_dp, 0.055661259823171286_dp, &
             0.09775570475722127_dp, 0.16832648559008362_dp, 0.2852019800763538_dp, &
             0.47643972768349957_dp, 0.7858625488052698_dp, 1.2816696415764266_dp, &
             2.0706307332283274_dp, 3.3248991112367183_dp, 5.347712267242679_dp, &
             8.825952375092989_dp, &
             ! w
             0.0013797888502005509_dp, 0.003481329651598812_dp, 0.0063868138275758575_dp, &
             0.011004275089362643_dp, 0.018806041634350512_dp, 0.032012926679498024_dp, &
             0.054007928408406826_dp, 0.09004611020303548_dp, 0.14829138721719914_dp, &
             0.2413494780456323_dp, 0.3886392366518693_dp, 0.6203667158670122_dp, &
             0.9853451315581527_dp, 1.570891642620387_dp, 2.573684140962671_dp, &
             4.728127112746264_dp]

      CASE (472)

         aw(:) = & ! a
            [0.0002992241208797295_dp, 0.001660263135655202_dp, 0.004513817000332386_dp, &
             0.009811141296951102_dp, 0.019461643837338276_dp, 0.036886468087944185_dp, &
             0.0679045674012652_dp, 0.12220413506291097_dp, 0.21569238652733375_dp, &
             0.37416770691201034_dp, 0.6390292853494859_dp, 1.076252660472402_dp, &
             1.7911592740081455_dp, 2.9558684609147226_dp, 4.87513923191185_dp, &
             8.236677844794322_dp, &
             ! w
             0.0007768847420693004_dp, 0.0020065660238064885_dp, 0.0038523102169686027_dp, &
             0.007051878544035466_dp, 0.012805153586703547_dp, 0.022990435633194663_dp, &
             0.040622319624940055_dp, 0.07056844316216015_dp, 0.12061532382449311_dp, &
             0.20309076737149567_dp, 0.33739500108413983_dp, 0.5541903301702226_dp, &
             0.903401381371959_dp, 1.4739942115085554_dp, 2.4637954099354933_dp, &
             4.602935860942909_dp]

      CASE (473)

         aw(:) = & ! a
            [0.00021347052176661868_dp, 0.0011944382121612197_dp, 0.0033030929481546053_dp, &
             0.007363357489812759_dp, 0.015044018797984875_dp, 0.02937269792814372_dp, &
             0.05559890919235446_dp, 0.10264072671139964_dp, 0.18541432635682942_dp, &
             0.3285173216429233_dp, 0.5720171780433249_dp, 0.9806147524370287_dp, &
             1.6587347438742526_dp, 2.778367566830957_dp, 4.645072034217995_dp, &
             7.947202683584637_dp, &
             ! w
             0.0005552467027450947_dp, 0.0014584383837619175_dp, 0.00289028275110568_dp, &
             0.005500374295153524_dp, 0.010351689150413703_dp, 0.019157144552366685_dp, &
             0.03474388578238406_dp, 0.061766102540009006_dp, 0.1077870257402971_dp, &
             0.18494941009745244_dp, 0.31259502169534753_dp, 0.5215821469062484_dp, &
             0.8624104031684324_dp, 1.4249765739881088_dp, 2.4078703105469663_dp, &
             4.539270216652968_dp]

      CASE (474)

         aw(:) = & ! a
            [0.00016801259282013013_dp, 0.0009465404888159955_dp, 0.0026538938383495574_dp, &
             0.006034351110706909_dp, 0.012604425731931665_dp, 0.025143940441556847_dp, &
             0.0485422220818331_dp, 0.09122333027409636_dp, 0.1674571159436081_dp, &
             0.3010446184572927_dp, 0.5311533323793718_dp, 0.9216025899928024_dp, &
             1.57617191456792_dp, 2.6667207963614876_dp, 4.499332232355596_dp, &
             7.762866885048903_dp, &
             ! w
             0.000437651070085769_dp, 0.0011653666641048587_dp, 0.002368045639112556_dp, &
             0.004637983434429464_dp, 0.00895184829397263_dp, 0.016919009702432652_dp, &
             0.031244175158437585_dp, 0.056436696333444086_dp, 0.09990333732068696_dp, &
             0.17365015462436198_dp, 0.2969613931046049_dp, 0.5008065845528724_dp, &
             0.8360592730493394_dp, 1.3932536792220334_dp, 2.371544562997579_dp, &
             4.497930679204095_dp]

      CASE (475)

         aw(:) = & ! a
            [0.00013955614082720468_dp, 0.0007908754043432692_dp, 0.0022437319728956973_dp, &
             0.005186144951081648_dp, 0.011026546539096826_dp, 0.022369463798109702_dp, &
             0.04384805790935869_dp, 0.08353122276508576_dp, 0.15521869940229618_dp, &
             0.28212529566655037_dp, 0.5027477680571566_dp, 0.8802381938407274_dp, &
             1.5178750221968238_dp, 2.587397219128016_dp, 4.395269353498463_dp, &
             7.630762334254393_dp, &
             ! w
             0.0003639842736266964_dp, 0.0009806443211126503_dp, 0.0020348019995182053_dp, &
             0.0040773537715263096_dp, 0.008024124570268312_dp, 0.015411167807786183_dp, &
             0.028853742267538167_dp, 0.052752952804675055_dp, 0.0943962899936183_dp, &
             0.16568218983720878_dp, 0.28584292951765694_dp, 0.4859200114387559_dp, &
             0.8170578475224641_dp, 1.370269754790146_dp, 2.3451567673310105_dp, &
             4.467906601942395_dp]

      CASE (476)

         aw(:) = & ! a
            [0.00011994063377784437_dp, 0.0006832918618089672_dp, 0.0019587716841422183_dp, &
             0.0045917452487347415_dp, 0.009908557035023382_dp, 0.020380871666707084_dp, &
             0.040446666413748104_dp, 0.07790186288971854_dp, 0.14618166223046783_dp, &
             0.2680422516895201_dp, 0.48145067435423805_dp, 0.8490259511837919_dp, &
             1.4736387032169467_dp, 2.5269188569364047_dp, 4.315626638157058_dp, &
             7.52937595194575_dp, &
             ! w
             0.0003131740893025812_dp, 0.0008525743777120363_dp, 0.0018013124768756677_dp, &
             0.003678443406853745_dp, 0.007353895406997607_dp, 0.014307935501779015_dp, &
             0.027086131721960074_dp, 0.050003945167185586_dp, 0.09025317138847347_dp, &
             0.1596439284804762_dp, 0.27736209279270935_dp, 0.47449946366194173_dp, &
             0.8024095949928362_dp, 1.3524864442578874_dp, 2.3246983701212196_dp, &
             4.444632015708798_dp]

      CASE (477)

         aw(:) = & ! a
            [0.00010553730746574328_dp, 0.000604114890345993_dp, 0.0017480808776888786_dp, &
             0.004148923513394785_dp, 0.009067774828527468_dp, 0.018870873650537016_dp, &
             0.03784054837713197_dp, 0.07355347439504066_dp, 0.13914997017207897_dp, &
             0.25701261969287315_dp, 0.4646737063801709_dp, 0.8243107746878443_dp, &
             1.4384517778893633_dp, 2.4786279831485563_dp, 4.251838403590527_dp, &
             7.4479905527901975_dp, &
             ! w
             0.00027584562442111983_dp, 0.0007580596968981613_dp, 0.0016273891617054745_dp, &
             0.0033773574743789_dp, 0.006841653021452152_dp, 0.013456040844322448_dp, &
             0.02570946297821904_dp, 0.04784700463715347_dp, 0.08698098534525467_dp, &
             0.15484690947378996_dp, 0.2705890951533227_dp, 0.4653363336543718_dp, &
             0.7906106418418388_dp, 1.3381198015624958_dp, 2.3081433515491243_dp, &
             4.425799819861813_dp]

      CASE (478)

         aw(:) = & ! a
            [9.44773691119612e-05_dp, 0.0005431933718162297_dp, 0.0015852907600157036_dp, &
             0.0038044511901996255_dp, 0.008408308889468244_dp, 0.0176766622119812_dp, &
             0.03576363220270861_dp, 0.07006420341351442_dp, 0.13347287505346608_dp, &
             0.24805897708410202_dp, 0.45098801346782985_dp, 0.8040623195001615_dp, &
             1.4095152073454302_dp, 2.4387882648489327_dp, 4.1990791711593305_dp, &
             7.380551388696279_dp, &
             ! w
             0.0002471688129707959_dp, 0.0006851572503816519_dp, 0.001492108730335673_dp, &
             0.00314045446537675_dp, 0.006434308779518773_dp, 0.012772732039276265_dp, &
             0.024597271852795737_dp, 0.04609359446526941_dp, 0.08430634115029434_dp, &
             0.15090661441253608_dp, 0.2650012689108123_dp, 0.4577473064340514_dp, &
             0.7808066064884537_dp, 1.3261526510468242_dp, 2.2943342451891664_dp, &
             4.410092262245033_dp]

      CASE (479)

         aw(:) = & ! a
            [8.569660149849527e-05_dp, 0.0004947372328309369_dp, 0.0014553149185678176_dp, &
             0.0035277193978652814_dp, 0.007874624380363664_dp, 0.016703189593832206_dp, &
             0.03405934243782461_dp, 0.06718393471622092_dp, 0.12876188160322527_dp, &
             0.24059411841847955_dp, 0.43953025508857146_dp, 0.7870474597304489_dp, &
             1.3851212211488515_dp, 2.4051111576427746_dp, 4.154383977803103_dp, &
             7.323329501473544_dp, &
             ! w
             0.00022439206057515914_dp, 0.0006270417067617411_dp, 0.0013834406617184846_dp, &
             0.0029481984836296285_dp, 0.006100685387801727_dp, 0.012208928667195511_dp, &
             0.023673922381027016_dp, 0.044630131267802894_dp, 0.08206346425748633_dp, &
             0.14758850802717696_dp, 0.26027810439683907_dp, 0.4513113627618849_dp, &
             0.7724689548824036_dp, 1.3159538709954992_dp, 2.282551732648174_dp, &
             4.396690573561224_dp]

      CASE (480)

         aw(:) = & ! a
            [7.854272602555263e-05_dp, 0.0004551924617865592_dp, 0.0013488674597178577_dp, &
             0.003299800617017515_dp, 0.007432161479352753_dp, 0.01589088903155438_dp, &
             0.0326288715086645_dp, 0.06475380759368868_dp, 0.12476876597066035_dp, &
             0.23424081600708707_dp, 0.4297430785088648_dp, 0.7724665951175793_dp, &
             1.36415811068297_dp, 2.3761018937333027_dp, 4.1158110002476525_dp, &
             7.2738777917555995_dp, &
             ! w
             0.00020582829623399666_dp, 0.0005795163034704685_dp, 0.0012939465354750554_dp, &
             0.0027884026266817487_dp, 0.005821143293247567_dp, 0.011733451054604781_dp, &
             0.02289102009240368_dp, 0.04338348628660218_dp, 0.08014502483782908_dp, &
             0.14473997264133082_dp, 0.2562100884259844_dp, 0.44575216929552997_dp, &
             0.765249583727319_dp, 1.3071066983128932_dp, 2.2723201246218485_dp, &
             4.3850533669438825_dp]

      CASE (481)

         aw(:) = & ! a
            [4.438030891276061e-05_dp, 0.00026521859119764385_dp, 0.0008308553234953406_dp, &
             0.0021681420195814446_dp, 0.005184894515915768_dp, 0.011675259442432782_dp, &
             0.025060075084169537_dp, 0.05167393749203477_dp, 0.1029490023128316_dp, &
             0.1990559138490817_dp, 0.3748926900570616_dp, 0.6898868043563599_dp, &
             1.2443377611462834_dp, 2.2090007328076124_dp, 3.8922422143769677_dp, &
             6.985966667401165_dp, &
             ! w
             0.000117061244040588_dp, 0.00034950794424142814_dp, 0.0008496206466648573_dp, &
             0.001969812141289176_dp, 0.004350855675227548_dp, 0.009179371194722275_dp, &
             0.018610919257334716_dp, 0.03646352165879841_dp, 0.06935185327196657_dp, &
             0.12852071337270213_dp, 0.23279774492844996_dp, 0.4134538405984293_dp, &
             0.7229685667095224_dp, 1.2549753888336732_dp, 2.2118227859778363_dp, &
             4.316251314870445_dp]

      CASE (482)

         aw(:) = & ! a
            [3.1869297987576294e-05_dp, 0.00019492414358100198_dp, 0.0006347353178457845_dp, &
             0.0017249961805337506_dp, 0.0042731255047116735_dp, 0.009908912660440592_dp, &
             0.0217985806983987_dp, 0.045899088408401324_dp, 0.09310959388200447_dp, &
             0.18289308993042971_dp, 0.3492826893011757_dp, 0.6507752647865163_dp, &
             1.1868817849172504_dp, 2.128034841130706_dp, 3.783018195545699_dp, &
             6.844468560787563_dp, &
             ! w
             8.448026199337955e-05_dp, 0.0002632874914218408_dp, 0.0006755667833759133_dp, &
             0.0016333352520148353_dp, 0.0037231451439136633_dp, 0.008056050505449877_dp, &
             0.016681572051630133_dp, 0.03327770276479761_dp, 0.06429041263785437_dp, &
             0.12078976154424555_dp, 0.22147582665240537_dp, 0.3976355572801984_dp, &
             0.7020384395715727_dp, 1.2289580237584161_dp, 2.1814899323970374_dp, &
             4.281756738279945_dp]

      CASE (483)

         aw(:) = & ! a
            [2.5237260590899675e-05_dp, 0.00015739213321288803_dp, 0.0005282742418487217_dp, &
             0.0014789294512643717_dp, 0.003755478032527643_dp, 0.008886464712803572_dp, &
             0.01987922539030577_dp, 0.042452297501102826_dp, 0.0871647656143441_dp, &
             0.1730235335659833_dp, 0.33349845244429505_dp, 0.6264730726762878_dp, &
             1.1509296666579922_dp, 2.0770723714927852_dp, 3.7139476866747994_dp, &
             6.754688452385413_dp, &
             ! w
             6.718393629837033e-05_dp, 0.0002168208345444966_dp, 0.0005788588947156977_dp, &
             0.0014407960268770614_dp, 0.0033559510750659906_dp, 0.007387535199394252_dp, &
             0.015516922014445595_dp, 0.031331077655121255_dp, 0.06116486142945538_dp, &
             0.11597110720426056_dp, 0.21436069380758427_dp, 0.3876228814720941_dp, &
             0.6887092107459465_dp, 1.2123118442248106_dp, 2.1620309084314937_dp, &
             4.259627929219993_dp]

      CASE (484)

         aw(:) = & ! a
            [2.1082873548577104e-05_dp, 0.0001337470439232748_dp, 0.00046030281417542055_dp, &
             0.001319089763942155_dp, 0.0034137383474493192_dp, 0.008202087463151552_dp, &
             0.018579493629608562_dp, 0.04009511197148702_dp, 0.08306468185959978_dp, &
             0.1661664950863389_dp, 0.3224617078849208_dp, 0.6093852721745535_dp, &
             1.125528328870905_dp, 2.040920061690538_dp, 3.664793009090272_dp, &
             6.6906493119640364_dp, &
             ! w
             5.6337542332766104e-05_dp, 0.0001873246783489965_dp, 0.0005159972380937614_dp, &
             0.001312978231978969_dp, 0.0031084065425590326_dp, 0.006931415407562781_dp, &
             0.014714390964448161_dp, 0.029978357818984606_dp, 0.05897697436502451_dp, &
             0.11257635906715585_dp, 0.20931967320582956_dp, 0.3804938243645329_dp, &
             0.6791790411475707_dp, 1.2003720104102569_dp, 2.1480477932275974_dp, &
             4.24372611094621_dp]

      CASE (485)

         aw(:) = & ! a
            [1.8216668673807486e-05_dp, 0.00011735525395160352_dp, 0.0004126460466781056_dp, &
             0.0012054450163029506_dp, 0.0031676622503733344_dp, 0.007704017636797183_dp, &
             0.017625154114990712_dp, 0.03835131607713491_dp, 0.08001204035182444_dp, &
             0.16103287923149287_dp, 0.3141590109309519_dp, 0.5964764926188472_dp, &
             1.1062697150766276_dp, 2.0134272666602406_dp, 3.627322823982276_dp, &
             6.6417493477040175_dp, &
             ! w
             4.8847988815998423e-05_dp, 0.00016674391497755278_dp, 0.00047127380690888777_dp, &
             0.001220550325011938_dp, 0.002927291261963308_dp, 0.006594638113225262_dp, &
             0.014117371583629381_dp, 0.02896560470841407_dp, 0.05732989281643629_dp, &
             0.11000837097467377_dp, 0.205490126818394_dp, 0.3750579142363867_dp, &
             0.6718894738706294_dp, 1.191217369036725_dp, 2.1373116504381717_dp, &
             4.2315165957104774_dp]

      CASE (486)

         aw(:) = & ! a
            [1.6110113710538974e-05_dp, 0.00010525743276785337_dp, 0.00037712446983756675_dp, &
             0.0011197396274799417_dp, 0.002980146520082011_dp, 0.007321198137565308_dp, &
             0.016886396920392316_dp, 0.036993331008405586_dp, 0.07762262425518779_dp, &
             0.15699688339935772_dp, 0.3076065127490934_dp, 0.5862549467587914_dp, &
             1.090976478991713_dp, 1.991542924984651_dp, 3.597440016489703_dp, &
             6.60269853853355_dp, &
             ! w
             4.333963424548889e-05_dp, 0.00015146749877325874_dp, 0.00043752467813129413_dp, &
             0.0011498795669194505_dp, 0.0027875032300416283_dp, 0.006332802126765144_dp, &
             0.013650409454938476_dp, 0.02816943790659501_dp, 0.05602936487631326_dp, &
             0.10797291643134632_dp, 0.2024444906981969_dp, 0.37072201387265813_dp, &
             0.6660605757138923_dp, 1.1838831818663529_dp, 2.1287009881566656_dp, &
             4.221724085650139_dp]

      CASE (487)

         aw(:) = & ! a
            [1.4491083002098063e-05_dp, 9.59247666285388e-05_dp, 0.00034948024161174515_dp, &
             0.0010523642863824415_dp, 0.002831438758254834_dp, 0.0070154178726742275_dp, &
             0.016292805497393058_dp, 0.035896769467608905_dp, 0.0756850447598009_dp, &
             0.15371220482820303_dp, 0.3022570004282441_dp, 0.5778871881638687_dp, &
             1.078427408303739_dp, 1.9735501359787677_dp, 3.572832988670633_dp, &
             6.570506571227025_dp, &
             ! w
             3.910366172801496e-05_dp, 0.00013962209346886036_dp, 0.0004109782817923266_dp, &
             0.0010936781095058597_dp, 0.0026754663501794324_dp, 0.006121672592111404_dp, &
             0.013272003692653493_dp, 0.027521546537137082_dp, 0.05496721303565856_dp, &
             0.10630530082841746_dp, 0.19994234611571116_dp, 0.3671512551051313_dp, &
             0.6612505016212189_dp, 1.1778214672945897_dp, 2.121577817947816_dp, &
             4.213623089833306_dp]

      CASE (488)

         aw(:) = & ! a
            [1.3204601153153415e-05_dp, 8.848400035473432e-05_dp, 0.0003272646796139465_dp, &
             0.0009977393519203389_dp, 0.0027099603181474944_dp, 0.00676409015229774_dp, &
             0.015802456580917938_dp, 0.03498711569460644_dp, 0.07407198144853272_dp, &
             0.15096927139632285_dp, 0.29777792146050547_dp, 0.570864828579562_dp, &
             1.067875166381263_dp, 1.9583953390177384_dp, 3.5520801635329433_dp, &
             6.543331606786412_dp, &
             ! w
             3.5736133476417654e-05_dp, 0.00013013362283395774_dp, 0.0003894441111243681_dp, &
             0.0010476591242880916_dp, 0.002583118648390216_dp, 0.005946749472787881_dp, &
             0.012957167327750894_dp, 0.026980578090526564_dp, 0.05407763866447738_dp, &
             0.10490492056995639_dp, 0.19783627015764157_dp, 0.36413960539165957_dp, &
             0.6571866185223857_dp, 1.1726933617135267_dp, 2.115547135173614_dp, &
             4.206764443958515_dp]

      CASE (489)

         aw(:) = & ! a
            [1.2155659454479203e-05_dp, 8.23983748572736e-05_dp, 0.00030896339837188365_dp, &
             0.0009523844991238111_dp, 0.00260842886057528_dp, 0.0065529068469690355_dp, &
             0.015388629583570622_dp, 0.03421662486963583_dp, 0.07270148210006128_dp, &
             0.14863264693165792_dp, 0.29395360397538245_dp, 0.5648571365830944_dp, &
             1.058832268452671_dp, 1.94538975252448_dp, 3.5342504978551266_dp, &
             6.519965789612569_dp, &
             ! w
             3.298926985269553e-05_dp, 0.0001223396739804258_dp, 0.00037155550618144374_dp, &
             0.0010091184074026995_dp, 0.002505333059976472_dp, 0.00579875266691652_dp, &
             0.012689822990371706_dp, 0.026519802978088165_dp, 0.05331793945992571_dp, &
             0.10370625210667844_dp, 0.19602994006852997_dp, 0.3615520738062023_dp, &
             0.6536898851571188_dp, 1.1682759321744662_dp, 2.110348795720481_dp, &
             4.200852320889001_dp]

      CASE (490)

         aw(:) = & ! a
            [7.1347100247199335e-06_dp, 5.2946703245104454e-05_dp, 0.000218126901539138_dp, &
             0.0007213605174662919_dp, 0.0020801549987173647_dp, 0.0054352820488359245_dp, &
             0.01316810092854344_dp, 0.030034571986660617_dp, 0.06519017840645336_dp, &
             0.13571934513026973_dp, 0.2726660803143451_dp, 0.5312071536236472_dp, &
             1.007909806861108_dp, 1.871824248249456_dp, 3.4330417706090337_dp, &
             6.38699885280502_dp, &
             ! w
             1.982579385525325e-05_dp, 8.402964621199333e-05_dp, 0.00028027397034065396_dp, &
             0.0008073287509403286_dp, 0.0020906122694838783_dp, 0.004998491640357875_dp, &
             0.011227464556565028_dp, 0.023974849535111926_dp, 0.04908696620605363_dp, &
             0.09698216582820675_dp, 0.18583284242914896_dp, 0.3468642560745065_dp, &
             0.6337485237556535_dp, 1.1429936952444755_dp, 2.0805350942159366_dp, &
             4.1669429466397085_dp]

      CASE (491)

         aw(:) = & ! a
            [5.292495143011322e-06_dp, 4.193432527579511e-05_dp, 0.00018273743890156237_dp, &
             0.0006278433584301215_dp, 0.0018598307089420518_dp, 0.004958188595925194_dp, &
             0.012202390742584767_dp, 0.028187800851603505_dp, 0.06183052391545048_dp, &
             0.129880299744985_dp, 0.2629500342427367_dp, 0.5157243578121021_dp, &
             0.9843175914828672_dp, 1.837545238816237_dp, 3.3856686377191174_dp, &
             6.324561223175189_dp, &
             ! w
             1.498996929777252e-05_dp, 6.932047742832677e-05_dp, 0.00024322090534047066_dp, &
             0.0007224733735920215_dp, 0.0019118771231445338_dp, 0.0046470131188411125_dp, &
             0.010575290014433768_dp, 0.022825308120321564_dp, 0.047155028503065945_dp, &
             0.09388295564999101_dp, 0.18109441566753678_dp, 0.3399906016391114_dp, &
             0.6243604836046777_dp, 1.1310366085919452_dp, 2.066397102625856_dp, &
             4.150861319105164_dp]

      CASE (492)

         aw(:) = & ! a
            [4.317737368786573e-06_dp, 3.603030473526196e-05_dp, 0.00016326116816759662_dp, &
             0.0005752261681002772_dp, 0.0017337950438442815_dp, 0.004681788157143037_dp, &
             0.011637270657435547_dp, 0.02709821874696711_dp, 0.05983478834447983_dp, &
             0.12639162975885107_dp, 0.2571161524618176_dp, 0.5063882303474644_dp, &
             0.9700396382615508_dp, 1.8167367587104382_dp, 3.3568431645157037_dp, &
             6.286505452972045_dp, &
             ! w
             1.2430307654015461e-05_dp, 6.129247538937486e-05_dp, 0.00022233856527017663_dp, &
             0.0006737259969843857_dp, 0.0018078222483448782_dp, 0.004440293218144445_dp, &
             0.01018855635085565_dp, 0.02213898574063119_dp, 0.04599491445051704_dp, &
             0.0920126454807105_dp, 0.17822252327341429_dp, 0.3358089920616348_dp, &
             0.6186312337608478_dp, 1.1237218380760685_dp, 2.057735863983757_dp, &
             4.141008826466872_dp]

      CASE (493)

         aw(:) = & ! a
            [3.7097586748543876e-06_dp, 3.2309597434583886e-05_dp, 0.0001507517917969462_dp, &
             0.0005409158661299352_dp, 0.0016506956662183882_dp, 0.004498014666298744_dp, &
             0.011259045818190115_dp, 0.026365064213282143_dp, 0.058485913189924764_dp, &
             0.12402482296873713_dp, 0.25314553559046893_dp, 0.5000163512983511_dp, &
             0.96027196458498_dp, 1.8024735420794404_dp, 3.3370542604812856_dp, &
             6.260351489602218_dp, &
             ! w
             1.0833600090655349e-05_dp, 5.616469138296074e-05_dp, 0.0002087064146430226_dp, &
             0.0006414984365406365_dp, 0.0017384232317709935_dp, 0.0043014936821916885_dp, &
             0.009927488029139939_dp, 0.02167361222899649_dp, 0.04520531727383878_dp, &
             0.09073556164946321_dp, 0.17625604547943877_dp, 0.33293876490726154_dp, &
             0.614690683320622_dp, 1.1186828635705408_dp, 2.051763837867852_dp, &
             4.134215156273755_dp]

      CASE (494)

         aw(:) = & ! a
            [3.2930743379663077e-06_dp, 2.973765279863682e-05_dp, 0.00014197671984500713_dp, &
             0.0005165767657839471_dp, 0.0015912686754121065_dp, 0.004365792626710519_dp, &
             0.010985622152762794_dp, 0.025833013257759798_dp, 0.05750390352753833_dp, &
             0.1222970878012441_dp, 0.25024036099655467_dp, 0.4953450245798517_dp, &
             0.953099040349521_dp, 1.7919846097539986_dp, 3.3224858350837563_dp, &
             6.241082239424807_dp, &
             ! w
             9.739213156547932e-06_dp, 5.2581999287346234e-05_dp, 0.00019902806642331698_dp, &
             0.0006184076325633332_dp, 0.0016883824697376546_dp, 0.004200925575806819_dp, &
             0.009737596138419492_dp, 0.0213340343901398_dp, 0.04462760416501207_dp, &
             0.08979901588188346_dp, 0.17481104481754212_dp, 0.33082601532273864_dp, &
             0.611785841611682_dp, 1.1149641284389784_dp, 2.0473536111747674_dp, &
             4.129198019606889_dp]

      CASE (495)

         aw(:) = & ! a
            [2.9896479541292816e-06_dp, 2.7851055603288865e-05_dp, 0.00013546361256197895_dp, &
             0.0004983536248044651_dp, 0.001546497654595369_dp, 0.004265715964919928_dp, &
             0.010777920291979073_dp, 0.02542766404892414_dp, 0.056753932908308456_dp, &
             0.12097490151331013_dp, 0.24801323563897704_dp, 0.4917585962955728_dp, &
             0.947584973821342_dp, 1.7839128692611583_dp, 3.3112654165046287_dp, &
             6.2262326184266_dp, &
             ! w
             8.94223582088384e-06_dp, 4.993085150060451e-05_dp, 0.0001917771510923769_dp, &
             0.0006009867661120243_dp, 0.0016504453883887842_dp, 0.0041244001976364015_dp, &
             0.009592675308628475_dp, 0.02107424869671021_dp, 0.044184736741917564_dp, &
             0.08907981609994198_dp, 0.17369970647447416_dp, 0.3291989846650942_dp, &
             0.6095463498692334_dp, 1.1120947290809624_dp, 2.043948954094459_dp, &
             4.125324747802907_dp]

      CASE (496)

         aw(:) = & ! a
            [2.7593120484618118e-06_dp, 2.6409810944394424e-05_dp, 0.00013043952644429906_dp, &
             0.0004841981765006107_dp, 0.0015115483186980657_dp, 0.004187306415159171_dp, &
             0.010614720740225778_dp, 0.02510842918438578_dp, 0.05616216075709516_dp, &
             0.11992993708019088_dp, 0.24625065481439828_dp, 0.4889169056938238_dp, &
             0.9432115524053757_dp, 1.7775055327488363_dp, 3.302352861111834_dp, &
             6.214431915762369_dp, &
             ! w
             8.337184243333709e-06_dp, 4.789063789498023e-05_dp, 0.00018614195585213908_dp, &
             0.0005873725697229844_dp, 0.0016206838870552707_dp, 0.004064191206663_dp, &
             0.009478388745528548_dp, 0.020868986305975362_dp, 0.04383425426498532_dp, &
             0.08850986241484268_dp, 0.17281794175043663_dp, 0.32790672887321026_dp, &
             0.6077661120092573_dp, 1.1098122379302886_dp, 2.041239624003451_dp, &
             4.122242450546229_dp]

      CASE (497)

         aw(:) = & ! a
            [2.5792221918675493e-06_dp, 2.5276685041551083e-05_dp, 0.00012645736570831166_dp, &
             0.00047291402596101655_dp, 0.0014835761709221854_dp, 0.004124363185313988_dp, &
             0.010483408537352659_dp, 0.024851089418092605_dp, 0.05568438846848649_dp, &
             0.11908518022232199_dp, 0.24482419469180772_dp, 0.48661494040162206_dp, &
             0.9396659210455454_dp, 1.7723074812037942_dp, 3.2951186186538286_dp, &
             6.204849848808785_dp, &
             ! w
             7.864064861781991e-06_dp, 4.627659460952986e-05_dp, 0.00018164800261044056_dp, &
             0.0005764666295898136_dp, 0.0015967684440860396_dp, 0.004015694667747015_dp, &
             0.009386161355321115_dp, 0.020703086972213142_dp, 0.04355061575240213_dp, &
             0.08804809802716361_dp, 0.17210286933754812_dp, 0.3268578982044417_dp, &
             0.6063202142506676_dp, 1.1079574175844626_dp, 2.039037244492888_dp, &
             4.119736850294895_dp]

      CASE (498)

         aw(:) = & ! a
            [2.4354136243317077e-06_dp, 2.436740109204857e-05_dp, 0.00012323987512766869_dp, &
             0.00046375324881587233_dp, 0.0014607921847590419_dp, 0.004072968279409224_dp, &
             0.010375983633463258_dp, 0.024640239423978902_dp, 0.055292431712294324_dp, &
             0.11839141619367698_dp, 0.24365163823066568_dp, 0.48472124409945827_dp, &
             0.936747200836527_dp, 1.7680261579625183_dp, 3.2891576433756384_dp, &
             6.196951887474531_dp, &
             ! w
             7.486209571735704e-06_dp, 4.4974493230726e-05_dp, 0.0001779984909799321_dp, &
             0.000567577081138818_dp, 0.0015772245267925906_dp, 0.00397598562426191_dp, &
             0.009310528653457151_dp, 0.020566865562055732_dp, 0.043317469396910266_dp, &
             0.08766818952983264_dp, 0.17151409332861423_dp, 0.3259937254252641_dp, &
             0.6051282005497428_dp, 1.106427609421274_dp, 2.0372203067713297_dp, &
             4.117669733570923_dp]

      CASE (499)

         aw(:) = & ! a
            [1.9085961715955434e-06_dp, 2.0995043960013904e-05_dp, 0.00011111398867441926_dp, &
             0.00042885255913484543_dp, 0.0013733360820866016_dp, 0.003874593789234121_dp, &
             0.009959559027290782_dp, 0.023820069278143882_dp, 0.0537634425665394_dp, &
             0.11567862202124427_dp, 0.23905729559633404_dp, 0.47728836137347114_dp, &
             0.9252740091801204_dp, 1.7511759472651802_dp, 3.2656740667243875_dp, &
             6.165816271073438_dp, &
             ! w
             6.101229797346719e-06_dp, 4.00837025306848e-05_dp, 0.0001640844444725366_dp, &
             0.00053340032030255_dp, 0.00150164908261384_dp, 0.003821757919080607_dp, &
             0.009015751682969649_dp, 0.020034430842349286_dp, 0.04240401234061361_dp, &
             0.08617667866638216_dp, 0.1691984867198657_dp, 0.3225898169116341_dp, &
             0.6004269137751367_dp, 1.1003880758298388_dp, 2.0300430086921573_dp, &
             4.109503927013779_dp]

      CASE (500)

         aw(:) = & ! a
            [1.908279945252509e-06_dp, 2.099299427495672e-05_dp, 0.00011110651444748947_dp, &
             0.00042883085533252464_dp, 0.0013732813892174006_dp, 0.0038744692237789734_dp, &
             0.00995929668391619_dp, 0.023819551138145992_dp, 0.05376247448222308_dp, &
             0.11567690110584063_dp, 0.23905437629412185_dp, 0.47728363290101083_dp, &
             0.9252667046693415_dp, 1.7511652086513543_dp, 3.265659074902631_dp, &
             6.1657963486574845_dp, &
             ! w
             6.10039728638634e-06_dp, 4.008069629625031e-05_dp, 0.00016407578363587837_dp, &
             0.000533378918013126_dp, 0.0015016015641512589_dp, 0.0038216606296993015_dp, &
             0.009015565175734293_dp, 0.02003409319715033_dp, 0.042403432006524695_dp, &
             0.08617572939827894_dp, 0.16919701115096455_dp, 0.32258764719545124_dp, &
             0.6004239149409639_dp, 1.100384212656272_dp, 2.0300383954052896_dp, &
             4.109498654374117_dp]

      CASE (501)

         aw(:) = & ! a
            [0.011763649966993063_dp, 0.062428205737342514_dp, 0.15544990740250392_dp, &
             0.29427287137907343_dp, 0.4843511156615427_dp, 0.7337793906644704_dp, &
             1.0541795508192826_dp, 1.4618861427706884_dp, 1.9794504252494227_dp, &
             2.63752178585389_dp, 3.4773350777344394_dp, 4.55440587644354_dp, &
             5.944785288108488_dp, 7.75703952819382_dp, 10.158682924645307_dp, &
             13.44791160523966_dp, 18.338684467263775_dp, &
             ! w
             0.030239769948628595_dp, 0.07141117296934633_dp, 0.11519834133086981_dp, &
             0.16334634556994845_dp, 0.2181445810861672_dp, 0.2826292106846991_dp, &
             0.3608685543158396_dp, 0.45827228192839375_dp, 0.5819243715708697_dp, &
             0.7410321325985205_dp, 0.9477492570824826_dp, 1.2188845101312764_dp, &
             1.57958056177065_dp, 2.071871446036535_dp, 2.7779101419370362_dp, &
             3.901959355820747_dp, 6.2568211133691465_dp]

      CASE (502)

         aw(:) = & ! a
            [0.008620957574467123_dp, 0.04582067200269628_dp, 0.11442264748254989_dp, &
             0.21754831945248493_dp, 0.36023713511117667_dp, 0.5501295551911949_dp, &
             0.7984222655089384_dp, 1.1211082157416934_dp, 1.540490816448991_dp, &
             2.0870398051392733_dp, 2.801872780790989_dp, 3.7405073580248507_dp, &
             4.979201454264444_dp, 6.626927101387456_dp, 8.851486285748587_dp, &
             11.950038790447127_dp, 16.629053519985472_dp, &
             ! w
             0.02216899543476805_dp, 0.05251363169828869_dp, 0.085202491882224_dp, &
             0.1218775877465484_dp, 0.16475754546913823_dp, 0.21687267548632566_dp, &
             0.2823530993770712_dp, 0.3667057338149049_dp, 0.47709784588892884_dp, &
             0.6227933403263984_dp, 0.8160260402127403_dp, 1.0737693823036798_dp, &
             1.421396979956243_dp, 1.9010620007560268_dp, 2.5944356141359597_dp, &
             3.703154472169265_dp, 6.026307922042811_dp]

      CASE (503)

         aw(:) = & ! a
            [0.006889377762839884_dp, 0.03666190803574594_dp, 0.09176037150749558_dp, &
             0.17507037815620646_dp, 0.2913192809176048_dp, 0.44778720445993453_dp, &
             0.6552926902022264_dp, 0.9294428610717238_dp, 1.2921313629933526_dp, &
             1.7733859123770381_dp, 2.413897575278871_dp, 3.2688815574094914_dp, &
             4.414534121752955_dp, 5.960042491831377_dp, 8.073518859029836_dp, &
             11.051769565150513_dp, 15.597025220046245_dp, &
             ! w
             0.017721171780941552_dp, 0.04208044998936166_dp, 0.06858925211169453_dp, &
             0.09880696703542288_dp, 0.13488771487291526_dp, 0.1798302894659236_dp, &
             0.23775709667732445_dp, 0.3141621682693045_dp, 0.4161787636643034_dp, &
             0.5530505032238192_dp, 0.7370736963599462_dp, 0.9854093744573996_dp, &
             1.3237121967377334_dp, 1.7943741588876145_dp, 2.478935025274032_dp, &
             3.5775455854296867_dp, 5.880876156058426_dp]

      CASE (504)

         aw(:) = & ! a
            [0.005779061268096959_dp, 0.030784849069456433_dp, 0.07719912472356236_dp, &
             0.14772533472747335_dp, 0.24684498770688412_dp, 0.38154420376776327_dp, &
             0.562310957764068_dp, 0.8043777658932391_dp, 1.1292008564500715_dp, &
             1.5663159266865059_dp, 2.155932832858683_dp, 2.9529057092101003_dp, &
             4.033296773451878_dp, 5.506429240336444_dp, 7.540703753038025_dp, &
             10.432830899152323_dp, 14.882266755973522_dp, &
             ! w
             0.01486866603880088_dp, 0.035379580456098146_dp, 0.057891335478331786_dp, &
             0.08389601665202415_dp, 0.11549092435956188_dp, 0.15563470586681194_dp, &
             0.20841320749677028_dp, 0.2792663830727313_dp, 0.3752605918790275_dp, &
             0.5056075720898942_dp, 0.6826650143959483_dp, 0.9237725888705499_dp, &
             1.254850152598101_dp, 1.7185396923469536_dp, 2.3963756361848145_dp, &
             3.487537801182031_dp, 5.776796760268069_dp]

      CASE (505)

         aw(:) = & ! a
            [0.005000767488016341_dp, 0.02666265026927288_dp, 0.06697428709127144_dp, &
             0.12849263546697673_dp, 0.21549854969570803_dp, 0.3347316344380094_dp, &
             0.49638900920800944_dp, 0.7153535743415229_dp, 1.0126577716550769_dp, &
             1.4173574351365212_dp, 1.9691940995077597_dp, 2.7226634022245193_dp, &
             3.7536719703944716_dp, 5.171634173848079_dp, 7.145208539203496_dp, &
             9.971124475461954_dp, 14.34686193502478_dp, &
             ! w
             0.012868864174744343_dp, 0.030675943104970887_dp, 0.05036525777390434_dp, &
             0.073372753832673_dp, 0.10174556517314705_dp, 0.13839960202350926_dp, &
             0.18737164472211038_dp, 0.2540325726980337_dp, 0.3453754544025245_dp, &
             0.47058211473370076_dp, 0.642068590905447_dp, 0.8773350492032141_dp, &
             1.202538866223134_dp, 1.6605588690218929_dp, 2.3329791005813414_dp, &
             3.418291325722145_dp, 5.696809904804326_dp]

      CASE (506)

         aw(:) = & ! a
            [0.0044220826490264055_dp, 0.02359598421362561_dp, 0.05936008899326881_dp, &
             0.1141499987320069_dp, 0.19207831536019496_dp, 0.2996733309241805_dp, &
             0.4468736581015505_dp, 0.648239768682221_dp, 0.9244066057150568_dp, &
             1.3039805390235892_dp, 1.826265661366115_dp, 2.545418158136293_dp, &
             3.5371843477475413_dp, 4.911040657049191_dp, 6.835874198167441_dp, &
             9.608488910004654_dp, 13.924877699901952_dp, &
             ! w
             0.011381759979217747_dp, 0.027174359344044833_dp, 0.04475154671090715_dp, &
             0.06550144903949855_dp, 0.09142650354599316_dp, 0.12539976786986748_dp, &
             0.17140391911969466_dp, 0.23473676962871573_dp, 0.32232050377890237_dp, &
             0.4433112013591309_dp, 0.6101784217262354_dp, 0.8405640412966287_dp, &
             1.1608360910097697_dp, 1.6140922256599763_dp, 2.281992414549576_dp, &
             3.3625148353540113_dp, 5.632440674010009_dp]

      CASE (507)

         aw(:) = & ! a
            [0.0039733564241690176_dp, 0.021216843249974877_dp, 0.05344768383817142_dp, &
             0.10299863204844592_dp, 0.1738382188340671_dp, 0.2723103794523001_dp, &
             0.40812171683281745_dp, 0.5955361059459275_dp, 0.8548199579708161_dp, &
             1.2141643235424078_dp, 1.7124696860301576_dp, 2.403575315218784_dp, &
             3.3630688929233137_dp, 4.700468631856519_dp, 6.584864495752914_dp, &
             9.313163003961266_dp, 13.580193642335534_dp, &
             ! w
             0.010228490409033503_dp, 0.024456161592103284_dp, 0.04038608186423737_dp, &
             0.05936493428426412_dp, 0.08335494359406384_dp, 0.11518728600334556_dp, &
             0.15878918103150655_dp, 0.21938622069978075_dp, 0.30383483930485183_dp, &
             0.42126916739472764_dp, 0.5842068690607539_dp, 0.8104146081742157_dp, &
             1.1264476636364262_dp, 1.5756047826778998_dp, 2.2396347826394383_dp, &
             3.316118284642388_dp, 5.578937934634807_dp]

      CASE (508)

         aw(:) = & ! a
            [0.0036142847601028185_dp, 0.019312197936351656_dp, 0.04871062565606132_dp, &
             0.09405361299570067_dp, 0.15918429328598857_dp, 0.25028350487151524_dp, &
             0.3768478515919124_dp, 0.5528681138914167_dp, 0.7982702270194616_dp, &
             1.140862874047967_dp, 1.6191750225339965_dp, 2.286751024183023_dp, &
             3.2190241844884033_dp, 4.525539770823869_dp, 6.3755685758961_dp, &
             9.06613362289836_dp, 13.291127529467497_dp, &
             ! w
             0.00930554527992734_dp, 0.022278887164099686_dp, 0.03688376103369464_dp, &
             0.05443042467660224_dp, 0.07684460681041379_dp, 0.10691705032991418_dp, &
             0.14852004495201218_dp, 0.20680991906152918_dp, 0.28858301637420974_dp, &
             0.402954506487906_dp, 0.5624848391441751_dp, 0.7850508763672497_dp, &
             1.0973753054331812_dp, 1.5429418443905634_dp, 2.2035942157384123_dp, &
             3.2765969845607468_dp, 5.533394223709508_dp]

      CASE (509)

         aw(:) = & ! a
            [0.0033198247340284217_dp, 0.017749631331696807_dp, 0.044821455936419015_dp, &
             0.08670170571507395_dp, 0.1471228721624465_dp, 0.2321198066274997_dp, &
             0.35099758950161103_dp, 0.5174948868641361_dp, 0.7512229986159361_dp, &
             1.0796383396633986_dp, 1.540927319309597_dp, 2.1883595496344137_dp, &
             3.0972191923966146_dp, 4.377067543193302_dp, 6.197337489998881_dp, &
             8.855174803475688_dp, 13.043700459848505_dp, &
             ! w
             0.008548602711241414_dp, 0.020491752428293177_dp, 0.03400477083242754_dp, &
             0.05036551539464545_dp, 0.07146632918399287_dp, 0.10005915841952623_dp, &
             0.13996286257638943_dp, 0.19626829904821647_dp, 0.27571710170098357_dp, &
             0.38740772532883344_dp, 0.5439384928815124_dp, 0.7632840773608892_dp, &
             1.0723178584501674_dp, 1.514694260200086_dp, 2.172354226390208_dp, &
             3.2423058243498173_dp, 5.493900885483596_dp]

      CASE (510)

         aw(:) = & ! a
            [0.0018882627435518682_dp, 0.010141716994752921_dp, 0.02583472761831859_dp, &
             0.05066872653408917_dp, 0.08769367251778923_dp, 0.14199794933436113_dp, &
             0.22157713402556642_dp, 0.3384059441793975_dp, 0.509905782056896_dp, &
             0.7610963644983121_dp, 1.1277658774087937_dp, 1.6611792947718913_dp, &
             2.4354020371117624_dp, 3.559892362248305_dp, 5.205100219607091_dp, &
             7.669306130543646_dp, 11.641887484729258_dp, &
             ! w
             0.00486734104806698_dp, 0.011774633045678043_dp, 0.019887079142862772_dp, &
             0.030277600427505283_dp, 0.04460586633113461_dp, 0.06531793316988077_dp, &
             0.09581248220465682_dp, 0.1407135799336433_dp, 0.20640432448587542_dp, &
             0.3018653172872778_dp, 0.43990111840158036_dp, 0.6390746417174834_dp, &
             0.9272349393169635_dp, 1.34924330660015_dp, 1.9879267827114793_dp, &
             3.039156813544293_dp, 5.260387230811638_dp]

      CASE (511)

         aw(:) = & ! a
            [0.0013527924373737048_dp, 0.0072886986335998_dp, 0.018681297463412815_dp, &
             0.036999273564417796_dp, 0.06493527017972293_dp, 0.10704888590724806_dp, &
             0.17056757629874036_dp, 0.26642384857892054_dp, 0.4107577410338992_dp, &
             0.6271579934161993_dp, 0.9499639456298539_dp, 1.4291576240431079_dp, &
             2.1379244969668703_dp, 3.1854886763815116_dp, 4.742811677117564_dp, &
             7.109007686904927_dp, 10.972131782040323_dp, &
             ! w
             0.003489557005875298_dp, 0.008495377107671717_dp, 0.014527041401980967_dp, &
             0.022546802751019696_dp, 0.03407194037548306_dp, 0.0513432024006064_dp, &
             0.07749186527362117_dp, 0.11687403369619569_dp, 0.1756713275412031_dp, &
             0.26277313531696966_dp, 0.39104830780991007_dp, 0.5793369494164639_dp, &
             0.8560194482301063_dp, 1.2666966821386287_dp, 1.894870709031309_dp, &
             2.936128075912932_dp, 5.142256045981233_dp]

      CASE (512)

         aw(:) = & ! a
            [0.0010664804747085547_dp, 0.00576045171825998_dp, 0.014836817086855152_dp, &
             0.029616562839403903_dp, 0.05255860019576755_dp, 0.0878641811279735_dp, &
             0.14223231648066315_dp, 0.22588150203936602_dp, 0.3540758486226656_dp, &
             0.5494155210276604_dp, 0.8452157720214318_dp, 1.2905193491966398_dp, &
             1.9578266197559235_dp, 2.956120434508262_dp, 4.456673816212555_dp, &
             6.75923087522096_dp, 10.551209538157856_dp, &
             ! w
             0.0027525568817549193_dp, 0.006734925870965941_dp, 0.011630694813380601_dp, &
             0.018328397086009855_dp, 0.028243588680285404_dp, 0.04346768068443954_dp, &
             0.066946825683937_dp, 0.10285810200166813_dp, 0.15724216557766652_dp, &
             0.23890847369320603_dp, 0.3607421430923924_dp, 0.5417453516089881_dp, &
             0.8106492863971688_dp, 1.2135804900968097_dp, 1.8345705538017225_dp, &
             2.8691445129195685_dp, 5.065557756580005_dp]

      CASE (513)

         aw(:) = & ! a
            [0.0008863304499981982_dp, 0.004797478687156452_dp, 0.012407937782549144_dp, &
             0.024933671543423255_dp, 0.04466350896406489_dp, 0.07553210536955951_dp, &
             0.1238433200134319_dp, 0.19928383025465143_dp, 0.31646429392274816_dp, &
             0.49723949290802594_dp, 0.7741388461621441_dp, 1.195466624778922_dp, &
             1.8331629335494573_dp, 2.7959834160555097_dp, 4.255410703916915_dp, &
             6.511688762572313_dp, 10.251886422940728_dp, &
             ! w
             0.0022886748517332673_dp, 0.00562368283450409_dp, 0.009792873303731632_dp, &
             0.015630407664363275_dp, 0.02447336577695618_dp, 0.03829819219676908_dp, &
             0.0599135544832279_dp, 0.09336464230798491_dp, 0.1445828113357693_dp, &
             0.22230646763359777_dp, 0.3394170592751431_dp, 0.5150227219028766_dp, &
             0.7781099899771061_dp, 1.1752094119195156_dp, 1.7907855704489537_dp, &
             2.8203860853302944_dp, 5.009777496447649_dp]

      CASE (514)

         aw(:) = & ! a
            [0.0007617277641131662_dp, 0.004130617806061752_dp, 0.010722167674873705_dp, &
             0.021672433881065812_dp, 0.03913841256538039_dp, 0.06684502062270839_dp, &
             0.11078483737003565_dp, 0.18022671433961276_dp, 0.28926629000646814_dp, &
             0.45916520195104404_dp, 0.7218187960719699_dp, 1.124924555369314_dp, &
             1.7399490705067022_dp, 2.675438183296255_dp, 4.103024447436759_dp, &
             6.323363228630736_dp, 10.023320633186344_dp, &
             ! w
             0.0019677354874388323_dp, 0.004853001746142512_dp, 0.008512641693891775_dp, &
             0.013738205589199897_dp, 0.02180334470995126_dp, 0.034592412327327654_dp, &
             0.05480672949392181_dp, 0.08638807202337442_dp, 0.13517816253357462_dp, &
             0.20985174763033818_dp, 0.32327734113038836_dp, 0.49463718513740845_dp, &
             0.7531150452829314_dp, 1.1455673736824923_dp, 1.7568239419680498_dp, &
             2.782491770486839_dp, 4.966453841342643_dp]

      CASE (515)

         aw(:) = & ! a
            [0.000670014476793155_dp, 0.003639256680814586_dp, 0.00947761441997857_dp, &
             0.01925754176216679_dp, 0.035029474705086584_dp, 0.06034701897970911_dp, &
             0.10094857952065324_dp, 0.16576236607034503_dp, 0.2684625388540078_dp, &
             0.42982118843709644_dp, 0.6812043753394367_dp, 1.0697957652746826_dp, &
             1.666652242217474_dp, 2.580126489097479_dp, 3.981962769209941_dp, &
             6.17316428323695_dp, 9.840477650921452_dp, &
             ! w
             0.0017314513798382411_dp, 0.004284407889281309_dp, 0.007564451405461784_dp, &
             0.012328362033896246_dp, 0.0197968832453667_dp, 0.031778415851714335_dp, &
             0.05088734010781185_dp, 0.08098084577261747_dp, 0.12782454895000084_dp, &
             0.20003572183631715_dp, 0.31046542437755614_dp, 0.47835010535418493_dp, &
             0.7330321104341379_dp, 1.12163962047183_dp, 1.7293175542134887_dp, &
             2.7517493678958247_dp, 4.931324161744377_dp]

      CASE (516)

         aw(:) = & ! a
            [0.0005994659006691419_dp, 0.003260929328746702_dp, 0.008517683368674414_dp, &
             0.017389897955788717_dp, 0.03183926194759523_dp, 0.05527562479185905_dp, &
             0.09322430615404195_dp, 0.15432815172523684_dp, 0.25190682909537176_dp, &
             0.4063175936906406_dp, 0.6484737266112314_dp, 1.0251141636009289_dp, &
             1.6069349856798574_dp, 2.502110635448886_dp, 3.8824710537171936_dp, &
             6.049319622362817_dp, 9.689334495884646_dp, &
             ! w
             0.0015496548496931517_dp, 0.0038461093612214814_dp, 0.006831009858755705_dp, &
             0.011231934737732102_dp, 0.018224474401726948_dp, 0.02955297127098162_dp, &
             0.04775941289319049_dp, 0.07662968471095458_dp, 0.12186323208611664_dp, &
             0.192025057361598_dp, 0.29994658510881267_dp, 0.4649051786447571_dp, &
             0.7163744310297717_dp, 1.1017145944498874_dp, 1.7063473740013237_dp, &
             2.7260403665683968_dp, 4.9019578207789705_dp]

      CASE (517)

         aw(:) = & ! a
            [0.0005433808812488209_dp, 0.0029599069426816398_dp, 0.007752684216764555_dp, &
             0.015897848233704857_dp, 0.02928149980759579_dp, 0.05119036260266149_dp, &
             0.08696747345379538_dp, 0.14501159846505923_dp, 0.23833799205350642_dp, &
             0.3869453058931429_dp, 0.6213522337346994_dp, 0.987906301446031_dp, &
             1.5569814737710108_dp, 2.436587099282762_dp, 3.7986202418785147_dp, &
             5.944647520478597_dp, 9.561311563070827_dp, &
             ! w
             0.0014051007766722966_dp, 0.0034970054384386865_dp, 0.006244986422830312_dp, &
             0.010351561402402542_dp, 0.016953108432719518_dp, 0.02773897849645473_dp, &
             0.04518956343570459_dp, 0.07302927511499638_dp, 0.11689902572804858_dp, &
             0.18531596227571526_dp, 0.2910909270989169_dp, 0.4535328802956833_dp, &
             0.702226457158761_dp, 1.0847337738721048_dp, 1.6867230015056864_dp, &
             2.7040488024480958_dp, 4.876845927033721_dp]

      CASE (518)

         aw(:) = & ! a
            [0.0004976403321071418_dp, 0.002714212750478709_dp, 0.007127380424846257_dp, &
             0.01467548508745452_dp, 0.027179113748376327_dp, 0.04781782740021432_dp, &
             0.08177621466071466_dp, 0.13724081513199407_dp, 0.22696117487490997_dp, &
             0.37062107412592427_dp, 0.5983902924464397_dp, 0.9562673114382849_dp, &
             1.5143353451316333_dp, 2.380450186248581_dp, 3.7265623220764432_dp, &
             5.85447219564017_dp, 9.450809001670763_dp, &
             ! w
             0.0012871872080564162_dp, 0.003211794161382628_dp, 0.005764831078351174_dp, &
             0.00962694414424929_dp, 0.015899986192882947_dp, 0.026225417129850274_dp, &
             0.04303030664412398_dp, 0.06998514403816393_dp, 0.11267837845960989_dp, &
             0.17958307908069895_dp, 0.28348923994622244_dp, 0.44373063669581314_dp, &
             0.6899874875249782_dp, 1.070000048658365_dp, 1.669658407541783_dp, &
             2.684904664172808_dp, 4.8549914587310425_dp]

      CASE (519)

         aw(:) = & ! a
            [0.00027873969184184674_dp, 0.00153509211230705_dp, 0.004110690792904548_dp, &
             0.008729310582552594_dp, 0.016827760292598675_dp, 0.030952553619796153_dp, &
             0.055356499116054905_dp, 0.09697227413991218_dp, 0.16695506247712322_dp, &
             0.28305998361524637_dp, 0.4732691834136238_dp, 0.7813315385337429_dp, &
             1.2753824306906016_dp, 2.0621523950909895_dp, 3.3137957600148757_dp, &
             5.333592025630548_dp, 8.808438491898944_dp, &
             ! w
             0.000722522787781202_dp, 0.0018383479865653803_dp, 0.0034284305708949456_dp, &
             0.006041941094276915_dp, 0.010569894375873_dp, 0.018372683734438747_dp, &
             0.031565549776009595_dp, 0.053486530005002426_dp, 0.08937927959131456_dp, &
             0.1474044567948892_dp, 0.24016684640466546_dp, 0.38709004953734244_dp, &
             0.6183966764795468_dp, 0.9829329898408707_dp, 1.568061487599537_dp, &
             2.5704877366016645_dp, 4.724484641752887_dp]

      CASE (520)

         aw(:) = & ! a
            [0.0001984953473166477_dp, 0.0011007426306861563_dp, 0.0029892040618499705_dp, &
             0.0064860166922959565_dp, 0.012839078088964787_dp, 0.024282422724072522_dp, &
             0.044610697853502225_dp, 0.08013208455353914_dp, 0.14118815552223068_dp, &
             0.24452183065262578_dp, 0.41693201821252734_dp, 0.7009081577243788_dp, &
             1.1634433815360945_dp, 1.9105417240992761_dp, 3.1143693332951914_dp, &
             5.0790073624075776_dp, 8.491728023012124_dp, &
             ! w
             0.0005152964505271766_dp, 0.0013294206789940207_dp, 0.00254676554648255_dp, &
             0.004649177220738149_dp, 0.008420105358948896_dp, 0.015083554228289199_dp, &
             0.02659967787776162_dp, 0.04612823689556763_dp, 0.07871505212539648_dp, &
             0.13232748859318125_dp, 0.21943270040383067_dp, 0.35945626577909584_dp, &
             0.5828720549538152_dp, 0.939113828916259_dp, 1.5163956265285226_dp, &
             2.5119855345932347_dp, 4.657820695446526_dp]

      CASE (521)

         aw(:) = & ! a
            [0.0001560036351335973_dp, 0.0008699543072648038_dp, 0.002389404106247081_dp, &
             0.00527339002010383_dp, 0.010650437240237379_dp, 0.02055766557308312_dp, &
             0.038500719124384054_dp, 0.0703892927052089_dp, 0.12603760027070512_dp, &
             0.22152163651719636_dp, 0.38284726583797_dp, 0.6516434522803097_dp, &
             1.0941030700653211_dp, 1.8156960891012024_dp, 2.988555833847681_dp, &
             4.917298397923595_dp, 8.28952974750375_dp, &
             ! w
             0.0004054777725141608_dp, 0.001057885055165663_dp, 0.0020701844217904607_dp, &
             0.0038805823916970897_dp, 0.007204052901643639_dp, 0.013179504777142824_dp, &
             0.023667219999698796_dp, 0.041707389259586766_dp, 0.07220904546186369_dp, &
             0.12300112756582732_dp, 0.20644444370191073_dp, 0.34194772283335984_dp, &
             0.5601360518319131_dp, 0.9108314795281586_dp, 1.4828393697982243_dp, &
             2.4738623767994983_dp, 4.614399607769071_dp]

      CASE (522)

         aw(:) = & ! a
            [0.00012942660369669388_dp, 0.0007252109691484227_dp, 0.0020112414165164147_dp, &
             0.004502205499976117_dp, 0.009241924957420352_dp, 0.018128200916790103_dp, &
             0.034461652810250175_dp, 0.06386657781717088_dp, 0.11577518861573856_dp, &
             0.2057748375160446_dp, 0.35928348369804525_dp, 0.6172838037451012_dp, &
             1.0453574126507788_dp, 1.748553561064432_dp, 2.8989593755718_dp, &
             4.80158540961176_dp, 8.14432591416385_dp, &
             ! w
             0.0003367471971412926_dp, 0.0008870248350176341_dp, 0.0017671250240551047_dp, &
             0.0033837180880403332_dp, 0.006403200196190485_dp, 0.011904536113767494_dp, &
             0.021675708252598944_dp, 0.03866812999138343_dp, 0.0676873864573128_dp, &
             0.1164553957031185_dp, 0.1972469731352802_dp, 0.3294489379855182_dp, &
             0.5437893337218036_dp, 0.8903748071315685_dp, 1.4584594866065748_dp, &
             2.4460979032403842_dp, 4.582786015781227_dp]

      CASE (523)

         aw(:) = & ! a
            [0.00011111963314320988_dp, 0.000625278687246286_dp, 0.0017489792522994028_dp, &
             0.00396340573031236_dp, 0.00824803957348528_dp, 0.016395107811091034_dp, &
             0.0315493517865167_dp, 0.05911635446494708_dp, 0.1082330020586242_dp, &
             0.1941055905637425_dp, 0.34168954280208236_dp, 0.5914540652570976_dp, &
             1.0084890439982894_dp, 1.6974976864246518_dp, 2.8305173366405514_dp, &
             4.712867256904743_dp, 8.032692308689725_dp, &
             ! w
             0.00028937880586602514_dp, 0.0007687334262258107_dp, 0.0015554121481513774_dp, &
             0.00303179736256024_dp, 0.0058274883277793475_dp, 0.010976070665113313_dp, &
             0.02020954884085334_dp, 0.03640939553666422_dp, 0.06429864525039648_dp, &
             0.1115124584041263_dp, 0.19025376465279834_dp, 0.3198864932433305_dp, &
             0.5312141007951445_dp, 0.8745650190632126_dp, 1.4395525665351276_dp, &
             2.424525925156707_dp, 4.558228384233685_dp]

      CASE (524)

         aw(:) = & ! a
            [9.768541498147083e-05_dp, 0.0005517983285910077_dp, 0.0015553737137321002_dp, &
             0.0035630585294778006_dp, 0.007503206642130452_dp, 0.015084302733820118_dp, &
             0.029327016433648698_dp, 0.05546166717945008_dp, 0.10238679885885461_dp, &
             0.18499904005737647_dp, 0.3278752751518903_dp, 0.5710612168073389_dp, &
             0.9792371997555712_dp, 1.656813253858836_dp, 2.775776972366038_dp, &
             4.641699008772861_dp, 7.942944658981498_dp, &
             ! w
             0.0002546024856868256_dp, 0.0006815428354254016_dp, 0.0013981202324835494_dp, &
             0.002767207087865029_dp, 0.0053892678391408485_dp, 0.010261856997860481_dp, &
             0.019071700100119625_dp, 0.03464297025325642_dp, 0.061630412276498785_dp, &
             0.10759654307172731_dp, 0.18468271032482161_dp, 0.31223039308565087_dp, &
             0.5211010386267269_dp, 0.8618030928018889_dp, 1.4242478034182446_dp, &
             2.4070372300275924_dp, 4.538322169383284_dp]

      CASE (525)

         aw(:) = & ! a
            [8.73752485729232e-05_dp, 0.0004953046592686002_dp, 0.0014059944165274103_dp, &
             0.0032523542512961734_dp, 0.006920780986677007_dp, 0.014051128308219434_dp, &
             0.027562034504967203_dp, 0.052538848222604025_dp, 0.09768179259767383_dp, &
             0.17762834439838698_dp, 0.31663675253655615_dp, 0.554393972308045_dp, &
             0.955230608435182_dp, 1.6233028811680008_dp, 2.730550113923314_dp, &
             4.582753592147178_dp, 7.868474720030727_dp, &
             ! w
             0.00022790231031207566_dp, 0.0006143635134771986_dp, 0.001276061383204444_dp, &
             0.002559718460344272_dp, 0.005041980338497405_dp, 0.009690807264611922_dp, &
             0.01815515505247096_dp, 0.03321092592577857_dp, 0.05945488947841082_dp, &
             0.10438728849400189_dp, 0.18009572483079425_dp, 0.3059001779584581_dp, &
             0.5127082722054082_dp, 0.8511788881930034_dp, 1.4114768032680935_dp, &
             2.3924251516095367_dp, 4.521692150398395_dp]

      CASE (526)

         aw(:) = & ! a
            [7.91937539895247e-05_dp, 0.00045040231500375445_dp, 0.0012868784814711026_dp, &
             0.0030032754553129806_dp, 0.006450716747020653_dp, 0.013211411470947522_dp, &
             0.026118015886734368_dp, 0.05013308761410789_dp, 0.0937880163778335_dp, &
             0.17149857922072664_dp, 0.3072490907067271_dp, 0.5404164110122928_dp, &
             0.9350268144743565_dp, 1.5950131357874375_dp, 2.6922684555193763_dp, &
             4.53275436155042_dp, 7.805208492281784_dp, &
             ! w
             0.00020670692329994715_dp, 0.0005608633909978252_dp, 0.00117822162433395_dp, &
             0.0023918307239149386_dp, 0.0047583847220433_dp, 0.009220914082914599_dp, &
             0.017396140411831254_dp, 0.032018442928032666_dp, 0.05763439977639338_dp, &
             0.10168990311563929_dp, 0.1762249592552838_dp, 0.3005391543097987_dp, &
             0.5055778600099088_dp, 0.8421284564034203_dp, 1.4005756769476354_dp, &
             2.3799387445050417_dp, 4.507482674005179_dp]

      CASE (527)

         aw(:) = & ! a
            [7.253109651073766e-05_dp, 0.00041378154246698155_dp, 0.0011894399545926847_dp, &
             0.002798526161215177_dp, 0.0060619459096750185_dp, 0.012512555917585542_dp, &
             0.024909165862831623_dp, 0.048108395498071804_dp, 0.09049532883191257_dp, &
             0.16629281149996789_dp, 0.29924575163888323_dp, 0.5284587376245806_dp, &
             0.9176892895564356_dp, 1.570671056991115_dp, 2.6592530593275407_dp, &
             4.489553818009151_dp, 7.750470715825014_dp, &
             ! w
             0.00018944048236774636_dp, 0.0005171521579185408_dp, 0.0010978019908693144_dp, &
             0.0022526568166520147_dp, 0.004521376664593447_dp, 0.008825576279679542_dp, &
             0.01675397683357181_dp, 0.031004652336597036_dp, 0.05608005851915602_dp, &
             0.09937797972756338_dp, 0.1728957766066036_dp, 0.29591376298043276_dp, &
             0.49940880628805856_dp, 0.8342799329605631_dp, 1.3911056054623188_dp, &
             2.3690809988090367_dp, 4.495127507755016_dp]

      CASE (528)

         aw(:) = & ! a
            [4.076405814605627e-05_dp, 0.0002382538909882903_dp, 0.0007172785036174207_dp, &
             0.0017887553118032778_dp, 0.004103604590222173_dp, 0.008917043928917007_dp, &
             0.01856727787107811_dp, 0.03729823641325223_dp, 0.0726364608741809_dp, &
             0.13765684708104833_dp, 0.2546598181053789_dp, 0.46108210773932384_dp, &
             0.8190039334641068_dp, 1.430877164081293_dp, 2.4682103026290094_dp, &
             4.238054139494266_dp, 7.4303818821050625_dp, &
             ! w
             0.0001070172373037324_dp, 0.0003062882624906517_dp, 0.0007012936016619437_dp, &
             0.0015459700494373915_dp, 0.003285238333576259_dp, 0.006718115008619782_dp, &
             0.013267499406418276_dp, 0.02541228267400262_dp, 0.04738390796270287_dp, &
             0.08627812215453308_dp, 0.15381430652300426_dp, 0.2691274981527009_dp, &
             0.46335419123057037_dp, 0.7880529629290074_dp, 1.3350005054020597_dp, &
             2.3045456706196203_dp, 4.4217074746843075_dp]

      CASE (529)

         aw(:) = & ! a
            [2.9160936668075905e-05_dp, 0.00017355365502172055_dp, 0.0005398265608048895_dp, &
             0.001397672386055173_dp, 0.003319019707423856_dp, 0.007429464246465446_dp, &
             0.01586691150562701_dp, 0.032577269870254215_dp, 0.0646612793531233_dp, &
             0.12461399088347623_dp, 0.23399280448585136_dp, 0.4293596632660751_dp, &
             0.771894119047728_dp, 1.3633337240128052_dp, 2.3749596565687208_dp, &
             4.114290731871903_dp, 7.271927422956754_dp, &
             ! w
             7.68502849403993e-05_dp, 0.00022768629688079903_dp, 0.0005477514153739001_dp, &
             0.0012592680559984917_dp, 0.0027636737709961353_dp, 0.005800766185918572_dp, &
             0.01171012014882713_dp, 0.022857962653192496_dp, 0.04333336440525826_dp, &
             0.08006903987145882_dp, 0.14462760292712087_dp, 0.25604969227661944_dp, &
             0.4455328439242189_dp, 0.764964512131641_dp, 1.3067570821118795_dp, &
             2.271915618884096_dp, 4.384593311974348_dp]

      CASE (530)

         aw(:) = & ! a
            [2.3021089475914032e-05_dp, 0.00013909831636292537_dp, 0.0004439894685442402_dp, &
             0.0011820562313644377_dp, 0.002876968546066814_dp, 0.0065746397526069145_dp, &
             0.014288246252458177_dp, 0.02977581333132268_dp, 0.05986665219462856_dp, &
             0.11668233589728014_dp, 0.22129675320222594_dp, 0.4096963661407671_dp, &
             0.7424605034446863_dp, 1.3208411240770856_dp, 2.315951441084732_dp, &
             4.035609898218205_dp, 7.170851898076781_dp, &
             ! w
             6.086526423575163e-05_dp, 0.00018549226025884883_dp, 0.0004630837126782273_dp, &
             0.0010964616570390648_dp, 0.0024605616099429603_dp, 0.005257849618763615_dp, &
             0.010774390164609474_dp, 0.021303199826252648_dp, 0.04083970904018302_dp, &
             0.07620781031516743_dp, 0.13886307130363254_dp, 0.24777738090808893_dp, &
             0.43417998023950005_dp, 0.7501681263270686_dp, 1.2885748376435835_dp, &
             2.250855744705992_dp, 4.360641433740663_dp]

      CASE (531)

         aw(:) = & ! a
            [1.918021108184618e-05_dp, 0.00011743496619441033_dp, 0.00038304286724531316_dp, &
             0.001042728254363479_dp, 0.002586697547420851_dp, 0.006005284015735613_dp, &
             0.0132238415003511_dp, 0.027866969228087193_dp, 0.05656974441752018_dp, &
             0.11118456754211066_dp, 0.21243440078169615_dp, 0.39588493368778044_dp, &
             0.7216727864412544_dp, 1.290687002879518_dp, 2.273908671258199_dp, &
             3.97937103189813_dp, 7.098438341837911_dp, &
             ! w
             5.085499126425964e-05_dp, 0.00015879189104924546_dp, 0.00040835568622275446_dp, &
             0.0009889473088433097_dp, 0.0022570984755315_dp, 0.004888731455877956_dp, &
             0.01013143228792217_dp, 0.020225156256024016_dp, 0.039096903340007934_dp, &
             0.07349025105641002_dp, 0.13478058532324966_dp, 0.24188623737314655_dp, &
             0.42605539338067017_dp, 0.7395353696295928_dp, 1.2754678823800256_dp, &
             2.2356474193399447_dp, 4.343345438395721_dp]

      CASE (532)

         aw(:) = & ! a
            [1.6533145622881964e-05_dp, 0.0001024413371155251_dp, 0.00034044882534500613_dp, &
             0.0009440705369768968_dp, 0.002378522981751988_dp, 0.005592411396363883_dp, &
             0.01244467176685519_dp, 0.026458341750016687_dp, 0.05411979840074476_dp, &
             0.1070742653542394_dp, 0.20577318067203482_dp, 0.3854548633406083_dp, &
             0.7059092574368645_dp, 1.267738457061893_dp, 2.2418152881945583_dp, &
             3.936337556123286_dp, 7.042931583759598_dp, &
             ! w
             4.395015726227889e-05_dp, 0.00014021020705245996_dp, 0.0003695888378725299_dp, &
             0.0009115007388702211_dp, 0.0021086912864314274_dp, 0.004616848849878659_dp, &
             0.009653998371810918_dp, 0.019419091110618163_dp, 0.03778592048110537_dp, &
             0.07143515934408562_dp, 0.13167871077424173_dp, 0.23739132583529082_dp, &
             0.4198334588281623_dp, 0.7313672054802236_dp, 1.2653750948198026_dp, &
             2.2239207327011843_dp, 4.330009366736664_dp]

      CASE (533)

         aw(:) = & ! a
            [1.458933897784122e-05_dp, 9.139030473073127e-05_dp, 0.00030878602244267865_dp, &
             0.0008699113552462914_dp, 0.0022203856912609544_dp, 0.005275934475200526_dp, &
             0.011842849219186052_dp, 0.02536325103370956_dp, 0.05220450013744886_dp, &
             0.10384529109056181_dp, 0.20051790022964933_dp, 0.37719525869318016_dp, &
             0.693384842204357_dp, 1.2494531716280413_dp, 2.2161817777173987_dp, &
             3.9019002052844005_dp, 6.998451151508137_dp, &
             ! w
             3.887610761530127e-05_dp, 0.00012644811783661284_dp, 0.00034043738029018326_dp, &
             0.0008524591235363151_dp, 0.0019944061590717213_dp, 0.0044058236614755355_dp, &
             0.009281013741210436_dp, 0.01878586883941478_dp, 0.03675107494621722_dp, &
             0.06980604654338901_dp, 0.1292105276995719_dp, 0.2338027267855387_dp, &
             0.4148514548104739_dp, 0.7248105725372748_dp, 1.2572582238153398_dp, &
             2.214479692495519_dp, 4.319272791830595_dp]

      CASE (534)

         aw(:) = & ! a
            [1.3096448252656754e-05_dp, 8.287482053602437e-05_dp, 0.0002842008919397817_dp, &
             0.0008117695702538987_dp, 0.0020952888205231376_dp, 0.00502367351697935_dp, &
             0.01136007599844034_dp, 0.024480027935377596_dp, 0.05065257579819165_dp, &
             0.1012183712955543_dp, 0.19622739726279395_dp, 0.3704310533161801_dp, &
             0.6831000445594243_dp, 1.234402220246752_dp, 2.195040469962191_dp, &
             3.873453146423682_dp, 6.961666232441807_dp, &
             ! w
             3.497670478725251e-05_dp, 0.00011579729066640441_dp, 0.00031757350177821347_dp, &
             0.0008056140241222235_dp, 0.001902964119042771_dp, 0.004235867285012855_dp, &
             0.008978988302363237_dp, 0.01827074951526243_dp, 0.035905874038777034_dp, &
             0.06847081379014039_dp, 0.1271813008090145_dp, 0.2308442081133539_dp, &
             0.4107342319356332_dp, 0.7193809493034429_dp, 1.2505260530980302_dp, &
             2.206642283152803_dp, 4.310359994604335_dp]

      CASE (535)

         aw(:) = & ! a
            [1.1910883015046781e-05_dp, 7.609212753561193e-05_dp, 0.00026448227563254345_dp, &
             0.000764736279561239_dp, 0.0019933032390468514_dp, 0.004816670511372321_dp, &
             0.010961751486023432_dp, 0.023747937579544923_dp, 0.049361123989206605_dp, &
             0.099024863792238_dp, 0.1926340592808602_dp, 0.3647510825372564_dp, &
             0.6744439026945592_dp, 1.2217093878950636_dp, 2.177181568447064_dp, &
             3.8493907944636194_dp, 6.930521344761309_dp, &
             ! w
             3.187836354302178e-05_dp, 0.0001072799478332931_dp, 0.00029907102497048253_dp, &
             0.0007673261744342137_dp, 0.0018276875216180584_dp, 0.004095171432554331_dp, &
             0.00872780547946678_dp, 0.017840663826624248_dp, 0.03519779958637537_dp, &
             0.06734888265256474_dp, 0.12547175393011545_dp, 0.2283459516778673_dp, &
             0.4072504095294596_dp, 0.7147786717212383_dp, 1.24481216406687_dp, &
             2.199985312173977_dp, 4.302789654588425_dp]

      CASE (536)

         aw(:) = & ! a
            [1.094467766532379e-05_dp, 7.05492793489514e-05_dp, 0.00024826530528285906_dp, &
             0.0007257578826090639_dp, 0.0019082025868475256_dp, 0.0046429506569553_dp, &
             0.010625881647910675_dp, 0.023128160320095827_dp, 0.04826405446679087_dp, &
             0.0971559894122433_dp, 0.18956460617133714_dp, 0.35988820424996204_dp, &
             0.6670182563676718_dp, 1.210802132037608_dp, 2.1618127797467053_dp, &
             3.8286597689957436_dp, 6.903666187020532_dp, &
             ! w
             2.9352097161866914e-05_dp, 0.00010029392135425999_dp, 0.00028373169352337215_dp, &
             0.0007353067079803949_dp, 0.001764340847949028_dp, 0.003976196228625831_dp, &
             0.008514548590324024_dp, 0.017474274911499204_dp, 0.034592820280502694_dp, &
             0.06638783728179978_dp, 0.12400402849715236_dp, 0.2261967623795507_dp, &
             0.4042480651027011_dp, 0.7108065227642008_dp, 1.239874982587707_dp, &
             2.194229503181971_dp, 4.29624415000418_dp]

      CASE (537)

         aw(:) = & ! a
            [6.324325770733568e-06_dp, 4.3784561254680955e-05_dp, 0.00016816473585988095_dp, &
             0.0005282005147507511_dp, 0.0014671088196661216_dp, 0.0037257757029787344_dp, &
             0.008825379661917565_dp, 0.01976288365694673_dp, 0.04224170211896209_dp, &
             0.08679945016337094_dp, 0.1724141912175066_dp, 0.33252003315428064_dp, &
             0.6249614244473011_dp, 1.1486866961091953_dp, 2.0738849996148088_dp, &
             3.7096192517530233_dp, 6.749054263778715_dp, &
             ! w
             1.7253691427870995e-05_dp, 6.61075093598061e-05_dp, 0.00020587498387287116_dp, &
             0.0005681844806581525_dp, 0.001427042343733492_dp, 0.0033327392817497216_dp, &
             0.0073462825544027685_dp, 0.015445168213430206_dp, 0.031210765117142113_dp, &
             0.060970923123981585_dp, 0.11567098057508378_dp, 0.21391601027046922_dp, &
             0.38699520861360387_dp, 0.6878714833474024_dp, 1.2112635921527724_dp, &
             2.1608041389741195_dp, 4.2582328395774445_dp]

      CASE (538)

         aw(:) = & ! a
            [4.628743966904852e-06_dp, 3.3796029923241924e-05_dp, 0.00013711304917296313_dp, &
             0.0004485391661440641_dp, 0.001283395088179801_dp, 0.0033337743947437393_dp, &
             0.008039532195749188_dp, 0.018268312876367567_dp, 0.03952752739923002_dp, &
             0.08207284400105036_dp, 0.16450119008048864_dp, 0.3197721747334692_dp, &
             0.6052088269338589_dp, 1.1193041273386288_dp, 2.0320425876350585_dp, &
             3.6527023575970827_dp, 6.674878528218103_dp, &
             ! w
             1.2804881247680028e-05_dp, 5.304913176856229e-05_dp, 0.00017440289685598898_dp, &
             0.0004979194173677241_dp, 0.001281263818003164_dp, 0.003048647428696953_dp, &
             0.006821446095669873_dp, 0.014520226380305402_dp, 0.029649758665777272_dp, &
             0.0584434988197207_dp, 0.11174582745534717_dp, 0.20808270110233404_dp, &
             0.37873991792831896_dp, 0.6768292319684563_dp, 1.197423087879505_dp, &
             2.1445908529215_dp, 4.2397947800554565_dp]

      CASE (539)

         aw(:) = & ! a
            [3.7296960525982587e-06_dp, 2.8437322158634238e-05_dp, 0.00012002756545776218_dp, &
             0.00040365517251802504_dp, 0.0011779424874718643_dp, 0.0031054719797459227_dp, &
             0.007576491381279091_dp, 0.017379198004902275_dp, 0.03789983438201302_dp, &
             0.07921877377529439_dp, 0.1596946828431425_dp, 0.31198888957058113_dp, &
             0.593094569512663_dp, 1.1012140816737752_dp, 2.006197922873577_dp, &
             3.617456818608499_dp, 6.6288616663883015_dp, &
             ! w
             1.044385300059981e-05_dp, 4.5928780491993075e-05_dp, 0.00015664140917764276_dp, &
             0.0004573836432214581_dp, 0.0011958645886293892_dp, 0.002880241996866597_dp, &
             0.006507334964394241_dp, 0.013962210461223206_dp, 0.028701568618929543_dp, &
             0.05689920818229356_dp, 0.1093351014641138_dp, 0.20448374251597093_dp, &
             0.3736264456356434_dp, 0.6699665335507886_dp, 1.1887992173029054_dp, &
             2.134473571055921_dp, 4.228288998074195_dp]

      CASE (540)

         aw(:) = & ! a
            [3.1670669624630847e-06_dp, 2.5052531441846564e-05_dp, 0.00010902841455310801_dp, &
             0.0003742726254346608_dp, 0.0011080238570622418_dp, 0.0029526032160513437_dp, &
             0.007264006197212855_dp, 0.01677531479403189_dp, 0.03678836441783677_dp, &
             0.0772609556461209_dp, 0.15638455439270008_dp, 0.3066103993953317_dp, &
             0.584698378696311_dp, 1.088644124244412_dp, 1.9882012310283597_dp, &
             3.5928725132260033_dp, 6.596725571953208_dp, &
             ! w
             8.965616478824352e-06_dp, 4.137395639487357e-05_dp, 0.00014500041419237566_dp, &
             0.0004304187360175022_dp, 0.0011384650973003435_dp, 0.0027661473380765607_dp, &
             0.006293153696944768_dp, 0.013579682026016221_dp, 0.028048615716789703_dp, &
             0.055831592359072896_dp, 0.10766278586866126_dp, 0.20197964422880343_dp, &
             0.3700592361637516_dp, 0.6651684395231704_dp, 1.1827595511959288_dp, &
             2.127381043276869_dp, 4.220222956712289_dp]

      CASE (541)

         aw(:) = & ! a
            [2.7796064722685634e-06_dp, 2.2703360781290608e-05_dp, 0.0001012779520563711_dp, &
             0.0003533032372163281_dp, 0.0010576479939282137_dp, 0.0028416576871572756_dp, &
             0.007035906475137151_dp, 0.01633242986390792_dp, 0.03597001222543971_dp, &
             0.0758146422949969_dp, 0.15393221922226166_dp, 0.30261578873069467_dp, &
             0.5784490537365227_dp, 1.0792708724990587_dp, 1.974760498324139_dp, &
             3.5744893746123694_dp, 6.572674541208949_dp, &
             ! w
             7.947346387490553e-06_dp, 3.8179758415793534e-05_dp, 0.00013668510884931506_dp, &
             0.0004109448392435954_dp, 0.0010966936455979088_dp, 0.0026826290778421682_dp, &
             0.0061356305282840974_dp, 0.013297243594125497_dp, 0.027564910849513305_dp, &
             0.055038445986010306_dp, 0.10641730381195505_dp, 0.2001106035329336_dp, &
             0.36739162135938086_dp, 0.6615745747054887_dp, 1.178230140095077_dp, &
             2.122058237685207_dp, 4.214169462968263_dp]

      CASE (542)

         aw(:) = & ! a
            [2.4955585457101115e-06_dp, 2.0969553191437247e-05_dp, 9.548563033736107e-05_dp, &
             0.0003374720932786499_dp, 0.0010193308013547494_dp, 0.002756789175571825_dp, &
             0.006860636528982101_dp, 0.015990877886852654_dp, 0.03533698150635536_dp, &
             0.07469297249050118_dp, 0.152026131592035_dp, 0.2995050097973982_dp, &
             0.5735743382538083_dp, 1.0719489212438271_dp, 1.9642486699868313_dp, &
             3.5600986070175926_dp, 6.553834178305341_dp, &
             ! w
             7.200732525066296e-06_dp, 3.580151936954856e-05_dp, 0.0001304028573155433_dp, &
             0.00039610583919052914_dp, 0.001064674088028023_dp, 0.0026183166339566895_dp, &
             0.006013887014916552_dp, 0.013078296756508874_dp, 0.02718898088292041_dp, &
             0.05442066236997839_dp, 0.10544533221889306_dp, 0.19864955351463343_dp, &
             0.3653032556774356_dp, 0.6587575962040341_dp, 1.1746764750032894_dp, &
             2.1178797923924426_dp, 4.209417368633295_dp]

      CASE (543)

         aw(:) = & ! a
            [2.2779459559839988e-06_dp, 1.963334880545524e-05_dp, 9.097408793239506e-05_dp, &
             0.00032503839568328165_dp, 0.0009890534104325496_dp, 0.00268941937378009_dp, &
             0.006721001303009458_dp, 0.015717968490801662_dp, 0.03482993687665034_dp, &
             0.07379268015183467_dp, 0.1504935234373866_dp, 0.29699991920016044_dp, &
             0.5696435388725413_dp, 1.066038009012073_dp, 1.9557545065077264_dp, &
             3.5484612807085503_dp, 6.538590448737748_dp, &
             ! w
             6.6286815373737414e-06_dp, 3.395477113685229e-05_dp, 0.00012546593608756427_dp, &
             0.00038436368481835803_dp, 0.0010392147491139615_dp, 0.0025669924358072654_dp, &
             0.005916443894659351_dp, 0.01290262550270768_dp, 0.026886734559403803_dp, &
             0.05392308933861789_dp, 0.10466128683302232_dp, 0.19746940405927732_dp, &
             0.36361441882205797_dp, 0.6564772809322426_dp, 1.1717976296141035_dp, &
             2.1144933129360592_dp, 4.205565929645237_dp]

      CASE (544)

         aw(:) = & ! a
            [2.105701716452156e-06_dp, 1.857006660990128e-05_dp, 8.735112149279849e-05_dp, &
             0.0003149834698083493_dp, 0.000964444494537918_dp, 0.002634453707931044_dp, &
             0.006606735170160016_dp, 0.015494100584332543_dp, 0.0344131706470937_dp, &
             0.0730514234285257_dp, 0.14922980636992003_dp, 0.2949317302040145_dp, &
             0.5663947413260328_dp, 1.061148060191027_dp, 1.9487219816729455_dp, &
             3.538820496257873_dp, 6.525956431119965_dp, &
             ! w
             6.17585407786133e-06_dp, 3.2475477772705476e-05_dp, 0.00012147159086186359_dp, &
             0.00037480876201814685_dp, 0.00101841511710842_dp, 0.002524934421154738_dp, &
             0.0058363992705457585_dp, 0.012758030884728199_dp, 0.026637534914135057_dp, &
             0.053512247550896994_dp, 0.1040130879997386_dp, 0.19649265059700774_dp, &
             0.3622152995158532_dp, 0.6545866113722596_dp, 1.169409211858116_dp, &
             2.1116827198148154_dp, 4.202369415333248_dp]

      CASE (545)

         aw(:) = & ! a
            [1.965911550315286e-06_dp, 1.770295804549473e-05_dp, 8.437297409462314e-05_dp, &
             0.0003066685457418877_dp, 0.0009440066551254706_dp, 0.002588657239420727_dp, &
             0.006511290198022752_dp, 0.01530672479228543_dp, 0.0340637495531545_dp, &
             0.07242905648873617_dp, 0.14816747455480803_dp, 0.29319128274260775_dp, &
             0.5636582621311574_dp, 1.057025981563586_dp, 1.9427898698613837_dp, &
             3.530684028828231_dp, 6.515289845676721_dp, &
             ! w
             5.808320991888156e-06_dp, 3.126204272475528e-05_dp, 0.0001181671116181515_dp, &
             0.0003668656543048002_dp, 0.001001065818698379_dp, 0.0024897630531041798_dp, &
             0.005769323981310001_dp, 0.01263665982819144_dp, 0.0264280618037184_dp, &
             0.05316647860380902_dp, 0.10346697694454568_dp, 0.1956689648535538_dp, &
             0.361034481345287_dp, 0.6529898473246921_dp, 1.167391014750996_dp, &
             2.1093070606847255_dp, 4.199667534273998_dp]

      CASE (546)

         aw(:) = & ! a
            [1.322039824529103e-06_dp, 1.3644660379054517e-05_dp, 7.009393117838189e-05_dp, &
             0.00026610080001424833_dp, 0.0008430526078221751_dp, 0.002360344948253793_dp, &
             0.006032017423532955_dp, 0.014360324623452917_dp, 0.03229034093158516_dp, &
             0.06925745438801997_dp, 0.14273483967194414_dp, 0.2842639156919321_dp, &
             0.5495851108359164_dp, 1.0357792776721013_dp, 1.9121561433509082_dp, &
             3.4886047381494176_dp, 6.4600675716736635_dp, &
             ! w
             4.114809371932191e-06_dp, 2.5478322550811068e-05_dp, 0.00010202607816647103_dp, &
             0.00032752443441371446_dp, 0.0009143026109553528_dp, 0.002312574577843141_dp, &
             0.0054294213552350526_dp, 0.012018640081376365_dp, 0.025357083262282858_dp, &
             0.05139248419554339_dp, 0.10065661073380844_dp, 0.19141889836816153_dp, &
             0.35492758163493465_dp, 0.6447156623336303_dp, 1.1569173598543177_dp, &
             2.0969675718938148_dp, 4.185633290543016_dp]

      CASE (547)

         aw(:) = & ! a
            [1.1414784913290507e-06_dp, 1.2479906043254122e-05_dp, 6.586792350915943e-05_dp, &
             0.0002538405276329722_dp, 0.0008120970203781402_dp, 0.0022895846593795665_dp, &
             0.005882241099735417_dp, 0.014062591248535634_dp, 0.03172936041406355_dp, &
             0.06824953475591365_dp, 0.14100154430186176_dp, 0.28140590790837133_dp, &
             0.5450664617927452_dp, 1.0289401016118418_dp, 1.9022745807980563_dp, &
             3.475008679695429_dp, 6.442203962408087_dp, &
             ! w
             3.6393993327764124e-06_dp, 2.377803001827817e-05_dp, 9.714146248334506e-05_dp, &
             0.00031542480855583033_dp, 0.0008873180020119566_dp, 0.0022569990672555566_dp, &
             0.005322094224988005_dp, 0.01182242228760336_dp, 0.02501548554271099_dp, &
             0.050824423584506936_dp, 0.09975361540313847_dp, 0.19004924595880676_dp, &
             0.3529544360619371_dp, 0.6420364207351131_dp, 1.1535202206350186_dp, &
             2.092961336457584_dp, 4.181076696726021_dp]

      CASE (548)

         aw(:) = & ! a
            [1.1242594032445634e-06_dp, 1.2367981585927119e-05_dp, 6.545818610240409e-05_dp, &
             0.00025264486136782104_dp, 0.0008090660491099191_dp, 0.002282635952329536_dp, &
             0.005867499650624025_dp, 0.01403323416656616_dp, 0.03167396391496785_dp, &
             0.06814987799660217_dp, 0.140829982868287_dp, 0.2811227618501116_dp, &
             0.5446184400971232_dp, 1.028261542994931_dp, 1.901293617264876_dp, &
             3.4736583783930577_dp, 6.440429281216567_dp, &
             ! w
             3.594037148536206e-06_dp, 2.3613463975539957e-05_dp, 9.666492767447006e-05_dp, &
             0.00031423914840777593_dp, 0.0008846656273099036_dp, 0.0022515238099725353_dp, &
             0.005311501090248518_dp, 0.01180302674525577_dp, 0.024981677330889934_dp, &
             0.050768141814753705_dp, 0.09966406660924938_dp, 0.18991331154444588_dp, &
             0.35275847212429345_dp, 0.6417701759163508_dp, 1.1531824876019705_dp, &
             2.0925629496356657_dp, 4.180623589040228_dp]

      CASE (549)

         aw(:) = & ! a
            [0.006504707620870862_dp, 0.03457638451151281_dp, 0.0863605588515674_dp, &
             0.16424368204284412_dp, 0.2720827762576999_dp, 0.4157314705178234_dp, &
             0.6037727276832092_dp, 0.8484615472103956_dp, 1.166857443526373_dp, &
             1.582174844949876_dp, 2.1255130141272733_dp, 2.8383150330853875_dp, &
             3.776184489691636_dp, 5.015317845923547_dp, 6.664555312407246_dp, &
             8.891550981317133_dp, 11.99335122869203_dp, 16.676470342449512_dp, &
             ! w
             0.016727416580251387_dp, 0.039632078965078775_dp, 0.06432780440458163_dp, &
             0.09207266697943958_dp, 0.12456890894406929_dp, 0.1641431837719012_dp, &
             0.21396298272289155_dp, 0.2782357700589099_dp, 0.36239707430805856_dp, &
             0.47338134062580534_dp, 0.6201412070315351_dp, 0.8146289484700561_dp, &
             1.073624694183775_dp, 1.4223973563857537_dp, 1.9030598837282355_dp, &
             2.5972903740046815_dp, 3.7067955624904116_dp, 6.030987538475906_dp]

      CASE (550)

         aw(:) = & ! a
            [0.005454939230439086_dp, 0.029022386751768166_dp, 0.07261074660919156_dp, &
             0.13845019379721252_dp, 0.23018463741887507_dp, 0.3534092543271198_dp, &
             0.5164149110038168_dp, 0.731130518398276_dp, 1.0142408538670358_dp, &
             1.3885292100345954_dp, 1.8846365428304332_dp, 2.54359362391212_dp, &
             3.4207312377096155_dp, 4.592187425556606_dp, 6.166964644487724_dp, &
             8.312948853281073_dp, 11.326920786084592_dp, 15.912264062176604_dp, &
             ! w
             0.01403076052185356_dp, 0.03330310464763135_dp, 0.05423910675979884_dp, &
             0.07803827495869796_dp, 0.10635069588915873_dp, 0.14146455058864316_dp, &
             0.1865163683259881_dp, 0.24567372242915428_dp, 0.32431567595192745_dp, &
             0.4293269788840337_dp, 0.5696635237812344_dp, 0.7573699308590284_dp, &
             1.0094100056379167_dp, 1.351286818847722_dp, 1.825296248222896_dp, &
             2.51301336523117_dp, 3.615041310021613_dp, 5.924599133293403_dp]

      CASE (551)

         aw(:) = & ! a
            [0.004719198693065434_dp, 0.025127630757328167_dp, 0.06295892659890262_dp, &
             0.12031784280556994_dp, 0.20067504660898328_dp, 0.309411065083213_dp, &
             0.4545645972722522_dp, 0.6477672584542331_dp, 0.9053458892365164_dp, &
             1.249664455458483_dp, 1.710918733883249_dp, 2.329738102077569_dp, &
             3.1611777285324156_dp, 4.281292057468559_dp, 5.799205871677062_dp, &
             7.883025436388123_dp, 10.829418604027625_dp, 15.339510264532255_dp, &
             ! w
             0.012140534751940763_dp, 0.028861839393117896_dp, 0.04714535342349926_dp, &
             0.0681419618356386_dp, 0.09345666238084763_dp, 0.125339489076594_dp, &
             0.1668879549209818_dp, 0.2222166633555844_dp, 0.2966388361651315_dp, &
             0.39698882246477685_dp, 0.532227240148291_dp, 0.7144836697800593_dp, &
             0.9608859859076535_dp, 1.2971489546989525_dp, 1.765749845406465_dp, &
             2.4482320396481305_dp, 3.544401097573219_dp, 5.842786282198943_dp]

      CASE (552)

         aw(:) = & ! a
            [0.004172227365416325_dp, 0.022230729039644317_dp, 0.05577357985498636_dp, &
             0.10680180401545937_dp, 0.1786412823496513_dp, 0.2764899558231186_dp, &
             0.40816545243668184_dp, 0.5850289259794452_dp, 0.8230733480317899_dp, &
             1.1442695863965273_dp, 1.578397776516414_dp, 2.1657119062532786_dp, &
             2.961003018513874_dp, 4.040231570611905_dp, 5.512616711937986_dp, &
             7.546467437808507_dp, 10.438420006766224_dp, 14.887879452056895_dp, &
             ! w
             0.010735123281046864_dp, 0.02555645084196436_dp, 0.04185658199776483_dp, &
             0.06074514237930913_dp, 0.08378759015379075_dp, 0.11319722246504531_dp, &
             0.15202936990628974_dp, 0.2043411997280821_dp, 0.2753797326933137_dp, &
             0.3719329029533199_dp, 0.5029671890258762_dp, 0.6806881808303616_dp, &
             0.9223683599982286_dp, 1.25391179651718_dp, 1.7179680619759234_dp, &
             2.3960872496711834_dp, 3.48746813284328_dp, 5.7769132915013195_dp]

      CASE (553)

         aw(:) = & ! a
            [0.0037481479791149834_dp, 0.01998370529062772_dp, 0.05019572951738094_dp, &
             0.09629745946207843_dp, 0.16149116096897648_dp, 0.2508165176467207_dp, &
             0.37189480031827843_dp, 0.535840488647031_dp, 0.7583376619109063_dp, &
             1.0609933975927763_dp, 1.4732045849654014_dp, 2.0348781672529332_dp, &
             2.80055616694921_dp, 3.8461009998400413_dp, 5.280805511730411_dp, &
             7.273163367832301_dp, 10.119825359903327_dp, 14.518835947441595_dp, &
             ! w
             0.009645363471445978_dp, 0.022991191495488852_dp, 0.037745580186778954_dp, &
             0.05498251401259031_dp, 0.07623233046456797_dp, 0.10367334795226314_dp, &
             0.14031795924770327_dp, 0.19016545993566092_dp, 0.25839972669382333_dp, &
             0.35176706596839663_dp, 0.47924011713604087_dp, 0.6530917220917539_dp, &
             0.8907218049245282_dp, 1.2182039602924128_dp, 1.678349191099112_dp, &
             2.352736388659292_dp, 3.440085680309329_dp, 5.722136250605784_dp]

      CASE (554)

         aw(:) = & ! a
            [0.0034088324466304944_dp, 0.018185090875107682_dp, 0.045727748399452345_dp, &
             0.0878744053671484_dp, 0.1477199920402149_dp, 0.23016494655539427_dp, &
             0.34265422432112574_dp, 0.49607641643219097_dp, 0.7058303459086763_dp, &
             0.9931876484911322_dp, 1.3871927775883186_dp, 1.927432411321234_dp, &
             2.6682148354966255_dp, 3.685303441973784_dp, 5.088049018579727_dp, &
             7.045113179204027_dp, 9.853188552772256_dp, 14.209212193237015_dp, &
             ! w
             0.008773340443438755_dp, 0.02093683847149185_dp, 0.03444861897423562_dp, &
             0.05035142585071726_dp, 0.0701440561722791_dp, 0.09597166675239711_dp, &
             0.13080419499020232_dp, 0.17858454783419062_dp, 0.24443766561939323_dp, &
             0.33507267593781265_dp, 0.4594680837150753_dp, 0.6299560570532854_dp, &
             0.8640493534648978_dp, 1.1879741820719463_dp, 1.6446922479162214_dp, &
             2.31582446184048_dp, 3.3997031182761037_dp, 5.6754851807026565_dp]

      CASE (555)

         aw(:) = & ! a
            [0.0031306056830560135_dp, 0.016709748015357007_dp, 0.042060375264030045_dp, &
             0.08095396529993411_dp, 0.13639096785774668_dp, 0.21314779812291593_dp, &
             0.3185096464168493_dp, 0.46315734143844034_dp, 0.6622260349962985_dp, &
             0.9366779619610529_dp, 1.3152330549324225_dp, 1.8371818862508218_dp, &
             2.5566133546364127_dp, 3.5491924578909004_dp, 4.924315664937931_dp, &
             6.850797628249299_dp, 9.625389361929418_dp, 13.944106543599815_dp, &
             ! w
             0.008058250952104717_dp, 0.01925096094766146_dp, 0.0317394427938591_dp, &
             0.046538725017352435_dp, 0.06511901428038042_dp, 0.08959406553832734_dp, &
             0.12289248368821457_dp, 0.16890315352568652_dp, 0.23269648792897377_dp, &
             0.3209482711730135_dp, 0.44264224272054753_dp, 0.610162954106963_dp, &
             0.8411236798163042_dp, 1.1618888828134089_dp, 1.6155610054874723_dp, &
             2.283811039826218_dp, 3.3646504591932693_dp, 5.635017031535249_dp]

      CASE (556)

         aw(:) = & ! a
            [0.0017784976328254598_dp, 0.00953050331705429_dp, 0.0241715763670107_dp, &
             0.047078713382615565_dp, 0.08067449545094924_dp, 0.12894431670112244_dp, &
             0.1980991783943776_dp, 0.2973741996928436_dp, 0.4400631372091785_dp, &
             0.6449848392880045_dp, 0.9386073149270318_dp, 1.3581198602013644_dp, &
             1.9559668797684324_dp, 2.806943867096501_dp, 4.020578184661381_dp, &
             5.766696293860253_dp, 8.342817860954147_dp, 12.440295670973242_dp, &
             ! w
             0.004582036896370416_dp, 0.01103391709664276_dp, 0.018471916277182893_dp, &
             0.027737429369125297_dp, 0.04010723661606406_dp, 0.0574545178903398_dp, &
             0.08237765899599424_dp, 0.11836408860959646_dp, 0.17011549140639595_dp, &
             0.24408915011169383_dp, 0.34928075268957354_dp, 0.4983722612841169_dp, &
             0.7095962619273725_dp, 1.010234642279125_dp, 1.4444207813624754_dp, &
             2.094410418461798_dp, 3.156654855251119_dp, 5.395389620425401_dp]

      CASE (557)

         aw(:) = & ! a
            [0.0012731020932926379_dp, 0.006840889911889719_dp, 0.017441871151260967_dp, &
             0.03425676267104925_dp, 0.059409560309849466_dp, 0.09645246374479774_dp, &
             0.15097458736251926_dp, 0.23135661755667683_dp, 0.34981080772217094_dp, &
             0.5239021241748318_dp, 0.7787609548262905_dp, 1.1502794202181457_dp, &
             1.6898140333634362_dp, 2.4714863686327306_dp, 3.604749420902687_dp, &
             5.259977275676536_dp, 7.735354575200575_dp, 11.72042732080812_dp, &
             ! w
             0.0032819955090503555_dp, 0.007946879349381084_dp, 0.013446306697748859_dp, &
             0.020529284744824725_dp, 0.030358207775712336_dp, 0.04464691816132566_dp, &
             0.06577726530888198_dp, 0.09699728556123041_dp, 0.14280235347746628_dp, &
             0.20951654252840063_dp, 0.3061070498052226_dp, 0.4453729615003168_dp, &
             0.6458550254409235_dp, 0.9353531013878362_dp, 1.358657352007492_dp, &
             1.9985335334287477_dp, 3.0509031139521676_dp, 5.2738872947947355_dp]

      CASE (558)

         aw(:) = & ! a
            [0.001003007640618_dp, 0.005401192853243206_dp, 0.013829006030721161_dp, &
             0.027343138381717924_dp, 0.04787398783803675_dp, 0.07868305861976885_dp, &
             0.12493283084954603_dp, 0.19441735623604497_dp, 0.2986096720559634_dp, &
             0.4542136973287632_dp, 0.6854260121554199_dp, 1.0272086576277664_dp, &
             1.5301066611493623_dp, 2.2676993471981066_dp, 3.34931803735087_dp, &
             4.945688630210126_dp, 7.355525423831127_dp, 11.267417319265014_dp, &
             ! w
             0.0025869658738650997_dp, 0.006291196597285421_dp, 0.010735193719682509_dp, &
             0.016607063768514875_dp, 0.02498881534373861_dp, 0.037477989189976496_dp, &
             0.0563035562499745_dp, 0.08455225510195966_dp, 0.12657933799322968_dp, &
             0.18860809792637548_dp, 0.279565535852876_dp, 0.4123050051435338_dp, &
             0.6055566517564234_dp, 0.8874705225249772_dp, 1.3033133509688122_dp, &
             1.9362701532184283_dp, 2.9820300811488276_dp, 5.194874374428528_dp]

      CASE (559)

         aw(:) = & ! a
            [0.0008331309433943816_dp, 0.004494533393081155_dp, 0.011548438889181107_dp, &
             0.02296365335753164_dp, 0.04053062382391568_dp, 0.06729580026343036_dp, &
             0.10810275820315267_dp, 0.17030803184533577_dp, 0.26483404269108474_dp, &
             0.4077390338229544_dp, 0.622509561219692_dp, 0.9433866330993251_dp, &
             1.4202692025580552_dp, 2.1262840620169734_dp, 3.1706302181916914_dp, &
             4.7242816837238015_dp, 7.086385390567176_dp, 10.944946439250204_dp, &
             ! w
             0.002149695374373156_dp, 0.005246886200683311_dp, 0.009017223092835954_dp, &
             0.014104263829788848_dp, 0.021528561526898454_dp, 0.03279756255363238_dp, &
             0.05002531062141126_dp, 0.07617992934245178_dp, 0.11551095755702746_dp, &
             0.17415921094997922_dp, 0.26100924839686485_dp, 0.38894044098202907_dp, &
             0.5768126546101896_dp, 0.8530354319361633_dp, 1.26324693268147_dp, &
             1.890984014251146_dp, 2.9318282815396244_dp, 5.137338630737624_dp]

      CASE (560)

         aw(:) = & ! a
            [0.0007156743708438212_dp, 0.003866973046724055_dp, 0.009966773656909301_dp, &
             0.019917211061448056_dp, 0.03540081508361522_dp, 0.05929543332843836_dp, &
             0.09619333707801735_dp, 0.15310711854368658_dp, 0.24052609212096823_dp, &
             0.3739971151430409_dp, 0.5764378796819639_dp, 0.8815030388162307_dp, &
             1.3385557188415094_dp, 2.0203347643968717_dp, 3.03590765912303_dp, &
             4.556433918219239_dp, 6.8814234769517055_dp, 10.698495570575057_dp, &
             ! w
             0.0018472812466389012_dp, 0.004523095116174931_dp, 0.007821851870583206_dp, &
             0.012352432321382304_dp, 0.019085978253845236_dp, 0.029457307511689484_dp, &
             0.04549005278246955_dp, 0.07005984566079591_dp, 0.10733141410643547_dp, &
             0.16337536445860176_dp, 0.24703471719863157_dp, 0.37120018459651805_dp, &
             0.554826538805612_dp, 0.8265266831834693_dp, 1.232241560519569_dp, &
             1.8558099395439571_dp, 2.8927683242146855_dp, 5.092604852951585_dp]

      CASE (561)

         aw(:) = & ! a
            [0.0006292469835733137_dp, 0.003404765182263968_dp, 0.008799834740885864_dp, &
             0.017663661343047414_dp, 0.031591910790609926_dp, 0.053325027865730824_dp, &
             0.0872499733034381_dp, 0.1400989014371246_dp, 0.22200741905845026_dp, &
             0.3481020164414586_dp, 0.5408282237151643_dp, 0.8333481626537103_dp, &
             1.2745678967141136_dp, 1.9368869444029295_dp, 2.929245971365592_dp, &
             4.42294883364497_dp, 6.71781677985601_dp, 10.50120072187859_dp, &
             ! w
             0.0016247092432591461_dp, 0.003989399431325142_dp, 0.006937405507990823_dp, &
             0.011049473564568906_dp, 0.017255655639838212_dp, 0.02693051891930886_dp, &
             0.04202413865776385_dp, 0.06533700289003663_dp, 0.10096305774931143_dp, &
             0.15491167253317023_dp, 0.23598628401178987_dp, 0.3570806030334209_dp, &
             0.537221991259609_dp, 0.8051888646508266_dp, 1.2071767417464814_dp, &
             1.8272882686574476_dp, 2.861049495849007_dp, 5.056298944057487_dp]

      CASE (562)

         aw(:) = & ! a
            [0.0005627823349452919_dp, 0.0030490190788550874_dp, 0.00790029086110356_dp, &
             0.015922394599248355_dp, 0.028638903303737875_dp, 0.048675131703015205_dp, &
             0.08024582715328522_dp, 0.12984825884578247_dp, 0.20732114251968206_dp, &
             0.3274359941701851_dp, 0.5122363225944793_dp, 0.7944607143227888_dp, &
             1.2226172507136812_dp, 1.8688040470778537_dp, 2.8418409276923855_dp, &
             4.313147383547011_dp, 6.582816326266341_dp, 10.338005655836904_dp, &
             ! w
             0.0014535131106379318_dp, 0.003578211128353728_dp, 0.006253890790180701_dp, &
             0.01003778483872021_dp, 0.015824937514605023_dp, 0.02493883699939866_dp, &
             0.03926815195450443_dp, 0.06155046121917452_dp, 0.09581900405954127_dp, &
             0.14802882375626922_dp, 0.226946010516923_dp, 0.34546224809925985_dp, &
             0.5226622305512622_dp, 0.7874628303833764_dp, 1.1862785054223426_dp, &
             1.8034459541421743_dp, 2.834501267371012_dp, 5.0259251227466635_dp]

      CASE (563)

         aw(:) = & ! a
            [0.0005099568208449972_dp, 0.002766060793701837_dp, 0.0071837947293005204_dp, &
             0.014532473440193957_dp, 0.026274445734568615_dp, 0.04493651798185145_dp, &
             0.07458604801491614_dp, 0.12151947948149408_dp, 0.1953210695123804_dp, &
             0.3104564089196821_dp, 0.48862004258264824_dp, 0.7621798194464763_dp, &
             1.1792917816666157_dp, 1.811783037470562_dp, 2.768358759293376_dp, &
             4.220533150129275_dp, 6.4686395215857635_dp, 10.199693737592291_dp, &
             ! w
             0.0013174240913447924_dp, 0.0032508514395187126_dp, 0.005708210041233193_dp, &
             0.009226639366647683_dp, 0.014670806696528513_dp, 0.023320152255546962_dp, &
             0.03701103207597337_dp, 0.058427146196618984_dp, 0.09154862030420802_dp, &
             0.14228168423189544_dp, 0.21935730563808947_dp, 0.3356619957096291_dp, &
             0.5103268370972778_dp, 0.77238695806063_dp, 1.1684483932541831_dp, &
             1.7830579341042165_dp, 2.811774344897853_dp, 4.999932993106821_dp]

      CASE (564)

         aw(:) = & ! a
            [0.0004668841916966001_dp, 0.0025351831509706916_dp, 0.006598421692551525_dp, &
             0.013394663714263032_dp, 0.02433332568028017_dp, 0.04185553599376796_dp, &
             0.0699004840051102_dp, 0.11459008560534774_dp, 0.18528689359114475_dp, &
             0.29618860170128614_dp, 0.4686822380294337_dp, 0.734806567718747_dp, &
             1.1424023311783487_dp, 1.7630505379545083_dp, 2.7053476877215945_dp, &
             4.140886916302708_dp, 6.370216716705846_dp, 10.080246883158214_dp, &
             ! w
             0.001206442622027735_dp, 0.0029835180690567413_dp, 0.005261449937462656_dp, &
             0.00855990693672605_dp, 0.013716816741912066_dp, 0.021973107576248586_dp, &
             0.03511983495236686_dp, 0.05579372725380304_dp, 0.08792773068803054_dp, &
             0.13738377236828653_dp, 0.21285979561082896_dp, 0.32723512527144366_dp, &
             0.49967909195525034_dp, 0.7593294858673185_dp, 1.1529622794480074_dp, &
             1.7653147148395656_dp, 2.7919762112360456_dp, 4.977297624597095_dp]

      CASE (565)

         aw(:) = & ! a
            [0.000260914461772336_dp, 0.0014284110053859973_dp, 0.003779313366150407_dp, &
             0.007875534265517174_dp, 0.014818727120990572_dp, 0.02654437101786918_dp, &
             0.04623735420014881_dp, 0.07899162477486178_dp, 0.1328484337557623_dp, &
             0.22037931923229523_dp, 0.3610645276411059_dp, 0.5848529330794866_dp, &
             0.9375236694343315_dp, 1.4889766904451114_dp, 2.3469686778446492_dp, &
             3.6834788445658697_dp, 5.800450143499236_dp, 9.3845111360647_dp, &
             ! w
             0.0006754355558846693_dp, 0.0016981110123511526_dp, 0.003093562039844899_dp, &
             0.0052777001406530315_dp, 0.008924677726232818_dp, 0.015046730066029211_dp, &
             0.025171297216547893_dp, 0.04165165016169422_dp, 0.06811847034128733_dp, &
             0.11013206715594186_dp, 0.17614288778329085_dp, 0.2789306706204916_dp, &
             0.4378452602827844_dp, 0.6826254815697036_dp, 1.0611210798041206_dp, &
             1.6593600203583403_dp, 2.6733428837255038_dp, 4.841796324444054_dp]

      CASE (566)

         aw(:) = & ! a
            [0.00018551321377360634_dp, 0.0010215028674021077_dp, 0.0027344893872422236_dp, &
             0.0058038700988552044_dp, 0.011181048177484084_dp, 0.020551830095233446_dp, &
             0.03673043057862141_dp, 0.06430201414145639_dp, 0.11064025372147042_dp, &
             0.1874733094281667_dp, 0.31326429093244756_dp, 0.5168126983136785_dp, &
             0.8427221509500933_dp, 1.3598834122689585_dp, 2.1754818299235206_dp, &
             3.4616215444359195_dp, 5.5210242574181025_dp, 9.040418361508078_dp, &
             ! w
             0.00048085270575301206_dp, 0.0012230540357134345_dp, 0.0022794773615028817_dp, &
             0.00401365797781974_dp, 0.007015403763875897_dp, 0.012184761187942284_dp, &
             0.020920071473765355_dp, 0.03542655912789011_dp, 0.05916587370855765_dp, &
             0.09751885164172235_dp, 0.1587743800332417_dp, 0.2556200593738798_dp, &
             0.40746039857466365_dp, 0.6443266992963216_dp, 1.0146526822324917_dp, &
             1.6052318670779617_dp, 2.612440353907917_dp, 4.772316603465554_dp]

      CASE (567)

         aw(:) = & ! a
            [0.00014562382710294662_dp, 0.0008055881529796759_dp, 0.0021769113040709318_dp, &
             0.0046881273958434425_dp, 0.009196001236600226_dp, 0.017228905560492613_dp, &
             0.03136764944560351_dp, 0.05587429246023262_dp, 0.09769241006638396_dp, &
             0.16799814339082067_dp, 0.28457797728806533_dp, 0.4754535138655539_dp, &
             0.7844166640229894_dp, 1.279641798278453_dp, 2.0678834725616793_dp, &
             3.3212938447939195_dp, 5.3431231415756_dp, 8.820257814253507_dp, &
             ! w
             0.0003778416599558625_dp, 0.0009700552711586841_dp, 0.0018409801051395853_dp, &
             0.0033204021060857123_dp, 0.005944000305659858_dp, 0.010541731118507081_dp, &
             0.01842984272446517_dp, 0.031715495560804136_dp, 0.05374476719644769_dp, &
             0.089772198289391_dp, 0.14796801202917811_dp, 0.24094326281169415_dp, &
             0.3881219089533589_dp, 0.6197175136938446_dp, 0.9845552848798075_dp, &
             1.5699678410487687_dp, 2.5726424686871687_dp, 4.726941044835544_dp]

      CASE (568)

         aw(:) = & ! a
            [0.00012069306501636694_dp, 0.0006703183122791028_dp, 0.0018259870924344271_dp, &
             0.003980667589060772_dp, 0.0079240714937168_dp, 0.01507315081217762_dp, &
             0.0278434453982068_dp, 0.05026645010562097_dp, 0.0889754985015726_dp, &
             0.15474407546557292_dp, 0.26485954501264436_dp, 0.4467628386148967_dp, &
             0.743631633383162_dp, 1.2230881055827414_dp, 1.9915418966860663_dp, &
             3.2211612799836016_dp, 5.215590858427921_dp, 8.661879729838459_dp, &
             ! w
             0.0003134246142158033_dp, 0.0008110913445565603_dp, 0.001562938469288158_dp, &
             0.0028744264048654856_dp, 0.005242599992383257_dp, 0.009448127503548561_dp, &
             0.0167483721901341_dp, 0.029178210827365967_dp, 0.049996895682636794_dp, &
             0.08436235623763434_dp, 0.14035158628733707_dp, 0.23051125331945824_dp, &
             0.374270678501655_dp, 0.6019710721015192_dp, 0.962727552491358_dp, &
             1.5442858258917591_dp, 2.54359484716486_dp, 4.693835053963997_dp]

      CASE (569)

         aw(:) = & ! a
            [0.00010353081360477903_dp, 0.000577010708614185_dp, 0.0015829797899359808_dp, &
             0.003487643090278658_dp, 0.00702981637575661_dp, 0.013542012955254482_dp, &
             0.02531434596654306_dp, 0.04620212975837872_dp, 0.08259960522442154_dp, &
             0.14496728229470254_dp, 0.2502012538239559_dp, 0.42528316526304255_dp, &
             0.7128998743123581_dp, 1.1802262631739215_dp, 1.9333849194731854_dp, &
             3.1445446516847144_dp, 5.117661524304808_dp, 8.539939250261368_dp, &
             ! w
             0.0002690595993960665_dp, 0.0007011712205890843_dp, 0.0013691799768125582_dp, &
             0.0025598301686818336_dp, 0.004740768169827436_dp, 0.0086554545194189_dp, &
             0.015515970318803389_dp, 0.02730051705055568_dp, 0.047199373682666385_dp, &
             0.08029272329388433_dp, 0.13458110045081412_dp, 0.22255593727666947_dp, &
             0.36364527296104165_dp, 0.5882861097353751_dp, 0.9458213024215231_dp, &
             1.5243296194546507_dp, 2.5209846078020965_dp, 4.6680728012172406_dp]

      CASE (570)

         aw(:) = & ! a
            [9.094343130849975e-05_dp, 0.0005084550515824451_dp, 0.0014038244792441586_dp, &
             0.0031221229770495226_dp, 0.006361736151585623_dp, 0.012388206466221354_dp, &
             0.02339197730376753_dp, 0.04308751795427685_dp, 0.07767658914909154_dp, &
             0.13736600522163228_dp, 0.2387325306496574_dp, 0.40838029574768436_dp, &
             0.6885894480169676_dp, 1.1461602509232325_dp, 1.8869698702994993_dp, &
             3.0831793456937975_dp, 5.039000247922106_dp, 8.44178006007088_dp, &
             ! w
             0.00023650765020381912_dp, 0.0006202374982483119_dp, 0.0012255391766448991_dp, &
             0.002324124178832601_dp, 0.004360281588623125_dp, 0.008048021240346297_dp, &
             0.014562986383659132_dp, 0.025837103982347814_dp, 0.0450037848576161_dp, &
             0.07707846816747213_dp, 0.12999708938848178_dp, 0.2162028653520236_dp, &
             0.35511913345417484_dp, 0.5772581722343985_dp, 0.9321489117777328_dp, &
             1.508148002041756_dp, 2.5026251803844644_dp, 4.647158214149157_dp]

      CASE (571)

         aw(:) = & ! a
            [8.128780459273251e-05_dp, 0.00045578417617948527_dp, 0.0012657578123163504_dp, &
             0.0028390119726776426_dp, 0.005840758669466495_dp, 0.01148166594744145_dp, &
             0.021870335147125514_dp, 0.04060496584282268_dp, 0.07372748394495345_dp, &
             0.13123282773857478_dp, 0.22942956229848835_dp, 0.394602877306536_dp, &
             0.6686870205875045_dp, 1.1181608634571265_dp, 1.8486875233181934_dp, &
             3.032415620913765_dp, 4.973772138903806_dp, 8.360237513682502_dp, &
             ! w
             0.0002115284740783797_dp, 0.0005579383421879089_dp, 0.0011142909499591146_dp, &
             0.002139849008725855_dp, 0.004059756005062852_dp, 0.007563907832157933_dp, &
             0.013797676055523244_dp, 0.024654104457232505_dp, 0.04321845009911772_dp, &
             0.07445088564571238_dp, 0.12623156194056043_dp, 0.21096100031431056_dp, &
             0.34805600038710355_dp, 0.5680899756605817_dp, 0.920748230015768_dp, &
             1.494625048261868_dp, 2.487264046018687_dp, 4.629662033992615_dp]

      CASE (572)

         aw(:) = & ! a
            [7.362902380322699e-05_dp, 0.0004139464765750486_dp, 0.0011557812297434123_dp, &
             0.0026124632635369602_dp, 0.0054213163235300055_dp, 0.010746931198410465_dp, &
             0.02062903374098843_dp, 0.03856750896122895_dp, 0.07046844263299985_dp, &
             0.1261458233762945_dp, 0.22167806260075487_dp, 0.38307528434830906_dp, &
             0.6519717385575523_dp, 1.094565508121492_dp, 1.816330140277326_dp, &
             2.9893991734035_dp, 4.918384968674218_dp, 8.290890939893337_dp, &
             ! w
             0.00019170870711407288_dp, 0.0005083673996572885_dp, 0.0010252753388815708_dp, &
             0.001991147955096542_dp, 0.0038150596475737433_dp, 0.007166653733954372_dp, &
             0.013165550359203417_dp, 0.023671412268265173_dp, 0.041727894784833346_dp, &
             0.07224710204648896_dp, 0.12306019974849018_dp, 0.20652945524229882_dp, &
             0.3420641412628419_dp, 0.5602885647626088_dp, 0.9110222459597854_dp, &
             1.48306650559867_dp, 2.4741208992697565_dp, 4.614694078252486_dp]

      CASE (573)

         aw(:) = & ! a
            [6.739449324435924e-05_dp, 0.00037984457242887693_dp, 0.0010659079537327376_dp, &
             0.002426542068960052_dp, 0.0050751785351652095_dp, 0.01013697193763941_dp, &
             0.01959256624028302_dp, 0.03685714980164952_dp, 0.06771926281964358_dp, &
             0.12183564859596094_dp, 0.21508388041400503_dp, 0.37323301422066457_dp, &
             0.6376531033583206_dp, 1.0742933968444754_dp, 1.788457502038564_dp, &
             2.952262375077832_dp, 4.870482497534054_dp, 8.230835304724469_dp, &
             ! w
             0.00017556987663771791_dp, 0.00046789829255630005_dp, 0.0009522284541816575_dp, &
             0.001868179296820573_dp, 0.003611085140121934_dp, 0.006833240477242143_dp, &
             0.012631955334979236_dp, 0.022837752985687437_dp, 0.04045778008255227_dp, &
             0.07036171339893307_dp, 0.12033713046284121_dp, 0.20271168814760507_dp, &
             0.33688662391872987_dp, 0.5535294427041705_dp, 0.9025767816825688_dp, &
             1.4730129882137049_dp, 2.462678834399958_dp, 4.601664805063239_dp]

      CASE (574)

         aw(:) = & ! a
            [3.770972671447743e-05_dp, 0.00021671815823933924_dp, 0.0006319647958778312_dp, &
             0.0015150185101379097_dp, 0.0033448466783742893_dp, 0.007025124585709654_dp, &
             0.014201414935093517_dp, 0.027801329556036924_dp, 0.052926656904140085_dp, &
             0.09830251130329612_dp, 0.17860015065735232_dp, 0.3181200064054065_dp, &
             0.5565968916143585_dp, 0.9584083064077357_dp, 1.6277447068619073_dp, &
             2.736552172466405_dp, 4.590583863831329_dp, 7.87837440091346_dp, &
             ! w
             9.864614157749192e-05_dp, 0.0002732211433595114_dp, 0.0005941975577521344_dp, &
             0.0012489573415471545_dp, 0.002556203182249123_dp, 0.005069908308872515_dp, &
             0.00975621380152378_dp, 0.018270493030267572_dp, 0.03339675488588097_dp, &
             0.059740515679277045_dp, 0.10481090900822707_dp, 0.18070308280191333_dp, &
             0.306740164242035_dp, 0.513823817457257_dp, 0.852592911322043_dp, &
             1.4131782077857424_dp, 2.39437286369401_dp, 4.523908764999524_dp]

      CASE (575)

         aw(:) = & ! a
            [2.6892519554907363e-05_dp, 0.0001567937858852912_dp, 0.00046989399004324_dp, &
             0.0011654522038541332_dp, 0.002659895069561157_dp, 0.005753840073969323_dp, &
             0.011934211172250079_dp, 0.02389280033840905_dp, 0.04639252185719773_dp, &
             0.08769012794026224_dp, 0.16183935850207254_dp, 0.2923748366935609_dp, &
             0.5181611748116954_dp, 0.9027178845025317_dp, 1.5496007583408882_dp, &
             2.6306172996201584_dp, 4.4520231849372305_dp, 7.7028603230062735_dp, &
             ! w
             7.05638795782659e-05_dp, 0.0002010028679147231_dp, 0.0004569474254904519_dp, &
             0.00100093647179562_dp, 0.002116528445009556_dp, 0.00431084332567352_dp, &
             0.008484575895053869_dp, 0.016203457881559215_dp, 0.030134877504812176_dp, &
             0.05474301588738865_dp, 0.09738330010494875_dp, 0.1700150284458082_dp, &
             0.2919005680926316_dp, 0.49404330049838874_dp, 0.8274397076545656_dp, &
             1.382839379424492_dp, 2.3595954111938178_dp, 4.484334415134187_dp]

      CASE (576)

         aw(:) = & ! a
            [2.117764284060119e-05_dp, 0.00012495595709202876_dp, 0.00038274936178464845_dp, &
             0.000973991658367676_dp, 0.0022768841713981217_dp, 0.00502884426710744_dp, &
             0.01061831718959387_dp, 0.021588822141837025_dp, 0.04248772649919_dp, &
             0.08127067048807829_dp, 0.1515902772342063_dp, 0.2764785036017604_dp, &
             0.4942222305733775_dp, 0.8677632859471236_dp, 1.5002201240187492_dp, &
             2.5632899998356247_dp, 4.3635548020944075_dp, 7.590418907169732_dp, &
             ! w
             5.570901072452278e-05_dp, 0.0001623667699557508_dp, 0.00038178255643626983_dp, &
             0.0008611922014131465_dp, 0.0018628123283926767_dp, 0.003864421172997219_dp, &
             0.007724785323416311_dp, 0.014951470752811016_dp, 0.02813530637477198_dp, &
             0.05164647162933681_dp, 0.0927364294632549_dp, 0.16326990553539597_dp, &
             0.2824618316106025_dp, 0.4813744803843905_dp, 0.8112354873409041_dp, &
             1.3632083348029826_dp, 2.337037609569253_dp, 4.458669560744956_dp]

      CASE (577)

         aw(:) = & ! a
            [1.7607027510129746e-05_dp, 0.00010497463258634355_dp, 0.0003275240571061572_dp, &
             0.0008508891384294316_dp, 0.0020267594369468207_dp, 0.00454854527962152_dp, &
             0.009735483724482805_dp, 0.020025950938082987_dp, 0.03981323330404025_dp, &
             0.07683612602150897_dp, 0.14445632795349211_dp, 0.2653385950247986_dp, &
             0.4773444815371746_dp, 0.8429864668461382_dp, 1.465052941517388_dp, &
             2.5151506872652627_dp, 4.30009786276047_dp, 7.509578251764822_dp, &
             ! w
             4.6418688781564724e-05_dp, 0.00013798379756057278_dp, 0.0003334522876238467_dp, &
             0.0007694194914053827_dp, 0.0016933362995822042_dp, 0.003562194356292001_dp, &
             0.007204639023093123_dp, 0.014086099235944071_dp, 0.026741496764881377_dp, &
             0.04947175447035527_dp, 0.0894508734055345_dp, 0.1584717931748057_dp, &
             0.2757109740714004_dp, 0.4722696710372406_dp, 0.7995424678739795_dp, &
             1.34899902892753_dp, 2.320682074663291_dp, 4.44006317705371_dp]

      CASE (578)

         aw(:) = & ! a
            [1.5148726194573733e-05_dp, 9.116593247691507e-05_dp, 0.0002890408736140009_dp, &
             0.0007640698684906546_dp, 0.0018481420000240024_dp, 0.004201670246314368_dp, &
             0.009091615193402356_dp, 0.01887638637434539_dp, 0.03783136781422241_dp, &
             0.07352849364924609_dp, 0.139104394958239_dp, 0.2569382141063124_dp, &
             0.46455875240209477_dp, 0.8241402125840703_dp, 1.438207991535087_dp, &
             2.4782925667780993_dp, 4.251394581677406_dp, 7.447423631712815_dp, &
             ! w
             4.0017293410861965e-05_dp, 0.00012105345936715976_dp, 0.00029936278219422834_dp, &
             0.0007035928891270861_dp, 0.0015701697517859468_dp, 0.00334027399899005_dp, &
             0.006819417528035018_dp, 0.013440467638629175_dp, 0.02569488217763655_dp, &
             0.04782938380103958_dp, 0.08695684825186281_dp, 0.1548128025198901_dp, &
             0.2705415374764525_dp, 0.4652722313536357_dp, 0.7905281489978131_dp, &
             1.3380193233848747_dp, 2.3080275343987338_dp, 4.42566810636102_dp]

      CASE (579)

         aw(:) = & ! a
            [1.3345094685759858e-05_dp, 8.100133456933608e-05_dp, 0.00026050551878049904_dp, &
             0.0006990271202941324_dp, 0.0017129267268830544_dp, 0.003936641186117154_dp, &
             0.008595728011419668_dp, 0.017984913280872282_dp, 0.03628522800112553_dp, &
             0.07093447544172429_dp, 0.1348875922544271_dp, 0.25029216032872337_dp, &
             0.4544059616545287_dp, 0.8091261466532933_dp, 1.416760821430585_dp, &
             2.448774725138738_dp, 4.212315572771937_dp, 7.39748146374_dp, &
             ! w
             3.53174020538624e-05_dp, 0.00010853921152648619_dp, 0.0002738195210148376_dp, &
             0.0006535800730987856_dp, 0.0014755921767249933_dp, 0.003168438229454705_dp, &
             0.006519064629723313_dp, 0.01293407952215318_dp, 0.02486972094779215_dp, &
             0.046528561218468496_dp, 0.08497336062732552_dp, 0.1518920884205169_dp, &
             0.26640150356795944_dp, 0.4596518597177716_dp, 0.7832699688296861_dp, &
             1.3291621588649858_dp, 2.297808666622791_dp, 4.414044285275878_dp]

      CASE (580)

         aw(:) = & ! a
            [1.1960898364376281e-05_dp, 7.317769815105737e-05_dp, 0.0002383972813355374_dp, &
             0.0006481775411288835_dp, 0.0016062704435372192_dp, 0.003725946065946414_dp, &
             0.008198851210705769_dp, 0.017267313986842944_dp, 0.03503441932495886_dp, &
             0.06882675294094925_dp, 0.13144806430515366_dp, 0.24485257610525005_dp, &
             0.44607090649407055_dp, 0.7967669300061988_dp, 1.3990644948095807_dp, &
             2.4243707572451285_dp, 4.179955615172348_dp, 7.356078282900798_dp, &
             ! w
             3.1708293644241954e-05_dp, 9.887110987799206e-05_dp, 0.000253846227836986_dp, &
             0.0006140092739336479_dp, 0.0014000922946545429_dp, 0.00303030469936904_dp, &
             0.006276220815551653_dp, 0.01252261978208976_dp, 0.024196345194016693_dp, &
             0.04546296242998983_dp, 0.08334300427837689_dp, 0.1494840336439967_dp, &
             0.26297881769337894_dp, 0.4549941272767069_dp, 0.777242691667036_dp, &
             1.321795719818751_dp, 2.2893023648446493_dp, 4.404368838990043_dp]

      CASE (581)

         aw(:) = & ! a
            [1.086236958166054e-05_dp, 6.695225038113829e-05_dp, 0.00022069965324013414_dp, &
             0.0006071441481048981_dp, 0.001519530395591377_dp, 0.003553430868596311_dp, &
             0.007872013931026682_dp, 0.016673432069369516_dp, 0.03399483044267095_dp, &
             0.06706841370437984_dp, 0.12856924808110373_dp, 0.2402864866488988_dp, &
             0.43905620101121207_dp, 0.7863417225293414_dp, 1.3841075559674645_dp, &
             2.4037097403756573_dp, 4.152522016546931_dp, 7.320943814292977_dp, &
             ! w
             2.884249165620708e-05_dp, 9.115179144103407e-05_dp, 0.00023772524880348903_dp, &
             0.0005817423884258551_dp, 0.0013380566711371329_dp, 0.002916125763845916_dp, &
             0.006074495196136801_dp, 0.01217938058716831_dp, 0.023632544691145112_dp, &
             0.04456786198719827_dp, 0.08196954278336645_dp, 0.14745015357154495_dp, &
             0.26008125555920814_dp, 0.45104293350498725_dp, 0.7721208601525964_dp, &
             1.3155277008666069_dp, 2.282059133476969_dp, 4.396130310968723_dp]

      CASE (582)

         aw(:) = & ! a
            [9.967611585535554e-06_dp, 6.186929142075591e-05_dp, 0.00020617003621056866_dp, &
             0.0005732111078902096_dp, 0.0014473024069595088_dp, 0.0034089209953625916_dp, &
             0.007596850079955685_dp, 0.01617128837200462_dp, 0.033112560068507886_dp, &
             0.06557132006336473_dp, 0.12611115961852515_dp, 0.23637786866043078_dp, &
             0.4330380932913855_dp, 0.777379946000164_dp, 1.371228004151114_dp, &
             2.385892431854792_dp, 4.128836800009076_dp, 7.290584309037711_dp, &
             ! w
             2.650715831062608e-05_dp, 8.482932543960835e-05_dp, 0.00022439089726813142_dp, &
             0.0005548115989066422_dp, 0.0012859342248887997_dp, 0.0028196916909736833_dp, &
             0.0059033849193231265_dp, 0.011887161404925883_dp, 0.023151011322468745_dp, &
             0.04380121306039398_dp, 0.08079022758313453_dp, 0.14569985265956242_dp, &
             0.2575826900435412_dp, 0.44762979378375634_dp, 0.7676898787855397_dp, &
             1.3100989864647254_dp, 2.275781795930811_dp, 4.388990568421244_dp]

      CASE (583)

         aw(:) = & ! a
            [5.696283282011203e-06_dp, 3.7394747769333536e-05_dp, 0.0001348082926064332_dp, &
             0.000402363370425018_dp, 0.0010752236274931557_dp, 0.002649879374199912_dp, &
             0.006127709733401064_dp, 0.013452712759965424_dp, 0.02827842220929969_dp, &
             0.05728235585657594_dp, 0.11237595298013743_dp, 0.21435935889926439_dp, &
             0.3988910474008291_dp, 0.726205544331862_dp, 1.2972724973010605_dp, &
             2.2831028096327666_dp, 3.9916826660137774_dp, 7.1143030231141795_dp, &
             ! w
             1.5341201465564108e-05_dp, 5.403787101728824e-05_dp, 0.00015718832128219762_dp, &
             0.0004150424006100662_dp, 0.0010095539551882964_dp, 0.002299668561776539_dp, &
             0.004967711953511185_dp, 0.01027011013570269_dp, 0.020458660533292164_dp, &
             0.03947552409429393_dp, 0.07408207511497567_dp, 0.1356715340876873_dp, &
             0.24317428843000152_dp, 0.4278346574865771_dp, 0.7418671103674677_dp, &
             1.2783452047799542_dp, 2.2389880177703896_dp, 4.347144558391107_dp]

      CASE (584)

         aw(:) = & ! a
            [4.131731825230856e-06_dp, 2.829535695862735e-05_dp, 0.00010735765528298308_dp, &
             0.0003340312073889461_dp, 0.0009212780132154344_dp, 0.0023269721702441413_dp, &
             0.005488219693965525_dp, 0.012246412491027748_dp, 0.026098018884433305_dp, &
             0.053490527392179514_dp, 0.10601487244590438_dp, 0.20405115588315595_dp, &
             0.3827514358066326_dp, 0.7018139874113792_dp, 1.261764629063452_dp, &
             2.2334468616577503_dp, 3.925101471939464_dp, 7.028424722729097_dp, &
             ! w
             1.1241382519879153e-05_dp, 4.235813973942612e-05_dp, 0.0001302574703258198_dp, &
             0.0003566095444644539_dp, 0.0008904772803443534_dp, 0.002070320729328212_dp, &
             0.00454706396508409_dp, 0.009531275821605133_dp, 0.019211259169201196_dp, &
             0.03744684079611271_dp, 0.07090207935618785_dp, 0.1308720023904189_dp, &
             0.23621960717189322_dp, 0.4182082223567253_dp, 0.7292298964037207_dp, &
             1.2627306979738906_dp, 2.220845935711008_dp, 4.326512619315605_dp]

      CASE (585)

         aw(:) = & ! a
            [3.3022628684719952e-06_dp, 2.3420630611943275e-05_dp, 9.230409449871005e-05_dp, &
             0.000295637516785855_dp, 0.0008330324547064296_dp, 0.0021388850653163097_dp, &
             0.005110845986392017_dp, 0.011526811995954624_dp, 0.0247853719976952_dp, &
             0.05118975431275387_dp, 0.10212869459707498_dp, 0.19771575316493134_dp, &
             0.372779679793518_dp, 0.686673954558246_dp, 1.2396360078823205_dp, &
             2.2023964419078634_dp, 3.883355737635126_dp, 6.97447560324605_dp, &
             ! w
             9.064938310095719e-06_dp, 3.601111583058008e-05_dp, 0.0001151038349590294_dp, &
             0.0003229219111561771_dp, 0.0008206449894361141_dp, 0.001934028030877805_dp, &
             0.0042943745617944_dp, 0.00908340287300126_dp, 0.01844919918242188_dp, &
             0.03619907581731908_dp, 0.06893451843829064_dp, 0.12788668213311002_dp, &
             0.23187347047208787_dp, 0.41216763904962367_dp, 0.7212724191626193_dp, &
             1.252872370489374_dp, 2.2093745269363225_dp, 4.313467128667264_dp]

      CASE (586)

         aw(:) = & ! a
            [2.782832013879544e-06_dp, 2.0342603732064615e-05_dp, 8.2625907295173e-05_dp, &
             0.0002705161068050067_dp, 0.0007744772173361553_dp, 0.0020126926988123786_dp, &
             0.004855389088415757_dp, 0.011036089952012563_dp, 0.023884657719413113_dp, &
             0.04960259574228993_dp, 0.09943549273473215_dp, 0.19330747767180548_dp, &
             0.3658165918240841_dp, 0.6760691117386491_dp, 1.2240942741386258_dp, &
             2.180539227329422_dp, 3.8539170196075587_dp, 6.936381923548616_dp, &
             ! w
             7.700874967225928e-06_dp, 3.195765846889435e-05_dp, 0.00010517748542706982_dp, &
             0.0003004829197892872_dp, 0.000773582704782333_dp, 0.0018413406871863173_dp, &
             0.004121261522515641_dp, 0.008774678444653202_dp, 0.017921132855175225_dp, &
             0.03533049138870044_dp, 0.06755938506374413_dp, 0.12579283396331942_dp, &
             0.2288155791403282_dp, 0.4079058107963534_dp, 0.7156450471221592_dp, &
             1.2458883318824714_dp, 2.2012394565082762_dp, 4.304215870787817_dp]

      CASE (587)

         aw(:) = & ! a
            [2.424709952197596e-06_dp, 1.8205641806588997e-05_dp, 7.580715872241364e-05_dp, &
             0.00025257376110559345_dp, 0.0007322079967535588_dp, 0.0019208383754019853_dp, &
             0.004668201569020592_dp, 0.010674534494176473_dp, 0.023217964280440775_dp, &
             0.04842317777538039_dp, 0.0974273510544483_dp, 0.19001073832058646_dp, &
             0.36059563542059514_dp, 0.6680993599217881_dp, 1.212391210597834_dp, &
             2.164053146177339_dp, 3.831683185696595_dp, 6.907584031603642_dp, &
             ! w
             6.7598998898177216e-06_dp, 2.911663206152118e-05_dp, 9.808144164118887e-05_dp, &
             0.00028423956571576854_dp, 0.000739214990519656_dp, 0.0017731956543435997_dp, &
             0.003993287166432996_dp, 0.008545407277659752_dp, 0.017527439560537577_dp, &
             0.03468074240385815_dp, 0.06652766613940835_dp, 0.12421777907099418_dp, &
             0.22651001207474478_dp, 0.4046859734903045_dp, 0.7113862270233584_dp, &
             1.240595852485868_dp, 2.195070117262737_dp, 4.297200094901468_dp]

      CASE (588)

         aw(:) = & ! a
            [2.1617612977125257e-06_dp, 1.662707882841798e-05_dp, 7.070738181995456e-05_dp, &
             0.0002390053522236388_dp, 0.0006999705975978091_dp, 0.0018503220278935197_dp, &
             0.004523742362597568_dp, 0.010394306552790484_dp, 0.022699368869302145_dp, &
             0.04750292832145165_dp, 0.09585631535628808_dp, 0.18742560438804043_dp, &
             0.3564932992821488_dp, 0.6618260190405989_dp, 1.2031650174356021_dp, &
             2.1510393992097123_dp, 3.8141142840292135_dp, 6.884811502764588_dp, &
             ! w
             6.0687294242645945e-06_dp, 2.700074746228437e-05_dp, 9.271128018315657e-05_dp, &
             0.00027182447699993916_dp, 0.000712764856522128_dp, 0.0017204697341860083_dp, &
             0.003893842534101108_dp, 0.008366609214019969_dp, 0.01721947990754413_dp, &
             0.03417114816545018_dp, 0.06571662733950154_dp, 0.1229770964564079_dp, &
             0.22469062235353063_dp, 0.40214108861742204_dp, 0.7080156566129415_dp, &
             1.2364029380934565_dp, 2.1901796686416946_dp, 4.291638686770147_dp]

      CASE (589)

         aw(:) = & ! a
            [1.9599124237347485e-06_dp, 1.5408796324721776e-05_dp, 6.672928881315168e-05_dp, &
             0.0002283231153317559_dp, 0.0006744124432153255_dp, 0.001794114301366692_dp, &
             0.004408101880317317_dp, 0.010169195236120976_dp, 0.02228155036743529_dp, &
             0.04675965518726107_dp, 0.09458467599906724_dp, 0.185329198231743_dp, &
             0.3531610539288955_dp, 0.6567229797403219_dp, 1.1956506445395527_dp, &
             2.1404291323927076_dp, 3.7997782811410152_dp, 6.866218343211515_dp, &
             ! w
             5.538015286397209e-06_dp, 2.5356016265939915e-05_dp, 8.84807914757732e-05_dp, &
             0.00026196458014929465_dp, 0.0006916394476519007_dp, 0.0016781748489766149_dp, &
             0.0038137914062414038_dp, 0.008222259952715933_dp, 0.016970238946182156_dp, &
             0.03375783763563263_dp, 0.06505760039970954_dp, 0.12196729387503166_dp, &
             0.22320764776371582_dp, 0.40006411712790924_dp, 0.7052618438013496_dp, &
             1.232974431798126_dp, 2.1861789066113344_dp, 4.287089030894947_dp]

      CASE (590)

         aw(:) = & ! a
            [1.7997487886600253e-06_dp, 1.4437416506510273e-05_dp, 6.352759509775112e-05_dp, &
             0.0002196577497868786_dp, 0.0006535572109381088_dp, 0.001748041695842587_dp, &
             0.004312973052403364_dp, 0.009983470619054852_dp, 0.021935992668723307_dp, &
             0.046143651360029715_dp, 0.09352888828088651_dp, 0.1835859264565479_dp, &
             0.350386332817856_dp, 0.652468662742327_dp, 1.1893795648402337_dp, &
             2.1315667047041567_dp, 3.7877956267574993_dp, 6.8506696917162_dp, &
             ! w
             5.116816287129631e-06_dp, 2.40362059065755e-05_dp, 8.504722458150927e-05_dp, &
             0.0002539074613698106_dp, 0.0006742946504348717_dp, 0.0016433224941597862_dp, &
             0.003747633275801227_dp, 0.00810267281450098_dp, 0.016763328206748914_dp, &
             0.033414112909622574_dp, 0.0645086782794706_dp, 0.12112505044683754_dp, &
             0.2219692502496186_dp, 0.39832784577851715_dp, 0.702957692850675_dp, &
             1.23010378957972_dp, 2.182827804206717_dp, 4.2832781644877125_dp]

      CASE (591)

         aw(:) = & ! a
            [1.6693645060820608e-06_dp, 1.3643146486114305e-05_dp, 6.088783522060366e-05_dp, &
             0.00021246429938838942_dp, 0.0006361564791437423_dp, 0.001709451765030174_dp, &
             0.004233050310459982_dp, 0.00982704453711614_dp, 0.021644342092136296_dp, &
             0.04562282514340107_dp, 0.09263486884498745_dp, 0.1821078041898917_dp, &
             0.3480309173484621_dp, 0.6488535836239001_dp, 1.184046085764974_dp, &
             2.124023770793666_dp, 3.7775910854683334_dp, 6.837422776978846_dp, &
             ! w
             4.773877589090357e-06_dp, 2.2950812678044945e-05_dp, 8.219558575957006e-05_dp, &
             0.00024717682984820225_dp, 0.0006597465544647993_dp, 0.0016139989585811682_dp, &
             0.003691831245964453_dp, 0.008001596714215905_dp, 0.01658813887916468_dp, &
             0.03312264491246706_dp, 0.06404259600944529_dp, 0.12040908345969413_dp, &
             0.22091544412851835_dp, 0.3968490483283259_dp, 0.7009937354717216_dp, &
             1.227655558898501_dp, 2.179968861224675_dp, 4.280026981213034_dp]

      CASE (592)

         aw(:) = & ! a
            [1.0515321068239267e-06_dp, 9.821093591985666e-06_dp, 4.7839586195355874e-05_dp, &
             0.00017615019972178953_dp, 0.000546950548985951_dp, 0.0015092946854455794_dp, &
             0.0038146738458003215_dp, 0.009002030655836032_dp, 0.020096495053154233_dp, &
             0.042844000310692514_dp, 0.08784307510117485_dp, 0.17415379677431764_dp, &
             0.33531195548043674_dp, 0.6292732085239225_dp, 1.15508235334966_dp, &
             2.082971031911345_dp, 3.721955351083245_dp, 6.765109279421807_dp, &
             ! w
             3.148188163122602e-06_dp, 1.7626578682896585e-05_dp, 6.777978796775255e-05_dp, &
             0.0002125503405426769_dp, 0.0005839809932350042_dp, 0.0014598459218290425_dp, &
             0.0033962665131134846_dp, 0.007462888404998172_dp, 0.01564949908707742_dp, &
             0.031553908557126566_dp, 0.06152411499343817_dp, 0.11652685139536159_dp, &
             0.21518370464952713_dp, 0.38878400274729047_dp, 0.6902582498684185_dp, &
             1.2142495238844528_dp, 2.16429814216164_dp, 4.26220623979455_dp]

      CASE (593)

         aw(:) = & ! a
            [8.347282146216849e-07_dp, 8.445609044597734e-06_dp, 4.295846963332234e-05_dp, &
             0.0001621772442246233_dp, 0.0005119326807149494_dp, 0.0014295422866934379_dp, &
             0.003646020808733424_dp, 0.008666316825666471_dp, 0.019461724107915235_dp, &
             0.0416968867194958_dp, 0.0858538110406806_dp, 0.17083559167621434_dp, &
             0.32998326007632633_dp, 0.6210393049021322_dp, 1.142863410912603_dp, &
             2.065605413079893_dp, 3.6983708639129658_dp, 6.734408153656111_dp, &
             ! w
             2.5772983376536385e-06_dp, 1.565429375462735e-05_dp, 6.222292762979304e-05_dp, &
             0.00019889915108187086_dp, 0.0005536417003478005_dp, 0.001397382607273151_dp, &
             0.003275368729985348_dp, 0.0072408211256767335_dp, 0.015260039647555644_dp, &
             0.030899361769210924_dp, 0.06046817347535602_dp, 0.11489215639079002_dp, &
             0.21276113132017072_dp, 0.3853639896416572_dp, 0.6856931427378729_dp, &
             1.2085366578097234_dp, 2.157612021347045_dp, 4.254602733848133_dp]

      CASE (594)

         aw(:) = & ! a
            [7.316974182351593e-07_dp, 7.781929137475314e-06_dp, 4.055569727018987e-05_dp, &
             0.00015520451039960237_dp, 0.0004942921590440873_dp, 0.0013890847744510468_dp, &
             0.003559999883048024_dp, 0.008494338776437845_dp, 0.01913537296945844_dp, &
             0.04110533461974308_dp, 0.08482530332484546_dp, 0.16911611957268938_dp, &
             0.3272165484785323_dp, 0.6167568714399279_dp, 1.1364989952199356_dp, &
             2.0565490922764194_dp, 3.6860593260682344_dp, 6.718370381403715_dp, &
             ! w
             2.305790352725447e-06_dp, 1.4687672181131984e-05_dp, 5.944769342524119e-05_dp, &
             0.00019200911058081936_dp, 0.000538216569141813_dp, 0.001365449054091066_dp, &
             0.003213289879430834_dp, 0.00712638348625628_dp, 0.015058734373588631_dp, &
             0.03056016399499463_dp, 0.05991974024033199_dp, 0.11404146352172441_dp, &
             0.21149824594524436_dp, 0.38357844041147665_dp, 0.6833067027828519_dp, &
             1.2055472917707855_dp, 2.1541114020539553_dp, 4.250621777795089_dp]

      CASE (595)

         aw(:) = & ! a
            [6.83039002664919e-07_dp, 7.465620502939985e-06_dp, 3.939791188408471e-05_dp, &
             0.00015182028159449893_dp, 0.0004856876672380213_dp, 0.0013692786766684858_dp, &
             0.0035177690374484398_dp, 0.008409717013647835_dp, 0.018974491763188005_dp, &
             0.04081325821750734_dp, 0.0843167978175517_dp, 0.16826500323032495_dp, &
             0.32584567313942464_dp, 0.6146330978111908_dp, 1.1333403062727645_dp, &
             2.052051529487636_dp, 3.6799420689868114_dp, 6.710398787959083_dp, &
             ! w
             2.1774829197749774e-06_dp, 1.4222934167466683e-05_dp, 5.8100167850820926e-05_dp, &
             0.00018864510331566967_dp, 0.0005306565260497579_dp, 0.001349752915044938_dp, &
             0.003182707093407167_dp, 0.0070699014087847595_dp, 0.014959222366659827_dp, &
             0.030392262780165366_dp, 0.059647953666310274_dp, 0.11361945826445594_dp, &
             0.21087120318738792_dp, 0.3826911935136782_dp, 0.682120087405447_dp, &
             1.2040601301845306_dp, 2.1523693922115394_dp, 4.248640732432654_dp]

      CASE (596)

         aw(:) = & ! a
            [6.722003459742127e-07_dp, 7.394878326781196e-06_dp, 3.913777831822341e-05_dp, &
             0.0001510576318092805_dp, 0.00048374465492350934_dp, 0.001364799496628946_dp, &
             0.00350820740770161_dp, 0.008390539676953471_dp, 0.018938004226196844_dp, &
             0.0407469730322969_dp, 0.08420133138459711_dp, 0.16807164831677698_dp, &
             0.32553411304757646_dp, 0.614150254525752_dp, 1.1326219530034776_dp, &
             2.051028424957739_dp, 3.678550234303169_dp, 6.7085847754388865_dp, &
             ! w
             2.1488928684308334e-06_dp, 1.4118608741100246e-05_dp, 5.779644587032146e-05_dp, &
             0.00018788516541697197_dp, 0.0005289460218468404_dp, 0.001346197377932109_dp, &
             0.0031757728853528105_dp, 0.007057085096263939_dp, 0.014936627666151461_dp, &
             0.030354118963994026_dp, 0.05958617995199161_dp, 0.11352350292151672_dp, &
             0.21072857617648197_dp, 0.3824893168833493_dp, 0.6818500228117441_dp, &
             1.2037215951561886_dp, 2.151972796095301_dp, 4.248189705220419_dp]

      CASE (597)

         aw(:) = & ! a
            [0.005165332477997096_dp, 0.027453090782803845_dp, 0.06855153379067212_dp, &
             0.13032338044494676_dp, 0.21577379193928328_dp, 0.3294542097505737_dp, &
             0.4780240766071933_dp, 0.6709722829701888_dp, 0.9214784312982677_dp, &
             1.2474223230855346_dp, 1.6726364905803062_dp, 2.228608038820973_dp, &
             2.9569665965904703_dp, 3.9133470930293974_dp, 5.173861728624551_dp, &
             6.847189934997828_dp, 9.100821151066391_dp, 12.231834752574837_dp, &
             16.947785452928883_dp, &
             ! w
             0.013282683019441681_dp, 0.031461954436631044_dp, 0.05104059411093964_dp, &
             0.07299734987692671_dp, 0.09865267984972562_dp, 0.12980597075173161_dp, &
             0.1689008819771996_dp, 0.21917552422581316_dp, 0.28479635641945156_dp, &
             0.37103591540574876_dp, 0.4846000732932077_dp, 0.6342194960650418_dp, &
             0.8316683182797598_dp, 1.093592784169909_dp, 1.4451639886413101_dp, &
             1.9284445503402172_dp, 2.6251732346689796_dp, 3.7374625254639873_dp, &
             6.066869248120491_dp]

      CASE (598)

         aw(:) = & ! a
            [0.0044677297018304385_dp, 0.02376189784088184_dp, 0.0594114160764658_dp, &
             0.11317093214998303_dp, 0.18789536211406485_dp, 0.28794844749529214_dp, &
             0.41976897974556754_dp, 0.5925893569356319_dp, 0.8192848729010562_dp, &
             1.1173745143468252_dp, 1.5102875368184387_dp, 2.0291096314075596_dp, &
             2.715135152022554_dp, 3.6237951049992803_dp, 4.831173118640435_dp, &
             6.446086093346339_dp, 8.636197950109521_dp, 11.698348017873675_dp, &
             16.33756949580318_dp, &
             ! w
             0.011490630911616305_dp, 0.02725513553317332_dp, 0.04433158603485956_dp, &
             0.06365656069161345_dp, 0.08651012831570848_dp, 0.11465712070877483_dp, &
             0.15050854605641303_dp, 0.19726168532262914_dp, 0.25902858102020215_dp, &
             0.3410272082701835_dp, 0.4499384688954574_dp, 0.5945251772892179_dp, &
             0.7866617733644203_dp, 1.0431535016530118_dp, 1.3893547388990992_dp, &
             1.86745854223555_dp, 2.559101435919116_dp, 3.665503059769784_dp, &
             5.9833141100858835_dp]

      CASE (599)

         aw(:) = & ! a
            [0.003949175377103119_dp, 0.021016882386231287_dp, 0.05260879735865853_dp, &
             0.10039031150584306_dp, 0.1670911065683335_dp, 0.2569161924165607_dp, &
             0.37611307512913306_dp, 0.5336840427318839_dp, 0.7422230579726513_dp, &
             1.0189113226927433_dp, 1.3868011096110702_dp, 1.8766039948207272_dp, &
             2.529299526096628_dp, 3.400119796208777_dp, 4.565105925112592_dp, &
             6.133187318271295_dp, 8.27219699809758_dp, 11.27883963479663_dp, &
             15.856211970486438_dp, &
             ! w
             0.010158393680043698_dp, 0.02412496650581039_dp, 0.039331705782722326_dp, &
             0.05667950869858235_dp, 0.07741342305966409_dp, 0.10326604898933696_dp, &
             0.1366142104363427_dp, 0.1806104453158263_dp, 0.23931023886372965_dp, &
             0.3178792354262361_dp, 0.4229766430140424_dp, 0.5633951382996059_dp, &
             0.751097950652108_dp, 1.0030307849707563_dp, 1.344713789847315_dp, &
             1.8184693928995797_dp, 2.5058806026915006_dp, 3.6074792807362663_dp, &
             5.916010788202849_dp]

      CASE (600)

         aw(:) = & ! a
            [0.0035471729434616347_dp, 0.01888800297150949_dp, 0.04732929776618154_dp, &
             0.09046100392857725_dp, 0.1509061231439061_dp, 0.23273276411601526_dp, &
             0.3420200251314302_dp, 0.4875623001214235_dp, 0.6816944583909494_dp, &
             0.9412849564184215_dp, 1.2890380629860994_dp, 1.7553205830818768_dp, &
             2.3808198111687657_dp, 3.2205776977543294_dp, 4.350584880671707_dp, &
             5.879863728649131_dp, 7.976406435902832_dp, 10.93684621410959_dp, &
             15.462735633741266_dp, &
             ! w
             0.009125498784984032_dp, 0.021696203027985347_dp, 0.035446687289527296_dp, &
             0.05124710107741155_dp, 0.0703116698276112_dp, 0.09434298192192263_dp, &
             0.12568361751443963_dp, 0.1674405452010151_dp, 0.22361398943119276_dp, &
             0.29932128320477897_dp, 0.4012029270053877_dp, 0.5380785623375758_dp, &
             0.7219895571621229_dp, 0.9700058375983625_dp, 1.3077973829622653_dp, &
             1.7778112511735331_dp, 2.4616068500429553_dp, 3.5591669824444745_dp, &
             5.860021619326129_dp]

      CASE (601)

         aw(:) = & ! a
            [0.003225554724122865_dp, 0.017184201050827182_dp, 0.04310121966398131_dp, &
             0.08250161959105298_dp, 0.137915957094343_dp, 0.2132924060260648_dp, &
             0.3145599254731431_dp, 0.4503238031101212_dp, 0.6326802072272908_dp, &
             0.878208923740481_dp, 1.209293959635216_dp, 1.6559849705361138_dp, &
             2.2586984151318545_dp, 3.072295727918567_dp, 4.172712399660881_dp, &
             5.669048301548914_dp, 7.729443854495566_dp, 10.650500818188634_dp, &
             15.132505473025887_dp, &
             ! w
             0.00829907171621874_dp, 0.01975153693123908_dp, 0.03233200976266627_dp, &
             0.04688374618524252_dp, 0.06459350616949455_dp, 0.08713589301190912_dp, &
             0.11681991816309456_dp, 0.15670769662898318_dp, 0.2107469020009805_dp, &
             0.28401083088203033_dp, 0.3831236427983252_dp, 0.5169289349957699_dp, &
             0.6975369065990421_dp, 0.9421279423087208_dp, 1.276507993989929_dp, &
             1.7432428305202894_dp, 2.4238875934879394_dp, 3.5179749485496643_dp, &
             5.812320538203648_dp]

      CASE (602)

         aw(:) = & ! a
            [0.002961864180714168_dp, 0.015786815087024474_dp, 0.039631457288065275_dp, &
             0.07596408434342933_dp, 0.1272340255084237_dp, 0.19728303997467087_dp, &
             0.29190484380803305_dp, 0.41953163457991377_dp, 0.5920391937559307_dp, &
             0.8257407447584977_dp, 1.1427253075563295_dp, 1.5727502568695093_dp, &
             2.15598061389256_dp, 2.947106721803515_dp, 4.022006234464495_dp, &
             5.489844122345034_dp, 7.518898371344817_dp, 10.405765620794869_dp, &
             14.849671080259483_dp, &
             ! w
             0.007621442842353354_dp, 0.01815595740871083_dp, 0.029773414721636585_dp, &
             0.04329325782769944_dp, 0.059877517744280344_dp, 0.08117463392642735_dp, &
             0.1094610878193479_dp, 0.14775561986267893_dp, 0.19995655418437847_dp, &
             0.27109714745190816_dp, 0.3677870191914155_dp, 0.4988908543475857_dp, &
             0.676579712238261_dp, 0.9181328491895517_dp, 1.2494802737705826_dp, &
             1.7133003460929643_dp, 2.391156903481232_dp, 3.4822059278807114_dp, &
             5.770927160681232_dp]

      CASE (603)

         aw(:) = & ! a
            [0.0016808496363439748_dp, 0.008990245536228625_dp, 0.022719025953813983_dp, &
             0.04399824644267399_dp, 0.07478375998620589_dp, 0.11824766620688688_dp, &
             0.17928978549413466_dp, 0.26514963680237874_dp, 0.3861629508899112_dp, &
             0.5567903239126517_dp, 0.7970794175700454_dp, 1.1347440178303774_dp, &
             1.6081340933944712_dp, 2.2706150896082855_dp, 3.1974823735060673_dp, &
             4.498198428702973_dp, 6.342000037109199_dp, 9.025883690750419_dp, &
             13.243550011832678_dp, &
             ! w
             0.004328598642613239_dp, 0.010384126789943299_dp, 0.017257580190711136_dp, &
             0.025620290592801503_dp, 0.036470805511324374_dp, 0.05126353915187415_dp, &
             0.07201501396525314_dp, 0.10141055674991709_dp, 0.1430022075061157_dp, &
             0.20155976382969545_dp, 0.2835890026511507_dp, 0.3980612231238237_dp, &
             0.5575035245976293_dp, 0.7798189698962098_dp, 1.0917817435362196_dp, &
             1.5369312378288846_dp, 2.197147988277921_dp, 3.2696589291009506_dp, &
             5.525504384643781_dp]

      CASE (604)

         aw(:) = & ! a
            [0.0012023276934156263_dp, 0.006446201408636427_dp, 0.016364897278616404_dp, &
             0.03192252431995431_dp, 0.054822603663368594_dp, 0.08787698176768015_dp, &
             0.13547622022660977_dp, 0.2041578642460905_dp, 0.30335004847620206_dp, &
             0.446429285736762_dp, 0.6522422871307547_dp, 0.9472699310767672_dp, &
             1.3687143312558199_dp, 1.9690313692588841_dp, 2.8230182294664274_dp, &
             4.04018536414647_dp, 5.790322365644983_dp, 8.370914401336751_dp, &
             12.4733976940383_dp, &
             ! w
             0.003097974462586997_dp, 0.007467725802530155_dp, 0.01252619138269576_dp, &
             0.01886677597849433_dp, 0.02739413310558101_dp, 0.03943711299491686_dp, &
             0.056838922266154654_dp, 0.08207843716980122_dp, 0.11851038043355511_dp, &
             0.170753769295393_dp, 0.24523539450065154_dp, 0.3509470243329158_dp, &
             0.5005717676750603_dp, 0.7123372377603621_dp, 1.0135106474518483_dp, &
             1.448203987048529_dp, 2.098657942341444_dp, 3.161354951047662_dp, &
             5.400811786336295_dp]

      CASE (605)

         aw(:) = & ! a
            [0.000946710334071774_dp, 0.0050852716789430924_dp, 0.012956849885372646_dp, &
             0.02542043796551092_dp, 0.044017129742518125_dp, 0.07131939341169745_dp, &
             0.11137026084970997_dp, 0.1702248746342488_dp, 0.2566902491099955_dp, &
             0.38340084008456893_dp, 0.5683723645798634_dp, 0.8372148082210225_dp, &
             1.2262949576500524_dp, 1.7873765690459487_dp, 2.594839065860491_dp, &
             3.7581658499459922_dp, 5.447516158593702_dp, 7.96079685841498_dp, &
             11.98818313201349_dp, &
             ! w
             0.0024403793946805316_dp, 0.005904858284596696_dp, 0.009977468741504241_dp, &
             0.015200659502200606_dp, 0.022414145315007575_dp, 0.03285545421008699_dp, &
             0.04824348878059537_dp, 0.07091596819765357_dp, 0.10409503054738367_dp, &
             0.15229253688991193_dp, 0.22186728407930348_dp, 0.3218037042202939_dp, &
             0.46486896417039536_dp, 0.6694974161260359_dp, 0.9632966712338786_dp, &
             1.3908032045867567_dp, 2.034575296781476_dp, 3.090712546128751_dp, &
             5.319610438419082_dp]

      CASE (606)

         aw(:) = & ! a
            [0.0007859972597901745_dp, 0.00422864508174495_dp, 0.010807191180201696_dp, &
             0.021306354928461373_dp, 0.037150467041811346_dp, 0.06073600511400965_dp, &
             0.09584674374420246_dp, 0.14817763607213577_dp, 0.22607314271643042_dp, &
             0.3416133201382743_dp, 0.5121865056933581_dp, 0.7627346252904954_dp, &
             1.1289701593483565_dp, 1.6620985323511748_dp, 2.4361419560409874_dp, &
             3.560528165360248_dp, 5.205685528678884_dp, 7.669879227625732_dp, &
             11.642476628696276_dp, &
             ! w
             0.0020268234806406437_dp, 0.004919747616795269_dp, 0.008364274194348474_dp, &
             0.012865817442588336_dp, 0.01921494067840693_dp, 0.028578322419148434_dp, &
             0.04258033439228271_dp, 0.06345440080080501_dp, 0.09432416333569144_dp, &
             0.13961749853097102_dp, 0.20563341013423364_dp, 0.301338501784395_dp, &
             0.4395500701480455_dp, 0.6388484681139391_dp, 0.9270974528650047_dp, &
             1.34916901046183_dp, 1.9878975548392614_dp, 3.039159900437875_dp, &
             5.260415972624009_dp]

      CASE (607)

         aw(:) = & ! a
            [0.0006749106811276842_dp, 0.0036359671789622104_dp, 0.009317270444530974_dp, &
             0.018447331323020285_dp, 0.03236085062677534_dp, 0.05331691112605414_dp, &
             0.0848951333994518_dp, 0.1325070880425022_dp, 0.20413372020293258_dp, &
             0.31141752683934143_dp, 0.4712470658078653_dp, 0.7080253033514385_dp, &
             1.0569284598192545_dp, 1.5686944224027135_dp, 2.3170340367907776_dp, &
             3.4113076576662587_dp, 5.022150513644707_dp, 7.448135962415542_dp, &
             11.37807004269694_dp, &
             ! w
             0.0017409056747026545_dp, 0.0042373726647584415_dp, 0.007242914212328824_dp, &
             0.011234272431903903_dp, 0.016962755023494452_dp, 0.025537748458784645_dp, &
             0.03850864084104038_dp, 0.058027528348821675_dp, 0.08714017870974422_dp, &
             0.13020508064061637_dp, 0.19346802487478867_dp, 0.2858736433521274_dp, &
             0.4202707496074738_dp, 0.615349485169685_dp, 0.8991767958325522_dp, &
             1.3169010232478595_dp, 1.9515988935832955_dp, 2.999009218436354_dp, &
             5.214350185617482_dp]

      CASE (608)

         aw(:) = & ! a
            [0.0005931923431999074_dp, 0.003199612449303557_dp, 0.008218636383443556_dp, &
             0.016334246861398966_dp, 0.02880927232199247_dp, 0.047791160121336454_dp, &
             0.07669276842752676_dp, 0.12069449660077518_dp, 0.1874807694527077_dp, &
             0.28833554144519097_dp, 0.43973450273256925_dp, 0.665630752192438_dp, &
             1.0007472532129744_dp, 1.4954200773895119_dp, 2.22308434925753_dp, &
             3.293028604961744_dp, 4.876053144425764_dp, 7.27100026846775_dp, &
             11.166264590427467_dp, &
             ! w
             0.0015305359756900181_dp, 0.0037344643826074083_dp, 0.006413946863512824_dp, &
             0.010022571076355022_dp, 0.01527915010348538_dp, 0.023245372856115606_dp, &
             0.035409259518815225_dp, 0.05385698229021926_dp, 0.08157013511548536_dp, &
             0.12284797170435875_dp, 0.18388845376767188_dp, 0.2736127817131254_dp, &
             0.40489005306629033_dp, 0.5964965514724115_dp, 0.8766662842850093_dp, &
             1.2907818526477846_dp, 1.9221351370573638_dp, 2.9663769540121385_dp, &
             5.176933586227603_dp]

      CASE (609)

         aw(:) = & ! a
            [0.0005303640593514309_dp, 0.002863875828720506_dp, 0.007372171384282706_dp, &
             0.014702784436254413_dp, 0.026059084455907586_dp, 0.04349518442903838_dp, &
             0.07028405219235231_dp, 0.11141242414631897_dp, 0.17431616280077114_dp, &
             0.2699773161196801_dp, 0.41452143643123013_dp, 0.631516777062393_dp, &
             0.9552945400406361_dp, 1.4358387761816822_dp, 2.146337308966376_dp, &
             3.1960058415130623_dp, 4.755780070315955_dp, 7.124740121900398_dp, &
             10.990967501145983_dp, &
             ! w
             0.0013687675835145836_dp, 0.003347167664642771_dp, 0.005773805080108536_dp, &
             0.009082995430859022_dp, 0.013965956169906004_dp, 0.021443794869635376_dp, &
             0.032953085265131986_dp, 0.05052499444706983_dp, 0.0770866644446477_dp, &
             0.11688563792535352_dp, 0.17607651378540942_dp, 0.26355683459933454_dp, &
             0.3922087439664238_dp, 0.5808782178139423_dp, 0.8579402313528373_dp, &
             1.2689802114182223_dp, 1.8974831878073657_dp, 2.9390436880013247_dp, &
             5.1456085823147975_dp]

      CASE (610)

         aw(:) = & ! a
            [0.00048043949565518106_dp, 0.0025969131851087024_dp, 0.006698261270896518_dp, &
             0.013401437886575702_dp, 0.023859457437051215_dp, 0.040046704515545414_dp, &
             0.06511639852607326_dp, 0.10388969035025705_dp, 0.16358969167239085_dp, &
             0.25493892371210936_dp, 0.39376004721705454_dp, 0.6032859114740475_dp, &
             0.9175033038061544_dp, 1.3860832943401065_dp, 2.081989581968695_dp, &
             3.1143657390872033_dp, 4.654260889694073_dp, 7.0009678541599_dp, &
             10.842323113643882_dp, &
             ! w
             0.0012402032728556424_dp, 0.0030389525605399796_dp, 0.005263112644710338_dp, &
             0.008330589975932237_dp, 0.012908711049654284_dp, 0.01998346855547739_dp, &
             0.03094747872394875_dp, 0.04778498019481216_dp, 0.07337590416452335_dp, &
             0.11192190570373758_dp, 0.16953797143187854_dp, 0.25509835045157353_dp, &
             0.3814933442739696_dp, 0.5676266444396509_dp, 0.8419945520361988_dp, &
             1.2503609824414232_dp, 1.8763860059088577_dp, 2.915629011144059_dp, &
             5.1187856317997875_dp]

      CASE (611)

         aw(:) = & ! a
            [0.0004397402753084449_dp, 0.002379146979898393_dp, 0.0061479118280827125_dp, &
             0.012336836108433634_dp, 0.02205550437562051_dp, 0.03720904397663541_dp, &
             0.06084651911069953_dp, 0.09764519545293227_dp, 0.1546430661551105_dp, &
             0.24233590782150838_dp, 0.37628025727117576_dp, 0.5794123567757681_dp, &
             0.8854121264367747_dp, 1.3436690831379876_dp, 2.0269420472401176_dp, &
             3.044304019820955_dp, 4.566900745885466_dp, 6.8942170928054125_dp, &
             10.713893661358698_dp, &
             ! w
             0.001135380846526007_dp, 0.002787345925072663_dp, 0.004845270506795661_dp, &
             0.007712839937737775_dp, 0.01203638389600089_dp, 0.018771095935044793_dp, &
             0.029271450212838973_dp, 0.04548093611798933_dp, 0.07023786898195965_dp, &
             0.10770266935820304_dp, 0.16395389873842114_dp, 0.2478431596680922_dp, &
             0.37226548984175384_dp, 0.5561732580829527_dp, 0.8281688204004571_dp, &
             1.2341752039531853_dp, 1.8580124017735813_dp, 2.8952193255576355_dp, &
             5.095413388496554_dp]

      CASE (612)

         aw(:) = & ! a
            [0.0002452565599847984_dp, 0.001336251823050387_dp, 0.003501471540929438_dp, &
             0.007185182036997784_dp, 0.013246418692518382_dp, 0.02318269575948486_dp, &
             0.0394296163815743_dp, 0.06581836011474483_dp, 0.10828972542414658_dp, &
             0.17597420011785742_dp, 0.2827944009101576_dp, 0.44982567053123507_dp, &
             0.7087746360197066_dp, 1.1071629823550355_dp, 1.7163201146111902_dp, &
             2.6447268869307314_dp, 4.064049145958035_dp, 6.275049984375717_dp, &
             9.964551282442011_dp, &
             ! w
             0.0006342268344939126_dp, 0.0015791476671186966_dp, 0.002822483197848424_dp, &
             0.004684406041435337_dp, 0.0076828924255215175_dp, 0.012588202719370847_dp, &
             0.020532509954255537_dp, 0.03321747100526211_dp, 0.05322084663019754_dp, &
             0.08443001798367637_dp, 0.13266627167354036_dp, 0.20659665811833028_dp, &
             0.31909344154293573_dp, 0.48936280251203956_dp, 0.746643309005604_dp, &
             1.1378801459629069_dp, 1.7480036471772586_dp, 2.7726434417237336_dp, &
             4.9552013805472646_dp]

      CASE (613)

         aw(:) = & ! a
            [0.0001741460746723223_dp, 0.0009534786680525147_dp, 0.0025232247863159526_dp, &
             0.005259664108906556_dp, 0.009900685826684632_dp, 0.01774286641177327_dp, &
             0.030920261624485742_dp, 0.052847162031634545_dp, 0.08891488519445367_dp, &
             0.1475536351631822_dp, 0.2418254321034717_dp, 0.39179749092196814_dp, &
             0.6280735734632387_dp, 0.997090807689619_dp, 1.5692989749944635_dp, &
             2.4527462114550977_dp, 3.8193114566860196_dp, 5.970494231428673_dp, &
             9.592943100394793_dp, &
             ! w
             0.0004508258133707264_dp, 0.0011336422409792784_dp, 0.0020660275741542743_dp, &
             0.00352662470860865_dp, 0.005967095404007063_dp, 0.010065762904314173_dp, &
             0.016846712842950907_dp, 0.027888329322930703_dp, 0.045626180291767605_dp, &
             0.07378963156087867_dp, 0.1180410005696795_dp, 0.1869178535193543_dp, &
             0.2932425855458447_dp, 0.45632059507388456_dp, 0.7057115787411917_dp, &
             1.0889288542900775_dp, 1.6915795941607348_dp, 2.709496127735778_dp, &
             4.883067088161423_dp]

      CASE (614)

         aw(:) = & ! a
            [0.00013655809986376583_dp, 0.0007506111348062843_dp, 0.0020021522024140184_dp, &
             0.00422582322624809_dp, 0.008083551134479257_dp, 0.014745206886610122_dp, &
             0.02615513568683237_dp, 0.04546393670455775_dp, 0.07771094306053655_dp, &
             0.1308709458375622_dp, 0.21743750607560622_dp, 0.3568011133504739_dp, &
             0.5788120768616463_dp, 0.929150761248229_dp, 1.4776339845111246_dp, &
             2.3319730133114933_dp, 3.6641596499525497_dp, 5.7762017388518485_dp, &
             9.354729635731525_dp, &
             ! w
             0.0003538233776194803_dp, 0.000896760446405615_dp, 0.0016597826113511904_dp, &
             0.0028949781278375048_dp, 0.005011224904955544_dp, 0.008629133943598814_dp, &
             0.01470470788060809_dp, 0.02473573889045818_dp, 0.04106161575848743_dp, &
             0.06730190835605325_dp, 0.10900497514904041_dp, 0.174610211143539_dp, &
             0.27689173728297317_dp, 0.4352058494881356_dp, 0.6793173395993287_dp, &
             1.0571253666352238_dp, 1.6547207576581278_dp, 2.668131723866405_dp, &
             4.835849410661586_dp]

      CASE (615)

         aw(:) = & ! a
            [0.00011308121399109941_dp, 0.000623634637975123_dp, 0.0016746960030130308_dp, &
             0.0035719327882889505_dp, 0.006923587329983111_dp, 0.012809842189313183_dp, &
             0.023040879793721047_dp, 0.04057974918190759_dp, 0.07021293758054123_dp, &
             0.11958451843089579_dp, 0.20077071735581575_dp, 0.33265963265552606_dp, &
             0.5445354247209374_dp, 0.8815016077516146_dp, 1.4128825664410358_dp, &
             2.2461144930477452_dp, 3.553256200333291_dp, 5.636699231526024_dp, &
             9.183109505873462_dp, &
             ! w
             0.0002932076769236164_dp, 0.0007481120070367528_dp, 0.0014028117638563895_dp, &
             0.002490340581151832_dp, 0.00438889273577708_dp, 0.007678445938318215_dp, &
             0.013266560967431732_dp, 0.02259210160528784_dp, 0.037922658195971584_dp, &
             0.06279446256056287_dp, 0.10266764145907013_dp, 0.1659029476447847_dp, &
             0.26523093116487967_dp, 0.42003727377417416_dp, 0.6602331680011628_dp, &
             1.034006152965516_dp, 1.6278213442595888_dp, 2.637883596163789_dp, &
             4.8013369479657255_dp]

      CASE (616)

         aw(:) = & ! a
            [9.692871428992666e-05_dp, 0.0005361166667958782_dp, 0.0014482287494156359_dp, &
             0.0031172124973499363_dp, 0.006110624576204801_dp, 0.011440671384384876_dp, &
             0.02081585035726263_dp, 0.037056288258413915_dp, 0.06475423314355687_dp, &
             0.11129755258357511_dp, 0.1884364659911477_dp, 0.31466329164292456_dp, &
             0.518812102943515_dp, 0.8455228759939528_dp, 1.363718309371896_dp, &
             2.1806024816750234_dp, 3.4682764293725907_dp, 5.529437384651954_dp, &
             9.050808091542981_dp, &
             ! w
             0.0002514859830440869_dp, 0.0006454345166225561_dp, 0.0012241052339225014_dp, &
             0.002205910494090055_dp, 0.003945601656522417_dp, 0.0069924961499716895_dp, &
             0.01221714691770993_dp, 0.02101246469495393_dp, 0.03558924924552499_dp, &
             0.05941703186696812_dp, 0.09788430469712169_dp, 0.1592864348880946_dp, &
             0.2563149746526521_dp, 0.4083735982934267_dp, 0.6454851278879378_dp, &
             1.016065286906939_dp, 1.606883214799993_dp, 2.614301806930561_dp, &
             4.774439568301108_dp]

      CASE (617)

         aw(:) = & ! a
            [8.508757405840501e-05_dp, 0.0004718588003697698_dp, 0.0012814541772765324_dp, &
             0.002780723343367501_dp, 0.0055049358203099925_dp, 0.010412383380469936_dp, &
             0.019130864807767222_dp, 0.03436651212901931_dp, 0.060555602889510285_dp, &
             0.10487887023776117_dp, 0.1788212031017138_dp, 0.30055061645761594_dp, &
             0.49852969740380165_dp, 0.8170125736689722_dp, 1.3245836332330116_dp, &
             2.1282463931469673_dp, 3.400129263283795_dp, 5.443182007912334_dp, &
             8.944192595254204_dp, &
             ! w
             0.00022088958216943074_dp, 0.0005699037544084404_dp, 0.0010918633813335535_dp, &
             0.0019934561476842214_dp, 0.0036107454977517797_dp, 0.006468811700138278_dp, &
             0.011408586421372484_dp, 0.019785608074966144_dp, 0.03376400181826369_dp, &
             0.05675798125314822_dp, 0.09409596810809624_dp, 0.15401750761816266_dp, &
             0.24917910679867586_dp, 0.3989957294671794_dp, 0.6335791266820889_dp, &
             1.0015325615248678_dp, 1.5898803365209977_dp, 2.5951273723316794_dp, &
             4.75257476360093_dp]

      CASE (618)

         aw(:) = & ! a
            [7.600822373816234e-05_dp, 0.00042251989138572483_dp, 0.0011530576109491102_dp, &
             0.002520537010995768_dp, 0.00503375307570398_dp, 0.009606823836148306_dp, &
             0.017801365585411767_dp, 0.03222959135625519_dp, 0.05719853610097257_dp, &
             0.09971631339025547_dp, 0.17104546310578572_dp, 0.28908070363547056_dp, &
             0.48196968203540597_dp, 0.7936371205258098_dp, 1.2923758135249086_dp, &
             2.0850128708834323_dp, 3.3436945551800945_dp, 5.371584770055354_dp, &
             8.855539374950604_dp, &
             ! w
             0.00019742194726119186_dp, 0.0005118116348137076_dp, 0.0009896106633899505_dp, &
             0.0018278066285044969_dp, 0.00334710232508817_dp, 0.0060527664083277226_dp, &
             0.010761229345322946_dp, 0.018796720675713002_dp, 0.03228394099920342_dp, &
             0.05459003681519023_dp, 0.09099185987169944_dp, 0.14968034692654478_dp, &
             0.24328029632329165_dp, 0.3912137314868303_dp, 0.6236655095108314_dp, &
             0.9893972810105117_dp, 1.5756526218352265_dp, 2.5790649790605737_dp, &
             4.734262357502428_dp]

      CASE (619)

         aw(:) = & ! a
            [6.880928317894288e-05_dp, 0.0003833503298649437_dp, 0.0010508770061426042_dp, &
             0.002312651921688992_dp, 0.004655223586962406_dp, 0.008955622505370815_dp, &
             0.01671983269130878_dp, 0.030480787690744485_dp, 0.05443587925776594_dp, &
             0.09544606234459735_dp, 0.16458342414963958_dp, 0.2795074718395376_dp, &
             0.46809341944311916_dp, 0.7739793587128425_dp, 1.2652023159621386_dp, &
             2.0484320743641677_dp, 3.2958263420011202_dp, 5.310734231725445_dp, &
             8.780079439268203_dp, &
             ! w
             0.0001788093141587951_dp, 0.0004656227515863443_dp, 0.0009079146920660995_dp, &
             0.0016944583618168278_dp, 0.003133034563393297_dp, 0.005712304059638192_dp, &
             0.01022793028426971_dp, 0.017977327533170728_dp, 0.03105120439534169_dp, &
             0.05277587836847496_dp, 0.08838313096087348_dp, 0.1460209186406147_dp, &
             0.23828514680469007_dp, 0.38460209116681837_dp, 0.6152181442361673_dp, &
             0.9790314957725786_dp, 1.5634775601762414_dp, 2.5653069289627086_dp, &
             4.718579708657027_dp]

      CASE (620)

         aw(:) = & ! a
            [6.295112287955062e-05_dp, 0.00035143926303216196_dp, 0.0009674449077851094_dp, &
             0.002142288054406022_dp, 0.004343464409125552_dp, 0.00841626764552616_dp, &
             0.01581901191689761_dp, 0.029016435920813478_dp, 0.052111201960826185_dp, &
             0.09183658041303387_dp, 0.1590987271080321_dp, 0.27135137173711127_dp, &
             0.4562303359239874_dp, 0.757120560817237_dp, 1.2418316253697426_dp, &
             2.016891435981481_dp, 3.2544646697665223_dp, 5.258063120748616_dp, &
             8.714676931404934_dp, &
             ! w
             0.00016365924209062137_dp, 0.00042794058548162057_dp, 0.0008409665962780909_dp, &
             0.0015844273205295669_dp, 0.0029550333285744625_dp, 0.0054272463089888685_dp, &
             0.00977879306129542_dp, 0.017283724934061266_dp, 0.030002969237911285_dp, &
             0.05122688543495452_dp, 0.08614730779022037_dp, 0.14287372210428936_dp, &
             0.23397553373032692_dp, 0.37888133830939713_dp, 0.6078902873632925_dp, &
             0.9700201717118178_dp, 1.552876616437419_dp, 2.5533176914480835_dp, &
             4.704915228054113_dp]

      CASE (621)

         aw(:) = & ! a
            [3.50923765549347e-05_dp, 0.00019906223521234002_dp, 0.0005658227010916817_dp, &
             0.0013112600862214884_dp, 0.002795691991344798_dp, 0.005686375699459768_dp, &
             0.011172361573755608_dp, 0.021327764214291525_dp, 0.039704519451766944_dp, &
             0.07228303182843565_dp, 0.12897767034865545_dp, 0.2259952894860738_dp, &
             0.3894999842443624_dp, 0.6612942621298318_dp, 1.107733975533177_dp, &
             1.834399485287558_dp, 3.0134332599269062_dp, 4.949343749742555_dp, &
             8.329664463657704_dp, &
             ! w
             9.154515676928922e-05_dp, 0.00024711067481745976_dp, 0.0005144625693660486_dp, &
             0.0010345563254320219_dp, 0.0020420553280663415_dp, 0.003931618264674404_dp, &
             0.007376458095215753_dp, 0.013510948790764584_dp, 0.024214894207119277_dp, &
             0.042556260434951816_dp, 0.07347469127524059_dp, 0.12482920693843112_dp, &
             0.20900384152891088_dp, 0.3454123871036579_dp, 0.5646508697430568_dp, &
             0.9164636544285858_dp, 1.4895357525021196_dp, 2.481478609821539_dp, &
             4.623073141582258_dp]

      CASE (622)

         aw(:) = & ! a
            [2.4961643626420166e-05_dp, 0.00014325538880899323_dp, 0.0004166190459596854_dp, &
             0.0009953067753841886_dp, 0.0021897853867406682_dp, 0.0045846919129257195_dp, &
             0.009242375349019675_dp, 0.018049263792974266_dp, 0.03428720929349682_dp, &
             0.0635603099919389_dp, 0.11527813392006443_dp, 0.2050013660740282_dp, &
             0.35811690427455783_dp, 0.6155736203439862_dp, 1.0429211305719979_dp, &
             1.7451864079195214_dp, 2.8944536047876546_dp, 4.795753346588278_dp, &
             8.136995520351189_dp, &
             ! w
             6.527874600971004e-05_dp, 0.0001803101026628384_dp, 0.0003903773935193581_dp, &
             0.0008169568416950697_dp, 0.0016661613185203136_dp, 0.0032951267881748455_dp, &
             0.006325424753489499_dp, 0.011820345517797207_dp, 0.021565549275927652_dp, &
             0.03851089525954203_dp, 0.06745861076716163_dp, 0.11612613576464842_dp, &
             0.19678430892254414_dp, 0.3288189265297978_dp, 0.5429632598598246_dp, &
             0.8893385370883415_dp, 1.4572221437311055_dp, 2.4446873321018328_dp, &
             4.581180141661304_dp]

      CASE (623)

         aw(:) = & ! a
            [1.9617012655046446e-05_dp, 0.00011366648314194275_dp, 0.0003366953613831632_dp, &
             0.0008232767446361525_dp, 0.0018534020117487947_dp, 0.003961157009342624_dp, &
             0.008130525125759252_dp, 0.016130337156362747_dp, 0.03107118199767957_dp, &
             0.058315897326562505_dp, 0.10694701127707384_dp, 0.1921022453393694_dp, &
             0.33865429849344675_dp, 0.5869804811235751_dp, 1.0020822693327578_dp, &
             1.6885999894712855_dp, 2.8185608307430305_dp, 4.69733858956372_dp, &
             8.013124704005667_dp, &
             ! w
             5.140590766964981e-05_dp, 0.0001446769568848332_dp, 0.0003228319226838376_dp, &
             0.0006952849249727678_dp, 0.001450829346359788_dp, 0.0029232870179219717_dp, &
             0.005701277073926892_dp, 0.010802074024944673_dp, 0.01994968391747334_dp, &
             0.03601567362546221_dp, 0.06370987485784774_dp, 0.11065265677969342_dp, &
             0.18903424695638463_dp, 0.3182142226840104_dp, 0.5290090154686774_dp, &
             0.8717862474618457_dp, 1.4362235296767876_dp, 2.4207239620574272_dp, &
             4.553900677754615_dp]

      CASE (624)

         aw(:) = & ! a
            [1.6281405779463337e-05_dp, 9.512659178743429e-05_dp, 0.0002861985923491392_dp, &
             0.0007131714186454165_dp, 0.0016348973811832423_dp, 0.0035503274738605136_dp, &
             0.007388507606706449_dp, 0.014835060986634086_dp, 0.02887837687796663_dp, &
             0.05470782386444532_dp, 0.10116907248925192_dp, 0.18309132265540146_dp, &
             0.3249694333826912_dp, 0.5667572812430093_dp, 0.9730460419880588_dp, &
             1.6481813652257802_dp, 2.7641390095504597_dp, 4.626543550118046_dp, &
             7.923809500369194_dp, &
             ! w
             4.2740140642576924e-05_dp, 0.00012224141038606366_dp, 0.0002796057724806588_dp, &
             0.0006158240808728694_dp, 0.0013077327168002295_dp, 0.0026727197246771343_dp, &
             0.005275780482944472_dp, 0.010100895050296082_dp, 0.018827103769685136_dp, &
             0.034268394430410846_dp, 0.06106598978688563_dp, 0.10676724019893825_dp, &
             0.18350025665191808_dp, 0.31060142549762465_dp, 0.5189443281390518_dp, &
             0.8590760280967278_dp, 1.4209723957768996_dp, 2.4032913239127405_dp, &
             4.534058816117401_dp]

      CASE (625)

         aw(:) = & ! a
            [1.3987003772537218e-05_dp, 8.233126581495457e-05_dp, 0.0002510997106105938_dp, &
             0.0006358080815767015_dp, 0.0014795203840787647_dp, 0.0032548818712141514_dp, &
             0.006849514856610564_dp, 0.013885852841736962_dp, 0.027258888675702454_dp, &
             0.05202463400151399_dp, 0.09684566716970727_dp, 0.1763114107819294_dp, &
             0.31462144196703923_dp, 0.5513966927058243_dp, 0.9509032322906138_dp, &
             1.6172501275505806_dp, 2.7223672011166866_dp, 4.572074103144319_dp, &
             7.8549690956660685_dp, &
             ! w
             3.677497478363272e-05_dp, 0.00010669382102084559_dp, 0.0002492351278510938_dp, &
             0.0005590759870087453_dp, 0.001204146325198469_dp, 0.0024893763074105933_dp, &
             0.0049616395468219865_dp, 0.009579204486871943_dp, 0.01798617737629224_dp, &
             0.03295153021252843_dp, 0.059062460905788955_dp, 0.10380827298510432_dp, &
             0.17926684003424143_dp, 0.3047541003732394_dp, 0.5111859309069675_dp, &
             0.8492486338184633_dp, 1.4091536103577953_dp, 2.3897652210459017_dp, &
             4.5186650835191235_dp]

      CASE (626)

         aw(:) = & ! a
            [1.2304945119076357e-05_dp, 7.292351774261556e-05_dp, 0.0002251311557808773_dp, &
             0.0005780322481414275_dp, 0.0013623090196823618_dp, 0.003029923325619632_dp, &
             0.006435733107169324_dp, 0.013151901958734974_dp, 0.02599874033747604_dp, &
             0.04992512436887561_dp, 0.09344589123353757_dp, 0.17095614759273242_dp, &
             0.30641511798921023_dp, 0.53917140896192_dp, 0.9332235610030336_dp, &
             1.592483956251792_dp, 2.6888412505984496_dp, 4.528273242726516_dp, &
             7.799533805537941_dp, &
             ! w
             3.239903959562625e-05_dp, 9.522129762770216e-05_dp, 0.00022655377517401506_dp, &
             0.0005161132502868733_dp, 0.0011248542380179888_dp, 0.0023478039849185945_dp, &
             0.00471730288797133_dp, 0.00917089202986637_dp, 0.017324380030818508_dp, &
             0.031910090895016896_dp, 0.05747098937553483_dp, 0.10144850017198352_dp, &
             0.17587851178903252_dp, 0.30005880675241453_dp, 0.5049381379035155_dp, &
             0.8413154749308983_dp, 1.3995955064041223_dp, 2.3788154334443536_dp, &
             4.506204424483072_dp]

      CASE (627)

         aw(:) = & ! a
            [1.1014941760419353e-05_dp, 6.568982563375485e-05_dp, 0.00020505092676584668_dp, &
             0.0005329873492760394_dp, 0.0012701282168454617_dp, 0.0028515983747078757_dp, &
             0.006105447788126198_dp, 0.012562509991457547_dp, 0.024981431615936404_dp, &
             0.04822228515598638_dp, 0.09067699974014744_dp, 0.16657845331040444_dp, &
             0.2996844771289189_dp, 0.5291145835290616_dp, 0.9186411376260856_dp, &
             1.572008855111889_dp, 2.6610692537616547_dp, 4.491932171750274_dp, &
             7.753486004679083_dp, &
             ! w
             2.904116489237036e-05_dp, 8.637152633339092e-05_dp, 0.0002088694546870463_dp, &
             0.0004822216289545002_dp, 0.0010617214344441838_dp, 0.002234256469019856_dp, &
             0.004520138659835224_dp, 0.008839680889070757_dp, 0.016785077411169892_dp, &
             0.03105794313145522_dp, 0.056164005977826746_dp, 0.0995041412720214_dp, &
             0.173078303101407_dp, 0.29616801374363083_dp, 0.49974849055211_dp, &
             0.8347126288039628_dp, 1.391628152183219_dp, 2.369680397816382_dp, &
             4.4958095643032445_dp]

      CASE (628)

         aw(:) = & ! a
            [9.991798288158873e-06_dp, 5.993908357234806e-05_dp, 0.0001890049752569889_dp, &
             0.0004967255233517953_dp, 0.0011953526020508204_dp, 0.0027059451477993786_dp, &
             0.005834057234813996_dp, 0.012075697108730042_dp, 0.024137363623983686_dp, &
             0.04680378161915906_dp, 0.08836227219649267_dp, 0.16290722200140526_dp, &
             0.29402400765008296_dp, 0.5206353058786354_dp, 0.9063183425940465_dp, &
             1.5546721701240949_dp, 2.6375145679494887_dp, 4.46106808455522_dp, &
             7.714339299486175_dp, &
             ! w
             2.637658597963895e-05_dp, 7.931535280897783e-05_dp, 0.0001946322776046633_dp, &
             0.0004546559368349872_dp, 0.0010099612837769826_dp, 0.002140577104075753_dp, &
             0.004356622739783422_dp, 0.008563760318368138_dp, 0.016334034455533317_dp, &
             0.030342764510610694_dp, 0.05506366820534725_dp, 0.09786258663322465_dp, &
             0.1707081570466381_dp, 0.29286721601501975_dp, 0.49533685039076136_dp, &
             0.8290900288497042_dp, 1.384834863856846_dp, 2.361885958186954_dp, &
             4.486940642421327_dp]

      CASE (629)

         aw(:) = & ! a
            [9.158906924641075e-06_dp, 5.524761971790829e-05_dp, 0.00017585244309199227_dp, &
             0.000466802761859249_dp, 0.0011332274091289303_dp, 0.002584195953243527_dp, &
             0.005606011776588414_dp, 0.011664774554501928_dp, 0.02342205633786763_dp, &
             0.0455974819206607_dp, 0.08638774823864112_dp, 0.1597669501965131_dp, &
             0.2891702749993733_dp, 0.5133484804696632_dp, 0.8957077457524608_dp, &
             1.5397186886864043_dp, 2.6171682373404486_dp, 4.434376819763486_dp, &
             7.680456182068045_dp, &
             ! w
             2.4206493619596893e-05_dp, 7.354335277694569e-05_dp, 0.00018288288517215708_dp, &
             0.0004316998316850065_dp, 0.0009665545131583606_dp, 0.002061584008863792_dp, &
             0.004218111510571009_dp, 0.008329116753254096_dp, 0.015949150868050394_dp, &
             0.029730634518310185_dp, 0.05411931553202164_dp, 0.09645029113769904_dp, &
             0.16866451551898268_dp, 0.2900154622749679_dp, 0.49151866551395207_dp, &
             0.8242165544949457_dp, 1.378940088367101_dp, 2.3551182694931776_dp, &
             4.47924034669797_dp]

      CASE (630)

         aw(:) = & ! a
            [5.19029150262673e-06_dp, 3.2722110439752425e-05_dp, 0.00011160878348221042_dp, &
             0.0003172077653535399_dp, 0.0008154724337136965_dp, 0.0019488844868831275_dp, &
             0.0043954064518988666_dp, 0.009450834921083295_dp, 0.01951828044608681_dp, &
             0.038939339737308114_dp, 0.07537985113390941_dp, 0.14210327821060348_dp, &
             0.26164989903231645_dp, 0.47173646049880735_dp, 0.8347284889062151_dp, &
             1.4533004419994764_dp, 2.4990264931021193_dp, 4.278804468598045_dp, &
             7.4824156472824015_dp, &
             ! w
             1.3850204080335596e-05_dp, 4.5557645282158684e-05_dp, 0.0001241114216171513_dp, &
             0.00031337765278543686_dp, 0.0007376986503677276_dp, 0.0016376059312250481_dp, &
             0.003463555865930534_dp, 0.0070344574635327955_dp, 0.01380170555938623_dp, &
             0.026281356728176226_dp, 0.048750784417235085_dp, 0.08835746236989919_dp, &
             0.15686948805103113_dp, 0.27344954854889764_dp, 0.4692113290431111_dp, &
             0.7956056213247642_dp, 1.344206602536058_dp, 2.315160397334174_dp, &
             4.433781938609984_dp]

      CASE (631)

         aw(:) = & ! a
            [3.7405761415436863e-06_dp, 2.4384158496268192e-05_dp, 8.710652140685235e-05_dp, &
             0.0002579770260064795_dp, 0.0006852423353444291_dp, 0.0016807913247481282_dp, &
             0.0038718856995353275_dp, 0.008473358952434904_dp, 0.017763749251261285_dp, &
             0.03590018434265326_dp, 0.07028646100375693_dp, 0.13383136472943855_dp, &
             0.24862352930097553_dp, 0.45185156979719143_dp, 0.8053417121633903_dp, &
             1.4113461669566514_dp, 2.441312380852481_dp, 4.202425374639944_dp, &
             7.3848320020337805_dp, &
             ! w
             1.0057653987951491e-05_dp, 3.501859257618236e-05_dp, 0.00010081055898269805_dp, &
             0.0002643441865926919_dp, 0.000639754606065103_dp, 0.0014515420351936495_dp, &
             0.0031254935024120313_dp, 0.006444122035922192_dp, 0.012807489157265395_dp, &
             0.024662920705221372_dp, 0.04620174324194651_dp, 0.08447389137105397_dp, &
             0.15115508895524626_dp, 0.2653548805492421_dp, 0.4582286607218884_dp, &
             0.7814294855940629_dp, 1.3269138504295317_dp, 2.295213172808117_dp, &
             4.411092021398197_dp]

      CASE (632)

         aw(:) = & ! a
            [2.9729444908896258e-06_dp, 1.9928118363210715e-05_dp, 7.37343915613315e-05_dp, &
             0.0002248669506300079_dp, 0.000610908212331651_dp, 0.0015251186984546852_dp, &
             0.003563559012376032_dp, 0.00789079359717085_dp, 0.016707415166104307_dp, &
             0.03405432448901093_dp, 0.06716915238420072_dp, 0.12873439089024988_dp, &
             0.24054873851928732_dp, 0.43945953548812156_dp, 0.7869417494548696_dp, &
             1.3849691549002294_dp, 2.4049007974769654_dp, 4.1541044218279675_dp, &
             7.3229712784239736_dp, &
             ! w
             8.04642090027095e-06_dp, 2.9316225435898897e-05_dp, 8.77703165180533e-05_dp, &
             0.00023617949424319337_dp, 0.0005824427677241877_dp, 0.0013410842375099245_dp, &
             0.002922408580842406_dp, 0.006085917423373416_dp, 0.012198984994028331_dp, &
             0.023664860646291697_dp, 0.04461927021546603_dp, 0.08204853923114737_dp, &
             0.14756727473009365_dp, 0.2602482845261011_dp, 0.4512709038073349_dp, &
             0.7724165959426755_dp, 1.3158898250013635_dp, 2.2824777331994013_dp, &
             4.396606424953292_dp]

      CASE (633)

         aw(:) = & ! a
            [2.4924753233952614e-06_dp, 1.7118437447398564e-05_dp, 6.516270735243685e-05_dp, &
             0.0002032636733439574_dp, 0.0005616807883809547_dp, 0.001420778879204668_dp, &
             0.0033548614120475057_dp, 0.007493225943974826_dp, 0.015981492754211193_dp, &
             0.03277820330635238_dp, 0.06500272830332274_dp, 0.1251758373959574_dp, &
             0.23488816490635614_dp, 0.4307410628645935_dp, 0.7739549761250911_dp, &
             1.3663002690697095_dp, 2.3790690886567667_dp, 4.119759472952184_dp, &
             7.278942730043317_dp, &
             ! w
             6.786237192393222e-06_dp, 2.5684800488837132e-05_dp, 7.925374861155672e-05_dp, &
             0.0002174473475364847_dp, 0.0005438316508360725_dp, 0.0012659212987209333_dp, &
             0.0027830814293904353_dp, 0.005838471763570109_dp, 0.011776140410452742_dp, &
             0.022967732148922337_dp, 0.04350890268529781_dp, 0.08033985421409949_dp, &
             0.14503045836283687_dp, 0.2566258897659328_dp, 0.44632127494860924_dp, &
             0.765989503933193_dp, 1.3080142097696057_dp, 2.273370132073467_dp, &
             4.386247620457227_dp]

      CASE (634)

         aw(:) = & ! a
            [2.1612327950791364e-06_dp, 1.5169363683508745e-05_dp, 5.91347369930944e-05_dp, &
             0.00018785723895442617_dp, 0.0005261696532667618_dp, 0.0013448198586290755_dp, &
             0.0032017971578935465_dp, 0.007199834777418629_dp, 0.01544298272743366_dp, &
             0.03182728293261711_dp, 0.06338207141781795_dp, 0.1225046143457833_dp, &
             0.2306261667142152_dp, 0.42415903724317106_dp, 0.7641273078715459_dp, &
             1.3521434774519288_dp, 2.3594464670504784_dp, 4.093633630298469_dp, &
             7.245416838675533_dp, &
             ! w
             5.9167713670179756e-06_dp, 2.3144418341859984e-05_dp, 7.31751753839496e-05_dp, &
             0.0002038914059757577_dp, 0.0005156167035380022_dp, 0.0012105802201424465_dp, &
             0.00267986390463754_dp, 0.005654206800260613_dp, 0.011459864133867672_dp, &
             0.022444287315800417_dp, 0.04267234261190277_dp, 0.07904863376467589_dp, &
             0.14310826991559672_dp, 0.2538745345332626_dp, 0.4425538576771599_dp, &
             0.7610887472472486_dp, 1.302000767591705_dp, 2.2664106843402982_dp, &
             4.378332286111564_dp]

      CASE (635)

         aw(:) = & ! a
            [1.9179632828752384e-06_dp, 1.3730193938843923e-05_dp, 5.4631441731861836e-05_dp, &
             0.00017621433995165562_dp, 0.0004990840711641122_dp, 0.0012864580023894188_dp, &
             0.003083495603026411_dp, 0.0069719645187613435_dp, 0.015023004055443379_dp, &
             0.03108303893208337_dp, 0.06210974705947611_dp, 0.12040186186446464_dp, &
             0.22726317851473515_dp, 0.418954438008914_dp, 0.7563418097709473_dp, &
             1.3409102514006346_dp, 2.3438548644634984_dp, 4.072852111067282_dp, &
             7.218727924314732_dp, &
             ! w
             5.277843010683874e-06_dp, 2.1254847802264512e-05_dp, 6.857828797695864e-05_dp, &
             0.00019352596876269883_dp, 0.0004938745719860908_dp, 0.001167678403044698_dp, &
             0.002599455524566683_dp, 0.005510072785290523_dp, 0.011211603043860392_dp, &
             0.022032160457047822_dp, 0.04201193040081539_dp, 0.07802688108298363_dp, &
             0.14158401095078366_dp, 0.2516886428425565_dp, 0.43955574677944875_dp, &
             0.7571832383822327_dp, 1.2972034166993252_dp, 2.2608552926107777_dp, &
             4.3720139739413115_dp]

      CASE (636)

         aw(:) = & ! a
            [1.7311390122961277e-06_dp, 1.2619634650819772e-05_dp, 5.1120733626948977e-05_dp, &
             0.00016704890612187693_dp, 0.00047759721241367277_dp, 0.0012398790190682977_dp, &
             0.002988617608873906_dp, 0.006788476706331566_dp, 0.014683678782816192_dp, &
             0.030479976288756958_dp, 0.06107618593240521_dp, 0.11868994612099557_dp, &
             0.2245199450334367_dp, 0.41470167853917583_dp, 0.7499705090228661_dp, &
             1.3317053521940017_dp, 2.3310643319171858_dp, 4.055788930279161_dp, &
             7.19680024116619_dp, &
             ! w
             4.786935585102578e-06_dp, 1.9787215567739266e-05_dp, 6.49573720997415e-05_dp, &
             0.00018528644237853948_dp, 0.00047648091066119125_dp, 0.0011331870631449258_dp, &
             0.002534550576940787_dp, 0.005393338286523516_dp, 0.011009959727561592_dp, &
             0.021696590493243424_dp, 0.04147302510395615_dp, 0.07719150719698725_dp, &
             0.14033564657626993_dp, 0.24989565071474518_dp, 0.43709319611338593_dp, &
             0.7539717088272132_dp, 1.2932550884256893_dp, 2.256280836934653_dp, &
             4.366811384325731_dp]

      CASE (637)

         aw(:) = & ! a
            [1.5828081120538354e-06_dp, 1.1734060579971623e-05_dp, 4.829576361009475e-05_dp, &
             0.00015961151744334153_dp, 0.0004600467550791127_dp, 0.001201638045943823_dp, &
             0.0029104029744172143_dp, 0.006636702317913455_dp, 0.014402201982289036_dp, &
             0.029978506221926434_dp, 0.060214926580446464_dp, 0.1172607849475538_dp, &
             0.22222608071465083_dp, 0.41114044682633955_dp, 0.7446284577977828_dp, &
             1.3239789399342268_dp, 2.3203182187495126_dp, 4.0414424849479325_dp, &
             7.178353902534158_dp, &
             ! w
             4.397030575805541e-06_dp, 1.861003146690943e-05_dp, 6.201758328210884e-05_dp, &
             0.0001785450270747963_dp, 0.0004621727487643162_dp, 0.0011046957809974095_dp, &
             0.0024807549461468943_dp, 0.005296311129803276_dp, 0.01084195506502504_dp, &
             0.021416418908680006_dp, 0.04102226473662316_dp, 0.07649163972735575_dp, &
             0.13928827202045468_dp, 0.2483894008178307_dp, 0.43502212498019316_dp, &
             0.7512681427707796_dp, 1.289928844996622_dp, 2.2524255334444856_dp, &
             4.362426744267351_dp]

      CASE (638)

         aw(:) = & ! a
            [1.461966158609154e-06_dp, 1.1009731789564298e-05_dp, 4.5966283009810174e-05_dp, &
             0.0001534333044003547_dp, 0.0004453845157919598_dp, 0.0011695487715724902_dp, &
             0.0028445380837141933_dp, 0.006508520878403069_dp, 0.014163900032884008_dp, &
             0.02955307040462143_dp, 0.05948293544966251_dp, 0.11604421244170511_dp, &
             0.22027071499953227_dp, 0.40810100520469544_dp, 0.7400641860587582_dp, &
             1.3173712465885603_dp, 2.311120758915002_dp, 4.029155795065795_dp, &
             7.1625487254662_dp, &
             ! w
             4.079288005747578e-06_dp, 1.7642043099409914e-05_dp, 5.9574366010494085e-05_dp, &
             0.00017290493733516016_dp, 0.0004501462629445721_dp, 0.0010806618462506246_dp, &
             0.0024352436170679346_dp, 0.005214027218554562_dp, 0.01069918530764597_dp, &
             0.021177905946895486_dp, 0.04063792846934924_dp, 0.0758940798066644_dp, &
             0.1383929029311083_dp, 0.2471003408735595_dp, 0.4332479762560413_dp, &
             0.7489502885386315_dp, 1.2870753808887139_dp, 2.24911704847337_dp, &
             4.3586640362079185_dp]

      CASE (639)

         aw(:) = & ! a
            [8.859033513429025e-07_dp, 7.508053265009177e-06_dp, 3.4393231054237084e-05_dp, &
             0.00012201111318451792_dp, 0.00036947521089982025_dp, 0.001001122074039284_dp, &
             0.0024950331420367337_dp, 0.005822213888380374_dp, 0.012878358226281203_dp, &
             0.027243245295628875_dp, 0.055486604657474896_dp, 0.10936998327513861_dp, &
             0.20949745442067044_dp, 0.391291572322708_dp, 0.7147376255535827_dp, &
             1.2805996944375402_dp, 2.2598121231113746_dp, 3.9604803768674763_dp, &
             7.074082625656915_dp, &
             ! w
             2.563433245112962e-06_dp, 1.2875251273252788e-05_dp, 4.712950891677713e-05_dp, &
             0.000143579483008374_dp, 0.00038670724802844064_dp, 0.0009524678071029245_dp, &
             0.0021903022479781557_dp, 0.004767854649246107_dp, 0.009920104105446772_dp, &
             0.019869199452744225_dp, 0.03851893740706719_dp, 0.07258546456730267_dp, &
             0.1334165417375979_dp, 0.23991165744593146_dp, 0.42332462958913447_dp, &
             0.735953170908337_dp, 1.271044215651141_dp, 2.2305093221643633_dp, &
             4.337502149389349_dp]

      CASE (640)

         aw(:) = & ! a
            [6.771039542043292e-07_dp, 6.208318576240484e-06_dp, 2.9915017861882194e-05_dp, &
             0.0001094455076719625_dp, 0.0003383807820201672_dp, 0.0009308638879072167_dp, &
             0.00234713732657807_dp, 0.005528400690381031_dp, 0.012322654695363946_dp, &
             0.02623653999163814_dp, 0.05373250263226416_dp, 0.10642239085323879_dp, &
             0.20471380756311588_dp, 0.3837920823769138_dp, 0.7033908654656422_dp, &
             1.2640654304722831_dp, 2.2366706297925742_dp, 3.9294306889669364_dp, &
             7.034014847750328_dp, &
             ! w
             2.013522758229603e-06_dp, 1.1052967741072955e-05_dp, 4.21425407948508e-05_dp, &
             0.0001315016317352101_dp, 0.0003600774846699093_dp, 0.0008978682399497888_dp, &
             0.0020847573739226377_dp, 0.004573742829347724_dp, 0.009578393916507038_dp, &
             0.01929117118474441_dp, 0.03757731005632088_dp, 0.07110729012854762_dp, &
             0.1311826603475626_dp, 0.2366709672585805_dp, 0.4188344496775948_dp, &
             0.7300536173037036_dp, 1.2637500246407039_dp, 2.2220312818465495_dp, &
             4.32786062713236_dp]

      CASE (641)

         aw(:) = & ! a
            [5.68819045552737e-07_dp, 5.52351190658554e-06_dp, 2.749772742682204e-05_dp, &
             0.00010254187781768126_dp, 0.0003210816966437369_dp, 0.0008914085768859882_dp, &
             0.002263472522695641_dp, 0.005361204135596198_dp, 0.012004872256967385_dp, &
             0.025658455071349475_dp, 0.052721638878488616_dp, 0.1047184624992845_dp, &
             0.2019409791059891_dp, 0.37943461536749024_dp, 0.6967841385672375_dp, &
             1.2544206868495094_dp, 2.2231510458066617_dp, 3.9112689194008645_dp, &
             7.010557546281332_dp, &
             ! w
             1.728185389568122e-06_dp, 1.0075359372946279e-05_dp, 3.93997461739934e-05_dp, &
             0.00012476452050813863_dp, 0.0003450770099333706_dp, 0.0008668826924992558_dp, &
             0.0020245037333139794_dp, 0.004462386003686064_dp, 0.0093815566476513_dp, &
             0.01895703067376839_dp, 0.03703131170034896_dp, 0.07024785897794121_dp, &
             0.1298807355865842_dp, 0.23477824454275165_dp, 0.4162070570365943_dp, &
             0.72659608006411_dp, 1.2594699836156429_dp, 2.2170531791959536_dp, &
             4.322199408665857_dp]

      CASE (642)

         aw(:) = & ! a
            [5.036035637541466e-07_dp, 5.106227651566605e-06_dp, 2.6000730005218392e-05_dp, &
             9.821804965444753e-05_dp, 0.00031016127134223345_dp, 0.0008663553783451762_dp, &
             0.0022101049299463457_dp, 0.0052541616950661585_dp, 0.011800802792244033_dp, &
             0.025286274265090457_dp, 0.052069394216198474_dp, 0.10361692663814434_dp, &
             0.20014543357163195_dp, 0.37660878507469897_dp, 0.6924941226232905_dp, &
             1.2481509488714801_dp, 2.214354116476969_dp, 3.8994425712422096_dp, &
             6.9952746811971185_dp, &
             ! w
             1.5562488693691335e-06_dp, 9.472207387093138e-06_dp, 3.768077541188872e-05_dp, &
             0.00012050482291113932_dp, 0.00033553438555474543_dp, 0.0008470794934041222_dp, &
             0.001985852897771394_dp, 0.004390737661777333_dp, 0.009254586794852051_dp, &
             0.018741023985196545_dp, 0.03667768083340908_dp, 0.06969029922765917_dp, &
             0.12903486163065656_dp, 0.2335469142232332_dp, 0.4144958151279039_dp, &
             0.724341970728508_dp, 1.2566775844427207_dp, 2.2138039788911006_dp, &
             4.318504359669525_dp]

      CASE (643)

         aw(:) = & ! a
            [4.615328797105288e-07_dp, 4.834636405510112e-06_dp, 2.5015241173426133e-05_dp, &
             9.534960546035415e-05_dp, 0.0003028778584951889_dp, 0.000849580124538958_dp, &
             0.002174261355068368_dp, 0.005182091136243459_dp, 0.011663125458494919_dp, &
             0.02503474830437038_dp, 0.05162794837239468_dp, 0.1028704450055479_dp, &
             0.19892728185582043_dp, 0.37468977637068707_dp, 0.6895782831364758_dp, &
             1.2438863490740313_dp, 2.208366792647559_dp, 3.891389366282853_dp, &
             6.984864005542914_dp, &
             ! w
             1.4452745728353393e-06_dp, 9.07612623336141e-06_dp, 3.653991079215376e-05_dp, &
             0.00011766083620374121_dp, 0.0003291369586306074_dp, 0.0008337619089541059_dp, &
             0.0019597960118812523_dp, 0.004342337324110828_dp, 0.00916866955004171_dp, &
             0.01859464525062751_dp, 0.03643773729826663_dp, 0.06931156728942266_dp, &
             0.12845972266851516_dp, 0.23270895971698_dp, 0.4133303747293787_dp, &
             0.7228058158644598_dp, 1.2547736500819207_dp, 2.2115879620251437_dp, &
             4.315984273584482_dp]

      CASE (644)

         aw(:) = & ! a
            [4.339896887672607e-07_dp, 4.65564005221034e-06_dp, 2.436047208357923e-05_dp, &
             9.343355300147844e-05_dp, 0.0002979947693769855_dp, 0.0008383028157196583_dp, &
             0.0021501146036319338_dp, 0.005133457452291181_dp, 0.011570090523687961_dp, &
             0.024864581040360018_dp, 0.05132899232074033_dp, 0.10236447152028563_dp, &
             0.19810097526676756_dp, 0.37338718612642047_dp, 0.6875978968353142_dp, &
             1.2409884314500514_dp, 2.2042964890581773_dp, 3.885912770448123_dp, &
             6.977782469133344_dp, &
             ! w
             1.3725865537365244e-06_dp, 8.813403577432947e-06_dp, 3.57776170575683e-05_dp, &
             0.00011575276862931395_dp, 0.0003248326583234597_dp, 0.0008247824253947033_dp, &
             0.0019421971881413143_dp, 0.004309602390337023_dp, 0.009110492925937144_dp, &
             0.01849543001121996_dp, 0.036274963433171103_dp, 0.06905444676841588_dp, &
             0.12806899965551122_dp, 0.23213935259497862_dp, 0.41253774119811615_dp, &
             0.7217605914897661_dp, 1.2534777451029218_dp, 2.2100793483284416_dp, &
             4.314268659532059_dp]

      CASE (645)

         aw(:) = & ! a
            [4.168893445584919e-07_dp, 4.543981955251293e-06_dp, 2.3949771096339182e-05_dp, &
             9.222738086711947e-05_dp, 0.0002949132167679088_dp, 0.0008311731521956374_dp, &
             0.0021348272897127496_dp, 0.00510263283950298_dp, 0.01151106918653797_dp, &
             0.024756542592937466_dp, 0.051139059390178765_dp, 0.10204282971083611_dp, &
             0.19757543489665844_dp, 0.37255835497994483_dp, 0.6863372953335471_dp, &
             1.2391431566797617_dp, 2.2017039482112253_dp, 3.882423715547703_dp, &
             6.973270194799345_dp, &
             ! w
             1.3274403142716745e-06_dp, 8.648792584038916e-06_dp, 3.529765445708514e-05_dp, &
             0.00011454809230188996_dp, 0.0003221099203732998_dp, 0.0008190942110839771_dp, &
             0.0019310363238376645_dp, 0.0042888233184603585_dp, 0.009073535679280187_dp, &
             0.018432360824494634_dp, 0.03617143184432845_dp, 0.06889082403415221_dp, &
             0.12782024599376507_dp, 0.23177656890436393_dp, 0.41203273604471696_dp, &
             0.7210944589246808_dp, 1.2526516653596356_dp, 2.2091175528704823_dp, &
             4.313174894943084_dp]

      CASE (646)

         aw(:) = & ! a
            [4.084650583208934e-07_dp, 4.48881593984698e-06_dp, 2.37461945248974e-05_dp, &
             9.162823755329421e-05_dp, 0.0002933802995417213_dp, 0.0008276227416206382_dp, &
             0.0021272083108013606_dp, 0.005087260177700746_dp, 0.011481618441205635_dp, &
             0.024702608394729198_dp, 0.051044205276961574_dp, 0.1018821445890046_dp, &
             0.1973128087408642_dp, 0.3721440587183614_dp, 0.6857070321533267_dp, &
             1.2382203916174337_dp, 2.2004072822310556_dp, 3.8806784245790644_dp, &
             6.97101285460236_dp, &
             ! w
             1.3051940084803493e-06_dp, 8.56724960600381e-06_dp, 3.50592154782945e-05_dp, &
             0.00011394866392924853_dp, 0.00032075362480584244_dp, 0.000816258332583344_dp, &
             0.0019254683517616487_dp, 0.004278451367366953_dp, 0.00905507996361836_dp, &
             0.018400853118647822_dp, 0.036119692824198316_dp, 0.06880903089766681_dp, &
             0.12769586446342168_dp, 0.2315951282557219_dp, 0.4117801142134111_dp, &
             0.7207611779445834_dp, 1.2522383046923014_dp, 2.208636245347379_dp, &
             4.31262754694385_dp]

      CASE (647)

         aw(:) = & ! a
            [4.0740554113376073e-07_dp, 4.481870010855176e-06_dp, 2.3720531008039338e-05_dp, &
             9.155264818763086e-05_dp, 0.0002931868035281797_dp, 0.0008271744153249903_dp, &
             0.002126245940334511_dp, 0.0050853179281050084_dp, 0.011477896732699876_dp, &
             0.024695791449217387_dp, 0.05103221437097328_dp, 0.10186182915129238_dp, &
             0.1972796019132273_dp, 0.37209167052453734_dp, 0.6856273280530193_dp, &
             1.2381036862280144_dp, 2.2002432689570703_dp, 3.8804576345416524_dp, &
             6.970727248623774_dp, &
             ! w
             1.3023957441521602e-06_dp, 8.556972572385258e-06_dp, 3.5029131851504276e-05_dp, &
             0.00011387299161161442_dp, 0.00032058234018585044_dp, 0.0008159000869000376_dp, &
             0.0019247647814086837_dp, 0.004277140486126056_dp, 0.009052746988393966_dp, &
             0.0183968695537645_dp, 0.036113150525904275_dp, 0.06879868763860304_dp, &
             0.127680134900435_dp, 0.23157218079563613_dp, 0.4117481601944227_dp, &
             0.7207190146581787_dp, 1.2521859976389873_dp, 2.2085753256603984_dp, &
             4.312558258635529_dp]

      CASE (648)

         aw(:) = & ! a
            [0.004241772553326922_dp, 0.02253863119170212_dp, 0.056252654578538085_dp, &
             0.10686297889578697_dp, 0.1767480303461968_dp, 0.26949675659200595_dp, &
             0.3903413741971059_dp, 0.5467141952811087_dp, 0.7489117852179463_dp, &
             1.0108647945710778_dp, 1.3510691399479386_dp, 1.793807628117466_dp, &
             2.370861271523886_dp, 3.1240180104942112_dp, 4.108948832501714_dp, &
             5.401688717441938_dp, 7.110753186116057_dp, 9.403471755301473_dp, &
             12.577053574970028_dp, 17.340642837706508_dp, &
             ! w
             0.010907089591214697_dp, 0.025821555710638135_dp, 0.04184928755405365_dp, &
             0.05976267618514956_dp, 0.08059837177459536_dp, 0.10576178276629854_dp, &
             0.13715507702276258_dp, 0.17729395534104633_dp, 0.22940779422008004_dp, &
             0.29756098082666815_dp, 0.3868670008889286_dp, 0.5038662196520325_dp, &
             0.6571437130474139_dp, 0.8583463758021642_dp, 1.1240021612615954_dp, &
             1.4791765085968618_dp, 1.9658740246659219_dp, 2.6659270101860626_dp, &
             3.782037265499978_dp, 6.118868095133639_dp]

      CASE (649)

         aw(:) = & ! a
            [0.0037488236103774817_dp, 0.019930293094675297_dp, 0.04979367932847378_dp, &
             0.09474082604624894_dp, 0.15704128191906463_dp, 0.24014528833308854_dp, &
             0.3491180664641332_dp, 0.49119352491964474_dp, 0.6764290683010313_dp, &
             0.9184668186510055_dp, 1.2354696723048324_dp, 1.6513682763417639_dp, &
             2.1976141031064995_dp, 2.9157364131133_dp, 3.861262156315753_dp, &
             5.110222459112411_dp, 6.77125238546136_dp, 9.0118118124226_dp, &
             12.128876175292088_dp, 16.82945136817126_dp, &
             ! w
             0.009640765748083391_dp, 0.022848814032431896_dp, 0.03710791484432057_dp, &
             0.053159507527788_dp, 0.07200919439511491_dp, 0.09503373260213062_dp, &
             0.12410655429508652_dp, 0.16170807287053018_dp, 0.2110214333163202_dp, &
             0.2760608847838723_dp, 0.3619046982414856_dp, 0.47509300165250873_dp, &
             0.6242574247591812_dp, 0.8211393223377071_dp, 1.082398587279776_dp, &
             1.4332424319088186_dp, 1.9157658348468556_dp, 2.6116952045089157_dp, &
             3.722973616876625_dp, 6.050192571793496_dp]

      CASE (650)

         aw(:) = & ! a
            [0.003366711683945265_dp, 0.01790770400246352_dp, 0.044781949092794454_dp, &
             0.0853260304433367_dp, 0.141716925654248_dp, 0.21728571203971078_dp, &
             0.316951657717353_dp, 0.44777110422795546_dp, 0.6195826682748403_dp, &
             0.8457617284670842_dp, 1.144163390200643_dp, 1.538393878511835_dp, &
             2.059600705991394_dp, 2.749071598154428_dp, 3.6621932897973233_dp, &
             4.874982879934231_dp, 6.496177865342313_dp, 8.693366989786185_dp, &
             11.76337124369219_dp, 16.41148115443289_dp, &
             ! w
             0.008659086358856514_dp, 0.020542642029852693_dp, 0.03342496579440091_dp, &
             0.04802087594522252_dp, 0.06530883702481378_dp, 0.0866394015413651_dp, &
             0.11385780643252849_dp, 0.14940847567110874_dp, 0.19642850913376267_dp, &
             0.25888470832290356_dp, 0.3418237825042459_dp, 0.45178630917617296_dp, &
             0.5974451909198462_dp, 0.7906248922759673_dp, 1.0481026458644116_dp, &
             1.3952152791208259_dp, 1.8741487865464226_dp, 2.5665602251435766_dp, &
             3.673781879670697_dp, 5.993048686887826_dp]

      CASE (651)

         aw(:) = & ! a
            [0.0030610352418640446_dp, 0.016289177186963568_dp, 0.040769082545252106_dp, &
             0.07778121534986662_dp, 0.1294224655903178_dp, 0.19891986790782226_dp, &
             0.29106335536902744_dp, 0.4127488785997983_dp, 0.5736143080480764_dp, &
             0.7867888460422182_dp, 1.0698439857964348_dp, 1.4460876939879321_dp, &
             1.946387359192112_dp, 2.6118055197376453_dp, 3.4975936449624725_dp, &
             4.67974761147643_dp, 6.267092973678195_dp, 8.427342184767177_dp, &
             11.457213472814704_dp, 16.06058220891612_dp, &
             ! w
             0.007873717027041583_dp, 0.018696448865385552_dp, 0.03047316116125044_dp, &
             0.04389542682419414_dp, 0.05991766909579364_dp, 0.07986634142466169_dp, &
             0.10555939472911423_dp, 0.13940569332571925_dp, 0.1844978629111232_dp, &
             0.2447586471699034_dp, 0.3252065290407138_dp, 0.43238268938229696_dp, &
             0.5749963524427604_dp, 0.7649452820216595_dp, 1.0191115839910416_dp, &
             1.3629507887370282_dp, 1.8387382016027254_dp, 2.5280860401426817_dp, &
             3.6318216968458192_dp, 5.944342632641489_dp]

      CASE (652)

         aw(:) = & ! a
            [0.002810436815627241_dp, 0.014961887863999799_dp, 0.03747651589044989_dp, &
             0.07158581140689843_dp, 0.11931639930252208_dp, 0.18380327975121988_dp, &
             0.26972048155186484_dp, 0.3838179390096638_dp, 0.5355486564059357_dp, &
             0.7378144249753411_dp, 1.0079252585553353_dp, 1.3689145349419942_dp, &
             1.8513909726764064_dp, 2.4962069757188354_dp, 3.3584842422582812_dp, &
             4.514192584071912_dp, 6.072233996634713_dp, 8.200438614882817_dp, &
             11.195457749923326_dp, 15.759975208198068_dp, &
             ! w
             0.007229813962328594_dp, 0.017181904981950723_dp, 0.0280490206203563_dp, &
             0.040502182119883684_dp, 0.05547427727826987_dp, 0.07426947848061351_dp, &
             0.09867951729362494_dp, 0.13107871029321663_dp, 0.1745173777820914_dp, &
             0.23287765125272755_dp, 0.3111526534957363_dp, 0.41588403717792344_dp, &
             0.5558129803400383_dp, 0.7429022864397854_dp, 0.9941284651201723_dp, &
             1.3350562364091412_dp, 1.808047525332381_dp, 2.4946869354115937_dp, &
             3.595376226505887_dp, 5.902068778258447_dp]

      CASE (653)

         aw(:) = & ! a
            [0.0015934091103306748_dp, 0.008509049468496795_dp, 0.02143813248916623_dp, &
             0.04132179241341644_dp, 0.06976208604731934_dp, 0.10932110468507521_dp, &
             0.16391894913083357_dp, 0.2393122305717882_dp, 0.343666004319766_dp, &
             0.48829689150562916_dp, 0.6887039382594222_dp, 0.9660147447332132_dp, &
             1.3490093267327685_dp, 1.8769921238144998_dp, 2.6040387411362604_dp, &
             3.6057691428697556_dp, 4.991494853262373_dp, 6.929917169042379_dp, &
             9.717722394528794_dp, 14.051261273391312_dp, &
             ! w
             0.00410193054094625_dp, 0.009809035790519609_dp, 0.01620265422744076_dp, &
             0.023826697381147642_dp, 0.03347420139913778_dp, 0.046289223081208496_dp, &
             0.06385657144849423_dp, 0.08827874183675773_dp, 0.12229802015324023_dp, &
             0.16952318533991945_dp, 0.2347790375469691_dp, 0.3245925111476496_dp, &
             0.44787902068913005_dp, 0.616992015749277_dp, 0.8495175968623115_dp, &
             1.1717832983682304_dp, 1.6268516930985548_dp, 2.2963863207527395_dp, &
             3.37853613531568_dp, 5.651133287053622_dp]

      CASE (654)

         aw(:) = & ! a
            [0.0011390449921040572_dp, 0.0060955201406988_dp, 0.015419250308827093_dp, &
             0.02990826519226592_dp, 0.05094935194943481_dp, 0.08080106924314619_dp, &
             0.1229612992811983_dp, 0.18260687338320566_dp, 0.26714281943879_dp, &
             0.38695596655717285_dp, 0.5564845870412083_dp, 0.7957234718437416_dp, &
             1.132328903295979_dp, 1.6046003100378334_dp, 2.2658640492658004_dp, &
             3.191389682165693_dp, 4.490627547122133_dp, 6.332813436815695_dp, &
             9.014943119477653_dp, 13.230670944948036_dp, &
             ! w
             0.0029336693857008778_dp, 0.007045155559718517_dp, 0.011732149752537013_dp, &
             0.017472271568595172_dp, 0.024980140626674693_dp, 0.035299034077060434_dp, &
             0.04987532809486311_dp, 0.0706377004364661_dp, 0.10014776931950316_dp, &
             0.14186018661282754_dp, 0.20049505044558505_dp, 0.282543540531873_dp, &
             0.3969798598412203_dp, 0.55634100754055_dp, 0.7785426625143363_dp, &
             1.0903727561666794_dp, 1.5353838262114279_dp, 2.1954643644645224_dp, &
             3.267829696349966_dp, 5.523411180505148_dp]

      CASE (655)

         aw(:) = & ! a
            [0.000896430353358387_dp, 0.004805119176955691_dp, 0.012193651558632754_dp, &
             0.023770227053445305_dp, 0.04078388271252856_dp, 0.06529332852670133_dp, &
             0.10051033124932444_dp, 0.15121365150247307_dp, 0.22428734299024927_dp, &
             0.32948247592091867_dp, 0.48050771905089906_dp, 0.6965665705233484_dp, &
             1.0045115922971968_dp, 1.4418995293281531_dp, 2.061474082778248_dp, &
             2.938193998409961_dp, 4.1815556890979115_dp, 5.961164867318764_dp, &
             8.574326896911181_dp, 12.71314678078166_dp, &
             ! w
             0.0023096719274510653_dp, 0.005565099697046422_dp, 0.00932697130056929_dp, &
             0.014029831490400173_dp, 0.02033483456646612_dp, 0.029212591178159775_dp, &
             0.04200916174497004_dp, 0.06053237096664171_dp, 0.0872222769826313_dp, &
             0.12542753127673448_dp, 0.17978826736121206_dp, 0.2567550853442525_dp, &
             0.3653220398036224_dp, 0.5181328291292073_dp, 0.7333190532368281_dp, &
             1.0379927221561274_dp, 1.4760781532543799_dp, 2.1296885906061567_dp, &
             3.19552335247835_dp, 5.440133455439082_dp]

      CASE (656)

         aw(:) = & ! a
            [0.0007439418581085582_dp, 0.0039932450657249995_dp, 0.01016041215644215_dp, &
             0.01989032315884935_dp, 0.0343334526398138_dp, 0.05540249311027042_dp, &
             0.08609649749456823_dp, 0.13089688630363036_dp, 0.19629951199337_dp, &
             0.29158147617537883_dp, 0.4299035699271377_dp, 0.6298673381976871_dp, &
             0.9177032441082223_dp, 1.3303801631760857_dp, 1.9201651329774694_dp, &
             2.7617427757686337_dp, 3.964614476648414_dp, 5.698662188893201_dp, &
             8.261471019071442_dp, 12.34412488855432_dp, &
             ! w
             0.001917383127252153_dp, 0.004632732016633214_dp, 0.007806163909567718_dp, &
             0.01184107613662278_dp, 0.017358546055447416_dp, 0.02527294887811247_dp, &
             0.03685334833110203_dp, 0.05381752047687813_dp, 0.07851580781114899_dp, &
             0.11421603606997456_dp, 0.1654925894048662_dp, 0.23875601493646906_dp, &
             0.3430036016845556_dp, 0.4909482023914976_dp, 0.7008773996077388_dp, &
             1.000150454248501_dp, 1.4329899005282858_dp, 2.08171532337389_dp, &
             3.1427009603970153_dp, 5.379366047067225_dp]

      CASE (657)

         aw(:) = & ! a
            [0.0006385690974850548_dp, 0.0034317403777428604_dp, 0.00875196804462231_dp, &
             0.017196311234527913_dp, 0.029839877029217036_dp, 0.04848187410129415_dp, &
             0.07595415290631778_dp, 0.11650406456540177_dp, 0.17632240507939068_dp, &
             0.26431308457328057_dp, 0.39320244745486965_dp, 0.5811095604063918_dp, &
             0.853759528622832_dp, 1.2476356308222163_dp, 1.8146016852789122_dp, &
             2.629098453882577_dp, 3.8006115297650886_dp, 5.499239639730987_dp, &
             8.022817145154024_dp, 12.061699046719113_dp, &
             ! w
             0.0016462492734708235_dp, 0.003987211222706093_dp, 0.006749927848474149_dp, &
             0.010313786046778031_dp, 0.015268078039520693_dp, 0.022481702890543482_dp, &
             0.03316218958817866_dp, 0.04895682507679239_dp, 0.07214562675632333_dp, &
             0.1059310341581058_dp, 0.15483131157096713_dp, 0.2252190529196926_dp, &
             0.32608670198114226_dp, 0.4701949338490822_dp, 0.6759509282977448_dp, &
             0.9709121995668761_dp, 1.3995495469103152_dp, 2.0443694768055507_dp, &
             3.1015254823222613_dp, 5.332037723557218_dp]

      CASE (658)

         aw(:) = & ! a
            [0.0005610725711432293_dp, 0.0030184719671453096_dp, 0.007713921315130133_dp, &
             0.015206663999710929_dp, 0.026511569454176265_dp, 0.043335950204500394_dp, &
             0.06837521035458008_dp, 0.10568566283565914_dp, 0.16120938549632083_dp, &
             0.24354520431861942_dp, 0.36506197351819525_dp, 0.5434781501498771_dp, &
             0.8040946723737953_dp, 1.182981838612351_dp, 1.7316542057370126_dp, &
             2.524333383016714_dp, 3.6704779200261983_dp, 5.340362657551184_dp, &
             7.832044678316398_dp, 11.835330483075156_dp, &
             ! w
             0.0014468089890773762_dp, 0.0035116704407315387_dp, 0.005969689339056726_dp, &
             0.00918093559184832_dp, 0.01370852778480848_dp, 0.02038345914096959_dp, &
             0.03036259649524996_dp, 0.045236019613996474_dp, 0.06722629097370483_dp, &
             0.09948092471425347_dp, 0.14646915175447675_dp, 0.21452807430388857_dp, &
             0.3126410766661178_dp, 0.4536033319935106_dp, 0.6559174963072968_dp, &
             0.9473055307348146_dp, 1.3724504245122913_dp, 2.0140282495547255_dp, &
             3.0680352452904605_dp, 5.293568706378472_dp]

      CASE (659)

         aw(:) = & ! a
            [0.0005015028829405916_dp, 0.0027005909930681474_dp, 0.006914486955199627_dp, &
             0.01367152910902928_dp, 0.023936894632908164_dp, 0.039341265669578386_dp, &
             0.06246564620060502_dp, 0.09720625003680655_dp, 0.14929680764893682_dp, &
             0.2270798198518985_dp, 0.3426219915469064_dp, 0.513300567195081_dp, &
             0.7640519475501053_dp, 1.130587662424983_dp, 1.664114344190769_dp, &
             2.438654789416963_dp, 3.563634197053104_dp, 5.209474209237955_dp, &
             7.674432292794263_dp, 11.647886503479498_dp, &
             ! w
             0.001293480478970173_dp, 0.003145591393180043_dp, 0.005367581432016754_dp, &
             0.008303496086809293_dp, 0.012494314976958765_dp, 0.01873872475063063_dp, &
             0.028150920427841508_dp, 0.04227326683064738_dp, 0.06327996456401508_dp, &
             0.09427117790892246_dp, 0.13967258345880318_dp, 0.20578820712319962_dp, &
             0.30159011010787845_dp, 0.43989907236368175_dp, 0.6392962249306502_dp, &
             0.9276433398220163_dp, 1.3498083621591483_dp, 1.9886219964371001_dp, &
             3.0399647857026815_dp, 5.261342694789207_dp]

      CASE (660)

         aw(:) = & ! a
            [0.0004541767980519825_dp, 0.002447893212537064_dp, 0.006278273148409975_dp, &
             0.012447764284239307_dp, 0.021879565191843225_dp, 0.03613903396558967_dp, &
             0.05770927188182987_dp, 0.0903495739014426_dp, 0.1396154853300277_dp, &
             0.21362961592784094_dp, 0.32419806433545034_dp, 0.48840180253933013_dp, &
             0.7308582772942873_dp, 1.086962252899046_dp, 1.6076451348571192_dp, &
             2.366747876877618_dp, 3.4736587933681657_dp, 5.098924335828885_dp, &
             7.540983597428812_dp, 11.488870490534087_dp, &
             ! w
             0.001171649233791555_dp, 0.002854363744491754_dp, 0.004887525207822211_dp, &
             0.0076015761052112005_dp, 0.011518388693923184_dp, 0.017408639729659964_dp, &
             0.026349928628182206_dp, 0.039843964094377245_dp, 0.06002331710036192_dp, &
             0.0899465532002833_dp, 0.13400017629258867_dp, 0.19845737327731705_dp, &
             0.2922776803300323_dp, 0.4283012619476658_dp, 0.6251751945622182_dp, &
             0.9108823666733669_dp, 1.3304544089975425_dp, 1.9668638852113633_dp, &
             3.0159043047515124_dp, 5.233732854758063_dp]

      CASE (661)

         aw(:) = & ! a
            [0.00041560272898321613_dp, 0.002241813192412507_dp, 0.005758898463407207_dp, &
             0.011447193138584475_dp, 0.020193770835598295_dp, 0.03350733934525363_dp, &
             0.05378585498474616_dp, 0.08466962315807926_dp, 0.13155928186988688_dp, &
             0.2023856910924211_dp, 0.30872657532736725_dp, 0.4674017464982133_dp, &
             0.7027456289571763_dp, 1.0498697079903423_dp, 1.559456679494127_dp, &
             2.305180002998289_dp, 3.3963892157202724_dp, 5.0037385567561214_dp, &
             7.425832889926019_dp, 11.351423637829141_dp, &
             ! w
             0.001072335584192338_dp, 0.0026167011819126502_dp, 0.004494970236155722_dp, &
             0.00702582820415164_dp, 0.01071439556145332_dp, 0.016306731243294847_dp, &
             0.024848587534049695_dp, 0.037806435462356625_dp, 0.057276384992356724_dp, &
             0.08627992469107162_dp, 0.12916797666084973_dp, 0.19218491545905134_dp, &
             0.28427715878659077_dp, 0.4182997388385609_dp, 0.6129561237291932_dp, &
             0.8963356883659791_dp, 1.313616686843964_dp, 1.9479025693636614_dp, &
             2.9949203773370825_dp, 5.209662761405782_dp]

      CASE (662)

         aw(:) = & ! a
            [0.00023138954565206143_dp, 0.0012557354836193616_dp, 0.003264674188789077_dp, &
             0.006615259390547789_dp, 0.011987718051028634_dp, 0.020558775484834778_dp, &
             0.03422479630665998_dp, 0.05592679003656871_dp, 0.09015105663108257_dp, &
             0.14368197637518992_dp, 0.22670752132903546_dp, 0.3544262626829793_dp, &
             0.5493766307654294_dp, 0.8448259409096455_dp, 1.2897823199991467_dp, &
             1.9567226023270095_dp, 2.954615621729697_dp, 4.454728654494038_dp, &
             6.75680531467233_dp, 10.548255884539723_dp, &
             ! w
             0.0005978419885949287_dp, 0.0014767718631817715_dp, 0.002598448344073917_dp, &
             0.004213188499592838_dp, 0.006724277304707164_dp, 0.010728665934929072_dp, &
             0.01708536858417583_dp, 0.02705401971171137_dp, 0.04250716462384939_dp, &
             0.06622639694306004_dp, 0.10231973475161851_dp, 0.15681891336454068_dp, &
             0.23854515401437476_dp, 0.36039785241191397_dp, 0.5413913490296749_dp, &
             0.8102669926540256_dp, 1.2131602744747734_dp, 1.8341101331108358_dp, &
             2.8686436013263883_dp, 5.064992123138682_dp]

      CASE (663)

         aw(:) = & ! a
            [0.00016410687347897756_dp, 0.0008943560944810265_dp, 0.0023447971546885437_dp, &
             0.00481575108056489_dp, 0.00888843116175323_dp, 0.015576509779991123_dp, &
             0.02652969704976162_dp, 0.044345403227743795_dp, 0.0730554063217297_dp, &
             0.118861997867758_dp, 0.191229739550222_dp, 0.3044906682984813_dp, &
             0.48019573446875213_dp, 0.750565238695917_dp, 1.163617043503107_dp, &
             1.7910716090620555_dp, 2.7415868113584296_dp, 4.18670920782827_dp, &
             6.4268615884660525_dp, 10.149012221330613_dp, &
             ! w
             0.0004244010974768446_dp, 0.0010572758982993402_dp, 0.0018917249558589985_dp, &
             0.0031445443054133513_dp, 0.005166527520630457_dp, 0.008479543782697985_dp, &
             0.013851786363182749_dp, 0.022439566003445564_dp, 0.035996327160287896_dp, &
             0.05716806795048214_dp, 0.08991770235316487_dp, 0.14013975784179847_dp, &
             0.21655728802956828_dp, 0.33205806317432546_dp, 0.5057921887798438_dp, &
             0.7668399326093335_dp, 1.1618799561175959_dp, 1.775539474385565_dp, &
             2.8033893736182858_dp, 4.990346897743811_dp]

      CASE (664)

         aw(:) = & ! a
            [0.0001285687160662041_dp, 0.0007030265145218757_dp, 0.0018556066468578645_dp, &
             0.0038520805744970844_dp, 0.007211906522971537_dp, 0.012846033073725241_dp, &
             0.022249041166251503_dp, 0.0378016757009909_dp, 0.06324608416940793_dp, &
             0.10440835102683746_dp, 0.1702784124967893_dp, 0.2746109811926541_dp, &
             0.43828839519493173_dp, 0.6928091186450619_dp, 1.0854928248814808_dp, &
             1.6875078196578601_dp, 2.607259000024923_dp, 4.0164561880298075_dp, &
             6.215999416155189_dp, 9.892663886988673_dp, &
             ! w
             0.00033274122489221656_dp, 0.0008345346294245275_dp, 0.00151314890244546_dp, &
             0.0025641220402895077_dp, 0.004304308656322565_dp, 0.007208139365701322_dp, &
             0.011986914291511_dp, 0.01973026005032079_dp, 0.03211195340172582_dp, &
             0.05168477847723007_dp, 0.08230924705368046_dp, 0.12977945449930747_dp, &
             0.20274072172405352_dp, 0.3140586398153147_dp, 0.48296069776984113_dp, &
             0.7387477919649796_dp, 1.1284719744874216_dp, 1.7371875880089804_dp, &
             2.760554454781633_dp, 4.94138714425316_dp]

      CASE (665)

         aw(:) = & ! a
            [0.00010638513337661713_dp, 0.0005833704497286544_dp, 0.0015485833354592_dp, &
             0.003243857779284663_dp, 0.006145151631333021_dp, 0.011090719399673434_dp, &
             0.019465510810252128_dp, 0.03349659199969773_dp, 0.056718875461408656_dp, &
             0.09468643752912093_dp, 0.15604237902086526_dp, 0.2541149900337321_dp, &
             0.40928721170803795_dp, 0.6525116370013598_dp, 1.0305718769518786_dp, &
             1.614202496363769_dp, 2.511596469919324_dp, 3.8945710669852747_dp, &
             6.064390848408057_dp, 9.707739662116053_dp, &
             ! w
             0.0002755007456293087_dp, 0.0006949144419033539_dp, 0.001274169153691706_dp, &
             0.002193630728460829_dp, 0.0037457419701558727_dp, 0.006371425385050147_dp, &
             0.010741743538383912_dp, 0.017897978924549002_dp, 0.02945482683025403_dp, &
             0.04789478310608085_dp, 0.07699984237828225_dp, 0.12248529451509546_dp, &
             0.19293260035204676_dp, 0.3011832664673696_dp, 0.4665147559883756_dp, &
             0.7183877529337838_dp, 1.1041357993848417_dp, 1.7091472734101285_dp, &
             2.7291791887956056_dp, 4.905544722001163_dp]

      CASE (666)

         aw(:) = & ! a
            [9.11299266109851e-05_dp, 0.0005009551141945329_dp, 0.0013364798805110458_dp, &
             0.002821657430933893_dp, 0.005399551234210625_dp, 0.009853324512425465_dp, &
             0.01748490637696873_dp, 0.030404533385128128_dp, 0.05198841321460045_dp, &
             0.08758059860064037_dp, 0.14555419725765043_dp, 0.23890283571194187_dp, &
             0.38761418843942763_dp, 0.6222047474237472_dp, 0.9890249822958673_dp, &
             1.5584530834516235_dp, 2.438500055428932_dp, 3.801059007802197_dp, &
             5.947687483988444_dp, 9.565027952205712_dp, &
             ! w
             0.00023612333676443841_dp, 0.0005985629681840805_dp, 0.0011082635962737625_dp, &
             0.0019339921633778057_dp, 0.003349486082006665_dp, 0.005770347649989956_dp, &
             0.00983706815425649_dp, 0.016553457676045477_dp, 0.02748768823260326_dp, &
             0.045066259082733595_dp, 0.07300785235072305_dp, 0.1169631532290535_dp, &
             0.18545960822716298_dp, 0.2913150984442548_dp, 0.4538418132150399_dp, &
             0.7026237141263744_dp, 1.0852185580301998_dp, 1.6872881704391611_dp, &
             2.704685059280989_dp, 4.877573997791687_dp]

      CASE (667)

         aw(:) = & ! a
            [7.995132118180791e-05_dp, 0.00044048010027379365_dp, 0.0011804328722135864_dp, &
             0.0025097286666722054_dp, 0.004845367703405868_dp, 0.00892681999663504_dp, &
             0.015990188929398_dp, 0.028052723146404997_dp, 0.048363517911433784_dp, &
             0.0820972791919213_dp, 0.13740797572233093_dp, 0.22701582280163482_dp, &
             0.3705833787404193_dp, 0.598265757020446_dp, 0.9560512745983604_dp, &
             1.5140162498697851_dp, 2.3800124105653953_dp, 3.725988927950908_dp, &
             5.853747072378959_dp, 9.449915264805615_dp, &
             ! w
             0.00020725952314682136_dp, 0.0005277431684360765_dp, 0.000985682497105203_dp, &
             0.0017405679883607983_dp, 0.0030511858987081137_dp, 0.005313115155973688_dp, &
             0.009142495573192385_dp, 0.015512798367128966_dp, 0.025954084638724124_dp, &
             0.04284657521764742_dp, 0.06985613461186108_dp, 0.11257886286898659_dp, &
             0.17949549992255948_dp, 0.28340146512637426_dp, 0.44363426571246983_dp, &
             0.6898770636570903_dp, 1.0698728982917909_dp, 1.6695145014205042_dp, &
             2.6847452179286795_dp, 4.854810790543109_dp]

      CASE (668)

         aw(:) = & ! a
            [7.138319759578014e-05_dp, 0.00039407055741463057_dp, 0.0010603983810765622_dp, &
             0.002268877358945472_dp, 0.0044151612591759315_dp, 0.008202922843677896_dp, &
             0.014814330720832129_dp, 0.026190163148415824_dp, 0.0454744113356478_dp, &
             0.07770096275156518_dp, 0.13084055992928648_dp, 0.21738356683378604_dp, &
             0.35671768422829436_dp, 0.5786906301817148_dp, 0.9289803945008825_dp, &
             1.4774018721298756_dp, 2.3316652624744667_dp, 3.6637625466597186_dp, &
             5.77570286165078_dp, 9.354116581076887_dp, &
             ! w
             0.00018512990988537123_dp, 0.0004733136428774598_dp, 0.0008910302345945904_dp, &
             0.001590110314295655_dp, 0.0028170215892367894_dp, 0.004950980035778313_dp, &
             0.008588067479626081_dp, 0.014676426665476994_dp, 0.024714009713086252_dp, &
             0.04104177082339774_dp, 0.0672804167578405_dp, 0.10897889875029373_dp, &
             0.17457687560090157_dp, 0.2768486379035145_dp, 0.43515078112079053_dp, &
             0.6792487312638682_dp, 1.0570427273451812_dp, 1.6546249337670091_dp, &
             2.6680241617729963_dp, 4.8357267236022885_dp]

      CASE (669)

         aw(:) = & ! a
            [6.459190461722343e-05_dp, 0.00035724435900411127_dp, 0.0009649470914152303_dp, &
             0.0020766909397591077_dp, 0.004070207966574025_dp, 0.007619118305242462_dp, &
             0.013860295060617365_dp, 0.024670071758435087_dp, 0.04310343856843796_dp, &
             0.07407445849146964_dp, 0.12539723010997336_dp, 0.20936469832195895_dp, &
             0.3451273583100701_dp, 0.5622662571957021_dp, 0.906188494647872_dp, &
             1.4464787367027907_dp, 2.2907199059439574_dp, 3.610937461958469_dp, &
             5.709321201296618_dp, 9.272513232139142_dp, &
             ! w
             0.00016758500638978994_dp, 0.00043006514282330785_dp, 0.0008155017874436882_dp, &
             0.001469248097150156_dp, 0.002627386997774696_dp, 0.004655426682192709_dp, &
             0.008132509888542614_dp, 0.013985146277627593_dp, 0.023683660533506797_dp, &
             0.03953502236781608_dp, 0.06512060492073322_dp, 0.1059479409027388_dp, &
             0.17042006610748636_dp, 0.2712914564768583_dp, 0.42793345558814766_dp, &
             0.6701812710599258_dp, 1.046071187548854_dp, 1.641870595521293_dp, &
             2.653688472529268_dp, 4.819368382240202_dp]

      CASE (670)

         aw(:) = & ! a
            [5.90671705438591e-05_dp, 0.00032725560625953594_dp, 0.0008870654681564532_dp, &
             0.001919380669129223_dp, 0.0037865946804733056_dp, 0.0071366131501388996_dp, &
             0.013067530856092496_dp, 0.023400328296027062_dp, 0.04111322369837066_dp, &
             0.07101647927822646_dp, 0.12078794040720191_dp, 0.20254813669344618_dp, &
             0.3352395392683722_dp, 0.5482082740098587_dp, 0.8866215226553135_dp, &
             1.4198584985308402_dp, 2.2553865435031883_dp, 3.5652577795257527_dp, &
             5.651821361392682_dp, 9.201737270796274_dp, &
             ! w
             0.000153308858399679_dp, 0.0003948028739733491_dp, 0.0007536794802031423_dp, &
             0.0013697121315453817_dp, 0.0024700705801408646_dp, 0.004408554074507663_dp, &
             0.007749721299145884_dp, 0.01340127663813159_dp, 0.022809381067901244_dp, &
             0.03825113202338237_dp, 0.06327313847041809_dp, 0.10334606017832278_dp, &
             0.16683992578410292_dp, 0.2664906249918933_dp, 0.4216811064028229_dp, &
             0.6623068297126378_dp, 1.0365236624009746_dp, 1.6307550397909834_dp, &
             2.641185142411161_dp, 4.805103347322468_dp]

      CASE (671)

         aw(:) = & ! a
            [3.282244845679842e-05_dp, 0.00018428019757663435_dp, 0.0005131297932514203_dp, &
             0.0011553395530343967_dp, 0.002387152274167475_dp, 0.004712323365880187_dp, &
             0.009010683603238598_dp, 0.016787745478661017_dp, 0.03057773526449549_dp, &
             0.05458202416656862_dp, 0.09566826339550022_dp, 0.1649179422994952_dp, &
             0.2800025692621691_dp, 0.4688114688558269_dp, 0.7749976991407692_dp, &
             1.2666117185092605_dp, 2.050331632839722_dp, 3.2983146221139155_dp, &
             5.313900078330825_dp, 8.784007925670132_dp, &
             ! w
             8.543536907994339e-05_dp, 0.00022594418279482442_dp, 0.0004534407500189063_dp, &
             0.0008756622753532169_dp, 0.001669533056188414_dp, 0.0031234054784167746_dp, &
             0.005717707148168715_dp, 0.01024843812867291_dp, 0.018015553258346157_dp, &
             0.031112489242744585_dp, 0.05286827390244286_dp, 0.08851737705733385_dp, &
             0.14621020274352442_dp, 0.23854430050242964_dp, 0.38494581966192026_dp, &
             0.6156579968253519_dp, 0.9795718925153979_dp, 1.5641128247650342_dp, &
             2.566025112898437_dp, 4.719398313268424_dp]

      CASE (672)

         aw(:) = & ! a
            [2.3296240216139954e-05_dp, 0.00013205598109267732_dp, 0.00037483997796931476_dp, &
             0.0008670180330958671_dp, 0.0018448086515574628_dp, 0.0037452033977933554_dp, &
             0.00734592546659092_dp, 0.014001974700058187_dp, 0.026031328267862423_dp, &
             0.04733299962040928_dp, 0.0843648984201871_dp, 0.14767359121846338_dp, &
             0.2542643418782515_dp, 0.4312479420643727_dp, 0.7214499968181386_dp, &
             1.1921724790624288_dp, 1.9496202226863095_dp, 3.1659629223902988_dp, &
             5.1450686884499515_dp, 8.574095257167548_dp, &
             ! w
             6.0763619216622206e-05_dp, 0.0001637936305323157_dp, 0.0003401748566576009_dp, &
             0.0006823060326976989_dp, 0.0013438522290620373_dp, 0.002582735666319016_dp, &
             0.004838306668765428_dp, 0.008850115868929915_dp, 0.015842574525153197_dp, &
             0.02781225803715432_dp, 0.047970804615602824_dp, 0.08142156305547979_dp, &
             0.13618750655782294_dp, 0.22477638237901468_dp, 0.36661726952946033_dp, &
             0.592120682929402_dp, 0.9505653583803337_dp, 1.5299354639822842_dp, &
             2.5273395599440844_dp, 4.675313119453099_dp]

      CASE (673)

         aw(:) = & ! a
            [1.8276868638413513e-05_dp, 0.00010441734791908067_dp, 0.00030100095146226835_dp, &
             0.0007108478106689388_dp, 0.0015457136129836213_dp, 0.003201818623579705_dp, &
             0.006393972429505417_dp, 0.012383246948387211_dp, 0.023351000974342766_dp, &
             0.04300305996357401_dp, 0.07753274085079426_dp, 0.1371376078702122_dp, &
             0.23838371716071566_dp, 0.40786296778662046_dp, 0.6878424945034628_dp, &
             1.1451105980936995_dp, 1.8855365450248804_dp, 3.081280965420726_dp, &
             5.036563393505113_dp, 8.43873603575549_dp, &
             ! w
             4.7751051453804356e-05_dp, 0.00013072534819952663_dp, 0.0002788396785103831_dp, &
             0.000574965922084612_dp, 0.0011586413313408029_dp, 0.0022690307150705883_dp, &
             0.004319416175125798_dp, 0.008012921142966554_dp, 0.014524625638736806_dp, &
             0.025787112468465653_dp, 0.044933465258560235_dp, 0.07697790652636302_dp, &
             0.12985476862376266_dp, 0.21600595832397973_dp, 0.3548547440342577_dp, &
             0.5769157832693194_dp, 0.9317238561593227_dp, 1.5076443958234427_dp, &
             2.502053463991033_dp, 4.646507016588269_dp]

      CASE (674)

         aw(:) = & ! a
            [1.5147340401565904e-05_dp, 8.712436592340288e-05_dp, 0.00025446897357620836_dp, &
             0.0006112999997930726_dp, 0.001352405918572009_dp, 0.0028457170686157877_dp, &
             0.005762058855060931_dp, 0.011296221165483266_dp, 0.021532316189895696_dp, &
             0.04003756333306515_dp, 0.07281403028357235_dp, 0.12980520659833916_dp, &
             0.2272551001522987_dp, 0.39137224595415776_dp, 0.6640076136938744_dp, &
             1.1115623832190098_dp, 1.83964746537561_dp, 3.0204076691520507_dp, &
             4.958321444052936_dp, 8.340902576822307_dp, &
             ! w
             3.963142520908971e-05_dp, 0.00010994677153384957_dp, 0.00023974999696204358_dp, &
             0.0005052398564666033_dp, 0.0010362041171127665_dp, 0.002058660329900395_dp, &
             0.0039672605226802465_dp, 0.007438819612369857_dp, 0.01361250064614926_dp, &
             0.02437391757564572_dp, 0.042797988134900745_dp, 0.0738322409424064_dp, &
             0.1253436635886914_dp, 0.20972248093193413_dp, 0.346383680564076_dp, &
             0.565915026807196_dp, 0.9180391646116518_dp, 1.4914076613135867_dp, &
             2.4836068652550582_dp, 4.6254969283030345_dp]

      CASE (675)

         aw(:) = & ! a
            [1.299644711501509e-05_dp, 7.520388051868122e-05_dp, 0.00022219652968725686_dp, &
             0.000541590274135701_dp, 0.0012155018753602202_dp, 0.0025907114687332097_dp, &
             0.005304948907276802_dp, 0.010502765909656745_dp, 0.0201940756625822_dp, &
             0.03783969999766638_dp, 0.06929406639713083_dp, 0.1243034535758097_dp, &
             0.21886053539179673_dp, 0.37887295006304633_dp, 0.6458629128150952_dp, &
             1.0859232168150545_dp, 1.8044557310508904_dp, 2.9735873205278343_dp, &
             4.897999122899006_dp, 8.265342187736998_dp, &
             ! w
             3.404718524927942e-05_dp, 9.557191444988849e-05_dp, 0.00021238045075080948_dp, &
             0.0004556557980377821_dp, 0.0009479317314757141_dp, 0.0019053007114755301_dp, &
             0.003708152349482698_dp, 0.0070130092379226475_dp, 0.0129311616612344_dp, &
             0.023311550008499805_dp, 0.04118337445729427_dp, 0.07144133076391913_dp, &
             0.12189846354382541_dp, 0.20490263895759564_dp, 0.33985996859096346_dp, &
             0.5574132845540142_dp, 0.9074318573784056_dp, 1.4787944589267876_dp, &
             2.4692600202586_dp, 4.609158769496444_dp]

      CASE (676)

         aw(:) = & ! a
            [1.1420704827384798e-05_dp, 6.644841306144953e-05_dp, 0.0001983647056325913_dp, &
             0.0004896803425685338_dp, 0.001112575152543483_dp, 0.002397220375369971_dp, &
             0.004955211654653356_dp, 0.009891192823939993_dp, 0.01915581859674716_dp, &
             0.036124523403920084_dp, 0.06653271030106435_dp, 0.11996697540771133_dp, &
             0.21221560024727218_dp, 0.36894041220632945_dp, 0.63139368249158_dp, &
             1.0654133065902547_dp, 1.7762262715303474_dp, 2.935941135547897_dp, &
             4.849404223266529_dp, 8.2043854281906_dp, &
             ! w
             2.995382356019056e-05_dp, 8.498045566980717e-05_dp, 0.000192001219094592_dp, &
             0.0004182489471985938_dp, 0.0008805842181958542_dp, 0.001787234799634824_dp, &
             0.0035071677345569767_dp, 0.006680559720881566_dp, 0.01239613980992307_dp, &
             0.022473018064728997_dp, 0.039903011342209464_dp, 0.06953734554432704_dp, &
             0.11914429878216869_dp, 0.2010359691226609_dp, 0.33460966467006525_dp, &
             0.5505516804951869_dp, 0.8988505091143713_dp, 1.4685722300308928_dp, &
             2.4576216714489925_dp, 4.595906548055523_dp]

      CASE (677)

         aw(:) = & ! a
            [1.0212985366548821e-05_dp, 5.9722403366572045e-05_dp, 0.0001799680018765727_dp, &
             0.00044931076196723173_dp, 0.0010318630619396574_dp, 0.0022442905537817565_dp, &
             0.004676834053791326_dp, 0.009401364514424294_dp, 0.018319655076619996_dp, &
             0.03473641915443493_dp, 0.06428810618528928_dp, 0.11642807548110373_dp, &
             0.20677345280706258_dp, 0.3607794533247314_dp, 0.6194704946470453_dp, &
             1.0484681536117124_dp, 1.7528495505699622_dp, 2.9047052932410935_dp, &
             4.8090202468495535_dp, 8.153668809618228_dp, &
             ! w
             2.6814888671895574e-05_dp, 7.682109925659004e-05_dp, 0.0001761533954567387_dp, &
             0.0003888283075248282_dp, 0.0008271090726097902_dp, 0.0016927749136430965_dp, &
             0.0033453481688006294_dp, 0.0064114272292809605_dp, 0.01196092381642915_dp, &
             0.021787965875767177_dp, 0.03885292682690948_dp, 0.06797028506028897_dp, &
             0.11687020958774029_dp, 0.1978339264924517_dp, 0.33025026257463663_dp, &
             0.544840978502778_dp, 0.8916943748102493_dp, 1.4600351408849785_dp, &
             2.447894150285315_dp, 4.584831122752953_dp]

      CASE (678)

         aw(:) = & ! a
            [9.255636381254815e-06_dp, 5.437965503214943e-05_dp, 0.00016528995727415496_dp, &
             0.0004168855293001865_dp, 0.0009665567880418297_dp, 0.0021196974050103484_dp, &
             0.004448647970498137_dp, 0.008997689818335222_dp, 0.017627287867521872_dp, &
             0.03358218483555298_dp, 0.06241464880294735_dp, 0.1134643430422088_dp, &
             0.20220189272564937_dp, 0.3539051015943022_dp, 0.6094020408842701_dp, &
             1.0341270308398436_dp, 1.733026399382762_dp, 2.87817345225015_dp, &
             4.774671775009438_dp, 8.110488835063792_dp, &
             ! w
             2.4325539837231312e-05_dp, 7.032319584055394e-05_dp, 0.00016342457532755945_dp, &
             0.0003649611351340364_dp, 0.0007833701929401099_dp, 0.0016150077028122003_dp, &
             0.0032113984249499696_dp, 0.006187598075897911_dp, 0.011597467166838148_dp, &
             0.021213750490927972_dp, 0.03796980894210324_dp, 0.0666484162871906_dp, &
             0.11494667032143786_dp, 0.19511870264534836_dp, 0.3265452682814756_dp, &
             0.5399777956464662_dp, 0.8855899837911332_dp, 1.4527435663667752_dp, &
             2.4395801419510788_dp, 4.575365773037915_dp]

      CASE (679)

         aw(:) = & ! a
            [8.476697961661328e-06_dp, 5.0024283797197366e-05_dp, 0.00015327557410932326_dp, &
             0.00039018260856822564_dp, 0.0009124207524581635_dp, 0.002015783790247564_dp, &
             0.00425730835582959_dp, 0.008657599106267897_dp, 0.01704155339535864_dp, &
             0.03260212202884126_dp, 0.06081866885650867_dp, 0.1109321284586668_dp, &
             0.1982855745959643_dp, 0.3480019224842352_dp, 0.600737297535444_dp, &
             1.0217613928506457_dp, 1.7159047521703672_dp, 2.8552241413630974_dp, &
             4.7449266063584945_dp, 8.073063372241517_dp, &
             ! w
             2.2299256835178584e-05_dp, 6.501368491357918e-05_dp, 0.00015294213063525062_dp, &
             0.0003451303428258545_dp, 0.00074676518206715_dp, 0.0015495513359210218_dp, &
             0.003098115315442275_dp, 0.005997523818292902_dp, 0.011287704989923162_dp, &
             0.02072278731325668_dp, 0.03721253945710439_dp, 0.06551194352534437_dp, &
             0.11328896357247986_dp, 0.19277363412593645_dp, 0.3233390657908395_dp, &
             0.5357619930962504_dp, 0.8802904393288624_dp, 1.4464064256116258_dp, &
             2.4323501049508445_dp, 4.567135023797751_dp]

      CASE (680)

         aw(:) = & ! a
            [4.7716221797505265e-06_dp, 2.9166885633180936e-05_dp, 9.488213798553663e-05_dp, &
             0.0002575982926136552_dp, 0.0006375727582029251_dp, 0.0014774229588766239_dp, &
             0.0032482763223485823_dp, 0.00683617387254335_dp, 0.01386165477796786_dp, &
             0.027217146466065434_dp, 0.05195509451290508_dp, 0.09673321633202932_dp, &
             0.17613457437375343_dp, 0.31435090332573695_dp, 0.5509942722729587_dp, &
             0.9503220681177528_dp, 1.6164370030051054_dp, 2.721267610650231_dp, &
             4.570638705803211_dp, 7.853153537926083_dp, &
             ! w
             1.2647092064865676e-05_dp, 3.9370942470099255e-05_dp, 0.0001008859404259641_dp, &
             0.0002436690353121397_dp, 0.0005550198767466638_dp, 0.001200217079464494_dp, &
             0.0024840317756176177_dp, 0.004953203500815161_dp, 0.009565504281249658_dp, &
             0.017964199971957393_dp, 0.032917106613217804_dp, 0.05901000203418003_dp, &
             0.1037306466940394_dp, 0.17915556535870944_dp, 0.3046001281541802_dp, &
             0.5109813069925129_dp, 0.848989087313391_dp, 1.4088411497502464_dp, &
             2.389407421830208_dp, 4.518257900670752_dp]

      CASE (681)

         aw(:) = & ! a
            [3.4220012461253343e-06_dp, 2.147959778004825e-05_dp, 7.279310603676119e-05_dp, &
             0.00020565282870151205_dp, 0.0005261246167408688_dp, 0.0012524674143083458_dp, &
             0.0028156967270048655_dp, 0.006037955457609617_dp, 0.012441305150176067_dp, &
             0.02477143012365551_dp, 0.047869761917779176_dp, 0.09010240654908779_dp, &
             0.16566810319519037_dp, 0.2982822232230308_dp, 0.5270158751881723_dp, &
             0.9155935079377637_dp, 1.5677241559383712_dp, 2.655251198087576_dp, &
             4.484312251458938_dp, 7.743824550507109_dp, &
             ! w
             9.122675539963441e-06_dp, 2.9778293823584392e-05_dp, 8.047620084809404e-05_dp, &
             0.0002020545520884287_dp, 0.0004736622605107203_dp, 0.0010480074744703772_dp, &
             0.002210522982769574_dp, 0.00447925090915589_dp, 0.008771008342639677_dp, &
             0.01667307575204666_dp, 0.03088061808207692_dp, 0.05589150440317464_dp, &
             0.09909801695618253_dp, 0.17249244929011046_dp, 0.29535277483086253_dp, &
             0.4986596637980531_dp, 0.8333257561984203_dp, 1.3899532710740816_dp, &
             2.3677591608301287_dp, 4.493623441663668_dp]

      CASE (682)

         aw(:) = & ! a
            [2.7085752802366948e-06_dp, 1.7382467597510653e-05_dp, 6.080070171315584e-05_dp, &
             0.0001767937296843378_dp, 0.0004628802209121424_dp, 0.001122497823893844_dp, &
             0.002561977384804581_dp, 0.005563757984746035_dp, 0.011588196126166381_dp, &
             0.023288336126993475_dp, 0.04537147402492685_dp, 0.0860171390199267_dp, &
             0.15917661401386687_dp, 0.2882565665686904_dp, 0.5119750581035682_dp, &
             0.8937056737916433_dp, 1.5368944783266678_dp, 2.613322386215946_dp, &
             4.429328379529669_dp, 7.674044406590095_dp, &
             ! w
             7.256716009660908e-06_dp, 2.4610881258789402e-05_dp, 6.912830022162299e-05_dp, &
             0.00017828034510865412_dp, 0.00042625517714759857_dp, 0.0009579352414834601_dp, &
             0.002046594367698543_dp, 0.004192088581339982_dp, 0.008285086245170853_dp, &
             0.015876860886764228_dp, 0.029615502432639326_dp, 0.05394144604822861_dp, &
             0.0961839308156842_dp, 0.16827861342195652_dp, 0.2894763707362812_dp, &
             0.4907961767038131_dp, 0.823293620189285_dp, 1.377823046756512_dp, &
             2.3538353718762264_dp, 4.477780694181771_dp]

      CASE (683)

         aw(:) = & ! a
            [2.2625144552501487e-06_dp, 1.4803988063613585e-05_dp, 5.314161280887496e-05_dp, &
             0.00015804064684016368_dp, 0.00042114762589958213_dp, 0.0010356367729601466_dp, &
             0.0023906096253725638_dp, 0.005240607980956467_dp, 0.011002381447191449_dp, &
             0.022263167173878565_dp, 0.04363452639852565_dp, 0.08316223374806304_dp, &
             0.1546194098389664_dp, 0.281189489616428_dp, 0.5013340157200542_dp, &
             0.8781701413341072_dp, 1.5149498825876013_dp, 2.583405360334611_dp, &
             4.390020468388924_dp, 7.624087892633641_dp, &
             ! w
             6.088709124201673e-06_dp, 2.1330742418718056e-05_dp, 6.174883670218116e-05_dp, &
             0.0001625193575379217_dp, 0.000394389143487223_dp, 0.000896733927318685_dp, &
             0.0019342164075336617_dp, 0.0039937427909891_dp, 0.007947265479864692_dp, &
             0.015320156788974165_dp, 0.028726472441608926_dp, 0.05256489400045494_dp, &
             0.09411850623687122_dp, 0.1652810054934653_dp, 0.2852822627217757_dp, &
             0.4851674727364836_dp, 0.8160949153709705_dp, 1.3691026876884091_dp, &
             2.3438153680168_dp, 4.466380534326724_dp]

      CASE (684)

         aw(:) = & ! a
            [1.955203091601473e-06_dp, 1.3017745561090503e-05_dp, 4.776992770567534e-05_dp, &
             0.00014470449026816263_dp, 0.00039111269587157607_dp, 0.0009725079206708274_dp, &
             0.002265054036210234_dp, 0.0050022392641724395_dp, 0.010567760755692334_dp, &
             0.021498783040520778_dp, 0.04233376505798949_dp, 0.0810160083972915_dp, &
             0.15118169950727942_dp, 0.27584213892168424_dp, 0.4932603685705069_dp, &
             0.8663542475903337_dp, 1.498223917052817_dp, 2.5605617705376296_dp, &
             4.359963020458261_dp, 7.585847417979316_dp, &
             ! w
             5.28329779560503e-06_dp, 1.904173638959348e-05_dp, 5.6497487252631314e-05_dp, &
             0.00015113623424647707_dp, 0.00037113044344116075_dp, 0.0008516953482978613_dp, &
             0.0018509570175995998_dp, 0.0038459511280842847_dp, 0.007694309315140704_dp, &
             0.014901511564071684_dp, 0.028055376349691675_dp, 0.05152226438885108_dp, &
             0.09254934758140199_dp, 0.16299739161247373_dp, 0.2820792590846976_dp, &
             0.4808595080268391_dp, 0.8105752014421631_dp, 1.3624069435837665_dp, &
             2.336115802835121_dp, 4.457620854719254_dp]

      CASE (685)

         aw(:) = & ! a
            [1.7296013927399615e-06_dp, 1.1700131360704459e-05_dp, 4.376503999876392e-05_dp, &
             0.00013464613565762916_dp, 0.00036823798135210065_dp, 0.0009240475343137818_dp, &
             0.0021680463702764526_dp, 0.004817070602864445_dp, 0.010228584870823931_dp, &
             0.02089989086809332_dp, 0.041311088091684274_dp, 0.07932345711293552_dp, &
             0.14846330194635823_dp, 0.2716034403247835_dp, 0.4868467767714144_dp, &
             0.8569498485149786_dp, 1.4848891863594231_dp, 2.5423238818073304_dp, &
             4.335938426139878_dp, 7.5552567610618695_dp, &
             ! w
             4.6916084726441045e-06_dp, 1.734239403376335e-05_dp, 5.253447292715045e-05_dp, &
             0.0001424426143058947_dp, 0.0003532162460203515_dp, 0.0008167769905050501_dp, &
             0.0017860572308146631_dp, 0.003730224528022603_dp, 0.007495459864979605_dp, &
             0.014571290759966087_dp, 0.027524433103528985_dp, 0.05069516756155617_dp, &
             0.0913015702641553_dp, 0.16117755086261226_dp, 0.27952177438727244_dp, &
             0.47741383267728105_dp, 0.8061539107344943_dp, 1.3570377740022355_dp, &
             2.3299379343491493_dp, 4.450592641731354_dp]

      CASE (686)

         aw(:) = & ! a
            [1.5563818664758744e-06_dp, 1.0684116605952199e-05_dp, 4.064765302882785e-05_dp, &
             0.00012673907159957106_dp, 0.00035010787263785277_dp, 0.0008853850620702174_dp, &
             0.0020902356167353444_dp, 0.004667879917353914_dp, 0.00995427249034999_dp, &
             0.020413945873792114_dp, 0.04047891494084723_dp, 0.07794273501607235_dp, &
             0.14624079233392093_dp, 0.2681310841925191_dp, 0.48158344003487213_dp, &
             0.8492199669105642_dp, 1.4739137335306447_dp, 2.5272953041037853_dp, &
             4.316123002974757_dp, 7.530008488261143_dp, &
             ! w
             4.237033258935608e-06_dp, 1.6024488309619524e-05_dp, 4.941740594954536e-05_dp, &
             0.00013553643744074447_dp, 0.00033888506804604095_dp, 0.0007886899807933802_dp, &
             0.0017336212660410044_dp, 0.0036363718289343767_dp, 0.007333676153502949_dp, &
             0.014301869538584661_dp, 0.02709017440103725_dp, 0.05001719554531349_dp, &
             0.09027674862804429_dp, 0.15968023110151966_dp, 0.2774141864334041_dp, &
             0.47457029890766245_dp, 0.8025009118197225_dp, 1.3525976246796874_dp, &
             2.324826463838408_dp, 4.44477777581224_dp]

      CASE (687)

         aw(:) = & ! a
            [1.4188613779699578e-06_dp, 9.874365402192613e-06_dp, 3.8142086332189405e-05_dp, &
             0.0001203289370338298_dp, 0.0003353063482870663_dp, 0.0008536431572877572_dp, &
             0.0020260609343664237_dp, 0.00454436729619164_dp, 0.009726443901037062_dp, &
             0.02000923218967787_dp, 0.03978418230364895_dp, 0.07678761271750069_dp, &
             0.14437793895207782_dp, 0.2652157814584065_dp, 0.4771578976492372_dp, &
             0.8427119086546228_dp, 1.4646624471106946_dp, 2.514615229698963_dp, &
             4.299391055913488_dp, 7.508676914885726_dp, &
             ! w
             3.875962782844034e-06_dp, 1.4968654179692135e-05_dp, 4.688922692014158e-05_dp, &
             0.00012988734842581015_dp, 0.00032709239189894886_dp, 0.0007654708395712492_dp, &
             0.0016901090517996006_dp, 0.003558243765877453_dp, 0.007198632103587924_dp, &
             0.014076446205497922_dp, 0.02672607538528167_dp, 0.04944770505319626_dp, &
             0.08941447943853127_dp, 0.15841852898311726_dp, 0.2756358684711466_dp, &
             0.47216817317962617_dp, 0.7994118924742644_dp, 1.3488401461390522_dp, &
             2.3204990598628923_dp, 4.439854984621447_dp]

      CASE (688)

         aw(:) = & ! a
            [1.3068207621557417e-06_dp, 9.212301034193857e-06_dp, 3.6077782050654243e-05_dp, &
             0.00011500739057496833_dp, 0.0003229427492826557_dp, 0.0008269998002373854_dp, &
             0.0019719812010534125_dp, 0.004439942621358676_dp, 0.009533291178960537_dp, &
             0.0196653009055035_dp, 0.03919256823145737_dp, 0.07580215709742091_dp, &
             0.14278614824851238_dp, 0.26272111524073444_dp, 0.4733660643226754_dp, &
             0.8371294493366118_dp, 1.4567190605583722_dp, 2.5037186827496347_dp, &
             4.285002931524411_dp, 7.490324507832817_dp, &
             ! w
             3.5816674054485943e-06_dp, 1.4101262482462756e-05_dp, 4.478947228580631e-05_dp, &
             0.00012516093805595003_dp, 0.00031717470933051873_dp, 0.0007458650505276785_dp, &
             0.0016532480898984255_dp, 0.003491877014477954_dp, 0.007083648720003959_dp, &
             0.013884118766664413_dp, 0.026414876268207557_dp, 0.048960181920969924_dp, &
             0.08867526689094965_dp, 0.15733550325276405_dp, 0.2741076348432424_dp, &
             0.4701017642448502_dp, 0.7967523128574955_dp, 1.3456029417014697_dp, &
             2.316769496854295_dp, 4.435612368010766_dp]

      CASE (689)

         aw(:) = & ! a
            [7.719913676011955e-07_dp, 6.011880174868863e-06_dp, 2.5833115041406673e-05_dp, &
             8.793439483979573e-05_dp, 0.0002587987590789933_dp, 0.0006866221045972633_dp, &
             0.001683480458012479_dp, 0.0038771043195828425_dp, 0.008483137655494119_dp, &
             0.01778139263631939_dp, 0.0359308996111747_dp, 0.07033817179348717_dp, &
             0.13391568640227072_dp, 0.24875679620964244_dp, 0.4520556566223064_dp, &
             0.8056441782199937_dp, 1.4117790557247591_dp, 2.441909128739049_dp, &
             4.203216432929416_dp, 7.385843906709667_dp, &
             ! w
             2.175165277102975e-06_dp, 9.83686937874479e-06_dp, 3.409117948071228e-05_dp, &
             0.00010051617583215904_dp, 0.0002646118121177824_dp, 0.0006406344379419803_dp, &
             0.0014533580241736681_dp, 0.0031288694916891424_dp, 0.006450075747897291_dp, &
             0.01281757872600264_dp, 0.02467942475720512_dp, 0.046227844304853834_dp, &
             0.0845138014752303_dp, 0.15121400342164015_dp, 0.26543857578837204_dp, &
             0.45834250505310004_dp, 0.7815767477876203_dp, 1.3270937794223034_dp, &
             2.2954209095646343_dp, 4.411328312061586_dp]

      CASE (690)

         aw(:) = & ! a
            [5.767350027933078e-07_dp, 4.817969498146783e-06_dp, 2.18483421872876e-05_dp, &
             7.701733937490957e-05_dp, 0.00023221706761777325_dp, 0.000627214064392759_dp, &
             0.001559331226024905_dp, 0.003631570338125293_dp, 0.00801975002563171_dp, &
             0.01694196918554081_dp, 0.03446529360605312_dp, 0.0678648367359474_dp, &
             0.12987423262364467_dp, 0.24235782752728424_dp, 0.442240387974228_dp, &
             0.7910767630296573_dp, 1.390904281788192_dp, 2.4131024951612994_dp, &
             4.164997890797416_dp, 7.336925716603451_dp, &
             ! w
             1.660951372249949e-06_dp, 8.200747534699697e-06_dp, 2.9767649798932182e-05_dp, &
             9.023633700653982e-05_dp, 0.00024219817565194083_dp, 0.0005949966260560371_dp, &
             0.0013654767982160731_dp, 0.002967456950233177_dp, 0.006165636395223621_dp, &
             0.012334776477467158_dp, 0.023888104756525554_dp, 0.04497396126499374_dp, &
             0.08259314219416934_dp, 0.14837421782705118_dp, 0.2613984947819671_dp, &
             0.45284008435611717_dp, 0.7744514766121792_dp, 1.3183808151857417_dp, &
             2.28535676109961_dp, 4.399881040416567_dp]

      CASE (691)

         aw(:) = & ! a
            [4.7388642561389144e-07_dp, 4.1797058086417e-06_dp, 1.9662371656572396e-05_dp, &
             7.090495144500604e-05_dp, 0.00021711030969899913_dp, 0.0005930677103712356_dp, &
             0.0014873345502435555_dp, 0.003488144657184059_dp, 0.0077474283248112116_dp, &
             0.016446123987808585_dp, 0.03359572824940348_dp, 0.06639170171369385_dp, &
             0.12745893321145077_dp, 0.23852209419315182_dp, 0.4363411083235919_dp, &
             0.7823006219890122_dp, 1.37830238440943_dp, 2.395681972917204_dp, &
             4.141853570937379_dp, 7.3072720454936135_dp, &
             ! w
             1.3899318059443448e-06_dp, 7.30994082205168e-06_dp, 2.734395449496093e-05_dp, &
             8.437484078302796e-05_dp, 0.00022926593102704486_dp, 0.0005684251915448996_dp, &
             0.0013139378741909092_dp, 0.0028722257036821338_dp, 0.005996968727206591_dp, &
             0.01204723285079215_dp, 0.023415021757870382_dp, 0.04422182173025604_dp, &
             0.08143759376776773_dp, 0.14666111213623687_dp, 0.25895547493701826_dp, &
             0.4495057695676949_dp, 0.7701260580617737_dp, 1.313084434504986_dp, &
             2.279234394283262_dp, 4.39291748104259_dp]

      CASE (692)

         aw(:) = & ! a
            [4.100989232498314e-07_dp, 3.7792850055544382e-06_dp, 1.8265622126749353e-05_dp, &
             6.694535998747585e-05_dp, 0.00020722729743620524_dp, 0.0005705629536785965_dp, &
             0.0014396080104541477_dp, 0.0033926203121782852_dp, 0.007565347472925269_dp, &
             0.016113491491623846_dp, 0.03301072623762438_dp, 0.06539818512564702_dp, &
             0.12582644154342598_dp, 0.23592452788793758_dp, 0.4323392610500245_dp, &
             0.776338212526684_dp, 1.3697295133957719_dp, 2.38381787615515_dp, &
             4.126077347051756_dp, 7.28704571113364_dp, &
             ! w
             1.2217736426804093e-06_dp, 6.74353773355063e-06_dp, 2.5772627348365625e-05_dp, &
             8.053221944722147e-05_dp, 0.00022072210692964357_dp, 0.0005507667496675252_dp, &
             0.0012795253756694418_dp, 0.002808392942412839_dp, 0.005883541894372713_dp, &
             0.01185331971280634_dp, 0.023095203280387715_dp, 0.04371225453828949_dp, &
             0.08065321770760842_dp, 0.14549627510752375_dp, 0.2572917811972208_dp, &
             0.44723203630635106_dp, 0.7671731038907279_dp, 1.3094654749292953_dp, &
             2.2750490100353447_dp, 4.388157119115768_dp]

      CASE (693)

         aw(:) = & ! a
            [3.667257381192862e-07_dp, 3.5044607832491283e-06_dp, 1.729358850645397e-05_dp, &
             6.41620382205262e-05_dp, 0.00020023077513810377_dp, 0.0005545465431006062_dp, &
             0.0014055008492858977_dp, 0.003324126642405687_dp, 0.0074344280397085835_dp, &
             0.015873761135890722_dp, 0.03258826069259531_dp, 0.06467944602392317_dp, &
             0.12464362615961723_dp, 0.23403990318048892_dp, 0.4294322703842799_dp, &
             0.7720024197048408_dp, 1.3634896246569728_dp, 2.375175638978845_dp, &
             4.114578183657263_dp, 7.272296194245673_dp, &
             ! w
             1.1073897599659528e-06_dp, 6.3507294591073e-06_dp, 2.466746175720504e-05_dp, &
             7.780796963870743e-05_dp, 0.00021463130016068807_dp, 0.0005381251961009983_dp, &
             0.0012548070551591116_dp, 0.0027624157314976878_dp, 0.005801653548919791_dp, &
             0.011713045705808466_dp, 0.022863450319988144_dp, 0.04334243847262178_dp, &
             0.08008318804433762_dp, 0.14464872734719786_dp, 0.25607994984057725_dp, &
             0.44557427325753557_dp, 0.7650183902122979_dp, 1.3068231751546633_dp, &
             2.271992097074532_dp, 4.384680296598112_dp]

      CASE (694)

         aw(:) = & ! a
            [3.354963113466894e-07_dp, 3.3050371581884835e-06_dp, 1.6580517486360618e-05_dp, &
             6.210454841433054e-05_dp, 0.00019503097427437302_dp, 0.0005425956338420271_dp, &
             0.0013799720857646176_dp, 0.0032727313966212225_dp, 0.007335986456582696_dp, &
             0.01569318510958676_dp, 0.032269559816750004_dp, 0.06413653008195322_dp, &
             0.12374913088292729_dp, 0.232613219846265_dp, 0.4272296599438312_dp, &
             0.7687145982595106_dp, 1.3587546505204962_dp, 2.36861386762867_dp, &
             4.105843169048649_dp, 7.2610883454495765_dp, &
             ! w
             1.0250016238292935e-06_dp, 6.063313137015976e-06_dp, 2.385015349539664e-05_dp, &
             7.578115888234317e-05_dp, 0.000210080879734588_dp, 0.0005286508368144408_dp, &
             0.0012362350038488553_dp, 0.002727799407650043_dp, 0.005739892440563216_dp, &
             0.01160709192962607_dp, 0.02268817286993225_dp, 0.04306242402953251_dp, &
             0.079651141191263_dp, 0.1440057588408086_dp, 0.25515988558352654_dp, &
             0.4443147466825058_dp, 0.7633803090692989_dp, 1.3048134995846872_dp, &
             2.269666478851866_dp, 4.382035245212188_dp]

      CASE (695)

         aw(:) = & ! a
            [3.121687087187584e-07_dp, 3.15509043854656e-06_dp, 1.603965400395562e-05_dp, &
             6.05345347901637e-05_dp, 0.0001910464999026499_dp, 0.0005334095187725207_dp, &
             0.0013603019450858125_dp, 0.0032330539293758493_dp, 0.007259866704858265_dp, &
             0.015553365499813334_dp, 0.0320225029785451_dp, 0.0637152356941127_dp, &
             0.12305439969539929_dp, 0.23150428244429586_dp, 0.4255164170638614_dp, &
             0.7661556829190928_dp, 1.3550674449575255_dp, 2.3635017985331985_dp, &
             4.099035548971015_dp, 7.252351246292505_dp, &
             ! w
             9.634376112667552e-07_dp, 5.845738683567085e-06_dp, 2.322628639068072e-05_dp, &
             7.422680335171511e-05_dp, 0.00020657983736026953_dp, 0.0005213434569512512_dp, &
             0.0012218828159153358_dp, 0.0027010056665998637_dp, 0.005692023872490465_dp, &
             0.011524877022100855_dp, 0.022552030396178677_dp, 0.04284473861305753_dp, &
             0.0793150026366343_dp, 0.14350517179484892_dp, 0.25444311834639155_dp, &
             0.4433329866270304_dp, 0.7621028863375885_dp, 1.3032457454768733_dp, &
             2.267851897380674_dp, 4.379971436882569_dp]

      CASE (696)

         aw(:) = & ! a
            [2.9434580139190526e-07_dp, 3.039887473234396e-06_dp, 1.5621156041345503e-05_dp, &
             5.931388859807957e-05_dp, 0.0001879383731124897_dp, 0.000526226214364329_dp, &
             0.0013448910872077343_dp, 0.0032019203712171595_dp, 0.007200062439484416_dp, &
             0.015443397316210628_dp, 0.03182801442633341_dp, 0.0633833192167307_dp, &
             0.12250667357079748_dp, 0.23062945632772253_dp, 0.42416412336809184_dp, &
             0.7641349097696671_dp, 1.3521544378242327_dp, 2.359461670654601_dp, &
             4.093653884829975_dp, 7.245442841635324_dp, &
             ! w
             9.163844878516861e-07_dp, 5.677648154623573e-06_dp, 2.27411216336871e-05_dp, &
             7.30135459556111e-05_dp, 0.00020384006906185584_dp, 0.0005156139028101804_dp, &
             0.0012106122434219812_dp, 0.0026799383956946977_dp, 0.005654346159643645_dp, &
             0.011460106466806825_dp, 0.022444690284557655_dp, 0.04267298816712531_dp, &
             0.07904963176735205_dp, 0.14310975748912527_dp, 0.2538766660985174_dp, &
             0.4425567791501926_dp, 0.7610925505620193_dp, 1.3020054372000984_dp, &
             2.2664160903569384_dp, 4.378338434585723_dp]

      CASE (697)

         aw(:) = & ! a
            [2.805804082793107e-07_dp, 2.950491711590651e-06_dp, 1.5294524832459847e-05_dp, &
             5.835751095078004e-05_dp, 0.0001854966794063506_dp, 0.0005205720877313719_dp, &
             0.001332742491953696_dp, 0.003177347426631186_dp, 0.007152812858160808_dp, &
             0.015356440842625177_dp, 0.031674112010656084_dp, 0.06312050119151012_dp, &
             0.12207273231301714_dp, 0.2299360270721139_dp, 0.4230917648931671_dp, &
             0.7625318381111248_dp, 1.3498427892818028_dp, 2.356254680336709_dp, &
             4.089381045912591_dp, 7.239956935214296_dp, &
             ! w
             8.800313125655099e-07_dp, 5.546616092028536e-06_dp, 2.2360917667162985e-05_dp, &
             7.205995092666968e-05_dp, 0.00020168226711326395_dp, 0.0005110944183033088_dp, &
             0.0012017111041264781_dp, 0.002663283448126958_dp, 0.005624534587322935_dp, &
             0.011408821456866228_dp, 0.02235964579311844_dp, 0.04253683727063298_dp, &
             0.0788391630738978_dp, 0.14279601338564965_dp, 0.2534270362095085_dp, &
             0.4419404420333103_dp, 0.7602900749137744_dp, 1.3010200857113459_dp, &
             2.2652752854237397_dp, 4.377040955310725_dp]

      CASE (698)

         aw(:) = & ! a
            [2.5003705454345945e-07_dp, 2.7506585025356867e-06_dp, 1.455800325385011e-05_dp, &
             5.6188613414129365e-05_dp, 0.00017993755812697063_dp, 0.0005076618125728505_dp, &
             0.0013049410663844576_dp, 0.003121012466836724_dp, 0.007044329613188399_dp, &
             0.015156542145764723_dp, 0.031319935722695355_dp, 0.06251511350101867_dp, &
             0.121072355633586_dp, 0.228336294921602_dp, 0.4206162709239359_dp, &
             0.7588291323317048_dp, 1.344500825200748_dp, 2.348840610134819_dp, &
             4.079499633205373_dp, 7.227267122695254_dp, &
             ! w
             7.993194986512641e-07_dp, 5.251671423470672e-06_dp, 2.1498431524474387e-05_dp, &
             6.988727944617676e-05_dp, 0.00019675102255989173_dp, 0.0005007424166542977_dp, &
             0.001181286012366606_dp, 0.0026250095685235878_dp, 0.005555941237957179_dp, &
             0.011290694759850419_dp, 0.022163579406079995_dp, 0.04222269431934139_dp, &
             0.07835319878537338_dp, 0.1420711274231852_dp, 0.2523876013999611_dp, &
             0.4405149033915557_dp, 0.7584332213100451_dp, 1.2987393328305021_dp, &
             2.262634225050463_dp, 4.374037187868886_dp]

      CASE (699)

         aw(:) = & ! a
            [0.002912528593439275_dp, 0.01548387838805494_dp, 0.03868327142039642_dp, &
             0.07359696772760346_dp, 0.1219830715395445_dp, 0.18651336109281208_dp, &
             0.27110865226768727_dp, 0.3813659503887679_dp, 0.5250622903507938_dp, &
             0.7127369661872631_dp, 0.9583997086853228_dp, 1.2804583869211272_dp, &
             1.7029906701275614_dp, 2.2575275719668424_dp, 2.985629614050784_dp, &
             3.942813034813778_dp, 5.205057218138501_dp, 6.8809123894389_dp, &
             9.137755989603955_dp, 12.27260150378066_dp, 16.993101810540406_dp, &
             ! w
             0.0074900483223948955_dp, 0.017750820333776663_dp, 0.02882615963470279_dp, &
             0.041290287294848266_dp, 0.05592161752072061_dp, 0.0737852627229118_dp, &
             0.09632974562624953_dp, 0.1254714338211335_dp, 0.16366622513075452_dp, &
             0.21400291283514866_dp, 0.2803697297994583_dp, 0.3677314569449578_dp, &
             0.48254653787196883_dp, 0.6333911635673961_dp, 0.8319632752888728_dp, &
             1.094881276100746_dp, 1.447312868707311_dp, 1.9313305500629703_dp, &
             2.6286997142342043_dp, 3.741608814883541_dp, 6.071946179849819_dp]

      CASE (700)

         aw(:) = & ! a
            [0.00267378214367888_dp, 0.014219954786944013_dp, 0.035550495769850046_dp, &
             0.0677091084669288_dp, 0.11239276193583495_dp, 0.1721931043977421_dp, &
             0.25093061240883624_dp, 0.3540775849447265_dp, 0.48925482102077406_dp, &
             0.6668081866580396_dp, 0.90051910139462_dp, 1.2085451570071766_dp, &
             1.614713038391973_dp, 2.1503280006439773_dp, 2.856777537336051_dp, &
             3.7894827822055683_dp, 5.024421251113702_dp, 6.670224516794641_dp, &
             8.894364773819014_dp, 11.993728835725038_dp, 16.67465556084517_dp, &
             ! w
             0.006876665981029532_dp, 0.016309405974623603_dp, 0.026522810022282926_dp, &
             0.03807328171604787_dp, 0.051720381805334975_dp, 0.06850994848593445_dp, &
             0.08986884041935923_dp, 0.11768576157341548_dp, 0.15438141970985333_dp, &
             0.20300682324310207_dp, 0.26742008709387566_dp, 0.3525735848301676_dp, &
             0.4649375134490115_dp, 0.613127034319381_dp, 0.8089062540298723_dp, &
             1.06898173147446_dp, 1.418615523926716_dp, 1.8999430641846025_dp, &
             2.594670220646987_dp, 3.704516452733679_dp, 6.0288246751663515_dp]

      CASE (701)

         aw(:) = & ! a
            [0.0015146491737760147_dp, 0.008077578446770248_dp, 0.020299219623305678_dp, &
             0.03897162476889744_dp, 0.06542403207952556_dp, 0.10175367034403547_dp, &
             0.15113754543766353_dp, 0.21821167260561974_dp, 0.30951561954124396_dp, &
             0.4340455501733756_dp, 0.603997953286101_dp, 0.8357981794242986_dp, &
             1.151520119331662_dp, 1.580851318131861_dp, 2.1638751702975596_dp, &
             2.9552066015459304_dp, 4.0306582223099_dp, 5.499339294147026_dp, &
             7.529480986100923_dp, 10.417648515189345_dp, 14.86308806543036_dp, &
             ! w
             0.0038979756780079264_dp, 0.009296138987866744_dp, 0.015276566196327761_dp, &
             0.022285959542313886_dp, 0.030963148882563418_dp, 0.04221928459682924_dp, &
             0.05731383340559683_dp, 0.07791534299340988_dp, 0.10617973964302194_dp, &
             0.14489596550735878_dp, 0.1977230804501745_dp, 0.2695241987672367_dp, &
             0.3668223890330759_dp, 0.4984541081634047_dp, 0.6765908300356022_dp, &
             0.9185215674834927_dp, 1.250186506194148_dp, 1.714272601393396_dp, &
             2.392353892036537_dp, 3.4836105936009223_dp, 5.772628869195125_dp]

      CASE (702)

         aw(:) = & ! a
            [0.0010821191288824547_dp, 0.00578174027870024_dp, 0.014581468708472323_dp, &
             0.028149812106127912_dp, 0.047630855357227546_dp, 0.0748632213865804_dp, &
             0.11266955091322602_dp, 0.16520337502327978_dp, 0.23836733325808943_dp, &
             0.3403637601835302_dp, 0.48246148352122803_dp, 0.6800654271214926_dp, &
             0.9541932711298037_dp, 1.333518024241693_dp, 1.8572519632567737_dp, &
             2.579402937933566_dp, 3.575555741833236_dp, 4.955023113517167_dp, &
             6.886534745313462_dp, 9.666786433520393_dp, 13.991920002667486_dp, &
             ! w
             0.00278604825370396_dp, 0.00666942429734448_dp, 0.01103901261361446_dp, &
             0.016284704991226685_dp, 0.022978941257169292_dp, 0.03195047184585896_dp, &
             0.04434763310061658_dp, 0.061694821897650194_dp, 0.08598756830441465_dp, &
             0.11986583039894727_dp, 0.16687284958336243_dp, 0.2318045507930067_dp, &
             0.32118305078668047_dp, 0.4439333058024178_dp, 0.6124285789377593_dp, &
             0.8442838049251395_dp, 1.1658622068098023_dp, 1.6202614555945163_dp, &
             2.2891590291234873_dp, 3.3706338493558206_dp, 5.6420208133079734_dp]

      CASE (703)

         aw(:) = & ! a
            [0.0008512472313432425_dp, 0.0045548640886128025_dp, 0.011519454211287847_dp, &
             0.02233617991912903_dp, 0.038031227907940994_dp, 0.06027433547090194_dp, &
             0.09164991237657227_dp, 0.13598116475965577_dp, 0.198733636734997_dp, &
             0.2875666853675028_dp, 0.4131133991401827_dp, 0.59007180077374_dp, &
             0.8387139977059008_dp, 1.186977989659552_dp, 1.6734219663811696_dp, &
             2.3515700471370193_dp, 3.2967868832052964_dp, 4.618486149495916_dp, &
             6.485754181553083_dp, 9.195482690425992_dp, 13.441990875715812_dp, &
             ! w
             0.002192373569931531_dp, 0.005263723727113722_dp, 0.008761656098735913_dp, &
             0.013039335307264994_dp, 0.01862448511551076_dp, 0.02628712330705854_dp, &
             0.03709468783248873_dp, 0.05246974493351825_dp, 0.07429894626409213_dp, &
             0.10512138695293262_dp, 0.1483969314828238_dp, 0.20886431445073794_dp, &
             0.29302831654438866_dp, 0.4098559803558632_dp, 0.5718437172722399_dp, &
             0.7968250446598246_dp, 1.1114665234817016_dp, 1.5591858781562895_dp, &
             2.2218010666112384_dp, 3.296756650714444_dp, 5.556766892575554_dp]

      CASE (704)

         aw(:) = & ! a
            [0.000706181665637202_dp, 0.0037832605940279836_dp, 0.009590445123069941_dp, &
             0.01866449257764946_dp, 0.031947559450222186_dp, 0.050986704657108785_dp, &
             0.0781900087460827_dp, 0.11713410554001233_dp, 0.17295856953307964_dp, &
             0.25291832156219207_dp, 0.3671708931450052_dp, 0.5298801213301267_dp, &
             0.760746428151499_dp, 1.0871337311202325_dp, 1.5470769030245328_dp, &
             2.1936982168610255_dp, 3.102159686708988_dp, 4.381926793986992_dp, &
             6.202358299955479_dp, 8.86054490346503_dp, 13.049581916125424_dp, &
             ! w
             0.0018192654620688945_dp, 0.004378652711843947_dp, 0.007322936786695653_dp, &
             0.010978833423347286_dp, 0.01584083918432559_dp, 0.022633761587539332_dp, &
             0.03236256666552527_dp, 0.04637319722671109_dp, 0.06647194613474862_dp, &
             0.09512233609235062_dp, 0.1357183287037266_dp, 0.19294871528877855_dp, &
             0.2732955699077965_dp, 0.38574692949055506_dp, 0.542882823922086_dp, &
             0.7626975682168371_dp, 1.0720921461682396_dp, 1.5147451907325926_dp, &
             2.172617144585184_dp, 3.2427363246876157_dp, 5.494503309968946_dp]

      CASE (705)

         aw(:) = & ! a
            [0.0006059630517732656_dp, 0.0032497850589363313_dp, 0.008254857572833242_dp, &
             0.01611692190326042_dp, 0.027714088207360835_dp, 0.044498489502335536_dp, &
             0.06873997623329513_dp, 0.103821076432779_dp, 0.15462511699747838_dp, &
             0.2280890173902606_dp, 0.33399454826255653_dp, 0.4860796870810229_dp, &
             0.7035837752077377_dp, 1.0134011681732502_dp, 1.4531310223338998_dp, &
             2.075551951975214_dp, 2.95564083269993_dp, 4.202887971921243_dp, &
             5.986869699658992_dp, 8.604864682867454_dp, 12.749079488048517_dp, &
             ! w
             0.0015614572648853513_dp, 0.0037661449167832725_dp, 0.006324458258380648_dp, &
             0.009542795415729098_dp, 0.013889474376249412_dp, 0.020052830575055577_dp, &
             0.028987614866875185_dp, 0.041979430673447615_dp, 0.06077179965328263_dp, &
             0.08776803441847775_dp, 0.1263073361292365_dp, 0.1810342464369783_dp, &
             0.2584066380241655_dp, 0.3674224365145487_dp, 0.5207222466943435_dp, &
             0.7364255479986647_dp, 1.0416231554631343_dp, 1.4802133869629102_dp, &
             2.1342925843994744_dp, 3.2005943960052603_dp, 5.4459745084875015_dp]

      CASE (706)

         aw(:) = & ! a
            [0.000532273008995531_dp, 0.0028572600783288472_dp, 0.007270929683116969_dp, &
             0.014236638215925999_dp, 0.02458144975050972_dp, 0.03968089378992191_dp, &
             0.061692210640403054_dp, 0.09383947346201514_dp, 0.14079708166644644_dp, &
             0.20924239460212252_dp, 0.30864896478807224_dp, 0.4524027445775415_dp, &
             0.6593582837561238_dp, 0.9560136522539945_dp, 1.3795951312980856_dp, &
             1.9825811680572736_dp, 2.839779117516862_dp, 4.060687995093856_dp, &
             5.815063497550326_dp, 8.400358437200548_dp, 12.508099878674136_dp, &
             ! w
             0.0013718631357690669_dp, 0.0033150960728274923_dp, 0.005587367036493863_dp, &
             0.008478777888709727_dp, 0.012436201589577962_dp, 0.018117599305360377_dp, &
             0.026436169643263188_dp, 0.03862837475769106_dp, 0.056386700438688396_dp, &
             0.08206451686895867_dp, 0.11895408749309858_dp, 0.17166029934290608_dp, &
             0.24661686117090256_dp, 0.3528251858348808_dp, 0.5029717126255187_dp, &
             0.7152771534219993_dp, 1.0169910497627965_dp, 1.4522011191312796_dp, &
             2.103131043718737_dp, 3.166295213723693_dp, 5.406505108251993_dp]

      CASE (707)

         aw(:) = & ! a
            [0.0004756401331268577_dp, 0.002555412870813139_dp, 0.006513464784162022_dp, &
             0.01278672176443298_dp, 0.022160241472440705_dp, 0.03594585244615309_dp, &
             0.05620648606275404_dp, 0.08603338976071762_dp, 0.12992598514475306_dp, &
             0.19434390160992548_dp, 0.2885009228780811_dp, 0.42548429765391527_dp, &
             0.6238194389850323_dp, 0.9096622964489856_dp, 1.3199135679348086_dp, &
             1.9067852837180768_dp, 2.744928837289791_dp, 3.943841765535262_dp, &
             5.673431184795017_dp, 8.231310557456482_dp, 12.308468051499757_dp, &
             ! w
             0.0012261344973695885_dp, 0.002967989517341061_dp, 0.005018888492974004_dp, &
             0.007655451479431189_dp, 0.011306487871651802_dp, 0.01660407096271243_dp, &
             0.02442624854976384_dp, 0.035968419164715645_dp, 0.052880333531535886_dp, &
             0.07747275201189595_dp, 0.11299683760848225_dp, 0.16402160027423562_dp, &
             0.23695725624307062_dp, 0.34080478602001063_dp, 0.4882864055201595_dp, &
             0.6977070085715041_dp, 0.9964519980401344_dp, 1.4287753328272155_dp, &
             2.07701947561897_dp, 3.1375298304844623_dp, 5.373422854917393_dp]

      CASE (708)

         aw(:) = & ! a
            [0.00043065506645819203_dp, 0.002315517422070806_dp, 0.005910861730887108_dp, &
             0.011631499224766407_dp, 0.020227079342209225_dp, 0.032955252043211355_dp, &
             0.05179830098846239_dp, 0.07973385406557651_dp, 0.12111174177371571_dp, &
             0.18220510290012484_dp, 0.2720042348018754_dp, 0.4033378649152963_dp, &
             0.5944445770290845_dp, 0.8711798296951927_dp, 1.27015561887797_dp, &
             1.8433445206760664_dp, 2.6652536847868804_dp, 3.8453724749666325_dp, &
             5.5537382035583_dp, 8.088112683148747_dp, 12.139044831188432_dp, &
             ! w
             0.00111036369019524_dp, 0.0026919404252636367_dp, 0.004565885827633419_dp, &
             0.0069974035369530195_dp, 0.010399766180461586_dp, 0.015382590348554808_dp, &
             0.02279366702294027_dp, 0.03379337933152443_dp, 0.04999488149844764_dp, &
             0.07367181978477365_dp, 0.10803879981146143_dp, 0.15763210713416884_dp, &
             0.22883938894461284_dp, 0.33065869615942317_dp, 0.4758408266732641_dp, &
             0.6827621656764927_dp, 0.9789264997486924_dp, 1.408735765820335_dp, &
             2.0546432560808214_dp, 3.1128606984101044_dp, 5.345065392286852_dp]

      CASE (709)

         aw(:) = & ! a
            [0.0003939949106784257_dp, 0.0021199200095252226_dp, 0.005419082605229479_dp, &
             0.010687427388557852_dp, 0.018644192955283894_dp, 0.03050013370359804_dp, &
             0.0481674309363621_dp, 0.07452495424254636_dp, 0.11379254714935771_dp, &
             0.17208096789507601_dp, 0.2581851127309392_dp, 0.38470647150976683_dp, &
             0.5696299884851689_dp, 0.8385435956292172_dp, 1.2278001525977922_dp, &
             1.7891548298725095_dp, 2.596980705488982_dp, 3.7607544609508485_dp, &
             5.450627005441288_dp, 7.964497534517149_dp, 11.992549225640493_dp, &
             ! w
             0.0010160065520274173_dp, 0.002466728414763598_dp, 0.004195634291500846_dp, &
             0.006458082875483104_dp, 0.009653763951944621_dp, 0.014372540502803585_dp, &
             0.021435790597874885_dp, 0.031973548678985074_dp, 0.04756707324919973_dp, &
             0.0704571667488686_dp, 0.103825522086608_dp, 0.15217836717321465_dp, &
             0.2218817837045788_dp, 0.3219292605213171_dp, 0.4650948451783752_dp, &
             0.6698166547223037_dp, 0.963703094771255_dp, 1.3912893105721738_dp, &
             2.035132216082195_dp, 3.0913356591946104_dp, 5.320332409219109_dp]

      CASE (710)

         aw(:) = & ! a
            [0.0002190201476233999_dp, 0.0011847161401304554_dp, 0.003060084761396343_dp, &
             0.006136335172214816_dp, 0.0109597725831443_dp, 0.018467814404298503_dp, &
             0.030159993867000315_dp, 0.048334749228975196_dp, 0.07644683648695118_dp, &
             0.11964186116918914_dp, 0.1855350566645857_dp, 0.2853305828454066_dp, &
             0.4354219713649722_dp, 0.6596805705680682_dp, 0.9927514549523896_dp, &
             1.4849033418439166_dp, 2.2095246760539875_dp, 3.2758891700834396_dp, &
             4.854819166433461_dp, 7.245196580178857_dp, 11.13535800313594_dp, &
             ! w
             0.000565468130145394_dp, 0.0013876071557883598_dp, 0.0024100067566517175_dp, &
             0.0038312495526718225_dp, 0.005969093786937915_dp, 0.00929310973684461_dp, &
             0.01446849232337032_dp, 0.022448054187728918_dp, 0.0346209657962691_dp, &
             0.05301996131594804_dp, 0.08060759210882006_dp, 0.12167829753613657_dp, &
             0.18242930518473746_dp, 0.2717835378593405_dp, 0.4026166639626322_dp, &
             0.59372008416868_dp, 0.8733542859866111_dp, 1.2869382219003953_dp, &
             1.917797589111749_dp, 2.961572640926896_dp, 5.171428698629114_dp]

      CASE (711)

         aw(:) = & ! a
            [0.00015517344947997203_dp, 0.0008424335208536418_dp, 0.0021918115852921803_dp, &
             0.004446652568997501_dp, 0.008071288924584091_dp, 0.0138696235071661_dp, &
             0.02313830852773234_dp, 0.03789100023944764_dp, 0.061204348253109386_dp, &
             0.09773839417102638_dp, 0.15450101702605973_dp, 0.2419587075389428_dp, &
             0.3756424269293803_dp, 0.5784661086721002_dp, 0.8840901320899797_dp, &
             1.3418876792688652_dp, 2.024605927146014_dp, 3.041312973632217_dp, &
             4.563157580178045_dp, 6.8896323136008455_dp, 10.708369152416786_dp, &
             ! w
             0.0004009559297720304_dp, 0.0009911823752180814_dp, 0.0017466456327943532_dp, &
             0.002838439770140455_dp, 0.004542366816484977_dp, 0.0072667993199022575_dp, &
             0.0116005297484477_dp, 0.018409115416900507_dp, 0.028981752270524994_dp, &
             0.04523615263541352_dp, 0.07000724593704305_dp, 0.10745874153358762_dp, &
             0.16367285846708884_dp, 0.24750410051086227_dp, 0.3718503082852143_dp, &
             0.555667642795944_dp, 0.8275641592095919_dp, 1.233470550370284_dp, &
             1.8572143372678382_dp, 2.894334050531638_dp, 5.0944008838246715_dp]

      CASE (712)

         aw(:) = & ! a
            [0.00012147264200089475_dp, 0.0006613812626125256_dp, 0.0017307123666432847_dp, &
             0.003543804109616693_dp, 0.006514149721044517_dp, 0.01136169943762222_dp, &
             0.01925551859486475_dp, 0.03203007532635975_dp, 0.052522187576779344_dp, &
             0.08507990794097489_dp, 0.13631520282946505_dp, 0.21620654455954455_dp, &
             0.33970549965315155_dp, 0.5290723342685504_dp, 0.8172824781640463_dp, &
             1.253071068792377_dp, 1.908709813966847_dp, 2.893087621803284_dp, &
             4.37755416983146_dp, 6.662039749581136_dp, 10.433811187142528_dp, &
             ! w
             0.0003140777097500278_dp, 0.0007809486573442481_dp, 0.001392071768161044_dp, &
             0.0023012464750372277_dp, 0.0037571677173666358_dp, 0.006129282815059349_dp, &
             0.009958597763291138_dp, 0.016055003773824093_dp, 0.025641599315711278_dp, &
             0.040557817757522575_dp, 0.0635495940374247_dp, 0.09868686839246495_dp, &
             0.15196556657136023_dp, 0.23218224780602373_dp, 0.3522360971124796_dp, &
             0.5311823234924457_dp, 0.7978588520009688_dp, 1.1985522831317563_dp, &
             1.8174611604172524_dp, 2.850114187630499_dp, 5.0437886145056465_dp]

      CASE (713)

         aw(:) = & ! a
            [0.00010044700235109968_dp, 0.0005482349511701674_dp, 0.0014416441325587794_dp, &
             0.0029749923122860756_dp, 0.005526095076794716_dp, 0.009755543321313273_dp, &
             0.016742265881198613_dp, 0.028193907371639727_dp, 0.04677630232745053_dp, &
             0.0766128789691386_dp, 0.12402759069454374_dp, 0.19864018969200373_dp, &
             0.31497182291102216_dp, 0.49479165979551865_dp, 0.7705539131146174_dp, &
             1.1905002842753047_dp, 1.826525115183567_dp, 2.7873615929411506_dp, &
             4.244497893864711_dp, 6.498202967388789_dp, 10.235526693121074_dp, &
             ! w
             0.00025985429340269966_dp, 0.0006492964286643749_dp, 0.0011686354886819145_dp, &
             0.001959396137997961_dp, 0.0032507572779906795_dp, 0.0053845561304046675_dp, &
             0.008868118442681846_dp, 0.014471318074345456_dp, 0.02336858168319099_dp, &
             0.037340676750970296_dp, 0.05906581203253332_dp, 0.09254126300992202_dp, &
             0.14369418985372256_dp, 0.22127180608899147_dp, 0.3381670236492554_dp, &
             0.5135021981275897_dp, 0.7762837135098907_dp, 1.1730686840762399_dp, &
             1.788349203299166_dp, 2.8176773541740805_dp, 5.0066848839663445_dp]

      CASE (714)

         aw(:) = & ! a
            [8.599442188034258e-05_dp, 0.000470350900814204_dp, 0.0012421341212631414_dp, &
             0.002580754204023296_dp, 0.004837119681568471_dp, 0.008626855764683935_dp, &
             0.014960667614142279_dp, 0.025449977697599838_dp, 0.042630062483500444_dp, &
             0.07045147959964136_dp, 0.11501474612676975_dp, 0.18565910655166207_dp, &
             0.29656641066373673_dp, 0.4691153814629614_dp, 0.7353418333403009_dp, &
             1.1430865744051124_dp, 1.76393151870596_dp, 2.7064729119448008_dp, &
             4.14230107530687_dp, 6.3719596992717324_dp, 10.082359662962014_dp, &
             ! w
             0.00022257021619058302_dp, 0.0005585177734638947_dp, 0.0010137514175097924_dp, &
             0.0017204506800246545_dp, 0.0028928179412600332_dp, 0.004851772621472659_dp, &
             0.00807914994440924_dp, 0.013314000515191627_dp, 0.021692609285552854_dp, &
             0.034949206147216735_dp, 0.05570768038394759_dp, 0.08790625574110784_dp, &
             0.13741509522253004_dp, 0.21293868905721705_dp, 0.3273605237131328_dp, &
             0.4998518335187192_dp, 0.7595502031932437_dp, 1.153229578317593_dp, &
             1.7656243978325539_dp, 2.792323837820332_dp, 4.977696280828467_dp]

      CASE (715)

         aw(:) = & ! a
            [7.540797086138839e-05_dp, 0.00041323116155337535_dp, 0.00109547398488451_dp, &
             0.0022898773494122934_dp, 0.004326073041963684_dp, 0.007784032902941941_dp, &
             0.013620400601584133_dp, 0.02337015087416255_dp, 0.03946424807687831_dp, &
             0.06571422926866606_dp, 0.10803975588548556_dp, 0.17555156500650804_dp, &
             0.2821533143971902_dp, 0.44890140513320964_dp, 0.7074836019137458_dp, &
             1.1054040446449467_dp, 1.7139784119319859_dp, 2.6416810161420936_dp, &
             4.060180999003668_dp, 6.2702522708681_dp, 9.958712550509539_dp, &
             ! w
             0.000195252119193883_dp, 0.000491841469415062_dp, 0.000899463243933757_dp, &
             0.0015428467443871746_dp, 0.0026242036972888073_dp, 0.004447885595551735_dp, &
             0.007475499776797756_dp, 0.012421262454536815_dp, 0.02039032445087072_dp, &
             0.03307857916623063_dp, 0.053064785821927786_dp, 0.08423760254500576_dp, &
             0.13241862701918977_dp, 0.20627478095816168_dp, 0.31867881474028964_dp, &
             0.4888392749696272_dp, 0.7460001633741213_dp, 1.137115520208229_dp, &
             1.7471257751603202_dp, 2.7716629434799165_dp, 4.9540811400439475_dp]

      CASE (716)

         aw(:) = & ! a
            [6.729643786378191e-05_dp, 0.00036941713801818963_dp, 0.000982743648624791_dp, &
             0.002065553108023886_dp, 0.00393007800116474_dp, 0.007127084830737928_dp, &
             0.012568949878724436_dp, 0.02172788237991511_dp, 0.03694877587872289_dp, &
             0.06192782168070114_dp, 0.10243380613489918_dp, 0.167385819384109_dp, &
             0.2704529437371129_dp, 0.4324182512644755_dp, 0.6846724893717545_dp, &
             1.0744302470741618_dp, 1.6727755057013247_dp, 2.5880728031271683_dp, &
             3.992054110974834_dp, 6.185690851191553_dp, 9.855736649525161_dp, &
             ! w
             0.00017431525736926246_dp, 0.0004406288539898687_dp, 0.0008113179024998627_dp, &
             0.0014049751563060245_dp, 0.0024139165345136693_dp, 0.004128942576256952_dp, &
             0.006995057826103786_dp, 0.011705825328993887_dp, 0.019340237024596_dp, &
             0.03156174211027009_dp, 0.05091064097770643_dp, 0.08123300118908505_dp, &
             0.1283081856801461_dp, 0.20076964370622172_dp, 0.311478928072386_dp, &
             0.4796740514736403_dp, 0.7346878745529786_dp, 1.12362789176655_dp, &
             1.7316133945678072_dp, 2.7543213284758723_dp, 4.934265345112631_dp]

      CASE (717)

         aw(:) = & ! a
            [6.0868989705923946e-05_dp, 0.00033466527259304297_dp, 0.0008931609466182841_dp, &
             0.0018867521999880012_dp, 0.0036130815932197253_dp, 0.0065984003015285885_dp, &
             0.01171793245060595_dp, 0.02039107070830741_dp, 0.03488995726268834_dp, &
             0.05881281922101638_dp, 0.09779970387493107_dp, 0.16060542497340377_dp, &
             0.26069706572986057_dp, 0.418621190967853_dp, 0.6655101626463257_dp, &
             1.0483250193870877_dp, 1.6379448418403901_dp, 2.5426344047167566_dp, &
             3.9341770026864693_dp, 6.113716577533663_dp, 9.767962121034467_dp, &
             ! w
             0.00015772147273873319_dp, 0.0003999596931299437_dp, 0.0007410568415063558_dp, &
             0.0012944266899096982_dp, 0.002244029472612469_dp, 0.0038693093037720627_dp, &
             0.0066012947792468855_dp, 0.011115964328758596_dp, 0.01846985189215927_dp, &
             0.03029838650471412_dp, 0.049108460638350844_dp, 0.078708886835214_dp, &
             0.12484171636309181_dp, 0.1961102483160019_dp, 0.3053647957640326_dp, &
             0.47186724243669576_dp, 0.7250262838503878_dp, 1.1120827118535555_dp, &
             1.7183136759905717_dp, 2.739441351104487_dp, 4.91726629929826_dp]

      CASE (718)

         aw(:) = & ! a
            [5.564168523389408e-05_dp, 0.0003063767219819402_dp, 0.000820112933650957_dp, &
             0.001740547561579862_dp, 0.003352848507146081_dp, 0.006162292801037561_dp, &
             0.011012318975462503_dp, 0.019277015860940546_dp, 0.03316587125100979_dp, &
             0.05619238773363049_dp, 0.09388481739853197_dp, 0.15485475069686305_dp, &
             0.2523924548090117_dp, 0.4068366053084554_dp, 0.6490913884563478_dp, &
             1.0258926665366521_dp, 1.6079360107994058_dp, 2.5033948043719034_dp, &
             3.8840950722938197_dp, 6.051333589748005_dp, 9.691788275891206_dp, &
             ! w
             0.00014422331203087961_dp, 0.000366817828320172_dp, 0.0006836027200453245_dp, &
             0.0012035377444222029_dp, 0.0021034007134757096_dp, 0.0036529349024215893_dp, &
             0.006271171436251541_dp, 0.010618842583150186_dp, 0.017732875959945806_dp, &
             0.029224115132592453_dp, 0.047569993405388866_dp, 0.07654627577754439_dp, &
             0.12186163689246807_dp, 0.19209196968868958_dp, 0.3000765188025008_dp, &
             0.4650968948346712_dp, 0.7166276244326747_dp, 1.1020270430120542_dp, &
             1.7067134247371432_dp, 2.7264535603396656_dp, 4.902431804326037_dp]

      CASE (719)

         aw(:) = & ! a
            [3.083377978935453e-05_dp, 0.0001716918817843317_dp, 0.00047016234008688887_dp, &
             0.0010330577445665406_dp, 0.0020756637538341624_dp, 0.003985744196797509_dp, &
             0.007428346797028242_dp, 0.013520613055502718_dp, 0.02411162341928319_dp, &
             0.04222055329478319_dp, 0.07271445563788197_dp, 0.12334656570201324_dp, &
             0.20633331010896683_dp, 0.3407331791900437_dp, 0.5560235495363512_dp, &
             0.8975058425831819_dp, 1.4346744226900925_dp, 2.275061638453358_dp, &
             3.5907050145294876_dp, 5.683864561369749_dp, 9.241189399766812_dp, &
             ! w
             8.011632840203526e-05_dp, 0.00020840708614676008_dp, 0.00040557197218575364_dp, &
             0.0007550993662412823_dp, 0.0013930606728961821_dp, 0.0025351052964841412_dp, &
             0.004531749822278851_dp, 0.007953849399608116_dp, 0.013720351878088203_dp, &
             0.02329179212333578_dp, 0.038962544293425896_dp, 0.06429900804198811_dp, &
             0.10479269780388148_dp, 0.16883229334972397_dp, 0.2691642694828368_dp, &
             0.4251652810015576_dp, 0.6666972571667594_dp, 1.0418492388024758_dp, &
             1.6369572152923662_dp, 2.648162791712954_dp, 4.813063877831505_dp]

      CASE (720)

         aw(:) = & ! a
            [2.1843886267148253e-05_dp, 0.0001226115344584977_dp, 0.0003412445362479306_dp, &
             0.0007677866770065332_dp, 0.0015851381803132431_dp, 0.0031267123806553525_dp, &
             0.005974531545979752_dp, 0.01112401268704387_dp, 0.02024997142908746_dp, &
             0.036127839190946226_dp, 0.06329257560393403_dp, 0.10905857618050065_dp, &
             0.1850834948879888_dp, 0.3097474479418764_dp, 0.5117566118040039_dp, &
             0.8356186163043576_dp, 1.3501405787534337_dp, 2.1624585651509625_dp, &
             3.444683730574412_dp, 5.499599959573985_dp, 9.01395053028111_dp, &
             ! w
             5.685567373893893e-05_dp, 0.0001502882705863255_dp, 0.00030133892991527587_dp, &
             0.0005813318543203359_dp, 0.0011073624809622472_dp, 0.002070120939024701_dp, &
             0.0037871123913737793_dp, 0.006784157156423117_dp, 0.011919636682893794_dp, &
             0.02057535159513807_dp, 0.03494775205842657_dp, 0.05848847775239787_dp, &
             0.09656611035083078_dp, 0.15745738782834195_dp, 0.2538423170885136_dp, &
             0.40512892282713914_dp, 0.6413709506048875_dp, 1.0110486304571666_dp, &
             1.601018295015051_dp, 2.607690435600062_dp, 4.766900034742349_dp]

      CASE (721)

         aw(:) = & ! a
            [1.7112421593039018e-05_dp, 9.667887908343418e-05_dp, 0.0002725993795411437_dp, &
             0.0006247505069310555_dp, 0.001316254650473843_dp, 0.0026473835104980285_dp, &
             0.005149190726116361_dp, 0.009741454936799017_dp, 0.01798935036656839_dp, &
             0.03251317395340915_dp, 0.05763422839237732_dp, 0.10038164215363311_dp, &
             0.17204611425473573_dp, 0.29055757932899445_dp, 0.4841046235435331_dp, &
             0.7966550263577111_dp, 1.2965398845963385_dp, 2.0906096632109947_dp, &
             3.351008538943259_dp, 5.380872387007378_dp, 8.86704752750853_dp, &
             ! w
             4.4602503512288176e-05_dp, 0.0001194339909226873_dp, 0.0002451466441127063_dp, &
             0.0004855039030383321_dp, 0.0009460416633712225_dp, 0.0018021710205014941_dp, &
             0.0033505882612748705_dp, 0.006088166580896316_dp, 0.010833865291437199_dp, &
             0.018917650224075534_dp, 0.03247072571972182_dp, 0.05486721783516996_dp, &
             0.0913912127189436_dp, 0.1502403237361068_dp, 0.24404374097076204_dp, &
             0.3922227747900555_dp, 0.6249528853887316_dp, 0.9909750652499932_dp, &
             1.5775040719445264_dp, 2.581156129169046_dp, 4.73664626737336_dp]

      CASE (722)

         aw(:) = & ! a
            [1.41649915911182e-05_dp, 8.047399413750059e-05_dp, 0.00022943618302964997_dp, &
             0.0005339007069024527_dp, 0.0011432774315932592_dp, 0.0023348741837511846_dp, &
             0.004604210346816311_dp, 0.008817847557679646_dp, 0.01646313255382169_dp, &
             0.03004934363197619_dp, 0.053743722552847054_dp, 0.0943682075397624_dp, &
             0.16294507398423763_dp, 0.27707255577447965_dp, 0.46455489062468663_dp, &
             0.7689553326998437_dp, 1.2582438544762815_dp, 2.0390486035596584_dp, &
             3.2835296248615986_dp, 5.295084122334434_dp, 8.760654763508924_dp, &
             ! w
             3.6964108046494346e-05_dp, 0.00010008079766551082_dp, 0.0002094623058182378_dp, &
             0.00042356799284893395_dp, 0.0008399481991473502_dp, 0.0016233635819141474_dp, &
             0.0030556998175141504_dp, 0.005612973525604237_dp, 0.010085500817981104_dp, &
             0.01776528598319665_dp, 0.030735351685566282_dp, 0.05231204673070483_dp, &
             0.08771573443395678_dp, 0.1450832612696229_dp, 0.23700291350613867_dp, &
             0.3829018510939877_dp, 0.6130422282941844_dp, 0.9763576546764727_dp, &
             1.5603337233482597_dp, 2.561752368669122_dp, 4.7145282901975465_dp]

      CASE (723)

         aw(:) = & ! a
            [1.2140731745168948e-05_dp, 6.931545434710285e-05_dp, 0.00019955632380778618_dp, &
             0.000470471943586975_dp, 0.0010212335336768094_dp, 0.002112002440951563_dp, &
             0.004211619357874433_dp, 0.008146400232491498_dp, 0.015344430201433154_dp, &
             0.028229924153002437_dp, 0.05085139521147477_dp, 0.0898702238679796_dp, &
             0.15609954680476318_dp, 0.26687773206277093_dp, 0.4497061311932802_dp, &
             0.747827038143223_dp, 1.2289212611940055_dp, 1.999435581055973_dp, &
             3.231537289651325_dp, 5.22882940246789_dp, 8.678342096912912_dp, &
             ! w
             3.171504409223711e-05_dp, 8.671189740124787e-05_dp, 0.00018455221870117844_dp, &
             0.00037970179145758684_dp, 0.0007637686985134399_dp, 0.0014935084964177556_dp, &
             0.002839499680691824_dp, 0.0052616976430052784_dp, 0.009528222158219348_dp, &
             0.016901482204415088_dp, 0.02942669999607208_dp, 0.050374577444331735_dp, &
             0.0849146953860281_dp, 0.14113483109292968_dp, 0.23158915664320742_dp, &
             0.3757070376842399_dp, 0.6038167131347285_dp, 0.9650029321138218_dp, &
             1.5469674946355616_dp, 2.5466306343287823_dp, 4.697294562839441_dp]

      CASE (724)

         aw(:) = & ! a
            [1.0658693363771162e-05_dp, 6.112716585570099e-05_dp, 0.00017752753948224734_dp, &
             0.00042336093012651987_dp, 0.0009297709482899602_dp, 0.0019434695866266387_dp, &
             0.003912268389572392_dp, 0.007630569595321666_dp, 0.014479204405932394_dp, &
             0.026814217807765356_dp, 0.048588528520964645_dp, 0.08633366861709256_dp, &
             0.15069290097592897_dp, 0.2587925858841355_dp, 0.43788580983815056_dp, &
             0.7309503643078487_dp, 1.2054270297429985_dp, 1.9676098500571977_dp, &
             3.1896685541669147_dp, 5.175374708251614_dp, 8.611837673854831_dp, &
             ! w
             2.7870025895968315e-05_dp, 7.687434278540856e-05_dp, 0.00016605276560382017_dp, &
             0.0003467205740320005_dp, 0.0007058403459088256_dp, 0.001393846710751239_dp, &
             0.0026722805230958197_dp, 0.004988175081253384_dp, 0.009091702553352312_dp, &
             0.016221225232550988_dp, 0.028391097046579633_dp, 0.04883453628449765_dp, &
             0.08267913787571335_dp, 0.13797169422191538_dp, 0.2272371892571991_dp, &
             0.3699052175608948_dp, 0.5963567025351203_dp, 0.955799820659188_dp, &
             1.5361153935084064_dp, 2.534342063449899_dp, 4.683291704847945_dp]

      CASE (725)

         aw(:) = & ! a
            [9.523422321680008e-06_dp, 5.4841980611249935e-05_dp, 0.00016054759716969163_dp, &
             0.0003868065692700672_dp, 0.0008582460905608223_dp, 0.001810653111185749_dp, &
             0.003674681486140931_dp, 0.0072185623456420146_dp, 0.013784195879131197_dp, &
             0.025671228041798885_dp, 0.04675318103172375_dp, 0.08345334093249553_dp, &
             0.14627284656826584_dp, 0.2521600357776338_dp, 0.42815871953217866_dp, &
             0.7170227339266184_dp, 1.185988421129549_dp, 1.9412182141698122_dp, &
             3.15488142596253_dp, 5.130891571680298_dp, 8.5564298943405_dp, &
             ! w
             2.4923318573050286e-05_dp, 6.930447246731093e-05_dp, 0.0001517000611901249_dp, &
             0.0003208560198861701_dp, 0.0006599729645970442_dp, 0.00131431767943587_dp, &
             0.0025379692153977384_dp, 0.00476723591416322_dp, 0.008737332561899699_dp, &
             0.015666500095688544_dp, 0.027543154559394063_dp, 0.047568873302204266_dp, &
             0.08083561242324293_dp, 0.13535508794664095_dp, 0.22362682710014248_dp, &
             0.36507951277924044_dp, 0.5901374351727159_dp, 0.9481124845424012_dp, &
             1.5270376416443954_dp, 2.5240548807382743_dp, 4.671570784369025_dp]

      CASE (726)

         aw(:) = & ! a
            [8.623950337068422e-06_dp, 4.985304939579605e-05_dp, 0.00014701795356450215_dp, &
             0.0003575054466307091_dp, 0.0008005138019903064_dp, 0.0017027202952697095_dp, &
             0.003480415080248477_dp, 0.00687982354848507_dp, 0.01320998071247325_dp, &
             0.02472274943534449_dp, 0.04522416038575763_dp, 0.08104519395603621_dp, &
             0.14256542105576472_dp, 0.2465804324709031_dp, 0.41995388605358513_dp, &
             0.705246103249046_dp, 1.16951589655553_dp, 1.9188103036024668_dp, &
             3.125296457362658_dp, 5.093009953967638_dp, 8.509197628080098_dp, &
             ! w
             2.2587687287924285e-05_dp, 6.32822858450486e-05_dp, 0.00014019613962026013_dp, &
             0.0002999269735419499_dp, 0.0006225473158553827_dp, 0.00124898815400276_dp, &
             0.0024270174634388017_dp, 0.004583833185422012_dp, 0.00844189944275594_dp, &
             0.015202245018859951_dp, 0.026831019702146748_dp, 0.04650253354001216_dp, &
             0.0792778855474222_dp, 0.13313821296807732_dp, 0.22056050745066122_dp, &
             0.36097185952843863_dp, 0.5848331240804111_dp, 0.9415452227973429_dp, &
             1.5192730308901936_dp, 2.515250046209582_dp, 4.661539767234837_dp]

      CASE (727)

         aw(:) = & ! a
            [7.892434153316376e-06_dp, 4.5788817789895694e-05_dp, 0.00013595708158935886_dp, &
             0.0003334197450044836_dp, 0.0007527599584934758_dp, 0.001612903287966996_dp, &
             0.0033178729718671596_dp, 0.0065950281879726185_dp, 0.012725129681698973_dp, &
             0.023918803241542497_dp, 0.043923665303214225_dp, 0.07899058578567784_dp, &
             0.13939334710808485_dp, 0.24179426233703347_dp, 0.4128993522007532_dp, &
             0.6950990600003744_dp, 1.1552956846103808_dp, 1.8994336900335373_dp, &
             3.0996769096229384_dp, 5.060167666236354_dp, 8.46821297892514_dp, &
             ! w
             2.068746536878967e-05_dp, 5.8366174355884795e-05_dp, 0.0001307403941695114_dp, &
             0.0002825766070063738_dp, 0.0005912922369900919_dp, 0.001194107145135195_dp, &
             0.002333350602644015_dp, 0.004428340557145883_dp, 0.008190478884113032_dp, &
             0.014805816746590095_dp, 0.026221068552197097_dp, 0.04558666377887202_dp, &
             0.07793656848444662_dp, 0.13122487471400268_dp, 0.21790838526485265_dp, &
             0.3574121831292319_dp, 0.5802285309385055_dp, 0.935836061127093_dp, &
             1.5125157488527603_dp, 2.507583126606479_dp, 4.6528058410856445_dp]

      CASE (728)

         aw(:) = & ! a
            [4.418388821010736e-06_dp, 2.6370994115626698e-05_dp, 8.243143951282628e-05_dp, &
             0.00021459032298976595_dp, 0.0005120728769743897_dp, 0.0011509921719024172_dp, &
             0.0024667437155003764_dp, 0.005079773663539318_dp, 0.010108773048559434_dp, &
             0.019525542332661537_dp, 0.0367361458456207_dp, 0.06751879269641345_dp, &
             0.1215176047961805_dp, 0.21459459998503147_dp, 0.37250040508925586_dp, &
             0.6365849539892988_dp, 1.0727785358893198_dp, 1.7863717442812288_dp, &
             2.9494801311566308_dp, 4.8668903872545135_dp, 8.226328780379669_dp, &
             ! w
             1.1651168641595598e-05_dp, 3.47038903064258e-05_dp, 8.409306993493593e-05_dp, &
             0.0001944654163319163_dp, 0.00042869180768247654_dp, 0.0009030286038907991_dp, &
             0.0018284541239802245_dp, 0.0035783353694742244_dp, 0.00679891264113387_dp, &
             0.012587133020572096_dp, 0.022772912159559053_dp, 0.040361596251024597_dp, &
             0.07022015730768517_dp, 0.12013312348093634_dp, 0.20242563801366176_dp, &
             0.33649835211553897_dp, 0.553022018683502_dp, 0.9019421184630673_dp, &
             1.472256887039661_dp, 2.461817935667705_dp, 4.600684550507741_dp]

      CASE (729)

         aw(:) = & ! a
            [3.156256121923772e-06_dp, 1.9242434704272226e-05_dp, 6.233304382419193e-05_dp, &
             0.0001685050178445759_dp, 0.0004155405754293711_dp, 0.0009600208002422959_dp, &
             0.0021054253031310994_dp, 0.004421586945633307_dp, 0.008949267732670357_dp, &
             0.017543818174283535_dp, 0.03344261601417565_dp, 0.062187606675033674_dp, &
             0.11310450423892396_dp, 0.20164593839289838_dp, 0.3530678878135004_dp, &
             0.6081742263737608_dp, 1.032376143553304_dp, 1.7306037439272326_dp, &
             2.8749280837256266_dp, 4.770467335940883_dp, 8.10520062782195_dp, &
             ! w
             8.36087612544397e-06_dp, 2.5903843244298178e-05_dp, 6.599816207385965e-05_dp, &
             0.00015871880324276895_dp, 0.0003603606461260279_dp, 0.0007772667060587587_dp, &
             0.0016052292293345705_dp, 0.003195022250005465_dp, 0.006160410671487519_dp, &
             0.011553343842763275_dp, 0.021143966741668314_dp, 0.0378623291094048_dp, &
             0.06648730136214098_dp, 0.11471189430044366_dp, 0.19478687103085798_dp, &
             0.326091944155697_dp, 0.5393821415026232_dp, 0.8848416471702162_dp, &
             1.4518491064025734_dp, 2.4385598969989633_dp, 4.574204286858721_dp]

      CASE (730)

         aw(:) = & ! a
            [2.490228753343335e-06_dp, 1.545310683301487e-05_dp, 5.147561661405814e-05_dp, &
             0.00014306438741912617_dp, 0.00036111419445202165_dp, 0.0008503450649730482_dp, &
             0.0018946259298567522_dp, 0.004032367859369506_dp, 0.008255517085201074_dp, &
             0.016345891277611956_dp, 0.03143356623105696_dp, 0.05890916538856094_dp, &
             0.10789312492987933_dp, 0.19357241929071503_dp, 0.340879871340005_dp, &
             0.5902599111639669_dp, 1.0067787286326135_dp, 1.6951227225099983_dp, &
             2.8273265139418666_dp, 4.708723869264849_dp, 8.027471977464657_dp, &
             ! w
             6.621978886421658e-06_dp, 2.118259133959032e-05_dp, 5.600483674307416e-05_dp, &
             0.00013842413222700403_dp, 0.0003207530943875525_dp, 0.0007031741340204698_dp, &
             0.0014719300135582409_dp, 0.002963468047849752_dp, 0.005770798227948436_dp, &
             0.010916894304410379_dp, 0.020133123016274835_dp, 0.03630029727846262_dp, &
             0.06413915502511082_dp, 0.11128157698947921_dp, 0.18992746343708172_dp, &
             0.31943978130485484_dp, 0.5306255577235727_dp, 0.8738237267596163_dp, &
             1.4386647512978625_dp, 2.4235121477717914_dp, 4.557074436966679_dp]

      CASE (731)

         aw(:) = & ! a
            [2.0743271806678598e-06_dp, 1.3073066405216227e-05_dp, 4.45673502581949e-05_dp, &
             0.00012660782684807668_dp, 0.00032535893157724117_dp, 0.0007773352089554563_dp, &
             0.0017527248978991834_dp, 0.0037678609628291116_dp, 0.007780175804185964_dp, &
             0.015519208114120065_dp, 0.030038358779849415_dp, 0.056619610071248064_dp, &
             0.1042353544726883_dp, 0.1878801249953549_dp, 0.33225157016566387_dp, &
             0.5775311972238041_dp, 0.9885311982867303_dp, 1.66975650498371_dp, &
             2.7932114967129467_dp, 4.664386299049412_dp, 7.9715740003446784_dp, &
             ! w
             5.534883336124387e-06_dp, 1.819505046125667e-05_dp, 4.953744928212405e-05_dp, &
             0.00012502596147360274_dp, 0.0002942194229029913_dp, 0.000652966679398471_dp, &
             0.0013807422709053692_dp, 0.002803780477493753_dp, 0.005500215110149796_dp, &
             0.010472143028510765_dp, 0.019422847999362777_dp, 0.03519729563841886_dp, &
             0.06247364102101911_dp, 0.10883861931331701_dp, 0.18645399322934625_dp, &
             0.31466901581062295_dp, 0.5243270159619963_dp, 0.8658789178529748_dp, &
             1.4291400665938667_dp, 2.412630288083889_dp, 4.544688126332677_dp]

      CASE (732)

         aw(:) = & ! a
            [1.7880661850330878e-06_dp, 1.1426865797785554e-05_dp, 3.973656979913143e-05_dp, &
             0.0001149450882641018_dp, 0.00029970866407865885_dp, 0.0007244194457284172_dp, &
             0.001648992249613456_dp, 0.0035730901919225648_dp, 0.007427963404383562_dp, &
             0.014903325150778038_dp, 0.028993943040980897_dp, 0.05489842559042218_dp, &
             0.10147516741946602_dp, 0.18357002192647398_dp, 0.32569831798670973_dp, &
             0.567836953219493_dp, 0.9745995041907134_dp, 1.6503478224655286_dp, &
             2.767060663594186_dp, 4.630349040912813_dp, 7.928615027058954_dp, &
             ! w
             4.7859595646738815e-06_dp, 1.6115511465714447e-05_dp, 4.495185373907567e-05_dp, &
             0.00011537812847398559_dp, 0.00027489721014209244_dp, 0.0006160826551585012_dp, &
             0.0013132657702431912_dp, 0.0026848857613003463_dp, 0.005297675606344371_dp, &
             0.010137669852892417_dp, 0.018886462065650853_dp, 0.034361219641097944_dp, &
             0.0612069270332057_dp, 0.10697494243849981_dp, 0.18379681313796506_dp, &
             0.3110102715485752_dp, 0.5194858898921757_dp, 0.859761042211407_dp, &
             1.421795342285459_dp, 2.4042326017504716_dp, 4.535130114389243_dp]

      CASE (733)

         aw(:) = & ! a
            [1.578070264923161e-06_dp, 1.0214076570537404e-05_dp, 3.614356330924491e-05_dp, &
             0.00010617241489885746_dp, 0.00028022041228022534_dp, 0.0006838794615998448_dp, &
             0.0015689678224966714_dp, 0.003421953785852066_dp, 0.007153285144333728_dp, &
             0.014420926998650425_dp, 0.028172763274896116_dp, 0.05354054794868486_dp, &
             0.0992910195412903_dp, 0.1801501757402018_dp, 0.320485972250225_dp, &
             0.5601094200660045_dp, 0.9634724618523698_dp, 1.6348197568083445_dp, &
             2.7461079442704808_dp, 4.603045531450228_dp, 7.894124987549242_dp, &
             ! w
             4.236143441905124e-06_dp, 1.457490935621372e-05_dp, 4.15010284890946e-05_dp, &
             0.00010802588165451714_dp, 0.00026003860422588784_dp, 0.0005875180406029854_dp, &
             0.001260703863713063_dp, 0.002591812190466208_dp, 0.005138444343316438_dp, &
             0.0098737293898921_dp, 0.018461780606812684_dp, 0.03369729207784295_dp, &
             0.06019833694058569_dp, 0.1054874375322657_dp, 0.18167130366703524_dp, &
             0.3080777879106673_dp, 0.5155989054542306_dp, 0.8548416659310465_dp, &
             1.4158828947318824_dp, 2.39746842019197_dp, 4.527431730864511_dp]

      CASE (734)

         aw(:) = & ! a
            [1.4169226226006832e-06_dp, 9.27985251379176e-06_dp, 3.335228720132515e-05_dp, &
             9.929072196347318e-05_dp, 0.0002648027651686561_dp, 0.0006515824131924677_dp, &
             0.00150484501946044_dp, 0.003300259537112201_dp, 0.006931195169317874_dp, &
             0.014029479989673037_dp, 0.02750430351154978_dp, 0.05243211475719941_dp, &
             0.09750366347132532_dp, 0.17734536401362122_dp, 0.31620245301651473_dp, &
             0.553747467240439_dp, 0.9542970104972216_dp, 1.6219971007975837_dp, &
             2.7287849968060702_dp, 4.580450233040749_dp, 7.865562116292796_dp, &
             ! w
             3.813949284825861e-06_dp, 1.3382219027813696e-05_dp, 3.8792892066697445e-05_dp, &
             0.00010219479266568074_dp, 0.0002481649340811796_dp, 0.0005645569714198488_dp, &
             0.0012182479746439114_dp, 0.0025163249797076574_dp, 0.0050088425717549016_dp, &
             0.009658236738479332_dp, 0.018114100805076434_dp, 0.03315241383257671_dp, &
             0.05936877304060796_dp, 0.10426152561847961_dp, 0.17991641689107252_dp, &
             0.30565268416072566_dp, 0.5123798012920056_dp, 0.8507625950882847_dp, &
             1.4109759008476357_dp, 2.3918517249250004_dp, 4.521039589644589_dp]

      CASE (735)

         aw(:) = & ! a
            [1.2890398735339655e-06_dp, 8.535915852543813e-06_dp, 3.1112500056327803e-05_dp, &
             9.372136779051264e-05_dp, 0.000252233454939128_dp, 0.000625093790119734_dp, &
             0.0014519939877786728_dp, 0.0031995411178805957_dp, 0.006746735710423984_dp, &
             0.013703364725020893_dp, 0.026945918034351805_dp, 0.05150401624859208_dp, &
             0.09600394733322477_dp, 0.17498749451122197_dp, 0.3125954064908613_dp, &
             0.5483820745310821_dp, 0.9465483343520941_dp, 1.611155459243661_dp, &
             2.71412351189476_dp, 4.561310900812292_dp, 7.841353471041189_dp, &
             ! w
             3.4787187450341595e-06_dp, 1.242812986473455e-05_dp, 3.660032698179687e-05_dp, &
             9.743089600232888e-05_dp, 0.00023840162371620172_dp, 0.0005455818205371704_dp, &
             0.0011830172342520767_dp, 0.0024534655033035357_dp, 0.004900596483189334_dp, &
             0.009477780089911857_dp, 0.017822272956183854_dp, 0.032694118591802365_dp, &
             0.05866972986833659_dp, 0.10322675413309865_dp, 0.17843288505979196_dp, &
             0.3035997453960333_dp, 0.5096513882405694_dp, 0.8473017462604504_dp, &
             1.4068093946681641_dp, 2.3870805960095005_dp, 4.515609946895473_dp]

      CASE (736)

         aw(:) = & ! a
            [1.1848840139033052e-06_dp, 7.928088911749566e-06_dp, 2.9269703388652913e-05_dp, &
             8.910420219507101e-05_dp, 0.00024174577469616895_dp, 0.000602876041777134_dp, &
             0.001407473526206281_dp, 0.0031143929338972405_dp, 0.006590314556407938_dp, &
             0.013426088394369429_dp, 0.02647005871107781_dp, 0.05071147138822162_dp, &
             0.09472094736384139_dp, 0.17296707080961687_dp, 0.3095000722938025_dp, &
             0.543771813076156_dp, 0.9398824410794905_dp, 1.601819250330051_dp, &
             2.7014868869158017_dp, 4.544803335824005_dp, 7.820462911779673_dp, &
             ! w
             3.2055517151208895e-06_dp, 1.1645329797178361e-05_dp, 3.4781949406695064e-05_dp, &
             9.34486899719997e-05_dp, 0.00023019445382517747_dp, 0.0005295612623495888_dp, &
             0.0011531655722754395_dp, 0.002400042410783118_dp, 0.004808360895478562_dp, &
             0.009323665567699509_dp, 0.017572545368782437_dp, 0.032301238353315775_dp, &
             0.058069502790228454_dp, 0.1023369677099046_dp, 0.1771555391128928_dp, &
             0.3018300360012234_dp, 0.5072969281932966_dp, 0.844312595932769_dp, &
             1.4032083732522427_dp, 2.3829555075306437_dp, 4.510915652022847_dp]

      CASE (737)

         aw(:) = & ! a
            [6.878751123100988e-07_dp, 4.9948767416867265e-06_dp, 2.015699004918653e-05_dp, &
             6.568802767519742e-05_dp, 0.00018743455161509808_dp, 0.0004858708901612455_dp, &
             0.0011697753797923571_dp, 0.002654543750041718_dp, 0.005737287713786191_dp, &
             0.011901236423373663_dp, 0.023833846557235198_dp, 0.0462923293979264_dp, &
             0.08752575769190367_dp, 0.16157779501740915_dp, 0.29197053807820966_dp, &
             0.5175542867838526_dp, 0.9018343271794528_dp, 1.5483557793555442_dp, &
             2.6289235916220406_dp, 4.449801582463132_dp, 7.700040386038932_dp, &
             ! w
             1.9000957811429553e-06_dp, 7.810564821052788e-06_dp, 2.554752626950506e-05_dp, &
             7.270906778891163e-05_dp, 0.0001866792005936189_dp, 0.0004434249568882415_dp, &
             0.0009908189388700367_dp, 0.0021066906488315744_dp, 0.004297675246827283_dp, &
             0.00846418906901924_dp, 0.016170946368608767_dp, 0.030083639167527126_dp, &
             0.0546642209806848_dp, 0.09726561097215576_dp, 0.1698448319003003_dp, &
             0.2916631565740895_dp, 0.493725512461792_dp, 0.8270341645085848_dp, &
             1.3823489161549267_dp, 2.3590323614542648_dp, 4.483693775727391_dp]

      CASE (738)

         aw(:) = & ! a
            [5.061640205986907e-07_dp, 3.901463019636847e-06_dp, 1.6621231641870892e-05_dp, &
             5.625200743355926e-05_dp, 0.00016488607412711224_dp, 0.00043614441685021045_dp, &
             0.001066838479367986_dp, 0.002452293136064197_dp, 0.0053571838404891815_dp, &
             0.01121413923570821_dp, 0.022634398225602648_dp, 0.04426448571974849_dp, &
             0.08419908424199538_dp, 0.1562765354008451_dp, 0.2837620898658193_dp, &
             0.505211424593177_dp, 0.8838359412734615_dp, 1.5229591564512728_dp, &
             2.594331429778299_dp, 4.404383614480078_dp, 7.642348953780425_dp, &
             ! w
             1.421847853168403e-06_dp, 6.343941882114338e-06_dp, 2.1818316695028234e-05_dp, &
             6.403312137967499e-05_dp, 0.0001680209625352613_dp, 0.0004057818334235415_dp, &
             0.0009187664110742089_dp, 0.0019748059914727178_dp, 0.004065544646288029_dp, &
             0.008069779773294754_dp, 0.015522364908592219_dp, 0.02904982931900183_dp, &
             0.05306617908861603_dp, 0.09487147002913786_dp, 0.16637487522935984_dp, &
             0.286814102187985_dp, 0.48722488066562347_dp, 0.8187279211409486_dp, &
             1.3722938354689747_dp, 2.347483116177906_dp, 4.470553404219168_dp]

      CASE (739)

         aw(:) = & ! a
            [4.100938733228521e-07_dp, 3.3155394914994624e-06_dp, 1.4676907157299696e-05_dp, &
             5.094618385319114e-05_dp, 0.00015199147272580924_dp, 0.00040733605745022275_dp, &
             0.0010065839612636732_dp, 0.0023328989731728326_dp, 0.005131201803264373_dp, &
             0.010803164170675303_dp, 0.021913206172056034_dp, 0.04303960746488527_dp, &
             0.08218152686751376_dp, 0.15304984050498888_dp, 0.278749798912173_dp, &
             0.4976528599748241_dp, 0.8727858248165798_dp, 1.5073319699678924_dp, &
             2.5730056001144685_dp, 4.376341217310294_dp, 7.606688885079443_dp, &
             ! w
             1.1687630938559981e-06_dp, 5.54424000123924e-06_dp, 1.9718654817081773e-05_dp, &
             5.9052016080665624e-05_dp, 0.00015716176210840287_dp, 0.0003836425515132985_dp, &
             0.0008760300901730673_dp, 0.001896030458208616_dp, 0.003926062860727896_dp, &
             0.00783156643042515_dp, 0.015128870890417048_dp, 0.02842011464034791_dp, &
             0.05208931261756143_dp, 0.09340327280130209_dp, 0.16424079622420096_dp, &
             0.2838241160233927_dp, 0.48320731465291844_dp, 0.8135844997577173_dp, &
             1.366058408159859_dp, 2.340315333753395_dp, 4.462398553417637_dp]

      CASE (740)

         aw(:) = & ! a
            [3.501357202075409e-07_dp, 2.9459693996539737e-06_dp, 1.3426973328470624e-05_dp, &
             4.7482089803544514e-05_dp, 0.00014347589354796487_dp, 0.00038814453276351515_dp, &
             0.0009661661178699956_dp, 0.00225235994199906_dp, 0.0049780455709448985_dp, &
             0.010523517221249844_dp, 0.021420777857497974_dp, 0.042200739412231465_dp, &
             0.08079610790521065_dp, 0.1508288892607995_dp, 0.275292530420094_dp, &
             0.49242945014670403_dp, 0.8651367590576851_dp, 1.496498733615495_dp, &
             2.558203541609602_dp, 4.356857920045129_dp, 7.581894954949414_dp, &
             ! w
             1.010724147932948e-06_dp, 5.033108794495802e-06_dp, 1.834657966789752e-05_dp, &
             5.575410701856889e-05_dp, 0.00014990614428587735_dp, 0.00036874629611273956_dp, &
             0.0008471132800042783_dp, 0.001842479847515402_dp, 0.003830870702682751_dp, &
             0.007668440384675904_dp, 0.014858610778179576_dp, 0.027986479957459145_dp, &
             0.05141505042147568_dp, 0.09238775326185764_dp, 0.16276191003943608_dp, &
             0.2817485787224856_dp, 0.4804142828131193_dp, 0.810004236404613_dp, &
             1.3617138652122631_dp, 2.3353185233346876_dp, 4.456713823855711_dp]

      CASE (741)

         aw(:) = & ! a
            [3.0896545243513254e-07_dp, 2.689963616574446e-06_dp, 1.2548136186107512e-05_dp, &
             4.5018010816848565e-05_dp, 0.0001373671633367787_dp, 0.00037428901750429135_dp, &
             0.0009368387923668359_dp, 0.0021936808706294357_dp, 0.004866077902129518_dp, &
             0.010318483237691118_dp, 0.021058832141188238_dp, 0.041582807118282966_dp, &
             0.07977360902341836_dp, 0.14918693273175038_dp, 0.27273266775554955_dp, &
             0.4885566260535694_dp, 0.8594586174226914_dp, 1.48844838856608_dp, &
             2.5471940640599224_dp, 4.342356259757842_dp, 7.563430853695242_dp, &
             ! w
             9.021648612002999e-07_dp, 4.675258174932395e-06_dp, 1.73699450116295e-05_dp, &
             5.3384031239987054e-05_dp, 0.00014465682102896688_dp, 0.0003579138519076522_dp, &
             0.0008259988813220216_dp, 0.0018032458225963229_dp, 0.0037609282204627793_dp, &
             0.007548288532020997_dp, 0.014659121192710555_dp, 0.02766579114029773_dp, &
             0.050915567587335495_dp, 0.0916343334401505_dp, 0.1616632194125819_dp, &
             0.2802047362952904_dp, 0.4783344993993839_dp, 0.8073358190857092_dp, &
             1.3584735869404352_dp, 2.3315903376554292_dp, 4.452472466100813_dp]

      CASE (742)

         aw(:) = & ! a
            [2.788872510079027e-07_dp, 2.5015172326619714e-06_dp, 1.1893342531609396e-05_dp, &
             4.31652429499386e-05_dp, 0.00013274363548895136_dp, 0.000363750189970843_dp, &
             0.0009144450456379911_dp, 0.0021487335321277484_dp, 0.004780087465197236_dp, &
             0.010160668353126024_dp, 0.020779708411780732_dp, 0.04110547655044329_dp, &
             0.07898260306123993_dp, 0.14791505559610219_dp, 0.2707474633014034_dp, &
             0.48555008592337023_dp, 0.8550465147165376_dp, 1.4821879623223857_dp, &
             2.5386265936917582_dp, 4.331065042303935_dp, 7.549048672876_dp, &
             ! w
             8.228289636356286e-07_dp, 4.409511560022702e-06_dp, 1.6635225176741835e-05_dp, &
             5.15877295813827e-05_dp, 0.00014065768585962049_dp, 0.00034962869166831225_dp, &
             0.0008097986366680321_dp, 0.0017730647829614257_dp, 0.0037070063448228553_dp, &
             0.007455483174206394_dp, 0.014504782020311642_dp, 0.027417324229650517_dp, &
             0.05052807349179981_dp, 0.09104916136954654_dp, 0.16080899140673677_dp, &
             0.2790032823400865_dp, 0.4767146260880311_dp, 0.8052560273436701_dp, &
             1.3559467494679023_dp, 2.3286821687664747_dp, 4.449164055024924_dp]

      CASE (743)

         aw(:) = & ! a
            [2.5594146230861006e-07_dp, 2.3568086255116493e-06_dp, 1.1385426908918622e-05_dp, &
             4.171736356288635e-05_dp, 0.0001291113166587167_dp, 0.0003554378166730387_dp, &
             0.0008967273264527089_dp, 0.0021130821685664664_dp, 0.004711739083787147_dp, &
             0.010035009363270925_dp, 0.020557119684993828_dp, 0.040724321256635096_dp, &
             0.07835023484039691_dp, 0.14689720130541953_dp, 0.26915728169764686_dp, &
             0.48313981333743866_dp, 0.8515068485727192_dp, 1.4771622424488413_dp, &
             2.53174510914876_dp, 4.321991887238555_dp, 7.537488086518538_dp, &
             ! w
             7.622892617923496e-07_dp, 4.203915192172493e-06_dp, 1.6060830272205736e-05_dp, &
             5.0175009129337795e-05_dp, 0.0001374994520457507_dp, 0.0003430649600182608_dp, &
             0.0007969320018546216_dp, 0.0017490445726059275_dp, 0.0036640165515419727_dp, &
             0.007381382186796765_dp, 0.014381387969839263_dp, 0.027218446148589447_dp, &
             0.05021759712263839_dp, 0.0905798667403026_dp, 0.16012335442029796_dp, &
             0.2780382329813487_dp, 0.47541263454856736_dp, 0.803583445387465_dp, &
             1.353913800841705_dp, 2.3263418789695307_dp, 4.446501713609481_dp]

      CASE (744)

         aw(:) = & ! a
            [2.3787548644039865e-07_dp, 2.2422090666165663e-06_dp, 1.0979732188292768e-05_dp, &
             4.055372358667582e-05_dp, 0.00012617930889981951_dp, 0.00034870618930574226_dp, &
             0.0008823424119592437_dp, 0.002084077450254962_dp, 0.0046560382254079895_dp, &
             0.009932454853533485_dp, 0.02037523163456637_dp, 0.04041252326401662_dp, &
             0.07783244275952018_dp, 0.14606306310202252_dp, 0.26785313629327046_dp, &
             0.48116176478051725_dp, 0.8486002047790405_dp, 1.473033155862348_dp, &
             2.526088849016823_dp, 4.314531545890505_dp, 7.527980019747692_dp, &
             ! w
             7.146119090789806e-07_dp, 4.040047459793898e-06_dp, 1.559904231355918e-05_dp, &
             4.903367044229666e-05_dp, 0.000134939208860833_dp, 0.00033773024565015994_dp, &
             0.0007864529826234841_dp, 0.0017294485078885825_dp, 0.0036288946888346526_dp, &
             0.007320768903245919_dp, 0.014280346242891416_dp, 0.027055440964129335_dp, &
             0.04996291063509008_dp, 0.09019461219010494_dp, 0.15956012075629677_dp, &
             0.2772449899633981_dp, 0.4743418630659874_dp, 0.8022072760680569_dp, &
             1.3522405587136235_dp, 2.3244153110048256_dp, 4.444310050128928_dp]

      CASE (745)

         aw(:) = & ! a
            [2.2331049379156777e-07_dp, 2.1493362107956597e-06_dp, 1.0648525332767434e-05_dp, &
             3.959877089846541e-05_dp, 0.00012376428800077265_dp, 0.00034314638375491495_dp, &
             0.0008704363112021186_dp, 0.002060029594783014_dp, 0.004609790775546505_dp, &
             0.009847202998571773_dp, 0.020223874460895987_dp, 0.04015282821336016_dp, &
             0.07740083420588519_dp, 0.1453672737761853_dp, 0.2667646134676391_dp, &
             0.4795098377342551_dp, 0.8461715776313841_dp, 1.4695816306250906_dp, &
             2.5213590183078898_dp, 4.308291294087543_dp, 7.520025232375245_dp, &
             ! w
             6.761641402555621e-07_dp, 3.906504271368701e-06_dp, 1.5219967735416549e-05_dp, &
             4.809290314842326e-05_dp, 0.0001328228608435159_dp, 0.0003333109289956591_dp, &
             0.0007777571311883246_dp, 0.0017131639888326245_dp, 0.0035996733075395857_dp, &
             0.007270287177887106_dp, 0.014196119050221367_dp, 0.02691945566967618_dp, &
             0.04975029370021081_dp, 0.08987279454658673_dp, 0.15908936697521675_dp, &
             0.2765816598168249_dp, 0.47344605898772424_dp, 0.8010555456821149_dp, &
             1.3508398088544546_dp, 2.322802236825114_dp, 4.442475034650325_dp]

      CASE (746)

         aw(:) = & ! a
            [1.6096310157758968e-07_dp, 1.745556264548495e-06_dp, 9.17905978498298e-06_dp, &
             3.530256560731041e-05_dp, 0.00011279352524850985_dp, 0.00031770738024748126_dp, &
             0.0008156537639023174_dp, 0.0019488792610287153_dp, 0.004395231445814714_dp, &
             0.009450434185452402_dp, 0.01951752866336194_dp, 0.03893802589373031_dp, &
             0.07537764977804465_dp, 0.14209971246675993_dp, 0.26164430114180565_dp, &
             0.4717279409198301_dp, 0.834715933004493_dp, 1.4532825603151018_dp, &
             2.499001945379153_dp, 4.278772035515032_dp, 7.4823742587314275_dp, &
             ! w
             5.114221428998643e-07_dp, 3.316705865711376e-06_dp, 1.3513345930342916e-05_dp, &
             4.3811406375243465e-05_dp, 0.000123118420931326_dp, 0.00031293028733250377_dp, &
             0.000737472025888166_dp, 0.0016374415939337899_dp, 0.003463369106033315_dp, &
             0.007034181505999811_dp, 0.013801265341094505_dp, 0.026280653893201885_dp, &
             0.04874968704417199_dp, 0.08835579903077295_dp, 0.1568670496936143_dp, &
             0.2734461051177345_dp, 0.469206669371532_dp, 0.7955996198724832_dp, &
             1.3441992933320523_dp, 2.315151972966625_dp, 4.43377235389112_dp]

      CASE (747)

         aw(:) = & ! a
            [1.5525223911801902e-07_dp, 1.7079307630782255e-06_dp, 9.039312804974884e-06_dp, &
             3.488846949388754e-05_dp, 0.0001117263004569822_dp, 0.00031521588043408836_dp, &
             0.0008102601724417631_dp, 0.0019378898924732528_dp, 0.0043739443366095625_dp, &
             0.009410954103944588_dp, 0.01944706780961549_dp, 0.03881657996249205_dp, &
             0.07517500107350235_dp, 0.14177186841364273_dp, 0.26112979105316997_dp, &
             0.4709449345802823_dp, 0.8335619068659411_dp, 1.4516389079132597_dp, &
             2.4967454014343855_dp, 4.275790504660655_dp, 7.478569449640916_dp, &
             ! w
             4.963110691970876e-07_dp, 3.260852427458791e-06_dp, 1.3348743445357634e-05_dp, &
             4.339420456576796e-05_dp, 0.00012216606605914423_dp, 0.0003109195086290201_dp, &
             0.0007334806436291085_dp, 0.0016299132169146405_dp, 0.0034497783792017413_dp, &
             0.007010581401141866_dp, 0.013761712713972672_dp, 0.02621654426105584_dp, &
             0.04864909895289677_dp, 0.08820307161296066_dp, 0.15664301130907557_dp, &
             0.2731296194225316_dp, 0.4687783130371601_dp, 0.7950478463846157_dp, &
             1.3435272619684118_dp, 2.314377459819345_dp, 4.432891311377889_dp]

      CASE (748)

         aw(:) = & ! a
            [0.0025498370150328175_dp, 0.013548969777634706_dp, 0.033817943783834084_dp, &
             0.06424953119334481_dp, 0.10627975304616633_dp, 0.16207613815785146_dp, &
             0.23479913882388248_dp, 0.32893749312111326_dp, 0.450705961835072_dp, &
             0.6085016447686721_dp, 0.8134463091063422_dp, 1.0800779942862897_dp, &
             1.4272779198638608_dp, 1.8795392499056383_dp, 2.468734525435586_dp, &
             3.2366628180256325_dp, 4.238943312512284_dp, 5.551503932825093_dp, &
             7.2827145663903305_dp, 9.599777136767443_dp, 12.799970662820616_dp, &
             17.59344480624227_dp, &
             ! w
             0.006556575820857464_dp, 0.015523085992047754_dp, 0.02516140409226648_dp, &
             0.03593802829197264_dp, 0.04847931133583091_dp, 0.0636345758902718_dp, &
             0.08255289856653057_dp, 0.10675206175174021_dp, 0.1381755122195695_dp, &
             0.17925816196500918_dp, 0.23303896694606002_dp, 0.30335047363213585_dp, &
             0.3951027279091373_dp, 0.5146907878671897_dp, 0.6706040029739012_dp, &
             0.8744204749875732_dp, 1.142609350548909_dp, 1.5001857181231342_dp, &
             1.9891334481256275_dp, 2.6913589917013168_dp, 3.8099502339343427_dp, &
             6.151546796176026_dp]

      CASE (749)

         aw(:) = & ! a
            [0.0014433340386207543_dp, 0.007688392310249193_dp, 0.019279234224944517_dp, &
             0.03688917533394893_dp, 0.06163378396171419_dp, 0.09525074537776242_dp, &
             0.14034568784044757_dp, 0.200695519607924_dp, 0.28160142129290044_dp, &
             0.3903123150009398_dp, 0.5365742231545787_dp, 0.7333763702377785_dp, &
             0.9979697637341145_dp, 1.353254099028081_dp, 1.8296850115231906_dp, &
             2.4679765042769186_dp, 3.3231453968588354_dp, 4.471098972104022_dp, &
             6.020717964130783_dp, 8.139833862536845_dp, 11.125057670998743_dp, &
             15.678729852361213_dp, &
             ! w
             0.003713462441096966_dp, 0.00883562839083754_dp, 0.014456236618062607_dp, &
             0.020946604959131625_dp, 0.028828100701419928_dp, 0.03883500473592093_dp, &
             0.051978528412584706_dp, 0.0695977507610946_dp, 0.09341322433495981_dp, &
             0.12562063830520065_dp, 0.16905128676816808_dp, 0.22740539901027826_dp, &
             0.3055668201409863_dp, 0.41003392517468645_dp, 0.5495486747937728_dp, &
             0.7360999452634414_dp, 0.9867036394465815_dp, 1.32696824727902_dp, &
             1.7992917757189784_dp, 2.4852614735754917_dp, 3.585169939050862_dp, &
             5.890302419371061_dp]

      CASE (750)

         aw(:) = & ! a
            [0.0010306343501664476_dp, 0.005499231563027791_dp, 0.01383349827107724_dp, &
             0.0265993900808953_dp, 0.04475207822999643_dp, 0.06980682492527533_dp, &
             0.10406842817267943_dp, 0.15090869184117758_dp, 0.215093367125172_dp, &
             0.30319499719802995_dp, 0.42415335184399106_dp, 0.5900495327971398_dp, &
             0.8171651863150591_dp, 1.1274243926114875_dp, 1.550374562931633_dp, &
             2.1259821473449616_dp, 2.9087815074786754_dp, 3.9745523714679707_dp, &
             5.432422039917516_dp, 7.450671490328925_dp, 10.325865977094331_dp, &
             14.756856074517449_dp, &
             ! w
             0.0026526748248829027_dp, 0.006332949644865581_dp, 0.010428000997806431_dp, &
             0.015260141588040224_dp, 0.02129391608002459_dp, 0.029195076534516938_dp, &
             0.03988474153491644_dp, 0.05458335544802141_dp, 0.07487223431702189_dp, &
             0.10280824658475067_dp, 0.14110454665011496_dp, 0.19337774294779675_dp, &
             0.26447517091195055_dp, 0.36092237375507297_dp, 0.4915723524824677_dp, &
             0.6686264207948882_dp, 0.909416270057912_dp, 1.2399335441278_dp, &
             1.702912155379959_dp, 2.379933761144888_dp, 3.470040102233809_dp, &
             5.756937427107729_dp]

      CASE (751)

         aw(:) = & ! a
            [0.0008104199790971963_dp, 0.004329875360730728_dp, 0.010919012319056941_dp, &
             0.02107673237142483_dp, 0.03565649285758547_dp, 0.05602932398536984_dp, &
             0.08429943257418211_dp, 0.12356253950538487_dp, 0.17821692803580538_dp, &
             0.25437202738110054_dp, 0.3604160241889657_dp, 0.5078046150652303_dp, &
             0.7121420980421576_dp, 0.9946562131749794_dp, 1.3842276646310172_dp, &
             1.920252111992597_dp, 2.656868488960321_dp, 3.669712288097814_dp, &
             5.068051260362004_dp, 7.020499406454845_dp, 9.823704410141259_dp, &
             14.174435088645948_dp, &
             ! w
             0.002086505465145412_dp, 0.004994391287343446_dp, 0.008265217758225094_dp, &
             0.012189748117177721_dp, 0.017194645140954118_dp, 0.02389742801595452_dp, &
             0.03315367691473142_dp, 0.04609892897663443_dp, 0.06421876876729789_dp, &
             0.08947647592157501_dp, 0.12450301352326126_dp, 0.1728506675634372_dp, &
             0.23932904425005275_dp, 0.3304660870304807_dp, 0.4551747091761119_dp, &
             0.6257905360208326_dp, 0.8598614155317791_dp, 1.183658805451631_dp, &
             1.6401869362244077_dp, 2.3110916977896885_dp, 3.3946746991979184_dp, &
             5.669797142951594_dp]

      CASE (752)

         aw(:) = & ! a
            [0.000672087511475086_dp, 0.0035947054183472003_dp, 0.009083881993347333_dp, &
             0.017591439976140398_dp, 0.029898578901409817_dp, 0.047272106253395535_dp, &
             0.07166850076409409_dp, 0.10597777312236537_dp, 0.15432400098495969_dp, &
             0.2224716732390262_dp, 0.3183974030771918_dp, 0.45308673892441814_dp, &
             0.6416285695762037_dp, 0.9047123865908624_dp, 1.2706934035218946_dp, &
             1.7785055184801077_dp, 2.481953467258232_dp, 3.456530860901502_dp, &
             4.811592320437186_dp, 6.7160119585640246_dp, 9.466554998203438_dp, &
             13.758573565893395_dp, &
             ! w
             0.0017307848794568584_dp, 0.0041519802524565335_dp, 0.006899934997819808_dp, &
             0.010242768374466787_dp, 0.014579206722752402_dp, 0.020489899917730776_dp, &
             0.028779679355755518_dp, 0.0405195853742844_dp, 0.05712422522558132_dp, &
             0.08048743941881376_dp, 0.11317694631593994_dp, 0.15869195987723836_dp, &
             0.22180615220011649_dp, 0.30903994739476703_dp, 0.42934193007211885_dp, &
             0.595141807058348_dp, 0.8241488988880186_dp, 1.1428536067336916_dp, &
             1.5944846584505448_dp, 2.260772273647255_dp, 3.3395202020108496_dp, &
             5.60610623943483_dp]

      CASE (753)

         aw(:) = & ! a
            [0.0005765417988712976_dp, 0.0030865704442351234_dp, 0.007813843607651923_dp, &
             0.015174746106365579_dp, 0.025895555165994438_dp, 0.04116276242083353_dp, &
             0.06281741993107001_dp, 0.09358771867519423_dp, 0.1373820811580077_dp, &
             0.19969393822672132_dp, 0.2881755389745978_dp, 0.41343927741874476_dp, &
             0.5901608400036042_dp, 0.838592901620649_dp, 1.1866579105505821_dp, &
             1.6729021616499282_dp, 2.3508411237133755_dp, 3.2958337313062858_dp, &
             4.617290591434739_dp, 6.484297007429214_dp, 9.193743827243535_dp, &
             13.43994218702936_dp, &
             ! w
             0.0014850500763126902_dp, 0.0035692204781799576_dp, 0.005953042044958433_dp, &
             0.008887288898291679_dp, 0.012748807576191208_dp, 0.018088626199249804_dp, &
             0.025670638224564724_dp, 0.036514754419242124_dp, 0.051980110570449566_dp, &
             0.0739057929346307_dp, 0.10480800535825094_dp, 0.14814057397158564_dp, &
             0.20864374420622286_dp, 0.29282603012925423_dp, 0.4096583256642417_dp, &
             0.5716404839383008_dp, 0.7966090917926576_dp, 1.1112335165618066_dp, &
             1.5589339261948736_dp, 2.2215297269504686_dp, 3.296463714715325_dp, &
             5.556432822862914_dp]

      CASE (754)

         aw(:) = & ! a
            [0.0005063013743488893_dp, 0.002712787611711358_dp, 0.006878560171515061_dp, &
             0.013392060728198618_dp, 0.022935896645317134_dp, 0.03663196032652959_dp, &
             0.05622748586918589_dp, 0.08431860056070759_dp, 0.12463785106108936_dp, &
             0.1824576214136149_dp, 0.26516475573963166_dp, 0.38306435128385913_dp, &
             0.5504889380824046_dp, 0.7873248203387914_dp, 1.1211272303598854_dp, &
             1.5901086505618087_dp, 2.2475295833277014_dp, 3.1686224666594622_dp, &
             4.462834516292547_dp, 6.299427453335221_dp, 8.975416451876987_dp, &
             13.184308089492959_dp, &
             ! w
             0.001304372659245205_dp, 0.0031402228978207516_dp, 0.005254432516770554_dp, &
             0.007883898331350975_dp, 0.011387629548733064_dp, 0.016292044710639664_dp, &
             0.02332703893194826_dp, 0.033470688296525344_dp, 0.04803709678108123_dp, &
             0.06882036328494114_dp, 0.09829319312262905_dp, 0.13986970465204632_dp, &
             0.1982592331882949_dp, 0.27995640866067345_dp, 0.3939463420383559_dp, &
             0.5527834423655349_dp, 0.7744083228662891_dp, 1.0856411205501726_dp, &
             1.5300686203625449_dp, 2.1895987210126213_dp, 3.26139926305612_dp, &
             5.516009839239283_dp]

      CASE (755)

         aw(:) = & ! a
            [0.0004523289977004399_dp, 0.0024254194065155017_dp, 0.006158787526149647_dp, &
             0.012018096801116705_dp, 0.0206500938360119_dp, 0.03312309952528277_dp, &
             0.051105906164364036_dp, 0.07708394352279445_dp, 0.11464245014478185_dp, &
             0.16886857694460544_dp, 0.24692599297021087_dp, 0.358859892600339_dp, &
             0.5187104768314617_dp, 0.7460493629003966_dp, 1.0681133406343073_dp, &
             1.522821908005346_dp, 2.1632076305889085_dp, 3.0643843956594616_dp, &
             4.335823138607256_dp, 6.146936170408139_dp, 8.79485826575428_dp, &
             12.97245147386671_dp, &
             ! w
             0.0011655236713745728_dp, 0.002810185591219798_dp, 0.0047159090990930705_dp, &
             0.007108136113797475_dp, 0.010330911156312197_dp, 0.014889707152187976_dp, &
             0.02148556041142088_dp, 0.03106147671895964_dp, 0.04489394877623434_dp, &
             0.06473901307010645_dp, 0.09303172450515705_dp, 0.13315089896929913_dp, &
             0.1897772814141164_dp, 0.269390808048385_dp, 0.38098557607825356_dp, &
             0.5371597389033855_dp, 0.7559412204496332_dp, 1.0642800197454287_dp, &
             1.5059102626486158_dp, 2.162825505365705_dp, 3.231976654581783_dp, &
             5.4821116124334_dp]

      CASE (756)

         aw(:) = & ! a
            [0.000409463955316871_dp, 0.0021970790340996075_dp, 0.005586346426732146_dp, &
             0.010923891312009569_dp, 0.018826289437449243_dp, 0.030316403448747947_dp, &
             0.04699602765556811_dp, 0.0712558970704633_dp, 0.10655532330180247_dp, &
             0.15782297329595119_dp, 0.23203091529480346_dp, 0.33900004214102764_dp, &
             0.4925166866746626_dp, 0.7118772545499205_dp, 1.0240377826249645_dp, &
             1.4666568437638203_dp, 2.092560913459095_dp, 2.976752662119821_dp, &
             4.228717789526811_dp, 6.0179993387075_dp, 8.641845819757558_dp, &
             12.792588922338119_dp, &
             ! w
             0.0010552368597852137_dp, 0.00254778311387449_dp, 0.004286976393703696_dp, &
             0.006488578301847026_dp, 0.009483805400773166_dp, 0.013759974060421567_dp, &
             0.019993202611892955_dp, 0.02909652811460109_dp, 0.04231433279342542_dp, &
             0.06136973960078679_dp, 0.08866461988383019_dp, 0.12754601573509738_dp, &
             0.18266815854648735_dp, 0.260496080777035_dp, 0.3700293238752068_dp, &
             0.5239019326716725_dp, 0.7402165971352499_dp, 1.0460370834507968_dp, &
             1.4852294779224067_dp, 2.1398693863702456_dp, 3.206732007236893_dp, &
             5.453041831528074_dp]

      CASE (757)

         aw(:) = & ! a
            [0.0003745365136890047_dp, 0.002010938742561351_dp, 0.005119314038479357_dp, &
             0.010030061600758183_dp, 0.01733388889706926_dp, 0.028014388712027473_dp, &
             0.0436151707268864_dp, 0.06644465469390803_dp, 0.09985274691474497_dp, &
             0.1486302704296793_dp, 0.21958209355794206_dp, 0.32233244541629946_dp, &
             0.47044384389332883_dp, 0.6829685773053592_dp, 0.9866119113627806_dp, &
             1.4187973417742634_dp, 2.0321632884481837_dp, 2.901607861554979_dp, &
             4.136625105064002_dp, 5.906873105924336_dp, 8.509708045070505_dp, &
             12.637016008506167_dp, &
             ! w
             0.000965363281953965_dp, 0.0023337582508529296_dp, 0.003936547907743397_dp, &
             0.005981161609432484_dp, 0.008787633586983897_dp, 0.012827313323225745_dp, &
             0.018754505971305256_dp, 0.027456238360757803_dp, 0.04014899712892212_dp, &
             0.058526961594539344_dp, 0.08496234944321378_dp, 0.12277332511237751_dp, &
             0.17658944834174675_dp, 0.2528609560266091_dp, 0.360590351441224_dp, &
             0.5124416667367092_dp, 0.72658263876052_dp, 1.0301780477641367_dp, &
             1.4672133875969238_dp, 2.119842512679787_dp, 3.184695372753126_dp, &
             5.427677356202323_dp]

      CASE (758)

         aw(:) = & ! a
            [0.00020791631177922895_dp, 0.0011215576115341443_dp, 0.002881277587342122_dp, &
             0.0057277275160863275_dp, 0.010105412362063743_dp, 0.016770405324037792_dp, &
             0.026924485919518657_dp, 0.042392190719159964_dp, 0.06588055696093978_dp, &
             0.10136496839774164_dp, 0.1546479615518937_dp, 0.2341547740325862_dp, &
             0.35205766361810475_dp, 0.5258613793548387_dp, 0.7806445176312687_dp, &
             1.1522621668013824_dp, 1.6920459781761579_dp, 2.4740999448194474_dp, &
             3.6078621869746055_dp, 5.263691913493065_dp, 7.739760682851753_dp, &
             11.725620514661394_dp, &
             ! w
             0.0005364677150234348_dp, 0.00130915525385517_dp, 0.002249112270561428_dp, &
             0.003516048983454188_dp, 0.0053631362663547014_dp, 0.008164303355271952_dp, &
             0.012443719513996662_dp, 0.018936537096784868_dp, 0.02869359385043377_dp, &
             0.04322927420950235_dp, 0.0647224871099812_dp, 0.09629387004026087_dp, &
             0.14239269869895335_dp, 0.20934229131747628_dp, 0.30612380147438845_dp, &
             0.4455505203477931_dp, 0.6461730120898519_dp, 0.9357959485392984_dp, &
             1.3592103902597157_dp, 1.9991818670317985_dp, 3.051637588010361_dp, &
             5.274743376553333_dp]

      CASE (759)

         aw(:) = & ! a
            [0.0001471710349114242_dp, 0.0007964283238371479_dp, 0.0020589666400482724_dp, &
             0.0041346670676657244_dp, 0.007399362011036264_dp, 0.012498745872237612_dp, &
             0.020466740916424406_dp, 0.032890613598826925_dp, 0.05216105943866098_dp, &
             0.08184646328125295_dp, 0.12723849744538981_dp, 0.19613666759968496_dp, &
             0.2999693795086896_dp, 0.4553912580033871_dp, 0.686560914454681_dp, &
             1.0284151057074777_dp, 1.5314829201664935_dp, 2.2693330000065535_dp, &
             3.3512877011453703_dp, 4.94806350992283_dp, 7.358365821367246_dp, &
             11.270786599832105_dp, &
             ! w
             0.0003800055385784594_dp, 0.000933338699389642_dp, 0.0016239022722212787_dp, &
             0.0025885536580075577_dp, 0.004046539472539789_dp, 0.006321896533186526_dp, &
             0.009874557653625002_dp, 0.015365506374564575_dp, 0.023761180611981632_dp, &
             0.036478724592218015_dp, 0.05558690158181038_dp, 0.08408802094333777_dp, &
             0.12631592475527947_dp, 0.1885008895204224_dp, 0.27958327696553686_dp, &
             0.4124277695295831_dp, 0.6057717894232834_dp, 0.8877687175607093_dp, &
             1.3036856034687678_dp, 1.936706722484125_dp, 2.9825243016709715_dp, &
             5.195448539479865_dp]

      CASE (760)

         aw(:) = & ! a
            [0.000115126529151328_dp, 0.0006245898803638764_dp, 0.0016228126960023878_dp, &
             0.003285072370393567_dp, 0.005944847791810198_dp, 0.010178627753497673_dp, &
             0.016914798187963254_dp, 0.027591536967412126_dp, 0.04439965148672822_dp, &
             0.07064742847500395_dp, 0.11129557181084923_dp, 0.17373237765032498_dp, &
             0.2688902586273879_dp, 0.412848323366157_dp, 0.6291319815031875_dp, &
             0.952033554436149_dp, 1.4315041901002497_dp, 2.1407086591738773_dp, &
             3.1888541267228843_dp, 4.746884550329156_dp, 7.113897113042695_dp, &
             10.977950238749214_dp, &
             ! w
             0.0002974321961854475_dp, 0.0007342487847945801_dp, 0.0012903520200170308_dp, &
             0.0020883247126737008_dp, 0.00332556285051951_dp, 0.005294186756249974_dp, &
             0.008413884982695352_dp, 0.013298756645919474_dp, 0.020860101876359848_dp, &
             0.032449552558703146_dp, 0.050059844405449846_dp, 0.07660961255208805_dp, &
             0.11634840799236064_dp, 0.1754344892901406_dp, 0.26276898087098216_dp, &
             0.3912386670448714_dp, 0.5796971670712566_dp, 0.8565316777219318_dp, &
             1.267344101703483_dp, 1.8956350842413046_dp, 2.936996041080307_dp, &
             5.1432642736468015_dp]

      CASE (761)

         aw(:) = & ! a
            [9.514367291620342e-05_dp, 0.0005172709654815866_dp, 0.0013496509500609878_dp, &
             0.002750636162709005_dp, 0.005024098350995886_dp, 0.008697667742000297_dp, &
             0.014625188214634419_dp, 0.024139481461350202_dp, 0.0392893016178502_dp, &
             0.06319642222628762_dp, 0.10058182898448903_dp, 0.15853295688946847_dp, &
             0.2476156128409093_dp, 0.3834790112794715_dp, 0.5891701416639776_dp, &
             0.898487922099274_dp, 1.3609332693224756_dp, 2.0493467890519708_dp, &
             3.0728292294129096_dp, 4.602485817216761_dp, 6.937721321765447_dp, &
             10.766253760503595_dp, &
             ! w
             0.0002459219473295276_dp, 0.0006096815755691054_dp, 0.0010804821743334398_dp, &
             0.001770829053446422_dp, 0.002862400284638159_dp, 0.004624582015563799_dp, &
             0.007448725891021077_dp, 0.011915481323164235_dp, 0.01889585291374632_dp, &
             0.029692673830559676_dp, 0.04624119184447866_dp, 0.07139577990665456_dp, &
             0.1093399115507425_dp, 0.16617337311744232_dp, 0.2507618426717478_dp, &
             0.3760017690770574_dp, 0.5608278246799532_dp, 0.8338002926317234_dp, &
             1.2407776166400746_dp, 1.8655141384575193_dp, 2.9035563143259813_dp, &
             5.1049609072768725_dp]

      CASE (762)

         aw(:) = & ! a
            [8.141328637267183e-05_dp, 0.00044343818179942875_dp, 0.001161275340697676_dp, &
             0.002380709613670433_dp, 0.004383348496563784_dp, 0.007659843854727517_dp, &
             0.013007635848118211_dp, 0.021679734442014414_dp, 0.03561671563861088_dp, &
             0.0577972661403761_dp, 0.0927570244102207_dp, 0.14734898846890754_dp, &
             0.23185106838224884_dp, 0.3615721613357245_dp, 0.5591772338408921_dp, &
             0.858067272951867_dp, 1.3073753795964227_dp, 1.9796715683556463_dp, &
             2.983959508297886_dp, 4.49146636044111_dp, 6.801849035522098_dp, &
             10.602590209977295_dp, &
             ! w
             0.00021051858126274736_dp, 0.0005238499922859899_dp, 0.000935189048688427_dp, &
             0.0015494004128264332_dp, 0.0025360990043404555_dp, 0.004147394701591875_dp, &
             0.006753235875429415_dp, 0.010908663640999414_dp, 0.01745327047155068_dp, &
             0.02765134310205097_dp, 0.04339223248680795_dp, 0.0674784216164884_dp, &
             0.1040392329762628_dp, 0.15912535396557376_dp, 0.2415706624490167_dp, &
             0.3642750229259834_dp, 0.5462336346350323_dp, 0.8161426734692004_dp, &
             1.2200674681958341_dp, 1.8419738091950792_dp, 2.8773909389926593_dp, &
             5.075004378711427_dp]

      CASE (763)

         aw(:) = & ! a
            [7.13592602953326e-05_dp, 0.00038931505362841393_dp, 0.0010229001663757625_dp, &
             0.0021080853469690884_dp, 0.003908912280027379_dp, 0.0068867293285242845_dp, &
             0.01179429779960341_dp, 0.019821310659832368_dp, 0.032822110060808846_dp, &
             0.05366061188096963_dp, 0.08672283428904878_dp, 0.13867133500502646_dp, &
             0.21954868776753408_dp, 0.3443838061331869_dp, 0.535525179860289_dp, &
             0.8260415203975644_dp, 1.2647556730487695_dp, 1.9240055871194643_dp, &
             2.9127062906594374_dp, 4.402181196223899_dp, 6.692300183988522_dp, &
             10.470374439791838_dp, &
             ! w
             0.00018458807343583818_dp, 0.0004608466693838815_dp, 0.0008280986466169222_dp, &
             0.0013851374779097951_dp, 0.0022919144304074947_dp, 0.003786825675332275_dp, &
             0.006222876025953849_dp, 0.01013457518150216_dp, 0.01633597442955237_dp, &
             0.026059708114393242_dp, 0.041157129296882936_dp, 0.06438737977541568_dp, &
             0.09983403206068221_dp, 0.15350552677575705_dp, 0.23420708813959973_dp, &
             0.3548385374544674_dp, 0.5344423777298078_dp, 0.8018257422487299_dp, &
             1.2032266345637266_dp, 1.8227918917133776_dp, 2.8560488027248003_dp, &
             5.050579148637396_dp]

      CASE (764)

         aw(:) = & ! a
            [6.365799830191801e-05_dp, 0.0003478169048345856_dp, 0.0009166066561902296_dp, &
             0.001898055401445788_dp, 0.0035418660944247023_dp, 0.0062853925901175355_dp, &
             0.010844826708455227_dp, 0.01835794773058513_dp, 0.030608084855568513_dp, &
             0.05036413606026543_dp, 0.08188758244511092_dp, 0.13168163292022234_dp, &
             0.20959091523837184_dp, 0.33040757848315827_dp, 0.5162109490278204_dp, &
             0.7997852856212547_dp, 1.229685619612_dp, 1.8780468865264222_dp, &
             2.853702817705598_dp, 4.328055203454556_dp, 6.601157813351916_dp, &
             10.360192164702253_dp, &
             ! w
             0.0001647211769773475_dp, 0.0004124820806133462_dp, 0.0007455878854792925_dp, &
             0.0012578449826419862_dp, 0.0021012228613021514_dp, 0.0035028835688452687_dp, &
             0.005801959792489643_dp, 0.009515955450776624_dp, 0.015437527062570414_dp, &
             0.024772589209176202_dp, 0.03934020589753908_dp, 0.061862422691484134_dp, &
             0.0963832904369785_dp, 0.14887422325566507_dp, 0.22811445008599435_dp, &
             0.34700171576787453_dp, 0.524616719284114_dp, 0.7898597755434382_dp, &
             1.1891166228486785_dp, 1.8066922274652772_dp, 2.8381208629034544_dp, &
             5.030067629091043_dp]

      CASE (765)

         aw(:) = & ! a
            [5.755729517796371e-05_dp, 0.00031491430664052906_dp, 0.0008321884827675827_dp, &
             0.001730806101816026_dp, 0.003248464343239904_dp, 0.00580238149711106_dp, &
             0.01007807148427162_dp, 0.017169684489143626_dp, 0.028800630731831745_dp, &
             0.04765926024805869_dp, 0.07790099269977076_dp, 0.1258927158926364_dp, &
             0.20130895928078743_dp, 0.31873748434527993_dp, 0.5000241249188767_dp, &
             0.777705013695571_dp, 1.2000999317444572_dp, 1.8391634936360362_dp, &
             2.8036544724020978_dp, 4.265039993527441_dp, 6.523535304394_dp, &
             10.266221168342534_dp, &
             ! w
             0.00014898005064997846_dp, 0.000374093905974257_dp, 0.0006798784942225079_dp, &
             0.0011559408540031825_dp, 0.0019475047774989323_dp, 0.003272307141776719_dp, &
             0.005457832330867657_dp, 0.009007153870759038_dp, 0.014694605156260414_dp, &
             0.023703072414747874_dp, 0.037823637396316695_dp, 0.05974600182803676_dp, &
             0.09347949120894637_dp, 0.144962599288783_dp, 0.22295080691524882_dp, &
             0.3403385431765515_dp, 0.5162381156354404_dp, 0.779629764330004_dp, &
             1.1770280079879114_dp, 1.792878053201223_dp, 2.822726602032779_dp, &
             5.012459431238408_dp]

      CASE (766)

         aw(:) = & ! a
            [5.259694976356289e-05_dp, 0.0002881402996578436_dp, 0.0007633887292551905_dp, &
             0.0015941662418338454_dp, 0.003007917514492729_dp, 0.005404637613018257_dp, &
             0.009443606087814827_dp, 0.01618159434941184_dp, 0.027290489775856935_dp, &
             0.04538909497904027_dp, 0.07454088104296895_dp, 0.12099408847367578_dp, &
             0.19427462426170708_dp, 0.3087909560785084_dp, 0.48618322335785347_dp, &
             0.7587679233895701_dp, 1.174655343174268_dp, 1.8056379928380988_dp, &
             2.7604051692599185_dp, 4.210479468821175_dp, 6.456219840123295_dp, &
             10.184627120582004_dp, &
             ! w
             0.00013617892489091487_dp, 0.0003428252348200111_dp, 0.0006261917491684965_dp, &
             0.0010722809011884272_dp, 0.0018205124597653688_dp, 0.0030805661377643332_dp, &
             0.005169950838676587_dp, 0.008579264374559897_dp, 0.014066876001072053_dp, &
             0.022795501947998915_dp, 0.03653159876412404_dp, 0.05793625577970115_dp, &
             0.09098787643388874_dp, 0.14159534642345972_dp, 0.21849231140971936_dp, &
             0.33456911582344556_dp, 0.5089647311618987_dp, 0.7707290917457327_dp, &
             1.166490627864167_dp, 1.7808204404334202_dp, 2.8092810452644454_dp, &
             4.997083558490311_dp]

      CASE (767)

         aw(:) = & ! a
            [2.907628904077249e-05_dp, 0.00016082076404549538_dp, 0.0004344226262905925_dp, &
             0.000935039866616583_dp, 0.0018328848707020412_dp, 0.0034315182108896193_dp, &
             0.006243306390460532_dp, 0.011113960485139207_dp, 0.019420565118187023_dp, &
             0.033378358807301514_dp, 0.05650996713571822_dp, 0.09435652357765235_dp, &
             0.1555466129045531_dp, 0.253391770241248_dp, 0.4082558539800291_dp, &
             0.6510708114984333_dp, 1.028599996026718_dp, 1.6115614693892595_dp, &
             2.508139912548048_dp, 3.8901562541340384_dp, 6.05888858834258_dp, &
             9.701018208484912_dp, &
             ! w
             7.543962146410753e-05_dp, 0.00019361118472649178_dp, 0.00036718303617062903_dp, &
             0.0006616570874122843_dp, 0.0011834229834969839_dp, 0.002097215495283438_dp, &
             0.0036640937154858028_dp, 0.006301740769991501_dp, 0.010673039318204436_dp, &
             0.017818070112319753_dp, 0.029351245221424006_dp, 0.04775394580328058_dp, &
             0.07680618496930319_dp, 0.122220852210936_dp, 0.19257729406487864_dp, &
             0.3007162043507918_dp, 0.4659168788385335_dp, 0.7176458827644393_dp, &
             1.1032472256367993_dp, 1.7081218801678342_dp, 2.728030963374664_dp, &
             4.904233360385957_dp]

      CASE (768)

         aw(:) = & ! a
            [2.0565388925150248e-05_dp, 0.00011452086738863314_dp, 0.0003136409940865938_dp, &
             0.0006892618540574246_dp, 0.0013851734754525018_dp, 0.002660386561993515_dp, &
             0.004959184594424867_dp, 0.009027974061423995_dp, 0.016102354309045712_dp, &
             0.028200033096785246_dp, 0.048573946082239555_dp, 0.08240631462638882_dp, &
             0.13786227352142752_dp, 0.22767678515526849_dp, 0.3715304050376075_dp, &
             0.599598522246047_dp, 0.9578899455685994_dp, 1.5164982000898621_dp, &
             2.383284206644635_dp, 3.73019408610862_dp, 5.859015232690362_dp, &
             9.456376372561994_dp, &
             ! w
             5.3436320129668144e-05_dp, 0.00013902007350660616_dp, 0.0002705992840788777_dp, &
             0.0005039390287103258_dp, 0.0009299301932313514_dp, 0.001692647409765458_dp, &
             0.003026314668002718_dp, 0.005312442079680487_dp, 0.009165238802257964_dp, &
             0.01556098295787941_dp, 0.026033389716835276_dp, 0.04296634623718816_dp, &
             0.07002935076265354_dp, 0.11282199043147696_dp, 0.1798279128849883_dp, &
             0.2838440038224664_dp, 0.44420648856691275_dp, 0.6905930126272187_dp, &
             1.0707361369610693_dp, 1.6705153985027792_dp, 2.6858687149928993_dp, &
             4.856093234006202_dp]

      CASE (769)

         aw(:) = & ! a
            [1.609052231918399e-05_dp, 9.009247051933442e-05_dp, 0.00024948071463684515_dp, &
             0.0005572536975209218_dp, 0.0011410895881533065_dp, 0.002232864066776534_dp, &
             0.004235203306907991_dp, 0.007833073623532659_dp, 0.014173444552420823_dp, &
             0.025148827489848043_dp, 0.04383934716025634_dp, 0.07519483722635223_dp, &
             0.12707730021135805_dp, 0.21184071986381642_dp, 0.34870934356364736_dp, &
             0.5673474201799109_dp, 0.9132468032319611_dp, 1.45606441575003_dp, &
             2.303423381715056_dp, 3.627339038363811_dp, 5.729944660907998_dp, &
             9.29787769290753_dp, &
             ! w
             4.185826305307244e-05_dp, 0.00011009520112019157_dp, 0.00021872915447843373_dp, &
             0.0004174813429819902_dp, 0.0007877670580220323_dp, 0.0014610780035391713_dp, &
             0.002654850403603292_dp, 0.004727453228189115_dp, 0.008261528702862685_dp, &
             0.014191502731255152_dp, 0.023997545648579352_dp, 0.039998042618849663_dp, &
             0.0657871441778338_dp, 0.1068856464861666_dp, 0.1717082822357783_dp, &
             0.27301595559947645_dp, 0.4301756299678513_dp, 0.6730008855624591_dp, &
             1.0494855277800719_dp, 1.6458419594230096_dp, 2.6581535061540267_dp, &
             4.824463093496843_dp]

      CASE (770)

         aw(:) = & ! a
            [1.3305128161768152e-05_dp, 7.484483849903708e-05_dp, 0.00020921431279245214_dp, &
             0.00047366931164595977_dp, 0.0009847260553730605_dp, 0.001955478929988737_dp, &
             0.003759587969985003_dp, 0.0070389264410185_dp, 0.012877741347257892_dp, &
             0.023079205235352323_dp, 0.04059917197933613_dp, 0.0702191345034593_dp, &
             0.119579941024401_dp, 0.2007556469243858_dp, 0.33263274752569577_dp, &
             0.544494103839673_dp, 0.8814421372321949_dp, 1.4128003770418798_dp, &
             2.2460045209204793_dp, 3.553113386180909_dp, 5.636518976441996_dp, &
             9.182887253811643_dp, &
             ! w
             3.4646946406306446e-05_dp, 9.198028959622438e-05_dp, 0.00018589042112813993_dp, &
             0.0003618585743323598_dp, 0.0006947409657712479_dp, 0.00130729901127845_dp, &
             0.0024050786957438806_dp, 0.004329824182817451_dp, 0.007641295559829068_dp, &
             0.013243340979253797_dp, 0.02257669593810452_dp, 0.0379110871371775_dp, &
             0.06278413125383121_dp, 0.10265672379663088_dp, 0.16589002088145904_dp, &
             0.26521480370347716_dp, 0.420016957188921_dp, 0.6602079616764603_dp, &
             1.0339757942811347_dp, 1.6277861052124318_dp, 2.637844018849805_dp, &
             4.801291850775574_dp]

      CASE (771)

         aw(:) = & ! a
            [1.1393402164739862e-05_dp, 6.435537671953513e-05_dp, 0.00018138473526022527_dp, &
             0.0004154658049068593_dp, 0.0008747846519680304_dp, 0.0017584267754467177_dp, &
             0.0034183517063473734_dp, 0.0064639213400214295_dp, 0.011931731255979431_dp, &
             0.02155664651632615_dp, 0.03819894044585809_dp, 0.06650990178628932_dp, &
             0.11395839375472008_dp, 0.19239959940803186_dp, 0.3204544521635997_dp, &
             0.5271041170550562_dp, 0.8571407364769187_dp, 1.3796195206015454_dp, &
             2.2018218543046926_dp, 3.4958359614134924_dp, 5.564258785183875_dp, &
             9.093791701582534_dp, &
             ! w
             2.9694924612056838e-05_dp, 7.948303904951842e-05_dp, 0.00016302622437119695_dp, &
             0.00032261147439344574_dp, 0.0006282090311303426_dp, 0.001196044810455543_dp, &
             0.0022226182809036185_dp, 0.00403689706422689_dp, 0.00718093917005208_dp, &
             0.012534794592718989_dp, 0.021508319513862697_dp, 0.036332889688642844_dp, &
             0.060501229605834685_dp, 0.09942620520485417_dp, 0.1614253186545218_dp, &
             0.259203569677637_dp, 0.4121593960237102_dp, 0.650279574358099_dp, &
             1.021905185920963_dp, 1.613705112733399_dp, 2.6219887573832_dp, &
             4.783206506353981_dp]

      CASE (772)

         aw(:) = & ! a
            [9.994542740935452e-06_dp, 5.666436362998256e-05_dp, 0.00016089650336480974_dp, &
             0.0003723340067523509_dp, 0.0007926338891327812_dp, 0.0016099040749552114_dp, &
             0.00315902992422393_dp, 0.006023642179884833_dp, 0.011202408166728408_dp, &
             0.02037554156616011_dp, 0.03632647691947301_dp, 0.06360135519041037_dp, &
             0.10952954268536703_dp, 0.18578792300483257_dp, 0.3107801275293393_dp, &
             0.5132392549917393_dp, 0.8377008664344392_dp, 1.3529965228826508_dp, &
             2.166276717918943_dp, 3.4496505228681706_dp, 5.5058835064249285_dp, &
             9.021714441715968_dp, &
             ! w
             2.6069739402847784e-05_dp, 7.029728726324608e-05_dp, 0.00014608462923361386_dp, &
             0.0002931967587981263_dp, 0.0005777823809891892_dp, 0.001110924663326483_dp, &
             0.002081910677238233_dp, 0.0038094440103938506_dp, 0.0068212874581884365_dp, &
             0.011978179782407453_dp, 0.020664796632037394_dp, 0.035081077350493484_dp, &
             0.0586827296947812_dp, 0.09684271394637756_dp, 0.15784181259572214_dp, &
             0.25436254930586366_dp, 0.40581209123918516_dp, 0.6422376969978276_dp, &
             1.012105986377126_dp, 1.602254840102212_dp, 2.609084595329209_dp, &
             4.768489877216812_dp]

      CASE (773)

         aw(:) = & ! a
            [8.923520940479506e-06_dp, 5.0765152004158595e-05_dp, 0.00014512398861934574_dp, &
             0.0003389345899264853_dp, 0.0007285553678155295_dp, 0.0014931851601751318_dp, &
             0.0029538002043550055_dp, 0.005672964011455442_dp, 0.010618142015169094_dp, &
             0.019424393503979655_dp, 0.03481141243349775_dp, 0.06123778597314709_dp, &
             0.10591630117205575_dp, 0.18037429604053778_dp, 0.3028324674436274_dp, &
             0.5018142441476405_dp, 0.8216373007259332_dp, 1.3309420417081383_dp, &
             2.1367655022625502_dp, 3.4112320115164585_dp, 5.4572497151715815_dp, &
             8.961594707558222_dp, &
             ! w
             2.329302182997228e-05_dp, 6.323607697802891e-05_dp, 0.0001329671336277957_dp, &
             0.00027019249350873344_dp, 0.0005379658526191249_dp, 0.001043178249812817_dp, &
             0.0019691737815543857_dp, 0.003626147327327001_dp, 0.006529960129177719_dp, &
             0.011525208320803339_dp, 0.019975432669217648_dp, 0.03405407285550966_dp, &
             0.057185480813742605_dp, 0.09470860379235609_dp, 0.15487261540100639_dp, &
             0.2503401511976684_dp, 0.40052464004716265_dp, 0.6355234486161233_dp, &
             1.003909043893088_dp, 1.592663469711052_dp, 2.59826755537669_dp, &
             4.7561552432941925_dp]

      CASE (774)

         aw(:) = & ! a
            [8.07533230400323e-06_dp, 4.608564416217443e-05_dp, 0.00013257085245278527_dp, &
             0.0003122109154513104_dp, 0.0006769508137699142_dp, 0.0013985669226681742_dp, &
             0.0027864069410812966_dp, 0.005385343782269933_dp, 0.010136534449276044_dp, &
             0.01863682403125464_dp, 0.033551777660902685_dp, 0.05926539091365332_dp, &
             0.10289082164951309_dp, 0.17582721190773956_dp, 0.2961379458275407_dp, &
             0.4921655358394929_dp, 0.8080389027396179_dp, 1.3122319478421456_dp, &
             2.111681724544952_dp, 3.378523977173361_dp, 5.415789890596947_dp, &
             8.910291894033268_dp, &
             ! w
             2.10932070918042e-05_dp, 5.76236612613594e-05_dp, 0.0001224724628077001_dp, &
             0.00025162299940528375_dp, 0.0005055558308953711_dp, 0.0009876537809585014_dp, &
             0.0018762422407666616_dp, 0.0034742949748948655_dp, 0.006287535961091847_dp, &
             0.011146764137828545_dp, 0.019397392996334844_dp, 0.03319005037958525_dp, &
             0.05592198483939459_dp, 0.09290259021370115_dp, 0.1523533496481229_dp, &
             0.2469190863149249_dp, 0.39601781128894525_dp, 0.6297893804124222_dp, &
             0.9968973940013939_dp, 1.584449251744576_dp, 2.5889978932065185_dp, &
             4.745586299698636_dp]

      CASE (775)

         aw(:) = & ! a
            [7.385803069679391e-06_dp, 4.227574694811219e-05_dp, 0.0001223191163656762_dp, &
             0.0002902801054977791_dp, 0.000634352507020876_dp, 0.0013200016688163828_dp, &
             0.0026466560925759008_dp, 0.005144039699098433_dp, 0.0097306998692072_dp, &
             0.01797053349015137_dp, 0.032482296228305806_dp, 0.05758529670594118_dp, &
             0.10030605859938087_dp, 0.17193195077297738_dp, 0.29038882231680097_dp, &
             0.4838605491838165_dp, 0.796309975940986_dp, 1.2960638180350468_dp, &
             2.0899698717806348_dp, 3.350172551857285_dp, 5.379810930581977_dp, &
             8.86573240555866_dp, &
             ! w
             1.930427853367265e-05_dp, 5.304586990760832e-05_dp, 0.00011386068161043624_dp, &
             0.00023626199362835186_dp, 0.00047854702457381354_dp, 0.0009411025287812773_dp, &
             0.001797934521557631_dp, 0.003345775885001045_dp, 0.006081561709059225_dp, &
             0.010824092794473432_dp, 0.018902971335607945_dp, 0.032448864559875425_dp, &
             0.05483522091516178_dp, 0.09134536694924299_dp, 0.15017619103932486_dp, &
             0.2439564036196247_dp, 0.39210740982256_dp, 0.6248057509645889_dp, &
             0.9907947822223216_dp, 1.5772925510685056_dp, 2.5809172423866595_dp, &
             4.736373936739432_dp]

      CASE (776)

         aw(:) = & ! a
            [4.115776479310656e-06_dp, 2.411091715911636e-05_dp, 7.289598787314097e-05_dp, &
             0.00018270555863340862_dp, 0.00042113531761360246_dp, 0.000918895701336235_dp, &
             0.0019201355511383707_dp, 0.003869038599387592_dp, 0.00755495901793931_dp, &
             0.014351494332714721_dp, 0.026604353815777797_dp, 0.04825198768892436_dp, &
             0.08580625861471436_dp, 0.1498846529828027_dp, 0.25758129461707013_dp, &
             0.43611143227920507_dp, 0.7284124343155312_dp, 1.2018882628721763_dp, &
             1.9628093615255051_dp, 3.1833455512262274_dp, 5.167294070117107_dp, &
             8.601776929190557_dp, &
             ! w
             1.0810345188889222e-05_dp, 3.107711443421804e-05_dp, 7.162742980568071e-05_dp, &
             0.00015881869889964977_dp, 0.00033901739593960846_dp, 0.0006958031127464125_dp, &
             0.0013783701049361041_dp, 0.002647137342734728_dp, 0.004947366366394651_dp, &
             0.009026606995623458_dp, 0.016119619786181386_dp, 0.028236086579038488_dp, &
             0.04860351975916543_dp, 0.0823430936734954_dp, 0.1374952980316822_dp, &
             0.22658057386609046_dp, 0.3690284268622359_dp, 0.5952276920259331_dp, &
             0.9544053187516136_dp, 1.534469552748104_dp, 2.532477479787969_dp, &
             4.681167161159148_dp]

      CASE (777)

         aw(:) = & ! a
            [2.9305924895869842e-06_dp, 1.746588021850344e-05_dp, 5.4458374568486694e-05_dp, &
             0.00014137766009695115_dp, 0.00033653439896220847_dp, 0.0007548527496616238_dp, &
             0.0016149024336884344_dp, 0.0033205479987406367_dp, 0.006599245851076902_dp, &
             0.012732075450476141_dp, 0.02393020827089087_dp, 0.043942071931938026_dp, &
             0.07901966773713946_dp, 0.13943828233577243_dp, 0.2418621302578669_dp, &
             0.4129994860663833_dp, 0.6952432263860158_dp, 1.1554978966979625_dp, &
             1.8997094388903577_dp, 3.1000417425438385_dp, 5.06063560401213_dp, &
             8.468797162886261_dp, &
             ! w
             7.725516861387516e-06_dp, 2.2948552619922402e-05_dp, 5.5404525147651974e-05_dp, &
             0.0001277430275140598_dp, 0.0002809706840034871_dp, 0.0005907844113491288_dp, &
             0.001194401745754514_dp, 0.002334424663937724_dp, 0.004430404753644415_dp, &
             0.008193961165177384_dp, 0.0148113872318636_dp, 0.026229688892851245_dp, &
             0.0455996442864698_dp, 0.07795561204486247_dp, 0.1312520748882399_dp, &
             0.2179461285374548_dp, 0.3574628892842476_dp, 0.5802941745418465_dp, &
             0.9359175063050069_dp, 1.5126121941158353_dp, 2.5076925839219726_dp, &
             4.652930527115194_dp]

      CASE (778)

         aw(:) = & ! a
            [2.306174184144005e-06_dp, 1.3942077704310813e-05_dp, 4.4542827301993846e-05_dp, &
             0.00011870317245919812_dp, 0.00028914976624303294_dp, 0.0006612451346436_dp, &
             0.0014378812907588148_dp, 0.0029979405922768224_dp, 0.006030161184301492_dp, &
             0.011757257796546421_dp, 0.022304869364607207_dp, 0.04129974306377381_dp, &
             0.07482633777651287_dp, 0.13293784043221613_dp, 0.23201756887371683_dp, &
             0.39844020405143754_dp, 0.6742374529450357_dp, 1.125979071636211_dp, &
             1.859389038737105_dp, 3.046619981435971_dp, 4.992038216815227_dp, &
             8.383085655430445_dp, &
             ! w
             6.09799434345543e-06_dp, 1.860312946108542e-05_dp, 4.650264066094958e-05_dp, &
             0.00011021580996852004_dp, 0.0002475195510616622_dp, 0.000529229248136808_dp, &
             0.0010850393997767066_dp, 0.002146259565366962_dp, 0.004116013484315798_dp, &
             0.007682782253863377_dp, 0.014001409375520133_dp, 0.024977915861741543_dp, &
             0.04371248294521763_dp, 0.07518165279510011_dp, 0.12728180337618636_dp, &
             0.21242588020667255_dp, 0.35003287938612626_dp, 0.5706592850049449_dp, &
             0.923946461035049_dp, 1.4984214987473523_dp, 2.491578285380307_dp, &
             4.634575679349051_dp]

      CASE (779)

         aw(:) = & ! a
            [1.9167341710834677e-06_dp, 1.1732939415540843e-05_dp, 3.825587180532255e-05_dp, &
             0.00010410275220852359_dp, 0.000258167492293394_dp, 0.0005992089507552246_dp, &
             0.001319197775231019_dp, 0.002779480057935501_dp, 0.005641434848854611_dp, &
             0.011086290577696107_dp, 0.021178560626502488_dp, 0.03945759882288879_dp, &
             0.07188696226609412_dp, 0.1283588310449827_dp, 0.2250519566356205_dp, &
             0.38809690292211496_dp, 0.6592596319494779_dp, 1.104861800226036_dp, &
             1.830460683345014_dp, 3.008196909414185_dp, 4.942601492904754_dp, &
             8.321222911182769_dp, &
             ! w
             5.081836203782143e-06_dp, 1.5861174079219656e-05_dp, 4.0769226803539744e-05_dp, &
             9.869784994404195e-05_dp, 0.00022519886548116018_dp, 0.0004876597434992643_dp, &
             0.0010104445186403786_dp, 0.0020168119876538_dp, 0.0038981070805253325_dp, &
             0.007326130843566232_dp, 0.013432935660525535_dp, 0.024094683916204878_dp, &
             0.042374485546041556_dp, 0.07320623685794946_dp, 0.12444304818760343_dp, &
             0.20846432479715432_dp, 0.34468295739329824_dp, 0.5637011756413218_dp, &
             0.9152796880265074_dp, 1.4881287118735915_dp, 2.47987867707641_dp, &
             4.621251074311755_dp]

      CASE (780)

         aw(:) = & ! a
            [1.648951545391286e-06_dp, 1.0207279492425605e-05_dp, 3.387212006975242e-05_dp, &
             9.379265637238277e-05_dp, 0.00023602160189721934_dp, 0.000554396087801814_dp, &
             0.0012326922793165042_dp, 0.0026190217439584673_dp, 0.005354011431551406_dp, &
             0.010587280574386267_dp, 0.020336579278475413_dp, 0.0380741487753279_dp, &
             0.06967036938497831_dp, 0.12489290807845509_dp, 0.21976177084751816_dp, &
             0.3802173983142883_dp, 0.6478179384326487_dp, 1.0886900017827261_dp, &
             1.8082585177804609_dp, 2.9786525452726442_dp, 4.904531146112459_dp, &
             8.273529961762002_dp, &
             ! w
             4.382497054527811e-06_dp, 1.3957069349999956e-05_dp, 3.6719443766788946e-05_dp, &
             9.043266655744782e-05_dp, 0.00020899192642057725_dp, 0.0004571960868449991_dp, &
             0.0009553580618671491_dp, 0.001920589519674794_dp, 0.003735202836930827_dp, &
             0.007058155421475286_dp, 0.013003880346014178_dp, 0.02342536463770226_dp, &
             0.04135682311212029_dp, 0.07169874625232926_dp, 0.12227011332108968_dp, &
             0.20542348912656067_dp, 0.3405660565885237_dp, 0.5583347438851695_dp, &
             0.9085828710304654_dp, 1.4801643254499464_dp, 2.4708188991495628_dp, &
             4.6109339171688815_dp]

      CASE (781)

         aw(:) = & ! a
            [1.4526713724223414e-06_dp, 9.08473901898181e-06_dp, 3.061944212328142e-05_dp, &
             8.6060184739444e-05_dp, 0.00021924418970420545_dp, 0.0005201525745205668_dp, &
             0.0011661063155848053_dp, 0.002494742688117069_dp, 0.005130198087643422_dp, &
             0.010196883032712201_dp, 0.01967513142984135_dp, 0.03698332459219403_dp, &
             0.06791685338341993_dp, 0.12214290048939122_dp, 0.2155530107827441_dp, &
             0.3739333319546899_dp, 0.6386728183862534_dp, 1.075738619559729_dp, &
             1.7904466363100695_dp, 2.9549150974802854_dp, 4.873906826343713_dp, &
             8.235130859142439_dp, &
             ! w
             3.869508185825249e-06_dp, 1.2549273531393551e-05_dp, 3.3681257015489096e-05_dp, &
             8.415120164832965e-05_dp, 0.00019655669995435456_dp, 0.00043364660810554524_dp, &
             0.000912509887585908_dp, 0.001845347890115005_dp, 0.0036072326385272876_dp, &
             0.006846792419674942_dp, 0.012664245883225053_dp, 0.02289382449863763_dp, &
             0.040546279994912476_dp, 0.07049487030033823_dp, 0.12053059943964924_dp, &
             0.2029837898330989_dp, 0.33725637860402813_dp, 0.5540128513490986_dp, &
             0.9031814650529043_dp, 1.473733373092902_dp, 2.4634990612961887_dp, &
             4.60259878286634_dp]

      CASE (782)

         aw(:) = & ! a
            [1.302152015147328e-06_dp, 8.220987115411935e-06_dp, 2.809776578399628e-05_dp, &
             8.000931575594767e-05_dp, 0.00020600219135109138_dp, 0.0004929277931648305_dp, &
             0.001112843794565352_dp, 0.00239481399463653_dp, 0.004949431534257316_dp, &
             0.009880341950542684_dp, 0.019136975259070224_dp, 0.03609311628878412_dp, &
             0.06648191824305334_dp, 0.1198869780210468_dp, 0.21209274232730374_dp, &
             0.36875642251731294_dp, 0.6311252111737147_dp, 1.0650321948005745_dp, &
             1.7757010433136946_dp, 2.9352399395067517_dp, 4.848498303123656_dp, &
             8.203248314261058_dp, &
             ! w
             3.475861712822614e-06_dp, 1.1461307719794296e-05_dp, 3.130309493200322e-05_dp, &
             7.918023841905698e-05_dp, 0.00018663691027749014_dp, 0.0004147429619161429_dp, &
             0.0008779365270941332_dp, 0.0017843690308584762_dp, 0.003503123593767545_dp, &
             0.006674261491088326_dp, 0.012386181104375116_dp, 0.02245747680967902_dp, &
             0.03987928446027479_dp, 0.06950202202471_dp, 0.11909312697425786_dp, &
             0.2009640190312887_dp, 0.3345118292149534_dp, 0.5504236556455473_dp, &
             0.8986902230626256_dp, 1.468381139745119_dp, 2.457404013321746_dp, &
             4.595658720409666_dp]

      CASE (783)

         aw(:) = & ! a
            [1.1827724789269656e-06_dp, 7.533824825883719e-06_dp, 2.607793421357581e-05_dp, &
             7.512249679929652e-05_dp, 0.00019522749095903492_dp, 0.00047063632726025115_dp, &
             0.0010690036045667734_dp, 0.0023121971521445875_dp, 0.004799410404167728_dp, &
             0.00961676624548491_dp, 0.018687556492606776_dp, 0.035347766901986766_dp, &
             0.06527769440901537_dp, 0.11798981395869498_dp, 0.20917727098173616_dp, &
             0.36438714033118746_dp, 0.6247452529803975_dp, 1.0559695971825915_dp, &
             1.7632042510757646_dp, 2.918548094308634_dp, 4.8269244311471695_dp, &
             8.176160678265346_dp, &
             ! w
             3.16347496014358e-06_dp, 1.059233516825719e-05_dp, 2.9381883356749218e-05_dp, &
             7.512631084275109e-05_dp, 0.00017849152415551152_dp, 0.0003991373727952283_dp, &
             0.000849268614427134_dp, 0.0017336154988135998_dp, 0.0034161896648225287_dp, &
             0.006529781078865992_dp, 0.01215273334440355_dp, 0.022090309940213923_dp, &
             0.03931688523480803_dp, 0.06866330876504984_dp, 0.11787675211142076_dp, &
             0.199252260617322_dp, 0.3321825552745324_dp, 0.5473737497547564_dp, &
             0.8948698149733284_dp, 1.4638247884739426_dp, 2.452213116837003_dp, &
             4.589748443532981_dp]

      CASE (784)

         aw(:) = & ! a
            [1.085589752657538e-06_dp, 6.9728537617290735e-06_dp, 2.4418713178803406e-05_dp, &
             7.10783551714023e-05_dp, 0.00018625175765404317_dp, 0.00045196433874997654_dp, &
             0.0010321132651102676_dp, 0.002242407990007849_dp, 0.004672261828590809_dp, &
             0.00939273150226986_dp, 0.018304590395138598_dp, 0.03471120028933651_dp, &
             0.06424716020214687_dp, 0.11636335404415295_dp, 0.20667373180833964_dp, &
             0.360629672694155_dp, 0.6192513576853955_dp, 1.0481563321712526_dp, &
             1.7524189111155486_dp, 2.904129346601461_dp, 4.808275070590371_dp, &
             8.152732459150297_dp, &
             ! w
             2.9090419927337045e-06_dp, 9.880348014549353e-06_dp, 2.77915178824644e-05_dp, &
             7.174257694053725e-05_dp, 0.0001716519919076494_dp, 0.0003859723114971427_dp, &
             0.0008249907138330951_dp, 0.0016904933160286398_dp, 0.0033421181002970674_dp, &
             0.006406371814138363_dp, 0.011952893071099027_dp, 0.021775382022870965_dp, &
             0.0388336452846908_dp, 0.06794148598001919_dp, 0.11682836622087606_dp, &
             0.19777493463890497_dp, 0.3301698525099383_dp, 0.5447355297055014_dp, &
             0.8915621148708586_dp, 1.459877249681947_dp, 2.4477141753149776_dp, &
             4.584626217811039_dp]

      CASE (785)

         aw(:) = & ! a
            [6.224431281538362e-07_dp, 4.27249258996541e-06_dp, 1.6252943498189566e-05_dp, &
             5.0672355137218765e-05_dp, 0.00013997020529883417_dp, 0.000353952707704477_dp, &
             0.0008355872390364983_dp, 0.0018659615297569223_dp, 0.003979065943914597_dp, &
             0.008159972875517616_dp, 0.016180116573415067_dp, 0.031154373878512873_dp, &
             0.05845186877074167_dp, 0.10716378586598481_dp, 0.19243913316093703_dp, &
             0.3391644085997884_dp, 0.5877323407461922_dp, 1.0031592819519326_dp, &
             1.6900961601207674_dp, 2.8205718888427906_dp, 4.699951071061254_dp, &
             8.016417246117232_dp, &
             ! w
             1.694475270981636e-06_dp, 6.407588365259449e-06_dp, 1.9757970289128656e-05_dp, &
             5.4186085421936505e-05_dp, 0.0001354758978179595_dp, 0.0003152817236420061_dp, &
             0.0006929994120358675_dp, 0.001453568245346649_dp, 0.002931426723551464_dp, &
             0.00571664630206478_dp, 0.010828077297736145_dp, 0.019991546704514262_dp, &
             0.036080807768958834_dp, 0.06380822909502767_dp, 0.11079684246112612_dp, &
             0.1892391105384695_dp, 0.3184954007460031_dp, 0.5293799939902119_dp, &
             0.8722539267280729_dp, 1.4367839721200395_dp, 2.4213641138161845_dp, &
             4.55462934975407_dp]

      CASE (786)

         aw(:) = & ! a
            [4.532562799594293e-07_dp, 3.26878734999866e-06_dp, 1.310279521964712e-05_dp, &
             4.249209846715176e-05_dp, 0.00012082093518252708_dp, 0.00031235986624583327_dp, &
             0.0007504521403401499_dp, 0.0017000678880343262_dp, 0.0036691213185990135_dp, &
             0.007601850177671225_dp, 0.015207747118217482_dp, 0.029510726995139624_dp, &
             0.05575087781395988_dp, 0.10284335228442375_dp, 0.18570808382078036_dp, &
             0.32895131097217595_dp, 0.5726520401592662_dp, 0.9815230690057081_dp, &
             1.6599978550476977_dp, 2.7800681697240037_dp, 4.647284785520889_dp, &
             7.949995037481279_dp, &
             ! w
             1.249717643725271e-06_dp, 5.08693360054884e-06_dp, 1.6531295536842016e-05_dp, &
             4.685988762883603e-05_dp, 0.00011996777012040138_dp, 0.00028433812458919413_dp, &
             0.0006342293017384891_dp, 0.0013465546833093518_dp, 0.0027436382184886014_dp, &
             0.005397887809564109_dp, 0.010303335618756094_dp, 0.019152382420540383_dp, &
             0.03477605767786949_dp, 0.06183590292962856_dp, 0.10790095757559305_dp, &
             0.18511790591877123_dp, 0.31283032173907516_dp, 0.5218952916023468_dp, &
             0.8628071245651988_dp, 1.425453395965989_dp, 2.408415787742415_dp, &
             4.539891226053357_dp]

      CASE (787)

         aw(:) = & ! a
            [3.63740571041785e-07_dp, 2.7312576145227483e-06_dp, 1.137336857547309e-05_dp, &
             3.789505483325795e-05_dp, 0.0001098600318011658_dp, 0.00028820684073496054_dp, &
             0.0007004376614657953_dp, 0.0016016735442746317_dp, 0.0034838006608538365_dp, &
             0.0072658277810501385_dp, 0.014618801473853648_dp, 0.028509940886325216_dp, &
             0.05409859250720024_dp, 0.10018935036277664_dp, 0.18155774890926699_dp, &
             0.32263269714411774_dp, 0.5632938703862731_dp, 0.9680601811056361_dp, &
             1.6412249262644083_dp, 2.7547540453008534_dp, 4.6143157467676685_dp, &
             7.9083648631221655_dp, &
             ! w
             1.0140963202412089e-06_dp, 4.368314114037337e-06_dp, 1.4715669445840793e-05_dp, &
             4.264716986550399e-05_dp, 0.00011091372630792967_dp, 0.00026605898017014904_dp, &
             0.0005991797662263173_dp, 0.0012822231428428461_dp, 0.0026299791274983914_dp, &
             0.005203820117594278_dp, 0.009982205853548788_dp, 0.0186364744122212_dp, &
             0.03397061525051465_dp, 0.06061384263455983_dp, 0.10610063141309317_dp, &
             0.18254800907116142_dp, 0.3092879761443027_dp, 0.517203741948951_dp, &
             0.8568735416170536_dp, 1.418325657440262_dp, 2.400263528805926_dp, &
             4.530612827850915_dp]

      CASE (788)

         aw(:) = & ! a
            [3.077840821133388e-07_dp, 2.392009243763714e-06_dp, 1.0261226862316773e-05_dp, &
             3.488932893869941e-05_dp, 0.00010260157530157011_dp, 0.00027205396680495913_dp, &
             0.0006667250573443441_dp, 0.0015349201867330272_dp, 0.0033573902089459665_dp, &
             0.00703555583758743_dp, 0.014213578734120918_dp, 0.027818919981082824_dp, &
             0.05295415884281253_dp, 0.09834596716185788_dp, 0.17866785975512275_dp, &
             0.3182231472326581_dp, 0.556749934505825_dp, 0.9586289559770258_dp, &
             1.6280530290652069_dp, 2.736968693465074_dp, 4.5911271573872705_dp, &
             7.879061189229068_dp, &
             ! w
             8.666908627455346e-07_dp, 3.909108559965584e-06_dp, 1.3527428823570999e-05_dp, &
             3.984903963722437e-05_dp, 0.0001048373469555314_dp, 0.00025369313971816075_dp, &
             0.0005753153350118134_dp, 0.0012381854596213628_dp, 0.002551818588850117_dp, &
             0.005069836800714532_dp, 0.009759731363646357_dp, 0.018277963193603128_dp, &
             0.033409368173551036_dp, 0.05976018738276161_dp, 0.1048402268183189_dp, &
             0.18074518466516518_dp, 0.30679841931444884_dp, 0.5139011879036711_dp, &
             0.8526909757713815_dp, 1.4132961911822142_dp, 2.3945079194948695_dp, &
             4.524062469931456_dp]

      CASE (789)

         aw(:) = & ! a
            [2.6926977919311533e-07_dp, 2.1566268938473792e-06_dp, 9.477903514513002e-06_dp, &
             3.274516769511483e-05_dp, 9.737388638862769e-05_dp, 0.0002603345121092903_dp, &
             0.0006421221340668872_dp, 0.0014859712320849632_dp, 0.0032643240984918517_dp, &
             0.006865445025918062_dp, 0.0139133395421499_dp, 0.027305600239076418_dp, &
             0.05210207861220421_dp, 0.0969706898911303_dp, 0.17650788556957064_dp, &
             0.3149219298090834_dp, 0.551843579752377_dp, 0.9515485315463394_dp, &
             1.6181528965309748_dp, 2.723587921962672_dp, 4.573667525440482_dp, &
             7.856984440823327_dp, &
             ! w
             7.651775550683405e-07_dp, 3.5872288207795254e-06_dp, 1.2679176671994036e-05_dp, &
             3.782937784152284e-05_dp, 0.00010041754572504688_dp, 0.00024464509670881596_dp, &
             0.0005577702254588655_dp, 0.0012056803701601768_dp, 0.0024939324169612983_dp, &
             0.004970319673215902_dp, 0.009594066991817172_dp, 0.01801040005199763_dp, &
             0.03298965947840887_dp, 0.05912065929547732_dp, 0.10389443439413525_dp, &
             0.17939036379430198_dp, 0.3049250197419523_dp, 0.511413065000559_dp, &
             0.8495367152420874_dp, 1.4095004062022107_dp, 2.3901623266247776_dp, &
             4.519116999694553_dp]

      CASE (790)

         aw(:) = & ! a
            [2.4103814806544383e-07_dp, 1.9828863468972035e-06_dp, 8.892458440490006e-06_dp, &
             3.112621582982953e-05_dp, 9.339672340844572e-05_dp, 0.0002513668688873825_dp, &
             0.0006232099401055233_dp, 0.0014482037763767143_dp, 0.0031922931500065793_dp, &
             0.0067334334125224715_dp, 0.01367980882935721_dp, 0.02690553004910012_dp, &
             0.051436807555299474_dp, 0.09589523094628315_dp, 0.17481640954167654_dp, &
             0.3123334626238051_dp, 0.5479921460740338_dp, 0.9459848232969446_dp, &
             1.6103665524174429_dp, 2.713056115376807_dp, 4.559916945373085_dp, &
             7.839589785978144_dp, &
             ! w
             6.907365523060062e-07_dp, 3.3475800016893185e-06_dp, 1.2038335949315703e-05_dp, &
             3.6290280018094995e-05_dp, 9.702897169395816e-05_dp, 0.00023767590692235968_dp, &
             0.0005442057794583839_dp, 0.0011804723784007742_dp, 0.0024489235489897404_dp, &
             0.004892766329469883_dp, 0.009464710820660667_dp, 0.017801114144700203_dp, &
             0.032660856364312145_dp, 0.05861894792059943_dp, 0.1031515205888081_dp, &
             0.17832494275276503_dp, 0.303450271132607_dp, 0.5094526130277873_dp, &
             0.8470494821875164_dp, 1.4065055786951166_dp, 2.3867326188554623_dp, &
             4.515213948548148_dp]

      CASE (791)

         aw(:) = & ! a
            [2.1940425684898104e-07_dp, 1.8489302148069205e-06_dp, 8.436257044826029e-06_dp, &
             2.985398050897613e-05_dp, 9.025192235315142e-05_dp, 0.00024424265728247696_dp, &
             0.0006081296472161716_dp, 0.0014179975675192727_dp, 0.0031345379416331687_dp, &
             0.006627358143039101_dp, 0.013491812648343057_dp, 0.026582945820073048_dp, &
             0.05089962089663782_dp, 0.09502572650655733_dp, 0.1734473028442072_dp, &
             0.3102361789668795_dp, 0.5448686960025444_dp, 0.9414690604327733_dp, &
             1.604042266031964_dp, 2.7044966813823037_dp, 4.548736083629629_dp, &
             7.825440765989207_dp, &
             ! w
             6.336743537208999e-07_dp, 3.1614202966509735e-06_dp, 1.1534502049589774e-05_dp, &
             3.5071675117424465e-05_dp, 9.433280382570607e-05_dp, 0.0002321098656265053_dp, &
             0.000533339583218343_dp, 0.0011602283047821111_dp, 0.0024127013554946916_dp, &
             0.004830239497664379_dp, 0.009360252756743556_dp, 0.017631874101152776_dp, &
             0.03239463583866337_dp, 0.0582122731713007_dp, 0.10254872177966001_dp, &
             0.17745966745714423_dp, 0.3022515700774478_dp, 0.507857954639636_dp, &
             0.8450250809198161_dp, 1.40406690406701_dp, 2.383939109892107_dp, &
             4.512034965549613_dp]

      CASE (792)

         aw(:) = & ! a
            [2.0227057615346106e-07_dp, 1.742254553449587e-06_dp, 8.069610841042915e-06_dp, &
             2.8824179706593397e-05_dp, 8.769316173762933e-05_dp, 0.0002384233050036885_dp, &
             0.0005957733895477248_dp, 0.0013931855242482302_dp, 0.003086997285041969_dp, &
             0.006539888142586533_dp, 0.013336552833242787_dp, 0.0263161776361762_dp, &
             0.05045485799901971_dp, 0.09430506579637335_dp, 0.1723114972063162_dp, &
             0.30849481371549203_dp, 0.5422733475117794_dp, 0.9377142755853102_dp, &
             1.5987806117765542_dp, 2.6973718606371424_dp, 4.539425449111783_dp, &
             7.813654956416135_dp, &
             ! w
             5.88470159408468e-07_dp, 3.0121956852682855e-06_dp, 1.1126518316635822e-05_dp, &
             3.4079082072637576e-05_dp, 9.21276649285828e-05_dp, 0.00022754323464210395_dp, &
             0.0005244020663760335_dp, 0.0011435428491755769_dp, 0.002382794184140662_dp, &
             0.00477853592166646_dp, 0.009273762682713809_dp, 0.0174915828904431_dp, &
             0.03217372492434475_dp, 0.05787449962652711_dp, 0.10204763281564107_dp, &
             0.17673984317098201_dp, 0.30125368565431776_dp, 0.5065296431584461_dp, &
             0.8433379427439142_dp, 1.402033728528429_dp, 2.381609607811765_dp, &
             4.509384060380935_dp]

      CASE (793)

         aw(:) = & ! a
            [1.8835198768753053e-07_dp, 1.655163141434324e-06_dp, 7.767853488699547e-06_dp, &
             2.7971422271936744e-05_dp, 8.556491566287217e-05_dp, 0.00023356693127181766_dp, &
             0.0005854348419665418_dp, 0.001372381054699316_dp, 0.003047064776996418_dp, &
             0.0064663060617234135_dp, 0.013205775084944834_dp, 0.026091220699612793_dp, &
             0.05007943009714656_dp, 0.09369621133269618_dp, 0.1713511447392023_dp, &
             0.30702139855422717_dp, 0.5400759543368066_dp, 0.9345334201280188_dp, &
             1.5943209939161322_dp, 2.6913305237500067_dp, 4.53152801808003_dp, &
             7.8036555420105485_dp, &
             ! w
             5.517399553949089e-07_dp, 2.889653492614516e-06_dp, 1.0788562895210595e-05_dp, &
             3.325274142870607e-05_dp, 9.028547059373886e-05_dp, 0.00022371808461766243_dp, &
             0.0005168998098551514_dp, 0.0011295123042554256_dp, 0.0023576084416140485_dp, &
             0.004734939322710739_dp, 0.009200753072023286_dp, 0.017373041759390886_dp, &
             0.03198690023257358_dp, 0.057588620343763344_dp, 0.10162322946489928_dp, &
             0.17612978922007602_dp, 0.30040748688745306_dp, 0.5054026696846113_dp, &
             0.8419059152590296_dp, 1.4003074290664757_dp, 2.3796313556839412_dp, &
             4.5071328990104815_dp]

      CASE (794)

         aw(:) = & ! a
            [1.2346376587197007e-07_dp, 1.2422701485189172e-06_dp, 6.300998213159383e-06_dp, &
             2.3749595418446697e-05_dp, 7.489006051760966e-05_dp, 0.00020896895769364996_dp, &
             0.0005326662805833826_dp, 0.0012655321399984989_dp, 0.0028409147850449144_dp, &
             0.006084773330042828_dp, 0.012525109756709202_dp, 0.024916510596851297_dp, &
             0.04811326408925896_dp, 0.09049931156486629_dp, 0.1662969836519555_dp, &
             0.29925100116945247_dp, 0.528465936613629_dp, 0.9176993713274656_dp, &
             1.5706850182920673_dp, 2.6592718941092843_dp, 4.489578414164911_dp, &
             7.750501858575766_dp, &
             ! w
             3.803623428754792e-07_dp, 2.297765919865831e-06_dp, 9.113818320269844e-06_dp, &
             2.909728582097957e-05_dp, 8.092623608086364e-05_dp, 0.0002041322787473526_dp, &
             0.0004782461633713341_dp, 0.0010568512125217107_dp, 0.0022266112574506887_dp, &
             0.004507338297061447_dp, 0.008818361313405623_dp, 0.016750399606735494_dp, &
             0.031003096001689958_dp, 0.056079761981500255_dp, 0.0993786263009129_dp, &
             0.17289728361889742_dp, 0.29591616630774_dp, 0.49941218028353873_dp, &
             0.8342843183732713_dp, 1.3911109483610424_dp, 2.3690871525765855_dp, &
             4.4951345242729905_dp]

      CASE (795)

         aw(:) = & ! a
            [1.0297967656240002e-07_dp, 1.1085496889876147e-06_dp, 5.8097120229033355e-06_dp, &
             2.2302564430945705e-05_dp, 7.117198100206764e-05_dp, 0.0002002989439588021_dp, &
             0.000513894394260242_dp, 0.0012272378227995335_dp, 0.002766574972144618_dp, &
             0.005946471176210231_dp, 0.012277270054106542_dp, 0.0244871162678801_dp, &
             0.04739210559800999_dp, 0.08932317821659377_dp, 0.16443252815740209_dp, &
             0.29637758102037026_dp, 0.5241633249195753_dp, 0.9114486796705003_dp, &
             1.5618937917651499_dp, 2.647330755963641_dp, 4.47393507086348_dp, &
             7.730663602175375_dp, &
             ! w
             3.261718111907601e-07_dp, 2.1010489906176792e-06_dp, 8.539149721413985e-06_dp, &
             2.764557002766996e-05_dp, 7.76156655549409e-05_dp, 0.00019713889179205033_dp, &
             0.0004643409697944965_dp, 0.0010305518984628523_dp, 0.002178953401583239_dp, &
             0.004424170574123751_dp, 0.008678097182038987_dp, 0.01652124188516167_dp, &
             0.030639934264252204_dp, 0.055521291114271565_dp, 0.09854583123662566_dp, &
             0.17169533993315336_dp, 0.2942428763766681_dp, 0.4971764802826427_dp, &
             0.8314357065956599_dp, 1.3876699254604818_dp, 2.36513945874929_dp, &
             4.490642595795109_dp]

      CASE (796)

         aw(:) = & ! a
            [9.744958304131219e-08_dp, 1.0720432579948353e-06_dp, 5.673844907917884e-06_dp, &
             2.189898390128409e-05_dp, 7.012897187268582e-05_dp, 0.00019785641975160915_dp, &
             0.0005085885147987631_dp, 0.0012163852625058323_dp, 0.002745461114725889_dp, &
             0.0059071183974160855_dp, 0.012206637633445712_dp, 0.024364573926812943_dp, &
             0.047186050009072215_dp, 0.08898676399255895_dp, 0.1638987215187362_dp, &
             0.2955541979246799_dp, 0.5229294621817844_dp, 0.9096549458515194_dp, &
             1.559369508336902_dp, 2.6439002808649326_dp, 4.469439189655929_dp, &
             7.724960396502579_dp, &
             ! w
             3.115272361025735e-07_dp, 2.046789561440139e-06_dp, 8.378811642172586e-06_dp, &
             2.7237910200027115e-05_dp, 7.668186167310419e-05_dp, 0.00019515965051577553_dp, &
             0.0004603951182157937_dp, 0.0010230727758135927_dp, 0.002165375536984058_dp, &
             0.004400438903430516_dp, 0.008638019146467266_dp, 0.01645568652631624_dp, &
             0.03053593511670295_dp, 0.055361210290825226_dp, 0.09830691482753337_dp, &
             0.17135025594719783_dp, 0.2937621357261106_dp, 0.49653376650515635_dp, &
             0.8306163712072967_dp, 1.3866798085553493_dp, 2.364003304613395_dp, &
             4.489349822212133_dp]

      CASE (797)

         aw(:) = & ! a
            [0.0013784527567334275_dp, 0.0073354844632425595_dp, 0.018359969705865502_dp, &
             0.03502942550932789_dp, 0.058289702515891675_dp, 0.08959653490759337_dp, &
             0.1311105964523301_dp, 0.18594175167845706_dp, 0.25843374169755695_dp, &
             0.35449679430290737_dp, 0.4820233176813449_dp, 0.6514394721188445_dp, &
             0.8764497569134696_dp, 1.1750392718170326_dp, 1.5708251549437284_dp, &
             2.0949094237780543_dp, 2.7885121062494354_dp, 3.706942131142376_dp, &
             4.926129947502942_dp, 6.554719238492812_dp, 8.760213250865423_dp, &
             11.839415233776139_dp, 16.497920468553467_dp, &
             ! w
             0.00354572268123833_dp, 0.008419694079078813_dp, 0.013723899364581463_dp, &
             0.019770325752220336_dp, 0.026989584770299464_dp, 0.0359800906589701_dp, &
             0.047561610147702445_dp, 0.06281920344881087_dp, 0.08314301912201755_dp, &
             0.11028942885960276_dp, 0.14648904740151114_dp, 0.19461084589556935_dp, &
             0.25838505593068395_dp, 0.34269989517715693_dp, 0.4540128542158546_dp, &
             0.6009620726696401_dp, 0.7953586283730821_dp, 1.0539711347571572_dp, &
             1.4021274023180144_dp, 1.8820102292923664_dp, 2.575302758463919_dp, &
             3.683469654055837_dp, 6.004428599991783_dp]

      CASE (798)

         aw(:) = & ! a
            [0.0009838433750908838_dp, 0.0052434702103699975_dp, 0.013161195082932432_dp, &
             0.025220671773101215_dp, 0.042227902831257706_dp, 0.06544576314488094_dp, &
             0.09677690802056031_dp, 0.1389878363126464_dp, 0.19596994186479394_dp, &
             0.2730571178152942_dp, 0.3774436108993491_dp, 0.5187535576303994_dp, &
             0.7098147054429543_dp, 0.9677002381150995_dp, 1.3151332568378022_dp, &
             1.782409081280922_dp, 2.4101128036163164_dp, 3.253179002717379_dp, &
             4.387491340512361_dp, 5.921965585700212_dp, 8.024496067416576_dp, &
             10.991668108635901_dp, 15.525222502428994_dp, &
             ! w
             0.002531568227368319_dp, 0.006029734418473026_dp, 0.009884808828892902_dp, &
             0.014366254036005877_dp, 0.019855613958281632_dp, 0.02689346062497072_dp, &
             0.036225608183600176_dp, 0.04883963416459039_dp, 0.06600672021458406_dp, &
             0.08935730037176266_dp, 0.12100695788899504_dp, 0.16373374434827284_dp, &
             0.22121102210315727_dp, 0.29831579902789673_dp, 0.4015553421818241_dp, &
             0.5396953344107225_dp, 0.7247638198482657_dp, 0.9738304012176817_dp, &
             1.312565267657188_dp, 1.7834173081309153_dp, 2.467966286243429_dp, &
             3.566291736228225_dp, 5.868424131929509_dp]

      CASE (799)

         aw(:) = & ! a
            [0.0007733451749411847_dp, 0.004126446422581911_dp, 0.010380457947273032_dp, &
             0.01996053388869254_dp, 0.033584305845319976_dp, 0.052390344963260546_dp, &
             0.07811061891535163_dp, 0.11327898502880308_dp, 0.1614767991746906_dp, &
             0.22764293851389547_dp, 0.31849418636000465_dp, 0.44310446733722775_dp, &
             0.6136936382370711_dp, 0.846691605275164_dp, 1.1641760565010792_dp, &
             1.5958420934257882_dp, 2.181781266464886_dp, 2.976610706787062_dp, &
             4.056132561496171_dp, 5.529429791006581_dp, 7.564698381371439_dp, &
             10.458494106139828_dp, 14.910230229878495_dp, &
             ! w
             0.0019904626321219174_dp, 0.004752115279114086_dp, 0.007825331882399034_dp, &
             0.011452292907696131_dp, 0.015982109728513053_dp, 0.021915147950907186_dp, &
             0.02994368605027909_dp, 0.04098489007279177_dp, 0.056226992719640136_dp, &
             0.07721521067732742_dp, 0.10598660941777005_dp, 0.14525308457861175_dp, &
             0.19863996923630922_dp, 0.27100336432018013_dp, 0.36886956796242293_dp, &
             0.5010780810072235_dp, 0.6797979865901491_dp, 0.9223087517289797_dp, &
             1.2545353217250466_dp, 1.719149774910329_dp, 2.397727246126472_dp, &
             3.489512104808864_dp, 5.779479043787369_dp]

      CASE (800)

         aw(:) = & ! a
            [0.0006411479248635202_dp, 0.003424399445834742_dp, 0.008630339909711753_dp, &
             0.01664311878123391_dp, 0.02811776389910474_dp, 0.04410343671493797_dp, &
             0.0662070484554279_dp, 0.09678960535069649_dp, 0.13920111363677218_dp, &
             0.19808610472624708_dp, 0.2798055840310687_dp, 0.3930214628429474_dp, &
             0.5494939208505794_dp, 0.7651598432960415_dp, 1.061593990328445_dp, &
             1.4680140159913153_dp, 2.0241075362987018_dp, 2.7842179701404697_dp, &
             3.824056597073975_dp, 5.252818375214606_dp, 7.238940571624984_dp, &
             10.079015448393468_dp, 14.47085627829025_dp, &
             ! w
             0.00165057711461096_dp, 0.003948383830537736_dp, 0.0065261492191724885_dp, &
             0.009606527488024248_dp, 0.013514866435068726_dp, 0.01872103203952206_dp, &
             0.025875861766723386_dp, 0.03584261034388069_dp, 0.04974756898427098_dp, &
             0.06907295452251423_dp, 0.09579652320540147_dp, 0.1325775235222265_dp, &
             0.18299955104276386_dp, 0.2518952848356535_dp, 0.3457966560788414_dp, &
             0.4735906749253684_dp, 0.647547528461446_dp, 0.8851055905287044_dp, &
             1.2123916202617506_dp, 1.6722672690680787_dp, 2.346339184988275_dp, &
             3.433280456586048_dp, 5.714423112450276_dp]

      CASE (801)

         aw(:) = & ! a
            [0.0005498583024251807_dp, 0.0029392878054117252_dp, 0.007419601975889576_dp, &
             0.014344127043174977_dp, 0.024320418024767258_dp, 0.038328989240943516_dp, &
             0.05787939326900643_dp, 0.08519686785179664_dp, 0.12344942293814261_dp, &
             0.17705030316606155_dp, 0.2520806034549217_dp, 0.3568763739092084_dp, &
             0.5028311343464295_dp, 0.7054842662408387_dp, 0.9859998577862228_dp, &
             1.3731996617639493_dp, 1.9064293570682802_dp, 2.6397921904362427_dp, &
             3.64890762639226_dp, 5.043049696225617_dp, 6.9908548830475175_dp, &
             9.788978588800319_dp, 14.134048054849288_dp, &
             ! w
             0.0014158323923938018_dp, 0.0033925732894293783_dp, 0.005625620447388377_dp, &
             0.008322706733836411_dp, 0.011790664474124317_dp, 0.016474978191613785_dp, &
             0.022993010863222533_dp, 0.032165010362384235_dp, 0.04506869825792983_dp, &
             0.0631369087896174_dp, 0.088299820935649_dp, 0.12317274045783322_dp, &
             0.17130240020923765_dp, 0.23749794674789146_dp, 0.32829024148621777_dp, &
             0.45259883871637807_dp, 0.6227708590109541_dp, 0.8563712770439723_dp, &
             1.1796933367531552_dp, 1.6357627605315581_dp, 2.3062327611809628_dp, &
             3.38935576249735_dp, 5.663655386078455_dp]

      CASE (802)

         aw(:) = & ! a
            [0.0004827586136186306_dp, 0.002582523947211592_dp, 0.00652828862179465_dp, &
             0.012649103360050854_dp, 0.021514853424688632_dp, 0.034051007309957686_dp, &
             0.05168819520599923_dp, 0.07654094288171073_dp, 0.111628798217135_dp, &
             0.1611764619157916_dp, 0.23103640646860754_dp, 0.3292771698582933_dp, &
             0.4669890452263814_dp, 0.6593797617232361_dp, 0.9272672650833411_dp, &
             1.299135552154824_dp, 1.8140338856516904_dp, 2.5258516610325015_dp, &
             3.5101187216445453_dp, 4.876166853234648_dp, 6.7928011209831025_dp, &
             9.556751007429314_dp, 13.863719858601991_dp, &
             ! w
             0.0012432682282453626_dp, 0.002983537501195038_dp, 0.004961553773013737_dp, &
             0.007373136756858046_dp, 0.010510088879566464_dp, 0.014797707747711709_dp, &
             0.020825501037925_dp, 0.02937835649869728_dp, 0.04149452516174276_dp, &
             0.058566527662603306_dp, 0.08248485775889107_dp, 0.11582703647554439_dp, &
             0.16210683219429933_dp, 0.22611061566030374_dp, 0.31436455122518836_dp, &
             0.43581138861819096_dp, 0.6028592074271457_dp, 0.8331774026770082_dp, &
             1.153200496001535_dp, 1.606098551942477_dp, 2.2735776855316536_dp, &
             3.3535651290761983_dp, 5.622321088044751_dp]

      CASE (803)

         aw(:) = & ! a
            [0.00043120767724843894_dp, 0.0023082969961048766_dp, 0.005842560341090457_dp, &
             0.011343280774860383_dp, 0.01934946584046834_dp, 0.030741052589663507_dp, &
             0.04688281625252862_dp, 0.0697965247979113_dp, 0.1023773579293564_dp, &
             0.1486921077225977_dp, 0.21440133766987204_dp, 0.3073479572801685_dp, &
             0.43836486430739935_dp, 0.6223761586956191_dp, 0.8799012584423649_dp, &
             1.2391299618116849_dp, 1.7388499853847126_dp, 2.4327580127959054_dp, &
             3.3962976302833914_dp, 4.738843753941742_dp, 6.629346875364611_dp, &
             9.364614445783106_dp, 13.639604346080795_dp, &
             ! w
             0.0011106766042351182_dp, 0.0026689417035429827_dp, 0.004449888870698982_dp, &
             0.006639522605736512_dp, 0.009517081808323325_dp, 0.013490723214619424_dp, &
             0.019126258123919746_dp, 0.027178820484671094_dp, 0.03865370429698751_dp, &
             0.05490953385749471_dp, 0.07780278985833888_dp, 0.10987783078707603_dp, &
             0.15461864598778818_dp, 0.21678988601948262_dp, 0.30291096660304506_dp, &
             0.4219415981812148_dp, 0.5863395969159725_dp, 0.8138626836365093_dp, &
             1.131067798319505_dp, 1.5812537188815419_dp, 2.246181767770242_dp, &
             3.323518954350447_dp, 5.587643191260664_dp]

      CASE (804)

         aw(:) = & ! a
            [0.00039027158064795385_dp, 0.002090439481548534_dp, 0.0052973435751850475_dp, &
             0.010303760621263385_dp, 0.017622761982581427_dp, 0.028095734426868503_dp, &
             0.04303129339429646_dp, 0.06437187847629526_dp, 0.09490634413203773_dp, &
             0.13856646706454756_dp, 0.200848327800129_dp, 0.28940059333293233_dp, &
             0.41483331627901554_dp, 0.5918234614832505_dp, 0.8406284613179826_dp, &
             1.1891777620404644_dp, 1.6760251186984416_dp, 2.3546918373842125_dp, &
             3.3005389718557225_dp, 4.622974102042791_dp, 6.4910751080232325_dp, &
             9.20172789790954_dp, 13.449272821141973_dp, &
             ! w
             0.0010053760724013485_dp, 0.0024188771323381017_dp, 0.00404251597165331_dp, &
             0.006054016537216315_dp, 0.008721881869050135_dp, 0.012439434628672922_dp, &
             0.017751980335480574_dp, 0.02538915439297688_dp, 0.03632814474153547_dp, &
             0.05189843339969331_dp, 0.07392676308606565_dp, 0.10492794145033446_dp, &
             0.14835882544547835_dp, 0.2089634726309885_dp, 0.2932533745167012_dp, &
             0.4102008435701887_dp, 0.5723051864213542_dp, 0.797400282775444_dp, &
             1.1121508696421945_dp, 1.5599717859461693_dp, 2.222679865928236_dp, &
             3.2977284201069734_dp, 5.557892989269995_dp]

      CASE (805)

         aw(:) = & ! a
            [0.00035692023310357295_dp, 0.0019128752334149788_dp, 0.004852633616926041_dp, &
             0.00945491631376296_dp, 0.016210590129236452_dp, 0.02592780369047276_dp, &
             0.03986646023375293_dp, 0.059900050843840584_dp, 0.08872505438366671_dp, &
             0.1301559011363722_dp, 0.18954536231254243_dp, 0.2743721725499398_dp, &
             0.3950504478761171_dp, 0.5660386399682379_dp, 0.8073610964746365_dp, &
             1.1467139624836917_dp, 1.6224394080489497_dp, 2.2878978340828704_dp, &
             3.218371500254624_dp, 4.523292909063757_dp, 6.371852918156335_dp, &
             9.061014679349572_dp, 13.284596435150553_dp, &
             ! w
             0.0009195778520630333_dp, 0.0022149609710024475_dp, 0.003709826622298377_dp, &
             0.005574782458897983_dp, 0.008068998522385486_dp, 0.011572767041690859_dp, &
             0.016613404423587844_dp, 0.023898367419202288_dp, 0.03438046582522217_dp, &
             0.049363688639634355_dp, 0.07064838010330315_dp, 0.10072270488821705_dp, &
             0.14301861890330705_dp, 0.20226068962866398_dp, 0.2849518213563298_dp, &
             0.4000737709446778_dp, 0.5601610685629225_dp, 0.7831142642081589_dp, &
             1.0956942346302607_dp, 1.541421433114326_dp, 2.2021674780100557_dp, &
             3.275206621507311_dp, 5.531925426035051_dp]

      CASE (806)

         aw(:) = & ! a
            [0.00019789189844647632_dp, 0.0010649878171288582_dp, 0.0027234651961959575_dp, &
             0.005374548717725806_dp, 0.009384283220862263_dp, 0.015369429211464058_dp, &
             0.024304580834858498_dp, 0.037657962063126806_dp, 0.057583691066766536_dp, &
             0.08720448876336913_dp, 0.13101920601121675_dp, 0.19547855085950874_dp, &
             0.28979052809416106_dp, 0.42704294498751244_dp, 0.6257676531360138_dp, &
             0.9121310962247107_dp, 1.3230460435749691_dp, 1.910735194796546_dp, &
             2.7498559549133126_dp, 3.9499044644464085_dp, 5.680778760748135_dp, &
             8.24008277332823_dp, 12.318831434261737_dp, &
             ! w
             0.0005103331140406049_dp, 0.001239524790388038_dp, 0.002109968177448312_dp, &
             0.003251753194812805_dp, 0.004868736593361478_dp, 0.00726180067785983_dp, &
             0.01085008506003253_dp, 0.016211062141110212_dp, 0.024154157048959544_dp, &
             0.03582779518743183_dp, 0.05286353718083628_dp, 0.0775714003961799_dp, &
             0.11320820677637615_dp, 0.16434942012955056_dp, 0.2374103604805345_dp, &
             0.3413949149535763_dp, 0.4890254624467801_dp, 0.6986038745110555_dp, &
             0.9975093266876792_dp, 1.4299875628567857_dp, 2.07837499462709_dp, &
             3.1390257639054786_dp, 5.375144375514531_dp]

      CASE (807)

         aw(:) = & ! a
            [0.00013996012907137545_dp, 0.0007553529554818944_dp, 0.0019423634055118642_dp, &
             0.0038671946033628658_dp, 0.006837743957151046_dp, 0.011378627141444005_dp, &
             0.01832468563502512_dp, 0.028945478778156647_dp, 0.045129013545104325_dp, &
             0.06965504563013547_dp, 0.10659117476438364_dp, 0.16185730083750358_dp, &
             0.24402375620195257_dp, 0.3654355201113753_dp, 0.543793856419_dp, &
             0.8043881771437233_dp, 1.1832813270735483_dp, 1.731983053099483_dp, &
             2.5247116633074644_dp, 3.6709230752851725_dp, 5.340889692053782_dp, &
             7.832666548614906_dp, 11.836060856280595_dp, &
             ! w
             0.0003611662950579334_dp, 0.0008822319573707234_dp, 0.0015185931673152625_dp, &
             0.002381116213671972_dp, 0.0036458905000068408_dp, 0.005573017590205212_dp, &
             0.008527700693144891_dp, 0.013024072938109452_dp, 0.019799793571297068_dp, &
             0.029920949184449853_dp, 0.04492487252084988_dp, 0.06701733613497259_dp, &
             0.09934715278462358_dp, 0.1463900696284206_dp, 0.21448958268011928_dp, &
             0.31263394697361147_dp, 0.45362164626001367_dp, 0.6559573637613074_dp, &
             0.9473640989598018_dp, 1.372525198685822_dp, 2.0141168666286617_dp, &
             3.0681362610503005_dp, 5.293686964509947_dp]

      CASE (808)

         aw(:) = & ! a
            [0.00010941640926740673_dp, 0.0005918230362322702_dp, 0.0015285156545133051_dp, &
             0.0030646401423472134_dp, 0.00547242071880935_dp, 0.009218909467412637_dp, &
             0.01505113987332291_dp, 0.02411393970450653_dp, 0.03812767312753826_dp, &
             0.059653980645693425_dp, 0.09248269052068217_dp, 0.14218708167089744_dp, &
             0.21691520940588246_dp, 0.32851301847789_dp, 0.494115228513199_dp, &
             0.7384028363074433_dp, 1.0968370085897698_dp, 1.6204083947121235_dp, &
             2.3829973908178337_dp, 3.493999488072853_dp, 5.123931585986614_dp, &
             7.571189815542324_dp, 11.52488375748675_dp, &
             ! w
             0.0002824892025218642_dp, 0.0006931357485954232_dp, 0.0012036148565504033_dp, &
             0.0019128614084413366_dp, 0.002979163271259069_dp, 0.0046364280084281485_dp, &
             0.00721593819731318_dp, 0.01119203499519941_dp, 0.01725603356047215_dp, &
             0.026419123606108694_dp, 0.040154210570503696_dp, 0.06059417807957928_dp, &
             0.0908103436881852_dp, 0.13520413437375134_dp, 0.20006106528557996_dp, &
             0.2943474886925393_dp, 0.43090193737392596_dp, 0.6283583257636025_dp, &
             0.9146730531333217_dp, 1.33484088250569_dp, 1.9718018663295296_dp, &
             3.02136860618871_dp, 5.240003704493937_dp]

      CASE (809)

         aw(:) = & ! a
            [9.037755850786657e-05_dp, 0.000489751888332228_dp, 0.0012695476399375437_dp, &
             0.0025604674384136746_dp, 0.004609893379999404_dp, 0.007844323272598373_dp, &
             0.012948790175995175_dp, 0.02098006948683223_dp, 0.03353970625295641_dp, &
             0.05303351539408618_dp, 0.08305097980689792_dp, 0.12891290356153357_dp, &
             0.1984567102137522_dp, 0.30315754192766375_dp, 0.45972468008999984_dp, &
             0.6923768859292504_dp, 1.0361118510975211_dp, 1.541512797682683_dp, &
             2.282185453474102_dp, 3.367461582867472_dp, 4.968034929819057_dp, &
             7.382574192857667_dp, 11.29972777109869_dp, &
             ! w
             0.00023343207909754665_dp, 0.0005749109301438361_dp, 0.0010056920640549231_dp, &
             0.0016163277989382093_dp, 0.002552329974714457_dp, 0.004028881825258552_dp, &
             0.0063533565414724364_dp, 0.009971877630473854_dp, 0.015542154829216576_dp, &
             0.024034745367677076_dp, 0.03687412483277579_dp, 0.05613759897836612_dp, &
             0.0848363994203121_dp, 0.12731293386377773_dp, 0.18980450153444506_dp, &
             0.2812550459994511_dp, 0.41452674601066225_dp, 0.6083454493292306_dp, &
             0.8908418849813289_dp, 1.3072512917110222_dp, 1.9407287816980878_dp, &
             2.9869788339832106_dp, 5.20055654459162_dp]

      CASE (810)

         aw(:) = & ! a
            [7.7300456987927e-05_dp, 0.000419563322849588_dp, 0.0010910900148242133_dp, &
             0.0022118842816014413_dp, 0.004010702133159692_dp, 0.006883389233119795_dp, &
             0.011468076893774755_dp, 0.018754911531701814_dp, 0.03025516498807919_dp, &
             0.04825546556733338_dp, 0.07619089536610027_dp, 0.11918619601999758_dp, &
             0.18483573362537256_dp, 0.28432227683445416_dp, 0.43401694673280855_dp, &
             0.6577678433173347_dp, 0.9901971311996591_dp, 1.4815533121200783_dp, &
             2.2052107528928055_dp, 3.27043895819725_dp, 4.8480676638924765_dp, &
             7.23699173845095_dp, 11.125529684989406_dp, &
             ! w
             0.00019972772702581205_dp, 0.0004935012692875426_dp, 0.0008688229862886769_dp, &
             0.0014099142447940786_dp, 0.002252492316877194_dp, 0.003597470247005841_dp, &
             0.005734167757991338_dp, 0.009087220565894907_dp, 0.014288287980816163_dp, &
             0.022275971953164032_dp, 0.034436245226751705_dp, 0.05280176817476985_dp, &
             0.08033491379405078_dp, 0.12132927248085588_dp, 0.18198099916899307_dp, &
             0.27121258100739354_dp, 0.4019009995490989_dp, 0.5928419231794434_dp, &
             0.8723039184348136_dp, 1.2857173127406785_dp, 1.916418490271973_dp, &
             2.960044346544269_dp, 5.169677212000795_dp]

      CASE (811)

         aw(:) = & ! a
            [6.772773294215754e-05_dp, 0.0003681330933123696_dp, 0.0009600833165707886_dp, &
             0.0019552449376276843_dp, 0.003567713091430119_dp, 0.006169065398236549_dp, &
             0.010360290808617492_dp, 0.01707873599614676_dp, 0.027763859202613803_dp, &
             0.04460692601859343_dp, 0.0709187089702277_dp, 0.11166511932544645_dp, &
             0.17424241515195532_dp, 0.26959358500617825_dp, 0.41381062527233786_dp, &
             0.6304335224229871_dp, 0.9537697039665054_dp, 1.43378406743243_dp, &
             2.1436511658590613_dp, 3.1925811542392095_dp, 4.7515127354614215_dp, &
             7.119533572817901_dp, 10.98471358527973_dp, &
             ! w
             0.00017504982451554808_dp, 0.0004337767668035506_dp, 0.0007680401893588274_dp, &
             0.0012570470205922682_dp, 0.0020286716203614854_dp, 0.0032724733594189375_dp, &
             0.005263494761006937_dp, 0.008409222390791752_dp, 0.013320223444780142_dp, &
             0.02090894585054932_dp, 0.032529588757453924_dp, 0.05017767149599278_dp, &
             0.07677453271404558_dp, 0.11657222338192556_dp, 0.17573107324937454_dp, &
             0.26315342393439656_dp, 0.39172577731792474_dp, 0.5802991932790549_dp, &
             0.857255440873771_dp, 1.268188460658404_dp, 1.8965911680238352_dp, &
             2.938056828544457_dp, 5.144479743330254_dp]

      CASE (812)

         aw(:) = & ! a
            [6.0397135406301614e-05_dp, 0.00032871433800046613_dp, 0.0008595068400582489_dp, &
             0.001757706352904466_dp, 0.0032254636863865284_dp, 0.005614498264353481_dp, &
             0.009495406762391986_dp, 0.015762292284963245_dp, 0.02579558236249796_dp, &
             0.041707759672396094_dp, 0.0667063565455527_dp, 0.10562468858539437_dp, &
             0.16569278135499657_dp, 0.25765138513470287_dp, 0.39735580760118655_dp, &
             0.6080832360815576_dp, 0.9238708226189551_dp, 1.3944377508884656_dp, &
             2.0927827862906323_dp, 3.1280599913739775_dp, 4.6712980071542_dp, &
             7.02175217649485_dp, 10.867298043308807_dp, &
             ! w
             0.00015614819092974898_dp, 0.0003879516988416004_dp, 0.0006904572032205239_dp, &
             0.0011387637173699851_dp, 0.0018542679667531696_dp, 0.0030172114402072705_dp, &
             0.004890963723966337_dp, 0.007868862647868726_dp, 0.012543866122957988_dp, &
             0.019806400926925403_dp, 0.030983736591709405_dp, 0.048039705818347815_dp, &
             0.07386036683039637_dp, 0.11266167870504212_dp, 0.17057228032319388_dp, &
             0.25647570162543065_dp, 0.3832645837381113_dp, 0.5698353295131204_dp, &
             0.844665248097364_dp, 1.2534888200926722_dp, 1.8799365962676922_dp, &
             2.9195734213407776_dp, 5.1233050498492485_dp]

      CASE (813)

         aw(:) = & ! a
            [5.459147845499838e-05_dp, 0.00029747088887302085_dp, 0.000779670018209787_dp, &
             0.0016005319095796884_dp, 0.0029522226375846665_dp, 0.005169805372288527_dp, &
             0.008798388562048485_dp, 0.014695770486066111_dp, 0.024192650709405416_dp, &
             0.03933484928834712_dp, 0.0632421417068238_dp, 0.10063463071936081_dp, &
             0.15859978684311307_dp, 0.24770416281717136_dp, 0.3835982601919537_dp, &
             0.5893307355179446_dp, 0.8987023544282555_dp, 1.3612157840316859_dp, &
             2.049712915507117_dp, 3.0732949168990396_dp, 4.603066338615398_dp, &
             6.938430645880391_dp, 10.76710713186747_dp, &
             ! w
             0.00014117583442286598_dp, 0.00035159534032034256_dp, 0.0006287212174442838_dp, &
             0.001044201420024743_dp, 0.0017139571050605786_dp, 0.0028104008428370573_dp, &
             0.004587115887159796_dp, 0.007425474593725979_dp, 0.011903394370300264_dp, &
             0.018892363300372656_dp, 0.02969636382923338_dp, 0.04625167578871546_dp, &
             0.07141347943047267_dp, 0.10936586916341903_dp, 0.1662090944495639_dp, &
             0.2508091237258758_dp, 0.3760624629387318_dp, 0.5609035169310178_dp, &
             0.8338918987720528_dp, 1.2408850149454318_dp, 1.8656361544308304_dp, &
             2.903691913137475_dp, 5.105116219860401_dp]

      CASE (814)

         aw(:) = & ! a
            [4.987207666488572e-05_dp, 0.00027205475767331497_dp, 0.0007146345422466652_dp, &
             0.0014722191397521602_dp, 0.0027284598111320772_dp, 0.0048041780793437175_dp, &
             0.00822269297766136_dp, 0.013810729833984306_dp, 0.022856290787743434_dp, &
             0.037347730632978596_dp, 0.060328871577925096_dp, 0.0964214485525208_dp, &
             0.1525885955589112_dp, 0.23924434617093937_dp, 0.3718591684287968_dp, &
             0.5732799635662582_dp, 0.877097767550666_dp, 1.3326219091134706_dp, &
             2.0125526853465634_dp, 3.025941368805157_dp, 4.5439577131822535_dp, &
             6.866137665752377_dp, 10.680072188887511_dp, &
             ! w
             0.00012900282840194976_dp, 0.00032199367417021875_dp, 0.0005783176855982976_dp, &
             0.0009666664699148756_dp, 0.001598248366470722_dp, 0.0026387756507622756_dp, &
             0.004333464445542502_dp, 0.007053372705401387_dp, 0.011363343009572492_dp, &
             0.01811830505874965_dp, 0.028601785950045554_dp, 0.044725747349435414_dp, &
             0.06931796399000324_dp, 0.10653403873953471_dp, 0.16244852673889768_dp, &
             0.24591097260205097_dp, 0.3698201597375351_dp, 0.5531429340627326_dp, &
             0.8245109549522822_dp, 1.2298906861496806_dp, 1.8531461123791817_dp, &
             2.8898126482562834_dp, 5.0892243556771914_dp]

      CASE (815)

         aw(:) = & ! a
            [2.75112874260035e-05_dp, 0.00015132248585800022_dp, 0.0004041877235074668_dp, &
             0.0008549328027316614_dp, 0.0016398749507422741_dp, 0.0030001755058056264_dp, &
             0.005337327169366603_dp, 0.009303281079111313_dp, 0.01594298286835429_dp, &
             0.026913147758594918_dp, 0.04481329749825336_dp, 0.07368219492456587_dp, &
             0.11973662615023326_dp, 0.19246329199547735_dp, 0.30622341037334644_dp, &
             0.4826028639725151_dp, 0.7538600970056132_dp, 1.1680498487977171_dp, &
             1.7969214159595308_dp, 2.7491452473384066_dp, 4.196258205003464_dp, &
             6.438657287329958_dp, 10.163323639802494_dp, &
             ! w
             7.129263674426213e-05_dp, 0.00018093736070606692_dp, 0.0003357867862829417_dp, &
             0.0005878240594227028_dp, 0.0010213725350442346_dp, 0.0017646805505146953_dp, &
             0.0030159602052222314_dp, 0.005086543041365371_dp, 0.008463586378211357_dp, &
             0.013901902468198004_dp, 0.02255946912538166_dp, 0.03619662958374437_dp, &
             0.05746714450315467_dp, 0.09034135764961466_dp, 0.14072034440494527_dp, &
             0.21733190704459512_dp, 0.33306516219511023_dp, 0.5070659345337035_dp, &
             0.7684024059794611_dp, 1.1637330762849774_dp, 1.7776625641173411_dp, &
             2.8057583068033236_dp, 4.993055600329781_dp]

      CASE (816)

         aw(:) = & ! a
            [1.943077545695233e-05_dp, 0.00010749914754601933_dp, 0.0002905375350137754_dp, &
             0.000625845877103975_dp, 0.0012279962234922016_dp, 0.002301395524603954_dp, &
             0.004191270004871865_dp, 0.0074678572675019005_dp, 0.013060313322173217_dp, &
             0.022464241892482667_dp, 0.03805925427102047_dp, 0.06359039011556974_dp, &
             0.1048918559039895_dp, 0.17096678270696955_dp, 0.27558771893430817_dp, &
             0.43965871656865374_dp, 0.6947023252260974_dp, 1.0880619339494948_dp, &
             1.6909249126686423_dp, 2.611705180921732_dp, 4.022107275350649_dp, &
             6.223014736639974_dp, 9.901207941028282_dp, &
             ! w
             5.041676706466184e-05_dp, 0.00012945846815702803_dp, 0.0002457626343830027_dp, &
             0.0004434351520031084_dp, 0.0007941201289321444_dp, 0.0014088488489885562_dp, &
             0.0024637631977260477_dp, 0.004240896723975039_dp, 0.007188115369875548_dp, &
             0.012008502670013273_dp, 0.01979389776733372_dp, 0.032223128583752016_dp, &
             0.051854207104282866_dp, 0.08255253338096369_dp, 0.13011654607334777_dp, &
             0.20319485641760857_dp, 0.31465430013839985_dp, 0.4837201142383853_dp, &
             0.7396859310632805_dp, 1.1295911124194844_dp, 1.7384751475926694_dp, &
             2.7619940863025993_dp, 4.943032210081897_dp]

      CASE (817)

         aw(:) = & ! a
            [1.5186028788487787e-05_dp, 8.440689165581798e-05_dp, 0.00023028964370770395_dp, &
             0.0005032170072102935_dp, 0.0010045305117883968_dp, 0.0019162708437991153_dp, &
             0.0035493344020672743_dp, 0.006423614760193411_dp, 0.011396058560892755_dp, &
             0.019860601987159356_dp, 0.03405654040598402_dp, 0.057539526657774384_dp, &
             0.09589457766863309_dp, 0.1578066878477717_dp, 0.25665716234505115_dp, &
             0.41289199266412385_dp, 0.6575333669388179_dp, 1.0374337878643975_dp, &
             1.6233840757330766_dp, 2.523605279932863_dp, 3.9099015937720365_dp, &
             6.083490615602792_dp, 9.731065378730635_dp, &
             ! w
             3.944278813817094e-05_dp, 0.00010222950025951529_dp, 0.00019757345912072447_dp, &
             0.00036470563524428187_dp, 0.0006674907691885711_dp, 0.0012065052478867878_dp, &
             0.0021441921604794805_dp, 0.0037439458410990214_dp, 0.006428251061381941_dp, &
             0.010866386448031513_dp, 0.018106484768346254_dp, 0.029772871054980683_dp, &
             0.0483586944734761_dp, 0.07765711952438033_dp, 0.12339430747578661_dp, &
             0.19416052189741978_dp, 0.302800925064658_dp, 0.4685871070120731_dp, &
             0.7209596705006001_dp, 1.1072161383206613_dp, 1.7127015405508177_dp, &
             2.733159034676653_dp, 4.910090392492674_dp]

      CASE (818)

         aw(:) = & ! a
            [1.254573522126483e-05_dp, 7.000771758757615e-05_dp, 0.00019254082708660645_dp, &
             0.00042578076534042055_dp, 0.0008619147234744453_dp, 0.0016675160825434612_dp, &
             0.0031296665520386957_dp, 0.005733059254379714_dp, 0.010283700465173956_dp, &
             0.018103197704366195_dp, 0.03133023713881733_dp, 0.05338366921363135_dp, &
             0.08966722278247792_dp, 0.14863281371340806_dp, 0.24337306308632814_dp, &
             0.39399344154356175_dp, 0.6311410700612039_dp, 1.001296744089423_dp, &
             1.574946468154319_dp, 2.46015699578551_dp, 3.8287995465905684_dp, &
             5.982343497975724_dp, 9.607440868469775_dp, &
             ! w
             3.2612992923360705e-05_dp, 8.519990467040972e-05_dp, 0.00016714535926317734_dp, &
             0.00031426486684134715_dp, 0.000585025383471879_dp, 0.0010727738947416787_dp, &
             0.0019303032966276204_dp, 0.003407663506622901_dp, 0.005908988748277583_dp, &
             0.01007892147011934_dp, 0.016933492678545287_dp, 0.02805666761951195_dp, &
             0.04589312978699421_dp, 0.07418148319009715_dp, 0.1185924698841046_dp, &
             0.18767032784355986_dp, 0.29424061083351266_dp, 0.4576056637352122_dp, &
             0.7073128951841113_dp, 1.0908528291173565_dp, 1.6938045862241802_dp, &
             2.711990371906971_dp, 4.885915259856555_dp]

      CASE (819)

         aw(:) = & ! a
            [1.0734659652527657e-05_dp, 6.011026071221395e-05_dp, 0.00016648760829417894_dp, &
             0.0003719813375304187_dp, 0.0007619522845491065_dp, 0.0014914463707654522_dp, &
             0.0028297398206932776_dp, 0.00523503399267206_dp, 0.009474728436500116_dp, &
             0.016815255869661296_dp, 0.029318089726597784_dp, 0.050296492285577254_dp, &
             0.08501350172895192_dp, 0.14173917660383176_dp, 0.2333396709232522_dp, &
             0.37965180331063897_dp, 0.6110249258532348_dp, 0.9736424828994826_dp, &
             1.5377439138027684_dp, 2.4112675709302316_dp, 3.7661331760527372_dp, &
             5.90401068753362_dp, 9.511533838288086_dp, &
             ! w
             2.7925976875382294e-05_dp, 7.346483205748955e-05_dp, 0.00014600688389127689_dp, &
             0.0002787958296714323_dp, 0.0005262713017950987_dp, 0.0009763858739287935_dp, &
             0.0017746191441236445_dp, 0.003160790314520077_dp, 0.005524864054564826_dp, &
             0.009492344431487299_dp, 0.016054175025394548_dp, 0.026762583363885387_dp, &
             0.04402388122227869_dp, 0.07153312618965908_dp, 0.11491633833226951_dp, &
             0.18267985972736092_dp, 0.28763166513636484_dp, 0.4490961062652346_dp, &
             0.6967033324407155_dp, 1.07809674232456_dp, 1.6790443155713535_dp, &
             2.6954392088766355_dp, 4.867018268732641_dp]

      CASE (820)

         aw(:) = & ! a
            [9.410117883516404e-06_dp, 5.2858585086734156e-05_dp, 0.00014733052988380674_dp, &
             0.0003321923374009351_dp, 0.0006874574809019757_dp, 0.0013591459030404722_dp, &
             0.002602548429823122_dp, 0.0048549414563873365_dp, 0.008853055000827214_dp, &
             0.01581925797475768_dp, 0.02775305790155189_dp, 0.04788258820815112_dp, &
             0.08135695572382737_dp, 0.13629831261354758_dp, 0.22538782142458622_dp, &
             0.3682418508356122_dp, 0.5949641847990454_dp, 0.9514915617504976_dp, &
             1.5078570567158924_dp, 2.3718892054824887_dp, 3.715544574069324_dp, &
             5.840658930949642_dp, 9.433860134519659_dp, &
             ! w
             2.4496693437567276e-05_dp, 6.484786415367783e-05_dp, 0.00013037460720919538_dp, &
             0.00025229015677215826_dp, 0.0004818808577720569_dp, 0.0009028670546788508_dp, &
             0.0016549153246979227_dp, 0.002969640998663022_dp, 0.005225582928230421_dp, &
             0.00903273748103435_dp, 0.015361621868829443_dp, 0.025738493324956084_dp, &
             0.04253809442491115_dp, 0.0694194275313286_dp, 0.11197116436683657_dp, &
             0.1786675112597719_dp, 0.2823006252472037_dp, 0.4422114131620818_dp, &
             0.6880968881446058_dp, 1.0677262708282482_dp, 1.6670252661896332_dp, &
             2.6819509150656184_dp, 4.85162133575408_dp]

      CASE (821)

         aw(:) = & ! a
            [8.396449618539869e-06_dp, 4.7299942977187724e-05_dp, 0.00013259888830115856_dp, &
             0.000301435909697078_dp, 0.0006294867032171851_dp, 0.001255451484100915_dp, &
             0.002423245339842898_dp, 0.004553041242587073_dp, 0.008356379161672634_dp, &
             0.015019275824332221_dp, 0.026489907775937602_dp, 0.0459256179079216_dp, &
             0.07838042132873657_dp, 0.13185257585053184_dp, 0.2188677236000601_dp, &
             0.35885620376869615_dp, 0.5817137152407771_dp, 0.9331669069409293_dp, &
             1.4830718592406307_dp, 2.3391612781264732_dp, 3.673420704474708_dp, &
             5.787826645703125_dp, 9.369008329913077_dp, &
             ! w
             2.187131530456291e-05_dp, 5.822974375369402e-05_dp, 0.00011829205553857099_dp, &
             0.0002316136862413449_dp, 0.00044692545301078815_dp, 0.0008445070703273048_dp, &
             0.00155924662976456_dp, 0.002815968261578136_dp, 0.004983709465311969_dp, &
             0.008659518493222097_dp, 0.014796788751371505_dp, 0.024899915561152387_dp, &
             0.04131694727645881_dp, 0.06767623938641835_dp, 0.10953449527159823_dp, &
             0.17533807676757446_dp, 0.2778648136563897_dp, 0.4364685190907497_dp, &
             0.6809018898691851_dp, 1.0590405721773954_dp, 1.6569453010953754_dp, &
             2.670631005092024_dp, 4.838701714305561_dp]

      CASE (822)

         aw(:) = & ! a
            [7.594003119761652e-06_dp, 4.289314581392432e-05_dp, 0.00012088578739431088_dp, &
             0.00027686621999750867_dp, 0.0005828982286734941_dp, 0.0011715880733756995_dp, &
             0.0022773530052228688_dp, 0.00430602466958367_dp, 0.007947932548778496_dp, &
             0.014358367188053034_dp, 0.025441965659642483_dp, 0.04429582728569558_dp, &
             0.0758927874432561_dp, 0.1281250091308677_dp, 0.21338454524641007_dp, &
             0.35094142560648356_dp, 0.5705114026547935_dp, 0.9176386810359661_dp, &
             1.462024711633191_dp, 2.311317276654545_dp, 3.637525303622565_dp, &
             5.7427471606945195_dp, 9.31361786078264_dp, &
             ! w
             1.9792311519439543e-05_dp, 5.2973704754217626e-05_dp, 0.00010864071545772197_dp, &
             0.00021496087882984724_dp, 0.00041853955279144467_dp, 0.0007967841134313631_dp, &
             0.001480555178191933_dp, 0.002688920157372317_dp, 0.004782831517145299_dp, &
             0.008348281205274673_dp, 0.014323990350402018_dp, 0.024195555932691012_dp, &
             0.0402879854528591_dp, 0.06620306483034866_dp, 0.10746961965759644_dp, &
             0.1725094849375673_dp, 0.2740874235246064_dp, 0.43156758317826366_dp, &
             0.6747500915886272_dp, 1.051602479017448_dp, 1.6483032621541625_dp, &
             2.6609201823709774_dp, 4.827620073809154_dp]

      CASE (823)

         aw(:) = & ! a
            [6.941897803062561e-06_dp, 3.930718144472374e-05_dp, 0.00011132880383520508_dp, &
             0.0002567325037691151_dp, 0.0005445129662364841_dp, 0.0011020985023357935_dp, &
             0.0021558143141816766_dp, 0.004099226505195036_dp, 0.007604457397456052_dp, &
             0.013800334018344265_dp, 0.024553878934543985_dp, 0.04290999587795706_dp, &
             0.07377099149949082_dp, 0.12493660635158689_dp, 0.20868220713701893_dp, &
             0.34413740821626987_dp, 0.5608598663503145_dp, 0.9042329141327881_dp, &
             1.443820952904812_dp, 2.287195584921532_dp, 3.606385003553035_dp, &
             5.703594743292729_dp, 9.265468359287246_dp, &
             ! w
             1.8102306402496735e-05_dp, 4.868973008587396e-05_dp, 0.00010073248747868423_dp, &
             0.00020121333400762914_dp, 0.00039493378905936774_dp, 0.0007568535682459059_dp, &
             0.00141437258556732_dp, 0.0025815877463692726_dp, 0.004612447683791073_dp, &
             0.008083337543304418_dp, 0.013920191319419782_dp, 0.023592172436583077_dp, &
             0.039404079626994624_dp, 0.06493430660962199_dp, 0.10568701217030595_dp, &
             0.1700621442417167_dp, 0.27081247100109584_dp, 0.4273105898193215_dp, &
             0.6693977463190739_dp, 1.0451220792581715_dp, 1.6407663510306043_dp, &
             2.652446791729949_dp, 4.817951665848005_dp]

      CASE (824)

         aw(:) = & ! a
            [3.853270779220833e-06_dp, 2.2241728520864382e-05_dp, 6.540535214803702e-05_dp, &
             0.00015848095993119255_dp, 0.0003536244659221865_dp, 0.0007498239134427326_dp, &
             0.0015284896933536677_dp, 0.0030142094326381257_dp, 0.00577545291759329_dp, &
             0.010788712861079184_dp, 0.019702267183033417_dp, 0.03525456088102581_dp, &
             0.06193000221398213_dp, 0.10697581653408518_dp, 0.18196357847991113_dp, &
             0.3051681669273905_dp, 0.5051751930610341_dp, 0.8263670489211143_dp, &
             1.337441035048714_dp, 2.145468106363411_dp, 3.4225682931292067_dp, &
             5.471607451639996_dp, 8.979350067431634_dp, &
             ! w
             1.008923347743312e-05_dp, 2.8184249820085405e-05_dp, 6.215059000915936e-05_dp, &
             0.00013238189012683877_dp, 0.0002738301923412137_dp, 0.0005478253821267155_dp, &
             0.0010619800865001483_dp, 0.0020015524541288367_dp, 0.003679414729906491_dp, &
             0.006615037197040588_dp, 0.011657836051224385_dp, 0.020177631353302125_dp, &
             0.03435572824472115_dp, 0.05762579147969522_dp, 0.09533688650333176_dp, &
             0.15574761784426483_dp, 0.2515266054875526_dp, 0.4020855285971333_dp, &
             0.6375069898886988_dp, 1.0063320893911387_dp, 1.5954999890268287_dp, &
             2.6014672997866284_dp, 4.759803740089366_dp]

      CASE (825)

         aw(:) = & ! a
            [2.7362333478713395e-06_dp, 1.6018569703292616e-05_dp, 4.8370196970770565e-05_dp, &
             0.00012105776534796837_dp, 0.0002786559313463596_dp, 0.0006072901874353698_dp, &
             0.0012677047096366792_dp, 0.002552144881886733_dp, 0.0049796485001332436_dp, &
             0.009452945428134405_dp, 0.017512908651776883_dp, 0.03174554692684556_dp, &
             0.05642483340555305_dp, 0.0985164532120049_dp, 0.1692291202896758_dp, &
             0.28639169186936664_dp, 0.47807591103041425_dp, 0.788126912084251_dp, &
             1.2847667026913803_dp, 2.0747790909381907_dp, 3.330313861838757_dp, &
             5.354586342504829_dp, 8.834470380703861_dp, &
             ! w
             7.1858684275017036e-06_dp, 2.0630979485376434e-05_dp, 4.745986258647168e-05_dp, &
             0.00010505591731147411_dp, 0.00022396334666135973_dp, 0.0004591816028414762_dp, &
             0.0009088204482574491_dp, 0.0017440343322827272_dp, 0.003257314375616538_dp, &
             0.005939501541007813_dp, 0.010601002767896068_dp, 0.018560306569222466_dp, &
             0.03193400287429269_dp, 0.0540787024973376_dp, 0.09025922312502486_dp, &
             0.14865489391753492_dp, 0.2418827597820275_dp, 0.3893662735237943_dp, &
             0.621307584280122_dp, 0.9865063325718852_dp, 1.5722591866906825_dp, &
             2.5752315823883256_dp, 4.729892490545373_dp]

      CASE (826)

         aw(:) = & ! a
            [2.1485725653763096e-06_dp, 1.2725610767311947e-05_dp, 3.9245142611661685e-05_dp, &
             0.00010064280220769361_dp, 0.0002369377809828602_dp, 0.0005264866822571787_dp, &
             0.0011174066214606883_dp, 0.002281957430450397_dp, 0.004508325425233879_dp, &
             0.00865279241950343_dp, 0.016188048305457656_dp, 0.029602600787565818_dp, &
             0.05303486610867192_dp, 0.09326792686007114_dp, 0.16127356372682886_dp, &
             0.2745873962730799_dp, 0.46094083148020853_dp, 0.7638201245999824_dp, &
             1.251126090041441_dp, 2.0294437662643596_dp, 3.270935367983161_dp, &
             5.2790476004623095_dp, 8.740743253412464_dp, &
             ! w
             5.65649586565214e-06_dp, 1.6605705580539614e-05_dp, 3.9446425858471386e-05_dp, &
             8.974499020101461e-05_dp, 0.0001954003495832924_dp, 0.00040751067567562703_dp, &
             0.0008182291847597183_dp, 0.0015897833190846396_dp, 0.0030016549532205315_dp, &
             0.005526264207570665_dp, 0.009948722859607129_dp, 0.01755397631792136_dp, &
             0.03041598678412126_dp, 0.05184015834737924_dp, 0.08703470054229626_dp, &
             0.14412476230304141_dp, 0.23569058190270062_dp, 0.3811600453758225_dp, &
             0.6108113783709254_dp, 0.9736145851829946_dp, 1.5571070220101144_dp, &
             2.5581032573117777_dp, 4.710369279446002_dp]

      CASE (827)

         aw(:) = & ! a
            [1.7824706410759602e-06_dp, 1.0664679395860483e-05_dp, 3.3477483043565994e-05_dp, &
             8.755432504560044e-05_dp, 0.0002097898001019664_dp, 0.0004731857022262164_dp, &
             0.0010170805983356738_dp, 0.0020997284782958876_dp, 0.004187543363553217_dp, &
             0.008103823087121502_dp, 0.015272563497365_dp, 0.028112274082148475_dp, &
             0.05066357019125376_dp, 0.08957716662770761_dp, 0.15565229251041415_dp, &
             0.266210000526289_dp, 0.4487314096726897_dp, 0.7464373080973531_dp, &
             1.2269890470068183_dp, 1.9968210991849522_dp, 3.2281010731629847_dp, &
             5.224445722641455_dp, 8.672891426699495_dp, &
             ! w
             4.702774535850405e-06_dp, 1.4072069665725788e-05_dp, 3.4308600212246764e-05_dp, &
             7.973113777740445e-05_dp, 0.00017642252796476598_dp, 0.0003727491433089244_dp, &
             0.0007566482109370002_dp, 0.0014839867603136827_dp, 0.0028249227856203076_dp, &
             0.005238599767886166_dp, 0.009491799679652798_dp, 0.01684503147251877_dp, &
             0.029341031589271156_dp, 0.05024746706993103_dp, 0.0847305187997003_dp, &
             0.1408746551083355_dp, 0.23123171364003703_dp, 0.3752311332166879_dp, &
             0.6032054963076499_dp, 0.964249624854869_dp, 1.5460798447272006_dp, &
             2.5456258671900445_dp, 4.696149561069354_dp]

      CASE (828)

         aw(:) = & ! a
            [1.530969024572701e-06_dp, 9.243367038141875e-06_dp, 2.946628948285831e-05_dp, &
             7.834419463255329e-05_dp, 0.00019045677030714647_dp, 0.00043482081400868667_dp, &
             0.0009441973479883247_dp, 0.001966281526467527_dp, 0.003950985645108461_dp, &
             0.007696489018375142_dp, 0.014589544412845094_dp, 0.026994909609096007_dp, &
             0.04887781674526017_dp, 0.08678658792989523_dp, 0.1513865061347294_dp, &
             0.25983146914073235_dp, 0.43940687341477486_dp, 0.7331250077245878_dp, &
             1.2084580462003072_dp, 1.9717201202961518_dp, 3.1950808210583053_dp, &
             5.182289784995443_dp, 8.620445690669941_dp, &
             ! w
             4.047053285540483e-06_dp, 1.2316301360213714e-05_dp, 3.069280520435761e-05_dp, &
             7.257162203634314e-05_dp, 0.00016268737647273898_dp, 0.0003473471893284819_dp, &
             0.0007112859777609981_dp, 0.001405515556370155_dp, 0.002693044360165157_dp, &
             0.005022791643608978_dp, 0.009147365586094932_dp, 0.01630830300015645_dp, &
             0.028523997119572805_dp, 0.04903256228390705_dp, 0.08296708359282406_dp, &
             0.138379730474653_dp, 0.22779935023168185_dp, 0.37065558922133773_dp, &
             0.5973225901805147_dp, 0.9569924889245729_dp, 1.53752271459837_dp, &
             2.535936240703898_dp, 4.685108173592886_dp]

      CASE (829)

         aw(:) = & ! a
            [1.3467681462972046e-06_dp, 8.198860357499511e-06_dp, 2.6496696210333228e-05_dp, &
             7.145685836364103e-05_dp, 0.00017585481273804296_dp, 0.00040558881091113774_dp, &
             0.0008882435632021239_dp, 0.001863163506477378_dp, 0.0037671538655925565_dp, &
             0.007378366175564499_dp, 0.014053754977722729_dp, 0.026114933731968496_dp, &
             0.0474664461277602_dp, 0.08457394532975825_dp, 0.1479942673747438_dp, &
             0.25474559325113605_dp, 0.4319539299558102_dp, 0.7224611849900392_dp, &
             1.193584228439049_dp, 1.9515374993058465_dp, 3.168490743879836_dp, &
             5.1483017379991205_dp, 8.578122934982396_dp, &
             ! w
             3.5664597412040825e-06_dp, 1.102053066184845e-05_dp, 2.798851972323109e-05_dp, &
             6.71465542046445e-05_dp, 0.0001521757617941444_dp, 0.00032775436577814956_dp, &
             0.0006760695251468474_dp, 0.0013442548368026391_dp, 0.0025895869655016717_dp, &
             0.004852760329243902_dp, 0.008874943793593926_dp, 0.015882313076166738_dp, &
             0.02787348652034074_dp, 0.048062485834429604_dp, 0.08155529805703034_dp, &
             0.1363774769490829_dp, 0.22503864306235727_dp, 0.3669679588738693_dp, &
             0.5925727957129064_dp, 0.9511243439282931_dp, 1.530595683200551_dp, &
             2.528087817945579_dp, 4.676165651686418_dp]

      CASE (830)

         aw(:) = & ! a
            [1.2056086566165549e-06_dp, 7.395998417494144e-06_dp, 2.4198989398328045e-05_dp, &
             6.608078408161546e-05_dp, 0.00016435917474465897_dp, 0.00038240350788677925_dp, &
             0.0008435809795176921_dp, 0.0017804039481713273_dp, 0.0036189157814293057_dp, &
             0.007120771396602508_dp, 0.013618310096621584_dp, 0.025397410685136106_dp, &
             0.046312226172941895_dp, 0.08275960583307228_dp, 0.14520591872701982_dp, &
             0.25055586613040787_dp, 0.4258018517685003_dp, 0.7136425612531333_dp, &
             1.181263825065014_dp, 1.9347953472822745_dp, 3.146405991156378_dp, &
             5.120044118051449_dp, 8.542909338548004_dp, &
             ! w
             3.197935511315449e-06_dp, 1.0020745253548173e-05_dp, 2.5877276250159568e-05_dp, &
             6.286385949393449e-05_dp, 0.000143808063283762_dp, 0.00031205503907536533_dp, &
             0.0006476972812668882_dp, 0.0012946693449989635_dp, 0.0025055055840059446_dp, &
             0.004714075943443775_dp, 0.008652031823475624_dp, 0.015532734519292617_dp, &
             0.027338261303341025_dp, 0.04726242212384382_dp, 0.08038839026014187_dp, &
             0.13471919527595133_dp, 0.2227479941113065_dp, 0.3639030921931474_dp, &
             0.5886192861002347_dp, 0.9462338945128798_dp, 1.5248174476065022_dp, &
             2.5215377926357765_dp, 4.6687030721757825_dp]

      CASE (831)

         aw(:) = & ! a
            [1.0937200431281287e-06_dp, 6.757875439333789e-06_dp, 2.2361754485768456e-05_dp, &
             6.174836519347853e-05_dp, 0.00015502577053533255_dp, 0.00036345744186304054_dp, &
             0.0008068841819623697_dp, 0.0017120856652685565_dp, 0.0034960475627543635_dp, &
             0.006906502806685575_dp, 0.01325496615293238_dp, 0.02479701642967891_dp, &
             0.045343989472682325_dp, 0.08123415438054458_dp, 0.142856705141712_dp, &
             0.24701934504276746_dp, 0.4206000401836095_dp, 0.7061745071020519_dp, &
             1.1708157118380706_dp, 1.92057992926266_dp, 3.1276345272017316_dp, &
             5.096005408320577_dp, 8.51293407219721_dp, &
             ! w
             2.905667024865261e-06_dp, 9.223353906958435e-06_dp, 2.4175588191019807e-05_dp, &
             5.937848615719115e-05_dp, 0.0001369491567359643_dp, 0.00029911380166042293_dp, &
             0.0006242000370141447_dp, 0.0012534394577407703_dp, 0.0024353491334064537_dp, &
             0.004598003893336091_dp, 0.008464954437487003_dp, 0.01523863056475835_dp, &
             0.02688696693034401_dp, 0.046586449656555486_dp, 0.07940063950628212_dp, &
             0.13331311810140323_dp, 0.22080268949288895_dp, 0.3612965987525656_dp, &
             0.5852528227111651_dp, 0.9420652196573274_dp, 1.5198881571634328_dp, &
             2.5159477760293925_dp, 4.6623346330091735_dp]

      CASE (832)

         aw(:) = & ! a
            [1.0026839856435724e-06_dp, 6.237374911349773e-06_dp, 2.085488117019199e-05_dp, &
             5.8169860962251726e-05_dp, 0.00014726530103885307_dp, 0.00034761468217947886_dp, &
             0.0007760506242668873_dp, 0.0016544475048729074_dp, 0.0033920199690078592_dp, &
             0.006724528433187649_dp, 0.012945540976798798_dp, 0.024284472752256092_dp, &
             0.04451562010911718_dp, 0.07992648877663189_dp, 0.1408392826829582_dp, &
             0.2439773649648799_dp, 0.41611901860654604_dp, 0.6997326251964361_dp, &
             1.161792368043248_dp, 1.9082899184919446_dp, 3.111390784105834_dp, &
             5.075188373470443_dp, 8.486961793889565_dp, &
             ! w
             2.6677497664863237e-06_dp, 8.570867779929502e-06_dp, 2.27697678977458e-05_dp, &
             5.647449772745657e-05_dp, 0.0001311983698388498_dp, 0.00028820979562128906_dp, &
             0.0006043208498015859_dp, 0.001218436354839996_dp, 0.0023756073959409715_dp, &
             0.004498898436014983_dp, 0.008304842477225361_dp, 0.014986380937835705_dp, &
             0.026499148704522437_dp, 0.0460045332196876_dp, 0.07854895830488497_dp, &
             0.13209894950407378_dp, 0.21912062008531333_dp, 0.35904004517695376_dp, &
             0.582335157967353_dp, 0.9384489894823416_dp, 1.5156092124431433_dp, &
             2.511093527097568_dp, 4.656804703604224_dp]

      CASE (833)

         aw(:) = & ! a
            [5.695482478296897e-07_dp, 3.7387017738497234e-06_dp, 1.347689628824476e-05_dp, &
             4.0221708631463127e-05_dp, 0.00010747729778809903_dp, 0.00026486531922520533_dp, &
             0.0006124658291870136_dp, 0.0013445619054028325_dp, 0.0028262791776858076_dp, &
             0.0057249298443552985_dp, 0.011230766633054922_dp, 0.021421627311030516_dp, &
             0.03985594956400731_dp, 0.07252373374554023_dp, 0.12935242860677965_dp, &
             0.2265656641001338_dp, 0.39034764448334164_dp, 0.6625228293427873_dp, &
             1.1094676161248844_dp, 1.8367762213694017_dp, 3.0165921840800327_dp, &
             4.953410355808327_dp, 8.33475527994059_dp, &
             ! w
             1.5338807594779788e-06_dp, 5.4023566807865e-06_dp, 1.5713159764209495e-05_dp, &
             4.1486613858198684e-05_dp, 0.00010090782806823178_dp, 0.00022985012460898297_dp, &
             0.0004965042419198246_dp, 0.0010264330911921925_dp, 0.0020446666858529397_dp, &
             0.003945128002854194_dp, 0.0074032682164944_dp, 0.013556132913603127_dp, &
             0.024286436303487514_dp, 0.042665437452962626_dp, 0.07363642754530832_dp, &
             0.1250620781292975_dp, 0.20932925335779837_dp, 0.3458523034842012_dp, &
             0.5652235213079986_dp, 0.9171774324449022_dp, 1.4903838853512916_dp, &
             2.482442933394029_dp, 4.624171367544714_dp]

      CASE (834)

         aw(:) = & ! a
            [4.1164904819495056e-07_dp, 2.813586205810931e-06_dp, 1.0651434826377284e-05_dp, &
             3.3082761700612936e-05_dp, 9.112484119257757e-05_dp, 0.0002299316740026587_dp, &
             0.0005418624311686956_dp, 0.001208311396446073_dp, 0.002573568614363697_dp, &
             0.0052722665767247_dp, 0.010444876531852484_dp, 0.020095585931105338_dp, &
             0.037677130745476024_dp, 0.06903275657764923_dp, 0.12389379924342069_dp, &
             0.21823385468524384_dp, 0.3779376665647566_dp, 0.644502363140509_dp, &
             1.083997113679983_dp, 1.801807658345806_dp, 2.9700592959374554_dp, &
             4.893448560821683_dp, 8.259637315234952_dp, &
             ! w
             1.1194489831490663e-06_dp, 4.205376538166022e-06_dp, 1.2901940067937311e-05_dp, &
             3.526918985534942e-05_dp, 8.797446712170179e-05_dp, 0.00020436565991084073_dp, &
             0.00044854676342248823_dp, 0.000939688100172495_dp, 0.0018931321660813649_dp, &
             0.0036885731095298766_dp, 0.006981232024143092_dp, 0.012880402988628177_dp, &
             0.023232297578308286_dp, 0.0410626624397191_dp, 0.07126217027911405_dp, &
             0.12163973033975671_dp, 0.20453992579830765_dp, 0.339368109347748_dp, &
             0.5567712192770106_dp, 0.9066296495958464_dp, 1.4778395493609238_dp, &
             2.468173248830966_dp, 4.607921241408812_dp]

      CASE (835)

         aw(:) = & ! a
            [3.2815106747013573e-07_dp, 2.3190388059084265e-06_dp, 9.105727958600882e-06_dp, &
             2.9083697560823515e-05_dp, 8.17846074303215e-05_dp, 0.00020966506586372546_dp, &
             0.0005003797625227516_dp, 0.0011274098319519973_dp, 0.002422169665104748_dp, &
             0.004998980872010807_dp, 0.009967215710163558_dp, 0.019284836586730216_dp, &
             0.036337952140351105_dp, 0.06687692803473942_dp, 0.12050849505479717_dp, &
             0.21304676333872524_dp, 0.37018468401267685_dp, 0.6332087786116111_dp, &
             1.0679893680205574_dp, 1.7797757896739255_dp, 2.940679102110993_dp, &
             4.855524732089155_dp, 8.212067192145357_dp, &
             ! w
             8.999600465203641e-07_dp, 3.556322522670299e-06_dp, 1.1325322843811366e-05_dp, &
             3.1699581990752516e-05_dp, 8.042516831768671e-05_dp, 0.00018929780097494032_dp, &
             0.00041989246827121634_dp, 0.0008873994705141611_dp, 0.0018010964747775297_dp, &
             0.003531724472102502_dp, 0.00672171495324129_dp, 0.012462738070578742_dp, &
             0.022577719603031936_dp, 0.04006323357801831_dp, 0.0697760365043427_dp, &
             0.11949011551442518_dp, 0.2015221588442651_dp, 0.33527066815854195_dp, &
             0.5514165114504993_dp, 0.8999331123660025_dp, 1.469862746831253_dp, &
             2.459091523474792_dp, 4.5975801498593105_dp]

      CASE (836)

         aw(:) = & ! a
            [2.7594881331818964e-07_dp, 2.0071743154717217e-06_dp, 8.11348505740982e-06_dp, &
             2.6472042546690133e-05_dp, 7.560063764703147e-05_dp, 0.00019610109104827052_dp, &
             0.0004723729023173509_dp, 0.0010723935089931224_dp, 0.00231858238510834_dp, &
             0.004811016795118102_dp, 0.009637183157578435_dp, 0.018722410693781336_dp, &
             0.0354056347865589_dp, 0.0653712807509657_dp, 0.11813738322454047_dp, &
             0.20940423059591254_dp, 0.3647275214801527_dp, 0.6252425984794626_dp, &
             1.0566764814931602_dp, 1.7641795077178264_dp, 2.9198513093858764_dp, &
             4.8286094086064715_dp, 8.178276854090898_dp, &
             ! w
             7.625962204237235e-07_dp, 3.1423807227117575e-06_dp, 1.0294757811612861e-05_dp, &
             2.9328060814257294e-05_dp, 7.535206068659267e-05_dp, 0.0001790822453176794_dp, &
             0.0004003253885991469_dp, 0.0008514774801044953_dp, 0.001737542145001489_dp, &
             0.0034229293515005083_dp, 0.006540997996161008_dp, 0.01217087807441986_dp, &
             0.022118876234280778_dp, 0.03936067958064679_dp, 0.0687286718990807_dp, &
             0.117971616113168_dp, 0.19938584721339075_dp, 0.3323644416301962_dp, &
             0.5476120341444782_dp, 0.8951684302967229_dp, 1.4641810458591635_dp, &
             2.452619061954174_dp, 4.5902106376365115_dp]

      CASE (837)

         aw(:) = & ! a
            [2.399987656078681e-07_dp, 1.7908417819181487e-06_dp, 7.415121447458101e-06_dp, &
             2.4609109256061034e-05_dp, 7.114315807040905e-05_dp, 0.00018624383111229658_dp, &
             0.0004518857442932671_dp, 0.0010319306769572963_dp, 0.002242049578594529_dp, &
             0.004671601900334482_dp, 0.009391564194309055_dp, 0.01830259110610384_dp, &
             0.03470787276365137_dp, 0.06424176777919703_dp, 0.11635483590568536_dp, &
             0.20666061005569386_dp, 0.36060996507206033_dp, 0.6192225246328895_dp, &
             1.0481153036894775_dp, 1.7523622480359342_dp, 2.9040535629398865_dp, &
             4.808177018157149_dp, 8.152609250158763_dp, &
             ! w
             6.679279540216612e-07_dp, 2.8525235443726298e-06_dp, 9.559109241978822e-06_dp, &
             2.7614312929129997e-05_dp, 7.165438530048072e-05_dp, 0.000171586701895398_dp, &
             0.0003858906672767459_dp, 0.0008248581438845777_dp, 0.0016902654389970753_dp, &
             0.0033417297848367285_dp, 0.006405725820790926_dp, 0.011951846762332956_dp, &
             0.02177373205219112_dp, 0.03883111160830619_dp, 0.0679376986512245_dp, &
             0.1168228617332441_dp, 0.19776717320551254_dp, 0.33015927239239035_dp, &
             0.5447216544955404_dp, 0.8915447113133756_dp, 1.4598564730371169_dp, &
             2.447690492480869_dp, 4.584599254453571_dp]

      CASE (838)

         aw(:) = & ! a
            [2.1362409300071513e-07_dp, 1.6311281636796907e-06_dp, 6.893176859606963e-06_dp, &
             2.320153079525968e-05_dp, 6.774691203990355e-05_dp, 0.00017868447935486168_dp, &
             0.00043609280506814156_dp, 0.0010006058866571556_dp, 0.0021825883100299727_dp, &
             0.004562953020253901_dp, 0.009199639171096712_dp, 0.01797378094055829_dp, &
             0.03416024251330929_dp, 0.06335364565520049_dp, 0.11495091609566763_dp, &
             0.20449652369965488_dp, 0.3573577666309494_dp, 0.6144618441497989_dp, &
             1.0413377140747122_dp, 1.7429979275258771_dp, 2.8915247368561734_dp, &
             4.791961865329579_dp, 8.132229446127386_dp, &
             ! w
             5.984358998815298e-07_dp, 2.6367879582450793e-06_dp, 9.002944281445811e-06_dp, &
             2.6306011923403513e-05_dp, 6.881220908810481e-05_dp, 0.00016579495326258533_dp, &
             0.00037468954278266624_dp, 0.0008041286795896982_dp, 0.001653338234998219_dp, &
             0.0032781404014528208_dp, 0.0062995485176757515_dp, 0.011779577101252508_dp, &
             0.021501782066665597_dp, 0.03841316784413099_dp, 0.06731252682000821_dp, &
             0.11591368382941206_dp, 0.1964845080528559_dp, 0.3284099185074698_dp, &
             0.5424264689547013_dp, 0.8886648138160013_dp, 1.456417454638106_dp, &
             2.443769844988774_dp, 4.580135594434126_dp]

      CASE (839)

         aw(:) = & ! a
            [1.9338981142090421e-07_dp, 1.5079122297723746e-06_dp, 6.4862253216743955e-06_dp, &
             2.2094005371900003e-05_dp, 6.50560938942071e-05_dp, 0.00017266325261119984_dp, &
             0.00042345976653957937_dp, 0.000975461245820272_dp, 0.0021347186892812634_dp, &
             0.004475266206955318_dp, 0.009044407959691296_dp, 0.01770733114255511_dp, &
             0.03371572799835314_dp, 0.06263167233628789_dp, 0.11380810513473674_dp, &
             0.2027327896200115_dp, 0.35470431413994125_dp, 0.6105737876909518_dp, &
             1.0357975429291284_dp, 1.7353373332630773_dp, 2.881268581990868_dp, &
             4.778680975529953_dp, 8.115530917313851_dp, &
             ! w
             5.450992049821696e-07_dp, 2.4691653230740935e-06_dp, 8.565118732811697e-06_dp, &
             2.5267819351171584e-05_dp, 6.654417104791554e-05_dp, 0.00016115323332037033_dp, &
             0.00036568127555640373_dp, 0.0007874091578990533_dp, 0.0016234810549921952_dp, &
             0.003226616624170021_dp, 0.006213357796533992_dp, 0.011639505232965874_dp, &
             0.02128033574113745_dp, 0.0380723900413563_dp, 0.0668021708239874_dp, &
             0.11517067089278998_dp, 0.19543522625129223_dp, 0.3269775763556101_dp, &
             0.5405457115997144_dp, 0.8863033397608341_dp, 1.4535960963058718_dp, &
             2.440552486753776_dp, 4.576472737007622_dp]

      CASE (840)

         aw(:) = & ! a
            [1.77340963615581e-07_dp, 1.4096900680093299e-06_dp, 6.15879625453549e-06_dp, &
             2.119591756433595e-05_dp, 6.286130165443841e-05_dp, 0.00016772986189202482_dp, &
             0.00041307208959420803_dp, 0.0009547252951730203_dp, 0.0020951456015534846_dp, &
             0.004402625614127159_dp, 0.008915580756347925_dp, 0.017485853302600614_dp, &
             0.03334572257314234_dp, 0.06202996555337387_dp, 0.11285459726972447_dp, &
             0.20125972689810276_dp, 0.352486141506121_dp, 0.6073208654612043_dp, &
             1.0311589741038818_dp, 1.7289192611880289_dp, 2.872671201433578_dp, &
             4.767543120650199_dp, 8.101522290281896_dp, &
             ! w
             5.027801262076685e-07_dp, 2.334695847127648e-06_dp, 8.209937050114295e-06_dp, &
             2.4419902121800118e-05_dp, 6.46830634447534e-05_dp, 0.00015733051136388254_dp, &
             0.00035824078237663506_dp, 0.0007735659945399227_dp, 0.0015987096026518109_dp, &
             0.0031837934200623823_dp, 0.006141610693025735_dp, 0.011522746445694513_dp, &
             0.021095520278357115_dp, 0.03778766917975653_dp, 0.06637534135235426_dp, &
             0.11454869777127678_dp, 0.1945561508240568_dp, 0.32577668041186647_dp, &
             0.5389678132435999_dp, 0.8843210297074802_dp, 1.451226754708991_dp, &
             2.4378499788117503_dp, 4.573396094152261_dp]

      CASE (841)

         aw(:) = & ! a
            [1.642797958173549e-07_dp, 1.3293855129189586e-06_dp, 5.888879200396203e-06_dp, &
             2.0450536928130472e-05_dp, 6.103049181883815e-05_dp, 0.0001635987345737612_dp, &
             0.0004043470549977173_dp, 0.000937264853526142_dp, 0.0020617540488895377_dp, &
             0.004341222878829549_dp, 0.008806516244467185_dp, 0.017298098709055992_dp, &
             0.03303168247513459_dp, 0.061518728279676645_dp, 0.11204368275724312_dp, &
             0.20000588030509914_dp, 0.350596606796668_dp, 0.6045479497995547_dp, &
             1.027202405571357_dp, 1.7234418124625002_dp, 2.865330402770221_dp, &
             4.7580295734761116_dp, 8.089553289290992_dp, &
             ! w
             4.6832912902611977e-07_dp, 2.224128312700839e-06_dp, 7.915040539097015e-06_dp, &
             2.3711823649761866e-05_dp, 6.312260513406812e-05_dp, 0.00015411538082922845_dp, &
             0.0003519672887267903_dp, 0.0007618699251709589_dp, 0.001577743567749907_dp, &
             0.003147494012590352_dp, 0.006080713471153958_dp, 0.011423528828298082_dp, &
             0.020938307009986876_dp, 0.03754524465754437_dp, 0.06601161090853291_dp, &
             0.11401826336980933_dp, 0.1938059263435469_dp, 0.32475115466366694_dp, &
             0.5376195808085394_dp, 0.8826264458112499_dp, 1.4492005982831313_dp, &
             2.4355384691127506_dp, 4.570764632337974_dp]

      CASE (842)

         aw(:) = & ! a
            [1.0236764121615025e-07_dp, 9.425968155937635e-07_dp, 4.553432239128545e-06_dp, &
             1.6683963942562624e-05_dp, 5.1634730582147975e-05_dp, 0.000142147023612431_dp, &
             0.0003586178348423604_dp, 0.0008450564941822454_dp, 0.0018842947714475182_dp, &
             0.004013136165950289_dp, 0.008221044247124321_dp, 0.016286094130054978_dp, &
             0.031332887620934474_dp, 0.05874430817058738_dp, 0.10763026421256834_dp, &
             0.19316406939810196_dp, 0.34026188251061096_dp, 0.5893495405675959_dp, &
             1.0054753168931867_dp, 1.6933128661726842_dp, 2.8248947704965333_dp, &
             4.705565890786283_dp, 8.023492871028985_dp, &
             ! w
             3.048838482564978e-07_dp, 1.6813054036109682e-06_dp, 6.423195826201865e-06_dp, &
             2.0066197946680867e-05_dp, 5.498887148884597e-05_dp, 0.00013719787601013992_dp, &
             0.00031870589417680125_dp, 0.0006994681188301581_dp, 0.0014652879406124022_dp, &
             0.0029519016672341267_dp, 0.0057512664670887345_dp, 0.010884874190899094_dp, &
             0.020082099271589843_dp, 0.03622121562957775_dp, 0.06401995184162773_dp, &
             0.11110700807552402_dp, 0.1896796134233492_dp, 0.3190998055068231_dp, &
             0.530177226613226_dp, 0.8732587638764073_dp, 1.4379879356005678_dp, &
             2.422739199643848_dp, 4.55619459265961_dp]

      CASE (843)

         aw(:) = & ! a
            [8.046371021012933e-08_dp, 8.02108847484878e-07_dp, 4.0489926441448694e-06_dp, &
             1.521986893157837e-05_dp, 4.790740389735278e-05_dp, 0.00013350658054668902_dp, &
             0.000339978350380571_dp, 0.0008071084564914788_dp, 0.0018106762238904546_dp, &
             0.003876105908789264_dp, 0.007975084303408493_dp, 0.015858779524321617_dp, &
             0.030612358509774795_dp, 0.05756286716240572_dp, 0.10574417462454092_dp, &
             0.19023081518969803_dp, 0.33581831427883096_dp, 0.582797742111891_dp, &
             0.9960873023195477_dp, 1.680267873386839_dp, 2.807356803893998_dp, &
             4.682779200862605_dp, 7.994770949519867_dp, &
             ! w
             2.4697645676339603e-07_dp, 1.4783354679370506e-06_dp, 5.842470798410027e-06_dp, &
             1.861419518956962e-05_dp, 5.169736429525192e-05_dp, 0.00013026854204282543_dp, &
             0.00030495029513828297_dp, 0.0006734558822313016_dp, 0.0014180968944220984_dp, &
             0.0028693519960576084_dp, 0.0056115292727673734_dp, 0.010655395374307581_dp, &
             0.019715908812675716_dp, 0.035652956396335714_dp, 0.0631624440936236_dp, &
             0.10984996331403596_dp, 0.18789326409600282_dp, 0.3166474639418603_dp, &
             0.5269409416836507_dp, 0.8691780769919247_dp, 1.4330971028809125_dp, &
             2.417152295738832_dp, 4.549835191259431_dp]

      CASE (844)

         aw(:) = & ! a
            [6.97509166551491e-08_dp, 7.322783968639456e-07_dp, 3.7929649175433556e-06_dp, &
             1.446607224553548e-05_dp, 4.596920541297255e-05_dp, 0.0001289804171133798_dp, &
             0.00033015837797235295_dp, 0.0007870237419688569_dp, 0.0017715634116700508_dp, &
             0.003803068161853224_dp, 0.007843623154709661_dp, 0.01562983643225872_dp, &
             0.030225499458253757_dp, 0.05692734377673081_dp, 0.10472789421740605_dp, &
             0.18864789650982447_dp, 0.33341708642435836_dp, 0.5792529112294669_dp, &
             0.9910023644134035_dp, 1.673195341828092_dp, 2.7978405347031763_dp, &
             4.670406762121503_dp, 7.9791682579659255_dp, &
             ! w
             2.1862331324695392e-07_dp, 1.3758059479346798e-06_dp, 5.54328351617483e-06_dp, &
             1.7857786196301592e-05_dp, 4.99694394589399e-05_dp, 0.00012660965398606224_dp, &
             0.0002976533088853177_dp, 0.0006596047169180028_dp, 0.0013928882717624761_dp, &
             0.0028251352857489717_dp, 0.005536503651580631_dp, 0.010531930574906107_dp, &
             0.01951852560501222_dp, 0.035346146522593914_dp, 0.06269877224808298_dp, &
             0.10916933024048832_dp, 0.18692484253419805_dp, 0.31531650946748363_dp, &
             0.5251827836269845_dp, 0.8669593404119038_dp, 1.430436218907681_dp, &
             2.4141116777556366_dp, 4.546374261265121_dp]

      CASE (845)

         aw(:) = & ! a
            [6.417150621734594e-08_dp, 6.955307285698317e-07_dp, 3.656575263586883e-06_dp, &
             1.4061256330510396e-05_dp, 4.4922524237699306e-05_dp, 0.0001265261447907051_dp, &
             0.00032481668790693645_dp, 0.0007760705667137015_dp, 0.001750188344007474_dp, &
             0.003763082183390906_dp, 0.007771542172269596_dp, 0.01550413863323877_dp, &
             0.030012851959179694_dp, 0.05657764885399067_dp, 0.10416817042945352_dp, &
             0.18777536661262895_dp, 0.33209249870152535_dp, 0.5772961583543643_dp, &
             0.9881937758018039_dp, 1.6692868642511767_dp, 2.7925792087923744_dp, &
             4.663563847069891_dp, 7.970536451977698_dp, &
             ! w
             2.0384316972412146e-07_dp, 1.3213269320747846e-06_dp, 5.382552054907922e-06_dp, &
             1.7448904426729824e-05_dp, 4.903139047955951e-05_dp, 0.00012461690541656648_dp, &
             0.00029366896419342624_dp, 0.0006520257820507237_dp, 0.0013790706502032032_dp, &
             0.0028008623244754095_dp, 0.005495264273797778_dp, 0.010463987842098016_dp, &
             0.01940979502032141_dp, 0.03517698304271777_dp, 0.06244291062300987_dp, &
             0.1087934656357535_dp, 0.18638969044989154_dp, 0.314580569994382_dp, &
             0.5242100979284817_dp, 0.8657312824871871_dp, 1.4289629306473244_dp, &
             2.4124278226777585_dp, 4.544457676956261_dp]

      CASE (846)

         aw(:) = & ! a
            [6.18945930124138e-08_dp, 6.804478465714669e-07_dp, 3.6002341044276897e-06_dp, &
             1.389333245591711e-05_dp, 4.4487105670113755e-05_dp, 0.00012550303103618475_dp, &
             0.000322586289763116_dp, 0.0007714911737015719_dp, 0.0017412420911464264_dp, &
             0.0037463314131483427_dp, 0.007741322802124294_dp, 0.015451405258495074_dp, &
             0.029923588152076325_dp, 0.05643077900876626_dp, 0.1039329790448533_dp, &
             0.18740858146843706_dp, 0.3315354698817613_dp, 0.5764730027615071_dp, &
             0.9870119122788135_dp, 1.6676417211694918_dp, 2.790364125973689_dp, &
             4.660682377307578_dp, 7.966901235678559_dp, &
             ! w
             1.9780802726756973e-07_dp, 1.2988511544257806e-06_dp, 5.315866431603908e-06_dp, &
             1.727872621458838e-05_dp, 4.864011446011469e-05_dp, 0.00012378432386470057_dp, &
             0.0002920021094461445_dp, 0.0006488517443326375_dp, 0.0013732786881359863_dp, &
             0.002790679977260903_dp, 0.005477953160111422_dp, 0.010435450902035993_dp, &
             0.01936410311610274_dp, 0.03510586245937061_dp, 0.06233529479125346_dp, &
             0.10863531632941754_dp, 0.18616444154994063_dp, 0.31427071171484716_dp, &
             0.5238004476447038_dp, 0.865213962844427_dp, 1.4283422050067183_dp, &
             2.4117183262453987_dp, 4.543650151818711_dp]

      CASE (847)

         aw(:) = & ! a
            [6.179038412892102e-08_dp, 6.797562684401005e-07_dp, 3.5976455420632004e-06_dp, &
             1.3885607229494045e-05_dp, 4.446705672522031e-05_dp, 0.0001254558908957228_dp, &
             0.0003224834721762476_dp, 0.0007712799863038676_dp, 0.0017408293801960475_dp, &
             0.0037455584454521174_dp, 0.007739927988342882_dp, 0.015448970771089355_dp, &
             0.02991946643576507_dp, 0.056423996239399504_dp, 0.10392211580052599_dp, &
             0.18739163782269105_dp, 0.3315097348903452_dp, 0.5764349685681808_dp, &
             0.9869572987208006_dp, 1.6675656932846914_dp, 2.790261751781642_dp, &
             4.6605491967257295_dp, 7.966733208853524_dp, &
             ! w
             1.975317596154078e-07_dp, 1.2978189319536624e-06_dp, 5.312798439624158e-06_dp, &
             1.7270889128557884e-05_dp, 4.8622083044750136e-05_dp, 0.00012374593586876733_dp, &
             0.000291925224357549_dp, 0.0006487052905521958_dp, 0.0013730113661025365_dp, &
             0.0027902099111700982_dp, 0.005477153831574969_dp, 0.01043413298656968_dp, &
             0.019361992591081827_dp, 0.035102576901136155_dp, 0.0623303226240599_dp, &
             0.10862800851281051_dp, 0.18615403206967085_dp, 0.3142563908089402_dp, &
             0.5237815129929865_dp, 0.8651900497697063_dp, 1.4283135102178455_dp, &
             2.41168552624284_dp, 4.5436128189152685_dp]

      CASE (848)

         aw(:) = & ! a
            [0.0013191708171382422_dp, 0.007013952244568126_dp, 0.017526845075845704_dp, &
             0.0333571762122411_dp, 0.05531426154998188_dp, 0.08462988169142081_dp, &
             0.12311406077479778_dp, 0.17335279408223145_dp, 0.23894016949454994_dp, &
             0.32474521457344047_dp, 0.4372340609281586_dp, 0.5848856078879786_dp, &
             0.7787449431160711_dp, 1.0331614665744948_dp, 1.3667708081197638_dp, &
             1.8038106398102907_dp, 2.375923866086619_dp, 3.1247325963124015_dp, &
             4.105750881454465_dp, 5.3948814186995575_dp, 7.100534729218748_dp, &
             9.389955610016083_dp, 12.560267378276638_dp, 17.320448524523357_dp, &
             ! w
             0.0033925596897429067_dp, 0.008042037219228835_dp, 0.01306564813422215_dp, &
             0.018728055881360426_dp, 0.02538880935359407_dp, 0.03354078196264304_dp, &
             0.04385470099486025_dp, 0.05721738360336377_dp, 0.0747635996265248_dp, &
             0.09791735828377998_dp, 0.12846436813436518_dp, 0.16866740509917058_dp, &
             0.2214266697433495_dp, 0.2904907275587165_dp, 0.3807381363095051_dp, &
             0.498573667268664_dp, 0.6525272048530176_dp, 0.854240467429174_dp, &
             1.1202609140421145_dp, 1.475681670900564_dp, 1.9625327292416328_dp, &
             2.662659520035786_dp, 3.778741804491519_dp, 6.115254808272038_dp]

      CASE (849)

         aw(:) = & ! a
            [0.0009411302711210234_dp, 0.005010776748467259_dp, 0.012553305018747674_dp, &
             0.02398551551351663_dp, 0.03999407580327986_dp, 0.061642358579819756_dp, &
             0.090517072942732_dp, 0.12891042621147536_dp, 0.18003192722958608_dp, &
             0.24825861535947816_dp, 0.3394531443872367_dp, 0.4613895163423435_dp, &
             0.6243271076486784_dp, 0.8417776615581216_dp, 1.1315256799297877_dp, &
             1.5169952236339346_dp, 2.0291179682075584_dp, 2.7089825914562047_dp, &
             3.611821627047202_dp, 4.813554899249381_dp, 6.422876441315519_dp, &
             8.607357522526408_dp, 11.663737122712822_dp, 16.296795135086885_dp, &
             ! w
             0.0024211008057564223_dp, 0.005754967263036798_dp, 0.009398336695779457_dp, &
             0.013578768450602307_dp, 0.01861312264968949_dp, 0.02494483622588476_dp, &
             0.03318311473900336_dp, 0.044134415578475084_dp, 0.058832891141752885_dp, &
             0.07859059331877351_dp, 0.1050846660033124_dp, 0.14048532852853143_dp, &
             0.18762501598265888_dp, 0.2502177421819362_dp, 0.33315178590832834_dp, &
             0.44289962052824366_dp, 0.588130089211663_dp, 0.7807015575852408_dp, &
             1.0374444892928816_dp, 1.3837539542389072_dp, 1.8618615509861853_dp, &
             2.55342180378167_dp, 3.6596083700818394_dp, 5.976718686024244_dp]

      CASE (850)

         aw(:) = & ! a
            [0.0007395263543945543_dp, 0.003941572518727916_dp, 0.009894400012631439_dp, &
             0.018963475200702556_dp, 0.03175820110588902_dp, 0.04923397194387599_dp, &
             0.07283085095946515_dp, 0.10464331336284902_dp, 0.14761801054081408_dp, &
             0.205794626102374_dp, 0.28462292647256027_dp, 0.39139439965443834_dp, &
             0.5358267908257189_dp, 0.7308464630310925_dp, 0.9936315582964159_dp, &
             1.347011993582268_dp, 1.8213829134058839_dp, 2.457410280015274_dp, &
             3.3100773627989186_dp, 4.4552735151891225_dp, 6.001876066367415_dp, &
             8.117720744584359_dp, 11.099407203813982_dp, 15.649154276631815_dp, &
             ! w
             0.0019029291244265622_dp, 0.004532913211473161_dp, 0.007432503720489143_dp, &
             0.010805526095993041_dp, 0.014940827705365648_dp, 0.020247879918085653_dp, &
             0.027291697480156557_dp, 0.0368204077486961_dp, 0.0497970766767931_dp, &
             0.06745734254118603_dp, 0.09140486632412054_dp, 0.12374480952309476_dp, &
             0.16725719970540492_dp, 0.22562218675340426_dp, 0.3037217122755203_dp, &
             0.4080611028787368_dp, 0.5473934713361166_dp, 0.7337212829333106_dp, &
             0.984075704843862_dp, 1.3240808647498064_dp, 1.7961472991589371_dp, &
             2.4818626849087813_dp, 3.5814796050492324_dp, 5.886040667021983_dp]

      CASE (851)

         aw(:) = & ! a
            [0.0006129501526138582_dp, 0.00326980917654684_dp, 0.008221722031198595_dp, &
             0.015798203310283643_dp, 0.026554002530065395_dp, 0.04136707111295181_dp, &
             0.061570449432561965_dp, 0.08911208566379174_dp, 0.1267424836037067_dp, &
             0.1782496201537085_dp, 0.24877547351043894_dp, 0.34525075918403986_dp, &
             0.47698504141795145_dp, 0.6564582447340319_dp, 0.9003790198621467_dp, &
             1.2311086464346916_dp, 1.678608871418133_dp, 2.2831918955304893_dp, &
             3.099616300088342_dp, 4.203717226586259_dp, 5.704494140694199_dp, &
             7.769992887680097_dp, 10.696761820106_dp, 15.185234602997046_dp, &
             ! w
             0.0015775444698891115_dp, 0.003764457801212348_dp, 0.006193191407435528_dp, &
             0.009050604552440167_dp, 0.012605099621571506_dp, 0.017240539524123223_dp, &
             0.023487797087749663_dp, 0.0320500661731545_dp, 0.043836870994993685_dp, &
             0.060027237298224644_dp, 0.08216985901718306_dp, 0.11231926553983515_dp, &
             0.15321187264621117_dp, 0.20849593377631764_dp, 0.2830406454125993_dp, &
             0.3833675689982923_dp, 0.5182858023813587_dp, 0.6999020681545637_dp, &
             0.9454003082021856_dp, 1.280587343651824_dp, 1.7480292773401191_dp, &
             2.4292968405980573_dp, 3.524004248152637_dp, 5.81939298104682_dp]

      CASE (852)

         aw(:) = & ! a
            [0.0005255458047953153_dp, 0.002805672135654855_dp, 0.007064817121119869_dp, &
             0.013605545168793197_dp, 0.022941333066032688_dp, 0.03589101920315231_dp, &
             0.05370484869622658_dp, 0.07821618311476665_dp, 0.11202153708005351_dp, &
             0.15871142054471787_dp, 0.22318640936332948_dp, 0.3120934307647829_dp, &
             0.43441901082993284_dp, 0.6022868454952361_dp, 0.8320273707670233_dp, &
             1.1456203189097798_dp, 1.5726703556808248_dp, 2.153194011006976_dp, &
             2.941758085517256_dp, 4.014147522339121_dp, 5.479463783513662_dp, &
             7.505943041875229_dp, 10.3901426953622_dp, 14.831182888623278_dp, &
             ! w
             0.00135282748473527_dp, 0.0032331393567288235_dp, 0.0053345250470015525_dp, &
             0.007830940115213342_dp, 0.010974985777160386_dp, 0.015130232240431314_dp, &
             0.020800045019433548_dp, 0.028651616905510078_dp, 0.039552323432727854_dp, &
             0.05463701810080754_dp, 0.07541100247651644_dp, 0.103887658437725_dp, &
             0.14276652295957712_dp, 0.195667203922333_dp, 0.2674448217510964_dp, &
             0.36462989912591764_dp, 0.4960730866531824_dp, 0.6739634053798428_dp, &
             0.9156096684986644_dp, 1.2469728334528365_dp, 1.7107578345227685_dp, &
             2.3885436800643545_dp, 3.4794705364059135_dp, 5.767859411086518_dp]

      CASE (853)

         aw(:) = & ! a
            [0.00046131770047957684_dp, 0.002464432680446064_dp, 0.006213450082941195_dp, &
             0.011989713237440542_dp, 0.0202739670923448_dp, 0.03183770990271786_dp, &
             0.047864160735242814_dp, 0.07009325128549257_dp, 0.10099574452427537_dp, &
             0.14400101711456137_dp, 0.20381247218744628_dp, 0.28684429110282955_dp, &
             0.40181627215137466_dp, 0.5605554522652554_dp, 0.779074123523175_dp, &
             1.0790272450715859_dp, 1.489710948988265_dp, 2.0508806536274333_dp, &
             2.8169263178406574_dp, 3.863574512674365_dp, 5.299999835657104_dp, &
             7.29459539669274_dp, 10.143942604884918_dp, 14.546123831909066_dp, &
             ! w
             0.0011876772426390693_dp, 0.0028422624266297245_dp, 0.00470164567550287_dp, &
             0.0069294826805314_dp, 0.009765569875563148_dp, 0.013556709647476461_dp, &
             0.01878334347268186_dp, 0.02608292084610775_dp, 0.0362884493956485_dp, &
             0.05049880066327403_dp, 0.07018333845808819_dp, 0.09732040012787682_dp, &
             0.1345770042517659_dp, 0.18554614609096862_dp, 0.2550676629093684_dp, &
             0.3496754575275062_dp, 0.4782507116559213_dp, 0.6530477944791613_dp, &
             0.8914784297039305_dp, 1.2196343781360022_dp, 1.6803434771868369_dp, &
             2.355205115670181_dp, 3.4429887645006825_dp, 5.7256520566153215_dp]

      CASE (854)

         aw(:) = & ! a
            [0.0004119800147578133_dp, 0.002202187666631192_dp, 0.0055586298765522325_dp, &
             0.01074538586480867_dp, 0.018216412349394823_dp, 0.028704141988056424_dp, &
             0.04333598952502643_dp, 0.06377367777025975_dp, 0.09238261701141717_dp, &
             0.13245726067082253_dp, 0.188535709332668_dp, 0.2668362894363366_dp, &
             0.37585308949485263_dp, 0.5271604084466373_dp, 0.7364974248909912_dp, &
             1.0252378751335003_dp, 1.4224079498254034_dp, 1.9675313996547419_dp, &
             2.714837375648203_dp, 3.739993769227867_dp, 5.152232757324615_dp, &
             7.120084317099094_dp, 9.940166171561_dp, 14.309718815359435_dp, &
             ! w
             0.0010608016880277566_dp, 0.0025417051448631085_dp, 0.004214205772075051_dp, &
             0.006233486698446084_dp, 0.00882867531556558_dp, 0.012332378058316955_dp, &
             0.01720551672303631_dp, 0.024060421813113842_dp, 0.03370136370923063_dp, &
             0.04719708174663011_dp, 0.06598639370043116_dp, 0.0920172060027153_dp, &
             0.1279276068925254_dp, 0.1772861016555703_dp, 0.24491720575747278_dp, &
             0.33735512811167173_dp, 0.4635046132408782_dp, 0.6356740259080225_dp, &
             0.8713627347727518_dp, 1.1967765174677258_dp, 1.6548538849467997_dp, &
             2.3272215313387874_dp, 3.4123494670797423_dp, 5.690227703015799_dp]

      CASE (855)

         aw(:) = & ! a
            [0.00037280652043157516_dp, 0.001993884363213998_dp, 0.005038114242461587_dp, &
             0.009755171799740833_dp, 0.016576544251215717_dp, 0.02620165616937378_dp, &
             0.03971041814160165_dp, 0.058697706705116655_dp, 0.08543885609538102_dp, &
             0.12311299579062895_dp, 0.17611672519469254_dp, 0.2505001125542848_dp, &
             0.3545623636916611_dp, 0.49965827477887637_dp, 0.7012882771117724_dp, &
             0.9805784667246793_dp, 1.3663152525082587_dp, 1.8978141990917545_dp, &
             2.6291570446010635_dp, 3.6359537976308074_dp, 5.02748284753467_dp, &
             6.972395137797034_dp, 9.767351389887494_dp, 14.108891181317151_dp, &
             ! w
             0.000960054683003753_dp, 0.0023028514079451246_dp, 0.003826260898935329_dp, &
             0.005678328736677647_dp, 0.008079091011558888_dp, 0.011348896077825356_dp, &
             0.015931753892874975_dp, 0.022418403276656024_dp, 0.03158859520947007_dp, &
             0.044485269569242814_dp, 0.0625207250741233_dp, 0.08761597891937416_dp, &
             0.12238306665951883_dp, 0.17036786604818624_dp, 0.23637986246268636_dp, &
             0.3269515880659933_dp, 0.45100637886459044_dp, 0.620898175320341_dp, &
             0.8542023716461618_dp, 1.1772256461783464_dp, 1.6330070779763841_dp, &
             2.3032044844109545_dp, 3.386039706743681_dp, 5.659826015269823_dp]

      CASE (856)

         aw(:) = & ! a
            [0.00034089498965815616_dp, 0.001824133600156173_dp, 0.004613646570599141_dp, &
             0.008946853774764174_dp, 0.015236030518234703_dp, 0.024152193288095838_dp, &
             0.03673410832337178_dp, 0.05451857619209025_dp, 0.07970266880066451_dp, &
             0.11536532576269588_dp, 0.16578001208385867_dp, 0.23684987689224068_dp, &
             0.3367031868087777_dp, 0.4765012070768546_dp, 0.6715326750881394_dp, &
             0.9427027157804668_dp, 1.3185821764209442_dp, 1.8382976097143267_dp, &
             2.555794682986875_dp, 3.546626869756141_dp, 4.920110196125663_dp, &
             6.845003706638745_dp, 9.618014659823633_dp, 13.93508725407195_dp, &
             ! w
             0.0008779770928030986_dp, 0.0021081160334263773_dp, 0.003509543421451002_dp, &
             0.00522417896258082_dp, 0.0074641751489964185_dp, 0.010539131229068377_dp, &
             0.014878201275783298_dp, 0.02105329693480702_dp, 0.029822905091615504_dp, &
             0.042207487086141526_dp, 0.059595941453445_dp, 0.08388520698504935_dp, &
             0.11766363072685189_dp, 0.16445609334096556_dp, 0.229057495231711_dp, &
             0.3179973760891294_dp, 0.44021398789701405_dp, 0.6081003912027022_dp, &
             0.8392988888820658_dp, 1.1602065365684422_dp, 1.6139545135007922_dp, &
             2.282233801677403_dp, 3.3630564778542484_dp, 5.633281022925557_dp]

      CASE (857)

         aw(:) = & ! a
            [0.0001887958055683052_dp, 0.0010140011514231941_dp, 0.002583003922601531_dp, &
             0.005065843636089984_dp, 0.008767303672868059_dp, 0.014195806645690045_dp, &
             0.022150085281441592_dp, 0.033825817559567886_dp, 0.05096071305489695_dp, &
             0.07604460262093993_dp, 0.11262094036163765_dp, 0.16571019421414088_dp, &
             0.2423970234164884_dp, 0.35264021704239473_dp, 0.5103880232179784_dp, &
             0.7351166780546029_dp, 1.0539684383384267_dp, 1.5047765109252977_dp, &
             2.1405059805604374_dp, 3.0362342895097014_dp, 4.301436353372956_dp, &
             6.105565182872631_dp, 8.745790105013906_dp, 12.91480089982088_dp, &
             ! w
             0.0004866548172524495_dp, 0.0011772553374289525_dp, 0.001988302817545131_dp, &
             0.0030270108282469885_dp, 0.004459200392617905_dp, 0.006529200493232991_dp, &
             0.009576285769530103_dp, 0.014061355427484788_dp, 0.020618451757895194_dp, &
             0.030132944360771286_dp, 0.04384682754072329_dp, 0.06349915029377412_dp, &
             0.09151598975211848_dp, 0.13126986512357425_dp, 0.18743679329973525_dp, &
             0.2664949553638483_dp, 0.37744243279562967_dp, 0.5328904322696564_dp, &
             0.7508917684920442_dp, 1.0584332354854833_dp, 1.4992910285546268_dp, &
             2.155484367516968_dp, 3.223907162182144_dp, 5.472819232890789_dp]

      CASE (858)

         aw(:) = & ! a
            [0.00013342797987973343_dp, 0.0007184332548415464_dp, 0.0018390692702492176_dp, &
             0.0036350651574475375_dp, 0.006361497859811503_dp, 0.010449127026494677_dp, &
             0.016579552724153_dp, 0.025781310602525833_dp, 0.039567123007227666_dp, &
             0.06013626034170526_dp, 0.09066648681820992_dp, 0.13572698761687055_dp, &
             0.201856747454321_dp, 0.2983708670958517_dp, 0.4384820926388917_dp, &
             0.6408613830131488_dp, 0.9318208285622838_dp, 1.3484125488338308_dp, &
             1.9429738003820318_dp, 2.7902294585597747_dp, 3.9996767562662234_dp, &
             5.741148624572861_dp, 8.312178747319939_dp, 12.404009589084724_dp, &
             ! w
             0.0003441303682860819_dp, 0.0008367064277683689_dp, 0.0014271485282657704_dp, &
             0.0022063350723832945_dp, 0.0033170826663852255_dp, 0.004970229512062173_dp, &
             0.007459814830044763_dp, 0.01119250331316113_dp, 0.01674087326122238_dp, &
             0.024920250826889152_dp, 0.036892239616004956_dp, 0.05430535765067142_dp, &
             0.07948827337017293_dp, 0.11571649985817099_dp, 0.16758205278780863_dp, &
             0.24151137898409777_dp, 0.3465094381247945_dp, 0.4952839640656329_dp, &
             0.7061011756092086_dp, 1.006282080220728_dp, 1.4400007613120491_dp, &
             2.0895417976739252_dp, 3.1513302933553375_dp, 5.38929409080123_dp]

      CASE (859)

         aw(:) = & ! a
            [0.00010425047596871953_dp, 0.0005624348465868834_dp, 0.0014452921523250921_dp, &
             0.0028743854555867534_dp, 0.0050744449677817835_dp, 0.008427914037653294_dp, &
             0.013542889158904291_dp, 0.021343015765992533_dp, 0.03319959616368126_dp, &
             0.05112789056756592_dp, 0.07807138857580394_dp, 0.11830635507395201_dp, &
             0.17801276250383263_dp, 0.2660763393912661_dp, 0.3952121551997967_dp, &
             0.5835378573858977_dp, 0.8567851515927054_dp, 1.2514483302756385_dp, &
             1.819402229050183_dp, 2.635093881857171_dp, 3.808006307894955_dp, &
             5.508225843603138_dp, 8.033573159828725_dp, 12.074433686364182_dp, &
             ! w
             0.00026899663267416335_dp, 0.0006566251673895259_dp, 0.00112870041866376_dp, &
             0.0017660182423269914_dp, 0.0026967294542568783_dp, 0.004110114750662556_dp, &
             0.006271625788925743_dp, 0.009553959411800093_dp, 0.014490415308751625_dp, &
             0.021850120738124184_dp, 0.032740222014243536_dp, 0.04874654977830935_dp, &
             0.0721287903700858_dp, 0.10609175301845192_dp, 0.15516329894602404_dp, &
             0.22572540588939047_dp, 0.32677691084377064_dp, 0.47108079534345704_dp, &
             0.6770418744422565_dp, 0.972210621290131_dp, 1.4010476557873797_dp, &
             2.0460514237227136_dp, 3.103385390621538_dp, 5.334177993889868_dp]

      CASE (860)

         aw(:) = & ! a
            [8.607026921566776e-05_dp, 0.0004651149399599326_dp, 0.0011990714173133468_dp, &
             0.0023970742868090343_dp, 0.004262797199943254_dp, 0.007144731139314655_dp, &
             0.011599103119600454_dp, 0.018475562086382267_dp, 0.02904539171788074_dp, &
             0.04519283422904983_dp, 0.06969307828951081_dp, 0.10660994270366297_dp, &
             0.16186050755153258_dp, 0.24401302932384153_dp, 0.36541091229727196_dp, &
             0.5437541677448378_dp, 0.8043311718104402_dp, 1.1832039084244395_dp, &
             1.731881450340041_dp, 2.524581668394068_dp, 3.6707603384419523_dp, &
             5.340690019952558_dp, 7.832425996094568_dp, 11.83577476996332_dp, &
             ! w
             0.0002221683929426592_dp, 0.0005441123408987849_dp, 0.0009413819308082749_dp, &
             0.0014877198010867876_dp, 0.002300785473681304_dp, 0.0035544114405579833_dp, &
             0.005493850588446286_dp, 0.008467812372742653_dp, 0.012981331176203046_dp, &
             0.019769567948437095_dp, 0.02989896794631233_dp, 0.044907858676198946_dp, &
             0.06700288575139283_dp, 0.09933347646179794_dp, 0.1463757770523572_dp, &
             0.2144735415612446_dp, 0.3126152110717114_dp, 0.4535994422732931_dp, &
             0.6559311255358178_dp, 0.9473335289111675_dp, 1.3724903135207094_dp, &
             2.014077934557657_dp, 3.0680933792197753_dp, 5.293637803543986_dp]

      CASE (861)

         aw(:) = & ! a
            [7.3586964524901e-05_dp, 0.00039822243961714813_dp, 0.0010295070020226668_dp, &
             0.0020673884420290708_dp, 0.003699797621461879_dp, 0.006249607757407593_dp, &
             0.010233822311661282_dp, 0.016446140097843794_dp, 0.026082004011254094_dp, &
             0.04092572169984396_dp, 0.06362318238372444_dp, 0.09807377219478362_dp, &
             0.14998958273020963_dp, 0.2276894816490094_dp, 0.34322268142947926_dp, &
             0.5139559274251155_dp, 0.7648206230244912_dp, 1.131526861707749_dp, &
             1.6652814845024666_dp, 2.440107332000138_dp, 3.565427954738515_dp, &
             5.211660996073357_dp, 7.677059397500033_dp, 11.651007329094579_dp, &
             ! w
             0.00019000655349626021_dp, 0.000466679698561009_dp, 0.0008119727488911588_dp, &
             0.0012943200462733895_dp, 0.0020233541924675608_dp, 0.0031611071563114697_dp, &
             0.004937564800240142_dp, 0.00768324601351367_dp, 0.011881399666067948_dp, &
             0.01824059797471018_dp, 0.027795058927595203_dp, 0.04204512216950357_dp, &
             0.06315456192273859_dp, 0.0942274356570296_dp, 0.13969666064059613_dp, &
             0.20587267084008198_dp, 0.30173206945464476_dp, 0.44009822587996245_dp, &
             0.6395529875161727_dp, 0.9279571211505386_dp, 1.3501763236166355_dp, &
             1.989039213231245_dp, 3.040428518294955_dp, 5.261876769535011_dp]

      CASE (862)

         aw(:) = & ! a
            [6.445145739993479e-05_dp, 0.00034922573864358704_dp, 0.000905098491367133_dp, &
             0.00182487183604718_dp, 0.0032841141209512636_dp, 0.005585441589281539_dp, &
             0.009214804580489082_dp, 0.014921600245856427_dp, 0.023841045810697594_dp, &
             0.03767767689185784_dp, 0.05897354931366518_dp, 0.09149519837154177_dp, &
             0.14078810334579403_dp, 0.21496721497254054_dp, 0.32583971635823444_dp, &
             0.4904963926650961_dp, 0.7335709216685262_dp, 1.090477379205174_dp, &
             1.6121653554439899_dp, 2.3724879281399867_dp, 3.4808347512663413_dp, &
             5.107741028138104_dp, 7.551630076446481_dp, 11.501561931691924_dp, &
             ! w
             0.00016646513159794063_dp, 0.00040990072343832025_dp, 0.0007167644763677765_dp, &
             0.0011512986965996007_dp, 0.0018167165349732256_dp, 0.0028656421619792493_dp, &
             0.004515985656032691_dp, 0.0070838041047658114_dp, 0.011034785918661583_dp, &
             0.01705581702372685_dp, 0.026154602882558043_dp, 0.03979999061796447_dp, &
             0.06011991500298289_dp, 0.09018010732401763_dp, 0.1343764044627956_dp, &
             0.19898978208377358_dp, 0.29298466878551144_dp, 0.4292023522421472_dp, &
             0.6262863980504009_dp, 0.9122110915261565_dp, 1.3319955279595996_dp, &
             1.9686010802961946_dp, 3.0178281260791664_dp, 5.235941524008776_dp]

      CASE (863)

         aw(:) = & ! a
            [5.745739111426376e-05_dp, 0.0003116846230207128_dp, 0.0008096351308322715_dp, &
             0.0016383481863358011_dp, 0.0029633399761531728_dp, 0.005070666195328971_dp, &
             0.008420873912538204_dp, 0.013727106346483556_dp, 0.022075154600468955_dp, &
             0.03510378419086121_dp, 0.05526900301692333_dp, 0.08622668736665065_dp, &
             0.1333828376238029_dp, 0.2046808406849177_dp, 0.3117232073994729_dp, &
             0.4713660948883066_dp, 0.7079885342023235_dp, 1.0567496189304995_dp, &
             1.5683752168401752_dp, 2.316568116587947_dp, 3.410683911516137_dp, &
             5.021355794532988_dp, 7.447156548657511_dp, 11.376888292852547_dp, &
             ! w
             0.00014843875083885981_dp, 0.0003663546575915057_dp, 0.0006435287111380763_dp, &
             0.0010407777395191687_dp, 0.0016560170122496632_dp, 0.002634136311516203_dp, &
             0.004183175949970894_dp, 0.006607302973256726_dp, 0.010357597191680452_dp, &
             0.016102734849353623_dp, 0.024828006719809642_dp, 0.03797546832093804_dp, &
             0.057642355526055275_dp, 0.08686129138751228_dp, 0.12999568821754237_dp, &
             0.19330015768557157_dp, 0.28572708896859966_dp, 0.4201311760551017_dp, &
             0.6152069967628054_dp, 0.8990250692892582_dp, 1.3167367756404205_dp, &
             1.951421155464166_dp, 2.9988171987348653_dp, 5.21413327144737_dp]

      CASE (864)

         aw(:) = & ! a
            [5.191948445990932e-05_dp, 0.00028193831027236463_dp, 0.0007338912543137364_dp, &
             0.0014900425313638492_dp, 0.0027075193658428515_dp, 0.004658497172473613_dp, &
             0.007782223947844696_dp, 0.012761430705027917_dp, 0.02064033216368051_dp, &
             0.03300214027133325_dp, 0.05222985655396758_dp, 0.08188505572838813_dp, &
             0.12725436384374947_dp, 0.1961337030973845_dp, 0.29994889193047153_dp, &
             0.455352600399399_dp, 0.6865018888482628_dp, 1.028332253712091_dp, &
             1.5313717936899758_dp, 2.2691884814391297_dp, 3.351104402230329_dp, &
             4.947836194469503_dp, 7.358089601556577_dp, 11.270455891922184_dp, &
             ! w
             0.00013416311270914299_dp, 0.00033182000463506087_dp, 0.0005852923839563806_dp, &
             0.0009525257121162873_dp, 0.0015269589511221894_dp, 0.002446973731176722_dp, &
             0.003912341162469911_dp, 0.00621720431912609_dp, 0.009800201957918683_dp, &
             0.015314384395885948_dp, 0.023725699706732616_dp, 0.03645296215968129_dp, &
             0.055566622901007556_dp, 0.08407021267920062_dp, 0.1262983648264082_dp, &
             0.18848185272607848_dp, 0.27956136796808734_dp, 0.4124018598879348_dp, &
             0.6057410370231056_dp, 0.8877326398841636_dp, 1.3036441484932781_dp, &
             1.9366602165510485_dp, 2.982472960133359_dp, 5.195389787024652_dp]

      CASE (865)

         aw(:) = & ! a
            [4.7418643294369677e-05_dp, 0.00025774669745857575_dp, 0.0006722154232610096_dp, &
             0.0013690484998241299_dp, 0.002498229525311891_dp, 0.004320072604535644_dp, &
             0.0072556219168702215_dp, 0.011961599431667937_dp, 0.019446569033982664_dp, &
             0.031245928732123392_dp, 0.04967959675089279_dp, 0.07822734872513695_dp, &
             0.12207187243213447_dp, 0.18888021292811724_dp, 0.2899231791889234_dp, &
             0.4416742098213092_dp, 0.6680939923915001_dp, 1.0039191866663002_dp, &
             1.4995008048615612_dp, 2.228284598653206_dp, 3.299560036031712_dp, &
             4.884116013550487_dp, 7.280777885983118_dp, 11.177960883088987_dp, &
             ! w
             0.0001225590896420599_dp, 0.0003037117142893077_dp, 0.000537776362192978_dp, &
             0.000880243545699582_dp, 0.0014207003340747865_dp, 0.0022919513418779218_dp, &
             0.0036867037086758146_dp, 0.005890483891323854_dp, 0.009331142025327265_dp, &
             0.014648091876885297_dp, 0.022790320946691468_dp, 0.03515617241771672_dp, &
             0.05379238772587816_dp, 0.08167658856617606_dp, 0.12311755994413606_dp, &
             0.18432433390541256_dp, 0.27422633207022185_dp, 0.405696473576583_dp, &
             0.5975095853432105_dp, 0.877892506855791_dp, 1.2922160174644357_dp, &
             1.9237605372451725_dp, 2.968181765164774_dp, 5.179004985329604_dp]

      CASE (866)

         aw(:) = & ! a
            [2.610837918220881e-05_dp, 0.00014294152953711235_dp, 0.00037823829830779283_dp, &
             0.0007883291636858416_dp, 0.0014836670604052187_dp, 0.0026583205059523204_dp, &
             0.004631674459068387_dp, 0.00791463449746751_dp, 0.013313792801150573_dp, &
             0.02209008059205638_dp, 0.03619642308260961_dp, 0.0586305139732096_dp, &
             0.09395470218112094_dp, 0.14905715670184888_dp, 0.23425987030445888_dp, &
             0.3649244441932152_dp, 0.5637754403243422_dp, 0.8642762772190523_dp, &
             1.315618090876835_dp, 1.9904139347500776_dp, 2.997683363901554_dp, &
             4.5086347976046985_dp, 6.822885222851859_dp, 10.62795212409125_dp, &
             ! w
             6.758819619053168e-05_dp, 0.00016994182958350685_dp, 0.0003096606452866372_dp, &
             0.0005284497656785211_dp, 0.0008939097374504265_dp, 0.0015075545451875359_dp, &
             0.0025226054448607266_dp, 0.004175159641001831_dp, 0.0068294356160370035_dp, &
             0.011042869269580353_dp, 0.01766093958916291_dp, 0.027955193660039738_dp, &
             0.0438230950360496_dp, 0.068075890272548_dp, 0.10485184235327691_dp, &
             0.1602097043710146_dp, 0.24298868411754887_dp, 0.36608844939550544_dp, &
             0.5484950112040988_dp, 0.8188834122309723_dp, 1.2232864492484932_dp, &
             1.8456362557073374_dp, 2.8814637006610293_dp, 5.0796664597325_dp]

      CASE (867)

         aw(:) = & ! a
            [1.8416687750042485e-05_dp, 0.00010133824836742902_dp, 0.00027089196657057704_dp, &
             0.0005736944887975803_dp, 0.0011021428620727463_dp, 0.0020197910232608593_dp, &
             0.003599193320138108_dp, 0.006283462534768932_dp, 0.010783661573626497_dp, &
             0.018228384440102482_dp, 0.030390269197808767_dp, 0.05002593152632277_dp, &
             0.08138213463949241_dp, 0.13094406849250767_dp, 0.20853441581974552_dp, &
             0.32891928525221653_dp, 0.5141484657577192_dp, 0.7969749765790031_dp, &
             1.2259243497387693_dp, 1.8731090193558748_dp, 2.847353490609118_dp, &
             4.320067886379148_dp, 6.591326168156173_dp, 10.348296583711107_dp, &
             ! w
             4.772900699701755e-05_dp, 0.00012122903472759187_dp, 0.00022532396215831503_dp, &
             0.00039527422396702915_dp, 0.0006882803215319564_dp, 0.0011914420915486407_dp, &
             0.0020396288849587915_dp, 0.003444984760152902_dp, 0.005739822200722312_dp, &
             0.009439520785622884_dp, 0.015335425407593311_dp, 0.024631475028841925_dp, &
             0.03914384125939362_dp, 0.061590900140885974_dp, 0.09601259400560788_dp, &
             0.14837638042963047_dp, 0.22745864953556036_dp, 0.34615685917191813_dp, &
             0.5235557930050255_dp, 0.788565881874275_dp, 1.187589036874192_dp, &
             1.8049477083780807_dp, 2.836177413417193_dp, 5.02784446982278_dp]

      CASE (868)

         aw(:) = & ! a
            [1.4379476875660705e-05_dp, 7.944072677434636e-05_dp, 0.00021408793237907966_dp, &
             0.0004591364105742203_dp, 0.0008960284465084502_dp, 0.001669765608058009_dp, &
             0.003024395620001795_dp, 0.005361428416498812_dp, 0.009332618282514442_dp, &
             0.015983422381147253_dp, 0.02697194804698383_dp, 0.04490003834434733_dp, &
             0.07380978782009179_dp, 0.11992243664630477_dp, 0.1927303597797634_dp, &
             0.30660168304523033_dp, 0.4831302611675917_dp, 0.754583091101484_dp, &
             1.1690231093810213_dp, 1.7982059981854133_dp, 2.7508049978432276_dp, &
             4.198354868072518_dp, 6.4412469792254665_dp, 10.166465347935803_dp, &
             ! w
             3.729872339597181e-05_dp, 9.550233077636349e-05_dp, 0.0001803067692627087_dp, &
             0.00032300059771685007_dp, 0.0005743849804293405_dp, 0.0010128151727083604_dp, &
             0.0017618544393789915_dp, 0.00301850376221718_dp, 0.00509455509525917_dp, &
             0.008478010392251779_dp, 0.013924585500662114_dp, 0.02259322051719583_dp, &
             0.036245322842248864_dp, 0.05753583167231124_dp, 0.09043633888597612_dp, &
             0.1408491093142117_dp, 0.21750283201225123_dp, 0.33328682182437386_dp, &
             0.5073459035430433_dp, 0.7687455716329866_dp, 1.164139885864926_dp, &
             1.7781285068318022_dp, 2.806278126354343_dp, 4.993649960979745_dp]

      CASE (869)

         aw(:) = & ! a
            [1.1869868659392883e-05_dp, 6.579873725198031e-05_dp, 0.00017854787249337202_dp, &
             0.00038696563339825655_dp, 0.0007649275248015224_dp, 0.0014446147759121677_dp, &
             0.0026503399256163815_dp, 0.004754611765203933_dp, 0.008367479644380128_dp, &
             0.014475441545644327_dp, 0.024654734736066126_dp, 0.041395736394096064_dp, &
             0.06859208178274945_dp, 0.1122722094970158_dp, 0.18168553543521138_dp, &
             0.2909053494701547_dp, 0.46118515796801196_dp, 0.7244260174106174_dp, &
             1.128338801713982_dp, 1.744403106065123_dp, 2.6811688492959522_dp, &
             4.110257448035766_dp, 6.332301390958184_dp, 10.034173171220958_dp, &
             ! w
             3.0811805276188275e-05_dp, 7.943163795834799e-05_dp, 0.00015194604998942248_dp, &
             0.0002768673144293447_dp, 0.0005005463277249662_dp, 0.0008953032650243461_dp, &
             0.0015767804865357874_dp, 0.002731180786330725_dp, 0.004655501963583216_dp, &
             0.007817846205637882_dp, 0.01294783208509387_dp, 0.021171179313085778_dp, &
             0.03420852514002672_dp, 0.054667132055999455_dp, 0.08646661201864665_dp, &
             0.13545877645622853_dp, 0.21033398807356385_dp, 0.32397199966905077_dp, &
             0.49555889181324647_dp, 0.754273716590566_dp, 1.146960090641393_dp, &
             1.7584316253358412_dp, 2.784292793904677_dp, 4.968516083047481_dp]

      CASE (870)

         aw(:) = & ! a
            [1.0149343128049566e-05_dp, 5.6428745973685895e-05_dp, 0.00015404878224141425_dp, &
             0.0003369236083818571_dp, 0.000673292683860254_dp, 0.0012857889696723_dp, &
             0.0023839957230107796_dp, 0.004318648232629378_dp, 0.0076682672710112355_dp, &
             0.01337448312659542_dp, 0.022950847308523405_dp, 0.03880189745024021_dp, &
             0.06470630940435321_dp, 0.10654242341823435_dp, 0.17336949804441074_dp, &
             0.27902880296664756_dp, 0.4445044556725669_dp, 0.7014059797540094_dp, &
             1.0971615044945617_dp, 1.703026144992393_dp, 2.627445828493471_dp, &
             4.042106577426718_dp, 6.2478341042021786_dp, 9.931427920239633_dp, &
             ! w
             2.63626597308959e-05_dp, 6.836851844340043e-05_dp, 0.00013228172607082886_dp, &
             0.0002445269228531047_dp, 0.0004481288998405546_dp, 0.0008109145563114448_dp, &
             0.0014425486858809112_dp, 0.0025209788035202886_dp, 0.004331803005911657_dp, &
             0.007327694851387208_dp, 0.012217919552863814_dp, 0.020102131853497938_dp, &
             0.03266877601095872_dp, 0.05248719846128431_dp, 0.08343530880580467_dp, &
             0.13132392232767895_dp, 0.20481143544267916_dp, 0.31676792382147095_dp, &
             0.48640994227747353_dp, 0.743005052876109_dp, 1.1335476679526284_dp, &
             1.743024965158016_dp, 2.7670800362918144_dp, 4.94884391877996_dp]

      CASE (871)

         aw(:) = & ! a
            [8.89159792421572e-06_dp, 4.956797593859735e-05_dp, 0.00013605354182909428_dp, &
             0.0002999774816598003_dp, 0.0006051679566362878_dp, 0.00116678643116254_dp, &
             0.0021828650626226297_dp, 0.003986975454291722_dp, 0.007132642404241459_dp, &
             0.012525733218187958_dp, 0.021629584892391147_dp, 0.036779651382950494_dp, &
             0.0616616759570113_dp, 0.10203213364836669_dp, 0.16679522934986216_dp, &
             0.26960221458280226_dp, 0.4312155609159948_dp, 0.6830036801795011_dp, &
             1.0721593594555405_dp, 1.6697490515158864_dp, 2.584128848182079_dp, &
             3.9870352267500606_dp, 6.179454382406481_dp, 9.848135697648654_dp, &
             ! w
             2.3109035961115926e-05_dp, 6.02520826232466e-05_dp, 0.00011776415169459174_dp, &
             0.0002204230294433494_dp, 0.0004086458426170141_dp, 0.0007467421454615169_dp, &
             0.0013396412087189713_dp, 0.002358684841842503_dp, 0.004080293702975978_dp, &
             0.006944658969234459_dp, 0.011644500486560797_dp, 0.01925818170073456_dp, &
             0.03144771528388423_dp, 0.050751140066791114_dp, 0.0810116930296267_dp, &
             0.128005765901508_dp, 0.2003643814385786_dp, 0.3109482560832928_dp, &
             0.47899756088003065_dp, 0.7338517483047018_dp, 1.1226297817929844_dp, &
             1.7304644346169935_dp, 2.753036319086578_dp, 4.932797214633451_dp]

      CASE (872)

         aw(:) = & ! a
            [7.929434928114258e-06_dp, 4.4311988689360874e-05_dp, 0.00012222843891632857_dp, &
             0.00027146259708249633_dp, 0.0005522664255651875_dp, 0.0010737459743259787_dp, &
             0.002024550788931481_dp, 0.0037242473147021406_dp, 0.00670586488633591_dp, &
             0.011845814045816874_dp, 0.020565898751369605_dp, 0.03514420892375717_dp, &
             0.05918903235653767_dp, 0.0983549064639916_dp, 0.16141587184931835_dp, &
             0.26186308186077817_dp, 0.4202715432916509_dp, 0.6678048631194581_dp, &
             1.0514548697306447_dp, 1.6421256394906163_dp, 2.5480942412941467_dp, &
             3.9411378844433047_dp, 6.122379518032944_dp, 9.778533004563165_dp, &
             ! w
             2.0619233161851453e-05_dp, 5.4023242759847456e-05_dp, 0.00010656012328395307_dp, &
             0.0002016636896892839_dp, 0.0003776356304696957_dp, 0.0006959322006467628_dp, &
             0.001257600403685012_dp, 0.0022285230603323373_dp, 0.0038774991616097702_dp, &
             0.00663431088389791_dp, 0.011177825800732752_dp, 0.018568509445219934_dp, &
             0.030446061273980904_dp, 0.04932196657477935_dp, 0.07900988776636443_dp, &
             0.12525661891464945_dp, 0.196669278768012_dp, 0.30609968290632206_dp, &
             0.47280694186672_dp, 0.7261906390086652_dp, 1.113475405336972_dp, &
             1.719919120471269_dp, 2.741238171384873_dp, 4.91931883283547_dp]

      CASE (873)

         aw(:) = & ! a
            [7.168036277522962e-06_dp, 4.014728612425409e-05_dp, 0.00011124558155015777_dp, &
             0.0002487153937607191_dp, 0.0005098323852779322_dp, 0.0009986641968841884_dp, &
             0.001896036493131769_dp, 0.003509789687226038_dp, 0.006355720494434169_dp, &
             0.011285375555404565_dp, 0.019685375538270426_dp, 0.03378505281723869_dp, &
             0.057126649620420064_dp, 0.09527751239102887_dp, 0.15690000622272873_dp, &
             0.2553474997095138_dp, 0.4110331337431813_dp, 0.6549430321725481_dp, &
             1.033894096652381_dp, 1.6186481204262502_dp, 2.517411678399561_dp, &
             3.9019956934471325_dp, 6.073641922963699_dp, 9.719038565175081_dp, &
             ! w
             1.8648369855753105e-05_dp, 4.907985243349969e-05_dp, 9.762282118324278e-05_dp, &
             0.00018658617991906153_dp, 0.00035251087539629804_dp, 0.0006544760325581405_dp, &
             0.0011902638684413126_dp, 0.0021211361119376467_dp, 0.0037094133648498473_dp, &
             0.006375999069790026_dp, 0.01078790450375352_dp, 0.017990223151059102_dp, &
             0.029603420890950242_dp, 0.04811600332941984_dp, 0.07731591669208952_dp, &
             0.12292404654529307_dp, 0.19352631964924183_dp, 0.3019661395373008_dp, &
             0.46751820227887037_dp, 0.7196335050220782_dp, 1.1056281465924653_dp, &
             1.7108694771197424_dp, 2.7311077480719073_dp, 4.907747464825127_dp]

      CASE (874)

         aw(:) = & ! a
            [6.549491470164241e-06_dp, 3.675992118587304e-05_dp, 0.00010229150016471586_dp, &
             0.0002300989740473544_dp, 0.00047493000524356923_dp, 0.0009365735438397159_dp, &
             0.0017891969325523956_dp, 0.003330624821140441_dp, 0.006061879975247759_dp, &
             0.010813119748880438_dp, 0.01894060303116302_dp, 0.032631462568626646_dp, &
             0.055370618240384466_dp, 0.09264953682645319_dp, 0.15303313601953122_dp, &
             0.24975422887365265_dp, 0.40308392826372863_dp, 0.6438521352879277_dp, &
             1.0187211865744763_dp, 1.5983264375784603_dp, 2.4908108720262887_dp, &
             3.8680138685284535_dp, 6.031281971845975_dp, 9.667284640639497_dp, &
             ! w
             1.7046848045904e-05_dp, 4.5053300950849274e-05_dp, 9.030890460672315e-05_dp, &
             0.00017416234574369624_dp, 0.00033165942589621856_dp, 0.0006198574580607246_dp, &
             0.0011337386013460233_dp, 0.0020305792700609517_dp, 0.0035670932629317644_dp, &
             0.006156476235375433_dp, 0.010455416007863942_dp, 0.017495580981029508_dp, &
             0.028880586770290206_dp, 0.04707873752414965_dp, 0.07585527931046517_dp, &
             0.12090809746721247_dp, 0.19080410426804453_dp, 0.29837876715472794_dp, &
             0.46291986847412764_dp, 0.7139231268669681_dp, 1.098785068946494_dp, &
             1.7029702014986956_dp, 2.7222607571156687_dp, 4.8976434161101405_dp]

      CASE (875)

         aw(:) = & ! a
            [3.6231638489974144e-06_dp, 2.0666258959103596e-05_dp, 5.9386798192876944e-05_dp, &
             0.00013966450003425118_dp, 0.0003023929535973138_dp, 0.0006239132203842525_dp, &
             0.0012415834836647572_dp, 0.002397144375103838_dp, 0.0045078532857277245_dp, &
             0.008281207668000756_dp, 0.014897519896363329_dp, 0.02629696756823826_dp, &
             0.04562587704763145_dp, 0.07792345533646255_dp, 0.13116863985229227_dp, &
             0.21786277904917078_dp, 0.35740709728432757_dp, 0.5796646568492807_dp, &
             0.9303290701340091_dp, 1.479228481229943_dp, 2.3340803353211097_dp, &
             3.666874529654296_dp, 5.779609662224813_dp, 9.358915696134543_dp, &
             ! w
             9.462829356948298e-06_dp, 2.582393719143882e-05_dp, 5.478764776929219e-05_dp, &
             0.00011235789917428121_dp, 0.00022541050884424443_dp, 0.000439823252088218_dp, &
             0.000834658513466712_dp, 0.0015441503518772801_dp, 0.0027922257555908062_dp, &
             0.004946544716387514_dp, 0.00860219334755444_dp, 0.014709922383457669_dp, &
             0.024770725090628422_dp, 0.041128475924466626_dp, 0.06740672052351113_dp, &
             0.10915712459144415_dp, 0.1748216337102886_dp, 0.27717575500987374_dp, &
             0.43557523319367486_dp, 0.6797814172535712_dp, 1.0576866281492552_dp, &
             1.6553728925553162_dp, 2.668864528354028_dp, 4.8366857728249535_dp]

      CASE (876)

         aw(:) = & ! a
            [2.5668709485767075e-06_dp, 1.4814047860829722e-05_dp, 4.354977911304416e-05_dp, &
             0.0001054830813475211_dp, 0.00023527920077971498_dp, 0.0004987161664713565_dp, &
             0.0010163147078496892_dp, 0.0020036725802634477_dp, 0.003838312404310137_dp, &
             0.007168624053973503_dp, 0.013088905936750479_dp, 0.0234169831350074_dp, &
             0.041129298418742875_dp, 0.071035384896434_dp, 0.12081306080965104_dp, &
             0.20258331502155236_dp, 0.3352894288899701_dp, 0.5482785725387223_dp, &
             0.8867190554907874_dp, 1.4199910810974723_dp, 2.2555625506289085_dp, &
             3.5654854406299905_dp, 5.652108093475424_dp, 9.202090376235223_dp, &
             ! w
             6.720754064709411e-06_dp, 1.8768558315887263e-05_dp, 4.1366860990526335e-05_dp, &
             8.807052852907412e-05_dp, 0.00018210431008171_dp, 0.0003642068566959797_dp, &
             0.0007058451367603335_dp, 0.0013300309146330846_dp, 0.0024444809051307687_dp, &
             0.0043940258680745135_dp, 0.007742447448539034_dp, 0.013398821459845719_dp, &
             0.02281056032708231_dp, 0.038255613474036575_dp, 0.06328116523818518_dp, &
             0.10335828612475916_dp, 0.16685729887724446_dp, 0.2665142613433028_dp, &
             0.42171210872629666_dp, 0.6623460252619595_dp, 1.0365712927147028_dp, &
             1.6308105671358557_dp, 2.641247645709985_dp, 4.8051746648398295_dp]

      CASE (877)

         aw(:) = & ! a
            [2.011895465153751e-06_dp, 1.1723364824408138e-05_dp, 3.5095911795379164e-05_dp, &
             8.6933699715748e-05_dp, 0.000198162808962633_dp, 0.00042819888773423977_dp, &
             0.0008873107078046534_dp, 0.0017749933315776947_dp, 0.0034440403924321595_dp, &
             0.006505714062992247_dp, 0.011999826233963971_dp, 0.02166608795082084_dp, &
             0.03837168245094283_dp, 0.06677738979294004_dp, 0.11436466679637546_dp, &
             0.19300477352150924_dp, 0.321338201704145_dp, 0.5283683943596655_dp, &
             0.8589104739875735_dp, 1.3820396045324734_dp, 2.205048749977677_dp, &
             3.500024113698465_dp, 5.569547496757411_dp, 9.10031727857536_dp, &
             ! w
             5.278413113034664e-06_dp, 1.5018920759248883e-05_dp, 3.408515870926816e-05_dp, &
             7.45495484921219e-05_dp, 0.00015745171307314743_dp, 0.00032038145105777865_dp, &
             0.0006300591094180497_dp, 0.0012024073387134865_dp, 0.002234813343462219_dp, &
             0.004057440936514008_dp, 0.007213800137548638_dp, 0.012585769386347333_dp, &
             0.021585518740865547_dp, 0.03644727448239819_dp, 0.0606670935229573_dp, &
             0.09966141461388588_dp, 0.1617510054591899_dp, 0.2596428303540104_dp, &
             0.41273447392037604_dp, 0.6510072177861492_dp, 1.0227908517919093_dp, &
             1.61473916165854_dp, 2.6231536107639766_dp, 4.784535081917654_dp]

      CASE (878)

         aw(:) = & ! a
            [1.666508049577257e-06_dp, 9.79195922229786e-06_dp, 2.976719639569971e-05_dp, &
             7.508899432805849e-05_dp, 0.00017412054971556627_dp, 0.00038190131085698016_dp, &
             0.0008015896454444582_dp, 0.0016214196504195668_dp, 0.003176764750821056_dp, &
             0.006052561661062765_dp, 0.011249758614268303_dp, 0.02045204387106796_dp, &
             0.03644783683230977_dp, 0.0637901575616464_dp, 0.10981753798220796_dp, &
             0.18621860962502862_dp, 0.31141135429804784_dp, 0.5141452889209098_dp, &
             0.8389729980451124_dp, 1.3547409306661051_dp, 2.1686083562292353_dp, &
             3.452683059026712_dp, 5.50971944265975_dp, 9.026453576778804_dp, &
             ! w
             4.379962540155109e-06_dp, 1.2663911715794544e-05_dp, 2.943582064293583e-05_dp, &
             6.57479994815269e-05_dp, 0.0001411440051568916_dp, 0.0002910168337339391_dp, &
             0.0005787330395056623_dp, 0.0011151701949330154_dp, 0.002090319268985992_dp, &
             0.0038237812425822963_dp, 0.006844385901186699_dp, 0.01201420933272734_dp, &
             0.020719622038716173_dp, 0.035162659118903865_dp, 0.058801490767019744_dp, &
             0.0970117352483137_dp, 0.15807663136259326_dp, 0.254680227579877_dp, &
             0.40622915836632156_dp, 0.6427667192153472_dp, 1.0127512301683719_dp, &
             1.603009332455124_dp, 2.609935199620177_dp, 4.769459884767857_dp]

      CASE (879)

         aw(:) = & ! a
            [1.4294369505550361e-06_dp, 8.461654243821268e-06_dp, 2.6069840110085106e-05_dp, &
             6.678141969526743e-05_dp, 0.00015706198009605657_dp, 0.0003486996097534264_dp, &
             0.000739534350906486_dp, 0.0015093235042276068_dp, 0.002980253585059486_dp, &
             0.005717233409854979_dp, 0.010691509920087367_dp, 0.01954377654537907_dp, &
             0.03500176793262573_dp, 0.061535169496355194_dp, 0.10637158574329325_dp, &
             0.1810573989342988_dp, 0.3038366412503757_dp, 0.5032595320476884_dp, &
             0.8236716393551725_dp, 1.3337379024434237_dp, 2.14050999882195_dp, &
             3.416110428669821_dp, 5.463429116456053_dp, 8.969237099098958_dp, &
             ! w
             3.762806664908035e-06_dp, 1.103491926399002e-05_dp, 2.6174947287183062e-05_dp, &
             5.947853042401969e-05_dp, 0.00012938139605028366_dp, 0.0002696247824178566_dp, &
             0.0005410306668561241_dp, 0.0010506282558337906_dp, 0.001982740164044871_dp, &
             0.0036488362853055968_dp, 0.006566397881725015_dp, 0.011582129914532113_dp, &
             0.020062294297729683_dp, 0.034183727722939775_dp, 0.05737479993056737_dp, &
             0.09497882294674534_dp, 0.15524904075307505_dp, 0.2508506768719537_dp, &
             0.4011964177312313_dp, 0.6363772790557829_dp, 1.004952215817743_dp, &
             1.5938847828412488_dp, 2.599645339090537_dp, 4.757726239614608_dp]

      CASE (880)

         aw(:) = & ! a
            [1.255930727790156e-06_dp, 7.485090992357303e-06_dp, 2.3338069007555833e-05_dp, &
             6.0586242535196434e-05_dp, 0.00014421698572414032_dp, 0.00032347734989359025_dp, &
             0.0006920276156423274_dp, 0.001422928393350737_dp, 0.0028279012496830623_dp, &
             0.005455896976071931_dp, 0.01025441015651299_dp, 0.018829635036160114_dp, &
             0.03386045932107693_dp, 0.05974928969115446_dp, 0.10363391727551965_dp, &
             0.17694522226743298_dp, 0.29778558218846346_dp, 0.4945424083818464_dp, &
             0.8113915309260751_dp, 1.3168483088441403_dp, 2.1178747891163106_dp, &
             3.386604034305876_dp, 5.426036692072949_dp, 8.922975819603801_dp, &
             ! w
             3.310831078414613e-06_dp, 9.834627350156096e-06_dp, 2.3743161673730403e-05_dp, &
             5.4742290332050724e-05_dp, 0.00012040403243127354_dp, 0.0002531654467926627_dp, &
             0.0005118255426143177_dp, 0.0010003409616212845_dp, 0.001898491496210142_dp, &
             0.0035112072124122446_dp, 0.00634681078723501_dp, 0.011239561453289684_dp, &
             0.01953938600509518_dp, 0.0334025813261772_dp, 0.05623313295897414_dp, &
             0.09334779037508346_dp, 0.15297494677240175_dp, 0.24776390329859116_dp, &
             0.3971316051305622_dp, 0.6312074278898153_dp, 0.9986323715834984_dp, &
             1.5864826416512963_dp, 2.591293052132225_dp, 4.748203054927835_dp]

      CASE (881)

         aw(:) = & ! a
            [1.123051521845402e-06_dp, 6.735175067953931e-06_dp, 2.1228122658260092e-05_dp, &
             5.576207167200362e-05_dp, 0.0001341306634962567_dp, 0.000303522628339759_dp, &
             0.0006541962837252704_dp, 0.0013537380177139813_dp, 0.0027052824362745133_dp, &
             0.005244641966659691_dp, 0.009899696057067961_dp, 0.018248068926044962_dp, &
             0.03292809048949714_dp, 0.058286167403033495_dp, 0.10138515551790359_dp, &
             0.1735593393230106_dp, 0.29279234287128547_dp, 0.48733473565256846_dp, &
             0.8012192487942377_dp, 1.302834613475526_dp, 2.0990663328650316_dp, &
             3.3620553289265547_dp, 5.394895315466798_dp, 8.884418636802783_dp, &
             ! w
             2.9644873941763704e-06_dp, 8.909827183515595e-06_dp, 2.1849428898659183e-05_dp, &
             5.1012933748903805e-05_dp, 0.00011327396489001808_dp, 0.00024000370765297463_dp, &
             0.0004883388181706352_dp, 0.0009597021969695117_dp, 0.0018301155744900168_dp, &
             0.0033990828226785775_dp, 0.006167306146808942_dp, 0.010958660858319591_dp, &
             0.019109407862763588_dp, 0.03275861213423219_dp, 0.05528973775239062_dp, &
             0.09199709101900228_dp, 0.15108793776849816_dp, 0.245197819485015_dp, &
             0.3937467859571869_dp, 0.6268960204926854_dp, 0.9933553650364108_dp, &
             1.5802962648163439_dp, 2.5843092449168634_dp, 4.74024089376434_dp]

      CASE (882)

         aw(:) = & ! a
            [1.0177859464865612e-06_dp, 6.139647310366034e-06_dp, 1.954370433810159e-05_dp, &
             5.1882646722191915e-05_dp, 0.00012595984081034686_dp, 0.0002872515944652215_dp, &
             0.0006231741524940217_dp, 0.0012967235790294577_dp, 0.0026038107357220872_dp, &
             0.005069163701355182_dp, 0.00960407022854281_dp, 0.017761930364441_dp, &
             0.03214660904090592_dp, 0.05705682536175276_dp, 0.09949149353444463_dp, &
             0.1707023084949433_dp, 0.2885711521422992_dp, 0.48123109738900294_dp, &
             0.7925916919754576_dp, 1.2909322814816275_dp, 2.083071764050479_dp, &
             3.341157065314179_dp, 5.368361805850402_dp, 8.85154518963746_dp, &
             ! w
             2.689975573339747e-06_dp, 8.173187931425067e-06_dp, 2.0326441017578754e-05_dp, &
             4.7984559746393564e-05_dp, 0.00010744088445708726_dp, 0.0002291728557876196_dp, &
             0.0004689169883025968_dp, 0.0009259557155188311_dp, 0.001773127433173225_dp, &
             0.003305327936176008_dp, 0.006016772240260923_dp, 0.010722475542111462_dp, &
             0.01874701156005967_dp, 0.03221467309663146_dp, 0.05449128515398358_dp, &
             0.09085179907927976_dp, 0.14948516681838267_dp, 0.243014842555163_dp, &
             0.3908631954487179_dp, 0.6232184019024157_dp, 0.9888493267743138_dp, &
             1.5750095951043104_dp, 2.5783386829131816_dp, 4.73343442147549_dp]

      CASE (883)

         aw(:) = & ! a
            [9.321825065344991e-07_dp, 5.6542705004737865e-06_dp, 1.8164164292428563e-05_dp, &
             4.868435806347234e-05_dp, 0.00011917940460300414_dp, 0.0002736711878849238_dp, &
             0.0005971531854435529_dp, 0.0012486956410471333_dp, 0.002518013953726868_dp, &
             0.004920305859126883_dp, 0.00935256125525346_dp, 0.01734726177328151_dp, &
             0.03147845316954256_dp, 0.05600351983890705_dp, 0.09786585523202557_dp, &
             0.16824531735274884_dp, 0.28493513576069945_dp, 0.4759658094889152_dp, &
             0.7851391108561293_dp, 1.2806383949936875_dp, 2.069223747075352_dp, &
             3.3230467688977554_dp, 5.34535083343202_dp, 8.82301987561055_dp, &
             ! w
             2.4666340152198957e-06_dp, 7.571121223884459e-06_dp, 1.907074078403665e-05_dp, &
             4.54661847163987e-05_dp, 0.00010255839531299155_dp, 0.0002200603944884196_dp, &
             0.00045250663772550433_dp, 0.0008973371311270423_dp, 0.0017246436783749544_dp, &
             0.0032253374428767826_dp, 0.005888012258500257_dp, 0.01051999061798292_dp, &
             0.018435678660655788_dp, 0.031746491251632536_dp, 0.05380284110291157_dp, &
             0.08986271981527012_dp, 0.14809896124004462_dp, 0.24112426177007623_dp, &
             0.3883627540985915_dp, 0.6200259387896603_dp, 0.9849341399869239_dp, &
             1.570413029025758_dp, 2.573145656790633_dp, 4.727514713964398_dp]

      CASE (884)

         aw(:) = & ! a
            [5.255908925744213e-07_dp, 3.3304383238004107e-06_dp, 1.1443109624783916e-05_dp, &
             3.274314854585307e-05_dp, 8.463509922732927e-05_dp, 0.00020315325962123426_dp, &
             0.00045982128598097156_dp, 0.0009916500017658595_dp, 0.002053220067040241_dp, &
             0.00410523807758926_dp, 0.007962336583849743_dp, 0.015035684133530595_dp, &
             0.027725300435719146_dp, 0.050045899108959103_dp, 0.0886130494935645_dp, &
             0.15418005146046893_dp, 0.26401070026274287_dp, 0.44551914501472983_dp, &
             0.7418549215172512_dp, 1.2206148402757315_dp, 1.9881924548156118_dp, &
             3.216756274228061_dp, 5.209968530749456_dp, 8.65488645350951_dp, &
             ! w
             1.4041187144128398e-06_dp, 4.65945746373288e-06_dp, 1.2808883635109877e-05_dp, &
             3.254706865100145e-05_dp, 7.697394301879186e-05_dp, 0.0001715035993932324_dp, &
             0.0003638326660916254_dp, 0.0007408314505770373_dp, 0.0014567119340833138_dp, &
             0.00277917966751328_dp, 0.0051638676664433536_dp, 0.00937269395086945_dp, &
             0.016659663050867045_dp, 0.029059174847929973_dp, 0.04982878864662892_dp, &
             0.08412338360307972_dp, 0.1400164244118149_dp, 0.2300519591079407_dp, &
             0.3736595968476675_dp, 0.6011862346544015_dp, 0.9617600219211764_dp, &
             1.5431454519436787_dp, 2.542303828767266_dp, 4.692363952999028_dp]

      CASE (885)

         aw(:) = & ! a
            [3.7774678648482005e-07_dp, 2.4736741274966162e-06_dp, 8.889106086152047e-06_dp, &
             2.645955995169582e-05_dp, 7.055878390956706e-05_dp, 0.00017360501573076073_dp, &
             0.00040091829666080155_dp, 0.0008792012945017072_dp, 0.0018464100101047538_dp, &
             0.00373718127951489_dp, 0.007326353218524106_dp, 0.013965944349060873_dp, &
             0.025970410593154174_dp, 0.047234212273879084_dp, 0.08420924677686167_dp, &
             0.14743427869652181_dp, 0.2539048399756715_dp, 0.43072029620078806_dp, &
             0.7206940248780189_dp, 1.1911168284756521_dp, 1.9481863345109296_dp, &
             3.1640721870991877_dp, 5.142650229948148_dp, 8.571082168791298_dp, &
             ! w
             1.0167569611090138e-06_dp, 3.566831543882181e-06_dp, 1.0337803281361749e-05_dp, &
             2.7230318663633734e-05_dp, 6.611901304740002e-05_dp, 0.00015040557419417_dp, &
             0.0003245395042251765_dp, 0.0006703139684105778_dp, 0.0013342325720729195_dp, &
             0.0025726267479815514_dp, 0.004824827146280093_dp, 0.008830100585281596_dp, &
             0.015812063472740504_dp, 0.027766033422029608_dp, 0.04790200308733762_dp, &
             0.08132141759645975_dp, 0.136045354853115_dp, 0.22458017769344552_dp, &
             0.3663549105011692_dp, 0.5917824245359105_dp, 0.9501471066912243_dp, &
             1.529441431923682_dp, 2.526779629080599_dp, 4.674675165636894_dp]

      CASE (886)

         aw(:) = & ! a
            [2.9966407897336536e-07_dp, 2.016765294564868e-06_dp, 7.498124840720449e-06_dp, &
             2.2956278962220296e-05_dp, 6.25512438204073e-05_dp, 0.0001565168360911557_dp, &
             0.0003663881428105752_dp, 0.0008125263631454981_dp, 0.001722587190363661_dp, &
             0.003514953682145494_dp, 0.006939513652529784_dp, 0.013311010271228278_dp, &
             0.024889724812769225_dp, 0.04549364407386586_dp, 0.08147014396008116_dp, &
             0.14322042524415524_dp, 0.24756729363735042_dp, 0.4214065474802654_dp, &
             0.707333082751402_dp, 1.1724374904804709_dp, 1.922787539352113_dp, &
             3.1305508751701203_dp, 5.099741320746874_dp, 8.517593757798599_dp, &
             ! w
             8.118402379748681e-07_dp, 2.9767643746759905e-06_dp, 8.958773407829602e-06_dp, &
             2.4188732065655904e-05_dp, 5.9798319523643864e-05_dp, 0.00013794980257671275_dp, &
             0.0003010774357778089_dp, 0.000627802625559086_dp, 0.0012597845443616017_dp, &
             0.0024461671298049736_dp, 0.004615926724791454_dp, 0.008493873895452728_dp, &
             0.015284139459586966_dp, 0.02695686044632157_dp, 0.04669122323293095_dp, &
             0.07955384903484494_dp, 0.13353135931176155_dp, 0.22110480988315032_dp, &
             0.361701630907671_dp, 0.5857762060870976_dp, 0.9427135896538691_dp, &
             1.5206550628946045_dp, 2.516817619908326_dp, 4.6633255815492225_dp]

      CASE (887)

         aw(:) = & ! a
            [2.508770003629448e-07_dp, 1.729068468961415e-06_dp, 6.607740100941875e-06_dp, &
             2.0674684595674792e-05_dp, 5.7260600901523145e-05_dp, 0.00014509500912665858_dp, &
             0.00034308853971090245_dp, 0.000767180190633804_dp, 0.0016378076010595578_dp, &
             0.0033619147940743906_dp, 0.006671762625750891_dp, 0.012855668170377534_dp, &
             0.0241353830487588_dp, 0.04427434286100904_dp, 0.07954515300427452_dp, &
             0.14025033414148916_dp, 0.24308844289971918_dp, 0.414808411277222_dp, &
             0.6978469777919645_dp, 1.159149156469379_dp, 1.9046874825976505_dp, &
             3.106626816750737_dp, 5.069080437067063_dp, 8.479338721964988_dp, &
             ! w
             6.836563523641364e-07_dp, 2.6014659035089563e-06_dp, 8.059929137868434e-06_dp, &
             2.217141044768141e-05_dp, 5.5554101192827904e-05_dp, 0.00012950532288842745_dp, &
             0.00028504570826151426_dp, 0.0005985616762051011_dp, 0.0012082846777621118_dp, &
             0.002358253947292533_dp, 0.00447006649626699_dp, 0.008258196383373868_dp, &
             0.014912797502291792_dp, 0.026385886747062_dp, 0.04583440446545356_dp, &
             0.07829971929200558_dp, 0.13174331431614134_dp, 0.21862753274127927_dp, &
             0.358378061505983_dp, 0.5814786679503154_dp, 0.9373868478064913_dp, &
             1.5143519070265659_dp, 2.5096668658111927_dp, 4.6551795101428715_dp]

      CASE (888)

         aw(:) = & ! a
            [2.1728628866521742e-07_dp, 1.5296913194615717e-06_dp, 5.98223462019019e-06_dp, &
             1.9049793125336064e-05_dp, 5.3450681471443056e-05_dp, 0.00013679682464523029_dp, &
             0.0003260389387176323_dp, 0.0007337994323717608_dp, 0.0015750826592510598_dp, &
             0.003248194367929552_dp, 0.006472046026216785_dp, 0.01251488944111331_dp, &
             0.023569151083515167_dp, 0.04335665500288706_dp, 0.07809285122171793_dp, &
             0.13800467337591324_dp, 0.23969530945360687_dp, 0.4098007172174164_dp, &
             0.6906356729347475_dp, 1.1490325618861899_dp, 1.890889778769048_dp, &
             3.0883692710021853_dp, 5.045660764277268_dp, 8.450098856242718_dp, &
             ! w
             5.953215905020066e-07_dp, 2.339164841584035e-06_dp, 7.419366788383969e-06_dp, &
             2.0714540158103964e-05_dp, 5.246015082552893e-05_dp, 0.00012330449836582942_dp, &
             0.0002732034028538731_dp, 0.0005768540388194546_dp, 0.0011698891334093017_dp, &
             0.002292466919512161_dp, 0.004360559463745213_dp, 0.008080742612090732_dp, &
             0.014632465175912447_dp, 0.025953832193567106_dp, 0.04518466048878026_dp, &
             0.07734682134968444_dp, 0.13038230283575483_dp, 0.21673879561873047_dp, &
             0.3558403070138537_dp, 0.5781928825717653_dp, 0.9333096339380905_dp, &
             1.5095233424647316_dp, 2.504186584547245_dp, 4.6489368095774015_dp]

      CASE (889)

         aw(:) = & ! a
            [1.9264176595371223e-07_dp, 1.3825832395128835e-06_dp, 5.5153140488003255e-06_dp, &
             1.7823143088041275e-05_dp, 5.054859269604775e-05_dp, 0.00013043097252956172_dp, &
             0.0003128844221271794_dp, 0.0007079223298258678_dp, 0.0015262624924254136_dp, &
             0.003159378216683541_dp, 0.006315598415537916_dp, 0.012247235884980208_dp, &
             0.0231233786773664_dp, 0.042632676394778427_dp, 0.07694493696959968_dp, &
             0.13622663593188175_dp, 0.2370045571917654_dp, 0.4058239952180038_dp, &
             0.6849016600652033_dp, 1.1409791450984343_dp, 1.8798947902726888_dp, &
             3.07380771882253_dp, 5.026968967980853_dp, 8.426749577323196_dp, &
             ! w
             5.304678113603172e-07_dp, 2.1441982380002112e-06_dp, 6.935522600851609e-06_dp, &
             1.960233077880767e-05_dp, 5.0080402524234486e-05_dp, 0.00011850729898150323_dp, &
             0.0002639983610371424_dp, 0.0005599137777205965_dp, 0.001139824428678728_dp, &
             0.002240802569465352_dp, 0.004274338514400611_dp, 0.007940702969032808_dp, &
             0.01441078297259345_dp, 0.025611536253611074_dp, 0.04466903136518609_dp, &
             0.07658944767779852_dp, 0.1292990290313732_dp, 0.21523354513821583_dp, &
             0.3538154427413664_dp, 0.5755684418219095_dp, 0.930050217100013_dp, &
             1.5056607768670003_dp, 2.499801166835287_dp, 4.643941527524527_dp]

      CASE (890)

         aw(:) = & ! a
            [1.7373084862634115e-07_dp, 1.2691297575193673e-06_dp, 5.1515382360636816e-06_dp, &
             1.6858324267231058e-05_dp, 4.82488018910974e-05_dp, 0.00012535655887271248_dp, &
             0.0003023488508882721_dp, 0.0006871160477069907_dp, 0.0014868795933107414_dp, &
             0.0030875282832190255_dp, 0.006188725338760255_dp, 0.01202970985668649_dp, &
             0.022760397622308103_dp, 0.04204214690297607_dp, 0.07600716699089691_dp, &
             0.13477206454827015_dp, 0.23480051861067744_dp, 0.40256283760540196_dp, &
             0.680194494981714_dp, 1.1343617183819898_dp, 1.8708528023600781_dp, &
             3.0618242056306104_dp, 5.0115776480482594_dp, 8.407514949627538_dp, &
             ! w
             4.806743575133496e-07_dp, 1.9928502104087824e-06_dp, 6.554768435048832e-06_dp, &
             1.8719347323413663e-05_dp, 4.817938442300351e-05_dp, 0.00011465673013568091_dp, &
             0.00025658093172343534_dp, 0.0005462188254748389_dp, 0.0011154518054161202_dp, &
             0.002198818870109298_dp, 0.0042041252090034295_dp, 0.007826449014750245_dp, &
             0.014229615664935273_dp, 0.02533137472536298_dp, 0.04424641963525323_dp, &
             0.07596792128380592_dp, 0.12840903533674083_dp, 0.2139955628637744_dp, &
             0.35214851683483284_dp, 0.5734060974435716_dp, 0.9273627892049743_dp, &
             1.5024743621647667_dp, 2.4961824043033047_dp, 4.6398196737055555_dp]

      CASE (891)

         aw(:) = & ! a
            [1.587263654680947e-07_dp, 1.1787014995977406e-06_dp, 4.858967632000977e-06_dp, &
             1.6075951245502484e-05_dp, 4.637194117668958e-05_dp, 0.00012119463725898148_dp, &
             0.0002936731950986282_dp, 0.0006699263121040091_dp, 0.001454251875631771_dp, &
             0.003027860898446394_dp, 0.006083146632572001_dp, 0.011848364809338856_dp, &
             0.022457303678593613_dp, 0.041548336869757015_dp, 0.0752219725541138_dp, &
             0.1335527246419867_dp, 0.23295094888976645_dp, 0.3998235177027954_dp, &
             0.6762370903852104_dp, 1.1287939436407128_dp, 1.8632397688782183_dp, &
             3.0517285374352277_dp, 4.998604825529031_dp, 8.391296943389394_dp, &
             ! w
             4.4114835426100035e-07_dp, 1.8715082472192741e-06_dp, 6.245881402788989e-06_dp, &
             1.799765165662706e-05_dp, 4.661744283529861e-05_dp, 0.0001114801293754326_dp, &
             0.00025044164391512457_dp, 0.0005348526073269111_dp, 0.0010951762751999087_dp, &
             0.002163822016300428_dp, 0.004145492749612261_dp, 0.007730889707929212_dp, &
             0.014077878131712247_dp, 0.025096425967905383_dp, 0.043891600728433386_dp, &
             0.07544554743651996_dp, 0.12766030293670247_dp, 0.2129531582622325_dp, &
             0.35074381202803634_dp, 0.5715826161287677_dp, 0.9250951612461602_dp, &
             1.4997845033493375_dp, 2.4931268493041765_dp, 4.636339435798124_dp]

      CASE (892)

         aw(:) = & ! a
            [1.465093527715584e-07_dp, 1.104765369990518e-06_dp, 4.61781450163179e-06_dp, &
             1.5426407928811306e-05_dp, 4.480505746825221e-05_dp, 0.00011770507939170671_dp, &
             0.00028637398065797174_dp, 0.0006554227878941288_dp, 0.0014266572050182665_dp, &
             0.0029772946861361263_dp, 0.005993513620315467_dp, 0.011694169130008907_dp, &
             0.02219923087295149_dp, 0.04112735852385193_dp, 0.07455184399663503_dp, &
             0.13251102962603944_dp, 0.2313694058851084_dp, 0.3974792277873134_dp, &
             0.6728478405632048_dp, 1.1240223182008329_dp, 1.856711476234256_dp, &
             3.043066972808633_dp, 4.98747026753634_dp, 8.377372807978487_dp, &
             ! w
             4.089527360579679e-07_dp, 1.771766178634674e-06_dp, 5.989339334151199e-06_dp, &
             1.7394373224397768e-05_dp, 4.530586031890046e-05_dp, 0.00010880335423639561_dp, &
             0.00024525370455501315_dp, 0.0005252250685532874_dp, 0.0010779678226346802_dp, &
             0.002134067583029784_dp, 0.004095567387984674_dp, 0.007649411506065918_dp, &
             0.013948343834735324_dp, 0.024895638638926576_dp, 0.04358807360109462_dp, &
             0.07499828521684905_dp, 0.12701870298893356_dp, 0.21205923460732037_dp, &
             0.3495383727121509_dp, 0.5700168596564025_dp, 0.9231470428842837_dp, &
             1.4974727731915656_dp, 2.490500299376918_dp, 4.633347908597025_dp]

      CASE (893)

         aw(:) = & ! a
            [8.83557997605878e-08_dp, 7.476213250134589e-07_dp, 3.420943251523543e-06_dp, &
             1.2127504869383644e-05_dp, 3.670730785970545e-05_dp, 9.942693908157206e-05_dp, &
             0.0002477292775619158_dp, 0.0005779578063836622_dp, 0.0012781790562788173_dp, &
             0.002703491888580084_dp, 0.005505511046375136_dp, 0.010850615936809553_dp, &
             0.020781376406521064_dp, 0.038805683225496386_dp, 0.07084344830010995_dp, &
             0.12672857536342025_dp, 0.22256557731930715_dp, 0.3843961800833152_dp, &
             0.653889333621267_dp, 1.0972757549508916_dp, 1.8200511535755035_dp, &
             2.9943509949545093_dp, 4.9247661507819975_dp, 8.298885375184273_dp, &
             ! w
             2.555293260969674e-07_dp, 1.2810163275365013e-06_dp, 4.684831369496089e-06_dp, &
             1.4264538683883918e-05_dp, 3.840469132181553e-05_dp, 9.456489514034605e-05_dp, &
             0.0002174140317560235_dp, 0.00047318215464642603_dp, 0.0009843651188738444_dp, &
             0.001971351097648105_dp, 0.003821255399580891_dp, 0.007199864383142737_dp, &
             0.013230984500068333_dp, 0.02377994099873053_dp, 0.041896344862730604_dp, &
             0.07249850403177445_dp, 0.1234236275312518_dp, 0.2070386570876527_dp, &
             0.342753920314871_dp, 0.5611879619355967_dp, 0.9121448015943489_dp, &
             1.484401674574574_dp, 2.4756397828331553_dp, 4.616423772883867_dp]

      CASE (894)

         aw(:) = & ! a
            [6.73037114169863e-08_dp, 6.150839283045243e-07_dp, 2.9579920066772968e-06_dp, &
             1.0809320024687083e-05_dp, 3.3393911429821674e-05_dp, 9.181253556911713e-05_dp, &
             0.00023140128944992754_dp, 0.0005448492708841132_dp, 0.0012141085798527968_dp, &
             0.002584376070646481_dp, 0.0052917119047038936_dp, 0.010478768888306074_dp, &
             0.02015297090975294_dp, 0.037771710505760975_dp, 0.06918471366540907_dp, &
             0.12413199592129741_dp, 0.21859824461020386_dp, 0.37848151893692467_dp, &
             0.6452935409989419_dp, 1.0851172251044725_dp, 1.8033476977103298_dp, &
             2.9721111718644626_dp, 4.896095224989115_dp, 8.262955422511705_dp, &
             ! w
             1.999055905481422e-07_dp, 1.093467199972253e-06_dp, 4.1627088549732715e-06_dp, &
             1.2977597964210251e-05_dp, 3.551324730754912e-05_dp, 8.851309228664333e-05_dp, &
             0.00020544454619325942_dp, 0.0004505933008500562_dp, 0.0009434105470464506_dp, &
             0.0018996643448384969_dp, 0.0036996752462780985_dp, 0.006999557208772001_dp, &
             0.012909832266855032_dp, 0.023278331362169436_dp, 0.04113282583934989_dp, &
             0.07136633688524384_dp, 0.12179018414807075_dp, 0.2047508638810021_dp, &
             0.33965417285495386_dp, 0.5571446626274162_dp, 0.9070962569494883_dp, &
             1.4783949942011378_dp, 2.4688054050477204_dp, 4.6086410889496_dp]

      CASE (895)

         aw(:) = & ! a
            [5.637566905316404e-08_dp, 5.451412979973509e-07_dp, 2.7076988495492482e-06_dp, &
             1.0084001391821436e-05_dp, 3.154783868772073e-05_dp, 8.753036775597828e-05_dp, &
             0.00022215155841846787_dp, 0.0005259823855587696_dp, 0.001177418607738147_dp, &
             0.0025158803381232827_dp, 0.005168329134193652_dp, 0.010263505757182758_dp, &
             0.019788181537049913_dp, 0.03717001579327548_dp, 0.06821733135608407_dp, &
             0.12261465737440959_dp, 0.21627573734908737_dp, 0.3750134096418765_dp, &
             0.640245931716625_dp, 1.0779681129759702_dp, 1.7935148190838959_dp, &
             2.9590062450557983_dp, 4.879187275234374_dp, 8.241754080516271_dp, &
             ! w
             1.7100456010361993e-07_dp, 9.926971823345007e-07_dp, 3.875173981192119e-06_dp, &
             1.2258858785979523e-05_dp, 3.3882578450733986e-05_dp, 8.507468118673443e-05_dp, &
             0.00019860359066267353_dp, 0.0004376200824290833_dp, 0.0009197929818837754_dp, &
             0.0018581786910477557_dp, 0.0036291002929185476_dp, 0.00688296869424028_dp, &
             0.012722456858343813_dp, 0.022985037749146005_dp, 0.04068552296418643_dp, &
             0.0707018891871148_dp, 0.12082999746152481_dp, 0.20340404725924074_dp, &
             0.3378269224946423_dp, 0.5547583804229727_dp, 0.9041137292525309_dp, &
             1.4748437939726093_dp, 2.4647632473887784_dp, 4.604038280587225_dp]

      CASE (896)

         aw(:) = & ! a
            [4.9772733442355435e-08_dp, 5.023627036539797e-07_dp, 2.5521009941598543e-06_dp, &
             9.627965328776978e-06_dp, 3.0377916882720864e-05_dp, 8.480061372498682e-05_dp, &
             0.0002162280934113343_dp, 0.0005138554691296738_dp, 0.0011537633995521374_dp, &
             0.0024716044334467937_dp, 0.005088396256815142_dp, 0.010123777822512099_dp, &
             0.019550990554601085_dp, 0.03677819071667573_dp, 0.06758651152968595_dp, &
             0.12162400289281462_dp, 0.21475771435539226_dp, 0.3727443278690883_dp, &
             0.6369404310716952_dp, 1.0732826008768912_dp, 1.7870657505421097_dp, &
             2.9504058932368378_dp, 4.868085651331815_dp, 8.227828346148627_dp, &
             ! w
             1.5352793724805392e-07_dp, 9.302920906943039e-07_dp, 3.694296677365706e-06_dp, &
             1.1802709208025934e-05_dp, 3.28412887152216e-05_dp, 8.286877595237555e-05_dp, &
             0.0001941985153917856_dp, 0.00042924083999099956_dp, 0.0009044996597311365_dp, &
             0.0018312562516105324_dp, 0.003583213002283877_dp, 0.00680703685979662_dp, &
             0.012600241261229146_dp, 0.022793481565913754_dp, 0.04039302751764794_dp, &
             0.07026692412795217_dp, 0.12020080496789606_dp, 0.20252069802280734_dp, &
             0.3366274759254646_dp, 0.553190825239167_dp, 0.902153291225225_dp, &
             1.4725084903030499_dp, 2.4621044257781532_dp, 4.6010107637177535_dp]

      CASE (897)

         aw(:) = & ! a
            [4.548311535444276e-08_dp, 4.7431190303889844e-07_dp, 2.4488855777714307e-06_dp, &
             9.323083947427675e-06_dp, 2.959152889272304e-05_dp, 8.295840786605074e-05_dp, &
             0.00021221815277281344_dp, 0.0005056255123491113_dp, 0.0011376765280935734_dp, &
             0.0024414416916590386_dp, 0.005033860581514005_dp, 0.010028321149009319_dp, &
             0.019388764249011235_dp, 0.03650992838727325_dp, 0.06715422612198901_dp, &
             0.12094457155633342_dp, 0.21371581583113633_dp, 0.37118588707023387_dp, &
             0.6346687765700535_dp, 1.0700607932001203_dp, 1.7826291658394597_dp, &
             2.9444869212915545_dp, 4.860442713785682_dp, 8.21823878370653_dp, &
             ! w
             1.4216594295285535e-07_dp, 8.890016107028812e-07_dp, 3.573330747081158e-06_dp, &
             1.1495804458554209e-05_dp, 3.213775787908108e-05_dp, 8.13736727282973e-05_dp, &
             0.00019120537506522943_dp, 0.0004235356494682756_dp, 0.000894068906357723_dp, &
             0.0018128668017023798_dp, 0.0035518293712454943_dp, 0.006755046345475746_dp, &
             0.01251647640402811_dp, 0.022662073691906266_dp, 0.040192212333251924_dp, &
             0.06996807513591659_dp, 0.11976821864961937_dp, 0.20191299987072983_dp, &
             0.3358018608113742_dp, 0.5521112983035065_dp, 0.9008026385171825_dp, &
             1.47089907493926_dp, 2.460271746843561_dp, 4.598923989939878_dp]

      CASE (898)

         aw(:) = & ! a
            [4.263327905827938e-08_dp, 4.555442061369081e-07_dp, 2.3792518347357195e-06_dp, &
             9.116263349047226e-06_dp, 2.9056050443521292e-05_dp, 8.170048898751679e-05_dp, &
             0.0002094741174576538_dp, 0.0004999839009847217_dp, 0.0011266331853945625_dp, &
             0.0024207104010501605_dp, 0.004996338354509343_dp, 0.009962584423797492_dp, &
             0.01927695706724967_dp, 0.036324909521959395_dp, 0.06685589234985742_dp, &
             0.12047540676637648_dp, 0.21299598672344616_dp, 0.37010868482805165_dp, &
             0.6330979340995081_dp, 1.0678320789361795_dp, 1.779559095270237_dp, &
             2.9403898906362462_dp, 4.855151166518804_dp, 8.211598370792768_dp, &
             ! w
             1.3461276471971353e-07_dp, 8.611940074775547e-07_dp, 3.4912528480249925e-06_dp, &
             1.1286685298132144e-05_dp, 3.1656985894881816e-05_dp, 8.034971335483267e-05_dp, &
             0.00018915187167193505_dp, 0.0004196159018889861_dp, 0.0008868938798695513_dp, &
             0.0018002042472326056_dp, 0.003530200096345398_dp, 0.006719187013547815_dp, &
             0.012458661301896497_dp, 0.022571318613227937_dp, 0.040053444121103794_dp, &
             0.0697614577762599_dp, 0.11946899912327133_dp, 0.20149247709176296_dp, &
             0.3352303212847525_dp, 0.5513637313713607_dp, 0.8998670503124884_dp, &
             1.4697840051404651_dp, 2.4590018441782022_dp, 4.597478038703463_dp]

      CASE (899)

         aw(:) = & ! a
            [4.080271175927372e-08_dp, 4.434269286640962e-07_dp, 2.334031152203996e-06_dp, &
             8.981441874797085e-06_dp, 2.8706075007825823e-05_dp, 8.087676976506513e-05_dp, &
             0.00020767458980701272_dp, 0.0004962797607419927_dp, 0.0011193752924824542_dp, &
             0.002407074122647782_dp, 0.004971640066457287_dp, 0.00991928770960574_dp, &
             0.019203276510454535_dp, 0.03620292378667997_dp, 0.0666591113420866_dp, &
             0.12016582518865584_dp, 0.21252083559855292_dp, 0.36939740939117655_dp, &
             0.6320604102806445_dp, 1.0663596587075976_dp, 1.7775303649550438_dp, &
             2.937682012304261_dp, 4.8516532430587_dp, 8.207208284855648_dp, &
             ! w
             1.297585858502892e-07_dp, 8.431569872572622e-07_dp, 3.4377399975620104e-06_dp, &
             1.1149949755047355e-05_dp, 3.134199480869196e-05_dp, 7.967782501106015e-05_dp, &
             0.0001878028258622622_dp, 0.00041703832133176957_dp, 0.0008821718075183771_dp, &
             0.0017918648608192483_dp, 0.0035159467081084276_dp, 0.006695543630437011_dp, &
             0.012420523638627168_dp, 0.022511426760790414_dp, 0.039961831999048444_dp, &
             0.06962500553936993_dp, 0.11927132882455928_dp, 0.20121459114557888_dp, &
             0.3348525429620856_dp, 0.5508694871650516_dp, 0.899248377573688_dp, &
             1.469046539929834_dp, 2.4581619122879825_dp, 4.596521674378948_dp]

      CASE (900)

         aw(:) = & ! a
            [3.9797259054851646e-08_dp, 4.367491128819479e-07_dp, 2.3090182209582046e-06_dp, &
             8.906690655073264e-06_dp, 2.851171753567362e-05_dp, 8.041877569823685e-05_dp, &
             0.00020667311752804656_dp, 0.0004942168102835682_dp, 0.0011153306788409163_dp, &
             0.0023994711022991416_dp, 0.0049578632395921715_dp, 0.009895127299066433_dp, &
             0.019162147413299226_dp, 0.036134809886564014_dp, 0.06654920404127064_dp, &
             0.11999287404461234_dp, 0.2122553292353876_dp, 0.3689998821842513_dp, &
             0.6314804412918881_dp, 1.065536453465059_dp, 1.7763959759375296_dp, &
             2.936167687364232_dp, 4.8496969121493505_dp, 8.204752805376222_dp, &
             ! w
             1.2709145236350234e-07_dp, 8.331874358562598e-07_dp, 3.408067202210578e-06_dp, &
             1.1073993365466279e-05_dp, 3.116679950719382e-05_dp, 7.93037753836723e-05_dp, &
             0.0001870512355599003_dp, 0.0004156014090353547_dp, 0.0008795380745097793_dp, &
             0.0017872115492601082_dp, 0.0035079904326112306_dp, 0.006682341468459542_dp, &
             0.012399221796251825_dp, 0.022477965276555635_dp, 0.03991063628242366_dp, &
             0.06954873530458963_dp, 0.11916081888306569_dp, 0.20105920773844818_dp, &
             0.334641268995126_dp, 0.5505930392537741_dp, 0.898902290488593_dp, &
             1.4686339627311775_dp, 2.457691986856393_dp, 4.595986610299396_dp]

      CASE (901)

         aw(:) = & ! a
            [3.9553391093700604e-08_dp, 4.3512701604117224e-07_dp, 2.302932448530375e-06_dp, &
             8.888484166138473e-06_dp, 2.8464346060959095e-05_dp, 8.030708924025138e-05_dp, &
             0.00020642879968401902_dp, 0.0004937133696804489_dp, 0.0011143433683227335_dp, &
             0.0023976147411339946_dp, 0.004954498813196079_dp, 0.009889226135376591_dp, &
             0.019152100287430633_dp, 0.03611816885552533_dp, 0.06652234941493979_dp, &
             0.11995061108518992_dp, 0.21219044329517284_dp, 0.36890272417691433_dp, &
             0.6313386823977204_dp, 1.065335227969929_dp, 1.7761186703751048_dp, &
             2.9357974929020836_dp, 4.849218651837246_dp, 8.204152504629592_dp, &
             ! w
             1.2644444670913428e-07_dp, 8.307625908807098e-07_dp, 3.4008397351421422e-06_dp, &
             1.1055477945937366e-05_dp, 3.112407025450718e-05_dp, 7.92125092015585e-05_dp, &
             0.00018686778963648376_dp, 0.0004152505968138721_dp, 0.000878894923058918_dp, &
             0.001786074996990426_dp, 0.0035060468295567814_dp, 0.006679115968754238_dp, &
             0.01239401686072571_dp, 0.022469788335589352_dp, 0.03989812431819047_dp, &
             0.06953009360685869_dp, 0.11913380626768146_dp, 0.20102122327344038_dp, &
             0.33458961786871366_dp, 0.5505254519244984_dp, 0.8988176770561613_dp, &
             1.468533093256879_dp, 2.4575770930727106_dp, 4.595855787712363_dp]

      CASE (902)

         aw(:) = & ! a
            [0.001264792186980667_dp, 0.006719746651533019_dp, 0.016768017908939923_dp, &
             0.031844448136599694_dp, 0.0526472378282337_dp, 0.08022815891214524_dp, &
             0.11611779125345187_dp, 0.16248635442730888_dp, 0.22233443343108342_dp, &
             0.29971065725666435_dp, 0.3999670587299389_dp, 0.5300785700219749_dp, &
             0.6990609349677402_dp, 0.9185230328881094_dp, 1.2033945239351396_dp, &
             1.57288547941923_dp, 2.05176801026866_dp, 2.6721351455965645_dp, &
             3.475925118934264_dp, 4.518789245187692_dp, 5.876568283988842_dp, &
             7.657453476949997_dp, 10.028500579165467_dp, 13.287269360867699_dp, &
             18.14623369152218_dp, &
             ! w
             0.003252145480014078_dp, 0.0076975105222907006_dp, 0.012470420302847464_dp, &
             0.01779733205666143_dp, 0.023981454966389173_dp, 0.03143278617471357_dp, &
             0.04070499044555699_dp, 0.052528630015347126_dp, 0.06783811086572537_dp, &
             0.08780110197318536_dp, 0.11386730855277_dp, 0.1478492255578618_dp, &
             0.19203831857080877_dp, 0.24935836314489407_dp, 0.32356528303352233_dp, &
             0.41951663072333456_dp, 0.5435564428716237_dp, 0.7041057265580195_dp, &
             0.9126477303176014_dp, 1.1855344800342673_dp, 1.5476658838496187_dp, &
             2.040972683230338_dp, 2.7475175625354233_dp, 3.871228739305233_dp, &
             6.223067895896287_dp]

      CASE (903)

         aw(:) = & ! a
            [0.0009019824790200289_dp, 0.004798124254710645_dp, 0.012000746348039203_dp, &
             0.022871751440315988_dp, 0.03800119549154519_dp, 0.05829287604467917_dp, &
             0.0850825523222433_dp, 0.12028681558582045_dp, 0.16657694336046427_dp, &
             0.22758018137914815_dp, 0.308127055853057_dp, 0.4145747903547037_dp, &
             0.5552391756094838_dp, 0.7409680922083405_dp, 0.9858974236592485_dp, &
             1.3084480812745232_dp, 1.7326562975658777_dp, 2.28999239287103_dp, &
             3.021951489169225_dp, 3.9839817826240718_dp, 5.2519896863678_dp, &
             6.934448454158269_dp, 9.19866879249285_dp, 12.341660976789266_dp, &
             17.071365106394726_dp, &
             ! w
             0.0023199217015820485_dp, 0.005504737687714373_dp, 0.008959830143646782_dp, &
             0.012879116787825596_dp, 0.017528426762849436_dp, 0.023275060735463298_dp, &
             0.03062082762942329_dp, 0.04022960422654227_dp, 0.052951146875945564_dp, &
             0.06985510396167391_dp, 0.09229106033536737_dp, 0.1219809080285301_dp, &
             0.1611433783359864_dp, 0.21265402736794287_dp, 0.2802529626775681_dp, &
             0.368824700890271_dp, 0.4847948235861328_dp, 0.6367303498539257_dp, &
             0.8363269121288506_dp, 1.100199228179841_dp, 1.453509336286211_dp, &
             1.9383294376081197_dp, 2.6364512391097206_dp, 3.750184287401272_dp, &
             6.082028458308234_dp]

      CASE (904)

         aw(:) = & ! a
            [0.0007085512156654903_dp, 0.0037727899369717727_dp, 0.00945329942157289_dp, &
             0.01806669347522852_dp, 0.03013508634097651_dp, 0.04646782774258337_dp, &
             0.068273597307657_dp, 0.0972994741805902_dp, 0.13599257725520975_dp, &
             0.1876913601172234_dp, 0.2568693114902722_dp, 0.34946123924173056_dp, &
             0.4733023835428611_dp, 0.6387127549098096_dp, 0.8592688031239546_dp, &
             1.1528235835592535_dp, 1.5428697612523479_dp, 2.060401317965205_dp, &
             2.7465549468899155_dp, 3.6565886495831417_dp, 4.8664192619846345_dp, &
             6.484710858886317_dp, 8.678997458851105_dp, 11.746043311320129_dp, &
             16.391005542660675_dp, &
             ! w
             0.001822815327019953_dp, 0.0043335625369764685_dp, 0.007079293766997166_dp, &
             0.010233166183811936_dp, 0.014036631513600822_dp, 0.018827963291972654_dp, &
             0.025072076341200674_dp, 0.03338440128882293_dp, 0.04455422758869428_dp, &
             0.059583552116233604_dp, 0.07975424053254619_dp, 0.1067258461620191_dp, &
             0.14266372370616048_dp, 0.19040297254662666_dp, 0.2536622653768278_dp, &
             0.3373323285973978_dp, 0.4478828664017129_dp, 0.5939726136047219_dp, &
             0.7874413987581261_dp, 1.0450934815760544_dp, 1.392294376947921_dp, &
             1.8712540835771283_dp, 2.563641274471628_dp, 3.670766294275969_dp, &
             5.989685252015426_dp]

      CASE (905)

         aw(:) = & ! a
            [0.0005871234347755582_dp, 0.0031287200652076437_dp, 0.0078512444963358_dp, &
             0.015039656424595605_dp, 0.025168156255180565_dp, 0.038978665802516604_dp, &
             0.05758766442879519_dp, 0.08261725534928242_dp, 0.11634769905813049_dp, &
             0.1619020956225047_dp, 0.22348801795368534_dp, 0.3067253573068742_dp, &
             0.41908918122991323_dp, 0.5705001351482432_dp, 0.7741061157332458_dp, &
             1.0473185683450876_dp, 1.4131997008186337_dp, 1.902357383842213_dp, &
             2.5556273324806322_dp, 3.428094379825971_dp, 4.595661373412782_dp, &
             6.167131087440881_dp, 8.310227643324977_dp, 11.321593639412928_dp, &
             15.904445742481778_dp, &
             ! w
             0.0015107073128955751_dp, 0.0035973055122298626_dp, 0.0058943513015849635_dp, &
             0.008560208866273868_dp, 0.011818649758528189_dp, 0.015986251841427174_dp, &
             0.021499578975118144_dp, 0.028936632844159926_dp, 0.03904081582912841_dp, &
             0.05276384156990716_dp, 0.07133734501495034_dp, 0.09637343934591731_dp, &
             0.12999476767136203_dp, 0.17500140425029448_dp, 0.23508969587799808_dp, &
             0.31514794383224387_dp, 0.42167211672068877_dp, 0.5633847121589651_dp, &
             0.7522324969102645_dp, 1.0051667327810028_dp, 1.347718400914871_dp, &
             1.8222228967949816_dp, 2.5102890093994676_dp, 3.6125290036140374_dp, &
             5.922063719269106_dp]

      CASE (906)

         aw(:) = & ! a
            [0.0005033005162814257_dp, 0.0026838748935644904_dp, 0.0067436604480959695_dp, &
             0.012943878721158942_dp, 0.021722534424108673_dp, 0.03377009549375688_dp, &
             0.050131782998313765_dp, 0.07233200735790644_dp, 0.10251988216546759_dp, &
             0.14364921181216436_dp, 0.19971865085104373_dp, 0.27610017043062096_dp, &
             0.37998374875131485_dp, 0.5209713497311376_dp, 0.7118654647677526_dp, &
             0.9697174563253291_dp, 1.317233727837776_dp, 1.784697485333644_dp, &
             2.41268492469926_dp, 3.2561234205708383_dp, 4.390889818710778_dp, &
             5.925892769488113_dp, 8.029019820702668_dp, 10.996853858103664_dp, &
             15.531155632547962_dp, &
             ! w
             0.00129522910762061_dp, 0.003088455134587591_dp, 0.005073804394134264_dp, &
             0.007398379256116653_dp, 0.010272289671075673_dp, 0.0139949534766784_dp, &
             0.01898009812868718_dp, 0.025775603126840345_dp, 0.03508851237777908_dp, &
             0.0478311785748681_dp, 0.06519569243079291_dp, 0.0887558818423784_dp, &
             0.12059861057074878_dp, 0.16349309148898908_dp, 0.22111395054698038_dp, &
             0.29834301748885456_dp, 0.4016926345538529_dp, 0.5399324516461382_dp, &
             0.7250929585393414_dp, 0.9742447330088337_dp, 1.313057863089418_dp, &
             1.7839809406018379_dp, 2.468595184950132_dp, 3.5669890239886457_dp, &
             5.869240907099177_dp]

      CASE (907)

         aw(:) = & ! a
            [0.0004417080509811216_dp, 0.0023568550576626873_dp, 0.005928750741980419_dp, &
             0.011399959195168034_dp, 0.019179858286514417_dp, 0.029917869648884148_dp, &
             0.04460177540868037_dp, 0.06467657455632461_dp, 0.0921844276675197_dp, &
             0.12994111778154407_dp, 0.1817749780205105_dp, 0.2528553526857663_dp, &
             0.35013801637260483_dp, 0.4829612992348114_dp, 0.663839589671581_dp, &
             0.9095210876064569_dp, 1.2424093852661684_dp, 1.6925076615145789_dp, &
             2.3001639500572266_dp, 3.1201617914958137_dp, 4.228345829166758_dp, &
             5.733708343533177_dp, 7.80427923143218_dp, 10.736615909318328_dp, &
             15.231332635232285_dp, &
             ! w
             0.001136880339581742_dp, 0.002714170371144598_dp, 0.004469230300280583_dp, &
             0.006540200735350515_dp, 0.00912617932193502_dp, 0.012512488697308382_dp, &
             0.017093863177163406_dp, 0.02339322113866336_dp, 0.03208786436364899_dp, &
             0.044058159386898936_dp, 0.060463772641010945_dp, 0.08284644560289782_dp, &
             0.11326219989230159_dp, 0.1544526957921216_dp, 0.2100719158630194_dp, &
             0.2849932222053517_dp, 0.3857393193911573_dp, 0.5211162173285876_dp, &
             0.7032224241921012_dp, 0.9492284420468897_dp, 1.2849241319990852_dp, &
             1.7528605803247912_dp, 2.4346096953010528_dp, 3.529847802726062_dp, &
             5.82619538482212_dp]

      CASE (908)

         aw(:) = & ! a
            [0.0003944011783031881_dp, 0.00210558048336692_dp, 0.005302121761635809_dp, &
             0.010211423512976926_dp, 0.01721946680329371_dp, 0.026941872657959182_dp, &
             0.040318709919254125_dp, 0.05872858774632043_dp, 0.08412407363671993_dp, &
             0.11920538010035349_dp, 0.16765816365395186_dp, 0.23448156986977955_dp, &
             0.32643371560778855_dp, 0.4526290798252106_dp, 0.6253357171540862_dp, &
             0.8610405991982784_dp, 1.1818840656874818_dp, 1.617623187059093_dp, &
             2.2084023780994877_dp, 3.0088729888040646_dp, 4.09484425855509_dp, &
             5.5753762481850355_dp, 7.618624491863564_dp, 10.521141381041872_dp, &
             14.98260707029159_dp, &
             ! w
             0.0010152470094515799_dp, 0.002426433465236499_dp, 0.004003756686968509_dp, &
             0.005877995462771927_dp, 0.008239097751466277_dp, 0.011360490710259425_dp, &
             0.01562073925801767_dp, 0.021521615636505727_dp, 0.029715489471120807_dp, &
             0.04105599392776666_dp, 0.05667541918092883_dp, 0.07808790119445315_dp, &
             0.10732234149479694_dp, 0.1470955382787312_dp, 0.20104201730126_dp, &
             0.2740256943249734_dp, 0.3725757391003214_dp, 0.5055270532595258_dp, &
             0.6850349048732122_dp, 0.9283554242631128_dp, 1.2613835033327732_dp, &
             1.7267636377166722_dp, 2.4060695004391017_dp, 3.4986422065853766_dp, &
             5.790053777216295_dp]

      CASE (909)

         aw(:) = & ! a
            [0.0003568445960538921_dp, 0.001906021750027994_dp, 0.004804123843670017_dp, &
             0.009265906322241734_dp, 0.01565774957593582_dp, 0.024566768283189974_dp, &
             0.03689249143951277_dp, 0.053956836496252415_dp, 0.07763577237950191_dp, &
             0.11053074040323893_dp, 0.1562053713199403_dp, 0.21951284606823127_dp, &
             0.30704102720951854_dp, 0.427710544376966_dp, 0.5935748537060425_dp, &
             0.8208918460215603_dp, 1.13156926952956_dp, 1.5551447630636983_dp, &
             2.131578889002727_dp, 2.9154008069773427_dp, 3.9823828315454435_dp, &
             5.441641058802126_dp, 7.461443156731813_dp, 10.338349251707424_dp, &
             14.771258855040374_dp, &
             ! w
             0.0009186749228765929_dp, 0.002197812921295569_dp, 0.0036334141481814044_dp, &
             0.005350063262318096_dp, 0.00752993104010816_dp, 0.010436199123805083_dp, &
             0.014433420012780513_dp, 0.0200051392912658_dp, 0.027782419513356697_dp, &
             0.038596060714744626_dp, 0.053554727414258006_dp, 0.074148345440924_dp, &
             0.10238161546834482_dp, 0.14094870250362668_dp, 0.19346583670300996_dp, &
             0.2647870529449249_dp, 0.36144534470564016_dp, 0.492299101720872_dp, &
             0.6695519380865473_dp, 0.910534663228173_dp, 1.2412355861346278_dp, &
             1.7043848762493803_dp, 2.381564965363442_dp, 3.471837410652139_dp, &
             5.759027195564568_dp]

      CASE (910)

         aw(:) = & ! a
            [0.0003262535595199603_dp, 0.0017434199537284323_dp, 0.004398099842318326_dp, &
             0.008494298409170108_dp, 0.014381655852945128_dp, 0.022622793807727856_dp, &
             0.03408218435149266_dp, 0.050032524995280984_dp, 0.07228323247045655_dp, &
             0.1033499551626597_dp, 0.1466902323003018_dp, 0.2070299965141243_dp, &
             0.2908080924552534_dp, 0.406774672911357_dp, 0.5667934798265516_dp, &
             0.7869187233713909_dp, 1.0888500069802247_dp, 1.5019270616584413_dp, &
             2.0659428465931056_dp, 2.8353129484242574_dp, 3.8857721810642007_dp, &
             5.326483675070013_dp, 7.325816120062528_dp, 10.18034555502146_dp, &
             14.588305981135823_dp, &
             ! w
             0.0008400077217026571_dp, 0.0020114539866088552_dp, 0.0033311567594055294_dp, &
             0.004918390568617975_dp, 0.006948597149564836_dp, 0.009675992360220152_dp, &
             0.013452813534252008_dp, 0.01874667507622429_dp, 0.02617014374857054_dp, &
             0.0365341784293806_dp, 0.05092670607027173_dp, 0.07081610661771791_dp, &
             0.09818523679512947_dp, 0.13570750377963123_dp, 0.18698192797385246_dp, &
             0.2568525291351039_dp, 0.3518542662263154_dp, 0.48086496152888847_dp, &
             0.656130073792686_dp, 0.8950464506676085_dp, 1.2236864381239394_dp, &
             1.6848592606126291_dp, 2.3601606496817835_dp, 3.448414412325281_dp, &
             5.731928513162785_dp]

      CASE (911)

         aw(:) = & ! a
            [0.000180504034187792_dp, 0.0009677909232647818_dp, 0.002457067830655028_dp, &
             0.004793386594545536_dp, 0.008233109949766969_dp, 0.013199466782144407_dp, &
             0.020353344898168105_dp, 0.030678758517746343_dp, 0.045594545452208084_dp, &
             0.0671126972368621_dp, 0.09806431502621629_dp, 0.1424154516907746_dp, &
             0.20570194812171275_dp, 0.2956238088012079_dp, 0.4228553011226151_dp, &
             0.6021489503388819_dp, 0.8538453320227214_dp, 1.205958439293759_dp, &
             1.6971190764874222_dp, 2.380906744587034_dp, 3.3327097985491454_dp, &
             4.661924210433044_dp, 6.537583830945129_dp, 9.256546322555412_dp, &
             13.513357840422676_dp, &
             ! w
             0.0004650985558755235_dp, 0.001121199495154307_dp, 0.0018809030351167075_dp, &
             0.002833526802666685_dp, 0.004115270502048328_dp, 0.005926196770047863_dp, &
             0.008543753564650484_dp, 0.012340956696035143_dp, 0.01782223348030816_dp, &
             0.02568072389205102_dp, 0.03687636827959418_dp, 0.05273877038388262_dp, &
             0.0751042367877035_dp, 0.10650038716651011_dp, 0.15039657389883376_dp, &
             0.2115466914021172_dp, 0.29646715921927974_dp, 0.4141277064806729_dp, &
             0.5770147946970628_dp, 0.8029372794911709_dp, 1.118523719568458_dp, &
             1.5671495268263922_dp, 2.2306121771715315_dp, 3.306436300834295_dp, &
             5.567937308843291_dp]

      CASE (912)

         aw(:) = & ! a
            [0.00012748242557475265_dp, 0.0006850527494782257_dp, 0.0017468339770502412_dp, &
             0.003431466926960379_dp, 0.0059524621630289605_dp, 0.009666989141616962_dp, &
             0.015137072366623847_dp, 0.023205423454124772_dp, 0.03509977712727163_dp, &
             0.05258469302395082_dp, 0.07817965599693827_dp, 0.11546571121454734_dp, &
             0.16951138907591248_dp, 0.2474609338166582_dp, 0.3593442669486373_dp, &
             0.5191911379339048_dp, 0.7465667188905931_dp, 1.068704231242412_dp, &
             1.5235213888504655_dp, 2.164049766002809_dp, 3.0654021846471866_dp, &
             4.337047808735618_dp, 6.148396389496952_dp, 8.7965806917738_dp, &
             12.97446814102286_dp, &
             ! w
             0.0003286476870063008_dp, 0.0007958627181726459_dp, 0.0013469131692074481_dp, &
             0.0020571136160102803_dp, 0.0030433861619813497_dp, 0.004478112460201092_dp, &
             0.0066008146679634614_dp, 0.009738022407722997_dp, 0.014341194024062346_dp, &
             0.021043513267904906_dp, 0.030736255679443006_dp, 0.044670945053259144_dp, &
             0.06459781068431443_dp, 0.09295463409942827_dp, 0.13312513044835475_dp, &
             0.18979456872590492_dp, 0.2694463340509089_dp, 0.3810768384968753_dp, &
             0.5372855064162958_dp, 0.7561005957214038_dp, 1.0644716593771493_dp, &
             1.506131941775474_dp, 2.1630745122045925_dp, 3.2322525293459914_dp, &
             5.4824309616088165_dp]

      CASE (913)

         aw(:) = & ! a
            [9.955390749628301e-05_dp, 0.0005359138839270567_dp, 0.0013712169422480686_dp, &
             0.0027083199385988_dp, 0.004734686293982663_dp, 0.007766568196366141_dp, &
             0.012304068333119863_dp, 0.019101261186273897_dp, 0.029265946400809005_dp, &
             0.044406537963453376_dp, 0.06684386833669606_dp, 0.09991049648239901_dp, &
             0.1483693879065867_dp, 0.218996526744258_dp, 0.32138902710012773_dp, &
             0.4690841703677293_dp, 0.6811105307080174_dp, 0.9841512807395667_dp, &
             1.4156103270278033_dp, 2.0281107035520822_dp, 2.896541558358517_dp, &
             4.130396263161919_dp, 5.8993400355237124_dp, 8.500736158455918_dp, &
             12.626440449344825_dp, &
             ! w
             0.00025675042802745477_dp, 0.000623956515071796_dp, 0.0010632771032706336_dp, &
             0.0016414252161794363_dp, 0.0024631061065698897_dp, 0.003682879416989593_dp, &
             0.005516179860252556_dp, 0.008260430709806157_dp, 0.01233353074178115_dp, &
             0.01832954402112065_dp, 0.027093669693146402_dp, 0.039823847996241696_dp, &
             0.05820999607973761_dp, 0.08462543028283542_dp, 0.12239034301438824_dp, &
             0.17613596669689047_dp, 0.2523139947275561_dp, 0.35992885090826865_dp, &
             0.5116477957889787_dp, 0.7256438403125711_dp, 1.0290892961832774_dp, &
             1.4659783509963051_dp, 2.1184706996036184_dp, 3.1831868209091585_dp, &
             5.425942401291094_dp]

      CASE (914)

         aw(:) = & ! a
            [8.215807403170786e-05_dp, 0.000442916595190489_dp, 0.001136510726862081_dp, &
             0.0022550228849497565_dp, 0.003967903510412303_dp, 0.006562734133348864_dp, &
             0.010495962956598773_dp, 0.016459162257057024_dp, 0.025475446392116217_dp, &
             0.03904237926444667_dp, 0.059338613769440315_dp, 0.08951722749355775_dp, &
             0.13411840461361346_dp, 0.1996471072194891_dp, 0.2953788694181808_dp, &
             0.434481401989287_dp, 0.6355762302772984_dp, 0.9249241228448893_dp, &
             1.3395283561287001_dp, 1.9316861809734545_dp, 2.7760989845340425_dp, &
             3.9822636487988174_dp, 5.720035712729033_dp, 8.286973026241188_dp, &
             12.374237998029473_dp, &
             ! w
             0.0002119563365174039_dp, 0.0005166159869332959_dp, 0.0008854373156153062_dp, &
             0.0013791388260692866_dp, 0.0020937195766920504_dp, 0.0031709519917747994_dp, &
             0.004809199260521118_dp, 0.007285375173981507_dp, 0.010993396355784456_dp, &
             0.016498765841585675_dp, 0.024612445032308566_dp, 0.0364919995762332_dp, &
             0.053781289501141485_dp, 0.07880363228033524_dp, 0.11482890342125626_dp, &
             0.16644398360248885_dp, 0.2400718981074724_dp, 0.34471756529244935_dp, &
             0.4930940621929616_dp, 0.7034801866821311_dp, 1.003217276757435_dp, &
             1.436504337343956_dp, 2.0856438400662087_dp, 3.147035856880362_dp, &
             5.384355430178487_dp]

      CASE (915)

         aw(:) = & ! a
            [7.021686258873252e-05_dp, 0.0003790200863573012_dp, 0.0009749682115072822_dp, &
             0.0019421944995459464_dp, 0.003436718189804632_dp, 0.005724532076937601_dp, &
             0.009229092137382197_dp, 0.01459473604792678_dp, 0.022780464801540973_dp, &
             0.03519950199269977_dp, 0.0539215934422907_dp, 0.08196134134582536_dp, &
             0.12368579889839859_dp, 0.1853877831707111_dp, 0.2760892627216751_dp, &
             0.40866432004628184_dp, 0.6014084191549361_dp, 0.880240819769707_dp, &
             1.2818371864757045_dp, 1.8582239487895156_dp, 2.6839407484972315_dp, &
             3.868477812270853_dp, 5.581840581871271_dp, 8.121754189657269_dp, &
             12.17886862474053_dp, &
             ! w
             0.0001812012633421688_dp, 0.0004427802336149792_dp, 0.0007626824708471096_dp, &
             0.0011971290226881104_dp, 0.001835475089161348_dp, 0.002809710643856534_dp, &
             0.00430527166119313_dp, 0.006583561761706038_dp, 0.010020115577166918_dp, &
             0.015158188785670825_dp, 0.02278174339558203_dp, 0.03401618988560197_dp, &
             0.050468346925082015_dp, 0.0744208606308063_dp, 0.10910205213131634_dp, &
             0.15906126553322625_dp, 0.23069575478300078_dp, 0.33300740031422793_dp, &
             0.4787424522310096_dp, 0.6862618646184032_dp, 0.9830422725672653_dp, &
             1.4134508792715863_dp, 2.0599144949988957_dp, 3.118675692803574_dp, &
             5.351750118071792_dp]

      CASE (916)

         aw(:) = & ! a
            [6.148028464317523e-05_dp, 0.00033223361925589014_dp, 0.0008565041655540711_dp, &
             0.0017122522792201729_dp, 0.003044973414624125_dp, 0.005103615118848573_dp, &
             0.008285524083416975_dp, 0.01319765852571231_dp, 0.02074818140075173_dp, &
             0.032283136854842416_dp, 0.049785028325928925_dp, 0.07615686947469912_dp, &
             0.11562537536868181_dp, 0.17431039435158174_dp, 0.26102593262051577_dp, &
             0.38840377671798154_dp, 0.5744685620188198_dp, 0.844853786959293_dp, &
             1.2359585054786772_dp, 1.7995775399889744_dp, 2.6101084615600976_dp, &
             3.7770294393517774_dp, 5.470468492797603_dp, 7.98829702894451_dp, &
             12.020767113719403_dp, &
             ! w
             0.00015869568265594475_dp, 0.0003886624475751539_dp, 0.0006724378555591735_dp, &
             0.0010627006799714058_dp, 0.001643502496417106_dp, 0.002539022751823654_dp, &
             0.003924461747988385_dp, 0.006048927640241453_dp, 0.009273204159843512_dp, &
             0.014122468734473325_dp, 0.021358540255837232_dp, 0.03208025681664322_dp, &
             0.04786359147467303_dp, 0.07095701916061443_dp, 0.10455353819661095_dp, &
             0.1531699951474543_dp, 0.22318037347626207_dp, 0.3235817256226352_dp, &
             0.46714546010899227_dp, 0.6722989795002908_dp, 0.9666311605954334_dp, &
             1.3946516333613748_dp, 2.0388971282442068_dp, 3.095491916157144_dp, &
             5.325108518389744_dp]

      CASE (917)

         aw(:) = & ! a
            [5.479313180147018e-05_dp, 0.00029639667003451424_dp, 0.000765642361268275_dp, &
             0.0015355194375918956_dp, 0.0027429825108886735_dp, 0.004623060530367564_dp, &
             0.007551747850685785_dp, 0.012105431466630439_dp, 0.019150609287267164_dp, &
             0.02997804151925011_dp, 0.04649808399619563_dp, 0.07152100484783801_dp, &
             0.10915630129096324_dp, 0.16537864168032657_dp, 0.2488266191663795_dp, &
             0.37192660408869965_dp, 0.5524723719439752_dp, 0.8158523396706655_dp, &
             1.1982265657121662_dp, 1.7511876687098333_dp, 2.549006543630126_dp, &
             3.7011463253498205_dp, 5.37783801889113_dp, 7.877082757123006_dp, &
             11.888812505829732_dp, &
             ! w
             0.00014146662501784236_dp, 0.0003471734807843872_dp, 0.0006030668066653917_dp, &
             0.000958937532686493_dp, 0.0014944660094504613_dp, 0.0023274014491732964_dp, &
             0.0036245741707887974_dp, 0.005625011498626868_dp, 0.008677270896814525_dp, &
             0.013291393984229799_dp, 0.020210525195289996_dp, 0.03051095506798608_dp, &
             0.045742318511552436_dp, 0.06812369995285124_dp, 0.10081742950258604_dp, &
             0.1483117280118188_dp, 0.2169594244414413_dp, 0.3157518120202655_dp, &
             0.45748002004462734_dp, 0.6606267834863961_dp, 0.9528765192769223_dp, &
             1.3788621749647323_dp, 2.021218892451167_dp, 3.0759788788493503_dp, &
             5.302693748973033_dp]

      CASE (918)

         aw(:) = & ! a
            [4.9499305956758506e-05_dp, 0.00026800832441533794_dp, 0.0006935783426909829_dp, &
             0.001395084929017318_dp, 0.002502366702353727_dp, 0.004238797545465045_dp, &
             0.00696247384781783_dp, 0.011224152795523258_dp, 0.017855328690749968_dp, &
             0.02810014301069517_dp, 0.043807849872026156_dp, 0.0677098299052879_dp, &
             0.10381545286687396_dp, 0.157974853630235_dp, 0.23867552253285154_dp, &
             0.35816611016937044_dp, 0.5340397295755684_dp, 0.7914706652789757_dp, &
             1.1664088503127215_dp, 1.710267569806086_dp, 2.497203415964356_dp, &
             3.6366628444969025_dp, 5.298965176412673_dp, 7.782227946821637_dp, &
             11.776118747184311_dp, &
             ! w
             0.00012782536954313628_dp, 0.0003142817856536682_dp, 0.0005479372478421675_dp, &
             0.0008761669299397903_dp, 0.0013749613804741672_dp, 0.002156652108508143_dp, &
             0.003381055960707149_dp, 0.0052787238950210025_dp, 0.008187833155272854_dp, &
             0.012605467966736391_dp, 0.019258690631970676_dp, 0.029204274503283847_dp, &
             0.04396893519473523_dp, 0.06574602482814026_dp, 0.09767081131996845_dp, &
             0.14420595561955962_dp, 0.2116849238731724_dp, 0.30909277838655835_dp, &
             0.4492364929938319_dp, 0.650645899027967_dp, 0.9410883282520589_dp, &
             1.3653053208077086_dp, 2.006021060190052_dp, 3.0591941971640715_dp, &
             5.283419302910287_dp]

      CASE (919)

         aw(:) = & ! a
            [4.519762499227449e-05_dp, 0.00024492676244518096_dp, 0.000634920373209974_dp, &
             0.0012805772860692618_dp, 0.0023056852387015048_dp, 0.003923665897377026_dp, &
             0.006477322234611608_dp, 0.010495505480897805_dp, 0.01677973026454833_dp, &
             0.02653407379121565_dp, 0.041555077867808786_dp, 0.0645058070978417_dp, &
             0.09930859853038251_dp, 0.15170494294108228_dp, 0.23005002124932783_dp, &
             0.3464362769980205_dp, 0.5182797372274011_dp, 0.770564860926563_dp, &
             1.1390542991010253_dp, 1.6750003018694268_dp, 2.452455323014001_dp, &
             3.5808482696575616_dp, 5.230575427684372_dp, 7.699859800511608_dp, &
             11.678146110650195_dp, &
             ! w
             0.00011673919420422583_dp, 0.000287519310039334_dp, 0.0005029812091582358_dp, &
             0.0008084384255100748_dp, 0.0012767086380551036_dp, 0.0020154753965507134_dp, &
             0.0031785651431472544_dp, 0.004989260348386967_dp, 0.00777676022864231_dp, &
             0.012026863707474748_dp, 0.018452556050915593_dp, 0.028093451708589783_dp, &
             0.04245602410857346_dp, 0.06371077890544444_dp, 0.09496879577587958_dp, &
             0.14066966253479712_dp, 0.2071290100905957_dp, 0.30332545733815075_dp, &
             0.44207895382865575_dp, 0.6419601085917797_dp, 0.9308093097081083_dp, &
             1.3534650043970295_dp, 1.9927326529385034_dp, 3.0445109282514466_dp, &
             5.266562685130558_dp]

      CASE (920)

         aw(:) = & ! a
            [2.484332176096168e-05_dp, 0.0001354838557265014_dp, 0.0003556905958051852_dp, &
             0.0007321049569446749_dp, 0.0013551882864956477_dp, 0.0023828939807499327_dp, &
             0.004072700686233577_dp, 0.006830967552477453_dp, 0.01129010648414186_dp, &
             0.018425398262854744_dp, 0.029728381433897944_dp, 0.04746171597611269_dp, &
             0.07503122259748701_dp, 0.1175256395953582_dp, 0.18249533262100145_dp, &
             0.2810704130355611_dp, 0.42956112326673557_dp, 0.6517486848449706_dp, &
             0.9821872250810612_dp, 1.4710642071055975_dp, 2.1917112831730408_dp, &
             3.25338566333437_dp, 4.826940950216266_dp, 7.211313348183739_dp, &
             11.094765631090855_dp, &
             ! w
             6.425761918825681e-05_dp, 0.0001602983441448431_dp, 0.00028758587207726986_dp, &
             0.00047992902442630243_dp, 0.0007920547073874592_dp, 0.0013054715655175509_dp, &
             0.002140608446084336_dp, 0.003479419697334665_dp, 0.005598573393288911_dp, &
             0.008916435620920012_dp, 0.014060558132811712_dp, 0.021964918134372163_dp, &
             0.03400947391181634_dp, 0.05221945092226801_dp, 0.07954876007137085_dp, &
             0.1202823192252337_dp, 0.18060880791731518_dp, 0.26944662692190885_dp, &
             0.3996750916931916_dp, 0.5901020972718635_dp, 0.8690208340566449_dp, &
             1.2818969352786433_dp, 1.912100215651556_dp, 2.955257183411623_dp, &
             5.1641904136068115_dp]

      CASE (921)

         aw(:) = & ! a
            [1.7504632425402243e-05_dp, 9.588228037295344e-05_dp, 0.0002539585500662228_dp, &
             0.0005301091865182882_dp, 0.0009996688673526097_dp, 0.0017950980258740318_dp, &
             0.003134648109397982_dp, 0.0053680023748393126_dp, 0.009048130082153548_dp, &
             0.015040810297903306_dp, 0.024688868504038952_dp, 0.04005623092609417_dp, &
             0.06428793471158471_dp, 0.10213760788259175_dp, 0.1607355277888786_dp, &
             0.2507009403812659_dp, 0.3877465674855937_dp, 0.5949907021003643_dp, &
             0.9063063234562462_dp, 1.3712624222858332_dp, 2.062749140324004_dp, &
             3.0898841778228925_dp, 4.6237491273505045_dp, 6.963702005016118_dp, &
             10.797508556154693_dp, &
             ! w
             4.53199621096933e-05_dp, 0.00011406055090649387_dp, 0.00020822848309713925_dp, &
             0.0003563008015662955_dp, 0.0006044408684831644_dp, 0.0010220525044236806_dp, &
             0.0017141565075007399_dp, 0.0028429279333777543_dp, 0.00465894535083049_dp, &
             0.007546236104315827_dp, 0.012087988084290301_dp, 0.019162288909922127_dp, &
             0.0300808307482645_dp, 0.04678870027833724_dp, 0.0721507688135849_dp, &
             0.11036105583197345_dp, 0.1675286515558319_dp, 0.25252506112199496_dp, &
             0.3782457917618129_dp, 0.5636138113868089_dp, 0.8371637004007547_dp, &
             1.2447152801480643_dp, 1.869984078121831_dp, 2.9085216325360026_dp, &
             5.110647135297838_dp]

      CASE (922)

         aw(:) = & ! a
            [1.36555435483062e-05_dp, 7.5059513460151e-05_dp, 0.00020021022018859206_dp, &
             0.0004225693467615781_dp, 0.0008083243678852621_dp, 0.0014744527353370161_dp, &
             0.002615371454441563_dp, 0.004546107209999381_dp, 0.0077705227356405894_dp, &
             0.01308591474450488_dp, 0.021741055644824904_dp, 0.03567284430600727_dp, &
             0.057857931353621196_dp, 0.09283142357255449_dp, 0.14744697848833743_dp, &
             0.23198410812198425_dp, 0.36175406089778367_dp, 0.5594247015197464_dp, &
             0.8584001781546909_dp, 1.3078165873758825_dp, 1.980246180222191_dp, &
             2.9846934342846145_dp, 4.492384477726292_dp, 6.802974056884105_dp, &
             10.603946678932024_dp, &
             ! w
             3.538163340451778e-05_dp, 8.967393364458034e-05_dp, 0.00016597317356243112_dp, &
             0.00028948610446630376_dp, 0.0005010971046826218_dp, 0.0008628615603581639_dp, &
             0.0014703690551964592_dp, 0.002473376128385094_dp, 0.004105762391891723_dp, &
             0.006729290995616865_dp, 0.010898068413369903_dp, 0.017453145631474484_dp, &
             0.027660458007677805_dp, 0.04341069946271799_dp, 0.06750740640052809_dp, &
             0.10408074270093545_dp, 0.15918205149134254_dp, 0.24164563507240974_dp, &
             0.3643714346865716_dp, 0.5463542126175686_dp, 0.8162890470614628_dp, &
             1.220239543641718_dp, 1.8421696978261015_dp, 2.8776088376333826_dp, &
             5.075253830687111_dp]

      CASE (923)

         aw(:) = & ! a
            [1.1264255419591642e-05_dp, 6.209747765944256e-05_dp, 0.00016662427381156723_dp, &
             0.0003549575704893148_dp, 0.0006869798909584008_dp, 0.0012689773472266972_dp, &
             0.002278893846679679_dp, 0.004007665131127903_dp, 0.006924733464710715_dp, &
             0.011778984949636095_dp, 0.019752171049436267_dp, 0.03268998821484311_dp, &
             0.05344731770236428_dp, 0.0864001366970387_dp, 0.13819927492725134_dp, &
             0.21887340544380418_dp, 0.34343521255423776_dp, 0.5342149781239465_dp, &
             0.8242623270666383_dp, 1.2623822438794168_dp, 1.9208992263493259_dp, &
             2.9087230063930716_dp, 4.3971823059618655_dp, 6.686159196138108_dp, &
             10.462955777325103_dp, &
             ! w
             2.920455382460861e-05_dp, 7.445663532265556e-05_dp, 0.00013940490970779088_dp, &
             0.000246977148074537_dp, 0.00043438011979209844_dp, 0.0007586007724358414_dp, &
             0.0013086554502651927_dp, 0.0022254850183210024_dp, 0.0037309676582622645_dp, &
             0.006170727647565829_dp, 0.010077629077303293_dp, 0.016265462257876492_dp, &
             0.02596621390714508_dp, 0.041029870377742375_dp, 0.06421359995014883_dp, &
             0.09959865728823639_dp, 0.1531912267289187_dp, 0.23379495604406195_dp, &
             0.3543096555544806_dp, 0.5337804796639635_dp, 0.8010208436801388_dp, &
             1.2022786064563087_dp, 1.8217110566480992_dp, 2.854845724237211_dp, &
             5.0492026348538435_dp]

      CASE (924)

         aw(:) = & ! a
            [9.625621197267401e-06_dp, 5.3200405097665054e-05_dp, 0.00014349661230098943_dp, &
             0.00030815694262115926_dp, 0.0006023745650754684_dp, 0.0011244777467333904_dp, &
             0.002040134070104985_dp, 0.0036222337162464145_dp, 0.006314262364038193_dp, &
             0.01082835666948158_dp, 0.018295074817417798_dp, 0.030490042563441757_dp, &
             0.050174065815643425_dp, 0.08159952187154287_dp, 0.13125886519845312_dp, &
             0.20898386237354977_dp, 0.3295515377180903_dp, 0.5150241381800297_dp, &
             0.7981678999836872_dp, 1.227520856309657_dp, 1.8752049848705_dp, &
             2.850048703655369_dp, 4.323458591436008_dp, 6.59550002276312_dp, &
             10.353346856870566_dp, &
             ! w
             2.4970096291700152e-05_dp, 6.399030364795696e-05_dp, 0.00012101402966049065_dp, &
             0.00021725883981540083_dp, 0.00038717866936081424_dp, 0.000683991187346278_dp, &
             0.0011917727844063721_dp, 0.002044746251339028_dp, 0.0034555629272043204_dp, &
             0.005757359869922115_dp, 0.009466468614968169_dp, 0.015375338149256223_dp, &
             0.02468921589383284_dp, 0.039225813987042814_dp, 0.06170525978069048_dp, &
             0.09616935770063212_dp, 0.14858731742712708_dp, 0.22773677957310562_dp, &
             0.3465153525842633_dp, 0.5240061037741408_dp, 0.7891151784732309_dp, &
             1.1882376229147227_dp, 1.8056884588729358_dp, 2.837002666182713_dp, &
             5.028788497014593_dp]

      CASE (925)

         aw(:) = & ! a
            [8.428236454978627e-06_dp, 4.668969366533895e-05_dp, 0.00012652449963876722_dp, &
             0.00027365605111452846_dp, 0.0005396111928525136_dp, 0.0010164946681599195_dp, &
             0.001860359770054331_dp, 0.003329899389879533_dp, 0.005848063104184197_dp, &
             0.010097754300465372_dp, 0.017168602220495682_dp, 0.028779945673755965_dp, &
             0.04761668029406841_dp, 0.07783101530822158_dp, 0.12578650565394875_dp, &
             0.20115397121480333_dp, 0.31851696184495_dp, 0.49971662762732477_dp, &
             0.7772841825310924_dp, 1.1995347612743323_dp, 1.8384194071762552_dp, &
             2.8026953750002894_dp, 4.263831014620897_dp, 6.5220447078557875_dp, &
             10.26441536482964_dp, &
             ! w
             2.187487012797571e-05_dp, 5.6317690403177536e-05_dp, 0.00010745650770815373_dp, &
             0.00019516169580212746_dp, 0.00035172622808923783_dp, 0.0006274213219197743_dp, &
             0.001102423236242906_dp, 0.0019055922588346733_dp, 0.003242164731832401_dp, &
             0.005435191506611393_dp, 0.008987586962551785_dp, 0.014674397938780542_dp, &
             0.023678961009454193_dp, 0.03779240269909609_dp, 0.05970415471749457_dp, &
             0.09342308035693977_dp, 0.144887156862035_dp, 0.22285147161677768_dp, &
             0.340210422747661_dp, 0.5160769449714985_dp, 0.7794328359570524_dp, &
             1.1767951257091047_dp, 1.7926117723462922_dp, 2.822429781117_dp, &
             5.012119978262623_dp]

      CASE (926)

         aw(:) = & ! a
            [7.512580561417466e-06_dp, 4.1704435446505547e-05_dp, 0.00011349619678488191_dp, &
             0.0002470641693577536_dp, 0.0004909654345888203_dp, 0.0009322633472187245_dp, &
             0.0017192114459924361_dp, 0.0030989373690991366_dp, 0.005477581254938056_dp, &
             0.009514008372088262_dp, 0.016264049607268155_dp, 0.027400378177425907_dp, &
             0.045544717315352055_dp, 0.07476561284049602_dp, 0.12131857432058243_dp, &
             0.1947390316468148_dp, 0.3094471835275971_dp, 0.48709677326247686_dp, &
             0.7600189163696878_dp, 1.1763379076590836_dp, 1.807857148305462_dp, &
             2.7632706563262546_dp, 4.214097370303493_dp, 6.460686588381629_dp, &
             10.190044226357147_dp, &
             ! w
             1.9507216928530496e-05_dp, 5.0433516999490687e-05_dp, 9.700709497699288e-05_dp, &
             0.0001780000156580277_dp, 0.00032395033806006707_dp, 0.0005827426613712303_dp, &
             0.001031364959433689_dp, 0.0017942545266352841_dp, 0.0030704985561677686_dp, &
             0.005174748322406993_dp, 0.008598700244570351_dp, 0.014102793544488573_dp, &
             0.0228518972990506_dp, 0.03661463254359253_dp, 0.05805431757214181_dp, &
             0.09115161791195821_dp, 0.14181752606515852_dp, 0.21878723973479378_dp, &
             0.3349514506224301_dp, 0.5094474083491244_dp, 0.7713204278639866_dp, &
             1.1671913259264175_dp, 1.7816227303743433_dp, 2.8101759625871567_dp, &
             4.998106870410477_dp]

      CASE (927)

         aw(:) = & ! a
            [6.788220616785301e-06_dp, 3.775607197218715e-05_dp, 0.00010315408198960777_dp, &
             0.00022587685905806756_dp, 0.00045201154936995707_dp, 0.0008644288519574218_dp, &
             0.0016048856583915864_dp, 0.0029108406101766085_dp, 0.005174321506209261_dp, &
             0.009033932557723268_dp, 0.015516912763127826_dp, 0.026256323325625823_dp, &
             0.04382008964988874_dp, 0.07220528690649423_dp, 0.11757484307922589_dp, &
             0.1893478208892097_dp, 0.3018035788096929_dp, 0.47643373859054905_dp, &
             0.7453956600440688_dp, 1.1566471369949367_dp, 1.7818618045367982_dp, &
             2.729676800890097_dp, 4.1716534951909745_dp, 6.408255316735139_dp, &
             10.126430487092456_dp, &
             ! w
             1.7633710611559317e-05_dp, 4.576657298449183e-05_dp, 8.868164885956802e-05_dp, &
             0.00016423238197371738_dp, 0.00030149488368642847_dp, 0.0005463684214304922_dp, &
             0.0009731662541954849_dp, 0.001702587537451059_dp, 0.0029284990000374976_dp, &
             0.004958395303550078_dp, 0.008274382297088961_dp, 0.013624368119035097_dp, &
             0.02215732425793583_dp, 0.0356224262033329_dp, 0.05666033470304331_dp, &
             0.08922712145359457_dp, 0.13921006076825926_dp, 0.2153265756769695_dp, &
             0.3304634232890878_dp, 0.5037781424568426_dp, 0.7643705365664168_dp, &
             1.1589514851766647_dp, 1.7721842839296196_dp, 2.7996456767036895_dp, &
             4.986066807310459_dp]

      CASE (928)

         aw(:) = & ! a
            [6.199943375647319e-06_dp, 3.4546040998317437e-05_dp, 9.47282552509405e-05_dp, &
             0.00020855662035306_dp, 0.00042002157026893344_dp, 0.0008084350623379755_dp, &
             0.0015100310249675942_dp, 0.002754020247234398_dp, 0.004920347655474828_dp, &
             0.00863020980732986_dp, 0.014886201713503853_dp, 0.02528713981625875_dp, &
             0.042354308956978254_dp, 0.07002265281685803_dp, 0.1143744006234503_dp, &
             0.18472691356373172_dp, 0.29523612394053766_dp, 0.4672511645069706_dp, &
             0.7327761804620658_dp, 1.1396215489399737_dp, 1.7593452868711055_dp, &
             2.700532870414355_dp, 4.134781931896993_dp, 6.362656872099852_dp, &
             10.071059348264148_dp, &
             ! w
             1.6111804268762397e-05_dp, 4.196739261169908e-05_dp, 8.187596326793673e-05_dp, &
             0.00015290724980524182_dp, 0.0002828950741753467_dp, 0.0005160526712214646_dp, &
             0.000924404125599723_dp, 0.0016254291085378556_dp, 0.0028084809469429_dp, &
             0.004774847665721209_dp, 0.007998292739743871_dp, 0.013215791628827338_dp, &
             0.021562403825986034_dp, 0.03477023312980401_dp, 0.05545997676611259_dp, &
             0.08756593983229598_dp, 0.13695426666173635_dp, 0.2123263245374691_dp, &
             0.3265648551837351_dp, 0.4988446579906151_dp, 0.7583130721280241_dp, &
             1.151760343195254_dp, 1.7639393660613107_dp, 2.7904427623104366_dp, &
             4.975546041382247_dp]

      CASE (929)

         aw(:) = & ! a
            [3.4196571099608765e-06_dp, 1.9317302510179668e-05_dp, 5.445369716840836e-05_dp, &
             0.00012475344197668988_dp, 0.000262734115419447_dp, 0.0005282417074206697_dp, &
             0.0010270920854476512_dp, 0.0019425136588267248_dp, 0.003586237275142373_dp, &
             0.00648003763457304_dp, 0.011484245398160816_dp, 0.01999800366146877_dp, &
             0.034268057646653756_dp, 0.05786028096787109_dp, 0.09637325777616398_dp, &
             0.1585094055170075_dp, 0.25767155812951503_dp, 0.41433102524223286_dp, &
             0.6595377813431746_dp, 1.0401717421626346_dp, 1.6270460558522717_dp, &
             2.528392873063054_dp, 3.9160111572814453_dp, 6.091099892632757_dp, &
             9.740355969773836_dp, &
             ! w
             8.912885151808805e-06_dp, 2.3860262170370725e-05_dp, 4.895236482292354e-05_dp, &
             9.689967627691996e-05_dp, 0.00018873577688507327_dp, 0.0003594069343782077_dp, &
             0.0006680055012211502_dp, 0.0012134721038989397_dp, 0.0021588532320399546_dp, &
             0.0037688838189787396_dp, 0.006467701799409704_dp, 0.010926598083724216_dp, &
             0.018196197876542582_dp, 0.02990388435521927_dp, 0.04854643172153566_dp, &
             0.07792104974793211_dp, 0.12375797387870123_dp, 0.19465080643623195_dp, &
             0.30344605212452697_dp, 0.469412888467795_dp, 0.7219838904242313_dp, &
             1.108442261642557_dp, 1.714115857034585_dp, 2.7347424409424503_dp, &
             4.91189896583111_dp]

      CASE (930)

         aw(:) = & ! a
            [2.4178342671032523e-06_dp, 1.3793600161505062e-05_dp, 3.9651293173406853e-05_dp, &
             9.329433174436365e-05_dp, 0.00020209210542904068_dp, 0.00041715146506397885_dp, &
             0.0008304548609486434_dp, 0.0016039289347853883_dp, 0.0030171351041716863_dp, &
             0.005544196902588216_dp, 0.00997624439304987_dp, 0.01761395218599895_dp, &
             0.030566915580203875_dp, 0.052214346684091015_dp, 0.0879072805507543_dp, &
             0.1460296442250238_dp, 0.23958953064092517_dp, 0.38859229995347366_dp, &
             0.6235743075610253_dp, 0.9909059713709528_dp, 1.560982049607978_dp, &
             2.441822143032943_dp, 3.8053160010776277_dp, 5.953007546134318_dp, &
             9.571540544314749_dp, &
             ! w
             6.315038114132999e-06_dp, 1.723970717190751e-05_dp, 3.659736253832021e-05_dp, &
             7.509919605637736e-05_dp, 0.00015073801047822553_dp, 0.00029424204225850757_dp, &
             0.0005585823076901842_dp, 0.0010337150233773018_dp, 0.0018697341758998729_dp, &
             0.003313120790727442_dp, 0.00576289031876162_dp, 0.00985663271978894_dp, &
             0.01660106563163279_dp, 0.02756834942144904_dp, 0.04518890067311967_dp, &
             0.07318515806968974_dp, 0.11721130730735613_dp, 0.18579763184019873_dp, &
             0.29176335478087956_dp, 0.4544192560929065_dp, 0.703343735002373_dp, &
             1.0860842192772073_dp, 1.688289771755533_dp, 2.705808149821475_dp, &
             4.8788563281729385_dp]

      CASE (931)

         aw(:) = & ! a
            [1.8921011509200616e-06_dp, 1.0881467761801985e-05_dp, 3.1773621500070025e-05_dp, &
             7.630233280031173e-05_dp, 0.00016874940270679398_dp, 0.0003549713312482043_dp, &
             0.0007185582412634302_dp, 0.0014083627889416826_dp, 0.0026839909315566697_dp, &
             0.004989722895211553_dp, 0.00907296942152873_dp, 0.016171690208676752_dp, &
             0.02830746452159138_dp, 0.048738848167106885_dp, 0.0826557465732628_dp, &
             0.13823335951581975_dp, 0.22821929003378577_dp, 0.37230919051449174_dp, &
             0.6006952916060142_dp, 0.9594033428482396_dp, 1.5185410608793113_dp, &
             2.3859769431647955_dp, 3.7336546141261318_dp, 5.863350082554545_dp, &
             9.461692378674249_dp, &
             ! w
             4.950338701686396e-06_dp, 1.3729662236592414e-05_dp, 2.992564176340581e-05_dp, &
             6.30371474210757e-05_dp, 0.0001292395102943294_dp, 0.00025669245032003835_dp, &
             0.000494556221473099_dp, 0.0009271258165516515_dp, 0.001696259532275589_dp, &
             0.003036734365152051_dp, 0.005331327715950111_dp, 0.009195675468341744_dp, &
             0.015607668596901976_dp, 0.026102920507376722_dp, 0.043067548974399766_dp, &
             0.07017356422101467_dp, 0.11302314188789121_dp, 0.1801021561997044_dp, &
             0.28420859934645554_dp, 0.44467758143980607_dp, 0.6911821759343525_dp, &
             1.0714463122754596_dp, 1.6713386816653828_dp, 2.6867927599871515_dp, &
             4.857148002851002_dp]

      CASE (932)

         aw(:) = & ! a
            [1.565215043869209e-06_dp, 9.064118577487203e-06_dp, 2.6820097046204684e-05_dp, &
             6.54917636237459e-05_dp, 0.00014724624144701632_dp, 0.000314336088855382_dp, &
             0.0006445459518849094_dp, 0.0012776079741076665_dp, 0.0024591013270511912_dp, &
             0.004612186745379088_dp, 0.008453144388013931_dp, 0.01517502237720394_dp, &
             0.026736032851169943_dp, 0.04630742523338557_dp, 0.07896194439723125_dp, &
             0.13272229581434794_dp, 0.22014485727579744_dp, 0.360696787605786_dp, &
             0.5843150352946623_dp, 0.9367679737971693_dp, 1.487946929523944_dp, &
             2.345603815090952_dp, 3.681718599945985_dp, 5.798239875445144_dp, &
             9.381796113013031_dp, &
             ! w
             4.101112570124687e-06_dp, 1.1529361186157911e-05_dp, 2.5681643827430185e-05_dp, &
             5.522119889632262e-05_dp, 0.00011508128698872398_dp, 0.0002316373352874172_dp, &
             0.0004513633812348636_dp, 0.000854531910453043_dp, 0.001577112280826013_dp, &
             0.002845462769165114_dp, 0.00503061436727277_dp, 0.00873223424771311_dp, &
             0.014907131604148268_dp, 0.02506404331452572_dp, 0.04155631506896698_dp, &
             0.06801838977907979_dp, 0.11001333434185726_dp, 0.17599307687390944_dp, &
             0.27873835614392195_dp, 0.437600510760203_dp, 0.6823212732034437_dp, &
             1.0607552093108712_dp, 1.658936184008103_dp, 2.6728673613641423_dp, &
             4.841253907411157_dp]

      CASE (933)

         aw(:) = & ! a
            [1.3410130601911737e-06_dp, 7.813781121306122e-06_dp, 2.339002765023402e-05_dp, &
             5.7932250027808746e-05_dp, 0.0001320426941584555_dp, 0.00028530059234273996_dp, &
             0.0005911566177196971_dp, 0.001182489380618414_dp, 0.0022942761740794224_dp, &
             0.004333631780317512_dp, 0.007993072218003263_dp, 0.014431211609569198_dp, &
             0.02555747987291958_dp, 0.0444756556842344_dp, 0.07616760651368647_dp, &
             0.12853732469886306_dp, 0.21399178222383325_dp, 0.3518189383407402_dp, &
             0.5717547015363916_dp, 0.9193637480428258_dp, 1.4643649219676804_dp, &
             2.314415612906338_dp, 3.6415222055393612_dp, 5.747769410581729_dp, &
             9.31979138318213_dp, &
             ! w
             3.5182527084724947e-06_dp, 1.0009821368838289e-05_dp, 2.2714226693335415e-05_dp, &
             4.9673908522294475e-05_dp, 0.00010490404149752618_dp, 0.00021344252867228221_dp, &
             0.0004197284937268107_dp, 0.0008009692872186537_dp, 0.0014886251113481908_dp, &
             0.0027025786779391737_dp, 0.004804786467920308_dp, 0.008382527533226498_dp, &
             0.014376188902870263_dp, 0.02427348339352188_dp, 0.04040200550517215_dp, &
             0.06636652481667049_dp, 0.10769900273417435_dp, 0.17282404355434658_dp, &
             0.2745079057337243_dp, 0.4321136178448005_dp, 0.6754360280113525_dp, &
             1.0524323808612344_dp, 1.6492679529729144_dp, 2.6620044434664645_dp, &
             4.828857325382887_dp]

      CASE (934)

         aw(:) = & ! a
            [1.1770336386849396e-06_dp, 6.896821855037611e-06_dp, 2.0860236453671273e-05_dp, &
             5.230944600227784e-05_dp, 0.00012062834944874861_dp, 0.00026330985295438717_dp, &
             0.0005504038875768702_dp, 0.0011093824068565774_dp, 0.0021668194446520118_dp, &
             0.00411705772566568_dp, 0.007633625239027381_dp, 0.013847525560533498_dp, &
             0.024628944381699634_dp, 0.04302721296359989_dp, 0.07395064432748309_dp, &
             0.12520686750509108_dp, 0.209081224496581_dp, 0.34471535007899884_dp, &
             0.5616804536733545_dp, 0.9053736803421107_dp, 1.4453712243449235_dp, &
             2.2892512761899986_dp, 3.609040424403979_dp, 5.706935012308736_dp, &
             9.26957773706765_dp, &
             ! w
             3.0917006189614628e-06_dp, 8.891752461680934e-06_dp, 2.050716539384258e-05_dp, &
             4.5495880413877914e-05_dp, 9.715863102838666e-05_dp, 0.00019947975822162597_dp, &
             0.00039528247967581784_dp, 0.0007593289113317334_dp, 0.0014194678739123067_dp, &
             0.002590376574557617_dp, 0.0046266912281892805_dp, 0.008105663900971633_dp, &
             0.013954345345853788_dp, 0.023643317494993175_dp, 0.03947911704406992_dp, &
             0.06504214855183323_dp, 0.10583869387098636_dp, 0.1702705908315587_dp, &
             0.2710916553905505_dp, 0.4276737839581186_dp, 0.6698547175243028_dp, &
             1.0456756886247989_dp, 1.6414104946022983_dp, 2.653171132358519_dp, &
             4.818778121140606_dp]

      CASE (935)

         aw(:) = & ! a
            [1.0515230698430059e-06_dp, 6.193284963619977e-06_dp, 1.8909367455026698e-05_dp, &
             4.794081020888575e-05_dp, 0.00011168809622300402_dp, 0.0002459561429710672_dp, &
             0.0005180304955812717_dp, 0.0010509683154563325_dp, 0.0020644547843532168_dp, &
             0.003942325401769708_dp, 0.00734243730560358_dp, 0.013372939979062666_dp, &
             0.023871446462197663_dp, 0.04184198012993085_dp, 0.07213149535930713_dp, &
             0.12246704203855552_dp, 0.20503200997425766_dp, 0.3388450763072245_dp, &
             0.5533386757466094_dp, 0.8937683355617237_dp, 1.4295891129820861_dp, &
             2.2683112303530715_dp, 3.581977287577037_dp, 5.672877761332947_dp, &
             9.227665237039842_dp, &
             ! w
             2.765043811429758e-06_dp, 8.031378984039238e-06_dp, 1.8792430330132604e-05_dp, &
             4.2214492815960355e-05_dp, 9.102159373611647e-05_dp, 0.00018833860372654003_dp, &
             0.00037566202324382627_dp, 0.0007257387022092869_dp, 0.0013634313110295066_dp, &
             0.002499099546916466_dp, 0.004481290330671292_dp, 0.00787889139526301_dp, &
             0.013607798112903054_dp, 0.023124223277326637_dp, 0.03871698850785757_dp, &
             0.06394593142042317_dp, 0.10429556087522458_dp, 0.16814826721490694_dp, &
             0.2682469792189059_dp, 0.4239705698851686_dp, 0.6651924848903303_dp, &
             1.0400246290070736_dp, 1.634832838180292_dp, 2.645773126258939_dp, &
             4.810337528358309_dp]

      CASE (936)

         aw(:) = & ! a
            [9.521463573240072e-07_dp, 5.635022214116851e-06_dp, 1.73541554048195e-05_dp, &
             4.443471059314448e-05_dp, 0.00010446170856415127_dp, 0.00023183717592642708_dp, &
             0.0004915396723030764_dp, 0.0010029277527925564_dp, 0.0019798957207170713_dp, &
             0.0037974198526104217_dp, 0.007100108112721859_dp, 0.012976739019737641_dp, &
             0.02323725277249668_dp, 0.040847095624930864_dp, 0.07060086849358448_dp, &
             0.1201567273095932_dp, 0.20161071610021764_dp, 0.3338759476272081_dp, &
             0.5462654384871978_dp, 0.8839125184070861_dp, 1.4161673165181583_dp, &
             2.250480662290421_dp, 3.5589081895105137_dp, 5.64382142479517_dp, &
             9.191883407493348_dp, &
             ! w
             2.506280958467133e-06_dp, 7.346832820297207e-06_dp, 1.7416251103630507e-05_dp, &
             3.955580840399161e-05_dp, 8.601112097616309e-05_dp, 0.00017918744218258932_dp, &
             0.0003594645617786005_dp, 0.0006978876350378568_dp, 0.001316790703744899_dp, &
             0.00242286774713162_dp, 0.00435948272105389_dp, 0.007688387030609795_dp, &
             0.013315936206723009_dp, 0.02268602574414834_dp, 0.03807225385183725_dp, &
             0.06301673356943775_dp, 0.10298514004651639_dp, 0.16634293720739177_dp, &
             0.2658233920796385_dp, 0.42081101922678027_dp, 0.6612096679774873_dp, &
             1.0351919853771312_dp, 1.629203444772178_dp, 2.6394391294330917_dp, &
             4.803111512471225_dp]

      CASE (937)

         aw(:) = & ! a
            [8.713697777228801e-07_dp, 5.180339365844703e-06_dp, 1.608208485260699e-05_dp, &
             4.1549376214282944e-05_dp, 9.847676556022646e-05_dp, 0.00022007585250690613_dp, &
             0.00046936041944032844_dp, 0.0009625283024460607_dp, 0.001908510357182973_dp, &
             0.003674669429975828_dp, 0.0068942001898652915_dp, 0.01263915773096621_dp, &
             0.02269554288139395_dp, 0.03999536534874257_dp, 0.0692877658323846_dp, &
             0.11817097085610004_dp, 0.1986649094461807_dp, 0.3295905350452975_dp, &
             0.5401564033072582_dp, 0.8753887012448163_dp, 1.4045452560527425_dp, &
             2.235024265184526_dp, 3.5388922053039833_dp, 5.618591524728985_dp, &
             9.160795815577712_dp, &
             ! w
             2.295859306466071e-06_dp, 6.787921332668822e-06_dp, 1.628370937713575e-05_dp, &
             3.734919457310432e-05_dp, 8.182464937546473e-05_dp, 0.00017150062274318687_dp, &
             0.00034579871317058034_dp, 0.0006742998163093093_dp, 0.0012771568116403166_dp, &
             0.0023578943800193616_dp, 0.004255385640779721_dp, 0.007525185781024726_dp, &
             0.013065350328789586_dp, 0.022309036385143068_dp, 0.03751654143470807_dp, &
             0.06221445436658023_dp, 0.10185190224664521_dp, 0.16477939833589778_dp, &
             0.26372153712762997_dp, 0.41806749638128915_dp, 0.6577474812725869_dp, &
             1.0309872197173404_dp, 1.624302170941023_dp, 2.6339224676647457_dp, &
             4.796818296375671_dp]

      CASE (938)

         aw(:) = & ! a
            [4.883241996554254e-07_dp, 3.0088259189500505e-06_dp, 9.912782840669955e-06_dp, &
             2.7255308553705595e-05_dp, 6.818239858452662e-05_dp, 0.00015938602642772903_dp, &
             0.0003529842589666131_dp, 0.0007474503545214432_dp, 0.0015236095434014677_dp, &
             0.0030053388936021335_dp, 0.005760131050612457_dp, 0.010763076516090155_dp, &
             0.019660443729506462_dp, 0.035187848837047085_dp, 0.06182581634012212_dp, &
             0.10681639895365452_dp, 0.18172453204217945_dp, 0.3048169670462201_dp, &
             0.504669992471773_dp, 0.825656303591553_dp, 1.3364646854441895_dp, &
             2.1441610304547405_dp, 3.420866033318728_dp, 5.469451883364541_dp, &
             8.97668473307791_dp, &
             ! w
             1.2965326194500702e-06_dp, 4.094893563823615e-06_dp, 1.0671835210228798e-05_dp, &
             2.610140425245674e-05_dp, 6.0011742804439186e-05_dp, 0.0001307469077612145_dp, &
             0.0002722840208901892_dp, 0.0005458090319333597_dp, 0.0010588672406050298_dp, &
             0.0019965242746077944_dp, 0.0036713095709842744_dp, 0.006602182035206125_dp, &
             0.011637849501482691_dp, 0.020147197451855975_dp, 0.034310355575618234_dp, &
             0.05755959492334501_dp, 0.09524246641781352_dp, 0.1556161641446703_dp, &
             0.2513484158600284_dp, 0.4018511702505057_dp, 0.6372092532083357_dp, &
             1.0059684723607225_dp, 1.595074405573704_dp, 2.6009872518894688_dp, &
             4.759256271268356_dp]

      CASE (939)

         aw(:) = & ! a
            [3.494055937256088e-07_dp, 2.2114814069415686e-06_dp, 7.585908048970869e-06_dp, &
             2.1673361572348394e-05_dp, 5.595347991731915e-05_dp, 0.00013417650648593873_dp, &
             0.000303455529009877_dp, 0.0006539952880561716_dp, 0.001353333950728148_dp, &
             0.0027045469138299257_dp, 0.005243362995942394_dp, 0.009897539799091578_dp, &
             0.018244525283890357_dp, 0.03292239953232392_dp, 0.05827722417952691_dp, &
             0.10137139304861773_dp, 0.17353859448178985_dp, 0.2927617188659314_dp, &
             0.48729048929129704_dp, 0.8011567505704275_dp, 1.3027484481055165_dp, &
             2.098950608193601_dp, 3.3619041985041136_dp, 5.394703508456835_dp, &
             8.884181070207115_dp, &
             ! w
             9.331985562425762e-07_dp, 3.090578413305247e-06_dp, 8.478807889943423e-06_dp, &
             2.1514021970201806e-05_dp, 5.082785395888855e-05_dp, 0.0001131550825464377_dp, &
             0.00023988814054951837_dp, 0.0004881788194108687_dp, 0.0009594463998562833_dp, &
             0.0018296947522543699_dp, 0.0033983967087050015_dp, 0.0061662086304422326_dp, &
             0.010956942332442057_dp, 0.019106774630023145_dp, 0.03275466413558302_dp, &
             0.05528394799262497_dp, 0.09198879339303162_dp, 0.1510763348393308_dp, &
             0.24518202765952338_dp, 0.39372593950341966_dp, 0.6268694491763945_dp, &
             0.9933228240773159_dp, 1.5802581000769693_dp, 2.5842661511936806_dp, &
             4.740191765046587_dp]

      CASE (940)

         aw(:) = & ! a
            [2.7615131946838356e-07_dp, 1.7873605963745182e-06_dp, 6.324608984772807e-06_dp, &
             1.8578275533588103e-05_dp, 4.9032766988937864e-05_dp, 0.0001196631980974643_dp, &
             0.0002745304652941211_dp, 0.0005987518627191242_dp, 0.0012516274037316089_dp, &
             0.002523247664265817_dp, 0.004929393260365425_dp, 0.00936793165809435_dp, &
             0.017372630590614946_dp, 0.03151937117817798_dp, 0.056068084221832265_dp, &
             0.09796558624451036_dp, 0.1683961676870817_dp, 0.285158532328525_dp, &
             0.47628951831731486_dp, 0.7855975649464767_dp, 1.2812719730458917_dp, &
             2.07007648050415_dp, 3.324162415431041_dp, 5.346768840269197_dp, &
             8.824778127075332_dp, &
             ! w
             7.412926137200279e-07_dp, 2.5504213561144046e-06_dp, 7.261943474643088e-06_dp, &
             1.8902257628434627e-05_dp, 4.5500663411229354e-05_dp, 0.0001028007839742903_dp, &
             0.00022058931549919753_dp, 0.00045349415536332825_dp, 0.0008990772802701902_dp, &
             0.0017276031281521092_dp, 0.0032302297398446958_dp, 0.005895897957864721_dp, &
             0.010532404907462488_dp, 0.018454784311403145_dp, 0.031775246494517635_dp, &
             0.05384515706078658_dp, 0.08992355744815264_dp, 0.14818428113251036_dp, &
             0.24124069489667668_dp, 0.38851682950570454_dp, 0.6202227505928178_dp, &
             0.9851756039340018_dp, 1.570696600172808_dp, 2.573466074545741_dp, &
             4.72787995919342_dp]

      CASE (941)

         aw(:) = & ! a
            [2.3042764731077874e-07_dp, 1.5207981128567643e-06_dp, 5.519939135940997e-06_dp, &
             1.656990649286297e-05_dp, 4.447512403910891e-05_dp, 0.00010998837641112904_dp, &
             0.00025505329787449795_dp, 0.0005612363368639505_dp, 0.0011820570597910107_dp, &
             0.0023984522841291726_dp, 0.004712080960225412_dp, 0.008999570487557137_dp, &
             0.01676354114391053_dp, 0.03053537746215978_dp, 0.05451320443713787_dp, &
             0.09556065883308584_dp, 0.16475419989132453_dp, 0.27975919838802754_dp, &
             0.4684578970304113_dp, 0.7744959039132902_dp, 1.2659170149411658_dp, &
             2.049395210405526_dp, 3.2970879168615475_dp, 5.312339306259277_dp, &
             8.782071149061263_dp, &
             ! w
             6.213641483196901e-07_dp, 2.2078986035867742e-06_dp, 6.471823739924832e-06_dp, &
             1.7175112070393368e-05_dp, 4.193130995889504e-05_dp, 9.579170687508408e-05_dp, &
             0.0002074149436250891_dp, 0.00042964721493660304_dp, 0.0008573152755713464_dp, &
             0.0016565970771438379_dp, 0.003112708358217391_dp, 0.0057061879880921425_dp, &
             0.010233309135432661_dp, 0.017993840335993303_dp, 0.031080613444186164_dp, &
             0.05282174345081988_dp, 0.08845060655332937_dp, 0.14611652685413618_dp, &
             0.23841631374346434_dp, 0.38477621909678056_dp, 0.6154410588823396_dp, &
             0.9793054213913431_dp, 1.5637996067067725_dp, 2.5656710256056474_dp, &
             4.718994713588816_dp]

      CASE (942)

         aw(:) = & ! a
            [1.989660885217588e-07_dp, 1.3363071522929592e-06_dp, 4.956012478133773e-06_dp, &
             1.5143152525005721e-05_dp, 4.1199822961100135e-05_dp, 0.00010297020851642495_dp, &
             0.00024081519320478184_dp, 0.0005336342576661464_dp, 0.0011305878939681375_dp, &
             0.0023056858031964603_dp, 0.004549866868901205_dp, 0.00872358881068555_dp, &
             0.01630569801707754_dp, 0.02979353398083965_dp, 0.053337825484258757_dp, &
             0.09373828282769495_dp, 0.1619883148053872_dp, 0.2756503753551097_dp, &
             0.4624870983819765_dp, 0.7660177682525956_dp, 1.2541728822543312_dp, &
             2.0335560026192434_dp, 3.276328447296113_dp, 5.285915697887929_dp, &
             8.749271846258091_dp, &
             ! w
             5.38764749393541e-07_dp, 1.9690394197865286e-06_dp, 5.910196188091973e-06_dp, &
             1.593001731685406e-05_dp, 3.93322547015549e-05_dp, 9.064793650861029e-05_dp, &
             0.00019768430932069284_dp, 0.0004119377906748169_dp, 0.0008261562183939421_dp, &
             0.0016034022666685883_dp, 0.0030243488529720372_dp, 0.0055630946542458955_dp, &
             0.010007057415738223_dp, 0.017644246506384773_dp, 0.030552528064701042_dp, &
             0.052042010230603425_dp, 0.0873261229815852_dp, 0.14453504142778112_dp, &
             0.23625246820085735_dp, 0.3819059940444843_dp, 0.6117669679358907_dp, &
             0.9747897929195796_dp, 1.5584896126719443_dp, 2.5596669549539905_dp, &
             4.712151463512326_dp]

      CASE (943)

         aw(:) = & ! a
            [1.7589363273051972e-07_dp, 1.200321436168315e-06_dp, 4.535854384189571e-06_dp, &
             1.4068070764342894e-05_dp, 3.8708530121047465e-05_dp, 9.759139636872547e-05_dp, &
             0.000229835181526437_dp, 0.0005122382102541937_dp, 0.001090515764274835_dp, &
             0.0022331872518521196_dp, 0.0044226737405069685_dp, 0.00850655598246814_dp, &
             0.015944708418161507_dp, 0.029207249888167825_dp, 0.05240694509988302_dp, &
             0.09229221084305227_dp, 0.15978970412362042_dp, 0.2723790162491887_dp, &
             0.4577263012776407_dp, 0.7592487301299083_dp, 1.2447849524984094_dp, &
             2.0208811121410784_dp, 3.2597011379119265_dp, 5.264736079648253_dp, &
             8.72296724982036_dp, &
             ! w
             4.781432195113995e-07_dp, 1.7918176562910505e-06_dp, 5.4867872292795165e-06_dp, &
             1.498061037447775e-05_dp, 3.733437738375037e-05_dp, 8.666904462109297e-05_dp, &
             0.00019011856102414416_dp, 0.0003981087908701062_dp, 0.0008017342120717062_dp, &
             0.001561573845482263_dp, 0.0029546708382295907_dp, 0.0054499673247172834_dp, &
             0.009827776552011789_dp, 0.017366655373305964_dp, 0.030132415521708696_dp, &
             0.0514206290887026_dp, 0.08642858047752035_dp, 0.14327088057221193_dp, &
             0.23452047191665357_dp, 0.37960578501331743_dp, 0.6088193474640177_dp, &
             0.9711637442308856_dp, 1.5542228309512058_dp, 2.5548407486824356_dp, &
             4.706651015882577_dp]

      CASE (944)

         aw(:) = & ! a
            [1.581933390102616e-07_dp, 1.095524825999186e-06_dp, 4.2089779872259085e-06_dp, &
             1.3223567205929488e-05_dp, 3.6736036057456705e-05_dp, 9.330570642318666e-05_dp, &
             0.00022104151684224522_dp, 0.0004950291140424004_dp, 0.0010581681510273982_dp, &
             0.002174480688500322_dp, 0.004319396078342461_dp, 0.008329905579934179_dp, &
             0.015650255514149947_dp, 0.028728106578962985_dp, 0.05164485523121274_dp, &
             0.09110647635564031_dp, 0.1579843176584447_dp, 0.2696892151934273_dp, &
             0.4538071483552247_dp, 0.7536702842513017_dp, 1.237040635439005_dp, &
             2.0104162095589486_dp, 3.2459627443411345_dp, 5.247225764457689_dp, &
             8.701209982391411_dp, &
             ! w
             4.3160656150673826e-07_dp, 1.6544385818668417e-06_dp, 5.154038692588179e-06_dp, &
             1.4227382427343776e-05_dp, 3.573866662984666e-05_dp, 8.347448330429102e-05_dp, &
             0.0001840182772406414_dp, 0.0003869184532871937_dp, 0.0007819113619844507_dp, &
             0.0015275317795930615_dp, 0.0028978297582352786_dp, 0.00535748821485203_dp, &
             0.00968094275084217_dp, 0.017138917402181396_dp, 0.02978721856305442_dp, &
             0.050909331511268995_dp, 0.08568908591381005_dp, 0.1422280791537895_dp, &
             0.23309018476805846_dp, 0.37770436745739533_dp, 0.6063806011871382_dp, &
             0.9681614640380704_dp, 1.5506880990692586_dp, 2.550841413995747_dp, &
             4.70209317185444_dp]

      CASE (945)

         aw(:) = & ! a
            [1.4415102193683324e-07_dp, 1.0120441484769141e-06_dp, 3.946372750697708e-06_dp, &
             1.25393881549676e-05_dp, 3.5127121567954065e-05_dp, 8.979106399806525e-05_dp, &
             0.0002137983143303452_dp, 0.00048080268062155365_dp, 0.0010313446627607103_dp, &
             0.0021256707588983817_dp, 0.004233330543343691_dp, 0.008182395966955351_dp, &
             0.015403931292023772_dp, 0.02832662905368292_dp, 0.051005359104810875_dp, &
             0.09011016263389932_dp, 0.15646550487263564_dp, 0.26742387313259175_dp, &
             0.45050312102187107_dp, 0.7489630809044053_dp, 1.2305004226369403_dp, &
             2.001571950959985_dp, 3.2343446829328895_dp, 5.23241042188484_dp, &
             8.682794317030885_dp, &
             ! w
             3.946669357819225e-07_dp, 1.5444201736033776e-06_dp, 4.884351699517509e-06_dp, &
             1.3611942316502256e-05_dp, 3.442739144149634e-05_dp, 8.083768528462951e-05_dp, &
             0.0001789648403772619_dp, 0.00037762027441350636_dp, 0.0007653975141056307_dp, &
             0.0014991082669216654_dp, 0.00285027585314387_dp, 0.005279982442694939_dp, &
             0.009557687859677614_dp, 0.0169474762594242_dp, 0.029496660898145444_dp, &
             0.05047845164945305_dp, 0.08506521879796616_dp, 0.14134744532157856_dp, &
             0.23188120769732948_dp, 0.3760958075359139_dp, 0.6043159341893328_dp, &
             0.9656181158182136_dp, 1.547692307273907_dp, 2.54745103071101_dp, &
             4.698229470975889_dp]

      CASE (946)

         aw(:) = & ! a
            [1.327175541448639e-07_dp, 9.438178623818151e-07_dp, 3.7301080641544722e-06_dp, &
             1.197175820749871e-05_dp, 3.378435404421081e-05_dp, 8.68440254762962e-05_dp, &
             0.00020770178727482063_dp, 0.0004687908408362699_dp, 0.0010086367977198087_dp, &
             0.0020842561167125134_dp, 0.004160160557644024_dp, 0.008056769636071534_dp, &
             0.01519382393948342_dp, 0.027983703321766397_dp, 0.050458440198246905_dp, &
             0.08925710773010012_dp, 0.15516372659811797_dp, 0.2654803964351467_dp, &
             0.4476660762195086_dp, 0.7449179872238292_dp, 1.2248761465465403_dp, &
             1.993961517107183_dp, 3.2243420305535375_dp, 5.219649489119213_dp, &
             8.666927126017093_dp, &
             ! w
             3.645758498225645e-07_dp, 1.454069433897389e-06_dp, 4.66052516524013e-06_dp, &
             1.3097561078829885e-05_dp, 3.3325986685701564e-05_dp, 7.861436387739844e-05_dp, &
             0.00017469049517061734_dp, 0.0003697350960989417_dp, 0.0007513620746125136_dp, &
             0.0014749038726824797_dp, 0.0028097117261707597_dp, 0.0052137686462714284_dp, &
             0.009452246750517389_dp, 0.016783501875886783_dp, 0.02924751145081051_dp, &
             0.050108598529512466_dp, 0.08452920684627144_dp, 0.14059016891056764_dp, &
             0.23084075335332718_dp, 0.37471046698337523_dp, 0.6025366401491944_dp, &
             0.9634251258767191_dp, 1.5451081737746009_dp, 2.544525914054537_dp, &
             4.694896120185015_dp]

      CASE (947)

         aw(:) = & ! a
            [7.824228034379207e-08_dp, 6.14390227895463e-07_dp, 2.658050135176706e-06_dp, &
             9.08891549512133e-06_dp, 2.6834245382791158e-05_dp, 7.136192529806615e-05_dp, &
             0.0001752884328006077_dp, 0.00040429236411606114_dp, 0.0008856811096467559_dp, &
             0.0018583921709820058_dp, 0.0037586079088100044_dp, 0.007363533558817367_dp, &
             0.014028716512106448_dp, 0.026073731156568792_dp, 0.04740024952863694_dp, &
             0.08447000809928956_dp, 0.14783469829156862_dp, 0.2545060543494268_dp, &
             0.43160250139637685_dp, 0.7219578290940919_dp, 1.1928814921416666_dp, &
             1.9505831494992472_dp, 3.1672325221206323_dp, 5.14669252797818_dp, &
             8.576118245844166_dp, &
             ! w
             2.2100376069550073e-07_dp, 1.0102842261332273e-06_dp, 3.5223676703537554e-06_dp, &
             1.0423201994401794e-05_dp, 2.7509262422083426e-05_dp, 6.67291914543849e-05_dp, &
             0.00015161429727435628_dp, 0.0003268107207837203_dp, 0.0006744147438423764_dp, &
             0.0013413898139805786_dp, 0.0025847473095681845_dp, 0.00484479499877843_dp, &
             0.008862160963782872_dp, 0.015862292711392945_dp, 0.027842869284740633_dp, &
             0.04801676591984226_dp, 0.08148867749131779_dp, 0.13628288479648168_dp, &
             0.22490808257206743_dp, 0.3667933956678423_dp, 0.5923477622167398_dp, &
             0.9508461286688416_dp, 1.5302670925041908_dp, 2.5277154163634417_dp, &
             4.675741353633826_dp]

      CASE (948)

         aw(:) = & ! a
            [5.840592583951511e-08_dp, 4.916703151625725e-07_dp, 2.2417233564691276e-06_dp, &
             7.929265495686039e-06_dp, 2.3963473362602307e-05_dp, 6.483566728336821e-05_dp, &
             0.00016140266993259147_dp, 0.0003762941870474549_dp, 0.0008317139481037687_dp, &
             0.0017583179582547426_dp, 0.0035792326700519552_dp, 0.007051638914805053_dp, &
             0.013501191152936928_dp, 0.025204049606147565_dp, 0.04600064342971184_dp, &
             0.0822690576872953_dp, 0.14445095621049703_dp, 0.24942000853528443_dp, &
             0.4241320312955493_dp, 0.7112464248136418_dp, 1.177912998073071_dp, &
             1.9302380643886932_dp, 3.140390083170787_dp, 5.112342211638947_dp, &
             8.533307303081411_dp, &
             ! w
             1.6862702362113172e-07_dp, 8.402310287250543e-07_dp, 3.0637180095673287e-06_dp, &
             9.312116495041397e-06_dp, 2.5040614930263316e-05_dp, 6.160184206200606e-05_dp, &
             0.0001415270130801681_dp, 0.00030784100460559157_dp, 0.0006400925472906312_dp, &
             0.0012813582160076616_dp, 0.0024828876460424123_dp, 0.004676694754118752_dp, &
             0.008591836701199815_dp, 0.015438175755906672_dp, 0.027193265253614513_dp, &
             0.04704537776049037_dp, 0.08007142333704705_dp, 0.13426822927461904_dp, &
             0.22212438890624178_dp, 0.3630679014738208_dp, 0.5875410058362586_dp, &
             0.9448991073360613_dp, 1.523239503085925_dp, 2.5197485731732305_dp, &
             4.666664661722717_dp]

      CASE (949)

         aw(:) = & ! a
            [4.797259934972725e-08_dp, 4.2611130176298447e-07_dp, 2.013544848930888e-06_dp, &
             7.280885294148539e-06_dp, 2.2334892645476155e-05_dp, 6.109244626183169e-05_dp, &
             0.00015336910283064213_dp, 0.0003599815262866277_dp, 0.000800085847069344_dp, &
             0.0016993749496613528_dp, 0.0034731258152949328_dp, 0.0068664457783231036_dp, &
             0.013186918886482048_dp, 0.024684398491991735_dp, 0.04516212071882386_dp, &
             0.08094726103781892_dp, 0.1424143827890205_dp, 0.24635278078863534_dp, &
             0.419618678094027_dp, 0.7047643966936252_dp, 1.168841393396041_dp, &
             1.917891902715953_dp, 3.124082931039517_dp, 5.091455108538851_dp, &
             8.50725804911902_dp, &
             ! w
             1.410520219565089e-07_dp, 7.47713633297645e-07_dp, 2.8070283540942364e-06_dp, &
             8.679948143286063e-06_dp, 2.36198038761506e-05_dp, 5.86247976093059e-05_dp, &
             0.00013562876076147245_dp, 0.0002966843128944375_dp, 0.000619807163942743_dp, &
             0.0012457276600424778_dp, 0.002422207714204915_dp, 0.0045762278144994666_dp, &
             0.008429805558661086_dp, 0.015183298738260029_dp, 0.026801957755254974_dp, &
             0.04645897612094422_dp, 0.07921418090921313_dp, 0.1330474426281743_dp, &
             0.22043481216294972_dp, 0.36080329900644204_dp, 0.5846152516564919_dp, &
             0.9412752601277751_dp, 1.5189536593481665_dp, 2.5148877742852918_dp, &
             4.661127063231731_dp]

      CASE (950)

         aw(:) = & ! a
            [4.150950764855222e-08_dp, 3.85010170455952e-07_dp, 1.8678724835592033e-06_dp, &
             6.861351903548152e-06_dp, 2.1270953966330377e-05_dp, 5.8629353909800667e-05_dp, &
             0.00014805297972890843_dp, 0.00034913731839949434_dp, 0.0007789802444671708_dp, &
             0.0016599148385464128_dp, 0.0034018933783112864_dp, 0.00674181798690482_dp, &
             0.012974971447894985_dp, 0.024333271534402123_dp, 0.04459456058834502_dp, &
             0.08005120791900255_dp, 0.1410318403484341_dp, 0.24426791323571406_dp, &
             0.41654728116069145_dp, 0.7003486409486049_dp, 1.1626556800687506_dp, &
             1.9094662990896458_dp, 3.112946206783876_dp, 5.077182331378875_dp, &
             8.489450123986295_dp, &
             ! w
             1.2395827989269563e-07_dp, 6.889331943773851e-07_dp, 2.6408328220476147e-06_dp, &
             8.266195091028286e-06_dp, 2.2682853714529978e-05_dp, 5.665029337309497e-05_dp, &
             0.00013169882542370915_dp, 0.000289222658039948_dp, 0.0006061969871994565_dp, &
             0.0012217565087754216_dp, 0.0023812868698032674_dp, 0.0045083334433175795_dp, &
             0.008320102103830555_dp, 0.015010444247216922_dp, 0.026536175130133163_dp, &
             0.0460601323748071_dp, 0.07863038746753989_dp, 0.13221510835872902_dp, &
             0.21928163450326807_dp, 0.3592561633346285_dp, 0.5826147213618769_dp, &
             0.9387956211652504_dp, 1.5160194854872515_dp, 2.5115590322407466_dp, &
             4.657334992680688_dp]

      CASE (951)

         aw(:) = & ! a
            [3.712037650520289e-08_dp, 3.568259288097215e-07_dp, 1.7666012337828326e-06_dp, &
             6.566820230376265e-06_dp, 2.0518832797606245e-05_dp, 5.6879134195059974e-05_dp, &
             0.00014426020887652044_dp, 0.0003413753051789524_dp, 0.0007638324619059357_dp, &
             0.0016315287847380496_dp, 0.003350550324003669_dp, 0.006651833667716771_dp, &
             0.012821707481600027_dp, 0.02407902008766382_dp, 0.044183090988583716_dp, &
             0.07940087660728272_dp, 0.14002743291452321_dp, 0.24275190588455509_dp, &
             0.4143120891131387_dp, 0.6971327082491973_dp, 1.1581476979382113_dp, &
             1.9033223179404801_dp, 3.1048211714574956_dp, 5.066765076052821_dp, &
             8.476448710908302_dp, &
             ! w
             1.1234280797000141e-07_dp, 6.482085269901344e-07_dp, 2.52410436624846e-06_dp, &
             7.973328969427062e-06_dp, 2.2016063964458232e-05_dp, 5.523934749634044e-05_dp, &
             0.00012888137813312476_dp, 0.000283858869672979_dp, 0.0005963912130770893_dp, &
             0.0012044524310158167_dp, 0.0023516974611879206_dp, 0.004459166794179124_dp, &
             0.008240553759556441_dp, 0.014884955032631592_dp, 0.02634301487637832_dp, &
             0.04576998580121252_dp, 0.07820531691430241_dp, 0.13160857533931408_dp, &
             0.21844066962931943_dp, 0.358127134199607_dp, 0.581153945399882_dp, &
             0.936984086504266_dp, 1.5138750788894504_dp, 2.5091257730395573_dp, &
             4.6545631255610775_dp]

      CASE (952)

         aw(:) = & ! a
            [3.396498908348899e-08_dp, 3.364000182089361e-07_dp, 1.6924135758298016e-06_dp, &
             6.34943601888739e-06_dp, 1.9960802204662824e-05_dp, 5.5575511474248536e-05_dp, &
             0.00014142665344340705_dp, 0.00033556215257177934_dp, 0.000752464899316656_dp, &
             0.0016101900392317841_dp, 0.003311896993295394_dp, 0.006584002209949997_dp, &
             0.012706043754934965_dp, 0.023886950339659367_dp, 0.04387197213051493_dp, &
             0.07890874890054307_dp, 0.13926680126641144_dp, 0.24160307051249105_dp, &
             0.412617217154555_dp, 0.6946928205949546_dp, 1.1547258431080636_dp, &
             1.8986565795935988_dp, 3.0986487027271377_dp, 5.058848838095611_dp, &
             8.466566486549759_dp, &
             ! w
             1.0398771986306039e-07_dp, 6.184509515321934e-07_dp, 2.437922753225062e-06_dp, &
             7.755833970564e-06_dp, 2.1518858085930608e-05_dp, 5.418399391780073e-05_dp, &
             0.00012676882181915776_dp, 0.00027982892309008075_dp, 0.0005890114036855777_dp, &
             0.0011914105150821932_dp, 0.002329368093239334_dp, 0.004422022484157516_dp, &
             0.008180397457966285_dp, 0.014789973149375116_dp, 0.026196696189944333_dp, &
             0.04555003998018787_dp, 0.07788287744440527_dp, 0.13114820615682488_dp, &
             0.21780200649392137_dp, 0.35726926903155937_dp, 0.5800435120451357_dp, &
             0.935606498797549_dp, 1.5122439009693132_dp, 2.507274598186067_dp, &
             4.652454388470879_dp]

      CASE (953)

         aw(:) = & ! a
            [3.1612854822741895e-08_dp, 3.2107033331068644e-07_dp, 1.6362538534874947e-06_dp, &
             6.183907003003412e-06_dp, 1.9534145081157408e-05_dp, 5.457577347432427e-05_dp, &
             0.00013924850743530424_dp, 0.0003310851244660279_dp, 0.0007436963725540934_dp, &
             0.0015937082610193096_dp, 0.003282007590387444_dp, 0.006531498138982228_dp, &
             0.012616437306134987_dp, 0.023738035072972136_dp, 0.043630587497355836_dp, &
             0.07852668649629166_dp, 0.13867594973459876_dp, 0.2407102037521426_dp, &
             0.41129935416707897_dp, 0.6927948535434808_dp, 1.1520629902218513_dp, &
             1.895024518396796_dp, 3.0938423315088226_dp, 5.052683188836677_dp, &
             8.458868261331752_dp, &
             ! w
             9.775633140917107e-08_dp, 5.959686367302335e-07_dp, 2.3722828241973046e-06_dp, &
             7.589422531515905e-06_dp, 2.113722583831963e-05_dp, 5.337201147486623e-05_dp, &
             0.00012514034086664972_dp, 0.00027671756372598327_dp, 0.0005833062935775884_dp, &
             0.0011813169077679445_dp, 0.0023120697740161636_dp, 0.004393222549542441_dp, &
             0.008133719612125006_dp, 0.01471622233420591_dp, 0.026083013688602275_dp, &
             0.04537905695236958_dp, 0.07763208876631139_dp, 0.130789969692671_dp, &
             0.21730481694371984_dp, 0.35660117346493947_dp, 0.5791784200792646_dp, &
             0.9345329663511375_dp, 1.5109724754041298_dp, 2.5058315293263207_dp, &
             4.650810565166964_dp]

      CASE (954)

         aw(:) = & ! a
            [2.982106784308046e-08_dp, 3.0932568032478317e-07_dp, 1.5929268813081832e-06_dp, &
             6.055603649998105e-06_dp, 1.920236790341532e-05_dp, 5.3796499648438656e-05_dp, &
             0.00013754754060265042_dp, 0.0003275836834337331_dp, 0.0007368301186982478_dp, &
             0.001580788619338084_dp, 0.0032585570347941576_dp, 0.006490272423031586_dp, &
             0.012546030697340487_dp, 0.02362095603187834_dp, 0.043440703815289354_dp, &
             0.07822599098436221_dp, 0.13821072222785305_dp, 0.2400068881207067_dp, &
             0.4102608829785299_dp, 0.6912987609483021_dp, 1.149963335490959_dp, &
             1.8921598826586348_dp, 3.090050647174203_dp, 5.047818294699353_dp, &
             8.452793282421238_dp, &
             ! w
             9.300716132232233e-08_dp, 5.78650057645162e-07_dp, 2.321394795134007e-06_dp, &
             7.459945071388382e-06_dp, 2.083955196991135e-05_dp, 5.273746579287848e-05_dp, &
             0.00012386581156585802_dp, 0.0002742794720718043_dp, 0.000578831108452957_dp, &
             0.0011733923335334414_dp, 0.002298478325315668_dp, 0.004370578942205484_dp, &
             0.008096997753042652_dp, 0.014658170819925504_dp, 0.02599348737426282_dp, &
             0.04524434649703489_dp, 0.07743442325432275_dp, 0.13050751239835914_dp, &
             0.21691266760187455_dp, 0.35607406367618916_dp, 0.5784956999607217_dp, &
             0.9336855543225105_dp, 1.509968682887292_dp, 2.504692121104411_dp, &
             4.649512662805106_dp]

      CASE (955)

         aw(:) = & ! a
            [2.8443374693489808e-08_dp, 3.002519531842274e-07_dp, 1.5592635600572266e-06_dp, &
             5.955543062267902e-06_dp, 1.8942955744544263e-05_dp, 5.318603868181537e-05_dp, &
             0.0001362130896252716_dp, 0.0003248334613956672_dp, 0.0007314317060541503_dp, &
             0.0015706224811796632_dp, 0.003240091265248646_dp, 0.006457789778995354_dp, &
             0.012490525532144055_dp, 0.02352861183379133_dp, 0.04329087110861864_dp, &
             0.07798862644950662_dp, 0.13784334847258925_dp, 0.23945132580692272_dp, &
             0.4094403348998895_dp, 0.6901163096126491_dp, 1.148303457122562_dp, &
             1.8898947731718565_dp, 3.087051968101036_dp, 5.0439703017080895_dp, &
             8.447987625269661_dp, &
             ! w
             8.935397327863303e-08_dp, 5.652103855965392e-07_dp, 2.2817025333755027e-06_dp, &
             7.358663604454459e-06_dp, 2.0606239145207474e-05_dp, 5.2239370765696735e-05_dp, &
             0.0001228641628927642_dp, 0.00027236151733490494_dp, 0.0005753077728344892_dp, &
             0.0011671489174657262_dp, 0.0022877637374243954_dp, 0.004352718719159734_dp, &
             0.008068019538090526_dp, 0.014612341341718776_dp, 0.025922782623229175_dp, &
             0.04513791980818469_dp, 0.07727820981652639_dp, 0.13028422360905742_dp, &
             0.2166025820511699_dp, 0.35565715940102066_dp, 0.5779556035652323_dp, &
             0.9330150500328237_dp, 1.5091743379939138_dp, 2.5037903930166054_dp, &
             4.648485512471723_dp]

      CASE (956)

         aw(:) = & ! a
            [2.554579889604458e-08_dp, 2.810294220765616e-07_dp, 1.4873628395563149e-06_dp, &
             5.740681200604643e-06_dp, 1.838386987009761e-05_dp, 5.186681866269023e-05_dp, &
             0.00013332328701137438_dp, 0.0003188677616442055_dp, 0.0007197053941881947_dp, &
             0.0015485139572249372_dp, 0.003199892941004378_dp, 0.006387016060749789_dp, &
             0.012369496823590988_dp, 0.02332711804159323_dp, 0.04296373775101195_dp, &
             0.07747009634664107_dp, 0.13704040753166502_dp, 0.23823652291098327_dp, &
             0.40764536615910757_dp, 0.6875286965581676_dp, 1.1446698405502618_dp, &
             1.8849347747858303_dp, 3.080483974473673_dp, 5.035540331415703_dp, &
             8.437458047722917_dp, &
             ! w
             8.166491643944941e-08_dp, 5.36553041282175e-07_dp, 2.19645287891008e-06_dp, &
             7.140247229758977e-06_dp, 2.0101668784085108e-05_dp, 5.115987742795437e-05_dp, &
             0.00012068969051501977_dp, 0.00026819210586740566_dp, 0.0005676395931593945_dp, &
             0.001153547314942272_dp, 0.0022644014268544885_dp, 0.0043137466160405435_dp, &
             0.008004745060257093_dp, 0.014512211808681529_dp, 0.025768221137273818_dp, &
             0.0449051551032205_dp, 0.07693640317514334_dp, 0.12979544911834126_dp, &
             0.21592355403333233_dp, 0.3547439046991982_dp, 0.5767721283796079_dp, &
             0.9315454465159397_dp, 1.507432974391191_dp, 2.5018134278609336_dp, &
             4.646233602095899_dp]

      CASE (957)

         aw(:) = & ! a
            [0.0006800749873795562_dp, 0.0036180609518548304_dp, 0.009051034040537087_dp, &
             0.017255218557269508_dp, 0.028681546356864825_dp, 0.0440216450314727_dp, &
             0.064298921450207_dp, 0.09098305996381034_dp, 0.12612360081366136_dp, &
             0.17250484994749662_dp, 0.23383687530980404_dp, 0.31500578433073784_dp, &
             0.4224077355675651_dp, 0.5643913046944888_dp, 0.7518375867825707_dp, &
             0.9989185642237394_dp, 1.3240935264429308_dp, 1.7514366936810328_dp, &
             2.3124519410747997_dp, 3.0486587241920478_dp, 4.015515745006521_dp, &
             5.2889233348583575_dp, 6.977334310897264_dp, 9.24803865935702_dp, &
             12.398072495443248_dp, 17.135640902026932_dp, &
             ! w
             0.0017492124137767726_dp, 0.004151428434220628_dp, 0.006759790942916095_dp, &
             0.009722649029950382_dp, 0.013243753829833964_dp, 0.017605088369397736_dp, &
             0.023192187728296527_dp, 0.030515135053755513_dp, 0.04022687334263273_dp, &
             0.05314985672890279_dp, 0.07032303477487614_dp, 0.09307350437086273_dp, &
             0.12311225260130998_dp, 0.16265589937224748_dp, 0.21458226628508223_dp, &
             0.28263445811521504_dp, 0.371698214670462_dp, 0.4881968860169695_dp, &
             0.6406907276970475_dp, 0.8408637117595529_dp, 1.1053148462533489_dp, &
             1.4591894859327104_dp, 1.9445496150321826_dp, 2.6432012332347354_dp, &
             3.7575497780492193_dp, 6.09060522815089_dp]

      CASE (958)

         aw(:) = & ! a
            [0.0005633988017409629_dp, 0.002999507406231127_dp, 0.007513857739307147_dp, &
             0.014354642661966652_dp, 0.023930518590500636_dp, 0.03687395055560716_dp, &
             0.05412818023599751_dp, 0.0770553297202473_dp, 0.10756206451038458_dp, &
             0.14824775993630188_dp, 0.20259227209399158_dp, 0.27520646053315306_dp, &
             0.37216860136055147_dp, 0.5014708154554379_dp, 0.6736057579721793_dp, &
             0.9023357442796528_dp, 1.2057058656117183_dp, 1.6073955838420657_dp, &
             2.138564901528592_dp, 2.840477131471388_dp, 3.7684586586283975_dp, &
             4.998423622481113_dp, 6.638967722626174_dp, 8.85754099227774_dp, &
             11.950981664059256_dp, 16.625396046433927_dp, &
             ! w
             0.001449352926846467_dp, 0.0034447791946423272_dp, 0.005624563401801641_dp, &
             0.008124072237332277_dp, 0.011131632594439013_dp, 0.014910623249740238_dp, &
             0.019822771899184428_dp, 0.026346879498449504_dp, 0.03509674789928704_dp, &
             0.046850497845360384_dp, 0.06260145127849621_dp, 0.08363260096924342_dp, &
             0.11161402044660416_dp, 0.14872650657761394_dp, 0.19782042727834628_dp, &
             0.26262488322047406_dp, 0.34803178637849824_dp, 0.46049847654577_dp, &
             0.6086539564534136_dp, 0.804288896153698_dp, 1.0641411475539202_dp, &
             1.413503320514965_dp, 1.894533795125528_dp, 2.5889391931258023_dp, &
             3.6983716437752316_dp, 6.021770425521848_dp]

      CASE (959)

         aw(:) = & ! a
            [0.000482868650475416_dp, 0.0025723714812009393_dp, 0.006451425203361023_dp, &
             0.012347223026310378_dp, 0.02063653528583523_dp, 0.03190679395449839_dp, &
             0.04703949320634453_dp, 0.0673129250297651_dp, 0.09452160320639286_dp, &
             0.13111951708025688_dp, 0.1804056680544861_dp, 0.24677450069657977_dp, &
             0.3360533672450059_dp, 0.45595110229230046_dp, 0.6166489144395308_dp, &
             0.8315772230424938_dp, 1.118441519134094_dp, 1.5005929832257932_dp, &
             2.008900372471863_dp, 2.6844038388074605_dp, 3.5823063935310624_dp, &
             4.778520545565925_dp, 6.38175364033621_dp, 8.559597226954661_dp, &
             11.608770253008302_dp, 16.233797296706665_dp, &
             ! w
             0.0012423657316500449_dp, 0.0029565153785674637_dp, 0.004838766072050065_dp, &
             0.007014616100407565_dp, 0.009660513652046008_dp, 0.013025195915537968_dp, &
             0.017451277120328422_dp, 0.02339233265703512_dp, 0.03143098560382007_dp, &
             0.0423105530809966_dp, 0.05698858695878928_dp, 0.07671289981587993_dp, &
             0.10311989223847585_dp, 0.13835964417127153_dp, 0.18525750368309649_dp, &
             0.24752753568022848_dp, 0.3300623386101616_dp, 0.4393423474253336_dp, &
             0.5840484835859676_dp, 0.7760558681498742_dp, 1.0322160091516717_dp, &
             1.3779466028542828_dp, 1.8554960292106963_dp, 2.5465111854590297_dp, &
             3.652075387805127_dp, 5.967977078570323_dp]

      CASE (960)

         aw(:) = & ! a
            [0.0004237037834910963_dp, 0.0022584250258391057_dp, 0.005669926553394399_dp, &
             0.010868911143328828_dp, 0.018206964975942954_dp, 0.02823567299372345_dp, &
             0.04178689225729039_dp, 0.06007091736987142_dp, 0.08479086560189315_dp, &
             0.11828218712891493_dp, 0.16369648584732052_dp, 0.22525157385900133_dp, &
             0.3085691852960936_dp, 0.4211246104207476_dp, 0.5728403866115461_dp, &
             0.776868940166554_dp, 1.0506285914083997_dp, 1.4171898092611077_dp, &
             1.9071687571092122_dp, 2.5614078786888856_dp, 3.434995540350199_dp, &
             4.603833669091058_dp, 6.176719596270857_dp, 8.321370353332446_dp, &
             11.33443073360128_dp, 15.91917469743004_dp, &
             ! w
             0.0010902788963688841_dp, 0.0025974522054153897_dp, 0.004260002737132316_dp, &
             0.006195586302948424_dp, 0.00857110307043823_dp, 0.011623310939915346_dp, &
             0.01567895848254813_dp, 0.021170669423382915_dp, 0.02865545855115553_dp, &
             0.03884833339107419_dp, 0.052677659828228876_dp, 0.07136205274301125_dp, &
             0.09650926148254177_dp, 0.13024244971693058_dp, 0.1753641278888878_dp, &
             0.23557328055506793_dp, 0.31576015614127284_dp, 0.4224214843120374_dp, &
             0.5642790237556757_dp, 0.7532771428412752_dp, 1.006363166804727_dp, &
             1.349063291104767_dp, 1.8237091035609123_dp, 2.5119110565578304_dp, &
             3.6143024283761602_dp, 5.9241235182278835_dp]

      CASE (961)

         aw(:) = & ! a
            [0.0003782669796188801_dp, 0.0020172331287092054_dp, 0.005069119686178599_dp, &
             0.009731235748105319_dp, 0.016334609195701042_dp, 0.025401361290988092_dp, &
             0.037722209480860816_dp, 0.05445069114683172_dp, 0.07721339376379627_dp, &
             0.10824650532415606_dp, 0.15057821335956886_dp, 0.20827817350771893_dp, &
             0.2867949825724002_dp, 0.3934062169633969_dp, 0.5378140330377112_dp, &
             0.7329323304659365_dp, 0.9959310718753651_dp, 1.3496357175092744_dp, &
             1.8244394085419915_dp, 2.4610070435796634_dp, 3.314320267516937_dp, &
             4.460265045300612_dp, 6.007713152691113_dp, 8.124494075689427_dp, &
             11.107207013050493_dp, 15.658103924584907_dp, &
             ! w
             0.0009734706766828854_dp, 0.002321471243960268_dp, 0.0038145436905808665_dp, &
             0.005563909138165681_dp, 0.00772855429135239_dp, 0.010535159498998865_dp, &
             0.014296981759005418_dp, 0.019428854353673957_dp, 0.026466260006582223_dp, &
             0.036100551006574624_dp, 0.0492355729938984_dp, 0.06706502306889205_dp, &
             0.09117170497025595_dp, 0.1236548221025004_dp, 0.16729594494773173_dp, &
             0.22577934030863508_dp, 0.30399113034440967_dp, 0.40843979736958863_dp, &
             0.5478801676084779_dp, 0.7343148739219543_dp, 0.9847738403081407_dp, &
             1.324879035386037_dp, 1.7970391136650552_dp, 2.4828429723726506_dp, &
             3.582556064398403_dp, 5.887293715164641_dp]

      CASE (962)

         aw(:) = & ! a
            [0.0003421989583335503_dp, 0.0018257085835073476_dp, 0.004591736677656845_dp, &
             0.008826434220953463_dp, 0.014843624961940874_dp, 0.02314062269237656_dp, &
             0.03447325637711267_dp, 0.04994665535974386_dp, 0.07112199266415632_dp, &
             0.10015062275900183_dp, 0.1399552377517428_dp, 0.19447860488209762_dp, &
             0.2690204942807069_dp, 0.37068771079581253_dp, 0.5089910086382617_dp, &
             0.6966358669709843_dp, 0.9505737785372849_dp, 1.2934130109358752_dp, &
             1.755347096540008_dp, 2.3768798048940627_dp, 3.2128931351947743_dp, &
             4.3392537332948615_dp, 5.864896048508325_dp, 7.957750984839813_dp, &
             10.914390629911459_dp, 15.436209910892218_dp, &
             ! w
             0.0008807402834673146_dp, 0.0021022300401999684_dp, 0.0034602274566588615_dp, &
             0.005060545410177445_dp, 0.007055458888449444_dp, 0.009662991250291022_dp, &
             0.013184714534354523_dp, 0.01802009401257509_dp, 0.024686169357615355_dp, &
             0.033854107476560605_dp, 0.04640669245488154_dp, 0.06351591519186055_dp, &
             0.08674249204042286_dp, 0.11816404128490503_dp, 0.16054283890141222_dp, &
             0.21754896946404417_dp, 0.29406337043978603_dp, 0.396603109143591_dp, &
             0.5339503928386523_dp, 0.7181578449236157_dp, 0.9663277769718858_dp, &
             1.3041677661571567_dp, 1.7741579608285538_dp, 2.457875518507209_dp, &
             3.5552775883497483_dp, 5.855665566184162_dp]

      CASE (963)

         aw(:) = & ! a
            [0.0003128233455421415_dp, 0.001669673082754051_dp, 0.004202590323833114_dp, &
             0.008088244445134112_dp, 0.013625780354074402_dp, 0.02129122886077477_dp, &
             0.031810297971517354_dp, 0.04624612856927586_dp, 0.0661030672949113_dp, &
             0.09345878680220564_dp, 0.13114431580517746_dp, 0.18299194836082766_dp, &
             0.2541714354460333_dp, 0.3516397443271608_dp, 0.4847388725624787_dp, &
             0.665989565060541_dp, 0.9121485459810947_dp, 1.2456291627103782_dp, &
             1.696444332912379_dp, 2.3049503123129553_dp, 3.1259357340551794_dp, &
             4.235245837328704_dp, 5.741868646575056_dp, 7.813827252112669_dp, &
             10.747679303707777_dp, 15.244086992898186_dp, &
             ! w
             0.0008052105618199396_dp, 0.0019235456225993144_dp, 0.0031711267284408634_dp, &
             0.00464913495010319_dp, 0.006504049546727353_dp, 0.008946334909568856_dp, &
             0.012267295895854357_dp, 0.016852937982288426_dp, 0.023204264803511575_dp, &
             0.03197492692120719_dp, 0.04402930212447864_dp, 0.060520172804849934_dp, &
             0.08298844397008205_dp, 0.11349210127434721_dp, 0.15477555393617257_dp, &
             0.2104952983462872_dp, 0.2855264524289269_dp, 0.3863923757324024_dp, &
             0.5218983468942295_dp, 0.7041405495761667_dp, 0.9502855859216035_dp, &
             1.2861183798558435_dp, 1.7541856586724158_dp, 2.4360596337389633_dp, &
             3.5314340789662326_dp, 5.828034228091876_dp]

      CASE (964)

         aw(:) = & ! a
            [0.00017291384829261748_dp, 0.0009257017264827352_dp, 0.002343428682416174_dp, &
             0.004550887296601435_dp, 0.0077657271212931145_dp, 0.012343486742842325_dp, &
             0.01883581627466424_dp, 0.0280601401141889_dp, 0.04118739304737253_dp, &
             0.05986307934498813_dp, 0.08637863703453268_dp, 0.12391006673044795_dp, &
             0.17684452734831274_dp, 0.25122321899461836_dp, 0.3553395162863297_dp, &
             0.5005457238316426_dp, 0.7023425951686122_dp, 0.9818585585484546_dp, &
             1.367883443668756_dp, 1.899736310727328_dp, 2.6315013931260287_dp, &
             3.6387891962644465_dp, 5.0308760443364715_dp, 6.976408925336727_dp, &
             9.772046782104916_dp, 14.114347646235071_dp, &
             ! w
             0.00044538864678129233_dp, 0.001070442099632409_dp, 0.0017853069225190193_dp, &
             0.002665130248769398_dp, 0.0038228260935726658_dp, 0.00542348973304381_dp, &
             0.007695984976677009_dp, 0.010945841708647623_dp, 0.015580097131435197_dp, &
             0.02214925242788893_dp, 0.03140581949728218_dp, 0.044380943055469295_dp, &
             0.0624850007126613_dp, 0.0876413379578901_dp, 0.12246539457248491_dp, &
             0.1705061984424023_dp, 0.23657567214516587_dp, 0.3272079129865313_dp, &
             0.451326853417839_dp, 0.6212859750175638_dp, 0.8546591396119484_dp, &
             1.1777506269719031_dp, 1.6335969803040233_dp, 2.303855263440203_dp, &
             3.386754092935651_dp, 5.660652309580738_dp]

      CASE (965)

         aw(:) = & ! a
            [0.00012204736632389048_dp, 0.0006547137273899111_dp, 0.0016639000148910713_dp, &
             0.003251243782910262_dp, 0.0055971716804018344_dp, 0.009000548470427925_dp, &
             0.013928979534077724_dp, 0.02107983727274523_dp, 0.03146103095517637_dp, &
             0.04650634895170061_dp, 0.06824037715051341_dp, 0.09950928232473315_dp, &
             0.1442990837276884_dp, 0.20817146698932965_dp, 0.29885839998364283_dp, &
             0.42707194755339606_dp, 0.6076073469310884_dp, 0.8608509722899242_dp, &
             1.2148626443424309_dp, 1.708314515620925_dp, 2.3948171373116023_dp, &
             3.3497734374647177_dp, 4.682573434159799_dp, 6.5622280936143556_dp, &
             9.285580848932232_dp, 13.547287670945307_dp, &
             ! w
             0.0003145128504221444_dp, 0.0007589894136388903_dp, 0.0012758739204803116_dp, &
             0.0019282168286111723_dp, 0.0028126071752766673_dp, 0.004071087399604989_dp, &
             0.00590069576977076_dp, 0.00856711682788079_dp, 0.012431523524338491_dp, &
             0.017992747443828772_dp, 0.025944436118419567_dp, 0.037250543967667894_dp, &
             0.05324622803435168_dp, 0.07577388445162916_dp, 0.10736719834183205_dp, &
             0.15150102153058_dp, 0.21293342752906466_dp, 0.29818262367612935_dp, &
             0.4162163619168194_dp, 0.5795138909103924_dp, 0.8058705985837914_dp, &
             1.1218958671286532_dp, 1.5709443976412079_dp, 2.234803707306745_dp, &
             3.3110364585859893_dp, 5.573243754665258_dp]

      CASE (966)

         aw(:) = & ! a
            [9.526509192089255e-05_dp, 0.000511849317855352_dp, 0.0013047972405718026_dp, &
             0.0025619436154127573_dp, 0.004441178863289631_dp, 0.007206419360121439_dp, &
             0.011272760908284383_dp, 0.01726231298621627_dp, 0.02608077662266949_dp, &
             0.03902871102265026_dp, 0.05796133129339388_dp, 0.08551305403153013_dp, &
             0.12540912878216595_dp, 0.18289549332952543_dp, 0.26532958134866874_dp, &
             0.38299052704205755_dp, 0.5501896143879951_dp, 0.7867960945495058_dp, &
             1.1203515692013901_dp, 1.589057834152725_dp, 2.2461675629862587_dp, &
             3.1669084116597372_dp, 4.460726006347194_dp, 6.296883118054435_dp, &
             8.97239575407156_dp, 13.180758650806741_dp, &
             ! w
             0.0002455835345996807_dp, 0.0005945321076836713_dp, 0.001005590506513917_dp, &
             0.0015344097783881383_dp, 0.0022672945223508384_dp, 0.003331465910298092_dp, &
             0.004903660968025277_dp, 0.007224570182614312_dp, 0.010626490699482138_dp, &
             0.015574890455045828_dp, 0.022724298768935162_dp, 0.03299295754449749_dp, &
             0.04766380478583325_dp, 0.06852184803420519_dp, 0.09804127527987248_dp, &
             0.13964079380481373_dp, 0.19803431686386955_dp, 0.2797202571589431_dp, &
             0.3936868428581616_dp, 0.5524913223814614_dp, 0.7740772341332859_dp, &
             1.0852678812413343_dp, 1.529653162676978_dp, 2.18914280855039_dp, &
             3.26090124218018_dp, 5.51543804685709_dp]

      CASE (967)

         aw(:) = & ! a
            [7.858861003020861e-05_dp, 0.0004228020464457426_dp, 0.0010805462378518911_dp, &
             0.002130252817788983_dp, 0.0037142745052382445_dp, 0.006072113312843861_dp, &
             0.009581896655499004_dp, 0.014812684077224287_dp, 0.022598093415146167_dp, &
             0.03414433920987659_dp, 0.0511858287410983_dp, 0.07620462665802546_dp, &
             0.11273665224621121_dp, 0.16579649471749044_dp, 0.24246462859737236_dp, &
             0.35269700416871586_dp, 0.5104400080582252_dp, 0.7351686953234271_dp, &
             1.0540245992710164_dp, 1.5048404927542525_dp, 2.140581184434082_dp, &
             3.036323893616461_dp, 4.3015432912995335_dp, 6.105692100995696_dp, &
             8.745939425896696_dp, 12.914975468664835_dp, &
             ! w
             0.00020265339583361557_dp, 0.0004918982582778122_dp, 0.0008362766563318414_dp, &
             0.0012862999440733855_dp, 0.0019209731163454416_dp, 0.0028568827096219947_dp, &
             0.0042563322914238935_dp, 0.006342407652999645_dp, 0.009426891026795546_dp, &
             0.01395106428566226_dp, 0.02054056376541806_dp, 0.030079374296996947_dp, &
             0.04381079506645361_dp, 0.0634756690751343_dp, 0.09150161349300447_dp, &
             0.131262299953859_dp, 0.1874345607168912_dp, 0.26649713497987665_dp, &
             0.37744847034119006_dp, 0.5328999922547795_dp, 0.7509046167323155_dp, &
             1.0584491474299464_dp, 1.499309735077973_dp, 2.1555055775468994_dp, &
             3.2239307912772914_dp, 5.4728466747818_dp]

      CASE (968)

         aw(:) = & ! a
            [6.714426849047052e-05_dp, 0.00036164102959078935_dp, 0.0009262784506638192_dp, &
             0.0018325623909533754_dp, 0.0032112886223927104_dp, 0.005283616037873613_dp, &
             0.00839975673005728_dp, 0.01308869472633626_dp, 0.020129540913442486_dp, &
             0.03065688468541363_dp, 0.0463128535174269_dp, 0.06946232702680057_dp, &
             0.10349456355542663_dp, 0.15324367637099853_dp, 0.22557255051246414_dp, &
             0.3301813868362005_dp, 0.48072560674769427_dp, 0.6963633182667466_dp, &
             1.0039113022008568_dp, 1.4408991611387025_dp, 2.060050881627079_dp, &
             2.9363121518368174_dp, 4.179172457541649_dp, 5.958235594916836_dp, &
             8.570806240608334_dp, 12.70897453511031_dp, &
             ! w
             0.00017318657061009572_dp, 0.0004213320140074014_dp, 0.0007194962326474239_dp, &
             0.0011143450627180186_dp, 0.0016793292440978331_dp, 0.002522894677866273_dp, &
             0.0037963911442654955_dp, 0.005709609204305061_dp, 0.00855868509856927_dp, &
             0.012766158004867703_dp, 0.01893497524969928_dp, 0.02792194469050403_dp, &
             0.04093862503708093_dp, 0.05969013643384049_dp, 0.0865659582207842_dp, &
             0.12490223520242975_dp, 0.17934393610775015_dp, 0.2563511945376545_dp, &
             0.3649273526603343_dp, 0.5177240866121879_dp, 0.732879960475268_dp, &
             1.0375133842387705_dp, 1.475554452783257_dp, 2.1291202716182727_dp, &
             3.1949073591976322_dp, 5.439431394436815_dp]

      CASE (969)

         aw(:) = & ! a
            [5.877315524615591e-05_dp, 0.0003168711662988032_dp, 0.0008131992640388029_dp, &
             0.0016138929285165722_dp, 0.0028407114967937673_dp, 0.004700355781807101_dp, &
             0.007520959244613406_dp, 0.011799804781544607_dp, 0.018272833050634645_dp, &
             0.02801769164159514_dp, 0.0426027571868882_dp, 0.06429871390008898_dp, &
             0.09637624933257341_dp, 0.1435227319529147_dp, 0.21242311104366365_dp, &
             0.3125672220038115_dp, 0.4573696852254068_dp, 0.665724556106723_dp, &
             0.9641759960011973_dp, 1.3899966830196846_dp, 1.9957036283725886_dp, &
             2.8561257255148917_dp, 4.080758897727319_dp, 5.839331470206991_dp, &
             8.4292690716787_dp, 12.542192007323566_dp, &
             ! w
             0.00015162903684708888_dp, 0.00036963089478142606_dp, 0.0006337009534260716_dp, &
             0.0009874834034994374_dp, 0.0015000048648289156_dp, 0.0022732061293618917_dp, &
             0.0034497504917524137_dp, 0.005228907684193824_dp, 0.007894311564724681_dp, &
             0.011853330682507974_dp, 0.01769036325055092_dp, 0.026239826104261222_dp, &
             0.03868693657247448_dp, 0.05670693263215636_dp, 0.08265708247627895_dp, &
             0.11984141269223443_dp, 0.1728769971924686_dp, 0.24820661328798427_dp, &
             0.35483544642595605_dp, 0.5054464541904394_dp, 0.7182479592359079_dp, &
             1.0204680489230487_dp, 1.4561678358074204_dp, 2.107552574061816_dp, &
             3.171167079399473_dp, 5.4121122040770615_dp]

      CASE (970)

         aw(:) = & ! a
            [5.236704996319027e-05_dp, 0.0002825881255930106_dp, 0.000726501835385866_dp, &
             0.0014459244001574544_dp, 0.0025552945385815713_dp, 0.004249523006647992_dp, &
             0.006838695021021226_dp, 0.010794177343872474_dp, 0.016816569310599196_dp, &
             0.025936739410602375_dp, 0.03966219561819804_dp, 0.06018549854548548_dp, &
             0.09067851669750725_dp, 0.13570573156226848_dp, 0.20180237122472047_dp, &
             0.298280397685904_dp, 0.43834978275159403_dp, 0.6406789305357604_dp, &
             0.9315775615488228_dp, 1.348095684361383_dp, 1.9425688768701441_dp, &
             2.7897210073366367_dp, 3.99904918585352_dp, 5.7403870801623125_dp, &
             8.311269185020835_dp, 12.402935022288148_dp, &
             ! w
             0.00013512937937597535_dp, 0.0003300084672626687_dp, 0.0005677892969331586_dp, &
             0.0008896577635894677_dp, 0.0013610003019452943_dp, 0.002078397292795582_dp, &
             0.003177401426481483_dp, 0.004848673307520871_dp, 0.007365521409612725_dp, &
             0.011122652393007337_dp, 0.016688860382451792_dp, 0.024879609005058326_dp, &
             0.03685768601441181_dp, 0.05427271905426708_dp, 0.07945415275827052_dp, &
             0.11567797414127165_dp, 0.1675364723188058_dp, 0.24145627805588288_dp, &
             0.346442555192769_dp, 0.4952033681167655_dp, 0.7060054816300838_dp, &
             1.0061706977203289_dp, 1.4398740382193465_dp, 2.089400751365336_dp, &
             3.1511750506850524_dp, 5.389115656189673_dp]

      CASE (971)

         aw(:) = & ! a
            [4.729664275456446e-05_dp, 0.0002554372921586266_dp, 0.0006577650984095654_dp, &
             0.0013125261902318568_dp, 0.0023280695025991497_dp, 0.003889443955632863_dp, &
             0.006291611504771252_dp, 0.009984221292039549_dp, 0.015638213125297302_dp, &
             0.024245068850249414_dp, 0.03726083665287899_dp, 0.056811748319294884_dp, &
             0.08598542050725545_dp, 0.12924113074742682_dp, 0.1929854034335601_dp, &
             0.2863766572597064_dp, 0.4224474368212489_dp, 0.6196695877367286_dp, &
             0.9041475232578073_dp, 1.3127349678868334_dp, 1.8976060372032333_dp, &
             2.733389180009217_dp, 3.9295792573513793_dp, 5.656100817459108_dp, &
             8.210587167709864_dp, 12.283961153847606_dp, &
             ! w
             0.00012206820286482532_dp, 0.0002986063874566603_dp, 0.0005154370795698593_dp, &
             0.0008116935386320558_dp, 0.0012496935459225039_dp, 0.0019214969912302185_dp, &
             0.0029566943589554753_dp, 0.004538721617540457_dp, 0.006932147823841688_dp, &
             0.010520867636437416_dp, 0.015860265062596017_dp, 0.02374943136179607_dp, &
             0.035331681887801045_dp, 0.05223428821109438_dp, 0.07676224676312167_dp, &
             0.11216668895608872_dp, 0.16301761827733965_dp, 0.2357266266099482_dp, &
             0.33929766687775437_dp, 0.48645944894469645_dp, 0.6955287493215354_dp, &
             0.993908969243626_dp, 1.4258757892032492_dp, 2.07378760223158_dp, &
             3.133970205705443_dp, 5.369331945406666_dp]

      CASE (972)

         aw(:) = & ! a
            [4.317719967734884e-05_dp, 0.00023336680795860756_dp, 0.0006018336980996464_dp, &
             0.0012038098681020043_dp, 0.0021424731842022656_dp, 0.0035944592715023886_dp, &
             0.0058418112137322625_dp, 0.009315624184933058_dp, 0.014661460733321648_dp, &
             0.02283700051662502_dp, 0.035253966060629735_dp, 0.0539812412007093_dp, &
             0.08203333029208569_dp, 0.12377788723451821_dp, 0.18550894397332324_dp, &
             0.27625022632293883_dp, 0.40887804274180545_dp, 0.6016904475676462_dp, &
             0.880609486600425_dp, 1.2823135418640648_dp, 1.8588313164004495_dp, &
             2.684703826327194_dp, 3.8694213674528846_dp, 5.582988110894329_dp, &
             8.123127725101146_dp, 12.180494361547703_dp, &
             ! w
             0.0001114553663670069_dp, 0.00027306329577101926_dp, 0.0004727669329327756_dp, &
             0.0007479507039135482_dp, 0.001158296425411531_dp, 0.0017919822969572356_dp, &
             0.0027735042629080794_dp, 0.004280115138194904_dp, 0.0065688452893215565_dp, &
             0.010014192680666882_dp, 0.015159823746232407_dp, 0.022790460929301412_dp, &
             0.0340322562933408_dp, 0.05049268292856977_dp, 0.07445496359297314_dp, &
             0.10914792741732254_dp, 0.15912134288705304_dp, 0.23077275756915933_dp, &
             0.3331041335784785_dp, 0.4788614828660507_dp, 0.6864050949448307_dp, &
             0.9832104763402225_dp, 1.4136434040592538_dp, 2.0601296079172973_dp, &
             3.1189129246260654_dp, 5.352022810567828_dp]

      CASE (973)

         aw(:) = & ! a
            [2.3696529954394216e-05_dp, 0.0001287991639836404_dp, 0.0003358900143682699_dp, &
             0.0006840018203583861_dp, 0.0012479643063968945_dp, 0.0021576439182584258_dp, &
             0.00362307508597025_dp, 0.005971815423400454_dp, 0.009706836302634192_dp, &
             0.015593724727794812_dp, 0.02478907709106132_dp, 0.039027475868502096_dp, &
             0.06089192350092822_dp, 0.09420272871142639_dp, 0.14457376171485573_dp, &
             0.22020433186469504_dp, 0.33300222359717174_dp, 0.5001733066799745_dp, &
             0.7464757579764995_dp, 1.107447768762324_dp, 1.6341471400324594_dp, &
             2.4004987655686034_dp, 3.5159072377751035_dp, 5.1508589046691355_dp, &
             7.60370514440916_dp, 11.56363867454749_dp, &
             ! w
             6.124599029230103e-05_dp, 0.0001517616577131681_dp, 0.00026868279454353525_dp, &
             0.00043969140770755015_dp, 0.0007094771072731152_dp, 0.001144285532736754_dp, &
             0.0018402242460228186_dp, 0.002939659068074601_dp, 0.004655881337856208_dp, &
             0.007307603500267008_dp, 0.011367772916235558_dp, 0.017533157983462947_dp, &
             0.026823343966772453_dp, 0.04072093632178502_dp, 0.061369278265663324_dp, &
             0.09185045423350256_dp, 0.1365760999125589_dp, 0.2018398104771673_dp, &
             0.2966114304084227_dp, 0.4337250840314629_dp, 0.6317987962384561_dp, &
             0.9187594711215165_dp, 1.3395618189041059_dp, 1.9771109985323623_dp, &
             3.02724046742181_dp, 5.246741506122488_dp]

      CASE (974)

         aw(:) = & ! a
            [1.6679761052170168e-05_dp, 9.101192188301173e-05_dp, 0.00023919164954058672_dp, &
             0.0004931572545336922_dp, 0.0009149548256836711_dp, 0.0016130186534847929_dp, &
             0.0027643325460259298_dp, 0.004648738213819286_dp, 0.007702635771125785_dp, &
             0.012600335960416894_dp, 0.020374962475382938_dp, 0.03259640993397566_dp, &
             0.05163145666661519_dp, 0.08102156631683173_dp, 0.1260284038388582_dp, &
             0.19441731882678862_dp, 0.2975776063742286_dp, 0.452119881100164_dp, &
             0.6821552622417691_dp, 1.0225728478099656_dp, 1.5238595292012491_dp, &
             2.259554964896021_dp, 3.3389738577108115_dp, 4.932849838085886_dp, &
             7.339916404870769_dp, 11.248722848339217_dp, &
             ! w
             4.314752437063996e-05_dp, 0.0001077518470897124_dp, 0.00019372208534351962_dp, &
             0.0003242819198625428_dp, 0.0005370339651032575_dp, 0.0008880358909871951_dp, &
             0.001460355558412263_dp, 0.0023798609437167477_dp, 0.0038383451961807357_dp, &
             0.006126335148697903_dp, 0.009680314796525412_dp, 0.015150863373628917_dp, &
             0.023500546187326967_dp, 0.0361437896419946_dp, 0.05514576524182976_dp, &
             0.08350412873760803_dp, 0.12554761090545138_dp, 0.18750202639477245_dp, &
             0.2783055524409113_dp, 0.41082511709594727_dp, 0.6038071990387519_dp, &
             0.8854226612224724_dp, 1.3009630601208435_dp, 1.933635218726777_dp, &
             2.979122332070239_dp, 5.19154796123303_dp]

      CASE (975)

         aw(:) = & ! a
            [1.3001984766087793e-05_dp, 7.116127252637867e-05_dp, 0.00018817390586289379_dp, &
             0.00039177798215826384_dp, 0.000736317267557241_dp, 0.001317226634587859_dp, &
             0.002291437482409679_dp, 0.003909705064102591_dp, 0.006567450381777662_dp, &
             0.010882153716399596_dp, 0.01780919686112579_dp, 0.02881367441365547_dp, &
             0.04612336665840598_dp, 0.07309871650984166_dp, 0.11477052968045261_dp, &
             0.17861695314795428_dp, 0.2756807345400839_dp, 0.4221704741533257_dp, &
             0.6417552650517414_dp, 0.9688747801678594_dp, 1.453613451097974_dp, &
             2.1692304640732014_dp, 3.224961686972853_dp, 4.791700519824695_dp, &
             7.168453178851233_dp, 11.043391139785637_dp, &
             ! w
             3.3656483278626174e-05_dp, 8.456823687052525e-05_dp, 0.00015389367787899035_dp, &
             0.00026213567417783024_dp, 0.00044252524733419586_dp, 0.0007449246515256214_dp, &
             0.00124446031871493_dp, 0.0020567162013199157_dp, 0.003359798043574568_dp, &
             0.005426034221366777_dp, 0.0086681019808671_dp, 0.013706102010484613_dp, &
             0.021464561703110372_dp, 0.03331189149601876_dp, 0.051259805563100766_dp, &
             0.07824729413002185_dp, 0.11854418715925116_dp, 0.17832617800528477_dp, &
             0.2665044110441916_dp, 0.39596175844669834_dp, 0.5855263623893628_dp, &
             0.8635326580612234_dp, 1.2755057756967882_dp, 1.9048723028837835_dp, &
             2.947242456506994_dp, 5.155005283166501_dp]

      CASE (976)

         aw(:) = & ! a
            [1.0718315150846706e-05_dp, 5.881320145962588e-05_dp, 0.00015632942277764856_dp, &
             0.0003281518552075014_dp, 0.00062332794816116_dp, 0.001128323720864188_dp, &
             0.001986229916290794_dp, 0.0034276341421111277_dp, 0.0058193228837554595_dp, &
             0.009738731357604379_dp, 0.016086019185961465_dp, 0.026251272351910014_dp, &
             0.042362065679764256_dp, 0.06764743024649592_dp, 0.10696943276619618_dp, &
             0.16759486682665498_dp, 0.26030968200090215_dp, 0.40102240616981333_dp, &
             0.6130695622943815_dp, 0.9305495355846601_dp, 1.403237379436583_dp, &
             2.104172046054719_dp, 3.1425207112169447_dp, 4.689291880522299_dp, &
             7.043702668151617_dp, 10.893671223171374_dp, &
             ! w
             2.7760738237263172e-05_dp, 7.01153213129313e-05_dp, 0.0001288943151317459_dp, &
             0.00022271096190231769_dp, 0.00038174758961133826_dp, 0.0006515914028297886_dp, &
             0.0011018585433675834_dp, 0.0018408678817666408_dp, 0.00303692172873389_dp, &
             0.004949194003149661_dp, 0.007973020736397898_dp, 0.012706134825041512_dp, &
             0.020044933254893073_dp, 0.03132351809116701_dp, 0.04851338525188155_dp, &
             0.07450890842713345_dp, 0.11353445966618125_dp, 0.1717260603688044_dp, &
             0.2579717028831469_dp, 0.38516284347180335_dp, 0.5721858020223305_dp, &
             0.8474963588483271_dp, 1.2567971101894204_dp, 1.8836871237354142_dp, &
             2.9237369830113087_dp, 5.1280743071765436_dp]

      CASE (977)

         aw(:) = & ! a
            [9.15411230314752e-06_dp, 5.034266515471185e-05_dp, 0.0001344215205311866_dp, &
             0.0002841762993812937_dp, 0.0005447221182910752_dp, 0.0009958535694640329_dp, &
             0.001770359728477814_dp, 0.003083753793747608_dp, 0.005281282444557467_dp, &
             0.00891005811317354_dp, 0.014828163199826862_dp, 0.024368193450410778_dp, &
             0.03958049176175854_dp, 0.06359225971157109_dp, 0.1011341216286402_dp, &
             0.15930737165974232_dp, 0.24869580555977153_dp, 0.38497041841672236_dp, &
             0.591202881998948_dp, 0.9012178771234863_dp, 1.3645400663240932_dp, &
             2.05402736409595_dp, 3.078786447927888_dp, 4.609914264036999_dp, &
             6.946799170983685_dp, 10.777175748831938_dp, &
             ! w
             2.3721059578624046e-05_dp, 6.018269104938307e-05_dp, 0.00011161473856337134_dp, &
             0.00019521590515585212_dp, 0.0003388839801873327_dp, 0.0005850277011660742_dp, &
             0.0009991379266343222_dp, 0.0016840167695220487_dp, 0.002800448281015556_dp, &
             0.0045974499581123195_dp, 0.00745689003858124_dp, 0.011959029457061846_dp, &
             0.018978162261342146_dp, 0.029821265806689926_dp, 0.04642782113319274_dp, &
             0.07165639789647535_dp, 0.10969449346539786_dp, 0.16664536319575102_dp, &
             0.25137686407541915_dp, 0.3767852098891963_dp, 0.5618010406082176_dp, &
             0.83497568188825_dp, 1.2421540703958742_dp, 1.8670769420552227_dp, &
             2.9052924737232733_dp, 5.106949117695199_dp]

      CASE (978)

         aw(:) = & ! a
            [8.011538951299322e-06_dp, 4.414727291794558e-05_dp, 0.00011835753345467715_dp, &
             0.00025180037890242717_dp, 0.00048652048813320784_dp, 0.0008970977310091758_dp, &
             0.0016082625192058758_dp, 0.0028236899068712793_dp, 0.0048716196079025935_dp, &
             0.008275092041820052_dp, 0.013858615599877711_dp, 0.022908694129468363_dp, &
             0.03741346973416002_dp, 0.06041777457032838_dp, 0.0965454566542163_dp, &
             0.15276279943191157_dp, 0.23948797977879843_dp, 0.3721965199962738_dp, &
             0.5737410815468109_dp, 0.8777187458387989_dp, 1.3334444502119562_dp, &
             2.0136226230150567_dp, 3.027306012240332_dp, 4.545662487348452_dp, &
             6.868224112057717_dp, 10.682585450364686_dp, &
             ! w
             2.0769397324140803e-05_dp, 5.2906323736100554e-05_dp, 9.889262192397104e-05_dp, &
             0.00017481475500315155_dp, 0.0003067753173183288_dp, 0.0005346999147658447_dp, &
             0.0009208330644032769_dp, 0.0015635853005832032_dp, 0.002617710352322344_dp, &
             0.004324036000417397_dp, 0.007053519028158293_dp, 0.011372198534568688_dp, &
             0.018136291512382954_dp, 0.028630480476273423_dp, 0.044767788499736005_dp, &
             0.06937700611935885_dp, 0.10661470164401311_dp, 0.16255626657905964_dp, &
             0.24605178510545586_dp, 0.37000002030091594_dp, 0.5533669086075808_dp, &
             0.8247820343947609_dp, 1.2302086902725924_dp, 1.853507615443971_dp, &
             2.890214488065904_dp, 5.089684427285232_dp]

      CASE (979)

         aw(:) = & ! a
            [7.138085793298004e-06_dp, 3.9405631758569e-05_dp, 0.0001060352833027681_dp, &
             0.00022687560276351_dp, 0.00044148675932495095_dp, 0.0008202264039839037_dp, &
             0.0014812941305292087_dp, 0.0026187369149824405_dp, 0.004546897907731144_dp, &
             0.007769057361543425_dp, 0.013082044991362951_dp, 0.02173421319378046_dp, &
             0.0356620224735299_dp, 0.05784161484001245_dp, 0.09280747681831969_dp, &
             0.14741250169574235_dp, 0.2319352727234425_dp, 0.36168594382991653_dp, &
             0.5593311504242302_dp, 0.8582737567184606_dp, 1.3076486668608143_dp, &
             1.9800272468824947_dp, 2.9844136452994183_dp, 4.492034370694095_dp, &
             6.802544985878319_dp, 10.603429293059833_dp, &
             ! w
             1.8512370629960755e-05_dp, 4.732945859321389e-05_dp, 8.909826890297781e-05_dp, &
             0.00015899978707284963_dp, 0.0002816773836864989_dp, 0.0004950469671479302_dp, &
             0.0008587066944865137_dp, 0.0014674528621459308_dp, 0.002471046150477078_dp, &
             0.004103503672681867_dp, 0.006726672049493583_dp, 0.010894669757205598_dp, &
             0.01744850144125048_dp, 0.02765401000420138_dp, 0.04340175997229905_dp, &
             0.0674951310697116_dp, 0.10406411793062145_dp, 0.15915990674959807_dp, &
             0.24161669534906116_dp, 0.364334431323379_dp, 0.5463080659088638_dp, &
             0.8162331100436446_dp, 1.220173835507811_dp, 1.8420949280175116_dp, &
             2.8775256863897587_dp, 5.075158651624107_dp]

      CASE (980)

         aw(:) = & ! a
            [6.447315515366079e-06_dp, 3.565176486703362e-05_dp, 9.626007413356028e-05_dp, &
             0.00020703770332841388_dp, 0.00040548020721979184_dp, 0.0007584353512550411_dp, &
             0.001378668354058478_dp, 0.0024521881184671364_dp, 0.0042816885080338716_dp, &
             0.007353819035777134_dp, 0.012442027858609615_dp, 0.020762327251525197_dp, &
             0.034207224651822514_dp, 0.05569426004567119_dp, 0.08968144901899952_dp, &
             0.14292435773222398_dp, 0.22558138492200114_dp, 0.35281985768272145_dp, &
             0.5471451926965387_dp, 0.8417912748197468_dp, 1.285735580345591_dp, &
             1.9514321579354552_dp, 2.9478409553323104_dp, 4.446238090069938_dp, &
             6.746387183468815_dp, 10.535681985640965_dp, &
             ! w
             1.672697610229953e-05_dp, 4.290869969643856e-05_dp, 8.130280692501406e-05_dp, &
             0.0001463337856486874_dp, 0.0002614283887292092_dp, 0.0004628322923203653_dp, &
             0.0008079289993928696_dp, 0.0013884657112219609_dp, 0.0023499696871174336_dp, &
             0.003920662117809293_dp, 0.006454610341780787_dp, 0.010495719371635402_dp, &
             0.016871915675052147_dp, 0.026832788191691517_dp, 0.04224945081571206_dp, &
             0.0659031778945015_dp, 0.10190071388430866_dp, 0.15627187321146097_dp, &
             0.23783650335833062_dp, 0.3594948520759777_dp, 0.5402662973560344_dp, &
             0.8089030734470476_dp, 1.211557287848049_dp, 1.8322851959845272_dp, &
             2.8666136837008023_dp, 5.062669240266597_dp]

      CASE (981)

         aw(:) = & ! a
            [5.886464532610611e-06_dp, 3.26009883977612e-05_dp, 8.830085016020185e-05_dp, &
             0.00019083628569058127_dp, 0.0003759514078148174_dp, 0.000707515919470363_dp, &
             0.0012936792774264547_dp, 0.002313602302956732_dp, 0.004060017906661753_dp, &
             0.007005305248647172_dp, 0.011902784199625784_dp, 0.019940542093872384_dp, &
             0.03297302481367222_dp, 0.05386688560504225_dp, 0.08701355932587622_dp, &
             0.13908364884683921_dp, 0.22013035661059766_dp, 0.345195649969885_dp, &
             0.5366430226212573_dp, 0.8275570983865407_dp, 1.2667756919189468_dp, &
             1.926648038438562_dp, 2.916093569090425_dp, 4.406431131413416_dp, &
             6.697520241556893_dp, 10.47667986222305_dp, &
             ! w
             1.5277060504852664e-05_dp, 3.931172288617379e-05_dp, 7.493636143456648e-05_dp, &
             0.0001359306286734141_dp, 0.0002446866870411191_dp, 0.00043603315339804124_dp, &
             0.0007654621403872697_dp, 0.0013220987812271629_dp, 0.00224781435948968_dp, &
             0.0037658086315521485_dp, 0.006223389032138102_dp, 0.010155559272038444_dp, &
             0.016378813919777503_dp, 0.026128490277382923_dp, 0.041258591043675796_dp, &
             0.06453086923997488_dp, 0.10003143553506957_dp, 0.1537709906727131_dp, &
             0.23455630469152183_dp, 0.3552873327074355_dp, 0.5350044018773913_dp, &
             0.8025093600096891_dp, 1.2040318800128793_dp, 1.8237099628961988_dp, &
             2.857070745862338_dp, 5.0517485608055965_dp]

      CASE (982)

         aw(:) = & ! a
            [3.238290342763316e-06_dp, 1.8147006909029444e-05_dp, 5.033876912627858e-05_dp, &
             0.0001127211083700331_dp, 0.00023147402305268534_dp, 0.0004542039985673368_dp, &
             0.0008637266102421281_dp, 0.0016011912287843474_dp, 0.002903345669071525_dp, &
             0.005161413976376238_dp, 0.009012995229989893_dp, 0.0154839918407676_dp, &
             0.02620565577794412_dp, 0.04374347600981119_dp, 0.07209129932837946_dp, &
             0.11740787283271652_dp, 0.1891070004539759_dp, 0.3014616684326893_dp, &
             0.47595615272918557_dp, 0.7447399264392535_dp, 1.155763209038512_dp, &
             1.7806937116698627_dp, 2.728165947692135_dp, 4.169743178819503_dp, &
             6.405894023548097_dp, 10.123564208163382_dp, &
             ! w
             8.425712395489665e-06_dp, 2.2199090084818415e-05_dp, 4.424309111678013e-05_dp, &
             8.475802931549623e-05_dp, 0.00016046069122317423_dp, 0.0002984272062873808_dp, &
             0.0005435392147754277_dp, 0.0009698932108728927_dp, 0.0016981275601559143_dp, &
             0.0029219684484493415_dp, 0.004948645838038015_dp, 0.008259865616146308_dp, &
             0.013602988254368778_dp, 0.022126276621676415_dp, 0.03557803136637922_dp, &
             0.05659788853283017_dp, 0.08914080316166212_dp, 0.1390929679984203_dp, &
             0.2151709890217123_dp, 0.3302614294763991_dp, 0.5035227312131432_dp, &
             0.7640571560951969_dp, 1.158579669857566_dp, 1.7717581608167268_dp, &
             2.7991701379927383_dp, 4.9855231331525784_dp]

      CASE (983)

         aw(:) = & ! a
            [2.285579515756328e-06_dp, 1.2916168150052104e-05_dp, 3.643855776314386e-05_dp, &
             8.357358643218068e-05_dp, 0.00017621980068586836_dp, 0.00035470375292651714_dp, &
             0.0006903830063139463_dp, 0.0013069087103258074_dp, 0.0024147882876834545_dp, &
             0.00436657453629266_dp, 0.007743883762308679_dp, 0.013493047752426329_dp, &
             0.023134360466563225_dp, 0.03908169856021381_dp, 0.06512631804527541_dp, &
             0.10716304411356753_dp, 0.174272077117968_dp, 0.28032030082697834_dp, &
             0.44632164343163794_dp, 0.7039179735406376_dp, 1.1005688895743868_dp, &
             1.7075546101068342_dp, 2.63333285070587_dp, 4.049582659154745_dp, &
             6.257108262455116_dp, 9.942716594325063_dp, &
             ! w
             5.957569970239744e-06_dp, 1.5961399544905822e-05_dp, 3.279326246751011e-05_dp, &
             6.501372582428369e-05_dp, 0.0001267968088939736_dp, 0.00024172022284842305_dp, &
             0.00044968778967707324_dp, 0.0008175543699416209_dp, 0.0014555522999045497_dp, &
             0.0025427533904299904_dp, 0.004366185129628688_dp, 0.007380323404093707_dp, &
             0.012296733668891757_dp, 0.020217978976422968_dp, 0.03283608171899166_dp, &
             0.05272460273645505_dp, 0.08376609702851401_dp, 0.1317759663965641_dp, &
             0.20541621778515945_dp, 0.31755809083044245_dp, 0.4874148637360738_dp, &
             0.7442443638411245_dp, 1.1350242822854533_dp, 1.7447224041208844_dp, &
             2.7689771696044665_dp, 4.95101187969963_dp]

      CASE (984)

         aw(:) = & ! a
            [1.7861558372834072e-06_dp, 1.0162721058852052e-05_dp, 2.9060588508942103e-05_dp, &
             6.789596292856617e-05_dp, 0.00014600269803134486_dp, 0.0002993435657128908_dp, &
             0.0005923499518536362_dp, 0.0011379739592923908_dp, 0.0021305031806454285_dp, &
             0.0038983505999002992_dp, 0.006987871231312497_dp, 0.012294860199886357_dp, &
             0.02126858894879571_dp, 0.03622517010223238_dp, 0.06082429418572556_dp, &
             0.10078820422628014_dp, 0.16497745141785863_dp, 0.26698965911671196_dp, &
             0.42752457973177455_dp, 0.6778821952071029_dp, 1.0651882557933707_dp, &
             1.660455177790823_dp, 2.572012927187589_dp, 3.971611807802713_dp, &
             6.160283663308771_dp, 9.82476529573032_dp, &
             ! w
             4.662533480362757e-06_dp, 1.2661413094988166e-05_dp, 2.6636099529206454e-05_dp, &
             5.415225318786296e-05_dp, 0.0001078642327690994_dp, 0.0002092323166044308_dp, &
             0.0003950756349935989_dp, 0.0007276994384505561_dp, 0.0013107362017076303_dp, &
             0.0023138839050105974_dp, 0.004011141473828312_dp, 0.006839284695947886_dp, &
             0.011486429548608422_dp, 0.01902496489031138_dp, 0.031109447558056426_dp, &
             0.050269047843786725_dp, 0.08033713804466817_dp, 0.12708025941636275_dp, &
             0.1991215155361568_dp, 0.309318746436775_dp, 0.4769185658253315_dp, &
             0.7312806126357808_dp, 1.1195591691660303_dp, 1.7269286628173137_dp, &
             2.749081262408883_dp, 4.9282786480308785_dp]

      CASE (985)

         aw(:) = & ! a
            [1.4758868942766727e-06_dp, 8.446492268214608e-06_dp, 2.4431064615535108e-05_dp, &
             5.795432708718084e-05_dp, 0.00012659498285636373_dp, 0.0002633258955025601_dp, &
             0.0005277997978086226_dp, 0.0010255246365668323_dp, 0.0019394145847126698_dp, &
             0.0035808371077294867_dp, 0.006471089731273009_dp, 0.011469843264226893_dp, &
             0.019975321354642264_dp, 0.03423300201914935_dp, 0.05780704553303091_dp, &
             0.09629377525783753_dp, 0.15839271284010523_dp, 0.2575031184071795_dp, &
             0.41409210152319825_dp, 0.6592050291523002_dp, 1.0397172730720725_dp, &
             1.626438282013283_dp, 2.5275983692066166_dp, 3.9149973674907232_dp, &
             6.089837346283289_dp, 9.738814548838203_dp, &
             ! w
             3.857394121672818e-06_dp, 1.0596286281071076e-05_dp, 2.273234190005156e-05_dp, &
             4.7144763035542114e-05_dp, 9.54502029898154e-05_dp, 0.00018764450035005142_dp, &
             0.0003583786230253613_dp, 0.0006667312998957075_dp, 0.0012116235799954208_dp, &
             0.002156021174408096_dp, 0.0037645117030672562_dp, 0.006461013683953383_dp, &
             0.010916516308120102_dp, 0.0181812491808032_dp, 0.029882098168377766_dp, &
             0.048515241942874286_dp, 0.07787722288582287_dp, 0.12369760320915187_dp, &
             0.19456943351685493_dp, 0.30333899766238975_dp, 0.46927587476273364_dp, &
             0.721813971920214_dp, 1.1082388668891696_dp, 1.713881260267566_dp, &
             2.7344798060252042_dp, 4.911598980700791_dp]

      CASE (986)

         aw(:) = & ! a
            [1.2632307074857689e-06_dp, 7.266934048077429e-06_dp, 2.1231093018397825e-05_dp, &
             5.102147488996831e-05_dp, 0.0001129188489552758_dp, 0.00023768181673153532_dp, &
             0.00048140325847356225_dp, 0.0009440086819948514_dp, 0.001799831769511048_dp, &
             0.0033473108154019318_dp, 0.006088643646214689_dp, 0.010855841001611832_dp, &
             0.01900787431297589_dp, 0.03273569167944661_dp, 0.055529422917049535_dp, &
             0.0928874504817337_dp, 0.15338358943165653_dp, 0.2502616758295631_dp, &
             0.4038058261954771_dp, 0.6448602648483447_dp, 1.0201015230231507_dp, &
             1.6001765955681635_dp, 2.4932343566937982_dp, 3.871111623317058_dp, &
             6.035145324235829_dp, 9.672006487991332_dp, &
             ! w
             3.3052157871270256e-06_dp, 9.172138558821055e-06_dp, 2.001039493141242e-05_dp, &
             4.218870922717441e-05_dp, 8.655763935026894e-05_dp, 0.0001720190531432748_dp, &
             0.00033158480873263235_dp, 0.0006218784772221762_dp, 0.0011382179445821935_dp, &
             0.002038396212362449_dp, 0.0035797375440436013_dp, 0.006176195114248862_dp, &
             0.010485428152813057_dp, 0.017540347507657792_dp, 0.028946119166021154_dp, &
             0.04717290321269717_dp, 0.07598803217368134_dp, 0.1210915092965627_dp, &
             0.19105200515603055_dp, 0.29870573599169054_dp, 0.4633393091107281_dp, &
             0.7144443635602686_dp, 1.0994100549118209_dp, 1.7036919506981927_dp, &
             2.723069267020809_dp, 4.8985667550359295_dp]

      CASE (987)

         aw(:) = & ! a
            [1.1077877649427223e-06_dp, 6.402638149833196e-06_dp, 1.8874666145988735e-05_dp, &
             4.587683636768642e-05_dp, 0.00010267991565522593_dp, 0.00021831670940809556_dp, &
             0.0004460914127549677_dp, 0.0008815325120941069_dp, 0.0016921821445671226_dp, &
             0.0031662002330419266_dp, 0.005790540391736413_dp, 0.010375055881095956_dp, &
             0.018247167221152834_dp, 0.03155385038925613_dp, 0.05372535506370829_dp, &
             0.09018060831193814_dp, 0.1493911462461239_dp, 0.2444739725204163_dp, &
             0.3955634096033509_dp, 0.6333384633681044_dp, 1.0043116535466552_dp, &
             1.5789951495220627_dp, 2.4654691452956343_dp, 3.835599417899475_dp, &
             5.9908338975379865_dp, 9.617827300722704_dp, &
             ! w
             2.901379205756704e-06_dp, 8.125550269009479e-06_dp, 1.7990728087923447e-05_dp, &
             3.846683146136776e-05_dp, 7.980897314073075e-05_dp, 0.0001600595942316087_dp, &
             0.0003109308943834929_dp, 0.0005870898272973621_dp, 0.0010809710716045255_dp, &
             0.0019462127601313614_dp, 0.003434284018913239_dp, 0.005951077062542574_dp, &
             0.010143432300365619_dp, 0.017030157415800114_dp, 0.028198670235951587_dp, &
             0.04609779007725842_dp, 0.07447078432757803_dp, 0.1189931535703577_dp, &
             0.18821312745614455_dp, 0.2949580395663386_dp, 0.45852775317354455_dp, &
             0.7084607088626661_dp, 1.0922310401482862_dp, 1.695397799392969_dp, &
             2.71377602104654_dp, 4.887954245341748_dp]

      CASE (988)

         aw(:) = & ! a
            [9.888733985288175e-07_dp, 5.74002216462327e-06_dp, 1.7060007329726906e-05_dp, &
             4.188794192471453e-05_dp, 9.467972856984587e-05_dp, 0.0002030735162441387_dp, &
             0.0004181098126805016_dp, 0.0008317313588524678_dp, 0.0016059190654697207_dp, &
             0.003020385608146233_dp, 0.005549514510266851_dp, 0.009984832401099344_dp, &
             0.01762758936282482_dp, 0.03058819013548879_dp, 0.05224695542647836_dp, &
             0.08795639776927838_dp, 0.14610235348381703_dp, 0.23969529198447734_dp, &
             0.38874339091245025_dp, 0.6237861263297512_dp, 0.99119702971201_dp, &
             1.561373437747705_dp, 2.442336289583095_dp, 3.8059748204385166_dp, &
             5.953830851239487_dp, 9.572548356913956_dp, &
             ! w
             2.592295438554287e-06_dp, 7.321069125294464e-06_dp, 1.6424893403698695e-05_dp, &
             3.555101524088326e-05_dp, 7.447450809421382e-05_dp, 0.00015053823685526132_dp, &
             0.0002943885633982488_dp, 0.0005590811882369939_dp, 0.001034668424396264_dp, &
             0.0018713443838382553_dp, 0.003315710175503711_dp, 0.005766936723572871_dp, &
             0.009862818460719718_dp, 0.016610338226467725_dp, 0.027581990351462787_dp, &
             0.045208594237883294_dp, 0.07321304498198991_dp, 0.11724999628923338_dp, &
             0.1858501272373051_dp, 0.29183284277388516_dp, 0.45450868928519855_dp, &
             0.7034551963018295_dp, 1.0862181889293159_dp, 1.688444754345431_dp, &
             2.7059819163393084_dp, 4.879054728719187_dp]

      CASE (989)

         aw(:) = & ! a
            [8.947636130149592e-07_dp, 5.214597499473495e-06_dp, 1.561519388498142e-05_dp, &
             3.8692490861714585e-05_dp, 8.822697561939921e-05_dp, 0.00019069889371064647_dp, &
             0.0003952618727737954_dp, 0.0007908578539835812_dp, 0.0015347972971197878_dp, &
             0.002899676352956112_dp, 0.005349258411696654_dp, 0.009659545770968513_dp, &
             0.017109565689671723_dp, 0.029778596990985087_dp, 0.05100437305502471_dp, &
             0.08608263843003333_dp, 0.14332582569947142_dp, 0.23565296489369597_dp, &
             0.3829637258732242_dp, 0.6156772738119386_dp, 0.980046946954404_dp, &
             1.5463703557783701_dp, 2.422616426058423_dp, 3.7806938830078076_dp, &
             5.922225606334609_dp, 9.533848506240155_dp, &
             ! w
             2.3475786970260737e-06_dp, 6.681630123747764e-06_dp, 1.5170588667482192e-05_dp, &
             3.319369377286438e-05_dp, 7.012829583568524e-05_dp, 0.00014273257257013498_dp, &
             0.0002807565986174396_dp, 0.0005358964081365482_dp, 0.0009961880634045113_dp, &
             0.0018089029630711964_dp, 0.00321650037110133_dp, 0.005612418514941724_dp, &
             0.009626717990427764_dp, 0.01625624814538481_dp, 0.027060683383513008_dp, &
             0.044455339715985004_dp, 0.07214551123276362_dp, 0.11576776498789267_dp, &
             0.18383743736550168_dp, 0.2891667847929794_dp, 0.4510751906606115_dp, &
             0.6991735685923384_dp, 1.0810695034186257_dp, 1.6824864575889973_dp, &
             2.6993003080865337_dp, 4.871426234542646_dp]

      CASE (990)

         aw(:) = & ! a
            [8.183008324342162e-07_dp, 4.7869328118446346e-06_dp, 1.4434782215019975e-05_dp, &
             3.606718198820443e-05_dp, 8.2892943372501e-05_dp, 0.00018041064363924077_dp, &
             0.0003761684980123291_dp, 0.0007565463301234727_dp, 0.0014748546162566816_dp, &
             0.0027975778218020234_dp, 0.005179335987080592_dp, 0.009382735228561147_dp, &
             0.016667585631734243_dp, 0.029086196168817708_dp, 0.04993932932066799_dp, &
             0.08447335815914907_dp, 0.14093676109851508_dp, 0.23216875614889984_dp, &
             0.377974124098394_dp, 0.6086665999330783_dp, 0.9703939297636615_dp, &
             1.5333657922460842_dp, 2.4055048323204384_dp, 3.7587362918930065_dp, &
             5.894754155450519_dp, 9.500190793802554_dp, &
             ! w
             2.14867181339874e-06_dp, 6.16002456580921e-06_dp, 1.4140116837101387e-05_dp, &
             3.1241003267484476e-05_dp, 6.65034489516716e-05_dp, 0.00013618689718180696_dp, &
             0.00026927291480151815_dp, 0.0005162882307712134_dp, 0.0009635305101728781_dp, &
             0.0017557451601807163_dp, 0.003131803773929586_dp, 0.0054801682991530536_dp, &
             0.009424172034020486_dp, 0.015951831579185092_dp, 0.026611623497497155_dp, &
             0.04380529144029536_dp, 0.07122267990949183_dp, 0.11448442283946264_dp, &
             0.18209225345061755_dp, 0.2868519276029316_dp, 0.4480902878467663_dp, &
             0.6954472580415753_dp, 1.0765844994587044_dp, 1.6772927565040117_dp, &
             2.6934741516834713_dp, 4.864774978505916_dp]

      CASE (991)

         aw(:) = & ! a
            [4.562469115792869e-07_dp, 2.7489989909837295e-06_dp, 8.733256639479415e-06_dp, &
             2.313553852401061e-05_dp, 5.6065343224443786e-05_dp, 0.00012766173192385065_dp, &
             0.00027659892703900174_dp, 0.0005749289666160984_dp, 0.0011533647274757695_dp, &
             0.0022435484153307266_dp, 0.004247543879449043_dp, 0.007850371478578814_dp, &
             0.014199759029268406_dp, 0.025189678388356212_dp, 0.04390248052004506_dp, &
             0.07529110942395219_dp, 0.12722169511208972_dp, 0.21205345763526556_dp, &
             0.34901695658559384_dp, 0.567783550996382_dp, 0.9138523278999139_dp, &
             1.45688636804011_dp, 2.3045122105770037_dp, 3.6287443136582156_dp, &
             5.731711127649081_dp, 9.300049739455408_dp, &
             ! w
             1.2055417783624751e-06_dp, 3.6548938859666544e-06_dp, 9.064243779199115e-06_dp, &
             2.1351301895744585e-05_dp, 4.772868408283207e-05_dp, 0.00010167189451844998_dp, &
             0.00020780488107187674_dp, 0.00040995984916220176_dp, 0.000784397091197657_dp, &
             0.0014611613352540625_dp, 0.0026580886076477362_dp, 0.004734256192807689_dp, &
             0.0082730012717923_dp, 0.01420947197842467_dp, 0.024024657367570274_dp, &
             0.04003789476951579_dp, 0.06584441041165796_dp, 0.10696612228341267_dp, &
             0.17181875284754664_dp, 0.2731637461533346_dp, 0.4303676823388346_dp, &
             0.6732422871885089_dp, 1.0497777366867178_dp, 1.6461817469712032_dp, &
             2.6585354781080484_dp, 4.824898945578199_dp]

      CASE (992)

         aw(:) = & ! a
            [3.25265730426002e-07_dp, 2.0035191019575433e-06_dp, 6.597545356466585e-06_dp, &
             1.8131385561348473e-05_dp, 4.5339723074456194e-05_dp, 0.00010595343922598583_dp, &
             0.00023458629322630387_dp, 0.0004966271796323126_dp, 0.001012131192490257_dp, &
             0.001996101875632395_dp, 0.0038252220799851826_dp, 0.007146658088996184_dp, &
             0.01305293395579405_dp, 0.023359346104199475_dp, 0.041038814112132095_dp, &
             0.07089608918057082_dp, 0.1206026947029687_dp, 0.2022716353192338_dp, &
             0.33483653817640013_dp, 0.5476336463177783_dp, 0.8858200806040595_dp, &
             1.4187664281412384_dp, 2.2539351357253308_dp, 3.563379356119668_dp, &
             5.649454848852348_dp, 9.198822480060773_dp, &
             ! w
             8.635439899868441e-07_dp, 2.725856315656924e-06_dp, 7.0994213722346605e-06_dp, &
             1.7355782909820106e-05_dp, 3.989010526416555e-05_dp, 8.688395110476006e-05_dp, &
             0.00018089644979700931_dp, 0.0003625463451390064_dp, 0.0007032195398968159_dp, &
             0.0013257444468314594_dp, 0.00243752696376032_dp, 0.0043829361927625364_dp, &
             0.007725107679712813_dp, 0.013372248408964655_dp, 0.022770646233599017_dp, &
             0.03819685882258097_dp, 0.06319644851448913_dp, 0.10323876059224307_dp, &
             0.1666925649378866_dp, 0.26629302896736046_dp, 0.42142359618353004_dp, &
             0.6619822257699717_dp, 1.0361297559113467_dp, 1.630296139192321_dp, &
             2.6406687738665062_dp, 4.804514283790612_dp]

      CASE (993)

         aw(:) = & ! a
            [2.563074156685936e-07_dp, 1.607980393135116e-06_dp, 5.445152672881355e-06_dp, &
             1.537236590890338e-05_dp, 3.9304209698318675e-05_dp, 9.352137528947918e-05_dp, &
             0.00021016538522819494_dp, 0.0004505297160373182_dp, 0.0009280651390003391_dp, &
             0.0018473900916196446_dp, 0.003569242801571125_dp, 0.006716868690485611_dp, &
             0.012347717844607661_dp, 0.022226847547468974_dp, 0.039256966691641006_dp, &
             0.06814731540406455_dp, 0.11644343855962974_dp, 0.19609825174043236_dp, &
             0.32585143878115014_dp, 0.5348192758282327_dp, 0.8679331030458826_dp, &
             1.3943687963570828_dp, 2.2214775102875945_dp, 3.5213348462304936_dp, &
             5.5964459408466345_dp, 9.133494928157374_dp, &
             ! w
             6.83208305844842e-07_dp, 2.2280942089216503e-06_dp, 6.0156296650280774e-06_dp, &
             1.5093320782926295e-05_dp, 3.5364324331927196e-05_dp, 7.821442437169257e-05_dp, &
             0.0001649202372588765_dp, 0.00033408963376083275_dp, 0.0006540375465176137_dp, &
             0.001243017515602559_dp, 0.002301784838692949_dp, 0.004165279337555477_dp, &
             0.0073836186025581195_dp, 0.012847561151804607_dp, 0.021980806472522638_dp, &
             0.037031915687568376_dp, 0.061513749884552535_dp, 0.10086076190661072_dp, &
             0.1634101460233071_dp, 0.2618786620996169_dp, 0.41565939025123094_dp, &
             0.6547056217611429_dp, 1.0272899267715916_dp, 1.6199898554700722_dp, &
             2.6290672484301876_dp, 4.791280175104064_dp]

      CASE (994)

         aw(:) = & ! a
            [2.1331448551822771e-07_dp, 1.3598455236373702e-06_dp, 4.712455049255939e-06_dp, &
             1.358923138754769e-05_dp, 3.5344829206273556e-05_dp, 8.526240947760915e-05_dp, &
             0.00019376961238278562_dp, 0.00041930195390112846_dp, 0.00087067515213978_dp, &
             0.0017451825269790492_dp, 0.0033922651238883997_dp, 0.006418150525220796_dp, &
             0.01185524278272199_dp, 0.021432598978481203_dp, 0.03800245231158692_dp, &
             0.0662051645142036_dp, 0.11349515567901422_dp, 0.19170919473803671_dp, &
             0.3194458205559337_dp, 0.5256606865849021_dp, 0.8551196101465752_dp, &
             1.3768549287486878_dp, 2.1981347341403645_dp, 3.4910495322750164_dp, &
             5.558213601175604_dp, 9.086331820346_dp, &
             ! w
             5.70641618758739e-07_dp, 1.9133732731057717e-06_dp, 5.314856848340553e-06_dp, &
             1.3602560595794284e-05_dp, 3.234090980634155e-05_dp, 7.236001203013158e-05_dp, &
             0.000154035080434313_dp, 0.0003145533052232731_dp, 0.0006200497020551953_dp, &
             0.0011855162537368045_dp, 0.002206947866273621_dp, 0.004012509785061605_dp, &
             0.007142933415833243_dp, 0.012476353250468444_dp, 0.021420069321076148_dp, &
             0.03620224262900055_dp, 0.0603118072107132_dp, 0.09915754900371249_dp, &
             0.16105323210691005_dp, 0.2587016001330439_dp, 0.41150205574261634_dp, &
             0.6494476549208797_dp, 1.020892402854168_dp, 1.6125224809282726_dp, &
             2.6206564269384836_dp, 4.781686937387278_dp]

      CASE (995)

         aw(:) = & ! a
            [1.8375820925369917e-07_dp, 1.1883674593561328e-06_dp, 4.200364909200512e-06_dp, &
             1.2326397351919616e-05_dp, 3.250763952402195e-05_dp, 7.928614665016776e-05_dp, &
             0.00018180859832414765_dp, 0.00039636371152964374_dp, 0.0008282704410917589_dp, &
             0.001669274836896903_dp, 0.00326023348222402_dp, 0.0061944025468820205_dp, &
             0.01148504137737065_dp, 0.020833617151675997_dp, 0.037053584356171496_dp, &
             0.06473226411744676_dp, 0.11125374568748922_dp, 0.1883649530710197_dp, &
             0.3145549898003143_dp, 0.5186546592454111_dp, 0.8453007388013818_dp, &
             1.3634132043342633_dp, 2.180194536094159_dp, 3.4677459537840876_dp, &
             5.528766597916794_dp, 9.049979630108789_dp, &
             ! w
             4.931811703889375e-07_dp, 1.6944267257900563e-06_dp, 4.818342486373775e-06_dp, &
             1.2530725715353673e-05_dp, 3.0143990971915996e-05_dp, 6.807061610069214e-05_dp, &
             0.0001460051036632169_dp, 0.00030005753977652193_dp, 0.0005947042427846028_dp, &
             0.0011424473242232771_dp, 0.0021356366319871317_dp, 0.003897235434408643_dp, &
             0.006960748620845674_dp, 0.012194566131649908_dp, 0.020993296281494177_dp, &
             0.035569269156639605_dp, 0.05939279361041922_dp, 0.09785259737115816_dp, &
             0.15924401731649981_dp, 0.2562585686271115_dp, 0.4083001522796776_dp, &
             0.6453923593695172_dp, 1.0159524018923725_dp, 1.6067513927821933_dp, &
             2.614153295042183_dp, 4.774270236934533_dp]

      CASE (996)

         aw(:) = & ! a
            [1.6209697952431938e-07_dp, 1.0621207532674978e-06_dp, 3.8196403038495165e-06_dp, &
             1.1377044908742176e-05_dp, 3.035407021338941e-05_dp, 7.47136878588221e-05_dp, &
             0.00017259682600429232_dp, 0.00037859984505164506_dp, 0.0007952756107111344_dp, &
             0.0016099686831430103_dp, 0.0031567057394169093_dp, 0.006018397360265397_dp, &
             0.011192999370617062_dp, 0.020359878718521313_dp, 0.03630136651204807_dp, &
             0.06356213519623669_dp, 0.10946962130897177_dp, 0.18569824830714887_dp, &
             0.31064864885118676_dp, 0.5130504936656559_dp, 0.837435790235141_dp, &
             1.3526329915060715_dp, 2.165790757845518_dp, 3.449018426920561_dp, &
             5.505083896037599_dp, 9.020726506086614_dp, &
             ! w
             4.363663601255748e-07_dp, 1.53228932356218e-06_dp, 4.444927563068069e-06_dp, &
             1.1714959912480046e-05_dp, 2.845757318003008e-05_dp, 6.47558740625986e-05_dp, &
             0.00013976549835195825_dp, 0.00028874116660658553_dp, 0.0005748381707210744_dp, &
             0.0011085704019482992_dp, 0.002079370038052941_dp, 0.00380602720092813_dp, &
             0.006816237183809477_dp, 0.011970539851458245_dp, 0.020653298327029196_dp, &
             0.03506403608486011_dp, 0.05865795844322412_dp, 0.0968074771379298_dp, &
             0.15779286524713165_dp, 0.2542963296261624_dp, 0.4057251488289527_dp, &
             0.6421274079977688_dp, 1.0119714582931185_dp, 1.6020975259562231_dp, &
             2.60890723646412_dp, 4.768287623198237_dp]

      CASE (997)

         aw(:) = & ! a
            [1.4548769708294038e-07_dp, 9.649238594586138e-07_dp, 3.52396628978383e-06_dp, &
             1.0632696637442083e-05_dp, 2.8651706643134223e-05_dp, 7.107506253927093e-05_dp, &
             0.00016522603089883108_dp, 0.0003643205171899337_dp, 0.000768648619943535_dp, &
             0.0015619452582290375_dp, 0.0030726233715244328_dp, 0.005875072879543896_dp, &
             0.010954622182863515_dp, 0.01997237135423031_dp, 0.03568488532802797_dp, &
             0.0626014751055641_dp, 0.10800253533701812_dp, 0.18350220032226405_dp, &
             0.30742742266032674_dp, 0.5084235114230242_dp, 0.8309349123140191_dp, &
             1.3437133866736237_dp, 2.1538622392390256_dp, 3.4334971941326153_dp, &
             5.485443511592956_dp, 8.996454959398125_dp, &
             ! w
             3.9277237845474694e-07_dp, 1.4068069447460971e-06_dp, 4.1520367428851375e-06_dp, &
             1.1068675428329213e-05_dp, 2.7111946716117225e-05_dp, 6.209619387714888e-05_dp, &
             0.00013473599415195008_dp, 0.0002795841058033691_dp, 0.0005587091239921457_dp, &
             0.0010809858769694016_dp, 0.0020334364689885737_dp, 0.0037313975152670374_dp, &
             0.006697748163016097_dp, 0.011786509636260054_dp, 0.020373523234000927_dp, &
             0.034647641258095364_dp, 0.0580514591008084_dp, 0.09594373203843533_dp, &
             0.1565920739683185_dp, 0.25267078182453884_dp, 0.40358976981355427_dp, &
             0.6394173901097732_dp, 1.0086646099383487_dp, 1.5982295231443286_dp, &
             2.6045457640345857_dp, 4.763314074668027_dp]

      CASE (998)

         aw(:) = & ! a
            [1.3231595968024283e-07_dp, 8.875591673696206e-07_dp, 3.2867773220415873e-06_dp, &
             1.0030562310873942e-05_dp, 2.7264844118823558e-05_dp, 6.809378235644936e-05_dp, &
             0.00015915841466512785_dp, 0.0003525195788011725_dp, 0.00074656944811823_dp, &
             0.0015220088479918155_dp, 0.0030025231086009308_dp, 0.0057553141549131544_dp, &
             0.010755041218315603_dp, 0.019647347958266347_dp, 0.03516696791923152_dp, &
             0.061793212232695426_dp, 0.10676651655546301_dp, 0.18164974037126463_dp, &
             0.30470709379363564_dp, 0.5045119499446022_dp, 0.8254339706771882_dp, &
             1.3361592727943141_dp, 2.1437521537472293_dp, 3.4203335048211754_dp, &
             5.468777504526625_dp, 8.975850887107116_dp, &
             ! w
             3.581803142844648e-07_dp, 1.3064538139210716e-06_dp, 3.915021908000313e-06_dp, &
             1.054117682635547e-05_dp, 2.6006915327676618e-05_dp, 5.990163099387182e-05_dp, &
             0.0001305697934128753_dp, 0.0002719737988436433_dp, 0.0005452664603964992_dp, &
             0.0010579387999411062_dp, 0.0019949747715444846_dp, 0.0036687858568409896_dp, &
             0.006598165891971828_dp, 0.011631598832631533_dp, 0.020137676855801638_dp, &
             0.03429616127146599_dp, 0.05753888689407997_dp, 0.09521293098976666_dp, &
             0.1555750464147666_dp, 0.25129268146553796_dp, 0.40177786696241224_dp, &
             0.6371161177166244_dp, 1.0058547084705969_dp, 1.5949412318941256_dp, &
             2.6008370421146063_dp, 4.759085080436081_dp]

      CASE (999)

         aw(:) = & ! a
            [1.2159442913104213e-07_dp, 8.243730182204329e-07_dp, 3.091676765701303e-06_dp, &
             9.531572935721358e-06_dp, 2.6108405278076098e-05_dp, 6.559538893843423e-05_dp, &
             0.00015405278147180495_dp, 0.00034255570175291975_dp, 0.0007278732562795857_dp, &
             0.0014881067872836008_dp, 0.0029428845606593017_dp, 0.005653230886874244_dp, &
             0.010584623121125201_dp, 0.019369386563771403_dp, 0.03472342134992166_dp, &
             0.061100129874337744_dp, 0.10570540127826548_dp, 0.18005771856776853_dp, &
             0.3023669316346305_dp, 0.501144015187033_dp, 0.8206936733885926_dp, &
             1.329644896459046_dp, 2.1350278995423517_dp, 3.408967849246569_dp, &
             5.4543813593328565_dp, 8.958046902931521_dp, &
             ! w
             3.3000837018863036e-07_dp, 1.2241351358880593e-06_dp, 3.7185437627628266e-06_dp, &
             1.0100606974943695e-05_dp, 2.5079067604936404e-05_dp, 5.80513034808428e-05_dp, &
             0.000127045158412295_dp, 0.0002655170011660814_dp, 0.0005338333043321268_dp, &
             0.0010382949937234127_dp, 0.0019621306562343024_dp, 0.003615229082370892_dp, &
             0.006512856391182795_dp, 0.011498709238735225_dp, 0.01993510486244627_dp, &
             0.03399392534961634_dp, 0.057097665976302565_dp, 0.09458324993504132_dp, &
             0.15469795948682544_dp, 0.25010322473547075_dp, 0.40021281493423944_dp, &
             0.6351270437338562_dp, 1.0034246567292198_dp, 1.592096297274434_dp, &
             2.5976276791918864_dp, 4.755425643242011_dp]

      CASE (1000)

         aw(:) = & ! a
            [7.052975884312649e-08_dp, 5.197955677855847e-07_dp, 2.1276690374112563e-06_dp, &
             7.004433907177465e-06_dp, 2.0132835741658847e-05_dp, 5.2476952485410094e-05_dp, &
             0.00012689199981142868_dp, 0.00028897064246147426_dp, 0.0006263938324777338_dp, &
             0.0013026214844699955_dp, 0.0026143105547792767_dp, 0.00508734177304204_dp, &
             0.009634732952540989_dp, 0.017812413864436525_dp, 0.03222785254022472_dp, &
             0.0571847581607159_dp, 0.09968874160836598_dp, 0.1710001547350933_dp, &
             0.28901155294857744_dp, 0.4818683466151774_dp, 0.7934930317010055_dp, &
             1.2921764681762131_dp, 2.084744586227854_dp, 3.3433437135151185_dp, &
             5.3711390804443955_dp, 8.854986996732828_dp, &
             ! w
             1.9561193405027e-07_dp, 8.211326080308579e-07_dp, 2.722329317890698e-06_dp, &
             7.812623642594039e-06_dp, 2.0177918058892047e-05_dp, 4.81469185038245e-05_dp, &
             0.00010797283819659017_dp, 0.0002302580302968393_dp, 0.00047090901116113_dp, &
             0.0009294424341572963_dp, 0.0017790338160186688_dp, 0.0033150627005034804_dp, &
             0.006032423815980427_dp, 0.010747060733819303_dp, 0.018784772556782105_dp, &
             0.032271402281262886_dp, 0.05457462789889467_dp, 0.09097143674482834_dp, &
             0.14965271129052868_dp, 0.24324318676508597_dp, 0.39116500381675556_dp, &
             0.6236035183214605_dp, 0.9893214014091501_dp, 1.5755636315243469_dp, &
             2.5789644963330822_dp, 4.73414783068008_dp]

      CASE (1001)

         aw(:) = & ! a
            [5.1912254806258856e-08_dp, 4.064298044289749e-07_dp, 1.754108690626556e-06_dp, &
             5.988375017342114e-06_dp, 1.7660392002570632e-05_dp, 4.6926305057114624e-05_dp, &
             0.00011519181314117514_dp, 0.0002655441644202491_dp, 0.0005814741123172845_dp, &
             0.001219638527807168_dp, 0.0024659474839121794_dp, 0.00482973871865107_dp, &
             0.009199197671456494_dp, 0.017093905302257184_dp, 0.031069490931963935_dp, &
             0.055357775373296086_dp, 0.09686776962063608_dp, 0.16673478703167277_dp, &
             0.2826970123716483_dp, 0.4727211648145708_dp, 0.7805418912730915_dp, &
             1.2742826268895935_dp, 2.0606665734772056_dp, 3.311847974107182_dp, &
             5.331113569377905_dp, 8.805363270581289_dp, &
             ! w
             1.4650135732549935e-07_dp, 6.671451570675598e-07_dp, 2.321073342098677e-06_dp, &
             6.8596035229181975e-06_dp, 1.8087820827534897e-05_dp, 4.3845772229369525e-05_dp, &
             9.956761668810095e-05_dp, 0.00021452752182226473_dp, 0.0004425410914476478_dp, &
             0.0008799233091916595_dp, 0.0016950768608085711_dp, 0.0031764546788083437_dp, &
             0.005809175795710938_dp, 0.010395799466371885_dp, 0.01824442637696643_dp, &
             0.03145847261894628_dp, 0.05337876156198433_dp, 0.08925271012636814_dp, &
             0.14724306915049712_dp, 0.2399557494148676_dp, 0.3868158574542994_dp, &
             0.6180492833977588_dp, 0.9825083162880089_dp, 1.5675635676961452_dp, &
             2.569925577614872_dp, 4.723844207232645_dp]

      CASE (1002)

         aw(:) = & ! a
            [4.208569882627062e-08_dp, 3.457314509885065e-07_dp, 1.5488722536902445e-06_dp, &
             5.417942675409361e-06_dp, 1.624963127815529e-05_dp, 4.371951649112457e-05_dp, &
             0.00010836516024425826_dp, 0.0002517647716722562_dp, 0.0005548731198006199_dp, &
             0.001170212469173054_dp, 0.00237713718336983_dp, 0.0046748601241764325_dp, &
             0.008936322779653746_dp, 0.01665873098993253_dp, 0.030365722071732895_dp, &
             0.05424464408595678_dp, 0.09514460917677328_dp, 0.1641232208353827_dp, &
             0.27882250115931784_dp, 0.4670975787863619_dp, 0.772565451446009_dp, &
             1.263244285855089_dp, 2.0457921726220456_dp, 3.292367483805231_dp, &
             5.306332827678124_dp, 8.774617165689104_dp, &
             ! w
             1.205499917804331e-07_dp, 5.832343398605094e-07_dp, 2.0955623571610993e-06_dp, &
             6.313921164805509e-06_dp, 1.6875372279390854e-05_dp, 4.13256010622087e-05_dp, &
             9.460286095384661e-05_dp, 0.00020517350569730435_dp, 0.0004255763339626193_dp, &
             0.0008501642625591446_dp, 0.0016444056200206384_dp, 0.0030924821585722037_dp, &
             0.005673468154296855_dp, 0.010181624646802669_dp, 0.017914050513205427_dp, &
             0.030960182960912516_dp, 0.05264405649928463_dp, 0.08819453273453204_dp, &
             0.14575660929741757_dp, 0.23792414696099684_dp, 0.38412372964310065_dp, &
             0.6146062145487327_dp, 0.978279757373532_dp, 1.5625938599305607_dp, &
             2.564307883856202_dp, 4.717441017149218_dp]

      CASE (1003)

         aw(:) = & ! a
            [3.596098619284216e-08_dp, 3.074731120664885e-07_dp, 1.4170469172369373e-06_dp, &
             5.046019234421252e-06_dp, 1.5319646787208583e-05_dp, 4.1587863609753575e-05_dp, &
             0.00010379727352356543_dp, 0.0002424950009272166_dp, 0.0005368975213073445_dp, &
             0.0011366851852783028_dp, 0.0023166953133367477_dp, 0.004569149287853463_dp, &
             0.00875644149255929_dp, 0.01636026865790445_dp, 0.029882055905283526_dp, &
             0.0534782249002233_dp, 0.09395617081720738_dp, 0.16231929459815117_dp, &
             0.2761424430009272_dp, 0.4632026656870715_dp, 0.7670344878695877_dp, &
             1.2555820952923231_dp, 2.0354575820580743_dp, 3.278821835460838_dp, &
             5.289090538849389_dp, 8.75321380519321_dp, &
             ! w
             1.0436269930271532e-07_dp, 5.296379626445037e-07_dp, 1.948422782434836e-06_dp, &
             5.953398109913807e-06_dp, 1.6067301184286198e-05_dp, 3.9634688369127006e-05_dp, &
             9.125382179028327e-05_dp, 0.0001988355585252619_dp, 0.0004140383710344944_dp, &
             0.0008298592025667272_dp, 0.0016097341693117417_dp, 0.0030348812273952334_dp, &
             0.00558017247924242_dp, 0.0100340901652149_dp, 0.017686058540336145_dp, &
             0.030615746220543045_dp, 0.05213543243522042_dp, 0.08746095529297568_dp, &
             0.1447248064447004_dp, 0.2365122818688088_dp, 0.38225082838017044_dp, &
             0.6122086132987152_dp, 0.9753328353500473_dp, 1.5591283926766577_dp, &
             2.5603893555790145_dp, 4.7129748111549175_dp]

      CASE (1004)

         aw(:) = & ! a
            [3.176084280897815e-08_dp, 2.809925323987775e-07_dp, 1.324451685229261e-06_dp, &
             4.781828008114672e-06_dp, 1.4653660456718943e-05_dp, 4.005196116209431e-05_dp, &
             0.00010049013875048396_dp, 0.00023575742892126763_dp, 0.0005237896368062412_dp, &
             0.001112169171704815_dp, 0.0022723929180991443_dp, 0.004491503788190833_dp, &
             0.008624072974309631_dp, 0.01614027846450238_dp, 0.02952502824348903_dp, &
             0.05291172072212475_dp, 0.09307666086087783_dp, 0.16098281103195272_dp, &
             0.27415484620782915_dp, 0.46031143260773066_dp, 0.7629253436396418_dp, &
             1.2498852706890915_dp, 2.0277686778396182_dp, 3.2687381314669053_dp, &
             5.2762489956821605_dp, 8.737267639880121_dp, &
             ! w
             9.325552834814655e-08_dp, 4.921458007962682e-07_dp, 1.8438493417411335e-06_dp, &
             5.694808038649421e-06_dp, 1.5483978069610705e-05_dp, 3.8408086668917325e-05_dp, &
             8.881487965420219e-05_dp, 0.00019420500704480522_dp, 0.0004055856320857199_dp, &
             0.0008149487690622053_dp, 0.0015842221208138153_dp, 0.002992420685533204_dp, &
             0.005511289000777368_dp, 0.009925002897200755_dp, 0.01751726042631872_dp, &
             0.03036043237653618_dp, 0.05175800375726225_dp, 0.08691605289906251_dp, &
             0.14395767502064197_dp, 0.23546168996008884_dp, 0.38085609780539714_dp, &
             0.6104219238921639_dp, 0.9731355360730988_dp, 1.5565433607320946_dp, &
             2.557465715674885_dp, 4.709642666425965_dp]

      CASE (1005)

         aw(:) = & ! a
            [2.8696648550828397e-08_dp, 2.615205495534795e-07_dp, 1.2555474964211257e-06_dp, &
             4.583490060110546e-06_dp, 1.4150516874439488e-05_dp, 3.8886103533575105e-05_dp, &
             9.797046053858556e-05_dp, 0.00023060866916394578_dp, 0.0005137476899219403_dp, &
             0.0010933475538228887_dp, 0.0022383184289575192_dp, 0.004431688365059123_dp, &
             0.008521956642064303_dp, 0.015970352164543313_dp, 0.029248938687012213_dp, &
             0.05247319548815225_dp, 0.09239520967289142_dp, 0.15994641805058724_dp, &
             0.27261234960576525_dp, 0.45806607711909814_dp, 0.7597321006855186_dp, &
             1.2454556701617865_dp, 2.0217870652719485_dp, 3.260890043369849_dp, &
             5.266250952792777_dp, 8.724849119816746_dp, &
             ! w
             8.514842125767189e-08_dp, 4.6433452833849786e-07_dp, 1.7653108380368226e-06_dp, &
             5.499213090653743e-06_dp, 1.5040568300805852e-05_dp, 3.747216759239332e-05_dp, &
             8.69483092306617e-05_dp, 0.0001906523567526129_dp, 0.0003990869500618729_dp, &
             0.000803464640920814_dp, 0.0015645417769315675_dp, 0.0029596208203120536_dp, &
             0.005458012570879126_dp, 0.009840538606075607_dp, 0.017386432662911767_dp, &
             0.03016237039977218_dp, 0.05146496663872449_dp, 0.08649266526856679_dp, &
             0.14336119680033282_dp, 0.23464428097593612_dp, 0.3797702952294165_dp, &
             0.609030255084238_dp, 0.9714232922916174_dp, 1.5545283268289154_dp, &
             2.555186349271952_dp, 4.7070448888563785_dp]

      CASE (1006)

         aw(:) = & ! a
            [2.63629942518937e-08_dp, 2.465886670719638e-07_dp, 1.20218377151429e-06_dp, &
             4.428782422283755e-06_dp, 1.3756059945379344e-05_dp, 3.796862002021246e-05_dp, &
             9.598169174090065e-05_dp, 0.00022653502804612425_dp, 0.0005057867803999663_dp, &
             0.0010784012161165883_dp, 0.0022112203200817205_dp, 0.004384059006560526_dp, &
             0.008440552936315522_dp, 0.01583475716342679_dp, 0.029028431628207785_dp, &
             0.0521226707289792_dp, 0.09185010744839554_dp, 0.15911683782326935_dp, &
             0.27137690396431086_dp, 0.45626668009451715_dp, 0.7571717792396536_dp, &
             1.2419024183772074_dp, 2.0169869066185813_dp, 3.254589883209914_dp, &
             5.258222633422846_dp, 8.714875078860178_dp, &
             ! w
             7.897145761935208e-08_dp, 4.428499802736919e-07_dp, 1.7040299780572567e-06_dp, &
             5.3457257109024575e-06_dp, 1.4691234037280902e-05_dp, 3.673259075181486e-05_dp, &
             8.546977159075211e-05_dp, 0.00018783268768212885_dp, 0.0003939204854436187_dp, &
             0.0007943216970930887_dp, 0.0015488540267759939_dp, 0.00293344644289406_dp, &
             0.005415456330307561_dp, 0.009773011010973085_dp, 0.017281755600357603_dp, &
             0.030003784221982032_dp, 0.05123018024687254_dp, 0.08615323446197692_dp, &
             0.1428827324660327_dp, 0.23398826173595816_dp, 0.37889846858221404_dp, &
             0.6079123771692972_dp, 0.9700474323685409_dp, 1.5529087485463002_dp, &
             2.553354067550396_dp, 4.704956696476055_dp]

      CASE (1007)

         aw(:) = & ! a
            [2.4529438521748676e-08_dp, 2.347855059623563e-07_dp, 1.159647112516243e-06_dp, &
             4.3047305664605045e-06_dp, 1.3438442960354012e-05_dp, 3.722756287504612e-05_dp, &
             9.43714515440851e-05_dp, 0.00022323026280345162_dp, 0.0004993179253988689_dp, &
             0.0010662394015379197_dp, 0.002189144434327567_dp, 0.004345216819607754_dp, &
             0.00837410677652943_dp, 0.015723987043677844_dp, 0.02884816375222955_dp, &
             0.05183592255546102_dp, 0.09140391771716953_dp, 0.15843742059919239_dp, &
             0.270364584073878_dp, 0.4547915907843429_dp, 0.7550720406467301_dp, &
             1.2389872848104957_dp, 2.0130474969555907_dp, 3.2494179900106244_dp, &
             5.2516305626510755_dp, 8.706683964235475_dp, &
             ! w
             7.411625542002126e-08_dp, 4.2575968774200727e-07_dp, 1.654879557989815e-06_dp, &
             5.222044083970586e-06_dp, 1.4408820218246699e-05_dp, 3.613321511172463e-05_dp, &
             8.426916196528803e-05_dp, 0.0001855393418506389_dp, 0.0003897126816957136_dp, &
             0.0007868665877583842_dp, 0.0015360493528890713_dp, 0.00291206329888623_dp, &
             0.005380662431807382_dp, 0.009717761247105052_dp, 0.017196055695904995_dp, &
             0.029873872409969925_dp, 0.051037743041279685_dp, 0.08587489091272248_dp, &
             0.14249019947069447_dp, 0.2334498384656949_dp, 0.3781826520305404_dp, &
             0.6069942309717372_dp, 0.9689170794659356_dp, 1.5515778885860558_dp, &
             2.551848256294776_dp, 4.703240602102145_dp]

      CASE (1008)

         aw(:) = & ! a
            [2.3055037366271908e-08_dp, 2.2524315032163081e-07_dp, 1.1250109310969348e-06_dp, &
             4.203214111811963e-06_dp, 1.317761524365091e-05_dp, 3.661742410741837e-05_dp, &
             9.304300021456431e-05_dp, 0.00022049937019305994_dp, 0.0004939651430452657_dp, &
             0.0010561643210257171_dp, 0.0021708382921015743_dp, 0.004312979713275232_dp, &
             0.008318917875782683_dp, 0.015631921370997275_dp, 0.0286982448022773_dp, &
             0.05159731927205979_dp, 0.09103245868669334_dp, 0.15787154071157386_dp, &
             0.2695210848474995_dp, 0.45356203334603384_dp, 0.7533212082709582_dp, &
             1.2365557978018051_dp, 2.0097607725344036_dp, 3.2451019728972086_dp, &
             5.246128345665338_dp, 8.699846097028562_dp, &
             ! w
             7.021058268659199e-08_dp, 4.118673016298507e-07_dp, 1.6146494462719496e-06_dp, &
             5.120413282241957e-06_dp, 1.4176126219971765e-05_dp, 3.56383432270676e-05_dp, &
             8.327626010540199e-05_dp, 0.00018364019629097647_dp, 0.000386224208480141_dp, &
             0.0007806799510242194_dp, 0.0015254144218179238_dp, 0.002894290355702903_dp, &
             0.005351723868468956_dp, 0.009671782079669023_dp, 0.017124697583850754_dp, &
             0.029765648532727274_dp, 0.05087736061187036_dp, 0.08564281703539799_dp, &
             0.14216279517834943_dp, 0.2330005949209192_dp, 0.37758520954061753_dp, &
             0.6062277044764331_dp, 0.9679731690831382_dp, 1.5504663513126933_dp, &
             2.550590484693577_dp, 4.7018072065809235_dp]

      CASE (1009)

         aw(:) = & ! a
            [1.6950925300941922e-08_dp, 1.8512469702023082e-07_dp, 9.765948837475134e-07_dp, &
             3.762570822068381e-06_dp, 1.2035276919132553e-05_dp, 3.392743547937393e-05_dp, &
             8.715580480355002e-05_dp, 0.0002083466216701058_dp, 0.00047006246038659797_dp, &
             0.0010110429565318437_dp, 0.0020886480676058712_dp, 0.00416792578404592_dp, &
             0.008070110703677505_dp, 0.01521615022368248_dp, 0.028020163637497274_dp, &
             0.050516619758622934_dp, 0.08934789714247594_dp, 0.1553023354280069_dp, &
             0.2656874176626726_dp, 0.44796839837719643_dp, 0.7453491951934768_dp, &
             1.225475888063848_dp, 1.994773280523565_dp, 3.2254092151877547_dp, &
             5.221011220048849_dp, 8.668620569340153_dp, &
             ! w
             5.4019823518473496e-08_dp, 3.525903920767117e-07_dp, 1.4399367103706628e-06_dp, &
             4.674617165176197e-06_dp, 1.3148314237770694e-05_dp, 3.3440956246506917e-05_dp, &
             7.884900549956665e-05_dp, 0.00017514302608273168_dp, 0.00037057106886326626_dp, &
             0.0007528513867465941_dp, 0.0014774740912708991_dp, 0.002814022028611975_dp, &
             0.0052208088038653636_dp, 0.009463464159725534_dp, 0.016800955649814573_dp, &
             0.029274044503458354_dp, 0.05014800390154651_dp, 0.08458633970195358_dp, &
             0.14067091819215147_dp, 0.23095173902804203_dp, 0.3748582906825795_dp, &
             0.602726555350653_dp, 0.963659252649939_dp, 1.5453841081675022_dp, &
             2.5448382872983824_dp, 4.6952520673997595_dp]

      CASE (1010)

         aw(:) = & ! a
            [1.6638079120434184e-08_dp, 1.8303560195222553e-07_dp, 9.687254852298032e-07_dp, &
             3.7389290792019776e-06_dp, 1.1973489220109e-05_dp, 3.3781069709670776e-05_dp, &
             8.68339986656858e-05_dp, 0.00020767987051717866_dp, 0.0004687470535353466_dp, &
             0.0010085534493369784_dp, 0.0020841033078915854_dp, 0.00415988949808476_dp, &
             0.008056302924136456_dp, 0.015193041898562215_dp, 0.027982425468244553_dp, &
             0.050456401034446936_dp, 0.08925392663621129_dp, 0.155158872857305_dp, &
             0.2654731527003931_dp, 0.4476555074111315_dp, 0.7449029267545827_dp, &
             1.224855217726335_dp, 1.9939332097233924_dp, 3.2243048377395063_dp, &
             5.2196020530995515_dp, 8.666868147708719_dp, &
             ! w
             5.318868744626686e-08_dp, 3.4945918122391136e-07_dp, 1.4305587424661624e-06_dp, &
             4.6504720775060675e-06_dp, 1.3092298484954104e-05_dp, 3.3320635920296754e-05_dp, &
             7.860568651872461e-05_dp, 0.00017467461085176437_dp, 0.0003697059802303021_dp, &
             0.0007513100691572613_dp, 0.0014748138594482874_dp, 0.0028095605349709373_dp, &
             0.0052135215681561615_dp, 0.009451853132067834_dp, 0.016782889801361688_dp, &
             0.029246581804004673_dp, 0.05010721933494469_dp, 0.0845272095628814_dp, &
             0.14058734964400246_dp, 0.23083688331264704_dp, 0.37470531773777044_dp, &
             0.6025300293209096_dp, 0.9634169796585248_dp, 1.545098576519601_dp, &
             2.5445150510338057_dp, 4.694883722789503_dp]

      CASE (1011)

         aw(:) = & ! a
            [4.133114668827588e-05_dp, 0.00022288422647064897_dp, 0.0005722500412080675_dp, &
             0.0011365051344008781_dp, 0.0020024427143605157_dp, 0.003317529283367159_dp, &
             0.005315993003722334_dp, 0.008353073591163749_dp, 0.012955066872823568_dp, &
             0.0198935766090916_dp, 0.03029293738457279_dp, 0.04578283545975285_dp, &
             0.06871321925219903_dp, 0.1024553158776816_dp, 0.15182147045059854_dp, &
             0.22364855693380606_dp, 0.32760630033354343_dp, 0.47731544725862374_dp, &
             0.6918962569705502_dp, 0.9981265915587121_dp, 1.433499530894329_dp, &
             2.050709804377275_dp, 2.9246867888040753_dp, 4.164921201624245_dp, &
             5.9410346956099644_dp, 8.550348782636469_dp, 12.684884831267697_dp, &
             ! w
             0.00010663577847824833_dp, 0.0002600677285126925_dp, 0.0004462614740991329_dp, &
             0.0006963595781430342_dp, 0.001059665527236895_dp, 0.001608995574289536_dp, &
             0.0024463241525708797_dp, 0.0037143355279016365_dp, 0.005616489210303001_dp, &
             0.008445415779015034_dp, 0.012621394172473247_dp, 0.01874511978764135_dp, &
             0.027670626460677458_dp, 0.040606044213690164_dp, 0.05925246976764002_dp, &
             0.08599496648369329_dp, 0.12416524566715859_dp, 0.17840443722654367_dp, &
             0.2551703458356085_dp, 0.36346673323426393_dp, 0.5159498918773728_dp, &
             0.7307684564801469_dp, 1.0350565159676461_dp, 1.4727627382852992_dp, &
             2.126016463967528_dp, 3.191491831744027_dp, 5.4355002435163255_dp]

      CASE (1012)

         aw(:) = & ! a
            [2.265199677518572e-05_dp, 0.00012276899544090978_dp, 0.0003183419308487753_dp, &
             0.0006423569924739833_dp, 0.0011572917670961626_dp, 0.0019709003624358437_dp, &
             0.0032563154740661536_dp, 0.0052807693555526536_dp, 0.008449529858618562_dp, &
             0.013371806281620383_dp, 0.02095712759715817_dp, 0.032554484153668795_dp, &
             0.05015179288902161_dp, 0.07666027133969103_dp, 0.11631783981237262_dp, &
             0.17525874343944897_dp, 0.2623146833760234_dp, 0.390138323022376_dp, &
             0.5767778193874423_dp, 0.8478916342891016_dp, 1.239903109976466_dp, &
             1.8046275240320706_dp, 2.6164751246680287_dp, 3.7849253510314402_dp, &
             5.480095553876378_dp, 7.999844070821326_dp, 12.034456801672523_dp, &
             ! w
             5.8508820083244256e-05_dp, 0.00014414371476422158_dp, 0.00025230438272487633_dp, &
             0.0004058732799564563_dp, 0.0006416322037341139_dp, 0.0010139785941349164_dp, &
             0.001600673042559488_dp, 0.002514700378198619_dp, 0.003922777968903744_dp, &
             0.006071085147745618_dp, 0.009321122344390921_dp, 0.014200250553697977_dp, &
             0.021472956898305726_dp, 0.03224097897434797_dp, 0.048083391826744905_dp, &
             0.07125193872093943_dp, 0.10494290980217887_dp, 0.15367615432779677_dp, &
             0.22382783186584307_dp, 0.32439554826773487_dp, 0.46814862907720634_dp, &
             0.6735087489637548_dp, 0.9680549833035281_dp, 1.3962844112141117_dp, &
             2.0407239117987688_dp, 3.0975076650112148_dp, 5.3274244904589585_dp]

      CASE (1013)

         aw(:) = & ! a
            [1.5930005105884995e-05_dp, 8.663411175959597e-05_dp, 0.0002261837933722728_dp, &
             0.0004614286035206294_dp, 0.0008439518082670551_dp, 0.0014633743227644115_dp, &
             0.002464830928593016_dp, 0.004075124572650641_dp, 0.006643334297919195_dp, &
             0.010702033121911687_dp, 0.017057545864950126_dp, 0.02692160183418139_dp, &
             0.042102054413649725_dp, 0.06527754003404436_dp, 0.10039083939974913_dp, &
             0.15320935650564652_dp, 0.232120090055218_dp, 0.34925322966436173_dp, &
             0.5220677314182667_dp, 0.7755941878823863_dp, 1.145640874222958_dp, &
             1.6834993820955622_dp, 2.463247734197893_dp, 3.594319334805111_dp, &
             5.247091844654251_dp, 7.719762398705776_dp, 11.701828983405496_dp, &
             ! w
             4.1177814957352994e-05_dp, 0.00010215054011081087_dp, 0.00018125542596989232_dp, &
             0.0002976097990460291_dp, 0.0004820949764557738_dp, 0.0007805172353090585_dp, &
             0.0012595332509567295_dp, 0.002018233079954551_dp, 0.003205473162508219_dp, &
             0.005044159087528887_dp, 0.007865700792910102_dp, 0.012159223700477275_dp, &
             0.018641678072609098_dp, 0.028357239422026687_dp, 0.042817563296153396_dp, &
             0.06419888682736664_dp, 0.0956183006446552_dp, 0.1415211119286643_dp, &
             0.20822739362048467_dp, 0.30471744005396384_dp, 0.44380815317579525_dp, &
             0.6440603134910088_dp, 0.9332965752824595_dp, 1.356331750428908_dp, &
             1.9959513198168084_dp, 3.048068109993711_dp, 5.270645994332936_dp]

      CASE (1014)

         aw(:) = & ! a
            [1.240889092145128e-05_dp, 6.76671091268666e-05_dp, 0.00017762160245782312_dp, &
             0.00036550248004403347_dp, 0.0006763538692074322_dp, 0.001188816546413399_dp, &
             0.0020310587228498335_dp, 0.0034053040718067177_dp, 0.005626177878735291_dp, &
             0.0091787448466855_dp, 0.014804637047488132_dp, 0.02362862298990212_dp, &
             0.03734331499822549_dp, 0.05847703452884025_dp, 0.09077998560884071_dp, &
             0.13977793255427248_dp, 0.21356213175752478_dp, 0.32391186013012424_dp, &
             0.48788615771042404_dp, 0.7300844292903402_dp, 1.0858865799076862_dp, &
             1.606212352488915_dp, 2.3648949144149687_dp, 3.4713195980057203_dp, &
             5.096034856403628_dp, 7.537483826966416_dp, 11.484690958024949_dp, &
             ! w
             3.209523556566068e-05_dp, 8.005307973745117e-05_dp, 0.00014357689284989373_dp, &
             0.0002394976888656755_dp, 0.0003950589838930408_dp, 0.0006508310711966158_dp, &
             0.0010667311727420943_dp, 0.0017332493022486793_dp, 0.0027879358542838_dp, &
             0.0044387353063556545_dp, 0.006997490531757815_dp, 0.010928201154780473_dp, &
             0.016916248555250018_dp, 0.02596715717716741_dp, 0.03954676484811979_dp, &
             0.0597792801010827_dp, 0.08972626093635792_dp, 0.13377939308098147_dp, &
             0.19821628079861398_dp, 0.2919998895221959_dp, 0.4279734467837606_dp, &
             0.6247874756186418_dp, 0.9104292207224619_dp, 1.3299354585510081_dp, &
             1.9662831184567857_dp, 3.0152638779267087_dp, 5.232999568682095_dp]

      CASE (1015)

         aw(:) = & ! a
            [1.0223540928768244e-05_dp, 5.5876327283269576e-05_dp, 0.00014733957605593857_dp, &
             0.0003053918596320136_dp, 0.0005705924768091414_dp, 0.0010140142990323277_dp, &
             0.0017521274582417957_dp, 0.002970151217082029_dp, 0.00495870735070803_dp, &
             0.008169490356635136_dp, 0.013298325293877252_dp, 0.021407936085750322_dp, &
             0.034108116238316104_dp, 0.05381840802569134_dp, 0.08414870439032592_dp, &
             0.1304475271025096_dp, 0.2005877378754264_dp, 0.3060876737670577_dp, &
             0.4637068009925585_dp, 0.6977188845167406_dp, 1.043177174915754_dp, &
             1.5507141223286136_dp, 2.293969059792534_dp, 3.3822821296037353_dp, &
             4.986326271706307_dp, 7.404736942557135_dp, 11.32621477560715_dp, &
             ! w
             2.6456127295513986e-05_dp, 6.62889792871176e-05_dp, 0.00011996267680836016_dp, &
             0.00020272597601243074_dp, 0.0003392847646235578_dp, 0.0005665923061765598_dp, &
             0.0009399058016811227_dp, 0.0015436703493629995_dp, 0.0025073688411316026_dp, &
             0.00402816319932474_dp, 0.006403683131362698_dp, 0.010079542136917337_dp, &
             0.015717848121982025_dp, 0.024295409254463527_dp, 0.0372437371039372_dp, &
             0.0566476893397916_dp, 0.0855263277738046_dp, 0.12822961246401526_dp, &
             0.19100103793311565_dp, 0.2827876396394856_dp, 0.41644916662806325_dp, &
             0.610700804133677_dp, 0.8936525855943015_dp, 1.3105108102105707_dp, &
             1.9444042040818155_dp, 2.9910487933809717_dp, 5.205223868786751_dp]

      CASE (1016)

         aw(:) = & ! a
            [8.727273674917548e-06_dp, 4.7792436114632105e-05_dp, 0.00012652378347591833_dp, &
             0.00026390047051243524_dp, 0.0004971583487991961_dp, 0.0008917450698057281_dp, &
             0.001555432572989849_dp, 0.002660754440823174_dp, 0.004480319340654578_dp, &
             0.007440616143799667_dp, 0.012202652833705552_dp, 0.019781715049912406_dp, &
             0.031723918509701556_dp, 0.050364715904059866_dp, 0.07920496621241908_dp, &
             0.12345474533656438_dp, 0.1908154389871886_dp, 0.29259941227369274_dp, &
             0.4453283336858533_dp, 0.6730158762720717_dp, 1.0104524774319095_dp, &
             1.5080371702737234_dp, 2.2392489882684807_dp, 3.313386474160821_dp, &
             4.9012191295438825_dp, 7.301539793246007_dp, 11.202810314075098_dp, &
             ! w
             2.2593943669260506e-05_dp, 5.683645477453563e-05_dp, 0.0001036613454234445_dp, &
             0.0001771360696133468_dp, 0.00030006391159141297_dp, 0.0005067076799329426_dp, &
             0.0008488470371810808_dp, 0.0014063541860811727_dp, 0.0023025418091215436_dp, &
             0.003726265079658037_dp, 0.005964138785013461_dp, 0.009447446330501801_dp, &
             0.014820052117746589_dp, 0.023036120406281942_dp, 0.03549992286424345_dp, &
             0.054264866181846255_dp, 0.08231575009495752_dp, 0.12396848107981792_dp, &
             0.1854380898065737_dp, 0.27565716822563574_dp, 0.407496608300226_dp, &
             0.5997213168950433_dp, 0.8805384358917568_dp, 1.2952907778060012_dp, &
             1.9272326688375536_dp, 2.9720291907799243_dp, 5.183415666643532_dp]

      CASE (1017)

         aw(:) = & ! a
            [7.634695215101426e-06_dp, 4.18825900310198e-05_dp, 0.00011127155589772385_dp, &
             0.00023338846923811062_dp, 0.000442878044053161_dp, 0.0008007935114935354_dp, &
             0.0014081088941739503_dp, 0.002427412201719985_dp, 0.004117118046229363_dp, &
             0.006883746435844753_dp, 0.01136058020349976_dp, 0.01852494717865621_dp, &
             0.02987177824680481_dp, 0.04766863771121763_dp, 0.07532798299212069_dp, &
             0.11794717177489715_dp, 0.18308739394395435_dp, 0.28189189308881496_dp, &
             0.4306862020648562_dp, 0.6532684048738324_dp, 0.9842098231967747_dp, &
             1.4737133706638015_dp, 2.1951215282037992_dp, 3.257694636436478_dp, &
             4.832280258914759_dp, 7.217804103278141_dp, 11.102542953703383_dp, &
             ! w
             1.9773009620194447e-05_dp, 4.9916067561900825e-05_dp, 9.16728514439909e-05_dp, &
             0.00015818390169340446_dp, 0.0002707564834876735_dp, 0.000461551434310484_dp, &
             0.0007796197133890599_dp, 0.0013012040914587013_dp, 0.0021446785646228863_dp, &
             0.0034922118403545896_dp, 0.005621509848737885_dp, 0.008952212246681059_dp, &
             0.014113290945894693_dp, 0.022040336976209742_dp, 0.0341151650393001_dp, &
             0.05236510536911372_dp, 0.07974635467222059_dp, 0.12054611971679802_dp, &
             0.18095503549814676_dp, 0.26989258207035016_dp, 0.4002375169727702_dp, &
             0.5907946627938913_dp, 0.869850986907248_dp, 1.2828631790667058_dp, &
             1.9131925658213558_dp, 2.9564682389481747_dp, 5.165578431314193_dp]

      CASE (1018)

         aw(:) = & ! a
            [6.799711235012341e-06_dp, 3.7361354633700115e-05_dp, 9.957947254448849e-05_dp, &
             0.0002099227073059153_dp, 0.00040094161028759475_dp, 0.0007301330868347708_dp, &
             0.0012929669202120357_dp, 0.002243954252451219_dp, 0.0038299313825976068_dp, &
             0.0064410546017474035_dp, 0.010687788549118879_dp, 0.017516094336635833_dp, &
             0.02837844324702902_dp, 0.04548587637015833_dp, 0.07217698601496424_dp, &
             0.1134546161511556_dp, 0.17676199097776782_dp, 0.2730995400527115_dp, &
             0.4186265473792023_dp, 0.6369576060949826_dp, 0.9624766463202936_dp, &
             1.4452177877127192_dp, 2.158404668439873_dp, 3.2112624524878743_dp, &
             4.774703705183174_dp, 7.1477688750728365_dp, 11.018586146362084_dp, &
             ! w
             1.7616646117932288e-05_dp, 4.461493874379961e-05_dp, 8.245248664300384e-05_dp, &
             0.00014351648725796493_dp, 0.00024789737902234326_dp, 0.00042605516793132347_dp, &
             0.0007248221610441795_dp, 0.0012174613056593114_dp, 0.002018264346258695_dp, &
             0.0033038475127682198_dp, 0.005344489195173677_dp, 0.008550083694002198_dp, &
             0.01353708991379313_dp, 0.02122543080342086_dp, 0.03297789647354285_dp, &
             0.050799622269392795_dp, 0.07762232191991819_dp, 0.11770844687256254_dp, &
             0.17722730637565837_dp, 0.26508641687843854_dp, 0.39417025816712203_dp, &
             0.583316698481399_dp, 0.8608801601929675_dp, 1.2724147545931694_dp, &
             1.9013749164935645_dp, 2.9433634809003704_dp, 5.1505602940292095_dp]

      CASE (1019)

         aw(:) = & ! a
            [6.1395417881484415e-06_dp, 3.378330751979996e-05_dp, 9.030955268263924e-05_dp, &
             0.00019126343967728574_dp, 0.00036745700300121214_dp, 0.000673432056752497_dp, &
             0.0012000820294391706_dp, 0.0020951839599493162_dp, 0.0035958826736603193_dp, &
             0.006078582798442805_dp, 0.010134501349156695_dp, 0.016683048305770786_dp, &
             0.02714063496044963_dp, 0.0436701463813738_dp, 0.06954705096992769_dp, &
             0.10969317959060103_dp, 0.17145033759904488_dp, 0.26569579709253305_dp, &
             0.40844501522698434_dp, 0.6231532848612138_dp, 0.9440411837113356_dp, &
             1.4209948993777566_dp, 2.127132943760494_dp, 3.171647951945379_dp, &
             4.725507832636439_dp, 7.087853839608335_dp, 10.946691631028461_dp, &
             ! w
             1.591137768041326e-05_dp, 4.041483047325665e-05_dp, 7.512051138507467e-05_dp, &
             0.00013178708504117987_dp, 0.00022948982935111316_dp, 0.00039727576552600846_dp, &
             0.0006801247749489536_dp, 0.0011487909519981932_dp, 0.0019141097499693535_dp, &
             0.003147978035748912_dp, 0.005114339469158271_dp, 0.00821474908851756_dp, &
             0.013054922396612592_dp, 0.020541283992213315_dp, 0.03202017473589182_dp, &
             0.04947746291488608_dp, 0.07582350771117832_dp, 0.11529902939205904_dp, &
             0.17405441003731065_dp, 0.2609861860864805_dp, 0.38898308357892836_dp, &
             0.5769109678235358_dp, 0.8531824556277925_dp, 1.2634365984125957_dp, &
             1.8912101693813979_dp, 2.9320864250312906_dp, 5.137639333493501_dp]

      CASE (1020)

         aw(:) = & ! a
            [5.603667028771084e-06_dp, 3.087640355676106e-05_dp, 8.276575797200318e-05_dp, &
             0.00017603762482989614_dp, 0.00034003045822112956_dp, 0.0006267796957423577_dp, &
             0.00112329481386439_dp, 0.0019716225618062015_dp, 0.0034006318025334974_dp, &
             0.005774942705550669_dp, 0.009669223839769881_dp, 0.01597998728856444_dp, &
             0.02609245562819452_dp, 0.04212774759630914_dp, 0.06730643643223058_dp, &
             0.10647971524037368_dp, 0.16690072741036277_dp, 0.2593387916882687_dp, &
             0.3996829422383991_dp, 0.6112480215113266_dp, 0.9281100739740955_dp, &
             1.4000237909644244_dp, 2.100013494392458_dp, 3.1372416013711932_dp, &
             4.682723964609551_dp, 7.035691606989607_dp, 10.884047055781624_dp, &
             ! w
             1.4526899396791722e-05_dp, 3.699892843585587e-05_dp, 6.913756455117421e-05_dp, &
             0.00012216632300348179_dp, 0.00021429669440599125_dp, 0.00037337750744536295_dp, &
             0.0006428098938279197_dp, 0.0010911941164545888_dp, 0.0018263843904255505_dp, &
             0.0030161932306861203_dp, 0.004919065493233738_dp, 0.007929296216031927_dp, &
             0.012643220556482752_dp, 0.019955442197383218_dp, 0.031197851672182128_dp, &
             0.04833933300292733_dp, 0.07427134369500363_dp, 0.11321526310864316_dp, &
             0.17130447153509062_dp, 0.2574253707582148_dp, 0.3844698762941206_dp, &
             0.5713280031973806_dp, 0.8464633891520393_dp, 1.2555902479324408_dp, &
             1.882319099902882_dp, 2.9222183918906004_dp, 5.126334767750104_dp]

      CASE (1021)

         aw(:) = & ! a
            [3.07556759937867e-06_dp, 1.7120389080799702e-05_dp, 4.6853128112836807e-05_dp, &
             0.00010285090387188242_dp, 0.0002064251564884343_dp, 0.0003959434946189406_dp, &
             0.0007371624578557768_dp, 0.001340455688058299_dp, 0.002388384628065489_dp, &
             0.0041788443496681375_dp, 0.007191765955163697_dp, 0.012191217406834163_dp, &
             0.02038025663073623_dp, 0.033633774376380626_dp, 0.05484578928210916_dp, &
             0.08844357053828202_dp, 0.14114347983097553_dp, 0.2230554104887353_dp, &
             0.3492889284995764_dp, 0.5422841074022949_dp, 0.8352061696480149_dp, &
             1.276968415268258_dp, 1.9399768148925256_dp, 2.9331728491314037_dp, &
             4.427852418603856_dp, 6.7238232066155925_dp, 10.508444012839721_dp, &
             ! w
             7.990802942553031e-06_dp, 2.0773608757718692e-05_dp, 4.0379093148952514e-05_dp, &
             7.506951892876223e-05_dp, 0.0001383080586862562_dp, 0.00025140850434764015_dp, &
             0.0004489776719220188_dp, 0.0007873351786187929_dp, 0.0013570877229921278_dp, &
             0.0023021589521644795_dp, 0.0038485068239294635_dp, 0.006347110013377911_dp, &
             0.010337784120358736_dp, 0.016643193773484617_dp, 0.026506365965223846_dp, &
             0.04179053873944084_dp, 0.06526800790339253_dp, 0.10103603842558948_dp, &
             0.15511568011466423_dp, 0.23632081159919982_dp, 0.3575516109837995_dp, &
             0.5378371660619029_dp, 0.8059525909923124_dp, 1.208085676942769_dp, &
             1.8283301837108623_dp, 2.862212837539241_dp, 5.057632827916134_dp]

      CASE (1022)

         aw(:) = & ! a
            [2.1673609335192107e-06_dp, 1.215224051225794e-05_dp, 3.374660631523813e-05_dp, &
             7.568699457885438e-05_dp, 0.00015570215911308465_dp, 0.0003060569597332415_dp, &
             0.0005829440518138641_dp, 0.0010822478805588526_dp, 0.00196496222304419_dp, &
             0.0034973790885279733_dp, 0.0061138619253917675_dp, 0.01051385631305701_dp, &
             0.017810334922367325_dp, 0.029754857523013108_dp, 0.04907586053909264_dp, &
             0.07998308288769955_dp, 0.12891418626142448_dp, 0.20563260607688139_dp, &
             0.32483332415567084_dp, 0.508484949133822_dp, 0.7892543212254228_dp, &
             1.2155853711805809_dp, 1.859528113177115_dp, 2.8298812799801762_dp, &
             4.298077884309647_dp, 6.56424790623646_dp, 10.315523815734926_dp, &
             ! w
             5.6399182085561305e-06_dp, 1.4875539568427877e-05_dp, 2.9706600466251618e-05_dp, &
             5.7042709795691895e-05_dp, 0.00010821411645355367_dp, 0.00020160556376608125_dp, &
             0.00036773763782613987_dp, 0.0006570506588635632_dp, 0.0011517408890466632_dp, &
             0.001983909900924021_dp, 0.003363214566695296_dp, 0.005618603943735307_dp, &
             0.009260765054366163_dp, 0.015074789762713156_dp, 0.02425664980358899_dp, &
             0.03861280525257212_dp, 0.060850405521274874_dp, 0.09499735173872459_dp, &
             0.14700970906083427_dp, 0.22565568898541846_dp, 0.3438316948268284_dp, &
             0.520633598989393_dp, 0.7849996799831586_dp, 1.1833765833329497_dp, &
             1.8001353196528784_dp, 2.830815316819402_dp, 5.0217109329159_dp]

      CASE (1023)

         aw(:) = & ! a
            [1.691730705048351e-06_dp, 9.54071766830379e-06_dp, 2.6806041134225265e-05_dp, &
             6.113045145136684e-05_dp, 0.0001280997244778255_dp, 0.00025632773264514105_dp, &
             0.0004962458692015434_dp, 0.0009349135417337102_dp, 0.0017200427656529288_dp, &
             0.00309826726310608_dp, 0.005475344082345973_dp, 0.009509803511055041_dp, &
             0.016257124566554013_dp, 0.0273895590216702_dp, 0.045528294014298316_dp, &
             0.0747411835192999_dp, 0.12128285278497943_dp, 0.19468762759706384_dp, &
             0.30937437548335495_dp, 0.4869953114925414_dp, 0.7598799169444655_dp, &
             1.1761509244737791_dp, 1.807610521952662_dp, 2.7629522010321566_dp, &
             4.213695305352366_dp, 6.460190203652438_dp, 10.189442245039029_dp, &
             ! w
             4.407722571378931e-06_dp, 1.176115655626714e-05_dp, 2.39884409008301e-05_dp, &
             4.717891004010905e-05_dp, 9.138709728583761e-05_dp, 0.0001732345639523079_dp, &
             0.0003207228941348957_dp, 0.0005806111503735416_dp, 0.0010297763010637384_dp, &
             0.001792777731541497_dp, 0.0030687864783234832_dp, 0.005172467240627817_dp, &
             0.008595471468440743_dp, 0.014098145427171844_dp, 0.022845222442037023_dp, &
             0.036605148453676545_dp, 0.0580410333780952_dp, 0.09113331477230503_dp, &
             0.14179276505536956_dp, 0.21875441800278994_dp, 0.33490893200321337_dp, &
             0.5093937512720524_dp, 0.7712547059516909_dp, 1.1671134596324746_dp, &
             1.7815335816178006_dp, 2.810076525420082_dp, 4.997993167777528_dp]

      CASE (1024)

         aw(:) = & ! a
            [1.3964676148401791e-06_dp, 7.914733199522862e-06_dp, 2.2459042745629518e-05_dp, &
             5.192676184508766e-05_dp, 0.00011043851426979227_dp, 0.00022411075108363782_dp, &
             0.0004394103291024476_dp, 0.0008372733563690645_dp, 0.0015561229648987992_dp, &
             0.0028287456266441343_dp, 0.00504061602155553_dp, 0.008821079052363718_dp, &
             0.015184370432406967_dp, 0.025745545690322027_dp, 0.043048050035191944_dp, &
             0.07105638251543377_dp, 0.11589119331008618_dp, 0.18691830015098546_dp, &
             0.2983524845018051_dp, 0.47161086678594194_dp, 0.73877075630945_dp, &
             1.1477129954359122_dp, 1.7700509507293882_dp, 2.71439494346417_dp, &
             4.152325422262254_dp, 6.384358529639266_dp, 10.097417649445605_dp, &
             ! w
             3.642289081161589e-06_dp, 9.815110147232306e-06_dp, 2.0373579861892006e-05_dp, &
             4.084095814010183e-05_dp, 8.040068758765127e-05_dp, 0.00015446011969151734_dp, &
             0.00028925627836834194_dp, 0.0005289433004641359_dp, 0.0009466080619320775_dp, &
             0.0016614027961660265_dp, 0.0028649367693126273_dp, 0.004861515254945349_dp, &
             0.008128908421998236_dp, 0.013409318745026535_dp, 0.021844447823907917_dp, &
             0.035174549175898254_dp, 0.056029854950862816_dp, 0.08835507587768701_dp, &
             0.1380264727693674_dp, 0.21375312503577767_dp, 0.32841975553615993_dp, &
             0.5011929946321245_dp, 0.7611975412622868_dp, 1.15518574690366_dp, &
             1.767867622333308_dp, 2.7948279390360464_dp, 4.980558958520095_dp]

      CASE (1025)

         aw(:) = & ! a
            [1.1942234450519989e-06_dp, 6.798219946527863e-06_dp, 1.9459054656655945e-05_dp, &
             4.55242257725863e-05_dp, 9.803164701017266e-05_dp, 0.00020125084716356384_dp, &
             0.0003987014727301214_dp, 0.000766737481734197_dp, 0.0014367880712821733_dp, &
             0.002631155676368643_dp, 0.004719881066623696_dp, 0.008309996697234727_dp, &
             0.014384069954118965_dp, 0.024513058207663523_dp, 0.04118023330524608_dp, &
             0.06826979454430288_dp, 0.1117979189054743_dp, 0.18099853910797675_dp, &
             0.28992610432002736_dp, 0.4598123019397362_dp, 0.7225346404759638_dp, &
             1.1257812310755833_dp, 1.7410136867637176_dp, 2.676773715708587_dp, &
             4.104688097668651_dp, 6.325404890426148_dp, 10.025790179284742_dp, &
             ! w
             3.1177034786852475e-06_dp, 8.474775154204022e-06_dp, 1.7859243856699156e-05_dp, &
             3.637321832747758e-05_dp, 7.255733868504639e-05_dp, 0.00014091503560671306_dp, &
             0.00026635256685154994_dp, 0.0004910450960806109_dp, 0.0008851850047830547_dp, &
             0.001563775367008409_dp, 0.0027125966070728116_dp, 0.00462793382879038_dp, &
             0.0077767650188767514_dp, 0.012887127682260247_dp, 0.02108266638710988_dp, &
             0.03408143313603231_dp, 0.05448763969262228_dp, 0.0862175473337454_dp, &
             0.13511968719572154_dp, 0.2098818897227727_dp, 0.3233831950464604_dp, &
             0.4948122242697113_dp, 0.7533552427288275_dp, 1.1458680534841634_dp, &
             1.7571781838064222_dp, 2.7828929483013316_dp, 4.9669160397426655_dp]

      CASE (1026)

         aw(:) = & ! a
            [1.0464712042947327e-06_dp, 5.980765144499402e-06_dp, 1.7252912521418522e-05_dp, &
             4.0783161490628e-05_dp, 8.876738629846045e-05_dp, 0.00018403740756034773_dp, &
             0.0003678080831135085_dp, 0.000712830378331724_dp, 0.0013450064289208113_dp, &
             0.002478316125274995_dp, 0.004470498011029695_dp, 0.007910732144522105_dp, &
             0.013756158386157673_dp, 0.023542208712553594_dp, 0.039703526423184164_dp, &
             0.06605922025338419_dp, 0.10854053067202926_dp, 0.17627385180523522_dp, &
             0.2831825041882665_dp, 0.4503460322167156_dp, 0.7094774230739503_dp, &
             1.1081053799498974_dp, 1.717565143051345_dp, 2.646340127338585_dp, &
             4.066093876570879_dp, 6.277583419982657_dp, 9.967632598319918_dp, &
             ! w
             2.734274551489556e-06_dp, 7.490860523813445e-06_dp, 1.599759316674891e-05_dp, &
             3.3027373538925784e-05_dp, 6.662155784032936e-05_dp, 0.0001305754517744195_dp, &
             0.0002487420572117283_dp, 0.0004617213609950396_dp, 0.0008373921287557189_dp, &
             0.001487428059682766_dp, 0.002592915139280559_dp, 0.0044436564421224_dp, &
             0.007497877837941563_dp, 0.012472090964712896_dp, 0.020475197790728546_dp, &
             0.03320705940605024_dp, 0.053250483139833285_dp, 0.08449821410559463_dp, &
             0.1327757008439797_dp, 0.20675283025921906_dp, 0.3193033205979091_dp, &
             0.4896331534634429_dp, 0.7469786775266661_dp, 1.138280811182018_dp, &
             1.7484648203675952_dp, 2.7731592386245523_dp, 4.9557911742029015_dp]

      CASE (1027)

         aw(:) = & ! a
            [9.334940484340804e-07_dp, 5.354497799872867e-06_dp, 1.555605701411221e-05_dp, &
             3.711403620499326e-05_dp, 8.154528636899452e-05_dp, 0.00017052106298692595_dp, &
             0.0003433880813302074_dp, 0.0006699632905117456_dp, 0.001271628922970902_dp, &
             0.0023555330779712674_dp, 0.00426928090829956_dp, 0.0075873007430232705_dp, &
             0.01324565918686754_dp, 0.02275026452866146_dp, 0.03849524042073429_dp, &
             0.06424532184076445_dp, 0.10586062217346191_dp, 0.17237726034028275_dp, &
             0.2776082029524206_dp, 0.4425046157711749_dp, 0.6986402497193854_dp, &
             1.0934084035518217_dp, 1.6980363736627428_dp, 2.620956977737536_dp, &
             4.033863893529919_dp, 6.237606604502866_dp, 9.918976607722271_dp, &
             ! w
             2.440963409080761e-06_dp, 6.735292344884374e-06_dp, 1.4556990779779755e-05_dp, &
             3.0412437742943997e-05_dp, 6.194078700143371e-05_dp, 0.0001223622925883734_dp, &
             0.00023466745662608176_dp, 0.0004381604321687307_dp, 0.0007988099305001273_dp, &
             0.0014255320748368298_dp, 0.0024955126739356678_dp, 0.0042931535862008405_dp, &
             0.007269367128492974_dp, 0.012131006670921922_dp, 0.019974585716419146_dp, &
             0.0324846342799765_dp, 0.0522258613200382_dp, 0.0830710485959085_dp, &
             0.13082593709502557_dp, 0.20414491725289174_dp, 0.3158967544682297_dp, &
             0.48530159829683467_dp, 0.7416377516096644_dp, 1.1319181146456272_dp, &
             1.741151348786439_dp, 2.7649857955486956_dp, 4.946450781717605_dp]

      CASE (1028)

         aw(:) = & ! a
            [8.441211641730451e-07_dp, 4.858207942250597e-06_dp, 1.4206528664414055e-05_dp, &
             3.417966429409421e-05_dp, 7.573188643794205e-05_dp, 0.0001595718551868552_dp, &
             0.00032349106371144484_dp, 0.0006348540020901873_dp, 0.0012112510891073285_dp, &
             0.002254080489963278_dp, 0.004102393417179351_dp, 0.007318131840783266_dp, &
             0.012819480113685465_dp, 0.02208723394180414_dp, 0.03748097367526316_dp, &
             0.06271898000087829_dp, 0.10360047076902992_dp, 0.16908411408601043_dp, &
             0.2728879977895823_dp, 0.4358526631543596_dp, 0.6894315779878049_dp, &
             1.0809007047240158_dp, 1.6813933600106372_dp, 2.5992977217281976_dp, &
             4.006332775717232_dp, 6.203428036405983_dp, 9.87734953214134_dp, &
             ! w
             2.208843275547235e-06_dp, 6.135260827516492e-06_dp, 1.340496256663769e-05_dp, &
             2.830282523663812e-05_dp, 5.8135050065726424e-05_dp, 0.00011564223621869513_dp, &
             0.00022309042213660292_dp, 0.0004186911465422955_dp, 0.0007667977662685246_dp, &
             0.0013739877507456377_dp, 0.002414130428439646_dp, 0.004167023204621714_dp, &
             0.007077328591464908_dp, 0.011843626812300917_dp, 0.019551794107026794_dp, &
             0.031873163586931284_dp, 0.051356822052345354_dp, 0.08185826147561384_dp, &
             0.12916607293781923_dp, 0.20192103328817493_dp, 0.31298730066731956_dp, &
             0.4815968788875952_dp, 0.7370639953435186_dp, 1.1264636930067569_dp, &
             1.7348771730028434_dp, 2.757971240663341_dp, 4.938435612332332_dp]

      CASE (1029)

         aw(:) = & ! a
            [7.715351902207404e-07_dp, 4.454488657860146e-06_dp, 1.3105081443546808e-05_dp, &
             3.177251563809805e-05_dp, 7.093507588008882e-05_dp, 0.00015048606873562504_dp, &
             0.00030689522743429385_dp, 0.0006054352971308557_dp, 0.0011604522626860166_dp, &
             0.002168410367058879_dp, 0.003961001827656396_dp, 0.007089400993579225_dp, &
             0.012456337353068097_dp, 0.021520857873617932_dp, 0.036612567967373125_dp, &
             0.061409361772741256_dp, 0.10165741874646757_dp, 0.1662478270145436_dp, &
             0.268815734552769_dp, 0.43010478939785496_dp, 0.6814628764322307_dp, &
             1.0700626854907365_dp, 1.666954483379765_dp, 2.5804865543829907_dp, &
             3.98239954584697_dp, 6.173693340567358_dp, 9.841113471805645_dp, &
             ! w
             2.02025512306358e-06_dp, 5.646195978302999e-06_dp, 1.2459981376118931e-05_dp, &
             2.6558616120894655e-05_dp, 5.496677329540227e-05_dp, 0.00011001661381376335_dp, &
             0.0002133536129728687_dp, 0.00040225031783798357_dp, 0.0007396681720342874_dp, &
             0.0013301645286480231_dp, 0.0023447373376501383_dp, 0.0040591887385540365_dp, &
             0.006912746797741015_dp, 0.011596783031411219_dp, 0.019187885775472535_dp, &
             0.031345840306920725_dp, 0.05060603076581172_dp, 0.08080873986051144_dp, &
             0.127727410418948_dp, 0.19999069795789906_dp, 0.3104584670174118_dp, &
             0.4783728335700245_dp, 0.7330793153515601_dp, 1.121707471323126_dp, &
             1.7294025536587003_dp, 2.751848599129929_dp, 4.931440243933937_dp]

      CASE (1030)

         aw(:) = & ! a
            [4.282997688066794e-07_dp, 2.5345161992402055e-06_dp, 7.804107618589548e-06_dp, &
             1.9977993343585435e-05_dp, 4.695744035623908e-05_dp, 0.00010419816341491701_dp, &
             0.00022088984011488607_dp, 0.00045064587806774786_dp, 0.0008895356267617178_dp, &
             0.0017059657107671494_dp, 0.0031894197252649243_dp, 0.005828823769633867_dp, &
             0.010436905268064723_dp, 0.018345184374114178_dp, 0.03170635866252156_dp, &
             0.053958479122247334_dp, 0.09053083668174715_dp, 0.14990832371153687_dp, &
             0.24522452570721986_dp, 0.39663337274530724_dp, 0.634835526371784_dp, &
             1.0063650387875287_dp, 1.581751822372174_dp, 2.4690851649384156_dp, &
             3.8402271206551823_dp, 5.9966110593710225_dp, 9.624893612606934_dp, &
             ! w
             1.1273646363230135e-06_dp, 3.30407491325403e-06_dp, 7.830421910507519e-06_dp, &
             1.7780384248535214e-05_dp, 3.865532573862084e-05_dp, 8.051945551775281e-05_dp, &
             0.0001615093363451519_dp, 0.0003135324773607453_dp, 0.0005915273858953615_dp, &
             0.001088311890188869_dp, 0.001958068346765551_dp, 0.0034530298414055904_dp, &
             0.0059801398994152625_dp, 0.010187650886489156_dp, 0.017096213065079784_dp, &
             0.028295565972774144_dp, 0.046237324565874476_dp, 0.07466791417101255_dp, &
             0.11926605834805211_dp, 0.18858268715811125_dp, 0.29544632977911794_dp, &
             0.45915514943809826_dp, 0.7092414841649864_dp, 1.0931683345070318_dp, &
             1.6964811418306054_dp, 2.7149901181714964_dp, 4.88934061869101_dp]

      CASE (1031)

         aw(:) = & ! a
            [3.044077869225635e-07_dp, 1.8345797051288298e-06_dp, 5.8306518190610556e-06_dp, &
             1.5452902624503837e-05_dp, 3.746195190021638e-05_dp, 8.532865882390207e-05_dp, &
             0.00018492717447337773_dp, 0.00038447068756841023_dp, 0.0007714380154039876_dp, &
             0.0015008747152174746_dp, 0.0028419264854728896_dp, 0.0052532011229795865_dp, &
             0.009503152656455776_dp, 0.016859985737617073_dp, 0.029387797103631402_dp, &
             0.05040351645324855_dp, 0.08517511727583636_dp, 0.14197907169434135_dp, &
             0.23368954926580837_dp, 0.38015290552527153_dp, 0.611729093898482_dp, &
             0.9746121803917456_dp, 1.5390504533360434_dp, 2.4129869268184634_dp, &
             3.7683396667726585_dp, 5.906771478610304_dp, 9.514916534301372_dp, &
             ! w
             8.043791158110075e-07_dp, 2.4397787368067945e-06_dp, 6.054236773947436e-06_dp, &
             1.4267550169056792e-05_dp, 3.190457591131693e-05_dp, 6.798194633826852e-05_dp, &
             0.00013897861724333242_dp, 0.00027423236514752215_dp, 0.0005247921696292247_dp, &
             0.00097772032108461_dp, 0.0017788682611212536_dp, 0.0031686768506202726_dp, &
             0.005537784752788189_dp, 0.009512468968782824_dp, 0.016084612913245176_dp, &
             0.02680759793413749_dp, 0.04408911600333397_dp, 0.07162578735056811_dp, &
             0.11504524024551752_dp, 0.18285518937714626_dp, 0.28786426742283683_dp, &
             0.44939607977744067_dp, 0.6970778609773493_dp, 1.0785475713046209_dp, &
             1.6795664194936384_dp, 2.696024912444383_dp, 4.867686914030128_dp]

      CASE (1032)

         aw(:) = & ! a
            [2.392806012063374e-07_dp, 1.4640669800963385e-06_dp, 4.77029549570211e-06_dp, &
             1.2971799915495292e-05_dp, 3.214975810349573e-05_dp, 7.458285899582335e-05_dp, &
             0.0001641313533637601_dp, 0.0003456955183275747_dp, 0.0007014385500812336_dp, &
             0.001378075285586933_dp, 0.002631983387012906_dp, 0.004902617785847488_dp, &
             0.008930302768378003_dp, 0.015942799349190424_dp, 0.027947325215813817_dp, &
             0.04818268331975345_dp, 0.08181231490338949_dp, 0.13697702864538128_dp, &
             0.22638136969925596_dp, 0.3696696252832951_dp, 0.5969767421433326_dp, &
             0.9542708407154293_dp, 1.511611335932235_dp, 2.3768408963523915_dp, &
             3.7219115701391203_dp, 5.848638064442507_dp, 9.443648530012794_dp, &
             ! w
             6.343440691502285e-07_dp, 1.978296825743782e-06_dp, 5.0801425036158145e-06_dp, &
             1.2289730260022029e-05_dp, 2.802660675530703e-05_dp, 6.0664864739766854e-05_dp, &
             0.00012565503531272474_dp, 0.0002507279450051139_dp, 0.00048448296908444144_dp, &
             0.000910333753249535_dp, 0.0016688182926602518_dp, 0.002992813546784709_dp, &
             0.005262443217918938_dp, 0.009089738617475543_dp, 0.015447838854484948_dp, &
             0.025866309579826947_dp, 0.04272390819565634_dp, 0.06968422652033_dp, &
             0.11234070289454058_dp, 0.1791716678399734_dp, 0.2829713527152451_dp, &
             0.4430786436162086_dp, 0.6891821372320907_dp, 1.0690350994295952_dp, &
             1.668543120096164_dp, 2.6836548658754666_dp, 4.853566250440072_dp]

      CASE (1033)

         aw(:) = & ! a
            [1.9872290995100557e-07_dp, 1.2320479928210371e-06_dp, 4.098304154463825e-06_dp, &
             1.137482340280105e-05_dp, 2.8679349116364172e-05_dp, 6.747187816793742e-05_dp, &
             0.00015021843719519121_dp, 0.00031950970238100903_dp, 0.000653780674403012_dp, &
             0.001293872102374736_dp, 0.0024871155604689293_dp, 0.004659337070576442_dp, &
             0.0085307640725342_dp, 0.01530015616462431_dp, 0.026933802420064816_dp, &
             0.04661410282101341_dp, 0.07942879353027087_dp, 0.13342012741836712_dp, &
             0.22116903812549918_dp, 0.3621720802476913_dp, 0.5863991104964001_dp, &
             0.9396517597170191_dp, 1.491849410464224_dp, 2.3507592133887694_dp, &
             3.6883566201161844_dp, 5.806567956794808_dp, 9.39202130324481_dp, &
             ! w
             5.283353751060703e-07_dp, 1.6873028954141597e-06_dp, 4.452982770378373e-06_dp, &
             1.0991723602931945e-05_dp, 2.544492068026009e-05_dp, 5.573859282224164e-05_dp, &
             0.00011660074493091565_dp, 0.00023462701275100932_dp, 0.0004566776617926573_dp, &
             0.0008635640585389771_dp, 0.0015920184391628952_dp, 0.002869478656358298_dp, &
             0.005068479068145049_dp, 0.008790732002078987_dp, 0.014995750805662462_dp, &
             0.02519572629194617_dp, 0.04174822470849311_dp, 0.06829253924420949_dp, &
             0.11039680309103479_dp, 0.17651736754983788_dp, 0.279437266163447_dp, &
             0.43850583064257825_dp, 0.6834560157675487_dp, 1.0621255598884007_dp, &
             1.6605269303073051_dp, 2.6746540314901255_dp, 4.843293051470446_dp]

      CASE (1034)

         aw(:) = & ! a
            [1.708663148980379e-07_dp, 1.0719407612707742e-06_dp, 3.629880565729153e-06_dp, &
             1.0247447633606835e-05_dp, 2.6200412588732798e-05_dp, 6.234120102184512e-05_dp, &
             0.00014009464704303065_dp, 0.0003003174052814775_dp, 0.0006186324716746418_dp, &
             0.0012314319838720523_dp, 0.0023791716502456874_dp, 0.004477284013418298_dp, &
             0.008230624112551932_dp, 0.014815705600642588_dp, 0.02616734026765008_dp, &
             0.04542444214765721_dp, 0.07761624061811427_dp, 0.13070864439986096_dp, &
             0.21718660976337473_dp, 0.3564317055692259_dp, 0.5782849494949315_dp, &
             0.9284176256739851_dp, 1.47663898094982_dp, 2.330656116732063_dp, &
             3.6624618440001444_dp, 5.7740696850698265_dp, 9.352110178540393_dp, &
             ! w
             4.554568714186842e-07_dp, 1.4853146580114608e-06_dp, 4.010109783706538e-06_dp, &
             1.0061276109700429e-05_dp, 2.357374234051719e-05_dp, 5.213699289131018e-05_dp, &
             0.00010993342107847236_dp, 0.00022269782569833403_dp, 0.0004359666166167894_dp, &
             0.0008285632929813975_dp, 0.0015343035723806068_dp, 0.002776444652921308_dp, &
             0.004921670982814177_dp, 0.008563719353161213_dp, 0.014651543788182222_dp, &
             0.024683833967299302_dp, 0.04100164172617943_dp, 0.06722526041781025_dp, &
             0.10890294768949611_dp, 0.17447365764087136_dp, 0.2767113164346024_dp, &
             0.434972958995465_dp, 0.6790257660043966_dp, 1.0567733191226292_dp, &
             1.6543120445112234_dp, 2.6676726498929257_dp, 4.8353256005667715_dp]

      CASE (1035)

         aw(:) = & ! a
            [1.5046600842380787e-07_dp, 9.542091743770965e-07_dp, 3.282386897052522e-06_dp, &
             9.402124660146259e-06_dp, 2.4323478144930646e-05_dp, 5.842447532946064e-05_dp, &
             0.00013231283468887587_dp, 0.0002854785131857045_dp, 0.0005913199720108167_dp, &
             0.0011826986439803994_dp, 0.0022945971607639024_dp, 0.0043341531128298005_dp, &
             0.007993923204493958_dp, 0.014432583881101056_dp, 0.025559655102631463_dp, &
             0.04447904108994307_dp, 0.0761727790922896_dp, 0.12854508379167245_dp, &
             0.21400320755658808_dp, 0.35183544654773324_dp, 0.5717780879330294_dp, &
             0.9193961924705142_dp, 1.4644089308699892_dp, 2.314473872998031_dp, &
             3.6415973555746186_dp, 5.747863832738934_dp, 9.31990744402916_dp, &
             ! w
             4.020435349049668e-07_dp, 1.3360185007046638e-06_dp, 3.677937554062206e-06_dp, &
             9.354777083526352e-06_dp, 2.2140148908614583e-05_dp, 4.935819599174935e-05_dp, &
             0.00010475930715355836_dp, 0.0002133943822876141_dp, 0.00041974486395866855_dp, &
             0.0008010457006218791_dp, 0.0014887759343406993_dp, 0.0027028352627257233_dp, &
             0.004805199475603059_dp, 0.008383171971295142_dp, 0.01437717115673257_dp, &
             0.024274949643709864_dp, 0.04040415055730206_dp, 0.06636959953403641_dp, &
             0.10770331687412875_dp, 0.1728299588562054_dp, 0.2745158119623451_dp, &
             0.432123883647418_dp, 0.6754489227647645_dp, 1.0524479806828497_dp, &
             1.649286085341666_dp, 2.662024822646241_dp, 4.828880580255117_dp]

      CASE (1036)

         aw(:) = & ! a
            [1.3483327719929097e-07_dp, 8.636623673712753e-07_dp, 3.013026640776208e-06_dp, &
             8.740768544062583e-06_dp, 2.2842797919313344e-05_dp, 5.531318070879643e-05_dp, &
             0.00012609546846510718_dp, 0.0002735647687929553_dp, 0.0005692993510592034_dp, &
             0.001143263875482272_dp, 0.0022259397121263524_dp, 0.004217627406270247_dp, &
             0.00780072652380558_dp, 0.014119154597089926_dp, 0.025061468426791706_dp, &
             0.04370250697165623_dp, 0.07498506712165497_dp, 0.1267619865516396_dp, &
             0.21137570939225736_dp, 0.34803661544585673_dp, 0.5663933525277494_dp, &
             0.9119219090702436_dp, 1.4542657015593066_dp, 2.301040356501029_dp, &
             3.6242631250794344_dp, 5.7260778508230965_dp, 9.29312278000876_dp, &
             ! w
             3.6108505326641745e-07_dp, 1.2206636594598906e-06_dp, 3.4179800936920884e-06_dp, &
             8.79610918590675e-06_dp, 2.0997982030650414e-05_dp, 4.713122236305409e-05_dp, &
             0.00010059249655814108_dp, 0.0002058711628328325_dp, 0.00040658022071106577_dp, &
             0.0007786439055636062_dp, 0.0014516091552813555_dp, 0.002642593964458383_dp, &
             0.004709665757474432_dp, 0.008234778512255804_dp, 0.01415124081315477_dp, &
             0.023937679004776856_dp, 0.039910526447777483_dp, 0.06566164979656054_dp, &
             0.10670942689099573_dp, 0.1714664426624184_dp, 0.27269242820113926_dp, &
             0.42975518987978567_dp, 0.6724723712141347_dp, 1.0488457304122376_dp, &
             1.645097944008935_dp, 2.6573170966013433_dp, 4.823508711705557_dp]

      CASE (1037)

         aw(:) = & ! a
            [1.2244238590252792e-07_dp, 7.916553732779728e-07_dp, 2.797295392425009e-06_dp, &
             8.206737559771293e-06_dp, 2.163854105266043e-05_dp, 5.276758321201203e-05_dp, &
             0.00012098328108035421_dp, 0.0002637277506440087_dp, 0.0005510519884690576_dp, &
             0.0011104844409749947_dp, 0.0021687130694728117_dp, 0.004120265828114491_dp, &
             0.007638951954037942_dp, 0.013856187180338079_dp, 0.024642745114058458_dp, &
             0.04304877443075282_dp, 0.07398369418795818_dp, 0.12525658445561255_dp, &
             0.20915462117943184_dp, 0.34482164841651886_dp, 0.5618313659445509_dp, &
             0.9055834564062765_dp, 1.445656280064054_dp, 2.28962923681839_dp, &
             3.6095286193874356_dp, 5.7075490823946735_dp, 9.270333167914828_dp, &
             ! w
             3.2860085327856715e-07_dp, 1.128541850454422e-06_dp, 3.2080114878353213e-06_dp, &
             8.340817679428347e-06_dp, 2.0061146327110713e-05_dp, 4.529536944721871e-05_dp, &
             9.714319556288765e-05_dp, 0.00019962141911422605_dp, 0.0003956106141301545_dp, &
             0.0007599273953992115_dp, 0.0014204830296747515_dp, 0.0025920366672210413_dp, &
             0.004629336304109948_dp, 0.00810978571737166_dp, 0.013960636986767397_dp, &
             0.02365273068745493_dp, 0.03949292178335002_dp, 0.06506198367811626_dp, &
             0.10586658786156224_dp, 0.1703089189789739_dp, 0.2711429850192632_dp, &
             0.42774055306794734_dp, 0.6699387197252387_dp, 1.0457774484685272_dp, &
             1.6415288899622589_dp, 2.6533042648466525_dp, 4.818930022908951_dp]

      CASE (1038)

         aw(:) = & ! a
            [1.1236083484313051e-07_dp, 7.328909531736803e-07_dp, 2.62009489348869e-06_dp, &
             7.764867659479818e-06_dp, 2.063575920856369e-05_dp, 5.0636754203679484e-05_dp, &
             0.00011668548489616494_dp, 0.0002554276253367132_dp, 0.0005356074653546425_dp, &
             0.0010826649055450436_dp, 0.002120030038940355_dp, 0.004037265202660264_dp, &
             0.0075007793354132145_dp, 0.013631203651731203_dp, 0.02428395192499402_dp, &
             0.04248782177814056_dp, 0.07312333577965469_dp, 0.12396164928213008_dp, &
             0.2072419840715197_dp, 0.34205038714986946_dp, 0.557895371990459_dp, &
             0.9001101548018628_dp, 1.4382163086880388_dp, 2.2797614573200717_dp, &
             3.596779537816259_dp, 5.691509413645565_dp, 9.250597946672741_dp, &
             ! w
             3.021567212669974e-07_dp, 1.0530708400970992e-06_dp, 3.0342324625713977e-06_dp, &
             7.96103199823831e-06_dp, 1.9275266697830992e-05_dp, 4.374854303521141e-05_dp, &
             9.422638341582306e-05_dp, 0.00019432024327497953_dp, 0.0003862812771383207_dp, &
             0.0007439726118163856_dp, 0.0013938952501706453_dp, 0.002548771646479404_dp, &
             0.004560479797197636_dp, 0.00800248421553169_dp, 0.013796787031162794_dp, &
             0.02340747128432762_dp, 0.039133067219051576_dp, 0.06454469181817421_dp, &
             0.10513880558145955_dp, 0.16930849011625215_dp, 0.26980268817526193_dp, &
             0.42599650407617173_dp, 0.6677438591497824_dp, 1.0431179254762268_dp, &
             1.6384340141027387_dp, 2.6498238231617117_dp, 4.81495898517923_dp]

      CASE (1039)

         aw(:) = & ! a
            [6.439693341715039e-08_dp, 4.502819297916944e-07_dp, 1.7482289018145977e-06_dp, &
             5.53670926896511e-06_dp, 1.547280917746472e-05_dp, 3.947823696344194e-05_dp, &
             9.386249835099445e-05_dp, 0.000210830651240247_dp, 0.00045178774147190497_dp, &
             0.0009303673560010658_dp, 0.0018514771580851032_dp, 0.0035763009582114973_dp, &
             0.006728754373286237_dp, 0.012367272315853876_dp, 0.022258325510475113_dp, &
             0.03930660181991045_dp, 0.06822403817692177_dp, 0.11655974256953242_dp, &
             0.1962711661218106_dp, 0.3261034969387611_dp, 0.5351792658275734_dp, &
             0.8684362472695254_dp, 1.3950558834185884_dp, 2.222392532587447_dp, &
             3.5225211903063993_dp, 5.597942744224213_dp, 9.135340584239298_dp, &
             ! w
             1.7612730922301005e-07_dp, 6.850472968731309e-07_dp, 2.157059548465582e-06_dp, &
             5.994847488728639e-06_dp, 1.5132238090479849e-05_dp, 3.547726848834456e-05_dp, &
             7.844604663600577e-05_dp, 0.00016535511280926297_dp, 0.0003348701672504547_dp, &
             0.0006553927766784215_dp, 0.001245305119248374_dp, 0.00230554962030188_dp, &
             0.004171331845030701_dp, 0.007393136947647742_dp, 0.01286221703738584_dp, &
             0.022002911822893913_dp, 0.037064577553237055_dp, 0.061561006228935934_dp, &
             0.10092764698021395_dp, 0.1635026001105081_dp, 0.2620031614990439_dp, &
             0.4158221528766768_dp, 0.6549113061391172_dp, 1.0275400176778966_dp, &
             1.6202816220933072_dp, 2.62939579200687_dp, 4.79165492595066_dp]

      CASE (1040)

         aw(:) = & ! a
            [4.692145622912571e-08_dp, 3.453726723137099e-07_dp, 1.4120419658352095e-06_dp, &
             4.644659065216893e-06_dp, 1.3342153037171206e-05_dp, 3.4761117866121466e-05_dp, &
             8.402438233137593e-05_dp, 0.00019129325375463404_dp, 0.0004145616772277768_dp, &
             0.0008619287326724236_dp, 0.0017295529763624714_dp, 0.0033651217300925706_dp, &
             0.0063722155908638835_dp, 0.01177933576631991_dp, 0.02130991999737814_dp, &
             0.0378083100384182_dp, 0.06590408267384229_dp, 0.11303736863200164_dp, &
             0.19102669703365602_dp, 0.3184484077141203_dp, 0.5242328503414031_dp, &
             0.8531197161722289_dp, 1.374118625810681_dp, 2.194484447273052_dp, &
             3.486309919495051_dp, 5.552226517971586_dp, 9.078942675712826_dp, &
             ! w
             1.3009025718548029e-07_dp, 5.451311569451456e-07_dp, 1.805287286636814e-06_dp, &
             5.177331270434758e-06_dp, 1.3365177498646305e-05_dp, 3.187908771076395e-05_dp, &
             7.14699447019722e-05_dp, 0.00015237652289038387_dp, 0.00031156693651894995_dp, &
             0.0006148379670237801_dp, 0.001176673973986873_dp, 0.002192327312638174_dp, &
             0.003988904463008586_dp, 0.007105667338441195_dp, 0.012418770683599542_dp, &
             0.0213329383455694_dp, 0.03607312097482517_dp, 0.06012447964156094_dp, &
             0.09889174257967095_dp, 0.16068495436194105_dp, 0.25820460645737253_dp, &
             0.4108510425715523_dp, 0.6486235337002515_dp, 1.0198889019105355_dp, &
             1.6113505071374186_dp, 2.6193359980392366_dp, 4.780180966203478_dp]

      CASE (1041)

         aw(:) = & ! a
            [3.769025339720575e-08_dp, 2.892293662887212e-07_dp, 1.2275616354232278e-06_dp, &
             4.143923800207455e-06_dp, 1.2124950008347731e-05_dp, 3.2029192476915924e-05_dp, &
             7.826379843796882e-05_dp, 0.0001797496325970079_dp, 0.00039239902316021494_dp, &
             0.0008209176117875542_dp, 0.001656077005404812_dp, 0.003237223335389175_dp, &
             0.006155326859559991_dp, 0.011420268223210758_dp, 0.020728638895969025_dp, &
             0.036887031659902474_dp, 0.0644733710392377_dp, 0.11085927201776558_dp, &
             0.1877757047257223_dp, 0.31369231695675587_dp, 0.5174176872289247_dp, &
             0.8435655874175582_dp, 1.3610359192316241_dp, 2.177019387180813_dp, &
             3.463619038073598_dp, 5.523549121267125_dp, 9.04353624914506_dp, &
             ! w
             1.0573738585540205e-07_dp, 4.690145806438172e-07_dp, 1.6076130131872165e-06_dp, &
             4.7084240586299786e-06_dp, 1.2336949577033776e-05_dp, 2.97619620671435e-05_dp, &
             6.732817995011226e-05_dp, 0.00014461298568028174_dp, 0.0002975380330592539_dp, &
             0.0005902879414245153_dp, 0.0011349258369968832_dp, 0.0021231577679816913_dp, &
             0.0038770268380961077_dp, 0.006928757867333576_dp, 0.012145012359387941_dp, &
             0.020918144424417976_dp, 0.03545766845761392_dp, 0.05923057540951733_dp, &
             0.09762201262557958_dp, 0.1589240170757832_dp, 0.2558260746622655_dp, &
             0.4077328490421953_dp, 0.6446733326221191_dp, 1.01507598040183_dp, &
             1.6057270693344223_dp, 2.6129987726487443_dp, 4.772953585782717_dp]

      CASE (1042)

         aw(:) = & ! a
            [3.192688511226688e-08_dp, 2.5381515538577897e-07_dp, 1.108986911961877e-06_dp, &
             3.8168657094130495e-06_dp, 1.1320216847351783e-05_dp, 3.0206065300549993e-05_dp, &
             7.439080500091955e-05_dp, 0.00017194106197723167_dp, 0.0003773304168625865_dp, &
             0.0007929116378581136_dp, 0.0016057107721317047_dp, 0.0031492597498130478_dp, &
             0.006005718908476013_dp, 0.01117193311394401_dp, 0.020325663351667567_dp, &
             0.03624697696426884_dp, 0.06347744157648616_dp, 0.1093403658603171_dp, &
             0.18550488617051494_dp, 0.31036517774846006_dp, 0.5126435227116273_dp, &
             0.8368642656986087_dp, 1.3518491553949243_dp, 2.164742896560698_dp, &
             3.447655400232127_dp, 5.50335958744786_dp, 9.018596028920532_dp, &
             ! w
             9.051900758375318e-08_dp, 4.203887118531679e-07_dp, 1.4784016574047624e-06_dp, &
             4.397599618263773e-06_dp, 1.164865445535051e-05_dp, 2.8334024418389487e-05_dp, &
             6.451762027438563e-05_dp, 0.00013931799719955234_dp, 0.0002879286642202782_dp, &
             0.0005734094505042307_dp, 0.0011061300978658056_dp, 0.002075310889819803_dp, &
             0.003799438502918505_dp, 0.0068057853417516695_dp, 0.011954319281190876_dp, &
             0.02062865619967732_dp, 0.03502738454414652_dp, 0.058604605623970504_dp, &
             0.0967315367741743_dp, 0.15768734573052112_dp, 0.2541535518492713_dp, &
             0.4055376711363269_dp, 0.6418895697728475_dp, 1.0116813328986147_dp, &
             1.6017582471360277_dp, 2.6085247198738886_dp, 4.767851415182237_dp]

      CASE (1043)

         aw(:) = & ! a
            [2.7964172749856977e-08_dp, 2.292558043774731e-07_dp, 1.025516306411912e-06_dp, &
             3.5837922248453173e-06_dp, 1.074148426471467e-05_dp, 2.888578436551974e-05_dp, &
             7.157052689936585e-05_dp, 0.00016622925266215853_dp, 0.0003662664074968937_dp, &
             0.0007722822578704207_dp, 0.0015685073023445051_dp, 0.003084126013028135_dp, &
             0.005894700349223595_dp, 0.01098729694119632_dp, 0.02002553209601744_dp, &
             0.03576952223362426_dp, 0.06273345532409448_dp, 0.10820421768565977_dp, &
             0.18380426814943254_dp, 0.30787073879604404_dp, 0.5090606012889078_dp, &
             0.8318304152966333_dp, 1.344942563151507_dp, 2.1555066448143143_dp, &
             3.43563752253542_dp, 5.488152526096072_dp, 8.99980337859715_dp, &
             ! w
             8.004834733492082e-08_dp, 3.8631659932858306e-07_dp, 1.386267509720853e-06_dp, &
             4.173643813437147e-06_dp, 1.114909313394502e-05_dp, 2.729181617848311e-05_dp, &
             6.245702168564406e-05_dp, 0.0001354213933719994_dp, 0.00028083469058846135_dp, &
             0.0005609151396578768_dp, 0.0010847632042291317_dp, 0.002039732928343927_dp, &
             0.0037416368875741166_dp, 0.006714018427169055_dp, 0.011811798340824064_dp, &
             0.02041199470592566_dp, 0.034704934431409096_dp, 0.05813495677201824_dp, &
             0.09606270744405344_dp, 0.15675755544577297_dp, 0.25289489904071155_dp, &
             0.403884297920705_dp, 0.6397913114271438_dp, 1.0091210177192986_dp, &
             1.5987634991060216_dp, 2.6051479326905813_dp, 4.7640007348228295_dp]

      CASE (1044)

         aw(:) = & ! a
            [2.506231356064682e-08_dp, 2.1113777525198703e-07_dp, 9.631716903097764e-07_dp, &
             3.407993565741137e-06_dp, 1.0301817554137037e-05_dp, 2.7877267164845783e-05_dp, &
             6.940690332605619e-05_dp, 0.00016183191675560055_dp, 0.0003577235548645196_dp, &
             0.0007563138536316547_dp, 0.0015396472707742605_dp, 0.003033503861178697_dp, &
             0.005808271877715008_dp, 0.010843341955159265_dp, 0.01979121468863471_dp, &
             0.03539631206039053_dp, 0.06215126305945643_dp, 0.10731424798130924_dp, &
             0.18247090314363906_dp, 0.3059133185528314_dp, 0.5062468435108554_dp, &
             0.8278743955344539_dp, 1.3395112993009697_dp, 2.148239229000104_dp, &
             3.4261768105937573_dp, 5.476176477499033_dp, 8.984999125160064_dp, &
             ! w
             7.237677372735283e-08_dp, 3.609608852035123e-07_dp, 1.316742490004197e-06_dp, &
             4.003257285933263e-06_dp, 1.0766846352372117e-05_dp, 2.649085541623535e-05_dp, &
             6.0867829164729415e-05_dp, 0.00013240747364775282_dp, 0.00027533418133416564_dp, &
             0.0005512068025827516_dp, 0.0010681299869222435_dp, 0.0020119917780688362_dp, &
             0.0036965016864726357_dp, 0.0066422670774409385_dp, 0.0117002308706876_dp, &
             0.020242205558766137_dp, 0.03445199249647781_dp, 0.05776621053751845_dp, &
             0.0955371316573405_dp, 0.1560263441077744_dp, 0.2519043522986348_dp, &
             0.4025822610720924_dp, 0.638137969771434_dp, 1.0071026176143199_dp, &
             1.5964017770696073_dp, 2.602484432456299_dp, 4.760963550075687_dp]

      CASE (1045)

         aw(:) = & ! a
            [2.2840880666416185e-08_dp, 1.9717751651393164e-07_dp, 9.146283371366514e-07_dp, &
             3.2700008155245947e-06_dp, 9.954670460404866e-06_dp, 2.7077432740156864e-05_dp, &
             6.7684972273779e-05_dp, 0.00015832231387490975_dp, 0.0003508891527388636_dp, &
             0.0007435131416554811_dp, 0.001516471985465016_dp, 0.0029927911556568436_dp, &
             0.00573866843967504_dp, 0.01072727140712659_dp, 0.01960208111233498_dp, &
             0.03509477436298562_dp, 0.06168045900903803_dp, 0.10659396841055092_dp, &
             0.18139096861967813_dp, 0.3043268616646848_dp, 0.5039649147311501_dp, &
             0.8246642748304757_dp, 1.3351018183188557_dp, 2.142336338304016_dp, &
             3.4184894315619343_dp, 5.466442149200285_dp, 8.972963095718418_dp, &
             ! w
             6.65014938459946e-08_dp, 3.412769350164443e-07_dp, 1.2621488875694028e-06_dp, &
             3.868571067015773e-06_dp, 1.0463283543017929e-05_dp, 2.5852507338959915e-05_dp, &
             5.9597671275086006e-05_dp, 0.0001299929395057802_dp, 0.000270918821674927_dp, &
             0.0005434004432453843_dp, 0.001054735505246438_dp, 0.0019896229047767933_dp, &
             0.0036600646446727867_dp, 0.0065842822196466935_dp, 0.011609983250722626_dp, &
             0.020104742814970756_dp, 0.03424704571382418_dp, 0.05746721450301402_dp, &
             0.09511068344611326_dp, 0.1554326746749852_dp, 0.25109966666466427_dp, &
             0.40152397792438266_dp, 0.6367935256444012_dp, 1.0054606830034698_dp, &
             1.5944800078704515_dp, 2.6003167797528004_dp, 4.758491848336381_dp]

      CASE (1046)

         aw(:) = & ! a
            [2.108348323503444e-08_dp, 1.8606901940485092e-07_dp, 8.756509397244363e-07_dp, &
             3.1584432704932897e-06_dp, 9.672645518882276e-06_dp, 2.642523726192473e-05_dp, &
             6.62768089630148e-05_dp, 0.00015544545804077703_dp, 0.0003452759242653766_dp, &
             0.0007329821232710457_dp, 0.001497378448240879_dp, 0.0029592067382946076_dp, &
             0.0056811878037465865_dp, 0.010631321697895069_dp, 0.019445594453750353_dp, &
             0.03484508486295183_dp, 0.061290322147537275_dp, 0.10599670172474181_dp, &
             0.18049492216461438_dp, 0.303009802614152_dp, 0.5020695072741082_dp, &
             0.8219966430476814_dp, 1.3314359545290304_dp, 2.1374270687373023_dp, &
             3.412093993743771_dp, 5.45834165307861_dp, 8.962945240578694_dp, &
             ! w
             6.185175707675386e-08_dp, 3.2551110500076984e-07_dp, 1.2180003503230362e-06_dp, &
             3.759048194362658e-06_dp, 1.0215479642775028e-05_dp, 2.5329874131444367e-05_dp, &
             5.8555303582220226e-05_dp, 0.00012800756953579488_dp, 0.0002672822889685553_dp, &
             0.000536961969098837_dp, 0.0010436744990435214_dp, 0.001971130899968773_dp, &
             0.0036299135717349886_dp, 0.006536258975730945_dp, 0.011535181110241271_dp, &
             0.019990724621369062_dp, 0.03407694156741163_dp, 0.05721890059934172_dp, &
             0.09475632440215155_dp, 0.1549391087725357_dp, 0.25043034886872095_dp, &
             0.4006433440619474_dp, 0.6356743394393334_dp, 1.0040934131898342_dp, &
             1.5928793384845792_dp, 2.5985110893333885_dp, 4.756432927411025_dp]

      CASE (1047)

         aw(:) = & ! a
            [1.9657634446563594e-08_dp, 1.7700877117699972e-07_dp, 8.436085924054425e-07_dp, &
             3.066197163932294e-06_dp, 9.43846448130806e-06_dp, 2.5881981725346932e-05_dp, &
             6.510097076310213e-05_dp, 0.00015303844387372722_dp, 0.0003405716397025741_dp, &
             0.0007241439465780422_dp, 0.0014813346880753638_dp, 0.0029309567299341845_dp, &
             0.005632791790008266_dp, 0.010550468959010304_dp, 0.019313630762516264_dp, &
             0.03463438117209301_dp, 0.06096089704944261_dp, 0.1054920956922182_dp, &
             0.17973749918952178_dp, 0.3018959738693043_dp, 0.5004658785371005_dp, &
             0.8197387740111778_dp, 1.3283320795922557_dp, 2.1332690916641397_dp, &
             3.4066758177063945_dp, 5.451477451164133_dp, 8.954454893740664_dp, &
             ! w
             5.807796213947274e-08_dp, 3.1257763684768025e-07_dp, 1.1814849204825377e-06_dp, &
             3.668034015752292e-06_dp, 1.000887614257514e-05_dp, 2.4893043706635218e-05_dp, &
             5.7682323503348534e-05_dp, 0.00012634209100652221_dp, 0.00026422745677494324_dp, &
             0.0005315469486941688_dp, 0.0010343620807832697_dp, 0.001955547967111518_dp, &
             0.0036044850072998384_dp, 0.006495727881305546_dp, 0.01147200712260352_dp, &
             0.019894372755884956_dp, 0.033933114546447915_dp, 0.057008838839305874_dp, &
             0.094456413106498_dp, 0.15452119924401148_dp, 0.24986340186171588_dp, &
             0.3998971307163926_dp, 0.6347256815232175_dp, 1.0029341600492385_dp, &
             1.591521926539097_dp, 2.596979655447138_dp, 4.754686758063712_dp]

      CASE (1048)

         aw(:) = & ! a
            [1.3077493553669045e-08_dp, 1.3446428461721632e-07_dp, 6.894775445454599e-07_dp, &
             2.614771927043685e-06_dp, 8.278412819111448e-06_dp, 2.3166284975862367e-05_dp, &
             5.918104565603476e-05_dp, 0.00014084993201930442_dp, 0.00031663592227642397_dp, &
             0.0006789915688675132_dp, 0.0013990828051893673_dp, 0.0027856827527221305_dp, &
             0.00538324428105995_dp, 0.010132556783363701_dp, 0.01863005822568714_dp, &
             0.03354079810465206_dp, 0.059248090669115176_dp, 0.10286419746112514_dp, &
             0.175787113058364_dp, 0.29607881519989404_dp, 0.4920801981687992_dp, &
             0.8079184922400721_dp, 1.312066104038825_dp, 2.1114591846486728_dp, &
             3.378233572831204_dp, 5.415421552586376_dp, 8.909835894750403_dp, &
             ! w
             4.064100105566295e-08_dp, 2.5073515305060495e-07_dp, 1.0026614671897668e-06_dp, &
             3.2161935234769364e-06_dp, 8.973372235973011e-06_dp, 2.2687699878913255e-05_dp, &
             5.324951565336133e-05_dp, 0.0001178448071382357_dp, 0.00024857907404769936_dp, &
             0.0005037129306199607_dp, 0.0009863512624435078_dp, 0.0018749966254274046_dp, &
             0.0034727302074631277_dp, 0.006285277663788842_dp, 0.011143361160208523_dp, &
             0.019392255752118715_dp, 0.033182397498854106_dp, 0.05591079878196936_dp, &
             0.09288659185447329_dp, 0.15233101253679696_dp, 0.24688872314683835_dp, &
             0.39597777284640007_dp, 0.6297383942856121_dp, 0.9968350013211065_dp, &
             1.5843761176989153_dp, 2.588915338343739_dp, 4.745492178729012_dp]

      CASE (1049)

         aw(:) = & ! a
            [1.1168676780778662e-08_dp, 1.2180640938260491e-07_dp, 6.42168305551965e-07_dp, &
             2.4732613601403548e-06_dp, 7.909425136687957e-06_dp, 2.2293132150776055e-05_dp, &
             5.726170006022515e-05_dp, 0.00013687153406210996_dp, 0.0003087795790773896_dp, &
             0.0006641014961387852_dp, 0.0013718484409475933_dp, 0.0027374115157595986_dp, &
             0.0053000676633829236_dp, 0.009992877648448003_dp, 0.018401021216170845_dp, &
             0.033173564254514905_dp, 0.058671749775199934_dp, 0.10197830271938904_dp, &
             0.17445314028708012_dp, 0.2941114245403274_dp, 0.48924009838623994_dp, &
             0.8039099679980727_dp, 1.3065435035515134_dp, 2.104046661822854_dp, &
             3.368558366679237_dp, 5.403147594518706_dp, 8.894638694743145_dp, &
             ! w
             3.5571636295387056e-08_dp, 2.31885701828602e-07_dp, 9.465595172037144e-07_dp, &
             3.072114962558827e-06_dp, 8.639440026864074e-06_dp, 2.1970436349182598e-05_dp, &
             5.1798036627847485e-05_dp, 0.0001150470429434874_dp, 0.00024340284217001595_dp, &
             0.0004944693115174225_dp, 0.0009703519870024377_dp, 0.0018480720795338971_dp, &
             0.0034285721559179512_dp, 0.006214574348251618_dp, 0.011032707828043933_dp, &
             0.019222860545356376_dp, 0.0329286751042282_dp, 0.05553907272594601_dp, &
             0.09235433563597871_dp, 0.15158736673964276_dp, 0.24587739838392697_dp, &
             0.3946436974261759_dp, 0.6280390312929831_dp, 0.9947549539603935_dp, &
             1.5819375458963196_dp, 2.5861623872104706_dp, 4.742353579150264_dp]

      CASE (1050)

         aw(:) = & ! a
            [1.0922730895902452e-08_dp, 1.2016102516355542e-07_dp, 6.359585057110366e-07_dp, &
             2.454569217019536e-06_dp, 7.860474879922026e-06_dp, 2.2176931492332475e-05_dp, &
             5.700564421868153e-05_dp, 0.0001363397401881642_dp, 0.00030772771266482525_dp, &
             0.0006621051667921726_dp, 0.0013681927984210268_dp, 0.0027309254735746026_dp, &
             0.0052888814206166895_dp, 0.009974077440709614_dp, 0.018370171673379145_dp, &
             0.033124068583551186_dp, 0.05859402503065693_dp, 0.10185876781344852_dp, &
             0.1742730574112346_dp, 0.2938457126503781_dp, 0.4888563624925515_dp, &
             0.8033681592277302_dp, 1.3057967934848087_dp, 2.103044115506054_dp, &
             3.367249456232268_dp, 5.401486766267579_dp, 8.892581990434447_dp, &
             ! w
             3.4917836241702025e-08_dp, 2.2941642585284216e-07_dp, 9.391473806911131e-07_dp, &
             3.0529880030367227e-06_dp, 8.594961885280855e-06_dp, 2.187466137505282e-05_dp, &
             5.16038403352657e-05_dp, 0.00011467212020671116_dp, 0.00024270824710016293_dp, &
             0.0004932274820766183_dp, 0.000968200417363755_dp, 0.0018444481009988747_dp, &
             0.003422623944082986_dp, 0.006205043746542174_dp, 0.01101778264918649_dp, &
             0.01919999899424585_dp, 0.032894414774324136_dp, 0.05548885404115657_dp, &
             0.09228239800037577_dp, 0.1514868173024687_dp, 0.24574060407320872_dp, &
             0.3944631850526886_dp, 0.6278090221906757_dp, 0.9944733474269393_dp, &
             1.5816073386937182_dp, 2.58578957328932_dp, 4.741928546959684_dp]

      CASE (1051)

         aw(:) = & ! a
            [1.7269308367952925e-08_dp, 1.4409917834746486e-07_dp, 6.529178049482427e-07_dp, &
             2.3003976110457863e-06_dp, 6.9335280303051496e-06_dp, 1.872244518172413e-05_dp, &
             4.6536901677980466e-05_dp, 0.00010836353287044556_dp, 0.00023927174031028936_dp, &
             0.0005054117792115246_dp, 0.0010280658785693684_dp, 0.0020241623971503264_dp, &
             0.003873331220201111_dp, 0.007227148098818627_dp, 0.01318458630956751_dp, &
             0.023570158372693398_dp, 0.041369635210886456_dp, 0.07140521350827259_dp, &
             0.12137137993196333_dp, 0.2034102602720391_dp, 0.33649068642728003_dp, &
             0.5499887380182457_dp, 0.8891023139229457_dp, 1.423237033551061_dp, &
             2.2598751952697516_dp, 3.571065635766467_dp, 5.659137075269493_dp, &
             9.210746809665743_dp, &
             ! w
             4.971556930715025e-08_dp, 2.4512493877760624e-07_dp, 8.891604477379925e-07_dp, &
             2.694261409494383e-06_dp, 7.229461600687299e-06_dp, 1.7756493210436763e-05_dp, &
             4.074312476992274e-05_dp, 8.853098139463149e-05_dp, 0.00018392439212625014_dp, &
             0.0003679174154176068_dp, 0.0007124640923785046_dp, 0.001341235454485254_dp, &
             0.0024628585175915647_dp, 0.004423429117921111_dp, 0.007788460877492792_dp, &
             0.013469340669403152_dp, 0.02291646325370977_dp, 0.03841146380316371_dp, &
             0.06350581851904569_dp, 0.10367515954543258_dp, 0.16729391138716596_dp, &
             0.2671004770572327_dp, 0.42247643359100207_dp, 0.6633096115860233_dp, &
             1.0377405893818756_dp, 1.6321727347271557_dp, 2.642780361790678_dp, &
             4.806923220998057_dp]

      CASE (1052)

         aw(:) = & ! a
            [1.0895118188773762e-08_dp, 1.0361461369511614e-07_dp, 5.099387053585073e-07_dp, &
             1.8889980533211041e-06_dp, 5.88889989698775e-06_dp, 1.6297328885123057e-05_dp, &
             4.128203713269053e-05_dp, 9.759101349023368e-05_dp, 0.0002181813642563272_dp, &
             0.00046570807815313593_dp, 0.0009558250785549132_dp, 0.0018966229932148376_dp, &
             0.0036541807006673865_dp, 0.006859768638436293_dp, 0.012582620547436744_dp, &
             0.022604694878683863_dp, 0.03985234994690154_dp, 0.06906703563584927_dp, &
             0.11783683036588014_dp, 0.19816876208730336_dp, 0.32886814809215037_dp, &
             0.5391258007622018_dp, 0.8739496829893695_dp, 1.402581890053244_dp, &
             2.232411606340228_dp, 3.5355070894776253_dp, 5.614322846490277_dp, &
             9.155534456578883_dp, &
             ! w
             3.2839863601841547e-08_dp, 1.8739359803730697e-07_dp, 7.263651992676576e-07_dp, &
             2.2884015454775067e-06_dp, 6.307323394138954e-06_dp, 1.5804182030425448e-05_dp, &
             3.683499130551101e-05_dp, 8.106008232173656e-05_dp, 0.00017018933037044374_dp, &
             0.0003435054727216734_dp, 0.0006703536787260965_dp, 0.0012705239221892889_dp, &
             0.002347008096574626_dp, 0.004237921456370455_dp, 0.007497771484212962_dp, &
             0.013023208384202923_dp, 0.02224556894310608_dp, 0.03742289330933707_dp, &
             0.062079132132665645_dp, 0.10166059521771431_dp, 0.1645152428331757_dp, &
             0.263366174652487_dp, 0.4176033322538294_dp, 0.6571613706264396_dp, &
             1.030275025843508_dp, 1.6234716823769342_dp, 2.632987540204431_dp, &
             4.795751925353884_dp]

      CASE (1053)

         aw(:) = & ! a
            [8.686257796942393e-09_dp, 8.919696282086086e-08_dp, 4.570716481175889e-07_dp, &
             1.7327705015742483e-06_dp, 5.4846862714460755e-06_dp, 1.5345733427853057e-05_dp, &
             3.919746772902721e-05_dp, 9.327982031192661e-05_dp, 0.00020967903002880602_dp, &
             0.00044960263715763813_dp, 0.0009263647086384991_dp, 0.0018443690285837458_dp, &
             0.003564023785290968_dp, 0.006708078244491314_dp, 0.012333253475007086_dp, &
             0.022203560424990262_dp, 0.039220242900969364_dp, 0.06809054434939908_dp, &
             0.11635737134162893_dp, 0.19597028083401266_dp, 0.3256648803141617_dp, &
             0.5345528134164331_dp, 0.8675606542040445_dp, 1.3938601548717462_dp, &
             2.2208000955914096_dp, 3.520456525260697_dp, 5.595337727613425_dp, &
             9.132128390894607_dp, &
             ! w
             2.6980081083251742e-08_dp, 1.6624567460090667e-07_dp, 6.644782645212181e-07_dp, &
             2.1308307661703125e-06_dp, 5.944038862967199e-06_dp, 1.5026472087285735e-05_dp, &
             3.526437353714725e-05_dp, 7.803580146688705e-05_dp, 0.00016459526397817446_dp, &
             0.0003335108211481309_dp, 0.0006530344972615959_dp, 0.0012413251317080235_dp, &
             0.0022989997312832505_dp, 0.004160801491081188_dp, 0.0073765759148882734_dp, &
             0.012836716065087616_dp, 0.02196444734149504_dp, 0.03700774200099543_dp, &
             0.06147877161491803_dp, 0.10081125096924243_dp, 0.16334170296863154_dp, &
             0.26178649001588755_dp, 0.41553888302241315_dp, 0.654553327700573_dp, &
             1.0271047446927402_dp, 1.6197738071528085_dp, 2.628823963269949_dp, &
             4.7910026746157115_dp]

      CASE (1054)

         aw(:) = & ! a
            [7.671273313771633e-09_dp, 8.246653720266085e-08_dp, 4.319208228961266e-07_dp, &
             1.657499869973599e-06_dp, 5.288224970217288e-06_dp, 1.4880232520420768e-05_dp, &
             3.8172620472173735e-05_dp, 9.11517264793093e-05_dp, 0.0002054680831908991_dp, &
             0.00044160358972318255_dp, 0.0009116972086216568_dp, 0.0018182982406914155_dp, &
             0.0035189584951300233_dp, 0.006632129546944244_dp, 0.012208214225227915_dp, &
             0.02200215105480042_dp, 0.03890247551718217_dp, 0.06759910518692695_dp, &
             0.11561204500293476_dp, 0.19486168724660805_dp, 0.32404822946051837_dp, &
             0.5322430581587142_dp, 0.8643313100206577_dp, 1.3894488613828202_dp, &
             2.2149238058950504_dp, 3.5128360373358216_dp, 5.585721189765881_dp, &
             9.120268878396354_dp, &
             ! w
             2.4283524094186823e-08_dp, 1.5622667959094901e-07_dp, 6.346474379082405e-07_dp, &
             2.0541345596170542e-06_dp, 5.766014636954599e-06_dp, 1.4643412114020715e-05_dp, &
             3.448763147103683e-05_dp, 7.653519444362885e-05_dp, 0.00016181183908035672_dp, &
             0.000328525972612215_dp, 0.0006443786762640081_dp, 0.001226705580173265_dp, &
             0.002274923610931946_dp, 0.004122069991043817_dp, 0.0073156290368224285_dp, &
             0.012742821283291895_dp, 0.02182275498434219_dp, 0.0367982863242317_dp, &
             0.06117559347037739_dp, 0.10038197186455844_dp, 0.16274809915012775_dp, &
             0.2609868667420739_dp, 0.41449318247761086_dp, 0.6532315053067045_dp, &
             1.025497177480197_dp, 1.617898036350469_dp, 2.6267115695473024_dp, &
             4.788593230920815_dp]

      CASE (1055)

         aw(:) = & ! a
            [7.255806770508953e-09_dp, 7.968751690254371e-08_dp, 4.2143545323811723e-07_dp, &
             1.6259224036456526e-06_dp, 5.205451325965015e-06_dp, 1.4683486874898775e-05_dp, &
             3.77384074915608e-05_dp, 9.024831669157842e-05_dp, 0.0002036775745890012_dp, &
             0.00043819771692071505_dp, 0.0009054446740473256_dp, 0.0018071732899671274_dp, &
             0.00349971087349948_dp, 0.006599665482911743_dp, 0.012154728167146342_dp, &
             0.021915941528508658_dp, 0.03876638107698634_dp, 0.06738851702935157_dp, &
             0.11529250579510457_dp, 0.19438619212069377_dp, 0.3233545319787663_dp, &
             0.5312515751248327_dp, 0.8629446025039647_dp, 1.3875540183136033_dp, &
             2.212398978202057_dp, 3.509561007426195_dp, 5.581587519480091_dp, &
             9.11517031241905_dp, &
             ! w
             2.317868606338727e-08_dp, 1.520580396166381e-07_dp, 6.221302037560175e-07_dp, &
             2.021798450107116e-06_dp, 5.690708668685123e-06_dp, 1.4480970181453707e-05_dp, &
             3.415759527617647e-05_dp, 7.589656396991277e-05_dp, 0.00016062566679841584_dp, &
             0.0003263992006667284_dp, 0.0006406820000084565_dp, 0.0012204564655341576_dp, &
             0.0022646242678760324_dp, 0.004105489753371218_dp, 0.00728952234115358_dp, &
             0.012702578220765676_dp, 0.021761994131306424_dp, 0.03670842371830825_dp, &
             0.06104546332957794_dp, 0.10019764087915804_dp, 0.16249311066481548_dp, &
             0.2606432599319929_dp, 0.41404368979692385_dp, 0.6526631609713396_dp, &
             1.0248058077707594_dp, 1.617091179753227_dp, 2.625802848770382_dp, &
             4.787556742587263_dp]

      CASE (1056)

         aw(:) = & ! a
            [7.22467210844994e-09_dp, 7.947864361800511e-08_dp, 4.2064485674580354e-07_dp, &
             1.6235366122649494e-06_dp, 5.199188870596663e-06_dp, 1.4668586543248157e-05_dp, &
             3.770549693885375e-05_dp, 9.01798003233954e-05_dp, 0.0002035417075392494_dp, &
             0.00043793915809042555_dp, 0.0009049698286859924_dp, 0.0018063281399514854_dp, &
             0.0034982482550036105_dp, 0.006597197956103633_dp, 0.012150661885505764_dp, &
             0.021909386078247235_dp, 0.03875603037140906_dp, 0.06737249787230323_dp, &
             0.11526819485575907_dp, 0.19435001020984807_dp, 0.32330173873475254_dp, &
             0.531176109293425_dp, 0.862839042021819_dp, 1.3874097611544784_dp, &
             2.2122067395532_dp, 3.509311624364863_dp, 5.5812727234009305_dp, &
             9.11478199860046_dp, &
             ! w
             2.30958621408986e-08_dp, 1.517439261365876e-07_dp, 6.211844102513912e-07_dp, &
             2.019351431824963e-06_dp, 5.685003937205477e-06_dp, 1.4468654689242246e-05_dp, &
             3.4132557441709815e-05_dp, 7.584808953199462e-05_dp, 0.00016053559252139304_dp, &
             0.0003262376392288452_dp, 0.0006404010916881625_dp, 0.00121998147852537_dp, &
             0.002263841254220773_dp, 0.004104228949625574_dp, 0.007287536693001169_dp, &
             0.012699516800963333_dp, 0.02175737106728934_dp, 0.03670158526817247_dp, &
             0.061035558951216755_dp, 0.10018360909762485_dp, 0.16247369770890274_dp, &
             0.2606170972448855_dp, 0.4140094610464392_dp, 0.6526198767561904_dp, &
             1.0247531480619902_dp, 1.6170297160254488_dp, 2.62573361583358_dp, &
             4.7874777650385605_dp]

      CASE (1057)

         aw(:) = & ! a
            [4.890645884174265e-09_dp, 5.346835755696374e-08_dp, 2.821978351887377e-07_dp, &
             1.0875205442970588e-06_dp, 3.47922372434538e-06_dp, 9.809111444203125e-06_dp, &
             2.5200818614422162e-05_dp, 6.0247106978709434e-05_dp, 0.00013593488916905727_dp, &
             0.0002923926196665419_dp, 0.0006040603417527699_dp, 0.0012054545016190379_dp, &
             0.0023341252566220017_dp, 0.004401104084490726_dp, 0.008104724263775355_dp, &
             0.0146120449234824_dp, 0.02584447795884748_dp, 0.04492239656634126_dp, &
             0.07685004766215034_dp, 0.1295606932646822_dp, 0.21549818705519827_dp, &
             0.35399477999268597_dp, 0.5748361381704128_dp, 0.9236374349617548_dp, &
             1.4701603719652196_dp, 2.3220859527032984_dp, 3.6514142021278686_dp, &
             5.760196125491916_dp, 9.335063986506384_dp, &
             ! w
             1.5592747619787515e-08_dp, 1.0187252224427987e-07_dp, 4.1618033732532483e-07_dp, &
             1.3513577701673915e-06_dp, 3.8014751828602334e-06_dp, 9.669478418697101e-06_dp, &
             2.280096644078986e-05_dp, 5.0649612218189246e-05_dp, 0.00010717079798347242_dp, &
             0.00021773713981556933_dp, 0.0004273255465689339_dp, 0.0008139170850266182_dp, &
             0.001510088797487049_dp, 0.002737318892475366_dp, 0.004859798056923315_dp, &
             0.008467857400591146_dp, 0.014505934426272083_dp, 0.024466934224819093_dp, &
             0.04068482019793737_dp, 0.06677171331936499_dp, 0.10826729851379355_dp, &
             0.17360299064162635_dp, 0.27554870456585295_dp, 0.43346466405956713_dp, &
             0.6771326482716927_dp, 1.0544845043191737_dp, 1.651652876061848_dp, &
             2.6646846836519025_dp, 4.8319158236313715_dp]

      CASE (1058)

         aw(:) = & ! a
            [4.812855434062239e-09_dp, 5.294584081007918e-08_dp, 2.802176646601784e-07_dp, &
             1.0815365309667173e-06_dp, 3.463491803738295e-06_dp, 9.771617479404539e-06_dp, &
             2.5117856498827222e-05_dp, 6.007405570457737e-05_dp, 0.0001355910160282717_dp, &
             0.0002917367421943136_dp, 0.0006028528525753048_dp, 0.0012032995465519022_dp, &
             0.0023303847416192157_dp, 0.004394772732234518_dp, 0.00809425233897039_dp, &
             0.01459509293922564_dp, 0.025817586509564217_dp, 0.04488055587535862_dp, &
             0.07678615792194951_dp, 0.1294649221516802_dp, 0.21535726033569366_dp, &
             0.3537912915321314_dp, 0.5745480419161182_dp, 0.9232379782668054_dp, &
             1.469618810399968_dp, 2.321369345978102_dp, 3.650490208013779_dp, &
             5.75903554438727_dp, 9.333637781576762_dp, &
             ! w
             1.538570715301579e-08_dp, 1.0108614145881527e-07_dp, 4.138088432326047e-07_dp, &
             1.3452112553195433e-06_dp, 3.7871188384446483e-06_dp, 9.638423671631228e-06_dp, &
             2.273769628946928e-05_dp, 5.052683465004879e-05_dp, 0.00010694208007246154_dp, &
             0.0002173257706342454_dp, 0.00042660811481252763_dp, 0.0008126998545068835_dp, &
             0.001508074577296664_dp, 0.0027340618254820637_dp, 0.004854643770594597_dp, &
             0.008459866600320706_dp, 0.014493789755704934_dp, 0.024448833854797996_dp, &
             0.040658368282798354_dp, 0.06673382869610278_dp, 0.10821418048318349_dp, &
             0.17353020474521405_dp, 0.2754514774056555_dp, 0.43333848626001725_dp, &
             0.6769742313120813_dp, 1.0542929284628535_dp, 1.6514302615746546_dp, &
             2.6644345207444156_dp, 4.831630351194341_dp]

      CASE (1059)

         aw(:) = & ! a
            [4.81276110248754e-09_dp, 5.2945202046964774e-08_dp, 2.802152366619104e-07_dp, &
             1.081529201523824e-06_dp, 3.4634726099545752e-06_dp, 9.771571870936095e-06_dp, &
             2.51177557007713e-05_dp, 6.0073845395588636e-05_dp, 0.000135590598185538_dp, &
             0.0002917359450692641_dp, 0.0006028513853327951_dp, 0.0012032969339251878_dp, &
             0.0023303802262736044_dp, 0.004394765119569222_dp, 0.008094239769377532_dp, &
             0.014595072602902532_dp, 0.025817554263529607_dp, 0.044880505700158856_dp, &
             0.07678608122763449_dp, 0.1294648069914634_dp, 0.2153570906165855_dp, &
             0.35379104633164854_dp, 0.574547694999163_dp, 0.9232374980633955_dp, &
             1.4696181607851027_dp, 2.3213684879705543_dp, 3.6504891031702975_dp, &
             5.759034158684595_dp, 9.333636080397362_dp, &
             ! w
             1.5385454907059592e-08_dp, 1.0108517789341093e-07_dp, 4.138059309175606e-07_dp, &
             1.3452037515986742e-06_dp, 3.787101394888971e-06_dp, 9.638385999605268e-06_dp, &
             2.2737619415454654e-05_dp, 5.052668544105883e-05_dp, 0.00010694180216120721_dp, &
             0.00021732527046511958_dp, 0.0004266072444953311_dp, 0.000812698388008673_dp, &
             0.001508072166863835_dp, 0.002734057931151093_dp, 0.004854637594291182_dp, &
             0.008459857023848858_dp, 0.01449377520265558_dp, 0.024448812122661354_dp, &
             0.040658336417427283_dp, 0.06673378293661669_dp, 0.10821411631716732_dp, &
             0.17353011705599297_dp, 0.27545136075760873_dp, 0.4333383355637861_dp, &
             0.6769740426793559_dp, 1.0542927003696698_dp, 1.6514299968479609_dp, &
             2.664434224189006_dp, 4.831630010270239_dp]

      CASE (1060)

         aw(:) = & ! a
            [3.8043751813390454e-09_dp, 3.935982447242644e-08_dp, 2.024349668685244e-07_dp, &
             7.690245020711723e-07_dp, 2.437458631967741e-06_dp, 6.826421584037535e-06_dp, &
             1.744949652962304e-05_dp, 4.154950959262124e-05_dp, 9.344161633358045e-05_dp, &
             0.00020044148569920938_dp, 0.000413131852722665_dp, 0.000822777831420505_dp, &
             0.001590331469624568_dp, 0.0029939508716928733_dp, 0.005505707351288482_dp, &
             0.009913753347021088_dp, 0.017514514361170186_dp, 0.030411643913131066_dp, &
             0.05197623310626743_dp, 0.0875485011299808_dp, 0.14549840038460096_dp, &
             0.23881662863734882_dp, 0.3874879198295163_dp, 0.6220257788034336_dp, &
             0.9887778185345365_dp, 1.558119903967937_dp, 2.438061819083998_dp, &
             3.80049706438324_dp, 5.946984950127678_dp, 9.564167753816543_dp, &
             ! w
             1.1852704834294101e-08_dp, 7.356055305793737e-08_dp, 2.9482860272177023e-07_dp, &
             9.469397224767065e-07_dp, 2.6443397183894273e-06_dp, 6.6900939628792655e-06_dp, &
             1.5709942271319508e-05_dp, 3.478116800331403e-05_dp, 7.339101807309114e-05_dp, &
             0.0001487591636135046_dp, 0.0002913650137459645_dp, 0.0005539849958262953_dp, &
             0.0010262412137048546_dp, 0.001857696988019733_dp, 0.0032940568223934585_dp, &
             0.00573325015168462_dp, 0.009811397221272458_dp, 0.016533288720877535_dp, &
             0.02746864794482401_dp, 0.04504494630168281_dp, 0.0729812821283275_dp, &
             0.11692841356831539_dp, 0.18541372406062892_dp, 0.2912550997315879_dp, &
             0.4537650240165651_dp, 0.7025282586173512_dp, 1.0851039652607064_dp, &
             1.6871556804263157_dp, 2.7045365564158415_dp, 4.87740447877141_dp]

      CASE (1061)

         aw(:) = & ! a
            [3.5809735952284415e-09_dp, 3.787489063278248e-08_dp, 1.968724594469968e-07_dp, &
             7.523314856628645e-07_dp, 2.3937582207948603e-06_dp, 6.722542793112886e-06_dp, &
             1.7220005807681834e-05_dp, 4.107120165645433e-05_dp, 9.249137894620668e-05_dp, &
             0.00019862861020103427_dp, 0.00040979200951305356_dp, 0.00081681087237226_dp, &
             0.0015799588274374335_dp, 0.0029763606007094916_dp, 0.005476546102478341_dp, &
             0.009866415656144939_dp, 0.01743917239759041_dp, 0.030293958537710636_dp, &
             0.0517956953143832_dp, 0.0872763809930841_dp, 0.14509534222877543_dp, &
             0.23823004558070357_dp, 0.3866495306253082_dp, 0.6208499089154033_dp, &
             0.9871614270503206_dp, 1.5559455525075347_dp, 2.4352045751962716_dp, &
             3.7968348342019396_dp, 5.9424073568190146_dp, 9.55856333864412_dp, &
             ! w
             1.125862999455319e-08_dp, 7.134668980105558e-08_dp, 2.8821659353862047e-07_dp, &
             9.298825682838519e-07_dp, 2.6046046931407997e-06_dp, 6.60426602004619e-06_dp, &
             1.55351881523148e-05_dp, 3.444205334146766e-05_dp, 7.275896803176011e-05_dp, &
             0.00014762125495544588_dp, 0.00028937767210740755_dp, 0.0005506068874225079_dp, &
             0.0010206382552741572_dp, 0.0018486111479582928_dp, 0.003279629672353728_dp, &
             0.0057107926839843415_dp, 0.009777100480438212_dp, 0.016481877007924722_dp, &
             0.027392990829464767_dp, 0.044935671557869906_dp, 0.07282647393180927_dp, &
             0.11671354435489172_dp, 0.18512205408422636_dp, 0.29086886468224243_dp, &
             0.45326774768488076_dp, 0.7019082997887867_dp, 1.0843586128473048_dp, &
             1.6862932529495267_dp, 2.703569508418598_dp, 4.876300367187299_dp]

      CASE (1062)

         aw(:) = & ! a
            [3.4218081238374378e-09_dp, 3.6811598558139175e-08_dp, 1.9286695269885455e-07_dp, &
             7.402667779447524e-07_dp, 2.362094506491098e-06_dp, 6.647136437188911e-06_dp, &
             1.7053177892824726e-05_dp, 4.072309582341635e-05_dp, 9.179915073818454e-05_dp, &
             0.00019730690462171858_dp, 0.00040735535774842434_dp, 0.0008124549292242439_dp, &
             0.0015723826713520172_dp, 0.0029635066413960004_dp, 0.005455227711640598_dp, &
             0.009831796087695964_dp, 0.017384053218071887_dp, 0.030207834056209516_dp, &
             0.05166353553803208_dp, 0.08707712613411644_dp, 0.1448001369546049_dp, &
             0.23780032521492653_dp, 0.3860352107713265_dp, 0.6199881348330097_dp, &
             0.9859765891006897_dp, 1.554351461688508_dp, 2.433109529412097_dp, &
             3.7941492064581297_dp, 5.939050123343085_dp, 9.554452707358008_dp, &
             ! w
             1.083513946628243e-08_dp, 6.97543876679885e-08_dp, 2.834372774876385e-07_dp, &
             9.17518668251658e-07_dp, 2.5757466615159235e-06_dp, 6.541841222994413e-06_dp, &
             1.540793816507618e-05_dp, 3.419488799874634e-05_dp, 7.229792943829625e-05_dp, &
             0.0001467906612950698_dp, 0.0002879261960369404_dp, 0.0005481383756672386_dp, &
             0.0010165420848826945_dp, 0.0018419660037770243_dp, 0.0032690741029568794_dp, &
             0.00569435619540819_dp, 0.009751991106685171_dp, 0.016444226636692653_dp, &
             0.027337570175861043_dp, 0.04485560558219221_dp, 0.07271301962139284_dp, &
             0.11655603983026591_dp, 0.1849082106933659_dp, 0.290585637633567_dp, &
             0.45290303336271104_dp, 0.7014535408561374_dp, 1.0838118079578096_dp, &
             1.6856605036749266_dp, 2.7028599681823913_dp, 4.875490268176401_dp]

      CASE (1063)

         aw(:) = & ! a
            [3.313449696441369e-09_dp, 3.6084985092898974e-08_dp, 1.9011848245333975e-07_dp, &
             7.319662375298938e-07_dp, 2.3402700804732265e-06_dp, 6.5950924450102756e-06_dp, &
             1.6937917269831713e-05_dp, 4.0482391279588363e-05_dp, 9.13201674642981e-05_dp, &
             0.00019639182767276086_dp, 0.00040566751230668245_dp, 0.0008094363001220381_dp, &
             0.001567130457787965_dp, 0.002954592531642921_dp, 0.005440439084450796_dp, &
             0.009807773782816203_dp, 0.01734579678708997_dp, 0.030148044184490956_dp, &
             0.05157176751102783_dp, 0.08693874245394853_dp, 0.14459507860339366_dp, &
             0.23750177950252874_dp, 0.38560835060933335_dp, 0.619389246882463_dp, &
             0.9851530823734852_dp, 1.5532433791700915_dp, 2.4316530721335226_dp, &
             3.792282014868708_dp, 5.936715825050437_dp, 9.551594408780353_dp, &
             ! w
             1.0546704808173441e-08_dp, 6.866271368160781e-08_dp, 2.801488375769307e-07_dp, &
             9.08994352329221e-07_dp, 2.5558224442953628e-06_dp, 6.498696158846561e-06_dp, &
             1.5319915663512187e-05_dp, 3.4023800023001035e-05_dp, 7.197861560344868e-05_dp, &
             0.0001462151133580047_dp, 0.0002869199902212454_dp, 0.0005464264937328414_dp, &
             0.0010137005027435165_dp, 0.0018373547896258224_dp, 0.003261747397035587_dp, &
             0.005682944719915628_dp, 0.009734554365967813_dp, 0.01641807568086544_dp, &
             0.027299069180751687_dp, 0.04479997363467851_dp, 0.07263417576679286_dp, &
             0.11644656718752751_dp, 0.18475955910637776_dp, 0.29038872887164047_dp, &
             0.45264944180886185_dp, 0.7011373065965806_dp, 1.0834315327128092_dp, &
             1.6852204304497582_dp, 2.702366471407691_dp, 4.874926835685036_dp]

      CASE (1064)

         aw(:) = & ! a
            [3.2277855412992547e-09_dp, 3.550888030440663e-08_dp, 1.8793259673205821e-07_dp, &
             7.253516753213124e-07_dp, 2.3228551361853644e-06_dp, 6.5535225680290114e-06_dp, &
             1.6845782954862133e-05_dp, 4.028986398950312e-05_dp, 9.093685818876606e-05_dp, &
             0.00019565921694746929_dp, 0.0004043157244581796_dp, 0.0008070179204678331_dp, &
             0.0015629214597109933_dp, 0.002947447197809483_dp, 0.005428582214486993_dp, &
             0.009788509829781904_dp, 0.01731511252382413_dp, 0.030100080494222314_dp, &
             0.05149813928704509_dp, 0.08682769703252141_dp, 0.1444305083555501_dp, &
             0.23726215112893237_dp, 0.3852656914762467_dp, 0.6189084434260947_dp, &
             0.9844918857635078_dp, 1.552353618967661_dp, 2.4304834878336483_dp, &
             3.7907824997861077_dp, 5.934841081920019_dp, 9.549298726046416_dp, &
             ! w
             1.0318598382924394e-08_dp, 6.779502854837367e-08_dp, 2.775281747463395e-07_dp, &
             9.021908920584991e-07_dp, 2.539903997731448e-06_dp, 6.464198494570064e-06_dp, &
             1.5249491483104842e-05_dp, 3.3886848279523055e-05_dp, 7.172290430559349e-05_dp, &
             0.0001457540387501337_dp, 0.0002861136612870025_dp, 0.0005450542951357844_dp, &
             0.0010114222190332068_dp, 0.001833656864684644_dp, 0.0032558706298341_dp, &
             0.005673789895536592_dp, 0.009720563478145415_dp, 0.01639708946423466_dp, &
             0.027268167609023647_dp, 0.04475531655415134_dp, 0.07257087824084146_dp, &
             0.11635867039221255_dp, 0.18464019275001078_dp, 0.2902305971479002_dp, &
             0.4524457720746747_dp, 0.7008833068892046_dp, 1.0831260766103628_dp, &
             1.6848669247367565_dp, 2.701970038647615_dp, 4.874474217948612_dp]

      CASE (1065)

         aw(:) = & ! a
            [3.1845944709614195e-09_dp, 3.062565373553227e-08_dp, 1.516597991871884e-07_dp, &
             5.638281921788058e-07_dp, 1.7619102914330155e-06_dp, 4.884415430972787e-06_dp, &
             1.2388762683313728e-05_dp, 2.93177603694674e-05_dp, 6.560108052179082e-05_dp, &
             0.00014012625836920818_dp, 0.0002877736974402615_dp, 0.0005713275090074288_dp, &
             0.0011012788665161126_dp, 0.002068220925307573_dp, 0.003795069536110126_dp, &
             0.006820127692169873_dp, 0.012027609022794141_dp, 0.0208504486006639_dp, &
             0.03558219035144457_dp, 0.05985204657293338_dp, 0.09934214331853568_dp, &
             0.16286180047593_dp, 0.2639456215156731_dp, 0.42321955832027414_dp, &
             0.6719029559685561_dp, 1.057042430222852_dp, 1.6495864416105692_dp, &
             2.5578338120635125_dp, 3.9535508586525485_dp, 6.137823273670679_dp, &
             9.797373954167822_dp, &
             ! w
             9.639585515201435e-09_dp, 5.564426171986406e-08_dp, 2.1671657013311257e-07_dp, &
             6.846527461220924e-07_dp, 1.8906083708667851e-06_dp, 4.743878011087015e-06_dp, &
             1.106861044577319e-05_dp, 2.437922508847552e-05_dp, 5.122249008306732e-05_dp, &
             0.00010344971163983378_dp, 0.00020198998743228388_dp, 0.0003830096568233367_dp, &
             0.0007078137690951454_dp, 0.0012785437951132554_dp, 0.002262752860781596_dp, &
             0.003931437728304965_dp, 0.006717275169726854_dp, 0.011302927563505908_dp, &
             0.018753737879374607_dp, 0.030715478540337324_dp, 0.049706865029238434_dp, &
             0.07954962370710848_dp, 0.12599863832538027_dp, 0.19766759346596147_dp, &
             0.3074108139011983_dp, 0.4744823575866795_dp, 0.7282655279429118_dp, &
             1.1159562098786406_dp, 1.7227781046723978_dp, 2.7444375108318373_dp, &
             4.922973584671284_dp]

      CASE (1066)

         aw(:) = & ! a
            [2.9458881131674385e-09_dp, 2.906530240635332e-08_dp, 1.4593184459039277e-07_dp, &
             5.468622061749822e-07_dp, 1.717874188294932e-06_dp, 4.7803525638991805e-06_dp, &
             1.2159817558190746e-05_dp, 2.8841990049930792e-05_dp, 6.465780353615138e-05_dp, &
             0.00013832902351679867_dp, 0.00028446502699414815_dp, 0.0005654173438094224_dp, &
             0.0010910018997140835_dp, 0.002050779640113027_dp, 0.0037661198641204194_dp, &
             0.006773053331664959_dp, 0.011952519903678717_dp, 0.020732831285982415_dp, &
             0.035401137878023126_dp, 0.05957800997713933_dp, 0.09893418889025589_dp, &
             0.16226444347439548_dp, 0.26308546276071054_dp, 0.42200220093309565_dp, &
             0.6702110375329138_dp, 1.0547360233206402_dp, 1.646507348990728_dp, &
             2.5538148960088454_dp, 3.9484294866487373_dp, 6.131452160191908_dp, &
             9.789602034397536_dp, &
             ! w
             9.005642291737438e-09_dp, 5.3354603431575985e-08_dp, 2.1000122734816408e-07_dp, &
             6.674934099465561e-07_dp, 1.8508769006338002e-06_dp, 4.658408264081491e-06_dp, &
             1.0895068739912581e-05_dp, 2.4043073728842693e-05_dp, 5.059662456362531e-05_dp, &
             0.00010232341264223534_dp, 0.00020002265023824425_dp, 0.0003796633444003026_dp, &
             0.0007022570399218758_dp, 0.001269517625173466_dp, 0.002248388129060536_dp, &
             0.003909013252393282_dp, 0.006682906607055666_dp, 0.011251182211780652_dp, &
             0.018677181014705547_dp, 0.030604177120792064_dp, 0.049547909487141245_dp, &
             0.07932678455799597_dp, 0.12569235682460742_dp, 0.19725561148141868_dp, &
             0.3068698475112988_dp, 0.47379121393155127_dp, 0.7274097295352671_dp, &
             1.1149331242713187_dp, 1.7215991698050224_dp, 2.743118289521767_dp, &
             4.921466557937733_dp]

      CASE (1067)

         aw(:) = & ! a
            [2.760833096851384e-09_dp, 2.7848245178789095e-08_dp, 1.414301330826447e-07_dp, &
             5.334593130728016e-07_dp, 1.6829612219911188e-06_dp, 4.6976293931347205e-06_dp, &
             1.1977444871006663e-05_dp, 2.8462370133722462e-05_dp, 6.390411829217585e-05_dp, &
             0.00013689134274264954_dp, 0.0002818156288265608_dp, 0.0005606806672925807_dp, &
             0.0010827591019067014_dp, 0.002036780960648585_dp, 0.0037428699902304777_dp, &
             0.006735226147189484_dp, 0.011892150445733848_dp, 0.020638226464835572_dp, &
             0.035255447215746576_dp, 0.059357409446227416_dp, 0.09860566460337492_dp, &
             0.1617832319207386_dp, 0.2623923300188903_dp, 0.4210209461010672_dp, &
             0.6688468953039179_dp, 1.0528759789801148_dp, 1.6440236015044547_dp, &
             2.550572394457297_dp, 3.9442968039747077_dp, 6.1263102785928645_dp, &
             9.783328939338427_dp, &
             ! w
             8.513909789892513e-09_dp, 5.155824710662649e-08_dp, 2.0469529777710544e-07_dp, &
             6.538810374831202e-07_dp, 1.8192703833677264e-06_dp, 4.590273504568213e-06_dp, &
             1.0756493342918919e-05_dp, 2.3774284291928655e-05_dp, 5.009560114978866e-05_dp, &
             0.000101420889588625_dp, 0.0001984448391114151_dp, 0.0003769775691457162_dp, &
             0.0006977941816459571_dp, 0.0012622639505348897_dp, 0.002236837977819785_dp, &
             0.0038909736588884115_dp, 0.006655245989882334_dp, 0.011209519060695982_dp, &
             0.01861551705303814_dp, 0.030514495716251963_dp, 0.04941978826954711_dp, &
             0.07914711646943993_dp, 0.12544534058228762_dp, 0.19692325828600513_dp, &
             0.306433331451065_dp, 0.4732333900737137_dp, 0.7267188735102531_dp, &
             1.1141070855616537_dp, 1.720647183722726_dp, 2.7420529582771356_dp, &
             4.920249585692938_dp]

      CASE (1068)

         aw(:) = & ! a
            [2.6147440821759635e-09_dp, 2.6882371212172404e-08_dp, 1.3783484470706793e-07_dp, &
             5.227096107067842e-07_dp, 1.6548771756333932e-06_dp, 4.630942195871751e-06_dp, &
             1.1830177155876862e-05_dp, 2.8155406645748684e-05_dp, 6.329399724863264e-05_dp, &
             0.00013572640878117768_dp, 0.00027966710013594423_dp, 0.0005568367211986735_dp, &
             0.0010760656322005085_dp, 0.0020254071263625813_dp, 0.0037239701244381335_dp, &
             0.006704462384005059_dp, 0.011843033408944017_dp, 0.020561226208445593_dp, &
             0.03513682631671154_dp, 0.059177739752580706_dp, 0.09833801657969836_dp, &
             0.1613910826071551_dp, 0.26182733830146143_dp, 0.42022090973196824_dp, &
             0.6677344410204007_dp, 1.0513588140698789_dp, 1.6419973379111372_dp, &
             2.547926704284454_dp, 3.9409243088453842_dp, 6.122113743341559_dp, &
             9.778208718461197_dp, &
             ! w
             8.125520493617971e-09_dp, 5.0125607829228896e-08_dp, 2.0043912351079124e-07_dp, &
             6.429260485340704e-07_dp, 1.7937761155751635e-06_dp, 4.535220384762442e-06_dp, &
             1.0644371000741364e-05_dp, 2.355656187392857e-05_dp, 4.968938614274313e-05_dp, &
             0.00010068856348158864_dp, 0.00019716367789800087_dp, 0.0003747954193808425_dp, &
             0.0006941662049026264_dp, 0.001256364364506214_dp, 0.002227439819893695_dp, &
             0.0038762892634982083_dp, 0.0066327217270079905_dp, 0.011175580971311345_dp, &
             0.01856527098844093_dp, 0.03044139888774772_dp, 0.0493153321348833_dp, &
             0.07900059787474112_dp, 0.12524385266697957_dp, 0.19665210293258134_dp, &
             0.3060771211809809_dp, 0.47277810488549005_dp, 0.7261549179045669_dp, &
             1.1134326871119178_dp, 1.7198698825468328_dp, 2.7411830675375377_dp, &
             4.919255886803094_dp]

      CASE (1069)

         aw(:) = & ! a
            [2.178746404425297e-09_dp, 2.3968396986536237e-08_dp, 1.2685398822346647e-07_dp, &
             4.896103947130975e-07_dp, 1.5679208564265335e-06_dp, 4.423609788102947e-06_dp, &
             1.137085736563612e-05_dp, 2.719554786584364e-05_dp, 6.138213025977332e-05_dp, &
             0.00013206943564910908_dp, 0.0002729120068142774_dp, 0.0005447348862068517_dp, &
             0.0010549677041431554_dp, 0.001989518778811104_dp, 0.003664278076414874_dp, &
             0.006607217019589959_dp, 0.011687651808745097_dp, 0.02031746277357493_dp, &
             0.03476105701251126_dp, 0.05860823625502674_dp, 0.09748917366088942_dp, &
             0.16014674372927154_dp, 0.2600336914185814_dp, 0.4176799538807555_dp, &
             0.6641997811550416_dp, 1.0465364281199305_dp, 1.6355545486589613_dp, &
             2.5395117933590634_dp, 3.9301948929695447_dp, 6.108759809155273_dp, &
             9.761912794347234_dp, &
             ! w
             6.965025659840846e-09_dp, 4.5761458646468266e-08_dp, 1.87330758040871e-07_dp, &
             6.089763817399534e-07_dp, 1.714428243517235e-06_dp, 4.3633162828523385e-06_dp, &
             1.0293364992813606e-05_dp, 2.2873529794168543e-05_dp, 4.8412764001083314e-05_dp, &
             9.838357701864868e-05_dp, 0.00019312593782735892_dp, 0.0003679101562852507_dp, &
             0.0006827072257299311_dp, 0.001237713346657453_dp, 0.0021977036641679646_dp, &
             0.0038297921004975173_dp, 0.006561350807065886_dp, 0.011067975509946986_dp, &
             0.01840586518757632_dp, 0.030209372602138566_dp, 0.04898359609782936_dp, &
             0.07853505838980228_dp, 0.12460337477334948_dp, 0.19578981614238863_dp, &
             0.3049439242826075_dp, 0.4713292208497163_dp, 0.7243596505986653_dp, &
             1.111285289765135_dp, 1.7173943643863407_dp, 2.738412414617402_dp, &
             4.916090972334704_dp]

      CASE (1070)

         aw(:) = & ! a
            [2.767021311761996e-09_dp, 2.4708975786972106e-08_dp, 1.1715183051303062e-07_dp, &
             4.2447516079544257e-07_dp, 1.303898160934604e-06_dp, 3.570081051242626e-06_dp, &
             8.96932126783791e-06_dp, 2.106524604322009e-05_dp, 4.6842552847563644e-05_dp, &
             9.953509164964869e-05_dp, 0.00020349990649677944_dp, 0.0004024498461211982_dp, &
             0.0007731124503172073_dp, 0.001447531076082164_dp, 0.00264895239862103_dp, &
             0.004748820747172448_dp, 0.008356206843702673_dp, 0.014456578827786474_dp, &
             0.024624941477455798_dp, 0.041350098897905585_dp, 0.0685236446115503_dp, &
             0.11217139021960207_dp, 0.18153944901278746_dp, 0.29069710037013596_dp, &
             0.46089321106042186_dp, 0.7240238333311021_dp, 1.1277950057862964_dp, &
             1.743682505412404_dp, 2.680234509236382_dp, 4.109073574690564_dp, &
             6.33083549640918_dp, 10.032391398713935_dp, &
             ! w
             8.150999104992031e-09_dp, 4.346572438827398e-08_dp, 1.6361499621751024e-07_dp, &
             5.067302734016744e-07_dp, 1.3804038928014288e-06_dp, 3.4289509286285853e-06_dp, &
             7.937899351333974e-06_dp, 1.737282025704644e-05_dp, 3.6309224313942096e-05_dp, &
             7.300292161573111e-05_dp, 0.00014199196745738165_dp, 0.00026833559290603755_dp, &
             0.0004944161454058322_dp, 0.0008907052621467972_dp, 0.0015725935638293178_dp, &
             0.00272640025274994_dp, 0.004649149834610824_dp, 0.007808815366611893_dp, &
             0.012934741800478789_dp, 0.021152242809125036_dp, 0.03418142604174922_dp, &
             0.05462891717035486_dp, 0.08641362306696478_dp, 0.13538666259343973_dp, &
             0.2102378635395208_dp, 0.3238468296717728_dp, 0.49540018312551176_dp, &
             0.7540785094856343_dp, 1.146728014134649_dp, 1.758165262786302_dp, &
             2.7839953281688774_dp, 4.968176073119719_dp]

      CASE (1071)

         aw(:) = & ! a
            [2.534503186791505e-09_dp, 2.321563973802394e-08_dp, 1.1179910132934571e-07_dp, &
             4.088780633960722e-07_dp, 1.2638608972796655e-06_dp, 3.4762066051290014e-06_dp, &
             8.763972379425872e-06_dp, 2.0640328828488407e-05_dp, 4.600277497162487e-05_dp, &
             9.793877985922164e-05_dp, 0.00020056592776003925_dp, 0.0003972144126349182_dp, &
             0.0007640133514451698_dp, 0.001432088971669627_dp, 0.0026233091800368915_dp, &
             0.004707083842309034_dp, 0.008289536495750269_dp, 0.014351944503419773_dp, &
             0.0244634668264697_dp, 0.04110491761386775_dp, 0.06815721111979134_dp, &
             0.11163224331081062_dp, 0.18075853362830807_dp, 0.28958393904764407_dp, &
             0.4593325088020153_dp, 0.7218735044673129_dp, 1.1248870617375066_dp, &
             1.7398284893138491_dp, 2.6752366118626445_dp, 4.102740081812762_dp, &
             6.322992411987514_dp, 10.022857464887233_dp, &
             ! w
             7.534212937325643e-09_dp, 4.131370544223117e-08_dp, 1.5744656277320748e-07_dp, &
             4.911625241211665e-07_dp, 1.3446489720354782e-06_dp, 3.3524734770322477e-06_dp, &
             7.783254550485835e-06_dp, 1.7074161012157894e-05_dp, 3.575431911249668e-05_dp, &
             7.200566953718451e-05_dp, 0.00014025130058356822_dp, 0.0002653753058103866_dp, &
             0.0004894986022176733_dp, 0.0008827102049387624_dp, 0.0015598515721405553_dp, &
             0.0027064691707863763_dp, 0.004618522428644687_dp, 0.007762548714526475_dp, &
             0.012866005347172915_dp, 0.02105179548009364_dp, 0.03403705747879731_dp, &
             0.054424929786357476_dp, 0.08613049685815177_dp, 0.13500113812854536_dp, &
             0.20972379415000655_dp, 0.3231772520379866_dp, 0.494551019145088_dp, &
             0.7530338844949883_dp, 1.1454859192433933_dp, 1.756739529152332_dp, &
             2.7824030356415865_dp, 4.966356070261009_dp]

      CASE (1072)

         aw(:) = & ! a
            [2.351316369674525e-09_dp, 2.2031837668929383e-08_dp, 1.0751939234791454e-07_dp, &
             3.9633088222135546e-07_dp, 1.231512663859948e-06_dp, 3.400114440501601e-06_dp, &
             8.597100030461891e-06_dp, 2.0294320873264912e-05_dp, 4.5317783094516824e-05_dp, &
             9.663481573201287e-05_dp, 0.00019816629177047226_dp, 0.00039292780841978354_dp, &
             0.0007565561390789995_dp, 0.0014194224681530778_dp, 0.002602258934193113_dp, &
             0.004672798585379616_dp, 0.008234734640004161_dp, 0.014265887012674639_dp, &
             0.024330589637776113_dp, 0.040903059113273056_dp, 0.06785538790234294_dp, &
             0.11118797215180493_dp, 0.18011478607007228_dp, 0.2886659690892859_dp, &
             0.4580450361225862_dp, 0.7200990711365218_dp, 1.1224867542980737_dp, &
             1.7366464248115516_dp, 2.6711091328279615_dp, 4.097508551162299_dp, &
             6.316512866248624_dp, 10.014980018418834_dp, &
             ! w
             7.048046035826438e-09_dp, 3.9596819693179754e-08_dp, 1.5248324703066553e-07_dp, &
             4.785751156699287e-07_dp, 1.3156409112993806e-06_dp, 3.290266962060145e-06_dp, &
             7.657208330985624e-06_dp, 1.683032154704649e-05_dp, 3.530062350139257e-05_dp, &
             7.118931180164291e-05_dp, 0.00013882486393646069_dp, 0.00026294714555961_dp, &
             0.0004854616505473604_dp, 0.000876141927764261_dp, 0.0015493764134011563_dp, &
             0.0026900737994686964_dp, 0.0045933140377506215_dp, 0.00772444845110307_dp, &
             0.012809374251232783_dp, 0.020969001418636966_dp, 0.03391801195004029_dp, &
             0.054256657912296294_dp, 0.0858968578988372_dp, 0.13468289130413266_dp, &
             0.20929930051770487_dp, 0.3226241849933072_dp, 0.4938494269633953_dp, &
             0.7521705952490426_dp, 1.1444592379802128_dp, 1.7555608911029867_dp, &
             2.7810866134099888_dp, 4.964851421352252_dp]

      CASE (1073)

         aw(:) = & ! a
            [2.203466922630409e-09_dp, 2.107110658930233e-08_dp, 1.0402038526903055e-07_dp, &
             3.8601902338045693e-07_dp, 1.2048298953592203e-06_dp, 3.3371778788600243e-06_dp, &
             8.458784680003031e-06_dp, 2.0007032197518567e-05_dp, 4.4748225524209116e-05_dp, &
             9.554928256944187e-05_dp, 0.00019616654394425314_dp, 0.00038935229668379345_dp, &
             0.0007503309859977787_dp, 0.001408841118250019_dp, 0.0025846626582887514_dp, &
             0.004644122249148056_dp, 0.008188873854339522_dp, 0.014193835136289927_dp, &
             0.02421928830097043_dp, 0.0407339078163885_dp, 0.06760237307329621_dp, &
             0.11081541355739924_dp, 0.17957477283025997_dp, 0.28789568715314023_dp, &
             0.4569643927462949_dp, 0.7186093046317348_dp, 1.1204710328876515_dp, &
             1.7339736139458237_dp, 2.667641530602976_dp, 4.093112664163765_dp, &
             6.3110675575914605_dp, 10.008359228463513_dp, &
             ! w
             6.655484261298561e-09_dp, 3.819567852823219e-08_dp, 1.4840343020765387e-07_dp, &
             4.681858428409442e-07_dp, 1.2916301230405307e-06_dp, 3.238665369641527e-06_dp, &
             7.552469975064844e-06_dp, 1.6627415282919965e-05_dp, 3.492263914968088e-05_dp, &
             7.050849025297844e-05_dp, 0.0001376341953743411_dp, 0.00026091873774473836_dp, &
             0.0004820869630520325_dp, 0.0008706477514164255_dp, 0.0015406092948257455_dp, &
             0.002676344744483253_dp, 0.004572195261905831_dp, 0.0076925154710571165_dp, &
             0.012761891048165845_dp, 0.020899555736440662_dp, 0.03381812493130791_dp, &
             0.05411542124864559_dp, 0.08570069714606018_dp, 0.13441561927194157_dp, &
             0.20894270560018569_dp, 0.322159468563459_dp, 0.49325978005594023_dp, &
             0.7514449082903101_dp, 1.1435960617333973_dp, 1.7545698415752757_dp, &
             2.7799796446010485_dp, 4.963586196670519_dp]

      CASE (1074)

         aw(:) = & ! a
            [1.5590197220129705e-09_dp, 1.6812521913433517e-08_dp, 8.818391606430967e-08_dp, &
             3.386777849441256e-07_dp, 1.081107279119066e-06_dp, 3.0431937273604944e-06_dp, &
             7.808974574632904e-06_dp, 1.8651064556230853e-05_dp, 4.204961818310387e-05_dp, &
             9.038915343256561e-05_dp, 0.00018663394058807532_dp, 0.0003722663790061768_dp, &
             0.0007205191180170731_dp, 0.0013580699568548778_dp, 0.002500086356581223_dp, &
             0.004506073690889394_dp, 0.007967783986710286_dp, 0.013846027584983215_dp, &
             0.023681372283795212_dp, 0.03991550066149132_dp, 0.06637695508261407_dp, &
             0.10900930018314296_dp, 0.17695455161127305_dp, 0.28415510038379477_dp, &
             0.4517126448115705_dp, 0.7113641614453075_dp, 1.1106616368066464_dp, &
             1.7209588171500714_dp, 2.6507477180992725_dp, 4.071686600754953_dp, &
             6.284516586003154_dp, 9.976067400722108_dp, &
             ! w
             4.941688849815975e-09_dp, 3.188453405153541e-08_dp, 1.2966459529352903e-07_dp, &
             4.199349641021897e-07_dp, 1.1792515540793927e-06_dp, 2.995732540485393e-06_dp, &
             7.057074391486454e-06_dp, 1.5664023586120457e-05_dp, 3.3122203675255304e-05_dp, &
             6.725662263132082e-05_dp, 0.00013193347072925867_dp, 0.0002511865724024704_dp, &
             0.0004658650661061385_dp, 0.0008441931610852995_dp, 0.001498331163696893_dp, &
             0.0026100467203816675_dp, 0.004470082816615264_dp, 0.00753793480587778_dp, &
             0.012531787667315511_dp, 0.020562686208356728_dp, 0.033333139034003735_dp, &
             0.053429073318249316_dp, 0.08474666813689917_dp, 0.13311475165594086_dp, &
             0.20720585207003916_dp, 0.3198944994834913_dp, 0.49038418594386834_dp, &
             0.7479039917633125_dp, 1.1393824285088334_dp, 1.749730456221594_dp, &
             2.7745733663071763_dp, 4.957407313118424_dp]

      CASE (1075)

         aw(:) = & ! a
            [1.479683649697902e-09_dp, 1.6278008122146516e-08_dp, 8.615220829458904e-08_dp, &
             3.3251628207071743e-07_dp, 1.0648450488017557e-06_dp, 3.004270865022241e-06_dp, &
             7.722456822337955e-06_dp, 1.8469710681822967e-05_dp, 4.168734522971689e-05_dp, &
             8.969425081797155e-05_dp, 0.0001853467303496863_dp, 0.00036995378510729657_dp, &
             0.0007164756713893742_dp, 0.0013511710277301961_dp, 0.0024885748322353617_dp, &
             0.004487256006984869_dp, 0.007937605727405291_dp, 0.013798493564046305_dp, &
             0.023607772592359827_dp, 0.03980340519027969_dp, 0.06620894875596875_dp, &
             0.10876145620961274_dp, 0.17659469006056397_dp, 0.28364096700054375_dp, &
             0.4509902833033633_dp, 0.7103669453765807_dp, 1.1093106441791758_dp, &
             1.719165355243869_dp, 2.648418555980675_dp, 4.0687313008763955_dp, &
             6.280853112112665_dp, 9.971610596180174_dp, &
             ! w
             4.730259456981354e-09_dp, 3.1078649871532e-08_dp, 1.2722468500418828e-07_dp, &
             4.13583047959329e-07_dp, 1.1643447419542326e-06_dp, 2.963322847463187e-06_dp, &
             6.990683711042888e-06_dp, 1.5534435365376526e-05_dp, 3.287926976195174e-05_dp, &
             6.681668055560068e-05_dp, 0.00013116044841311768_dp, 0.00024986421384895415_dp, &
             0.00046365695555379473_dp, 0.0008405863897573469_dp, 0.0014925586633026797_dp, &
             0.0026009826759276194_dp, 0.004456105483221565_dp, 0.007516752041572211_dp, &
             0.012500223434600246_dp, 0.02051643253409508_dp, 0.03326648949207963_dp, &
             0.05333467369887276_dp, 0.08461535079090401_dp, 0.132935564560582_dp, &
             0.206966449248565_dp, 0.3195821076971402_dp, 0.4899873486084814_dp, &
             0.7474150925905267_dp, 1.138800404131442_dp, 1.749061796714666_dp, &
             2.7738262678122343_dp, 4.956553485781041_dp]

      CASE (1076)

         aw(:) = & ! a
            [2.46193468149273e-09_dp, 2.0517012422421106e-08_dp, 9.287924072309443e-08_dp, &
             3.270507112276594e-07_dp, 9.85364443448603e-07_dp, 2.659993261324006e-06_dp, &
             6.610267053741911e-06_dp, 1.5389598264383788e-05_dp, 3.397595269753531e-05_dp, &
             7.17582291905975e-05_dp, 0.00014594898237369207_dp, 0.00028733283396248875_dp, &
             0.0005497802425245835_dp, 0.0010257462855167148_dp, 0.0018711613306756434_dp, &
             0.003344888018347371_dp, 0.005870527663059019_dp, 0.01013217833576613_dp, &
             0.01722129639833083_dp, 0.02885982141492794_dp, 0.047736208655418395_dp, &
             0.07800739824654102_dp, 0.12604305799805285_dp, 0.2015216621653775_dp, &
             0.31903598360084906_dp, 0.5004377357962535_dp, 0.7782693884430489_dp, &
             1.2008567827791354_dp, 1.8401592102840052_dp, 2.8049374107325544_dp, &
             4.266656832885215_dp, 6.525528510362614_dp, 10.268635671946859_dp, &
             ! w
             7.084603793567241e-09_dp, 3.487794902176719e-08_dp, 1.2641993365178165e-07_dp, &
             3.828954860328335e-07_dp, 1.0270959992861877e-06_dp, 2.522090631405802e-06_dp, &
             5.785992814360229e-06_dp, 1.2570535838393628e-05_dp, 2.611220585614889e-05_dp, &
             5.2228599126264525e-05_dp, 0.00010113024587310934_dp, 0.0001903654408516313_dp, &
             0.00034953570526057763_dp, 0.0006277452204999214_dp, 0.0011052254890226505_dp, &
             0.001911272019678177_dp, 0.0032516344763093307_dp, 0.005449942858830312_dp, &
             0.009009806228701298_dp, 0.01470713146731143_dp, 0.023726317579465396_dp, &
             0.037859771247372616_dp, 0.059798399966042054_dp, 0.09355265288967414_dp, &
             0.1450620168584118_dp, 0.223082680822343_dp, 0.3405092257401168_dp, &
             0.5164531885374843_dp, 0.7798927693583374_dp, 1.1773391573927072_dp, &
             1.7932339000678958_dp, 2.8231233080187206_dp, 5.012913153382189_dp]

      CASE (1077)

         aw(:) = & ! a
            [2.240246009407978e-09_dp, 1.911716572413108e-08_dp, 8.798785735352122e-08_dp, &
             3.130606824505942e-07_dp, 9.499146463451981e-07_dp, 2.577649827664465e-06_dp, &
             6.43139936519414e-06_dp, 1.5021452072385664e-05_dp, 3.325136856381082e-05_dp, &
             7.038525110519707e-05_dp, 0.00014343154246222533_dp, 0.00028284854831156904_dp, &
             0.0005419958546086785_dp, 0.0010125442131242175_dp, 0.0018492419502684682_dp, &
             0.0033092017804164714_dp, 0.0058134792438991915_dp, 0.010042532824218934_dp, &
             0.017082707058294365_dp, 0.028648893597049602_dp, 0.04742002130513469_dp, &
             0.07754043333365111_dp, 0.1253635545171844_dp, 0.20054753702799782_dp, &
             0.31766067416630916_dp, 0.4985266434210604_dp, 0.7756580384179792_dp, &
             1.1973522766790134_dp, 1.8355467445790619_dp, 2.7989929149691033_dp, &
             4.259163922447078_dp, 6.516290305825923_dp, 10.257443882247975_dp, &
             ! w
             6.497191821722683e-09_dp, 3.289792516604182e-08_dp, 1.2089209590426751e-07_dp, &
             3.6914665131848886e-07_dp, 9.958240964149661e-07_dp, 2.455668881337165e-06_dp, &
             5.6523795315326065e-06_dp, 1.2313502555937632e-05_dp, 2.563603611678634e-05_dp, &
             5.13746641325023e-05_dp, 9.964189363278133e-05_dp, 0.00018783641677618933_dp, &
             0.0003453358496901147_dp, 0.0006209155076820697_dp, 0.0010943327010421104_dp, &
             0.0018942117366776293_dp, 0.003225370091752421_dp, 0.005410168831079731_dp, &
             0.008950527050063316_dp, 0.014620155973246188_dp, 0.023600684391197304_dp, &
             0.037681149860563035_dp, 0.05954855927837189_dp, 0.09320916138627404_dp, &
             0.1445984396918395_dp, 0.22246965860778173_dp, 0.339716909680212_dp, &
             0.5154554335925028_dp, 0.7786729731013715_dp, 1.1758962184769854_dp, &
             1.791583740402769_dp, 2.821283719181297_dp, 5.010809265448467_dp]

      CASE (1078)

         aw(:) = & ! a
            [2.064815945181132e-09_dp, 1.8002849655204384e-08_dp, 8.405879228342665e-08_dp, &
             3.017456102797898e-07_dp, 9.211005326016554e-07_dp, 2.510469276566396e-06_dp, &
             6.28503922911925e-06_dp, 1.4719492534982855e-05_dp, 3.2655867959765656e-05_dp, &
             6.925495480983726e-05_dp, 0.0001413560414825351_dp, 0.00027914674866869087_dp, &
             0.0005355625244674178_dp, 0.0010016224548443596_dp, 0.001831092066095918_dp, &
             0.003279628172100598_dp, 0.0057661670231325644_dp, 0.009968135560880038_dp, &
             0.01696761827824179_dp, 0.028473630397756598_dp, 0.04715715407721535_dp, &
             0.07715201955908634_dp, 0.12479808988868402_dp, 0.19973654204614608_dp, &
             0.31651521192340626_dp, 0.4969343359224441_dp, 0.7734815141073117_dp, &
             1.1944303687195488_dp, 1.8316999277316117_dp, 2.7940338698600864_dp, &
             4.252911727925997_dp, 6.508580368362125_dp, 10.248102184709344_dp, &
             ! w
             6.032132749024786e-09_dp, 3.1311457805993405e-08_dp, 1.1642005548157246e-07_dp, &
             3.579613224696849e-07_dp, 9.702831771593452e-07_dp, 2.401257434534262e-06_dp, &
             5.542663964784802e-06_dp, 1.210202427465438e-05_dp, 2.524360612108806e-05_dp, &
             5.066989430900562e-05_dp, 9.841199126095584e-05_dp, 0.00018574424920047322_dp, &
             0.00034185805095052657_dp, 0.0006152549921834014_dp, 0.0010852974611181047_dp, &
             0.0018800504388205874_dp, 0.0032035541664586996_dp, 0.005377111138690664_dp, &
             0.00890122992125947_dp, 0.014547788202126216_dp, 0.02349610023859743_dp, &
             0.037532387110019355_dp, 0.05934039291287655_dp, 0.0929228500596027_dp, &
             0.14421188680078023_dp, 0.2219583101552472_dp, 0.3390557869483171_dp, &
             0.5146226388160611_dp, 0.7776545775661803_dp, 1.1746912601382353_dp, &
             1.7902055218005835_dp, 2.819747171042936_dp, 5.009052002052361_dp]

      CASE (1079)

         aw(:) = & ! a
            [1.9224136769688107e-09_dp, 1.7093474902647872e-08_dp, 8.082681499804011e-08_dp, &
             2.92382923262524e-07_dp, 8.97157038715647e-07_dp, 2.454466770639545e-06_dp, &
             6.162726671862985e-06_dp, 1.446663427982735e-05_dp, 3.2156358612527e-05_dp, &
             6.83054948001992e-05_dp, 0.00013961043957989898_dp, 0.00027602997238821526_dp, &
             0.0005301407121021795_dp, 0.0009924100643079732_dp, 0.001815771054318544_dp, &
             0.0032546465754189646_dp, 0.005726175810288809_dp, 0.009905213876563405_dp, &
             0.016870229645094017_dp, 0.028325248764686145_dp, 0.046934503292588414_dp, &
             0.07682289014332723_dp, 0.12431874389350313_dp, 0.19904880595912292_dp, &
             0.3155435086442959_dp, 0.49558313817916533_dp, 0.7716340131507327_dp, &
             1.1919494826458479_dp, 1.8284329190814448_dp, 2.789821334752695_dp, &
             4.247599687785534_dp, 6.5020287544030415_dp, 10.240162992183818_dp, &
             ! w
             5.6544697646017435e-09_dp, 3.000926193822514e-08_dp, 1.127188533033903e-07_dp, &
             3.4865970299635663e-07_dp, 9.489727080444396e-07_dp, 2.3557429586618757e-06_dp, &
             5.450701804999332e-06_dp, 1.1924468974900441e-05_dp, 2.491365948607449e-05_dp, &
             5.007661931457484e-05_dp, 9.73755642817493e-05_dp, 0.00018397955532893916_dp, &
             0.0003389221740777272_dp, 0.0006104729520254755_dp, 0.0010776592794382507_dp, &
             0.0018680714472541708_dp, 0.003185089761906451_dp, 0.005349117507557122_dp, &
             0.008859464546060155_dp, 0.01448644971121428_dp, 0.023407418607862676_dp, &
             0.03740619561761807_dp, 0.059163746882612736_dp, 0.09267980904419874_dp, &
             0.14388364900742137_dp, 0.22152397363562518_dp, 0.3384940771515369_dp, &
             0.5139148919915394_dp, 0.7767889053510988_dp, 1.1736668141837074_dp, &
             1.7890336164847918_dp, 2.818440552988326_dp, 5.007557728900876_dp]

      CASE (1080)

         aw(:) = & ! a
            [1.261030077714601e-09_dp, 1.2793937998754653e-08_dp, 6.516649566813368e-08_dp, &
             2.4621038055808034e-07_dp, 7.775933394609482e-07_dp, 2.172185169231488e-06_dp, &
             5.541673297875507e-06_dp, 1.3175074497154998e-05_dp, 2.9592307805657116e-05_dp, &
             6.341130181996039e-05_dp, 0.00013057974442088255_dp, 0.00025985454092306183_dp, &
             0.0005019237795876626_dp, 0.0009443458540415964_dp, 0.0017356564226109629_dp, &
             0.003123750292819559_dp, 0.005516245220402361_dp, 0.009574351312611156_dp, &
             0.016357329788737593_dp, 0.0275426673392691_dp, 0.04575865116702402_dp, &
             0.07508255383048805_dp, 0.12178118273769663_dp, 0.19540415195153182_dp, &
             0.3103888029196236_dp, 0.4884085796170796_dp, 0.7618156715683803_dp, &
             1.1787545186576538_dp, 1.8110441627913738_dp, 2.7673853965454445_dp, &
             4.219291917049886_dp, 6.4670992068777755_dp, 10.197820536344981_dp, &
             ! w
             3.897830549575035e-09_dp, 2.3738634233636e-08_dp, 9.445515329659807e-08_dp, &
             3.021123361058843e-07_dp, 8.412797520957919e-07_dp, 2.1240098832730693e-06_dp, &
             4.979685373667982e-06_dp, 1.1010584593813334e-05_dp, 2.3208374123614428e-05_dp, &
             4.6999454942670736e-05_dp, 9.198323953647233e-05_dp, 0.00017477313588458253_dp, &
             0.00032356845720150034_dp, 0.0005854098146343251_dp, 0.0010375478210232424_dp, &
             0.0018050514332640386_dp, 0.0030877909185145374_dp, 0.005201380646780763_dp, &
             0.008638738470225492_dp, 0.014161860293578658_dp, 0.02293756737668983_dp, &
             0.036736854113909534_dp, 0.05822579313669984_dp, 0.09138802916483149_dp, &
             0.14213741531143836_dp, 0.21921127457447104_dp, 0.3355007314007439_dp, &
             0.5101405228480381_dp, 0.7721693129240349_dp, 1.1681969906276446_dp, &
             1.7827740424660827_dp, 2.8114601073233194_dp, 4.9995752593999_dp]

      CASE (1081)

         aw(:) = & ! a
            [1.0566655865248073e-09_dp, 1.1429446718821667e-08_dp, 6.003152656757956e-08_dp, &
             2.307312190308448e-07_dp, 7.368890784250718e-07_dp, 2.0749868676338605e-06_dp, &
             5.325929053394356e-06_dp, 1.2723204286686793e-05_dp, 2.868994076455378e-05_dp, &
             6.168028886922652e-05_dp, 0.00012737199574422252_dp, 0.0002540874507998262_dp, &
             0.000491830300236805_dp, 0.0009271023133570745_dp, 0.0017068386891889334_dp, &
             0.0030765538769786777_dp, 0.005440387856828313_dp, 0.009454558847580249_dp, &
             0.016171290642443614_dp, 0.027258332659636383_dp, 0.0453307659406808_dp, &
             0.07444834300772185_dp, 0.12085520773362532_dp, 0.19407252797301192_dp, &
             0.3085032661998795_dp, 0.4857813537342025_dp, 0.7582167058469943_dp, &
             1.1739133416178547_dp, 1.804658907142145_dp, 2.7591405783368064_dp, &
             4.208882545107766_dp, 6.4542480086803975_dp, 10.182235566361792_dp, &
             ! w
             3.3536290243901775e-09_dp, 2.1697913917655122e-08_dp, 8.832811534902565e-08_dp, &
             2.862265243919859e-07_dp, 8.040858507665811e-07_dp, 2.043252732175792e-06_dp, &
             4.814364128606313e-06_dp, 1.0687940871855503e-05_dp, 2.2603367111934036e-05_dp, &
             4.5903135360071666e-05_dp, 9.005506725395092e-05_dp, 0.00017147055947368438_dp, &
             0.0003180449760528093_dp, 0.0005763703232462657_dp, 0.0010230474492465636_dp, &
             0.0017822217168556931_dp, 0.0030524755923080583_dp, 0.005147663784744834_dp, &
             0.00855835199223758_dp, 0.014043468966128462_dp, 0.022765951832222037_dp, &
             0.036492052556536404_dp, 0.05788232888727132_dp, 0.09091445383858077_dp, &
             0.14149654191415614_dp, 0.2183616481154842_dp, 0.33440002020304904_dp, &
             0.5087514244690141_dp, 0.7704678570840019_dp, 1.166181126239908_dp, &
             1.7804660879041752_dp, 2.8088857975140793_dp, 4.996631628712299_dp]

      CASE (1082)

         aw(:) = & ! a
            [1.0108113209759398e-09_dp, 1.1119941118942126e-08_dp, 5.8852869505338446e-08_dp, &
             2.2715073464609125e-07_dp, 7.27424034697187e-07_dp, 2.0522975046032946e-06_dp, &
             5.275416096488924e-06_dp, 1.261715168325287e-05_dp, 2.8477736732801676e-05_dp, &
             6.127253362733354e-05_dp, 0.0001266152921605661_dp, 0.00025272529228248535_dp, &
             0.0004894436299305439_dp, 0.0009230209463726673_dp, 0.001700011801374611_dp, &
             0.0030653641684659986_dp, 0.0054223899133845005_dp, 0.009426117969241535_dp, &
             0.016127094667807193_dp, 0.027190747324971835_dp, 0.04522900620842918_dp, &
             0.07429744201903152_dp, 0.12063478680154299_dp, 0.19375541282834224_dp, &
             0.30805406608466984_dp, 0.48515523003175626_dp, 0.7573587063924442_dp, &
             1.1727588376783455_dp, 1.8031357461248292_dp, 2.757173333766668_dp, &
             4.206398288168869_dp, 6.451180444982384_dp, 10.178514936439111_dp, &
             ! w
             3.231366266745103e-09_dp, 2.1230653971588304e-08_dp, 8.691057314412012e-08_dp, &
             2.8252960348951036e-07_dp, 7.953949280707434e-07_dp, 2.024324824377468e-06_dp, &
             4.7755223828501555e-06_dp, 1.0611986881536505e-05_dp, 2.2460705609533583e-05_dp, &
             4.564425558157518e-05_dp, 8.959919860757453e-05_dp, 0.0001706889050477756_dp, &
             0.0003167364258501587_dp, 0.000574226969130065_dp, 0.0010196066020646871_dp, &
             0.0017768005530869207_dp, 0.0030440841831127097_dp, 0.005134892358176567_dp, &
             0.0085392292962598_dp, 0.014015291234860975_dp, 0.022725087209662195_dp, &
             0.03643373542318754_dp, 0.057800474416534686_dp, 0.0908015476745432_dp, &
             0.1413436944826071_dp, 0.2181589446681848_dp, 0.334137330397612_dp, &
             0.5084198143667269_dp, 0.7700615767490158_dp, 1.1656996689709882_dp, &
             1.7799147858399724_dp, 2.808270824290397_dp, 4.9959284427881485_dp]

      CASE (1083)

         aw(:) = & ! a
            [2.2274857143791788e-09_dp, 1.7452329222319866e-08_dp, 7.536813911823567e-08_dp, &
             2.5740414429863387e-07_dp, 7.59327680166685e-07_dp, 2.0180698598548547e-06_dp, &
             4.954642624566124e-06_dp, 1.1423118799152858e-05_dp, 2.5016452846667024e-05_dp, &
             5.2476709506426854e-05_dp, 0.00010610938292451702_dp, 0.00020783739218017472_dp, &
             0.0003958918029415031_dp, 0.0007356841401316799_dp, 0.001337227084379473_dp, &
             0.002382692155023381_dp, 0.004169493694660976_dp, 0.007176921221194825_dp, &
             0.012168162938270537_dp, 0.020345069658743716_dp, 0.03358089387804767_dp, &
             0.05476746737966397_dp, 0.08832919918046878_dp, 0.14097880029878562_dp, &
             0.2228216410192318_dp, 0.3489618913715647_dp, 0.5418335052445565_dp, &
             0.8345952495177218_dp, 1.2761543471145376_dp, 1.9389121348256646_dp, &
             2.9318081385079755_dp, 4.426139751694001_dp, 6.721718216312974_dp, &
             10.50589803308251_dp, &
             ! w
             6.287579033693602e-09_dp, 2.866027383361553e-08_dp, 9.976565058340684e-08_dp, &
             2.9493834459446365e-07_dp, 7.778875343938408e-07_dp, 1.8859609309719048e-06_dp, &
             4.283334530991183e-06_dp, 9.229859066286541e-06_dp, 1.904171264820599e-05_dp, &
             3.786445165425619e-05_dp, 7.294677515836348e-05_dp, 0.00013670536272714038_dp, &
             0.000250023354478363_dp, 0.00044744998077084636_dp, 0.0007852983999726469_dp, &
             0.0013541178012168918_dp, 0.0022977068553191042_dp, 0.00384182561573767_dp, &
             0.006337172298157243_dp, 0.01032318575822768_dp, 0.016622042719138596_dp, &
             0.02647615953244465_dp, 0.04174803928941701_dp, 0.0652091373356819_dp, &
             0.10095582752341163_dp, 0.155008329527622_dp, 0.23617994440111334_dp, &
             0.3573708113077927_dp, 0.5376108754377592_dp, 0.8056773157854522_dp, &
             1.207761153615302_dp, 1.8279594993512716_dp, 2.861798789933721_dp, &
             5.057156122401313_dp]

      CASE (1084)

         aw(:) = & ! a
            [2.0172967327387257e-09_dp, 1.614584278026562e-08_dp, 7.091977385242614e-08_dp, &
             2.4493553200837293e-07_dp, 7.281868505114004e-07_dp, 1.946502272582263e-06_dp, &
             4.800437717927625e-06_dp, 1.1107729506508856e-05_dp, 2.4398784467258136e-05_dp, &
             5.1310922225251035e-05_dp, 0.00010397846799368052_dp, 0.0002040507513929573_dp, &
             0.0003893303445185409_dp, 0.0007245702524773872_dp, 0.0013187890502140368_dp, &
             0.0023526830218396724_dp, 0.004121513439091418_dp, 0.0071014787708213786_dp, &
             0.012051402514466228_dp, 0.020167076430592148_dp, 0.033313491362902015_dp, &
             0.05437143049614822_dp, 0.08775084355458419_dp, 0.14014598703409886_dp, &
             0.22163937938586792_dp, 0.34730795733959097_dp, 0.5395548448635472_dp, &
             0.8315063932629486_dp, 1.2720395180689776_dp, 1.9335328679491852_dp, &
             2.9249173396389962_dp, 4.417499937632058_dp, 6.711113328194943_dp, &
             10.49309693909053_dp, &
             ! w
             5.7312966788233454e-09_dp, 2.6845770352473356e-08_dp, 9.484372746430734e-08_dp, &
             2.8289630426777587e-07_dp, 7.507997105105137e-07_dp, 1.8288916424757699e-06_dp, &
             4.16924093818577e-06_dp, 9.011419652462668e-06_dp, 1.863853183129267e-05_dp, &
             3.7143455330995374e-05_dp, 7.169277810745862e-05_dp, 0.0001345777259838475_dp, &
             0.0002464933392115592_dp, 0.0004417118522174695_dp, 0.0007761456117064183_dp, &
             0.0013397738522708156_dp, 0.0022755989551174164_dp, 0.003808288677964405_dp, &
             0.0062870719408963_dp, 0.01024945338475834_dp, 0.01651512565938768_dp, &
             0.026323408045312316_dp, 0.041533084788832456_dp, 0.0649113684912127_dp, &
             0.10055014878836191_dp, 0.15446549283188105_dp, 0.23546786482502932_dp, &
             0.3564573544201329_dp, 0.5364684913069616_dp, 0.8042892994278837_dp, &
             1.2061277693188093_dp, 1.8260989681173887_dp, 2.8597297398894663_dp, &
             5.054791277265716_dp]

      CASE (1085)

         aw(:) = & ! a
            [1.850740981998257e-09_dp, 1.5104833047859603e-08_dp, 6.734253109129304e-08_dp, &
             2.34833911324831e-07_dp, 7.028186850557079e-07_dp, 1.8879565627881793e-06_dp, &
             4.673871366494548e-06_dp, 1.0848165308288041e-05_dp, 2.3889290043365354e-05_dp, &
             5.034743644885774e-05_dp, 0.00010221437304761533_dp, 0.00020091132242531327_dp, &
             0.00038388324266710403_dp, 0.0007153330520107875_dp, 0.0013034482225701264_dp, &
             0.00232769075061419_dp, 0.004081519203355378_dp, 0.007038542259076679_dp, &
             0.011953924173265984_dp, 0.020018373854763606_dp, 0.03308994743271452_dp, &
             0.05404014862204927_dp, 0.08726677595442342_dp, 0.13944856767678987_dp, &
             0.2206488134502364_dp, 0.3459215170569635_dp, 0.5376438169727374_dp, &
             0.8289147083512008_dp, 1.2685854776524752_dp, 1.9290155002563718_dp, &
             2.9191281683182515_dp, 4.410238259336714_dp, 6.702196038231376_dp, &
             10.482327477721661_dp, &
             ! w
             5.290271368600501e-09_dp, 2.5390682297024868e-08_dp, 9.085501648061086e-08_dp, &
             2.730761637996616e-07_dp, 7.286126873507325e-07_dp, 1.7819896586576813e-06_dp, &
             4.075219300882112e-06_dp, 8.831003852374763e-06_dp, 1.8304896314491113e-05_dp, &
             3.654584137471835e-05_dp, 7.065187625415539e-05_dp, 0.00013280939129141288_dp, &
             0.00024355610352215322_dp, 0.00043693238706051744_dp, 0.0007685148366324585_dp, &
             0.0013278049087297543_dp, 0.002257137023290618_dp, 0.003780262078984825_dp, &
             0.006245174895475462_dp, 0.010187754457293476_dp, 0.016425604231342008_dp, &
             0.02619543691888099_dp, 0.04135290427531485_dp, 0.06466164128832476_dp, &
             0.10020975076138158_dp, 0.15400978423284883_dp, 0.2348697877185054_dp, &
             0.3556897674585268_dp, 0.5355080591362693_dp, 0.8031217534772147_dp, &
             1.2047530557904302_dp, 1.824532061180195_dp, 2.8579857864634843_dp, &
             5.052795637732665_dp]

      CASE (1086)

         aw(:) = & ! a
            [1.7153091124854105e-09_dp, 1.425409001814337e-08_dp, 6.43951904177628e-08_dp, &
             2.26457302430374e-07_dp, 6.816830018404275e-07_dp, 1.839003514606777e-06_dp, &
             4.5677424478737965e-06_dp, 1.0630010929910116e-05_dp, 2.3460250638202357e-05_dp, &
             4.9534760534089635e-05_dp, 0.0001007242847591298_dp, 0.00019825621447659207_dp, &
             0.0003792713775865615_dp, 0.0007075045239472364_dp, 0.0012904353066398916_dp, &
             0.0023064738671743855_dp, 0.004047541623344702_dp, 0.0069850377660518395_dp, &
             0.011871003073216316_dp, 0.019891805728292994_dp, 0.03289957696430025_dp, &
             0.053757888818087775_dp, 0.08685414910409205_dp, 0.13885382054779094_dp, &
             0.21980373605319486_dp, 0.3447382639025349_dp, 0.536012282976087_dp, &
             0.8267013421055617_dp, 1.2656347466472995_dp, 1.9251553237466008_dp, &
             2.91417999676147_dp, 4.404030177913078_dp, 6.694571226204766_dp, &
             10.473117699040598_dp, &
             ! w
             4.9315025944171035e-09_dp, 2.4194703352116735e-08_dp, 8.754662294957071e-08_dp, &
             2.6488715039529833e-07_dp, 7.100413220259457e-07_dp, 1.7426177386695847e-06_dp, &
             3.9961099923564505e-06_dp, 8.678912283447885e-06_dp, 1.802318404797261e-05_dp, &
             3.6040529796384376e-05_dp, 6.977067222889229e-05_dp, 0.00013131074992817489_dp, &
             0.00024106444632794165_dp, 0.0004328744687631082_dp, 0.000762031002841193_dp, &
             0.001317627711659764_dp, 0.0022414285890036724_dp, 0.0037564011550864466_dp, &
             0.006209485282369086_dp, 0.010135169635674476_dp, 0.01634927008738069_dp, &
             0.026086267938169604_dp, 0.04119913191272156_dp, 0.06444843086150295_dp, &
             0.09991901963568436_dp, 0.15362043082797508_dp, 0.23435862795266732_dp, &
             0.35503353191620346_dp, 0.5346867248054958_dp, 0.8021230559895891_dp, &
             1.2035769133212262_dp, 1.8231912933627146_dp, 2.8564934198830527_dp, &
             5.051087937156654_dp]

      CASE (1087)

         aw(:) = & ! a
            [1.0758659340255967e-09_dp, 1.0167164015240067e-08_dp, 4.9859592286185096e-08_dp, &
             1.8431400593006791e-07_dp, 5.738015222033274e-07_dp, 1.5863937162236038e-06_dp, &
             4.015360098808803e-06_dp, 9.486576708604261e-06_dp, 2.119830152313252e-05_dp, &
             4.5228863817178745e-05_dp, 9.279513732526955e-05_dp, 0.00018407435527092067_dp, &
             0.00035455547351989603_dp, 0.0006654246358606789_dp, 0.0012203003184695523_dp, &
             0.0021918438571634814_dp, 0.003863560773416426_dp, 0.006694733518895417_dp, &
             0.01142024757192155_dp, 0.01920259575640197_dp, 0.031861275677595516_dp, &
             0.05221611486432177_dp, 0.0845971409884596_dp, 0.13559641313576798_dp, &
             0.2151696546607952_dp, 0.3382423823171786_dp, 0.5270459012716615_dp, &
             0.8145253874758794_dp, 1.2493876688137333_dp, 1.9038830603298444_dp, &
             2.886891897531776_dp, 4.3697720246036935_dp, 6.652472907190046_dp, &
             10.422247573899798_dp, &
             ! w
             3.235159378159741e-09_dp, 1.8339242831941008e-08_dp, 7.088982438417943e-08_dp, &
             2.2297988277124975e-07_dp, 6.13909232292796e-07_dp, 1.5370220364486574e-06_dp, &
             3.580100476231876e-06_dp, 7.87446076916954e-06_dp, 1.6525813148385928e-05_dp, &
             3.334332260153284e-05_dp, 6.504971734511139e-05_dp, 0.0001232558121611228_dp, &
             0.00022763336577209098_dp, 0.00041094352775341956_dp, 0.0007269064798609755_dp, &
             0.0012623766488428546_dp, 0.0021559807094720745_dp, 0.0036263710068930814_dp, &
             0.006014667749255997_dp, 0.009847678812396432_dp, 0.01593133030777213_dp, &
             0.025487740131110808_dp, 0.04035498769146014_dp, 0.06327659551707043_dp, &
             0.09831931929717358_dp, 0.15147580801769658_dp, 0.23154027931032156_dp, &
             0.3514119476229571_dp, 0.5301501843862094_dp, 0.7966027741178908_dp, &
             1.1970718345468874_dp, 1.815772457762836_dp, 2.8482340034483533_dp, &
             5.0416374859708_dp]

      CASE (1088)

         aw(:) = & ! a
            [8.530766154470298e-10_dp, 8.70284226105799e-09_dp, 4.4451066257575195e-08_dp, &
             1.6820644584206475e-07_dp, 5.317809778430233e-07_dp, 1.48660479151414e-06_dp, &
             3.7947395852828838e-06_dp, 9.025815552131148e-06_dp, 2.0280074380309348e-05_dp, &
             4.346995032147937e-05_dp, 8.953868442507957e-05_dp, 0.00017822251684367218_dp, &
             0.0003443145843865128_dp, 0.0006479244821443401_dp, 0.0011910355495978132_dp, &
             0.0021438689242639325_dp, 0.003786350128625508_dp, 0.00657259701154879_dp, &
             0.011230169121998363_dp, 0.018911345617281236_dp, 0.03142164025632888_dp, &
             0.05156210536058579_dp, 0.08363810242925038_dp, 0.13421008768472792_dp, &
             0.21319449779576682_dp, 0.33546983718241036_dp, 0.523213948281257_dp, &
             0.8093154919519803_dp, 1.242428071386227_dp, 1.894761638454777_dp, &
             2.875180343614518_dp, 4.355057581604841_dp, 6.634379367985747_dp, &
             10.400373073436226_dp, &
             ! w
             2.6427031908949146e-09_dp, 1.6181094404493005e-08_dp, 6.451780298222103e-08_dp, &
             2.0660472994990585e-07_dp, 5.757877767181744e-07_dp, 1.4545748849486805e-06_dp, &
             3.4117817238528944e-06_dp, 7.546587060800534e-06_dp, 1.591175588447046e-05_dp, &
             3.223137637007654e-05_dp, 6.30945186273881e-05_dp, 0.00011990633629638359_dp, &
             0.00022202824003445852_dp, 0.0004017616623647714_dp, 0.0007121579415223659_dp, &
             0.0012391155524192623_dp, 0.002119919181919329_dp, 0.0035713719081832155_dp, &
             0.005932095426075794_dp, 0.009725594268101923_dp, 0.015753533712217164_dp, &
             0.025232695523749075_dp, 0.03999472044455227_dp, 0.06277574334983835_dp, &
             0.09763465396721004_dp, 0.15055673187308719_dp, 0.2303310159894199_dp, &
             0.34985629290085185_dp, 0.5281995078641347_dp, 0.7942269487213767_dp, &
             1.1942700872213858_dp, 1.8125754547977413_dp, 2.8446738520936137_dp, &
             5.037564323693374_dp]

      CASE (1089)

         aw(:) = & ! a
            [7.485517113163199e-10_dp, 8.004464374515241e-09_dp, 4.182104218009189e-08_dp, &
             1.6027133146953048e-07_dp, 5.108935884752593e-07_dp, 1.4366722946534533e-06_dp, &
             3.683775648135436e-06_dp, 8.793107317512489e-06_dp, 1.981472855754801e-05_dp, &
             4.257596613387601e-05_dp, 8.787942807692114e-05_dp, 0.00017523434466571028_dp, &
             0.00033907515416720774_dp, 0.0006389557687380277_dp, 0.0011760145270830638_dp, &
             0.0021192101718956412_dp, 0.003746614335982069_dp, 0.006509667929861093_dp, &
             0.011132130017987636_dp, 0.018760976882557942_dp, 0.031194456333704244_dp, &
             0.05122385794200837_dp, 0.08314170804254446_dp, 0.1334920064851158_dp, &
             0.21217071784296496_dp, 0.3340318306728112_dp, 0.5212252858817166_dp, &
             0.8066102233296886_dp, 1.2388124123187383_dp, 1.8900206583520773_dp, &
             2.8690905795689696_dp, 4.347403635360103_dp, 6.624964954718267_dp, &
             10.388988746922239_dp, &
             ! w
             2.364271870423896e-09_dp, 1.5136190827701523e-08_dp, 6.137755396060629e-08_dp, &
             1.9845357947137618e-07_dp, 5.566795770926899e-07_dp, 1.4130310746080372e-06_dp, &
             3.326614800119239e-06_dp, 7.380121200576499e-06_dp, 1.5599097942649512e-05_dp, &
             3.1663823009091906e-05_dp, 6.209443482620178e-05_dp, 0.00011818987687018821_dp, &
             0.00021915108769775198_dp, 0.00039704152484541045_dp, 0.0007045659357474019_dp, &
             0.001227126947942397_dp, 0.002101312533897999_dp, 0.0035429648673508704_dp, &
             0.005889406304828762_dp, 0.00966242210021254_dp, 0.01566145812776196_dp, &
             0.025100514361634532_dp, 0.03980787201150227_dp, 0.06251580736786437_dp, &
             0.09727909653041707_dp, 0.1500791576366265_dp, 0.22970230377822035_dp, &
             0.34904706896485305_dp, 0.5271843216790469_dp, 0.7929899889654172_dp, &
             1.1928108742711225_dp, 1.8109099762845895_dp, 2.8428189713625165_dp, &
             5.035442247417427_dp]

      CASE (1090)

         aw(:) = & ! a
            [7.015759952264556e-10_dp, 7.687605461805616e-09_dp, 4.06153328751681e-08_dp, &
             1.5660890498596987e-07_dp, 5.012083723973226e-07_dp, 1.4134403460572587e-06_dp, &
             3.6320117485286108e-06_dp, 8.684320848399275e-06_dp, 1.9596808212053662e-05_dp, &
             4.215669691359067e-05_dp, 8.710026616696479e-05_dp, 0.0001738295916548252_dp, &
             0.00033660967382591756_dp, 0.000634731765933465_dp, 0.0011689345559288055_dp, &
             0.0021075793996146617_dp, 0.0037278601949364072_dp, 0.006479949846403211_dp, &
             0.01108580642816379_dp, 0.01868989224086861_dp, 0.031087009113049235_dp, &
             0.05106381475153055_dp, 0.08290674373042546_dp, 0.1331519826617938_dp, &
             0.21168577277828624_dp, 0.3333504543600378_dp, 0.5202827061155677_dp, &
             0.8053276295681061_dp, 1.2370977503636218_dp, 1.8877718031140505_dp, &
             2.866201329763345_dp, 4.34377161068584_dp, 6.620496869060699_dp, &
             10.383585113156837_dp, &
             ! w
             2.238997466106155e-09_dp, 1.4658206914195193e-08_dp, 5.992792460965658e-08_dp, &
             1.9467135835313114e-07_dp, 5.477814545209847e-07_dp, 1.3936332861904557e-06_dp, &
             3.286763847132626e-06_dp, 7.302093832260751e-06_dp, 1.5452332489261882e-05_dp, &
             3.139707453996938e-05_dp, 6.162389023876517e-05_dp, 0.00011738150729535318_dp, &
             0.00021779494577127908_dp, 0.0003948150140722096_dp, 0.0007009823090453667_dp, &
             0.0012214645018813195_dp, 0.002092519275230718_dp, 0.003529533071018548_dp, &
             0.005869211749643338_dp, 0.009632524485049173_dp, 0.015617863248112181_dp, &
             0.025037906506793727_dp, 0.03971933879556791_dp, 0.06239260161295121_dp, &
             0.09711051373150156_dp, 0.14985265414969443_dp, 0.22940403474769006_dp, &
             0.34866306244184087_dp, 0.52670246342252_dp, 0.7924027422427473_dp, &
             1.1921179927454306_dp, 1.8101190551276467_dp, 2.8419380516825403_dp, &
             5.034434452504777_dp]

      CASE (1091)

         aw(:) = & ! a
            [6.943813104846991e-10_dp, 7.638891284727516e-09_dp, 4.0429229618250355e-08_dp, &
             1.5604216589299267e-07_dp, 4.99707050869597e-07_dp, 1.4098345541979897e-06_dp, &
             3.623969652411993e-06_dp, 8.667406245892608e-06_dp, 1.9562902790099188e-05_dp, &
             4.209142824273307e-05_dp, 8.697891449886091e-05_dp, 0.00017361071664949315_dp, &
             0.0003362253890632743_dp, 0.0006340731772189557_dp, 0.0011678303568035191_dp, &
             0.0021057649743316032_dp, 0.003724933805080419_dp, 0.00647531162120873_dp, &
             0.011078575029474128_dp, 0.01867879337102774_dp, 0.031070229762573637_dp, &
             0.051038817718966456_dp, 0.08287003922973965_dp, 0.13309885893502246_dp, &
             0.21160999712408324_dp, 0.33324397176230375_dp, 0.5201353864290272_dp, &
             0.805127145634904_dp, 1.2368297018297942_dp, 1.8874202127485686_dp, &
             2.865749581935815_dp, 4.343203687458757_dp, 6.619798175972809_dp, &
             10.382740088943475_dp, &
             ! w
             2.219801091458909e-09_dp, 1.4584488395432053e-08_dp, 5.970358786870748e-08_dp, &
             1.940849185829006e-07_dp, 5.463999525709266e-07_dp, 1.3906186037408768e-06_dp, &
             3.2805654981323565e-06_dp, 7.289949663565937e-06_dp, 1.5429477471535556e-05_dp, &
             3.135551577137874e-05_dp, 6.15505511414413e-05_dp, 0.0001172554713986625_dp, &
             0.00021758344005324697_dp, 0.0003944676661812053_dp, 0.0007004230977715815_dp, &
             0.0012205806915926256_dp, 0.0020911465046260947_dp, 0.0035274357281613535_dp, &
             0.005866057824506221_dp, 0.00962785434967134_dp, 0.015611052450282553_dp, &
             0.0250281238814589_dp, 0.0397055033524259_dp, 0.06237334520989919_dp, &
             0.09708416182808391_dp, 0.1498172442136133_dp, 0.22935740028956064_dp, &
             0.3486030165635429_dp, 0.5266271094625212_dp, 0.7923109000349835_dp, &
             1.1920096234079312_dp, 1.8099953470798475_dp, 2.8418002644657556_dp, &
             5.034276822946402_dp]

      CASE (1092)

         aw(:) = & ! a
            [2.0404909110384475e-09_dp, 1.5151442097142887e-08_dp, 6.245416599662895e-08_dp, &
             2.0661940652827375e-07_dp, 5.959822274864856e-07_dp, 1.557575484071925e-06_dp, &
             3.7741601101992155e-06_dp, 8.60944555155634e-06_dp, 1.868879219385792e-05_dp, &
             3.8910961485681584e-05_dp, 7.81734938233369e-05_dp, 0.00015225985758064739_dp, &
             0.00028858964675804725_dp, 0.0005339159787935496_dp, 0.0009666276456836273_dp, &
             0.0017161644907638827_dp, 0.002993306589107769_dp, 0.005136925782329796_dp, &
             0.008685437210940068_dp, 0.014485001245983453_dp, 0.023851978787347662_dp, &
             0.03881489489843682_dp, 0.06247213087021271_dp, 0.09951701756574863_dp, &
             0.15700376412760972_dp, 0.2454584857259318_dp, 0.3804841923044579_dp, &
             0.5850764943121439_dp, 0.8929811589653015_dp, 1.3536502872809886_dp, &
             2.039888878054595_dp, 3.06078528102233_dp, 4.587461301561305_dp, &
             6.919354993331615_dp, 10.744151118313793_dp, &
             ! w
             5.671009581738294e-09_dp, 2.4055717463455767e-08_dp, 8.027675008113348e-08_dp, &
             2.3130933927897619e-07_dp, 5.99118177555492e-07_dp, 1.4326872059077528e-06_dp, &
             3.218491669573828e-06_dp, 6.873427000836394e-06_dp, 1.4074003413378896e-05_dp, &
             2.7806758956751627e-05_dp, 5.3272221379320055e-05_dp, 9.934598752612586e-05_dp, &
             0.00018090655818156425_dp, 0.00032249604517037945_dp, 0.0005640087632846268_dp, &
             0.0009694349082615356_dp, 0.0016401651235089072_dp, 0.0027350465332988154_dp, &
             0.004500354345088165_dp, 0.007314218975682474_dp, 0.011751982671632126_dp, &
             0.01868166624087445_dp, 0.029402505102421705_dp, 0.045844798655652684_dp, &
             0.07085679716737465_dp, 0.10861530420845496_dp, 0.1652139785801667_dp, &
             0.24951462076740005_dp, 0.3744144724200437_dp, 0.5588565709507807_dp, &
             0.8314194980761831_dp, 1.2379892279383173_dp, 1.8623478799739805_dp, &
             2.900038675886075_dp, 5.100932905035487_dp]

      CASE (1093)

         aw(:) = & ! a
            [1.8414082779546135e-09_dp, 1.3931677417941283e-08_dp, 5.8405519526959223e-08_dp, &
             1.955092377265184e-07_dp, 5.686647765262082e-07_dp, 1.4955263063950401e-06_dp, &
             3.641667355155677e-06_dp, 8.340388270437216e-06_dp, 1.8164859681222616e-05_dp, &
             3.792663009296004e-05_dp, 7.638092416773742e-05_dp, 0.0001490839167418985_dp, &
             0.0002830992770653307_dp, 0.0005246328918177588_dp, 0.0009512465838597859_dp, &
             0.0016911508892508056_dp, 0.0029533273516591818_dp, 0.00507405727687672_dp, &
             0.008588082206283019_dp, 0.014336434862997982_dp, 0.023628435431101668_dp, &
             0.038483108388421575_dp, 0.0619862563727028_dp, 0.09881490841524607_dp, &
             0.1560026667117371_dp, 0.24405037483113298_dp, 0.3785312246764298_dp, &
             0.5824074291231462_dp, 0.8893900612672502_dp, 1.3488992685003742_dp, &
             2.033716683950777_dp, 3.052922474979899_dp, 4.57764928621845_dp, &
             6.907357064194191_dp, 10.729709098125946_dp, &
             ! w
             5.14483019500244e-09_dp, 2.23907492239133e-08_dp, 7.589542727641687e-08_dp, &
             2.2078098212940023e-07_dp, 5.75722876573559e-07_dp, 1.3838421833046814e-06_dp, &
             3.1215204051033388e-06_dp, 6.688786577258879e-06_dp, 1.3734685587828229e-05_dp, &
             2.7202048261579674e-05_dp, 5.222328846046373e-05_dp, 9.756988591002205e-05_dp, &
             0.00017796405082823_dp, 0.00031771730894266043_dp, 0.0005563894187961383_dp, &
             0.0009574931010947457_dp, 0.0016217487250190772_dp, 0.0027070782912054733_dp, &
             0.00445850262869856_dp, 0.007252483630532394_dp, 0.011662191041651403_dp, &
             0.018552886077995906_dp, 0.029220402385046033_dp, 0.04559100416651338_dp, &
             0.0705084021081063_dp, 0.10814468767541524_dp, 0.16458927850486596_dp, &
             0.24870127649892812_dp, 0.373378326575567_dp, 0.5575688629611708_dp, &
             0.8298634094680106_dp, 1.2361659821700268_dp, 1.8602769703068671_dp, &
             2.8977376266341373_dp, 5.098298094971871_dp]

      CASE (1094)

         aw(:) = & ! a
            [1.6835931905098269e-09_dp, 1.2959813708235182e-08_dp, 5.515027188658148e-08_dp, &
             1.8650630747666386e-07_dp, 5.463970782554214e-07_dp, 1.4447154912042303e-06_dp, &
             3.532775386726778e-06_dp, 8.118594924447018e-06_dp, 1.7731876517863316e-05_dp, &
             3.7111417549382796e-05_dp, 7.489356395869152e-05_dp, 0.0001464443973232522_dp, &
             0.0002785295904198779_dp, 0.0005168964109596914_dp, 0.0009384129952013451_dp, &
             0.001670257841492673_dp, 0.002919901415196955_dp, 0.0050214469659123915_dp, &
             0.008506545297089656_dp, 0.014211912321501622_dp, 0.023440937304335538_dp, &
             0.038204637084348285_dp, 0.06157820642378461_dp, 0.09822491920184509_dp, &
             0.15516098055593858_dp, 0.24286588733419276_dp, 0.37688763013216325_dp, &
             0.5801601779368206_dp, 0.8863652427608161_dp, 1.3448959054597227_dp, &
             2.0285139770113605_dp, 3.0462926407885687_dp, 4.569373671372515_dp, &
             6.897235563583895_dp, 10.71752363827283_dp, &
             ! w
             4.727489893110114e-09_dp, 2.1056037768907683e-08_dp, 7.234400997472033e-08_dp, &
             2.1218843662219005e-07_dp, 5.565374096637542e-07_dp, 1.34363808524887e-06_dp, &
             3.0414648634102525e-06_dp, 6.535975352097481e-06_dp, 1.3453266870529135e-05_dp, &
             2.6699605298279875e-05_dp, 5.1350354896232e-05_dp, 9.608970013646915e-05_dp, &
             0.00017550868949081015_dp, 0.00031372516835964886_dp, 0.0005500176586624825_dp, &
             0.0009474972034126543_dp, 0.001606319906477992_dp, 0.0026836284428096327_dp, &
             0.004423386277139935_dp, 0.007200647969842769_dp, 0.011586749733714279_dp, &
             0.01844462234217734_dp, 0.02906722486014722_dp, 0.04537740928792174_dp, &
             0.07021504448069832_dp, 0.10774822984772503_dp, 0.1640627838290144_dp, &
             0.2480155078759422_dp, 0.3725043652642452_dp, 0.5564823319726203_dp, &
             0.8285500186244518_dp, 1.234626709626042_dp, 1.858528291438609_dp, &
             2.895794450094336_dp, 5.096073142430604_dp]

      CASE (1095)

         aw(:) = & ! a
            [1.5552015829991145e-09_dp, 1.2165458305132021e-08_dp, 5.246790769760238e-08_dp, &
             1.7903697443025552e-07_dp, 5.278274428525224e-07_dp, 1.4021753681530081e-06_dp, &
             3.4413213768023774e-06_dp, 7.931839259075196e-06_dp, 1.7366505278388265e-05_dp, &
             3.642223298591174e-05_dp, 7.36341292184873e-05_dp, 0.00014420622070141634_dp, &
             0.000274649900170601_dp, 0.0005103207645559094_dp, 0.0009274940676298663_dp, &
             0.0016524656159344706_dp, 0.0028914126171122705_dp, 0.004976573092407527_dp, &
             0.008436949507025988_dp, 0.014105556864197068_dp, 0.023280697102297358_dp, &
             0.03796651503914784_dp, 0.06122909764046382_dp, 0.09771990332921691_dp, &
             0.15444018660992323_dp, 0.24185109094921042_dp, 0.37547892904451774_dp, &
             0.5782333621224738_dp, 0.8837708177359374_dp, 1.341461050138272_dp, &
             2.024048765020729_dp, 3.040601087855359_dp, 4.562267623789521_dp, &
             6.888542855611844_dp, 10.707056779390536_dp, &
             ! w
             4.387796674462411e-09_dp, 1.9959072281083754e-08_dp, 6.939677013157643e-08_dp, &
             2.0501552132463639e-07_dp, 5.404552447976642e-07_dp, 1.3098295095475104e-06_dp, &
             2.973970850137846e-06_dp, 6.406865977391905e-06_dp, 1.3215066049896975e-05_dp, &
             2.627365716243625e-05_dp, 5.06093080325915e-05_dp, 9.4831624379329e-05_dp, &
             0.00017341950865974504_dp, 0.00031032508213488913_dp, 0.000544586059463598_dp, &
             0.0009389693317150078_dp, 0.0015931472595080536_dp, 0.002663594024486951_dp, &
             0.004393365605889235_dp, 0.007156308081079714_dp, 0.01152218238536826_dp, &
             0.018351916051089393_dp, 0.028935995824406258_dp, 0.045194337267774165_dp, &
             0.06996350079953981_dp, 0.10740814493439677_dp, 0.1636109813522357_dp, &
             0.2474268182319067_dp, 0.37175387620169315_dp, 0.5555490237003874_dp, &
             0.827421543263676_dp, 1.2333038674610606_dp, 1.857025254461482_dp, &
             2.8941241139784153_dp, 5.094160649830687_dp]

      CASE (1096)

         aw(:) = & ! a
            [9.459673535851904e-10_dp, 8.334293569930787e-09_dp, 3.917884706182252e-08_dp, &
             1.4122296908165727e-07_dp, 4.322980090687376e-07_dp, 1.1806338342062016e-06_dp, &
             2.960390093021729e-06_dp, 6.941898331244286e-06_dp, 1.5416820042415313e-05_dp, &
             3.272363662119428e-05_dp, 6.684179387016864e-05_dp, 0.00013208300653220852_dp, &
             0.00025355449893918686_dp, 0.0004744433496388714_dp, 0.0008677344348137625_dp, &
             0.0015548140641988004_dp, 0.0027346518092018373_dp, 0.004729070704264146_dp, &
             0.008052259415847064_dp, 0.013516497030495265_dp, 0.02239153661791747_dp, &
             0.036642900701831065_dp, 0.059285413846926246_dp, 0.09490393769073507_dp, &
             0.15041534012062396_dp, 0.23617702410022706_dp, 0.36759258852036963_dp, &
             0.5674338722755372_dp, 0.8692136769184071_dp, 1.3221689353918424_dp, &
             1.9989465693744168_dp, 3.0085785280197155_dp, 4.522258339234745_dp, &
             6.839571562391387_dp, 10.648063728448992_dp, &
             ! w
             2.7735062176295402e-09_dp, 1.4568310665755517e-08_dp, 5.446397060387528e-08_dp, &
             1.68002423961422e-07_dp, 4.5639610286526923e-07_dp, 1.1313606164769208e-06_dp, &
             2.614833864841551e-06_dp, 5.715318145586774e-06_dp, 1.1932023457972336e-05_dp, &
             2.3968219348236865e-05_dp, 4.658141851780609e-05_dp, 8.796787071664377e-05_dp, &
             0.00016198334441270982_dp, 0.000291657010094577_dp, 0.0005146826469277505_dp, &
             0.0008919026609983487_dp, 0.0015202793220082452_dp, 0.002552535131022295_dp, &
             0.004226624603766521_dp, 0.006909589259674551_dp, 0.011162306031160234_dp, &
             0.017834386703605473_dp, 0.028202329224935544_dp, 0.04416940427443703_dp, &
             0.06855337933855238_dp, 0.10549931378412311_dp, 0.16107214464250502_dp, &
             0.24411514375427973_dp, 0.36752769984961553_dp, 0.5502884646716867_dp, &
             0.8210557274560507_dp, 1.2258366023930047_dp, 1.8485367622980344_dp, &
             2.8846886380833974_dp, 5.083358231256422_dp]

      CASE (1097)

         aw(:) = & ! a
            [7.273102982139781e-10_dp, 6.92144323944422e-09_dp, 3.407649049671692e-08_dp, &
             1.2625900830114456e-07_dp, 3.936654500762542e-07_dp, 1.0895684649343947e-06_dp, &
             2.7601558219033445e-06_dp, 6.52543820354918e-06_dp, 1.4589484764414247e-05_dp, &
             3.114260304069562e-05_dp, 6.391983656376198e-05_dp, 0.0001268388230454835_dp, &
             0.0002443843898432227_dp, 0.000458779248541801_dp, 0.0008415406084465976_dp, &
             0.0015118588711423079_dp, 0.00266547158729582_dp, 0.004619520499331146_dp, &
             0.007881521101534278_dp, 0.013254392156440174_dp, 0.02199497422207769_dp, &
             0.03605128830337178_dp, 0.0584148875534893_dp, 0.09364034750240999_dp, &
             0.14860609252418394_dp, 0.23362218152255373_dp, 0.3640361075094539_dp, &
             0.5625565704529129_dp, 0.8626304237072151_dp, 1.3134334337127707_dp, &
             1.9875672719429849_dp, 2.9940473014673312_dp, 4.504086900619768_dp, &
             6.8173135920763315_dp, 10.621235577151532_dp, &
             ! w
             2.192793031414204e-09_dp, 1.2521328564769824e-08_dp, 4.8548435607827576e-08_dp, &
             1.5297649793173104e-07_dp, 4.216837431485938e-07_dp, 1.0566964068481496e-06_dp, &
             2.4630156199409405e-06_dp, 5.420465198614438e-06_dp, 1.1381011734306003e-05_dp, &
             2.2971974106109578e-05_dp, 4.483142438962657e-05_dp, 8.49715438163735e-05_dp, &
             0.000156969743231988_dp, 0.0002834417190395827_dp, 0.0005014775802709292_dp, &
             0.0008710531515594637_dp, 0.0014879075446847592_dp, 0.002503066290644606_dp, &
             0.004152171578196498_dp, 0.006799174417995733_dp, 0.011000908768639673_dp, &
             0.017601826515960557_dp, 0.02787203366917738_dp, 0.0437071783022296_dp, &
             0.0679163991946226_dp, 0.10463572640173835_dp, 0.15992186392307833_dp, &
             0.24261266592466305_dp, 0.3656078912739175_dp, 0.547896008140713_dp, &
             0.818157663971773_dp, 1.2224342607686673_dp, 1.8446668174442742_dp, &
             2.8803857371110673_dp, 5.078432499169791_dp]

      CASE (1098)

         aw(:) = & ! a
            [6.155715975531782e-10_dp, 6.18671200607816e-09_dp, 3.1361740021939036e-08_dp, &
             1.1816893758191133e-07_dp, 3.7254358065230806e-07_dp, 1.0393624829688067e-06_dp, &
             2.6490425728200133e-06_dp, 6.2931198677288194e-06_dp, 1.4125946250596634e-05_dp, &
             3.0253504183507454e-05_dp, 6.227142811943758e-05_dp, 0.00012387211147251288_dp, &
             0.00023918398478598303_dp, 0.0004498766336403791_dp, 0.0008266242024370201_dp, &
             0.0014873539703357177_dp, 0.0026259420175185058_dp, 0.004556830740916472_dp, &
             0.007783683492592172_dp, 0.013104010119872675_dp, 0.021767182298726106_dp, &
             0.035711088299006565_dp, 0.05791379584797851_dp, 0.09291231473586366_dp, &
             0.14756275297218502_dp, 0.23214766219496497_dp, 0.36198190772071465_dp, &
             0.5597374357109449_dp, 0.8588226670364908_dp, 1.3083776705465153_dp, &
             1.9809776460701942_dp, 2.985628165411167_dp, 4.493554083058828_dp, &
             6.804407417058434_dp, 10.605675030165191_dp, &
             ! w
             1.8955688764928513e-09_dp, 1.1438287508592789e-08_dp, 4.53486886574274e-08_dp, &
             1.4474615889350026e-07_dp, 4.025034044092842e-07_dp, 1.0151662455799979e-06_dp, &
             2.378124001564282e-06_dp, 5.2548778291870785e-06_dp, 1.1070438405226986e-05_dp, &
             2.240869431511857e-05_dp, 4.383928613823005e-05_dp, 8.326875182364188e-05_dp, &
             0.00015411449203861708_dp, 0.00027875418150330873_dp, 0.0004939299553289538_dp, &
             0.0008591174996412999_dp, 0.0014693491827789818_dp, 0.0024746689583404714_dp, &
             0.00410938010334063_dp, 0.006735642321050696_dp, 0.010907943885760923_dp, &
             0.017467740010951468_dp, 0.027681420718614114_dp, 0.04344019773107903_dp, &
             0.06754818146101546_dp, 0.1041361313656899_dp, 0.15925593297187876_dp, &
             0.24174224857445048_dp, 0.36449500489368925_dp, 0.5465083372092596_dp, &
             0.8164758809924726_dp, 1.220459018850422_dp, 1.8424194425880962_dp, &
             2.877886579657062_dp, 5.07557175163324_dp]

      CASE (1099)

         aw(:) = & ! a
            [5.501377972845272e-10_dp, 5.751093865121371e-09_dp, 2.9728149442449127e-08_dp, &
             1.1325193515840639e-07_dp, 3.5961701638322006e-07_dp, 1.0084790431472421e-06_dp, &
             2.580420963636574e-06_dp, 6.1491842289532385e-06_dp, 1.3837993391706146e-05_dp, &
             2.969995165906444e-05_dp, 6.124314953114723e-05_dp, 0.0001220183601617238_dp, &
             0.0002359296853952235_dp, 0.0004442982069209724_dp, 0.0008172664042291088_dp, &
             0.0014719643281044014_dp, 0.0026010922800610695_dp, 0.004517386432482301_dp, &
             0.00772207362828217_dp, 0.013009240400590979_dp, 0.021623528561297456_dp, &
             0.03549640594646649_dp, 0.05759739083532943_dp, 0.09245235085315245_dp, &
             0.14690323107411135_dp, 0.2312151166379835_dp, 0.3606821436626231_dp, &
             0.5579528938458069_dp, 0.8564113398200852_dp, 1.305174822193762_dp, &
             1.9768016625798621_dp, 2.980291156210381_dp, 4.486875422479056_dp, &
             6.796222076734361_dp, 10.595804565593246_dp, &
             ! w
             1.72129509413978e-09_dp, 1.0788757961992678e-08_dp, 4.34033893993919e-08_dp, &
             1.3970375372857102e-07_dp, 3.906892431770502e-07_dp, 9.894818001539318e-07_dp, &
             2.325453493751884e-06_dp, 5.1518693164691434e-06_dp, 1.0876809771460424e-05_dp, &
             2.2056850075465586e-05_dp, 4.321854354441453e-05_dp, 8.22018408755955e-05_dp, &
             0.00015232319192638562_dp, 0.0002758099724183504_dp, 0.0004891844131686903_dp, &
             0.0008516059003545024_dp, 0.0014576595299728397_dp, 0.002456767642845342_dp, &
             0.0040823850792124005_dp, 0.0066955357450820185_dp, 0.01084921972511023_dp, &
             0.017382989899991905_dp, 0.027560875868666485_dp, 0.04327126948202192_dp, &
             0.06731508254199497_dp, 0.10381971847361003_dp, 0.15883398979585_dp, &
             0.24119051507594835_dp, 0.36378930949645155_dp, 0.5456280929355098_dp, &
             0.8154087470717531_dp, 1.2192053642258545_dp, 1.8409928150905468_dp, &
             2.8762999850320745_dp, 5.0737556591194934_dp]

      CASE (1100)

         aw(:) = & ! a
            [5.103392705636164e-10_dp, 5.4837891074081695e-09_dp, 2.8715753824103955e-08_dp, &
             1.1018476697347952e-07_dp, 3.5151739685525435e-07_dp, 9.89063980843569e-07_dp, &
             2.5371712118435913e-06_dp, 6.058280067003997e-06_dp, 1.365582403053583e-05_dp, &
             2.9349250624813884e-05_dp, 6.059088192000604e-05_dp, 0.00012084120303786125_dp, &
             0.00023386119724053424_dp, 0.0004407494637846186_dp, 0.0008113088742741943_dp, &
             0.0014621599692364894_dp, 0.002585251239896863_dp, 0.004492227377804558_dp, &
             0.007682755917594446_dp, 0.012948731679237024_dp, 0.021531767123932975_dp, &
             0.03535921649569577_dp, 0.05739511852391548_dp, 0.09215819713751497_dp, &
             0.14648131393641453_dp, 0.23061834805380185_dp, 0.3598501317054071_dp, &
             0.556810245725112_dp, 0.8548669592473823_dp, 1.303123006072089_dp, &
             1.974125850818753_dp, 2.976870737530385_dp, 4.482594437506625_dp, &
             6.79097459073463_dp, 10.589476091186905_dp, &
             ! w
             1.6151901908964372e-09_dp, 1.0387078192676779e-08_dp, 4.218972510259385e-08_dp, &
             1.3654209896602066e-07_dp, 3.8325595304660485e-07_dp, 9.73279325409775e-07_dp, &
             2.2921587537684575e-06_dp, 5.086644118998708e-06_dp, 1.0754029742495814e-05_dp, &
             2.183347468867218e-05_dp, 4.2824037483162775e-05_dp, 8.152314991849026e-05_dp, &
             0.00015118276077657951_dp, 0.0002739341609224345_dp, 0.0004861589264290202_dp, &
             0.0008468140363175547_dp, 0.001450198240015965_dp, 0.002445335751127185_dp, &
             0.004065137790895596_dp, 0.006669900243161303_dp, 0.010811668932834194_dp, &
             0.017328776465540815_dp, 0.027483737721658463_dp, 0.043163134221858246_dp, &
             0.06716582328218623_dp, 0.103617051375306_dp, 0.15856365399234945_dp, &
             0.24083693155832608_dp, 0.36333694858012344_dp, 0.5450637190057266_dp, &
             0.814724414665393_dp, 1.218401291304493_dp, 1.8400776962217131_dp, &
             2.875282198772647_dp, 5.072590676326598_dp]

      CASE (1101)

         aw(:) = & ! a
            [4.87790941234244e-10_dp, 5.331448753653054e-09_dp, 2.813510455722938e-08_dp, &
             1.08418382727191e-07_dp, 3.4683968744854786e-07_dp, 9.778281630095731e-07_dp, &
             2.512101857352496e-06_dp, 6.005520634981572e-06_dp, 1.3549983541859966e-05_dp, &
             2.9145310648380686e-05_dp, 6.02112827489952e-05_dp, 0.00012015567545716495_dp, &
             0.00023265588372084164_dp, 0.0004386805145719425_dp, 0.0008078339417896351_dp, &
             0.0014564388034235037_dp, 0.002576003885409963_dp, 0.004477535333173045_dp, &
             0.00765978820403723_dp, 0.012913374453879503_dp, 0.021478132975173543_dp, &
             0.03527900912358578_dp, 0.05727683215631881_dp, 0.09198614093419698_dp, &
             0.14623447448778035_dp, 0.23026914428744572_dp, 0.3593631836539147_dp, &
             0.5561413776918341_dp, 0.8539627853472211_dp, 1.3019215705036256_dp, &
             1.9725588243293637_dp, 2.9748674076764856_dp, 4.48008681495523_dp, &
             6.787900567913543_dp, 10.585768567804074_dp, &
             ! w
             1.5550314477459792e-09_dp, 1.0157002220578889e-08_dp, 4.14907038463106e-08_dp, &
             1.3471541252004687e-07_dp, 3.789519445318463e-07_dp, 9.63882474405851e-07_dp, &
             2.2728241317929927e-06_dp, 5.048727236394103e-06_dp, 1.068259184943716e-05_dp, &
             2.1703408558028543e-05_dp, 4.2594175610540045e-05_dp, 8.112747816561795e-05_dp, &
             0.00015051755836461726_dp, 0.0002728395175187649_dp, 0.0004843926519706005_dp, &
             0.000844015502277498_dp, 0.0014458392157007788_dp, 0.0024386549167706656_dp, &
             0.004055055480207437_dp, 0.006654910353169939_dp, 0.01078970627313776_dp, &
             0.01729706073523714_dp, 0.027438600743000037_dp, 0.04309984638435192_dp, &
             0.06707845002367935_dp, 0.10349839258939435_dp, 0.15840534887615287_dp, &
             0.2406298442451036_dp, 0.3630719694107384_dp, 0.544733080756475_dp, &
             0.814323450442657_dp, 1.2179301220284808_dp, 1.8395414186031285_dp, &
             2.874685735849042_dp, 5.071907959157792_dp]

      CASE (1102)

         aw(:) = & ! a
            [4.795969592006824e-10_dp, 5.275916434711567e-09_dp, 2.7922751250213468e-08_dp, &
             1.0777103411958934e-07_dp, 3.451229428888808e-07_dp, 9.737002394193796e-07_dp, &
             2.5028841708080844e-06_dp, 5.986109014347054e-06_dp, 1.3511021003903979e-05_dp, &
             2.9070201116827335e-05_dp, 6.007142473066248e-05_dp, 0.0001199030165548222_dp, &
             0.0002322115188391347_dp, 0.00043791754846338155_dp, 0.0008065521849923825_dp, &
             0.0014543280492845764_dp, 0.002572591516461448_dp, 0.004472112843042736_dp, &
             0.007651309958081472_dp, 0.012900320784959299_dp, 0.021458328785233186_dp, &
             0.03524938899765936_dp, 0.05723314434049473_dp, 0.0919225865556483_dp, &
             0.1461432868418258_dp, 0.23014012825153402_dp, 0.3591832601184768_dp, &
             0.5558942145512157_dp, 0.8536286437535189_dp, 1.3014775411438124_dp, &
             1.971979639341572_dp, 2.9741269163378576_dp, 4.479159872789606_dp, &
             6.786764206435387_dp, 10.584397976122634_dp, &
             ! w
             1.5331612000906034e-09_dp, 1.0072915099476297e-08_dp, 4.1234512351720566e-08_dp, &
             1.3404486818299088e-07_dp, 3.7737028264744505e-07_dp, 9.60426405860645e-07_dp, &
             2.265708386077519e-06_dp, 5.03476516006834e-06_dp, 1.0656274589640479e-05_dp, &
             2.16554746900241e-05_dp, 4.250943544095629e-05_dp, 8.098156850232206e-05_dp, &
             0.00015027219186875497_dp, 0.00027243565379040773_dp, 0.00048374085651024286_dp, &
             0.0008429825829358832_dp, 0.0014442300502727236_dp, 0.0024361882438229685_dp, &
             0.004051332376981885_dp, 0.0066493742661886_dp, 0.010781593959763056_dp, &
             0.017285344551948564_dp, 0.027421924727322945_dp, 0.04307646202520369_dp, &
             0.06704616311542803_dp, 0.10345454072656213_dp, 0.15834684010811978_dp, &
             0.24055329956564758_dp, 0.3629740189971625_dp, 0.5446108507648367_dp, &
             0.8141752133576832_dp, 1.2177559212708298_dp, 1.8393431388605066_dp, &
             2.874465199741433_dp, 5.071655532936565_dp]

      CASE (1103)

         aw(:) = & ! a
            [4.795656899618547e-10_dp, 5.275703815294485e-09_dp, 2.7921936961144855e-08_dp, &
             1.0776855138181483e-07_dp, 3.4511636432440086e-07_dp, 9.736844316472588e-07_dp, &
             2.502848878810929e-06_dp, 5.986034679665068e-06_dp, 1.351087179459349e-05_dp, &
             2.9069913436937345e-05_dp, 6.0070889018443205e-05_dp, 0.00011990204924813175_dp, &
             0.00023220981941233054_dp, 0.0004379146334968274_dp, 0.0008065472901235645_dp, &
             0.0014543199897438439_dp, 0.0025725784880714616_dp, 0.004472092139150253_dp, &
             0.0076512775787345375_dp, 0.012900270912393746_dp, 0.021458253095558853_dp, &
             0.035249275775026304_dp, 0.057232977352526034_dp, 0.09192234367092901_dp, &
             0.14614293842620601_dp, 0.2301396354369227_dp, 0.35918257311636737_dp, &
             0.5558932713143581_dp, 0.8536273695258033_dp, 1.301475849667882_dp, &
             1.9719774360937354_dp, 2.9741241036848662_dp, 4.479156356463695_dp, &
             6.78675989973554_dp, 10.584392785622473_dp, &
             ! w
             1.5330775975217554e-09_dp, 1.0072592777311047e-08_dp, 4.1233529030227604e-08_dp, &
             1.340422980430436e-07_dp, 3.7736422726281444e-07_dp, 9.604131786140613e-07_dp, &
             2.265681137565567e-06_dp, 5.034711687200461e-06_dp, 1.0656173793936654e-05_dp, &
             2.1655291044486802e-05_dp, 4.2509110928099235e-05_dp, 8.098101071113425e-05_dp, &
             0.00015027125548183564_dp, 0.0002724341129320892_dp, 0.0004837383684753617_dp, &
             0.0008429786397978206_dp, 0.0014442239070924347_dp, 0.0024361788226602782_dp, &
             0.00405131814699903_dp, 0.006649353095299916_dp, 0.010781562935127904_dp, &
             0.017285299762391196_dp, 0.02742186100680197_dp, 0.043076372704118585_dp, &
             0.06704603983523175_dp, 0.10345437338343318_dp, 0.1583466170193326_dp, &
             0.2405530080379608_dp, 0.3629736465907005_dp, 0.5446103872659257_dp, &
             0.8141746527651287_dp, 1.217755263636201_dp, 1.8393423906044235_dp, &
             2.87446436738254_dp, 5.071654580814584_dp]

      CASE (1104)

         aw(:) = & ! a
            [3.392583358853396e-10_dp, 3.705407847421689e-09_dp, 1.9547969316324468e-08_dp, &
             7.531481753745472e-08_dp, 2.4091140464501275e-07_dp, 6.791352064866371e-07_dp, &
             1.7446347371894621e-06_dp, 4.170585443286809e-06_dp, 9.409532747229403e-06_dp, &
             2.0238751058915135e-05_dp, 4.181005522936697e-05_dp, 8.343276424802206e-05_dp, &
             0.00016154635685945883_dp, 0.0003045953808677941_dp, 0.0005609055984931827_dp, &
             0.0010112379381757267_dp, 0.0017885524256270916_dp, 0.0031087711443333254_dp, &
             0.005318161337419406_dp, 0.008965610923402157_dp, 0.014911879199690583_dp, &
             0.024493347835706163_dp, 0.0397655254397142_dp, 0.06386247830017364_dp, &
             0.10152373304283478_dp, 0.15986182947659336_dp, 0.24947433688911422_dp, &
             0.3860484753741254_dp, 0.5926740547659806_dp, 0.9031945550767589_dp, &
             1.3671519011160778_dp, 2.0574165664603616_dp, 3.0830995364239744_dp, &
             4.6152917919308_dp, 6.953369872081639_dp, 10.785080428126326_dp, &
             ! w
             1.081197541465692e-09_dp, 7.057546012719769e-09_dp, 2.8822924800821273e-08_dp, &
             9.357227178628371e-08_dp, 2.631937344316969e-07_dp, 6.694025142445665e-07_dp, &
             1.5783645920007254e-06_dp, 3.505951732103351e-06_dp, 7.417989674212712e-06_dp, &
             1.5070419626413703e-05_dp, 2.9575843759692954e-05_dp, 5.633078152174802e-05_dp, &
             0.00010450977666290101_dp, 0.0001894392058605559_dp, 0.0003363207835330652_dp, &
             0.0005860042494920928_dp, 0.001003841266152317_dp, 0.0016931309203139233_dp, &
             0.0028153507775786333_dp, 0.004620337979483453_dp, 0.007490954836592828_dp, &
             0.012008696996148273_dp, 0.01904939166402658_dp, 0.029921884640777753_dp, &
             0.04656785882493066_dp, 0.07184834939142913_dp, 0.10995340151492985_dp, &
             0.16698854965561868_dp, 0.2518230788634322_dp, 0.3773529364199611_dp, &
             0.5625057841387796_dp, 0.835826435070735_dp, 1.2431500487196163_dp, &
             1.8682075366993838_dp, 2.906548354818295_dp, 5.108387324537728_dp]

      CASE (1105)

         aw(:) = & ! a
            [3.329073031698242e-10_dp, 3.6623150640078627e-09_dp, 1.9382993680284026e-08_dp, &
             7.48113264049199e-08_dp, 2.3957464540027657e-07_dp, 6.759172398097704e-07_dp, &
             1.7374404403366709e-06_dp, 4.155416206586593e-06_dp, 9.379046204241336e-06_dp, &
             2.0179901540328174e-05_dp, 4.170031783927232e-05_dp, 8.323421939139891e-05_dp, &
             0.00016119660234165963_dp, 0.0003039938231823944_dp, 0.0005598931291495909_dp, &
             0.001009567301849729_dp, 0.0017858457174757762_dp, 0.0031044598305894236_dp, &
             0.005311403002232281_dp, 0.008955175680694129_dp, 0.014895997532480078_dp, &
             0.024469510832743475_dp, 0.039730228487644875_dp, 0.06381090106606593_dp, &
             0.10144935329028434_dp, 0.15975597769636998_dp, 0.24932571530014258_dp, &
             0.3858426913050339_dp, 0.592393256664666_dp, 0.9028173072566857_dp, &
             1.3666534775893948_dp, 2.0567698482885395_dp, 3.0822765850638008_dp, &
             4.614265807197474_dp, 6.952116306300628_dp, 10.783572426341268_dp, &
             ! w
             1.0642395834767485e-09_dp, 6.992244154735406e-09_dp, 2.8623703419626118e-08_dp, &
             9.305016929459931e-08_dp, 2.6196062853812135e-07_dp, 6.667045307747711e-07_dp, &
             1.5728021344251456e-06_dp, 3.4950219596987456e-06_dp, 7.3973573593283105e-06_dp, &
             1.5032781049426562e-05_dp, 2.9509192705111625e-05_dp, 5.621581293353092e-05_dp, &
             0.00010431606930205269_dp, 0.00018911970745805805_dp, 0.00033580397950069526_dp, &
             0.0005851832330558694_dp, 0.0010025587522760112_dp, 0.0016911590786602886_dp, &
             0.002812364638773913_dp, 0.004615881158040535_dp, 0.007484396522804887_dp, &
             0.011999179760414362_dp, 0.019035770839226068_dp, 0.029902662275653615_dp, &
             0.046541118930219186_dp, 0.0718117070372261_dp, 0.10990398684444458_dp, &
             0.16692305902537857_dp, 0.25173793712740533_dp, 0.37724462006065657_dp, &
             0.5623713382170766_dp, 0.8356641468037543_dp, 1.2429600692307807_dp, &
             1.8679918889394456_dp, 2.906308815012812_dp, 5.108113006634463_dp]

      CASE (1106)

         aw(:) = & ! a
            [2.7050727917365697e-10_dp, 2.813067805178994e-09_dp, 1.4504351493006907e-08_dp, &
             5.5177418752615494e-08_dp, 1.7504726861967963e-07_dp, 4.905635849651036e-07_dp, &
             1.2545879366100444e-06_dp, 2.9885144120138355e-06_dp, 6.723103566240445e-06_dp, &
             1.4425612830626782e-05_dp, 2.9739611597429572e-05_dp, 5.924009312529937e-05_dp, &
             0.00011452404284250141_dp, 0.00021563581813658367_dp, 0.00039659700822371413_dp, &
             0.0007142142731290889_dp, 0.0012619365805488006_dp, 0.002191411211738429_dp, &
             0.003745667097083974_dp, 0.0063097044048837135_dp, 0.010486927767345493_dp, &
             0.017213655832383212_dp, 0.027929326261874537_dp, 0.04482761898970188_dp, &
             0.07122440580773219_dp, 0.112093406541656_dp, 0.1748414232337484_dp, &
             0.2704246996399173_dp, 0.41495096922847297_dp, 0.6319777889621487_dp, &
             0.955830673316784_dp, 1.4364909250163393_dp, 2.147144751405847_dp, &
             3.1970058826101844_dp, 4.7570068157244645_dp, 7.126223965513759_dp, &
             10.99274096225511_dp, &
             ! w
             8.445518525777438e-10_dp, 5.267220940668875e-09_dp, 2.115025761353554e-08_dp, &
             6.800360859927778e-08_dp, 1.9003765150769885e-07_dp, 4.810434430050926e-07_dp, &
             1.1300681762979004e-06_dp, 2.5027497194110852e-06_dp, 5.2824410404067726e-06_dp, &
             1.0709658988930224e-05_dp, 2.098048164453259e-05_dp, 3.9898003499085404e-05_dp, &
             7.392117665509561e-05_dp, 0.00013382995052025227_dp, 0.00023733540463421102_dp, &
             0.00041312381816434416_dp, 0.0007070557331304373_dp, 0.0011915748379394996_dp, &
             0.0019798582465593113_dp, 0.0032469164023683655_dp, 0.005260805535715079_dp, &
             0.008428455313462943_dp, 0.013362485190068375_dp, 0.020977985759496356_dp, &
             0.03263185739311718_dp, 0.050322340971325895_dp, 0.07697351747735423_dp, &
             0.11684008476052864_dp, 0.17608462030417454_dp, 0.26361077305794905_dp, &
             0.39230462037391517_dp, 0.5810141285826121_dp, 0.8581145953532153_dp, &
             1.2691905053037102_dp, 1.8977256131017224_dp, 2.9393154023620687_dp, &
             5.145921832303291_dp]

      CASE (1107)

         aw(:) = & ! a
            [2.3223808372042365e-10_dp, 2.5548524933308593e-09_dp, 1.352169035539464e-08_dp, &
             5.21888214756931e-08_dp, 1.671286828966234e-07_dp, 4.7152384661760095e-07_dp, &
             1.2120486846226164e-06_dp, 2.898842825760203e-06_dp, 6.542877894049853e-06_dp, &
             1.4077618231887907e-05_dp, 2.9090387487037663e-05_dp, 5.8064682039956405e-05_dp, &
             0.00011245169998753053_dp, 0.00021206788311506223_dp, 0.0003905847476943591_dp, &
             0.0007042801017601814_dp, 0.001245816500656955_dp, 0.0021656894751080928_dp, &
             0.003705266013527944_dp, 0.0062471831665418965_dp, 0.010391534812560785_dp, &
             0.01707007100730302_dp, 0.027716024271449913_dp, 0.04451478842379094_dp, &
             0.07077137484275098_dp, 0.11144556697019102_dp, 0.17392672265563247_dp, &
             0.2691499056140394_dp, 0.41319811866157097_dp, 0.6296015323757118_dp, &
             0.9526575840859893_dp, 1.4323221568607232_dp, 2.1417634024963474_dp, &
             3.1901894767114403_dp, 4.748542400949159_dp, 7.115915909769255_dp, &
             10.98037251245892_dp, &
             ! w
             7.424197720244335e-10_dp, 4.877830620145493e-09_dp, 1.9968063906317793e-08_dp, &
             6.491234583634297e-08_dp, 1.8274527705182694e-07_dp, 4.650970076759751e-07_dp, &
             1.0971960349620669e-06_dp, 2.438147913012728e-06_dp, 5.160440087698477e-06_dp, &
             1.0486956660894084e-05_dp, 2.0585786770717878e-05_dp, 3.921648249045094e-05_dp, &
             7.277150492564249e-05_dp, 0.00013193102291630016_dp, 0.00023425883596034345_dp, &
             0.0004082272734054518_dp, 0.0006993908948265893_dp, 0.0011797625327401736_dp, &
             0.001961922092036988_dp, 0.0032200656121570376_dp, 0.005221157776993632_dp, &
             0.008370692584741291_dp, 0.013279440882055578_dp, 0.020860171387635646_dp, &
             0.03246695574405004_dp, 0.05009472233518375_dp, 0.07666388171703159_dp, &
             0.11642539783099821_dp, 0.17553859785687212_dp, 0.2629052483022534_dp, &
             0.3914121629172119_dp, 0.579912127418803_dp, 0.8567904474507609_dp, &
             1.2676462071162014_dp, 1.8959773030782618_dp, 2.9373758246484116_dp, &
             5.143699491009028_dp]

      CASE (1108)

         aw(:) = & ! a
            [2.2945118174358155e-10_dp, 2.236789221523896e-09_dp, 1.1158861937845354e-08_dp, &
             4.166303566457208e-08_dp, 1.3056072174388643e-07_dp, 3.6268035326635414e-07_dp, &
             9.213263312298933e-07_dp, 2.1829937398352367e-06_dp, 4.88958605943292e-06_dp, &
             1.0453197063540192e-05_dp, 2.1483026489780754e-05_dp, 4.267783321941021e-05_dp, &
             8.231026427586508e-05_dp, 0.0001546557105438077_dp, 0.0002839087187801501_dp, &
             0.0005104141762953134_dp, 0.000900460949269449_dp, 0.0015615026271285923_dp, &
             0.0026655683888016915_dp, 0.004484925717776435_dp, 0.0074459215148885895_dp, &
             0.012209588937764982_dp, 0.019791381285845417_dp, 0.0317377177518386_dp, &
             0.050384497973931804_dp, 0.07923318813331695_dp, 0.12349465225281009_dp, &
             0.1908712633967883_dp, 0.2926765806444099_dp, 0.44543365732944873_dp, &
             0.6731576869426142_dp, 1.0106406489859432_dp, 1.5082829523572383_dp, &
             2.2395645825450092_dp, 3.3137843396874542_dp, 4.901711167282969_dp, &
             7.302136973110393_dp, 11.203524953934489_dp, &
             ! w
             6.981670824547048e-10_dp, 4.086491860878355e-09_dp, 1.600596587105048e-08_dp, &
             5.073216769045445e-08_dp, 1.4040510764370505e-07_dp, 3.528823060596037e-07_dp, &
             8.244147487705258e-07_dp, 1.8176967643122893e-06_dp, 3.822385115608636e-06_dp, &
             7.725351436438007e-06_dp, 1.5093509940158535e-05_dp, 2.8635653696217524e-05_dp, &
             5.294508343880973e-05_dp, 9.567727491034868e-05_dp, 0.00016939399254351406_dp, &
             0.00029441781083090263_dp, 0.0005032039362582628_dp, 0.0008469703718650195_dp, &
             0.0014056617262180801_dp, 0.0023027961295353136_dp, 0.003727413362833333_dp, &
             0.005966280010262805_dp, 0.009450815808112583_dp, 0.014825022852009371_dp, &
             0.023043216169087073_dp, 0.035509837831885896_dp, 0.05427848561208271_dp, &
             0.0823341648692915_dp, 0.12399298590855426_dp, 0.18547015078704393_dp, &
             0.27569834144769756_dp, 0.4075483905474138_dp, 0.5997849192414254_dp, &
             0.8806145036673687_dp, 1.2953791544550293_dp, 1.9273324514130719_dp, &
             2.9721397505598137_dp, 5.18354241622047_dp]

      CASE (1109)

         aw(:) = & ! a
            [1.667868801552446e-10_dp, 1.8179755283104173e-09_dp, 9.582028773001437e-09_dp, &
             3.6899335536766786e-08_dp, 1.1799254481612924e-07_dp, 3.325469907748598e-07_dp, &
             8.541323230445451e-07_dp, 2.0415378203865073e-06_dp, 4.605526599093449e-06_dp, &
             9.904985321538931e-06_dp, 2.04604812199647e-05_dp, 4.0826432683718924e-05_dp, &
             7.904518533587914e-05_dp, 0.0001490314955754176_dp, 0.0002744249021451099_dp, &
             0.0004947299293117908_dp, 0.0008749824800004178_dp, 0.0015207954156303217_dp, &
             0.0026015322154239397_dp, 0.0043856540122633484_dp, 0.007294148510614474_dp, &
             0.01198061004907101_dp, 0.019450317074772808_dp, 0.03123598509396637_dp, &
             0.0496553622318882_dp, 0.07818631080208505_dp, 0.12200963976991717_dp, &
             0.18879039281179869_dp, 0.2897971624412768_dp, 0.4415009319542647_dp, &
             0.6678597557838679_dp, 1.0036076534695024_dp, 1.499093291970609_dp, &
             2.227760796385059_dp, 3.2988991774321867_dp, 4.8832982475140065_dp, &
             7.27978491606707_dp, 11.17677218952431_dp, &
             ! w
             5.310806901437774e-10_dp, 3.4602732626352495e-09_dp, 1.4122259498450536e-08_dp, &
             4.582981965578465e-08_dp, 1.2887411461676403e-07_dp, 3.27715064754208e-07_dp, &
             7.725982593837445e-07_dp, 1.7159395957068919e-06_dp, 3.6302848469173863e-06_dp, &
             7.374705946718379e-06_dp, 1.4471926082771447e-05_dp, 2.7561869909759182e-05_dp, &
             5.113245482373218e-05_dp, 9.268062441872776e-05_dp, 0.00016453347225450877_dp, &
             0.0002866715289531349_dp, 0.0004910588996422571_dp, 0.0008282188585073472_dp, &
             0.0013771277497967788_dp, 0.002259974584058029_dp, 0.0036640034283707008_dp, &
             0.005873595599580233_dp, 0.00931706103177629_dp, 0.014634432430750365_dp, &
             0.02277508348154464_dp, 0.03513749254536177_dp, 0.053768346106742776_dp, &
             0.08164508630215679_dp, 0.12307625877351171_dp, 0.184270671198136_dp, &
             0.2741576322215528_dp, 0.40561018211817773_dp, 0.5974036382187602_dp, &
             0.8777657942402292_dp, 1.2920687784313114_dp, 1.9235942705223896_dp, &
             2.9679975310133493_dp, 5.178793799115431_dp]

      CASE (1110)

         aw(:) = & ! a
            [1.6277733834662105e-10_dp, 1.7907144010725795e-09_dp, 9.477449484401406e-09_dp, &
             3.657951808843438e-08_dp, 1.1714168835989649e-07_dp, 3.3049443352270925e-07_dp, &
             8.49533582798658e-07_dp, 2.031819645447088e-06_dp, 4.585949852531391e-06_dp, &
             9.86710316848169e-06_dp, 2.038966035366554e-05_dp, 4.069795031501298e-05_dp, &
             7.88181995596325e-05_dp, 0.00014863989361694926_dp, 0.0002737636386295888_dp, &
             0.0004936349519791294_dp, 0.0008732016799527758_dp, 0.0015179472129912212_dp, &
             0.002597047389455791_dp, 0.0043786952176379365_dp, 0.007283500637663834_dp, &
             0.011964533298539777_dp, 0.019426353539209358_dp, 0.03120070914226171_dp, &
             0.04960406569583599_dp, 0.07811261681931812_dp, 0.12190504557181894_dp, &
             0.18864375376230252_dp, 0.28959414978753883_dp, 0.4412235276882574_dp, &
             0.6674858909202261_dp, 1.0031111467828642_dp, 1.4984442875264588_dp, &
             2.226926886976337_dp, 3.297847252842361_dp, 4.881996672326622_dp, &
             7.278204542232129_dp, 11.1748803415977_dp, &
             ! w
             5.203673382195436e-10_dp, 3.418906377706841e-09_dp, 1.3995758758829816e-08_dp, &
             4.549752714791722e-08_dp, 1.2808747053452296e-07_dp, 3.259898155891666e-07_dp, &
             7.690325388753919e-07_dp, 1.7089152876870955e-06_dp, 3.6169893148701564e-06_dp, &
             7.350382819228773e-06_dp, 1.4428724957840905e-05_dp, 2.748711262097471e-05_dp, &
             5.1006067435867694e-05_dp, 9.24713959880647e-05_dp, 0.000164193690718227_dp, &
             0.00028612941043763536_dp, 0.0004902080706933056_dp, 0.0008269039775944339_dp, &
             0.0013751251794088329_dp, 0.002256966875123335_dp, 0.0036595463058615237_dp, &
             0.005867076220147428_dp, 0.009307646646742773_dp, 0.014621009430905047_dp, &
             0.022756188408826362_dp, 0.03511123943293656_dp, 0.05373235898834015_dp, &
             0.08159645251637021_dp, 0.12301152821140963_dp, 0.1841859384793427_dp, &
             0.27404875006424667_dp, 0.4054731564483023_dp, 0.5972352301490466_dp, &
             0.8775642675303071_dp, 1.291834533847968_dp, 1.9233297080266942_dp, &
             2.9677043498389613_dp, 5.178457710695912_dp]

      CASE (1111)

         aw(:) = & ! a
            [2.0110245477708717e-10_dp, 1.8338324051049543e-09_dp, 8.807415909952976e-09_dp, &
             3.215940396296987e-08_dp, 9.929980776260292e-08_dp, 2.7290863658649594e-07_dp, &
             6.876284979298008e-07_dp, 1.6186873301191621e-06_dp, 3.6062907206598513e-06_dp, &
             7.6751849367455e-06_dp, 1.5713371678780415e-05_dp, 3.111224801348558e-05_dp, &
             5.982936132480508e-05_dp, 0.00011212459871159381_dp, 0.00020535557778846914_dp, &
             0.00036841925059189753_dp, 0.0006487242102097525_dp, 0.0011230143308221144_dp, &
             0.0019139989312549359_dp, 0.003215661316028958_dp, 0.005331432869723926_dp, &
             0.008731295357066737_dp, 0.01413653243150883_dp, 0.022644598474384287_dp, &
             0.03591182759727779_dp, 0.05641904104445589_dp, 0.08785531525625187_dp, &
             0.13566934692892502_dp, 0.20785760044585336_dp, 0.3160862927213594_dp, &
             0.4772848999702004_dp, 0.7159122693612_dp, 1.067207640356472_dp, &
             1.581967430173526_dp, 2.3339420796308366_dp, 3.432498353923494_dp, &
             5.048238919789414_dp, 7.479689259581741_dp, 11.415730590902372_dp, &
             ! w
             5.968418225139929e-10_dp, 3.256902745695496e-09_dp, 1.23857558738077e-08_dp, &
             3.8590143681169086e-08_dp, 1.0555800971710985e-07_dp, 2.630108793498219e-07_dp, &
             6.103173029981139e-07_dp, 1.3383222923915242e-06_dp, 2.801599169012276e-06_dp, &
             5.640559130663436e-06_dp, 1.0983914271100391e-05_dp, 2.0778719289679447e-05_dp, &
             3.832025901328514e-05_dp, 6.909117742429718e-05_dp, 0.00012207380160977687_dp, &
             0.00021177912777321258_dp, 0.00036135095050280117_dp, 0.0006072687863640266_dp, &
             0.0010064088604687794_dp, 0.0016465564417445658_dp, 0.002661933533266697_dp, &
             0.004255965622935348_dp, 0.006734430985166784_dp, 0.010553412411735606_dp, &
             0.0163882354602878_dp, 0.02523200972218895_dp, 0.038535745364557325_dp, &
             0.05840668195591196_dp, 0.08788809845020105_dp, 0.1313538016715104_dp, &
             0.19506688722482382_dp, 0.28798374295812257_dp, 0.4229550808489394_dp, &
             0.61865966279584_dp, 0.9031378946925505_dp, 1.3214995288270444_dp, &
             1.9567862623295684_dp, 3.004755479292111_dp, 5.220944579946335_dp]

      CASE (1112)

         aw(:) = & ! a
            [1.3437407957628294e-10_dp, 1.394435117442241e-09_dp, 7.182416370397351e-09_dp, &
             2.730762635219069e-08_dp, 8.659943689850386e-08_dp, 2.426266359935003e-07_dp, &
             6.203769710828572e-07_dp, 1.4775414461536802e-06_dp, 3.3235077272525825e-06_dp, &
             7.130386890620804e-06_dp, 1.4698503575742955e-05_dp, 2.9276426895387725e-05_dp, &
             5.659367477926323e-05_dp, 0.00010655272455203315_dp, 0.00019596046011447957_dp, &
             0.00035287861780045976_dp, 0.0006234680691090676_dp, 0.0010826353365572625_dp, &
             0.0018504214257018725_dp, 0.00311698677585132_dp, 0.005180358845163511_dp, &
             0.008502982778956166_dp, 0.0137957746998753_dp, 0.022142129183443573_dp, &
             0.03517959919216362_dp, 0.055364325593348813_dp, 0.08635353841042134_dp, &
             0.133555692755207_dp, 0.20491765353690566_dp, 0.31204635865545577_dp, &
             0.47180320835663564_dp, 0.7085730265790277_dp, 1.0575207708426249_dp, &
             1.569377494464927_dp, 2.3178494973959336_dp, 3.4122931937413927_dp, &
             5.0233395086820165_dp, 7.44955770067443_dp, 11.379755680958926_dp, &
             ! w
             4.191664346120608e-10_dp, 2.6089588688151355e-09_dp, 1.0468130286223671e-08_dp, &
             3.3643036769489804e-08_dp, 9.398846452500873e-08_dp, 2.3786195133169826e-07_dp, &
             5.586918258276425e-07_dp, 1.2371613813655595e-06_dp, 2.6109285954306044e-06_dp, &
             5.2929147333749965e-06_dp, 1.0368104898601239e-05_dp, 1.9715340761791163e-05_dp, &
             3.652538653572135e-05_dp, 6.612338220154368e-05_dp, 0.00011725801191083363_dp, &
             0.00020409884978220548_dp, 0.00034929796569896756_dp, 0.0005886366895568365_dp, &
             0.0009780138452581346_dp, 0.0016038658744798612_dp, 0.002598580623374784_dp, &
             0.004163127844701338_dp, 0.006600052344705271_dp, 0.0103612575059776_dp, &
             0.016116780142495173_dp, 0.0248532042239162_dp, 0.038013738061316395_dp, &
             0.057696674716920054_dp, 0.086935627515981_dp, 0.1300949069929207_dp, &
             0.19342983663305913_dp, 0.2858931576619781_dp, 0.4203393077502191_dp, &
             0.6154617200199481_dp, 0.8993286990143969_dp, 1.3170885493613784_dp, &
             1.9518175383752772_dp, 2.9992559944657904_dp, 5.214636570200131_dp]

      CASE (1113)

         aw(:) = & ! a
            [1.1601607697221639e-10_dp, 1.2704051621815988e-09_dp, 6.709810724833609e-09_dp, &
             2.5868120732746033e-08_dp, 8.277902011352086e-08_dp, 2.3342444980558035e-07_dp, &
             5.997786660087458e-07_dp, 1.434035101913015e-06_dp, 3.235884890734147e-06_dp, &
             6.960825626056568e-06_dp, 1.4381431714953747e-05_dp, 2.8700956490162444e-05_dp, &
             5.557641247886324e-05_dp, 0.00010479641048718491_dp, 0.00019299206506064997_dp, &
             0.00034795811606454376_dp, 0.0006154559752169072_dp, 0.0010698031215018529_dp, &
             0.0018301840171428576_dp, 0.0030855304671364597_dp, 0.0051321311971811966_dp, &
             0.00843000383938626_dp, 0.013686721810593954_dp, 0.021981142149472045_dp, &
             0.0349447509548975_dp, 0.055025709484088366_dp, 0.08587094225863463_dp, &
             0.13287587097068862_dp, 0.203971282332162_dp, 0.31074488117228116_dp, &
             0.47003595964861594_dp, 0.7062052771722638_dp, 1.054393612911002_dp, &
             1.5653107093639502_dp, 2.3126484315515774_dp, 3.405759743358449_dp, &
             5.01528471270009_dp, 7.439806889162383_dp, 11.36811066495547_dp, &
             ! w
             3.701448362615787e-10_dp, 2.4217785264924756e-09_dp, 9.898899960853414e-09_dp, &
             3.2151799391594205e-08_dp, 9.046363029736844e-08_dp, 2.3013807393874382e-07_dp, &
             5.427349791770456e-07_dp, 1.2057301946878856e-06_dp, 2.551426558459768e-06_dp, &
             5.184019707557201e-06_dp, 1.0174583326118999e-05_dp, 1.938021291072514e-05_dp, &
             3.5958289512253096e-05_dp, 6.518356214403441e-05_dp, 0.00011572984843005127_dp, &
             0.00020165716359981325_dp, 0.0003454595782129568_dp, 0.0005826938045045975_dp, &
             0.0009689438534826649_dp, 0.0015902112379390992_dp, 0.002578291755396367_dp, &
             0.004133361667212425_dp, 0.006556919953469987_dp, 0.010299516896652976_dp, &
             0.016029475388628138_dp, 0.024731262521026646_dp, 0.037845553006875705_dp, &
             0.05746773107432987_dp, 0.08662826274276725_dp, 0.1296883599347844_dp, &
             0.1929008003830909_dp, 0.28521711302156244_dp, 0.4194929148394175_dp, &
             0.6144263786960392_dp, 0.8980948640390688_dp, 1.3156592261664193_dp, &
             1.9502070389285024_dp, 2.9974732342938153_dp, 5.212591840105367_dp]

      CASE (1114)

         aw(:) = & ! a
            [1.1461086022774257e-10_dp, 1.2608347250132962e-09_dp, 6.673033667984663e-09_dp, &
             2.575549013295909e-08_dp, 8.247898723430502e-08_dp, 2.3269978923779006e-07_dp, &
             5.981531480027921e-07_dp, 1.4305959676725715e-06_dp, 3.228948684387716e-06_dp, &
             6.947387310005722e-06_dp, 1.4356277138447397e-05_dp, 2.8655261720711888e-05_dp, &
             5.5495574585592494e-05_dp, 0.00010465674615477689_dp, 0.00019275586749572236_dp, &
             0.0003475663673951903_dp, 0.0006148177608579328_dp, 0.00106878047475534_dp, &
             0.0018285705285470502_dp, 0.0030830215211253426_dp, 0.005128283157424972_dp, &
             0.008424178921800203_dp, 0.013678014816836058_dp, 0.021968284790503695_dp, &
             0.03492598935918093_dp, 0.05499865095545739_dp, 0.08583236888122585_dp, &
             0.13282152092465668_dp, 0.20389560571516557_dp, 0.31064078687415264_dp, &
             0.46989458452296523_dp, 0.7060158287076141_dp, 1.0541433595112948_dp, &
             1.5649852096369112_dp, 2.3122320851617952_dp, 3.4052366718264238_dp, &
             5.014639770016525_dp, 7.439026076814212_dp, 11.367178106006026_dp, &
             ! w
             3.6638852279553266e-10_dp, 2.407238083730855e-09_dp, 9.854356896796465e-09_dp, &
             3.2034624171816796e-08_dp, 9.018586815469553e-08_dp, 2.295281062121795e-07_dp, &
             5.414726900926578e-07_dp, 1.2032403206360924e-06_dp, 2.5467075030109347e-06_dp, &
             5.175374725183649e-06_dp, 1.0159206712131733e-05_dp, 1.9353564498412823e-05_dp, &
             3.591316518515371e-05_dp, 6.510873482882477e-05_dp, 0.00011560811164678087_dp, &
             0.00020146255728214422_dp, 0.00034515351392672405_dp, 0.0005822197351342528_dp, &
             0.000968220054250909_dp, 0.0015891211893639553_dp, 0.00257667155972031_dp, &
             0.004130983911249265_dp, 0.006553473489839261_dp, 0.010294582215337335_dp, &
             0.016022495685632254_dp, 0.024721511373378826_dp, 0.037832100920601575_dp, &
             0.057449415326799494_dp, 0.08660366815696913_dp, 0.1296558226514676_dp, &
             0.19285845203915086_dp, 0.2851629875173099_dp, 0.4194251399966698_dp, &
             0.6143434617793996_dp, 0.8979960380691173_dp, 1.3155447308950585_dp, &
             1.950078022130876_dp, 2.9973304142716617_dp, 5.212428037833929_dp]

      CASE (1115)

         aw(:) = & ! a
            [1.80034357509866e-10_dp, 1.5405013996706422e-09_dp, 7.103368535354762e-09_dp, &
             2.530280891820828e-08_dp, 7.683577248880414e-08_dp, 2.0861752148453446e-07_dp, &
             5.207433286367151e-07_dp, 1.216699994028169e-06_dp, 2.694061531798567e-06_dp, &
             5.704086825668886e-06_dp, 1.1626269985033553e-05_dp, 2.293131953098188e-05_dp, &
             4.394817784536967e-05_dp, 8.211470613775864e-05_dp, 0.00014998792027380388_dp, &
             0.0002684330537629704_dp, 0.0004716224577744553_dp, 0.0008147854703367764_dp, &
             0.001386101183773497_dp, 0.002324776683503002_dp, 0.0038482861854482455_dp, &
             0.006293075598961585_dp, 0.01017491953505094_dp, 0.01627776928678679_dp, &
             0.025783645557273145_dp, 0.040461310593182864_dp, 0.06293864916806174_dp, &
             0.09709364948569206_dp, 0.14861265890347927_dp, 0.2257837469788173_dp, &
             0.3406200422415268_dp, 0.5104474754575444_dp, 0.7601536649641217_dp, &
             1.1254053791908751_dp, 1.657372003368264_dp, 2.4300518688154322_dp, &
             3.5528639235298303_dp, 5.196243193932063_dp, 7.658467241843419_dp, &
             11.628871207180643_dp, &
             ! w
             5.226128671451475e-10_dp, 2.6546082845004817e-09_dp, 9.769882587329497e-09_dp, &
             2.985934795899477e-08_dp, 8.059979543357828e-08_dp, 1.9884881037203774e-07_dp, &
             4.5787085477116726e-07_dp, 9.97750720446966e-07_dp, 2.077775594598853e-06_dp, &
             4.164741167064495e-06_dp, 8.079041450382808e-06_dp, 1.5232341902667766e-05_dp, &
             2.8008486112910915e-05_dp, 5.036571962701152e-05_dp, 8.877711872629837e-05_dp, &
             0.00015368261208308713_dp, 0.00026170762634631685_dp, 0.0004390205403531246_dp, &
             0.0007263681152204182_dp, 0.0011865646089542917_dp, 0.0019155433926019802_dp, &
             0.003058546388584826_dp, 0.0048336765941428845_dp, 0.007565940102885534_dp, &
             0.011736148052913429_dp, 0.018050760286625812_dp, 0.02754110247998155_dp, &
             0.041703621225533005_dp, 0.06269733221320213_dp, 0.0936210214480219_dp, &
             0.13890243522706044_dp, 0.2048479169193073_dp, 0.3004324997835046_dp, &
             0.4384823823409449_dp, 0.6375886226729988_dp, 0.9256286967007789_dp, &
             1.3474906868449552_dp, 1.9860223285698502_dp, 3.0370935753224853_dp, &
             5.258049117224792_dp]

      CASE (1116)

         aw(:) = & ! a
            [1.1458994531303259e-10_dp, 1.116463899865673e-09_dp, 5.568167531773443e-09_dp, &
             2.078595619395463e-08_dp, 6.513033407237547e-08_dp, 1.809089198308219e-07_dp, &
             4.595395122036253e-07_dp, 1.088781509339231e-06_dp, 2.438613808513792e-06_dp, &
             5.213214035322539e-06_dp, 1.0713700095419553e-05_dp, 2.1283136860042053e-05_dp, &
             4.104666176512367e-05_dp, 7.712256170413337e-05_dp, 0.00014157505436499074_dp, &
             0.0002545212186276719_dp, 0.0004490141548139169_dp, 0.0007786320645723325_dp, &
             0.0013291508421305036_dp, 0.0022363252683017592_dp, 0.0037127336305645057_dp, &
             0.006087967115884585_dp, 0.009868326679217646_dp, 0.01582484332892341_dp, &
             0.025122159864475007_dp, 0.03950600835943753_dp, 0.06157424130511713_dp, &
             0.09516638518906102_dp, 0.14592054779743022_dp, 0.2220657865284112_dp, &
             0.3355452449512651_dp, 0.5036051402007276_dp, 0.751047197659662_dp, &
             1.1134529186647788_dp, 1.6419177490725296_dp, 2.410391428838679_dp, &
             3.5282835138536948_dp, 5.166063146430546_dp, 7.62205675937331_dp, &
             11.585504470872786_dp, &
             ! w
             3.4859746803257895e-10_dp, 2.0392735350586325e-09_dp, 7.985636080747364e-09_dp, &
             2.5307826252775974e-08_dp, 7.003516072061651e-08_dp, 1.7600901424967307e-07_dp, &
             4.111770384039354e-07_dp, 9.06539741021209e-07_dp, 1.9062734141219702e-06_dp, &
             3.852623588770603e-06_dp, 7.526930208082628e-06_dp, 1.4279908151413794e-05_dp, &
             2.6401932127648204e-05_dp, 4.77102316600423e-05_dp, 8.446837321735595e-05_dp, &
             0.00014680952312298912_dp, 0.0002509162182496244_dp, 0.0004223261113245025_dp, &
             0.0007008998079551283_dp, 0.0011482233082181546_dp, 0.0018585509068544917_dp, &
             0.002974861829426026_dp, 0.004712255870787982_dp, 0.007391818275602717_dp, &
             0.011489334718914918_dp, 0.017704959303613124_dp, 0.027062305026851263_dp, &
             0.04104868349187665_dp, 0.06181274584595445_dp, 0.09244222631772768_dp, &
             0.13735421432130254_dp, 0.20284663484600096_dp, 0.29789112748955615_dp, &
             0.43531920405882435_dp, 0.6337398638085372_dp, 0.9210634008158828_dp, &
             1.3422220745264866_dp, 1.9801016088953018_dp, 3.030547498797203_dp, &
             5.2505365021420545_dp]

      CASE (1117)

         aw(:) = & ! a
            [9.245083727335525e-11_dp, 9.688910054778095e-10_dp, 5.014330784873288e-09_dp, &
             1.9115313268073576e-08_dp, 6.07246983173125e-08_dp, 1.703442370131078e-07_dp, &
             4.3596742513438636e-07_dp, 1.0391124268281379e-06_dp, 2.338755437123553e-06_dp, &
             5.020225421778518e-06_dp, 1.0353151269849264e-05_dp, 2.0629153350512764e-05_dp, &
             3.989099826787706e-05_dp, 7.512748536417518e-05_dp, 0.00013820269755403296_dp, &
             0.0002489292349096817_dp, 0.0004399038126001696_dp, 0.0007640301633495673_dp, &
             0.0013061008746573146_dp, 0.002200455970262447_dp, 0.0036576644608222447_dp, &
             0.006004500511750603_dp, 0.00974336671949372_dp, 0.01563997089186818_dp, &
             0.02485178760258314_dp, 0.03911503835969579_dp, 0.0610151597635596_dp, &
             0.0943757602887648_dp, 0.14481496149506373_dp, 0.22053734477253553_dp, &
             0.3334569940318453_dp, 0.5007869876899175_dp, 0.7472932962754667_dp, &
             1.1085218708245348_dp, 1.6355372634188903_dp, 2.402268828897588_dp, &
             3.5181220568113494_dp, 5.153580212943063_dp, 7.606990193365446_dp, &
             11.567553175334984_dp, &
             ! w
             2.8956253028397253e-10_dp, 1.819215266391565e-09_dp, 7.3252633875284604e-09_dp, &
             2.3590056305292376e-08_dp, 6.599355487819895e-08_dp, 1.6718116077860641e-07_dp, &
             3.929813686110841e-07_dp, 8.707577124152452e-07_dp, 1.8386134310794138e-06_dp, &
             3.7288941755779278e-06_dp, 7.30714428864103e-06_dp, 1.3899361081172519e-05_dp, &
             2.5757923476109033e-05_dp, 4.66426117102103e-05_dp, 8.273146721881567e-05_dp, &
             0.00014403219992693294_dp, 0.00024654590820028633_dp, 0.00041555144189804946_dp, &
             0.0006905452465472577_dp, 0.0011326078580167934_dp, 0.0018353016463963595_dp, &
             0.0029406722023037454_dp, 0.004662578586453995_dp, 0.007320484213534334_dp, &
             0.01138809349121848_dp, 0.0175629458340195_dp, 0.026865452108242857_dp, &
             0.04077912610937432_dp, 0.061448303295889656_dp, 0.09195610714635562_dp, &
             0.13671516867336328_dp, 0.20201987195016258_dp, 0.29684039311082117_dp, &
             0.4340103960422105_dp, 0.6321462934497492_dp, 0.9191720144488631_dp, &
             1.3400382424449495_dp, 1.977646643885012_dp, 3.0278328170340365_dp, &
             5.247421245649557_dp]

      CASE (1118)

         aw(:) = & ! a
            [8.314049740536352e-11_dp, 9.058760469822048e-10_dp, 4.77376975901067e-09_dp, &
             1.838147477084065e-08_dp, 5.877451595962851e-08_dp, 1.6564107721766646e-07_dp, &
             4.2542736395888067e-07_dp, 1.0168245308230877e-06_dp, 2.2938147796789015e-06_dp, &
             4.933156773343243e-06_dp, 1.0190139600934417e-05_dp, 2.0332924516525508e-05_dp, &
             3.936667083188025e-05_dp, 7.42209943561043e-05_dp, 0.00013666841380029126_dp, &
             0.0002463821001504863_dp, 0.0004357496034025712_dp, 0.0007573653034506079_dp, &
             0.0012955704606728044_dp, 0.002184055307295319_dp, 0.0036324654238494387_dp, &
             0.005966279555878983_dp, 0.009686106621706868_dp, 0.015555203862622113_dp, &
             0.02472774423827018_dp, 0.03893556664084016_dp, 0.060758383521210706_dp, &
             0.09401246118976118_dp, 0.14430669831900414_dp, 0.21983437627635727_dp, &
             0.33249615632857743_dp, 0.49948980231472234_dp, 0.745564751810983_dp, &
             1.1062505063871784_dp, 1.6325973101262354_dp, 2.398525060479348_dp, &
             3.513437335577366_dp, 5.1478239162734925_dp, 7.600041184323437_dp, &
             11.559272431772172_dp, &
             ! w
             2.646906866942325e-10_dp, 1.7239867932574688e-09_dp, 7.035118550773094e-09_dp, &
             2.2828814981786882e-08_dp, 6.419178187974133e-08_dp, 1.6322791148399693e-07_dp, &
             3.8480401334441106e-07_dp, 8.546299428201422e-07_dp, 1.80804297513248e-06_dp, &
             3.672873287956493e-06_dp, 7.207451406342112e-06_dp, 1.3726473253333003e-05_dp, &
             2.546492717879969e-05_dp, 4.6156274202480474e-05_dp, 8.193934038194782e-05_dp, &
             0.00014276426488163613_dp, 0.0002445488239900861_dp, 0.0004124529355072001_dp, &
             0.0006858055882912367_dp, 0.001125454740981814_dp, 0.0018246442056247649_dp, &
             0.002924989494599287_dp, 0.0046397778169269335_dp, 0.00728772469460034_dp, &
             0.011341574278837124_dp, 0.017497658988115766_dp, 0.026774910712818296_dp, &
             0.04065508811484694_dp, 0.06128053085411256_dp, 0.09173222839956267_dp, &
             0.13642074581877267_dp, 0.20163882264094019_dp, 0.2963559482402296_dp, &
             0.4334067697728587_dp, 0.6314111169756155_dp, 0.9182992185276677_dp, &
             1.3390302837988404_dp, 1.9765133765398153_dp, 3.026579575125085_dp, &
             5.245983126533738_dp]

      CASE (1119)

         aw(:) = & ! a
            [8.105036917128686e-11_dp, 8.916353918126276e-10_dp, 4.7190267069906185e-09_dp, &
             1.8213731835744166e-08_dp, 5.8327376876995756e-08_dp, 1.645603184165264e-07_dp, &
             4.2300112342242466e-07_dp, 1.0116868884019828e-06_dp, 2.283443477517136e-06_dp, &
             4.913043750153079e-06_dp, 1.0152452173548074e-05_dp, 2.0264388298285887e-05_dp, &
             3.92452837680488e-05_dp, 7.401101339962555e-05_dp, 0.00013631282837049745_dp, &
             0.00024579150332763356_dp, 0.0004347859730760208_dp, 0.0007558186939911443_dp, &
             0.0012931259643688805_dp, 0.0021802468594143472_dp, 0.0036266120961004223_dp, &
             0.005957398933814257_dp, 0.009672798752113888_dp, 0.015535498218662989_dp, &
             0.02469890141113428_dp, 0.038893826398058474_dp, 0.06069865207635582_dp, &
             0.09392793392500058_dp, 0.14418842113098004_dp, 0.21967076097336_dp, &
             0.3322724852250308_dp, 0.4991877864016212_dp, 0.7451622457727589_dp, &
             1.1057215271516696_dp, 1.6319125355768132_dp, 2.3976529602163463_dp, &
             3.5123459308547855_dp, 5.146482746883756_dp, 7.598422005925609_dp, &
             11.557342842902521_dp, &
             ! w
             2.5910215815768987e-10_dp, 1.7023473117535032e-09_dp, 6.968790345311862e-09_dp, &
             2.26542014335368e-08_dp, 6.377751905955557e-08_dp, 1.623173723922451e-07_dp, &
             3.829179119127959e-07_dp, 8.509058288835969e-07_dp, 1.8009770883024804e-06_dp, &
             3.659914315003767e-06_dp, 7.184373707254334e-06_dp, 1.3686427013509861e-05_dp, &
             2.5397022735876466e-05_dp, 4.6043505365669545e-05_dp, 8.175558450552733e-05_dp, &
             0.00014247001207718467_dp, 0.00024408518283574346_dp, 0.0004117333417667165_dp, &
             0.0006847045048475441_dp, 0.0011237924899204165_dp, 0.0018221669371960608_dp, &
             0.0029213431990782026_dp, 0.004634475275940993_dp, 0.007280104436267471_dp, &
             0.011330751054239988_dp, 0.017482466213982608_dp, 0.026753836995530602_dp, &
             0.04062621275651319_dp, 0.06124146760061116_dp, 0.09168009304196569_dp, &
             0.1363521718502062_dp, 0.20155005930086692_dp, 0.29624308367084384_dp, &
             0.4332661205863603_dp, 0.631239796098427_dp, 0.9180958077289131_dp, &
             1.3387953539315174_dp, 1.9762492265218348_dp, 3.0262874562406465_dp, &
             5.245647926948447_dp]

      CASE (1120)

         aw(:) = & ! a
            [1.6364137384434621e-10_dp, 1.3210377486271154e-09_dp, 5.841176464767646e-09_dp, &
             2.026080632715249e-08_dp, 6.041472374937316e-08_dp, 1.6185001549675903e-07_dp, &
             3.998368441888749e-07_dp, 9.264595243070832e-07_dp, 2.0373470623000663e-06_dp, &
             4.288697454214027e-06_dp, 8.697987981261276e-06_dp, 1.708153989761005e-05_dp, &
             3.261243858267086e-05_dp, 6.0728323076307486e-05_dp, 0.00011058765397849812_dp, &
             0.0001973758610700513_dp, 0.0003459143773026472_dp, 0.0005962473693796777_dp, &
             0.0010122030565147162_dp, 0.0016943909235004747_dp, 0.0027997613670785186_dp, &
             0.004570801293535262_dp, 0.00737877934310612_dp, 0.011787340505097048_dp, &
             0.018645390474000836_dp, 0.029221888997446103_dp, 0.045400274230782384_dp, &
             0.06995729131769945_dp, 0.10696071821412645_dp, 0.16233388334639812_dp, &
             0.24465345780070938_dp, 0.3662732491751197_dp, 0.5449053129458524_dp, &
             0.8058510442237334_dp, 1.1851852851318976_dp, 1.7344281107172446_dp, &
             2.527804393083869_dp, 3.6747708749408012_dp, 5.345594588084227_dp, &
             7.838323572595179_dp, 11.84278077937327_dp, &
             ! w
             4.661586414112498e-10_dp, 2.2071904993484876e-09_dp, 7.842428587543773e-09_dp, &
             2.3472210508842365e-08_dp, 6.24437139345481e-08_dp, 1.5238245126579308e-07_dp, &
             3.478738057185164e-07_dp, 7.527694421132782e-07_dp, 1.5584826437792473e-06_dp, &
             3.1083643595302936e-06_dp, 6.003937668050791e-06_dp, 1.1277343727454584e-05_dp, &
             2.066715437902478e-05_dp, 3.705364026187104e-05_dp, 6.5137343685239e-05_dp, &
             0.00011248518204743692_dp, 0.0001911265907578534_dp, 0.0003199657632613632_dp, &
             0.0005283950291749178_dp, 0.0008616628070338599_dp, 0.0013887865161319691_dp, &
             0.0022141324149550163_dp, 0.003494245933736811_dp, 0.005462155552577028_dp, &
             0.008462255215462058_dp, 0.013000081745253118_dp, 0.019812962852473352_dp, &
             0.029969781075366728_dp, 0.04501121380589444_dp, 0.0671461362956841_dp, &
             0.09952607060488151_dp, 0.1466290177979314_dp, 0.2148001059484958_dp, &
             0.3130282725568901_dp, 0.4541112600230295_dp, 0.6565510790438718_dp, &
             0.9480658799901298_dp, 1.3733326062444926_dp, 2.0150222098293096_dp, &
             3.0691362783428544_dp, 5.294835469191273_dp]

      CASE (1121)

         aw(:) = & ! a
            [1.0077830992332914e-10_dp, 9.20122346669601e-10_dp, 4.4223948088085445e-09_dp, &
             1.615509671021659e-08_dp, 4.989748896385042e-08_dp, 1.37164262461527e-07_dp, &
             3.4566007610125516e-07_dp, 8.13795909153356e-07_dp, 1.8132606938181964e-06_dp, &
             3.859471481118592e-06_dp, 7.902093035736414e-06_dp, 1.5647085569197518e-05_dp, &
             3.0091385355972e-05_dp, 5.639642458921698e-05_dp, 0.00010329460389139393_dp, &
             0.00018532410144612415_dp, 0.0003263371025718005_dp, 0.0005649460165880816_dp, &
             0.0009628915875550053_dp, 0.0016177781193824183_dp, 0.002682282981534022_dp, &
             0.004392891408715603_dp, 0.007112545407594812_dp, 0.011393481130361175_dp, &
             0.0180691725708096_dp, 0.028387987232039877_dp, 0.04420628084531695_dp, &
             0.06826574331723989_dp, 0.10458958450300211_dp, 0.159045638835404_dp, &
             0.24014318746889385_dp, 0.3601567220071993_dp, 0.5367086843262416_dp, &
             0.7950047759958014_dp, 1.1710258634465203_dp, 1.7162117331911668_dp, &
             2.504735979884491_dp, 3.6460477026464178_dp, 5.310453314963071_dp, &
             7.796053043477965_dp, 11.792552497938217_dp, &
             ! w
             2.992281355932222e-10_dp, 1.6350488259739786e-09_dp, 6.221617930663424e-09_dp, &
             1.939126520186215e-08_dp, 5.3054601381100064e-08_dp, 1.3221520583809728e-07_dp, &
             3.06847509466454e-07_dp, 6.729387945836895e-07_dp, 1.4088368052337008e-06_dp, &
             2.8366819017337755e-06_dp, 5.524268233810102e-06_dp, 1.0451096801831577e-05_dp, &
             1.9274984784513243e-05_dp, 3.475427961456303e-05_dp, 6.140818172447243e-05_dp, &
             0.00010653769479052191_dp, 0.00018178760349040973_dp, 0.0003055130021794588_dp, &
             0.0005063324762600125_dp, 0.0008284181385067849_dp, 0.0013393096200459633_dp, &
             0.002141371105079261_dp, 0.0033884733109008007_dp, 0.005310124314353085_dp, &
             0.008246159600509327_dp, 0.012696320823873695_dp, 0.019390724083776085_dp, &
             0.02938950687949788_dp, 0.04422308491329402_dp, 0.0660888141440765_dp, &
             0.09812609690570866_dp, 0.1448014744538381_dp, 0.21245136929109476_dp, &
             0.3100618666542363_dp, 0.4504377136815248_dp, 0.6521019010479792_dp, &
             0.9428096046855696_dp, 1.3672863455656539_dp, 2.0082430342211115_dp, &
             3.061648751241772_dp, 5.286237610420216_dp]

      CASE (1122)

         aw(:) = & ! a
            [7.846157541054408e-11_dp, 7.737193081496072e-10_dp, 3.883615926892372e-09_dp, &
             1.4551039365540695e-08_dp, 4.5704749991175906e-08_dp, 1.2717353189330714e-07_dp, &
             3.234733879615845e-07_dp, 7.672141645629378e-07_dp, 1.7198709793056789e-06_dp, &
             3.6793781773241907e-06_dp, 7.56620688444088e-06_dp, 1.503862896814775e-05_dp, &
             2.901721079342104e-05_dp, 5.454327489954248e-05_dp, 0.000100163442658953_dp, &
             0.00018013295982322172_dp, 0.0003178793524299607_dp, 0.0005513863979727495_dp, &
             0.0009414764985133457_dp, 0.001584429453923439_dp, 0.0026310360839014793_dp, &
             0.0043151275598890355_dp, 0.006995958011963441_dp, 0.011220703217310593_dp, &
             0.017815982899270923_dp, 0.028021005869974868_dp, 0.043680066956110035_dp, &
             0.06751922484084998_dp, 0.10354179167794113_dp, 0.157590795617473_dp, &
             0.23814535618063415_dp, 0.35744441949108036_dp, 0.5330702113591295_dp, &
             0.7901854222813905_dp, 1.1647286011080535_dp, 1.7081032869109365_dp, &
             2.4944598079189606_dp, 3.633243612759346_dp, 5.2947787129555905_dp, &
             7.7771889702802985_dp, 11.770128154981425_dp, &
             ! w
             2.3980859641845776e-10_dp, 1.4200030799207322e-09_dp, 5.587868750138464e-09_dp, &
             1.7758965730621705e-08_dp, 4.923929088288768e-08_dp, 1.239210090031837e-07_dp, &
             2.8981208697283986e-07_dp, 6.395281683452343e-07_dp, 1.345789972458573e-06_dp, &
             2.721566525627505e-06_dp, 5.3200161923181364e-06_dp, 1.0097725892015597e-05_dp, &
             1.8677260675906158e-05_dp, 3.376360664441058e-05_dp, 5.979640826717305e-05_dp, &
             0.0001039597545595159_dp, 0.00017772896651963526_dp, 0.00029921676529703586_dp, &
             0.0004966995953831652_dp, 0.0008138728389760328_dp, 0.0013176205216000878_dp, &
             0.002109417274168953_dp, 0.0033419436587054977_dp, 0.005243139063836892_dp, &
             0.008150804885174239_dp, 0.012562093237366368_dp, 0.019203893202511212_dp, &
             0.029132423331587283_dp, 0.04387349335169384_dp, 0.0656192792247561_dp, &
             0.09750372057267062_dp, 0.14398817553994187_dp, 0.2114051014752916_dp, &
             0.30873922991086195_dp, 0.4487983729648066_dp, 0.6501148753728081_dp, &
             0.9404605202212373_dp, 1.364582712712996_dp, 2.00521050630963_dp, &
             3.0582987798040926_dp, 5.282391239948746_dp]

      CASE (1123)

         aw(:) = & ! a
            [6.736055010410299e-11_dp, 6.996060022595122e-10_dp, 3.6049826700700847e-09_dp, &
             1.3709320287733097e-08_dp, 4.3482234232362585e-08_dp, 1.2183768932790463e-07_dp, &
             3.115545624426518e-07_dp, 7.420720599734724e-07_dp, 1.66926758535568e-06_dp, &
             3.5814714777227945e-06_dp, 7.383084558625586e-06_dp, 1.4706079846854891e-05_dp, &
             2.842883935320496e-05_dp, 5.352624372223149e-05_dp, 9.844200897659978e-05_dp, &
             0.00017727447098699503_dp, 0.0003132153838979451_dp, 0.0005438991477460063_dp, &
             0.0009296372851136106_dp, 0.0015659721071756118_dp, 0.0026026431153605245_dp, &
             0.004272001283224277_dp, 0.006931242439755254_dp, 0.011124716031428426_dp, &
             0.01767521093264663_dp, 0.027816813845121375_dp, 0.04338707015647836_dp, &
             0.06710328577560774_dp, 0.10295762436319943_dp, 0.15677920703661025_dp, &
             0.23703023370344276_dp, 0.3559296982351572_dp, 0.5310372374298867_dp, &
             0.7874913648324828_dp, 1.161206820889257_dp, 1.703566726235335_dp, &
             2.4887082715715545_dp, 3.6260747952576873_dp, 5.286000172652444_dp, &
             7.766621598639381_dp, 11.757563945957756_dp, &
             ! w
             2.101968290573295e-10_dp, 1.309346847655089e-09_dp, 5.25519463781895e-09_dp, &
             1.689236995846742e-08_dp, 4.719772077382954e-08_dp, 1.194562708469918e-07_dp, &
             2.805985049513223e-07_dp, 6.213878708108684e-07_dp, 1.311447038076214e-06_dp, &
             2.6586857803801947e-06_dp, 5.208174901247038e-06_dp, 9.903819769918184e-06_dp, &
             1.8348648406913883e-05_dp, 3.3218036226099946e-05_dp, 5.890743024509087e-05_dp, &
             0.00010253589827897028_dp, 0.00017548442246896934_dp, 0.0002957306728195197_dp, &
             0.0004913602867259009_dp, 0.0008058025634577054_dp, 0.0013055753250753412_dp, &
             0.002091655955182345_dp, 0.003316059263228482_dp, 0.0052058465369273655_dp, &
             0.00809767978677353_dp, 0.012487259632443633_dp, 0.01909966510690306_dp, &
             0.028988915044160915_dp, 0.043678231828519355_dp, 0.06535687904669692_dp, &
             0.09715572156087553_dp, 0.14353319532603007_dp, 0.21081951471297253_dp, &
             0.30799863092399754_dp, 0.44788005636834544_dp, 0.6490013733053659_dp, &
             0.9391436904449139_dp, 1.363066727729098_dp, 2.0035097860710507_dp, &
             3.056419871367975_dp, 5.2802340131162335_dp]

      CASE (1124)

         aw(:) = & ! a
            [6.127539113848334e-11_dp, 6.585042222099449e-10_dp, 3.4484300709106907e-09_dp, &
             1.3232310461756044e-08_dp, 4.221524661655175e-08_dp, 1.1878261333646666e-07_dp, &
             3.0470726126448926e-07_dp, 7.275887381212208e-07_dp, 1.640051433625022e-06_dp, &
             3.5248369097909654e-06_dp, 7.276983263179379e-06_dp, 1.451312555434058e-05_dp, &
             2.8087020216229502e-05_dp, 5.293472915369195e-05_dp, 9.743980005645125e-05_dp, &
             0.0001756087604184208_dp, 0.0003104953254321021_dp, 0.0005395292218883107_dp, &
             0.0009227225233688956_dp, 0.0015551850449507135_dp, 0.0025860394521180752_dp, &
             0.0042467678555054185_dp, 0.006893357364449355_dp, 0.011068497013180502_dp, &
             0.01759272425742476_dp, 0.02769711437908127_dp, 0.04321524317747791_dp, &
             0.06685926694291575_dp, 0.10261478796306317_dp, 0.1563027396605973_dp, &
             0.2363753571565395_dp, 0.35503987925415303_dp, 0.5298426291172311_dp, &
             0.7859078646593438_dp, 1.1591362807959065_dp, 1.7008989439479196_dp, &
             2.48532527824902_dp, 3.6218573578541493_dp, 5.2808348646138255_dp, &
             7.760402871458198_dp, 11.750169294218498_dp, &
             ! w
             1.939421355848777e-10_dp, 1.247350919732893e-09_dp, 5.06662983529617e-09_dp, &
             1.6397924067160886e-08_dp, 4.6027544095260826e-08_dp, 1.1688834865531014e-07_dp, &
             2.7528475149811317e-07_dp, 6.109023831250881e-07_dp, 1.291558794658838e-06_dp, &
             2.6222126096896627e-06_dp, 5.143212365885017e-06_dp, 9.791052260931775e-06_dp, &
             1.8157333630852843e-05_dp, 3.290010084104736e-05_dp, 5.838891512931703e-05_dp, &
             0.00010170473910935861_dp, 0.00017417323697941343_dp, 0.0002936928463253703_dp, &
             0.0004882372055734686_dp, 0.0008010793567578624_dp, 0.001298521973689501_dp, &
             0.002081250179672232_dp, 0.003300887324287589_dp, 0.0051839781704531174_dp, &
             0.008066514250513147_dp, 0.012443341717747034_dp, 0.019038473649737864_dp, &
             0.028904632915824957_dp, 0.043563516891769176_dp, 0.0652026717087234_dp, &
             0.09695114785726025_dp, 0.14326565532908955_dp, 0.21047508118590955_dp, &
             0.3075629099559849_dp, 0.44733964894025197_dp, 0.64834596163264_dp, &
             0.9383684522042456_dp, 1.3621741063854518_dp, 2.002508284243861_dp, &
             3.0553133870543427_dp, 5.278963662175973_dp]

      CASE (1125)

         aw(:) = & ! a
            [5.821799210923045e-11_dp, 6.377036543776304e-10_dp, 3.368593265829029e-09_dp, &
             1.2987842519843727e-08_dp, 4.1563709332938546e-08_dp, 1.1720764865959467e-07_dp, &
             3.0117050781399276e-07_dp, 7.200962636596814e-07_dp, 1.624918075680271e-06_dp, &
             3.4954696759506704e-06_dp, 7.221914495413273e-06_dp, 1.441289723807534e-05_dp, &
             2.790933879240479e-05_dp, 5.2627058016903764e-05_dp, 9.691821271758948e-05_dp, &
             0.00017474141498668443_dp, 0.0003090783104521796_dp, 0.0005372517310495454_dp, &
             0.0009191173084452_dp, 0.001549558842442514_dp, 0.00257737656285679_dp, &
             0.004233598287722606_dp, 0.006873578984081914_dp, 0.011039139116332876_dp, &
             0.01754963815437458_dp, 0.02763457541627272_dp, 0.04312544887891895_dp, &
             0.06673171880890857_dp, 0.10243555169606595_dp, 0.15605359272998934_dp, &
             0.23603285704252502_dp, 0.3545744243765163_dp, 0.5292176407238489_dp, &
             0.7850792913680315_dp, 1.1580527066089683_dp, 1.6995026298751894_dp, &
             2.4835544082696717_dp, 3.6196494474972107_dp, 5.278130470390086_dp, &
             7.7571466843542805_dp, 11.746297140839042_dp, &
             ! w
             1.857675063474109e-10_dp, 1.215785716737123e-09_dp, 4.969980721408779e-09_dp, &
             1.6143537549502136e-08_dp, 4.542392365844714e-08_dp, 1.1556110718197491e-07_dp, &
             2.725340434942317e-07_dp, 6.054675818751546e-07_dp, 1.2812393921431467e-06_dp, &
             2.6032705371236807e-06_dp, 5.1094478778164505e-06_dp, 9.73240029063973e-06_dp, &
             1.8057766778338127e-05_dp, 3.273454482500827e-05_dp, 5.811877811137271e-05_dp, &
             0.00010127152381208313_dp, 0.00017348953975596007_dp, 0.00029262984833733574_dp, &
             0.00048660752939977045_dp, 0.0007986139043604258_dp, 0.0012948390987612738_dp, &
             0.002075815312285157_dp, 0.0032929610182812277_dp, 0.005172550588710973_dp, &
             0.008050224496665403_dp, 0.012420381403604831_dp, 0.019006476013628593_dp, &
             0.028860552182275415_dp, 0.04350350806636042_dp, 0.06512198946429831_dp, &
             0.09684409538723417_dp, 0.14312563023334907_dp, 0.21029478390104492_dp, &
             0.3073347943631397_dp, 0.44705668818426597_dp, 0.6480027419673723_dp, &
             0.9379624396131295_dp, 1.3617065768668402_dp, 2.001983694900147_dp, &
             3.0547337922970343_dp, 5.278298241750151_dp]

      CASE (1126)

         aw(:) = & ! a
            [5.755879823092453e-11_dp, 6.332046703734606e-10_dp, 3.3512688144233827e-09_dp, &
             1.2934682373119975e-08_dp, 4.142182880836962e-08_dp, 1.1686432105031416e-07_dp, &
             3.003989037464654e-07_dp, 7.184605850141597e-07_dp, 1.6216125406255824e-06_dp, &
             3.489052162841711e-06_dp, 7.2098758262520215e-06_dp, 1.4390978724193739e-05_dp, &
             2.7870470743602508e-05_dp, 5.255973660346555e-05_dp, 9.680405706570152e-05_dp, &
             0.00017455154495484018_dp, 0.0003087680515499572_dp, 0.0005367529790402491_dp, &
             0.0009183276649055452_dp, 0.0015483263573175146_dp, 0.002575478587183788_dp, &
             0.004230712551323576_dp, 0.006869244581091624_dp, 0.011032704635727027_dp, &
             0.01754019379139732_dp, 0.02762086564460487_dp, 0.04310576231832612_dp, &
             0.0667037525487406_dp, 0.10239624892073004_dp, 0.1559989555950486_dp, &
             0.23595774211151196_dp, 0.3544723364163357_dp, 0.5290805530078792_dp, &
             0.7848975367559231_dp, 1.1578150010565404_dp, 1.6991963008781352_dp, &
             2.483165887734602_dp, 3.619165019970507_dp, 5.277537087570941_dp, &
             7.756432205734298_dp, 11.745447481727286_dp, &
             ! w
             1.8400423042461785e-10_dp, 1.2089406877943307e-09_dp, 4.948963339797113e-09_dp, &
             1.6088130595999037e-08_dp, 4.529230680715883e-08_dp, 1.1527146806489142e-07_dp, &
             2.7193337222538263e-07_dp, 6.042801494537686e-07_dp, 1.2789837281255483e-06_dp, &
             2.599128504377676e-06_dp, 5.102062200854228e-06_dp, 9.719566961180106e-06_dp, &
             1.8035975438585433e-05_dp, 3.2698302600811254e-05_dp, 5.805962945051483e-05_dp, &
             0.00010117664980280834_dp, 0.0001733397842304803_dp, 0.0002923969743745046_dp, &
             0.00048625045893160304_dp, 0.0007980736370510741_dp, 0.001294031948557977_dp, &
             0.0020746240483449562_dp, 0.003291223464490599_dp, 0.005170045245666498_dp, &
             0.008046652836119885_dp, 0.01241534670040915_dp, 0.01899945900262857_dp, &
             0.028850884568196365_dp, 0.043490346122490676_dp, 0.06510429181364769_dp, &
             0.09682061174367765_dp, 0.1430949114342032_dp, 0.21025522761150742_dp, &
             0.30728474394210875_dp, 0.44699460076419695_dp, 0.6479274285893785_dp, &
             0.9378733433677441_dp, 1.3616039772992576_dp, 2.0018685702482233_dp, &
             3.0546065942337584_dp, 5.278152208062264_dp]

      CASE (1127)

         aw(:) = & ! a
            [1.504533796769453e-10_dp, 1.1531074953405321e-09_dp, 4.8888187124083735e-09_dp, &
             1.64913263438988e-08_dp, 4.82281745320224e-08_dp, 1.2735042553338537e-07_dp, &
             3.110894054663479e-07_dp, 7.143105683543449e-07_dp, 1.5590485153608082e-06_dp, &
             3.2610487264232e-06_dp, 6.57770699653839e-06_dp, 1.2856153509666872e-05_dp, &
             2.4442207132353652e-05_dp, 4.5344210115474104e-05_dp, 8.229565780832036e-05_dp, &
             0.00014643467321744517_dp, 0.0002559275716031921_dp, 0.00044002331987520266_dp, &
             0.000745257122854606_dp, 0.0012448547035166385_dp, 0.0020528658296695953_dp, &
             0.0033452339853937848_dp, 0.0053909681941535555_dp, 0.008597931626136578_dp, &
             0.013579648107021822_dp, 0.021252153578070875_dp, 0.032973558110725165_dp, &
             0.05074400507394785_dp, 0.07749061957506888_dp, 0.1174715178598124_dp, &
             0.17684597768695903_dp, 0.264475911062782_dp, 0.3930492819191733_dp, &
             0.5806538052951022_dp, 0.8529897882443226_dp, 1.2465211951108859_dp, &
             1.8130974497943952_dp, 2.62714991076857_dp, 3.7981600722550914_dp, &
             5.4962274554572_dp, 8.019188692235735_dp, 12.057386630438975_dp, &
             ! w
             4.2193209155725965e-10_dp, 1.868421442363326e-09_dp, 6.398083919571562e-09_dp, &
             1.8727899270213683e-08_dp, 4.90502922225399e-08_dp, 1.1829298380009044e-07_dp, &
             2.6753835248527113e-07_dp, 5.745234884851598e-07_dp, 1.181866972572199e-06_dp, &
             2.344371560782497e-06_dp, 4.50686817269553e-06_dp, 8.430264019191465e-06_dp, &
             1.5392695342439878e-05_dp, 2.7506328214873737e-05_dp, 4.8210326072618356e-05_dp, &
             8.302952009905674e-05_dp, 0.00014073015924529557_dp, 0.00023506402780935405_dp, &
             0.0003873774517226005_dp, 0.0006304822293962166_dp, 0.001014354946006773_dp, &
             0.0016144680116281528_dp, 0.0025438852570893893_dp, 0.003970711003739061_dp, &
             0.00614310995746551_dp, 0.009424977561012782_dp, 0.014346519338652794_dp, &
             0.021675605497769666_dp, 0.032517958983232106_dp, 0.048457237329562496_dp, &
             0.07175024148830268_dp, 0.10559849024163727_dp, 0.1545266611196524_dp, &
             0.2249144111603384_dp, 0.3257601636970398_dp, 0.46982968031426914_dp, &
             0.6755350077022726_dp, 0.9704388241513623_dp, 1.3990172699260266_dp, &
             2.0437808601951684_dp, 3.1008805126561834_dp, 5.331299844531794_dp]

      CASE (1128)

         aw(:) = & ! a
            [9.044089458167161e-11_dp, 7.760490006718084e-10_dp, 3.5852241336987436e-09_dp, &
             1.2785918280557138e-08_dp, 3.885739038534123e-08_dp, 1.0556377548653211e-07_dp, &
             2.6362329365754025e-07_dp, 6.16169999839811e-07_dp, 1.3647525613704547e-06_dp, &
             2.890290524653775e-06_dp, 5.892358609652293e-06_dp, 1.1624091192301527e-05_dp, &
             2.2281389262796003e-05_dp, 4.1637611128927765e-05_dp, 7.60638010614533e-05_dp, &
             0.0001361473161713841_dp, 0.0002392292240702679_dp, 0.000413338322662384_dp, &
             0.0007032285563654285_dp, 0.0011795570256516283_dp, 0.0019527135441420882_dp, &
             0.0031934877115869895_dp, 0.00516371555401082_dp, 0.008261394409421483_dp, &
             0.013086641419359156_dp, 0.020537501144712905_dp, 0.0319482574310672_dp, &
             0.04928794762318627_dp, 0.075443721507337_dp, 0.11462323037774363_dp, &
             0.17292335791129965_dp, 0.2591307884929151_dp, 0.38584560228432824_dp, &
             0.5710570093866103_dp, 0.8403610648529733_dp, 1.23012045204557_dp, &
             1.792099359670144_dp, 2.6006763190123077_dp, 3.765327463511408_dp, &
             5.456196605212914_dp, 7.9711746431386885_dp, 12.00046365729151_dp, &
             ! w
             2.6278378503237663e-10_dp, 1.3391746092584678e-09_dp, 4.936308615909203e-09_dp, &
             1.5100553792094086e-08_dp, 4.0787113989337976e-08_dp, 1.0067433172943575e-07_dp, &
             2.3190014168695236e-07_dp, 5.054891717619687e-07_dp, 1.0529270453645942e-06_dp, &
             2.110965019070496e-06_dp, 4.095750860647717e-06_dp, 7.723443816473219e-06_dp, &
             1.4203534920755094e-05_dp, 2.5544515812331387e-05_dp, 4.503124360806736e-05_dp, &
             7.796204090378747e-05_dp, 0.00013277503138524942_dp, 0.0002227527089083657_dp, &
             0.0003685785445475256_dp, 0.0006021395740165651_dp, 0.0009721380957671119_dp, &
             0.0015523125515765094_dp, 0.002453397758842146_dp, 0.0038404122697334613_dp, &
             0.005957492042884414_dp, 0.009163357139122481_dp, 0.013981685679326982_dp, &
             0.021172290552966412_dp, 0.03183120668304537_dp, 0.04753082814504019_dp, &
             0.07051544704190335_dp, 0.10397360028499768_dp, 0.1524179408004863_dp, &
             0.22221937993089996_dp, 0.3223742400855906_dp, 0.4656570762873271_dp, &
             0.6705038371587703_dp, 0.9645180053399289_dp, 1.3922279294523348_dp, &
             2.0361850872427834_dp, 3.092499183883131_dp, 5.321670259310118_dp]

      CASE (1129)

         aw(:) = & ! a
            [6.881173825328228e-11_dp, 6.364256825799476e-10_dp, 3.0824462290764994e-09_dp, &
             1.1311705960135936e-08_dp, 3.504442940095042e-08_dp, 9.654719833516433e-08_dp, &
             2.4371527929356525e-07_dp, 5.745596794209253e-07_dp, 1.2816258757479816e-06_dp, &
             2.7304438152078568e-06_dp, 5.594913921514471e-06_dp, 1.1086255520485083e-05_dp, &
             2.133326226202394e-05_dp, 4.000374460754579e-05_dp, 7.330543477184387e-05_dp, &
             0.00013157681669769035_dp, 0.00023178507220642596_dp, 0.0004014048422558821_dp, &
             0.0006843792526008033_dp, 0.001150193708924114_dp, 0.001907565200672008_dp, &
             0.0031249232867409577_dp, 0.0050608135721216425_dp, 0.008108700126489182_dp, &
             0.012862529994572014_dp, 0.020212054983542893_dp, 0.03148056055540311_dp, &
             0.04862270474587203_dp, 0.07450713225703838_dp, 0.11331810135604625_dp, &
             0.17112353314624698_dp, 0.25667514341424347_dp, 0.38253210122761966_dp, &
             0.5666376896830858_dp, 0.8345392720353884_dp, 1.222552126941264_dp, &
             1.7824004295652038_dp, 2.5884377617471768_dp, 3.750137484521136_dp, &
             5.43766397303759_dp, 7.948933714776576_dp, 11.974084247772131_dp, &
             ! w
             2.0527502282012817e-10_dp, 1.1373966330849592e-09_dp, 4.354181517054271e-09_dp, &
             1.3618821403927912e-08_dp, 3.735120132005803e-08_dp, 9.32482588300979e-08_dp, &
             2.1671508545175283e-07_dp, 4.7580976772531757e-07_dp, 9.970722003870901e-07_dp, &
             2.009197838396392e-06_dp, 3.915482380907003e-06_dp, 7.411962397202441e-06_dp, &
             1.3677158792718827e-05_dp, 2.4672650004352243e-05_dp, 4.361327873908445e-05_dp, &
             7.569433348145283e-05_dp, 0.0001292043318463938_dp, 0.00021721133229393022_dp, &
             0.00036009530617037454_dp, 0.0005893190569577606_dp, 0.000952999242436352_dp, &
             0.00152407601665399_dp, 0.0024122101754994463_dp, 0.0037809949840906716_dp, &
             0.005872702643858514_dp, 0.009043655306095556_dp, 0.013814502064009433_dp, &
             0.020941310911667343_dp, 0.031515606482521934_dp, 0.047104529580163465_dp, &
             0.06994652582668863_dp, 0.1032240480401869_dp, 0.15144409029024405_dp, &
             0.2209734111079275_dp, 0.32080725936161336_dp, 0.46372419113246266_dp, &
             0.6681712238899777_dp, 0.9617708566736385_dp, 1.389075888676877_dp, &
             2.032657165550234_dp, 3.088605679386896_dp, 5.317197375968195_dp]

      CASE (1130)

         aw(:) = & ! a
            [5.762373767788402e-11_dp, 5.629107211167953e-10_dp, 2.811377347251901e-09_dp, &
             1.0503383546400876e-08_dp, 3.292872025179952e-08_dp, 9.149952681487838e-08_dp, &
             2.3249276531468532e-07_dp, 5.509711519840702e-07_dp, 1.2342818569905515e-06_dp, &
             2.6390439790362314e-06_dp, 5.424255353985955e-06_dp, 1.0776752565502037e-05_dp, &
             2.0786215444323352e-05_dp, 3.905882734189739e-05_dp, 7.170681379795403e-05_dp, &
             0.00012892290216883612_dp, 0.00022745500767569078_dp, 0.00039445238047264725_dp, &
             0.0006733814992680403_dp, 0.0011330382366301264_dp, 0.0018811541443374616_dp, &
             0.003084767193026275_dp, 0.00500048106464862_dp, 0.008019082133498577_dp, &
             0.012730869956791837_dp, 0.020020690224051913_dp, 0.031205316721924403_dp, &
             0.0482308879450105_dp, 0.07395507683783181_dp, 0.11254826214212527_dp, &
             0.17006116892388293_dp, 0.25522473622348213_dp, 0.3805738076803729_dp, &
             0.5640243369852944_dp, 0.8310946871129161_dp, 1.218071872742981_dp, &
             1.776656175990301_dp, 2.5811862357137687_dp, 3.7411336804669477_dp, &
             5.42667509490555_dp, 7.935742278921181_dp, 11.958434690487499_dp, &
             ! w
             1.7547680610057865e-10_dp, 1.029262673268843e-09_dp, 4.034852590918621e-09_dp, &
             1.2795085152759469e-08_dp, 3.542322398839987e-08_dp, 8.90517327428308e-08_dp, &
             2.080854390793545e-07_dp, 4.5886473846068296e-07_dp, 9.650581740496518e-07_dp, &
             1.950673309912423e-06_dp, 3.811511368288111e-06_dp, 7.231852895076139e-06_dp, &
             1.3372095480342655e-05_dp, 2.4166323036365803e-05_dp, 4.2788287375899815e-05_dp, &
             7.43727293858309e-05_dp, 0.00012712014690236992_dp, 0.0002139722970962429_dp, &
             0.00035513020579731023_dp, 0.0005818063011564883_dp, 0.0009417712842132405_dp, &
             0.001507493319028869_dp, 0.002387997647814448_dp, 0.0037460334311680814_dp, &
             0.005822768207706197_dp, 0.008973101753806037_dp, 0.013715885020269089_dp, &
             0.02080496111996761_dp, 0.03132917282248027_dp, 0.046852534403805435_dp, &
             0.06961000817705658_dp, 0.10278041710909547_dp, 0.15086737212038878_dp, &
             0.22023513795728097_dp, 0.3198782963588352_dp, 0.4625777544219739_dp, &
             0.6667870934207869_dp, 0.9601401256634218_dp, 1.3872042326406353_dp, &
             2.0305618662066713_dp, 3.086293033875503_dp, 5.3145407411872245_dp]

      CASE (1131)

         aw(:) = & ! a
            [5.089521599084929e-11_dp, 5.181216769631e-10_dp, 2.6435859823586346e-09_dp, &
             9.997572176326138e-09_dp, 3.159475878675708e-08_dp, 8.82990666770166e-08_dp, &
             2.2534605843279803e-07_dp, 5.358965687113875e-07_dp, 1.2039375677991995e-06_dp, &
             2.580318023464704e-06_dp, 5.314370903218271e-06_dp, 1.0577098193041535e-05_dp, &
             2.043274762989286e-05_dp, 3.8447388286763005e-05_dp, 7.06710167383026e-05_dp, &
             0.0001272013039563542_dp, 0.00022464304939132193_dp, 0.0003899329563819428_dp, &
             0.0006662259327764216_dp, 0.0011218668215417723_dp, 0.0018639421876792204_dp, &
             0.003058578645844583_dp, 0.004961107398406696_dp, 0.007960559286633173_dp, &
             0.012644841325021102_dp, 0.01989557952346713_dp, 0.03102527234977699_dp, &
             0.04797446218368935_dp, 0.07359361194659655_dp, 0.11204397559573867_dp, &
             0.16936496816638244_dp, 0.2542738576039283_dp, 0.3792894741313159_dp, &
             0.5623097717260181_dp, 0.82883400272175_dp, 1.2151305440363507_dp, &
             1.772883907686348_dp, 2.5764228514217984_dp, 3.7352178290350966_dp, &
             5.419453455750257_dp, 7.92707165151411_dp, 11.948146926479781_dp, &
             ! w
             1.5753152478296264e-10_dp, 9.625783116042642e-10_dp, 3.834979234259651e-09_dp, &
             1.2275120803701542e-08_dp, 3.41990576126701e-08_dp, 8.63752632552421e-08_dp, &
             2.02562111731916e-07_dp, 4.479877210264443e-07_dp, 9.444582371991111e-07_dp, &
             1.912936169902662e-06_dp, 3.744348365349662e-06_dp, 7.115320493878184e-06_dp, &
             1.3174436965123362e-05_dp, 2.3837843202007873e-05_dp, 4.2252458510012674e-05_dp, &
             7.35134545875828e-05_dp, 0.00012576376409151433_dp, 0.00021186248607131292_dp, &
             0.00035189345954491947_dp, 0.0005769050486782259_dp, 0.0009344411284773096_dp, &
             0.0014966602341250804_dp, 0.002372170461358687_dp, 0.003723166662028694_dp, &
             0.005790090605055332_dp, 0.008926907103906643_dp, 0.013651284572048878_dp, &
             0.02071560222094273_dp, 0.03120693732150626_dp, 0.04668724475406355_dp, &
             0.0693891909163507_dp, 0.10248920425314961_dp, 0.15048866105954362_dp, &
             0.21975017441404748_dp, 0.31926787513071525_dp, 0.4618242065611505_dp, &
             0.6658770637201433_dp, 0.959067708983269_dp, 1.3859731414100134_dp, &
             2.0291834900622616_dp, 3.084771589112346_dp, 5.312793052529753_dp]

      CASE (1132)

         aw(:) = & ! a
            [4.655839914813421e-11_dp, 4.889698221313238e-10_dp, 2.5331445046805452e-09_dp, &
             9.662149649143539e-09_dp, 3.070559126027258e-08_dp, 8.615763974696274e-08_dp, &
             2.2055006470124057e-07_dp, 5.257562795657575e-07_dp, 1.1834855313394206e-06_dp, &
             2.540670836491189e-06_dp, 5.240079217912787e-06_dp, 1.0441945397265341e-05_dp, &
             2.019320959107207e-05_dp, 3.803262149157397e-05_dp, 6.996776906459235e-05_dp, &
             0.00012603150354540974_dp, 0.00022273097791536772_dp, 0.0003868578019120708_dp, &
             0.0006613540895256288_dp, 0.001114256505479876_dp, 0.0018522107248818784_dp, &
             0.003040720146173731_dp, 0.004934245482550136_dp, 0.00792061620950118_dp, &
             0.01258610149879416_dp, 0.019810122582362363_dp, 0.030902249361405374_dp, &
             0.047799189761202086_dp, 0.07334646495025084_dp, 0.11169907294526635_dp, &
             0.1688886723984487_dp, 0.25362315290944704_dp, 0.3784103561757928_dp, &
             0.5611358805980543_dp, 0.8272858555420428_dp, 1.2131158539936244_dp, &
             1.770299547351948_dp, 2.573158892461925_dp, 3.7311635210453504_dp, &
             5.414503555528428_dp, 7.921127878404948_dp, 11.941093935556074_dp, &
             ! w
             1.459519286039505e-10_dp, 9.187961403763315e-10_dp, 3.7024140041922947e-09_dp, &
             1.1928270145762675e-08_dp, 3.337919500387715e-08_dp, 8.457733572715138e-08_dp, &
             1.9884291740602865e-07_dp, 4.4064921654876637e-07_dp, 9.305370405032079e-07_dp, &
             1.8873980159757483e-06_dp, 3.6988411515984825e-06_dp, 7.036277648517482e-06_dp, &
             1.3040239213795488e-05_dp, 2.361463536991283e-05_dp, 4.188807226539795e-05_dp, &
             7.292870124178898e-05_dp, 0.00012484012632902355_dp, 0.0002104249483473028_dp, &
             0.0003496868724815439_dp, 0.0005735620280894835_dp, 0.0009294390606530346_dp, &
             0.001489264531929705_dp, 0.0023613608564222316_dp, 0.003707543140745257_dp, &
             0.005767755781304221_dp, 0.0088953226513683_dp, 0.013607101242358957_dp, &
             0.02065446656495418_dp, 0.031123284379568673_dp, 0.04657409584345657_dp, &
             0.06923799057155386_dp, 0.10228975157703117_dp, 0.1502292182896008_dp, &
             0.21941786579199918_dp, 0.3188495100516651_dp, 0.4613076433366639_dp, &
             0.6652531174287252_dp, 0.9583323081025146_dp, 1.3851288222950382_dp, &
             2.0282380741255848_dp, 3.0837280019645634_dp, 5.31159430763875_dp]

      CASE (1133)

         aw(:) = & ! a
            [4.372239454152232e-11_dp, 4.697671392539843e-10_dp, 2.459813915175251e-09_dp, &
             9.438274748140831e-09_dp, 3.0110000452788115e-08_dp, 8.471947271280103e-08_dp, &
             2.173225428041362e-07_dp, 5.18921061235575e-07_dp, 1.1696808260313077e-06_dp, &
             2.5138791240331714e-06_dp, 5.189826891395849e-06_dp, 1.0350446890300948e-05_dp, &
             2.003091936508714e-05_dp, 3.775142187796361e-05_dp, 6.949069916674238e-05_dp, &
             0.00012523749838643484_dp, 0.00022143250720734305_dp, 0.0003847685389857988_dp, &
             0.0006580427654987771_dp, 0.0011090818747393008_dp, 0.001844231055670228_dp, &
             0.003028568842131297_dp, 0.004915962358070604_dp, 0.007893421665313708_dp, &
             0.012546098556564696_dp, 0.019751909774421636_dp, 0.030818426409970907_dp, &
             0.04767973875169655_dp, 0.0731779937420558_dp, 0.11146391699799875_dp, &
             0.1685638690187674_dp, 0.2531793320495862_dp, 0.3778106387808482_dp, &
             0.5603349425450836_dp, 0.8262294008796733_dp, 1.2117408304091315_dp, &
             1.7685354857072824_dp, 2.570930666710737_dp, 3.7283954359788978_dp, &
             5.4111236777482645_dp, 7.917069040928632_dp, 11.936277336374067_dp, &
             ! w
             1.3837271096309108e-10_dp, 8.897755906779012e-10_dp, 3.6139251181519142e-09_dp, &
             1.1695820444441597e-08_dp, 3.282822673894336e-08_dp, 8.336656541627032e-08_dp, &
             1.963341880560622e-07_dp, 4.35692456418518e-07_dp, 9.211233931964809e-07_dp, &
             1.8701121634323854e-06_dp, 3.6680130998339836e-06_dp, 6.982691988128684e-06_dp, &
             1.2949202704007787e-05_dp, 2.3463127696214187e-05_dp, 4.16406052383363e-05_dp, &
             7.253138405150644e-05_dp, 0.0001242122736899331_dp, 0.0002094473707311157_dp, &
             0.00034818575204545503_dp, 0.0005712870135694963_dp, 0.0009260339210774173_dp, &
             0.0014842284163924088_dp, 0.0023539979595969696_dp, 0.003696898451172256_dp, &
             0.005752534714690219_dp, 0.008873792935425322_dp, 0.013576976715839338_dp, &
             0.020612775004942234_dp, 0.031066225671774996_dp, 0.04649690354837202_dp, &
             0.06913482014328017_dp, 0.1021536329927305_dp, 0.1500521296989211_dp, &
             0.21919100554349588_dp, 0.3185638588147285_dp, 0.4609548960985716_dp, &
             0.664826988125453_dp, 0.9578300054461429_dp, 1.384552074354948_dp, &
             2.0275922284381322_dp, 3.0830150729932084_dp, 5.310775395194252_dp]

      CASE (1134)

         aw(:) = & ! a
            [4.196916904268143e-11_dp, 4.5783472123597864e-10_dp, 2.41399857284035e-09_dp, &
             9.29791109918092e-09_dp, 2.973568347957108e-08_dp, 8.38140179703636e-08_dp, &
             2.152877572862184e-07_dp, 5.146070917978946e-07_dp, 1.1609602620872351e-06_dp, &
             2.496941605069929e-06_dp, 5.158036877724326e-06_dp, 1.0292531092268829e-05_dp, &
             1.9928142718505796e-05_dp, 3.7573261224508415e-05_dp, 6.918831840510557e-05_dp, &
             0.00012473405135775677_dp, 0.00022060892548471232_dp, 0.00038344297861418275_dp, &
             0.0006559412656486614_dp, 0.001105796998364909_dp, 0.0018391643187098841_dp, &
             0.0030208515858810023_dp, 0.004904348391726479_dp, 0.00787614355411739_dp, &
             0.012520677971531575_dp, 0.019714911066134683_dp, 0.030765141888046618_dp, &
             0.0476037946361518_dp, 0.07307086831794665_dp, 0.11131436850591339_dp, &
             0.16835728216060383_dp, 0.25289701117266594_dp, 0.37742910565713156_dp, &
             0.5598253393857198_dp, 0.8255571538365273_dp, 1.210865785733112_dp, &
             1.7674127620344793_dp, 2.5695124126420223_dp, 3.7266334345293606_dp, &
             5.408972107272138_dp, 7.914485119175988_dp, 11.93321088148437_dp, &
             ! w
             1.3368401267913223e-10_dp, 8.716647770404902e-10_dp, 3.5584412392256325e-09_dp, &
             1.1549681242661814e-08_dp, 3.2481194665526896e-08_dp, 8.260288707227677e-08_dp, &
             1.9475009743785357e-07_dp, 4.325597872882583e-07_dp, 9.151694842648144e-07_dp, &
             1.859172219422295e-06_dp, 3.6484916025828558e-06_dp, 6.948742843301368e-06_dp, &
             1.2891501458472161e-05_dp, 2.3367060772298477e-05_dp, 4.148363771368323e-05_dp, &
             7.227928621271513e-05_dp, 0.00012381378416486148_dp, 0.00020882674815508886_dp, &
             0.00034723251707318326_dp, 0.0005698420104246942_dp, 0.0009238706409401791_dp, &
             0.00148102833828441_dp, 0.0023493185060400558_dp, 0.003690132077069927_dp, &
             0.005742857726751972_dp, 0.008860102997741797_dp, 0.01355781882146208_dp, &
             0.020586257253336994_dp, 0.031029928866411937_dp, 0.04644779292006761_dp, &
             0.06906917400549814_dp, 0.10206701237630725_dp, 0.14993942503239616_dp, &
             0.219046609681775_dp, 0.3183820249021022_dp, 0.46073033119928253_dp, &
             0.6645556845480565_dp, 0.9575101816141885_dp, 1.3841848287751488_dp, &
             2.027180968220947_dp, 3.082561087569542_dp, 5.310253926156578_dp]

      CASE (1135)

         aw(:) = & ! a
            [4.1127165009997967e-11_dp, 4.5208631863548186e-10_dp, 2.3918566820888186e-09_dp, &
             9.229936702536655e-09_dp, 2.9554159088576507e-08_dp, 8.337447035045251e-08_dp, &
             2.142992013689694e-07_dp, 5.125099115642874e-07_dp, 1.1567186454703263e-06_dp, &
             2.488699668053342e-06_dp, 5.142561697888398e-06_dp, 1.0264328683802903e-05_dp, &
             1.987808043390075e-05_dp, 3.74864569460461e-05_dp, 6.904095656648396e-05_dp, &
             0.00012448865027706285_dp, 0.00022020740016839407_dp, 0.00038279660713390665_dp, &
             0.0006549163636929012_dp, 0.001104194723962221_dp, 0.0018366925581560902_dp, &
             0.003017086308797732_dp, 0.0048986812140641624_dp, 0.007867711538061789_dp, &
             0.012508270977330033_dp, 0.01969685135727524_dp, 0.03073913035713723_dp, &
             0.04756671825396208_dp, 0.07301856465958849_dp, 0.11124134613871574_dp, &
             0.16825640122175003_dp, 0.2527591378391157_dp, 0.37724276880047053_dp, &
             0.5595764385890466_dp, 0.825228794655026_dp, 1.2104383459243773_dp, &
             1.7668643078198798_dp, 2.56881955756283_dp, 3.725772612498451_dp, &
             5.407920922554632_dp, 7.913222662323158_dp, 11.931712629557587_dp, &
             ! w
             1.3143126672878202e-10_dp, 8.629177951457761e-10_dp, 3.531570877967938e-09_dp, &
             1.1478797513811431e-08_dp, 3.2312688613887935e-08_dp, 8.223177286556583e-08_dp, &
             1.9397980590007926e-07_dp, 4.310356755677059e-07_dp, 9.122715091381536e-07_dp, &
             1.853845375244453e-06_dp, 3.6389831719940646e-06_dp, 6.932202357501217e-06_dp, &
             1.2863381514108447e-05_dp, 2.3320233207470337e-05_dp, 4.1407108656970464e-05_dp, &
             7.2156353789366e-05_dp, 0.00012361943258451088_dp, 0.00020852401038785965_dp, &
             0.00034676746491693047_dp, 0.0005691369463581905_dp, 0.0009228149747480054_dp, &
             0.0014794665401182345_dp, 0.0023470344500257597_dp, 0.003686829050845064_dp, &
             0.0057381334206000854_dp, 0.008853418961766443_dp, 0.013548464281687904_dp, &
             0.020573307939622178_dp, 0.031012202811739747_dp, 0.04642380728805961_dp, &
             0.06903711018081043_dp, 0.10202470106976738_dp, 0.14988436902844351_dp, &
             0.21897606833248978_dp, 0.3182931889721274_dp, 0.460620612987102_dp, &
             0.6644231242659094_dp, 0.9573539074420375_dp, 1.3840053768626657_dp, &
             2.0269800038927603_dp, 3.0823392426401535_dp, 5.309999105325884_dp]

      CASE (1136)

         aw(:) = & ! a
            [4.1042231238385694e-11_dp, 4.5150581518412125e-10_dp, 2.389618110441018e-09_dp, &
             9.223059339947427e-09_dp, 2.953578396074567e-08_dp, 8.33299598979284e-08_dp, &
             2.1419906731112746e-07_dp, 5.122974323031782e-07_dp, 1.1562888172343277e-06_dp, &
             2.4878643293756714e-06_dp, 5.140993037585987e-06_dp, 1.0261469569084476e-05_dp, &
             1.987300466262588e-05_dp, 3.747765510632244e-05_dp, 6.90260130050227e-05_dp, &
             0.00012446376291921877_dp, 0.0002201666766304586_dp, 0.000382731046607637_dp, &
             0.000654812403274749_dp, 0.001104032189261055_dp, 0.001836441810203125_dp, &
             0.0030167043221156747_dp, 0.004898106254849544_dp, 0.007866856039358602_dp, &
             0.012507012135573328_dp, 0.01969501891226128_dp, 0.030736490983514164_dp, &
             0.04756295601613861_dp, 0.07301325710891807_dp, 0.11123393592919985_dp, &
             0.16824616368318404_dp, 0.2527451458982139_dp, 0.3772238581214361_dp, &
             0.5595511779250323_dp, 0.825195469112096_dp, 1.2103949636647597_dp, &
             1.7668086422937717_dp, 2.5687492346830636_dp, 3.7256852398990254_dp, &
             5.407814226706766_dp, 7.913094520736236_dp, 11.931560553320608_dp, &
             ! w
             1.3120399239960082e-10_dp, 8.620336734661605e-10_dp, 3.5288522175428966e-09_dp, &
             1.1471621686348695e-08_dp, 3.229562346824973e-08_dp, 8.219417792281092e-08_dp, &
             1.9390175497733175e-07_dp, 4.3088121336615584e-07_dp, 9.119777652275903e-07_dp, &
             1.8533053638262234e-06_dp, 3.6380191370926036e-06_dp, 6.9305251879707836e-06_dp, &
             1.2860529951376536e-05_dp, 2.3315484169088386e-05_dp, 4.139934685353226e-05_dp, &
             7.214388477946586e-05_dp, 0.00012359971833064958_dp, 0.00020849330013087046_dp, &
             0.0003467202867342341_dp, 0.00056906541620143_dp, 0.0009227078704740659_dp, &
             0.0014793080787531686_dp, 0.0023468026986220897_dp, 0.003686493897354143_dp, &
             0.005737654035057934_dp, 0.008852740695821823_dp, 0.013547514995340426_dp, &
             0.020571993821374316_dp, 0.031010403890985017_dp, 0.04642137305143006_dp, &
             0.06903385602688432_dp, 0.10202040679502959_dp, 0.14987878113375322_dp, &
             0.21896890859689308_dp, 0.31828417219614324_dp, 0.4606094764361201_dp, &
             0.6644096689252987_dp, 0.9573380447451075_dp, 1.3839871612209713_dp, &
             2.026959604444003_dp, 3.0823167237718936_dp, 5.3099732398171104_dp]

      CASE (1137)

         aw(:) = & ! a
            [3.10545537311889e-11_dp, 3.346340554085941e-10_dp, 1.7545763471630164e-09_dp, &
             6.737276684474854e-09_dp, 2.150360182657336e-08_dp, 6.052469332341586e-08_dp, &
             1.552984431174646e-07_dp, 3.708967673002915e-07_dp, 8.361653185378072e-07_dp, &
             1.7973400337252573e-06_dp, 3.7109985039504026e-06_dp, 7.401880146578872e-06_dp, &
             1.4325945593652393e-05_dp, 2.7001673439224997e-05_dp, 4.970674008905796e-05_dp, &
             8.958826230235682e-05_dp, 0.00015841037064703037_dp, 0.0002752739083577955_dp, &
             0.0004708048663513996_dp, 0.0007935422996163246_dp, 0.0013195929390722794_dp, &
             0.0021671001535556907_dp, 0.0035177567986392736_dp, 0.005648553754494596_dp, &
             0.008978306572985588_dp, 0.014135388686909328_dp, 0.022055720661463402_dp, &
             0.03412369112852952_dp, 0.052373681691766094_dp, 0.07977673101625049_dp, &
             0.12064628356553592_dp, 0.18120989022276213_dp, 0.2704116150070408_dp, &
             0.40103519608638083_dp, 0.5912754586331583_dp, 0.8669457852098937_dp, &
             1.2646196593875323_dp, 1.8362382787010891_dp, 2.656289385731902_dp, &
             3.8342594017837275_dp, 5.540199405785691_dp, 8.071887986996131_dp, &
             12.11982472014633_dp, &
             ! w
             9.840264038207425e-11_dp, 6.344570985208832e-10_dp, 2.5794668330224168e-09_dp, &
             8.352680359622475e-09_dp, 2.345345379836522e-08_dp, 5.957602882478861e-08_dp, &
             1.403358366707645e-07_dp, 3.1147804383129733e-07_dp, 6.586079729526731e-07_dp, &
             1.3373004962829441e-06_dp, 2.6232342304396444e-06_dp, 4.994225642311579e-06_dp, &
             9.262383645075728e-06_dp, 1.6784033106693948e-05_dp, 2.9788937507394094e-05_dp, &
             5.18906213431086e-05_dp, 8.886896205263378e-05_dp, 0.00014985846489258854_dp, &
             0.0002491359898728472_dp, 0.00040878738166126485_dp, 0.0006626533594396859_dp, &
             0.0010621246763700659_dp, 0.001684593335556114_dp, 0.0026456954355616224_dp, &
             0.004116935598120336_dp, 0.006350908057508755_dp, 0.009717186007342433_dp, &
             0.014753120253674189_dp, 0.02223538118512112_dp, 0.03328024790854573_dp, &
             0.049483618823955945_dp, 0.07311584793214533_dp, 0.10739246451482792_dp, &
             0.15685104574566022_dp, 0.22788050781156885_dp, 0.32948124607777635_dp, &
             0.47440913290516173_dp, 0.6810499766380471_dp, 0.9769220613937061_dp, &
             1.4064451222486212_dp, 2.0520860311054285_dp, 3.1100421959705855_dp, &
             5.3418276718834585_dp]

      CASE (1138)

         aw(:) = & ! a
            [2.9379946937066885e-11_dp, 3.2320896344761557e-10_dp, 1.710600328606682e-09_dp, &
             6.602296956973786e-09_dp, 2.1143094678081747e-08_dp, 5.965148020884071e-08_dp, &
             1.5333370425464905e-07_dp, 3.667264473986085e-07_dp, 8.277255817544218e-07_dp, &
             1.7809295729577364e-06_dp, 3.6801631126905677e-06_dp, 7.345639550504187e-06_dp, &
             1.4226025622753398e-05_dp, 2.6828257264985585e-05_dp, 4.941204643947861e-05_dp, &
             8.909697901638683e-05_dp, 0.0001576055978224511_dp, 0.00027397677208900655_dp, &
             0.0004687453241805737_dp, 0.0007903178431903183_dp, 0.0013146108812577577_dp, &
             0.0021594979468122708_dp, 0.003506293363601258_dp, 0.005631463127937535_dp, &
             0.008953103391180694_dp, 0.014098613264398261_dp, 0.022002609997768464_dp, &
             0.03404776156560792_dp, 0.05226620993209978_dp, 0.07962612398356159_dp, &
             0.12043733588243713_dp, 0.18092294329834657_dp, 0.27002165696499053_dp, &
             0.40051098041258104_dp, 0.5905787738311084_dp, 0.8660310759118847_dp, &
             1.2634342679299708_dp, 1.8347236084714764_dp, 2.654383207143485_dp, &
             3.831899193378691_dp, 5.5373258050536185_dp, 8.068445375165187_dp, &
             12.115747170747774_dp, &
             ! w
             9.392194892318008e-11_dp, 6.170839924286718e-10_dp, 2.526117352831834e-09_dp, &
             8.211922994890131e-09_dp, 2.3118716811540165e-08_dp, 5.8838434394543414e-08_dp, &
             1.3880393939685158e-07_dp, 3.0844491247662714e-07_dp, 6.528363114262537e-07_dp, &
             1.3266826053076486e-06_dp, 2.6042641434195615e-06_dp, 4.961193869489435e-06_dp, &
             9.206168451436119e-06_dp, 1.6690313349969386e-05_dp, 2.9635587508434175e-05_dp, &
             5.1643964757712046e-05_dp, 8.847845543404601e-05_dp, 0.00014924924916106822_dp, &
             0.0002481985867932157_dp, 0.0004073636224097017_dp, 0.0006605174183376301_dp, &
             0.0010589578546577512_dp, 0.001679951018487681_dp, 0.002638964472664498_dp, &
             0.004107280296349304_dp, 0.006337203133073932_dp, 0.009697935375440784_dp, &
             0.014726361685292882_dp, 0.02219857897159231_dp, 0.03323017922433627_dp, &
             0.049416265081960814_dp, 0.07302631079514021_dp, 0.10727493796018868_dp, &
             0.15669889052534391_dp, 0.22768649056170318_dp, 0.3292380151173698_dp, &
             0.47410998998016085_dp, 0.6806899375074603_dp, 0.9764990293686424_dp, &
             1.4059606569400547_dp, 2.0515445017069798_dp, 3.109444895405208_dp, &
             5.341141253466358_dp]

      CASE (1139)

         aw(:) = & ! a
            [2.5805537426128395e-11_dp, 2.6406325197349135e-10_dp, 1.3507894738968888e-09_dp, &
             5.115870470539107e-09_dp, 1.6182759856989415e-08_dp, 4.525738309601122e-08_dp, &
             1.1556022754932956e-07_dp, 2.749273709828951e-07_dp, 6.178556887429758e-07_dp, &
             1.3245813970072105e-06_dp, 2.728736755231757e-06_dp, 5.432086783030678e-06_dp, &
             1.0495573936508608e-05_dp, 1.9752246553631183e-05_dp, 3.631230245788907e-05_dp, &
             6.536730604008212e-05_dp, 0.00011545515448355912_dp, 0.00020042740637144532_dp, &
             0.0003424771841065342_dp, 0.0005767544971012292_dp, 0.0009583380771902763_dp, &
             0.0015726785852723665_dp, 0.002551118615796569_dp, 0.004093786973327022_dp, &
             0.006503135718149682_dp, 0.010232743990415958_dp, 0.01595789572043291_dp, &
             0.024677042459505517_dp, 0.037856848642687665_dp, 0.05763843655350069_dp, &
             0.08712917903177249_dp, 0.13081358543228738_dp, 0.1951294242999519_dp, &
             0.2892726485035215_dp, 0.42631936650166596_dp, 0.6247898385811417_dp, &
             0.9108391834696583_dp, 1.3213697343223358_dp, 1.9085957303934402_dp, &
             2.747169495915087_dp, 3.9465867173599785_dp, 5.676749574460742_dp, &
             8.235266633116058_dp, 12.313137637726658_dp, &
             ! w
             8.00399863396705e-11_dp, 4.915251541355609e-10_dp, 1.9620551367438805e-09_dp, &
             6.2871713631879725e-09_dp, 1.7529479476054223e-08_dp, 4.4297934733352e-08_dp, &
             1.0392926100669545e-07_dp, 2.2992971986952449e-07_dp, 4.848812709009576e-07_dp, &
             9.82330661925483e-07_dp, 1.9231954467389874e-06_dp, 3.6552753821897653e-06_dp, &
             6.769037631888964e-06_dp, 1.2249645092187482e-05_dp, 2.1715205753471587e-05_dp, &
             3.778582813092096e-05_dp, 6.464922259182427e-05_dp, 0.0001089190037601165_dp, &
             0.00018092528741374622_dp, 0.0002966388177395374_dp, 0.0004805168139765504_dp, &
             0.0007696806955460541_dp, 0.001220006019266378_dp, 0.001914941187212991_dp, &
             0.002978197840805819_dp, 0.0045919046894181475_dp, 0.007022431181136046_dp, &
             0.010656925964886156_dp, 0.016054755456341857_dp, 0.02401957575343736_dp, &
             0.03569987399641648_dp, 0.05272869049457537_dp, 0.07741723596715053_dp, &
             0.11302291966439043_dp, 0.16412137506311983_dp, 0.23712803558036213_dp, &
             0.3410472856916833_dp, 0.48860286424216137_dp, 0.698099292412088_dp, &
             0.9969198390552333_dp, 1.429315216662463_dp, 2.0776254656572872_dp, &
             3.1382001146497958_dp, 5.374195315582562_dp]

      CASE (1140)

         aw(:) = & ! a
            [2.1111243580084996e-11_dp, 2.3224491017554419e-10_dp, 1.2291683218745666e-09_dp, &
             4.7441439887589486e-09_dp, 1.519257405943804e-08_dp, 4.286314488841236e-08_dp, &
             1.1017940807788809e-07_dp, 2.635148162029348e-07_dp, 5.9477017832161e-07_dp, &
             1.2797040745453973e-06_dp, 2.6444166021018683e-06_dp, 5.2782799523408055e-06_dp, &
             1.0222247542299618e-05_dp, 1.9277702300756025e-05_dp, 3.5505501229057255e-05_dp, &
             6.402149123184618e-05_dp, 0.00011324901820679112_dp, 0.0001968686447577509_dp, &
             0.00033682146117394897_dp, 0.0005678904875798317_dp, 0.0009446262924561243_dp, &
             0.0015517280179759845_dp, 0.0025194808110439865_dp, 0.0040465419548871145_dp, &
             0.006433338391767066_dp, 0.01013069352623149_dp, 0.015810184605045314_dp, &
             0.024465337502942924_dp, 0.03755635503141191_dp, 0.05721600089228407_dp, &
             0.08654100248424336_dp, 0.13000254970793015_dp, 0.1940220897840922_dp, &
             0.28777607587357007_dp, 0.4243180518620017_dp, 0.622143244212191_dp, &
             0.9073805817689972_dp, 1.3169073103665563_dp, 1.902916975492621_dp, &
             2.740049599784258_dp, 3.9378004595057625_dp, 5.666083276187741_dp, &
             8.222519349247909_dp, 12.298068641427866_dp, &
             ! w
             6.748852008911724e-11_dp, 4.4341163915523165e-10_dp, 1.8151659238252474e-09_dp, &
             5.900756259714016e-09_dp, 1.6612176342286474e-08_dp, 4.2278922987722326e-08_dp, &
             9.973890576048785e-08_dp, 2.216360586555885e-07_dp, 4.69101810871081e-07_dp, &
             9.533005466400577e-07_dp, 1.8713190490457361e-06_dp, 3.564913573353679e-06_dp, &
             6.615180888514306e-06_dp, 1.1992985189735323e-05_dp, 2.1294936446995888e-05_dp, &
             3.7109267600874355e-05_dp, 6.357704515763241e-05_dp, 0.0001072444834913886_dp, &
             0.00017834548173518552_dp, 0.00029271504882903205_dp, 0.00047462114318726244_dp, &
             0.0007609243484303307_dp, 0.0012071449484875507_dp, 0.0018962532470387734_dp, &
             0.0029513255368186966_dp, 0.004553657400690409_dp, 0.006968541513819855_dp, &
             0.010581757063146131_dp, 0.01595096052204758_dp, 0.023877718930558203_dp, &
             0.03550803481637499_dp, 0.05247209436887822_dp, 0.07707798312611479_dp, &
             0.11257992616613019_dp, 0.16355069094257252_dp, 0.236403757928437_dp, &
             0.34014331201786646_dp, 0.48749567521889686_dp, 0.6967717046242862_dp, &
             0.9953650608175667_dp, 1.4275393309821744_dp, 2.0756440017949287_dp, &
             3.1360163129443808_dp, 5.371684431915519_dp]

      CASE (1141)

         aw(:) = & ! a
            [2.2372988225912986e-11_dp, 2.155311024643934e-10_dp, 1.0683399847572335e-09_dp, &
             3.9739825349021844e-09_dp, 1.2422863442524694e-08_dp, 3.444807931563984e-08_dp, &
             8.739122593186021e-08_dp, 2.068427418199086e-07_dp, 4.628897235545483e-07_dp, &
             9.888579907017114e-07_dp, 2.0309834216653173e-06_dp, 4.032513775733553e-06_dp, &
             7.773545358434298e-06_dp, 1.459978113798453e-05_dp, 2.6791302797117927e-05_dp, &
             4.814917379398309e-05_dp, 8.491722460683491e-05_dp, 0.00014721437400774806_dp, &
             0.00025123728604219104_dp, 0.0004226152253907819_dp, 0.0007014742221237098_dp, &
             0.0011500176701666695_dp, 0.0018637850253244044_dp, 0.0029882501950574264_dp, &
             0.004743122360075808_dp, 0.007457690327954169_dp, 0.01162191009540732_dp, &
             0.01795981321943177_dp, 0.0275343982411992_dp, 0.041896710661668196_dp, &
             0.0632966577684987_dp, 0.09497970184804179_dp, 0.1416025612739261_dp, &
             0.20981332955217188_dp, 0.30905839614943154_dp, 0.45270264706401014_dp, &
             0.6595855013522461_dp, 0.9561945344912952_dp, 1.379748852234404_dp, &
             1.982721994895184_dp, 2.839917905447953_dp, 4.060833137310887_dp, &
             5.815221529842535_dp, 8.400534530235255_dp, 12.508298773674282_dp, &
             ! w
             6.776665965861506e-11_dp, 3.918804188974133e-10_dp, 1.527368063806098e-09_dp, &
             4.827323088099226e-09_dp, 1.3334082786027613e-08_dp, 3.3464770762276443e-08_dp, &
             7.809435517474159e-08_dp, 1.7203020780136926e-07_dp, 3.6148795940862717e-07_dp, &
             7.301353394892479e-07_dp, 1.425736223288309e-06_dp, 2.703646134304637e-06_dp, &
             4.99673497200199e-06_dp, 9.026245725497744e-06_dp, 1.597535232451133e-05_dp, &
             2.775776001522263e-05_dp, 4.7429005640715064e-05_dp, 7.981012027905184e-05_dp, &
             0.00013242472618964552_dp, 0.0002168951901198007_dp, 0.0003510062684536178_dp, &
             0.0005617327479209903_dp, 0.0008896504066323469_dp, 0.0013953200440996291_dp, &
             0.0021684696795750535_dp, 0.003341123556732734_dp, 0.005106269458631146_dp, &
             0.007744259638564807_dp, 0.01165996096307984_dp, 0.017434781143235333_dp, &
             0.025899201421230506_dp, 0.038233483671874485_dp, 0.05610700360381818_dp, &
             0.08187054481901408_dp, 0.11882155063596944_dp, 0.17157128053823664_dp, &
             0.24655878355217412_dp, 0.35278941428064936_dp, 0.5029523565279985_dp, &
             0.7152701616532843_dp, 1.0169935591848016_dp, 1.4522110109881194_dp, &
             2.103146700225878_dp, 3.1663155716664577_dp, 5.406530821702442_dp]

      CASE (1142)

         aw(:) = & ! a
            [1.590531594341684e-11_dp, 1.7214376693616277e-10_dp, 9.044087524731547e-10_dp, &
             3.476622052672392e-09_dp, 1.110441929902987e-08_dp, 3.127083500900928e-08_dp, &
             8.026802071879804e-08_dp, 1.917617219326265e-07_dp, 4.3242417699210244e-07_dp, &
             9.296921646417572e-07_dp, 1.919894389123156e-06_dp, 3.829973643341747e-06_dp, &
             7.4137151691578905e-06_dp, 1.3975115501859041e-05_dp, 2.572921796685171e-05_dp, &
             4.6377182839343526e-05_dp, 8.201152607407125e-05_dp, 0.00014252499046154098_dp, &
             0.000243780363664504_dp, 0.0004109198863804765_dp, 0.0006833672580321727_dp, &
             0.0011223242275402755_dp, 0.001821917139636936_dp, 0.0029256475083544466_dp, &
             0.004650500664676603_dp, 0.007322043973169291_dp, 0.011425204379043083_dp, &
             0.01767729380019354_dp, 0.02713243830590747_dp, 0.041330116239260084_dp, &
             0.06250536211815287_dp, 0.09388480855834652_dp, 0.14010175302256056_dp, &
             0.20777575943757826_dp, 0.3063193383321254_dp, 0.4490584205879896_dp, &
             0.6547895039284874_dp, 0.9499556102932464_dp, 1.3717335769178085_dp, &
             1.9725625611697524_dp, 2.8272266775493593_dp, 4.045222830935186_dp, &
             5.7963249341851135_dp, 8.378004991685499_dp, 12.481716620987017_dp, &
             ! w
             5.0492941975379845e-11_dp, 3.268684389604195e-10_dp, 1.3308885023039828e-09_dp, &
             4.3132288952946936e-09_dp, 1.211793381251175e-08_dp, 3.079447379730047e-08_dp, &
             7.256189422819233e-08_dp, 1.610938516283908e-07_dp, 3.406988441231785e-07_dp, &
             6.919114419463749e-07_dp, 1.3574550584674181e-06_dp, 2.584727945903791e-06_dp, &
             4.794251786129428e-06_dp, 8.68840717399617e-06_dp, 1.5421973280739284e-05_dp, &
             2.6866492696977943e-05_dp, 4.601570151623402e-05_dp, 7.760115605664705e-05_dp, &
             0.00012901846589119607_dp, 0.00021170894721520657_dp, 0.00034320417163573053_dp, &
             0.0005501288119703177_dp, 0.0008725799040271002_dp, 0.0013704710479607478_dp, &
             0.002132665562364094_dp, 0.0032900465466762585_dp, 0.005034115507046455_dp, &
             0.007643316492220549_dp, 0.011520105464033497_dp, 0.017242899858765075_dp, &
             0.02563855277474636_dp, 0.03788304255132047_dp, 0.05564086444252724_dp, &
             0.08125750998569554_dp, 0.11802509999886159_dp, 0.17055020234244153_dp, &
             0.2452688293886317_dp, 0.3511863756356012_dp, 0.5009968467533765_dp, &
             0.7129339452525164_dp, 1.0142662291940763_dp, 1.4491037748599194_dp, &
             2.099685850421519_dp, 3.162504202594817_dp, 5.40214639440978_dp]

      CASE (1143)

         aw(:) = & ! a
            [1.5225259437665765e-11_dp, 1.6749313167827746e-10_dp, 8.864661145809686e-10_dp, &
             3.421437743950288e-09_dp, 1.0956760016736019e-08_dp, 3.0912549364976967e-08_dp, &
             7.94604874025738e-08_dp, 1.900447275345126e-07_dp, 4.289433800948182e-07_dp, &
             9.22912090194094e-07_dp, 1.9071315730773318e-06_dp, 3.8066522325071943e-06_dp, &
             7.372201167341749e-06_dp, 1.3902920987800809e-05_dp, 2.560627662512332e-05_dp, &
             4.617177488940133e-05_dp, 8.167426375702172e-05_dp, 0.00014198005265164286_dp, &
             0.00024291287557809692_dp, 0.0004095579607967174_dp, 0.0006812567302045216_dp, &
             0.0011190935095500695_dp, 0.0018170288756849953_dp, 0.0029183328232831466_dp, &
             0.004639670770038319_dp, 0.007306172756936884_dp, 0.011402174427105598_dp, &
             0.017644197214806537_dp, 0.027085322931629918_dp, 0.04126366785420287_dp, &
             0.062412514369447246_dp, 0.09375627578927048_dp, 0.13992548787529538_dp, &
             0.20753634890227643_dp, 0.305997370775741_dp, 0.44862988431163076_dp, &
             0.6542253160446062_dp, 0.9492214205096632_dp, 1.370790033591107_dp, &
             1.9713662396413763_dp, 2.8257317984924697_dp, 4.043383646883342_dp, &
             5.794098067299848_dp, 8.375349506181639_dp, 12.47858300048998_dp, &
             ! w
             4.8672172849902926e-11_dp, 3.197848414370642e-10_dp, 1.3090827586866968e-09_dp, &
             4.2555770852212076e-09_dp, 1.198056569937119e-08_dp, 3.0491213403272416e-08_dp, &
             7.193088204567227e-08_dp, 1.5984210950418523e-07_dp, 3.383123809725665e-07_dp, &
             6.875125364292405e-07_dp, 1.3495799532157325e-06_dp, 2.5709864604590263e-06_dp, &
             4.770814291005097e-06_dp, 8.64924269614402e-06_dp, 1.535773365886325e-05_dp, &
             2.67628995974332e-05_dp, 4.585124367229309e-05_dp, 7.734384180062181e-05_dp, &
             0.00012862129785617453_dp, 0.00021110369024164206_dp, 0.00034229287145775915_dp, &
             0.0005487723914995019_dp, 0.0008705830206509073_dp, 0.0013675622509684727_dp, &
             0.0021284716597606288_dp, 0.003284060007185324_dp, 0.005025653737391251_dp, &
             0.007631472021827377_dp, 0.011503686557388615_dp, 0.01722036202252093_dp, &
             0.025607923280691152_dp, 0.037841842791074394_dp, 0.055586039054075276_dp, &
             0.0811853777273192_dp, 0.11793134947802157_dp, 0.17042996575507047_dp, &
             0.24511687720710656_dp, 0.35099747928443814_dp, 0.5007663434779905_dp, &
             0.7126584872624464_dp, 1.0139445753847953_dp, 1.4487372424708918_dp, &
             2.0992775499510032_dp, 3.1620545242536195_dp, 5.401629127223172_dp]

      CASE (1144)

         aw(:) = & ! a
            [1.9889665315999944e-11_dp, 1.80272869653762e-10_dp, 8.626178841896934e-10_dp, &
             3.142836551674726e-09_dp, 9.689973889134477e-09_dp, 2.6602784362871437e-08_dp, &
             6.697416579015003e-08_dp, 1.5755492817893477e-07_dp, 3.5082945536793354e-07_dp, &
             7.463246025867492e-07_dp, 1.5273559538067252e-06_dp, 3.023127121582227e-06_dp, &
             5.811807611893505e-06_dp, 1.0888900153813021e-05_dp, 1.993828379633436e-05_dp, &
             3.576281629102979e-05_dp, 6.2960197183316e-05_dp, 0.0001089720401914669_dp, &
             0.00018569556563489997_dp, 0.0003119361030773096_dp, 0.0005171059742431752_dp, &
             0.0008467569378957519_dp, 0.0013707907704543236_dp, 0.0021955552975368394_dp, &
             0.00348154167161361_dp, 0.005469107850667961_dp, 0.008515637512630765_dp, &
             0.013148903519834271_dp, 0.02014327571022081_dp, 0.030627975631431322_dp, &
             0.046240078975273546_dp, 0.06933972634015344_dp, 0.10331146632235541_dp, &
             0.15298443206035695_dp, 0.22521602407788194_dp, 0.32970031586674736_dp, &
             0.4800859438166058_dp, 0.6955237232472457_dp, 1.002823001945679_dp, &
             1.4395064418037677_dp, 2.0582924830264564_dp, 2.934123719159388_dp, &
             4.176489848221342_dp, 5.954998004572204_dp, 8.566956003577966_dp, &
             12.704441009453802_dp, &
             ! w
             5.890076009519511e-11_dp, 3.192909632224882e-10_dp, 1.2107064431521439e-09_dp, &
             3.765761904959274e-09_dp, 1.0288691082631399e-08_dp, 2.5613304538340012e-08_dp, &
             5.939542772773906e-08_dp, 1.3017260560550593e-07_dp, 2.7237465787572513e-07_dp, &
             5.481692992668862e-07_dp, 1.067098892137299e-06_dp, 2.018085433503081e-06_dp, &
             3.720807107625367e-06_dp, 6.707046618441869e-06_dp, 1.184789641903158e-05_dp, &
             2.0550407651383642e-05_dp, 3.505840889391953e-05_dp, 5.8908216673610795e-05_dp, &
             9.761277630780791e-05_dp, 0.0001596801146357906_dp, 0.0002581175622398492_dp, &
             0.00041263706758664265_dp, 0.0006528659468828749_dp, 0.0010229908786939144_dp, &
             0.001588433369504076_dp, 0.0024453890825492414_dp, 0.0037343838009584887_dp, &
             0.005659435354826038_dp, 0.008515003485146626_dp, 0.012723711521736862_dp, &
             0.01888890617165378_dp, 0.027867581663846647_dp, 0.040871168904073246_dp, &
             0.059604386695842414_dp, 0.08645612282862442_dp, 0.12476184832055617_dp, &
             0.17916591762214173_dp, 0.2561280995039498_dp, 0.36465186518237813_dp, &
             0.5173897931439986_dp, 0.7324823635948339_dp, 1.0370509473545133_dp, &
             1.475029135529334_dp, 2.128536330347539_dp, 3.1942648310066475_dp, &
             5.438691875088123_dp]

      CASE (1145)

         aw(:) = & ! a
            [1.3183147184500683e-11_dp, 1.3593052603235346e-10_dp, 6.979566040022932e-10_dp, &
             2.648977974688602e-09_dp, 8.39095154380992e-09_dp, 2.3489720091893182e-08_dp, &
             6.002383459497303e-08_dp, 1.4288697276326656e-07_dp, 3.2127266546265903e-07_dp, &
             6.890376338512807e-07_dp, 1.4199642016477966e-06_dp, 2.827571627783258e-06_dp, &
             5.464722487892563e-06_dp, 1.0286800531920756e-05_dp, 1.891510292076915e-05_dp, &
             3.405630952829122e-05_dp, 6.016234663928762e-05_dp, 0.00010445674011675562_dp, &
             0.00017851441896248906_dp, 0.0003006699661070718_dp, 0.000499655784478016_dp, &
             0.0008200521755115458_dp, 0.0013303874721282327_dp, 0.0021350877316750013_dp, &
             0.0033919825504173384_dp, 0.0053377814130711075_dp, 0.008324918346780634_dp, &
             0.012874522425464012_dp, 0.01975214372963747_dp, 0.030075431133482003_dp, &
             0.04546646439277627_dp, 0.06826621698732022_dp, 0.10183511271037193_dp, &
             0.15097245203247794_dp, 0.22249950878584412_dp, 0.3260677144005951_dp, &
             0.47527698457132245_dp, 0.6892248227523239_dp, 0.9946656720135671_dp, &
             1.4290706126479813_dp, 2.045116725059387_dp, 2.9177235135753703_dp, &
             4.15638238386079_dp, 5.930725714663616_dp, 8.538085212221711_dp, &
             12.670441154198503_dp, &
             ! w
             4.101598348423519e-11_dp, 2.5372970333361264e-10_dp, 1.0156816295282353e-09_dp, &
             3.2598814638029736e-09_dp, 9.098890590697723e-09_dp, 2.301179120234967e-08_dp, &
             5.4022443170164404e-08_dp, 1.1957719753347743e-07_dp, 2.5227148468324727e-07_dp, &
             5.112604839080281e-07_dp, 1.0012406883797258e-06_dp, 1.9034830805083944e-06_dp, &
             3.5257888211939555e-06_dp, 6.381787892520357e-06_dp, 1.1315225285045682e-05_dp, &
             1.969250434673992e-05_dp, 3.369780060418296e-05_dp, 5.6780938516992646e-05_dp, &
             9.433090316707594e-05_dp, 0.00015468004489564891_dp, 0.00025058943905977314_dp, &
             0.0004014295552305837_dp, 0.0006363592584588094_dp, 0.0009989295305552143_dp, &
             0.0015537087376587743_dp, 0.002395760663940541_dp, 0.0036641274790913687_dp, &
             0.005560907477568705_dp, 0.008378112017662732_dp, 0.012535291184962525_dp, &
             0.018632005099561783_dp, 0.027520685193489127_dp, 0.04040741498376785_dp, &
             0.05899087105200388_dp, 0.08565345459888354_dp, 0.12372420043293976_dp, &
             0.17784191574399924_dp, 0.2544629865318548_dp, 0.3625914057470307_dp, &
             0.5148862209068571_dp, 0.7295021077590301_dp, 1.0335825834858796_dp, &
             1.4710875091604472_dp, 2.124153641085798_dp, 3.1894417740910894_dp, &
             5.433140815438923_dp]

      CASE (1146)

         aw(:) = & ! a
            [1.1250678273887154e-11_dp, 1.2280860356642776e-10_dp, 6.477083214090759e-10_dp, &
             2.49514110472808e-09_dp, 7.980519733429738e-09_dp, 2.2495773766554698e-08_dp, &
             5.778659801627176e-08_dp, 1.3813450258994345e-07_dp, 3.116441256218406e-07_dp, &
             6.702898993114151e-07_dp, 1.3846804213269353e-06_dp, 2.7631003420070266e-06_dp, &
             5.349947183581236e-06_dp, 1.0087157991742216e-05_dp, 1.8575015919269084e-05_dp, &
             3.348785188071328e-05_dp, 5.922848686696023e-05_dp, 0.00010294687755937951_dp, &
             0.0001761090894494224_dp, 0.00029689050241014204_dp, 0.0004937933129369583_dp, &
             0.000811068554496121_dp, 0.0013167785720276848_dp, 0.0021146967745086197_dp, &
             0.003361748054071172_dp, 0.00529340065663193_dp, 0.008260403318440383_dp, &
             0.0127816212124882_dp, 0.019619596773133335_dp, 0.02988802956362016_dp, &
             0.045203878163093174_dp, 0.06790156584770467_dp, 0.10133326630738987_dp, &
             0.15028807098638022_dp, 0.22157488228226707_dp, 0.32483051866450047_dp, &
             0.4736381873069568_dp, 0.6870770937506279_dp, 0.9918828189205328_dp, &
             1.42550870210052_dp, 2.040617574078006_dp, 2.912120953027429_dp, &
             4.1495107804173506_dp, 5.922428087929023_dp, 8.528212849218558_dp, &
             12.658812342285152_dp, &
             ! w
             3.5846271010036296e-11_dp, 2.338626183751999e-10_dp, 9.549065296940625e-10_dp, &
             3.099717283726927e-09_dp, 8.718029147531816e-09_dp, 2.2172088131537334e-08_dp, &
             5.227672987676964e-08_dp, 1.1611609457586231e-07_dp, 2.4567482406656904e-07_dp, &
             4.991023239013811e-07_dp, 9.79473389874538e-07_dp, 1.8654931199680533e-06_dp, &
             3.460971905541957e-06_dp, 6.2734296873024125e-06_dp, 1.1137391942107859e-05_dp, &
             1.9405538669311967e-05_dp, 3.324187843212017e-05_dp, 5.6066958635853446e-05_dp, &
             9.322775548701746e-05_dp, 0.00015299701887247247_dp, 0.0002480521926634801_dp, &
             0.00039764766278854816_dp, 0.0006307828992383153_dp, 0.000990792390081867_dp, &
             0.0015419537088732782_dp, 0.0023789445246042646_dp, 0.0036403005042700883_dp, &
             0.00552746410779951_dp, 0.008331609530025114_dp, 0.012471235232284699_dp, &
             0.01854460460508332_dp, 0.027402585269080956_dp, 0.04024942642943273_dp, &
             0.058781729971175006_dp, 0.08537966716328438_dp, 0.1233700566063932_dp, &
             0.17738979014863024_dp, 0.2538940770043399_dp, 0.36188707020294253_dp, &
             0.514030018426745_dp, 0.7284824560785899_dp, 1.0323955017557676_dp, &
             1.4697380564426195_dp, 2.122652888528458_dp, 3.187790089105867_dp, &
             5.431239936861945_dp]

      CASE (1147)

         aw(:) = & ! a
            [1.101923605265126e-11_dp, 1.212226770123347e-10_dp, 6.415773490481731e-10_dp, &
             2.4762559051224134e-09_dp, 7.929923977907637e-09_dp, 2.237286972714994e-08_dp, &
             5.750930197539186e-08_dp, 1.3754433247188287e-07_dp, 3.104465559571273e-07_dp, &
             6.679550135935034e-07_dp, 1.38028108419894e-06_dp, 2.755053787714086e-06_dp, &
             5.335609718999313e-06_dp, 1.0062199662762204e-05_dp, 1.853247015479338e-05_dp, &
             3.3416691346808076e-05_dp, 5.9111517317791066e-05_dp, 0.00010275766175572183_dp, &
             0.00017580750772875497_dp, 0.0002964164187709001_dp, 0.000493057638379407_dp, &
             0.0008099407752923882_dp, 0.0013150695321341062_dp, 0.0021121351627158857_dp, &
             0.0033579486456704697_dp, 0.005287821900836725_dp, 0.008252291362738329_dp, &
             0.01276993694868505_dp, 0.019602922041636966_dp, 0.029864448371515518_dp, &
             0.04517082883195196_dp, 0.06785566067494088_dp, 0.10127007699301532_dp, &
             0.15020188129899056_dp, 0.22145841494913063_dp, 0.324674652157885_dp, &
             0.47343169091723036_dp, 0.686806426408576_dp, 0.9915320573058203_dp, &
             1.4250596814472636_dp, 2.040050328287654_dp, 2.911414505076273_dp, &
             4.148644218583927_dp, 5.921381592004945_dp, 8.526967644719207_dp, &
             12.657345496260316_dp, &
             ! w
             3.522634412785201e-11_dp, 2.314433744834323e-10_dp, 9.474449660105104e-10_dp, &
             3.079961852313683e-09_dp, 8.670900327815597e-09_dp, 2.2067928950071434e-08_dp, &
             5.2059771863957745e-08_dp, 1.1568527405557228e-07_dp, 2.448526285553359e-07_dp, &
             4.975852560182098e-07_dp, 9.767546778174784e-07_dp, 1.8607441691139815e-06_dp, &
             3.4528633065731547e-06_dp, 6.259864878117295e-06_dp, 1.1115116265568409e-05_dp, &
             1.9369572905323917e-05_dp, 3.318470806411213e-05_dp, 5.597738724706268e-05_dp, &
             9.308930183391209e-05_dp, 0.00015278570144603936_dp, 0.0002477335024242731_dp, &
             0.00039717247398146486_dp, 0.0006300820104651136_dp, 0.0009897693258403989_dp, &
             0.0015404753503734116_dp, 0.0023768290865569964_dp, 0.003637302346516419_dp, &
             0.005523254890595573_dp, 0.008325755325563726_dp, 0.012463169446471446_dp, &
             0.018533597020086723_dp, 0.027387708296702472_dp, 0.04022952091326529_dp, &
             0.05875537476623232_dp, 0.08534515943714915_dp, 0.12332541344238071_dp, &
             0.17733278637889285_dp, 0.2538223382994896_dp, 0.36179824173064734_dp, &
             0.5139220225121202_dp, 0.7283538278840112_dp, 1.0322457360422321_dp, &
             1.4695677900714799_dp, 2.122463519476907_dp, 3.1875816678034923_dp, &
             5.431000071386333_dp]

      CASE (1148)

         aw(:) = & ! a
            [1.798805909228713e-11_dp, 1.53839951225772e-10_dp, 7.091213507291229e-10_dp, &
             2.525406955289592e-09_dp, 7.667654857085414e-09_dp, 2.0816291223776836e-08_dp, &
             5.195656038061871e-08_dp, 1.2138678832062047e-07_dp, 2.6876438647203315e-07_dp, &
             5.690237179878281e-07_dp, 1.159758406368651e-06_dp, 2.287395644728178e-06_dp, &
             4.383691923689901e-06_dp, 8.190463170074449e-06_dp, 1.496006141312805e-05_dp, &
             2.677340899807644e-05_dp, 4.703851517601322e-05_dp, 8.12633246330919e-05_dp, &
             0.0001382416996407243_dp, 0.00023185621721525868_dp, 0.00038379448229264757_dp, &
             0.0006276081870520739_dp, 0.0010147318504401767_dp, 0.0016233425351401937_dp, &
             0.0025713121941002273_dp, 0.004035020524907112_dp, 0.006276512313312456_dp, &
             0.009682472081804859_dp, 0.014819848369208103_dp, 0.02251482142547398_dp, &
             0.03396434705988137_dp, 0.0508929604494615_dp, 0.07577220134585366_dp, &
             0.11212635188431745_dp, 0.16495674998495494_dp, 0.2413286133868193_dp, &
             0.3511804421153578_dp, 0.5084390984557957_dp, 0.7325586026113594_dp, &
             1.0506591102672675_dp, 1.5005535801686698_dp, 2.135191210468108_dp, &
             3.0296408310251155_dp, 4.293378275890483_dp, 6.095866185871022_dp, &
             8.734282193906077_dp, 12.901275925764192_dp, &
             ! w
             5.220770839682068e-11_dp, 2.6503060583671306e-10_dp, 9.75126931198741e-10_dp, &
             2.979744017690284e-09_dp, 8.042330881549597e-09_dp, 1.983960975029142e-08_dp, &
             4.5679716514156204e-08_dp, 9.953556843749864e-08_dp, 2.0726919689764147e-07_dp, &
             4.154387726112272e-07_dp, 8.058682800459535e-07_dp, 1.519350460702175e-06_dp, &
             2.793633715345739e-06_dp, 5.023479559556112e-06_dp, 8.854446711724928e-06_dp, &
             1.5327691927520336e-05_dp, 2.61012188530083e-05_dp, 4.378468821099094e-05_dp, &
             7.24415532487845e-05_dp, 0.00011833587076158645_dp, 0.0001910343527797288_dp, &
             0.00030502073785644215_dp, 0.00048204435818159327_dp, 0.0007545145969349059_dp, &
             0.0011703768387402353_dp, 0.0018000756858538765_dp, 0.002746443912959605_dp, &
             0.004158672545310963_dp, 0.006251947394237573_dp, 0.009334918648789629_dp, &
             0.013847953598539466_dp, 0.0204161764611164_dp, 0.029922716701901077_dp, &
             0.04360950073258756_dp, 0.06321553524304416_dp, 0.09116629633198774_dp, &
             0.1308332603778279_dp, 0.18689138028559418_dp, 0.26581835455420155_dp, &
             0.376613111117387_dp, 0.5318898499617621_dp, 0.7497071939172463_dp, &
             1.0570605884241022_dp, 1.4977361727541902_dp, 2.153759302748372_dp, &
             3.222010625928744_dp, 5.470635428748041_dp]

      CASE (1149)

         aw(:) = & ! a
            [1.1417759176069757e-11_dp, 1.1105938189470662e-10_dp, 5.533920542273237e-10_dp, &
             2.0647424803399273e-09_dp, 6.46741830218002e-09_dp, 1.7959772915624782e-08_dp, &
             4.5612323710476485e-08_dp, 1.0805260854793613e-07_dp, 2.419827654873018e-07_dp, &
             5.172522956870207e-07_dp, 1.0629144753439889e-06_dp, 2.111356188980847e-06_dp, &
             4.071690695400636e-06_dp, 7.649846683497861e-06_dp, 1.4042197417358748e-05_dp, &
             2.524362074270852e-05_dp, 4.453166315438664e-05_dp, 7.721897557378769e-05_dp, &
             0.00013181058989757794_dp, 0.00022176679215919567_dp, 0.0003681645109326792_dp, &
             0.0006036816325803273_dp, 0.00097851513923606_dp, 0.0015691069402872832_dp, &
             0.002490920116860705_dp, 0.003917023010412205_dp, 0.006104953358701777_dp, &
             0.00943532269596246_dp, 0.014466980951670576_dp, 0.022015424666555705_dp, &
             0.033263676229369606_dp, 0.049918322353087616_dp, 0.07442809450481169_dp, &
             0.1102887309765592_dp, 0.16246647365206343_dp, 0.23798428492382606_dp, &
             0.34673108518396406_dp, 0.5025774446101041_dp, 0.7249162078372728_dp, &
             1.0408050213248192_dp, 1.4879986939475671_dp, 2.119400478451778_dp, &
             3.010054429986924_dp, 4.2694397456277935_dp, 6.067048339605677_dp, &
             8.700083316873853_dp, 12.861075814593784_dp, &
             ! w
             3.47119814073856e-11_dp, 2.0271952627479283e-10_dp, 7.932891774749607e-10_dp, &
             2.5130694830641525e-09_dp, 6.952621844916025e-09_dp, 1.746951715737868e-08_dp, &
             4.0804454994412794e-08_dp, 8.995210083332826e-08_dp, 1.8913184829111012e-07_dp, &
             3.822064075152068e-07_dp, 7.466661694191297e-07_dp, 1.4164634351054509e-06_dp, &
             2.6187279551115562e-06_dp, 4.73198830469709e-06_dp, 8.377337450397964e-06_dp, &
             1.4559543829902424e-05_dp, 2.4883161984049447e-05_dp, 4.188026907499856e-05_dp, &
             6.950297320086579e-05_dp, 0.00011385727221732604_dp, 0.00018428788347498945_dp, &
             0.00029496999496710823_dp, 0.0004672285054874817_dp, 0.0007328949882252686_dp, &
             0.001139136446394179_dp, 0.0017553601696330965_dp, 0.0026830320197529806_dp, &
             0.004069563117950773_dp, 0.0061278512246060795_dp, 0.009163646701051343_dp, &
             0.013613700181815014_dp, 0.02009870876791137_dp, 0.0294965035582973_dp, &
             0.0430428470670722_dp, 0.062469839043384066_dp, 0.09019561586399491_dp, &
             0.12958447265743486_dp, 0.18530532021556598_dp, 0.2638324199416774_dp, &
             0.3741658329415729_dp, 0.5289277197994406_dp, 0.7461933717959262_dp, &
             1.052983635613201_dp, 1.4931141306490128_dp, 2.1486285463100314_dp, &
             3.2163682572792656_dp, 5.464138054959617_dp]

      CASE (1150)

         aw(:) = & ! a
            [9.18876775445597e-12_dp, 9.612182778583699e-11_dp, 4.97023330471738e-10_dp, &
             1.8937874605225484e-09_dp, 6.0141715197395455e-09_dp, 1.6867013812344426e-08_dp, &
             4.316075650419434e-08_dp, 1.0285786817457909e-07_dp, 2.3147853696017573e-07_dp, &
             4.968303047845597e-07_dp, 1.0245246442057487e-06_dp, 2.041272322345501e-06_dp, &
             3.947006852782393e-06_dp, 7.433071215594664e-06_dp, 1.3673034984383113e-05_dp, &
             2.462665125010715e-05_dp, 4.351810717182004e-05_dp, 7.558003681059848e-05_dp, &
             0.00012919893625357382_dp, 0.00021766151800224056_dp, 0.00036179334569629766_dp, &
             0.0005939121485742866_dp, 0.0009637041869080006_dp, 0.0015468944299197173_dp, &
             0.0024579495206692756_dp, 0.003868566571924585_dp, 0.006034415041797955_dp, &
             0.009333586593086278_dp, 0.014321567622109467_dp, 0.021809413450606982_dp, &
             0.03297434923781482_dp, 0.04951548753030204_dp, 0.07387205388896591_dp, &
             0.10952788036842959_dp, 0.16143455738562945_dp, 0.23659739221976878_dp, &
             0.3448845719074522_dp, 0.5001431045925608_dp, 0.7217402031028426_dp, &
             1.0367072930717973_dp, 1.482774755796338_dp, 2.1128265011704173_dp, &
             3.0018960995220336_dp, 4.259464072062987_dp, 6.055034580475648_dp, &
             8.68582152247754_dp, 12.84430684406525_dp, &
             ! w
             2.875801477412135e-11_dp, 1.8036215126170125e-10_dp, 7.257711630725427e-10_dp, &
             2.336374944815795e-09_dp, 6.534393560717241e-09_dp, 1.6550477769589058e-08_dp, &
             3.889849161311228e-08_dp, 8.618028780543101e-08_dp, 1.81953127236594e-07_dp, &
             3.6898954998635163e-07_dp, 7.230220183914937e-07_dp, 1.3752206148100382e-06_dp, &
             2.5483858439114003e-06_dp, 4.614413714992229e-06_dp, 8.184380898323216e-06_dp, &
             1.42481312342943e-05_dp, 2.4388260272038775e-05_dp, 4.110492084493808e-05_dp, &
             6.830433721562149e-05_dp, 0.00011202728371938108_dp, 0.00018152675245497956_dp, &
             0.0002908502867850502_dp, 0.00046114699785363637_dp, 0.0007240088831063442_dp, &
             0.0011262798666155092_dp, 0.0017369362477710885_dp, 0.0026568753963568184_dp, &
             0.004032767474599687_dp, 0.006076556853055943_dp, 0.0090927844627551_dp, &
             0.013516691133342256_dp, 0.019967124065479187_dp, 0.029319698195714807_dp, &
             0.042807595068950224_dp, 0.06216001878180788_dp, 0.08979202397790674_dp, &
             0.1290648842717654_dp, 0.1846449589515681_dp, 0.2630050420493166_dp, &
             0.3731456368266956_dp, 0.5276922065903086_dp, 0.7447270036153351_dp, &
             1.0512815184172557_dp, 1.4911837629668827_dp, 2.1464852048788896_dp, &
             3.2140109565498305_dp, 5.461423748252499_dp]

      CASE (1151)

         aw(:) = & ! a
            [8.23830627361467e-12_dp, 8.965198074504287e-11_dp, 4.721837990107785e-10_dp, &
             1.8175951713247128e-09_dp, 5.81058534960726e-09_dp, 1.6373351134337678e-08_dp, &
             4.204830352813864e-08_dp, 1.0049218832358239e-07_dp, 2.2668073844940317e-07_dp, &
             4.87479203221545e-07_dp, 1.0069083111788144e-06_dp, 2.009051827230209e-06_dp, &
             3.8895895271442995e-06_dp, 7.333098173487026e-06_dp, 1.3502558286685386e-05_dp, &
             2.4341397665836132e-05_dp, 4.3048982349740864e-05_dp, 7.48206962725485e-05_dp, &
             0.00012798781424880914_dp, 0.0002157561336814028_dp, 0.00035883396591783977_dp, &
             0.0005893709423154242_dp, 0.0009568148204614256_dp, 0.0015365555907910398_dp, &
             0.002442594097550494_dp, 0.0038459861542346286_dp, 0.006001527068966944_dp, &
             0.009286128978717997_dp, 0.01425370320016663_dp, 0.02171322449480679_dp, &
             0.032839201179569344_dp, 0.04932724176855373_dp, 0.07361211364797975_dp, &
             0.10917206295620524_dp, 0.16095180377427828_dp, 0.23594835433734912_dp, &
             0.3440201637610421_dp, 0.49900316933645056_dp, 0.7202525353412668_dp, &
             1.0347873558817557_dp, 1.480326516963952_dp, 2.109744817567887_dp, &
             2.9980708785777512_dp, 4.2547858269385825_dp, 6.049399578877714_dp, &
             8.679131131690859_dp, 12.836439403694241_dp, &
             ! w
             2.621416721450516e-11_dp, 1.7054737238216016e-10_dp, 6.956732371838825e-10_dp, &
             2.2569215998250536e-09_dp, 6.345193975779423e-09_dp, 1.6132824450148453e-08_dp, &
             3.80292116459328e-08_dp, 8.445494995672559e-08_dp, 1.7866124933775816e-07_dp, &
             3.6291600815704345e-07_dp, 7.121368882317043e-07_dp, 1.3562029237851986e-06_dp, &
             2.5159035914996503e-06_dp, 4.5600510112362096e-06_dp, 8.095060631845659e-06_dp, &
             1.4103825401915711e-05_dp, 2.415870629526777e-05_dp, 4.074496680765386e-05_dp, &
             6.774741762584867e-05_dp, 0.00011117637562570732_dp, 0.00018024197583741176_dp, &
             0.0002889320903274638_dp, 0.00045831361161084944_dp, 0.0007198664299315931_dp, &
             0.0011202832183961513_dp, 0.0017283384178999434_dp, 0.002644663026057907_dp, &
             0.004015579877626759_dp, 0.006052586256148459_dp, 0.009059655708906569_dp, &
             0.01347132048068198_dp, 0.019905559215678912_dp, 0.02923694588066883_dp, &
             0.042697449130802236_dp, 0.06201491157352632_dp, 0.08960293793483894_dp, &
             0.1288213789726416_dp, 0.18433539058766554_dp, 0.26261707155236497_dp, &
             0.3726671263857382_dp, 0.5271125643141326_dp, 0.7440389051711221_dp, &
             1.0504826431089604_dp, 1.4902776240303044_dp, 2.1454789921092536_dp, &
             3.212904255435092_dp, 5.460149490725463_dp]

      CASE (1152)

         aw(:) = & ! a
            [8.002495313787271e-12_dp, 8.80354955228065e-11_dp, 4.659324580989662e-10_dp, &
             1.7983303227504877e-09_dp, 5.758945479142516e-09_dp, 1.624784012452785e-08_dp, &
             4.176495709699171e-08_dp, 9.988876491102496e-08_dp, 2.25455476723559e-07_dp, &
             4.850886992965945e-07_dp, 1.0024009734373747e-06_dp, 2.0008015979043476e-06_dp, &
             3.874877688293248e-06_dp, 7.307467191827834e-06_dp, 1.3458828306281963e-05_dp, &
             2.4268190244420503e-05_dp, 4.292853332063449e-05_dp, 7.462565514986567e-05_dp, &
             0.00012767661525780135_dp, 0.00021526637596255068_dp, 0.00035807304936854206_dp, &
             0.0005882029617098168_dp, 0.0009550424142456422_dp, 0.001533895063468268_dp, &
             0.002438641679670952_dp, 0.0038401727305812446_dp, 0.005993058102266889_dp, &
             0.00927390570826476_dp, 0.014236220553352327_dp, 0.021688440606318277_dp, &
             0.03280437311969673_dp, 0.04927872227727763_dp, 0.07354510471708023_dp, &
             0.10908032452428991_dp, 0.16082732033276853_dp, 0.23578096985348862_dp, &
             0.34379720713636_dp, 0.49870910967209414_dp, 0.7198687288162936_dp, &
             1.0342919719646237_dp, 1.479694753692039_dp, 2.108949516854285_dp, &
             2.9970836009277115_dp, 4.253578286818751_dp, 6.047944976038668_dp, &
             8.677403984545812_dp, 12.834408288535888_dp, &
             ! w
             2.5582413575729417e-11_dp, 1.6808102799325176e-10_dp, 6.880625737859812e-10_dp, &
             2.2367594487166643e-09_dp, 6.2970644348170965e-09_dp, 1.6026383109840176e-08_dp, &
             3.78073470546418e-08_dp, 8.401406977235512e-08_dp, 1.7781922538385101e-07_dp, &
             3.6136113920869383e-07_dp, 7.093481546188499e-07_dp, 1.3513274854117128e-06_dp, &
             2.5075715228237105e-06_dp, 4.54609913916192e-06_dp, 8.072126391237208e-06_dp, &
             1.4066757099513577e-05_dp, 2.4099717120727124e-05_dp, 4.065243530935404e-05_dp, &
             6.760420603616111e-05_dp, 0.00011095749818905482_dp, 0.00017991140128515595_dp, &
             0.00028843840507035645_dp, 0.0004575842030906138_dp, 0.000718799774663469_dp, &
             0.0011187387801020441_dp, 0.0017261235848265283_dp, 0.0026415164558987224_dp, &
             0.004011150594044567_dp, 0.006046407885063582_dp, 0.009051115404763773_dp, &
             0.01345962244501853_dp, 0.019889683355693032_dp, 0.029215603238347407_dp, &
             0.04266903741894665_dp, 0.061977476710306054_dp, 0.08955415109324912_dp, &
             0.12875854343657084_dp, 0.1842554982938495_dp, 0.26251693420393774_dp, &
             0.3725436069419209_dp, 0.5269629243610809_dp, 0.7438612498755058_dp, &
             1.05027637017034_dp, 1.4900436389359122_dp, 2.145219151929823_dp, &
             3.212618456294743_dp, 5.459820418855417_dp]

      CASE (1153)

         aw(:) = & ! a
            [1.647616291142771e-11_dp, 1.3356977881862888e-10_dp, 5.924969580081976e-10_dp, &
             2.059413151029267e-09_dp, 6.14966574072076e-09_dp, 1.6492290190203246e-08_dp, &
             4.077629321682057e-08_dp, 9.454497442548673e-08_dp, 2.080244530351058e-07_dp, &
             4.381020206990912e-07_dp, 8.888753255188183e-07_dp, 1.7462203450910746e-06_dp, &
             3.334935465168507e-06_dp, 6.211736260168672e-06_dp, 1.1314460046508406e-05_dp, &
             2.019837991754068e-05_dp, 3.54060755092658e-05_dp, 6.104005634050375e-05_dp, &
             0.00010364036959251668_dp, 0.00017351706307441558_dp, 0.00028675561758322244_dp, &
             0.00046821061572620986_dp, 0.0007559404114482438_dp, 0.0012077290784531702_dp, &
             0.001910612833638326_dp, 0.0029947055145396815_dp, 0.004653140869502506_dp, &
             0.007170671426541634_dp, 0.010964455788847733_dp, 0.01664192377317624_dp, &
             0.025082459270423497_dp, 0.03755215390042441_dp, 0.05586428685780406_dp, &
             0.08260278078053697_dp, 0.1214320803737933_dp, 0.17752527086626296_dp, &
             0.2581536355813954_dp, 0.37349659326911194_dp, 0.5377535075510939_dp, &
             0.7706733145816889_dp, 1.0996760788236264_dp, 1.5628520770064933_dp, &
             2.213367635202292_dp, 3.1264060892795835_dp, 4.411423082941484_dp, &
             6.237739369248621_dp, 8.902415626591255_dp, 13.098694640986073_dp, &
             ! w
             4.699689056476382e-11_dp, 2.2369412140157897e-10_dp, 7.969990837687399e-10_dp, &
             2.389321592681589e-09_dp, 6.3636676070987175e-09_dp, 1.554278675849097e-08_dp, &
             3.550681442513094e-08_dp, 7.687646234301937e-08_dp, 1.5923377716266537e-07_dp, &
             3.1771458630983175e-07_dp, 6.138896417108705e-07_dp, 1.1534306838884694e-06_dp, &
             2.1143703720716197e-06_dp, 3.7917066644824396e-06_dp, 6.666949334327637e-06_dp, &
             1.1515347691971003e-05_dp, 1.9569517103637245e-05_dp, 3.276675277584921e-05_dp, &
             5.411952608387905e-05_dp, 8.826593605276049e-05_dp, 0.00014228120082721485_dp, &
             0.000226865297591241_dp, 0.0003580692245278835_dp, 0.000559787971399868_dp, &
             0.0008673387486631246_dp, 0.0013325671897858047_dp, 0.0020310939354332294_dp, &
             0.0030725449818923617_dp, 0.0046149223481792845_dp, 0.006884694707708151_dp, &
             0.010204757338739478_dp, 0.015033175836713802_dp, 0.022016653799727667_dp, &
             0.03206404071676786_dp, 0.04644705117565207_dp, 0.06693790111976468_dp, &
             0.0959971356610976_dp, 0.1370302427932984_dp, 0.19474035163116307_dp, &
             0.275620432681178_dp, 0.3886624029327335_dp, 0.5464405429548751_dp, &
             0.7669320501302382_dp, 1.0770099840493617_dp, 1.5203203212057514_dp, &
             2.1788046527703147_dp, 3.249542192327256_dp, 5.50234885850124_dp]

      CASE (1154)

         aw(:) = & ! a
            [1.0144967735136025e-11_dp, 9.296088115102934e-11_dp, 4.477696612109085e-10_dp, &
             1.6378258285365944e-09_dp, 5.0630430005705925e-09_dp, 1.392662229799864e-08_dp, &
             3.511256832931075e-08_dp, 8.269804923695543e-08_dp, 1.843217959131674e-07_dp, &
             3.924272194524353e-07_dp, 8.036587152126382e-07_dp, 1.591652435479185e-06_dp, &
             3.0614830835829437e-06_dp, 5.738623442809387e-06_dp, 1.051219247078839e-05_dp, &
             1.886258694429004e-05_dp, 3.321884641059549e-05_dp, 5.751346815133137e-05_dp, &
             9.803493346273935e-05_dp, 0.00016472522208323587_dp, 0.00027313705747823524_dp, &
             0.00044736182642954536_dp, 0.0007243754737116722_dp, 0.001160441860416062_dp, &
             0.0018404825040348601_dp, 0.0028916974167899775_dp, 0.004503243389424042_dp, &
             0.006954496791624054_dp, 0.01065541940172467_dp, 0.016203903932962286_dp, &
             0.02446682909718306_dp, 0.03669407339301907_dp, 0.054678150611630984_dp, &
             0.08097674391823168_dp, 0.11922164312176826_dp, 0.17454600660197928_dp, &
             0.2541733181316527_dp, 0.3682272395973995_dp, 0.5308442208865406_dp, &
             0.7617052900954628_dp, 1.0881616659602218_dp, 1.5482402129544741_dp, &
             2.1950579025978905_dp, 3.1037724269629003_dp, 4.383844659777786_dp, &
             6.204628132004833_dp, 8.863209579862483_dp, 13.05269192734497_dp, &
             ! w
             3.0161595988161546e-11_dp, 1.6545718429501698e-10_dp, 6.30668086185682e-10_dp, &
             1.967602553974376e-09_dp, 5.387056895645382e-09_dp, 1.3431688663890328e-08_dp, &
             3.118488658561038e-08_dp, 6.841267998890188e-08_dp, 1.4326418126058122e-07_dp, &
             2.885266139291353e-07_dp, 5.619980018342678e-07_dp, 1.063398332338151e-06_dp, &
             1.9615240343802775e-06_dp, 3.5372545323823513e-06_dp, 6.250817515743305e-06_dp, &
             1.084579753408801e-05_dp, 1.8508271548135743e-05_dp, 3.110792702080238e-05_dp, &
             5.156010417848469e-05_dp, 8.436483823439821e-05_dp, 0.00013640312459614474_dp, &
             0.00021810447074536128_dp, 0.00034514697362515354_dp, 0.0005409165951872263_dp, &
             0.0008400426338306469_dp, 0.0012934505109839817_dp, 0.001975542816452841_dp, &
             0.0029943509553587793_dp, 0.0045058138050219606_dp, 0.006733764494277592_dp, &
             0.009997778831469054_dp, 0.014751808269803065_dp, 0.021637552573414102_dp, &
             0.03155791290115861_dp, 0.045777734360736_dp, 0.06606159960488964_dp, &
             0.09486202580643768_dp, 0.13557675986523965_dp, 0.19290258409267164_dp, &
             0.27332914421918636_dp, 0.3858501997082122_dp, 0.5430494494902006_dp, &
             0.7629229959847406_dp, 1.0723720675072448_dp, 1.5150746254207113_dp, &
             2.1729908664851174_dp, 3.2431529088679847_dp, 5.494987613419884_dp]

      CASE (1155)

         aw(:) = & ! a
            [7.900769488235331e-12_dp, 7.814722663022084e-11_dp, 3.9287987365678147e-10_dp, &
             1.4733788538638435e-09_dp, 4.630659582379246e-09_dp, 1.2890389907385739e-08_dp, &
             3.2798283020215687e-08_dp, 7.781135343265963e-08_dp, 1.7446785647434285e-07_dp, &
             3.733121315195938e-07_dp, 7.677902264644754e-07_dp, 1.526266615797455e-06_dp, &
             2.945293159184414e-06_dp, 5.536802384663569e-06_dp, 1.0168742687482866e-05_dp, &
             1.8288893582839574e-05_dp, 3.2276723472626514e-05_dp, 5.5990348730190556e-05_dp, &
             9.560797757493799e-05_dp, 0.0001609099585620783_dp, 0.00026721467043683966_dp, &
             0.00043827726291568975_dp, 0.0007105961024765451_dp, 0.0011397633661777572_dp, &
             0.0018097649824182812_dp, 0.0028465101980149946_dp, 0.004437391975662489_dp, &
             0.00685939955767285_dp, 0.010519295663022352_dp, 0.01601072917567208_dp, &
             0.024195008540464732_dp, 0.03631478315093013_dp, 0.054153298931598634_dp, &
             0.08025651686368643_dp, 0.11824162476711915_dp, 0.17322391039407323_dp, &
             0.25240544261592784_dp, 0.3658848785640795_dp, 0.5277704240110832_dp, &
             0.7577125854972211_dp, 1.0830315973218378_dp, 1.5417257365369637_dp, &
             2.186889667291552_dp, 3.0936694161075984_dp, 4.3715280958646385_dp, &
             6.189833951409355_dp, 8.845685605106762_dp, 13.032123770941494_dp, &
             ! w
             2.41764490065197e-11_dp, 1.435939294168813e-10_dp, 5.65743785096857e-10_dp, &
             1.7992651282452442e-09_dp, 4.991091335333693e-09_dp, 1.2565526936477847e-08_dp, &
             2.939479244867424e-08_dp, 6.487969934322884e-08_dp, 1.365542950430527e-07_dp, &
             2.761937414929718e-07_dp, 5.39964658282401e-07_dp, 1.0250053724917532e-06_dp, &
             1.8960947867160589e-06_dp, 3.4279555476846832e-06_dp, 6.0715130793484074e-06_dp, &
             1.0556481821311543e-05_dp, 1.804851326959257e-05_dp, 3.038756769848336e-05_dp, &
             5.044620339024835e-05_dp, 8.2663546825465e-05_dp, 0.00013383477971527777_dp, &
             0.0002142697368704272_dp, 0.00033948129747542844_dp, 0.0005326296027157612_dp, &
             0.0008280383949904617_dp, 0.0012762238735087765_dp, 0.0019510463390754642_dp, &
             0.0029598264848269967_dp, 0.004457582650182253_dp, 0.006666970780363717_dp, &
             0.009906082454046554_dp, 0.01462702761636469_dp, 0.021469263852379374_dp, &
             0.03133302386686176_dp, 0.04548006722506051_dp, 0.06567154303770206_dp, &
             0.0943563509773263_dp, 0.13492874087553125_dp, 0.19208261475051983_dp, &
             0.27230608644363663_dp, 0.38459369901376694_dp, 0.5415333405985303_dp, &
             0.7611295807487598_dp, 1.0702963114337831_dp, 1.5127259268263606_dp, &
             2.170387119996248_dp, 3.2402911128162435_dp, 5.491690772171394_dp]

      CASE (1156)

         aw(:) = & ! a
            [6.7874596423246166e-12_dp, 7.066690751586158e-11_dp, 3.6456915931928396e-10_dp, &
             1.387332324791829e-09_dp, 4.402142765417573e-09_dp, 1.2338667407868526e-08_dp, &
             3.1558970897856514e-08_dp, 7.518232338120744e-08_dp, 1.6914602482103706e-07_dp, &
             3.629549477376612e-07_dp, 7.483009582293073e-07_dp, 1.4906520215624133e-06_dp, &
             2.8818696567134147e-06_dp, 5.426424113204663e-06_dp, 9.980581070194102e-06_dp, &
             1.797409915316835e-05_dp, 3.175902898517468e-05_dp, 5.515230539119803e-05_dp, &
             9.427102595246788e-05_dp, 0.00015880588910014394_dp, 0.000263945186338151_dp, &
             0.00043325727594663695_dp, 0.0007029750220659643_dp, 0.001128316935635593_dp, &
             0.001792748130694753_dp, 0.0028214588849297425_dp, 0.004400859119483845_dp, &
             0.006806606909183448_dp, 0.010443680085036633_dp, 0.01590335850693984_dp, &
             0.024043839598597495_dp, 0.0361037333195303_dp, 0.053861104520857885_dp, &
             0.0798553580325367_dp, 0.11769551112661945_dp, 0.17248684745124082_dp, &
             0.25141944207006134_dp, 0.3645779409783641_dp, 0.5260547155746327_dp, &
             0.7554831516313645_dp, 1.08016609172011_dp, 1.5380857532936762_dp, &
             2.1823242527675175_dp, 3.088021033404532_dp, 4.364640431005401_dp, &
             6.181558934760973_dp, 8.835881875610946_dp, 13.020615237565316_dp, &
             ! w
             2.1201322652306333e-11_dp, 1.3237337448242993e-10_dp, 5.317625571166145e-10_dp, &
             1.7101680920780329e-09_dp, 4.779881688897635e-09_dp, 1.2100788173296617e-08_dp, &
             2.842982516075328e-08_dp, 6.296791233281326e-08_dp, 1.329117199566872e-07_dp, &
             2.6948020620497256e-07_dp, 5.279418912568997e-07_dp, 1.0040117289776313e-06_dp, &
             1.860250575882954e-06_dp, 3.36797775877298e-06_dp, 5.972970810174391e-06_dp, &
             1.0397260463175537e-05_dp, 1.779517215201591e-05_dp, 2.9990167525947964e-05_dp, &
             4.983104104480311e-05_dp, 8.172306016039312e-05_dp, 0.0001324136679335978_dp, &
             0.00021214607844944128_dp, 0.0003363411351372884_dp, 0.0005280331056047854_dp, &
             0.0008213753124429223_dp, 0.0012666555857422572_dp, 0.0019374314313039156_dp, &
             0.0029406264936642124_dp, 0.004430744572865759_dp, 0.006629783282966314_dp, &
             0.009855003734538537_dp, 0.014557484967761034_dp, 0.02137542869902573_dp, &
             0.031207572173697046_dp, 0.045313944705023305_dp, 0.0654537686667965_dp, &
             0.09407391201025903_dp, 0.13456665809133175_dp, 0.1916242857426715_dp, &
             0.2717340392935814_dp, 0.38389088914906944_dp, 0.5406850610897284_dp, &
             0.7601258654666483_dp, 1.069134297590356_dp, 1.5114108617155213_dp, &
             2.168929051612694_dp, 3.238688438744156_dp, 5.4898445166407175_dp]

      CASE (1157)

         aw(:) = & ! a
            [6.181420193575556e-12_dp, 6.65470267569179e-11_dp, 3.487747118105563e-10_dp, &
             1.3389196427805697e-09_dp, 4.2728221224183854e-09_dp, 1.2025101065120915e-08_dp, &
             3.0852277587697435e-08_dp, 7.367916117844684e-08_dp, 1.6609650680467592e-07_dp, &
             3.570089714746149e-07_dp, 7.370943325441603e-07_dp, 1.4701443750551264e-06_dp, &
             2.8453038859193375e-06_dp, 5.362717267768913e-06_dp, 9.871872733789604e-06_dp, &
             1.779206737863755e-05_dp, 3.1459425243852194e-05_dp, 5.4666945609335756e-05_dp, &
             9.349618834699118e-05_dp, 0.0001575856916563184_dp, 0.00026204802447731164_dp, &
             0.0004303427681822409_dp, 0.0006985481098337754_dp, 0.001121664784629989_dp, &
             0.0017828542603032492_dp, 0.002806887493857699_dp, 0.004379600882473427_dp, &
             0.006775875628394275_dp, 0.010399647581412331_dp, 0.015840813209561596_dp, &
             0.023955752743147277_dp, 0.0359807160031879_dp, 0.05369073986369921_dp, &
             0.07962139653625894_dp, 0.11737692568596443_dp, 0.17205675969185794_dp, &
             0.2508439567530354_dp, 0.3638149635576707_dp, 0.5250528825796598_dp, &
             0.7541810745043206_dp, 1.0784921944730756_dp, 1.5359590485606536_dp, &
             2.179656394878735_dp, 3.0847198097759216_dp, 4.360614336028439_dp, &
             6.176721290320609_dp, 8.830149950295533_dp, 13.013886027101723_dp, &
             ! w
             1.9579369543349136e-11_dp, 1.2613112345911602e-10_dp, 5.126407796252794e-10_dp, &
             1.6597059663785438e-09_dp, 4.6597192597144355e-09_dp, 1.1835487814971327e-08_dp, &
             2.7877482485645675e-08_dp, 6.187120576671947e-08_dp, 1.3081828046753875e-07_dp, &
             2.656157527016099e-07_dp, 5.210118502031575e-07_dp, 9.918962193619926e-07_dp, &
             1.8395426320703812e-06_dp, 3.3332941440493052e-06_dp, 5.9159370334843796e-06_dp, &
             1.0305034657884553e-05_dp, 1.7648323762290914e-05_dp, 2.975966339392436e-05_dp, &
             4.947401032602271e-05_dp, 8.117690707835691e-05_dp, 0.0001315879744257946_dp, &
             0.0002109115843794137_dp, 0.00033451490037359713_dp, 0.0005253587477896196_dp, &
             0.0008174969823172447_dp, 0.001261084101443533_dp, 0.001929500769588309_dp, &
             0.0029294386721103087_dp, 0.004415100921404012_dp, 0.006608100302343785_dp, &
             0.009825212332646648_dp, 0.014516913084054821_dp, 0.021320669505527013_dp, &
             0.031134343623957814_dp, 0.04521695177668224_dp, 0.06532658789062631_dp, &
             0.09390892941086802_dp, 0.13435510663684844_dp, 0.19135644547707067_dp, &
             0.2713996777749537_dp, 0.38348002022274097_dp, 0.5401890635622258_dp, &
             0.7595388916231829_dp, 1.0684546596798659_dp, 1.5106416265419067_dp, &
             2.168076110013365_dp, 3.237750885409374_dp, 5.488764510376502_dp]

      CASE (1158)

         aw(:) = & ! a
            [5.884644058849317e-12_dp, 6.451507527815593e-11_dp, 3.4092646896038614e-10_dp, &
             1.3147474454606686e-09_dp, 4.2080403024173724e-09_dp, 1.1867643341205218e-08_dp, &
             3.04967471345055e-08_dp, 7.292179977482178e-08_dp, 1.645581147686142e-07_dp, &
             3.5400625695930574e-07_dp, 7.314298943591434e-07_dp, 1.4597705582495436e-06_dp, &
             2.826794249611054e-06_dp, 5.330448921876189e-06_dp, 9.81678009335476e-06_dp, &
             1.769976877566513e-05_dp, 3.1307443001346005e-05_dp, 5.4420630892320425e-05_dp, &
             9.310281610602178e-05_dp, 0.00015696599824466405_dp, 0.0002610842093686673_dp, &
             0.0004288616583924934_dp, 0.0006962977765647008_dp, 0.0011182823917793637_dp, &
             0.0017778223012139837_dp, 0.0027994748278307568_dp, 0.004368784117619784_dp, &
             0.0067602354308881_dp, 0.010377233469571985_dp, 0.015808969405048615_dp, &
             0.023910896858780305_dp, 0.03591806200385311_dp, 0.053603957243433184_dp, &
             0.07950219960890689_dp, 0.11721459114984535_dp, 0.17183757839091535_dp, &
             0.2505506383481965_dp, 0.3634260323981909_dp, 0.52454213094434_dp, &
             0.753517175640407_dp, 1.0776386186218405_dp, 1.5348744580233893_dp, &
             2.1782956912800984_dp, 3.0830359163653553_dp, 4.358560533936401_dp, &
             6.17425332478917_dp, 8.827225584684081_dp, 13.010452684714588_dp, &
             ! w
             1.8784318334205977e-11_dp, 1.2303422488568435e-10_dp, 5.030926749824205e-10_dp, &
             1.6344162169120857e-09_dp, 4.599345074656738e-09_dp, 1.170193569666764e-08_dp, &
             2.759901357191199e-08_dp, 6.13176081164999e-08_dp, 1.2976045602941865e-07_dp, &
             2.636612991714466e-07_dp, 5.175042826021328e-07_dp, 9.857599508737265e-07_dp, &
             1.8290481910737664e-06_dp, 3.3157076459943413e-06_dp, 5.887003780590426e-06_dp, &
             1.0258227894310132e-05_dp, 1.757376476467849e-05_dp, 2.9642586769473303e-05_dp, &
             4.9292606971954226e-05_dp, 8.089932493130112e-05_dp, 0.00013116819235664838_dp, &
             0.0002102837957726176_dp, 0.00033358594922685116_dp, 0.000523998052618739_dp, &
             0.0008155232645797145_dp, 0.0012582481122005613_dp, 0.0019254630936995666_dp, &
             0.002923741605113227_dp, 0.004407133395375212_dp, 0.006597054937974856_dp, &
             0.009810033996700551_dp, 0.014496238952709236_dp, 0.021292761744946733_dp, &
             0.031097017627568595_dp, 0.04516750574389687_dp, 0.06526174374278014_dp, &
             0.09382480093035586_dp, 0.1342472183678273_dp, 0.1912198346613916_dp, &
             0.2712291189947634_dp, 0.3832704126526934_dp, 0.5399360021557283_dp, &
             0.759239386651285_dp, 1.0681078456170228_dp, 1.5102490671112527_dp, &
             2.167640813862452_dp, 3.2372723958661553_dp, 5.488213314912204_dp]

      CASE (1159)

         aw(:) = & ! a
            [5.83095795619701e-12_dp, 6.414640198802457e-11_dp, 3.394981852848648e-10_dp, &
             1.310339884893137e-09_dp, 4.196212386529782e-09_dp, 1.1838866743066459e-08_dp, &
             3.04317225236678e-08_dp, 7.278319892053902e-08_dp, 1.6427644129172042e-07_dp, &
             3.5345624088742896e-07_dp, 7.303919482560881e-07_dp, 1.4578690730440504e-06_dp, &
             2.8234005571592047e-06_dp, 5.324531148914614e-06_dp, 9.806674281330163e-06_dp, &
             1.768283477501853e-05_dp, 3.127955378753774e-05_dp, 5.4375423856440906e-05_dp, &
             9.303060802652057e-05_dp, 0.00015685223020378528_dp, 0.00026090724168729187_dp, &
             0.00042858967628173916_dp, 0.0006958844919453802_dp, 0.0011176611329265708_dp, &
             0.0017768979672202988_dp, 0.0027981130478390097_dp, 0.004366796795620498_dp, &
             0.00675736167901063_dp, 0.010373114744755977_dp, 0.01580311748242563_dp, &
             0.02390265312905428_dp, 0.03590654651861522_dp, 0.05358800602525562_dp, &
             0.07948028910223924_dp, 0.11718474947243103_dp, 0.17179728447252826_dp, &
             0.2504967123963026_dp, 0.3633545247326136_dp, 0.5244482213727782_dp, &
             0.7533951022164782_dp, 1.0774816621314605_dp, 1.5346750146083534_dp, &
             2.1780454650823105_dp, 3.082726247403624_dp, 4.3581828278145816_dp, &
             6.17379944016873_dp, 8.826687752031601_dp, 13.009821235576737_dp, &
             ! w
             1.864043316009394e-11_dp, 1.2247097811815585e-10_dp, 5.013516270451582e-10_dp, &
             1.6297979791142854e-09_dp, 4.588308732343392e-09_dp, 1.1677503770088829e-08_dp, &
             2.7548039706751194e-08_dp, 6.121622151285229e-08_dp, 1.295666443487286e-07_dp, &
             2.6330308285735986e-07_dp, 5.168612110330132e-07_dp, 9.846346344802236e-07_dp, &
             1.8271231788685233e-06_dp, 3.3124810351678103e-06_dp, 5.881694343843819e-06_dp, &
             1.0249637045223772e-05_dp, 1.7560078111827703e-05_dp, 2.9621092067986227e-05_dp, &
             4.925929766290742e-05_dp, 8.084834883401663e-05_dp, 0.00013109109316280157_dp, &
             0.00021016848048991112_dp, 0.0003334152974781827_dp, 0.0005237480639104012_dp, &
             0.0008151606178285686_dp, 0.0012577269891679392_dp, 0.0019247210969071825_dp, &
             0.002922694585306613_dp, 0.004405669000090817_dp, 0.006595024711286768_dp, &
             0.009807243915711014_dp, 0.014492438400604482_dp, 0.021287631124761203_dp, &
             0.031090155157002225_dp, 0.04515841449464533_dp, 0.065249820762397_dp, &
             0.09380933136953794_dp, 0.13422737894488831_dp, 0.19119471240090385_dp, &
             0.27119775253938294_dp, 0.3832318635004756_dp, 0.5398894597654474_dp, &
             0.7591843008541924_dp, 1.0680440569907392_dp, 1.5101768630192531_dp, &
             2.1675607477189542_dp, 3.237184385123853_dp, 5.488111936208487_dp]

      CASE (1160)

         aw(:) = & ! a
            [1.523981414165121e-11_dp, 1.177285543502411e-10_dp, 5.02495921291138e-10_dp, &
             1.7027746664359129e-09_dp, 4.995617007450077e-09_dp, 1.3222795338285072e-08_dp, &
             3.236058750734665e-08_dp, 7.441693568858156e-08_dp, 1.6262476074441392e-07_dp, &
             3.4052080774817276e-07_dp, 6.874740985991028e-07_dp, 1.3447387824530376e-06_dp, &
             2.558419766477054e-06_dp, 4.749244199875857e-06_dp, 8.62428791192768e-06_dp, &
             1.5353615854749356e-05_dp, 2.6846301596008116e-05_dp, 4.6177107271236106e-05_dp, &
             7.823950082326327e-05_dp, 0.0001307358315801109_dp, 0.00021566578333120443_dp, &
             0.0003515457670434021_dp, 0.0005666930842790061_dp, 0.0009040509663667878_dp, &
             0.0014282296284987266_dp, 0.002235715531956006_dp, 0.003469584877760849_dp, &
             0.005340586712056983_dp, 0.008157187935287706_dp, 0.01236816659556134_dp, &
             0.01862269399510427_dp, 0.027854683957089687_dp, 0.04140067335346312_dp, &
             0.061163850150915945_dp, 0.08984135586693329_dp, 0.13123805441260544_dp, &
             0.19069813260415866_dp, 0.2756969999907959_dp, 0.396651324978626_dp, &
             0.5680271504185594_dp, 0.809860052816623_dp, 1.1498592809344097_dp, &
             1.6263798261669995_dp, 2.2927920900555394_dp, 3.2243828862445305_dp, &
             4.530581983119371_dp, 6.380571162921642_dp, 9.071307100128411_dp, &
             13.296645403533567_dp, &
             ! w
             4.283761682516475e-11_dp, 1.9169292794133834e-10_dp, 6.603846972866373e-10_dp, &
             1.940090168769456e-09_dp, 5.094382412589533e-09_dp, 1.230995231706309e-08_dp, &
             2.7884048340062445e-08_dp, 5.9955385634952e-08_dp, 1.2346697295633655e-07_dp, &
             2.451338271146058e-07_dp, 4.7162189600217717e-07_dp, 8.827971780301206e-07_dp, &
             1.6128771053305136e-06_dp, 2.883753724360601e-06_dp, 5.0568633550088786e-06_dp, &
             8.71303613114445e-06_dp, 1.477417005235584e-05_dp, 2.4686891711284213e-05_dp, &
             4.069735515045661e-05_dp, 6.625904260368932e-05_dp, 0.00010663330479374017_dp, &
             0.0001697673726179182_dp, 0.0002675693416386917_dp, 0.0004177477615683761_dp, &
             0.0006464497789185825_dp, 0.000992022744166735_dp, 0.0015103484366159829_dp, &
             0.0022823680679471_dp, 0.003424645237061723_dp, 0.0051041231981649145_dp, &
             0.007558648815516776_dp, 0.011125393625228294_dp, 0.016280049187729823_dp, &
             0.023690672062707127_dp, 0.034291389503501205_dp, 0.04938297559311435_dp, &
             0.07076976780760581_dp, 0.10094587269342901_dp, 0.1433488392722309_dp, &
             0.2027076366361728_dp, 0.28552798597457335_dp, 0.40079251032884006_dp, &
             0.5610342409504953_dp, 0.7841495614997074_dp, 1.0968927355997744_dp, &
             1.5427766788426247_dp, 2.2036690275227504_dp, 3.2768570953255374_dp, &
             5.533829138592649_dp]

      CASE (1161)

         aw(:) = & ! a
            [9.171862453399814e-12_dp, 7.935956684239246e-11_dp, 3.686812358305204e-10_dp, &
             1.3193611550921274e-09_dp, 4.019018426570814e-09_dp, 1.0937127827099127e-08_dp, &
             2.7349179714388934e-08_dp, 6.399105699565373e-08_dp, 1.4185679716377734e-07_dp, &
             3.006461361090227e-07_dp, 6.13303889928754e-07_dp, 1.2105489644998209e-06_dp, &
             2.3215279265235778e-06_dp, 4.3401277197026575e-06_dp, 7.931595500855107e-06_dp, &
             1.420173221147577e-05_dp, 2.4962178588126345e-05_dp, 4.3141796146557506e-05_dp, &
             7.341811091509271e-05_dp, 0.00012317738284507135_dp, 0.0002039614717264376_dp, &
             0.00033363035976710366_dp, 0.000539569324280392_dp, 0.0008634108014204965_dp, &
             0.0013679386051822506_dp, 0.002147117659954436_dp, 0.0033405746094531488_dp, &
             0.005154382484403879_dp, 0.00789072833855923_dp, 0.011990036444772874_dp, &
             0.018090474819803682_dp, 0.02711161411216969_dp, 0.04037150094901301_dp, &
             0.0597497731945415_dp, 0.08791397710672046_dp, 0.12863234571060359_dp, &
             0.18720452194524248_dp, 0.2710528845163671_dp, 0.3905325988331259_dp, &
             0.5600407345452626_dp, 0.7995389201344407_dp, 1.1366620200046942_dp, &
             1.6096971918449237_dp, 2.271962848075372_dp, 3.198719843456602_dp, &
             4.499405206091266_dp, 6.343236657443881_dp, 9.027196757629973_dp, &
             13.244979475465904_dp, &
             ! w
             2.672480146217006e-11_dp, 1.375119346040148e-10_dp, 5.091967480791288e-10_dp, &
             1.5618709323521029e-09_dp, 4.2265241300656694e-09_dp, 1.0446792035536836e-08_dp, &
             2.4090117717939432e-08_dp, 5.2557464629571976e-08_dp, 1.0955729415795742e-07_dp, &
             2.197841939059749e-07_dp, 4.266622768751085e-07_dp, 8.049476605147709e-07_dp, &
             1.4809326974205131e-06_dp, 2.664399298630926e-06_dp, 4.6985311355854115e-06_dp, &
             8.136998041692244e-06_dp, 1.3861766880547802e-05_dp, 2.3261423145107227e-05_dp, &
             3.8498668726830146e-05_dp, 6.290824376792089e-05_dp, 0.00010158424191235406_dp, &
             0.00016224060863732307_dp, 0.00025646319341620494_dp, 0.00040151966644799404_dp, &
             0.0006229597533905544_dp, 0.0009583290879128704_dp, 0.0014624442205305225_dp, &
             0.0022148455278740713_dp, 0.003330274312470514_dp, 0.004973330632986942_dp, &
             0.007378886278848547_dp, 0.010880388402258847_dp, 0.01594894084996979_dp, &
             0.023247055534286264_dp, 0.033702312342746676_dp, 0.04860797912833535_dp, &
             0.06976012258566133_dp, 0.09964424231224721_dp, 0.1416897125645108_dp, &
             0.20061898469283224_dp, 0.2829346904009106_dp, 0.3976220859745893_dp, &
             0.5572250754032572_dp, 0.7796610467562776_dp, 1.0917148924353726_dp, &
             1.5369335507169304_dp, 2.1972030221561467_dp, 3.2697554739339356_dp, &
             5.525642967342611_dp]

      CASE (1162)

         aw(:) = & ! a
            [6.988633461017063e-12_dp, 6.516740463036952e-11_dp, 3.1714156007152943e-10_dp, &
             1.167106657176365e-09_dp, 3.6225713100246434e-09_dp, 9.993743313237771e-09_dp, &
             2.5253622943197303e-08_dp, 5.9584968371123234e-08_dp, 1.3300222035241e-07_dp, &
             2.835172490275868e-07_dp, 5.812356257418256e-07_dp, 1.1522007121095305e-06_dp, &
             2.218004946568103e-06_dp, 4.160539112359366e-06_dp, 7.626298190862354e-06_dp, &
             1.3692192750639602e-05_dp, 2.412594971544557e-05_dp, 4.1790516414487614e-05_dp, &
             7.126564773286385e-05_dp, 0.00011979419662679074_dp, 0.00019870991520215393_dp, &
             0.00032557386537793783_dp, 0.0005273461886420195_dp, 0.0008450604585872709_dp, &
             0.0013406648019506037_dp, 0.0021069687352416985_dp, 0.0032820161013247173_dp, &
             0.005069731579596858_dp, 0.007769413400021589_dp, 0.011817638356748211_dp, &
             0.01784750083784359_dp, 0.02677195041683166_dp, 0.039900488965505285_dp, &
             0.059101859697937936_dp, 0.0870299032358171_dp, 0.12743587102815332_dp, &
             0.18559873441777747_dp, 0.268916235641865_dp, 0.38771493605751534_dp, &
             0.5563597891634203_dp, 0.7947779461329826_dp, 1.1305695382506247_dp, &
             1.601989987686797_dp, 2.2623333110206985_dp, 3.186848085104524_dp, &
             4.484974584519207_dp, 6.325947267807059_dp, 9.006760992083032_dp, &
             13.221035214239166_dp, &
             ! w
             2.0910867621963518e-11_dp, 1.1687927641506689e-10_dp, 4.4911134628503026e-10_dp, &
             1.4077705875525946e-09_dp, 3.866724283651438e-09_dp, 9.664034059265516e-09_dp, &
             2.247918708465719e-08_dp, 4.9388633116119586e-08_dp, 1.0355503570794405e-07_dp, &
             2.0877576978744188e-07_dp, 4.070296775411347e-07_dp, 7.707867824236703e-07_dp, &
             1.4227825816901133e-06_dp, 2.567348145705426e-06_dp, 4.53942985836113e-06_dp, &
             7.880410476369572e-06_dp, 1.3454151744680313e-05_dp, 2.2622867803609213e-05_dp, &
             3.751126959035773e-05_dp, 6.139994211057342e-05_dp, 9.930656628665506e-05_dp, &
             0.00015883833616310157_dp, 0.00025143341458750524_dp, 0.0003941571194114239_dp, &
             0.0006122845925389939_dp, 0.0009429924932020589_dp, 0.0014406064895041177_dp, &
             0.0021840206164458194_dp, 0.0032871343568628146_dp, 0.00491346403600738_dp, &
             0.007296503856254431_dp, 0.010767974409474121_dp, 0.015796850247012387_dp, &
             0.023043066713898913_dp, 0.03343115725901334_dp, 0.048250892216337875_dp, &
             0.06929447821730451_dp, 0.09904338962760993_dp, 0.14092316622056036_dp, &
             0.19965318504894958_dp, 0.28173458971844423_dp, 0.39615380818989837_dp, &
             0.5554597588185199_dp, 0.7775795857433198_dp, 1.0893124663076827_dp, &
             1.534221274532354_dp, 2.194200744291355_dp, 3.2664576810370085_dp, &
             5.5218419185363805_dp]

      CASE (1163)

         aw(:) = & ! a
            [5.862367860061174e-12_dp, 5.771288440054508e-11_dp, 2.894288292142189e-10_dp, &
             1.0838759242014664e-09_dp, 3.403314862796861e-09_dp, 9.467457778850611e-09_dp, &
             2.4076633317832437e-08_dp, 5.7096657436302904e-08_dp, 1.2797885985977153e-07_dp, &
             2.73762183887244e-07_dp, 5.629116513891386e-07_dp, 1.1187632807109833e-06_dp, &
             2.158527021482171e-06_dp, 4.057122246631518e-06_dp, 7.45012943484979e-06_dp, &
             1.3397619945930963e-05_dp, 2.364169148038263e-05_dp, 4.1006775900545765e-05_dp, &
             7.00154340124065e-05_dp, 0.00011782654149308082_dp, 0.00019565186687022762_dp, &
             0.0003208770984316964_dp, 0.0005202127407812014_dp, 0.0008343404349471974_dp, &
             0.001324716824281362_dp, 0.0020834714133614093_dp, 0.0032477158464714316_dp, &
             0.005020108683782832_dp, 0.007698244564104432_dp, 0.011716429972187855_dp, &
             0.01770476370365552_dp, 0.026572283864376354_dp, 0.039623441629689615_dp, &
             0.058720537039489276_dp, 0.08650930078533924_dp, 0.12673093061990426_dp, &
             0.18465215305999697_dp, 0.26765610902053044_dp, 0.38605239753504933_dp, &
             0.5541869156965814_dp, 0.7919663437802664_dp, 1.1269701745004448_dp, &
             1.5974349542361899_dp, 2.256640167417194_dp, 3.179827048420054_dp, &
             4.476437758055296_dp, 6.315716686767486_dp, 8.99466606204883_dp, &
             13.206861359273379_dp, &
             ! w
             1.7905942964070176e-11_dp, 1.0585043311070365e-10_dp, 4.162532511094832e-10_dp, &
             1.3223933284722865e-09_dp, 3.6655606959295506e-09_dp, 9.223361826503501e-09_dp, &
             2.1567274700065202e-08_dp, 4.7586703040610984e-08_dp, 1.0012889883822415e-07_dp, &
             2.024715669109687e-07_dp, 3.9575475220120004e-07_dp, 7.511192923618664e-07_dp, &
             1.389229313540147e-06_dp, 2.511236874172281e-06_dp, 4.447277753607622e-06_dp, &
             7.73155012234136e-06_dp, 1.3217317081756622e-05_dp, 2.2251338737197652e-05_dp, &
             3.693604062113894e-05_dp, 6.052021091090344e-05_dp, 9.79766260840673e-05_dp, &
             0.00015684969530113947_dp, 0.00024849065317020523_dp, 0.0003898456246194355_dp, &
             0.0006060278968016463_dp, 0.0009339964873335879_dp, 0.00142778732602891_dp, &
             0.00216591271136559_dp, 0.0032617746327260606_dp, 0.004878248571606872_dp, &
             0.0072480136385777464_dp, 0.010701768248449801_dp, 0.015707225591983422_dp, &
             0.022922793760205434_dp, 0.03327119925722925_dp, 0.048040136439762245_dp, &
             0.06901951887913914_dp, 0.09868842727801344_dp, 0.14047011776031584_dp, &
             0.199082131342755_dp, 0.2810247137160017_dp, 0.39528497162728177_dp, &
             0.5544147841793988_dp, 0.7763470771737865_dp, 1.0878895112188054_dp, &
             1.5326144441014489_dp, 2.1924218463826914_dp, 3.2645035716423605_dp, &
             5.519589731273068_dp]

      CASE (1164)

         aw(:) = & ! a
            [5.188432830955208e-12_dp, 5.3193549984654216e-11_dp, 2.723638173684846e-10_dp, &
             1.0320783300807111e-09_dp, 3.265855376280847e-09_dp, 9.135707616022516e-09_dp, &
             2.3331551241038266e-08_dp, 5.551605850474286e-08_dp, 1.247789143111087e-07_dp, &
             2.675331534133193e-07_dp, 5.511868129368236e-07_dp, 1.0973293022880076e-06_dp, &
             2.120339930945644e-06_dp, 3.990630475075324e-06_dp, 7.336717485719508e-06_dp, &
             1.3207764318715481e-05_dp, 2.3329253350957878e-05_dp, 4.050062846044796e-05_dp, &
             6.92073186192276e-05_dp, 0.0001165536443376527_dp, 0.0001936720809328994_dp, &
             0.0003178342540846652_dp, 0.0005155882191396092_dp, 0.0008273864853203947_dp, &
             0.0013143655559925453_dp, 0.002068211794237141_dp, 0.003225429101426508_dp, &
             0.004987850259914315_dp, 0.007651958372355067_dp, 0.0116505780667326_dp, &
             0.017611852173980407_dp, 0.02644226392264725_dp, 0.0394429643708382_dp, &
             0.058472042197392646_dp, 0.08616992557767737_dp, 0.12627123671615448_dp, &
             0.18403469087187022_dp, 0.2668338727703851_dp, 0.38496727621423954_dp, &
             0.5527683175058673_dp, 0.7901302643187645_dp, 1.1246190805473206_dp, &
             1.594458931824005_dp, 2.2529197619751375_dp, 3.1752379732686724_dp, &
             4.470856945760245_dp, 6.309027575681953_dp, 8.986756947641247_dp, &
             13.197591818339466_dp, &
             ! w
             1.6105167642374804e-11_dp, 9.908387245267458e-11_dp, 3.9580025217924434e-10_dp, &
             1.2688101775451393e-09_dp, 3.5385849904858562e-09_dp, 8.943999061256616e-09_dp, &
             2.098717942238369e-08_dp, 4.643720413346551e-08_dp, 9.793809507074102e-08_dp, &
             1.9843224276859163e-07_dp, 3.885177858119185e-07_dp, 7.384758816105438e-07_dp, &
             1.3676296042690652e-06_dp, 2.4750709703108308e-06_dp, 4.387815823127027e-06_dp, &
             7.635399183745375e-06_dp, 1.3064200257015965e-05_dp, 2.20109350503899e-05_dp, &
             3.6563536176693046e-05_dp, 5.995010142970852e-05_dp, 9.711417193341837e-05_dp, &
             0.0001555592594828477_dp, 0.00024657994729890265_dp, 0.0003870446464577182_dp, &
             0.0006019610687319593_dp, 0.0009281462079050425_dp, 0.00141944684653765_dp, &
             0.0021541259807089843_dp, 0.0032452605767469282_dp, 0.004855307284521639_dp, &
             0.007216412306910782_dp, 0.010658605516052997_dp, 0.015648774822376556_dp, &
             0.02284432869074211_dp, 0.03316681046192026_dp, 0.0479025545750647_dp, &
             0.0688399717177289_dp, 0.09845657277648631_dp, 0.1401741148468701_dp, &
             0.19870893170001816_dp, 0.2805606749437246_dp, 0.3947168892417226_dp, &
             0.5537313868743774_dp, 0.7755408772977592_dp, 1.0869585800386095_dp, &
             1.5315630759577659_dp, 2.191257787980056_dp, 3.2632248137146105_dp, &
             5.518115950087038_dp]

      CASE (1165)

         aw(:) = & ! a
            [4.758352104187281e-12_dp, 5.02812549159909e-11_dp, 2.6124641451742635e-10_dp, &
             9.98089095055812e-10_dp, 3.1752057634478207e-09_dp, 8.916126609802078e-09_dp, &
             2.2836985676559106e-08_dp, 5.446449173058987e-08_dp, 1.226459594160127e-07_dp, &
             2.633744659616569e-07_dp, 5.433481492280093e-07_dp, 1.0829823156445831e-06_dp, &
             2.094751970210271e-06_dp, 3.946034272089805e-06_dp, 7.260587125119189e-06_dp, &
             1.3080221357724968e-05_dp, 2.311921379066298e-05_dp, 4.0160147777441165e-05_dp, &
             6.866338624281474e-05_dp, 0.00011569640616305737_dp, 0.00019233811118498265_dp, &
             0.00031578303572344296_dp, 0.000512469402181349_dp, 0.000822694750727548_dp, &
             0.001307378995675964_dp, 0.0020579086054435917_dp, 0.0032103760763651_dp, &
             0.00496605503746454_dp, 0.007620675748318246_dp, 0.011606058924449991_dp, &
             0.017549021970815707_dp, 0.026354316442310066_dp, 0.039320856097354584_dp, &
             0.058303873906019994_dp, 0.08594020163866666_dp, 0.12596000069641874_dp, &
             0.18361655037005498_dp, 0.266276949666286_dp, 0.3842321535908943_dp, &
             0.5518071037427774_dp, 0.7888859578599832_dp, 1.123025488012803_dp, &
             1.5924414446010193_dp, 2.250397282509168_dp, 3.172126111781486_dp, &
             4.467072131710448_dp, 6.304490653875316_dp, 8.981392071731873_dp, &
             13.191303684726394_dp, &
             ! w
             1.495458075340686e-11_dp, 9.46863489529958e-11_dp, 3.823774505468782e-10_dp, &
             1.2334492989452416e-09_dp, 3.454467308507025e-09_dp, 8.758390102960478e-09_dp, &
             2.0600873443662086e-08_dp, 4.5670265872093145e-08_dp, 9.647408839793046e-08_dp, &
             1.957293024760697e-07_dp, 3.836694239586327e-07_dp, 7.299967427055192e-07_dp, &
             1.3531307272704275e-06_dp, 2.4507744868530743e-06_dp, 4.347839195177835e-06_dp, &
             7.570712566311302e-06_dp, 1.2961125455084394e-05_dp, 2.184900862254973e-05_dp, &
             3.631250004835195e-05_dp, 5.956570914640979e-05_dp, 9.653240447076888e-05_dp, &
             0.00015468842862689295_dp, 0.00024529002487285866_dp, 0.00038515299372704373_dp, &
             0.0005992135540312242_dp, 0.0009241924980992221_dp, 0.0014138084530812705_dp, &
             0.0021461554578853636_dp, 0.003234090148623008_dp, 0.0048397851829363745_dp, &
             0.0071950253332997245_dp, 0.010629386940442595_dp, 0.015609197916793181_dp, &
             0.022791188254524884_dp, 0.03309609803586847_dp, 0.047809338170506394_dp, &
             0.06871829837753712_dp, 0.09829942264558247_dp, 0.13997344865033182_dp, &
             0.1984558882157345_dp, 0.28024598675956275_dp, 0.3943315832245498_dp, &
             0.5532677998102198_dp, 0.7749939136243683_dp, 1.086326919644846_dp, &
             1.530849628587728_dp, 2.1904678165519647_dp, 3.2623569752309867_dp, &
             5.517115782009485_dp]

      CASE (1166)

         aw(:) = & ! a
            [4.4831050229698e-12_dp, 4.840408976111793e-11_dp, 2.5402570704586307e-10_dp, &
             9.759035946465057e-10_dp, 3.115835497348452e-09_dp, 8.771953879925353e-09_dp, &
             2.251163443902008e-08_dp, 5.377163902229442e-08_dp, 1.2123879660836997e-07_dp, &
             2.606278930761494e-07_dp, 5.381663219900275e-07_dp, 1.0734903630031233e-06_dp, &
             2.077810829508294e-06_dp, 3.916489355610189e-06_dp, 7.210121887165675e-06_dp, &
             1.299563181771782e-05_dp, 2.297984460921216e-05_dp, 3.993412811470049e-05_dp, &
             6.830216633818416e-05_dp, 0.0001151269135445876_dp, 0.00019145160696240312_dp, &
             0.0003144194429699228_dp, 0.0005103954846121004_dp, 0.0008195740245386402_dp, &
             0.0013027306448354515_dp, 0.0020510519276356947_dp, 0.0032003561063815742_dp, &
             0.004951543982733616_dp, 0.007599843754253221_dp, 0.01157640652672864_dp, &
             0.01750716550262878_dp, 0.0262957168422734_dp, 0.0392394813223067_dp, &
             0.058191785974600604_dp, 0.08578706192959946_dp, 0.12575249252096168_dp, &
             0.1833377274170802_dp, 0.2659055344415112_dp, 0.38374183340387324_dp, &
             0.5511659047041083_dp, 0.788055819483674_dp, 1.1219622072937283_dp, &
             1.5910951935925686_dp, 2.248713894266044_dp, 3.170049215746313_dp, &
             4.464545901435586_dp, 6.301462212229204_dp, 8.977810756760842_dp, &
             13.187105851121398_dp, &
             ! w
             1.4217515743582697e-11_dp, 9.183484842963641e-11_dp, 3.7361541111806454e-10_dp, &
             1.2102792214698335e-09_dp, 3.39920446946211e-09_dp, 8.636208883335744e-09_dp, &
             2.0346180806924746e-08_dp, 4.5163973693938275e-08_dp, 9.550659043550767e-08_dp, &
             1.9394140696525158e-07_dp, 3.804598549925049e-07_dp, 7.243797055709557e-07_dp, &
             1.3435199234295413e-06_dp, 2.4346602197877686e-06_dp, 4.321311980334397e-06_dp, &
             7.52776896456871e-06_dp, 1.2892668416397367e-05_dp, 2.1741423989003436e-05_dp, &
             3.614565134992204e-05_dp, 5.9310142548242016e-05_dp, 9.614549284322455e-05_dp, &
             0.00015410910629403543_dp, 0.0002444316711639904_dp, 0.00038389391388323174_dp, &
             0.0005973843814809526_dp, 0.0009215597066811564_dp, 0.001410053030616751_dp, &
             0.002140845669736006_dp, 0.0032266472384865344_dp, 0.004829440861238963_dp, &
             0.00718077005656887_dp, 0.01060990838511558_dp, 0.015582809840452043_dp, &
             0.02275575131233444_dp, 0.03304893634978204_dp, 0.04774715887525724_dp, &
             0.06863712637301411_dp, 0.09819456970892051_dp, 0.139839544901586_dp, &
             0.1982870137666854_dp, 0.2800359492232611_dp, 0.39407438554581403_dp, &
             0.5529583188915541_dp, 0.7746287410617761_dp, 1.085905169573679_dp, &
             1.5303732427614596_dp, 2.189940312611069_dp, 3.2617774642535675_dp, &
             5.516447906671223_dp]

      CASE (1167)

         aw(:) = & ! a
            [4.322358769244208e-12_dp, 4.7302501958125866e-11_dp, 2.49767087642843e-10_dp, &
             9.627768773537843e-10_dp, 3.0806300041069364e-09_dp, 8.686323998079472e-09_dp, &
             2.2318154255055267e-08_dp, 5.335919977397635e-08_dp, 1.204004511373159e-07_dp, &
             2.5899042168673674e-07_dp, 5.35075125009334e-07_dp, 1.0678250118633307e-06_dp, &
             2.06769469756875e-06_dp, 3.898839824289732e-06_dp, 7.1799638493387856e-06_dp, &
             1.2945064205151672e-05_dp, 2.2896504406129192e-05_dp, 3.979893490511676e-05_dp, &
             6.808604806099442e-05_dp, 0.00011478610488494115_dp, 0.0001909209686788475_dp, &
             0.0003136030653066242_dp, 0.000509153601524139_dp, 0.000817704968024298_dp, &
             0.0012999462007952894_dp, 0.002046944008354121_dp, 0.0031943521258368823_dp, &
             0.004942847712817239_dp, 0.007587357768655942_dp, 0.011558631641505931_dp, &
             0.017482071967307947_dp, 0.026260581539965383_dp, 0.03919068509023194_dp, &
             0.05812456564734397_dp, 0.08569521326459385_dp, 0.12562802340257795_dp, &
             0.18317046648522053_dp, 0.26568270961128326_dp, 0.3834476488170515_dp, &
             0.5507811644262712_dp, 0.7875576715795971_dp, 1.1213241097235687_dp, &
             1.5902872245356239_dp, 2.2477035240891805_dp, 3.1688025896237098_dp, &
             4.463029490617433_dp, 6.2996442609016565_dp, 8.975660844915549_dp, &
             13.184585773042572_dp, &
             ! w
             1.3786770818654527e-11_dp, 9.015485727207108e-11_dp, 3.684308488192036e-10_dp, &
             1.1965357154458432e-09_dp, 3.36636923625314e-09_dp, 8.563520329952762e-09_dp, &
             2.0194505265350052e-08_dp, 4.4862216150664746e-08_dp, 9.492954870598448e-08_dp, &
             1.928744280630934e-07_dp, 3.7854347032646576e-07_dp, 7.210243503130763e-07_dp, &
             1.3377765861079689e-06_dp, 2.42502702029644e-06_dp, 4.305448737513217e-06_dp, &
             7.502081197321026e-06_dp, 1.2851708158453496e-05_dp, 2.1677036411586744e-05_dp, &
             3.6045772463794974e-05_dp, 5.915712305463385e-05_dp, 9.591378545199521e-05_dp, &
             0.00015376210733665903_dp, 0.00024391745119296896_dp, 0.000383139506114624_dp, &
             0.0005962882220943784_dp, 0.0009199817395732322_dp, 0.0014078019067019325_dp, &
             0.0021376623969913185_dp, 0.0032221845898357158_dp, 0.004823237848938145_dp, &
             0.007172220873266745_dp, 0.010598225451549138_dp, 0.01556698107338649_dp, &
             0.022734492551465126_dp, 0.033020641220933934_dp, 0.047709850424832535_dp, &
             0.06858841781550795_dp, 0.09813164578964184_dp, 0.13975918069150015_dp, &
             0.19818565362212823_dp, 0.2799098733586033_dp, 0.3939199909477246_dp, &
             0.5527725269526996_dp, 0.7744095028278992_dp, 1.0856519516001806_dp, &
             1.5300872114499544_dp, 2.189623583875716_dp, 3.261429512052241_dp, &
             5.516046913725189_dp]

      CASE (1168)

         aw(:) = & ! a
            [4.26265995582096e-12_dp, 4.689233940875616e-11_dp, 2.4817731167175107e-10_dp, &
             9.578684108498303e-10_dp, 3.067450677592579e-09_dp, 8.654241356485026e-09_dp, &
             2.224561692596476e-08_dp, 5.320449258175509e-08_dp, 1.200858504062368e-07_dp, &
             2.583757156060082e-07_dp, 5.339143304509146e-07_dp, 1.0656970050098415e-06_dp, &
             2.063893991868123e-06_dp, 3.8922073569766415e-06_dp, 7.168628686492001e-06_dp, &
             1.2926054650769463e-05_dp, 2.2865169954659465e-05_dp, 3.974809734724841e-05_dp, &
             6.80047690269957e-05_dp, 0.00011465791596027819_dp, 0.00019072135631057437_dp, &
             0.0003132959329688714_dp, 0.0005086863425399075_dp, 0.0008170016703757389_dp, &
             0.0012988983666402758_dp, 0.0020453980027807486_dp, 0.0031920923700733098_dp, &
             0.004939574407919484_dp, 0.007582657682654787_dp, 0.011551940229954987_dp, &
             0.017472624850136252_dp, 0.026247353170634796_dp, 0.03917231239750282_dp, &
             0.058099254614594587_dp, 0.08566062699796839_dp, 0.12558115145841323_dp, &
             0.18310747739307298_dp, 0.2655987920404102_dp, 0.3833368521368437_dp, &
             0.5506362568594046_dp, 0.7873700436847625_dp, 1.1210837614286617_dp, &
             1.5899828819319632_dp, 2.2473229301952857_dp, 3.168332987649695_dp, &
             4.462458246189539_dp, 6.298959407080432_dp, 8.974850917122454_dp, &
             13.183636371936872_dp, &
             ! w
             1.3626739988852615e-11_dp, 8.95280392795631e-11_dp, 3.6649214639315293e-10_dp, &
             1.1913899998409993e-09_dp, 3.3540645871609373e-09_dp, 8.536263077699327e-09_dp, &
             2.0137599153540295e-08_dp, 4.4748953402598255e-08_dp, 9.471288248435095e-08_dp, &
             1.9247367932808747e-07_dp, 3.778235013035161e-07_dp, 7.197634798525846e-07_dp, &
             1.3356179183963622e-06_dp, 2.421405656187152e-06_dp, 4.299484348454738e-06_dp, &
             7.492421442847529e-06_dp, 1.283630313057241e-05_dp, 2.1652817363236087e-05_dp, &
             3.60081991412364e-05_dp, 5.909955257455414e-05_dp, 9.582660143969664e-05_dp, &
             0.0001536315305229179_dp, 0.0002437239315120955_dp, 0.0003828555715339194_dp, &
             0.0005958756311950754_dp, 0.0009193877542101927_dp, 0.001406954469962047_dp, &
             0.0021364639737968746_dp, 0.0032205044089997255_dp, 0.00482090228602368_dp, &
             0.00716900174802886_dp, 0.010593826104031544_dp, 0.015561020261175021_dp, &
             0.022726486521439153_dp, 0.03300998481315461_dp, 0.04769579884147676_dp, &
             0.06857007182065945_dp, 0.09810794467359252_dp, 0.13972890933513316_dp, &
             0.19814747222384305_dp, 0.2798623801669748_dp, 0.393861828081088_dp, &
             0.552702533990695_dp, 0.7743269072433034_dp, 1.0855565519349926_dp, &
             1.5299794467273327_dp, 2.1895042505869613_dp, 3.261298410876307_dp, &
             5.515895822833872_dp]

      CASE (1169)

         aw(:) = & ! a
            [4.2623674707089334e-12_dp, 4.6890328237747354e-11_dp, 2.481695111360063e-10_dp, &
             9.578443166693538e-10_dp, 3.0673859671339273e-09_dp, 8.65408380210017e-09_dp, &
             2.224526065651569e-08_dp, 5.3203732659946515e-08_dp, 1.2008430497960548e-07_dp, &
             2.5837269580739137e-07_dp, 5.339086277449182e-07_dp, 1.0656865503877187e-06_dp, &
             2.0638753192738845e-06_dp, 3.892174772057942e-06_dp, 7.168572997824025e-06_dp, &
             1.2925961259714086e-05_dp, 2.286501601636541e-05_dp, 3.9747847601812214e-05_dp, &
             6.800436974761707e-05_dp, 0.00011465728626345944_dp, 0.00019072037580911868_dp, &
             0.0003132944244066568_dp, 0.0005086840476111823_dp, 0.0008169982163831886_dp, &
             0.0012988932209681606_dp, 0.002045390411337644_dp, 0.003192081274886607_dp, &
             0.004939558337947742_dp, 0.007582634610654461_dp, 0.011551907386963234_dp, &
             0.01747257848794227_dp, 0.026247288261738944_dp, 0.0391722222618557_dp, &
             0.05809913046335926_dp, 0.08566045738748274_dp, 0.12558092165367757_dp, &
             0.1831071686510088_dp, 0.26559838084033643_dp, 0.3833363094106689_dp, &
             0.550635547310563_dp, 0.787369125333085_dp, 1.121082585564548_dp, &
             1.5899813936820664_dp, 2.2473210698708623_dp, 3.1683306929473307_dp, &
             4.462455455073878_dp, 6.298956060517784_dp, 8.974846958612142_dp, &
             13.183631730555868_dp, &
             ! w
             1.3625955795261923e-11_dp, 8.952496403354892e-11_dp, 3.6648262974738074e-10_dp, &
             1.1913647333055983e-09_dp, 3.354004157355252e-09_dp, 8.53612919622454e-09_dp, &
             2.013731961782847e-08_dp, 4.4748396994182064e-08_dp, 9.471181805423774e-08_dp, &
             1.924717104872792e-07_dp, 3.7781996410699387e-07_dp, 7.197572851839901e-07_dp, &
             1.3356073128871137e-06_dp, 2.421387864695149e-06_dp, 4.299455046460214e-06_dp, &
             7.492373987504376e-06_dp, 1.2836227453170956e-05_dp, 2.16526983918797e-05_dp, &
             3.600801457873812e-05_dp, 5.909926980082793e-05_dp, 9.582617323828705e-05_dp, &
             0.00015363088924561722_dp, 0.000243722981192243_dp, 0.000382854177337907_dp, &
             0.0005958736054641463_dp, 0.0009193848381956949_dp, 0.0014069503102072192_dp, &
             0.0021364580919847513_dp, 0.0032204961639994336_dp, 0.0048208908268743255_dp, &
             0.007168985956781234_dp, 0.010593804527839886_dp, 0.015560991033910256_dp, &
             0.022726447276589824_dp, 0.0330099325923249_dp, 0.04769573000673085_dp, &
             0.06856998198457157_dp, 0.09810782866773426_dp, 0.1397287612470656_dp, &
             0.1981472855458768_dp, 0.27986214810743426_dp, 0.39386154407520313_dp, &
             0.552702192407736_dp, 0.7743265042288483_dp, 1.08555608623663_dp, &
             1.5299789201383425_dp, 2.189503666889874_dp, 3.2612977693063283_dp, &
             5.5158950823749855_dp]

      CASE (1170)

         aw(:) = & ! a
            [2.8971984233107593e-09_dp, 1.6018256317594296e-08_dp, 4.323626682301341e-08_dp, &
             9.294984189607577e-08_dp, 1.8193672998957262e-07_dp, 3.4010208586812965e-07_dp, &
             6.178742809226445e-07_dp, 1.0984011113712622e-06_dp, 1.916935612425083e-06_dp, &
             3.2908483995013476e-06_dp, 5.565497650098839e-06_dp, 9.283684750724219e-06_dp, &
             1.528993586061381e-05_dp, 2.4885826950975822e-05_dp, 4.005960128009102e-05_dp, &
             6.382320999198082e-05_dp, 0.00010070372170101055_dp, 0.00015745526553687545_dp, &
             0.00024408428150661206_dp, 0.00037531752872562395_dp, 0.0005726926377076732_dp, &
             0.0008675197902760785_dp, 0.0013050567705704594_dp, 0.0019503666562434986_dp, &
             0.0028964990704414494_dp, 0.004275867060776388_dp, 0.00627600188586692_dp, &
             0.009161282972284718_dp, 0.013302793718586762_dp, 0.019219189644552874_dp, &
             0.02763244097308416_dp, 0.039543601837076295_dp, 0.05633546032683864_dp, &
             0.07991116496153607_dp, 0.11288087211038206_dp, 0.1588123416646962_dp, &
             0.22256654047584748_dp, 0.3107461626971102_dp, 0.4322943007658676_dp, &
             0.5992936640638288_dp, 0.8280364684727842_dp, 1.1404673279227049_dp, &
             1.5661598422748992_dp, 2.145105254051381_dp, 2.9318531540344512_dp, &
             4.002181540940412_dp, 5.465195792237991_dp, 7.489142732916083_dp, &
             10.370579762277501_dp, 14.808542693028459_dp, &
             ! w
             7.516273330055954e-09_dp, 1.9275233015036655e-08_dp, 3.650119433506242e-08_dp, &
             6.564715689895416e-08_dp, 1.1719290151600152e-07_dp, 2.0734445611739548e-07_dp, &
             3.617427423691262e-07_dp, 6.213650840776349e-07_dp, 1.0511858574918623e-06_dp, &
             1.753068009726688e-06_dp, 2.8849971761320662e-06_dp, 4.689596878317835e-06_dp, &
             7.536085061279275e-06_dp, 1.1981563320910117e-05_dp, 1.8859941685505676e-05_dp, &
             2.9410138416844563e-05_dp, 4.545978856495287e-05_dp, 6.96869931540966e-05_dp, &
             0.00010599123824890137_dp, 0.00016001630176424464_dp, 0.00023988379550906778_dp, &
             0.000357217348192642_dp, 0.0005285661498788828_dp, 0.000777375051289844_dp, &
             0.0011366997874062896_dp, 0.0016529342863116053_dp, 0.0023909077836055522_dp, &
             0.0034408295587463004_dp, 0.004927717583243525_dp, 0.0070241559274600435_dp, &
             0.009967499630343036_dp, 0.01408300465110598_dp, 0.019814830394375857_dp, &
             0.027767477820977583_dp, 0.03876103524137987_dp, 0.05390467688402352_dp, &
             0.07469430950156407_dp, 0.10314229252113714_dp, 0.14195017147121716_dp, &
             0.1947402383666068_dp, 0.26637047703676026_dp, 0.3633748596052032_dp, &
             0.49460882693554753_dp, 0.6722682699419169_dp, 0.9136712810183666_dp, &
             1.2447898270747142_dp, 1.708338830529429_dp, 2.3858988797016165_dp, &
             3.4765805913919876_dp, 5.764517478181698_dp]

      CASE (1171)

         aw(:) = & ! a
            [2.020467640052466e-09_dp, 1.1227309110408129e-08_dp, 3.061604492664945e-08_dp, &
             6.684897828997804e-08_dp, 1.3332293634907182e-07_dp, 2.54093752343165e-07_dp, &
             4.7022049094737206e-07_dp, 8.503198015555523e-07_dp, 1.5074260709731065e-06_dp, &
             2.6253050163839787e-06_dp, 4.499020599583709e-06_dp, 7.5968679822237795e-06_dp, &
             1.265407616894579e-05_dp, 2.081337990019884e-05_dp, 3.383422980805425e-05_dp, &
             5.440182882781105e-05_dp, 8.658041223081727e-05_dp, 0.00013647367353363424_dp, &
             0.00021318093966311312_dp, 0.0003301732657867455_dp, 0.0005072626202625829_dp, &
             0.0007734045481213732_dp, 0.0011706665334663597_dp, 0.001759819248089712_dp, &
             0.002628177291695103_dp, 0.003900544878665931_dp, 0.005754429982017154_dp, &
             0.008441103734431936_dp, 0.012314634574862045_dp, 0.017871763467403382_dp, &
             0.02580646598091358_dp, 0.037084345440884556_dp, 0.053043718417856364_dp, &
             0.0755325199428255_dp, 0.107093143225409_dp, 0.15121126858118003_dp, &
             0.21264994992176062_dp, 0.29789718735401477_dp, 0.415764680255332_dp, &
             0.578188782215448_dp, 0.8013045614215859_dp, 1.1068961538969189_dp, &
             1.524384813348897_dp, 2.093633124571856_dp, 2.869104001622676_dp, &
             3.9265476099665846_dp, 5.375106542105872_dp, 7.383105718317672_dp, &
             10.247112910932296_dp, 14.665641796325803_dp, &
             ! w
             5.247488161953065e-09_dp, 1.359376031187571e-08_dp, 2.624655033303169e-08_dp, &
             4.839062169710201e-08_dp, 8.846508123020649e-08_dp, 1.5974842091531878e-07_dp, &
             2.836674670467907e-07_dp, 4.949428375002212e-07_dp, 8.492377583704704e-07_dp, &
             1.4346894667201564e-06_dp, 2.3892643655718615e-06_dp, 3.9266764624159165e-06_dp, &
             6.374781637514752e-06_dp, 1.0232020696879596e-05_dp, 1.6249805226819115e-05_dp, &
             2.5551965370006366e-05_dp, 3.980683783589679e-05_dp, 6.147370953957672e-05_dp, &
             9.415374156849576e-05_dp, 0.00014308697239821477_dp, 0.00021585259251451536_dp, &
             0.00032335079175164805_dp, 0.00048117295104547835_dp, 0.0007115052124551078_dp, &
             0.0010457617078107712_dp, 0.0015282121363938384_dp, 0.0022209594195582855_dp, &
             0.0032107439365087346_dp, 0.004618210621213312_dp, 0.006610485994204436_dp, &
             0.00941818961203052_dp, 0.013358368788898735_dp, 0.018865323405030155_dp, &
             0.026531914662163154_dp, 0.037164776768565055_dp, 0.05185794501849601_dp, &
             0.07209089087003548_dp, 0.09985901567805544_dp, 0.13784769795335552_dp, &
             0.1896658754303073_dp, 0.2601638527649279_dp, 0.35587729691101555_dp, &
             0.4856769228868462_dp, 0.6617912253383094_dp, 0.9015894960986276_dp, &
             1.2311085631834784_dp, 1.6931239296481257_dp, 2.3692252191765855_dp, &
             3.4583363384533907_dp, 5.743406949498808_dp]

      CASE (1172)

         aw(:) = & ! a
            [1.5654654091958936e-09_dp, 8.733657541552575e-09_dp, 2.4009123525201713e-08_dp, &
             5.3057765370203266e-08_dp, 1.0732249153425836e-07_dp, 2.074726209224575e-07_dp, &
             3.8911278790635127e-07_dp, 7.123245724166093e-07_dp, 1.2769712044539662e-06_dp, &
             2.2467271966204283e-06_dp, 3.886352262754194e-06_dp, 6.61892548177663e-06_dp, &
             1.1112902846674573e-05_dp, 1.8413372052396573e-05_dp, 3.0138309017414293e-05_dp, &
             4.876975442989987e-05_dp, 7.80826686404344e-05_dp, 0.00012377221640541528_dp, &
             0.0001943653240365529_dp, 0.00030253717003741895_dp, 0.00046700136295859323_dp, &
             0.0007152086998147802_dp, 0.001087179984541876_dp, 0.0016409219463219581_dp, &
             0.002460043162020096_dp, 0.0036644141136555394_dp, 0.005425022022175467_dp, &
             0.0079845831320494_dp, 0.011686027174898467_dp, 0.01701170607001819_dp, &
             0.024637160757279262_dp, 0.03550458381245833_dp, 0.050922842321847454_dp, &
             0.07270320792568294_dp, 0.10334295397570854_dp, 0.14627295825442901_dp, &
             0.20619071876858838_dp, 0.2895072258633407_dp, 0.4049456966798739_dp, &
             0.5643436187382798_dp, 0.7837295428260371_dp, 1.0847784097162154_dp, &
             1.4968073242185163_dp, 2.059590349052438_dp, 2.8275300674441066_dp, &
             3.8763562715890534_dp, 5.315235868977332_dp, 7.312547126019548_dp, &
             10.164867743878013_dp, 14.570366677795155_dp, &
             ! w
             4.069275070874611e-09_dp, 1.0626028840153552e-08_dp, 2.0828638328653688e-08_dp, &
             3.912198092343268e-08_dp, 7.27607862272369e-08_dp, 1.3331883256629878e-07_dp, &
             2.3972690411598166e-07_dp, 4.2295055771061964e-07_dp, 7.330161792294444e-07_dp, &
             1.249691474107462e-06_dp, 2.098654355373026e-06_dp, 3.4757623244132207e-06_dp, &
             5.683172528470769e-06_dp, 9.182675391377619e-06_dp, 1.4673866283239451e-05_dp, &
             2.3207930446604414e-05_dp, 3.635218158924205e-05_dp, 5.6426521564584644e-05_dp, &
             8.684127020458473e-05_dp, 0.0001325771393121017_dp, 0.00020086355088729348_dp, &
             0.0003021324371185639_dp, 0.00045135296731425883_dp, 0.0006698907497687766_dp, &
             0.0009880861805908246_dp, 0.0014488150173013717_dp, 0.002112385460969776_dp, &
             0.0030632472379704062_dp, 0.0044191487872293515_dp, 0.006343590909230722_dp, &
             0.00906270500594333_dp, 0.012888052026687906_dp, 0.018247321568932242_dp, &
             0.025725545439724603_dp, 0.036120276117660534_dp, 0.05051528994029317_dp, &
             0.070378870657254_dp, 0.09769481132672062_dp, 0.1351373363902397_dp, &
             0.1863060427315367_dp, 0.25604564775417626_dp, 0.3508925145431276_dp, &
             0.47972724620169166_dp, 0.6548000642272692_dp, 0.8935148374323146_dp, &
             1.22195266139964_dp, 1.6829310241563493_dp, 2.358047450613727_dp, &
             3.4461027736215204_dp, 5.729255925549111_dp]

      CASE (1173)

         aw(:) = & ! a
            [1.2848413584997882e-09_dp, 7.19214133058497e-09_dp, 1.990605071014294e-08_dp, &
             4.44302561399753e-08_dp, 9.090329208782599e-08_dp, 1.777307101936104e-07_dp, &
             3.368541870249681e-07_dp, 6.225844591015914e-07_dp, 1.125818837492098e-06_dp, &
             1.9964707359966657e-06_dp, 3.4784266494167983e-06_dp, 5.963459202346142e-06_dp, &
             1.007356851360138e-05_dp, 1.6785600283198396e-05_dp, 2.7618244020302116e-05_dp, &
             4.4910400927415415e-05_dp, 7.22324619965349e-05_dp, 0.00011498966764116697_dp, &
             0.00018130135823660743_dp, 0.00028327418582150247_dp, 0.0004388347758590685_dp, &
             0.0006743526637272816_dp, 0.001028373968274981_dp, 0.0015569087398530194_dp, &
             0.0023408816049536344_dp, 0.0034965813282278354_dp, 0.005190250222190789_dp, &
             0.0076583653556052565_dp, 0.011235714997362227_dp, 0.016394111317041212_dp, &
             0.023795563848002854_dp, 0.03436504582252928_dp, 0.04938971952849293_dp, &
             0.07065378144378935_dp, 0.10062112004981584_dp, 0.14268198621668113_dp, &
             0.2014851861317065_dp, 0.28338439925588926_dp, 0.39703686035856084_dp, &
             0.5542061725477722_dp, 0.7708410977952374_dp, 1.0685345653726666_dp, &
             1.4765252078273992_dp, 2.034520096499247_dp, 2.7968757960512156_dp, &
             3.8393060313060765_dp, 5.270995394776077_dp, 7.260362225293443_dp, &
             10.103993021012_dp, 14.49980307899297_dp, &
             ! w
             3.3422348893791125e-09_dp, 8.786277273234188e-09_dp, 1.7439601165298908e-08_dp, &
             3.324980248463564e-08_dp, 6.267920972403724e-08_dp, 1.1615609708761664e-07_dp, &
             2.1091226283198226e-07_dp, 3.753340406261468e-07_dp, 6.555526120887345e-07_dp, &
             1.125522417795889e-06_dp, 1.902344731879224e-06_dp, 3.1693569335206797e-06_dp, &
             5.21062026388393e-06_dp, 8.462015655477187e-06_dp, 1.358637220319571e-05_dp, &
             2.1583146575369965e-05_dp, 3.3947470774765226e-05_dp, 5.289933074217751e-05_dp, &
             8.171182151767837e-05_dp, 0.00012517863913112458_dp, 0.0001902762729494208_dp, &
             0.00028709716175353076_dp, 0.0004301582368794696_dp, 0.0006402272598056988_dp, &
             0.0009468604144969061_dp, 0.0013919129993427701_dp, 0.0020343763607106533_dp, &
             0.002957015718316692_dp, 0.004275444754752398_dp, 0.006150486163115107_dp, &
             0.008804948871088894_dp, 0.012546323357120805_dp, 0.017797383901694972_dp, &
             0.025137327361273006_dp, 0.035356922242529606_dp, 0.04953226207330944_dp, &
             0.06912322714677144_dp, 0.09610485798928665_dp, 0.13314292406218137_dp, &
             0.18382987685531205_dp, 0.253006051737489_dp, 0.347208099636934_dp, &
             0.4753237890911045_dp, 0.6496194193143515_dp, 0.8875246823906293_dp, &
             1.215153966686021_dp, 1.6753566484876994_dp, 2.3497370465896026_dp, &
             3.4370055705630165_dp, 5.718734775936481_dp]

      CASE (1174)

         aw(:) = & ! a
            [1.0935995702844266e-09_dp, 6.139582756656664e-09_dp, 1.7093565949152064e-08_dp, &
             3.848012755939702e-08_dp, 7.94912182111222e-08_dp, 1.5688766694632112e-07_dp, &
             2.9993914856150185e-07_dp, 5.587236122447681e-07_dp, 1.017526984601223e-06_dp, &
             1.8160660873905199e-06_dp, 3.1826924762075238e-06_dp, 5.485784306982321e-06_dp, &
             9.312496206344414e-06_dp, 1.5588307839034354e-05_dp, 2.5756929488147883e-05_dp, &
             4.2048831725144e-05_dp, 6.787899067105368e-05_dp, 0.00010843178455852965_dp, &
             0.0001715152573561116_dp, 0.0002688007844930578_dp, 0.000417611057388918_dp, &
             0.0006434839971626446_dp, 0.0009838290931299095_dp, 0.0014931142772939224_dp, &
             0.0022501871812893102_dp, 0.0033685596200221843_dp, 0.005010788477389797_dp, &
             0.007408496355290323_dp, 0.010890126338978371_dp, 0.015919262961160682_dp, &
             0.02314733615726362_dp, 0.033485835561930605_dp, 0.04820490211644359_dp, &
             0.06906746802333394_dp, 0.09851116324667507_dp, 0.1398942363967866_dp, &
             0.1978270806513849_dp, 0.27861809313037844_dp, 0.3908722916175211_dp, &
             0.546294710377221_dp, 0.7607707464883958_dp, 1.0558280642810853_dp, &
             1.460642691371729_dp, 2.0148680712348273_dp, 2.772823708612851_dp, &
             3.810210076137468_dp, 5.236225511353124_dp, 7.219320300915957_dp, &
             10.056088835937166_dp, 14.444247813761926_dp, &
             ! w
             2.8465526586654025e-09_dp, 7.527113180975546e-09_dp, 1.510248349426318e-08_dp, &
             2.9157499057691058e-08_dp, 5.557870540083568e-08_dp, 1.0395806836051348e-07_dp, &
             1.9027422375447505e-07_dp, 3.4099830322016523e-07_dp, 5.993556039001359e-07_dp, &
             1.0349459878534372e-06_dp, 1.7584219346075544e-06_dp, 2.943673914331797e-06_dp, &
             4.861063963550479e-06_dp, 7.926796878656605e-06_dp, 1.277569766577836e-05_dp, &
             2.0367713250774472e-05_dp, 3.214270113030735e-05_dp, 5.0243947155235576e-05_dp, &
             7.78389604191002e-05_dp, 0.0001195772179718874_dp, 0.0001822397349123_dp, &
             0.00027565602617697497_dp, 0.0004139920758912902_dp, 0.0006175508129473733_dp, &
             0.0009152775690070634_dp, 0.0013482313584308148_dp, 0.001974374190235159_dp, &
             0.0028751523570634888_dp, 0.004164505339838644_dp, 0.006001152117803911_dp, &
             0.008605287565193418_dp, 0.01228119444243178_dp, 0.017447767099642575_dp, &
             0.02467958840568318_dp, 0.03476205213375144_dp, 0.04876515205850083_dp, &
             0.06814208047796765_dp, 0.09486089160623822_dp, 0.13158057871500253_dp, &
             0.1818878271027352_dp, 0.2506193700802144_dp, 0.34431194135343374_dp, &
             0.4718588486232527_dp, 0.6455390366806396_dp, 0.8828026621024403_dp, &
             1.2097906501532938_dp, 1.6693779987163408_dp, 2.3431749879404555_dp, &
             3.429821302238789_dp, 5.710427402984122_dp]

      CASE (1175)

         aw(:) = & ! a
            [9.544679365794432e-10_dp, 5.372539948764549e-09_dp, 1.5037059601143117e-08_dp, &
             3.410617904833659e-08_dp, 7.104627971922534e-08_dp, 1.4135614990748465e-07_dp, &
             2.722478851102195e-07_dp, 5.105249063342137e-07_dp, 9.353359629740198e-07_dp, &
             1.6784433485588236e-06_dp, 2.956035621961025e-06_dp, 5.118113342353053e-06_dp, &
             8.724373740415704e-06_dp, 1.465970515280082e-05_dp, 2.4308407653514743e-05_dp, &
             3.981482018117794e-05_dp, 6.44701733780019e-05_dp, 0.00010328259670501849_dp, &
             0.0001638111943862729_dp, 0.0002573785604928223_dp, 0.0004008225422931376_dp, &
             0.0006190122241539617_dp, 0.0009484413666191151_dp, 0.0014423334370367343_dp, &
             0.002177857312050465_dp, 0.0032662768926184946_dp, 0.00486716137845716_dp, &
             0.0072081924223179215_dp, 0.01061265493695416_dp, 0.015537436134726865_dp, &
             0.022625343047039684_dp, 0.03277686170197073_dp, 0.047248225289081826_dp, &
             0.06778497042498016_dp, 0.09680322010334876_dp, 0.13763498018998002_dp, &
             0.1948591061403351_dp, 0.2747467718141893_dp, 0.38586001664742803_dp, &
             0.5398556036877897_dp, 0.7525666296813017_dp, 1.0454668013489523_dp, &
             1.4476802848165953_dp, 1.9988159787345587_dp, 2.753162459148784_dp, &
             3.7864089327444557_dp, 5.207764878685831_dp, 7.185707028741695_dp, &
             10.016836883494937_dp, 14.39870908977672_dp, &
             ! w
             2.485800675562579e-09_dp, 6.607626062249647e-09_dp, 1.3384600016126528e-08_dp, &
             2.6122335177256528e-08_dp, 5.0265661320542476e-08_dp, 9.47619293806772e-08_dp, &
             1.74615678681107e-07_dp, 3.1480137821851603e-07_dp, 5.562649086461555e-07_dp, &
             9.65178945099696e-07_dp, 1.6471047919886778e-06_dp, 2.7684531332747186e-06_dp, &
             4.588709888773589e-06_dp, 7.508418701062635e-06_dp, 1.2140061440305523e-05_dp, &
             1.9411990898024066e-05_dp, 3.0719769718905206e-05_dp, 4.814509501596982e-05_dp, &
             7.47705295415567e-05_dp, 0.00011512931053296784_dp, 0.0001758446262598965_dp, &
             0.0002665333502780842_dp, 0.00040107717990732044_dp, 0.0005994019214037374_dp, &
             0.0008899565766803314_dp, 0.0013131522270637044_dp, 0.0019261121347955053_dp, &
             0.0028092063982820105_dp, 0.004075006570287405_dp, 0.005880510472801458_dp, &
             0.008443771449664175_dp, 0.012066440874068196_dp, 0.017164225950989688_dp, &
             0.024307914076785418_dp, 0.03427847114358359_dp, 0.04814085750256939_dp, &
             0.06734273620850309_dp, 0.09384637215315099_dp, 0.13030512534984537_dp, &
             0.1803008614854679_dp, 0.24866726028447472_dp, 0.3419410275652344_dp, &
             0.4690199431470963_dp, 0.6421933079763037_dp, 0.8789281396699508_dp, &
             1.2053873280979708_dp, 1.6644672130491116_dp, 2.3377833699292054_dp, &
             3.42391782589973_dp, 5.7036019868411305_dp]

      CASE (1176)

         aw(:) = & ! a
            [8.484592239796619e-10_dp, 4.7872314633360295e-09_dp, 1.3463066223313578e-08_dp, &
             3.0742656312585494e-08_dp, 6.451428406242278e-08_dp, 1.2927019790410144e-07_dp, &
             2.505762395754886e-07_dp, 4.726054784971917e-07_dp, 8.70364640836279e-07_dp, &
             1.5691807480047864e-06_dp, 2.775372200428833e-06_dp, 4.823984286787068e-06_dp, &
             8.252311796067599e-06_dp, 1.3912044401544471e-05_dp, 2.313878203975687e-05_dp, &
             3.80061095325226e-05_dp, 6.170340241174754e-05_dp, 9.90934470603192e-05_dp, &
             0.0001575296958082098_dp, 0.0002480461259072405_dp, 0.0003870787349000641_dp, &
             0.0005989414176497636_dp, 0.0009193667024520808_dp, 0.0014005421169081763_dp, &
             0.002118237222150866_dp, 0.0031818397958488153_dp, 0.004748422439409158_dp, &
             0.007042369737778079_dp, 0.010382646255800859_dp, 0.015220523499434547_dp, &
             0.02219157041312458_dp, 0.032187026879985425_dp, 0.04645142763333492_dp, &
             0.06671566226159496_dp, 0.09537772683555473_dp, 0.13574748672885315_dp, &
             0.19237716291304885_dp, 0.2715064610085603_dp, 0.3816610484611753_dp, &
             0.5344567914546203_dp, 0.7456824226812692_dp, 1.0367658164981954_dp, &
             1.4367870145748476_dp, 1.985316931490584_dp, 2.7366176668647855_dp, &
             3.7663686378901926_dp, 5.183788628269622_dp, 7.157376848606006_dp, &
             9.98374123363817_dp, 14.360300249878337_dp, &
             ! w
             2.2108412846680144e-09_dp, 5.904712467727348e-09_dp, 1.2063660499149484e-08_dp, &
             2.3770061054498662e-08_dp, 4.6116598205844305e-08_dp, 8.75343458059843e-08_dp, &
             1.622421097499549e-07_dp, 2.940018216472538e-07_dp, 5.219068098739647e-07_dp, &
             9.093364872941888e-07_dp, 1.5576917839919737e-06_dp, 2.627256082324202e-06_dp, &
             4.368585316848802e-06_dp, 7.169336653546598e-06_dp, 1.1623570200529639e-05_dp, &
             1.8633540084682295e-05_dp, 2.9558153088549894e-05_dp, 4.642805237548734e-05_dp, &
             7.22552707082319e-05_dp, 0.00011147638431167306_dp, 0.0001705831456792932_dp, &
             0.000259015095403091_dp, 0.0003904165293023874_dp, 0.0005843979183068667_dp, &
             0.0008689927328873541_dp, 0.0012840689416896624_dp, 0.0018860459221788166_dp, &
             0.0027543895825991255_dp, 0.004000520882495375_dp, 0.005779988331591739_dp, &
             0.008309039987808162_dp, 0.011887106526311914_dp, 0.016927202390303965_dp, &
             0.023996904491766717_dp, 0.033873428648811_dp, 0.04761746649648468_dp, &
             0.06667198456760932_dp, 0.09299432087766428_dp, 0.129233028737685_dp, &
             0.1789658432914383_dp, 0.24702379756180656_dp, 0.33994350602516027_dp, &
             0.46662647133931573_dp, 0.6393707210170568_dp, 0.8756575531278759_dp, &
             1.2016685318268894_dp, 1.6603182352885655_dp, 2.333226988861463_dp, &
             3.418928405466185_dp, 5.697833967072847_dp]

      CASE (1177)

         aw(:) = & ! a
            [7.648570176559014e-10_dp, 4.325011778282143e-09_dp, 1.2216683011129698e-08_dp, &
             2.8067861627191858e-08_dp, 5.92927508884103e-08_dp, 1.1955738274874954e-07_dp, &
             2.3307236853845445e-07_dp, 4.418378521489627e-07_dp, 8.17427706766632e-07_dp, &
             1.4798198729772703e-06_dp, 2.627106738126745e-06_dp, 4.581839115290662e-06_dp, &
             7.862553325754246e-06_dp, 1.3293082241118532e-05_dp, 2.216808196922212e-05_dp, &
             3.650154318959508e-05_dp, 5.939690439768001e-05_dp, 9.55941247535251e-05_dp, &
             0.00015227259749619492_dp, 0.00024022166745041212_dp, 0.00037553625503459586_dp, &
             0.0005820584117007001_dp, 0.0008948729041038866_dp, 0.0013652846886743075_dp, &
             0.0020678698166658634_dp, 0.003110414123743377_dp, 0.004647856228347492_dp, &
             0.006901759666320384_dp, 0.010187388785633895_dp, 0.014951200879609277_dp, &
             0.02182255412306927_dp, 0.03168474743802122_dp, 0.04577225960108093_dp, &
             0.06580337875028064_dp, 0.09416049223164195_dp, 0.1341343843846544_dp, &
             0.19025430239294736_dp, 0.26873278489824026_dp, 0.37806407040720974_dp, &
             0.5298286523888526_dp, 0.7397768521108868_dp, 1.0292968147277115_dp, &
             1.4274302793414164_dp, 1.9737151254627125_dp, 2.722390349421421_dp, &
             3.749126716968171_dp, 5.163150953154221_dp, 7.1329817578238215_dp, &
             9.955232932784092_dp, 14.32720600022558_dp, &
             ! w
             1.9939335977875353e-09_dp, 5.348705073041545e-09_dp, 1.1013271477065373e-08_dp, &
             2.188641075526738e-08_dp, 4.277189455827029e-08_dp, 8.167534206297861e-08_dp, &
             1.5216407546044626e-07_dp, 2.7699092869560053e-07_dp, 4.937032889570478e-07_dp, &
             8.633439098702137e-07_dp, 1.4838254815335317e-06_dp, 2.5102830523010176e-06_dp, &
             4.185754438358128e-06_dp, 6.88702759543274e-06_dp, 1.1192597581054517e-05_dp, &
             1.798263089134329e-05_dp, 2.8584963917258433e-05_dp, 4.498690185372185e-05_dp, &
             7.014052379344651e-05_dp, 0.00010840013113682017_dp, 0.0001661454610998152_dp, &
             0.00025266474075002435_dp, 0.00038139946295831707_dp, 0.0005716904201172243_dp, &
             0.0008512153336962456_dp, 0.0012593767139473337_dp, 0.0018519900851579976_dp, &
             0.0027077448975005984_dp, 0.0039370731021902625_dp, 0.005694276243754277_dp, &
             0.008194047666793852_dp, 0.011733903758161263_dp, 0.01672453549677282_dp, &
             0.023730746834438418_dp, 0.03352651113196975_dp, 0.047168824837974827_dp, &
             0.06609658355932758_dp, 0.09226284772682646_dp, 0.1283119880275131_dp, &
             0.1778181319283108_dp, 0.24560997889664674_dp, 0.33822401399918084_dp, &
             0.4645649107611907_dp, 0.6369382091914111_dp, 0.8728375543869173_dp, &
             1.1984607078188612_dp, 1.656738135436543_dp, 2.3292944674562635_dp, &
             3.414621774835842_dp, 5.692855765976414_dp]

      CASE (1178)

         aw(:) = & ! a
            [6.971427207957829e-10_dp, 3.950169233397704e-09_dp, 1.120338081261412e-08_dp, &
             2.5884806331687886e-08_dp, 5.501109451545534e-08_dp, 1.1155478912434455e-07_dp, &
             2.1858596777437735e-07_dp, 4.162703951202757e-07_dp, 7.732756147834929e-07_dp, &
             1.405039231969556e-06_dp, 2.50265497992533e-06_dp, 4.378020598102278e-06_dp, &
             7.533646789207944e-06_dp, 1.276952414651262e-05_dp, 2.1345205312505807e-05_dp, &
             3.522350543065931e-05_dp, 5.743395461356597e-05_dp, 9.261072706966718e-05_dp, &
             0.00014778310675686593_dp, 0.0002335291987821809_dp, 0.0003656490493785139_dp, &
             0.000567576316350868_dp, 0.0008738344988964744_dp, 0.0013349630330995502_dp, &
             0.0020245017165548365_dp, 0.0030488441828247113_dp, 0.004561072760367001_dp, &
             0.006780294861278975_dp, 0.0100185505125243_dp, 0.01471809848872215_dp, &
             0.021502876137418147_dp, 0.03124924598854935_dp, 0.04518289624631197_dp, &
             0.06501109097699859_dp, 0.09310255241668716_dp, 0.13273134743224493_dp, &
             0.18840658011867867_dp, 0.26631695137021877_dp, 0.37492910059240697_dp, &
             0.5257924332218084_dp, 0.7346234815854413_dp, 1.022775407401825_dp, &
             1.4192561816732332_dp, 1.9635745121011048_dp, 2.709948935745093_dp, &
             3.7340424757713677_dp, 5.145088758342605_dp, 7.111623649107541_dp, &
             9.930266330604871_dp, 14.29821616213302_dp, &
             ! w
             1.818199253647581e-09_dp, 4.8971244866689115e-09_dp, 1.0156040212268246e-08_dp, &
             2.0339383061164458e-08_dp, 4.0008559129406523e-08_dp, 7.681075547168506e-08_dp, &
             1.4376149540845354e-07_dp, 2.627561911884194e-07_dp, 4.700255156117404e-07_dp, &
             8.246177252505131e-07_dp, 1.421462075883971e-06_dp, 2.4112820937152264e-06_dp, &
             4.030662192954686e-06_dp, 6.647045189860147e-06_dp, 1.0825523127389315e-05_dp, &
             1.7427215535020944e-05_dp, 2.7753129371069887e-05_dp, 4.375309932249962e-05_dp, &
             6.83273046788897e-05_dp, 0.00010575874183174276_dp, 0.00016232996066544502_dp, &
             0.00024719776622827316_dp, 0.0003736273303989269_dp, 0.0005607247532524899_dp, &
             0.0008358578689202736_dp, 0.0012380233658222492_dp, 0.0018225098190965672_dp, &
             0.0026673285249526725_dp, 0.0038820469129028264_dp, 0.005619875612573132_dp, &
             0.008094146696627356_dp, 0.011600699055363968_dp, 0.01654818589201557_dp, &
             0.02349897703616451_dp, 0.03322419697663573_dp, 0.04677759225174249_dp, &
             0.06559447398011634_dp, 0.09162413083235546_dp, 0.1275072383744549_dp, &
             0.17681472675760665_dp, 0.24437321084673458_dp, 0.33671902363986606_dp, &
             0.4627595880305447_dp, 0.6348070195849047_dp, 0.8703658118847004_dp, &
             1.1956479978529793_dp, 1.6535980936631367_dp, 2.325844658971084_dp, &
             3.4108434965605143_dp, 5.688488626713276_dp]

      CASE (1179)

         aw(:) = & ! a
            [3.797656788518067e-10_dp, 2.1855796884320484e-09_dp, 6.390565194705541e-09_dp, &
             1.537336716242537e-08_dp, 3.405885891962638e-08_dp, 7.175674243545133e-08_dp, &
             1.454552300016464e-07_dp, 2.8543445453132507e-07_dp, 5.445480744712429e-07_dp, &
             1.0133146329212295e-06_dp, 1.8441142592424734e-06_dp, 3.2894910188104693e-06_dp, &
             5.762073692083049e-06_dp, 9.927302971926705e-06_dp, 1.684548887356859e-05_dp, &
             2.8187414464673295e-05_dp, 4.6558718491510744e-05_dp, 7.59839688794241e-05_dp, &
             0.00012262352442432534_dp, 0.00019582749865999677_dp, 0.00030967483364911617_dp, &
             0.00048520634356167175_dp, 0.0007536449327497368_dp, 0.001161012569219447_dp, &
             0.0017747134696766724_dp, 0.0026928716744231795_dp, 0.0040575092291607094_dp, &
             0.006073055761066172_dp, 0.009032227409792759_dp, 0.013352050521547393_dp, &
             0.019623796083314316_dp, 0.028681917233414254_dp, 0.0416988529903872_dp, &
             0.06031491948574401_dp, 0.08681564426238826_dp, 0.12437306209610567_dp, &
             0.17737302839731184_dp, 0.2518580159826437_dp, 0.35612491542598684_dp, &
             0.5015313918463108_dp, 0.7035849502555322_dp, 0.9834216003525784_dp, &
             1.3698388507368364_dp, 1.902162596429401_dp, 2.634481950442649_dp, &
             3.6424092565690755_dp, 5.035219053354037_dp, 6.9815538839681714_dp, &
             9.778070891495517_dp, 14.121352236774849_dp, &
             ! w
             9.937371094714112e-10_dp, 2.759940455541273e-09_dp, 6.029305603644532e-09_dp, &
             1.2728258066518587e-08_dp, 2.614130295794949e-08_dp, 5.199539246719203e-08_dp, &
             1.0029853952788665e-07_dp, 1.8822496204661985e-07_dp, 3.4469701532819195e-07_dp, &
             6.17611853645783e-07_dp, 1.0851053900317174e-06_dp, 1.8729125657222578e-06_dp, &
             3.180848304079989e-06_dp, 5.3228172684593016e-06_dp, 8.786730688004638e-06_dp, &
             1.4323496249649822e-05_dp, 2.3078168933190114e-05_dp, 3.678183795214505e-05_dp, &
             5.8030462282940665e-05_dp, 9.068744474313275e-05_dp, 0.0001404613155024009_dp, &
             0.0002157299050176025_dp, 0.00032870972722833945_dp, 0.0004971065001682968_dp, &
             0.0007464331648292713_dp, 0.0011132498705432301_dp, 0.0016496720370219352_dp, &
             0.0024296154903742533_dp, 0.003557411908211608_dp, 0.005179646632261267_dp, &
             0.0075013616040843735_dp, 0.010808151361566135_dp, 0.015496189396934975_dp, &
             0.022112895043372296_dp, 0.03141184089204085_dp, 0.0444266833264619_dp, &
             0.06257048982499858_dp, 0.08776903334900267_dp, 0.122639804320428_dp, &
             0.17073346087376526_dp, 0.2368630595334666_dp, 0.3275631946055268_dp, &
             0.4517574603995731_dp, 0.621797957724878_dp, 0.8552560184574121_dp, &
             1.1784324454971447_dp, 1.6343602478010495_dp, 2.3046953323053962_dp, &
             3.387674905556139_dp, 5.661716362065017_dp]

      CASE (1180)

         aw(:) = & ! a
            [2.667776766785319e-10_dp, 1.5525652036908868e-09_dp, 4.636956902756964e-09_dp, &
             1.1453241904917906e-08_dp, 2.6036446274787185e-08_dp, 5.612654955134197e-08_dp, &
             1.1606533461156362e-07_dp, 2.3176365178069678e-07_dp, 4.489876827397679e-07_dp, &
             8.469483093839204e-07_dp, 1.5602604121566675e-06_dp, 2.8139514266445153e-06_dp, &
             4.978587229591546e-06_dp, 8.65609319103419e-06_dp, 1.4811964950847971e-05_dp, &
             2.497696867274681e-05_dp, 4.155207467482561e-05_dp, 6.826557009418739e-05_dp, &
             0.00011085287588181096_dp, 0.0001780600516910158_dp, 0.000283114729427037_dp, &
             0.0004458679299997754_dp, 0.0006958932046701074_dp, 0.0010769443350734645_dp, &
             0.0016533309076961842_dp, 0.0025189878478409124_dp, 0.003810311032395408_dp, &
             0.005724233805268051_dp, 0.008543564983915937_dp, 0.012672345939317512_dp, &
             0.018684976158161067_dp, 0.027394187146493006_dp, 0.03994472386019245_dp, &
             0.05794196621493067_dp, 0.08362788249037112_dp, 0.12012090806154688_dp, &
             0.171741937878244_dp, 0.24445611199717118_dp, 0.3464702363819156_dp, &
             0.4890398478416475_dp, 0.6875606745902855_dp, 0.9630520888717407_dp, &
             1.344197955192573_dp, 1.870224754435161_dp, 2.5951503683186967_dp, &
             3.5945580594053004_dp, 4.977742954872777_dp, 6.913404931544084_dp, &
             9.698224629388175_dp, 14.028464323977419_dp, &
             ! w
             6.99732158015401e-10_dp, 1.986136763548785e-09_dp, 4.490734669374502e-09_dp, &
             9.789047175529042e-09_dp, 2.0620773912839688e-08_dp, 4.18698323769614e-08_dp, &
             8.219269802773589e-08_dp, 1.566122397927373e-07_dp, 2.9068248775411426e-07_dp, &
             5.271074969086278e-07_dp, 9.36127077950029e-07_dp, 1.6316220585477495e-06_dp, &
             2.7958206696143686e-06_dp, 4.716824091966227e-06_dp, 7.845078949080902e-06_dp, &
             1.2877636129369296e-05_dp, 2.0882862425442846e-05_dp, 3.348364245255381e-05_dp, &
             5.312460551020624e-05_dp, 8.345932407143168e-05_dp, 0.00012990782094933834_dp, &
             0.00020045451918988003_dp, 0.00030678388436795515_dp, 0.00046588799292918544_dp, &
             0.0007023305018739282_dp, 0.001051419485444909_dp, 0.0015636332735254098_dp, &
             0.0023107665851301064_dp, 0.003394429160549964_dp, 0.004957749198591217_dp, &
             0.0072014268078500155_dp, 0.010405671400153586_dp, 0.014960071755190277_dp, &
             0.021404128450474794_dp, 0.030482079761654787_dp, 0.04321685079793115_dp, &
             0.061009567997010926_dp, 0.0857733037779314_dp, 0.12011292419766467_dp, &
             0.167567903536084_dp, 0.23294358406629806_dp, 0.32277306930383476_dp, &
             0.44598801020442164_dp, 0.6149613961099453_dp, 0.8473000979467618_dp, &
             1.1693527149303544_dp, 1.6242006709252734_dp, 2.2935166137287446_dp, &
             3.3754248478086764_dp, 5.64756581020558_dp]

      CASE (1181)

         aw(:) = & ! a
            [2.0788515037461914e-10_dp, 1.2208535821673166e-09_dp, 3.7078928097538935e-09_dp, &
             9.343146011886314e-09_dp, 2.164350805502342e-08_dp, 4.742953981258455e-08_dp, &
             9.947741131534407e-08_dp, 2.0108834127971026e-07_dp, 3.9375972386881223e-07_dp, &
             7.49840873664612e-07_dp, 1.3930939220931038e-06_dp, 2.531633374487684e-06_dp, &
             4.5100261336954715e-06_dp, 7.890735662743826e-06_dp, 1.3580067336409994e-05_dp, &
             2.302098877794672e-05_dp, 3.848559799481203e-05_dp, 6.351489846260197e-05_dp, &
             0.00010357471883990563_dp, 0.00016702656959584098_dp, 0.0002665543620077789_dp, &
             0.00042124686789878876_dp, 0.0006596178746114524_dp, 0.0010239597161836426_dp, &
             0.0015765827304421972_dp, 0.002408708632689567_dp, 0.0036530804553574947_dp, &
             0.005501753110499239_dp, 0.008231073384719569_dp, 0.012236594774624222_dp, &
             0.018081665118395453_dp, 0.02656475801513414_dp, 0.03881240386903888_dp, &
             0.05640695973144655_dp, 0.08156163287397562_dp, 0.11735940035157187_dp, &
             0.1680781000655122_dp, 0.23963151412367328_dp, 0.3401665013148019_dp, &
             0.48087049302469875_dp, 0.6770645078673645_dp, 0.9496897605592978_dp, &
             1.3273537323868312_dp, 1.849215815557512_dp, 2.569245496043361_dp, &
             3.5630058271297957_dp, 4.93980519184067_dp, 6.86838189260828_dp, &
             9.645433498543792_dp, 13.967012349742465_dp, &
             ! w
             5.463103626921387e-10_dp, 1.5780151782569215e-09_dp, 3.6626695968702385e-09_dp, &
             8.170860301398346e-09_dp, 1.7524033273465807e-08_dp, 3.6103975475683465e-08_dp, &
             7.175180315942242e-08_dp, 1.3818211363965233e-07_dp, 2.5888654755437473e-07_dp, &
             4.733695044030837e-07_dp, 8.469776465881719e-07_dp, 1.486206889516327e-06_dp, &
             2.5622754843003266e-06_dp, 4.347057082580518e-06_dp, 7.267337425131048e-06_dp, &
             1.1986019356291125e-05_dp, 1.9522664803039106e-05_dp, 3.14310556122215e-05_dp, &
             5.005885845355374e-05_dp, 7.892477958021784e-05_dp, 0.0001232628565555843_dp, &
             0.00019080318698626745_dp, 0.0002928853554309296_dp, 0.0004460376495245814_dp, &
             0.000674205262363363_dp, 0.0010118785793547854_dp, 0.0015084643213162292_dp, &
             0.0022343656394901053_dp, 0.00328940259486988_dp, 0.004814425390238989_dp, &
             0.007007267067881109_dp, 0.010144573444234256_dp, 0.014611566455949366_dp, &
             0.02094248345073754_dp, 0.029875344128418385_dp, 0.04242590627544202_dp, &
             0.05998729818420502_dp, 0.08446405605840644_dp, 0.11845252189725437_dp, &
             0.165484563317089_dp, 0.2303601878943901_dp, 0.3196112856981757_dp, &
             0.44217466292240915_dp, 0.610437077570398_dp, 0.8420290736445079_dp, &
             1.163331321491052_dp, 1.6174580275860424_dp, 2.286093846947345_dp, &
             3.36728915108151_dp, 5.638169861685973_dp]

      CASE (1182)

         aw(:) = & ! a
            [1.7143797004198514e-10_dp, 1.014690701162741e-09_dp, 3.125384467841787e-09_dp, &
             8.003730311345055e-09_dp, 1.8818763118055096e-08_dp, 4.177068799798216e-08_dp, &
             8.857141456674936e-08_dp, 1.8073596231959345e-07_dp, 3.568222846190476e-07_dp, &
             6.844296494831037e-07_dp, 1.2797710646021103e-06_dp, 2.339145120925692e-06_dp, &
             4.188882506336469e-06_dp, 7.363665677060335e-06_dp, 1.272798891954287e-05_dp, &
             2.1662604921961583e-05_dp, 3.63480232581809e-05_dp, 6.01917808671481e-05_dp, &
             9.846708787458141e-05_dp, 0.00015926003500387522_dp, 0.0002548642208370207_dp, &
             0.0004038200464739175_dp, 0.0006338773247306185_dp, 0.000986272783624162_dp, &
             0.0015218700064878934_dp, 0.0023299237960094942_dp, 0.003540524571361351_dp, &
             0.005342178827901393_dp, 0.008006525656164836_dp, 0.011922925574632436_dp, &
             0.017646650950740618_dp, 0.025965742158885155_dp, 0.03799338246987254_dp, &
             0.05529503528277229_dp, 0.08006277518995542_dp, 0.1153534906568273_dp, &
             0.16541329764481885_dp, 0.2361180891693015_dp, 0.33557043628937067_dp, &
             0.47490738153729745_dp, 0.6693945809072406_dp, 0.9399152287808964_dp, &
             1.3150199838311396_dp, 1.8338181907275959_dp, 2.5502430642142873_dp, &
             3.539842278183927_dp, 4.911933768449103_dp, 6.835284350302015_dp, &
             9.606604742615675_dp, 13.921793649942067_dp, &
             ! w
             4.5127411939303475e-10_dp, 1.3230517857491095e-09_dp, 3.1370685453963587e-09_dp, &
             7.126182751271468e-09_dp, 1.5497456773103163e-08_dp, 3.2289485926399965e-08_dp, &
             6.478132019880978e-08_dp, 1.2578063575944464e-07_dp, 2.3734235826067154e-07_dp, &
             4.367319146197643e-07_dp, 7.858578736911673e-07_dp, 1.3860076609327713e-06_dp, &
             2.400607312456006e-06_dp, 4.090008588636433e-06_dp, 6.864147184332484e-06_dp, &
             1.1361539481447843e-05_dp, 1.8566802603612374e-05_dp, 2.9984118277693852e-05_dp, &
             4.789139440111278e-05_dp, 7.571009585392944e-05_dp, 0.00011853988138100735_dp, &
             0.0001839267278082631_dp, 0.0002829600677018369_dp, 0.0004318311795304515_dp, &
             0.0006540350706545242_dp, 0.000983465923964454_dp, 0.001468747848453517_dp, &
             0.002179266111323009_dp, 0.0032135296711479494_dp, 0.004710717372285053_dp, &
             0.00686655555764585_dp, 0.009955067635294627_dp, 0.014358257193054263_dp, &
             0.020606476596574656_dp, 0.0294331479466191_dp, 0.041848720714550945_dp, &
             0.05924038783876884_dp, 0.08350633581043443_dp, 0.11723654594905822_dp, &
             0.16395718153126024_dp, 0.22846420497551284_dp, 0.3172884950848487_dp, &
             0.4393705592406897_dp, 0.6071072578832923_dp, 0.838146634033823_dp, &
             1.158893184983642_dp, 1.6124856233004554_dp, 2.2806179159873943_dp, &
             3.3612864199176813_dp, 5.631238179518659_dp]

      CASE (1183)

         aw(:) = & ! a
            [1.4652804678971417e-10_dp, 8.732819622787592e-10_dp, 2.722866770330731e-09_dp, &
             7.068706921833249e-09_dp, 1.682621304851421e-08_dp, 3.7741333793139044e-08_dp, &
             8.074203076246538e-08_dp, 1.6602051197445262e-07_dp, 3.29947677231248e-07_dp, &
             6.365739477055299e-07_dp, 1.1964512666506885e-06_dp, 2.196987636081006e-06_dp, &
             3.950750102348182e-06_dp, 6.971394293846574e-06_dp, 1.209168593919944e-05_dp, &
             2.064505121519545e-05_dp, 3.4742171795060674e-05_dp, 5.768861395541599e-05_dp, &
             9.461013310417009e-05_dp, 0.00015338158660816706_dp, 0.000245996697415349_dp, &
             0.00039057385655666063_dp, 0.0006142740191413314_dp, 0.0009575190539736801_dp, &
             0.0014800541520289736_dp, 0.002269611552963568_dp, 0.0034542256436912496_dp, &
             0.0052196491111724435_dp, 0.007833863159567975_dp, 0.01168141076284789_dp, &
             0.0173112752798039_dp, 0.025503363032700066_dp, 0.03736044039094562_dp, &
             0.05443477218644712_dp, 0.07890190563664955_dp, 0.11379830511498974_dp, &
             0.16334522690529737_dp, 0.23338884056244727_dp, 0.331996937072406_dp, &
             0.4702669434322886_dp, 0.6634209375036618_dp, 0.9322963822382414_dp, &
             1.3053990775983235_dp, 1.8217987751859235_dp, 2.535399895686932_dp, &
             3.5217377682506044_dp, 4.890137692459445_dp, 6.809388961076662_dp, &
             9.5762130207107_dp, 13.886388817925129_dp, &
             ! w
             3.862714352269292e-10_dp, 1.1474079894196364e-09_dp, 2.7701574435172473e-09_dp, &
             6.386952636029124e-09_dp, 1.4048030134546941e-08_dp, 2.9538047301932982e-08_dp, &
             5.971752265891002e-08_dp, 1.1671589510819929e-07_dp, 2.2150955072689904e-07_dp, &
             4.096772711967926e-07_dp, 7.405294353561165e-07_dp, 1.3114059124841184e-06_dp, &
             2.279811098809911e-06_dp, 3.897319062362611e-06_dp, 6.560999101688676e-06_dp, &
             1.089070655069496e-05_dp, 1.784426431014209e-05_dp, 2.8887750982807336e-05_dp, &
             4.6245389945240664e-05_dp, 7.326368585487069e-05_dp, 0.00011493854390758498_dp, &
             0.00017867356589382322_dp, 0.00027536446935656935_dp, 0.00042094119508221376_dp, &
             0.000638549184787924_dp, 0.0009616190822996391_dp, 0.0014381657735840839_dp, &
             0.0021367812030670713_dp, 0.0031549512784841763_dp, 0.004630549185953724_dp, &
             0.00675765365325639_dp, 0.009808234820647004_dp, 0.014161773522773471_dp, &
             0.02034557324948698_dp, 0.029089443535328493_dp, 0.04139965729805845_dp, &
             0.05865873058605824_dp, 0.08275983753308636_dp, 0.11628792656504627_dp, &
             0.16276463067578859_dp, 0.22698267334685174_dp, 0.31547207039163055_dp, &
             0.4371761713039822_dp, 0.6044997284994543_dp, 0.8351045340642531_dp, &
             1.1554138796152078_dp, 1.608585899420891_dp, 2.276322143161544_dp, &
             3.35657691574868_dp, 5.625800469207035_dp]

      CASE (1184)

         aw(:) = & ! a
            [1.2836028869794193e-10_dp, 7.698239722878696e-10_dp, 2.4264640665525623e-09_dp, &
             6.374152421618995e-09_dp, 1.533311433313415e-08_dp, 3.469837505049082e-08_dp, &
             7.478928959635721e-08_dp, 1.5476661563376848e-07_dp, 3.092894234926464e-07_dp, &
             5.996210563847102e-07_dp, 1.1318541217905446e-06_dp, 2.0863741525646494e-06_dp, &
             3.7648494652214294e-06_dp, 6.664247468508839e-06_dp, 1.1592098903055572e-05_dp, &
             1.9844115602986895e-05_dp, 3.347523222203949e-05_dp, 5.570946829391531e-05_dp, &
             9.15544704494083e-05_dp, 0.00014871564803457026_dp, 0.00023894582857952006_dp, &
             0.00038002394202069176_dp, 0.000598636671828566_dp, 0.0009345488016630823_dp, &
             0.0014466026409759069_dp, 0.002221299914241333_dp, 0.003385011609334365_dp, &
             0.00512126025014117_dp, 0.007695061848733614_dp, 0.01148705049171396_dp, &
             0.01704110212245322_dp, 0.025130511470685582_dp, 0.03684957051985645_dp, &
             0.053739797790896326_dp, 0.07796327278295757_dp, 0.11253980322473123_dp, &
             0.1616703547892305_dp, 0.2311768164858303_dp, 0.3290985439059154_dp, &
             0.46650054637237104_dp, 0.6585692072168399_dp, 0.9261044909836009_dp, &
             1.2975753681504805_dp, 1.8120190295055012_dp, 2.5233161500020467_dp, &
             3.5069917878482024_dp, 4.8723771671631315_dp, 6.788280006096587_dp, &
             9.551430773097731_dp, 13.857511066903283_dp, &
             ! w
             3.3883094919989544e-10_dp, 1.0184138618138462e-09_dp, 2.497602718637774e-09_dp, &
             5.831576996841072e-09_dp, 1.2949493168914275e-08_dp, 2.743811346857385e-08_dp, &
             5.58301700702295e-08_dp, 1.0972206215099128e-07_dp, 2.0923989147970786e-07_dp, &
             3.8862897990162344e-07_dp, 7.051402766367773e-07_dp, 1.2529772535302986e-06_dp, &
             2.184929257170578e-06_dp, 3.745567191959547e-06_dp, 6.321675986392452e-06_dp, &
             1.0518168953547077e-05_dp, 1.7271379829191565e-05_dp, 2.8016781670454967e-05_dp, &
             4.4935415101877726e-05_dp, 7.131340658764376e-05_dp, 0.00011206298364740646_dp, &
             0.00017447278188515407_dp, 0.0002692819141245657_dp, 0.00041220881315835647_dp, &
             0.00062611570506774_dp, 0.0009440572910394045_dp, 0.0014135538696240526_dp, &
             0.002102552718712119_dp, 0.0031077076176104852_dp, 0.004565828778540794_dp, &
             0.0066696523377161605_dp, 0.009689473770517217_dp, 0.014002714445046196_dp, &
             0.020134186543888155_dp, 0.028810744789256906_dp, 0.0410352423742927_dp, &
             0.05818636139541698_dp, 0.08215316116175507_dp, 0.11551645062716445_dp, &
             0.16179412577974606_dp, 0.22577622122095414_dp, 0.31399200204311106_dp, &
             0.4353870989158665_dp, 0.6023726886219761_dp, 0.8326218072510309_dp, &
             1.1525731643716244_dp, 1.6054008864030795_dp, 2.272812896348087_dp, &
             3.3527293534476454_dp, 5.621358324975687_dp]

      CASE (1185)

         aw(:) = & ! a
            [1.1448694113833941e-10_dp, 6.906000180313356e-10_dp, 2.1981767363736587e-09_dp, &
             5.835107638691404e-09_dp, 1.4165555715481128e-08_dp, 3.230300297678373e-08_dp, &
             7.007646947686432e-08_dp, 1.45812634146777e-07_dp, 2.9278187012912134e-07_dp, &
             5.699801462405915e-07_dp, 1.0798630527260862e-06_dp, 1.9970754686939545e-06_dp, &
             3.6143577031551778e-06_dp, 6.4149803191747106e-06_dp, 1.1185725832244928e-05_dp, &
             1.9191246536103405e-05_dp, 3.2440496923125296e-05_dp, 5.409014128799717e-05_dp, &
             8.905014216021358e-05_dp, 0.0001448855885175784_dp, 0.00023314958665410525_dp, &
             0.00037133931318020854_dp, 0.0005857473900269736_dp, 0.0009155920967342629_dp, &
             0.0014189640908001111_dp, 0.0021813397555258515_dp, 0.003327702824581891_dp, &
             0.005039714339534541_dp, 0.007579913324509009_dp, 0.011325666027703038_dp, &
             0.016816575381070873_dp, 0.02482040063951146_dp, 0.03642433367981422_dp, &
             0.05316088317651415_dp, 0.07718082940224165_dp, 0.11148999592437607_dp, &
             0.16027230179259772_dp, 0.2293292233596801_dp, 0.3266761937420213_dp, &
             0.46335092943261263_dp, 0.6545097383224936_dp, 0.9209209616935903_dp, &
             1.2910224783886153_dp, 1.8038239533479334_dp, 2.513185942075596_dp, &
             3.4946247378480413_dp, 4.8574764820621965_dp, 6.7705644275811805_dp, &
             9.530626809267849_dp, 13.833263726833003_dp, &
             ! w
             3.0258287354316334e-10_dp, 9.192998602268231e-10_dp, 2.2860646170310637e-09_dp, &
             5.396331452716896e-09_dp, 1.2082145670835873e-08_dp, 2.5770303219192244e-08_dp, &
             5.272749840699552e-08_dp, 1.0411623698325842e-07_dp, 1.9936868189393665e-07_dp, &
             3.71639289817253e-07_dp, 6.764905881823154e-07_dp, 1.2055495921713635e-06_dp, &
             2.1077253478382046e-06_dp, 3.621815746959837e-06_dp, 6.126114653373329e-06_dp, &
             1.0213180967437786e-05_dp, 1.680155669331504e-05_dp, 2.730134322706121e-05_dp, &
             4.3857740330248396e-05_dp, 6.970670501056455e-05_dp, 0.00010969086087831555_dp, &
             0.00017100311637052387_dp, 0.0002642520603560146_dp, 0.0004049796785410494_dp, &
             0.0006158117188580768_dp, 0.0009294887113706864_dp, 0.001393117292472464_dp, &
             0.0020741051232759716_dp, 0.003068408908170096_dp, 0.004511947722780805_dp, &
             0.00659633129933144_dp, 0.009590449048554835_dp, 0.013869991877866918_dp, &
             0.019957677028702228_dp, 0.02857787250027279_dp, 0.04073055098069153_dp, &
             0.057791161975314836_dp, 0.08164529131423894_dp, 0.11487024828379233_dp, &
             0.1609807619204961_dp, 0.22476457629690214_dp, 0.31275029342739086_dp, &
             0.43388543344980673_dp, 0.6005865584509453_dp, 0.830536166248424_dp, &
             1.150185973639186_dp, 1.6027236398784586_dp, 2.269862571582528_dp, &
             3.349494367928898_dp, 5.6176236814091425_dp]

      CASE (1186)

         aw(:) = & ! a
            [1.0352399357156693e-10_dp, 6.278372271859005e-10_dp, 2.0163739640109897e-09_dp, &
             5.402889242149487e-09_dp, 1.3223160905011337e-08_dp, 3.035834198409525e-08_dp, &
             6.623134839681738e-08_dp, 1.3847585747680198e-07_dp, 2.792052460287841e-07_dp, &
             5.455218114585414e-07_dp, 1.0368369500093258e-06_dp, 1.922981136251969e-06_dp, &
             3.489193766976314e-06_dp, 6.207219004020023e-06_dp, 1.0846352528025186e-05_dp, &
             1.864503262763551e-05_dp, 3.1573356361536435e-05_dp, 5.273099752553439e-05_dp, &
             8.694517229662529e-05_dp, 0.00014166197993569388_dp, 0.00022826499557235898_dp, &
             0.0003640120053544916_dp, 0.000574860517072244_dp, 0.0008995636598289583_dp, &
             0.0013955718106657623_dp, 0.0021474872568206325_dp, 0.0032791102286099335_dp, &
             0.004970512555739379_dp, 0.007482117122777099_dp, 0.011188496554535056_dp, &
             0.0166255985192328_dp, 0.02455644412344187_dp, 0.03606214416280607_dp, &
             0.052667486021790284_dp, 0.07651356108858898_dp, 0.11059419484144617_dp, &
             0.1590786722658311_dp, 0.22775093574648608_dp, 0.32460585723208635_dp, &
             0.46065768049149197_dp, 0.6510368300511262_dp, 0.916484411188488_dp, &
             1.2854115006807851_dp, 1.7968040073795304_dp, 2.504505089306459_dp, &
             3.484023415222912_dp, 4.844699305130709_dp, 6.75536939217473_dp, &
             9.512778720512173_dp, 13.812457635492118_dp, &
             ! w
             2.739238494815766e-10_dp, 8.40537808785961e-10_dp, 2.11644018677323e-09_dp, &
             5.044351168609757e-09_dp, 1.137618976817688e-08_dp, 2.440587907563567e-08_dp, &
             5.017837974316622e-08_dp, 9.949364992262913e-08_dp, 1.9120270841127184e-07_dp, &
             3.5754459192551266e-07_dp, 6.526623255196929e-07_dp, 1.1660131532293156e-06_dp, &
             2.0432332075853456e-06_dp, 3.5182440221540273e-06_dp, 5.96215767276273e-06_dp, &
             9.957070675961157e-06_dp, 1.6406440583525848e-05_dp, 2.669883443579783e-05_dp, &
             4.2949000743751474e-05_dp, 6.835023031845342e-05_dp, 0.00010768589872836877_dp, &
             0.00016806735901757656_dp, 0.0002599919072592476_dp, 0.00039885096052611895_dp, &
             0.0006070683185121759_dp, 0.0009171160189437985_dp, 0.0013757469313808387_dp, &
             0.002049906917675778_dp, 0.003034955734587034_dp, 0.004466048825597342_dp, &
             0.006533830064803823_dp, 0.009505982431380651_dp, 0.013756711103993896_dp, &
             0.01980693347348617_dp, 0.02837887961377136_dp, 0.04047004388957898_dp, &
             0.05745309222360975_dp, 0.0812106164449792_dp, 0.11431690530964281_dp, &
             0.16028394894050954_dp, 0.22389750179274442_dp, 0.31168557271940767_dp, &
             0.43259728392123986_dp, 0.599053811703288_dp, 0.8287457879005939_dp, &
             1.1481361363900706_dp, 1.6004242076479345_dp, 2.2673282129235117_dp, &
             3.3467153256091597_dp, 5.614415600849092_dp]

      CASE (1187)

         aw(:) = & ! a
            [9.462820612267824e-11_dp, 5.767904762867008e-10_dp, 1.867797490547165e-09_dp, &
             5.047482092068457e-09_dp, 1.2443651473191909e-08_dp, 2.8741520106590857e-08_dp, &
             6.302040296008635e-08_dp, 1.3232599675992837e-07_dp, 2.6778764706833865e-07_dp, &
             5.248936371282116e-07_dp, 1.0004557866098026e-06_dp, 1.860186249252722e-06_dp, &
             3.3828979526461943e-06_dp, 6.030445540702249e-06_dp, 1.0557101382934006e-05_dp, &
             1.8178755264986153e-05_dp, 3.083204231766837e-05_dp, 5.156750723676196e-05_dp, &
             8.514096339800124e-05_dp, 0.00013889573694533239_dp, 0.00022406884831402253_dp, &
             0.0003577109665322941_dp, 0.000565489436448305_dp, 0.0008857543310408699_dp, &
             0.0013754008273826471_dp, 0.0021182727554219763_dp, 0.0032371427513807264_dp, &
             0.004910701934328088_dp, 0.007397533588026998_dp, 0.011069780393545456_dp, &
             0.016460208842427396_dp, 0.024327714212999032_dp, 0.035748109316760146_dp, &
             0.05223945111718081_dp, 0.0759343812277249_dp, 0.10981625623926022_dp, &
             0.15804158585541675_dp, 0.22637899787140514_dp, 0.32280539628727284_dp, &
             0.4583145014200057_dp, 0.6480140933608375_dp, 0.9126214340207093_dp, &
             1.2805241185523246_dp, 1.7906872216508682_dp, 2.496938618229322_dp, &
             3.4747802442008666_dp, 4.833556032368586_dp, 6.742114324576782_dp, &
             9.497206209966738_dp, 13.794301296440665_dp, &
             ! w
             2.5065749800079785e-10_dp, 7.762972469590782e-10_dp, 1.976952575675056e-09_dp, &
             4.752717795230882e-09_dp, 1.0787934589602114e-08_dp, 2.326381198663677e-08_dp, &
             4.803666690131858e-08_dp, 9.559733389020767e-08_dp, 1.8430034108424388e-07_dp, &
             3.45601248058648e-07_dp, 6.324262050893737e-07_dp, 1.1323696439102905e-06_dp, &
             1.9882539740158723e-06_dp, 3.429803279942067e-06_dp, 5.821940570677704e-06_dp, &
             9.737736436365159e-06_dp, 1.606762188347712e-05_dp, 2.618154984344969e-05_dp, &
             4.2167923625132856e-05_dp, 6.718308999858363e-05_dp, 0.00010595908177992591_dp, &
             0.00016553652496595765_dp, 0.00025631613001898617_dp, 0.0003935585530854464_dp, &
             0.0005995120974412179_dp, 0.0009064153289070719_dp, 0.0013607133048378472_dp, &
             0.0020289498374461727_dp, 0.003005964671320218_dp, 0.004426247659691359_dp, &
             0.006479600355131993_dp, 0.009432653027950088_dp, 0.013658313712651905_dp, &
             0.019675927506170872_dp, 0.028205855739324787_dp, 0.04024342510064967_dp, &
             0.0571588650852403_dp, 0.08083214466343623_dp, 0.11383490350253989_dp, &
             0.15967672543780584_dp, 0.22314161040711153_dp, 0.3107570312809365_dp, &
             0.4314734935771462_dp, 0.5977161948171613_dp, 0.8271828759811768_dp, &
             1.1463462743394115_dp, 1.598416000008677_dp, 2.2651145292711674_dp, &
             3.3442877828957895_dp, 5.611613426069749_dp]

      CASE (1188)

         aw(:) = & ! a
            [5.266595321120371e-11_dp, 3.3399870328314053e-10_dp, 1.1489595907315604e-09_dp, &
             3.291194676991691e-09_dp, 8.514582382515992e-09_dp, 2.0452226072644907e-08_dp, &
             4.6318485394919494e-08_dp, 9.993815467144849e-08_dp, 2.070070561380891e-07_dp, &
             4.1403853568986233e-07_dp, 8.032979766199053e-07_dp, 1.5173215058971754e-06_dp, &
             2.798569706855337e-06_dp, 5.052690772966723e-06_dp, 8.948201070025004e-06_dp, &
             1.5571760669583898e-05_dp, 2.6667510881412685e-05_dp, 4.500238991406956e-05_dp, &
             7.491868121745718e-05_dp, 0.00012316267702130069_dp, 0.0002001175710493163_dp, &
             0.0003216240835726255_dp, 0.0005116500991201278_dp, 0.000806179248046686_dp, &
             0.0012588393986136502_dp, 0.0019490009991618166_dp, 0.002993363055327382_dp, &
             0.00456243932387813_dp, 0.006903896583598299_dp, 0.010375430605003802_dp, &
             0.01549086010632034_dp, 0.02298446254819233_dp, 0.03390038445922832_dp, &
             0.049716386922591616_dp, 0.07251443769800478_dp, 0.10521500746671854_dp, &
             0.1518977394081935_dp, 0.2182389629264437_dp, 0.31210710733490526_dp, &
             0.4443717689621982_dp, 0.6300035136704071_dp, 0.8895748153615481_dp, &
             1.2513303177072732_dp, 1.7541078411474331_dp, 2.451641406854549_dp, &
             3.419390917215014_dp, 4.766721368335335_dp, 6.662552166440162_dp, &
             9.403672848315978_dp, 13.685190663468623_dp, &
             ! w
             1.407235527281632e-10_dp, 4.676520494203022e-10_dp, 1.2874535665985023e-09_dp, &
             3.274706508969593e-09_dp, 7.750471012804181e-09_dp, 1.7278802895655377e-08_dp, &
             3.667351948593109e-08_dp, 7.470465656840379e-08_dp, 1.469447689753995e-07_dp, &
             2.8043328331256906e-07_dp, 5.21200355311459e-07_dp, 9.462321200380404e-07_dp, &
             1.6822554987585234e-06_dp, 2.9348842649418216e-06_dp, 5.033354893002764e-06_dp, &
             8.498508549543201e-06_dp, 1.4145146211612971e-05_dp, 2.323481344351603e-05_dp, &
             3.770203117793626e-05_dp, 6.0486759025517355e-05_dp, 9.60195193704635e-05_dp, &
             0.0001509245611929735_dp, 0.00023503264058734465_dp, 0.0003628310444686375_dp, &
             0.0005555278719436638_dp, 0.0008439749648081943_dp, 0.0012727854281407518_dp, &
             0.0019061055987502087_dp, 0.0028356684294227733_dp, 0.004191979829347834_dp, &
             0.006159790305343869_dp, 0.008999405936471517_dp, 0.013075927763048554_dp, &
             0.018899220055233038_dp, 0.02717835359538528_dp, 0.038895530770646745_dp, &
             0.0554061947414209_dp, 0.07857435193277673_dp, 0.11095546049174261_dp, &
             0.15604432283164849_dp, 0.21861403199829468_dp, 0.30518847457045245_dp, &
             0.4247261751999236_dp, 0.5896763873144145_dp, 0.8177797915321899_dp, &
             1.135568773311409_dp, 1.5863157645165344_dp, 2.2517703473413273_dp, &
             3.32965195063848_dp, 5.594721719820455_dp]

      CASE (1189)

         aw(:) = & ! a
            [3.756330199349711e-11_dp, 2.453430728312633e-10_dp, 8.786535105491705e-10_dp, &
             2.6078914578752837e-09_dp, 6.938815478277077e-09_dp, 1.7042437828133182e-08_dp, &
             3.930131199849581e-08_dp, 8.608491192531718e-08_dp, 1.80606504600645e-07_dp, &
             3.6523855070370734e-07_dp, 7.154763593762313e-07_dp, 1.362985121522505e-06_dp, &
             2.533053033722411e-06_dp, 4.6046054351957115e-06_dp, 8.205145153064144e-06_dp, &
             1.4359189552487758e-05_dp, 2.4717847474038504e-05_dp, 4.191034566461705e-05_dp, &
             7.00772883348433e-05_dp, 0.00011567259746546179_dp, 0.0001886597304976398_dp, &
             0.0003042824410028688_dp, 0.0004856672493297479_dp, 0.0007676223987614935_dp, &
             0.0012021478990912654_dp, 0.001866378875900649_dp, 0.002873970837877085_dp, &
             0.004391328828064864_dp, 0.006660620449719588_dp, 0.0100322457503328_dp, &
             0.01501043178836766_dp, 0.022316965089955974_dp, 0.03297988851924328_dp, &
             0.04845642562827152_dp, 0.07080266115292679_dp, 0.10290687596632388_dp, &
             0.14880928249187145_dp, 0.21413876035269364_dp, 0.30670784884396324_dp, &
             0.4373220424563533_dp, 0.6208808683298574_dp, 0.8778816327091056_dp, &
             1.2364945224352224_dp, 1.735490793786183_dp, 2.428555034180448_dp, &
             3.391124544194174_dp, 4.73257465610981_dp, 6.621861630565766_dp, &
             9.355796110366072_dp, 13.629301517573115_dp, &
             ! w
             1.0104548798594975e-10_dp, 3.52949533555899e-10_dp, 1.0190155567356863e-09_dp, &
             2.6772392918064228e-09_dp, 6.488519218145113e-09_dp, 1.473818565952187e-08_dp, &
             3.176336113083653e-08_dp, 6.553917307973461e-08_dp, 1.3034129207025693e-07_dp, &
             2.5113354195760847e-07_dp, 4.706809377644482e-07_dp, 8.609130509750774e-07_dp, &
             1.540837760253797e-06_dp, 2.704441123478718e-06_dp, 4.663659832664683e-06_dp, &
             7.91389415580951e-06_dp, 1.3232941034384158e-05_dp, 2.1829083178232657e-05_dp, &
             3.556093940624237e-05_dp, 5.726134274998528e-05_dp, 9.121102644739427e-05_dp, &
             0.00014382670138019224_dp, 0.00022465414391250086_dp, 0.0003477927937862384_dp, &
             0.000533927588182235_dp, 0.0008132110071993453_dp, 0.001229329802055074_dp, &
             0.0018452148766447362_dp, 0.0027510202183478067_dp, 0.0040752220282258556_dp, &
             0.005999991251810654_dp, 0.008782396488447928_dp, 0.01278353261181859_dp, &
             0.018508386858424836_dp, 0.026660207191128774_dp, 0.03821440890901302_dp, &
             0.05451876518561173_dp, 0.07742897231197515_dp, 0.10949202457418529_dp, &
             0.15419493739429427_dp, 0.21630497205431679_dp, 0.30234393031455087_dp, &
             0.42127424287020593_dp, 0.5855574058976074_dp, 0.8129562345563649_dp, &
             1.1300341162787315_dp, 1.5800964440827936_dp, 2.244907677648448_dp, &
             3.322123286456614_dp, 5.58603449017686_dp]

      CASE (1190)

         aw(:) = & ! a
            [2.96337441031901e-11_dp, 1.9831907448526746e-10_dp, 7.323487452286773e-10_dp, &
             2.22988362367553e-09_dp, 6.050720571391184e-09_dp, 1.5091283224633598e-08_dp, &
             3.5235340728785044e-08_dp, 7.797308408344552e-08_dp, 1.650081142354883e-07_dp, &
             3.3618047271693546e-07_dp, 6.628242718084179e-07_dp, 1.269892780303462e-06_dp, &
             2.372028433104463e-06_dp, 4.331528726302746e-06_dp, 7.750289955344635e-06_dp, &
             1.361391003342467e-05_dp, 2.3515064267881918e-05_dp, 3.9996251369238014e-05_dp, &
             6.707074456743068e-05_dp, 0.00011100744534674412_dp, 0.000181503607090107_dp, &
             0.0002934235947481514_dp, 0.0004693581758603593_dp, 0.0007433657640196518_dp, &
             0.0011664060677159436_dp, 0.0018141833201414776_dp, 0.002798401699308042_dp, &
             0.004282827940950647_dp, 0.006506093327250441_dp, 0.00981389969010429_dp, &
             0.014704288977350474_dp, 0.021890982442327442_dp, 0.032391611686709836_dp, &
             0.047650105865841154_dp, 0.06970577452019477_dp, 0.10142601049305405_dp, &
             0.14682540853220435_dp, 0.21150198305704704_dp, 0.30323187245462174_dp, &
             0.4327787674528961_dp, 0.6149958027510869_dp, 0.8703311194288161_dp, &
             1.2269061150700988_dp, 1.723448324657578_dp, 2.413609741340885_dp, &
             3.3728125947338032_dp, 4.7104387560500545_dp, 6.59546857602645_dp, &
             9.324726923169584_dp, 13.593018552097414_dp, &
             ! w
             8.017407135648836e-11_dp, 2.913528537049821e-10_dp, 8.704308694607909e-10_dp, &
             2.3389832015083633e-09_dp, 5.762381940567701e-09_dp, 1.325760089891841e-08_dp, &
             2.8871816229533497e-08_dp, 6.00937673433541e-08_dp, 1.204013644570338e-07_dp, &
             2.334755876914974e-07_dp, 4.400547934920345e-07_dp, 8.089178678990586e-07_dp, &
             1.4542461473249906e-06_dp, 2.562731758494217e-06_dp, 4.435427322447133e-06_dp, &
             7.551683208175649e-06_dp, 1.2665892130209574e-05_dp, 2.0952569006918215e-05_dp, &
             3.422210449336468e-05_dp, 5.5239120650442386e-05_dp, 8.818878625159443e-05_dp, &
             0.00013935515578713227_dp, 0.0002181015247290695_dp, 0.00033827856124287516_dp, &
             0.0005202350928933188_dp, 0.0007936735888764145_dp, 0.0012016838340612817_dp, &
             0.0018064123386033698_dp, 0.0026969926487428264_dp, 0.0040005874137339055_dp, &
             0.005897695917779712_dp, 0.008643285914706523_dp, 0.012595849103107667_dp, &
             0.018257199289213065_dp, 0.02632679015568207_dp, 0.03777560758781426_dp, &
             0.05394641058769155_dp, 0.07668945182741158_dp, 0.1085461667877756_dp, &
             0.15299843644863775_dp, 0.2148096457810902_dp, 0.30050015120125_dp, &
             0.41903484378054057_dp, 0.5828831366660923_dp, 0.8098222639858026_dp, &
             1.1264359029869095_dp, 1.576051133134067_dp, 2.2404424420978732_dp, &
             3.3172240929661214_dp, 5.580382080201641_dp]

      CASE (1191)

         aw(:) = & ! a
            [2.469922099674294e-11_dp, 1.6881807750473117e-10_dp, 6.391014381076482e-10_dp, &
             1.985023983344086e-09_dp, 5.467672875871338e-09_dp, 1.3796403261648591e-08_dp, &
             3.251301092725624e-08_dp, 7.250165141436749e-08_dp, 1.5442069930889224e-07_dp, &
             3.163499325763861e-07_dp, 6.267209416052445e-07_dp, 1.205790658173054e-06_dp, &
             2.260731665450166e-06_dp, 4.142144278797675e-06_dp, 7.433869244468252e-06_dp, &
             1.3094002772168029e-05_dp, 2.267384717279306e-05_dp, 3.8654381459703e-05_dp, &
             6.4958402827253e-05_dp, 0.00010772313403941397_dp, 0.00017645608630808263_dp, &
             0.0002857508246514608_dp, 0.0004578151955637295_dp, 0.0007261710049044604_dp, &
             0.0011410325859024333_dp, 0.0017770777659002165_dp, 0.002744609469915561_dp, &
             0.004205497885292005_dp, 0.006395829665220602_dp, 0.009657922696360968_dp, &
             0.014485359893162133_dp, 0.02158604278500868_dp, 0.031970083872631165_dp, &
             0.04707180279672984_dp, 0.06891837342017781_dp, 0.10036206504432714_dp, &
             0.14539890914269543_dp, 0.20960453569883947_dp, 0.3007286614412287_dp, &
             0.42950460671743296_dp, 0.6107517700285942_dp, 0.8648825021344094_dp, &
             1.2199826398183893_dp, 1.714747799641849_dp, 2.402806117758009_dp, &
             3.3595687094355635_dp, 4.6944220945104895_dp, 6.576364149059323_dp, &
             9.302230297539024_dp, 13.566739665560622_dp, &
             ! w
             6.716779406172266e-11_dp, 2.523281329384768e-10_dp, 7.741389292323057e-10_dp, &
             2.1162226419046244e-09_dp, 5.278663163350065e-09_dp, 1.2262424265790456e-08_dp, &
             2.691392517086305e-08_dp, 5.6383727606050644e-08_dp, 1.1359301240986947e-07_dp, &
             2.2132455218917183e-07_dp, 4.188935127210476e-07_dp, 7.728603437434257e-07_dp, &
             1.3939997117754498e-06_dp, 2.4638441454128985e-06_dp, 4.275731337841935e-06_dp, &
             7.297613573108523e-06_dp, 1.2267233515745538e-05_dp, 2.0335043981243775e-05_dp, &
             3.327701683005087e-05_dp, 5.380902414715739e-05_dp, 8.604784244580878e-05_dp, &
             0.00013618246640560312_dp, 0.00021344526839330545_dp, 0.0003315082247112872_dp, &
             0.0005104784678772512_dp, 0.0007797345206772052_dp, 0.0011819359922578257_dp, &
             0.0017786635955224718_dp, 0.0026583141046532697_dp, 0.003947100873707664_dp, &
             0.005824313968308064_dp, 0.008543400068249841_dp, 0.012460964355099505_dp, &
             0.018076518691349238_dp, 0.026086761773008678_dp, 0.0374594599729924_dp, &
             0.05353372351094552_dp, 0.0761558383352315_dp, 0.10786318209462475_dp, &
             0.1521338777497926_dp, 0.21372845784005695_dp, 0.2991661861561817_dp, &
             0.4174136979608509_dp, 0.5809461245501593_dp, 0.8075511671815416_dp, &
             1.1238272788074999_dp, 1.573117387559291_dp, 2.2372034253039783_dp, &
             3.3136699624979644_dp, 5.576281838772698_dp]

      CASE (1192)

         aw(:) = & ! a
            [2.131188191226459e-11_dp, 1.4842800878200733e-10_dp, 5.738052636227673e-10_dp, &
             1.8113393696647257e-09_dp, 5.049760894019697e-09_dp, 1.2860507939971225e-08_dp, &
             3.053203079061829e-08_dp, 6.849770940067234e-08_dp, 1.466357984810959e-07_dp, &
             3.017083456184907e-07_dp, 5.999683648958933e-07_dp, 1.1581394820661017e-06_dp, &
             2.1777623749561416e-06_dp, 4.000601243970777e-06_dp, 7.196833523928876e-06_dp, &
             1.270371016217523e-05_dp, 2.204112744728664e-05_dp, 3.76433004381518e-05_dp, &
             6.336416261378233e-05_dp, 0.00010524059172987811_dp, 0.0001726353436069185_dp, &
             0.0002799351688292062_dp, 0.0004490551774391026_dp, 0.0007131065586330194_dp, &
             0.0011217327644842926_dp, 0.001748824827078042_dp, 0.002703610658842845_dp, &
             0.004146504318463626_dp, 0.006311637285375666_dp, 0.009538725508058092_dp, &
             0.01431792093700094_dp, 0.02135264415397123_dp, 0.03164721448018555_dp, &
             0.046628543073197734_dp, 0.06831444190429_dp, 0.09954550622160732_dp, &
             0.14430342943033744_dp, 0.20814654125621043_dp, 0.29880412490602215_dp, &
             0.4269859977980349_dp, 0.6074854321988228_dp, 0.8606870399611786_dp, &
             1.214649071249204_dp, 1.708042351730647_dp, 2.3944764520608017_dp, &
             3.349353790040508_dp, 4.682064423569568_dp, 6.561619805487643_dp, &
             9.284863659411766_dp, 13.546449188028616_dp, &
             ! w
             5.822961021760758e-11_dp, 2.2513245633185387e-10_dp, 7.058105863372246e-10_dp, &
             1.9561830374580434e-09_dp, 4.928066620809328e-09_dp, 1.1536155068934385e-08_dp, &
             2.547703250776919e-08_dp, 5.364805759332977e-08_dp, 1.0855240034255214e-07_dp, &
             2.1229674187812345e-07_dp, 4.0312259543706505e-07_dp, 7.459134734477842e-07_dp, &
             1.348864214355314e-06_dp, 2.389593495081258e-06_dp, 4.155577792830423e-06_dp, &
             7.106097716495594e-06_dp, 1.196621187536216e-05_dp, 1.986801996130303e-05_dp, &
             3.256121101581334e-05_dp, 5.272439048172419e-05_dp, 8.442199951905298e-05_dp, &
             0.0001337702207927308_dp, 0.0002099010529526293_dp, 0.0003263493564293606_dp, &
             0.0005030366489699277_dp, 0.000769092473702539_dp, 0.0011668455606459744_dp, &
             0.001757441103780886_dp, 0.0026287083414235583_dp, 0.0039061288685742325_dp, &
             0.005768059984154825_dp, 0.008466774236439096_dp, 0.012357419584438957_dp, &
             0.017937728526964028_dp, 0.025902268649762417_dp, 0.037216313945623865_dp, &
             0.05321614748823459_dp, 0.07574497859781909_dp, 0.10733703350023041_dp, &
             0.15146751083802257_dp, 0.21289471471668883_dp, 0.2981370386480893_dp, &
             0.416162442497864_dp, 0.5794504631035586_dp, 0.8057968970015219_dp, &
             1.1218116520614487_dp, 1.5708499759825103_dp, 2.234699654725105_dp, &
             3.3109224238827597_dp, 5.5731123242900455_dp]

      CASE (1193)

         aw(:) = & ! a
            [1.8832454645127752e-11_dp, 1.3341397920410822e-10_dp, 5.251856211473151e-10_dp, &
             1.6806274858529602e-09_dp, 4.7325599237053945e-09_dp, 1.2145350780997587e-08_dp, &
             2.901000431728642e-08_dp, 6.540745854558825e-08_dp, 1.406043543915497e-07_dp, &
             2.903271726167605e-07_dp, 5.791132406203874e-07_dp, 1.1208983827506908e-06_dp, &
             2.1127721496093733e-06_dp, 3.889504482374426e-06_dp, 7.010442433397599e-06_dp, &
             1.2396292493301988e-05_dp, 2.154199511520687e-05_dp, 3.684456502333815e-05_dp, &
             6.210309919038353e-05_dp, 0.00010327449382157264_dp, 0.0001696060219111398_dp, &
             0.0002753193029486315_dp, 0.00044209552586091334_dp, 0.0007027175163019843_dp, &
             0.0011063718770529844_dp, 0.0017263195788318753_dp, 0.0026709271139560313_dp, &
             0.004099441007738538_dp, 0.006244424140808396_dp, 0.009443503788981868_dp, &
             0.014184076230903665_dp, 0.021165961145848858_dp, 0.031388820042265446_dp, &
             0.0462736045408107_dp, 0.06783059199891159_dp, 0.09889097760022603_dp, &
             0.14342490397854343_dp, 0.2069767569155621_dp, 0.29725934203293913_dp, &
             0.4249635141091604_dp, 0.6048614544773474_dp, 0.8573153683569401_dp, &
             1.2103612061506908_dp, 1.7026497348761334_dp, 2.3877754809970066_dp, &
             3.3411337840574475_dp, 4.672117518664464_dp, 6.549749099815065_dp, &
             9.270879027172652_dp, 13.530107527099426_dp, &
             ! w
             5.168111021826878e-11_dp, 2.0496379428208522e-10_dp, 6.543711637566422e-10_dp, &
             1.8344882645903475e-09_dp, 4.659570606882869e-09_dp, 1.0976875058382018e-08_dp, &
             2.4365520825133506e-08_dp, 5.1523859502446726e-08_dp, 1.0462578317820291e-07_dp, &
             2.0524433853544371e-07_dp, 3.907721069968607e-07_dp, 7.247645287409938e-07_dp, &
             1.3133703441212676e-06_dp, 2.3310999777919722e-06_dp, 4.060769453980099e-06_dp, &
             6.954756483274607e-06_dp, 1.1728012271619507e-05_dp, 1.9497997407817443e-05_dp, &
             3.199341767034342e-05_dp, 5.186310065667579e-05_dp, 8.31296353613136e-05_dp, &
             0.00013185093281100463_dp, 0.00020707860422759718_dp, 0.0003222376210251494_dp, &
             0.0004971006454507531_dp, 0.0007605974228852468_dp, 0.0011547910025351915_dp, &
             0.001740476671181692_dp, 0.0026050274346139384_dp, 0.0038733363126851293_dp, &
             0.005723009943615866_dp, 0.008405375460942598_dp, 0.01227440682604414_dp, &
             0.017826402217761102_dp, 0.02575421037711793_dp, 0.03702109383710072_dp, &
             0.0529610524916278_dp, 0.07541480902703361_dp, 0.10691403997917456_dp, &
             0.15093157441713953_dp, 0.2122239052122146_dp, 0.29730870809954485_dp, &
             0.41515499576106113_dp, 0.5782458471379883_dp, 0.8043835870200232_dp, &
             1.1201873780401594_dp, 1.5690224417390282_dp, 2.232681345384161_dp, &
             3.308707489806258_dp, 5.570557326166346_dp]

      CASE (1194)

         aw(:) = & ! a
            [1.6933386381973608e-11_dp, 1.218531565884373e-10_dp, 4.873815119215195e-10_dp, &
             1.5780665536391437e-09_dp, 4.481887819752296e-09_dp, 1.1577003425606672e-08_dp, &
             2.7794928212615668e-08_dp, 6.293115422799762e-08_dp, 1.3575584645734518e-07_dp, &
             2.8115321090565297e-07_dp, 5.622627057336758e-07_dp, 1.0907452544716564e-06_dp, &
             2.0600531375304626e-06_dp, 3.7992336700230515e-06_dp, 6.858762411892525e-06_dp, &
             1.2145779486737926e-05_dp, 2.1134741749434062e-05_dp, 3.619210256583364e-05_dp, &
             6.107187104691637e-05_dp, 0.0001016651291962295_dp, 0.00016712405325515747_dp, &
             0.0002715341882156855_dp, 0.0004363838372737916_dp, 0.0006941848985890009_dp, &
             0.0010937468158727264_dp, 0.0017078101143075315_dp, 0.002644029356608802_dp, &
             0.004060685660349027_dp, 0.006189044238257731_dp, 0.009365003648255328_dp, &
             0.01407367837679615_dp, 0.02101190513574534_dp, 0.031175485299022553_dp, &
             0.045980429521803465_dp, 0.06743076579751651_dp, 0.09834988958975514_dp, &
             0.1426983554461946_dp, 0.20600896989259262_dp, 0.29598084971553934_dp, &
             0.42328908966690343_dp, 0.6026883342187473_dp, 0.854522148326354_dp, &
             1.206807920042408_dp, 1.698179706748384_dp, 2.3822194873168745_dp, &
             3.3343166776349453_dp, 4.66386646933231_dp, 6.539900397709868_dp, &
             9.259274656042267_dp, 13.516545598033831_dp, &
             ! w
             4.66614881886532e-11_dp, 1.8933557667873048e-10_dp, 6.139987638171744e-10_dp, &
             1.7381735950061203e-09_dp, 4.445809017791955e-09_dp, 1.0529555421262821e-08_dp, &
             2.3473186281847474e-08_dp, 4.981317873668608e-08_dp, 1.0145508869484276e-07_dp, &
             1.9953638242783558e-07_dp, 3.8075568297554297e-07_dp, 7.075813630226601e-07_dp, &
             1.2844853634428033e-06_dp, 2.2834280873828597e-06_dp, 3.983398288862977e-06_dp, &
             6.831099596993337e-06_dp, 1.1533167935055234e-05_dp, 1.9195010820715643e-05_dp, &
             3.1528044607795465e-05_dp, 5.1156543854231374e-05_dp, 8.20685659320012e-05_dp, &
             0.00013027391260347172_dp, 0.00020475778744905238_dp, 0.0003188543318686592_dp, &
             0.0004922131037115729_dp, 0.0007535985290744227_dp, 0.0011448537146930954_dp, &
             0.0017264841396220075_dp, 0.0025854847629249913_dp, 0.003846260672965591_dp, &
             0.005685795928467997_dp, 0.008354633210857176_dp, 0.012205771929840372_dp, &
             0.017734319076423512_dp, 0.025631695239573836_dp, 0.036859490837095894_dp, &
             0.052749806716034374_dp, 0.07514129594571878_dp, 0.10656351119382258_dp, &
             0.15048730524321635_dp, 0.21166765659632777_dp, 0.29662163379509127_dp, &
             0.41431911368221713_dp, 0.5772461106928231_dp, 0.8032103734233277_dp, &
             1.118838765656578_dp, 1.567504823155196_dp, 2.231005129349355_dp, &
             3.3068679113718926_dp, 5.568435421223672_dp]

      CASE (1195)

         aw(:) = & ! a
            [1.5428869517910908e-11_dp, 1.126502126605945e-10_dp, 4.5702575750145733e-10_dp, &
             1.4950615729696237e-09_dp, 4.277763711152499e-09_dp, 1.1111965699702537e-08_dp, &
             2.6796870153530614e-08_dp, 6.089063006168848e-08_dp, 1.317498152010162e-07_dp, &
             2.735557933149908e-07_dp, 5.48279851419471e-07_dp, 1.065679338241227e-06_dp, &
             2.0161593842636e-06_dp, 3.723968021280084e-06_dp, 6.732133500204096e-06_dp, &
             1.1936397308182001e-05_dp, 2.0793991582530223e-05_dp, 3.564565099397548e-05_dp, &
             6.020741444572833e-05_dp, 0.0001003149027429537_dp, 0.00016504010518315622_dp, &
             0.00026835376008125253_dp, 0.0004315813428526055_dp, 0.0006870059250760717_dp, &
             0.0010831182436916282_dp, 0.001692218831728957_dp, 0.002621360108104943_dp, &
             0.004028006309134849_dp, 0.006142324175789108_dp, 0.009298748338283109_dp, &
             0.013980460266523255_dp, 0.020881768733191038_dp, 0.03099520264973475_dp, &
             0.045732582473846664_dp, 0.0670926346086362_dp, 0.09789213547231863_dp, &
             0.1420835001837104_dp, 0.20518970218597785_dp, 0.2948982297006441_dp, &
             0.4218707846072679_dp, 0.6008471027837289_dp, 0.8521548957626351_dp, &
             1.2037957565105737_dp, 1.6943895145914585_dp, 2.3775074539008187_dp, &
             3.3285339265628604_dp, 4.656866074795981_dp, 6.531543188217418_dp, &
             9.249426336979775_dp, 13.505034693534814_dp, &
             ! w
             4.268204773233358e-11_dp, 1.7682408867467472e-10_dp, 5.813164511803607e-10_dp, &
             1.6596456444800294e-09_dp, 4.270640454730058e-09_dp, 1.0161555915719804e-08_dp, &
             2.2736741876760634e-08_dp, 4.83975917512049e-08_dp, 9.882538456748687e-08_dp, &
             1.9479301674636858e-07_dp, 3.724175698579294e-07_dp, 6.932554329247844e-07_dp, &
             1.260370388847176e-06_dp, 2.243579389330969e-06_dp, 3.918651413139414e-06_dp, &
             6.727512920020203e-06_dp, 1.1369793887665192e-05_dp, 1.8940739804411977e-05_dp, &
             3.113718104418048e-05_dp, 5.056266611801506e-05_dp, 8.117608681098094e-05_dp, &
             0.00012894659011806353_dp, 0.00020280323377375447_dp, 0.0003160033283752813_dp, &
             0.00048809225355894974_dp, 0.0007476944714632414_dp, 0.0011364668011871696_dp, &
             0.0017146691601525692_dp, 0.0025689760531452306_dp, 0.003823378811485668_dp, &
             0.005654333359363392_dp, 0.008311716678657191_dp, 0.0121477008480753_dp, &
             0.017656381232439927_dp, 0.025527965071051196_dp, 0.036722621562450666_dp, &
             0.05257083669572415_dp, 0.07490950266538153_dp, 0.10626636382407_dp, &
             0.1501105885198186_dp, 0.2111958618200197_dp, 0.2960387285926111_dp, &
             0.4136097922086062_dp, 0.5763975555871754_dp, 0.8022143749873136_dp, &
             1.1176936650250209_dp, 1.5662160436087877_dp, 2.2295815350756283_dp, &
             3.3053055146325363_dp, 5.566633291056069_dp]

      CASE (1196)

         aw(:) = & ! a
            [1.4205366813477368e-11_dp, 1.0513332027569587e-10_dp, 4.320373247626252e-10_dp, &
             1.4262571057956208e-09_dp, 4.107651873223626e-09_dp, 1.0722794242347856e-08_dp, &
             2.5958834309732795e-08_dp, 5.9172535456116426e-08_dp, 1.2836893752849081e-07_dp, &
             2.6713118010469304e-07_dp, 5.364349890094574e-07_dp, 1.0444135831300498e-06_dp, &
             1.978869628845653e-06_dp, 3.659948498092734e-06_dp, 6.624306807373153e-06_dp, &
             1.175792633729086e-05_dp, 2.050328073030349e-05_dp, 3.517905395774211e-05_dp, &
             5.9468710242622196e-05_dp, 9.916026318668874e-05_dp, 0.00016325683404955733_dp, &
             0.0002656305109796341_dp, 0.00042746679026842995_dp, 0.0006808519393939835_dp, &
             0.0010740024735728254_dp, 0.0016788401961063946_dp, 0.0026018990238441267_dp, &
             0.003999939525275833_dp, 0.006102181878503007_dp, 0.00924179881084461_dp, &
             0.013900304997082091_dp, 0.020769828745896623_dp, 0.03084007546327462_dp, &
             0.04551924920381181_dp, 0.06680149943922198_dp, 0.09749788664403933_dp, &
             0.14155379504544682_dp, 0.20448370176408534_dp, 0.2939650443477448_dp, &
             0.420647945539331_dp, 0.5992592473850633_dp, 0.8501129454459876_dp, &
             1.2011969613648552_dp, 1.6911188031151465_dp, 2.373440484195271_dp, &
             3.323541955840734_dp, 4.650822031023149_dp, 6.524326703053585_dp, &
             9.240921282877906_dp, 13.495092868493074_dp, &
             ! w
             3.944395947915385e-11_dp, 1.665520563448468e-10_dp, 5.542195993863368e-10_dp, &
             1.594132427822787e-09_dp, 4.123860704574339e-09_dp, 9.85214725364671e-09_dp, &
             2.2115841135501796e-08_dp, 4.720135325006146e-08_dp, 9.659880093327974e-08_dp, &
             1.907699677033464e-07_dp, 3.653351362874308e-07_dp, 6.810708595296777e-07_dp, &
             1.2398357553908765e-06_dp, 2.2096108681086434e-06_dp, 3.863405373435621e-06_dp, &
             6.639048276554992e-06_dp, 1.1230156717758909e-05_dp, 1.8723249590718648e-05_dp, &
             3.08026248422546e-05_dp, 5.005401389738086e-05_dp, 8.041122444378521e-05_dp, &
             0.0001278084222424459_dp, 0.0002011263338847422_dp, 0.00031355610596579424_dp, &
             0.0004845533685478475_dp, 0.0007426219597297248_dp, 0.001129258094100234_dp, &
             0.0017045099091785836_dp, 0.002554775447429477_dp, 0.003803688967103374_dp, &
             0.005627250465117789_dp, 0.008274762056376665_dp, 0.012097681147608032_dp, &
             0.017589228949295932_dp, 0.025438563883459917_dp, 0.03660462620941226_dp, &
             0.052416504982521366_dp, 0.07470956844188095_dp, 0.10600999498756812_dp, &
             0.14978549298534388_dp, 0.21078862443210603_dp, 0.29553547579022443_dp, &
             0.4129972728122464_dp, 0.5756646653434719_dp, 0.8013539917865502_dp, &
             1.116704334153313_dp, 1.5651024443878665_dp, 2.228351341309453_dp, &
             3.303955320081216_dp, 5.565075951473173_dp]

      CASE (1197)

         aw(:) = & ! a
            [8.398876653269376e-12_dp, 6.89042083248209e-11_dp, 3.0838587997088325e-10_dp, &
             1.078051906786543e-09_dp, 3.2319212225754043e-09_dp, 8.692687332804405e-09_dp, &
             2.1540759709164525e-08_dp, 5.003570214016123e-08_dp, 1.1025731894065299e-07_dp, &
             2.3249750292920072e-07_dp, 4.722328143993024e-07_dp, 9.285933432750631e-07_dp, &
             1.7749093128841369e-06_dp, 3.3084451046838776e-06_dp, 6.030228456613886e-06_dp, &
             1.0771535159743405e-05_dp, 1.8891929341752533e-05_dp, 3.258596487423732e-05_dp, &
             5.5353382494445585e-05_dp, 9.27132084815545e-05_dp, 0.00015327879943589334_dp, &
             0.00025036304722469993_dp, 0.0004043568050751485_dp, 0.0006462275460117554_dp, &
             0.0010226308177412693_dp, 0.0016033298409316003_dp, 0.002491899763354035_dp, &
             0.0038410811520282018_dp, 0.005874680596882282_dp, 0.008918646113698192_dp, &
             0.013444939829026809_dp, 0.020133181300227666_dp, 0.029956862715425678_dp, &
             0.044303402521743085_dp, 0.06514061976160614_dp, 0.09524666215311117_dp, &
             0.13852639311470552_dp, 0.2004452788148281_dp, 0.28862272575615205_dp, &
             0.4136419359223435_dp, 0.5901551555572233_dp, 0.8383969374383553_dp, &
             1.1862759131538265_dp, 1.6723280051898266_dp, 2.350061220853425_dp, &
             3.2948296777750756_dp, 4.616041603597973_dp, 6.482781806857452_dp, &
             9.191940664108245_dp, 13.437821320652803_dp, &
             ! w
             2.4047451006448983e-11_dp, 1.1615429745865822e-10_dp, 4.1699489533809246e-10_dp, &
             1.2557796645535653e-09_dp, 3.3551913077477594e-09_dp, 8.214290964290214e-09_dp, &
             1.8800357947725935e-08_dp, 4.0767171468226905e-08_dp, 8.454855235287015e-08_dp, &
             1.6888066955056285e-07_dp, 3.266189613860745e-07_dp, 6.14186818983472e-07_dp, &
             1.1266970845948102e-06_dp, 2.021828829627162e-06_dp, 3.557069369378233e-06_dp, &
             6.147153878869648e-06_dp, 1.0451743983453825e-05_dp, 1.750798819364048e-05_dp, &
             2.8929161797203043e-05_dp, 4.719986318372929e-05_dp, 7.611130360250284e-05_dp, &
             0.00012139850699300437_dp, 0.00019166668830450828_dp, 0.0002997293415192314_dp, &
             0.0004645291934489775_dp, 0.000713879928006679_dp, 0.0010883578208034843_dp, &
             0.0016467966172676355_dp, 0.0024740073851396106_dp, 0.0036915725584781003_dp, &
             0.005472869609165712_dp, 0.008063890651917026_dp, 0.01181197397575496_dp, &
             0.017205297315569704_dp, 0.024926961467297178_dp, 0.03592880071915235_dp, &
             0.05153181665966192_dp, 0.07356254172488792_dp, 0.10453805800075097_dp, &
             0.14791756955752258_dp, 0.2084470607150433_dp, 0.2926398660246615_dp, &
             0.4094707033299533_dp, 0.5714425485246447_dp, 0.7963947341525218_dp, &
             1.1109991837109405_dp, 1.5586783331955767_dp, 2.2212529020380547_dp, &
             3.296163741662596_dp, 5.556089857352023_dp]

      CASE (1198)

         aw(:) = & ! a
            [6.298464978866128e-12_dp, 5.5450934854751915e-11_dp, 2.605475331587664e-10_dp, &
             9.38892791386109e-10_dp, 2.8734913603883134e-09_dp, 7.84658415779471e-09_dp, &
             1.967284738136507e-08_dp, 4.612738548674639e-08_dp, 1.0243405786442538e-07_dp, &
             2.1741276527437143e-07_dp, 4.4406774401717306e-07_dp, 8.77462699018481e-07_dp, &
             1.6843640762100574e-06_dp, 3.151620051010987e-06_dp, 5.763983719110363e-06_dp, &
             1.032766445956254e-05_dp, 1.8164128435346706e-05_dp, 3.141073942335293e-05_dp, &
             5.348238416021424e-05_dp, 8.977357130191129e-05_dp, 0.00014871684487209616_dp, &
             0.00024336516414392493_dp, 0.0003937393086201775_dp, 0.0006302847873239445_dp, &
             0.000998927661861944_dp, 0.0015684207658963369_dp, 0.002440952194633685_dp, &
             0.003767375214701119_dp, 0.0057689513094582425_dp, 0.008768227367622955_dp, &
             0.013232662726897272_dp, 0.019835973337770655_dp, 0.029543990170556506_dp, &
             0.04373429828089402_dp, 0.06436224573207555_dp, 0.09419037385197995_dp, &
             0.13710431027491263_dp, 0.1985462283537795_dp, 0.28610792264947005_dp, &
             0.4103407145827731_dp, 0.5858612672754546_dp, 0.8328661837674213_dp, &
             1.1792261551335765_dp, 1.6634427756554122_dp, 2.3389980653397635_dp, &
             3.281233591573469_dp, 4.599561914855748_dp, 6.4630863656212805_dp, &
             9.16870957780673_dp, 13.410647892155138_dp, &
             ! w
             1.846193625822815e-11_dp, 9.689414612185396e-11_dp, 3.6210352329816684e-10_dp, &
             1.1167142457838506e-09_dp, 3.033204096277582e-09_dp, 7.51814970383236e-09_dp, &
             1.7374608141863734e-08_dp, 3.797341954030985e-08_dp, 7.927332557415118e-08_dp, &
             1.5923055066590317e-07_dp, 3.0944540716490855e-07_dp, 5.843573798318635e-07_dp, &
             1.075994090240436e-06_dp, 1.9373077484906326e-06_dp, 3.418642802133384e-06_dp, &
             5.9240782456395526e-06_dp, 1.0097568225985238e-05_dp, 1.6953369686520547e-05_dp, &
             2.8071755102589027e-05_dp, 4.589023244414527e-05_dp, 7.413348850230434e-05_dp, &
             0.000118443474448965_dp, 0.00018729641812362642_dp, 0.0002933287181629532_dp, &
             0.0004552422077101772_dp, 0.0007005259449722193_dp, 0.0010693228763083307_dp, &
             0.0016198939272755507_dp, 0.002436300599461667_dp, 0.0036391549373870295_dp, &
             0.005400592528715038_dp, 0.007965036066181094_dp, 0.011677868296436762_dp, &
             0.01702486948641206_dp, 0.02468625703170825_dp, 0.035610478006329764_dp, &
             0.05111467287379762_dp, 0.07302114791993895_dp, 0.10384262427037946_dp, &
             0.14703421471458286_dp, 0.20733871764400807_dp, 0.2912680947467692_dp, &
             0.40779866415413674_dp, 0.5694392240991495_dp, 0.7940400478830877_dp, &
             1.108288754077354_dp, 1.5556249136484792_dp, 2.217877909619131_dp, &
             3.2924587205127698_dp, 5.551817286742326_dp]

      CASE (1199)

         aw(:) = & ! a
            [5.200498962004415e-12_dp, 4.829384036726918e-11_dp, 2.344617458313477e-10_dp, &
             8.616178532328689e-10_dp, 2.6718494231456656e-09_dp, 7.365920670513124e-09_dp, &
             1.8603543338270428e-08_dp, 4.3876085185056985e-08_dp, 9.790426695012216e-08_dp, &
             2.0864000194791608e-07_dp, 4.2762565806712863e-07_dp, 8.47514652910655e-07_dp, &
             1.6311743837760601e-06_dp, 3.0592533989174246e-06_dp, 5.6068010330328565e-06_dp, &
             1.0065056472468247e-05_dp, 1.773269892114307e-05_dp, 3.071283864408177e-05_dp, &
             5.2369474648186694e-05_dp, 8.802235205542369e-05_dp, 0.00014599532925367868_dp, &
             0.00023918495754235116_dp, 0.00038738911432139846_dp, 0.000620738627852788_dp, &
             0.0009847193810743919_dp, 0.0015474740209835804_dp, 0.002410352316072785_dp, &
             0.0037230659933211495_dp, 0.005705336001843513_dp, 0.008677649117127675_dp, &
             0.013104735437890832_dp, 0.01965673028477948_dp, 0.029294815051225838_dp, &
             0.043390603262657944_dp, 0.06389186516317658_dp, 0.09355165343843032_dp, &
             0.13624389350462626_dp, 0.1973965799416015_dp, 0.28458468919193075_dp, &
             0.4083401123271333_dp, 0.5832578109184712_dp, 0.8295112258927908_dp, &
             1.1749478717160549_dp, 1.6580483491859062_dp, 2.332278757599492_dp, &
             3.27297293830819_dp, 4.589546045042946_dp, 6.4511127084093465_dp, &
             9.154583136574116_dp, 13.394121033576663_dp, &
             ! w
             1.5536914926207933e-11_dp, 8.646166827233514e-11_dp, 3.3160881322789117e-10_dp, &
             1.0383167284394085e-09_dp, 2.84981720924827e-09_dp, 7.118557353071206e-09_dp, &
             1.6551095166823054e-08_dp, 3.6351440682349606e-08_dp, 7.619737805523065e-08_dp, &
             1.5358267679290873e-07_dp, 2.9936172653832667e-07_dp, 5.667925474714513e-07_dp, &
             1.04606202647084e-06_dp, 1.8872974329310547e-06_dp, 3.3365677915870782e-06_dp, &
             5.791564673200419e-06_dp, 9.886814644166937e-06_dp, 1.662281842038474e-05_dp, &
             2.7559993512328815e-05_dp, 4.510748741978461e-05_dp, 7.294988200410512e-05_dp, &
             0.00011667296548465338_dp, 0.00018467506129516323_dp, 0.00028948551663583516_dp, &
             0.0004496604297794594_dp, 0.0006924923157318693_dp, 0.001057861568159926_dp, &
             0.0016036817937346288_dp, 0.0024135596915329256_dp, 0.0036075181175406605_dp, &
             0.005356938124734133_dp, 0.007905288238227472_dp, 0.01159676150060408_dp, &
             0.016915678640418055_dp, 0.024540500469752962_dp, 0.03541760903326102_dp, &
             0.05086178892896324_dp, 0.07269276540488516_dp, 0.10342059299722126_dp, &
             0.14649787837227557_dp, 0.2066654616543834_dp, 0.29043444825096576_dp, &
             0.40678211153374255_dp, 0.5682207824393413_dp, 0.7926073980627067_dp, &
             1.1066391587663271_dp, 1.5537661226046213_dp, 2.2158230274780246_dp, &
             3.290202758214508_dp, 5.549215911161244_dp]

      CASE (1200)

         aw(:) = & ! a
            [4.52600936077113e-12_dp, 4.3838575022849034e-11_dp, 2.1794222540029254e-10_dp, &
             8.12084888966259e-10_dp, 2.5414898303845082e-09_dp, 7.053193625502587e-09_dp, &
             1.790437313161217e-08_dp, 4.239813310626565e-08_dp, 9.492056984375533e-08_dp, &
             2.028451519109831e-07_dp, 4.1673830588642226e-07_dp, 8.27641825755446e-07_dp, &
             1.5958123917610186e-06_dp, 2.9977422075903074e-06_dp, 5.5019676086965e-06_dp, &
             9.889669563440174e-06_dp, 1.7444202445285553e-05_dp, 3.024561978460634e-05_dp, &
             5.162363897224276e-05_dp, 8.6847600632667e-05_dp, 0.00014416803658662917_dp, &
             0.0002363759043020696_dp, 0.00038311850131500725_dp, 0.0006143139494937664_dp, &
             0.0009751504287078737_dp, 0.0015333576823081698_dp, 0.002389717908334761_dp, &
             0.003693169605521691_dp, 0.00566238975002572_dp, 0.008616468420462505_dp, &
             0.013018284578188044_dp, 0.01953554393587264_dp, 0.02912627168771905_dp, &
             0.04315802621645722_dp, 0.06357342985830804_dp, 0.0931190868628312_dp, &
             0.13566096688984164_dp, 0.19661742143519167_dp, 0.283551984405233_dp, &
             0.4069833217356816_dp, 0.5814916176351641_dp, 0.8272345336249507_dp, &
             1.1720437861578876_dp, 1.6543856559913497_dp, 2.3277153705534577_dp, &
             3.267361479099094_dp, 4.582740879180665_dp, 6.442975896202103_dp, &
             9.144981936011286_dp, 13.382886987068508_dp, &
             ! w
             1.373753855902433e-11_dp, 7.988354838072296e-11_dp, 3.1205666205027604e-10_dp, &
             9.875677327497518e-10_dp, 2.7303107979781676e-09_dp, 6.856834801085614e-09_dp, &
             1.6009534289621234e-08_dp, 3.528124626421566e-08_dp, 7.416217022588563e-08_dp, &
             1.498367965741692e-07_dp, 2.926599199976308e-07_dp, 5.550972029411609e-07_dp, &
             1.026099569306192e-06_dp, 1.8538954365854816e-06_dp, 3.281677146299213e-06_dp, &
             5.702834631418154e-06_dp, 9.745540091073546e-06_dp, 1.640101506741568e-05_dp, &
             2.721627428084713e-05_dp, 4.458130786112161e-05_dp, 7.215358960207057e-05_dp, &
             0.00011548092267174342_dp, 0.00018290891087218222_dp, 0.00028689441798671197_dp, &
             0.0004458948155147638_dp, 0.0006870694060004753_dp, 0.0010501205464494075_dp, &
             0.00159272621925415_dp, 0.0023981844441039336_dp, 0.003586118014994989_dp, &
             0.005327395434394467_dp, 0.007864836795338226_dp, 0.011541826304437201_dp, &
             0.016841691983380043_dp, 0.02444169937632993_dp, 0.035286824721833386_dp, &
             0.05069024799393403_dp, 0.07246993538784788_dp, 0.1031341226827681_dp, &
             0.146133705298541_dp, 0.20620818327551377_dp, 0.2898680706807055_dp, &
             0.4060912823648196_dp, 0.5673925469978758_dp, 0.7916333355277967_dp, &
             1.10551737673369_dp, 1.5525018847726937_dp, 2.214425273665476_dp, &
             3.2886681607343244_dp, 5.547446408096967_dp]

      CASE (1201)

         aw(:) = & ! a
            [4.073664492830643e-12_dp, 4.081922102913296e-11_dp, 2.0660342888504193e-10_dp, &
             7.777882957371289e-10_dp, 2.4506785558588123e-09_dp, 6.834356448129326e-09_dp, &
             1.7413392958106678e-08_dp, 4.1357319910184756e-08_dp, 9.281442353159956e-08_dp, &
             1.987465020396972e-07_dp, 4.090245579778006e-07_dp, 8.135407360040196e-07_dp, &
             1.5706875421966694e-06_dp, 2.9539867395219704e-06_dp, 5.427316253957548e-06_dp, &
             9.764657683088674e-06_dp, 1.723838900747784e-05_dp, 2.9912039205200553e-05_dp, &
             5.109074228008123e-05_dp, 8.60076744544312e-05_dp, 0.00014286073071426004_dp, &
             0.00023436503349759345_dp, 0.0003800596898543041_dp, 0.0006097099303365672_dp, &
             0.0009682898646806463_dp, 0.0015232322178460565_dp, 0.0023749107665745057_dp, &
             0.003671707409718264_dp, 0.005631547387852265_dp, 0.0085725147450021_dp, &
             0.012956154677529587_dp, 0.019448421832621073_dp, 0.029005066032899925_dp, &
             0.042990721423218414_dp, 0.06334429709471913_dp, 0.09280774473621764_dp, &
             0.13524129240065813_dp, 0.19605633040171033_dp, 0.28280813023492873_dp, &
             0.40600580556397226_dp, 0.5802188639807403_dp, 0.825593562631136_dp, &
             1.1699501957032066_dp, 1.6517446850917723_dp, 2.324424387476326_dp, &
             3.2633140125968323_dp, 4.577831706858374_dp, 6.437105355695042_dp, &
             9.13805414096656_dp, 13.37478031050644_dp, &
             ! w
             1.2529372293673143e-11_dp, 7.538211694081257e-11_dp, 2.9851644347648603e-10_dp, &
             9.521834921176493e-10_dp, 2.6465907063574317e-09_dp, 6.672826198884552e-09_dp, &
             1.5627694072177836e-08_dp, 3.452491339922673e-08_dp, 7.272100934507552e-08_dp, &
             1.4717981921581675e-07_dp, 2.878993340916098e-07_dp, 5.467788014872245e-07_dp, &
             1.0118848820603522e-06_dp, 1.8300864395353793e-06_dp, 3.2425147187610067e-06_dp, &
             5.63947564702371e-06_dp, 9.644583072763804e-06_dp, 1.624239843124977e-05_dp, &
             2.6970311717147108e-05_dp, 4.420454933211684e-05_dp, 7.158310040228151e-05_dp, &
             0.0001146264537776706_dp, 0.00018164228845061697_dp, 0.0002850353061687006_dp, &
             0.0004431918048714337_dp, 0.0006831751540336526_dp, 0.0010445594549226887_dp, &
             0.0015848528944898983_dp, 0.002387130982952511_dp, 0.0035707280449198283_dp, &
             0.005306142892616442_dp, 0.00783572778350968_dp, 0.011502283187979297_dp, &
             0.01678842047853666_dp, 0.024370542052247438_dp, 0.03519260863010292_dp, &
             0.05056664062830545_dp, 0.07230933255258958_dp, 0.10292760462705913_dp, &
             0.14587111351192195_dp, 0.20587838769816816_dp, 0.289459509823114_dp, &
             0.40559285361131275_dp, 0.5667948760281161_dp, 0.7909303217216849_dp, &
             1.1047076383088907_dp, 1.5515892182791182_dp, 2.213416146996517_dp, &
             3.287560204892521_dp, 5.5461688844098616_dp]

      CASE (1202)

         aw(:) = & ! a
            [3.7547066838519105e-12_dp, 3.867217329187868e-11_dp, 1.9846156181929866e-10_dp, &
             7.53000202666059e-10_dp, 2.3847467111692373e-09_dp, 6.6749420370422115e-09_dp, &
             1.7054802846329026e-08_dp, 4.059556366398386e-08_dp, 9.127029185346383e-08_dp, &
             1.9573714815518227e-07_dp, 4.0335373959351745e-07_dp, 8.031628000409671e-07_dp, &
             1.5521785121483937e-06_dp, 2.921724947952763e-06_dp, 5.372231469564549e-06_dp, &
             9.672347212596773e-06_dp, 1.7086316116077005e-05_dp, 2.9665416104036208e-05_dp, &
             5.069654776232332e-05_dp, 8.538605379337611e-05_dp, 0.00014189275904268408_dp, &
             0.00023287547743079377_dp, 0.00037779295718005603_dp, 0.0006062968337578725_dp, &
             0.0009632021215448261_dp, 0.0015157207434579939_dp, 0.0023639227789501818_dp, &
             0.0036557761431143007_dp, 0.005608646823609468_dp, 0.008539870255888362_dp, &
             0.012909998940503013_dp, 0.019383683964783707_dp, 0.028914980931250816_dp, &
             0.04286634619080865_dp, 0.06317392289631009_dp, 0.0925761964495451_dp, &
             0.13492911642729163_dp, 0.19563888487550593_dp, 0.28225461398950186_dp, &
             0.40527829514671937_dp, 0.57927147361296_dp, 0.824371899445446_dp, &
             1.1683913440267673_dp, 1.6497779971109197_dp, 2.3219733370650184_dp, &
             3.2602992016078396_dp, 4.5741746624604165_dp, 6.432731751647573_dp, &
             9.132892478517155_dp, 13.368739931374455_dp, &
             ! w
             1.1676605989248299e-11_dp, 7.215704160125522e-11_dp, 2.8872903462259433e-10_dp, &
             9.26476948367063e-10_dp, 2.5855544704863635e-09_dp, 6.5383179896018955e-09_dp, &
             1.5347984782115387e-08_dp, 3.396992151956199e-08_dp, 7.166196462546705e-08_dp, &
             1.4522490716331027e-07_dp, 2.843928913548577e-07_dp, 5.406460275827774e-07_dp, &
             1.0013962359807482e-06_dp, 1.812505157301588e-06_dp, 3.2135763006735856e-06_dp, &
             5.592628585357823e-06_dp, 9.569894062773703e-06_dp, 1.6124991119850416e-05_dp, &
             2.6788163733106288e-05_dp, 4.3925415650882254e-05_dp, 7.11602595079101e-05_dp, &
             0.00011399288417472533_dp, 0.00018070277467316926_dp, 0.00028365584387855916_dp, &
             0.0004411855239398697_dp, 0.0006802838094429805_dp, 0.0010404293551557883_dp, &
             0.0015790039585189106_dp, 0.002378917464300399_dp, 0.003559289374981356_dp, &
             0.005290343136193394_dp, 0.00781408245993837_dp, 0.011472872830135959_dp, &
             0.01674879147350705_dp, 0.02431759730685334_dp, 0.03512249379792647_dp, &
             0.05047463647400319_dp, 0.07218977102571825_dp, 0.10277383575801964_dp, &
             0.14567556221586847_dp, 0.20563275254878668_dp, 0.2891551654124624_dp, &
             0.4052215142835455_dp, 0.5663495428337335_dp, 0.790406436050233_dp, &
             1.1041041618168157_dp, 1.5509089786719268_dp, 2.2126639724647386_dp, &
             3.286734348774233_dp, 5.545216656606194_dp]

      CASE (1203)

         aw(:) = & ! a
            [3.5241322899118167e-12_dp, 3.710946041207745e-11_dp, 1.924906273053455e-10_dp, &
             7.347307652261576e-10_dp, 2.3359865961952348e-09_dp, 6.556748703260132e-09_dp, &
             1.6788415210676133e-08_dp, 4.002878110028881e-08_dp, 9.011988474644778e-08_dp, &
             1.9349264736301856e-07_dp, 3.9912019158178314e-07_dp, 7.95408737822913e-07_dp, &
             1.5383390515607981e-06_dp, 2.897586653987624e-06_dp, 5.3309929000904745e-06_dp, &
             9.60320351908077e-06_dp, 1.6972353470518164e-05_dp, 2.9480516527912048e-05_dp, &
             5.040089027979122e-05_dp, 8.491964524058562e-05_dp, 0.00014116622706410273_dp, &
             0.0002317570970166119_dp, 0.00037609054653606287_dp, 0.0006037327299405726_dp, &
             0.0009593789152010876_dp, 0.0015100747999255485_dp, 0.002355661788147829_dp, &
             0.0036437960189680684_dp, 0.0055914222250052215_dp, 0.00851531185868082_dp, &
             0.012875269424959942_dp, 0.01933496362326585_dp, 0.028847173061739267_dp, &
             0.04277271236254311_dp, 0.06304563926909634_dp, 0.09240182535598795_dp, &
             0.13469399345371583_dp, 0.19532443213900547_dp, 0.28183760699820365_dp, &
             0.4047301354783838_dp, 0.5785575540829622_dp, 0.8234511918168674_dp, &
             1.16721638576263_dp, 1.6482954866699802_dp, 2.320125531862285_dp, &
             3.258026186531764_dp, 4.571417216842777_dp, 6.429433783426401_dp, &
             9.129000034041793_dp, 13.364184613538244_dp, &
             ! w
             1.1059608271159237e-11_dp, 6.979585172638759e-11_dp, 2.815148749030197e-10_dp, &
             9.0745649551362e-10_dp, 2.540273561438773e-09_dp, 6.438330503954199e-09_dp, &
             1.5139731049690393e-08_dp, 3.355617284451245e-08_dp, 7.0871581769451e-08_dp, &
             1.4376456349583492e-07_dp, 2.817714151963544e-07_dp, 5.360577996551975e-07_dp, &
             9.935442073504824e-07_dp, 1.7993359701531407e-06_dp, 3.19188900515401e-06_dp, &
             5.5575036932262996e-06_dp, 9.51387006107789e-06_dp, 1.603688979333398e-05_dp, &
             2.665143222941237e-05_dp, 4.3715810463679896e-05_dp, 7.084264357253321e-05_dp, &
             0.00011351684118721693_dp, 0.00017999666281010167_dp, 0.0002826188133541681_dp, &
             0.000439676909182359_dp, 0.0006781091789183688_dp, 0.001037322365369387_dp, &
             0.001574603023770262_dp, 0.002372736138432072_dp, 0.003550679275483157_dp, &
             0.00527844827088446_dp, 0.007797784012010508_dp, 0.011450723923182868_dp, &
             0.01671894232077786_dp, 0.024277712656357434_dp, 0.035069666993686674_dp, &
             0.050405308105438376_dp, 0.0720996654186585_dp, 0.1026579357641754_dp, &
             0.14552815180276382_dp, 0.2054475665042407_dp, 0.28892569272161306_dp, &
             0.40494149889637093_dp, 0.5660136984398277_dp, 0.7900113172351806_dp, &
             1.1036489801936284_dp, 1.5503958656077668_dp, 2.2120965731931475_dp, &
             3.286111353437484_dp, 5.544498329593884_dp]

      CASE (1204)

         aw(:) = & ! a
            [3.357157465916683e-12_dp, 3.5971623848115894e-11_dp, 1.8811771781117414e-10_dp, &
             7.213000884157045e-10_dp, 2.3000476297193427e-09_dp, 6.469466814460489e-09_dp, &
             1.659140518208623e-08_dp, 3.960911107500339e-08_dp, 8.92672346741006e-08_dp, &
             1.9182769922779613e-07_dp, 3.9597753794807777e-07_dp, 7.896491353078233e-07_dp, &
             1.5280536437156758e-06_dp, 2.8796384147072087e-06_dp, 5.300316098528265e-06_dp, &
             9.551747993566587e-06_dp, 1.6887513714391392e-05_dp, 2.934282192093872e-05_dp, &
             5.0180647100628755e-05_dp, 8.457210701411638e-05_dp, 0.00014062471964740027_dp, &
             0.00023092332951340163_dp, 0.00037482108814086045_dp, 0.0006018203146057763_dp, &
             0.00095652683865708_dp, 0.0015058621856158603_dp, 0.0023494969112316352_dp, &
             0.0036348541875670126_dp, 0.005578563932588871_dp, 0.008496976065787648_dp, &
             0.012849335955382974_dp, 0.01929857786527473_dp, 0.028796525609523685_dp, &
             0.042702766170792245_dp, 0.06294979770968015_dp, 0.09227153682716394_dp, &
             0.13451829269729013_dp, 0.19508942624448122_dp, 0.28152592655858305_dp, &
             0.4043203899566875_dp, 0.5780238562230996_dp, 0.8227628484211498_dp, &
             1.1663378869733307_dp, 1.6471869506734955_dp, 2.318743750534047_dp, &
             3.2563263231526336_dp, 4.569354952610676_dp, 6.426967141477912_dp, &
             9.126088642912041_dp, 13.36077730737932_dp, &
             ! w
             1.0612471377090294e-11_dp, 6.806876392959282e-11_dp, 2.762111137121767e-10_dp, &
             8.934323896507781e-10_dp, 2.5068199733122325e-09_dp, 6.364347620285935e-09_dp, &
             1.4985454897151917e-08_dp, 3.324936362244154e-08_dp, 7.028500288287937e-08_dp, &
             1.4268001376931023e-07_dp, 2.798233411666616e-07_dp, 5.326463628309604e-07_dp, &
             9.87703291814928e-07_dp, 1.7895355821220243e-06_dp, 3.1757433003726817e-06_dp, &
             5.531344838658541e-06_dp, 9.472133470556418e-06_dp, 1.597123700583799e-05_dp, &
             2.6549512707081397e-05_dp, 4.3559531590012285e-05_dp, 7.060577775512707e-05_dp, &
             0.00011316174877499471_dp, 0.00017946984820063356_dp, 0.00028184495850184073_dp, &
             0.00043855094362217146_dp, 0.00067648584973354_dp, 0.0010350026682585686_dp, &
             0.001571316755198527_dp, 0.0023681197422902177_dp, 0.0035442481080424407_dp, &
             0.005269562429736511_dp, 0.007785607020485599_dp, 0.011434173909712152_dp, &
             0.016696635992466787_dp, 0.024247903493353675_dp, 0.035030180892807544_dp, &
             0.050353482433641156_dp, 0.07203230138938421_dp, 0.10257127944299295_dp, &
             0.14541792582120228_dp, 0.20530908191299477_dp, 0.2887540758650365_dp, &
             0.40473206609509804_dp, 0.5657624910167334_dp, 0.789715753867823_dp, &
             1.1033084684258092_dp, 1.5500119996064643_dp, 2.21167208333539_dp, &
             3.2856452678511334_dp, 5.543960936180039_dp]

      CASE (1205)

         aw(:) = & ! a
            [3.2397250826178528e-12_dp, 3.516803232745436e-11_dp, 1.850158490560689e-10_dp, &
             7.117462914786814e-10_dp, 2.2744333339458994e-09_dp, 6.4071712097974315e-09_dp, &
             1.6450639084872275e-08_dp, 3.930898745494358e-08_dp, 8.865702488118343e-08_dp, &
             1.9063542499318127e-07_dp, 3.937258846135361e-07_dp, 7.855205852423613e-07_dp, &
             1.5206779493204314e-06_dp, 2.8667630284326906e-06_dp, 5.278302579135075e-06_dp, &
             9.514812922575608e-06_dp, 1.682659897272245e-05_dp, 2.9243933331043068e-05_dp, &
             5.002243873047172e-05_dp, 8.432240626744782e-05_dp, 0.00014023558009894154_dp, &
             0.00023032405775185542_dp, 0.00037390851005134556_dp, 0.0006004453172285065_dp, &
             0.0009544759373966216_dp, 0.0015028325148772515_dp, 0.0023450626131358565_dp, &
             0.003628421675971125_dp, 0.005569312943403241_dp, 0.00848378278744663_dp, &
             0.012830673903466996_dp, 0.019272391591749413_dp, 0.02876007193561765_dp, &
             0.042652417586169014_dp, 0.06288080316686598_dp, 0.09217773678806376_dp, &
             0.13439178853363823_dp, 0.1949202097311847_dp, 0.28130148419684703_dp, &
             0.4040253102685132_dp, 0.5776394867463892_dp, 0.8222670725606329_dp, &
             1.1657051159645198_dp, 1.6463884438188503_dp, 2.317748368725948_dp, &
             3.2551017527209054_dp, 4.567869253104574_dp, 6.425190061397887_dp, &
             9.123991090491648_dp, 13.358322418186328_dp, &
             ! w
             1.0297819439974801e-11_dp, 6.684480642244102e-11_dp, 2.7243817978784534e-10_dp, &
             8.834345946793629e-10_dp, 2.4829352173324413e-09_dp, 6.311466923230022e-09_dp, &
             1.4875085347681011e-08_dp, 3.302971256303185e-08_dp, 6.986480378681162e-08_dp, &
             1.419026870748678e-07_dp, 2.784264742827214e-07_dp, 5.301992241108315e-07_dp, &
             9.835119348841818e-07_dp, 1.7825007521552061e-06_dp, 3.164150437517696e-06_dp, &
             5.5125575303639026e-06_dp, 9.442151139928284e-06_dp, 1.5924063746201672e-05_dp, &
             2.6476266167802237e-05_dp, 4.344719774409618e-05_dp, 7.043548829995319e-05_dp, &
             0.00011290642174873464_dp, 0.00017909098817292204_dp, 0.00028128835995064615_dp, &
             0.0004377409797530387_dp, 0.0006753179599892068_dp, 0.0010333335839891695_dp, &
             0.0015689519299480997_dp, 0.0023647973906360245_dp, 0.0035396192209731932_dp, &
             0.005263166150108631_dp, 0.007776840866903007_dp, 0.011422258587668542_dp, &
             0.016680575006771205_dp, 0.024226438588028004_dp, 0.035001745643305715_dp, &
             0.050316158285623795_dp, 0.07198378328724706_dp, 0.10250886208020232_dp, &
             0.1453385264324728_dp, 0.20520932083363425_dp, 0.2886304400043743_dp, &
             0.4045811789746617_dp, 0.5655814984702185_dp, 0.789502794471978_dp, &
             1.1030631146791507_dp, 1.5497354009180733_dp, 2.211366209274297_dp, &
             3.2853094219597914_dp, 5.543573716653388_dp]

      CASE (1206)

         aw(:) = & ! a
            [3.125478264501053e-12_dp, 3.4383397202243596e-11_dp, 1.8197591880945277e-10_dp, &
             7.023610569524158e-10_dp, 2.249230330318897e-09_dp, 6.345803277574686e-09_dp, &
             1.63118420122016e-08_dp, 3.901284364471683e-08_dp, 8.805453978113409e-08_dp, &
             1.8945763738288497e-07_dp, 3.915006053813974e-07_dp, 7.814388239296914e-07_dp, &
             1.5133833768603852e-06_dp, 2.854025417360324e-06_dp, 5.256518712762183e-06_dp, &
             9.478254190066363e-06_dp, 1.67662914095672e-05_dp, 2.9146010354233887e-05_dp, &
             4.9865745606597965e-05_dp, 8.407505380104078e-05_dp, 0.00013985003764394093_dp, &
             0.00022973023594134484_dp, 0.000373004104002962_dp, 0.0005990824533249532_dp, &
             0.0009524428824913285_dp, 0.001499828857423095_dp, 0.002340665904943646_dp, &
             0.0036220430289072673_dp, 0.005560138512999057_dp, 0.008470697462728619_dp, &
             0.012812162895518484_dp, 0.019246415043401785_dp, 0.02872390727347424_dp, &
             0.042602464289439694_dp, 0.0628123452077373_dp, 0.09208465961166945_dp, &
             0.13426625067374842_dp, 0.19475227473753778_dp, 0.2810787274949731_dp, &
             0.40373242891340116_dp, 0.5772579584267157_dp, 0.8217749335296338_dp, &
             1.1650769527507772_dp, 1.6455957112729942_dp, 2.3167601384691117_dp, &
             3.2538859285301323_dp, 4.566394107979387_dp, 6.423425544485625_dp, &
             9.121908301973829_dp, 13.355884737234929_dp, &
             ! w
             9.991541948910744e-12_dp, 6.564620957415059e-11_dp, 2.6873169083171955e-10_dp, &
             8.735952086491843e-10_dp, 2.4593996497875046e-09_dp, 6.259310407701984e-09_dp, &
             1.4766146322107644e-08_dp, 3.281277679840137e-08_dp, 6.944958797109736e-08_dp, &
             1.4113424475977626e-07_dp, 2.7704505351087713e-07_dp, 5.277783504704841e-07_dp, &
             9.793643551758935e-07_dp, 1.7755375664409923e-06_dp, 3.1526728992760615e-06_dp, &
             5.493953088569735e-06_dp, 9.412454763155619e-06_dp, 1.5877331863776244e-05_dp, &
             2.6403692694172185e-05_dp, 4.333587870441741e-05_dp, 7.02667126956813e-05_dp, &
             0.00011265333032345016_dp, 0.00017871539779676555_dp, 0.0002807364991624313_dp, &
             0.00043693781994796037_dp, 0.0006741597579947089_dp, 0.001031678178264238_dp, &
             0.0015666062602483536_dp, 0.0023615016495277174_dp, 0.003535027009082381_dp, &
             0.005256820021572774_dp, 0.007768142756220681_dp, 0.011410434855261744_dp, &
             0.016664636317264607_dp, 0.024205135633823375_dp, 0.03497352302673261_dp, &
             0.05027911082051581_dp, 0.07193562179292327_dp, 0.1024468996652568_dp, &
             0.1452597010612601_dp, 0.20511027518939529_dp, 0.2885076838251999_dp, &
             0.4044313575287212_dp, 0.5654017758672052_dp, 0.7892913207814454_dp, &
             1.1028194639215534_dp, 1.5494607132121472_dp, 2.2110624397192726_dp, &
             3.284975878161778_dp, 5.543189139984311_dp]

      CASE (1207)

         aw(:) = & ! a
            [2.6953651372154835e-12_dp, 2.7997215379009496e-11_dp, 1.4427412546555194e-10_dp, &
             5.486742902801282e-10_dp, 1.740280646228406e-09_dp, 4.876355478096589e-09_dp, &
             1.2469601687670536e-08_dp, 2.970081187424903e-08_dp, 6.681150999691188e-08_dp, &
             1.433472637502857e-07_dp, 2.9550712467159207e-07_dp, 5.886117683527617e-07_dp, &
             1.1378704597779898e-06_dp, 2.142406820377834e-06_dp, 3.940188212630798e-06_dp, &
             7.09551447660667e-06_dp, 1.2536660328218212e-05_dp, 2.176998549966967e-05_dp, &
             3.720953136357181e-05_dp, 6.267949987241335e-05_dp, 0.00010417341249152186_dp, &
             0.0001709914357533802_dp, 0.00027743083804796133_dp, 0.00044527999151911185_dp, &
             0.0007074722061975458_dp, 0.0011134034885776788_dp, 0.0017366236548105182_dp, &
             0.0026858941686780263_dp, 0.004120996963909009_dp, 0.006275214353920286_dp, &
             0.00948713170977586_dp, 0.014245409532747949_dp, 0.021251519704697512_dp, &
             0.03150726114619474_dp, 0.04643632056901852_dp, 0.06805243370822688_dp, &
             0.09919111087718714_dp, 0.1438277977580932_dp, 0.20751328267423405_dp, &
             0.29796793807661226_dp, 0.4258913278846893_dp, 0.6060653213969585_dp, &
             0.8588624177314803_dp, 1.2123288139000439_dp, 1.7051244958266485_dp, &
             2.3908509067538746_dp, 3.3449066440276702_dp, 4.676683293178817_dp, &
             6.555198234341016_dp, 9.277298847303436_dp, 13.537609667402254_dp, &
             ! w
             8.411207579440785e-12_dp, 5.240031283795205e-11_dp, 2.103225476305145e-10_dp, &
             6.760796058043235e-10_dp, 1.889013991568849e-09_dp, 4.781103714808941e-09_dp, &
             1.1230742976502488e-08_dp, 2.4870760429663843e-08_dp, 5.2490373224539076e-08_dp, &
             1.0641384236556784e-07_dp, 2.084580020605577e-07_dp, 3.964034218301193e-07_dp, &
             7.344127461369051e-07_dp, 1.329570535536136e-06_dp, 2.3578092878837692e-06_dp, &
             4.104077888264411e-06_dp, 7.023914030070038e-06_dp, 1.1836897019814347e-05_dp, &
             1.9667192158763175e-05_dp, 3.225311770200931e-05_dp, 5.225714537933315e-05_dp, &
             8.372104446471076e-05_dp, 0.00013272943602406266_dp, 0.00020837078505902_dp, &
             0.00032412045024059346_dp, 0.0004998193591554492_dp, 0.0007644888987681849_dp, &
             0.001160313996688207_dp, 0.001748250479067397_dp, 0.0026158807025909174_dp, &
             0.0038883678022026626_dp, 0.0057436629628041065_dp, 0.008433527312094036_dp, &
             0.01231247380821109_dp, 0.017877459232714152_dp, 0.02582212168649262_dp, &
             0.03711064758133462_dp, 0.05307808566865877_dp, 0.07556630118211012_dp, &
             0.10710814247732663_dp, 0.1511775280797085_dp, 0.21253178424184316_dp, &
             0.29768891812914194_dp, 0.4156174606492107_dp, 0.5787988652723517_dp, &
             0.805032458867547_dp, 1.1209331522893193_dp, 1.5698615828219604_dp, &
             2.233608115708463_dp, 3.3097245630825047_dp, 5.571730544888998_dp]

      CASE (1208)

         aw(:) = & ! a
            [2.2987753325106644e-12_dp, 2.5288840390898755e-11_dp, 1.3384250895191774e-10_dp, &
             5.165835487381851e-10_dp, 1.6542992451455838e-09_dp, 4.6673110153412825e-09_dp, &
             1.1997289665618487e-08_dp, 2.8693778966047503e-08_dp, 6.476373617901964e-08_dp, &
             1.3934528009687168e-07_dp, 2.87947018037518e-07_dp, 5.747449066653641e-07_dp, &
             1.1130869837524305e-06_dp, 2.099123448797096e-06_dp, 3.866146963569737e-06_dp, &
             6.971215316939249e-06_dp, 1.2331535480769493e-05_dp, 2.14367657864907e-05_dp, &
             3.6676042474232985e-05_dp, 6.183684250022084e-05_dp, 0.00010285910495420848_dp, &
             0.00016896560709380476_dp, 0.000274342925470388_dp, 0.0004406225864954028_dp, &
             0.0007005176742539175_dp, 0.0011031177231567836_dp, 0.001721549783446806_dp, &
             0.002663997187387399_dp, 0.004089458129921753_dp, 0.006230161749339109_dp, &
             0.009423290788794258_dp, 0.014155654861944469_dp, 0.021126306817652908_dp, &
             0.03133391598150247_dp, 0.046198164157487766_dp, 0.06772772284885792_dp, &
             0.09875178326984561_dp, 0.14323802493203816_dp, 0.2067278593765609_dp, &
             0.29693057660691485_dp, 0.42453298444208437_dp, 0.6043027616985136_dp, &
             0.8565973286342782_dp, 1.2094478716968953_dp, 1.701500870630818_dp, &
             2.3863476340066283_dp, 3.339381984463105_dp, 4.669997395044592_dp, &
             6.547218624591303_dp, 9.267897634327612_dp, 13.52662337991838_dp, &
             ! w
             7.348735503278848e-12_dp, 4.828250551627061e-11_dp, 1.976510108046221e-10_dp, &
             6.425255250094272e-10_dp, 1.8088778555786032e-09_dp, 4.603695869468038e-09_dp, &
             1.0860437214810826e-08_dp, 2.413365658573958e-08_dp, 5.107987427098065e-08_dp, &
             1.0380363271600794e-07_dp, 2.0376545050230847e-07_dp, 3.88178713227328e-07_dp, &
             7.203182779276715e-07_dp, 1.3059002600173353e-06_dp, 2.3187774027460278e-06_dp, &
             4.040778961459714e-06_dp, 6.922820180908104e-06_dp, 1.167770966679104e-05_dp, &
             1.941980312844536e-05_dp, 3.187335369893633e-05_dp, 5.1680867186976667e-05_dp, &
             8.285604354915241e-05_dp, 0.00013144441225572458_dp, 0.00020648049704743952_dp, &
             0.00032136590214240266_dp, 0.000495841625390895_dp, 0.0007587948969656515_dp, &
             0.0011522322130658837_dp, 0.001736874355396618_dp, 0.0025999971548880656_dp, &
             0.0038663682301604183_dp, 0.005713434236635728_dp, 0.008392320759088466_dp, &
             0.01225675140556888_dp, 0.0178027183961876_dp, 0.025722703706221962_dp, &
             0.036979540522583336_dp, 0.052906741215271724_dp, 0.07534449727550047_dp, &
             0.10682394027712179_dp, 0.15081739241559536_dp, 0.21208095850110392_dp, &
             0.29713215931737724_dp, 0.4149402304517504_dp, 0.5779890054599727_dp, &
             0.8040822022741917_dp, 1.119840960756182_dp, 1.5686326362384764_dp, &
             2.2322508253323976_dp, 3.308235029586777_dp, 5.570012368388973_dp]

      CASE (1209)

         aw(:) = & ! a
            [2.3407904165357795e-12_dp, 2.3026093517431488e-11_dp, 1.1542721823709632e-10_dp, &
             4.3215763818412123e-10_dp, 1.3567393282614966e-09_dp, 3.7737965944957335e-09_dp, &
             9.596289885374707e-09_dp, 2.2755609613045154e-08_dp, 5.100251720993843e-08_dp, &
             1.0909538116614804e-07_dp, 2.2431363362120118e-07_dp, 4.4579838652841907e-07_dp, &
             8.600911740797687e-07_dp, 1.6165655355500464e-06_dp, 2.9684415630604836e-06_dp, &
             5.338052533630034e-06_dp, 9.419440759284204e-06_dp, 1.6337828568464053e-05_dp, &
             2.7894930316267464e-05_dp, 4.6942690398144265e-05_dp, 7.794758928467206e-05_dp, &
             0.00012783556529411395_dp, 0.00020724717958999061_dp, 0.0003323884710440661_dp, &
             0.000527741246048743_dp, 0.0008300058186735309_dp, 0.0012938009688005959_dp, &
             0.0019998556603703053_dp, 0.003066715761755708_dp, 0.004667384650730215_dp, &
             0.007052856589439585_dp, 0.010585235944186724_dp, 0.01578412984505154_dp, &
             0.023391344310993326_dp, 0.03446071997631776_dp, 0.05048236629075492_dp, &
             0.07355379503887585_dp, 0.1066147858782969_dp, 0.15376860683246257_dp, &
             0.2207199832210546_dp, 0.3153707508591033_dp, 0.4486287726021707_dp, &
             0.6355069646167577_dp, 0.896622545314332_dp, 1.260264401314268_dp, &
             1.765309814668173_dp, 2.4655220029413387_dp, 3.4363740695586737_dp, &
             4.78722466309487_dp, 6.686971237392607_dp, 9.432391132251375_dp, &
             13.718702399649647_dp, &
             ! w
             7.1474871786444975e-12_dp, 4.2218737550369666e-11_dp, 1.6597109826890778e-10_dp, &
             5.271763933642269e-10_dp, 1.4611060284927318e-09_dp, 3.6761290726879375e-09_dp, &
             8.595394873004495e-09_dp, 1.8964056000009034e-08_dp, 3.9901054662485086e-08_dp, &
             8.068103780505702e-08_dp, 1.5769521049535549e-07_dp, 2.992871916799416e-07_dp, &
             5.535304843818029e-07_dp, 1.0005641465926448e-06_dp, 1.7719122492174988e-06_dp, &
             3.0803928957737785e-06_dp, 5.2659311608670375e-06_dp, 8.865045925540588e-06_dp, &
             1.4715286390124864e-05_dp, 2.411086746735178e-05_dp, 3.903276672144136e-05_dp, &
             6.248638046522871e-05_dp, 9.899355466880988e-05_dp, 0.00015530484377200752_dp, &
             0.00024142414404246595_dp, 0.00037207399707328185_dp, 0.0005687793403844799_dp, &
             0.0008628149321845518_dp, 0.0012993532542658326_dp, 0.001943273557644775_dp, &
             0.0028872596313858252_dp, 0.004263038033031039_dp, 0.006256908567698866_dp, &
             0.009131119238071684_dp, 0.013253170951703716_dp, 0.0191358456449394_dp, &
             0.02749169275245853_dp, 0.03930696336159951_dp, 0.05594166842478813_dp, &
             0.07926475368655359_dp, 0.1118366951882809_dp, 0.15715689416887765_dp, &
             0.2200018604392296_dp, 0.30689664956232987_dp, 0.426797377874139_dp, &
             0.5921459313608238_dp, 0.8206697633418121_dp, 1.1388828160010713_dp, &
             1.590038012510657_dp, 2.2558763383971057_dp, 3.3341558442714043_dp, &
             5.599919295308823_dp]

      CASE (1210)

         aw(:) = & ! a
            [1.7215858121060493e-12_dp, 1.8832689733858825e-11_dp, 9.942207925282854e-11_dp, &
             3.8320311904829207e-10_dp, 1.2260672652448378e-09_dp, 3.456928378426848e-09_dp, &
             8.881721895414182e-09_dp, 2.1234203339607483e-08_dp, 4.791205454327192e-08_dp, &
             1.0306045049535306e-07_dp, 2.1291976809543524e-07_dp, 4.2490824711009216e-07_dp, &
             8.227655440499326e-07_dp, 1.5513875554632664e-06_dp, 2.8569517375818406e-06_dp, &
             5.150875061881271e-06_dp, 9.110506251001971e-06_dp, 1.5835858110930888e-05_dp, &
             2.7091020742147094e-05_dp, 4.567240485710637e-05_dp, 7.596537918525689e-05_dp, &
             0.0001247785908841864_dp, 0.00020258461107659117_dp, 0.0003253510871280861_dp, &
             0.0005172245427165007_dp, 0.000814437834406974_dp, 0.001270963559975791_dp, &
             0.001966644967579975_dp, 0.0030188240597172258_dp, 0.004598881678840358_dp, &
             0.006955643895803914_dp, 0.010448342935932704_dp, 0.015592815495983311_dp, &
             0.023125964478061006_dp, 0.03409531805627914_dp, 0.049982946117380966_dp, &
             0.07287624186604075_dp, 0.1057024174534087_dp, 0.15254936669880356_dp, &
             0.21910333872027388_dp, 0.31324443635711247_dp, 0.44585563186003696_dp, &
             0.6319222993674355_dp, 0.8920325639434107_dp, 1.2544465515925423_dp, &
             1.7580158995174553_dp, 2.4564848605211727_dp, 3.425317975392787_dp, &
             4.7738780356993455_dp, 6.671076769088245_dp, 9.413699412921002_dp, &
             13.696891853017377_dp, &
             ! w
             5.490266999958233e-12_dp, 3.588866339168042e-11_dp, 1.466442630997566e-10_dp, &
             4.76212944143142e-10_dp, 1.3397222359946064e-09_dp, 3.407916000914313e-09_dp, &
             8.036316634240367e-09_dp, 1.785230610998599e-08_dp, 3.777518278075605e-08_dp, &
             7.674899195756313e-08_dp, 1.506286774134826e-07_dp, 2.869039306683061e-07_dp, &
             5.32311013371492e-07_dp, 9.649265446692393e-07_dp, 1.7131379393274911e-06_dp, &
             2.9850545284375275e-06_dp, 5.113617823032839e-06_dp, 8.625106873006436e-06_dp, &
             1.4342217617640284e-05_dp, 2.3537840046897734e-05_dp, 3.816263550102344e-05_dp, &
             6.117930930743289e-05_dp, 9.705014244542327e-05_dp, 0.00015244332693338544_dp, &
             0.00023724988118379198_dp, 0.0003660390060814293_dp, 0.0005601291634564192_dp, &
             0.0008505195612525854_dp, 0.0012820182889245553_dp, 0.001919027244758816_dp, &
             0.0028536110397419465_dp, 0.004216701490491993_dp, 0.0061935898553211015_dp, &
             0.009045260217207588_dp, 0.013137651773019673_dp, 0.018981648121409175_dp, &
             0.027287535772292128_dp, 0.039038932574558885_dp, 0.055592879274736995_dp, &
             0.07881511065337202_dp, 0.11126284180339567_dp, 0.15643248688156441_dp, &
             0.2190983387539134_dp, 0.3057846990102029_dp, 0.4254492568597246_dp, &
             0.5905386952252027_dp, 0.8187890720724559_dp, 1.1367263205192033_dp, &
             1.5876160388514782_dp, 2.2532047797010355_dp, 3.331225433538529_dp, &
             5.596537488159656_dp]

      CASE (1211)

         aw(:) = & ! a
            [1.6957172281587293e-12_dp, 1.8654594769969503e-11_dp, 9.873041740395125e-11_dp, &
             3.810636083895109e-10_dp, 1.2203122638109545e-09_dp, 3.442893953721867e-09_dp, &
             8.849934341303335e-09_dp, 2.1166285628419753e-08_dp, 4.7773691292991845e-08_dp, &
             1.0278959781005634e-07_dp, 2.1240732482632804e-07_dp, 4.2396698151437014e-07_dp, &
             8.21081010314631e-07_dp, 1.5484417904115179e-06_dp, 2.851906365873008e-06_dp, &
             5.1423946188665366e-06_dp, 9.096494486770587e-06_dp, 1.581306902838385e-05_dp, &
             2.7054491202501697e-05_dp, 4.5614635572900754e-05_dp, 7.587516436429773e-05_dp, &
             0.0001246393619286835_dp, 0.0002023721144506896_dp, 0.0003250301583155292_dp, &
             0.0005167446643724814_dp, 0.0008137270744264817_dp, 0.0012699203713144205_dp, &
             0.0019651271937216547_dp, 0.003016634326143532_dp, 0.004595748162069503_dp, &
             0.006951195250998526_dp, 0.010442075940423415_dp, 0.015584053746590368_dp, &
             0.023113806294557088_dp, 0.034078571639434854_dp, 0.04996005006168302_dp, &
             0.07284516944410344_dp, 0.10566056383005452_dp, 0.15249341961765533_dp, &
             0.21902913552129866_dp, 0.3131468137772808_dp, 0.44572828029584644_dp, &
             0.6317576403461903_dp, 0.8918216781881242_dp, 1.2541791938653608_dp, &
             1.7576806417723927_dp, 2.4560693995063674_dp, 3.4248096121291662_dp, &
             4.7732642616985785_dp, 6.670345734859146_dp, 9.412839637395933_dp, &
             13.695888543187772_dp, &
             ! w
             5.420876587977142e-12_dp, 3.561612794657484e-11_dp, 1.4579946949290635e-10_dp, &
             4.739661100918526e-10_dp, 1.334338897271311e-09_dp, 3.39596753121497e-09_dp, &
             8.0113224674035e-09_dp, 1.7802460566594097e-08_dp, 3.7679638154747375e-08_dp, &
             7.657190577440994e-08_dp, 1.5030985395313515e-07_dp, 2.863443510577239e-07_dp, &
             5.313508000340871e-07_dp, 9.633118708546681e-07_dp, 1.7104719760773868e-06_dp, &
             2.980725604246212e-06_dp, 5.1066954070564375e-06_dp, 8.61419259307498e-06_dp, &
             1.43252340655935e-05_dp, 2.3511734351272024e-05_dp, 3.812296728192683e-05_dp, &
             6.111968332502579e-05_dp, 9.696143463496418e-05_dp, 0.00015231263829706517_dp, &
             0.00023705913694019496_dp, 0.0003657630974119519_dp, 0.0005597335063014509_dp, &
             0.0008499569215492174_dp, 0.0012812246978547211_dp, 0.0019179168034091294_dp, &
             0.002852069394027273_dp, 0.004214577751001961_dp, 0.006190686753373341_dp, &
             0.009041322342547788_dp, 0.013132351852100429_dp, 0.0189745714986648_dp, &
             0.02727816359208974_dp, 0.03902662468461708_dp, 0.05557685864776937_dp, &
             0.07879445222004604_dp, 0.11123647005249832_dp, 0.15639918836459726_dp, &
             0.21905679745857662_dp, 0.30573356369753335_dp, 0.4253872482923687_dp, &
             0.5904647548650498_dp, 0.8187025378204439_dp, 1.1366270832156777_dp, &
             1.5875045745528873_dp, 2.2530818232820407_dp, 3.331090565135075_dp, &
             5.5963818599073_dp]

      CASE (1212)

         aw(:) = & ! a
            [3.971595228677068e-12_dp, 3.1479279161697507e-11_dp, 1.3721526986594002e-10_dp, &
             4.71525378236041e-10_dp, 1.396953661117157e-09_dp, 3.7245368049346426e-09_dp, &
             9.166968052977739e-09_dp, 2.1177129112748784e-08_dp, 4.6454521416312363e-08_dp, &
             9.758379191671911e-08_dp, 1.9755519337216988e-07_dp, 3.8735999988762556e-07_dp, &
             7.385332393058731e-07_dp, 1.3735460544789624e-06_dp, 2.498495844765095e-06_dp, &
             4.4548409844368646e-06_dp, 7.800315646108175e-06_dp, 1.3434116076612935e-05_dp, &
             2.2788643282580256e-05_dp, 3.812045595824948e-05_dp, 6.294793496535208e-05_dp, &
             0.00010270428973320617_dp, 0.0001657044293456851_dp, 0.0002645662770567956_dp, &
             0.0004182859656057084_dp, 0.000655248313853429_dp, 0.0010175675793723366_dp, &
             0.0015673101992261867_dp, 0.002395366506127593_dp, 0.0036340328936942507_dp, &
             0.005474767148628013_dp, 0.008193124291152633_dp, 0.012183616809537605_dp, &
             0.018008235804269166_dp, 0.026463702904884496_dp, 0.038674308342865144_dp, &
             0.05621957512413827_dp, 0.08130916703727938_dp, 0.11702168873798299_dp, &
             0.16762966414512454_dp, 0.23904053142251583_dp, 0.33939373791543825_dp, &
             0.47986828723909164_dp, 0.6757759440136589_dp, 0.9480482315975118_dp, &
             1.325283148414386_dp, 1.8466317449252563_dp, 2.5660574746468865_dp, &
             3.559120850361908_dp, 4.935131869303874_dp, 6.862833589104965_dp, &
             9.638925722357582_dp, 13.959434789967498_dp, &
             ! w
             1.124996359137677e-11_dp, 5.204802211875482e-11_dp, 1.8266191981556888e-10_dp, &
             5.426649326312457e-10_dp, 1.4361965751109272e-09_dp, 3.491089674458725e-09_dp, &
             7.945211108901498e-09_dp, 1.714944373937754e-08_dp, 3.543019035766573e-08_dp, &
             7.053779164222903e-08_dp, 1.3603455535052165e-07_dp, 2.5516798884253746e-07_dp, &
             4.670611525383764e-07_dp, 8.364753231435185e-07_dp, 1.469020468731327e-06_dp, &
             2.5345908934361683e-06_dp, 4.303104765048507e-06_dp, 7.198491035426322e-06_dp, &
             1.1879522077469658e-05_dp, 1.9359846483355973e-05_dp, 3.118486007926449e-05_dp, &
             4.969044490229552e-05_dp, 7.837889260311442e-05_dp, 0.0001224615739629281_dp, &
             0.0001896375511364938_dp, 0.00029120427115872003_dp, 0.00044363328971990033_dp, &
             0.0006707940626331274_dp, 0.001007076732238086_dp, 0.0015017565095706302_dp, &
             0.002225065596255198_dp, 0.0032766039786190264_dp, 0.004796941464250815_dp, &
             0.0069835578519034025_dp, 0.010112659473229538_dp, 0.014568929193912077_dp, &
             0.020885953986038448_dp, 0.029800984319581697_dp, 0.04232889040199801_dp, &
             0.059861808835242586_dp, 0.08430321587272424_dp, 0.11824839303603972_dp, &
             0.16522825874115538_dp, 0.23004215043431114_dp, 0.31922179683392776_dp, &
             0.4417046315315419_dp, 0.6098791093421821_dp, 0.8413787034272895_dp, &
             1.1625880613030148_dp, 1.6166254652136325_dp, 2.285177091025828_dp, &
             3.3662842275340825_dp, 5.637009321568099_dp]

      CASE (1213)

         aw(:) = & ! a
            [3.3663517829368112e-12_dp, 2.765183022624092e-11_dp, 1.2387142566567383e-10_dp, &
             4.3328247016058214e-10_dp, 1.299473096117342e-09_dp, 3.496146347244391e-09_dp, &
             8.665558279946706e-09_dp, 2.0132416041734324e-08_dp, 4.437002970479056e-08_dp, &
             9.357430643964278e-08_dp, 1.9008269795876703e-07_dp, 3.738125387469709e-07_dp, &
             7.145643475158952e-07_dp, 1.3320542295278836e-06_dp, 2.4280697259635905e-06_dp, &
             4.33742089167271e-06_dp, 7.607718582141769e-06_dp, 1.3122929788666286e-05_dp, &
             2.2292803472004234e-05_dp, 3.734056405432479e-05_dp, 6.173603328289366e-05_dp, &
             0.00010084236326768166_dp, 0.00016287432118729264_dp, 0.00026030795514615096_dp, &
             0.000411940125312993_dp, 0.0006458781027650764_dp, 0.0010038527453011168_dp, &
             0.0015474051493847187_dp, 0.002366711592257903_dp, 0.0035931056903277282_dp, &
             0.0054167575879211765_dp, 0.008111514331213383_dp, 0.012069642005703944_dp, &
             0.017850203019252213_dp, 0.026246135929167812_dp, 0.03837689218431716_dp, &
             0.055815871353289845_dp, 0.08076507859477457_dp, 0.11629366549782594_dp, &
             0.16666266364954474_dp, 0.23776578693315165_dp, 0.3377264433786549_dp, &
             0.4777053908130308_dp, 0.6729943506201924_dp, 0.9445038446410954_dp, &
             1.3208113074056633_dp, 1.8410496814421369_dp, 2.5591693011131684_dp, &
             3.5507251228494447_dp, 4.92503061162409_dp, 6.850839151390944_dp, &
             9.624855202166582_dp, 13.943049594641773_dp, &
             ! w
             9.64227230524896e-12_dp, 4.6645141960189725e-11_dp, 1.6758649397132534e-10_dp, &
             5.049201871370448e-10_dp, 1.3494804597656808e-09_dp, 3.3046437219270578e-09_dp, &
             7.564907325461075e-09_dp, 1.6406488263947502e-08_dp, 3.403044714177637e-08_dp, &
             6.798130256698659e-08_dp, 1.3148993591039167e-07_dp, 2.472794653603076e-07_dp, &
             4.536563709343451e-07_dp, 8.141287465692833e-07_dp, 1.432409014453352e-06_dp, &
             2.475553912639328e-06_dp, 4.209287604640209e-06_dp, 7.051407928980766e-06_dp, &
             1.1651817082343397e-05_dp, 1.90114572884737e-05_dp, 3.065769483432128e-05_dp, &
             4.8901062154656264e-05_dp, 7.72085243408147e-05_dp, 0.00012074264706871639_dp, &
             0.000187135638612257_dp, 0.000287594138442643_dp, 0.00043846739319931377_dp, &
             0.0006634615115425548_dp, 0.0009967503538124039_dp, 0.001487325285061769_dp, &
             0.0022050494043753172_dp, 0.0032490473630866785_dp, 0.0047592831418505005_dp, &
             0.0069324730586019555_dp, 0.010043873250465357_dp, 0.014477000683115367_dp, &
             0.020764035207959792_dp, 0.029640562411450605_dp, 0.04211953067229467_dp, &
             0.05959092825290022_dp, 0.08395593369965056_dp, 0.11780752764961057_dp, &
             0.16467456614631365_dp, 0.22935492577582403_dp, 0.318379973228923_dp, &
             0.44068848581906594_dp, 0.6086725781407694_dp, 0.8399720587759099_dp, &
             1.160980199883009_dp, 1.614824152593401_dp, 2.283193454689839_dp, &
             3.3641098121297004_dp, 5.63449839771455_dp]

      CASE (1214)

         aw(:) = & ! a
            [2.9510431948146254e-12_dp, 2.5000115825922822e-11_dp, 1.1448979837776506e-10_dp, &
             4.060857178136842e-10_dp, 1.2295671126029085e-09_dp, 3.3313195221012567e-09_dp, &
             8.301880277375112e-09_dp, 1.9371575937620424e-08_dp, 4.284675975018178e-08_dp, &
             9.063580301387962e-08_dp, 1.845924238232895e-07_dp, 3.638368664835871e-07_dp, &
             6.96880411667933e-07_dp, 1.3013887010834244e-06_dp, 2.375937810111994e-06_dp, &
             4.250378458410427e-06_dp, 7.464762432450147e-06_dp, 1.2891674545050502e-05_dp, &
             2.1923919356369408e-05_dp, 3.675976638725279e-05_dp, 6.083265768797148e-05_dp, &
             9.945322284602193e-05_dp, 0.00016076110240369285_dp, 0.0002571258433079263_dp, &
             0.0004071946301755019_dp, 0.0006388661489392023_dp, 0.00099358296849749_dp, &
             0.0015324909672816101_dp, 0.002345228973187155_dp, 0.003562405612382992_dp, &
             0.005373220978842572_dp, 0.00805023486058384_dp, 0.011984019586724282_dp, &
             0.017731428490424583_dp, 0.026082545500455603_dp, 0.038153169439299486_dp, &
             0.05551207579713303_dp, 0.08035548425398154_dp, 0.11574540278883398_dp, &
             0.16593417519868425_dp, 0.23680513656505947_dp, 0.33646956097141384_dp, &
             0.47607439786850553_dp, 0.6708961939281376_dp, 0.941829563920597_dp, &
             1.317436359202448_dp, 1.836835792053625_dp, 2.553968240243261_dp, &
             3.544384431245667_dp, 4.917400441620472_dp, 6.841777477051363_dp, &
             9.614223630438296_dp, 13.930667702020337_dp, &
             ! w
             8.53794549297356e-12_dp, 4.286286177274973e-11_dp, 1.568622790685921e-10_dp, &
             4.778139795921825e-10_dp, 1.286790333439939e-09_dp, 3.1691667787154554e-09_dp, &
             7.28743543424965e-09_dp, 1.5862587688014296e-08_dp, 3.3002793223775e-08_dp, &
             6.60997730131223e-08_dp, 1.2813798857350358e-07_dp, 2.414501334822079e-07_dp, &
             4.4373396259704074e-07_dp, 7.975623238018532e-07_dp, 1.4052300542721897e-06_dp, &
             2.4316720439025562e-06_dp, 4.139473541030642e-06_dp, 6.941839937981538e-06_dp, &
             1.1482024288210282e-05_dp, 1.875143729396839e-05_dp, 3.026391157144943e-05_dp, &
             4.831093912877953e-05_dp, 7.63329349706316e-05_dp, 0.00011945576399484887_dp, &
             0.00018526133420926762_dp, 0.00028488792716247857_dp, 0.00043459267262092475_dp, &
             0.0006579585965267814_dp, 0.0009889965160358347_dp, 0.0014764837264214109_dp, &
             0.0021900048400690353_dp, 0.0032283257205098868_dp, 0.004730952876967884_dp, &
             0.006894025875433708_dp, 0.009992082763721237_dp, 0.014407759045432157_dp, &
             0.02067217029603053_dp, 0.029519642319596952_dp, 0.04196166861185848_dp, &
             0.059386610245222925_dp, 0.08369390440090496_dp, 0.11747478657997285_dp, &
             0.16425654670208595_dp, 0.22883594909517824_dp, 0.3177440790367206_dp, &
             0.439920723893764_dp, 0.6077607637716094_dp, 0.8389088027683151_dp, &
             1.159764647663979_dp, 1.613462176484786_dp, 2.2816934962629847_dp, &
             3.3624655373209307_dp, 5.632599719257643_dp]

      CASE (1215)

         aw(:) = & ! a
            [2.6474808446039293e-12_dp, 2.3045853339215338e-11_dp, 1.0749179914982693e-10_dp, &
             3.856129195740199e-10_dp, 1.1765959898380606e-09_dp, 3.205799069362708e-09_dp, &
             8.023841767019949e-09_dp, 1.8788043354852115e-08_dp, 4.167536650467684e-08_dp, &
             8.837098349272668e-08_dp, 1.803525819448485e-07_dp, 3.561200091732354e-07_dp, &
             6.831799687524375e-07_dp, 1.2775987552683994e-06_dp, 2.3354452161287404e-06_dp, &
             4.1826950017038435e-06_dp, 7.353488912380224e-06_dp, 1.2711504852608046e-05_dp, &
             2.1636279070580793e-05_dp, 3.630652808680465e-05_dp, 6.012717279117898e-05_dp, &
             9.83676435754131e-05_dp, 0.00015910862087442095_dp, 0.00025463602833342163_dp, &
             0.000403479473872285_dp, 0.0006333737200067829_dp, 0.0009855346815368007_dp, &
             0.0015207973931794122_dp, 0.0023283778146738023_dp, 0.003538313941877902_dp, &
             0.0053390420928226484_dp, 0.008002108189042765_dp, 0.011916750103026322_dp, &
             0.01763808017143525_dp, 0.025953931884014278_dp, 0.03797722368276972_dp, &
             0.05527308356679684_dp, 0.08003316631882493_dp, 0.11531384183781629_dp, &
             0.16536059517234455_dp, 0.23604856531616913_dp, 0.33547944142189934_dp, &
             0.47478926196666976_dp, 0.6692425790055402_dp, 0.9397214290721622_dp, &
             1.3147753358186416_dp, 1.8335126429896893_dp, 2.5498658385814017_dp, &
             3.5393822869430314_dp, 4.911380113948548_dp, 6.834626707561407_dp, &
             9.60583305997015_dp, 13.920894839311197_dp, &
             ! w
             7.730055696570207e-12_dp, 4.005112922636467e-11_dp, 1.4878762994629426e-10_dp, &
             4.5725211049939694e-10_dp, 1.2389868092411167e-09_dp, 3.0654475744824398e-09_dp, &
             7.074327170295351e-09_dp, 1.5443749078200803e-08_dp, 3.220966895978088e-08_dp, &
             6.464486175255942e-08_dp, 1.2554173327241957e-07_dp, 2.3692836395603808e-07_dp, &
             4.36027117748395e-07_dp, 7.846798138518815e-07_dp, 1.3840723608769647e-06_dp, &
             2.39747858080342e-06_dp, 4.08502479247293e-06_dp, 6.85631653478879e-06_dp, &
             1.1349391768561342e-05_dp, 1.8548181271299545e-05_dp, 2.9955891529412368e-05_dp, &
             4.7849057432922264e-05_dp, 7.564722807632412e-05_dp, 0.00011844741233481564_dp, &
             0.00018379195313486182_dp, 0.00028276534185128115_dp, 0.00043155219486490875_dp, &
             0.0006536386134366688_dp, 0.00098290697704637_dp, 0.0014679658897058887_dp, &
             0.002178180438375309_dp, 0.003212033570459503_dp, 0.004708670949860171_dp, &
             0.0068637770688982_dp, 0.009951323205949266_dp, 0.014353248930181792_dp, &
             0.020599829275466094_dp, 0.029424394760982426_dp, 0.04183728903905676_dp, &
             0.0592255866273777_dp, 0.08348734719490843_dp, 0.11721242483985654_dp, &
             0.16392686849599433_dp, 0.22842655910636905_dp, 0.31724235430581377_dp, &
             0.43931483420160966_dp, 0.6070410598232925_dp, 0.8380694230450183_dp, &
             1.15880489771286_dp, 1.6123866885563223_dp, 2.280508953831196_dp, &
             3.3611669824366546_dp, 5.631100293657594_dp]

      CASE (1216)

         aw(:) = & ! a
            [2.4155310209251444e-12_dp, 2.154180469696622e-11_dp, 1.0205096693836394e-10_dp, &
             3.6957566340727185e-10_dp, 1.1348775237583485e-09_dp, 3.106542605947829e-09_dp, &
             7.803281630367265e-09_dp, 1.8323949608601824e-08_dp, 4.074173592325564e-08_dp, &
             8.656256918567037e-08_dp, 1.7696181445254036e-07_dp, 3.4994002550255445e-07_dp, &
             6.721946948482938e-07_dp, 1.25850275312117e-06_dp, 2.3029102708231638e-06_dp, &
             4.128264462830786e-06_dp, 7.263931083493255e-06_dp, 1.2566388678112742e-05_dp, &
             2.140444314963113e-05_dp, 3.5940990281919195e-05_dp, 5.9557863210983845e-05_dp, &
             9.749112726820865e-05_dp, 0.0001577736946823427_dp, 0.00025262371317853294_dp, &
             0.00040047546107878894_dp, 0.0006289307480264824_dp, 0.0009790215966998682_dp, &
             0.001511330764763924_dp, 0.002314730914438932_dp, 0.003518796653325014_dp, &
             0.005311343893905052_dp, 0.007963094788657006_dp, 0.011862202683590457_dp, &
             0.0175623646397335_dp, 0.025849584300001335_dp, 0.03783443750120398_dp, &
             0.05507908506738866_dp, 0.07977146747090115_dp, 0.11496336605512598_dp, &
             0.1648946826152562_dp, 0.2354338827364608_dp, 0.3346748495270939_dp, &
             0.47374473331898753_dp, 0.667898312719422_dp, 0.9380073722844634_dp, &
             1.3126113844642175_dp, 1.8308098275739186_dp, 2.5465287491886115_dp, &
             3.535312761683076_dp, 4.906481652476629_dp, 6.828807846136792_dp, &
             9.59900470977928_dp, 13.912940959690431_dp, &
             ! w
             7.1122872174461776e-12_dp, 3.7871084590736715e-11_dp, 1.4246129124471554e-10_dp, &
             4.4104433006345785e-10_dp, 1.2011458999303874e-09_dp, 2.983078372720249e-09_dp, &
             6.904648081587444e-09_dp, 1.5109554032599778e-08_dp, 3.157568927050816e-08_dp, &
             6.348008862581203e-08_dp, 1.234604242588072e-07_dp, 2.33299149156462e-07_dp, &
             4.2983499485913164e-07_dp, 7.743194382644818e-07_dp, 1.3670423004051019e-06_dp, &
             2.369934358163303e-06_dp, 4.041132695408446e-06_dp, 6.7873291526438264e-06_dp, &
             1.1242338635466541e-05_dp, 1.8384032184151705e-05_dp, 2.9707004129081536e-05_dp, &
             4.747566232746593e-05_dp, 7.50926318135649e-05_dp, 0.00011763150851961013_dp, &
             0.00018260252269728088_dp, 0.0002810464934691837_dp, 0.00042908914565092834_dp, &
             0.0006501378403474838_dp, 0.0009779705868117813_dp, 0.0014610588708860164_dp, &
             0.00216858929440561_dp, 0.0031988147205270134_dp, 0.004690587277957857_dp, &
             0.006839221179680285_dp, 0.009918226426230718_dp, 0.0143089760425134_dp, &
             0.020541060620028202_dp, 0.029347000010307517_dp, 0.04173620103451951_dp, &
             0.05909468965105415_dp, 0.08331940240389561_dp, 0.11699906651888765_dp, &
             0.16365871778308938_dp, 0.22809351465772587_dp, 0.3168341260481743_dp, &
             0.4388217746095472_dp, 0.6064552955934907_dp, 0.8373861659716264_dp, &
             1.1580235713609988_dp, 1.6115110620763502_dp, 2.279544477551428_dp, &
             3.3601096393932304_dp, 5.62987939182264_dp]

      CASE (1217)

         aw(:) = & ! a
            [2.232400681192695e-12_dp, 2.0346684427822984e-11_dp, 9.768981218921208e-11_dp, &
             3.5663938481106003e-10_dp, 1.1010742276843559e-09_dp, 3.0258466458141576e-09_dp, &
             7.623491492541256e-09_dp, 1.794483276054229e-08_dp, 3.997769850256168e-08_dp, &
             8.508041628272105e-08_dp, 1.7417916311287992e-07_dp, 3.4486260741336273e-07_dp, &
             6.631601952304606e-07_dp, 1.242783703613254e-06_dp, 2.2761071525711605e-06_dp, &
             4.0833903131624114e-06_dp, 7.1900476027598405e-06_dp, 1.2446597296800715e-05_dp, &
             2.1212957864323096e-05_dp, 3.5638916169901323e-05_dp, 5.9087167733656905e-05_dp, &
             9.676611142137358e-05_dp, 0.00015666903651131148_dp, 0.00025095785387987734_dp, &
             0.00039798771868033886_dp, 0.0006252500587922309_dp, 0.0009736241823898729_dp, &
             0.0015034832878628535_dp, 0.0023034147998595746_dp, 0.0035026082171836052_dp, &
             0.005288363734314937_dp, 0.007930718588913532_dp, 0.011816924218728538_dp, &
             0.01749950048174647_dp, 0.02576292868456699_dp, 0.03771583536602568_dp, &
             0.054917911880348495_dp, 0.07955400684649136_dp, 0.11467208131893437_dp, &
             0.16450738779588087_dp, 0.2349228337315347_dp, 0.3340057995072415_dp, &
             0.47287602982167914_dp, 0.6667801582948729_dp, 0.9365814237008007_dp, &
             1.3108109177501013_dp, 1.828560724107347_dp, 2.5437515155328208_dp, &
             3.531925604285191_dp, 4.902404159332935_dp, 6.823963806752881_dp, &
             9.593319910648134_dp, 13.90631877677133_dp, &
             ! w
             6.624215483356366e-12_dp, 3.612765960717791e-11_dp, 1.3735750052993774e-10_dp, &
             4.279024743992062e-10_dp, 1.1703545545054274e-09_dp, 2.9158734353079526e-09_dp, &
             6.76590957920545e-09_dp, 1.483581573934773e-08_dp, 3.1055624957083805e-08_dp, &
             6.252338442927521e-08_dp, 1.2174900638174468e-07_dp, 2.3031199160206668e-07_dp, &
             4.247338902947156e-07_dp, 7.657778127433537e-07_dp, 1.3529918901155736e-06_dp, &
             2.3471947200219344e-06_dp, 4.004875354994867e-06_dp, 6.730310702919663e-06_dp, &
             1.1153814153254272e-05_dp, 1.8248230526353478e-05_dp, 2.950100839223046e-05_dp, &
             4.716649021319132e-05_dp, 7.463324986440142e-05_dp, 0.00011695543913012176_dp, &
             0.00018161661174884196_dp, 0.0002796212980951757_dp, 0.00042704627574648396_dp, &
             0.0006472334458563603_dp, 0.0009738740303164968_dp, 0.0014553254711016058_dp, &
             0.0021606258882720087_dp, 0.0031878366974186696_dp, 0.004675565733033794_dp, &
             0.006818818980611389_dp, 0.00989072238618451_dp, 0.014272177180044521_dp, &
             0.02049220385888125_dp, 0.029282646893599086_dp, 0.04165213242991627_dp, &
             0.058985812409136675_dp, 0.08317968689840534_dp, 0.11682154312634274_dp, &
             0.16343557121330307_dp, 0.22781632572946417_dp, 0.3164943161134268_dp, &
             0.43841129820392605_dp, 0.6059675840778288_dp, 0.8368172214717424_dp, &
             1.1573729080639534_dp, 1.6107818208395157_dp, 2.2787412100400823_dp, &
             3.359229026399312_dp, 5.628862607515617_dp]

      CASE (1218)

         aw(:) = & ! a
            [2.084151662958795e-12_dp, 1.9373623040283225e-11_dp, 9.41119538090363e-11_dp, &
             3.459690044263528e-10_dp, 1.0730850685262167e-09_dp, 2.958839334197943e-09_dp, &
             7.473865961753599e-09_dp, 1.7628752305444398e-08_dp, 3.933974158238866e-08_dp, &
             8.384126868032002e-08_dp, 1.7185018128860148e-07_dp, 3.406089002898318e-07_dp, &
             6.555849308640613e-07_dp, 1.2295935694325834e-06_dp, 2.253600841268673e-06_dp, &
             4.045686665711083e-06_dp, 7.127935126942065e-06_dp, 1.234583896223757e-05_dp, &
             2.1051820228524196e-05_dp, 3.5384604817615375e-05_dp, 5.869073522434417e-05_dp, &
             9.61552515098705e-05_dp, 0.00015573797980922792_dp, 0.0002495533240831586_dp, &
             0.0003958895821214297_dp, 0.0006221448892503967_dp, 0.0009690694528668295_dp, &
             0.0014968592770748684_dp, 0.0022938605519328584_dp, 0.003488937012684749_dp, &
             0.005268952524742089_dp, 0.00790336473942721_dp, 0.011778661749507228_dp, &
             0.017446366897128675_dp, 0.02568967259966634_dp, 0.037615554751463276_dp, &
             0.054781613254778566_dp, 0.07937007775558591_dp, 0.11442567280579251_dp, &
             0.16417971146069638_dp, 0.23449039091365_dp, 0.3334395798826655_dp, &
             0.47214074501586745_dp, 0.6658336137856403_dp, 0.9353741774032615_dp, &
             1.309286418081921_dp, 1.8266561440786924_dp, 2.541399463461661_dp, &
             3.529056729702866_dp, 4.898950275553098_dp, 6.81986028601439_dp, &
             9.588503823070681_dp, 13.900708213140026_dp, &
             ! w
             6.22886651279198e-12_dp, 3.470015068613236e-11_dp, 1.331472079305835e-10_dp, &
             4.1701467053914956e-10_dp, 1.1447679730444503e-09_dp, 2.859901022616392e-09_dp, &
             6.650149714191318e-09_dp, 1.4607074453053193e-08_dp, 3.0620502266576184e-08_dp, &
             6.17220746984639e-08_dp, 1.2031422280454428e-07_dp, 2.278056063000175e-07_dp, &
             4.2045063812481064e-07_dp, 7.586009166862682e-07_dp, 1.341179321494627e-06_dp, &
             2.3280664871821573e-06_dp, 3.974361064479222e-06_dp, 6.682301829485466e-06_dp, &
             1.1079246027793607e-05_dp, 1.813379380836607e-05_dp, 2.9327357366915684e-05_dp, &
             4.6905774129351475e-05_dp, 7.424574202030486e-05_dp, 0.00011638497476105744_dp, &
             0.0001807844685710961_dp, 0.0002784180614842563_dp, 0.0004253211236963455_dp, &
             0.0006447801686851748_dp, 0.0009704129705690895_dp, 0.0014504804382690687_dp, &
             0.002153894985570281_dp, 0.0031785559174684276_dp, 0.004662864173165724_dp, &
             0.006801564652158282_dp, 0.00986745792863999_dp, 0.014241045458099828_dp, &
             0.02045086458164323_dp, 0.029228187276653796_dp, 0.04158097780617558_dp, &
             0.05889364692200563_dp, 0.08306140031621793_dp, 0.11667122755880321_dp, &
             0.16324660072883013_dp, 0.22758156096008655_dp, 0.3162064807295937_dp, &
             0.43806356598935087_dp, 0.605554378995681_dp, 0.8363351455548758_dp, &
             1.156821540861148_dp, 1.61016381929861_dp, 2.278060433275034_dp, &
             3.3584826707635105_dp, 5.6280008363983_dp]

      CASE (1219)

         aw(:) = & ! a
            [1.4126454170489986e-12_dp, 1.4884205931628885e-11_dp, 7.722789668195569e-11_dp, &
             2.9482321064825933e-10_dp, 9.374510050269946e-10_dp, 2.6314722125102193e-09_dp, &
             6.7382057825845145e-09_dp, 1.6066682124630048e-08_dp, 3.617347304373929e-08_dp, &
             7.76689119370881e-08_dp, 1.6021297542558323e-07_dp, 3.192964329057949e-07_dp, &
             6.17538866206499e-07_dp, 1.16320487186043e-06_dp, 2.140102778745077e-06_dp, &
             3.855216188172163e-06_dp, 6.813655676313729e-06_dp, 1.183527166085015e-05_dp, &
             2.0234196375207006e-05_dp, 3.409260380144085e-05_dp, 5.667437477705017e-05_dp, &
             9.304491048609797e-05_dp, 0.00015099250795178675_dp, 0.00024238787000328366_dp, &
             0.00038517608304288045_dp, 0.000606276040160395_dp, 0.0009457743309836686_dp, &
             0.0014629555649175855_dp, 0.002244924416779311_dp, 0.00341886713152857_dp, &
             0.005169399519236967_dp, 0.007762991527451481_dp, 0.01158219430805189_dp, &
             0.017173389181864247_dp, 0.02531311474855688_dp, 0.03709982160377267_dp, &
             0.05408030275544872_dp, 0.0784232491061914_dp, 0.11315664671447767_dp, &
             0.16249142560229338_dp, 0.2322614006136645_dp, 0.33051989638430695_dp, &
             0.46834785049408456_dp, 0.6609491878828319_dp, 0.9291423046015346_dp, &
             1.301414276819522_dp, 1.8168183045149535_dp, 2.5292467382578883_dp, &
             3.514229683953308_dp, 4.881095440829646_dp, 6.798642631473516_dp, &
             9.563597200357071_dp, 13.871688390649938_dp, &
             ! w
             4.434335869956801e-12_dp, 2.8000306539685763e-11_dp, 1.1296038920755327e-10_dp, &
             3.641690344978627e-10_dp, 1.0195134574646586e-09_dp, 2.5841136636254924e-09_dp, &
             6.076822299051206e-09_dp, 1.3469358345647207e-08_dp, 2.8448536649874895e-08_dp, &
             5.770997868101266e-08_dp, 1.1311121708334478e-07_dp, 2.1519332663072326e-07_dp, &
             3.988520398421899e-07_dp, 7.223430726279125e-07_dp, 1.2814007811814024e-06_dp, &
             2.2311170218187767e-06_dp, 3.8194841198853e-06_dp, 6.438313606121647e-06_dp, &
             1.0699824769726176e-05_dp, 1.7550862503395514e-05_dp, 2.8441875423235514e-05_dp, &
             4.557504077561303e-05_dp, 7.226604745840388e-05_dp, 0.00011346810539719503_dp, &
             0.00017652617230568855_dp, 0.00027225610221770394_dp, 0.00041647999976762837_dp, &
             0.0006321989446164987_dp, 0.0009526519962842704_dp, 0.0014256020482026886_dp, &
             0.002119312655434694_dp, 0.0031308459303687184_dp, 0.0045975338038777925_dp, &
             0.006712771503786658_dp, 0.00974767678020524_dp, 0.014080682419707564_dp, &
             0.020237824351779937_dp, 0.02894740902614061_dp, 0.04121397015338734_dp, &
             0.05841807436984593_dp, 0.08245080416079444_dp, 0.11589500452040971_dp, &
             0.1622704105401599_dp, 0.22636838154157363_dp, 0.3147185548986559_dp, &
             0.43626544251659993_dp, 0.6034170640934511_dp, 0.8338409266330685_dp, &
             1.1539681537956248_dp, 1.6069649937104484_dp, 2.274536200508685_dp, &
             3.354618648234637_dp, 5.623539238610878_dp]

      CASE (1220)

         aw(:) = & ! a
            [1.2553971263831339e-12_dp, 1.3806713914489081e-11_dp, 7.306352192292575e-11_dp, &
             2.8197917705748934e-10_dp, 9.02965384022056e-10_dp, 2.54747498158223e-09_dp, &
             6.548108090564881e-09_dp, 1.5660735206773384e-08_dp, 3.534675271910539e-08_dp, &
             7.60508774690804e-08_dp, 1.5715193393434566e-07_dp, 3.136736985338853e-07_dp, &
             6.074750195311699e-07_dp, 1.1456027943444606e-06_dp, 2.109947030403563e-06_dp, &
             3.804513136206983e-06_dp, 6.72985006363681e-06_dp, 1.1698908463979875e-05_dp, &
             2.0015506022296003e-05_dp, 3.374656672683883e-05_dp, 5.6133658721574255e-05_dp, &
             9.220985968062528e-05_dp, 0.00014971708370474176_dp, 0.0002404600812476481_dp, &
             0.0003822909848122114_dp, 0.0006019987973497609_dp, 0.0009394901226963736_dp, &
             0.0014538022205215454_dp, 0.0022317026113920805_dp, 0.003399921717286713_dp, &
             0.005142464136307774_dp, 0.007724987058349524_dp, 0.01152897008612237_dp, &
             0.017099394280407906_dp, 0.025210985397966644_dp, 0.03695987019000328_dp, &
             0.05388989452196452_dp, 0.07816605546676393_dp, 0.11281177020217012_dp, &
             0.1620324025936082_dp, 0.23165510624537583_dp, 0.3297254040776691_dp, &
             0.46731533869161496_dp, 0.6596190402889114_dp, 0.9274446148888407_dp, &
             1.2992690328169374_dp, 1.814136561649976_dp, 2.525933042065085_dp, &
             3.5101857728347428_dp, 4.876224698766473_dp, 6.792853535216114_dp, &
             9.556800770386827_dp, 13.86376906694082_dp, &
             ! w
             4.0127696307911685e-12_dp, 2.6357869871456713e-11_dp, 1.0788955459443293e-10_dp, &
             3.507098392118496e-10_dp, 9.87305427412576e-10_dp, 2.512683182248576e-09_dp, &
             5.927475708027273e-09_dp, 1.3171604139518086e-08_dp, 2.787787296095129e-08_dp, &
             5.6652298243484345e-08_dp, 1.1120680822549588e-07_dp, 2.1185022248783528e-07_dp, &
             3.931139481242819e-07_dp, 7.126908557191123e-07_dp, 1.2654578601022403e-06_dp, &
             2.205217430227896e-06_dp, 3.7780462858951713e-06_dp, 6.372942142081516e-06_dp, &
             1.0598035205537437e-05_dp, 1.739428831784377e-05_dp, 2.820377101975927e-05_dp, &
             4.521683593761747e-05_dp, 7.173263584470525e-05_dp, 0.0001126814595449584_dp, &
             0.00017537676807696144_dp, 0.0002705915049314369_dp, 0.0004140898121727358_dp, &
             0.0006287951452666797_dp, 0.0009478435113679872_dp, 0.0014188622120375044_dp, &
             0.0021099380448511406_dp, 0.003117904939208309_dp, 0.004579803308937709_dp, &
             0.006688660140760133_dp, 0.0097151338153203_dp, 0.014037092138219234_dp, &
             0.0201798875205471_dp, 0.028871015644924027_dp, 0.041114071791819064_dp, &
             0.058288570383584676_dp, 0.08228446448349679_dp, 0.11568346275746656_dp, &
             0.16200427427718764_dp, 0.22603751999559327_dp, 0.3143126313294553_dp, &
             0.43577474746754147_dp, 0.602833653862242_dp, 0.8331599472530745_dp, &
             1.1531889884161004_dp, 1.606091425501809_dp, 2.273573784651251_dp, &
             3.353563611200909_dp, 5.622321460467435_dp]

      CASE (1221)

         aw(:) = & ! a
            [1.2544430440765798e-12_dp, 1.3800135126883162e-11_dp, 7.303793643071283e-11_dp, &
             2.818999492359324e-10_dp, 9.027520815325056e-10_dp, 2.5469543983692096e-09_dp, &
             6.5469281147114734e-09_dp, 1.5658212278191648e-08_dp, 3.534160943860049e-08_dp, &
             7.604080244500397e-08_dp, 1.5713285951216842e-07_dp, 3.136386385181874e-07_dp, &
             6.07412231380492e-07_dp, 1.1454929191502026e-06_dp, 2.109758706503128e-06_dp, &
             3.804196361721916e-06_dp, 6.729326277542975e-06_dp, 1.1698055895859524e-05_dp, &
             2.0014138291147205e-05_dp, 3.3744401902797e-05_dp, 5.613027504359523e-05_dp, &
             9.220463277752715e-05_dp, 0.0001497090984106373_dp, 0.0002404480088350214_dp, &
             0.0003822729135716618_dp, 0.0006019720007963777_dp, 0.0009394507451437946_dp, &
             0.001453744854205066_dp, 0.0022316197326344525_dp, 0.0033998029413293677_dp, &
             0.005142295241414632_dp, 0.007724748719836389_dp, 0.011528636250549236_dp, &
             0.0170989301001639_dp, 0.02521034463837112_dp, 0.03695899201725169_dp, &
             0.053888699584296734_dp, 0.07816444119844104_dp, 0.11280960531821103_dp, &
             0.1620295208129716_dp, 0.2316512993939981_dp, 0.32972041489774107_dp, &
             0.46730885393929833_dp, 0.659610685056489_dp, 0.9274339495186438_dp, &
             1.2992555541665303_dp, 1.8141197106392004_dp, 2.525912219144849_dp, &
             3.5101603617120714_dp, 4.876194095134997_dp, 6.792817168995609_dp, &
             9.556758089309067_dp, 13.863719352422802_dp, &
             ! w
             4.0102092802242075e-12_dp, 2.634779183775859e-11_dp, 1.0785827390757358e-10_dp, &
             3.506265609555229e-10_dp, 9.87105722234526e-10_dp, 2.512239577691751e-09_dp, &
             5.926547059879969e-09_dp, 1.3169750786991112e-08_dp, 2.7874317854774955e-08_dp, &
             5.6645704304334585e-08_dp, 1.1119492794408921e-07_dp, 2.118293555264692e-07_dp, &
             3.930781143440523e-07_dp, 7.126305517546412e-07_dp, 1.2653582136583923e-06_dp, &
             2.205055493161972e-06_dp, 3.7777871088913106e-06_dp, 6.372533143150381e-06_dp, &
             1.05973981728321e-05_dp, 1.7393308163702046e-05_dp, 2.820228011574313e-05_dp, &
             4.521459249514445e-05_dp, 7.172929434714882e-05_dp, 0.00011267653067062564_dp, &
             0.00017536956486354278_dp, 0.0002705810711105633_dp, 0.0004140748276460326_dp, &
             0.000628773802614195_dp, 0.0009478133561128633_dp, 0.0014188199382063843_dp, &
             0.0021098792362922105_dp, 0.003117823746374484_dp, 0.004579692050737401_dp, &
             0.006688508822100626_dp, 0.009714929553906827_dp, 0.014036818500035011_dp, &
             0.020179523773551273_dp, 0.028870535957192058_dp, 0.04111344442634647_dp, &
             0.058287756979028604_dp, 0.0822834195619817_dp, 0.11568213367154634_dp, &
             0.16200260189883334_dp, 0.2260354405500391_dp, 0.31431007981345865_dp, &
             0.43577166296551734_dp, 0.6028299867534292_dp, 0.8331556676509858_dp, &
             1.1531840936804112_dp, 1.6060859414429771_dp, 2.273567748405457_dp, &
             3.353557000325493_dp, 5.622313834768768_dp]

      END SELECT

   END SUBROUTINE get_minimax_coeff_low
END MODULE minimax_exp_k53
